
/* Copyright © 2015-2021 Intel Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/* This file generated from vk_entrypoints_gen.py, don't edit directly. */

/* This file is the only place we rely on undefined symbols to fall back to
 * NULL. Other files use regular symbol declarations.
 * See also comments on VK_ENTRY_WEAK.
 */
#define VK_ENTRY_USE_WEAK 1
#include "anv_entrypoints.h"

/* Weak aliases for all potential implementations. These will resolve to
 * NULL if they're not defined, which lets the resolve_entrypoint() function
 * either pick the correct entry point.
 *
 * MSVC uses different decorated names for 32-bit versus 64-bit. Declare
 * all argument sizes for 32-bit because computing the actual size would be
 * difficult.
 */




#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateInstance@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateInstance@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateInstance@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateInstance@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateInstance@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateInstance@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateInstance@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateInstance@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateInstance@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateInstance@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateInstance@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateInstance@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateInstance@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateInstance@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateInstance=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateInstance=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyInstance@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyInstance@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyInstance@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyInstance@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyInstance@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyInstance@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyInstance@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyInstance@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyInstance@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyInstance@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyInstance@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyInstance@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyInstance@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyInstance@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyInstance=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyInstance=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDevices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDevices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDevices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDevices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDevices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDevices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDevices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDevices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDevices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDevices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDevices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDevices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDevices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDevices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_EnumeratePhysicalDevices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_EnumeratePhysicalDevices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetInstanceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetInstanceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetInstanceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetInstanceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetInstanceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetInstanceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetInstanceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetInstanceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetInstanceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetInstanceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetInstanceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetInstanceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetInstanceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetInstanceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetInstanceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetInstanceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceVersion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceVersion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceVersion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceVersion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceVersion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceVersion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceVersion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceVersion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceVersion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceVersion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceVersion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceVersion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceVersion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceVersion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_EnumerateInstanceVersion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_EnumerateInstanceVersion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceLayerProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceLayerProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceLayerProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceLayerProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceLayerProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceLayerProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceLayerProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceLayerProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceLayerProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceLayerProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceLayerProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceLayerProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceLayerProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceLayerProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_EnumerateInstanceLayerProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_EnumerateInstanceLayerProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceExtensionProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceExtensionProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceExtensionProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceExtensionProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceExtensionProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceExtensionProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceExtensionProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceExtensionProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceExtensionProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceExtensionProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceExtensionProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceExtensionProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceExtensionProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateInstanceExtensionProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_EnumerateInstanceExtensionProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_EnumerateInstanceExtensionProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateAndroidSurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAndroidSurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAndroidSurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAndroidSurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAndroidSurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAndroidSurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAndroidSurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAndroidSurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAndroidSurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAndroidSurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAndroidSurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAndroidSurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAndroidSurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAndroidSurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateAndroidSurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateAndroidSurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_OHOS
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateSurfaceOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSurfaceOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSurfaceOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSurfaceOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSurfaceOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSurfaceOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSurfaceOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSurfaceOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSurfaceOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSurfaceOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSurfaceOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSurfaceOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSurfaceOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSurfaceOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateSurfaceOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateSurfaceOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_OHOS
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayPlaneSurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayPlaneSurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayPlaneSurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayPlaneSurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayPlaneSurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayPlaneSurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayPlaneSurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayPlaneSurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayPlaneSurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayPlaneSurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayPlaneSurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayPlaneSurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayPlaneSurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayPlaneSurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateDisplayPlaneSurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateDisplayPlaneSurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroySurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroySurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroySurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_VI_NN
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateViSurfaceNN@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateViSurfaceNN@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateViSurfaceNN@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateViSurfaceNN@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateViSurfaceNN@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateViSurfaceNN@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateViSurfaceNN@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateViSurfaceNN@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateViSurfaceNN@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateViSurfaceNN@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateViSurfaceNN@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateViSurfaceNN@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateViSurfaceNN@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateViSurfaceNN@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateViSurfaceNN=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateViSurfaceNN=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_VI_NN
#ifdef VK_USE_PLATFORM_WAYLAND_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateWaylandSurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWaylandSurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWaylandSurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWaylandSurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWaylandSurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWaylandSurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWaylandSurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWaylandSurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWaylandSurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWaylandSurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWaylandSurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWaylandSurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWaylandSurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWaylandSurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateWaylandSurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateWaylandSurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WAYLAND_KHR
#ifdef VK_USE_PLATFORM_UBM_SEC
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateUbmSurfaceSEC@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateUbmSurfaceSEC@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateUbmSurfaceSEC@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateUbmSurfaceSEC@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateUbmSurfaceSEC@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateUbmSurfaceSEC@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateUbmSurfaceSEC@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateUbmSurfaceSEC@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateUbmSurfaceSEC@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateUbmSurfaceSEC@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateUbmSurfaceSEC@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateUbmSurfaceSEC@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateUbmSurfaceSEC@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateUbmSurfaceSEC@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateUbmSurfaceSEC=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateUbmSurfaceSEC=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_UBM_SEC
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateWin32SurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWin32SurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWin32SurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWin32SurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWin32SurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWin32SurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWin32SurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWin32SurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWin32SurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWin32SurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWin32SurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWin32SurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWin32SurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateWin32SurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateWin32SurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateWin32SurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_XLIB_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateXlibSurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXlibSurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXlibSurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXlibSurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXlibSurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXlibSurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXlibSurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXlibSurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXlibSurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXlibSurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXlibSurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXlibSurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXlibSurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXlibSurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateXlibSurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateXlibSurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XLIB_KHR
#ifdef VK_USE_PLATFORM_XCB_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateXcbSurfaceKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXcbSurfaceKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXcbSurfaceKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXcbSurfaceKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXcbSurfaceKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXcbSurfaceKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXcbSurfaceKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXcbSurfaceKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXcbSurfaceKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXcbSurfaceKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXcbSurfaceKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXcbSurfaceKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXcbSurfaceKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateXcbSurfaceKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateXcbSurfaceKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateXcbSurfaceKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XCB_KHR
#ifdef VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateDirectFBSurfaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDirectFBSurfaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDirectFBSurfaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDirectFBSurfaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDirectFBSurfaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDirectFBSurfaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDirectFBSurfaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDirectFBSurfaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDirectFBSurfaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDirectFBSurfaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDirectFBSurfaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDirectFBSurfaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDirectFBSurfaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDirectFBSurfaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateDirectFBSurfaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateDirectFBSurfaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateImagePipeSurfaceFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImagePipeSurfaceFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImagePipeSurfaceFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImagePipeSurfaceFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImagePipeSurfaceFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImagePipeSurfaceFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImagePipeSurfaceFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImagePipeSurfaceFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImagePipeSurfaceFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImagePipeSurfaceFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImagePipeSurfaceFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImagePipeSurfaceFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImagePipeSurfaceFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImagePipeSurfaceFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateImagePipeSurfaceFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateImagePipeSurfaceFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_GGP
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateStreamDescriptorSurfaceGGP@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateStreamDescriptorSurfaceGGP@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateStreamDescriptorSurfaceGGP@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateStreamDescriptorSurfaceGGP@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateStreamDescriptorSurfaceGGP@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateStreamDescriptorSurfaceGGP@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateStreamDescriptorSurfaceGGP@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateStreamDescriptorSurfaceGGP@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateStreamDescriptorSurfaceGGP@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateStreamDescriptorSurfaceGGP@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateStreamDescriptorSurfaceGGP@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateStreamDescriptorSurfaceGGP@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateStreamDescriptorSurfaceGGP@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateStreamDescriptorSurfaceGGP@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateStreamDescriptorSurfaceGGP=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateStreamDescriptorSurfaceGGP=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_GGP
#ifdef VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateScreenSurfaceQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateScreenSurfaceQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateScreenSurfaceQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateScreenSurfaceQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateScreenSurfaceQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateScreenSurfaceQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateScreenSurfaceQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateScreenSurfaceQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateScreenSurfaceQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateScreenSurfaceQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateScreenSurfaceQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateScreenSurfaceQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateScreenSurfaceQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateScreenSurfaceQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateScreenSurfaceQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateScreenSurfaceQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateDebugReportCallbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugReportCallbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugReportCallbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugReportCallbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugReportCallbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugReportCallbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugReportCallbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugReportCallbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugReportCallbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugReportCallbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugReportCallbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugReportCallbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugReportCallbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugReportCallbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateDebugReportCallbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateDebugReportCallbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugReportCallbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugReportCallbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugReportCallbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugReportCallbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugReportCallbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugReportCallbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugReportCallbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugReportCallbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugReportCallbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugReportCallbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugReportCallbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugReportCallbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugReportCallbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugReportCallbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyDebugReportCallbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyDebugReportCallbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DebugReportMessageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugReportMessageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugReportMessageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugReportMessageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugReportMessageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugReportMessageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugReportMessageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugReportMessageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugReportMessageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugReportMessageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugReportMessageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugReportMessageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugReportMessageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugReportMessageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DebugReportMessageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DebugReportMessageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroups@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroups@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroups@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroups@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroups@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroups@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroups@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroups@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroups@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroups@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroups@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroups@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroups@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroups@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_EnumeratePhysicalDeviceGroups=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_EnumeratePhysicalDeviceGroups=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroupsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroupsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroupsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroupsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroupsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroupsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroupsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroupsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroupsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroupsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroupsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroupsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroupsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceGroupsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_EnumeratePhysicalDeviceGroupsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_EnumeratePhysicalDeviceGroupsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_IOS_MVK
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateIOSSurfaceMVK@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIOSSurfaceMVK@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIOSSurfaceMVK@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIOSSurfaceMVK@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIOSSurfaceMVK@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIOSSurfaceMVK@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIOSSurfaceMVK@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIOSSurfaceMVK@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIOSSurfaceMVK@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIOSSurfaceMVK@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIOSSurfaceMVK@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIOSSurfaceMVK@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIOSSurfaceMVK@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIOSSurfaceMVK@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateIOSSurfaceMVK=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateIOSSurfaceMVK=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_IOS_MVK
#ifdef VK_USE_PLATFORM_MACOS_MVK
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateMacOSSurfaceMVK@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMacOSSurfaceMVK@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMacOSSurfaceMVK@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMacOSSurfaceMVK@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMacOSSurfaceMVK@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMacOSSurfaceMVK@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMacOSSurfaceMVK@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMacOSSurfaceMVK@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMacOSSurfaceMVK@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMacOSSurfaceMVK@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMacOSSurfaceMVK@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMacOSSurfaceMVK@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMacOSSurfaceMVK@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMacOSSurfaceMVK@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateMacOSSurfaceMVK=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateMacOSSurfaceMVK=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_MACOS_MVK
#ifdef VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateMetalSurfaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMetalSurfaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMetalSurfaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMetalSurfaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMetalSurfaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMetalSurfaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMetalSurfaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMetalSurfaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMetalSurfaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMetalSurfaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMetalSurfaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMetalSurfaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMetalSurfaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMetalSurfaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateMetalSurfaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateMetalSurfaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateDebugUtilsMessengerEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugUtilsMessengerEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugUtilsMessengerEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugUtilsMessengerEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugUtilsMessengerEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugUtilsMessengerEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugUtilsMessengerEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugUtilsMessengerEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugUtilsMessengerEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugUtilsMessengerEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugUtilsMessengerEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugUtilsMessengerEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugUtilsMessengerEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDebugUtilsMessengerEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateDebugUtilsMessengerEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateDebugUtilsMessengerEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugUtilsMessengerEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugUtilsMessengerEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugUtilsMessengerEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugUtilsMessengerEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugUtilsMessengerEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugUtilsMessengerEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugUtilsMessengerEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugUtilsMessengerEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugUtilsMessengerEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugUtilsMessengerEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugUtilsMessengerEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugUtilsMessengerEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugUtilsMessengerEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDebugUtilsMessengerEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyDebugUtilsMessengerEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyDebugUtilsMessengerEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_SubmitDebugUtilsMessageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SubmitDebugUtilsMessageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SubmitDebugUtilsMessageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SubmitDebugUtilsMessageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SubmitDebugUtilsMessageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SubmitDebugUtilsMessageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SubmitDebugUtilsMessageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SubmitDebugUtilsMessageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SubmitDebugUtilsMessageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SubmitDebugUtilsMessageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SubmitDebugUtilsMessageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SubmitDebugUtilsMessageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SubmitDebugUtilsMessageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SubmitDebugUtilsMessageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_SubmitDebugUtilsMessageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_SubmitDebugUtilsMessageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateHeadlessSurfaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateHeadlessSurfaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateHeadlessSurfaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateHeadlessSurfaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateHeadlessSurfaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateHeadlessSurfaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateHeadlessSurfaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateHeadlessSurfaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateHeadlessSurfaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateHeadlessSurfaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateHeadlessSurfaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateHeadlessSurfaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateHeadlessSurfaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateHeadlessSurfaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateHeadlessSurfaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateHeadlessSurfaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetExternalComputeQueueDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetExternalComputeQueueDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetExternalComputeQueueDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetExternalComputeQueueDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetExternalComputeQueueDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetExternalComputeQueueDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetExternalComputeQueueDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetExternalComputeQueueDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetExternalComputeQueueDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetExternalComputeQueueDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetExternalComputeQueueDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetExternalComputeQueueDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetExternalComputeQueueDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetExternalComputeQueueDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetExternalComputeQueueDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetExternalComputeQueueDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif


const struct vk_instance_entrypoint_table anv_instance_entrypoints = {
    .CreateInstance = anv_CreateInstance,
    .DestroyInstance = anv_DestroyInstance,
    .EnumeratePhysicalDevices = anv_EnumeratePhysicalDevices,
    .GetInstanceProcAddr = anv_GetInstanceProcAddr,
    .EnumerateInstanceVersion = anv_EnumerateInstanceVersion,
    .EnumerateInstanceLayerProperties = anv_EnumerateInstanceLayerProperties,
    .EnumerateInstanceExtensionProperties = anv_EnumerateInstanceExtensionProperties,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .CreateAndroidSurfaceKHR = anv_CreateAndroidSurfaceKHR,
#elif defined(_MSC_VER)
    .CreateAndroidSurfaceKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_OHOS
    .CreateSurfaceOHOS = anv_CreateSurfaceOHOS,
#elif defined(_MSC_VER)
    .CreateSurfaceOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .CreateDisplayPlaneSurfaceKHR = anv_CreateDisplayPlaneSurfaceKHR,
    .DestroySurfaceKHR = anv_DestroySurfaceKHR,
#ifdef VK_USE_PLATFORM_VI_NN
    .CreateViSurfaceNN = anv_CreateViSurfaceNN,
#elif defined(_MSC_VER)
    .CreateViSurfaceNN = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_VI_NN
#ifdef VK_USE_PLATFORM_WAYLAND_KHR
    .CreateWaylandSurfaceKHR = anv_CreateWaylandSurfaceKHR,
#elif defined(_MSC_VER)
    .CreateWaylandSurfaceKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WAYLAND_KHR
#ifdef VK_USE_PLATFORM_UBM_SEC
    .CreateUbmSurfaceSEC = anv_CreateUbmSurfaceSEC,
#elif defined(_MSC_VER)
    .CreateUbmSurfaceSEC = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_UBM_SEC
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .CreateWin32SurfaceKHR = anv_CreateWin32SurfaceKHR,
#elif defined(_MSC_VER)
    .CreateWin32SurfaceKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_XLIB_KHR
    .CreateXlibSurfaceKHR = anv_CreateXlibSurfaceKHR,
#elif defined(_MSC_VER)
    .CreateXlibSurfaceKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XLIB_KHR
#ifdef VK_USE_PLATFORM_XCB_KHR
    .CreateXcbSurfaceKHR = anv_CreateXcbSurfaceKHR,
#elif defined(_MSC_VER)
    .CreateXcbSurfaceKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XCB_KHR
#ifdef VK_USE_PLATFORM_DIRECTFB_EXT
    .CreateDirectFBSurfaceEXT = anv_CreateDirectFBSurfaceEXT,
#elif defined(_MSC_VER)
    .CreateDirectFBSurfaceEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateImagePipeSurfaceFUCHSIA = anv_CreateImagePipeSurfaceFUCHSIA,
#elif defined(_MSC_VER)
    .CreateImagePipeSurfaceFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_GGP
    .CreateStreamDescriptorSurfaceGGP = anv_CreateStreamDescriptorSurfaceGGP,
#elif defined(_MSC_VER)
    .CreateStreamDescriptorSurfaceGGP = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_GGP
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .CreateScreenSurfaceQNX = anv_CreateScreenSurfaceQNX,
#elif defined(_MSC_VER)
    .CreateScreenSurfaceQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CreateDebugReportCallbackEXT = anv_CreateDebugReportCallbackEXT,
    .DestroyDebugReportCallbackEXT = anv_DestroyDebugReportCallbackEXT,
    .DebugReportMessageEXT = anv_DebugReportMessageEXT,
    .EnumeratePhysicalDeviceGroups = anv_EnumeratePhysicalDeviceGroups,
    .EnumeratePhysicalDeviceGroupsKHR = anv_EnumeratePhysicalDeviceGroupsKHR,
#ifdef VK_USE_PLATFORM_IOS_MVK
    .CreateIOSSurfaceMVK = anv_CreateIOSSurfaceMVK,
#elif defined(_MSC_VER)
    .CreateIOSSurfaceMVK = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_IOS_MVK
#ifdef VK_USE_PLATFORM_MACOS_MVK
    .CreateMacOSSurfaceMVK = anv_CreateMacOSSurfaceMVK,
#elif defined(_MSC_VER)
    .CreateMacOSSurfaceMVK = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_MACOS_MVK
#ifdef VK_USE_PLATFORM_METAL_EXT
    .CreateMetalSurfaceEXT = anv_CreateMetalSurfaceEXT,
#elif defined(_MSC_VER)
    .CreateMetalSurfaceEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CreateDebugUtilsMessengerEXT = anv_CreateDebugUtilsMessengerEXT,
    .DestroyDebugUtilsMessengerEXT = anv_DestroyDebugUtilsMessengerEXT,
    .SubmitDebugUtilsMessageEXT = anv_SubmitDebugUtilsMessageEXT,
    .CreateHeadlessSurfaceEXT = anv_CreateHeadlessSurfaceEXT,
    .GetExternalComputeQueueDataNV = anv_GetExternalComputeQueueDataNV,
};



#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceQueueFamilyProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceQueueFamilyProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceMemoryProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceMemoryProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceFormatProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceFormatProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceImageFormatProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceImageFormatProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceLayerProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceLayerProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceLayerProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceLayerProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceLayerProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceLayerProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceLayerProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceLayerProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceLayerProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceLayerProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceLayerProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceLayerProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceLayerProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceLayerProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_EnumerateDeviceLayerProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_EnumerateDeviceLayerProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceExtensionProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceExtensionProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceExtensionProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceExtensionProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceExtensionProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceExtensionProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceExtensionProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceExtensionProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceExtensionProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceExtensionProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceExtensionProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceExtensionProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceExtensionProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumerateDeviceExtensionProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_EnumerateDeviceExtensionProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_EnumerateDeviceExtensionProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceSparseImageFormatProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceSparseImageFormatProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceDisplayPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceDisplayPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlanePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlanePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlanePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlanePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlanePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlanePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlanePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlanePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlanePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlanePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlanePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlanePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlanePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlanePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceDisplayPlanePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceDisplayPlanePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneSupportedDisplaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneSupportedDisplaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneSupportedDisplaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneSupportedDisplaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneSupportedDisplaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneSupportedDisplaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneSupportedDisplaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneSupportedDisplaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneSupportedDisplaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneSupportedDisplaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneSupportedDisplaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneSupportedDisplaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneSupportedDisplaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneSupportedDisplaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDisplayPlaneSupportedDisplaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDisplayPlaneSupportedDisplaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDisplayModePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDisplayModePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayModeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayModeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayModeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayModeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayModeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayModeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayModeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayModeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayModeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayModeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayModeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayModeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayModeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDisplayModeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateDisplayModeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateDisplayModeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDisplayPlaneCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDisplayPlaneCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceSurfaceSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceSurfaceSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceSurfaceCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceSurfaceCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormatsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormatsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormatsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormatsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormatsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormatsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormatsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormatsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormatsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormatsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormatsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormatsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormatsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormatsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceSurfaceFormatsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceSurfaceFormatsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WAYLAND_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWaylandPresentationSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWaylandPresentationSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWaylandPresentationSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWaylandPresentationSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWaylandPresentationSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWaylandPresentationSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWaylandPresentationSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWaylandPresentationSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWaylandPresentationSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWaylandPresentationSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWaylandPresentationSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWaylandPresentationSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWaylandPresentationSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWaylandPresentationSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceWaylandPresentationSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceWaylandPresentationSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WAYLAND_KHR
#ifdef VK_USE_PLATFORM_UBM_SEC
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceUbmPresentationSupportSEC@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceUbmPresentationSupportSEC@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceUbmPresentationSupportSEC@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceUbmPresentationSupportSEC@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceUbmPresentationSupportSEC@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceUbmPresentationSupportSEC@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceUbmPresentationSupportSEC@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceUbmPresentationSupportSEC@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceUbmPresentationSupportSEC@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceUbmPresentationSupportSEC@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceUbmPresentationSupportSEC@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceUbmPresentationSupportSEC@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceUbmPresentationSupportSEC@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceUbmPresentationSupportSEC@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceUbmPresentationSupportSEC=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceUbmPresentationSupportSEC=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_UBM_SEC
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWin32PresentationSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWin32PresentationSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWin32PresentationSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWin32PresentationSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWin32PresentationSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWin32PresentationSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWin32PresentationSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWin32PresentationSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWin32PresentationSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWin32PresentationSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWin32PresentationSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWin32PresentationSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWin32PresentationSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceWin32PresentationSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceWin32PresentationSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceWin32PresentationSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_XLIB_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXlibPresentationSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXlibPresentationSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXlibPresentationSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXlibPresentationSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXlibPresentationSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXlibPresentationSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXlibPresentationSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXlibPresentationSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXlibPresentationSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXlibPresentationSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXlibPresentationSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXlibPresentationSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXlibPresentationSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXlibPresentationSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceXlibPresentationSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceXlibPresentationSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XLIB_KHR
#ifdef VK_USE_PLATFORM_XCB_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXcbPresentationSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXcbPresentationSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXcbPresentationSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXcbPresentationSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXcbPresentationSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXcbPresentationSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXcbPresentationSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXcbPresentationSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXcbPresentationSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXcbPresentationSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXcbPresentationSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXcbPresentationSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXcbPresentationSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceXcbPresentationSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceXcbPresentationSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceXcbPresentationSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XCB_KHR
#ifdef VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDirectFBPresentationSupportEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDirectFBPresentationSupportEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDirectFBPresentationSupportEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDirectFBPresentationSupportEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDirectFBPresentationSupportEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDirectFBPresentationSupportEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDirectFBPresentationSupportEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDirectFBPresentationSupportEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDirectFBPresentationSupportEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDirectFBPresentationSupportEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDirectFBPresentationSupportEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDirectFBPresentationSupportEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDirectFBPresentationSupportEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDirectFBPresentationSupportEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceDirectFBPresentationSupportEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceDirectFBPresentationSupportEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceScreenPresentationSupportQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceScreenPresentationSupportQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceScreenPresentationSupportQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceScreenPresentationSupportQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceScreenPresentationSupportQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceScreenPresentationSupportQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceScreenPresentationSupportQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceScreenPresentationSupportQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceScreenPresentationSupportQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceScreenPresentationSupportQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceScreenPresentationSupportQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceScreenPresentationSupportQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceScreenPresentationSupportQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceScreenPresentationSupportQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceScreenPresentationSupportQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceScreenPresentationSupportQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalImageFormatPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalImageFormatPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalImageFormatPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalImageFormatPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalImageFormatPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalImageFormatPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalImageFormatPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalImageFormatPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalImageFormatPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalImageFormatPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalImageFormatPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalImageFormatPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalImageFormatPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalImageFormatPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceExternalImageFormatPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceExternalImageFormatPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceFeatures2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceFeatures2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFeatures2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceFeatures2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceFeatures2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceFormatProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceFormatProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFormatProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceFormatProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceFormatProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceImageFormatProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceImageFormatProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceImageFormatProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceImageFormatProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceImageFormatProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceQueueFamilyProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceQueueFamilyProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceQueueFamilyProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceQueueFamilyProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceMemoryProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceMemoryProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMemoryProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceMemoryProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceMemoryProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceSparseImageFormatProperties2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceSparseImageFormatProperties2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSparseImageFormatProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceSparseImageFormatProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceSparseImageFormatProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceExternalBufferProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceExternalBufferProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalBufferPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceExternalBufferPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceExternalBufferPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphoreProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphoreProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphoreProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphoreProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphoreProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphoreProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphoreProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphoreProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphoreProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphoreProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphoreProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphoreProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphoreProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphoreProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceExternalSemaphoreProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceExternalSemaphoreProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalSemaphorePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceExternalSemaphorePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceExternalSemaphorePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFenceProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFenceProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFenceProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFenceProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFenceProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFenceProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFenceProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFenceProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFenceProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFenceProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFenceProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFenceProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFenceProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFenceProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceExternalFenceProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceExternalFenceProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFencePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFencePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFencePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFencePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFencePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFencePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFencePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFencePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFencePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFencePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFencePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFencePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFencePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalFencePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceExternalFencePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceExternalFencePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ReleaseDisplayEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseDisplayEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseDisplayEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseDisplayEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseDisplayEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseDisplayEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseDisplayEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseDisplayEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseDisplayEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseDisplayEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseDisplayEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseDisplayEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseDisplayEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseDisplayEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ReleaseDisplayEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ReleaseDisplayEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_AcquireXlibDisplayEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireXlibDisplayEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireXlibDisplayEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireXlibDisplayEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireXlibDisplayEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireXlibDisplayEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireXlibDisplayEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireXlibDisplayEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireXlibDisplayEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireXlibDisplayEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireXlibDisplayEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireXlibDisplayEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireXlibDisplayEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireXlibDisplayEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_AcquireXlibDisplayEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_AcquireXlibDisplayEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetRandROutputDisplayEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRandROutputDisplayEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRandROutputDisplayEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRandROutputDisplayEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRandROutputDisplayEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRandROutputDisplayEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRandROutputDisplayEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRandROutputDisplayEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRandROutputDisplayEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRandROutputDisplayEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRandROutputDisplayEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRandROutputDisplayEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRandROutputDisplayEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRandROutputDisplayEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetRandROutputDisplayEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetRandROutputDisplayEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_AcquireWinrtDisplayNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireWinrtDisplayNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireWinrtDisplayNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireWinrtDisplayNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireWinrtDisplayNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireWinrtDisplayNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireWinrtDisplayNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireWinrtDisplayNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireWinrtDisplayNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireWinrtDisplayNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireWinrtDisplayNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireWinrtDisplayNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireWinrtDisplayNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireWinrtDisplayNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_AcquireWinrtDisplayNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_AcquireWinrtDisplayNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetWinrtDisplayNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetWinrtDisplayNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetWinrtDisplayNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetWinrtDisplayNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetWinrtDisplayNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetWinrtDisplayNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetWinrtDisplayNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetWinrtDisplayNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetWinrtDisplayNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetWinrtDisplayNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetWinrtDisplayNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetWinrtDisplayNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetWinrtDisplayNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetWinrtDisplayNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetWinrtDisplayNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetWinrtDisplayNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceSurfaceCapabilities2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceSurfaceCapabilities2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDevicePresentRectanglesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDevicePresentRectanglesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDevicePresentRectanglesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDevicePresentRectanglesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDevicePresentRectanglesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDevicePresentRectanglesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDevicePresentRectanglesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDevicePresentRectanglesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDevicePresentRectanglesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDevicePresentRectanglesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDevicePresentRectanglesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDevicePresentRectanglesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDevicePresentRectanglesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDevicePresentRectanglesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDevicePresentRectanglesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDevicePresentRectanglesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMultisamplePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMultisamplePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMultisamplePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMultisamplePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMultisamplePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMultisamplePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMultisamplePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMultisamplePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMultisamplePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMultisamplePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMultisamplePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMultisamplePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMultisamplePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceMultisamplePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceMultisamplePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceMultisamplePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceCapabilities2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceSurfaceCapabilities2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceSurfaceCapabilities2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormats2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormats2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormats2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormats2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormats2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormats2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormats2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormats2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormats2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormats2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormats2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormats2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormats2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfaceFormats2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceSurfaceFormats2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceSurfaceFormats2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceDisplayProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceDisplayProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlaneProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlaneProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlaneProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlaneProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlaneProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlaneProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlaneProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlaneProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlaneProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlaneProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlaneProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlaneProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlaneProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDisplayPlaneProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceDisplayPlaneProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceDisplayPlaneProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModeProperties2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModeProperties2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModeProperties2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModeProperties2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModeProperties2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModeProperties2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModeProperties2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModeProperties2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModeProperties2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModeProperties2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModeProperties2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModeProperties2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModeProperties2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayModeProperties2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDisplayModeProperties2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDisplayModeProperties2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilities2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilities2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilities2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilities2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilities2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilities2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilities2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilities2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilities2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilities2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilities2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilities2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilities2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDisplayPlaneCapabilities2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDisplayPlaneCapabilities2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDisplayPlaneCapabilities2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceCalibrateableTimeDomainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceCalibrateableTimeDomainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCalibrateableTimeDomainsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceCalibrateableTimeDomainsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceCalibrateableTimeDomainsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceCooperativeMatrixPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceCooperativeMatrixPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolProperties@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolProperties@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolProperties@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolProperties@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolProperties@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolProperties@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolProperties@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolProperties@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolProperties@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolProperties@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolProperties@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolProperties@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolProperties@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolProperties@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceToolProperties=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceToolProperties=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceToolPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceToolPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceToolPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFragmentShadingRatesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFragmentShadingRatesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFragmentShadingRatesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFragmentShadingRatesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFragmentShadingRatesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFragmentShadingRatesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFragmentShadingRatesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFragmentShadingRatesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFragmentShadingRatesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFragmentShadingRatesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFragmentShadingRatesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFragmentShadingRatesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFragmentShadingRatesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceFragmentShadingRatesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceFragmentShadingRatesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceFragmentShadingRatesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceVideoCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceVideoCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoFormatPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoFormatPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoFormatPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoFormatPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoFormatPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoFormatPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoFormatPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoFormatPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoFormatPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoFormatPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoFormatPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoFormatPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoFormatPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoFormatPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceVideoFormatPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceVideoFormatPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_AcquireDrmDisplayEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireDrmDisplayEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireDrmDisplayEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireDrmDisplayEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireDrmDisplayEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireDrmDisplayEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireDrmDisplayEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireDrmDisplayEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireDrmDisplayEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireDrmDisplayEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireDrmDisplayEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireDrmDisplayEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireDrmDisplayEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireDrmDisplayEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_AcquireDrmDisplayEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_AcquireDrmDisplayEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDrmDisplayEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDrmDisplayEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDrmDisplayEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDrmDisplayEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDrmDisplayEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDrmDisplayEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDrmDisplayEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDrmDisplayEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDrmDisplayEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDrmDisplayEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDrmDisplayEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDrmDisplayEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDrmDisplayEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDrmDisplayEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDrmDisplayEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDrmDisplayEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceOpticalFlowImageFormatsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceOpticalFlowImageFormatsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceOpticalFlowImageFormatsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceOpticalFlowImageFormatsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceOpticalFlowImageFormatsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceOpticalFlowImageFormatsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceOpticalFlowImageFormatsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceOpticalFlowImageFormatsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceOpticalFlowImageFormatsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceOpticalFlowImageFormatsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceOpticalFlowImageFormatsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceOpticalFlowImageFormatsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceOpticalFlowImageFormatsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceOpticalFlowImageFormatsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceOpticalFlowImageFormatsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceOpticalFlowImageFormatsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeVectorPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeVectorPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeVectorPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeVectorPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeVectorPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeVectorPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeVectorPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeVectorPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeVectorPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeVectorPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeVectorPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeVectorPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeVectorPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceCooperativeVectorPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceCooperativeVectorPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceCooperativeVectorPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalTensorPropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalTensorPropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalTensorPropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalTensorPropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalTensorPropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalTensorPropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalTensorPropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalTensorPropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalTensorPropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalTensorPropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalTensorPropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalTensorPropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalTensorPropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceExternalTensorPropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceExternalTensorPropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceExternalTensorPropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDescriptorSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDescriptorSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDescriptorSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDescriptorSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDescriptorSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDescriptorSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDescriptorSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDescriptorSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDescriptorSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDescriptorSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDescriptorSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDescriptorSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDescriptorSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceDescriptorSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceDescriptorSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceDescriptorSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif


const struct vk_physical_device_entrypoint_table anv_physical_device_entrypoints = {
    .GetPhysicalDeviceProperties = anv_GetPhysicalDeviceProperties,
    .GetPhysicalDeviceQueueFamilyProperties = anv_GetPhysicalDeviceQueueFamilyProperties,
    .GetPhysicalDeviceMemoryProperties = anv_GetPhysicalDeviceMemoryProperties,
    .GetPhysicalDeviceFeatures = anv_GetPhysicalDeviceFeatures,
    .GetPhysicalDeviceFormatProperties = anv_GetPhysicalDeviceFormatProperties,
    .GetPhysicalDeviceImageFormatProperties = anv_GetPhysicalDeviceImageFormatProperties,
    .CreateDevice = anv_CreateDevice,
    .EnumerateDeviceLayerProperties = anv_EnumerateDeviceLayerProperties,
    .EnumerateDeviceExtensionProperties = anv_EnumerateDeviceExtensionProperties,
    .GetPhysicalDeviceSparseImageFormatProperties = anv_GetPhysicalDeviceSparseImageFormatProperties,
    .GetPhysicalDeviceDisplayPropertiesKHR = anv_GetPhysicalDeviceDisplayPropertiesKHR,
    .GetPhysicalDeviceDisplayPlanePropertiesKHR = anv_GetPhysicalDeviceDisplayPlanePropertiesKHR,
    .GetDisplayPlaneSupportedDisplaysKHR = anv_GetDisplayPlaneSupportedDisplaysKHR,
    .GetDisplayModePropertiesKHR = anv_GetDisplayModePropertiesKHR,
    .CreateDisplayModeKHR = anv_CreateDisplayModeKHR,
    .GetDisplayPlaneCapabilitiesKHR = anv_GetDisplayPlaneCapabilitiesKHR,
    .GetPhysicalDeviceSurfaceSupportKHR = anv_GetPhysicalDeviceSurfaceSupportKHR,
    .GetPhysicalDeviceSurfaceCapabilitiesKHR = anv_GetPhysicalDeviceSurfaceCapabilitiesKHR,
    .GetPhysicalDeviceSurfaceFormatsKHR = anv_GetPhysicalDeviceSurfaceFormatsKHR,
    .GetPhysicalDeviceSurfacePresentModesKHR = anv_GetPhysicalDeviceSurfacePresentModesKHR,
#ifdef VK_USE_PLATFORM_WAYLAND_KHR
    .GetPhysicalDeviceWaylandPresentationSupportKHR = anv_GetPhysicalDeviceWaylandPresentationSupportKHR,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceWaylandPresentationSupportKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WAYLAND_KHR
#ifdef VK_USE_PLATFORM_UBM_SEC
    .GetPhysicalDeviceUbmPresentationSupportSEC = anv_GetPhysicalDeviceUbmPresentationSupportSEC,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceUbmPresentationSupportSEC = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_UBM_SEC
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetPhysicalDeviceWin32PresentationSupportKHR = anv_GetPhysicalDeviceWin32PresentationSupportKHR,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceWin32PresentationSupportKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_XLIB_KHR
    .GetPhysicalDeviceXlibPresentationSupportKHR = anv_GetPhysicalDeviceXlibPresentationSupportKHR,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceXlibPresentationSupportKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XLIB_KHR
#ifdef VK_USE_PLATFORM_XCB_KHR
    .GetPhysicalDeviceXcbPresentationSupportKHR = anv_GetPhysicalDeviceXcbPresentationSupportKHR,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceXcbPresentationSupportKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XCB_KHR
#ifdef VK_USE_PLATFORM_DIRECTFB_EXT
    .GetPhysicalDeviceDirectFBPresentationSupportEXT = anv_GetPhysicalDeviceDirectFBPresentationSupportEXT,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceDirectFBPresentationSupportEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_DIRECTFB_EXT
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetPhysicalDeviceScreenPresentationSupportQNX = anv_GetPhysicalDeviceScreenPresentationSupportQNX,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceScreenPresentationSupportQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .GetPhysicalDeviceExternalImageFormatPropertiesNV = anv_GetPhysicalDeviceExternalImageFormatPropertiesNV,
    .GetPhysicalDeviceFeatures2 = anv_GetPhysicalDeviceFeatures2,
    .GetPhysicalDeviceFeatures2KHR = anv_GetPhysicalDeviceFeatures2KHR,
    .GetPhysicalDeviceProperties2 = anv_GetPhysicalDeviceProperties2,
    .GetPhysicalDeviceProperties2KHR = anv_GetPhysicalDeviceProperties2KHR,
    .GetPhysicalDeviceFormatProperties2 = anv_GetPhysicalDeviceFormatProperties2,
    .GetPhysicalDeviceFormatProperties2KHR = anv_GetPhysicalDeviceFormatProperties2KHR,
    .GetPhysicalDeviceImageFormatProperties2 = anv_GetPhysicalDeviceImageFormatProperties2,
    .GetPhysicalDeviceImageFormatProperties2KHR = anv_GetPhysicalDeviceImageFormatProperties2KHR,
    .GetPhysicalDeviceQueueFamilyProperties2 = anv_GetPhysicalDeviceQueueFamilyProperties2,
    .GetPhysicalDeviceQueueFamilyProperties2KHR = anv_GetPhysicalDeviceQueueFamilyProperties2KHR,
    .GetPhysicalDeviceMemoryProperties2 = anv_GetPhysicalDeviceMemoryProperties2,
    .GetPhysicalDeviceMemoryProperties2KHR = anv_GetPhysicalDeviceMemoryProperties2KHR,
    .GetPhysicalDeviceSparseImageFormatProperties2 = anv_GetPhysicalDeviceSparseImageFormatProperties2,
    .GetPhysicalDeviceSparseImageFormatProperties2KHR = anv_GetPhysicalDeviceSparseImageFormatProperties2KHR,
    .GetPhysicalDeviceExternalBufferProperties = anv_GetPhysicalDeviceExternalBufferProperties,
    .GetPhysicalDeviceExternalBufferPropertiesKHR = anv_GetPhysicalDeviceExternalBufferPropertiesKHR,
    .GetPhysicalDeviceExternalSemaphoreProperties = anv_GetPhysicalDeviceExternalSemaphoreProperties,
    .GetPhysicalDeviceExternalSemaphorePropertiesKHR = anv_GetPhysicalDeviceExternalSemaphorePropertiesKHR,
    .GetPhysicalDeviceExternalFenceProperties = anv_GetPhysicalDeviceExternalFenceProperties,
    .GetPhysicalDeviceExternalFencePropertiesKHR = anv_GetPhysicalDeviceExternalFencePropertiesKHR,
    .ReleaseDisplayEXT = anv_ReleaseDisplayEXT,
#ifdef VK_USE_PLATFORM_XLIB_XRANDR_EXT
    .AcquireXlibDisplayEXT = anv_AcquireXlibDisplayEXT,
#elif defined(_MSC_VER)
    .AcquireXlibDisplayEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef VK_USE_PLATFORM_XLIB_XRANDR_EXT
    .GetRandROutputDisplayEXT = anv_GetRandROutputDisplayEXT,
#elif defined(_MSC_VER)
    .GetRandROutputDisplayEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_XLIB_XRANDR_EXT
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireWinrtDisplayNV = anv_AcquireWinrtDisplayNV,
#elif defined(_MSC_VER)
    .AcquireWinrtDisplayNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetWinrtDisplayNV = anv_GetWinrtDisplayNV,
#elif defined(_MSC_VER)
    .GetWinrtDisplayNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetPhysicalDeviceSurfaceCapabilities2EXT = anv_GetPhysicalDeviceSurfaceCapabilities2EXT,
    .GetPhysicalDevicePresentRectanglesKHR = anv_GetPhysicalDevicePresentRectanglesKHR,
    .GetPhysicalDeviceMultisamplePropertiesEXT = anv_GetPhysicalDeviceMultisamplePropertiesEXT,
    .GetPhysicalDeviceSurfaceCapabilities2KHR = anv_GetPhysicalDeviceSurfaceCapabilities2KHR,
    .GetPhysicalDeviceSurfaceFormats2KHR = anv_GetPhysicalDeviceSurfaceFormats2KHR,
    .GetPhysicalDeviceDisplayProperties2KHR = anv_GetPhysicalDeviceDisplayProperties2KHR,
    .GetPhysicalDeviceDisplayPlaneProperties2KHR = anv_GetPhysicalDeviceDisplayPlaneProperties2KHR,
    .GetDisplayModeProperties2KHR = anv_GetDisplayModeProperties2KHR,
    .GetDisplayPlaneCapabilities2KHR = anv_GetDisplayPlaneCapabilities2KHR,
    .GetPhysicalDeviceCalibrateableTimeDomainsKHR = anv_GetPhysicalDeviceCalibrateableTimeDomainsKHR,
    .GetPhysicalDeviceCalibrateableTimeDomainsEXT = anv_GetPhysicalDeviceCalibrateableTimeDomainsEXT,
    .GetPhysicalDeviceCooperativeMatrixPropertiesNV = anv_GetPhysicalDeviceCooperativeMatrixPropertiesNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetPhysicalDeviceSurfacePresentModes2EXT = anv_GetPhysicalDeviceSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetPhysicalDeviceSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR = anv_EnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR,
    .GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR = anv_GetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR,
    .GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV = anv_GetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV,
    .GetPhysicalDeviceToolProperties = anv_GetPhysicalDeviceToolProperties,
    .GetPhysicalDeviceToolPropertiesEXT = anv_GetPhysicalDeviceToolPropertiesEXT,
    .GetPhysicalDeviceFragmentShadingRatesKHR = anv_GetPhysicalDeviceFragmentShadingRatesKHR,
    .GetPhysicalDeviceVideoCapabilitiesKHR = anv_GetPhysicalDeviceVideoCapabilitiesKHR,
    .GetPhysicalDeviceVideoFormatPropertiesKHR = anv_GetPhysicalDeviceVideoFormatPropertiesKHR,
    .GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR = anv_GetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR,
    .AcquireDrmDisplayEXT = anv_AcquireDrmDisplayEXT,
    .GetDrmDisplayEXT = anv_GetDrmDisplayEXT,
    .GetPhysicalDeviceOpticalFlowImageFormatsNV = anv_GetPhysicalDeviceOpticalFlowImageFormatsNV,
    .GetPhysicalDeviceCooperativeMatrixPropertiesKHR = anv_GetPhysicalDeviceCooperativeMatrixPropertiesKHR,
    .GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV = anv_GetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV,
    .GetPhysicalDeviceCooperativeVectorPropertiesNV = anv_GetPhysicalDeviceCooperativeVectorPropertiesNV,
    .EnumeratePhysicalDeviceShaderInstrumentationMetricsARM = anv_EnumeratePhysicalDeviceShaderInstrumentationMetricsARM,
    .GetPhysicalDeviceExternalTensorPropertiesARM = anv_GetPhysicalDeviceExternalTensorPropertiesARM,
    .GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM = anv_GetPhysicalDeviceQueueFamilyDataGraphPropertiesARM,
    .GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM = anv_GetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM,
    .EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM = anv_EnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM,
    .GetPhysicalDeviceDescriptorSizeEXT = anv_GetPhysicalDeviceDescriptorSizeEXT,
    .GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM = anv_GetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM,
};



#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceProcAddr@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceProcAddr@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceProcAddr@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceProcAddr@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceProcAddr@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceProcAddr@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceProcAddr@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceProcAddr@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceProcAddr@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceProcAddr@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceProcAddr@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceProcAddr@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceProcAddr@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceProcAddr@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceProcAddr=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceProcAddr=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDevice@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDevice@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDevice@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDevice@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDevice@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDevice@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDevice@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDevice@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDevice@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDevice@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDevice@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDevice@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDevice@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDevice@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyDevice=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyDevice=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceQueue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceQueue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceQueue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceQueue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceQueue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceQueue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceQueue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceQueue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceQueue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceQueue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceQueue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceQueue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceQueue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceQueue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceQueue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceQueue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceQueue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceQueue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceQueue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceQueue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceQueue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceQueue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceQueue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceQueue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceQueue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceQueue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_QueueSubmit=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_QueueSubmit=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_QueueSubmit=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_QueueSubmit=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_QueueSubmit=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_QueueSubmit=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_QueueSubmit=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_QueueSubmit=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_QueueSubmit=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_QueueSubmit=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_QueueSubmit=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_QueueSubmit=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_QueueSubmit=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_QueueSubmit=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_QueueSubmit=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_QueueSubmit=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_QueueSubmit=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_QueueSubmit=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_QueueSubmit=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_QueueSubmit=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_QueueSubmit=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_QueueSubmit=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_QueueSubmit=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_QueueSubmit=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_QueueSubmit=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_QueueSubmit=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_QueueWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_QueueWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_QueueWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_QueueWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_QueueWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_QueueWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_QueueWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_QueueWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_QueueWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_QueueWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_QueueWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_QueueWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_QueueWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_QueueWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_QueueWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_QueueWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_QueueWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_QueueWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_QueueWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_QueueWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_QueueWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_QueueWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_QueueWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_QueueWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_QueueWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_QueueWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_QueueWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_QueueWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_QueueWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_QueueWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_QueueWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_QueueWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_QueueWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_QueueWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_QueueWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DeviceWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeviceWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeviceWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeviceWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeviceWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeviceWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeviceWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeviceWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeviceWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeviceWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeviceWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeviceWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeviceWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeviceWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DeviceWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DeviceWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DeviceWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeviceWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeviceWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeviceWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeviceWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeviceWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeviceWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeviceWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeviceWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeviceWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeviceWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeviceWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeviceWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeviceWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DeviceWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DeviceWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DeviceWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeviceWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeviceWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeviceWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeviceWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeviceWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeviceWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeviceWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeviceWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeviceWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeviceWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeviceWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeviceWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeviceWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DeviceWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DeviceWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DeviceWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeviceWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeviceWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeviceWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeviceWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeviceWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeviceWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeviceWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeviceWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeviceWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeviceWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeviceWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeviceWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeviceWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DeviceWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DeviceWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DeviceWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeviceWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeviceWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeviceWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeviceWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeviceWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeviceWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeviceWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeviceWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeviceWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeviceWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeviceWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeviceWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeviceWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DeviceWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DeviceWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DeviceWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeviceWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeviceWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeviceWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeviceWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeviceWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeviceWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeviceWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeviceWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeviceWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeviceWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeviceWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeviceWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeviceWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DeviceWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DeviceWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DeviceWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeviceWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeviceWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeviceWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeviceWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeviceWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeviceWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeviceWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeviceWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeviceWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeviceWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeviceWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeviceWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeviceWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DeviceWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DeviceWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DeviceWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeviceWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeviceWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeviceWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeviceWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeviceWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeviceWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeviceWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeviceWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeviceWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeviceWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeviceWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeviceWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeviceWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DeviceWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DeviceWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DeviceWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeviceWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeviceWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeviceWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeviceWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeviceWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeviceWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeviceWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeviceWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeviceWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeviceWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeviceWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeviceWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeviceWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DeviceWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DeviceWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DeviceWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeviceWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeviceWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeviceWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeviceWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeviceWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeviceWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeviceWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeviceWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeviceWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeviceWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeviceWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeviceWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeviceWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DeviceWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DeviceWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeviceWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeviceWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeviceWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeviceWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeviceWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeviceWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeviceWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeviceWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeviceWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeviceWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeviceWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeviceWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeviceWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeviceWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DeviceWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DeviceWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DeviceWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeviceWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeviceWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeviceWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeviceWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeviceWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeviceWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeviceWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeviceWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeviceWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeviceWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeviceWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeviceWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeviceWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DeviceWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DeviceWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DeviceWaitIdle@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeviceWaitIdle@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeviceWaitIdle@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeviceWaitIdle@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeviceWaitIdle@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeviceWaitIdle@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeviceWaitIdle@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeviceWaitIdle@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeviceWaitIdle@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeviceWaitIdle@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeviceWaitIdle@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeviceWaitIdle@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeviceWaitIdle@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeviceWaitIdle@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DeviceWaitIdle=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DeviceWaitIdle=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_AllocateMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_AllocateMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_AllocateMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_AllocateMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_AllocateMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_AllocateMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_AllocateMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_AllocateMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_AllocateMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_AllocateMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_AllocateMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_AllocateMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_AllocateMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_AllocateMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_AllocateMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_AllocateMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_AllocateMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_AllocateMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_AllocateMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_AllocateMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_AllocateMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_AllocateMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_AllocateMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_AllocateMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_AllocateMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_AllocateMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_AllocateMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_AllocateMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_AllocateMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_AllocateMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_AllocateMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_AllocateMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_AllocateMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_AllocateMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_AllocateMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_FreeMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_FreeMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_FreeMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_FreeMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_FreeMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_FreeMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_FreeMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_FreeMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_FreeMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_FreeMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_FreeMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_FreeMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_FreeMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_FreeMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_FreeMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_FreeMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_FreeMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_FreeMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_FreeMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_FreeMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_FreeMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_FreeMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_FreeMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_FreeMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_FreeMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_FreeMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_FreeMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_FreeMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_FreeMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_FreeMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_FreeMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_FreeMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_FreeMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_FreeMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_FreeMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_MapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_MapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_MapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_MapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_MapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_MapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_MapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_MapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_MapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_MapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_MapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_MapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_MapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_MapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_MapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_MapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_MapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_MapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_MapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_MapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_MapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_MapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_MapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_MapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_MapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_MapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_MapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_UnmapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_UnmapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_UnmapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_UnmapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_UnmapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_UnmapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_UnmapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_UnmapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_UnmapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_UnmapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_UnmapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_UnmapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_UnmapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_UnmapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_UnmapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_UnmapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_UnmapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_UnmapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_UnmapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_UnmapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_UnmapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_UnmapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_UnmapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_UnmapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_UnmapMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_UnmapMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_FlushMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FlushMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FlushMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FlushMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FlushMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FlushMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FlushMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FlushMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FlushMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FlushMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FlushMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FlushMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FlushMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FlushMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_FlushMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_FlushMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_FlushMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FlushMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FlushMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FlushMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FlushMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FlushMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FlushMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FlushMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FlushMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FlushMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FlushMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FlushMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FlushMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FlushMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_FlushMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_FlushMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_FlushMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FlushMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FlushMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FlushMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FlushMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FlushMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FlushMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FlushMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FlushMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FlushMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FlushMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FlushMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FlushMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FlushMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_FlushMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_FlushMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_FlushMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FlushMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FlushMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FlushMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FlushMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FlushMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FlushMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FlushMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FlushMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FlushMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FlushMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FlushMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FlushMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FlushMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_FlushMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_FlushMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_FlushMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FlushMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FlushMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FlushMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FlushMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FlushMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FlushMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FlushMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FlushMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FlushMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FlushMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FlushMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FlushMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FlushMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_FlushMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_FlushMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_FlushMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FlushMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FlushMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FlushMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FlushMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FlushMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FlushMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FlushMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FlushMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FlushMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FlushMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FlushMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FlushMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FlushMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_FlushMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_FlushMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_FlushMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FlushMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FlushMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FlushMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FlushMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FlushMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FlushMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FlushMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FlushMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FlushMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FlushMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FlushMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FlushMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FlushMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_FlushMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_FlushMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_FlushMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FlushMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FlushMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FlushMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FlushMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FlushMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FlushMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FlushMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FlushMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FlushMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FlushMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FlushMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FlushMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FlushMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_FlushMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_FlushMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_FlushMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FlushMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FlushMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FlushMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FlushMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FlushMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FlushMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FlushMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FlushMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FlushMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FlushMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FlushMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FlushMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FlushMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_FlushMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_FlushMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_FlushMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FlushMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FlushMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FlushMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FlushMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FlushMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FlushMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FlushMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FlushMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FlushMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FlushMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FlushMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FlushMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FlushMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_FlushMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_FlushMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_FlushMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FlushMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FlushMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FlushMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FlushMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FlushMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FlushMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FlushMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FlushMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FlushMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FlushMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FlushMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FlushMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FlushMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_FlushMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_FlushMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_FlushMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FlushMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FlushMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FlushMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FlushMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FlushMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FlushMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FlushMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FlushMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FlushMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FlushMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FlushMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FlushMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FlushMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_FlushMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_FlushMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_FlushMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FlushMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FlushMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FlushMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FlushMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FlushMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FlushMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FlushMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FlushMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FlushMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FlushMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FlushMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FlushMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FlushMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_FlushMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_FlushMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_InvalidateMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InvalidateMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InvalidateMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InvalidateMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InvalidateMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InvalidateMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InvalidateMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InvalidateMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InvalidateMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InvalidateMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InvalidateMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InvalidateMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InvalidateMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InvalidateMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_InvalidateMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_InvalidateMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_InvalidateMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InvalidateMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InvalidateMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InvalidateMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InvalidateMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InvalidateMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InvalidateMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InvalidateMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InvalidateMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InvalidateMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InvalidateMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InvalidateMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InvalidateMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InvalidateMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_InvalidateMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_InvalidateMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_InvalidateMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InvalidateMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InvalidateMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InvalidateMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InvalidateMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InvalidateMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InvalidateMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InvalidateMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InvalidateMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InvalidateMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InvalidateMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InvalidateMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InvalidateMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InvalidateMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_InvalidateMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_InvalidateMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_InvalidateMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InvalidateMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InvalidateMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InvalidateMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InvalidateMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InvalidateMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InvalidateMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InvalidateMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InvalidateMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InvalidateMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InvalidateMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InvalidateMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InvalidateMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InvalidateMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_InvalidateMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_InvalidateMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_InvalidateMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InvalidateMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InvalidateMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InvalidateMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InvalidateMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InvalidateMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InvalidateMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InvalidateMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InvalidateMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InvalidateMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InvalidateMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InvalidateMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InvalidateMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InvalidateMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_InvalidateMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_InvalidateMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_InvalidateMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InvalidateMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InvalidateMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InvalidateMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InvalidateMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InvalidateMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InvalidateMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InvalidateMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InvalidateMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InvalidateMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InvalidateMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InvalidateMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InvalidateMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InvalidateMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_InvalidateMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_InvalidateMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_InvalidateMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InvalidateMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InvalidateMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InvalidateMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InvalidateMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InvalidateMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InvalidateMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InvalidateMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InvalidateMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InvalidateMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InvalidateMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InvalidateMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InvalidateMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InvalidateMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_InvalidateMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_InvalidateMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_InvalidateMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InvalidateMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InvalidateMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InvalidateMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InvalidateMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InvalidateMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InvalidateMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InvalidateMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InvalidateMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InvalidateMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InvalidateMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InvalidateMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InvalidateMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InvalidateMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_InvalidateMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_InvalidateMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_InvalidateMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InvalidateMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InvalidateMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InvalidateMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InvalidateMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InvalidateMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InvalidateMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InvalidateMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InvalidateMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InvalidateMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InvalidateMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InvalidateMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InvalidateMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InvalidateMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_InvalidateMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_InvalidateMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_InvalidateMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InvalidateMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InvalidateMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InvalidateMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InvalidateMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InvalidateMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InvalidateMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InvalidateMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InvalidateMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InvalidateMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InvalidateMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InvalidateMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InvalidateMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InvalidateMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_InvalidateMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_InvalidateMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_InvalidateMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InvalidateMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InvalidateMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InvalidateMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InvalidateMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InvalidateMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InvalidateMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InvalidateMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InvalidateMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InvalidateMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InvalidateMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InvalidateMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InvalidateMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InvalidateMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_InvalidateMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_InvalidateMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_InvalidateMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InvalidateMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InvalidateMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InvalidateMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InvalidateMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InvalidateMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InvalidateMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InvalidateMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InvalidateMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InvalidateMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InvalidateMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InvalidateMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InvalidateMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InvalidateMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_InvalidateMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_InvalidateMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_InvalidateMappedMemoryRanges@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InvalidateMappedMemoryRanges@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InvalidateMappedMemoryRanges@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InvalidateMappedMemoryRanges@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InvalidateMappedMemoryRanges@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InvalidateMappedMemoryRanges@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InvalidateMappedMemoryRanges@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InvalidateMappedMemoryRanges@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InvalidateMappedMemoryRanges@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InvalidateMappedMemoryRanges@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InvalidateMappedMemoryRanges@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InvalidateMappedMemoryRanges@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InvalidateMappedMemoryRanges@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InvalidateMappedMemoryRanges@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_InvalidateMappedMemoryRanges=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_InvalidateMappedMemoryRanges=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryCommitment@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryCommitment@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryCommitment@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryCommitment@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryCommitment@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryCommitment@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryCommitment@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryCommitment@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryCommitment@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryCommitment@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryCommitment@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryCommitment@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryCommitment@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryCommitment@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceMemoryCommitment=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceMemoryCommitment=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryCommitment@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryCommitment@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryCommitment@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryCommitment@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryCommitment@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryCommitment@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryCommitment@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryCommitment@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryCommitment@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryCommitment@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryCommitment@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryCommitment@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryCommitment@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryCommitment@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceMemoryCommitment=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceMemoryCommitment=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryCommitment@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryCommitment@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryCommitment@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryCommitment@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryCommitment@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryCommitment@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryCommitment@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryCommitment@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryCommitment@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryCommitment@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryCommitment@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryCommitment@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryCommitment@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryCommitment@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceMemoryCommitment=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceMemoryCommitment=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryCommitment@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryCommitment@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryCommitment@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryCommitment@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryCommitment@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryCommitment@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryCommitment@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryCommitment@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryCommitment@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryCommitment@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryCommitment@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryCommitment@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryCommitment@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryCommitment@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceMemoryCommitment=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceMemoryCommitment=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryCommitment@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryCommitment@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryCommitment@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryCommitment@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryCommitment@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryCommitment@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryCommitment@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryCommitment@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryCommitment@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryCommitment@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryCommitment@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryCommitment@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryCommitment@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryCommitment@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceMemoryCommitment=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceMemoryCommitment=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryCommitment@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryCommitment@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryCommitment@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryCommitment@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryCommitment@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryCommitment@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryCommitment@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryCommitment@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryCommitment@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryCommitment@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryCommitment@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryCommitment@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryCommitment@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryCommitment@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceMemoryCommitment=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceMemoryCommitment=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryCommitment@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryCommitment@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryCommitment@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryCommitment@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryCommitment@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryCommitment@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryCommitment@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryCommitment@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryCommitment@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryCommitment@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryCommitment@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryCommitment@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryCommitment@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryCommitment@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceMemoryCommitment=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceMemoryCommitment=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryCommitment@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryCommitment@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryCommitment@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryCommitment@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryCommitment@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryCommitment@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryCommitment@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryCommitment@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryCommitment@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryCommitment@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryCommitment@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryCommitment@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryCommitment@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryCommitment@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceMemoryCommitment=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceMemoryCommitment=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryCommitment@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryCommitment@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryCommitment@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryCommitment@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryCommitment@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryCommitment@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryCommitment@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryCommitment@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryCommitment@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryCommitment@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryCommitment@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryCommitment@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryCommitment@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryCommitment@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceMemoryCommitment=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceMemoryCommitment=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryCommitment@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryCommitment@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryCommitment@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryCommitment@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryCommitment@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryCommitment@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryCommitment@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryCommitment@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryCommitment@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryCommitment@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryCommitment@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryCommitment@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryCommitment@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryCommitment@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceMemoryCommitment=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceMemoryCommitment=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryCommitment@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryCommitment@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryCommitment@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryCommitment@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryCommitment@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryCommitment@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryCommitment@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryCommitment@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryCommitment@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryCommitment@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryCommitment@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryCommitment@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryCommitment@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryCommitment@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceMemoryCommitment=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceMemoryCommitment=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryCommitment@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryCommitment@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryCommitment@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryCommitment@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryCommitment@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryCommitment@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryCommitment@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryCommitment@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryCommitment@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryCommitment@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryCommitment@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryCommitment@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryCommitment@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryCommitment@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceMemoryCommitment=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceMemoryCommitment=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryCommitment@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryCommitment@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryCommitment@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryCommitment@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryCommitment@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryCommitment@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryCommitment@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryCommitment@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryCommitment@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryCommitment@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryCommitment@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryCommitment@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryCommitment@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryCommitment@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceMemoryCommitment=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceMemoryCommitment=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_BindBufferMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_BindBufferMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_BindBufferMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_BindBufferMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_BindBufferMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_BindBufferMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_BindBufferMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_BindBufferMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_BindBufferMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_BindBufferMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_BindBufferMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_BindBufferMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_BindBufferMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_BindBufferMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_BindBufferMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_BindBufferMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_BindBufferMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_BindBufferMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_BindBufferMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_BindBufferMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_BindBufferMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_BindBufferMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_BindBufferMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_BindBufferMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_BindBufferMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_BindBufferMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_BindImageMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_BindImageMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_BindImageMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_BindImageMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_BindImageMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_BindImageMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_BindImageMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_BindImageMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_BindImageMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_BindImageMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_BindImageMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_BindImageMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_BindImageMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_BindImageMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_BindImageMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_BindImageMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_BindImageMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_BindImageMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_BindImageMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_BindImageMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_BindImageMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_BindImageMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_BindImageMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_BindImageMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_BindImageMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_BindImageMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_QueueBindSparse@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBindSparse@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBindSparse@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBindSparse@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBindSparse@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBindSparse@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBindSparse@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBindSparse@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBindSparse@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBindSparse@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBindSparse@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBindSparse@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBindSparse@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBindSparse@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_QueueBindSparse=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_QueueBindSparse=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_QueueBindSparse@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBindSparse@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBindSparse@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBindSparse@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBindSparse@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBindSparse@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBindSparse@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBindSparse@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBindSparse@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBindSparse@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBindSparse@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBindSparse@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBindSparse@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBindSparse@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_QueueBindSparse=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_QueueBindSparse=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_QueueBindSparse@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBindSparse@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBindSparse@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBindSparse@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBindSparse@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBindSparse@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBindSparse@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBindSparse@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBindSparse@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBindSparse@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBindSparse@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBindSparse@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBindSparse@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBindSparse@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_QueueBindSparse=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_QueueBindSparse=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_QueueBindSparse@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBindSparse@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBindSparse@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBindSparse@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBindSparse@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBindSparse@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBindSparse@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBindSparse@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBindSparse@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBindSparse@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBindSparse@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBindSparse@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBindSparse@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBindSparse@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_QueueBindSparse=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_QueueBindSparse=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_QueueBindSparse@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBindSparse@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBindSparse@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBindSparse@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBindSparse@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBindSparse@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBindSparse@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBindSparse@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBindSparse@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBindSparse@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBindSparse@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBindSparse@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBindSparse@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBindSparse@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_QueueBindSparse=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_QueueBindSparse=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_QueueBindSparse@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBindSparse@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBindSparse@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBindSparse@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBindSparse@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBindSparse@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBindSparse@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBindSparse@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBindSparse@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBindSparse@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBindSparse@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBindSparse@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBindSparse@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBindSparse@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_QueueBindSparse=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_QueueBindSparse=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_QueueBindSparse@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBindSparse@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBindSparse@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBindSparse@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBindSparse@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBindSparse@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBindSparse@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBindSparse@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBindSparse@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBindSparse@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBindSparse@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBindSparse@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBindSparse@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBindSparse@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_QueueBindSparse=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_QueueBindSparse=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_QueueBindSparse@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBindSparse@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBindSparse@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBindSparse@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBindSparse@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBindSparse@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBindSparse@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBindSparse@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBindSparse@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBindSparse@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBindSparse@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBindSparse@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBindSparse@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBindSparse@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_QueueBindSparse=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_QueueBindSparse=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBindSparse@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBindSparse@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBindSparse@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBindSparse@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBindSparse@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBindSparse@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBindSparse@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBindSparse@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBindSparse@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBindSparse@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBindSparse@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBindSparse@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBindSparse@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBindSparse@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_QueueBindSparse=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_QueueBindSparse=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBindSparse@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBindSparse@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBindSparse@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBindSparse@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBindSparse@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBindSparse@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBindSparse@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBindSparse@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBindSparse@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBindSparse@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBindSparse@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBindSparse@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBindSparse@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBindSparse@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_QueueBindSparse=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_QueueBindSparse=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBindSparse@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBindSparse@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBindSparse@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBindSparse@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBindSparse@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBindSparse@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBindSparse@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBindSparse@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBindSparse@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBindSparse@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBindSparse@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBindSparse@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBindSparse@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBindSparse@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_QueueBindSparse=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_QueueBindSparse=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_QueueBindSparse@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBindSparse@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBindSparse@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBindSparse@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBindSparse@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBindSparse@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBindSparse@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBindSparse@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBindSparse@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBindSparse@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBindSparse@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBindSparse@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBindSparse@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBindSparse@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_QueueBindSparse=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_QueueBindSparse=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBindSparse@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBindSparse@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBindSparse@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBindSparse@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBindSparse@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBindSparse@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBindSparse@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBindSparse@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBindSparse@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBindSparse@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBindSparse@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBindSparse@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBindSparse@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBindSparse@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_QueueBindSparse=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_QueueBindSparse=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFence@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFence@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFence@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFence@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFence@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFence@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFence@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFence@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFence@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFence@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFence@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFence@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFence@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFence@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyFence=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyFence=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ResetFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ResetFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ResetFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ResetFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ResetFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ResetFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ResetFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ResetFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ResetFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ResetFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ResetFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ResetFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ResetFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ResetFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ResetFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ResetFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ResetFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ResetFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ResetFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ResetFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ResetFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ResetFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ResetFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ResetFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ResetFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ResetFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ResetFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ResetFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ResetFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ResetFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ResetFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ResetFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ResetFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ResetFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ResetFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetFenceStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetFenceStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetFenceStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetFenceStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetFenceStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetFenceStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetFenceStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetFenceStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetFenceStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetFenceStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetFenceStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetFenceStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetFenceStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetFenceStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetFenceStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetFenceStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetFenceStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetFenceStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetFenceStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetFenceStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetFenceStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetFenceStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetFenceStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetFenceStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetFenceStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetFenceStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetFenceStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_WaitForFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_WaitForFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_WaitForFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_WaitForFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_WaitForFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_WaitForFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_WaitForFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_WaitForFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_WaitForFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_WaitForFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_WaitForFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_WaitForFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_WaitForFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_WaitForFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_WaitForFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_WaitForFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_WaitForFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_WaitForFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_WaitForFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_WaitForFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_WaitForFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_WaitForFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_WaitForFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_WaitForFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_WaitForFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_WaitForFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_WaitForFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_WaitForFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_WaitForFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_WaitForFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_WaitForFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_WaitForFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_WaitForFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForFences@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForFences@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForFences@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForFences@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForFences@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForFences@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForFences@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForFences@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForFences@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForFences@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForFences@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForFences@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForFences@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForFences@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_WaitForFences=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_WaitForFences=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroySemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroySemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroySemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroySemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroySemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroySemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroySemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroySemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroySemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroySemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroySemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroySemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroySemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroySemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroySemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroySemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroySemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroySemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroySemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroySemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroySemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroySemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroySemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroySemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroySemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroySemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroySemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroySemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroySemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroySemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroySemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroySemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroySemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroySemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroySemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetEventStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEventStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEventStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEventStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEventStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEventStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEventStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEventStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEventStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEventStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEventStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEventStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEventStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEventStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetEventStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetEventStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetEventStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEventStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEventStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEventStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEventStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEventStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEventStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEventStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEventStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEventStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEventStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEventStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEventStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEventStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetEventStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetEventStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetEventStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEventStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEventStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEventStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEventStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEventStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEventStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEventStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEventStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEventStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEventStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEventStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEventStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEventStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetEventStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetEventStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetEventStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEventStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEventStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEventStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEventStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEventStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEventStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEventStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEventStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEventStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEventStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEventStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEventStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEventStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetEventStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetEventStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetEventStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEventStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEventStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEventStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEventStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEventStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEventStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEventStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEventStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEventStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEventStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEventStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEventStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEventStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetEventStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetEventStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetEventStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEventStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEventStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEventStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEventStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEventStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEventStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEventStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEventStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEventStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEventStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEventStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEventStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEventStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetEventStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetEventStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetEventStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEventStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEventStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEventStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEventStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEventStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEventStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEventStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEventStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEventStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEventStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEventStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEventStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEventStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetEventStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetEventStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetEventStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEventStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEventStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEventStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEventStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEventStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEventStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEventStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEventStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEventStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEventStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEventStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEventStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEventStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetEventStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetEventStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEventStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEventStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEventStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEventStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEventStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEventStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEventStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEventStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEventStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEventStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEventStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEventStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEventStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEventStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetEventStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetEventStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEventStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEventStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEventStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEventStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEventStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEventStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEventStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEventStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEventStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEventStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEventStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEventStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEventStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEventStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetEventStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetEventStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEventStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEventStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEventStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEventStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEventStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEventStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEventStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEventStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEventStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEventStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEventStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEventStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEventStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEventStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetEventStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetEventStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetEventStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEventStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEventStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEventStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEventStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEventStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEventStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEventStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEventStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEventStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEventStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEventStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEventStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEventStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetEventStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetEventStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEventStatus@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEventStatus@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEventStatus@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEventStatus@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEventStatus@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEventStatus@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEventStatus@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEventStatus@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEventStatus@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEventStatus@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEventStatus@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEventStatus@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEventStatus@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEventStatus@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetEventStatus=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetEventStatus=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_SetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_SetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_SetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_SetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_SetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_SetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_SetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_SetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_SetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_SetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_SetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_SetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_SetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_SetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_SetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_SetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_SetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_SetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_SetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_SetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_SetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_SetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_SetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_SetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_SetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_SetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_SetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_SetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_SetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_SetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_SetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_SetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_SetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_SetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_SetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_SetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_SetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_SetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPoolEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPoolEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPoolEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPoolEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPoolEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPoolEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPoolEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPoolEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPoolEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPoolEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPoolEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPoolEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPoolEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetQueryPoolEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ResetQueryPoolEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ResetQueryPoolEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPoolEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPoolEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPoolEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPoolEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPoolEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPoolEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPoolEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPoolEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPoolEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPoolEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPoolEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPoolEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPoolEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetQueryPoolEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ResetQueryPoolEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ResetQueryPoolEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPoolEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPoolEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPoolEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPoolEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPoolEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPoolEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPoolEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPoolEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPoolEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPoolEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPoolEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPoolEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPoolEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetQueryPoolEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ResetQueryPoolEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ResetQueryPoolEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPoolEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPoolEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPoolEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPoolEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPoolEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPoolEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPoolEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPoolEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPoolEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPoolEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPoolEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPoolEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPoolEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetQueryPoolEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ResetQueryPoolEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ResetQueryPoolEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPoolEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPoolEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPoolEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPoolEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPoolEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPoolEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPoolEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPoolEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPoolEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPoolEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPoolEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPoolEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPoolEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetQueryPoolEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ResetQueryPoolEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ResetQueryPoolEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPoolEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPoolEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPoolEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPoolEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPoolEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPoolEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPoolEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPoolEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPoolEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPoolEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPoolEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPoolEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPoolEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetQueryPoolEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ResetQueryPoolEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ResetQueryPoolEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPoolEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPoolEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPoolEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPoolEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPoolEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPoolEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPoolEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPoolEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPoolEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPoolEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPoolEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPoolEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPoolEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetQueryPoolEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ResetQueryPoolEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ResetQueryPoolEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPoolEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPoolEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPoolEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPoolEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPoolEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPoolEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPoolEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPoolEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPoolEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPoolEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPoolEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPoolEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPoolEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetQueryPoolEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ResetQueryPoolEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ResetQueryPoolEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPoolEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPoolEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPoolEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPoolEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPoolEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPoolEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPoolEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPoolEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPoolEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPoolEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPoolEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPoolEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPoolEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetQueryPoolEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ResetQueryPoolEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ResetQueryPoolEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPoolEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPoolEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPoolEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPoolEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPoolEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPoolEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPoolEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPoolEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPoolEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPoolEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPoolEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPoolEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPoolEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetQueryPoolEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ResetQueryPoolEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ResetQueryPoolEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPoolEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPoolEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPoolEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPoolEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPoolEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPoolEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPoolEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPoolEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPoolEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPoolEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPoolEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPoolEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPoolEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetQueryPoolEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ResetQueryPoolEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ResetQueryPoolEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPoolEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPoolEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPoolEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPoolEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPoolEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPoolEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPoolEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPoolEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPoolEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPoolEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPoolEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPoolEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPoolEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetQueryPoolEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ResetQueryPoolEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ResetQueryPoolEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPoolEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPoolEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPoolEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPoolEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPoolEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPoolEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPoolEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPoolEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPoolEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPoolEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPoolEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPoolEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPoolEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetQueryPoolEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ResetQueryPoolEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ResetQueryPoolEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyBufferView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyBufferView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImageView@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImageView@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImageView@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImageView@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImageView@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImageView@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImageView@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImageView@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImageView@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImageView@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImageView@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImageView@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImageView@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyImageView@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyImageView=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyImageView=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderModule@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderModule@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderModule@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderModule@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderModule@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderModule@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderModule@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderModule@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderModule@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderModule@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderModule@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderModule@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderModule@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderModule@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyShaderModule=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyShaderModule=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreatePipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreatePipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreatePipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreatePipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreatePipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreatePipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreatePipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreatePipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreatePipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreatePipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreatePipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreatePipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreatePipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreatePipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreatePipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreatePipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreatePipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreatePipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreatePipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreatePipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreatePipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreatePipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreatePipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreatePipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreatePipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreatePipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyPipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyPipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyPipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyPipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyPipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyPipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyPipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyPipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyPipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyPipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyPipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyPipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyPipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyPipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyPipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyPipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyPipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyPipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyPipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyPipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyPipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyPipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyPipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyPipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineCache@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineCache@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineCache@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineCache@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineCache@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineCache@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineCache@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineCache@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineCache@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineCache@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineCache@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineCache@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineCache@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineCache@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyPipelineCache=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyPipelineCache=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPipelineCacheData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineCacheData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineCacheData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineCacheData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineCacheData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineCacheData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineCacheData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineCacheData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineCacheData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineCacheData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineCacheData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineCacheData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineCacheData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineCacheData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPipelineCacheData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPipelineCacheData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineCacheData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineCacheData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineCacheData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineCacheData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineCacheData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineCacheData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineCacheData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineCacheData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineCacheData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineCacheData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineCacheData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineCacheData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineCacheData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineCacheData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetPipelineCacheData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetPipelineCacheData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineCacheData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineCacheData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineCacheData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineCacheData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineCacheData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineCacheData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineCacheData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineCacheData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineCacheData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineCacheData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineCacheData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineCacheData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineCacheData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineCacheData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetPipelineCacheData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetPipelineCacheData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineCacheData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineCacheData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineCacheData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineCacheData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineCacheData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineCacheData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineCacheData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineCacheData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineCacheData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineCacheData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineCacheData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineCacheData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineCacheData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineCacheData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetPipelineCacheData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetPipelineCacheData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineCacheData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineCacheData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineCacheData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineCacheData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineCacheData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineCacheData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineCacheData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineCacheData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineCacheData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineCacheData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineCacheData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineCacheData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineCacheData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineCacheData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetPipelineCacheData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetPipelineCacheData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineCacheData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineCacheData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineCacheData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineCacheData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineCacheData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineCacheData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineCacheData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineCacheData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineCacheData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineCacheData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineCacheData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineCacheData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineCacheData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineCacheData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetPipelineCacheData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetPipelineCacheData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineCacheData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineCacheData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineCacheData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineCacheData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineCacheData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineCacheData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineCacheData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineCacheData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineCacheData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineCacheData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineCacheData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineCacheData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineCacheData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineCacheData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetPipelineCacheData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetPipelineCacheData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineCacheData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineCacheData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineCacheData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineCacheData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineCacheData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineCacheData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineCacheData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineCacheData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineCacheData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineCacheData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineCacheData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineCacheData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineCacheData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineCacheData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetPipelineCacheData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetPipelineCacheData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineCacheData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineCacheData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineCacheData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineCacheData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineCacheData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineCacheData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineCacheData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineCacheData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineCacheData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineCacheData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineCacheData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineCacheData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineCacheData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineCacheData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetPipelineCacheData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetPipelineCacheData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineCacheData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineCacheData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineCacheData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineCacheData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineCacheData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineCacheData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineCacheData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineCacheData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineCacheData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineCacheData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineCacheData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineCacheData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineCacheData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineCacheData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetPipelineCacheData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetPipelineCacheData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineCacheData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineCacheData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineCacheData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineCacheData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineCacheData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineCacheData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineCacheData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineCacheData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineCacheData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineCacheData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineCacheData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineCacheData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineCacheData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineCacheData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetPipelineCacheData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetPipelineCacheData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineCacheData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineCacheData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineCacheData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineCacheData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineCacheData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineCacheData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineCacheData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineCacheData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineCacheData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineCacheData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineCacheData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineCacheData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineCacheData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineCacheData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetPipelineCacheData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetPipelineCacheData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineCacheData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineCacheData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineCacheData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineCacheData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineCacheData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineCacheData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineCacheData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineCacheData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineCacheData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineCacheData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineCacheData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineCacheData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineCacheData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineCacheData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetPipelineCacheData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetPipelineCacheData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_MergePipelineCaches@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergePipelineCaches@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergePipelineCaches@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergePipelineCaches@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergePipelineCaches@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergePipelineCaches@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergePipelineCaches@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergePipelineCaches@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergePipelineCaches@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergePipelineCaches@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergePipelineCaches@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergePipelineCaches@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergePipelineCaches@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergePipelineCaches@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_MergePipelineCaches=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_MergePipelineCaches=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_MergePipelineCaches@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergePipelineCaches@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergePipelineCaches@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergePipelineCaches@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergePipelineCaches@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergePipelineCaches@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergePipelineCaches@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergePipelineCaches@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergePipelineCaches@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergePipelineCaches@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergePipelineCaches@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergePipelineCaches@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergePipelineCaches@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergePipelineCaches@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_MergePipelineCaches=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_MergePipelineCaches=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_MergePipelineCaches@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergePipelineCaches@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergePipelineCaches@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergePipelineCaches@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergePipelineCaches@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergePipelineCaches@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergePipelineCaches@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergePipelineCaches@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergePipelineCaches@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergePipelineCaches@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergePipelineCaches@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergePipelineCaches@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergePipelineCaches@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergePipelineCaches@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_MergePipelineCaches=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_MergePipelineCaches=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_MergePipelineCaches@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergePipelineCaches@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergePipelineCaches@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergePipelineCaches@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergePipelineCaches@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergePipelineCaches@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergePipelineCaches@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergePipelineCaches@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergePipelineCaches@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergePipelineCaches@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergePipelineCaches@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergePipelineCaches@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergePipelineCaches@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergePipelineCaches@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_MergePipelineCaches=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_MergePipelineCaches=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_MergePipelineCaches@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergePipelineCaches@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergePipelineCaches@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergePipelineCaches@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergePipelineCaches@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergePipelineCaches@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergePipelineCaches@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergePipelineCaches@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergePipelineCaches@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergePipelineCaches@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergePipelineCaches@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergePipelineCaches@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergePipelineCaches@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergePipelineCaches@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_MergePipelineCaches=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_MergePipelineCaches=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_MergePipelineCaches@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergePipelineCaches@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergePipelineCaches@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergePipelineCaches@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergePipelineCaches@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergePipelineCaches@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergePipelineCaches@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergePipelineCaches@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergePipelineCaches@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergePipelineCaches@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergePipelineCaches@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergePipelineCaches@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergePipelineCaches@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergePipelineCaches@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_MergePipelineCaches=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_MergePipelineCaches=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_MergePipelineCaches@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergePipelineCaches@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergePipelineCaches@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergePipelineCaches@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergePipelineCaches@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergePipelineCaches@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergePipelineCaches@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergePipelineCaches@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergePipelineCaches@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergePipelineCaches@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergePipelineCaches@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergePipelineCaches@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergePipelineCaches@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergePipelineCaches@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_MergePipelineCaches=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_MergePipelineCaches=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_MergePipelineCaches@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergePipelineCaches@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergePipelineCaches@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergePipelineCaches@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergePipelineCaches@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergePipelineCaches@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergePipelineCaches@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergePipelineCaches@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergePipelineCaches@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergePipelineCaches@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergePipelineCaches@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergePipelineCaches@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergePipelineCaches@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergePipelineCaches@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_MergePipelineCaches=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_MergePipelineCaches=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_MergePipelineCaches@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergePipelineCaches@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergePipelineCaches@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergePipelineCaches@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergePipelineCaches@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergePipelineCaches@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergePipelineCaches@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergePipelineCaches@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergePipelineCaches@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergePipelineCaches@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergePipelineCaches@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergePipelineCaches@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergePipelineCaches@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergePipelineCaches@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_MergePipelineCaches=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_MergePipelineCaches=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_MergePipelineCaches@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergePipelineCaches@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergePipelineCaches@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergePipelineCaches@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergePipelineCaches@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergePipelineCaches@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergePipelineCaches@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergePipelineCaches@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergePipelineCaches@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergePipelineCaches@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergePipelineCaches@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergePipelineCaches@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergePipelineCaches@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergePipelineCaches@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_MergePipelineCaches=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_MergePipelineCaches=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergePipelineCaches@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergePipelineCaches@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergePipelineCaches@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergePipelineCaches@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergePipelineCaches@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergePipelineCaches@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergePipelineCaches@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergePipelineCaches@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergePipelineCaches@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergePipelineCaches@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergePipelineCaches@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergePipelineCaches@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergePipelineCaches@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergePipelineCaches@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_MergePipelineCaches=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_MergePipelineCaches=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_MergePipelineCaches@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergePipelineCaches@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergePipelineCaches@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergePipelineCaches@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergePipelineCaches@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergePipelineCaches@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergePipelineCaches@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergePipelineCaches@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergePipelineCaches@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergePipelineCaches@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergePipelineCaches@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergePipelineCaches@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergePipelineCaches@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergePipelineCaches@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_MergePipelineCaches=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_MergePipelineCaches=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_MergePipelineCaches@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergePipelineCaches@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergePipelineCaches@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergePipelineCaches@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergePipelineCaches@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergePipelineCaches@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergePipelineCaches@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergePipelineCaches@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergePipelineCaches@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergePipelineCaches@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergePipelineCaches@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergePipelineCaches@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergePipelineCaches@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergePipelineCaches@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_MergePipelineCaches=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_MergePipelineCaches=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineBinariesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineBinariesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineBinariesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineBinariesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineBinariesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineBinariesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineBinariesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineBinariesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineBinariesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineBinariesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineBinariesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineBinariesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineBinariesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineBinariesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreatePipelineBinariesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreatePipelineBinariesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineBinariesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineBinariesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineBinariesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineBinariesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineBinariesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineBinariesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineBinariesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineBinariesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineBinariesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineBinariesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineBinariesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineBinariesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineBinariesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineBinariesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreatePipelineBinariesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreatePipelineBinariesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineBinariesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineBinariesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineBinariesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineBinariesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineBinariesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineBinariesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineBinariesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineBinariesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineBinariesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineBinariesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineBinariesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineBinariesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineBinariesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineBinariesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreatePipelineBinariesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreatePipelineBinariesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineBinariesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineBinariesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineBinariesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineBinariesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineBinariesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineBinariesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineBinariesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineBinariesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineBinariesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineBinariesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineBinariesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineBinariesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineBinariesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineBinariesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreatePipelineBinariesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreatePipelineBinariesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineBinariesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineBinariesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineBinariesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineBinariesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineBinariesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineBinariesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineBinariesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineBinariesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineBinariesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineBinariesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineBinariesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineBinariesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineBinariesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineBinariesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreatePipelineBinariesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreatePipelineBinariesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineBinariesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineBinariesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineBinariesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineBinariesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineBinariesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineBinariesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineBinariesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineBinariesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineBinariesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineBinariesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineBinariesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineBinariesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineBinariesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineBinariesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreatePipelineBinariesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreatePipelineBinariesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineBinariesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineBinariesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineBinariesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineBinariesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineBinariesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineBinariesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineBinariesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineBinariesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineBinariesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineBinariesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineBinariesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineBinariesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineBinariesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineBinariesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreatePipelineBinariesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreatePipelineBinariesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineBinariesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineBinariesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineBinariesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineBinariesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineBinariesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineBinariesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineBinariesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineBinariesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineBinariesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineBinariesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineBinariesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineBinariesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineBinariesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineBinariesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreatePipelineBinariesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreatePipelineBinariesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineBinariesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineBinariesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineBinariesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineBinariesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineBinariesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineBinariesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineBinariesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineBinariesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineBinariesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineBinariesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineBinariesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineBinariesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineBinariesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineBinariesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreatePipelineBinariesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreatePipelineBinariesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineBinariesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineBinariesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineBinariesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineBinariesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineBinariesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineBinariesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineBinariesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineBinariesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineBinariesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineBinariesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineBinariesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineBinariesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineBinariesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineBinariesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreatePipelineBinariesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreatePipelineBinariesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineBinariesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineBinariesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineBinariesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineBinariesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineBinariesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineBinariesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineBinariesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineBinariesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineBinariesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineBinariesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineBinariesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineBinariesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineBinariesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineBinariesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreatePipelineBinariesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreatePipelineBinariesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineBinariesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineBinariesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineBinariesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineBinariesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineBinariesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineBinariesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineBinariesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineBinariesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineBinariesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineBinariesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineBinariesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineBinariesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineBinariesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineBinariesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreatePipelineBinariesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreatePipelineBinariesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineBinariesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineBinariesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineBinariesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineBinariesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineBinariesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineBinariesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineBinariesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineBinariesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineBinariesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineBinariesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineBinariesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineBinariesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineBinariesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineBinariesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreatePipelineBinariesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreatePipelineBinariesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineBinaryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineBinaryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineBinaryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineBinaryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineBinaryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineBinaryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineBinaryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineBinaryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineBinaryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineBinaryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineBinaryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineBinaryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineBinaryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineBinaryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyPipelineBinaryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyPipelineBinaryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineBinaryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineBinaryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineBinaryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineBinaryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineBinaryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineBinaryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineBinaryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineBinaryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineBinaryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineBinaryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineBinaryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineBinaryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineBinaryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineBinaryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyPipelineBinaryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyPipelineBinaryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineBinaryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineBinaryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineBinaryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineBinaryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineBinaryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineBinaryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineBinaryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineBinaryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineBinaryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineBinaryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineBinaryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineBinaryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineBinaryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineBinaryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyPipelineBinaryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyPipelineBinaryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineBinaryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineBinaryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineBinaryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineBinaryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineBinaryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineBinaryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineBinaryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineBinaryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineBinaryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineBinaryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineBinaryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineBinaryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineBinaryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineBinaryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyPipelineBinaryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyPipelineBinaryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineBinaryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineBinaryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineBinaryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineBinaryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineBinaryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineBinaryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineBinaryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineBinaryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineBinaryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineBinaryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineBinaryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineBinaryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineBinaryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineBinaryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyPipelineBinaryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyPipelineBinaryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineBinaryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineBinaryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineBinaryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineBinaryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineBinaryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineBinaryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineBinaryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineBinaryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineBinaryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineBinaryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineBinaryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineBinaryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineBinaryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineBinaryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyPipelineBinaryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyPipelineBinaryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineBinaryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineBinaryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineBinaryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineBinaryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineBinaryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineBinaryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineBinaryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineBinaryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineBinaryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineBinaryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineBinaryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineBinaryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineBinaryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineBinaryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyPipelineBinaryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyPipelineBinaryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineBinaryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineBinaryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineBinaryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineBinaryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineBinaryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineBinaryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineBinaryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineBinaryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineBinaryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineBinaryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineBinaryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineBinaryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineBinaryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineBinaryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyPipelineBinaryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyPipelineBinaryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineBinaryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineBinaryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineBinaryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineBinaryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineBinaryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineBinaryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineBinaryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineBinaryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineBinaryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineBinaryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineBinaryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineBinaryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineBinaryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineBinaryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyPipelineBinaryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyPipelineBinaryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineBinaryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineBinaryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineBinaryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineBinaryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineBinaryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineBinaryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineBinaryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineBinaryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineBinaryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineBinaryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineBinaryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineBinaryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineBinaryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineBinaryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyPipelineBinaryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyPipelineBinaryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineBinaryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineBinaryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineBinaryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineBinaryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineBinaryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineBinaryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineBinaryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineBinaryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineBinaryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineBinaryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineBinaryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineBinaryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineBinaryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineBinaryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyPipelineBinaryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyPipelineBinaryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineBinaryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineBinaryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineBinaryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineBinaryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineBinaryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineBinaryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineBinaryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineBinaryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineBinaryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineBinaryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineBinaryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineBinaryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineBinaryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineBinaryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyPipelineBinaryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyPipelineBinaryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineBinaryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineBinaryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineBinaryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineBinaryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineBinaryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineBinaryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineBinaryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineBinaryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineBinaryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineBinaryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineBinaryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineBinaryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineBinaryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineBinaryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyPipelineBinaryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyPipelineBinaryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPipelineKeyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineKeyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineKeyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineKeyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineKeyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineKeyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineKeyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineKeyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineKeyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineKeyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineKeyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineKeyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineKeyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineKeyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPipelineKeyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPipelineKeyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineKeyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineKeyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineKeyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineKeyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineKeyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineKeyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineKeyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineKeyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineKeyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineKeyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineKeyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineKeyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineKeyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineKeyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetPipelineKeyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetPipelineKeyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineKeyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineKeyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineKeyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineKeyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineKeyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineKeyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineKeyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineKeyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineKeyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineKeyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineKeyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineKeyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineKeyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineKeyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetPipelineKeyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetPipelineKeyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineKeyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineKeyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineKeyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineKeyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineKeyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineKeyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineKeyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineKeyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineKeyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineKeyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineKeyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineKeyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineKeyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineKeyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetPipelineKeyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetPipelineKeyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineKeyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineKeyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineKeyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineKeyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineKeyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineKeyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineKeyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineKeyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineKeyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineKeyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineKeyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineKeyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineKeyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineKeyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetPipelineKeyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetPipelineKeyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineKeyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineKeyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineKeyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineKeyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineKeyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineKeyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineKeyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineKeyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineKeyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineKeyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineKeyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineKeyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineKeyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineKeyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetPipelineKeyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetPipelineKeyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineKeyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineKeyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineKeyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineKeyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineKeyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineKeyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineKeyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineKeyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineKeyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineKeyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineKeyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineKeyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineKeyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineKeyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetPipelineKeyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetPipelineKeyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineKeyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineKeyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineKeyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineKeyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineKeyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineKeyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineKeyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineKeyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineKeyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineKeyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineKeyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineKeyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineKeyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineKeyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetPipelineKeyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetPipelineKeyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineKeyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineKeyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineKeyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineKeyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineKeyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineKeyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineKeyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineKeyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineKeyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineKeyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineKeyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineKeyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineKeyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineKeyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetPipelineKeyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetPipelineKeyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineKeyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineKeyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineKeyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineKeyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineKeyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineKeyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineKeyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineKeyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineKeyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineKeyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineKeyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineKeyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineKeyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineKeyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetPipelineKeyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetPipelineKeyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineKeyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineKeyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineKeyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineKeyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineKeyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineKeyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineKeyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineKeyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineKeyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineKeyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineKeyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineKeyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineKeyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineKeyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetPipelineKeyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetPipelineKeyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineKeyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineKeyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineKeyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineKeyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineKeyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineKeyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineKeyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineKeyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineKeyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineKeyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineKeyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineKeyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineKeyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineKeyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetPipelineKeyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetPipelineKeyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineKeyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineKeyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineKeyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineKeyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineKeyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineKeyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineKeyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineKeyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineKeyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineKeyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineKeyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineKeyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineKeyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineKeyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetPipelineKeyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetPipelineKeyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPipelineBinaryDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineBinaryDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineBinaryDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineBinaryDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineBinaryDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineBinaryDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineBinaryDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineBinaryDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineBinaryDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineBinaryDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineBinaryDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineBinaryDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineBinaryDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineBinaryDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPipelineBinaryDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPipelineBinaryDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineBinaryDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineBinaryDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineBinaryDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineBinaryDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineBinaryDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineBinaryDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineBinaryDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineBinaryDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineBinaryDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineBinaryDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineBinaryDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineBinaryDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineBinaryDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineBinaryDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetPipelineBinaryDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetPipelineBinaryDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineBinaryDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineBinaryDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineBinaryDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineBinaryDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineBinaryDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineBinaryDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineBinaryDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineBinaryDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineBinaryDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineBinaryDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineBinaryDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineBinaryDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineBinaryDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineBinaryDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetPipelineBinaryDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetPipelineBinaryDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineBinaryDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineBinaryDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineBinaryDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineBinaryDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineBinaryDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineBinaryDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineBinaryDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineBinaryDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineBinaryDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineBinaryDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineBinaryDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineBinaryDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineBinaryDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineBinaryDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetPipelineBinaryDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetPipelineBinaryDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineBinaryDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineBinaryDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineBinaryDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineBinaryDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineBinaryDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineBinaryDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineBinaryDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineBinaryDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineBinaryDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineBinaryDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineBinaryDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineBinaryDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineBinaryDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineBinaryDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetPipelineBinaryDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetPipelineBinaryDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineBinaryDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineBinaryDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineBinaryDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineBinaryDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineBinaryDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineBinaryDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineBinaryDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineBinaryDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineBinaryDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineBinaryDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineBinaryDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineBinaryDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineBinaryDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineBinaryDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetPipelineBinaryDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetPipelineBinaryDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineBinaryDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineBinaryDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineBinaryDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineBinaryDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineBinaryDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineBinaryDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineBinaryDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineBinaryDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineBinaryDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineBinaryDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineBinaryDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineBinaryDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineBinaryDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineBinaryDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetPipelineBinaryDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetPipelineBinaryDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineBinaryDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineBinaryDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineBinaryDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineBinaryDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineBinaryDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineBinaryDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineBinaryDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineBinaryDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineBinaryDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineBinaryDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineBinaryDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineBinaryDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineBinaryDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineBinaryDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetPipelineBinaryDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetPipelineBinaryDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineBinaryDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineBinaryDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineBinaryDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineBinaryDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineBinaryDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineBinaryDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineBinaryDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineBinaryDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineBinaryDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineBinaryDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineBinaryDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineBinaryDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineBinaryDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineBinaryDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetPipelineBinaryDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetPipelineBinaryDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineBinaryDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineBinaryDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineBinaryDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineBinaryDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineBinaryDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineBinaryDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineBinaryDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineBinaryDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineBinaryDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineBinaryDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineBinaryDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineBinaryDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineBinaryDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineBinaryDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetPipelineBinaryDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetPipelineBinaryDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineBinaryDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineBinaryDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineBinaryDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineBinaryDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineBinaryDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineBinaryDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineBinaryDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineBinaryDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineBinaryDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineBinaryDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineBinaryDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineBinaryDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineBinaryDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineBinaryDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetPipelineBinaryDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetPipelineBinaryDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineBinaryDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineBinaryDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineBinaryDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineBinaryDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineBinaryDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineBinaryDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineBinaryDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineBinaryDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineBinaryDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineBinaryDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineBinaryDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineBinaryDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineBinaryDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineBinaryDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetPipelineBinaryDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetPipelineBinaryDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineBinaryDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineBinaryDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineBinaryDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineBinaryDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineBinaryDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineBinaryDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineBinaryDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineBinaryDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineBinaryDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineBinaryDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineBinaryDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineBinaryDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineBinaryDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineBinaryDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetPipelineBinaryDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetPipelineBinaryDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ReleaseCapturedPipelineDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseCapturedPipelineDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseCapturedPipelineDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseCapturedPipelineDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseCapturedPipelineDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseCapturedPipelineDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseCapturedPipelineDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseCapturedPipelineDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseCapturedPipelineDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseCapturedPipelineDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseCapturedPipelineDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseCapturedPipelineDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseCapturedPipelineDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseCapturedPipelineDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ReleaseCapturedPipelineDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ReleaseCapturedPipelineDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseCapturedPipelineDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseCapturedPipelineDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseCapturedPipelineDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseCapturedPipelineDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseCapturedPipelineDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseCapturedPipelineDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseCapturedPipelineDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseCapturedPipelineDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseCapturedPipelineDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseCapturedPipelineDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseCapturedPipelineDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseCapturedPipelineDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseCapturedPipelineDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseCapturedPipelineDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ReleaseCapturedPipelineDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ReleaseCapturedPipelineDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseCapturedPipelineDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseCapturedPipelineDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseCapturedPipelineDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseCapturedPipelineDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseCapturedPipelineDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseCapturedPipelineDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseCapturedPipelineDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseCapturedPipelineDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseCapturedPipelineDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseCapturedPipelineDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseCapturedPipelineDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseCapturedPipelineDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseCapturedPipelineDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseCapturedPipelineDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ReleaseCapturedPipelineDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ReleaseCapturedPipelineDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseCapturedPipelineDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseCapturedPipelineDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseCapturedPipelineDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseCapturedPipelineDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseCapturedPipelineDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseCapturedPipelineDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseCapturedPipelineDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseCapturedPipelineDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseCapturedPipelineDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseCapturedPipelineDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseCapturedPipelineDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseCapturedPipelineDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseCapturedPipelineDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseCapturedPipelineDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ReleaseCapturedPipelineDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ReleaseCapturedPipelineDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseCapturedPipelineDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseCapturedPipelineDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseCapturedPipelineDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseCapturedPipelineDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseCapturedPipelineDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseCapturedPipelineDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseCapturedPipelineDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseCapturedPipelineDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseCapturedPipelineDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseCapturedPipelineDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseCapturedPipelineDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseCapturedPipelineDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseCapturedPipelineDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseCapturedPipelineDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ReleaseCapturedPipelineDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ReleaseCapturedPipelineDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseCapturedPipelineDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseCapturedPipelineDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseCapturedPipelineDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseCapturedPipelineDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseCapturedPipelineDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseCapturedPipelineDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseCapturedPipelineDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseCapturedPipelineDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseCapturedPipelineDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseCapturedPipelineDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseCapturedPipelineDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseCapturedPipelineDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseCapturedPipelineDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseCapturedPipelineDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ReleaseCapturedPipelineDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ReleaseCapturedPipelineDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseCapturedPipelineDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseCapturedPipelineDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseCapturedPipelineDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseCapturedPipelineDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseCapturedPipelineDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseCapturedPipelineDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseCapturedPipelineDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseCapturedPipelineDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseCapturedPipelineDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseCapturedPipelineDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseCapturedPipelineDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseCapturedPipelineDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseCapturedPipelineDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseCapturedPipelineDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ReleaseCapturedPipelineDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ReleaseCapturedPipelineDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseCapturedPipelineDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseCapturedPipelineDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseCapturedPipelineDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseCapturedPipelineDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseCapturedPipelineDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseCapturedPipelineDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseCapturedPipelineDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseCapturedPipelineDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseCapturedPipelineDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseCapturedPipelineDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseCapturedPipelineDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseCapturedPipelineDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseCapturedPipelineDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseCapturedPipelineDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ReleaseCapturedPipelineDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ReleaseCapturedPipelineDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseCapturedPipelineDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseCapturedPipelineDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseCapturedPipelineDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseCapturedPipelineDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseCapturedPipelineDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseCapturedPipelineDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseCapturedPipelineDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseCapturedPipelineDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseCapturedPipelineDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseCapturedPipelineDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseCapturedPipelineDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseCapturedPipelineDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseCapturedPipelineDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseCapturedPipelineDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ReleaseCapturedPipelineDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ReleaseCapturedPipelineDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseCapturedPipelineDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseCapturedPipelineDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseCapturedPipelineDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseCapturedPipelineDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseCapturedPipelineDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseCapturedPipelineDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseCapturedPipelineDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseCapturedPipelineDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseCapturedPipelineDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseCapturedPipelineDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseCapturedPipelineDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseCapturedPipelineDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseCapturedPipelineDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseCapturedPipelineDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ReleaseCapturedPipelineDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ReleaseCapturedPipelineDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseCapturedPipelineDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseCapturedPipelineDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseCapturedPipelineDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseCapturedPipelineDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseCapturedPipelineDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseCapturedPipelineDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseCapturedPipelineDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseCapturedPipelineDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseCapturedPipelineDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseCapturedPipelineDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseCapturedPipelineDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseCapturedPipelineDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseCapturedPipelineDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseCapturedPipelineDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ReleaseCapturedPipelineDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ReleaseCapturedPipelineDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseCapturedPipelineDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseCapturedPipelineDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseCapturedPipelineDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseCapturedPipelineDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseCapturedPipelineDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseCapturedPipelineDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseCapturedPipelineDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseCapturedPipelineDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseCapturedPipelineDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseCapturedPipelineDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseCapturedPipelineDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseCapturedPipelineDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseCapturedPipelineDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseCapturedPipelineDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ReleaseCapturedPipelineDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ReleaseCapturedPipelineDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseCapturedPipelineDataKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseCapturedPipelineDataKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseCapturedPipelineDataKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseCapturedPipelineDataKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseCapturedPipelineDataKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseCapturedPipelineDataKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseCapturedPipelineDataKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseCapturedPipelineDataKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseCapturedPipelineDataKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseCapturedPipelineDataKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseCapturedPipelineDataKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseCapturedPipelineDataKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseCapturedPipelineDataKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseCapturedPipelineDataKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ReleaseCapturedPipelineDataKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ReleaseCapturedPipelineDataKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateGraphicsPipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateGraphicsPipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateGraphicsPipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateGraphicsPipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateGraphicsPipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateGraphicsPipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateGraphicsPipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateGraphicsPipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateGraphicsPipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateGraphicsPipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateGraphicsPipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateGraphicsPipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateGraphicsPipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateGraphicsPipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateGraphicsPipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateGraphicsPipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateGraphicsPipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateGraphicsPipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateGraphicsPipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateGraphicsPipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateGraphicsPipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateGraphicsPipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateGraphicsPipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateGraphicsPipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateGraphicsPipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateGraphicsPipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateGraphicsPipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateGraphicsPipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateGraphicsPipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateGraphicsPipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateGraphicsPipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateGraphicsPipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateGraphicsPipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateGraphicsPipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateGraphicsPipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateGraphicsPipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateGraphicsPipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateGraphicsPipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateGraphicsPipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateGraphicsPipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateGraphicsPipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateGraphicsPipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateGraphicsPipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateGraphicsPipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateGraphicsPipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateGraphicsPipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateGraphicsPipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateGraphicsPipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateGraphicsPipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateGraphicsPipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateGraphicsPipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateGraphicsPipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateGraphicsPipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateGraphicsPipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateGraphicsPipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateGraphicsPipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateGraphicsPipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateGraphicsPipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateGraphicsPipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateGraphicsPipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateGraphicsPipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateGraphicsPipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateGraphicsPipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateGraphicsPipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateGraphicsPipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateGraphicsPipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateGraphicsPipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateGraphicsPipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateGraphicsPipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateGraphicsPipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateGraphicsPipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateGraphicsPipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateGraphicsPipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateGraphicsPipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateGraphicsPipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateGraphicsPipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateGraphicsPipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateGraphicsPipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateGraphicsPipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateGraphicsPipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateGraphicsPipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateGraphicsPipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateGraphicsPipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateGraphicsPipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateGraphicsPipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateGraphicsPipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateGraphicsPipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateGraphicsPipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateGraphicsPipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateGraphicsPipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateGraphicsPipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateGraphicsPipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateGraphicsPipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateGraphicsPipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateGraphicsPipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateGraphicsPipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateGraphicsPipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateGraphicsPipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateGraphicsPipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateGraphicsPipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateGraphicsPipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateGraphicsPipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateGraphicsPipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateGraphicsPipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateGraphicsPipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateGraphicsPipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateGraphicsPipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateGraphicsPipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateGraphicsPipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateGraphicsPipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateGraphicsPipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateGraphicsPipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateGraphicsPipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateGraphicsPipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateGraphicsPipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateGraphicsPipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateGraphicsPipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateGraphicsPipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateGraphicsPipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateGraphicsPipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateGraphicsPipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateGraphicsPipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateGraphicsPipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateGraphicsPipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateGraphicsPipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateGraphicsPipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateGraphicsPipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateGraphicsPipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateGraphicsPipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateGraphicsPipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateGraphicsPipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateGraphicsPipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateGraphicsPipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateGraphicsPipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateGraphicsPipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateGraphicsPipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateGraphicsPipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateGraphicsPipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateGraphicsPipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateGraphicsPipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateGraphicsPipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateGraphicsPipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateGraphicsPipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateGraphicsPipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateGraphicsPipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateGraphicsPipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateGraphicsPipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateGraphicsPipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateGraphicsPipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateGraphicsPipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateGraphicsPipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateGraphicsPipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateGraphicsPipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateGraphicsPipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateGraphicsPipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateGraphicsPipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateGraphicsPipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateGraphicsPipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateGraphicsPipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateGraphicsPipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateGraphicsPipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateGraphicsPipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateGraphicsPipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateGraphicsPipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateGraphicsPipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateGraphicsPipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateGraphicsPipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateGraphicsPipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateGraphicsPipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateGraphicsPipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateGraphicsPipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateGraphicsPipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateGraphicsPipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateGraphicsPipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateGraphicsPipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateGraphicsPipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateGraphicsPipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateGraphicsPipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateGraphicsPipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateGraphicsPipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateGraphicsPipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateGraphicsPipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateGraphicsPipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateGraphicsPipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateGraphicsPipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateGraphicsPipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateGraphicsPipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateGraphicsPipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateGraphicsPipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateGraphicsPipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateGraphicsPipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateGraphicsPipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateGraphicsPipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateGraphicsPipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateGraphicsPipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateGraphicsPipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateGraphicsPipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateGraphicsPipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateGraphicsPipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateGraphicsPipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateGraphicsPipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateGraphicsPipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateGraphicsPipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateGraphicsPipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateGraphicsPipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateGraphicsPipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateGraphicsPipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateGraphicsPipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateComputePipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateComputePipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateComputePipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateComputePipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateComputePipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateComputePipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateComputePipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateComputePipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateComputePipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateComputePipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateComputePipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateComputePipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateComputePipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateComputePipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateComputePipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateComputePipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateComputePipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateComputePipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateComputePipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateComputePipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateComputePipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateComputePipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateComputePipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateComputePipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateComputePipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateComputePipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateComputePipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateComputePipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateComputePipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateComputePipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateComputePipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateComputePipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateComputePipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateComputePipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateComputePipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateComputePipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateComputePipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateComputePipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateComputePipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateComputePipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateComputePipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateComputePipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateComputePipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateComputePipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateComputePipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateComputePipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateComputePipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateComputePipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateComputePipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateComputePipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateComputePipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateComputePipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateComputePipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateComputePipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateComputePipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateComputePipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateComputePipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateComputePipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateComputePipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateComputePipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateComputePipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateComputePipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateComputePipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateComputePipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateComputePipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateComputePipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateComputePipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateComputePipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateComputePipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateComputePipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateComputePipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateComputePipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateComputePipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateComputePipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateComputePipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateComputePipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateComputePipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateComputePipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateComputePipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateComputePipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateComputePipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateComputePipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateComputePipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateComputePipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateComputePipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateComputePipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateComputePipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateComputePipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateComputePipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateComputePipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateComputePipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateComputePipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateComputePipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateComputePipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateComputePipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateComputePipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateComputePipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateComputePipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateComputePipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateComputePipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateComputePipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateComputePipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateComputePipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateComputePipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateComputePipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateComputePipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateComputePipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateComputePipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateComputePipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateComputePipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateComputePipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateComputePipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateComputePipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateComputePipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateComputePipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateComputePipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateComputePipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateComputePipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateComputePipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateComputePipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateComputePipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateComputePipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateComputePipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateComputePipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateComputePipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateComputePipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateComputePipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateComputePipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateComputePipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateComputePipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateComputePipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateComputePipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateComputePipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateComputePipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateComputePipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateComputePipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateComputePipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateComputePipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateComputePipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateComputePipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateComputePipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateComputePipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateComputePipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateComputePipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateComputePipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateComputePipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateComputePipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateComputePipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateComputePipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateComputePipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateComputePipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateComputePipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateComputePipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateComputePipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateComputePipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateComputePipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateComputePipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateComputePipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateComputePipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateComputePipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateComputePipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateComputePipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateComputePipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateComputePipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateComputePipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateComputePipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateComputePipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateComputePipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateComputePipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateComputePipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateComputePipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateComputePipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateComputePipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateComputePipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateComputePipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateComputePipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateComputePipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateComputePipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateComputePipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateComputePipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateComputePipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateComputePipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateComputePipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateComputePipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateComputePipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateComputePipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateComputePipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateComputePipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateComputePipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateComputePipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateComputePipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateComputePipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateComputePipelines@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateComputePipelines@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateComputePipelines@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateComputePipelines@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateComputePipelines@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateComputePipelines@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateComputePipelines@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateComputePipelines@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateComputePipelines@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateComputePipelines@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateComputePipelines@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateComputePipelines@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateComputePipelines@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateComputePipelines@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateComputePipelines=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateComputePipelines=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreatePipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreatePipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreatePipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreatePipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreatePipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreatePipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreatePipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreatePipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreatePipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreatePipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreatePipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreatePipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreatePipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreatePipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreatePipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreatePipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreatePipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreatePipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreatePipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreatePipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreatePipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreatePipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreatePipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreatePipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreatePipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreatePipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyPipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyPipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyPipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyPipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyPipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyPipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyPipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyPipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyPipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyPipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyPipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyPipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyPipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyPipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyPipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyPipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyPipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyPipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyPipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyPipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyPipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyPipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyPipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyPipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPipelineLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyPipelineLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyPipelineLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateSampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateSampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateSampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateSampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateSampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateSampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateSampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateSampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateSampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateSampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateSampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateSampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateSampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateSampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateSampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateSampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateSampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateSampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateSampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateSampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateSampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateSampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateSampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateSampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateSampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateSampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateSampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateSampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateSampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateSampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateSampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateSampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateSampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateSampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateSampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroySampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroySampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroySampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroySampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroySampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroySampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroySampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroySampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroySampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroySampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroySampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroySampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroySampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroySampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroySampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroySampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroySampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroySampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroySampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroySampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroySampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroySampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroySampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroySampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroySampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroySampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroySampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroySampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroySampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroySampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroySampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroySampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroySampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySampler@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySampler@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySampler@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySampler@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySampler@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySampler@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySampler@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySampler@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySampler@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySampler@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySampler@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySampler@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySampler@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySampler@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroySampler=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroySampler=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorSetLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorSetLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorSetLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorSetLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorSetLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorSetLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorSetLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorSetLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorSetLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorSetLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorSetLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorSetLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorSetLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorSetLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyDescriptorSetLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyDescriptorSetLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ResetDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ResetDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ResetDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ResetDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ResetDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ResetDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ResetDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ResetDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ResetDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ResetDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ResetDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ResetDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ResetDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ResetDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ResetDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ResetDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ResetDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ResetDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ResetDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ResetDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ResetDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ResetDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ResetDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ResetDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ResetDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ResetDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ResetDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ResetDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ResetDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ResetDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ResetDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ResetDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ResetDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetDescriptorPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetDescriptorPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetDescriptorPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetDescriptorPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetDescriptorPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetDescriptorPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetDescriptorPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetDescriptorPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetDescriptorPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetDescriptorPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetDescriptorPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetDescriptorPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetDescriptorPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetDescriptorPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ResetDescriptorPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ResetDescriptorPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_AllocateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_AllocateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_AllocateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_AllocateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_AllocateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_AllocateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_AllocateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_AllocateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_AllocateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_AllocateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_AllocateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_AllocateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_AllocateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_AllocateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_AllocateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_AllocateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_AllocateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_AllocateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_AllocateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_AllocateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_AllocateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_AllocateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_AllocateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_AllocateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_AllocateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_AllocateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_AllocateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_AllocateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_AllocateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_AllocateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_AllocateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_AllocateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_AllocateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_AllocateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_AllocateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_FreeDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_FreeDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_FreeDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_FreeDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_FreeDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_FreeDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_FreeDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_FreeDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_FreeDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_FreeDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_FreeDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_FreeDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_FreeDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_FreeDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_FreeDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_FreeDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_FreeDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_FreeDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_FreeDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_FreeDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_FreeDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_FreeDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_FreeDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_FreeDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_FreeDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_FreeDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_FreeDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_FreeDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_FreeDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_FreeDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_FreeDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_FreeDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_FreeDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_FreeDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_FreeDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_UpdateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_UpdateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_UpdateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_UpdateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_UpdateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_UpdateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_UpdateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_UpdateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_UpdateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_UpdateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_UpdateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_UpdateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_UpdateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_UpdateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_UpdateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_UpdateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_UpdateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_UpdateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_UpdateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_UpdateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_UpdateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_UpdateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_UpdateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_UpdateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_UpdateDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_UpdateDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFramebuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFramebuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFramebuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFramebuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFramebuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFramebuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFramebuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFramebuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFramebuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFramebuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFramebuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFramebuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFramebuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyFramebuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyFramebuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyFramebuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetRenderAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetRenderAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetRenderAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetRenderAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetRenderAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetRenderAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetRenderAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetRenderAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetRenderAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetRenderAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetRenderAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetRenderAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetRenderAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetRenderAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetRenderAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetRenderAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetRenderAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetRenderAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetRenderAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetRenderAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetRenderAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetRenderAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetRenderAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetRenderAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetRenderAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetRenderAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetRenderAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetRenderingAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetRenderingAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetRenderingAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetRenderingAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetRenderingAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetRenderingAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetRenderingAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetRenderingAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetRenderingAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetRenderingAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetRenderingAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetRenderingAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetRenderingAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetRenderingAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetRenderingAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetRenderingAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetRenderingAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetRenderingAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetRenderingAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetRenderingAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetRenderingAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetRenderingAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetRenderingAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetRenderingAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularity@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularity@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularity@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularity@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularity@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularity@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularity@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularity@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularity@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularity@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularity@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularity@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularity@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularity@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetRenderingAreaGranularity=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetRenderingAreaGranularity=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRenderingAreaGranularityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetRenderingAreaGranularityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetRenderingAreaGranularityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRenderingAreaGranularityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetRenderingAreaGranularityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetRenderingAreaGranularityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRenderingAreaGranularityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetRenderingAreaGranularityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetRenderingAreaGranularityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRenderingAreaGranularityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetRenderingAreaGranularityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetRenderingAreaGranularityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRenderingAreaGranularityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetRenderingAreaGranularityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetRenderingAreaGranularityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRenderingAreaGranularityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetRenderingAreaGranularityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetRenderingAreaGranularityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRenderingAreaGranularityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetRenderingAreaGranularityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetRenderingAreaGranularityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRenderingAreaGranularityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetRenderingAreaGranularityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetRenderingAreaGranularityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRenderingAreaGranularityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetRenderingAreaGranularityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetRenderingAreaGranularityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRenderingAreaGranularityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetRenderingAreaGranularityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetRenderingAreaGranularityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRenderingAreaGranularityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetRenderingAreaGranularityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetRenderingAreaGranularityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRenderingAreaGranularityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetRenderingAreaGranularityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetRenderingAreaGranularityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRenderingAreaGranularityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetRenderingAreaGranularityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetRenderingAreaGranularityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ResetCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ResetCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ResetCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ResetCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ResetCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ResetCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ResetCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ResetCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ResetCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ResetCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ResetCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ResetCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ResetCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ResetCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ResetCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ResetCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ResetCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ResetCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ResetCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ResetCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ResetCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ResetCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ResetCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ResetCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ResetCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ResetCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ResetCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_AllocateCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AllocateCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_AllocateCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_AllocateCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_AllocateCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AllocateCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_AllocateCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_AllocateCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_AllocateCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AllocateCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_AllocateCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_AllocateCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_AllocateCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AllocateCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_AllocateCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_AllocateCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_AllocateCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AllocateCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_AllocateCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_AllocateCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_AllocateCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AllocateCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_AllocateCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_AllocateCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_AllocateCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AllocateCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_AllocateCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_AllocateCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_AllocateCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AllocateCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_AllocateCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_AllocateCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AllocateCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_AllocateCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_AllocateCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AllocateCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_AllocateCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_AllocateCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AllocateCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_AllocateCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_AllocateCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_AllocateCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AllocateCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_AllocateCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_AllocateCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AllocateCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_AllocateCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_AllocateCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_FreeCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_FreeCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_FreeCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_FreeCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_FreeCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_FreeCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_FreeCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_FreeCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_FreeCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_FreeCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_FreeCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_FreeCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_FreeCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_FreeCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_FreeCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_FreeCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_FreeCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_FreeCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_FreeCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_FreeCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_FreeCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_FreeCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_FreeCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_FreeCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_FreeCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_FreeCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_FreeCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_FreeCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_FreeCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_FreeCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_FreeCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_FreeCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_FreeCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_FreeCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_FreeCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_FreeCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_FreeCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_FreeCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_FreeCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_FreeCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_FreeCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_FreeCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_FreeCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_FreeCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_FreeCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeCommandBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeCommandBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeCommandBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeCommandBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeCommandBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeCommandBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeCommandBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeCommandBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeCommandBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeCommandBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeCommandBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeCommandBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeCommandBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_FreeCommandBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_FreeCommandBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_FreeCommandBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_BeginCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BeginCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BeginCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BeginCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BeginCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BeginCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BeginCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BeginCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BeginCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BeginCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BeginCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BeginCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BeginCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BeginCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_BeginCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_BeginCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_BeginCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BeginCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BeginCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BeginCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BeginCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BeginCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BeginCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BeginCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BeginCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BeginCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BeginCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BeginCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BeginCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BeginCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_BeginCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_BeginCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_BeginCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BeginCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BeginCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BeginCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BeginCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BeginCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BeginCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BeginCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BeginCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BeginCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BeginCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BeginCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BeginCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BeginCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_BeginCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_BeginCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_BeginCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BeginCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BeginCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BeginCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BeginCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BeginCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BeginCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BeginCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BeginCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BeginCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BeginCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BeginCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BeginCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BeginCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_BeginCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_BeginCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_BeginCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BeginCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BeginCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BeginCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BeginCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BeginCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BeginCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BeginCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BeginCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BeginCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BeginCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BeginCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BeginCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BeginCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_BeginCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_BeginCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_BeginCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BeginCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BeginCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BeginCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BeginCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BeginCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BeginCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BeginCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BeginCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BeginCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BeginCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BeginCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BeginCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BeginCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_BeginCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_BeginCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_BeginCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BeginCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BeginCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BeginCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BeginCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BeginCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BeginCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BeginCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BeginCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BeginCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BeginCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BeginCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BeginCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BeginCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_BeginCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_BeginCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_BeginCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BeginCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BeginCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BeginCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BeginCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BeginCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BeginCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BeginCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BeginCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BeginCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BeginCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BeginCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BeginCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BeginCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_BeginCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_BeginCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_BeginCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BeginCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BeginCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BeginCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BeginCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BeginCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BeginCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BeginCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BeginCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BeginCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BeginCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BeginCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BeginCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BeginCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_BeginCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_BeginCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_BeginCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BeginCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BeginCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BeginCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BeginCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BeginCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BeginCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BeginCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BeginCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BeginCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BeginCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BeginCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BeginCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BeginCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_BeginCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_BeginCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_BeginCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BeginCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BeginCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BeginCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BeginCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BeginCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BeginCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BeginCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BeginCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BeginCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BeginCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BeginCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BeginCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BeginCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_BeginCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_BeginCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_BeginCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BeginCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BeginCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BeginCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BeginCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BeginCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BeginCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BeginCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BeginCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BeginCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BeginCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BeginCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BeginCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BeginCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_BeginCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_BeginCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_BeginCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BeginCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BeginCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BeginCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BeginCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BeginCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BeginCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BeginCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BeginCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BeginCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BeginCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BeginCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BeginCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BeginCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_BeginCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_BeginCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_EndCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EndCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EndCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EndCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EndCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EndCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EndCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EndCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EndCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EndCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EndCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EndCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EndCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_EndCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_EndCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_EndCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_EndCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_EndCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_EndCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_EndCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_EndCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_EndCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_EndCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_EndCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_EndCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_EndCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_EndCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_EndCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_EndCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_EndCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_EndCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_EndCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_EndCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_EndCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_EndCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_EndCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_EndCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_EndCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_EndCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_EndCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_EndCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_EndCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_EndCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_EndCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_EndCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_EndCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_EndCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_EndCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_EndCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_EndCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_EndCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_EndCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_EndCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_EndCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_EndCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_EndCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_EndCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_EndCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_EndCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_EndCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_EndCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_EndCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_EndCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_EndCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_EndCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_EndCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_EndCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_EndCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_EndCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_EndCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_EndCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_EndCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_EndCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_EndCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_EndCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_EndCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_EndCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_EndCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_EndCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_EndCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_EndCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_EndCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_EndCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_EndCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_EndCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_EndCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_EndCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_EndCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_EndCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_EndCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_EndCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_EndCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_EndCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_EndCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_EndCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_EndCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_EndCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_EndCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_EndCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_EndCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_EndCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_EndCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_EndCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_EndCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_EndCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_EndCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_EndCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_EndCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_EndCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_EndCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_EndCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_EndCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_EndCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_EndCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_EndCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_EndCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_EndCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_EndCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_EndCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_EndCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_EndCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_EndCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_EndCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_EndCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_EndCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_EndCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_EndCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_EndCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_EndCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_EndCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_EndCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_EndCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_EndCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_EndCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_EndCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_EndCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_EndCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_EndCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_EndCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_EndCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_EndCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_EndCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_EndCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_EndCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_EndCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_EndCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_EndCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_EndCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_EndCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_EndCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_EndCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_EndCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_EndCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_EndCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_EndCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_EndCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_EndCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_EndCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_EndCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_EndCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_EndCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_EndCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_EndCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_EndCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_EndCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_EndCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_EndCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_EndCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_EndCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_EndCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_EndCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_EndCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_EndCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_EndCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_EndCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_EndCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_EndCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_EndCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_EndCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_EndCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_EndCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_EndCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_EndCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_EndCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_EndCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_EndCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_EndCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_EndCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_EndCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_EndCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_EndCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_EndCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_EndCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_EndCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_EndCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_EndCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_EndCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_EndCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_EndCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_EndCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_EndCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_EndCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_EndCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_EndCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_EndCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_EndCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_EndCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_EndCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ResetCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ResetCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ResetCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ResetCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ResetCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ResetCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ResetCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ResetCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ResetCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ResetCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ResetCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ResetCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ResetCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ResetCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ResetCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ResetCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ResetCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ResetCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ResetCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ResetCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ResetCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ResetCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ResetCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ResetCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ResetCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ResetCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ResetCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ResetCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ResetCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ResetCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ResetCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ResetCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ResetCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ResetCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ResetCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ResetCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ResetCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ResetCommandBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ResetCommandBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ResetCommandBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipeline@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipeline@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipeline@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipeline@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipeline@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipeline@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipeline@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipeline@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipeline@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipeline@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipeline@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipeline@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipeline@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipeline@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindPipeline=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindPipeline=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartIndexEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartIndexEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartIndexEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartIndexEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartIndexEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartIndexEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartIndexEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartIndexEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartIndexEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartIndexEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartIndexEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartIndexEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartIndexEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartIndexEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetPrimitiveRestartIndexEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetPrimitiveRestartIndexEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartIndexEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartIndexEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartIndexEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartIndexEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartIndexEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartIndexEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartIndexEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartIndexEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartIndexEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartIndexEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartIndexEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartIndexEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartIndexEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartIndexEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetPrimitiveRestartIndexEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetPrimitiveRestartIndexEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartIndexEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartIndexEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartIndexEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartIndexEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartIndexEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartIndexEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartIndexEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartIndexEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartIndexEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartIndexEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartIndexEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartIndexEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartIndexEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartIndexEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetPrimitiveRestartIndexEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetPrimitiveRestartIndexEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartIndexEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartIndexEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartIndexEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartIndexEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartIndexEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartIndexEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartIndexEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartIndexEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartIndexEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartIndexEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartIndexEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartIndexEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartIndexEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartIndexEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetPrimitiveRestartIndexEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetPrimitiveRestartIndexEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartIndexEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartIndexEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartIndexEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartIndexEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartIndexEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartIndexEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartIndexEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartIndexEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartIndexEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartIndexEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartIndexEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartIndexEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartIndexEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartIndexEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetPrimitiveRestartIndexEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetPrimitiveRestartIndexEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartIndexEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartIndexEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartIndexEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartIndexEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartIndexEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartIndexEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartIndexEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartIndexEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartIndexEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartIndexEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartIndexEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartIndexEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartIndexEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartIndexEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetPrimitiveRestartIndexEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetPrimitiveRestartIndexEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartIndexEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartIndexEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartIndexEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartIndexEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartIndexEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartIndexEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartIndexEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartIndexEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartIndexEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartIndexEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartIndexEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartIndexEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartIndexEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartIndexEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetPrimitiveRestartIndexEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetPrimitiveRestartIndexEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartIndexEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartIndexEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartIndexEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartIndexEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartIndexEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartIndexEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartIndexEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartIndexEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartIndexEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartIndexEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartIndexEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartIndexEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartIndexEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartIndexEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetPrimitiveRestartIndexEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetPrimitiveRestartIndexEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartIndexEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartIndexEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartIndexEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartIndexEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartIndexEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartIndexEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartIndexEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartIndexEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartIndexEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartIndexEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartIndexEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartIndexEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartIndexEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartIndexEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetPrimitiveRestartIndexEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetPrimitiveRestartIndexEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartIndexEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartIndexEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartIndexEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartIndexEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartIndexEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartIndexEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartIndexEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartIndexEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartIndexEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartIndexEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartIndexEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartIndexEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartIndexEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartIndexEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetPrimitiveRestartIndexEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetPrimitiveRestartIndexEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartIndexEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartIndexEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartIndexEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartIndexEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartIndexEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartIndexEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartIndexEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartIndexEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartIndexEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartIndexEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartIndexEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartIndexEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartIndexEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartIndexEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetPrimitiveRestartIndexEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetPrimitiveRestartIndexEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartIndexEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartIndexEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartIndexEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartIndexEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartIndexEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartIndexEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartIndexEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartIndexEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartIndexEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartIndexEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartIndexEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartIndexEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartIndexEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartIndexEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetPrimitiveRestartIndexEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetPrimitiveRestartIndexEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartIndexEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartIndexEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartIndexEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartIndexEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartIndexEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartIndexEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartIndexEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartIndexEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartIndexEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartIndexEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartIndexEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartIndexEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartIndexEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartIndexEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetPrimitiveRestartIndexEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetPrimitiveRestartIndexEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetAttachmentFeedbackLoopEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAttachmentFeedbackLoopEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAttachmentFeedbackLoopEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAttachmentFeedbackLoopEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAttachmentFeedbackLoopEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAttachmentFeedbackLoopEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAttachmentFeedbackLoopEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAttachmentFeedbackLoopEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAttachmentFeedbackLoopEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAttachmentFeedbackLoopEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAttachmentFeedbackLoopEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAttachmentFeedbackLoopEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAttachmentFeedbackLoopEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAttachmentFeedbackLoopEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetAttachmentFeedbackLoopEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetAttachmentFeedbackLoopEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAttachmentFeedbackLoopEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAttachmentFeedbackLoopEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAttachmentFeedbackLoopEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAttachmentFeedbackLoopEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAttachmentFeedbackLoopEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAttachmentFeedbackLoopEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAttachmentFeedbackLoopEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAttachmentFeedbackLoopEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAttachmentFeedbackLoopEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAttachmentFeedbackLoopEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAttachmentFeedbackLoopEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAttachmentFeedbackLoopEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAttachmentFeedbackLoopEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAttachmentFeedbackLoopEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetAttachmentFeedbackLoopEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetAttachmentFeedbackLoopEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAttachmentFeedbackLoopEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAttachmentFeedbackLoopEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAttachmentFeedbackLoopEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAttachmentFeedbackLoopEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAttachmentFeedbackLoopEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAttachmentFeedbackLoopEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAttachmentFeedbackLoopEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAttachmentFeedbackLoopEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAttachmentFeedbackLoopEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAttachmentFeedbackLoopEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAttachmentFeedbackLoopEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAttachmentFeedbackLoopEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAttachmentFeedbackLoopEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAttachmentFeedbackLoopEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetAttachmentFeedbackLoopEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetAttachmentFeedbackLoopEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAttachmentFeedbackLoopEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAttachmentFeedbackLoopEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAttachmentFeedbackLoopEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAttachmentFeedbackLoopEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAttachmentFeedbackLoopEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAttachmentFeedbackLoopEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAttachmentFeedbackLoopEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAttachmentFeedbackLoopEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAttachmentFeedbackLoopEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAttachmentFeedbackLoopEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAttachmentFeedbackLoopEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAttachmentFeedbackLoopEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAttachmentFeedbackLoopEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAttachmentFeedbackLoopEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetAttachmentFeedbackLoopEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetAttachmentFeedbackLoopEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAttachmentFeedbackLoopEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAttachmentFeedbackLoopEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAttachmentFeedbackLoopEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAttachmentFeedbackLoopEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAttachmentFeedbackLoopEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAttachmentFeedbackLoopEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAttachmentFeedbackLoopEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAttachmentFeedbackLoopEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAttachmentFeedbackLoopEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAttachmentFeedbackLoopEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAttachmentFeedbackLoopEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAttachmentFeedbackLoopEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAttachmentFeedbackLoopEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAttachmentFeedbackLoopEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetAttachmentFeedbackLoopEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetAttachmentFeedbackLoopEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAttachmentFeedbackLoopEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAttachmentFeedbackLoopEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAttachmentFeedbackLoopEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAttachmentFeedbackLoopEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAttachmentFeedbackLoopEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAttachmentFeedbackLoopEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAttachmentFeedbackLoopEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAttachmentFeedbackLoopEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAttachmentFeedbackLoopEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAttachmentFeedbackLoopEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAttachmentFeedbackLoopEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAttachmentFeedbackLoopEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAttachmentFeedbackLoopEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAttachmentFeedbackLoopEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetAttachmentFeedbackLoopEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetAttachmentFeedbackLoopEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAttachmentFeedbackLoopEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAttachmentFeedbackLoopEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAttachmentFeedbackLoopEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAttachmentFeedbackLoopEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAttachmentFeedbackLoopEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAttachmentFeedbackLoopEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAttachmentFeedbackLoopEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAttachmentFeedbackLoopEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAttachmentFeedbackLoopEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAttachmentFeedbackLoopEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAttachmentFeedbackLoopEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAttachmentFeedbackLoopEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAttachmentFeedbackLoopEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAttachmentFeedbackLoopEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetAttachmentFeedbackLoopEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetAttachmentFeedbackLoopEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAttachmentFeedbackLoopEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAttachmentFeedbackLoopEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAttachmentFeedbackLoopEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAttachmentFeedbackLoopEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAttachmentFeedbackLoopEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAttachmentFeedbackLoopEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAttachmentFeedbackLoopEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAttachmentFeedbackLoopEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAttachmentFeedbackLoopEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAttachmentFeedbackLoopEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAttachmentFeedbackLoopEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAttachmentFeedbackLoopEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAttachmentFeedbackLoopEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAttachmentFeedbackLoopEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetAttachmentFeedbackLoopEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetAttachmentFeedbackLoopEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAttachmentFeedbackLoopEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAttachmentFeedbackLoopEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAttachmentFeedbackLoopEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAttachmentFeedbackLoopEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAttachmentFeedbackLoopEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAttachmentFeedbackLoopEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAttachmentFeedbackLoopEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAttachmentFeedbackLoopEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAttachmentFeedbackLoopEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAttachmentFeedbackLoopEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAttachmentFeedbackLoopEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAttachmentFeedbackLoopEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAttachmentFeedbackLoopEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAttachmentFeedbackLoopEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetAttachmentFeedbackLoopEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetAttachmentFeedbackLoopEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAttachmentFeedbackLoopEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAttachmentFeedbackLoopEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAttachmentFeedbackLoopEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAttachmentFeedbackLoopEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAttachmentFeedbackLoopEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAttachmentFeedbackLoopEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAttachmentFeedbackLoopEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAttachmentFeedbackLoopEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAttachmentFeedbackLoopEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAttachmentFeedbackLoopEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAttachmentFeedbackLoopEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAttachmentFeedbackLoopEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAttachmentFeedbackLoopEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAttachmentFeedbackLoopEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetAttachmentFeedbackLoopEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetAttachmentFeedbackLoopEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAttachmentFeedbackLoopEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAttachmentFeedbackLoopEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAttachmentFeedbackLoopEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAttachmentFeedbackLoopEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAttachmentFeedbackLoopEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAttachmentFeedbackLoopEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAttachmentFeedbackLoopEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAttachmentFeedbackLoopEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAttachmentFeedbackLoopEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAttachmentFeedbackLoopEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAttachmentFeedbackLoopEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAttachmentFeedbackLoopEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAttachmentFeedbackLoopEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAttachmentFeedbackLoopEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetAttachmentFeedbackLoopEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetAttachmentFeedbackLoopEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAttachmentFeedbackLoopEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAttachmentFeedbackLoopEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAttachmentFeedbackLoopEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAttachmentFeedbackLoopEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAttachmentFeedbackLoopEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAttachmentFeedbackLoopEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAttachmentFeedbackLoopEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAttachmentFeedbackLoopEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAttachmentFeedbackLoopEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAttachmentFeedbackLoopEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAttachmentFeedbackLoopEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAttachmentFeedbackLoopEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAttachmentFeedbackLoopEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAttachmentFeedbackLoopEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetAttachmentFeedbackLoopEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetAttachmentFeedbackLoopEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAttachmentFeedbackLoopEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAttachmentFeedbackLoopEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAttachmentFeedbackLoopEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAttachmentFeedbackLoopEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAttachmentFeedbackLoopEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAttachmentFeedbackLoopEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAttachmentFeedbackLoopEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAttachmentFeedbackLoopEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAttachmentFeedbackLoopEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAttachmentFeedbackLoopEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAttachmentFeedbackLoopEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAttachmentFeedbackLoopEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAttachmentFeedbackLoopEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAttachmentFeedbackLoopEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetAttachmentFeedbackLoopEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetAttachmentFeedbackLoopEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetViewport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetViewport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetViewport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetViewport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetViewport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetViewport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetViewport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetViewport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetViewport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetViewport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetViewport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetViewport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetViewport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetViewport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetViewport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetViewport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetViewport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetViewport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetViewport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetViewport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetViewport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetViewport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetViewport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetViewport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetViewport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetViewport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissor@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissor@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissor@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissor@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissor@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissor@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissor@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissor@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissor@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissor@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissor@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissor@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissor@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissor@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetScissor=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetScissor=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissor@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissor@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissor@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissor@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissor@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissor@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissor@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissor@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissor@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissor@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissor@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissor@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissor@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissor@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetScissor=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetScissor=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissor@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissor@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissor@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissor@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissor@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissor@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissor@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissor@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissor@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissor@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissor@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissor@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissor@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissor@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetScissor=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetScissor=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissor@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissor@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissor@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissor@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissor@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissor@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissor@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissor@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissor@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissor@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissor@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissor@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissor@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissor@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetScissor=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetScissor=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissor@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissor@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissor@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissor@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissor@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissor@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissor@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissor@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissor@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissor@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissor@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissor@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissor@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissor@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetScissor=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetScissor=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissor@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissor@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissor@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissor@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissor@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissor@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissor@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissor@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissor@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissor@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissor@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissor@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissor@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissor@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetScissor=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetScissor=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissor@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissor@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissor@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissor@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissor@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissor@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissor@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissor@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissor@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissor@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissor@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissor@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissor@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissor@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetScissor=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetScissor=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissor@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissor@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissor@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissor@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissor@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissor@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissor@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissor@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissor@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissor@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissor@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissor@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissor@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissor@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetScissor=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetScissor=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissor@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissor@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissor@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissor@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissor@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissor@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissor@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissor@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissor@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissor@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissor@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissor@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissor@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissor@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetScissor=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetScissor=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissor@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissor@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissor@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissor@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissor@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissor@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissor@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissor@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissor@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissor@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissor@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissor@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissor@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissor@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetScissor=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetScissor=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissor@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissor@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissor@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissor@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissor@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissor@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissor@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissor@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissor@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissor@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissor@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissor@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissor@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissor@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetScissor=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetScissor=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissor@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissor@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissor@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissor@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissor@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissor@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissor@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissor@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissor@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissor@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissor@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissor@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissor@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissor@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetScissor=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetScissor=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissor@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissor@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissor@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissor@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissor@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissor@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissor@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissor@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissor@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissor@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissor@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissor@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissor@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissor@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetScissor=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetScissor=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineWidth@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineWidth@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineWidth@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineWidth@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineWidth@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineWidth@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineWidth@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineWidth@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineWidth@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineWidth@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineWidth@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineWidth@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineWidth@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineWidth@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetLineWidth=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetLineWidth=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineWidth@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineWidth@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineWidth@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineWidth@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineWidth@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineWidth@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineWidth@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineWidth@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineWidth@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineWidth@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineWidth@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineWidth@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineWidth@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineWidth@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetLineWidth=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetLineWidth=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineWidth@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineWidth@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineWidth@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineWidth@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineWidth@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineWidth@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineWidth@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineWidth@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineWidth@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineWidth@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineWidth@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineWidth@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineWidth@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineWidth@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetLineWidth=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetLineWidth=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineWidth@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineWidth@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineWidth@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineWidth@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineWidth@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineWidth@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineWidth@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineWidth@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineWidth@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineWidth@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineWidth@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineWidth@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineWidth@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineWidth@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetLineWidth=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetLineWidth=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineWidth@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineWidth@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineWidth@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineWidth@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineWidth@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineWidth@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineWidth@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineWidth@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineWidth@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineWidth@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineWidth@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineWidth@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineWidth@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineWidth@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetLineWidth=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetLineWidth=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineWidth@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineWidth@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineWidth@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineWidth@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineWidth@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineWidth@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineWidth@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineWidth@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineWidth@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineWidth@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineWidth@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineWidth@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineWidth@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineWidth@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetLineWidth=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetLineWidth=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineWidth@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineWidth@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineWidth@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineWidth@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineWidth@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineWidth@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineWidth@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineWidth@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineWidth@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineWidth@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineWidth@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineWidth@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineWidth@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineWidth@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetLineWidth=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetLineWidth=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineWidth@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineWidth@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineWidth@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineWidth@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineWidth@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineWidth@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineWidth@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineWidth@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineWidth@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineWidth@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineWidth@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineWidth@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineWidth@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineWidth@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetLineWidth=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetLineWidth=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineWidth@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineWidth@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineWidth@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineWidth@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineWidth@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineWidth@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineWidth@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineWidth@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineWidth@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineWidth@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineWidth@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineWidth@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineWidth@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineWidth@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetLineWidth=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetLineWidth=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineWidth@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineWidth@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineWidth@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineWidth@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineWidth@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineWidth@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineWidth@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineWidth@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineWidth@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineWidth@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineWidth@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineWidth@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineWidth@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineWidth@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetLineWidth=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetLineWidth=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineWidth@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineWidth@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineWidth@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineWidth@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineWidth@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineWidth@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineWidth@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineWidth@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineWidth@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineWidth@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineWidth@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineWidth@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineWidth@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineWidth@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetLineWidth=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetLineWidth=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineWidth@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineWidth@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineWidth@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineWidth@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineWidth@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineWidth@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineWidth@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineWidth@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineWidth@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineWidth@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineWidth@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineWidth@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineWidth@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineWidth@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetLineWidth=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetLineWidth=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineWidth@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineWidth@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineWidth@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineWidth@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineWidth@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineWidth@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineWidth@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineWidth@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineWidth@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineWidth@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineWidth@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineWidth@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineWidth@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineWidth@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetLineWidth=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetLineWidth=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDepthBias=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDepthBias=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDepthBias=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDepthBias=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDepthBias=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDepthBias=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDepthBias=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDepthBias=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDepthBias=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDepthBias=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDepthBias=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDepthBias=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDepthBias=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDepthBias=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDepthBias=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDepthBias=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDepthBias=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDepthBias=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDepthBias=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDepthBias=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDepthBias=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDepthBias=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDepthBias=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDepthBias=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDepthBias=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDepthBias=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetBlendConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetBlendConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetBlendConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetBlendConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetBlendConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetBlendConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetBlendConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetBlendConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetBlendConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetBlendConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetBlendConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetBlendConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetBlendConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetBlendConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetBlendConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetBlendConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetBlendConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetBlendConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetBlendConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetBlendConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetBlendConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetBlendConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetBlendConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetBlendConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetBlendConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetBlendConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetBlendConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetBlendConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetBlendConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetBlendConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetBlendConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetBlendConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetBlendConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetBlendConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetBlendConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetBlendConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetBlendConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetBlendConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetBlendConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetBlendConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetBlendConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetBlendConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetBlendConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetBlendConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetBlendConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetBlendConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetBlendConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetBlendConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetBlendConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetBlendConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetBlendConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetBlendConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetBlendConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetBlendConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetBlendConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetBlendConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetBlendConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetBlendConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetBlendConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetBlendConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetBlendConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetBlendConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetBlendConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetBlendConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetBlendConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetBlendConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetBlendConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetBlendConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetBlendConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetBlendConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetBlendConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetBlendConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetBlendConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetBlendConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetBlendConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetBlendConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetBlendConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetBlendConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetBlendConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetBlendConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetBlendConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetBlendConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetBlendConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetBlendConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetBlendConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetBlendConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetBlendConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetBlendConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetBlendConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetBlendConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetBlendConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetBlendConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetBlendConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetBlendConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetBlendConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetBlendConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetBlendConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetBlendConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetBlendConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetBlendConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetBlendConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetBlendConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetBlendConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetBlendConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetBlendConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetBlendConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetBlendConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetBlendConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetBlendConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetBlendConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetBlendConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetBlendConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetBlendConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetBlendConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetBlendConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetBlendConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetBlendConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetBlendConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetBlendConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetBlendConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetBlendConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetBlendConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetBlendConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetBlendConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetBlendConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetBlendConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetBlendConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetBlendConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetBlendConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetBlendConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetBlendConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetBlendConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetBlendConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetBlendConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetBlendConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetBlendConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetBlendConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetBlendConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetBlendConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetBlendConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetBlendConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetBlendConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetBlendConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetBlendConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetBlendConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetBlendConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetBlendConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetBlendConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetBlendConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetBlendConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetBlendConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetBlendConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetBlendConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetBlendConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetBlendConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetBlendConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetBlendConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetBlendConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetBlendConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetBlendConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetBlendConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetBlendConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetBlendConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetBlendConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetBlendConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetBlendConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetBlendConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetBlendConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetBlendConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetBlendConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetBlendConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetBlendConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetBlendConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetBlendConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetBlendConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetBlendConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetBlendConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetBlendConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetBlendConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetBlendConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetBlendConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetBlendConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetBlendConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetBlendConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetBlendConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetBlendConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetBlendConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetBlendConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetBlendConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetBlendConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetBlendConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetBlendConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetBlendConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetBlendConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetBlendConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetBlendConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetBlendConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetBlendConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetBlendConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetBlendConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetBlendConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetBlendConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetBlendConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetBlendConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetBlendConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetBlendConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetBlendConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetBlendConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBounds@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBounds@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBounds@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBounds@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBounds@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBounds@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBounds@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBounds@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBounds@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBounds@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBounds@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBounds@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBounds@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBounds@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDepthBounds=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDepthBounds=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBounds@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBounds@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBounds@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBounds@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBounds@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBounds@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBounds@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBounds@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBounds@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBounds@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBounds@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBounds@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBounds@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBounds@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDepthBounds=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDepthBounds=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBounds@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBounds@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBounds@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBounds@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBounds@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBounds@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBounds@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBounds@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBounds@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBounds@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBounds@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBounds@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBounds@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBounds@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDepthBounds=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDepthBounds=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBounds@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBounds@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBounds@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBounds@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBounds@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBounds@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBounds@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBounds@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBounds@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBounds@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBounds@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBounds@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBounds@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBounds@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDepthBounds=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDepthBounds=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBounds@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBounds@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBounds@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBounds@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBounds@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBounds@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBounds@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBounds@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBounds@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBounds@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBounds@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBounds@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBounds@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBounds@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDepthBounds=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDepthBounds=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBounds@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBounds@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBounds@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBounds@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBounds@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBounds@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBounds@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBounds@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBounds@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBounds@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBounds@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBounds@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBounds@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBounds@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDepthBounds=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDepthBounds=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBounds@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBounds@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBounds@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBounds@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBounds@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBounds@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBounds@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBounds@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBounds@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBounds@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBounds@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBounds@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBounds@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBounds@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDepthBounds=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDepthBounds=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBounds@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBounds@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBounds@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBounds@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBounds@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBounds@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBounds@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBounds@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBounds@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBounds@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBounds@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBounds@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBounds@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBounds@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDepthBounds=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDepthBounds=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBounds@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBounds@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBounds@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBounds@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBounds@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBounds@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBounds@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBounds@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBounds@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBounds@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBounds@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBounds@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBounds@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBounds@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDepthBounds=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDepthBounds=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBounds@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBounds@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBounds@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBounds@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBounds@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBounds@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBounds@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBounds@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBounds@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBounds@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBounds@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBounds@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBounds@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBounds@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDepthBounds=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDepthBounds=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBounds@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBounds@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBounds@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBounds@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBounds@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBounds@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBounds@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBounds@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBounds@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBounds@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBounds@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBounds@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBounds@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBounds@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDepthBounds=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDepthBounds=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBounds@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBounds@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBounds@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBounds@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBounds@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBounds@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBounds@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBounds@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBounds@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBounds@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBounds@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBounds@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBounds@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBounds@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDepthBounds=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDepthBounds=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBounds@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBounds@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBounds@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBounds@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBounds@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBounds@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBounds@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBounds@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBounds@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBounds@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBounds@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBounds@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBounds@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBounds@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDepthBounds=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDepthBounds=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilCompareMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilCompareMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilCompareMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilCompareMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilCompareMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilCompareMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilCompareMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilCompareMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilCompareMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilCompareMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilCompareMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilCompareMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilCompareMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilCompareMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetStencilCompareMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetStencilCompareMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilCompareMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilCompareMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilCompareMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilCompareMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilCompareMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilCompareMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilCompareMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilCompareMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilCompareMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilCompareMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilCompareMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilCompareMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilCompareMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilCompareMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetStencilCompareMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetStencilCompareMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilCompareMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilCompareMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilCompareMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilCompareMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilCompareMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilCompareMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilCompareMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilCompareMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilCompareMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilCompareMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilCompareMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilCompareMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilCompareMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilCompareMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetStencilCompareMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetStencilCompareMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilCompareMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilCompareMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilCompareMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilCompareMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilCompareMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilCompareMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilCompareMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilCompareMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilCompareMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilCompareMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilCompareMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilCompareMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilCompareMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilCompareMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetStencilCompareMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetStencilCompareMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilCompareMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilCompareMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilCompareMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilCompareMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilCompareMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilCompareMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilCompareMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilCompareMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilCompareMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilCompareMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilCompareMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilCompareMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilCompareMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilCompareMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetStencilCompareMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetStencilCompareMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilCompareMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilCompareMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilCompareMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilCompareMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilCompareMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilCompareMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilCompareMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilCompareMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilCompareMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilCompareMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilCompareMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilCompareMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilCompareMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilCompareMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetStencilCompareMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetStencilCompareMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilCompareMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilCompareMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilCompareMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilCompareMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilCompareMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilCompareMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilCompareMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilCompareMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilCompareMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilCompareMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilCompareMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilCompareMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilCompareMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilCompareMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetStencilCompareMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetStencilCompareMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilCompareMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilCompareMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilCompareMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilCompareMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilCompareMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilCompareMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilCompareMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilCompareMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilCompareMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilCompareMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilCompareMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilCompareMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilCompareMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilCompareMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetStencilCompareMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetStencilCompareMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilCompareMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilCompareMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilCompareMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilCompareMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilCompareMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilCompareMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilCompareMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilCompareMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilCompareMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilCompareMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilCompareMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilCompareMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilCompareMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilCompareMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetStencilCompareMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetStencilCompareMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilCompareMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilCompareMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilCompareMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilCompareMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilCompareMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilCompareMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilCompareMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilCompareMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilCompareMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilCompareMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilCompareMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilCompareMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilCompareMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilCompareMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetStencilCompareMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetStencilCompareMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilCompareMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilCompareMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilCompareMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilCompareMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilCompareMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilCompareMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilCompareMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilCompareMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilCompareMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilCompareMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilCompareMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilCompareMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilCompareMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilCompareMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetStencilCompareMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetStencilCompareMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilCompareMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilCompareMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilCompareMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilCompareMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilCompareMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilCompareMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilCompareMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilCompareMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilCompareMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilCompareMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilCompareMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilCompareMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilCompareMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilCompareMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetStencilCompareMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetStencilCompareMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilCompareMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilCompareMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilCompareMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilCompareMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilCompareMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilCompareMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilCompareMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilCompareMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilCompareMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilCompareMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilCompareMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilCompareMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilCompareMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilCompareMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetStencilCompareMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetStencilCompareMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilWriteMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilWriteMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilWriteMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilWriteMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilWriteMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilWriteMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilWriteMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilWriteMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilWriteMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilWriteMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilWriteMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilWriteMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilWriteMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilWriteMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetStencilWriteMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetStencilWriteMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilWriteMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilWriteMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilWriteMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilWriteMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilWriteMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilWriteMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilWriteMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilWriteMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilWriteMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilWriteMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilWriteMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilWriteMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilWriteMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilWriteMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetStencilWriteMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetStencilWriteMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilWriteMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilWriteMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilWriteMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilWriteMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilWriteMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilWriteMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilWriteMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilWriteMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilWriteMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilWriteMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilWriteMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilWriteMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilWriteMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilWriteMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetStencilWriteMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetStencilWriteMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilWriteMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilWriteMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilWriteMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilWriteMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilWriteMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilWriteMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilWriteMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilWriteMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilWriteMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilWriteMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilWriteMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilWriteMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilWriteMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilWriteMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetStencilWriteMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetStencilWriteMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilWriteMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilWriteMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilWriteMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilWriteMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilWriteMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilWriteMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilWriteMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilWriteMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilWriteMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilWriteMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilWriteMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilWriteMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilWriteMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilWriteMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetStencilWriteMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetStencilWriteMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilWriteMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilWriteMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilWriteMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilWriteMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilWriteMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilWriteMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilWriteMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilWriteMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilWriteMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilWriteMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilWriteMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilWriteMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilWriteMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilWriteMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetStencilWriteMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetStencilWriteMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilWriteMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilWriteMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilWriteMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilWriteMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilWriteMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilWriteMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilWriteMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilWriteMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilWriteMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilWriteMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilWriteMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilWriteMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilWriteMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilWriteMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetStencilWriteMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetStencilWriteMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilWriteMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilWriteMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilWriteMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilWriteMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilWriteMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilWriteMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilWriteMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilWriteMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilWriteMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilWriteMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilWriteMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilWriteMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilWriteMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilWriteMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetStencilWriteMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetStencilWriteMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilWriteMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilWriteMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilWriteMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilWriteMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilWriteMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilWriteMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilWriteMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilWriteMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilWriteMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilWriteMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilWriteMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilWriteMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilWriteMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilWriteMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetStencilWriteMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetStencilWriteMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilWriteMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilWriteMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilWriteMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilWriteMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilWriteMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilWriteMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilWriteMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilWriteMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilWriteMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilWriteMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilWriteMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilWriteMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilWriteMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilWriteMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetStencilWriteMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetStencilWriteMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilWriteMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilWriteMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilWriteMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilWriteMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilWriteMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilWriteMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilWriteMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilWriteMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilWriteMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilWriteMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilWriteMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilWriteMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilWriteMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilWriteMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetStencilWriteMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetStencilWriteMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilWriteMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilWriteMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilWriteMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilWriteMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilWriteMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilWriteMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilWriteMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilWriteMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilWriteMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilWriteMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilWriteMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilWriteMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilWriteMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilWriteMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetStencilWriteMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetStencilWriteMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilWriteMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilWriteMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilWriteMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilWriteMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilWriteMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilWriteMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilWriteMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilWriteMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilWriteMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilWriteMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilWriteMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilWriteMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilWriteMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilWriteMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetStencilWriteMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetStencilWriteMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilReference@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilReference@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilReference@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilReference@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilReference@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilReference@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilReference@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilReference@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilReference@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilReference@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilReference@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilReference@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilReference@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilReference@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetStencilReference=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetStencilReference=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilReference@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilReference@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilReference@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilReference@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilReference@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilReference@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilReference@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilReference@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilReference@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilReference@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilReference@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilReference@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilReference@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilReference@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetStencilReference=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetStencilReference=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilReference@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilReference@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilReference@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilReference@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilReference@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilReference@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilReference@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilReference@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilReference@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilReference@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilReference@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilReference@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilReference@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilReference@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetStencilReference=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetStencilReference=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilReference@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilReference@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilReference@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilReference@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilReference@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilReference@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilReference@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilReference@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilReference@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilReference@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilReference@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilReference@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilReference@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilReference@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetStencilReference=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetStencilReference=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilReference@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilReference@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilReference@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilReference@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilReference@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilReference@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilReference@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilReference@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilReference@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilReference@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilReference@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilReference@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilReference@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilReference@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetStencilReference=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetStencilReference=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilReference@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilReference@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilReference@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilReference@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilReference@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilReference@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilReference@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilReference@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilReference@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilReference@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilReference@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilReference@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilReference@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilReference@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetStencilReference=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetStencilReference=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilReference@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilReference@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilReference@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilReference@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilReference@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilReference@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilReference@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilReference@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilReference@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilReference@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilReference@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilReference@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilReference@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilReference@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetStencilReference=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetStencilReference=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilReference@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilReference@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilReference@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilReference@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilReference@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilReference@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilReference@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilReference@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilReference@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilReference@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilReference@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilReference@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilReference@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilReference@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetStencilReference=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetStencilReference=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilReference@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilReference@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilReference@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilReference@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilReference@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilReference@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilReference@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilReference@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilReference@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilReference@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilReference@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilReference@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilReference@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilReference@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetStencilReference=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetStencilReference=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilReference@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilReference@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilReference@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilReference@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilReference@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilReference@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilReference@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilReference@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilReference@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilReference@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilReference@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilReference@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilReference@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilReference@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetStencilReference=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetStencilReference=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilReference@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilReference@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilReference@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilReference@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilReference@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilReference@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilReference@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilReference@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilReference@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilReference@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilReference@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilReference@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilReference@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilReference@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetStencilReference=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetStencilReference=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilReference@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilReference@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilReference@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilReference@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilReference@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilReference@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilReference@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilReference@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilReference@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilReference@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilReference@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilReference@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilReference@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilReference@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetStencilReference=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetStencilReference=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilReference@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilReference@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilReference@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilReference@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilReference@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilReference@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilReference@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilReference@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilReference@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilReference@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilReference@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilReference@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilReference@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilReference@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetStencilReference=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetStencilReference=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindDescriptorSets=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindDescriptorSets=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindIndexBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindIndexBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindIndexBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindIndexBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindIndexBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindIndexBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindIndexBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindIndexBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindIndexBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindIndexBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindIndexBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindIndexBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindIndexBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindIndexBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindIndexBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindIndexBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindIndexBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindIndexBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindIndexBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindIndexBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindIndexBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindIndexBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindIndexBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindIndexBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindIndexBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindIndexBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindVertexBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindVertexBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindVertexBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindVertexBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindVertexBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindVertexBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindVertexBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindVertexBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindVertexBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindVertexBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindVertexBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindVertexBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindVertexBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindVertexBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindVertexBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindVertexBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindVertexBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindVertexBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindVertexBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindVertexBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindVertexBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindVertexBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindVertexBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindVertexBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindVertexBuffers=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindVertexBuffers=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDraw@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDraw@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDraw@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDraw@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDraw@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDraw@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDraw@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDraw@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDraw@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDraw@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDraw@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDraw@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDraw@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDraw@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDraw=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDraw=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDraw@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDraw@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDraw@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDraw@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDraw@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDraw@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDraw@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDraw@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDraw@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDraw@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDraw@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDraw@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDraw@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDraw@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDraw=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDraw=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDraw@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDraw@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDraw@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDraw@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDraw@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDraw@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDraw@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDraw@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDraw@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDraw@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDraw@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDraw@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDraw@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDraw@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDraw=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDraw=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDraw@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDraw@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDraw@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDraw@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDraw@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDraw@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDraw@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDraw@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDraw@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDraw@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDraw@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDraw@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDraw@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDraw@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDraw=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDraw=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDraw@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDraw@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDraw@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDraw@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDraw@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDraw@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDraw@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDraw@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDraw@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDraw@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDraw@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDraw@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDraw@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDraw@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDraw=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDraw=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDraw@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDraw@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDraw@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDraw@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDraw@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDraw@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDraw@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDraw@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDraw@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDraw@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDraw@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDraw@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDraw@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDraw@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDraw=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDraw=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDraw@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDraw@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDraw@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDraw@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDraw@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDraw@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDraw@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDraw@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDraw@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDraw@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDraw@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDraw@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDraw@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDraw@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDraw=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDraw=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDraw@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDraw@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDraw@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDraw@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDraw@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDraw@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDraw@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDraw@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDraw@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDraw@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDraw@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDraw@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDraw@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDraw@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDraw=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDraw=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDraw@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDraw@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDraw@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDraw@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDraw@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDraw@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDraw@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDraw@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDraw@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDraw@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDraw@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDraw@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDraw@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDraw@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDraw=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDraw=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDraw@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDraw@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDraw@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDraw@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDraw@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDraw@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDraw@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDraw@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDraw@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDraw@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDraw@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDraw@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDraw@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDraw@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDraw=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDraw=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDraw@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDraw@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDraw@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDraw@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDraw@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDraw@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDraw@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDraw@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDraw@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDraw@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDraw@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDraw@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDraw@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDraw@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDraw=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDraw=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDraw@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDraw@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDraw@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDraw@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDraw@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDraw@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDraw@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDraw@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDraw@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDraw@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDraw@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDraw@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDraw@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDraw@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDraw=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDraw=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDraw@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDraw@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDraw@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDraw@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDraw@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDraw@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDraw@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDraw@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDraw@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDraw@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDraw@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDraw@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDraw@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDraw@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDraw=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDraw=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexed@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexed@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexed@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexed@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexed@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexed@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexed@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexed@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexed@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexed@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexed@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexed@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexed@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexed@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawIndexed=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawIndexed=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexed@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexed@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexed@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexed@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexed@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexed@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexed@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexed@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexed@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexed@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexed@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexed@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexed@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexed@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawIndexed=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawIndexed=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexed@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexed@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexed@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexed@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexed@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexed@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexed@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexed@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexed@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexed@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexed@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexed@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexed@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexed@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawIndexed=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawIndexed=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexed@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexed@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexed@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexed@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexed@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexed@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexed@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexed@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexed@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexed@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexed@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexed@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexed@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexed@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawIndexed=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawIndexed=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexed@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexed@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexed@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexed@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexed@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexed@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexed@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexed@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexed@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexed@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexed@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexed@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexed@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexed@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawIndexed=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawIndexed=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexed@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexed@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexed@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexed@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexed@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexed@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexed@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexed@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexed@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexed@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexed@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexed@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexed@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexed@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawIndexed=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawIndexed=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexed@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexed@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexed@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexed@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexed@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexed@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexed@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexed@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexed@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexed@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexed@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexed@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexed@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexed@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawIndexed=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawIndexed=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexed@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexed@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexed@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexed@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexed@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexed@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexed@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexed@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexed@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexed@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexed@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexed@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexed@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexed@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawIndexed=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawIndexed=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexed@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexed@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexed@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexed@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexed@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexed@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexed@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexed@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexed@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexed@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexed@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexed@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexed@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexed@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawIndexed=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawIndexed=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexed@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexed@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexed@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexed@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexed@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexed@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexed@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexed@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexed@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexed@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexed@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexed@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexed@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexed@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawIndexed=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawIndexed=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexed@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexed@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexed@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexed@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexed@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexed@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexed@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexed@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexed@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexed@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexed@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexed@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexed@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexed@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawIndexed=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawIndexed=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexed@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexed@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexed@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexed@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexed@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexed@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexed@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexed@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexed@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexed@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexed@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexed@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexed@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexed@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawIndexed=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawIndexed=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexed@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexed@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexed@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexed@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexed@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexed@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexed@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexed@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexed@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexed@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexed@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexed@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexed@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexed@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawIndexed=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawIndexed=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawMultiEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawMultiEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawMultiEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawMultiEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawMultiEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawMultiEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawMultiEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawMultiEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawMultiEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawMultiEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawMultiEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawMultiEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawMultiEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawMultiEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawMultiEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawMultiEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawMultiEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawMultiEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawMultiEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawMultiEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawMultiEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawMultiEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawMultiEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawMultiEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawMultiEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawMultiEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMultiIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawMultiIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawMultiIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMultiIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawMultiIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawMultiIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMultiIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawMultiIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawMultiIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMultiIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawMultiIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawMultiIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMultiIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawMultiIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawMultiIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMultiIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawMultiIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawMultiIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMultiIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawMultiIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawMultiIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMultiIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawMultiIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawMultiIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMultiIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawMultiIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawMultiIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMultiIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawMultiIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawMultiIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMultiIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawMultiIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawMultiIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMultiIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawMultiIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawMultiIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMultiIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawMultiIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawMultiIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawIndexedIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawIndexedIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawIndexedIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawIndexedIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawIndexedIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawIndexedIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawIndexedIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawIndexedIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawIndexedIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawIndexedIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawIndexedIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawIndexedIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawIndexedIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawIndexedIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawIndexedIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawIndexedIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawIndexedIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawIndexedIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawIndexedIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawIndexedIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawIndexedIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawIndexedIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawIndexedIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawIndexedIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawIndexedIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawIndexedIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDispatch@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatch@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatch@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatch@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatch@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatch@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatch@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatch@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatch@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatch@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatch@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatch@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatch@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatch@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDispatch=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDispatch=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatch@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatch@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatch@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatch@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatch@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatch@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatch@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatch@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatch@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatch@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatch@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatch@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatch@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatch@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDispatch=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDispatch=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatch@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatch@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatch@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatch@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatch@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatch@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatch@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatch@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatch@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatch@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatch@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatch@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatch@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatch@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDispatch=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDispatch=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatch@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatch@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatch@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatch@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatch@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatch@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatch@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatch@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatch@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatch@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatch@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatch@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatch@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatch@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDispatch=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDispatch=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatch@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatch@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatch@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatch@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatch@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatch@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatch@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatch@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatch@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatch@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatch@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatch@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatch@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatch@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDispatch=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDispatch=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatch@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatch@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatch@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatch@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatch@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatch@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatch@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatch@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatch@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatch@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatch@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatch@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatch@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatch@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDispatch=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDispatch=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatch@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatch@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatch@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatch@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatch@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatch@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatch@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatch@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatch@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatch@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatch@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatch@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatch@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatch@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDispatch=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDispatch=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatch@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatch@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatch@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatch@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatch@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatch@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatch@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatch@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatch@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatch@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatch@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatch@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatch@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatch@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDispatch=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDispatch=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatch@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatch@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatch@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatch@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatch@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatch@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatch@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatch@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatch@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatch@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatch@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatch@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatch@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatch@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDispatch=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDispatch=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatch@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatch@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatch@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatch@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatch@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatch@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatch@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatch@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatch@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatch@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatch@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatch@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatch@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatch@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDispatch=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDispatch=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatch@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatch@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatch@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatch@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatch@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatch@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatch@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatch@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatch@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatch@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatch@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatch@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatch@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatch@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDispatch=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDispatch=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatch@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatch@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatch@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatch@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatch@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatch@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatch@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatch@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatch@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatch@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatch@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatch@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatch@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatch@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDispatch=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDispatch=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatch@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatch@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatch@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatch@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatch@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatch@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatch@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatch@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatch@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatch@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatch@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatch@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatch@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatch@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDispatch=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDispatch=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDispatchIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDispatchIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDispatchIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDispatchIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDispatchIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDispatchIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDispatchIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDispatchIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDispatchIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDispatchIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDispatchIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDispatchIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDispatchIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDispatchIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDispatchIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDispatchIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDispatchIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDispatchIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDispatchIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDispatchIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDispatchIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDispatchIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDispatchIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDispatchIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDispatchIndirect=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDispatchIndirect=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSubpassShadingHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSubpassShadingHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSubpassShadingHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSubpassShadingHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSubpassShadingHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSubpassShadingHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSubpassShadingHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSubpassShadingHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSubpassShadingHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSubpassShadingHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSubpassShadingHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSubpassShadingHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSubpassShadingHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSubpassShadingHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSubpassShadingHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSubpassShadingHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSubpassShadingHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSubpassShadingHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSubpassShadingHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSubpassShadingHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSubpassShadingHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSubpassShadingHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSubpassShadingHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSubpassShadingHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSubpassShadingHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSubpassShadingHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSubpassShadingHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSubpassShadingHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSubpassShadingHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSubpassShadingHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSubpassShadingHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSubpassShadingHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSubpassShadingHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSubpassShadingHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSubpassShadingHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSubpassShadingHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSubpassShadingHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSubpassShadingHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSubpassShadingHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSubpassShadingHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSubpassShadingHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSubpassShadingHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSubpassShadingHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSubpassShadingHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSubpassShadingHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSubpassShadingHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSubpassShadingHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSubpassShadingHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSubpassShadingHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSubpassShadingHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSubpassShadingHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSubpassShadingHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSubpassShadingHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSubpassShadingHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSubpassShadingHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSubpassShadingHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSubpassShadingHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSubpassShadingHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSubpassShadingHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSubpassShadingHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSubpassShadingHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSubpassShadingHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSubpassShadingHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSubpassShadingHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSubpassShadingHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSubpassShadingHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSubpassShadingHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSubpassShadingHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSubpassShadingHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSubpassShadingHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSubpassShadingHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSubpassShadingHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSubpassShadingHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSubpassShadingHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSubpassShadingHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSubpassShadingHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSubpassShadingHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSubpassShadingHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSubpassShadingHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSubpassShadingHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSubpassShadingHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSubpassShadingHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSubpassShadingHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSubpassShadingHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSubpassShadingHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSubpassShadingHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSubpassShadingHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSubpassShadingHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSubpassShadingHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSubpassShadingHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSubpassShadingHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSubpassShadingHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSubpassShadingHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSubpassShadingHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSubpassShadingHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSubpassShadingHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSubpassShadingHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSubpassShadingHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSubpassShadingHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSubpassShadingHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSubpassShadingHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSubpassShadingHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSubpassShadingHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSubpassShadingHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSubpassShadingHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSubpassShadingHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSubpassShadingHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSubpassShadingHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSubpassShadingHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSubpassShadingHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSubpassShadingHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSubpassShadingHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSubpassShadingHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSubpassShadingHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSubpassShadingHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSubpassShadingHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSubpassShadingHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSubpassShadingHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSubpassShadingHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSubpassShadingHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSubpassShadingHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSubpassShadingHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSubpassShadingHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSubpassShadingHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSubpassShadingHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSubpassShadingHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSubpassShadingHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSubpassShadingHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSubpassShadingHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSubpassShadingHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSubpassShadingHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSubpassShadingHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSubpassShadingHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSubpassShadingHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSubpassShadingHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSubpassShadingHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSubpassShadingHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSubpassShadingHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSubpassShadingHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSubpassShadingHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSubpassShadingHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSubpassShadingHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSubpassShadingHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSubpassShadingHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSubpassShadingHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSubpassShadingHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSubpassShadingHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSubpassShadingHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSubpassShadingHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSubpassShadingHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSubpassShadingHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSubpassShadingHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSubpassShadingHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSubpassShadingHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSubpassShadingHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSubpassShadingHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSubpassShadingHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSubpassShadingHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSubpassShadingHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSubpassShadingHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSubpassShadingHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSubpassShadingHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSubpassShadingHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSubpassShadingHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSubpassShadingHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSubpassShadingHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSubpassShadingHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSubpassShadingHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSubpassShadingHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSubpassShadingHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSubpassShadingHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSubpassShadingHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSubpassShadingHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSubpassShadingHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSubpassShadingHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSubpassShadingHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSubpassShadingHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSubpassShadingHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSubpassShadingHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSubpassShadingHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSubpassShadingHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSubpassShadingHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSubpassShadingHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSubpassShadingHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSubpassShadingHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSubpassShadingHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSubpassShadingHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSubpassShadingHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSubpassShadingHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSubpassShadingHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSubpassShadingHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSubpassShadingHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSubpassShadingHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSubpassShadingHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSubpassShadingHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSubpassShadingHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSubpassShadingHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSubpassShadingHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSubpassShadingHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSubpassShadingHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSubpassShadingHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSubpassShadingHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSubpassShadingHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSubpassShadingHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSubpassShadingHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSubpassShadingHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSubpassShadingHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSubpassShadingHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawClusterHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawClusterHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawClusterHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawClusterHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawClusterHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawClusterHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawClusterHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawClusterHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawClusterHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawClusterHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawClusterHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawClusterHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawClusterHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawClusterHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawClusterHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawClusterHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawClusterHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawClusterHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawClusterHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawClusterHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawClusterHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawClusterHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawClusterHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawClusterHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawClusterHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawClusterHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterIndirectHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterIndirectHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterIndirectHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterIndirectHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterIndirectHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterIndirectHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterIndirectHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterIndirectHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterIndirectHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterIndirectHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterIndirectHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterIndirectHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterIndirectHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawClusterIndirectHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawClusterIndirectHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawClusterIndirectHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterIndirectHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterIndirectHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterIndirectHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterIndirectHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterIndirectHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterIndirectHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterIndirectHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterIndirectHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterIndirectHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterIndirectHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterIndirectHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterIndirectHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterIndirectHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawClusterIndirectHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawClusterIndirectHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawClusterIndirectHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterIndirectHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterIndirectHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterIndirectHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterIndirectHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterIndirectHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterIndirectHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterIndirectHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterIndirectHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterIndirectHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterIndirectHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterIndirectHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterIndirectHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterIndirectHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawClusterIndirectHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawClusterIndirectHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawClusterIndirectHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterIndirectHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterIndirectHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterIndirectHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterIndirectHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterIndirectHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterIndirectHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterIndirectHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterIndirectHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterIndirectHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterIndirectHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterIndirectHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterIndirectHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterIndirectHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawClusterIndirectHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawClusterIndirectHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawClusterIndirectHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterIndirectHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterIndirectHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterIndirectHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterIndirectHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterIndirectHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterIndirectHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterIndirectHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterIndirectHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterIndirectHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterIndirectHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterIndirectHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterIndirectHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterIndirectHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawClusterIndirectHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawClusterIndirectHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawClusterIndirectHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterIndirectHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterIndirectHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterIndirectHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterIndirectHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterIndirectHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterIndirectHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterIndirectHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterIndirectHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterIndirectHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterIndirectHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterIndirectHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterIndirectHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterIndirectHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawClusterIndirectHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawClusterIndirectHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawClusterIndirectHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterIndirectHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterIndirectHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterIndirectHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterIndirectHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterIndirectHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterIndirectHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterIndirectHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterIndirectHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterIndirectHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterIndirectHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterIndirectHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterIndirectHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterIndirectHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawClusterIndirectHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawClusterIndirectHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawClusterIndirectHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterIndirectHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterIndirectHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterIndirectHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterIndirectHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterIndirectHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterIndirectHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterIndirectHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterIndirectHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterIndirectHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterIndirectHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterIndirectHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterIndirectHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterIndirectHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawClusterIndirectHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawClusterIndirectHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawClusterIndirectHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterIndirectHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterIndirectHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterIndirectHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterIndirectHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterIndirectHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterIndirectHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterIndirectHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterIndirectHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterIndirectHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterIndirectHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterIndirectHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterIndirectHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterIndirectHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawClusterIndirectHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawClusterIndirectHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawClusterIndirectHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterIndirectHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterIndirectHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterIndirectHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterIndirectHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterIndirectHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterIndirectHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterIndirectHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterIndirectHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterIndirectHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterIndirectHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterIndirectHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterIndirectHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterIndirectHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawClusterIndirectHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawClusterIndirectHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawClusterIndirectHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterIndirectHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterIndirectHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterIndirectHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterIndirectHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterIndirectHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterIndirectHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterIndirectHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterIndirectHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterIndirectHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterIndirectHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterIndirectHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterIndirectHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterIndirectHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawClusterIndirectHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawClusterIndirectHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawClusterIndirectHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterIndirectHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterIndirectHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterIndirectHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterIndirectHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterIndirectHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterIndirectHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterIndirectHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterIndirectHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterIndirectHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterIndirectHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterIndirectHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterIndirectHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterIndirectHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawClusterIndirectHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawClusterIndirectHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawClusterIndirectHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterIndirectHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterIndirectHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterIndirectHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterIndirectHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterIndirectHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterIndirectHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterIndirectHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterIndirectHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterIndirectHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterIndirectHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterIndirectHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterIndirectHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterIndirectHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawClusterIndirectHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawClusterIndirectHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawClusterIndirectHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdUpdatePipelineIndirectBufferNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdatePipelineIndirectBufferNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdatePipelineIndirectBufferNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdatePipelineIndirectBufferNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdatePipelineIndirectBufferNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdatePipelineIndirectBufferNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdatePipelineIndirectBufferNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdatePipelineIndirectBufferNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdatePipelineIndirectBufferNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdatePipelineIndirectBufferNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdatePipelineIndirectBufferNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdatePipelineIndirectBufferNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdatePipelineIndirectBufferNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdatePipelineIndirectBufferNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdUpdatePipelineIndirectBufferNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdUpdatePipelineIndirectBufferNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdatePipelineIndirectBufferNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdatePipelineIndirectBufferNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdatePipelineIndirectBufferNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdatePipelineIndirectBufferNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdatePipelineIndirectBufferNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdatePipelineIndirectBufferNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdatePipelineIndirectBufferNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdatePipelineIndirectBufferNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdatePipelineIndirectBufferNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdatePipelineIndirectBufferNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdatePipelineIndirectBufferNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdatePipelineIndirectBufferNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdatePipelineIndirectBufferNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdatePipelineIndirectBufferNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdUpdatePipelineIndirectBufferNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdUpdatePipelineIndirectBufferNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdatePipelineIndirectBufferNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdatePipelineIndirectBufferNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdatePipelineIndirectBufferNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdatePipelineIndirectBufferNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdatePipelineIndirectBufferNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdatePipelineIndirectBufferNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdatePipelineIndirectBufferNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdatePipelineIndirectBufferNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdatePipelineIndirectBufferNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdatePipelineIndirectBufferNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdatePipelineIndirectBufferNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdatePipelineIndirectBufferNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdatePipelineIndirectBufferNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdatePipelineIndirectBufferNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdUpdatePipelineIndirectBufferNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdUpdatePipelineIndirectBufferNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdatePipelineIndirectBufferNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdatePipelineIndirectBufferNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdatePipelineIndirectBufferNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdatePipelineIndirectBufferNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdatePipelineIndirectBufferNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdatePipelineIndirectBufferNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdatePipelineIndirectBufferNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdatePipelineIndirectBufferNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdatePipelineIndirectBufferNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdatePipelineIndirectBufferNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdatePipelineIndirectBufferNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdatePipelineIndirectBufferNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdatePipelineIndirectBufferNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdatePipelineIndirectBufferNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdUpdatePipelineIndirectBufferNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdUpdatePipelineIndirectBufferNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdatePipelineIndirectBufferNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdatePipelineIndirectBufferNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdatePipelineIndirectBufferNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdatePipelineIndirectBufferNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdatePipelineIndirectBufferNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdatePipelineIndirectBufferNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdatePipelineIndirectBufferNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdatePipelineIndirectBufferNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdatePipelineIndirectBufferNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdatePipelineIndirectBufferNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdatePipelineIndirectBufferNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdatePipelineIndirectBufferNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdatePipelineIndirectBufferNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdatePipelineIndirectBufferNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdUpdatePipelineIndirectBufferNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdUpdatePipelineIndirectBufferNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdatePipelineIndirectBufferNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdatePipelineIndirectBufferNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdatePipelineIndirectBufferNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdatePipelineIndirectBufferNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdatePipelineIndirectBufferNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdatePipelineIndirectBufferNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdatePipelineIndirectBufferNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdatePipelineIndirectBufferNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdatePipelineIndirectBufferNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdatePipelineIndirectBufferNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdatePipelineIndirectBufferNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdatePipelineIndirectBufferNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdatePipelineIndirectBufferNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdatePipelineIndirectBufferNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdUpdatePipelineIndirectBufferNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdUpdatePipelineIndirectBufferNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdatePipelineIndirectBufferNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdatePipelineIndirectBufferNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdatePipelineIndirectBufferNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdatePipelineIndirectBufferNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdatePipelineIndirectBufferNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdatePipelineIndirectBufferNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdatePipelineIndirectBufferNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdatePipelineIndirectBufferNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdatePipelineIndirectBufferNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdatePipelineIndirectBufferNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdatePipelineIndirectBufferNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdatePipelineIndirectBufferNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdatePipelineIndirectBufferNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdatePipelineIndirectBufferNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdUpdatePipelineIndirectBufferNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdUpdatePipelineIndirectBufferNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdatePipelineIndirectBufferNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdatePipelineIndirectBufferNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdatePipelineIndirectBufferNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdatePipelineIndirectBufferNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdatePipelineIndirectBufferNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdatePipelineIndirectBufferNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdatePipelineIndirectBufferNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdatePipelineIndirectBufferNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdatePipelineIndirectBufferNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdatePipelineIndirectBufferNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdatePipelineIndirectBufferNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdatePipelineIndirectBufferNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdatePipelineIndirectBufferNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdatePipelineIndirectBufferNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdUpdatePipelineIndirectBufferNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdUpdatePipelineIndirectBufferNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdatePipelineIndirectBufferNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdatePipelineIndirectBufferNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdatePipelineIndirectBufferNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdatePipelineIndirectBufferNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdatePipelineIndirectBufferNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdatePipelineIndirectBufferNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdatePipelineIndirectBufferNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdatePipelineIndirectBufferNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdatePipelineIndirectBufferNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdatePipelineIndirectBufferNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdatePipelineIndirectBufferNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdatePipelineIndirectBufferNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdatePipelineIndirectBufferNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdatePipelineIndirectBufferNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdUpdatePipelineIndirectBufferNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdUpdatePipelineIndirectBufferNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdatePipelineIndirectBufferNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdatePipelineIndirectBufferNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdatePipelineIndirectBufferNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdatePipelineIndirectBufferNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdatePipelineIndirectBufferNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdatePipelineIndirectBufferNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdatePipelineIndirectBufferNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdatePipelineIndirectBufferNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdatePipelineIndirectBufferNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdatePipelineIndirectBufferNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdatePipelineIndirectBufferNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdatePipelineIndirectBufferNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdatePipelineIndirectBufferNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdatePipelineIndirectBufferNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdUpdatePipelineIndirectBufferNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdUpdatePipelineIndirectBufferNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdatePipelineIndirectBufferNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdatePipelineIndirectBufferNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdatePipelineIndirectBufferNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdatePipelineIndirectBufferNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdatePipelineIndirectBufferNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdatePipelineIndirectBufferNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdatePipelineIndirectBufferNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdatePipelineIndirectBufferNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdatePipelineIndirectBufferNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdatePipelineIndirectBufferNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdatePipelineIndirectBufferNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdatePipelineIndirectBufferNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdatePipelineIndirectBufferNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdatePipelineIndirectBufferNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdUpdatePipelineIndirectBufferNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdUpdatePipelineIndirectBufferNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdatePipelineIndirectBufferNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdatePipelineIndirectBufferNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdatePipelineIndirectBufferNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdatePipelineIndirectBufferNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdatePipelineIndirectBufferNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdatePipelineIndirectBufferNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdatePipelineIndirectBufferNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdatePipelineIndirectBufferNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdatePipelineIndirectBufferNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdatePipelineIndirectBufferNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdatePipelineIndirectBufferNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdatePipelineIndirectBufferNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdatePipelineIndirectBufferNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdatePipelineIndirectBufferNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdUpdatePipelineIndirectBufferNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdUpdatePipelineIndirectBufferNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdatePipelineIndirectBufferNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdatePipelineIndirectBufferNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdatePipelineIndirectBufferNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdatePipelineIndirectBufferNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdatePipelineIndirectBufferNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdatePipelineIndirectBufferNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdatePipelineIndirectBufferNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdatePipelineIndirectBufferNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdatePipelineIndirectBufferNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdatePipelineIndirectBufferNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdatePipelineIndirectBufferNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdatePipelineIndirectBufferNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdatePipelineIndirectBufferNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdatePipelineIndirectBufferNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdUpdatePipelineIndirectBufferNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdUpdatePipelineIndirectBufferNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBlitImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBlitImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBlitImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBlitImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBlitImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBlitImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBlitImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBlitImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBlitImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBlitImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBlitImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBlitImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBlitImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBlitImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBlitImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBlitImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBlitImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBlitImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBlitImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBlitImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBlitImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBlitImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBlitImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBlitImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBlitImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBlitImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyBufferToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyBufferToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyBufferToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyBufferToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyBufferToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyBufferToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyBufferToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyBufferToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyBufferToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyBufferToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyBufferToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyBufferToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyBufferToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyBufferToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyBufferToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyBufferToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyBufferToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyBufferToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyBufferToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyBufferToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyBufferToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyBufferToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyBufferToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyBufferToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyBufferToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyBufferToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyImageToBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyImageToBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyImageToBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyImageToBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyImageToBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyImageToBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyImageToBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyImageToBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyImageToBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyImageToBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyImageToBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyImageToBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyImageToBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyImageToBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyImageToBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyImageToBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyImageToBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyImageToBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyImageToBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyImageToBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyImageToBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyImageToBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyImageToBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyImageToBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyImageToBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyImageToBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyMemoryIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyMemoryIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyMemoryIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyMemoryIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyMemoryIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyMemoryIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyMemoryIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyMemoryIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyMemoryIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyMemoryIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyMemoryIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyMemoryIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyMemoryIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyMemoryIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyMemoryIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyMemoryIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyMemoryIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyMemoryIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyMemoryIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyMemoryIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyMemoryIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyMemoryIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyMemoryIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyMemoryIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyMemoryIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyMemoryIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyMemoryIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyMemoryIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyMemoryIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyMemoryIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyMemoryIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyMemoryIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyMemoryIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyMemoryIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyMemoryIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyMemoryIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyMemoryIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyMemoryIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyMemoryIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyMemoryIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyMemoryIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyMemoryIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyMemoryIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyMemoryIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyMemoryIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyMemoryIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyMemoryIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyMemoryIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyMemoryIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyMemoryIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyMemoryIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyMemoryIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyMemoryToImageIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyMemoryToImageIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyMemoryToImageIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyMemoryToImageIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyMemoryToImageIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyMemoryToImageIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyMemoryToImageIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyMemoryToImageIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyMemoryToImageIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyMemoryToImageIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyMemoryToImageIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyMemoryToImageIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyMemoryToImageIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyMemoryToImageIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyMemoryToImageIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyMemoryToImageIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyMemoryToImageIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyMemoryToImageIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyMemoryToImageIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyMemoryToImageIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyMemoryToImageIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyMemoryToImageIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyMemoryToImageIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyMemoryToImageIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyMemoryToImageIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyMemoryToImageIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyMemoryToImageIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyMemoryToImageIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyMemoryToImageIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyMemoryToImageIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyMemoryToImageIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyMemoryToImageIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyMemoryToImageIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyMemoryToImageIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyMemoryToImageIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyMemoryToImageIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyMemoryToImageIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyMemoryToImageIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyMemoryToImageIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyMemoryToImageIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyMemoryToImageIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyMemoryToImageIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyMemoryToImageIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyMemoryToImageIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyMemoryToImageIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyMemoryToImageIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyMemoryToImageIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyMemoryToImageIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyMemoryToImageIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyMemoryToImageIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyMemoryToImageIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyMemoryToImageIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdUpdateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdUpdateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdUpdateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdUpdateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdUpdateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdUpdateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdUpdateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdUpdateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdUpdateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdUpdateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdUpdateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdUpdateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdUpdateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdUpdateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdUpdateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdUpdateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdUpdateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdUpdateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdUpdateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdUpdateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdUpdateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdUpdateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdUpdateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdUpdateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdUpdateBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdUpdateBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdFillBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdFillBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdFillBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdFillBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdFillBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdFillBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdFillBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdFillBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdFillBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdFillBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdFillBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdFillBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdFillBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdFillBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdFillBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdFillBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdFillBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdFillBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdFillBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdFillBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdFillBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdFillBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdFillBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdFillBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdFillBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillBuffer@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillBuffer@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillBuffer@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillBuffer@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillBuffer@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillBuffer@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillBuffer@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillBuffer@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillBuffer@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillBuffer@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillBuffer@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillBuffer@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillBuffer@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillBuffer@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdFillBuffer=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdFillBuffer=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdClearColorImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearColorImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearColorImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearColorImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearColorImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearColorImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearColorImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearColorImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearColorImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearColorImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearColorImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearColorImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearColorImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearColorImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdClearColorImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdClearColorImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearColorImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearColorImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearColorImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearColorImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearColorImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearColorImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearColorImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearColorImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearColorImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearColorImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearColorImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearColorImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearColorImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearColorImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdClearColorImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdClearColorImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearColorImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearColorImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearColorImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearColorImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearColorImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearColorImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearColorImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearColorImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearColorImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearColorImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearColorImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearColorImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearColorImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearColorImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdClearColorImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdClearColorImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearColorImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearColorImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearColorImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearColorImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearColorImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearColorImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearColorImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearColorImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearColorImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearColorImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearColorImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearColorImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearColorImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearColorImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdClearColorImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdClearColorImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearColorImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearColorImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearColorImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearColorImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearColorImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearColorImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearColorImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearColorImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearColorImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearColorImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearColorImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearColorImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearColorImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearColorImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdClearColorImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdClearColorImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearColorImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearColorImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearColorImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearColorImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearColorImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearColorImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearColorImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearColorImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearColorImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearColorImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearColorImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearColorImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearColorImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearColorImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdClearColorImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdClearColorImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearColorImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearColorImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearColorImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearColorImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearColorImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearColorImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearColorImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearColorImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearColorImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearColorImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearColorImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearColorImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearColorImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearColorImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdClearColorImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdClearColorImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearColorImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearColorImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearColorImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearColorImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearColorImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearColorImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearColorImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearColorImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearColorImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearColorImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearColorImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearColorImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearColorImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearColorImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdClearColorImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdClearColorImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearColorImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearColorImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearColorImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearColorImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearColorImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearColorImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearColorImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearColorImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearColorImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearColorImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearColorImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearColorImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearColorImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearColorImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdClearColorImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdClearColorImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearColorImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearColorImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearColorImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearColorImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearColorImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearColorImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearColorImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearColorImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearColorImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearColorImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearColorImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearColorImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearColorImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearColorImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdClearColorImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdClearColorImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearColorImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearColorImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearColorImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearColorImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearColorImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearColorImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearColorImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearColorImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearColorImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearColorImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearColorImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearColorImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearColorImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearColorImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdClearColorImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdClearColorImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearColorImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearColorImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearColorImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearColorImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearColorImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearColorImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearColorImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearColorImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearColorImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearColorImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearColorImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearColorImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearColorImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearColorImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdClearColorImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdClearColorImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearColorImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearColorImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearColorImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearColorImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearColorImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearColorImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearColorImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearColorImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearColorImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearColorImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearColorImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearColorImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearColorImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearColorImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdClearColorImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdClearColorImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdClearDepthStencilImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearDepthStencilImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearDepthStencilImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearDepthStencilImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearDepthStencilImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearDepthStencilImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearDepthStencilImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearDepthStencilImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearDepthStencilImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearDepthStencilImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearDepthStencilImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearDepthStencilImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearDepthStencilImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearDepthStencilImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdClearDepthStencilImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdClearDepthStencilImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearDepthStencilImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearDepthStencilImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearDepthStencilImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearDepthStencilImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearDepthStencilImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearDepthStencilImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearDepthStencilImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearDepthStencilImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearDepthStencilImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearDepthStencilImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearDepthStencilImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearDepthStencilImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearDepthStencilImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearDepthStencilImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdClearDepthStencilImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdClearDepthStencilImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearDepthStencilImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearDepthStencilImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearDepthStencilImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearDepthStencilImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearDepthStencilImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearDepthStencilImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearDepthStencilImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearDepthStencilImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearDepthStencilImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearDepthStencilImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearDepthStencilImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearDepthStencilImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearDepthStencilImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearDepthStencilImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdClearDepthStencilImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdClearDepthStencilImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearDepthStencilImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearDepthStencilImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearDepthStencilImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearDepthStencilImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearDepthStencilImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearDepthStencilImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearDepthStencilImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearDepthStencilImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearDepthStencilImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearDepthStencilImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearDepthStencilImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearDepthStencilImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearDepthStencilImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearDepthStencilImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdClearDepthStencilImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdClearDepthStencilImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearDepthStencilImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearDepthStencilImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearDepthStencilImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearDepthStencilImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearDepthStencilImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearDepthStencilImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearDepthStencilImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearDepthStencilImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearDepthStencilImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearDepthStencilImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearDepthStencilImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearDepthStencilImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearDepthStencilImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearDepthStencilImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdClearDepthStencilImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdClearDepthStencilImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearDepthStencilImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearDepthStencilImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearDepthStencilImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearDepthStencilImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearDepthStencilImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearDepthStencilImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearDepthStencilImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearDepthStencilImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearDepthStencilImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearDepthStencilImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearDepthStencilImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearDepthStencilImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearDepthStencilImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearDepthStencilImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdClearDepthStencilImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdClearDepthStencilImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearDepthStencilImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearDepthStencilImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearDepthStencilImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearDepthStencilImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearDepthStencilImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearDepthStencilImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearDepthStencilImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearDepthStencilImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearDepthStencilImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearDepthStencilImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearDepthStencilImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearDepthStencilImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearDepthStencilImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearDepthStencilImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdClearDepthStencilImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdClearDepthStencilImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearDepthStencilImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearDepthStencilImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearDepthStencilImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearDepthStencilImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearDepthStencilImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearDepthStencilImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearDepthStencilImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearDepthStencilImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearDepthStencilImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearDepthStencilImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearDepthStencilImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearDepthStencilImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearDepthStencilImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearDepthStencilImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdClearDepthStencilImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdClearDepthStencilImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearDepthStencilImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearDepthStencilImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearDepthStencilImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearDepthStencilImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearDepthStencilImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearDepthStencilImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearDepthStencilImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearDepthStencilImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearDepthStencilImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearDepthStencilImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearDepthStencilImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearDepthStencilImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearDepthStencilImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearDepthStencilImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdClearDepthStencilImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdClearDepthStencilImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearDepthStencilImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearDepthStencilImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearDepthStencilImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearDepthStencilImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearDepthStencilImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearDepthStencilImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearDepthStencilImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearDepthStencilImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearDepthStencilImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearDepthStencilImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearDepthStencilImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearDepthStencilImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearDepthStencilImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearDepthStencilImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdClearDepthStencilImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdClearDepthStencilImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearDepthStencilImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearDepthStencilImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearDepthStencilImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearDepthStencilImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearDepthStencilImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearDepthStencilImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearDepthStencilImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearDepthStencilImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearDepthStencilImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearDepthStencilImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearDepthStencilImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearDepthStencilImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearDepthStencilImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearDepthStencilImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdClearDepthStencilImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdClearDepthStencilImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearDepthStencilImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearDepthStencilImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearDepthStencilImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearDepthStencilImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearDepthStencilImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearDepthStencilImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearDepthStencilImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearDepthStencilImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearDepthStencilImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearDepthStencilImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearDepthStencilImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearDepthStencilImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearDepthStencilImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearDepthStencilImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdClearDepthStencilImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdClearDepthStencilImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearDepthStencilImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearDepthStencilImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearDepthStencilImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearDepthStencilImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearDepthStencilImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearDepthStencilImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearDepthStencilImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearDepthStencilImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearDepthStencilImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearDepthStencilImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearDepthStencilImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearDepthStencilImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearDepthStencilImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearDepthStencilImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdClearDepthStencilImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdClearDepthStencilImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdClearAttachments@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearAttachments@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearAttachments@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearAttachments@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearAttachments@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearAttachments@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearAttachments@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearAttachments@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearAttachments@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearAttachments@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearAttachments@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearAttachments@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearAttachments@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdClearAttachments@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdClearAttachments=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdClearAttachments=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearAttachments@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearAttachments@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearAttachments@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearAttachments@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearAttachments@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearAttachments@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearAttachments@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearAttachments@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearAttachments@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearAttachments@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearAttachments@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearAttachments@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearAttachments@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdClearAttachments@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdClearAttachments=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdClearAttachments=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearAttachments@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearAttachments@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearAttachments@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearAttachments@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearAttachments@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearAttachments@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearAttachments@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearAttachments@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearAttachments@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearAttachments@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearAttachments@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearAttachments@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearAttachments@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdClearAttachments@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdClearAttachments=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdClearAttachments=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearAttachments@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearAttachments@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearAttachments@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearAttachments@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearAttachments@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearAttachments@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearAttachments@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearAttachments@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearAttachments@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearAttachments@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearAttachments@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearAttachments@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearAttachments@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdClearAttachments@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdClearAttachments=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdClearAttachments=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearAttachments@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearAttachments@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearAttachments@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearAttachments@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearAttachments@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearAttachments@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearAttachments@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearAttachments@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearAttachments@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearAttachments@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearAttachments@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearAttachments@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearAttachments@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdClearAttachments@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdClearAttachments=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdClearAttachments=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearAttachments@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearAttachments@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearAttachments@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearAttachments@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearAttachments@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearAttachments@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearAttachments@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearAttachments@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearAttachments@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearAttachments@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearAttachments@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearAttachments@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearAttachments@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdClearAttachments@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdClearAttachments=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdClearAttachments=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearAttachments@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearAttachments@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearAttachments@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearAttachments@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearAttachments@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearAttachments@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearAttachments@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearAttachments@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearAttachments@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearAttachments@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearAttachments@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearAttachments@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearAttachments@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdClearAttachments@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdClearAttachments=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdClearAttachments=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearAttachments@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearAttachments@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearAttachments@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearAttachments@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearAttachments@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearAttachments@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearAttachments@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearAttachments@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearAttachments@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearAttachments@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearAttachments@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearAttachments@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearAttachments@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdClearAttachments@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdClearAttachments=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdClearAttachments=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearAttachments@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearAttachments@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearAttachments@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearAttachments@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearAttachments@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearAttachments@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearAttachments@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearAttachments@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearAttachments@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearAttachments@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearAttachments@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearAttachments@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearAttachments@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdClearAttachments@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdClearAttachments=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdClearAttachments=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearAttachments@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearAttachments@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearAttachments@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearAttachments@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearAttachments@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearAttachments@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearAttachments@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearAttachments@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearAttachments@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearAttachments@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearAttachments@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearAttachments@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearAttachments@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdClearAttachments@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdClearAttachments=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdClearAttachments=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearAttachments@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearAttachments@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearAttachments@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearAttachments@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearAttachments@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearAttachments@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearAttachments@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearAttachments@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearAttachments@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearAttachments@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearAttachments@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearAttachments@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearAttachments@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdClearAttachments@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdClearAttachments=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdClearAttachments=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearAttachments@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearAttachments@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearAttachments@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearAttachments@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearAttachments@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearAttachments@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearAttachments@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearAttachments@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearAttachments@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearAttachments@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearAttachments@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearAttachments@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearAttachments@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdClearAttachments@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdClearAttachments=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdClearAttachments=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearAttachments@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearAttachments@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearAttachments@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearAttachments@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearAttachments@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearAttachments@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearAttachments@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearAttachments@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearAttachments@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearAttachments@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearAttachments@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearAttachments@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearAttachments@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdClearAttachments@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdClearAttachments=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdClearAttachments=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdResolveImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdResolveImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdResolveImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdResolveImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdResolveImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdResolveImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdResolveImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdResolveImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdResolveImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdResolveImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdResolveImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdResolveImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdResolveImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdResolveImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdResolveImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdResolveImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdResolveImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdResolveImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdResolveImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdResolveImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdResolveImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdResolveImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdResolveImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdResolveImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdResolveImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdResolveImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdResetEvent=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdResetEvent=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdWaitEvents=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdWaitEvents=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdWaitEvents=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdWaitEvents=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdWaitEvents=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdWaitEvents=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdWaitEvents=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdWaitEvents=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdWaitEvents=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdWaitEvents=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdWaitEvents=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdWaitEvents=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdWaitEvents=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdWaitEvents=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdWaitEvents=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdWaitEvents=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdWaitEvents=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdWaitEvents=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdWaitEvents=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdWaitEvents=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdWaitEvents=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdWaitEvents=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdWaitEvents=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdWaitEvents=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdWaitEvents=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdWaitEvents=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdPipelineBarrier=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdPipelineBarrier=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdPipelineBarrier=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdPipelineBarrier=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdPipelineBarrier=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdPipelineBarrier=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdPipelineBarrier=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdPipelineBarrier=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdPipelineBarrier=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdPipelineBarrier=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdPipelineBarrier=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdPipelineBarrier=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdPipelineBarrier=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdPipelineBarrier=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdPipelineBarrier=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdPipelineBarrier=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdPipelineBarrier=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdPipelineBarrier=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdPipelineBarrier=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdPipelineBarrier=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdPipelineBarrier=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdPipelineBarrier=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdPipelineBarrier=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdPipelineBarrier=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdPipelineBarrier=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdPipelineBarrier=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBeginQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBeginQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBeginQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBeginQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBeginQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBeginQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBeginQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBeginQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBeginQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBeginQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBeginQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBeginQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBeginQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBeginQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBeginQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBeginQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBeginQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBeginQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBeginQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBeginQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBeginQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBeginQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBeginQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBeginQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBeginQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBeginQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdEndQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdEndQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdEndQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdEndQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdEndQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdEndQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdEndQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdEndQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdEndQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdEndQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdEndQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdEndQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdEndQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdEndQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdEndQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdEndQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdEndQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdEndQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdEndQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdEndQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdEndQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdEndQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdEndQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdEndQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdEndQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQuery@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQuery@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQuery@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQuery@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQuery@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQuery@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQuery@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQuery@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQuery@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQuery@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQuery@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQuery@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQuery@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQuery@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdEndQuery=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdEndQuery=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBeginConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBeginConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBeginConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBeginConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBeginConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBeginConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBeginConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBeginConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBeginConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBeginConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBeginConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBeginConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBeginConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBeginConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBeginConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBeginConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBeginConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBeginConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBeginConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBeginConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBeginConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBeginConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBeginConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBeginConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBeginConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBeginConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdEndConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdEndConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdEndConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdEndConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdEndConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdEndConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdEndConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdEndConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdEndConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdEndConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdEndConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdEndConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdEndConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdEndConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdEndConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdEndConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdEndConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdEndConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdEndConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdEndConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdEndConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdEndConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdEndConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdEndConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdEndConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndConditionalRenderingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndConditionalRenderingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndConditionalRenderingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndConditionalRenderingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndConditionalRenderingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndConditionalRenderingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndConditionalRenderingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndConditionalRenderingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndConditionalRenderingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndConditionalRenderingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndConditionalRenderingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndConditionalRenderingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndConditionalRenderingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndConditionalRenderingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdEndConditionalRenderingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdEndConditionalRenderingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBeginCustomResolveEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginCustomResolveEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginCustomResolveEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginCustomResolveEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginCustomResolveEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginCustomResolveEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginCustomResolveEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginCustomResolveEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginCustomResolveEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginCustomResolveEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginCustomResolveEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginCustomResolveEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginCustomResolveEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginCustomResolveEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBeginCustomResolveEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBeginCustomResolveEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginCustomResolveEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginCustomResolveEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginCustomResolveEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginCustomResolveEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginCustomResolveEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginCustomResolveEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginCustomResolveEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginCustomResolveEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginCustomResolveEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginCustomResolveEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginCustomResolveEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginCustomResolveEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginCustomResolveEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginCustomResolveEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBeginCustomResolveEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBeginCustomResolveEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginCustomResolveEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginCustomResolveEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginCustomResolveEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginCustomResolveEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginCustomResolveEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginCustomResolveEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginCustomResolveEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginCustomResolveEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginCustomResolveEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginCustomResolveEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginCustomResolveEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginCustomResolveEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginCustomResolveEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginCustomResolveEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBeginCustomResolveEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBeginCustomResolveEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginCustomResolveEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginCustomResolveEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginCustomResolveEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginCustomResolveEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginCustomResolveEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginCustomResolveEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginCustomResolveEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginCustomResolveEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginCustomResolveEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginCustomResolveEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginCustomResolveEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginCustomResolveEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginCustomResolveEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginCustomResolveEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBeginCustomResolveEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBeginCustomResolveEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginCustomResolveEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginCustomResolveEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginCustomResolveEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginCustomResolveEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginCustomResolveEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginCustomResolveEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginCustomResolveEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginCustomResolveEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginCustomResolveEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginCustomResolveEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginCustomResolveEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginCustomResolveEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginCustomResolveEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginCustomResolveEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBeginCustomResolveEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBeginCustomResolveEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginCustomResolveEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginCustomResolveEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginCustomResolveEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginCustomResolveEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginCustomResolveEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginCustomResolveEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginCustomResolveEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginCustomResolveEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginCustomResolveEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginCustomResolveEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginCustomResolveEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginCustomResolveEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginCustomResolveEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginCustomResolveEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBeginCustomResolveEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBeginCustomResolveEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginCustomResolveEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginCustomResolveEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginCustomResolveEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginCustomResolveEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginCustomResolveEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginCustomResolveEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginCustomResolveEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginCustomResolveEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginCustomResolveEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginCustomResolveEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginCustomResolveEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginCustomResolveEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginCustomResolveEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginCustomResolveEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBeginCustomResolveEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBeginCustomResolveEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginCustomResolveEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginCustomResolveEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginCustomResolveEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginCustomResolveEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginCustomResolveEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginCustomResolveEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginCustomResolveEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginCustomResolveEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginCustomResolveEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginCustomResolveEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginCustomResolveEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginCustomResolveEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginCustomResolveEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginCustomResolveEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBeginCustomResolveEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBeginCustomResolveEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginCustomResolveEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginCustomResolveEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginCustomResolveEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginCustomResolveEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginCustomResolveEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginCustomResolveEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginCustomResolveEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginCustomResolveEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginCustomResolveEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginCustomResolveEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginCustomResolveEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginCustomResolveEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginCustomResolveEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginCustomResolveEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBeginCustomResolveEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBeginCustomResolveEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginCustomResolveEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginCustomResolveEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginCustomResolveEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginCustomResolveEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginCustomResolveEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginCustomResolveEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginCustomResolveEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginCustomResolveEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginCustomResolveEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginCustomResolveEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginCustomResolveEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginCustomResolveEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginCustomResolveEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginCustomResolveEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBeginCustomResolveEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBeginCustomResolveEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginCustomResolveEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginCustomResolveEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginCustomResolveEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginCustomResolveEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginCustomResolveEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginCustomResolveEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginCustomResolveEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginCustomResolveEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginCustomResolveEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginCustomResolveEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginCustomResolveEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginCustomResolveEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginCustomResolveEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginCustomResolveEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBeginCustomResolveEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBeginCustomResolveEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginCustomResolveEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginCustomResolveEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginCustomResolveEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginCustomResolveEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginCustomResolveEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginCustomResolveEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginCustomResolveEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginCustomResolveEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginCustomResolveEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginCustomResolveEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginCustomResolveEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginCustomResolveEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginCustomResolveEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginCustomResolveEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBeginCustomResolveEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBeginCustomResolveEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginCustomResolveEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginCustomResolveEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginCustomResolveEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginCustomResolveEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginCustomResolveEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginCustomResolveEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginCustomResolveEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginCustomResolveEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginCustomResolveEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginCustomResolveEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginCustomResolveEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginCustomResolveEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginCustomResolveEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginCustomResolveEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBeginCustomResolveEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBeginCustomResolveEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetQueryPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetQueryPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetQueryPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetQueryPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetQueryPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetQueryPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetQueryPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetQueryPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetQueryPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetQueryPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetQueryPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetQueryPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetQueryPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetQueryPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdResetQueryPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdResetQueryPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdWriteTimestamp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdWriteTimestamp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdWriteTimestamp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdWriteTimestamp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdWriteTimestamp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdWriteTimestamp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdWriteTimestamp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdWriteTimestamp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdWriteTimestamp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdWriteTimestamp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdWriteTimestamp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdWriteTimestamp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdWriteTimestamp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdWriteTimestamp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdWriteTimestamp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdWriteTimestamp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdWriteTimestamp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdWriteTimestamp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdWriteTimestamp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdWriteTimestamp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdWriteTimestamp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdWriteTimestamp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdWriteTimestamp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdWriteTimestamp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdWriteTimestamp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdWriteTimestamp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResults@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResults@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResults@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResults@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResults@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResults@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResults@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResults@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResults@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResults@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResults@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResults@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResults@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResults@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyQueryPoolResults=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyQueryPoolResults=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdPushConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdPushConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdPushConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdPushConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdPushConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdPushConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdPushConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdPushConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdPushConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdPushConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdPushConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdPushConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdPushConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdPushConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdPushConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdPushConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdPushConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdPushConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdPushConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdPushConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdPushConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdPushConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdPushConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdPushConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdPushConstants=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdPushConstants=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBeginRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBeginRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBeginRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBeginRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBeginRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBeginRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBeginRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBeginRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBeginRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBeginRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBeginRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBeginRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBeginRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBeginRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBeginRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBeginRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBeginRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBeginRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBeginRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBeginRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBeginRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBeginRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBeginRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBeginRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBeginRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBeginRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdNextSubpass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdNextSubpass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdNextSubpass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdNextSubpass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdNextSubpass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdNextSubpass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdNextSubpass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdNextSubpass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdNextSubpass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdNextSubpass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdNextSubpass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdNextSubpass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdNextSubpass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdNextSubpass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdNextSubpass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdNextSubpass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdNextSubpass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdNextSubpass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdNextSubpass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdNextSubpass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdNextSubpass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdNextSubpass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdNextSubpass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdNextSubpass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdNextSubpass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdNextSubpass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdEndRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdEndRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdEndRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdEndRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdEndRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdEndRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdEndRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdEndRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdEndRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdEndRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdEndRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdEndRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdEndRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdEndRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdEndRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdEndRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdEndRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdEndRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdEndRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdEndRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdEndRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdEndRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdEndRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdEndRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdEndRenderPass=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdEndRenderPass=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteCommands@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteCommands@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteCommands@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteCommands@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteCommands@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteCommands@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteCommands@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteCommands@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteCommands@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteCommands@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteCommands@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteCommands@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteCommands@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteCommands@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdExecuteCommands=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdExecuteCommands=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteCommands@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteCommands@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteCommands@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteCommands@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteCommands@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteCommands@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteCommands@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteCommands@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteCommands@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteCommands@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteCommands@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteCommands@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteCommands@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteCommands@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdExecuteCommands=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdExecuteCommands=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteCommands@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteCommands@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteCommands@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteCommands@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteCommands@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteCommands@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteCommands@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteCommands@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteCommands@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteCommands@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteCommands@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteCommands@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteCommands@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteCommands@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdExecuteCommands=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdExecuteCommands=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteCommands@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteCommands@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteCommands@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteCommands@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteCommands@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteCommands@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteCommands@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteCommands@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteCommands@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteCommands@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteCommands@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteCommands@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteCommands@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteCommands@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdExecuteCommands=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdExecuteCommands=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteCommands@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteCommands@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteCommands@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteCommands@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteCommands@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteCommands@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteCommands@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteCommands@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteCommands@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteCommands@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteCommands@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteCommands@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteCommands@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteCommands@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdExecuteCommands=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdExecuteCommands=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteCommands@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteCommands@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteCommands@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteCommands@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteCommands@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteCommands@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteCommands@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteCommands@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteCommands@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteCommands@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteCommands@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteCommands@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteCommands@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteCommands@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdExecuteCommands=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdExecuteCommands=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteCommands@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteCommands@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteCommands@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteCommands@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteCommands@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteCommands@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteCommands@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteCommands@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteCommands@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteCommands@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteCommands@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteCommands@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteCommands@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteCommands@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdExecuteCommands=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdExecuteCommands=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteCommands@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteCommands@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteCommands@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteCommands@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteCommands@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteCommands@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteCommands@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteCommands@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteCommands@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteCommands@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteCommands@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteCommands@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteCommands@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteCommands@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdExecuteCommands=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdExecuteCommands=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteCommands@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteCommands@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteCommands@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteCommands@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteCommands@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteCommands@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteCommands@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteCommands@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteCommands@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteCommands@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteCommands@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteCommands@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteCommands@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteCommands@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdExecuteCommands=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdExecuteCommands=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteCommands@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteCommands@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteCommands@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteCommands@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteCommands@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteCommands@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteCommands@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteCommands@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteCommands@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteCommands@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteCommands@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteCommands@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteCommands@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteCommands@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdExecuteCommands=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdExecuteCommands=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteCommands@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteCommands@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteCommands@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteCommands@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteCommands@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteCommands@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteCommands@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteCommands@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteCommands@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteCommands@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteCommands@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteCommands@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteCommands@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteCommands@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdExecuteCommands=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdExecuteCommands=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteCommands@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteCommands@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteCommands@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteCommands@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteCommands@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteCommands@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteCommands@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteCommands@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteCommands@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteCommands@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteCommands@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteCommands@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteCommands@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteCommands@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdExecuteCommands=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdExecuteCommands=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteCommands@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteCommands@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteCommands@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteCommands@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteCommands@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteCommands@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteCommands@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteCommands@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteCommands@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteCommands@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteCommands@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteCommands@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteCommands@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteCommands@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdExecuteCommands=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdExecuteCommands=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateSharedSwapchainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSharedSwapchainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSharedSwapchainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSharedSwapchainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSharedSwapchainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSharedSwapchainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSharedSwapchainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSharedSwapchainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSharedSwapchainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSharedSwapchainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSharedSwapchainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSharedSwapchainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSharedSwapchainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSharedSwapchainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateSharedSwapchainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateSharedSwapchainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateSharedSwapchainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSharedSwapchainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSharedSwapchainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSharedSwapchainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSharedSwapchainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSharedSwapchainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSharedSwapchainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSharedSwapchainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSharedSwapchainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSharedSwapchainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSharedSwapchainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSharedSwapchainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSharedSwapchainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSharedSwapchainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateSharedSwapchainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateSharedSwapchainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateSharedSwapchainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSharedSwapchainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSharedSwapchainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSharedSwapchainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSharedSwapchainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSharedSwapchainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSharedSwapchainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSharedSwapchainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSharedSwapchainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSharedSwapchainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSharedSwapchainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSharedSwapchainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSharedSwapchainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSharedSwapchainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateSharedSwapchainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateSharedSwapchainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateSharedSwapchainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSharedSwapchainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSharedSwapchainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSharedSwapchainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSharedSwapchainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSharedSwapchainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSharedSwapchainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSharedSwapchainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSharedSwapchainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSharedSwapchainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSharedSwapchainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSharedSwapchainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSharedSwapchainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSharedSwapchainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateSharedSwapchainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateSharedSwapchainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateSharedSwapchainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSharedSwapchainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSharedSwapchainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSharedSwapchainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSharedSwapchainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSharedSwapchainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSharedSwapchainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSharedSwapchainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSharedSwapchainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSharedSwapchainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSharedSwapchainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSharedSwapchainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSharedSwapchainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSharedSwapchainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateSharedSwapchainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateSharedSwapchainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateSharedSwapchainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSharedSwapchainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSharedSwapchainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSharedSwapchainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSharedSwapchainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSharedSwapchainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSharedSwapchainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSharedSwapchainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSharedSwapchainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSharedSwapchainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSharedSwapchainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSharedSwapchainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSharedSwapchainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSharedSwapchainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateSharedSwapchainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateSharedSwapchainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateSharedSwapchainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSharedSwapchainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSharedSwapchainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSharedSwapchainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSharedSwapchainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSharedSwapchainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSharedSwapchainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSharedSwapchainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSharedSwapchainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSharedSwapchainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSharedSwapchainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSharedSwapchainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSharedSwapchainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSharedSwapchainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateSharedSwapchainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateSharedSwapchainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateSharedSwapchainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSharedSwapchainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSharedSwapchainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSharedSwapchainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSharedSwapchainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSharedSwapchainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSharedSwapchainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSharedSwapchainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSharedSwapchainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSharedSwapchainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSharedSwapchainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSharedSwapchainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSharedSwapchainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSharedSwapchainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateSharedSwapchainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateSharedSwapchainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSharedSwapchainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSharedSwapchainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSharedSwapchainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSharedSwapchainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSharedSwapchainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSharedSwapchainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSharedSwapchainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSharedSwapchainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSharedSwapchainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSharedSwapchainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSharedSwapchainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSharedSwapchainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSharedSwapchainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSharedSwapchainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateSharedSwapchainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateSharedSwapchainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSharedSwapchainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSharedSwapchainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSharedSwapchainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSharedSwapchainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSharedSwapchainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSharedSwapchainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSharedSwapchainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSharedSwapchainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSharedSwapchainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSharedSwapchainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSharedSwapchainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSharedSwapchainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSharedSwapchainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSharedSwapchainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateSharedSwapchainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateSharedSwapchainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSharedSwapchainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSharedSwapchainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSharedSwapchainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSharedSwapchainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSharedSwapchainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSharedSwapchainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSharedSwapchainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSharedSwapchainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSharedSwapchainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSharedSwapchainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSharedSwapchainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSharedSwapchainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSharedSwapchainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSharedSwapchainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateSharedSwapchainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateSharedSwapchainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateSharedSwapchainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSharedSwapchainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSharedSwapchainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSharedSwapchainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSharedSwapchainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSharedSwapchainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSharedSwapchainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSharedSwapchainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSharedSwapchainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSharedSwapchainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSharedSwapchainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSharedSwapchainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSharedSwapchainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSharedSwapchainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateSharedSwapchainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateSharedSwapchainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSharedSwapchainsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSharedSwapchainsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSharedSwapchainsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSharedSwapchainsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSharedSwapchainsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSharedSwapchainsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSharedSwapchainsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSharedSwapchainsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSharedSwapchainsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSharedSwapchainsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSharedSwapchainsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSharedSwapchainsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSharedSwapchainsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSharedSwapchainsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateSharedSwapchainsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateSharedSwapchainsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateSwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateSwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateSwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateSwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateSwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateSwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateSwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateSwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateSwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateSwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateSwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateSwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateSwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateSwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateSwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateSwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateSwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateSwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateSwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateSwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateSwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateSwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateSwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateSwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateSwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateSwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateSwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateSwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateSwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateSwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateSwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateSwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateSwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateSwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateSwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroySwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroySwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroySwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroySwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroySwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroySwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroySwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroySwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroySwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroySwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroySwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroySwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroySwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroySwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroySwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroySwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroySwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroySwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroySwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroySwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroySwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroySwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroySwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroySwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroySwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroySwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroySwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroySwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroySwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroySwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroySwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroySwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroySwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySwapchainKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySwapchainKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySwapchainKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySwapchainKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySwapchainKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySwapchainKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySwapchainKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySwapchainKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySwapchainKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySwapchainKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySwapchainKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySwapchainKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySwapchainKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySwapchainKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroySwapchainKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroySwapchainKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_AcquireNextImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_AcquireNextImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_AcquireNextImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_AcquireNextImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_AcquireNextImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_AcquireNextImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_AcquireNextImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_AcquireNextImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_AcquireNextImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_AcquireNextImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_AcquireNextImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_AcquireNextImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_AcquireNextImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_AcquireNextImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_AcquireNextImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_AcquireNextImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_AcquireNextImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_AcquireNextImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_AcquireNextImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_AcquireNextImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_AcquireNextImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_AcquireNextImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_AcquireNextImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_AcquireNextImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_AcquireNextImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_AcquireNextImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_QueuePresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueuePresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueuePresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueuePresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueuePresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueuePresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueuePresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueuePresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueuePresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueuePresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueuePresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueuePresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueuePresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueuePresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_QueuePresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_QueuePresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_QueuePresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueuePresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueuePresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueuePresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueuePresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueuePresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueuePresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueuePresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueuePresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueuePresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueuePresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueuePresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueuePresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueuePresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_QueuePresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_QueuePresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_QueuePresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueuePresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueuePresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueuePresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueuePresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueuePresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueuePresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueuePresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueuePresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueuePresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueuePresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueuePresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueuePresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueuePresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_QueuePresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_QueuePresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_QueuePresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueuePresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueuePresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueuePresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueuePresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueuePresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueuePresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueuePresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueuePresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueuePresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueuePresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueuePresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueuePresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueuePresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_QueuePresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_QueuePresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_QueuePresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueuePresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueuePresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueuePresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueuePresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueuePresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueuePresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueuePresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueuePresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueuePresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueuePresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueuePresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueuePresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueuePresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_QueuePresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_QueuePresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_QueuePresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueuePresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueuePresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueuePresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueuePresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueuePresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueuePresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueuePresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueuePresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueuePresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueuePresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueuePresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueuePresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueuePresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_QueuePresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_QueuePresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_QueuePresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueuePresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueuePresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueuePresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueuePresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueuePresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueuePresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueuePresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueuePresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueuePresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueuePresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueuePresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueuePresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueuePresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_QueuePresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_QueuePresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_QueuePresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueuePresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueuePresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueuePresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueuePresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueuePresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueuePresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueuePresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueuePresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueuePresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueuePresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueuePresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueuePresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueuePresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_QueuePresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_QueuePresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_QueuePresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueuePresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueuePresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueuePresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueuePresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueuePresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueuePresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueuePresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueuePresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueuePresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueuePresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueuePresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueuePresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueuePresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_QueuePresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_QueuePresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_QueuePresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueuePresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueuePresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueuePresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueuePresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueuePresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueuePresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueuePresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueuePresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueuePresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueuePresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueuePresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueuePresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueuePresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_QueuePresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_QueuePresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueuePresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueuePresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueuePresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueuePresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueuePresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueuePresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueuePresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueuePresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueuePresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueuePresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueuePresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueuePresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueuePresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueuePresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_QueuePresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_QueuePresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_QueuePresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueuePresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueuePresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueuePresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueuePresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueuePresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueuePresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueuePresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueuePresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueuePresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueuePresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueuePresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueuePresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueuePresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_QueuePresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_QueuePresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_QueuePresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueuePresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueuePresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueuePresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueuePresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueuePresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueuePresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueuePresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueuePresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueuePresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueuePresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueuePresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueuePresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueuePresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_QueuePresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_QueuePresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DebugMarkerSetObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DebugMarkerSetObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DebugMarkerSetObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DebugMarkerSetObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DebugMarkerSetObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DebugMarkerSetObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DebugMarkerSetObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DebugMarkerSetObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DebugMarkerSetObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DebugMarkerSetObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DebugMarkerSetObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DebugMarkerSetObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DebugMarkerSetObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DebugMarkerSetObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DebugMarkerSetObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DebugMarkerSetObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DebugMarkerSetObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DebugMarkerSetObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DebugMarkerSetObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DebugMarkerSetObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DebugMarkerSetObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DebugMarkerSetObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DebugMarkerSetObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DebugMarkerSetObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DebugMarkerSetObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DebugMarkerSetObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DebugMarkerSetObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DebugMarkerSetObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DebugMarkerSetObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DebugMarkerSetObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DebugMarkerSetObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DebugMarkerSetObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DebugMarkerSetObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DebugMarkerSetObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DebugMarkerSetObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DebugMarkerSetObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DebugMarkerSetObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DebugMarkerSetObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DebugMarkerSetObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DebugMarkerSetObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DebugMarkerSetObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DebugMarkerSetObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DebugMarkerSetObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DebugMarkerSetObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DebugMarkerSetObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DebugMarkerSetObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DebugMarkerSetObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DebugMarkerSetObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DebugMarkerSetObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DebugMarkerSetObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DebugMarkerSetObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DebugMarkerSetObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DebugMarkerSetObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DebugMarkerSetObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DebugMarkerSetObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DebugMarkerSetObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DebugMarkerSetObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DebugMarkerSetObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DebugMarkerSetObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DebugMarkerSetObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DebugMarkerSetObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DebugMarkerSetObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DebugMarkerSetObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DebugMarkerSetObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DebugMarkerSetObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerBeginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerBeginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerBeginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerBeginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerBeginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerBeginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerBeginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerBeginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerBeginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerBeginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerBeginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerBeginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerBeginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerBeginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDebugMarkerBeginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDebugMarkerBeginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerBeginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerBeginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerBeginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerBeginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerBeginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerBeginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerBeginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerBeginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerBeginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerBeginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerBeginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerBeginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerBeginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerBeginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDebugMarkerBeginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDebugMarkerBeginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerBeginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerBeginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerBeginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerBeginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerBeginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerBeginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerBeginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerBeginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerBeginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerBeginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerBeginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerBeginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerBeginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerBeginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDebugMarkerBeginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDebugMarkerBeginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerBeginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerBeginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerBeginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerBeginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerBeginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerBeginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerBeginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerBeginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerBeginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerBeginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerBeginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerBeginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerBeginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerBeginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDebugMarkerBeginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDebugMarkerBeginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerBeginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerBeginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerBeginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerBeginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerBeginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerBeginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerBeginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerBeginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerBeginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerBeginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerBeginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerBeginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerBeginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerBeginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDebugMarkerBeginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDebugMarkerBeginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerBeginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerBeginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerBeginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerBeginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerBeginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerBeginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerBeginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerBeginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerBeginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerBeginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerBeginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerBeginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerBeginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerBeginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDebugMarkerBeginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDebugMarkerBeginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerBeginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerBeginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerBeginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerBeginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerBeginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerBeginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerBeginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerBeginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerBeginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerBeginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerBeginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerBeginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerBeginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerBeginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDebugMarkerBeginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDebugMarkerBeginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerBeginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerBeginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerBeginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerBeginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerBeginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerBeginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerBeginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerBeginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerBeginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerBeginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerBeginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerBeginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerBeginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerBeginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDebugMarkerBeginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDebugMarkerBeginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerBeginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerBeginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerBeginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerBeginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerBeginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerBeginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerBeginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerBeginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerBeginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerBeginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerBeginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerBeginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerBeginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerBeginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDebugMarkerBeginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDebugMarkerBeginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerBeginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerBeginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerBeginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerBeginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerBeginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerBeginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerBeginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerBeginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerBeginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerBeginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerBeginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerBeginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerBeginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerBeginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDebugMarkerBeginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDebugMarkerBeginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerBeginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerBeginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerBeginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerBeginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerBeginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerBeginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerBeginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerBeginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerBeginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerBeginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerBeginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerBeginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerBeginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerBeginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDebugMarkerBeginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDebugMarkerBeginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerBeginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerBeginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerBeginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerBeginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerBeginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerBeginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerBeginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerBeginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerBeginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerBeginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerBeginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerBeginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerBeginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerBeginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDebugMarkerBeginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDebugMarkerBeginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerBeginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerBeginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerBeginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerBeginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerBeginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerBeginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerBeginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerBeginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerBeginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerBeginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerBeginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerBeginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerBeginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerBeginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDebugMarkerBeginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDebugMarkerBeginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerEndEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerEndEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerEndEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerEndEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerEndEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerEndEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerEndEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerEndEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerEndEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerEndEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerEndEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerEndEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerEndEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerEndEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDebugMarkerEndEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDebugMarkerEndEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerEndEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerEndEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerEndEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerEndEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerEndEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerEndEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerEndEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerEndEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerEndEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerEndEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerEndEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerEndEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerEndEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerEndEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDebugMarkerEndEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDebugMarkerEndEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerEndEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerEndEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerEndEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerEndEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerEndEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerEndEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerEndEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerEndEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerEndEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerEndEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerEndEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerEndEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerEndEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerEndEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDebugMarkerEndEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDebugMarkerEndEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerEndEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerEndEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerEndEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerEndEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerEndEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerEndEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerEndEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerEndEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerEndEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerEndEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerEndEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerEndEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerEndEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerEndEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDebugMarkerEndEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDebugMarkerEndEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerEndEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerEndEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerEndEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerEndEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerEndEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerEndEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerEndEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerEndEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerEndEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerEndEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerEndEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerEndEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerEndEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerEndEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDebugMarkerEndEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDebugMarkerEndEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerEndEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerEndEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerEndEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerEndEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerEndEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerEndEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerEndEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerEndEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerEndEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerEndEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerEndEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerEndEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerEndEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerEndEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDebugMarkerEndEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDebugMarkerEndEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerEndEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerEndEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerEndEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerEndEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerEndEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerEndEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerEndEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerEndEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerEndEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerEndEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerEndEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerEndEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerEndEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerEndEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDebugMarkerEndEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDebugMarkerEndEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerEndEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerEndEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerEndEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerEndEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerEndEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerEndEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerEndEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerEndEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerEndEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerEndEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerEndEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerEndEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerEndEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerEndEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDebugMarkerEndEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDebugMarkerEndEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerEndEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerEndEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerEndEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerEndEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerEndEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerEndEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerEndEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerEndEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerEndEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerEndEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerEndEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerEndEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerEndEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerEndEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDebugMarkerEndEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDebugMarkerEndEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerEndEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerEndEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerEndEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerEndEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerEndEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerEndEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerEndEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerEndEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerEndEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerEndEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerEndEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerEndEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerEndEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerEndEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDebugMarkerEndEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDebugMarkerEndEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerEndEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerEndEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerEndEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerEndEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerEndEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerEndEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerEndEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerEndEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerEndEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerEndEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerEndEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerEndEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerEndEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerEndEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDebugMarkerEndEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDebugMarkerEndEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerEndEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerEndEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerEndEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerEndEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerEndEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerEndEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerEndEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerEndEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerEndEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerEndEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerEndEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerEndEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerEndEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerEndEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDebugMarkerEndEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDebugMarkerEndEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerEndEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerEndEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerEndEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerEndEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerEndEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerEndEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerEndEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerEndEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerEndEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerEndEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerEndEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerEndEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerEndEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerEndEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDebugMarkerEndEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDebugMarkerEndEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerInsertEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerInsertEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerInsertEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerInsertEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerInsertEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerInsertEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerInsertEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerInsertEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerInsertEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerInsertEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerInsertEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerInsertEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerInsertEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDebugMarkerInsertEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDebugMarkerInsertEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDebugMarkerInsertEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerInsertEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerInsertEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerInsertEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerInsertEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerInsertEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerInsertEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerInsertEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerInsertEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerInsertEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerInsertEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerInsertEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerInsertEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerInsertEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDebugMarkerInsertEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDebugMarkerInsertEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDebugMarkerInsertEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerInsertEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerInsertEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerInsertEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerInsertEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerInsertEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerInsertEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerInsertEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerInsertEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerInsertEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerInsertEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerInsertEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerInsertEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerInsertEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDebugMarkerInsertEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDebugMarkerInsertEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDebugMarkerInsertEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerInsertEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerInsertEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerInsertEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerInsertEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerInsertEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerInsertEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerInsertEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerInsertEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerInsertEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerInsertEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerInsertEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerInsertEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerInsertEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDebugMarkerInsertEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDebugMarkerInsertEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDebugMarkerInsertEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerInsertEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerInsertEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerInsertEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerInsertEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerInsertEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerInsertEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerInsertEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerInsertEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerInsertEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerInsertEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerInsertEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerInsertEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerInsertEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDebugMarkerInsertEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDebugMarkerInsertEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDebugMarkerInsertEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerInsertEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerInsertEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerInsertEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerInsertEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerInsertEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerInsertEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerInsertEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerInsertEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerInsertEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerInsertEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerInsertEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerInsertEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerInsertEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDebugMarkerInsertEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDebugMarkerInsertEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDebugMarkerInsertEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerInsertEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerInsertEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerInsertEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerInsertEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerInsertEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerInsertEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerInsertEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerInsertEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerInsertEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerInsertEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerInsertEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerInsertEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerInsertEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDebugMarkerInsertEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDebugMarkerInsertEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDebugMarkerInsertEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerInsertEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerInsertEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerInsertEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerInsertEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerInsertEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerInsertEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerInsertEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerInsertEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerInsertEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerInsertEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerInsertEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerInsertEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerInsertEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDebugMarkerInsertEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDebugMarkerInsertEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDebugMarkerInsertEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerInsertEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerInsertEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerInsertEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerInsertEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerInsertEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerInsertEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerInsertEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerInsertEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerInsertEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerInsertEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerInsertEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerInsertEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerInsertEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDebugMarkerInsertEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDebugMarkerInsertEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDebugMarkerInsertEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerInsertEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerInsertEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerInsertEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerInsertEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerInsertEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerInsertEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerInsertEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerInsertEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerInsertEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerInsertEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerInsertEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerInsertEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerInsertEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDebugMarkerInsertEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDebugMarkerInsertEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDebugMarkerInsertEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerInsertEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerInsertEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerInsertEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerInsertEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerInsertEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerInsertEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerInsertEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerInsertEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerInsertEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerInsertEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerInsertEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerInsertEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerInsertEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDebugMarkerInsertEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDebugMarkerInsertEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDebugMarkerInsertEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerInsertEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerInsertEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerInsertEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerInsertEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerInsertEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerInsertEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerInsertEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerInsertEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerInsertEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerInsertEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerInsertEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerInsertEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerInsertEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDebugMarkerInsertEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDebugMarkerInsertEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDebugMarkerInsertEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerInsertEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerInsertEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerInsertEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerInsertEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerInsertEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerInsertEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerInsertEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerInsertEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerInsertEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerInsertEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerInsertEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerInsertEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerInsertEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDebugMarkerInsertEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDebugMarkerInsertEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDebugMarkerInsertEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetMemoryWin32HandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetMemoryWin32HandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetMemoryWin32HandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetMemoryWin32HandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetMemoryWin32HandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetMemoryWin32HandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetMemoryWin32HandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetMemoryWin32HandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetMemoryWin32HandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetMemoryWin32HandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetMemoryWin32HandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetMemoryWin32HandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetMemoryWin32HandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetMemoryWin32HandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetMemoryWin32HandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetMemoryWin32HandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetMemoryWin32HandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetMemoryWin32HandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetMemoryWin32HandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetMemoryWin32HandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetMemoryWin32HandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetMemoryWin32HandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetMemoryWin32HandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetMemoryWin32HandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetMemoryWin32HandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetMemoryWin32HandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdExecuteGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdExecuteGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdExecuteGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdExecuteGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdExecuteGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdExecuteGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdExecuteGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdExecuteGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdExecuteGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdExecuteGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdExecuteGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdExecuteGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdExecuteGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdExecuteGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdExecuteGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdExecuteGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdExecuteGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdExecuteGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdExecuteGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdExecuteGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdExecuteGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdExecuteGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdExecuteGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdExecuteGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdExecuteGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdExecuteGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdPreprocessGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdPreprocessGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdPreprocessGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdPreprocessGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdPreprocessGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdPreprocessGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdPreprocessGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdPreprocessGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdPreprocessGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdPreprocessGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdPreprocessGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdPreprocessGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdPreprocessGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdPreprocessGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdPreprocessGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdPreprocessGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdPreprocessGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdPreprocessGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdPreprocessGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdPreprocessGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdPreprocessGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdPreprocessGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdPreprocessGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdPreprocessGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdPreprocessGeneratedCommandsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdPreprocessGeneratedCommandsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipelineShaderGroupNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipelineShaderGroupNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipelineShaderGroupNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipelineShaderGroupNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipelineShaderGroupNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipelineShaderGroupNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipelineShaderGroupNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipelineShaderGroupNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipelineShaderGroupNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipelineShaderGroupNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipelineShaderGroupNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipelineShaderGroupNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipelineShaderGroupNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindPipelineShaderGroupNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindPipelineShaderGroupNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindPipelineShaderGroupNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipelineShaderGroupNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipelineShaderGroupNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipelineShaderGroupNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipelineShaderGroupNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipelineShaderGroupNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipelineShaderGroupNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipelineShaderGroupNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipelineShaderGroupNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipelineShaderGroupNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipelineShaderGroupNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipelineShaderGroupNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipelineShaderGroupNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipelineShaderGroupNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindPipelineShaderGroupNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindPipelineShaderGroupNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindPipelineShaderGroupNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipelineShaderGroupNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipelineShaderGroupNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipelineShaderGroupNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipelineShaderGroupNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipelineShaderGroupNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipelineShaderGroupNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipelineShaderGroupNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipelineShaderGroupNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipelineShaderGroupNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipelineShaderGroupNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipelineShaderGroupNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipelineShaderGroupNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipelineShaderGroupNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindPipelineShaderGroupNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindPipelineShaderGroupNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindPipelineShaderGroupNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipelineShaderGroupNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipelineShaderGroupNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipelineShaderGroupNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipelineShaderGroupNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipelineShaderGroupNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipelineShaderGroupNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipelineShaderGroupNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipelineShaderGroupNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipelineShaderGroupNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipelineShaderGroupNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipelineShaderGroupNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipelineShaderGroupNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipelineShaderGroupNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindPipelineShaderGroupNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindPipelineShaderGroupNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindPipelineShaderGroupNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipelineShaderGroupNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipelineShaderGroupNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipelineShaderGroupNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipelineShaderGroupNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipelineShaderGroupNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipelineShaderGroupNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipelineShaderGroupNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipelineShaderGroupNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipelineShaderGroupNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipelineShaderGroupNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipelineShaderGroupNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipelineShaderGroupNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipelineShaderGroupNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindPipelineShaderGroupNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindPipelineShaderGroupNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindPipelineShaderGroupNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipelineShaderGroupNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipelineShaderGroupNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipelineShaderGroupNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipelineShaderGroupNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipelineShaderGroupNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipelineShaderGroupNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipelineShaderGroupNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipelineShaderGroupNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipelineShaderGroupNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipelineShaderGroupNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipelineShaderGroupNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipelineShaderGroupNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipelineShaderGroupNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindPipelineShaderGroupNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindPipelineShaderGroupNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindPipelineShaderGroupNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipelineShaderGroupNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipelineShaderGroupNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipelineShaderGroupNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipelineShaderGroupNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipelineShaderGroupNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipelineShaderGroupNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipelineShaderGroupNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipelineShaderGroupNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipelineShaderGroupNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipelineShaderGroupNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipelineShaderGroupNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipelineShaderGroupNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipelineShaderGroupNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindPipelineShaderGroupNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindPipelineShaderGroupNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindPipelineShaderGroupNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipelineShaderGroupNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipelineShaderGroupNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipelineShaderGroupNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipelineShaderGroupNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipelineShaderGroupNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipelineShaderGroupNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipelineShaderGroupNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipelineShaderGroupNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipelineShaderGroupNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipelineShaderGroupNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipelineShaderGroupNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipelineShaderGroupNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipelineShaderGroupNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindPipelineShaderGroupNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindPipelineShaderGroupNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindPipelineShaderGroupNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipelineShaderGroupNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipelineShaderGroupNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipelineShaderGroupNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipelineShaderGroupNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipelineShaderGroupNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipelineShaderGroupNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipelineShaderGroupNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipelineShaderGroupNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipelineShaderGroupNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipelineShaderGroupNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipelineShaderGroupNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipelineShaderGroupNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipelineShaderGroupNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindPipelineShaderGroupNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindPipelineShaderGroupNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindPipelineShaderGroupNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipelineShaderGroupNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipelineShaderGroupNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipelineShaderGroupNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipelineShaderGroupNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipelineShaderGroupNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipelineShaderGroupNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipelineShaderGroupNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipelineShaderGroupNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipelineShaderGroupNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipelineShaderGroupNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipelineShaderGroupNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipelineShaderGroupNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipelineShaderGroupNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindPipelineShaderGroupNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindPipelineShaderGroupNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindPipelineShaderGroupNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipelineShaderGroupNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipelineShaderGroupNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipelineShaderGroupNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipelineShaderGroupNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipelineShaderGroupNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipelineShaderGroupNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipelineShaderGroupNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipelineShaderGroupNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipelineShaderGroupNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipelineShaderGroupNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipelineShaderGroupNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipelineShaderGroupNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipelineShaderGroupNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindPipelineShaderGroupNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindPipelineShaderGroupNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindPipelineShaderGroupNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipelineShaderGroupNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipelineShaderGroupNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipelineShaderGroupNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipelineShaderGroupNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipelineShaderGroupNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipelineShaderGroupNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipelineShaderGroupNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipelineShaderGroupNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipelineShaderGroupNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipelineShaderGroupNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipelineShaderGroupNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipelineShaderGroupNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipelineShaderGroupNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindPipelineShaderGroupNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindPipelineShaderGroupNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindPipelineShaderGroupNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipelineShaderGroupNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipelineShaderGroupNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipelineShaderGroupNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipelineShaderGroupNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipelineShaderGroupNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipelineShaderGroupNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipelineShaderGroupNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipelineShaderGroupNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipelineShaderGroupNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipelineShaderGroupNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipelineShaderGroupNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipelineShaderGroupNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipelineShaderGroupNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindPipelineShaderGroupNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindPipelineShaderGroupNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindPipelineShaderGroupNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetGeneratedCommandsMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetGeneratedCommandsMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetGeneratedCommandsMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetGeneratedCommandsMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetGeneratedCommandsMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetGeneratedCommandsMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetGeneratedCommandsMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetGeneratedCommandsMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetGeneratedCommandsMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetGeneratedCommandsMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetGeneratedCommandsMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetGeneratedCommandsMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetGeneratedCommandsMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetGeneratedCommandsMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetGeneratedCommandsMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetGeneratedCommandsMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetGeneratedCommandsMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetGeneratedCommandsMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetGeneratedCommandsMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetGeneratedCommandsMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetGeneratedCommandsMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetGeneratedCommandsMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetGeneratedCommandsMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetGeneratedCommandsMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetGeneratedCommandsMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetGeneratedCommandsMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyIndirectCommandsLayoutNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyIndirectCommandsLayoutNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdExecuteGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdExecuteGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdExecuteGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdExecuteGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdExecuteGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdExecuteGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdExecuteGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdExecuteGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdExecuteGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdExecuteGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdExecuteGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdExecuteGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdExecuteGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdExecuteGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdExecuteGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdExecuteGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdExecuteGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdExecuteGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdExecuteGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdExecuteGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdExecuteGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdExecuteGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdExecuteGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdExecuteGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdExecuteGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdExecuteGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdExecuteGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdExecuteGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdExecuteGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdExecuteGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdExecuteGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdExecuteGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdExecuteGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdExecuteGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdExecuteGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdExecuteGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdExecuteGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdExecuteGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdExecuteGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPreprocessGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdPreprocessGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdPreprocessGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPreprocessGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdPreprocessGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdPreprocessGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPreprocessGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdPreprocessGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdPreprocessGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPreprocessGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdPreprocessGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdPreprocessGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPreprocessGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdPreprocessGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdPreprocessGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPreprocessGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdPreprocessGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdPreprocessGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPreprocessGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdPreprocessGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdPreprocessGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPreprocessGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdPreprocessGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdPreprocessGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPreprocessGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdPreprocessGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdPreprocessGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPreprocessGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdPreprocessGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdPreprocessGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPreprocessGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdPreprocessGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdPreprocessGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPreprocessGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdPreprocessGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdPreprocessGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPreprocessGeneratedCommandsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdPreprocessGeneratedCommandsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdPreprocessGeneratedCommandsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetGeneratedCommandsMemoryRequirementsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetGeneratedCommandsMemoryRequirementsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetGeneratedCommandsMemoryRequirementsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetGeneratedCommandsMemoryRequirementsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetGeneratedCommandsMemoryRequirementsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetGeneratedCommandsMemoryRequirementsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetGeneratedCommandsMemoryRequirementsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetGeneratedCommandsMemoryRequirementsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetGeneratedCommandsMemoryRequirementsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetGeneratedCommandsMemoryRequirementsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetGeneratedCommandsMemoryRequirementsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetGeneratedCommandsMemoryRequirementsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetGeneratedCommandsMemoryRequirementsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetGeneratedCommandsMemoryRequirementsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetGeneratedCommandsMemoryRequirementsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetGeneratedCommandsMemoryRequirementsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetGeneratedCommandsMemoryRequirementsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetGeneratedCommandsMemoryRequirementsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetGeneratedCommandsMemoryRequirementsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetGeneratedCommandsMemoryRequirementsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetGeneratedCommandsMemoryRequirementsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetGeneratedCommandsMemoryRequirementsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetGeneratedCommandsMemoryRequirementsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetGeneratedCommandsMemoryRequirementsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetGeneratedCommandsMemoryRequirementsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetGeneratedCommandsMemoryRequirementsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetGeneratedCommandsMemoryRequirementsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetGeneratedCommandsMemoryRequirementsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetGeneratedCommandsMemoryRequirementsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetGeneratedCommandsMemoryRequirementsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetGeneratedCommandsMemoryRequirementsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetGeneratedCommandsMemoryRequirementsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetGeneratedCommandsMemoryRequirementsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetGeneratedCommandsMemoryRequirementsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetGeneratedCommandsMemoryRequirementsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetGeneratedCommandsMemoryRequirementsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetGeneratedCommandsMemoryRequirementsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetGeneratedCommandsMemoryRequirementsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetGeneratedCommandsMemoryRequirementsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectCommandsLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyIndirectCommandsLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyIndirectCommandsLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectExecutionSetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectExecutionSetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectExecutionSetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectExecutionSetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectExecutionSetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectExecutionSetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectExecutionSetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectExecutionSetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectExecutionSetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectExecutionSetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectExecutionSetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectExecutionSetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectExecutionSetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyIndirectExecutionSetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyIndirectExecutionSetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyIndirectExecutionSetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetPipelineEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetPipelineEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetPipelineEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetPipelineEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetPipelineEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetPipelineEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetPipelineEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetPipelineEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetPipelineEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetPipelineEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetPipelineEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetPipelineEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetPipelineEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetPipelineEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_UpdateIndirectExecutionSetPipelineEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_UpdateIndirectExecutionSetPipelineEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetPipelineEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetPipelineEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetPipelineEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetPipelineEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetPipelineEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetPipelineEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetPipelineEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetPipelineEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetPipelineEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetPipelineEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetPipelineEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetPipelineEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetPipelineEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetPipelineEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_UpdateIndirectExecutionSetPipelineEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_UpdateIndirectExecutionSetPipelineEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetPipelineEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetPipelineEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetPipelineEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetPipelineEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetPipelineEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetPipelineEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetPipelineEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetPipelineEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetPipelineEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetPipelineEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetPipelineEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetPipelineEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetPipelineEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetPipelineEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_UpdateIndirectExecutionSetPipelineEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_UpdateIndirectExecutionSetPipelineEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetPipelineEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetPipelineEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetPipelineEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetPipelineEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetPipelineEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetPipelineEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetPipelineEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetPipelineEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetPipelineEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetPipelineEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetPipelineEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetPipelineEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetPipelineEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetPipelineEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_UpdateIndirectExecutionSetPipelineEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_UpdateIndirectExecutionSetPipelineEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetPipelineEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetPipelineEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetPipelineEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetPipelineEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetPipelineEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetPipelineEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetPipelineEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetPipelineEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetPipelineEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetPipelineEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetPipelineEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetPipelineEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetPipelineEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetPipelineEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_UpdateIndirectExecutionSetPipelineEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_UpdateIndirectExecutionSetPipelineEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetPipelineEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetPipelineEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetPipelineEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetPipelineEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetPipelineEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetPipelineEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetPipelineEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetPipelineEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetPipelineEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetPipelineEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetPipelineEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetPipelineEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetPipelineEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetPipelineEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_UpdateIndirectExecutionSetPipelineEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_UpdateIndirectExecutionSetPipelineEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetPipelineEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetPipelineEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetPipelineEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetPipelineEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetPipelineEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetPipelineEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetPipelineEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetPipelineEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetPipelineEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetPipelineEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetPipelineEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetPipelineEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetPipelineEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetPipelineEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_UpdateIndirectExecutionSetPipelineEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_UpdateIndirectExecutionSetPipelineEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetPipelineEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetPipelineEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetPipelineEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetPipelineEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetPipelineEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetPipelineEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetPipelineEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetPipelineEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetPipelineEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetPipelineEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetPipelineEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetPipelineEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetPipelineEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetPipelineEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_UpdateIndirectExecutionSetPipelineEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_UpdateIndirectExecutionSetPipelineEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetPipelineEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetPipelineEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetPipelineEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetPipelineEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetPipelineEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetPipelineEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetPipelineEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetPipelineEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetPipelineEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetPipelineEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetPipelineEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetPipelineEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetPipelineEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetPipelineEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_UpdateIndirectExecutionSetPipelineEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_UpdateIndirectExecutionSetPipelineEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetPipelineEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetPipelineEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetPipelineEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetPipelineEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetPipelineEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetPipelineEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetPipelineEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetPipelineEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetPipelineEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetPipelineEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetPipelineEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetPipelineEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetPipelineEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetPipelineEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_UpdateIndirectExecutionSetPipelineEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_UpdateIndirectExecutionSetPipelineEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetPipelineEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetPipelineEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetPipelineEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetPipelineEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetPipelineEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetPipelineEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetPipelineEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetPipelineEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetPipelineEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetPipelineEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetPipelineEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetPipelineEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetPipelineEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetPipelineEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_UpdateIndirectExecutionSetPipelineEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_UpdateIndirectExecutionSetPipelineEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetPipelineEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetPipelineEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetPipelineEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetPipelineEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetPipelineEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetPipelineEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetPipelineEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetPipelineEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetPipelineEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetPipelineEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetPipelineEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetPipelineEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetPipelineEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetPipelineEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_UpdateIndirectExecutionSetPipelineEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_UpdateIndirectExecutionSetPipelineEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetPipelineEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetPipelineEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetPipelineEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetPipelineEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetPipelineEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetPipelineEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetPipelineEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetPipelineEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetPipelineEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetPipelineEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetPipelineEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetPipelineEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetPipelineEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetPipelineEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_UpdateIndirectExecutionSetPipelineEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_UpdateIndirectExecutionSetPipelineEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateIndirectExecutionSetShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_UpdateIndirectExecutionSetShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_UpdateIndirectExecutionSetShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateIndirectExecutionSetShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_UpdateIndirectExecutionSetShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_UpdateIndirectExecutionSetShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateIndirectExecutionSetShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_UpdateIndirectExecutionSetShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_UpdateIndirectExecutionSetShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateIndirectExecutionSetShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_UpdateIndirectExecutionSetShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_UpdateIndirectExecutionSetShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateIndirectExecutionSetShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_UpdateIndirectExecutionSetShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_UpdateIndirectExecutionSetShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateIndirectExecutionSetShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_UpdateIndirectExecutionSetShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_UpdateIndirectExecutionSetShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateIndirectExecutionSetShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_UpdateIndirectExecutionSetShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_UpdateIndirectExecutionSetShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateIndirectExecutionSetShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_UpdateIndirectExecutionSetShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_UpdateIndirectExecutionSetShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateIndirectExecutionSetShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_UpdateIndirectExecutionSetShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_UpdateIndirectExecutionSetShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateIndirectExecutionSetShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_UpdateIndirectExecutionSetShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_UpdateIndirectExecutionSetShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateIndirectExecutionSetShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_UpdateIndirectExecutionSetShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_UpdateIndirectExecutionSetShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateIndirectExecutionSetShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_UpdateIndirectExecutionSetShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_UpdateIndirectExecutionSetShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateIndirectExecutionSetShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_UpdateIndirectExecutionSetShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_UpdateIndirectExecutionSetShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdPushDescriptorSet=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdPushDescriptorSet=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdPushDescriptorSet=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdPushDescriptorSet=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdPushDescriptorSet=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdPushDescriptorSet=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdPushDescriptorSet=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdPushDescriptorSet=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdPushDescriptorSet=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdPushDescriptorSet=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdPushDescriptorSet=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdPushDescriptorSet=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdPushDescriptorSet=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdPushDescriptorSet=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdPushDescriptorSet=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdPushDescriptorSet=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdPushDescriptorSet=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdPushDescriptorSet=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdPushDescriptorSet=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdPushDescriptorSet=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdPushDescriptorSet=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdPushDescriptorSet=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdPushDescriptorSet=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdPushDescriptorSet=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdPushDescriptorSet=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdPushDescriptorSet=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdPushDescriptorSetKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdPushDescriptorSetKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdPushDescriptorSetKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdPushDescriptorSetKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdPushDescriptorSetKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdPushDescriptorSetKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdPushDescriptorSetKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdPushDescriptorSetKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdPushDescriptorSetKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdPushDescriptorSetKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdPushDescriptorSetKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdPushDescriptorSetKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdPushDescriptorSetKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdPushDescriptorSetKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdPushDescriptorSetKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdPushDescriptorSetKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdPushDescriptorSetKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdPushDescriptorSetKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdPushDescriptorSetKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdPushDescriptorSetKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdPushDescriptorSetKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdPushDescriptorSetKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdPushDescriptorSetKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdPushDescriptorSetKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdPushDescriptorSetKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdPushDescriptorSetKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_TrimCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_TrimCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_TrimCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_TrimCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_TrimCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_TrimCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_TrimCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_TrimCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_TrimCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_TrimCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_TrimCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_TrimCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_TrimCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_TrimCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_TrimCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_TrimCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_TrimCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_TrimCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_TrimCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_TrimCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_TrimCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_TrimCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_TrimCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_TrimCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPool@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPool@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPool@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPool@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPool@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPool@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPool@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPool@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPool@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPool@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPool@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPool@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPool@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPool@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_TrimCommandPool=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_TrimCommandPool=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPoolKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPoolKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPoolKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPoolKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPoolKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPoolKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPoolKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPoolKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPoolKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPoolKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPoolKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPoolKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPoolKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TrimCommandPoolKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_TrimCommandPoolKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_TrimCommandPoolKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPoolKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPoolKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPoolKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPoolKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPoolKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPoolKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPoolKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPoolKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPoolKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPoolKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPoolKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPoolKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPoolKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TrimCommandPoolKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_TrimCommandPoolKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_TrimCommandPoolKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPoolKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPoolKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPoolKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPoolKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPoolKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPoolKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPoolKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPoolKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPoolKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPoolKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPoolKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPoolKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPoolKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TrimCommandPoolKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_TrimCommandPoolKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_TrimCommandPoolKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPoolKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPoolKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPoolKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPoolKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPoolKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPoolKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPoolKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPoolKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPoolKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPoolKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPoolKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPoolKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPoolKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TrimCommandPoolKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_TrimCommandPoolKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_TrimCommandPoolKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPoolKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPoolKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPoolKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPoolKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPoolKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPoolKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPoolKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPoolKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPoolKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPoolKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPoolKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPoolKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPoolKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TrimCommandPoolKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_TrimCommandPoolKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_TrimCommandPoolKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPoolKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPoolKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPoolKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPoolKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPoolKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPoolKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPoolKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPoolKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPoolKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPoolKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPoolKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPoolKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPoolKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TrimCommandPoolKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_TrimCommandPoolKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_TrimCommandPoolKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPoolKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPoolKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPoolKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPoolKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPoolKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPoolKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPoolKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPoolKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPoolKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPoolKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPoolKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPoolKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPoolKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TrimCommandPoolKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_TrimCommandPoolKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_TrimCommandPoolKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPoolKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPoolKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPoolKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPoolKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPoolKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPoolKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPoolKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPoolKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPoolKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPoolKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPoolKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPoolKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPoolKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TrimCommandPoolKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_TrimCommandPoolKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_TrimCommandPoolKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPoolKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPoolKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPoolKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPoolKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPoolKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPoolKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPoolKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPoolKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPoolKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPoolKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPoolKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPoolKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPoolKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TrimCommandPoolKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_TrimCommandPoolKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_TrimCommandPoolKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPoolKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPoolKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPoolKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPoolKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPoolKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPoolKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPoolKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPoolKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPoolKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPoolKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPoolKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPoolKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPoolKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TrimCommandPoolKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_TrimCommandPoolKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_TrimCommandPoolKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPoolKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPoolKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPoolKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPoolKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPoolKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPoolKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPoolKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPoolKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPoolKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPoolKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPoolKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPoolKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPoolKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TrimCommandPoolKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_TrimCommandPoolKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_TrimCommandPoolKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPoolKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPoolKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPoolKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPoolKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPoolKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPoolKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPoolKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPoolKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPoolKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPoolKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPoolKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPoolKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPoolKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TrimCommandPoolKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_TrimCommandPoolKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_TrimCommandPoolKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPoolKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPoolKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPoolKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPoolKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPoolKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPoolKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPoolKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPoolKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPoolKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPoolKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPoolKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPoolKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPoolKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TrimCommandPoolKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_TrimCommandPoolKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_TrimCommandPoolKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetMemoryWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetMemoryWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetMemoryWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetMemoryWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetMemoryWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetMemoryWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetMemoryWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetMemoryWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetMemoryWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetMemoryWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetMemoryWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetMemoryWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetMemoryWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetMemoryWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetMemoryWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetMemoryWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetMemoryWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetMemoryWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetMemoryWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetMemoryWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetMemoryWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetMemoryWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetMemoryWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetMemoryWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetMemoryWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetMemoryWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandlePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandlePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandlePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandlePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandlePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandlePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandlePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandlePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandlePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandlePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandlePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandlePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandlePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryWin32HandlePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetMemoryWin32HandlePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetMemoryWin32HandlePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandlePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandlePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandlePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandlePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandlePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandlePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandlePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandlePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandlePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandlePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandlePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandlePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandlePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryWin32HandlePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetMemoryWin32HandlePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetMemoryWin32HandlePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandlePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandlePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandlePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandlePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandlePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandlePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandlePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandlePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandlePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandlePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandlePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandlePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandlePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryWin32HandlePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetMemoryWin32HandlePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetMemoryWin32HandlePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandlePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandlePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandlePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandlePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandlePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandlePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandlePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandlePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandlePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandlePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandlePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandlePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandlePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryWin32HandlePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetMemoryWin32HandlePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetMemoryWin32HandlePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandlePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandlePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandlePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandlePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandlePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandlePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandlePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandlePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandlePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandlePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandlePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandlePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandlePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryWin32HandlePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetMemoryWin32HandlePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetMemoryWin32HandlePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandlePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandlePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandlePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandlePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandlePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandlePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandlePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandlePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandlePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandlePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandlePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandlePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandlePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryWin32HandlePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetMemoryWin32HandlePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetMemoryWin32HandlePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandlePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandlePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandlePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandlePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandlePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandlePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandlePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandlePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandlePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandlePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandlePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandlePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandlePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryWin32HandlePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetMemoryWin32HandlePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetMemoryWin32HandlePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandlePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandlePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandlePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandlePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandlePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandlePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandlePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandlePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandlePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandlePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandlePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandlePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandlePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryWin32HandlePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetMemoryWin32HandlePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetMemoryWin32HandlePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandlePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandlePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandlePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandlePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandlePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandlePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandlePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandlePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandlePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandlePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandlePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandlePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandlePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryWin32HandlePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetMemoryWin32HandlePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetMemoryWin32HandlePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandlePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandlePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandlePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandlePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandlePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandlePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandlePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandlePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandlePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandlePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandlePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandlePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandlePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryWin32HandlePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetMemoryWin32HandlePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetMemoryWin32HandlePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandlePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandlePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandlePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandlePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandlePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandlePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandlePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandlePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandlePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandlePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandlePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandlePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandlePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryWin32HandlePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetMemoryWin32HandlePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetMemoryWin32HandlePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandlePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandlePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandlePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandlePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandlePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandlePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandlePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandlePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandlePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandlePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandlePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandlePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandlePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryWin32HandlePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetMemoryWin32HandlePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetMemoryWin32HandlePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandlePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandlePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandlePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandlePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandlePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandlePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandlePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandlePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandlePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandlePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandlePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandlePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandlePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryWin32HandlePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetMemoryWin32HandlePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetMemoryWin32HandlePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetMemoryFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetMemoryFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetMemoryFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetMemoryFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetMemoryFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetMemoryFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetMemoryFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetMemoryFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetMemoryFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetMemoryFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetMemoryFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetMemoryFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetMemoryFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetMemoryFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetMemoryFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetMemoryFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetMemoryFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetMemoryFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetMemoryFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetMemoryFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetMemoryFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetMemoryFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetMemoryFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetMemoryFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetMemoryFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetMemoryFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryFdPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetMemoryFdPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetMemoryFdPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryFdPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetMemoryFdPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetMemoryFdPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryFdPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetMemoryFdPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetMemoryFdPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryFdPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetMemoryFdPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetMemoryFdPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryFdPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetMemoryFdPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetMemoryFdPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryFdPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetMemoryFdPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetMemoryFdPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryFdPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetMemoryFdPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetMemoryFdPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryFdPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetMemoryFdPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetMemoryFdPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryFdPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetMemoryFdPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetMemoryFdPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryFdPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetMemoryFdPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetMemoryFdPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryFdPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetMemoryFdPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetMemoryFdPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryFdPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetMemoryFdPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetMemoryFdPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryFdPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetMemoryFdPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetMemoryFdPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetMemoryZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetMemoryZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetMemoryZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetMemoryZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetMemoryZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetMemoryZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetMemoryZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetMemoryZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetMemoryZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetMemoryZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetMemoryZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetMemoryZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetMemoryZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetMemoryZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetMemoryZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetMemoryZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetMemoryZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetMemoryZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetMemoryZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetMemoryZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetMemoryZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetMemoryZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetMemoryZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetMemoryZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetMemoryZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetMemoryZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandlePropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandlePropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandlePropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandlePropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandlePropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandlePropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandlePropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandlePropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandlePropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandlePropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandlePropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandlePropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandlePropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryZirconHandlePropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetMemoryZirconHandlePropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetMemoryZirconHandlePropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandlePropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandlePropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandlePropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandlePropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandlePropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandlePropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandlePropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandlePropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandlePropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandlePropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandlePropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandlePropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandlePropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryZirconHandlePropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetMemoryZirconHandlePropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetMemoryZirconHandlePropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandlePropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandlePropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandlePropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandlePropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandlePropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandlePropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandlePropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandlePropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandlePropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandlePropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandlePropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandlePropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandlePropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryZirconHandlePropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetMemoryZirconHandlePropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetMemoryZirconHandlePropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandlePropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandlePropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandlePropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandlePropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandlePropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandlePropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandlePropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandlePropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandlePropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandlePropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandlePropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandlePropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandlePropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryZirconHandlePropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetMemoryZirconHandlePropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetMemoryZirconHandlePropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandlePropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandlePropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandlePropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandlePropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandlePropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandlePropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandlePropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandlePropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandlePropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandlePropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandlePropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandlePropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandlePropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryZirconHandlePropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetMemoryZirconHandlePropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetMemoryZirconHandlePropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandlePropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandlePropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandlePropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandlePropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandlePropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandlePropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandlePropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandlePropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandlePropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandlePropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandlePropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandlePropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandlePropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryZirconHandlePropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetMemoryZirconHandlePropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetMemoryZirconHandlePropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandlePropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandlePropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandlePropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandlePropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandlePropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandlePropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandlePropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandlePropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandlePropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandlePropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandlePropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandlePropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandlePropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryZirconHandlePropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetMemoryZirconHandlePropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetMemoryZirconHandlePropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandlePropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandlePropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandlePropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandlePropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandlePropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandlePropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandlePropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandlePropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandlePropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandlePropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandlePropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandlePropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandlePropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryZirconHandlePropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetMemoryZirconHandlePropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetMemoryZirconHandlePropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandlePropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandlePropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandlePropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandlePropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandlePropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandlePropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandlePropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandlePropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandlePropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandlePropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandlePropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandlePropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandlePropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryZirconHandlePropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetMemoryZirconHandlePropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetMemoryZirconHandlePropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandlePropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandlePropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandlePropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandlePropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandlePropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandlePropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandlePropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandlePropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandlePropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandlePropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandlePropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandlePropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandlePropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryZirconHandlePropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetMemoryZirconHandlePropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetMemoryZirconHandlePropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandlePropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandlePropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandlePropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandlePropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandlePropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandlePropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandlePropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandlePropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandlePropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandlePropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandlePropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandlePropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandlePropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryZirconHandlePropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetMemoryZirconHandlePropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetMemoryZirconHandlePropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandlePropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandlePropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandlePropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandlePropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandlePropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandlePropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandlePropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandlePropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandlePropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandlePropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandlePropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandlePropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandlePropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryZirconHandlePropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetMemoryZirconHandlePropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetMemoryZirconHandlePropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandlePropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandlePropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandlePropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandlePropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandlePropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandlePropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandlePropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandlePropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandlePropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandlePropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandlePropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandlePropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandlePropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryZirconHandlePropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetMemoryZirconHandlePropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetMemoryZirconHandlePropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetMemoryRemoteAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryRemoteAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryRemoteAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryRemoteAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryRemoteAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryRemoteAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryRemoteAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryRemoteAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryRemoteAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryRemoteAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryRemoteAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryRemoteAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryRemoteAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryRemoteAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetMemoryRemoteAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetMemoryRemoteAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryRemoteAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryRemoteAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryRemoteAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryRemoteAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryRemoteAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryRemoteAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryRemoteAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryRemoteAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryRemoteAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryRemoteAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryRemoteAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryRemoteAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryRemoteAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryRemoteAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetMemoryRemoteAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetMemoryRemoteAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryRemoteAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryRemoteAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryRemoteAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryRemoteAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryRemoteAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryRemoteAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryRemoteAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryRemoteAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryRemoteAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryRemoteAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryRemoteAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryRemoteAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryRemoteAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryRemoteAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetMemoryRemoteAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetMemoryRemoteAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryRemoteAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryRemoteAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryRemoteAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryRemoteAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryRemoteAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryRemoteAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryRemoteAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryRemoteAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryRemoteAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryRemoteAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryRemoteAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryRemoteAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryRemoteAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryRemoteAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetMemoryRemoteAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetMemoryRemoteAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryRemoteAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryRemoteAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryRemoteAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryRemoteAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryRemoteAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryRemoteAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryRemoteAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryRemoteAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryRemoteAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryRemoteAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryRemoteAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryRemoteAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryRemoteAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryRemoteAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetMemoryRemoteAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetMemoryRemoteAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryRemoteAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryRemoteAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryRemoteAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryRemoteAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryRemoteAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryRemoteAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryRemoteAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryRemoteAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryRemoteAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryRemoteAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryRemoteAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryRemoteAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryRemoteAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryRemoteAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetMemoryRemoteAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetMemoryRemoteAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryRemoteAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryRemoteAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryRemoteAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryRemoteAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryRemoteAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryRemoteAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryRemoteAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryRemoteAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryRemoteAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryRemoteAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryRemoteAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryRemoteAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryRemoteAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryRemoteAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetMemoryRemoteAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetMemoryRemoteAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryRemoteAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryRemoteAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryRemoteAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryRemoteAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryRemoteAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryRemoteAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryRemoteAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryRemoteAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryRemoteAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryRemoteAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryRemoteAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryRemoteAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryRemoteAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryRemoteAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetMemoryRemoteAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetMemoryRemoteAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryRemoteAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryRemoteAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryRemoteAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryRemoteAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryRemoteAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryRemoteAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryRemoteAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryRemoteAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryRemoteAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryRemoteAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryRemoteAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryRemoteAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryRemoteAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryRemoteAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetMemoryRemoteAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetMemoryRemoteAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryRemoteAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryRemoteAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryRemoteAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryRemoteAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryRemoteAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryRemoteAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryRemoteAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryRemoteAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryRemoteAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryRemoteAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryRemoteAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryRemoteAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryRemoteAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryRemoteAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetMemoryRemoteAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetMemoryRemoteAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryRemoteAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryRemoteAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryRemoteAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryRemoteAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryRemoteAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryRemoteAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryRemoteAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryRemoteAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryRemoteAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryRemoteAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryRemoteAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryRemoteAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryRemoteAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryRemoteAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetMemoryRemoteAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetMemoryRemoteAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryRemoteAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryRemoteAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryRemoteAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryRemoteAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryRemoteAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryRemoteAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryRemoteAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryRemoteAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryRemoteAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryRemoteAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryRemoteAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryRemoteAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryRemoteAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryRemoteAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetMemoryRemoteAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetMemoryRemoteAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryRemoteAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryRemoteAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryRemoteAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryRemoteAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryRemoteAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryRemoteAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryRemoteAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryRemoteAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryRemoteAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryRemoteAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryRemoteAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryRemoteAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryRemoteAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryRemoteAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetMemoryRemoteAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetMemoryRemoteAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ImportSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ImportSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ImportSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ImportSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ImportSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ImportSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ImportSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ImportSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ImportSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ImportSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ImportSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ImportSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ImportSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ImportSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ImportSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ImportSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ImportSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ImportSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ImportSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ImportSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ImportSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ImportSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ImportSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ImportSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ImportSemaphoreWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ImportSemaphoreWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ImportSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ImportSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ImportSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ImportSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ImportSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ImportSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ImportSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ImportSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ImportSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ImportSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ImportSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ImportSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ImportSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ImportSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ImportSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ImportSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ImportSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ImportSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ImportSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ImportSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ImportSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ImportSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ImportSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ImportSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ImportSemaphoreFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ImportSemaphoreFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ImportSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ImportSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ImportSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ImportSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ImportSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ImportSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ImportSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ImportSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ImportSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ImportSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ImportSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ImportSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ImportSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ImportSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ImportSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ImportSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ImportSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ImportSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ImportSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ImportSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ImportSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ImportSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ImportSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ImportSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreZirconHandleFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreZirconHandleFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreZirconHandleFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreZirconHandleFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreZirconHandleFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreZirconHandleFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreZirconHandleFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreZirconHandleFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreZirconHandleFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreZirconHandleFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreZirconHandleFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreZirconHandleFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreZirconHandleFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportSemaphoreZirconHandleFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ImportSemaphoreZirconHandleFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ImportSemaphoreZirconHandleFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ImportFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ImportFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ImportFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ImportFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ImportFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ImportFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ImportFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ImportFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ImportFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ImportFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ImportFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ImportFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ImportFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ImportFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ImportFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ImportFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ImportFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ImportFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ImportFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ImportFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ImportFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ImportFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ImportFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ImportFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ImportFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceWin32HandleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceWin32HandleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceWin32HandleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceWin32HandleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceWin32HandleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceWin32HandleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceWin32HandleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceWin32HandleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceWin32HandleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceWin32HandleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceWin32HandleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceWin32HandleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceWin32HandleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceWin32HandleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ImportFenceWin32HandleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ImportFenceWin32HandleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ImportFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ImportFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ImportFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ImportFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ImportFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ImportFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ImportFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ImportFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ImportFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ImportFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ImportFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ImportFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ImportFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ImportFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ImportFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ImportFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ImportFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ImportFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ImportFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ImportFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ImportFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ImportFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ImportFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ImportFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ImportFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ImportFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ImportFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ImportFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ImportFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ImportFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ImportFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ImportFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ImportFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ImportFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ImportFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ImportFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ImportFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceFdKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceFdKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceFdKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceFdKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceFdKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceFdKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceFdKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceFdKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceFdKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceFdKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceFdKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceFdKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceFdKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ImportFenceFdKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ImportFenceFdKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ImportFenceFdKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DisplayPowerControlEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DisplayPowerControlEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DisplayPowerControlEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DisplayPowerControlEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DisplayPowerControlEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DisplayPowerControlEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DisplayPowerControlEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DisplayPowerControlEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DisplayPowerControlEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DisplayPowerControlEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DisplayPowerControlEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DisplayPowerControlEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DisplayPowerControlEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DisplayPowerControlEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DisplayPowerControlEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DisplayPowerControlEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DisplayPowerControlEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DisplayPowerControlEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DisplayPowerControlEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DisplayPowerControlEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DisplayPowerControlEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DisplayPowerControlEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DisplayPowerControlEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DisplayPowerControlEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DisplayPowerControlEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DisplayPowerControlEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DisplayPowerControlEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DisplayPowerControlEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DisplayPowerControlEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DisplayPowerControlEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DisplayPowerControlEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DisplayPowerControlEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DisplayPowerControlEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DisplayPowerControlEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DisplayPowerControlEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DisplayPowerControlEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DisplayPowerControlEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DisplayPowerControlEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DisplayPowerControlEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DisplayPowerControlEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DisplayPowerControlEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DisplayPowerControlEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DisplayPowerControlEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DisplayPowerControlEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DisplayPowerControlEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DisplayPowerControlEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DisplayPowerControlEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DisplayPowerControlEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DisplayPowerControlEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DisplayPowerControlEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DisplayPowerControlEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DisplayPowerControlEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DisplayPowerControlEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DisplayPowerControlEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DisplayPowerControlEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DisplayPowerControlEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DisplayPowerControlEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DisplayPowerControlEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DisplayPowerControlEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DisplayPowerControlEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DisplayPowerControlEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DisplayPowerControlEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DisplayPowerControlEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DisplayPowerControlEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DisplayPowerControlEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DisplayPowerControlEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DisplayPowerControlEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DisplayPowerControlEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DisplayPowerControlEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DisplayPowerControlEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DisplayPowerControlEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DisplayPowerControlEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DisplayPowerControlEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DisplayPowerControlEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DisplayPowerControlEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DisplayPowerControlEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DisplayPowerControlEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DisplayPowerControlEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DisplayPowerControlEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DisplayPowerControlEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DisplayPowerControlEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DisplayPowerControlEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DisplayPowerControlEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DisplayPowerControlEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DisplayPowerControlEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DisplayPowerControlEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DisplayPowerControlEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DisplayPowerControlEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DisplayPowerControlEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DisplayPowerControlEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DisplayPowerControlEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DisplayPowerControlEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DisplayPowerControlEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DisplayPowerControlEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DisplayPowerControlEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DisplayPowerControlEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DisplayPowerControlEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DisplayPowerControlEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DisplayPowerControlEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DisplayPowerControlEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DisplayPowerControlEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DisplayPowerControlEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DisplayPowerControlEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DisplayPowerControlEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DisplayPowerControlEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DisplayPowerControlEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DisplayPowerControlEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DisplayPowerControlEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DisplayPowerControlEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DisplayPowerControlEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DisplayPowerControlEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DisplayPowerControlEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DisplayPowerControlEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DisplayPowerControlEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DisplayPowerControlEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DisplayPowerControlEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DisplayPowerControlEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DisplayPowerControlEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DisplayPowerControlEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DisplayPowerControlEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DisplayPowerControlEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DisplayPowerControlEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DisplayPowerControlEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DisplayPowerControlEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DisplayPowerControlEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DisplayPowerControlEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DisplayPowerControlEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DisplayPowerControlEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DisplayPowerControlEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DisplayPowerControlEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DisplayPowerControlEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DisplayPowerControlEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DisplayPowerControlEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DisplayPowerControlEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DisplayPowerControlEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DisplayPowerControlEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DisplayPowerControlEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DisplayPowerControlEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DisplayPowerControlEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DisplayPowerControlEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DisplayPowerControlEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DisplayPowerControlEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DisplayPowerControlEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DisplayPowerControlEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DisplayPowerControlEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DisplayPowerControlEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DisplayPowerControlEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DisplayPowerControlEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DisplayPowerControlEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DisplayPowerControlEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DisplayPowerControlEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DisplayPowerControlEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DisplayPowerControlEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DisplayPowerControlEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DisplayPowerControlEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DisplayPowerControlEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DisplayPowerControlEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DisplayPowerControlEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DisplayPowerControlEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DisplayPowerControlEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DisplayPowerControlEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DisplayPowerControlEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DisplayPowerControlEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DisplayPowerControlEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DisplayPowerControlEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DisplayPowerControlEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DisplayPowerControlEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DisplayPowerControlEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DisplayPowerControlEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DisplayPowerControlEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DisplayPowerControlEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DisplayPowerControlEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DisplayPowerControlEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DisplayPowerControlEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DisplayPowerControlEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DisplayPowerControlEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DisplayPowerControlEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DisplayPowerControlEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DisplayPowerControlEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DisplayPowerControlEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DisplayPowerControlEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DisplayPowerControlEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DisplayPowerControlEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DisplayPowerControlEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DisplayPowerControlEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DisplayPowerControlEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DisplayPowerControlEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DisplayPowerControlEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DisplayPowerControlEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DisplayPowerControlEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DisplayPowerControlEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DisplayPowerControlEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DisplayPowerControlEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DisplayPowerControlEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DisplayPowerControlEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DisplayPowerControlEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DisplayPowerControlEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DisplayPowerControlEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DisplayPowerControlEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DisplayPowerControlEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DisplayPowerControlEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DisplayPowerControlEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DisplayPowerControlEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DisplayPowerControlEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DisplayPowerControlEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DisplayPowerControlEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DisplayPowerControlEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DisplayPowerControlEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_RegisterDeviceEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDeviceEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDeviceEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDeviceEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDeviceEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDeviceEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDeviceEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDeviceEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDeviceEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDeviceEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDeviceEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDeviceEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDeviceEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDeviceEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_RegisterDeviceEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_RegisterDeviceEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDeviceEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDeviceEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDeviceEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDeviceEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDeviceEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDeviceEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDeviceEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDeviceEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDeviceEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDeviceEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDeviceEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDeviceEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDeviceEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDeviceEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_RegisterDeviceEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_RegisterDeviceEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDeviceEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDeviceEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDeviceEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDeviceEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDeviceEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDeviceEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDeviceEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDeviceEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDeviceEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDeviceEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDeviceEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDeviceEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDeviceEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDeviceEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_RegisterDeviceEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_RegisterDeviceEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDeviceEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDeviceEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDeviceEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDeviceEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDeviceEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDeviceEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDeviceEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDeviceEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDeviceEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDeviceEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDeviceEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDeviceEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDeviceEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDeviceEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_RegisterDeviceEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_RegisterDeviceEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDeviceEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDeviceEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDeviceEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDeviceEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDeviceEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDeviceEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDeviceEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDeviceEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDeviceEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDeviceEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDeviceEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDeviceEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDeviceEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDeviceEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_RegisterDeviceEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_RegisterDeviceEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDeviceEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDeviceEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDeviceEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDeviceEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDeviceEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDeviceEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDeviceEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDeviceEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDeviceEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDeviceEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDeviceEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDeviceEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDeviceEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDeviceEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_RegisterDeviceEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_RegisterDeviceEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDeviceEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDeviceEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDeviceEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDeviceEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDeviceEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDeviceEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDeviceEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDeviceEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDeviceEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDeviceEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDeviceEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDeviceEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDeviceEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDeviceEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_RegisterDeviceEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_RegisterDeviceEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDeviceEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDeviceEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDeviceEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDeviceEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDeviceEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDeviceEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDeviceEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDeviceEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDeviceEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDeviceEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDeviceEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDeviceEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDeviceEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDeviceEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_RegisterDeviceEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_RegisterDeviceEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDeviceEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDeviceEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDeviceEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDeviceEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDeviceEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDeviceEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDeviceEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDeviceEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDeviceEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDeviceEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDeviceEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDeviceEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDeviceEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDeviceEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_RegisterDeviceEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_RegisterDeviceEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDeviceEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDeviceEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDeviceEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDeviceEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDeviceEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDeviceEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDeviceEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDeviceEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDeviceEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDeviceEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDeviceEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDeviceEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDeviceEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDeviceEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_RegisterDeviceEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_RegisterDeviceEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDeviceEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDeviceEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDeviceEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDeviceEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDeviceEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDeviceEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDeviceEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDeviceEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDeviceEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDeviceEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDeviceEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDeviceEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDeviceEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDeviceEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_RegisterDeviceEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_RegisterDeviceEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDeviceEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDeviceEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDeviceEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDeviceEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDeviceEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDeviceEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDeviceEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDeviceEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDeviceEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDeviceEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDeviceEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDeviceEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDeviceEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDeviceEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_RegisterDeviceEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_RegisterDeviceEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDeviceEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDeviceEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDeviceEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDeviceEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDeviceEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDeviceEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDeviceEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDeviceEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDeviceEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDeviceEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDeviceEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDeviceEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDeviceEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDeviceEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_RegisterDeviceEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_RegisterDeviceEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_RegisterDisplayEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDisplayEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDisplayEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDisplayEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDisplayEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDisplayEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDisplayEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDisplayEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDisplayEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDisplayEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDisplayEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDisplayEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDisplayEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterDisplayEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_RegisterDisplayEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_RegisterDisplayEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDisplayEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDisplayEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDisplayEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDisplayEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDisplayEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDisplayEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDisplayEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDisplayEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDisplayEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDisplayEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDisplayEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDisplayEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDisplayEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterDisplayEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_RegisterDisplayEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_RegisterDisplayEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDisplayEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDisplayEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDisplayEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDisplayEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDisplayEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDisplayEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDisplayEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDisplayEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDisplayEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDisplayEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDisplayEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDisplayEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDisplayEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterDisplayEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_RegisterDisplayEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_RegisterDisplayEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDisplayEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDisplayEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDisplayEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDisplayEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDisplayEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDisplayEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDisplayEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDisplayEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDisplayEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDisplayEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDisplayEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDisplayEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDisplayEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterDisplayEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_RegisterDisplayEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_RegisterDisplayEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDisplayEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDisplayEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDisplayEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDisplayEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDisplayEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDisplayEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDisplayEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDisplayEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDisplayEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDisplayEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDisplayEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDisplayEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDisplayEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterDisplayEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_RegisterDisplayEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_RegisterDisplayEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDisplayEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDisplayEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDisplayEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDisplayEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDisplayEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDisplayEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDisplayEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDisplayEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDisplayEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDisplayEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDisplayEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDisplayEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDisplayEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterDisplayEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_RegisterDisplayEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_RegisterDisplayEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDisplayEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDisplayEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDisplayEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDisplayEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDisplayEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDisplayEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDisplayEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDisplayEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDisplayEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDisplayEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDisplayEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDisplayEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDisplayEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterDisplayEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_RegisterDisplayEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_RegisterDisplayEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDisplayEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDisplayEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDisplayEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDisplayEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDisplayEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDisplayEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDisplayEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDisplayEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDisplayEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDisplayEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDisplayEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDisplayEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDisplayEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterDisplayEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_RegisterDisplayEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_RegisterDisplayEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDisplayEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDisplayEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDisplayEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDisplayEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDisplayEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDisplayEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDisplayEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDisplayEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDisplayEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDisplayEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDisplayEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDisplayEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDisplayEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterDisplayEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_RegisterDisplayEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_RegisterDisplayEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDisplayEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDisplayEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDisplayEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDisplayEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDisplayEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDisplayEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDisplayEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDisplayEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDisplayEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDisplayEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDisplayEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDisplayEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDisplayEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterDisplayEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_RegisterDisplayEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_RegisterDisplayEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDisplayEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDisplayEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDisplayEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDisplayEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDisplayEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDisplayEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDisplayEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDisplayEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDisplayEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDisplayEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDisplayEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDisplayEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDisplayEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterDisplayEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_RegisterDisplayEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_RegisterDisplayEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDisplayEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDisplayEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDisplayEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDisplayEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDisplayEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDisplayEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDisplayEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDisplayEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDisplayEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDisplayEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDisplayEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDisplayEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDisplayEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterDisplayEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_RegisterDisplayEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_RegisterDisplayEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDisplayEventEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDisplayEventEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDisplayEventEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDisplayEventEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDisplayEventEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDisplayEventEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDisplayEventEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDisplayEventEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDisplayEventEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDisplayEventEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDisplayEventEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDisplayEventEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDisplayEventEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterDisplayEventEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_RegisterDisplayEventEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_RegisterDisplayEventEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainCounterEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainCounterEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainCounterEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainCounterEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainCounterEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainCounterEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainCounterEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainCounterEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainCounterEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainCounterEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainCounterEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainCounterEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainCounterEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainCounterEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetSwapchainCounterEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetSwapchainCounterEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainCounterEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainCounterEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainCounterEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainCounterEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainCounterEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainCounterEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainCounterEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainCounterEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainCounterEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainCounterEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainCounterEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainCounterEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainCounterEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainCounterEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetSwapchainCounterEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetSwapchainCounterEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainCounterEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainCounterEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainCounterEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainCounterEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainCounterEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainCounterEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainCounterEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainCounterEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainCounterEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainCounterEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainCounterEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainCounterEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainCounterEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainCounterEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetSwapchainCounterEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetSwapchainCounterEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainCounterEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainCounterEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainCounterEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainCounterEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainCounterEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainCounterEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainCounterEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainCounterEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainCounterEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainCounterEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainCounterEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainCounterEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainCounterEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainCounterEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetSwapchainCounterEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetSwapchainCounterEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainCounterEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainCounterEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainCounterEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainCounterEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainCounterEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainCounterEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainCounterEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainCounterEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainCounterEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainCounterEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainCounterEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainCounterEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainCounterEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainCounterEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetSwapchainCounterEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetSwapchainCounterEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainCounterEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainCounterEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainCounterEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainCounterEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainCounterEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainCounterEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainCounterEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainCounterEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainCounterEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainCounterEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainCounterEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainCounterEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainCounterEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainCounterEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetSwapchainCounterEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetSwapchainCounterEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainCounterEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainCounterEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainCounterEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainCounterEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainCounterEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainCounterEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainCounterEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainCounterEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainCounterEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainCounterEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainCounterEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainCounterEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainCounterEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainCounterEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetSwapchainCounterEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetSwapchainCounterEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainCounterEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainCounterEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainCounterEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainCounterEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainCounterEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainCounterEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainCounterEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainCounterEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainCounterEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainCounterEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainCounterEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainCounterEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainCounterEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainCounterEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetSwapchainCounterEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetSwapchainCounterEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainCounterEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainCounterEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainCounterEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainCounterEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainCounterEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainCounterEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainCounterEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainCounterEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainCounterEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainCounterEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainCounterEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainCounterEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainCounterEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainCounterEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetSwapchainCounterEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetSwapchainCounterEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainCounterEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainCounterEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainCounterEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainCounterEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainCounterEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainCounterEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainCounterEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainCounterEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainCounterEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainCounterEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainCounterEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainCounterEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainCounterEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainCounterEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetSwapchainCounterEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetSwapchainCounterEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainCounterEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainCounterEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainCounterEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainCounterEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainCounterEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainCounterEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainCounterEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainCounterEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainCounterEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainCounterEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainCounterEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainCounterEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainCounterEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainCounterEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetSwapchainCounterEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetSwapchainCounterEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainCounterEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainCounterEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainCounterEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainCounterEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainCounterEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainCounterEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainCounterEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainCounterEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainCounterEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainCounterEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainCounterEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainCounterEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainCounterEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainCounterEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetSwapchainCounterEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetSwapchainCounterEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainCounterEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainCounterEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainCounterEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainCounterEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainCounterEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainCounterEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainCounterEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainCounterEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainCounterEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainCounterEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainCounterEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainCounterEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainCounterEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainCounterEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetSwapchainCounterEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetSwapchainCounterEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceGroupPeerMemoryFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceGroupPeerMemoryFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceGroupPeerMemoryFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceGroupPeerMemoryFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceGroupPeerMemoryFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceGroupPeerMemoryFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceGroupPeerMemoryFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceGroupPeerMemoryFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceGroupPeerMemoryFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceGroupPeerMemoryFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceGroupPeerMemoryFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceGroupPeerMemoryFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceGroupPeerMemoryFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceGroupPeerMemoryFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceGroupPeerMemoryFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceGroupPeerMemoryFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceGroupPeerMemoryFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceGroupPeerMemoryFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceGroupPeerMemoryFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceGroupPeerMemoryFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceGroupPeerMemoryFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceGroupPeerMemoryFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceGroupPeerMemoryFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceGroupPeerMemoryFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeatures@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeatures@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeatures@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeatures@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeatures@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeatures@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeatures@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeatures@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeatures@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeatures@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeatures@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeatures@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeatures@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeatures@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceGroupPeerMemoryFeatures=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceGroupPeerMemoryFeatures=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeaturesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeaturesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeaturesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeaturesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeaturesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeaturesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeaturesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeaturesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeaturesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeaturesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeaturesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeaturesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeaturesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPeerMemoryFeaturesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceGroupPeerMemoryFeaturesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceGroupPeerMemoryFeaturesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeaturesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeaturesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeaturesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeaturesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeaturesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeaturesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeaturesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeaturesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeaturesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeaturesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeaturesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeaturesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeaturesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPeerMemoryFeaturesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceGroupPeerMemoryFeaturesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceGroupPeerMemoryFeaturesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeaturesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeaturesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeaturesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeaturesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeaturesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeaturesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeaturesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeaturesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeaturesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeaturesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeaturesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeaturesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeaturesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPeerMemoryFeaturesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceGroupPeerMemoryFeaturesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceGroupPeerMemoryFeaturesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeaturesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeaturesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeaturesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeaturesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeaturesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeaturesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeaturesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeaturesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeaturesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeaturesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeaturesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeaturesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeaturesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPeerMemoryFeaturesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceGroupPeerMemoryFeaturesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceGroupPeerMemoryFeaturesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeaturesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeaturesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeaturesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeaturesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeaturesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeaturesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeaturesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeaturesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeaturesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeaturesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeaturesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeaturesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeaturesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPeerMemoryFeaturesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceGroupPeerMemoryFeaturesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceGroupPeerMemoryFeaturesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeaturesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeaturesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeaturesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeaturesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeaturesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeaturesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeaturesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeaturesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeaturesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeaturesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeaturesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeaturesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeaturesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPeerMemoryFeaturesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceGroupPeerMemoryFeaturesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceGroupPeerMemoryFeaturesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeaturesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeaturesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeaturesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeaturesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeaturesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeaturesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeaturesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeaturesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeaturesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeaturesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeaturesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeaturesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeaturesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPeerMemoryFeaturesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceGroupPeerMemoryFeaturesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceGroupPeerMemoryFeaturesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeaturesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeaturesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeaturesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeaturesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeaturesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeaturesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeaturesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeaturesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeaturesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeaturesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeaturesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeaturesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeaturesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPeerMemoryFeaturesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceGroupPeerMemoryFeaturesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceGroupPeerMemoryFeaturesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeaturesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeaturesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeaturesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeaturesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeaturesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeaturesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeaturesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeaturesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeaturesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeaturesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeaturesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeaturesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeaturesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPeerMemoryFeaturesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceGroupPeerMemoryFeaturesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceGroupPeerMemoryFeaturesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeaturesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeaturesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeaturesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeaturesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeaturesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeaturesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeaturesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeaturesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeaturesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeaturesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeaturesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeaturesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeaturesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPeerMemoryFeaturesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceGroupPeerMemoryFeaturesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceGroupPeerMemoryFeaturesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeaturesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeaturesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeaturesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeaturesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeaturesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeaturesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeaturesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeaturesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeaturesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeaturesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeaturesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeaturesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeaturesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPeerMemoryFeaturesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceGroupPeerMemoryFeaturesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceGroupPeerMemoryFeaturesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeaturesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeaturesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeaturesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeaturesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeaturesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeaturesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeaturesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeaturesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeaturesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeaturesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeaturesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeaturesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeaturesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPeerMemoryFeaturesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceGroupPeerMemoryFeaturesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceGroupPeerMemoryFeaturesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeaturesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeaturesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeaturesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeaturesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeaturesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeaturesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeaturesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeaturesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeaturesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeaturesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeaturesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeaturesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeaturesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPeerMemoryFeaturesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceGroupPeerMemoryFeaturesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceGroupPeerMemoryFeaturesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_BindBufferMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_BindBufferMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_BindBufferMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_BindBufferMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_BindBufferMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_BindBufferMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_BindBufferMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_BindBufferMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_BindBufferMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_BindBufferMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_BindBufferMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_BindBufferMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_BindBufferMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_BindBufferMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_BindBufferMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_BindBufferMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_BindBufferMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_BindBufferMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_BindBufferMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_BindBufferMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_BindBufferMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_BindBufferMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_BindBufferMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_BindBufferMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_BindBufferMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_BindBufferMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindBufferMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_BindBufferMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_BindBufferMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindBufferMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_BindBufferMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_BindBufferMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindBufferMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_BindBufferMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_BindBufferMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindBufferMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_BindBufferMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_BindBufferMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindBufferMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_BindBufferMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_BindBufferMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindBufferMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_BindBufferMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_BindBufferMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindBufferMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_BindBufferMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_BindBufferMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindBufferMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_BindBufferMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_BindBufferMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindBufferMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_BindBufferMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_BindBufferMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindBufferMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_BindBufferMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_BindBufferMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindBufferMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_BindBufferMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_BindBufferMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindBufferMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_BindBufferMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_BindBufferMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindBufferMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_BindBufferMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_BindBufferMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_BindImageMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_BindImageMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_BindImageMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_BindImageMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_BindImageMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_BindImageMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_BindImageMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_BindImageMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_BindImageMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_BindImageMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_BindImageMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_BindImageMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_BindImageMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_BindImageMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_BindImageMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_BindImageMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_BindImageMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_BindImageMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_BindImageMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_BindImageMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_BindImageMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_BindImageMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_BindImageMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_BindImageMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_BindImageMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_BindImageMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindImageMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_BindImageMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_BindImageMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindImageMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_BindImageMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_BindImageMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindImageMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_BindImageMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_BindImageMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindImageMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_BindImageMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_BindImageMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindImageMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_BindImageMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_BindImageMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindImageMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_BindImageMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_BindImageMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindImageMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_BindImageMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_BindImageMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindImageMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_BindImageMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_BindImageMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindImageMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_BindImageMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_BindImageMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindImageMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_BindImageMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_BindImageMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindImageMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_BindImageMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_BindImageMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindImageMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_BindImageMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_BindImageMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindImageMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_BindImageMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_BindImageMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDeviceMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDeviceMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDeviceMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDeviceMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDeviceMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDeviceMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDeviceMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDeviceMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDeviceMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDeviceMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDeviceMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDeviceMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDeviceMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDeviceMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDeviceMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDeviceMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDeviceMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDeviceMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDeviceMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDeviceMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDeviceMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDeviceMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDeviceMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDeviceMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMask@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMask@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMask@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMask@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMask@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMask@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMask@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMask@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMask@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMask@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMask@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMask@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMask@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMask@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDeviceMask=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDeviceMask=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMaskKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMaskKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMaskKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMaskKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMaskKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMaskKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMaskKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMaskKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMaskKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMaskKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMaskKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMaskKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMaskKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDeviceMaskKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDeviceMaskKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDeviceMaskKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMaskKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMaskKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMaskKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMaskKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMaskKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMaskKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMaskKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMaskKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMaskKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMaskKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMaskKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMaskKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMaskKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDeviceMaskKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDeviceMaskKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDeviceMaskKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMaskKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMaskKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMaskKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMaskKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMaskKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMaskKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMaskKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMaskKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMaskKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMaskKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMaskKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMaskKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMaskKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDeviceMaskKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDeviceMaskKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDeviceMaskKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMaskKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMaskKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMaskKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMaskKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMaskKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMaskKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMaskKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMaskKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMaskKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMaskKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMaskKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMaskKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMaskKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDeviceMaskKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDeviceMaskKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDeviceMaskKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMaskKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMaskKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMaskKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMaskKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMaskKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMaskKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMaskKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMaskKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMaskKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMaskKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMaskKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMaskKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMaskKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDeviceMaskKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDeviceMaskKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDeviceMaskKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMaskKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMaskKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMaskKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMaskKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMaskKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMaskKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMaskKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMaskKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMaskKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMaskKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMaskKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMaskKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMaskKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDeviceMaskKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDeviceMaskKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDeviceMaskKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMaskKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMaskKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMaskKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMaskKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMaskKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMaskKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMaskKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMaskKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMaskKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMaskKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMaskKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMaskKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMaskKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDeviceMaskKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDeviceMaskKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDeviceMaskKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMaskKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMaskKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMaskKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMaskKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMaskKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMaskKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMaskKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMaskKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMaskKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMaskKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMaskKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMaskKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMaskKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDeviceMaskKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDeviceMaskKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDeviceMaskKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMaskKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMaskKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMaskKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMaskKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMaskKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMaskKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMaskKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMaskKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMaskKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMaskKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMaskKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMaskKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMaskKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDeviceMaskKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDeviceMaskKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDeviceMaskKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMaskKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMaskKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMaskKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMaskKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMaskKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMaskKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMaskKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMaskKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMaskKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMaskKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMaskKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMaskKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMaskKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDeviceMaskKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDeviceMaskKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDeviceMaskKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMaskKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMaskKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMaskKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMaskKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMaskKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMaskKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMaskKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMaskKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMaskKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMaskKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMaskKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMaskKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMaskKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDeviceMaskKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDeviceMaskKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDeviceMaskKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMaskKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMaskKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMaskKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMaskKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMaskKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMaskKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMaskKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMaskKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMaskKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMaskKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMaskKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMaskKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMaskKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDeviceMaskKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDeviceMaskKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDeviceMaskKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMaskKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMaskKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMaskKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMaskKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMaskKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMaskKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMaskKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMaskKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMaskKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMaskKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMaskKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMaskKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMaskKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDeviceMaskKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDeviceMaskKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDeviceMaskKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPresentCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPresentCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPresentCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPresentCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPresentCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPresentCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPresentCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPresentCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPresentCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPresentCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPresentCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPresentCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPresentCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupPresentCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceGroupPresentCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceGroupPresentCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPresentCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPresentCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPresentCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPresentCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPresentCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPresentCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPresentCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPresentCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPresentCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPresentCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPresentCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPresentCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPresentCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupPresentCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceGroupPresentCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceGroupPresentCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPresentCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPresentCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPresentCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPresentCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPresentCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPresentCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPresentCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPresentCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPresentCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPresentCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPresentCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPresentCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPresentCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupPresentCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceGroupPresentCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceGroupPresentCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPresentCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPresentCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPresentCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPresentCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPresentCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPresentCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPresentCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPresentCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPresentCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPresentCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPresentCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPresentCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPresentCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupPresentCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceGroupPresentCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceGroupPresentCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPresentCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPresentCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPresentCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPresentCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPresentCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPresentCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPresentCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPresentCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPresentCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPresentCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPresentCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPresentCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPresentCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupPresentCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceGroupPresentCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceGroupPresentCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPresentCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPresentCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPresentCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPresentCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPresentCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPresentCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPresentCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPresentCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPresentCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPresentCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPresentCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPresentCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPresentCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupPresentCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceGroupPresentCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceGroupPresentCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPresentCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPresentCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPresentCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPresentCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPresentCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPresentCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPresentCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPresentCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPresentCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPresentCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPresentCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPresentCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPresentCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupPresentCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceGroupPresentCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceGroupPresentCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPresentCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPresentCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPresentCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPresentCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPresentCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPresentCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPresentCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPresentCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPresentCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPresentCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPresentCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPresentCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPresentCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupPresentCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceGroupPresentCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceGroupPresentCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPresentCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPresentCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPresentCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPresentCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPresentCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPresentCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPresentCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPresentCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPresentCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPresentCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPresentCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPresentCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPresentCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupPresentCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceGroupPresentCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceGroupPresentCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPresentCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPresentCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPresentCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPresentCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPresentCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPresentCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPresentCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPresentCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPresentCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPresentCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPresentCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPresentCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPresentCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupPresentCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceGroupPresentCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceGroupPresentCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPresentCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPresentCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPresentCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPresentCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPresentCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPresentCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPresentCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPresentCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPresentCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPresentCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPresentCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPresentCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPresentCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupPresentCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceGroupPresentCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceGroupPresentCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPresentCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPresentCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPresentCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPresentCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPresentCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPresentCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPresentCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPresentCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPresentCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPresentCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPresentCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPresentCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPresentCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupPresentCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceGroupPresentCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceGroupPresentCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPresentCapabilitiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPresentCapabilitiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPresentCapabilitiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPresentCapabilitiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPresentCapabilitiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPresentCapabilitiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPresentCapabilitiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPresentCapabilitiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPresentCapabilitiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPresentCapabilitiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPresentCapabilitiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPresentCapabilitiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPresentCapabilitiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupPresentCapabilitiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceGroupPresentCapabilitiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceGroupPresentCapabilitiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceGroupSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceGroupSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceGroupSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceGroupSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceGroupSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceGroupSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceGroupSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceGroupSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceGroupSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceGroupSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceGroupSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceGroupSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceGroupSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceGroupSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceGroupSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceGroupSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceGroupSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceGroupSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceGroupSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceGroupSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceGroupSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceGroupSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceGroupSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceGroupSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceGroupSurfacePresentModesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceGroupSurfacePresentModesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireNextImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_AcquireNextImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_AcquireNextImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireNextImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_AcquireNextImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_AcquireNextImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireNextImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_AcquireNextImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_AcquireNextImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireNextImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_AcquireNextImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_AcquireNextImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireNextImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_AcquireNextImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_AcquireNextImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireNextImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_AcquireNextImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_AcquireNextImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireNextImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_AcquireNextImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_AcquireNextImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireNextImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_AcquireNextImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_AcquireNextImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireNextImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_AcquireNextImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_AcquireNextImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireNextImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_AcquireNextImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_AcquireNextImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireNextImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_AcquireNextImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_AcquireNextImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireNextImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_AcquireNextImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_AcquireNextImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireNextImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_AcquireNextImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_AcquireNextImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBase@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBase@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBase@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBase@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBase@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBase@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBase@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBase@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBase@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBase@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBase@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBase@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBase@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBase@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDispatchBase=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDispatchBase=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBase@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBase@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBase@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBase@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBase@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBase@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBase@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBase@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBase@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBase@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBase@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBase@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBase@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBase@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDispatchBase=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDispatchBase=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBase@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBase@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBase@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBase@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBase@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBase@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBase@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBase@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBase@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBase@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBase@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBase@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBase@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBase@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDispatchBase=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDispatchBase=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBase@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBase@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBase@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBase@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBase@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBase@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBase@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBase@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBase@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBase@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBase@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBase@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBase@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBase@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDispatchBase=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDispatchBase=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBase@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBase@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBase@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBase@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBase@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBase@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBase@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBase@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBase@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBase@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBase@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBase@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBase@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBase@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDispatchBase=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDispatchBase=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBase@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBase@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBase@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBase@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBase@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBase@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBase@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBase@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBase@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBase@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBase@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBase@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBase@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBase@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDispatchBase=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDispatchBase=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBase@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBase@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBase@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBase@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBase@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBase@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBase@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBase@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBase@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBase@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBase@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBase@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBase@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBase@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDispatchBase=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDispatchBase=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBase@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBase@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBase@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBase@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBase@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBase@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBase@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBase@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBase@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBase@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBase@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBase@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBase@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBase@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDispatchBase=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDispatchBase=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBase@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBase@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBase@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBase@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBase@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBase@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBase@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBase@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBase@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBase@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBase@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBase@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBase@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBase@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDispatchBase=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDispatchBase=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBase@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBase@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBase@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBase@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBase@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBase@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBase@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBase@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBase@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBase@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBase@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBase@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBase@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBase@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDispatchBase=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDispatchBase=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBase@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBase@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBase@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBase@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBase@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBase@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBase@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBase@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBase@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBase@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBase@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBase@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBase@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBase@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDispatchBase=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDispatchBase=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBase@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBase@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBase@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBase@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBase@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBase@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBase@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBase@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBase@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBase@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBase@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBase@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBase@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBase@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDispatchBase=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDispatchBase=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBase@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBase@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBase@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBase@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBase@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBase@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBase@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBase@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBase@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBase@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBase@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBase@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBase@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBase@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDispatchBase=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDispatchBase=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBaseKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBaseKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBaseKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBaseKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBaseKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBaseKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBaseKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBaseKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBaseKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBaseKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBaseKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBaseKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBaseKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchBaseKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDispatchBaseKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDispatchBaseKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBaseKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBaseKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBaseKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBaseKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBaseKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBaseKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBaseKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBaseKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBaseKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBaseKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBaseKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBaseKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBaseKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchBaseKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDispatchBaseKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDispatchBaseKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBaseKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBaseKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBaseKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBaseKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBaseKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBaseKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBaseKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBaseKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBaseKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBaseKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBaseKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBaseKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBaseKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchBaseKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDispatchBaseKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDispatchBaseKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBaseKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBaseKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBaseKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBaseKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBaseKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBaseKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBaseKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBaseKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBaseKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBaseKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBaseKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBaseKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBaseKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchBaseKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDispatchBaseKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDispatchBaseKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBaseKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBaseKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBaseKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBaseKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBaseKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBaseKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBaseKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBaseKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBaseKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBaseKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBaseKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBaseKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBaseKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchBaseKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDispatchBaseKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDispatchBaseKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBaseKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBaseKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBaseKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBaseKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBaseKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBaseKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBaseKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBaseKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBaseKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBaseKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBaseKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBaseKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBaseKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchBaseKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDispatchBaseKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDispatchBaseKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBaseKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBaseKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBaseKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBaseKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBaseKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBaseKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBaseKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBaseKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBaseKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBaseKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBaseKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBaseKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBaseKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchBaseKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDispatchBaseKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDispatchBaseKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBaseKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBaseKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBaseKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBaseKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBaseKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBaseKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBaseKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBaseKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBaseKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBaseKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBaseKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBaseKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBaseKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchBaseKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDispatchBaseKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDispatchBaseKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBaseKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBaseKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBaseKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBaseKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBaseKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBaseKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBaseKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBaseKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBaseKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBaseKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBaseKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBaseKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBaseKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchBaseKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDispatchBaseKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDispatchBaseKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBaseKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBaseKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBaseKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBaseKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBaseKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBaseKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBaseKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBaseKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBaseKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBaseKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBaseKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBaseKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBaseKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchBaseKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDispatchBaseKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDispatchBaseKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBaseKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBaseKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBaseKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBaseKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBaseKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBaseKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBaseKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBaseKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBaseKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBaseKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBaseKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBaseKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBaseKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchBaseKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDispatchBaseKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDispatchBaseKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBaseKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBaseKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBaseKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBaseKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBaseKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBaseKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBaseKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBaseKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBaseKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBaseKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBaseKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBaseKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBaseKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchBaseKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDispatchBaseKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDispatchBaseKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBaseKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBaseKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBaseKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBaseKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBaseKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBaseKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBaseKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBaseKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBaseKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBaseKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBaseKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBaseKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBaseKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchBaseKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDispatchBaseKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDispatchBaseKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyDescriptorUpdateTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyDescriptorUpdateTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDescriptorUpdateTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyDescriptorUpdateTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyDescriptorUpdateTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_UpdateDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_UpdateDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_UpdateDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_UpdateDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_UpdateDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_UpdateDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_UpdateDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_UpdateDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_UpdateDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_UpdateDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_UpdateDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_UpdateDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_UpdateDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_UpdateDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_UpdateDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_UpdateDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_UpdateDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_UpdateDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_UpdateDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_UpdateDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_UpdateDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_UpdateDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_UpdateDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_UpdateDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_UpdateDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_UpdateDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_UpdateDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_UpdateDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_UpdateDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_UpdateDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_UpdateDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_UpdateDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_UpdateDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_UpdateDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_UpdateDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_UpdateDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_UpdateDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_UpdateDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_UpdateDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_UpdateDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_UpdateDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_UpdateDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_UpdateDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_UpdateDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_UpdateDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_UpdateDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_UpdateDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_UpdateDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_UpdateDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_UpdateDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_UpdateDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_UpdateDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdPushDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdPushDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdPushDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdPushDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdPushDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdPushDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdPushDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdPushDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdPushDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdPushDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdPushDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdPushDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdPushDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdPushDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdPushDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdPushDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdPushDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdPushDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdPushDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdPushDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdPushDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdPushDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdPushDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdPushDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdPushDescriptorSetWithTemplate=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdPushDescriptorSetWithTemplate=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdPushDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdPushDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdPushDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdPushDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdPushDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdPushDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdPushDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdPushDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdPushDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdPushDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdPushDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdPushDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdPushDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdPushDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdPushDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdPushDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdPushDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdPushDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdPushDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdPushDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdPushDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdPushDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdPushDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdPushDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdPushDescriptorSetWithTemplateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdPushDescriptorSetWithTemplateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_SetHdrMetadataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetHdrMetadataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetHdrMetadataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetHdrMetadataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetHdrMetadataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetHdrMetadataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetHdrMetadataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetHdrMetadataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetHdrMetadataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetHdrMetadataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetHdrMetadataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetHdrMetadataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetHdrMetadataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetHdrMetadataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_SetHdrMetadataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_SetHdrMetadataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_SetHdrMetadataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetHdrMetadataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetHdrMetadataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetHdrMetadataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetHdrMetadataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetHdrMetadataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetHdrMetadataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetHdrMetadataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetHdrMetadataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetHdrMetadataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetHdrMetadataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetHdrMetadataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetHdrMetadataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetHdrMetadataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_SetHdrMetadataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_SetHdrMetadataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_SetHdrMetadataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetHdrMetadataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetHdrMetadataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetHdrMetadataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetHdrMetadataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetHdrMetadataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetHdrMetadataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetHdrMetadataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetHdrMetadataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetHdrMetadataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetHdrMetadataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetHdrMetadataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetHdrMetadataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetHdrMetadataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_SetHdrMetadataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_SetHdrMetadataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_SetHdrMetadataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetHdrMetadataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetHdrMetadataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetHdrMetadataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetHdrMetadataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetHdrMetadataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetHdrMetadataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetHdrMetadataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetHdrMetadataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetHdrMetadataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetHdrMetadataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetHdrMetadataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetHdrMetadataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetHdrMetadataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_SetHdrMetadataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_SetHdrMetadataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_SetHdrMetadataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetHdrMetadataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetHdrMetadataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetHdrMetadataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetHdrMetadataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetHdrMetadataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetHdrMetadataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetHdrMetadataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetHdrMetadataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetHdrMetadataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetHdrMetadataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetHdrMetadataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetHdrMetadataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetHdrMetadataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_SetHdrMetadataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_SetHdrMetadataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_SetHdrMetadataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetHdrMetadataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetHdrMetadataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetHdrMetadataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetHdrMetadataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetHdrMetadataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetHdrMetadataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetHdrMetadataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetHdrMetadataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetHdrMetadataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetHdrMetadataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetHdrMetadataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetHdrMetadataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetHdrMetadataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_SetHdrMetadataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_SetHdrMetadataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_SetHdrMetadataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetHdrMetadataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetHdrMetadataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetHdrMetadataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetHdrMetadataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetHdrMetadataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetHdrMetadataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetHdrMetadataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetHdrMetadataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetHdrMetadataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetHdrMetadataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetHdrMetadataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetHdrMetadataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetHdrMetadataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_SetHdrMetadataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_SetHdrMetadataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_SetHdrMetadataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetHdrMetadataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetHdrMetadataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetHdrMetadataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetHdrMetadataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetHdrMetadataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetHdrMetadataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetHdrMetadataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetHdrMetadataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetHdrMetadataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetHdrMetadataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetHdrMetadataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetHdrMetadataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetHdrMetadataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_SetHdrMetadataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_SetHdrMetadataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_SetHdrMetadataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetHdrMetadataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetHdrMetadataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetHdrMetadataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetHdrMetadataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetHdrMetadataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetHdrMetadataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetHdrMetadataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetHdrMetadataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetHdrMetadataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetHdrMetadataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetHdrMetadataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetHdrMetadataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetHdrMetadataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_SetHdrMetadataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_SetHdrMetadataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_SetHdrMetadataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetHdrMetadataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetHdrMetadataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetHdrMetadataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetHdrMetadataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetHdrMetadataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetHdrMetadataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetHdrMetadataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetHdrMetadataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetHdrMetadataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetHdrMetadataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetHdrMetadataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetHdrMetadataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetHdrMetadataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_SetHdrMetadataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_SetHdrMetadataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetHdrMetadataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetHdrMetadataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetHdrMetadataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetHdrMetadataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetHdrMetadataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetHdrMetadataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetHdrMetadataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetHdrMetadataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetHdrMetadataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetHdrMetadataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetHdrMetadataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetHdrMetadataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetHdrMetadataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetHdrMetadataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_SetHdrMetadataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_SetHdrMetadataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_SetHdrMetadataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetHdrMetadataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetHdrMetadataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetHdrMetadataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetHdrMetadataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetHdrMetadataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetHdrMetadataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetHdrMetadataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetHdrMetadataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetHdrMetadataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetHdrMetadataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetHdrMetadataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetHdrMetadataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetHdrMetadataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_SetHdrMetadataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_SetHdrMetadataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_SetHdrMetadataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetHdrMetadataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetHdrMetadataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetHdrMetadataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetHdrMetadataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetHdrMetadataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetHdrMetadataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetHdrMetadataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetHdrMetadataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetHdrMetadataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetHdrMetadataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetHdrMetadataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetHdrMetadataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetHdrMetadataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_SetHdrMetadataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_SetHdrMetadataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainStatusKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainStatusKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainStatusKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainStatusKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainStatusKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainStatusKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainStatusKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainStatusKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainStatusKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainStatusKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainStatusKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainStatusKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainStatusKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainStatusKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetSwapchainStatusKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetSwapchainStatusKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainStatusKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainStatusKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainStatusKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainStatusKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainStatusKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainStatusKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainStatusKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainStatusKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainStatusKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainStatusKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainStatusKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainStatusKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainStatusKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainStatusKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetSwapchainStatusKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetSwapchainStatusKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainStatusKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainStatusKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainStatusKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainStatusKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainStatusKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainStatusKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainStatusKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainStatusKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainStatusKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainStatusKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainStatusKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainStatusKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainStatusKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainStatusKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetSwapchainStatusKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetSwapchainStatusKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainStatusKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainStatusKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainStatusKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainStatusKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainStatusKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainStatusKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainStatusKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainStatusKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainStatusKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainStatusKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainStatusKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainStatusKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainStatusKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainStatusKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetSwapchainStatusKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetSwapchainStatusKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainStatusKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainStatusKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainStatusKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainStatusKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainStatusKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainStatusKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainStatusKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainStatusKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainStatusKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainStatusKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainStatusKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainStatusKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainStatusKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainStatusKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetSwapchainStatusKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetSwapchainStatusKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainStatusKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainStatusKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainStatusKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainStatusKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainStatusKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainStatusKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainStatusKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainStatusKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainStatusKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainStatusKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainStatusKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainStatusKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainStatusKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainStatusKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetSwapchainStatusKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetSwapchainStatusKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainStatusKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainStatusKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainStatusKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainStatusKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainStatusKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainStatusKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainStatusKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainStatusKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainStatusKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainStatusKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainStatusKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainStatusKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainStatusKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainStatusKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetSwapchainStatusKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetSwapchainStatusKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainStatusKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainStatusKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainStatusKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainStatusKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainStatusKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainStatusKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainStatusKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainStatusKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainStatusKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainStatusKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainStatusKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainStatusKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainStatusKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainStatusKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetSwapchainStatusKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetSwapchainStatusKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainStatusKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainStatusKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainStatusKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainStatusKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainStatusKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainStatusKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainStatusKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainStatusKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainStatusKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainStatusKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainStatusKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainStatusKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainStatusKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainStatusKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetSwapchainStatusKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetSwapchainStatusKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainStatusKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainStatusKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainStatusKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainStatusKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainStatusKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainStatusKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainStatusKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainStatusKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainStatusKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainStatusKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainStatusKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainStatusKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainStatusKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainStatusKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetSwapchainStatusKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetSwapchainStatusKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainStatusKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainStatusKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainStatusKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainStatusKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainStatusKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainStatusKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainStatusKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainStatusKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainStatusKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainStatusKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainStatusKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainStatusKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainStatusKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainStatusKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetSwapchainStatusKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetSwapchainStatusKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainStatusKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainStatusKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainStatusKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainStatusKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainStatusKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainStatusKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainStatusKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainStatusKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainStatusKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainStatusKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainStatusKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainStatusKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainStatusKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainStatusKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetSwapchainStatusKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetSwapchainStatusKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainStatusKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainStatusKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainStatusKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainStatusKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainStatusKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainStatusKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainStatusKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainStatusKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainStatusKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainStatusKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainStatusKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainStatusKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainStatusKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainStatusKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetSwapchainStatusKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetSwapchainStatusKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetRefreshCycleDurationGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRefreshCycleDurationGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRefreshCycleDurationGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRefreshCycleDurationGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRefreshCycleDurationGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRefreshCycleDurationGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRefreshCycleDurationGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRefreshCycleDurationGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRefreshCycleDurationGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRefreshCycleDurationGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRefreshCycleDurationGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRefreshCycleDurationGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRefreshCycleDurationGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRefreshCycleDurationGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetRefreshCycleDurationGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetRefreshCycleDurationGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetRefreshCycleDurationGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRefreshCycleDurationGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRefreshCycleDurationGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRefreshCycleDurationGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRefreshCycleDurationGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRefreshCycleDurationGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRefreshCycleDurationGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRefreshCycleDurationGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRefreshCycleDurationGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRefreshCycleDurationGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRefreshCycleDurationGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRefreshCycleDurationGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRefreshCycleDurationGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRefreshCycleDurationGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetRefreshCycleDurationGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetRefreshCycleDurationGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetRefreshCycleDurationGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRefreshCycleDurationGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRefreshCycleDurationGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRefreshCycleDurationGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRefreshCycleDurationGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRefreshCycleDurationGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRefreshCycleDurationGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRefreshCycleDurationGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRefreshCycleDurationGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRefreshCycleDurationGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRefreshCycleDurationGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRefreshCycleDurationGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRefreshCycleDurationGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRefreshCycleDurationGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetRefreshCycleDurationGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetRefreshCycleDurationGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetRefreshCycleDurationGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRefreshCycleDurationGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRefreshCycleDurationGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRefreshCycleDurationGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRefreshCycleDurationGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRefreshCycleDurationGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRefreshCycleDurationGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRefreshCycleDurationGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRefreshCycleDurationGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRefreshCycleDurationGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRefreshCycleDurationGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRefreshCycleDurationGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRefreshCycleDurationGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRefreshCycleDurationGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetRefreshCycleDurationGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetRefreshCycleDurationGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetRefreshCycleDurationGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRefreshCycleDurationGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRefreshCycleDurationGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRefreshCycleDurationGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRefreshCycleDurationGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRefreshCycleDurationGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRefreshCycleDurationGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRefreshCycleDurationGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRefreshCycleDurationGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRefreshCycleDurationGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRefreshCycleDurationGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRefreshCycleDurationGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRefreshCycleDurationGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRefreshCycleDurationGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetRefreshCycleDurationGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetRefreshCycleDurationGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetRefreshCycleDurationGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRefreshCycleDurationGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRefreshCycleDurationGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRefreshCycleDurationGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRefreshCycleDurationGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRefreshCycleDurationGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRefreshCycleDurationGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRefreshCycleDurationGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRefreshCycleDurationGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRefreshCycleDurationGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRefreshCycleDurationGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRefreshCycleDurationGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRefreshCycleDurationGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRefreshCycleDurationGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetRefreshCycleDurationGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetRefreshCycleDurationGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetRefreshCycleDurationGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRefreshCycleDurationGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRefreshCycleDurationGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRefreshCycleDurationGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRefreshCycleDurationGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRefreshCycleDurationGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRefreshCycleDurationGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRefreshCycleDurationGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRefreshCycleDurationGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRefreshCycleDurationGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRefreshCycleDurationGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRefreshCycleDurationGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRefreshCycleDurationGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRefreshCycleDurationGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetRefreshCycleDurationGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetRefreshCycleDurationGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetRefreshCycleDurationGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRefreshCycleDurationGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRefreshCycleDurationGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRefreshCycleDurationGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRefreshCycleDurationGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRefreshCycleDurationGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRefreshCycleDurationGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRefreshCycleDurationGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRefreshCycleDurationGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRefreshCycleDurationGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRefreshCycleDurationGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRefreshCycleDurationGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRefreshCycleDurationGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRefreshCycleDurationGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetRefreshCycleDurationGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetRefreshCycleDurationGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRefreshCycleDurationGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRefreshCycleDurationGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRefreshCycleDurationGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRefreshCycleDurationGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRefreshCycleDurationGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRefreshCycleDurationGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRefreshCycleDurationGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRefreshCycleDurationGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRefreshCycleDurationGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRefreshCycleDurationGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRefreshCycleDurationGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRefreshCycleDurationGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRefreshCycleDurationGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRefreshCycleDurationGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetRefreshCycleDurationGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetRefreshCycleDurationGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRefreshCycleDurationGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRefreshCycleDurationGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRefreshCycleDurationGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRefreshCycleDurationGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRefreshCycleDurationGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRefreshCycleDurationGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRefreshCycleDurationGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRefreshCycleDurationGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRefreshCycleDurationGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRefreshCycleDurationGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRefreshCycleDurationGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRefreshCycleDurationGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRefreshCycleDurationGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRefreshCycleDurationGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetRefreshCycleDurationGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetRefreshCycleDurationGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRefreshCycleDurationGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRefreshCycleDurationGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRefreshCycleDurationGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRefreshCycleDurationGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRefreshCycleDurationGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRefreshCycleDurationGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRefreshCycleDurationGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRefreshCycleDurationGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRefreshCycleDurationGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRefreshCycleDurationGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRefreshCycleDurationGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRefreshCycleDurationGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRefreshCycleDurationGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRefreshCycleDurationGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetRefreshCycleDurationGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetRefreshCycleDurationGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetRefreshCycleDurationGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRefreshCycleDurationGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRefreshCycleDurationGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRefreshCycleDurationGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRefreshCycleDurationGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRefreshCycleDurationGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRefreshCycleDurationGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRefreshCycleDurationGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRefreshCycleDurationGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRefreshCycleDurationGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRefreshCycleDurationGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRefreshCycleDurationGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRefreshCycleDurationGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRefreshCycleDurationGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetRefreshCycleDurationGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetRefreshCycleDurationGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRefreshCycleDurationGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRefreshCycleDurationGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRefreshCycleDurationGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRefreshCycleDurationGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRefreshCycleDurationGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRefreshCycleDurationGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRefreshCycleDurationGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRefreshCycleDurationGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRefreshCycleDurationGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRefreshCycleDurationGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRefreshCycleDurationGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRefreshCycleDurationGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRefreshCycleDurationGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRefreshCycleDurationGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetRefreshCycleDurationGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetRefreshCycleDurationGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPastPresentationTimingGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPastPresentationTimingGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetPastPresentationTimingGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetPastPresentationTimingGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetPastPresentationTimingGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetPastPresentationTimingGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetPastPresentationTimingGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetPastPresentationTimingGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetPastPresentationTimingGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetPastPresentationTimingGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetPastPresentationTimingGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetPastPresentationTimingGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetPastPresentationTimingGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetPastPresentationTimingGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetPastPresentationTimingGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetPastPresentationTimingGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetPastPresentationTimingGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetPastPresentationTimingGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetPastPresentationTimingGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetPastPresentationTimingGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetPastPresentationTimingGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetPastPresentationTimingGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetPastPresentationTimingGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetPastPresentationTimingGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingGOOGLE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingGOOGLE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingGOOGLE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingGOOGLE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingGOOGLE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingGOOGLE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingGOOGLE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingGOOGLE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingGOOGLE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingGOOGLE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingGOOGLE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingGOOGLE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingGOOGLE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingGOOGLE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetPastPresentationTimingGOOGLE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetPastPresentationTimingGOOGLE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetViewportWScalingNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetViewportWScalingNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetViewportWScalingNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetViewportWScalingNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetViewportWScalingNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetViewportWScalingNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetViewportWScalingNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetViewportWScalingNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetViewportWScalingNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetViewportWScalingNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetViewportWScalingNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetViewportWScalingNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetViewportWScalingNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetViewportWScalingNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetViewportWScalingNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetViewportWScalingNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetViewportWScalingNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetViewportWScalingNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetViewportWScalingNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetViewportWScalingNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetViewportWScalingNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetViewportWScalingNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetViewportWScalingNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetViewportWScalingNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetViewportWScalingNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetViewportWScalingNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDiscardRectangleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDiscardRectangleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDiscardRectangleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDiscardRectangleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDiscardRectangleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDiscardRectangleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDiscardRectangleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDiscardRectangleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDiscardRectangleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDiscardRectangleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDiscardRectangleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDiscardRectangleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDiscardRectangleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDiscardRectangleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDiscardRectangleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDiscardRectangleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDiscardRectangleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDiscardRectangleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDiscardRectangleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDiscardRectangleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDiscardRectangleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDiscardRectangleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDiscardRectangleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDiscardRectangleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDiscardRectangleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDiscardRectangleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDiscardRectangleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDiscardRectangleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDiscardRectangleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDiscardRectangleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDiscardRectangleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDiscardRectangleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDiscardRectangleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDiscardRectangleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDiscardRectangleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDiscardRectangleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDiscardRectangleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDiscardRectangleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDiscardRectangleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDiscardRectangleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDiscardRectangleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDiscardRectangleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDiscardRectangleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDiscardRectangleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDiscardRectangleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDiscardRectangleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDiscardRectangleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDiscardRectangleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDiscardRectangleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDiscardRectangleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDiscardRectangleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDiscardRectangleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDiscardRectangleModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDiscardRectangleModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDiscardRectangleModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDiscardRectangleModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDiscardRectangleModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDiscardRectangleModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDiscardRectangleModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDiscardRectangleModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDiscardRectangleModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDiscardRectangleModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDiscardRectangleModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDiscardRectangleModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDiscardRectangleModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDiscardRectangleModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDiscardRectangleModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDiscardRectangleModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDiscardRectangleModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDiscardRectangleModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDiscardRectangleModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDiscardRectangleModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDiscardRectangleModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDiscardRectangleModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDiscardRectangleModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDiscardRectangleModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDiscardRectangleModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDiscardRectangleModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDiscardRectangleModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDiscardRectangleModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDiscardRectangleModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDiscardRectangleModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDiscardRectangleModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDiscardRectangleModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDiscardRectangleModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDiscardRectangleModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDiscardRectangleModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDiscardRectangleModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDiscardRectangleModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDiscardRectangleModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDiscardRectangleModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetSampleLocationsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetSampleLocationsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetSampleLocationsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetSampleLocationsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetSampleLocationsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetSampleLocationsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetSampleLocationsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetSampleLocationsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetSampleLocationsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetSampleLocationsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetSampleLocationsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetSampleLocationsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetSampleLocationsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetSampleLocationsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetSampleLocationsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetSampleLocationsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetSampleLocationsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetSampleLocationsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetSampleLocationsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetSampleLocationsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetSampleLocationsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetSampleLocationsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetSampleLocationsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetSampleLocationsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetSampleLocationsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetSampleLocationsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetBufferMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetBufferMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetBufferMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetBufferMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetBufferMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetBufferMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetBufferMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetBufferMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetBufferMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetBufferMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetBufferMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetBufferMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetBufferMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetBufferMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetBufferMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetBufferMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetBufferMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetBufferMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetBufferMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetBufferMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetBufferMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetBufferMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetBufferMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetBufferMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetBufferMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetBufferMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetBufferMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetBufferMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetBufferMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetBufferMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetBufferMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetBufferMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetBufferMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetBufferMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetBufferMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetBufferMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetBufferMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetBufferMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetBufferMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetBufferMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetBufferMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetBufferMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetBufferMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetBufferMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetBufferMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetBufferMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetBufferMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetBufferMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetBufferMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetBufferMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetBufferMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetBufferMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetImageMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetImageMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetImageMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetImageMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetImageMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetImageMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetImageMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetImageMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetImageMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetImageMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetImageMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetImageMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetImageMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetImageMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetImageMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetImageMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetImageMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetImageMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetImageMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetImageMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetImageMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetImageMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetImageMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetImageMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetImageMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetImageMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetImageMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetImageMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetImageMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetImageMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetImageMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetImageMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetImageMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetImageMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetImageMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetImageMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetImageMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetImageMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetImageMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetImageMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetImageMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetImageMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetImageMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetImageMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetImageMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetImageMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetImageMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetImageMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetImageMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetImageMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetImageMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetImageMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetImageSparseMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetImageSparseMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetImageSparseMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetImageSparseMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetImageSparseMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetImageSparseMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetImageSparseMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetImageSparseMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetImageSparseMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetImageSparseMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetImageSparseMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetImageSparseMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetImageSparseMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetImageSparseMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetImageSparseMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetImageSparseMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetImageSparseMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetImageSparseMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetImageSparseMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetImageSparseMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetImageSparseMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetImageSparseMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetImageSparseMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetImageSparseMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetImageSparseMemoryRequirements2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetImageSparseMemoryRequirements2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSparseMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetImageSparseMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetImageSparseMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSparseMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetImageSparseMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetImageSparseMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSparseMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetImageSparseMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetImageSparseMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSparseMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetImageSparseMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetImageSparseMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSparseMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetImageSparseMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetImageSparseMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSparseMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetImageSparseMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetImageSparseMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSparseMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetImageSparseMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetImageSparseMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSparseMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetImageSparseMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetImageSparseMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSparseMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetImageSparseMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetImageSparseMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSparseMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetImageSparseMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetImageSparseMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSparseMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetImageSparseMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetImageSparseMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSparseMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetImageSparseMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetImageSparseMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSparseMemoryRequirements2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetImageSparseMemoryRequirements2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetImageSparseMemoryRequirements2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceBufferMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceBufferMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceBufferMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceBufferMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceBufferMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceBufferMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceBufferMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceBufferMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceBufferMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceBufferMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceBufferMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceBufferMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceBufferMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceBufferMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceBufferMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceBufferMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceBufferMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceBufferMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceBufferMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceBufferMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceBufferMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceBufferMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceBufferMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceBufferMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceBufferMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceBufferMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceBufferMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceBufferMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceBufferMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceBufferMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceBufferMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceBufferMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceBufferMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceBufferMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceBufferMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceBufferMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceBufferMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceBufferMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceBufferMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceBufferMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceBufferMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceImageMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceImageMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceImageMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceImageMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceImageMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceImageMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceImageMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceImageMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceImageMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceImageMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceImageMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceImageMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceImageMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceImageMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceImageMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceImageMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceImageMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceImageMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceImageMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceImageMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceImageMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceImageMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceImageMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceImageMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceImageMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceImageMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceImageMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceImageMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirements@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirements@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirements@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirements@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirements@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirements@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirements@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirements@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirements@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirements@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirements@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirements@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirements@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirements@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceImageSparseMemoryRequirements=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceImageSparseMemoryRequirements=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSparseMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceImageSparseMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceImageSparseMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSparseMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceImageSparseMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceImageSparseMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSparseMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceImageSparseMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceImageSparseMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSparseMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceImageSparseMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceImageSparseMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSparseMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceImageSparseMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceImageSparseMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSparseMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceImageSparseMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceImageSparseMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSparseMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceImageSparseMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceImageSparseMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSparseMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceImageSparseMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceImageSparseMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSparseMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceImageSparseMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceImageSparseMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSparseMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceImageSparseMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceImageSparseMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSparseMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceImageSparseMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceImageSparseMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSparseMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceImageSparseMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceImageSparseMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSparseMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceImageSparseMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceImageSparseMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateSamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateSamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateSamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateSamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateSamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateSamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateSamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateSamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateSamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateSamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateSamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateSamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateSamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateSamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateSamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateSamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateSamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateSamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateSamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateSamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateSamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateSamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateSamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateSamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateSamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateSamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateSamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateSamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateSamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateSamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateSamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateSamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateSamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateSamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateSamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateSamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateSamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateSamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateSamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateSamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateSamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateSamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateSamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateSamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateSamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateSamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateSamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateSamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateSamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateSamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateSamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateSamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateSamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateSamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateSamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateSamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateSamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateSamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateSamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateSamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateSamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateSamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateSamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateSamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateSamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroySamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroySamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroySamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroySamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroySamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroySamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroySamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroySamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroySamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroySamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroySamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroySamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroySamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroySamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroySamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroySamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroySamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroySamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroySamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroySamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroySamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroySamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroySamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroySamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversion@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversion@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversion@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversion@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversion@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversion@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversion@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversion@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversion@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversion@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversion@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversion@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversion@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversion@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroySamplerYcbcrConversion=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroySamplerYcbcrConversion=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroySamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroySamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroySamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroySamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroySamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroySamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroySamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroySamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroySamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroySamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroySamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroySamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroySamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroySamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroySamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroySamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroySamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroySamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroySamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroySamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroySamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroySamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroySamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroySamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroySamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroySamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroySamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroySamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroySamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroySamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroySamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroySamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroySamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroySamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroySamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroySamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroySamplerYcbcrConversionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroySamplerYcbcrConversionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroySamplerYcbcrConversionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceQueue2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceQueue2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceQueue2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceQueue2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceQueue2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceQueue2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceQueue2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceQueue2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceQueue2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceQueue2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceQueue2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceQueue2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceQueue2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceQueue2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceQueue2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceQueue2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceQueue2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceQueue2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceQueue2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceQueue2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceQueue2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceQueue2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceQueue2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceQueue2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceQueue2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceQueue2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceQueue2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceQueue2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceQueue2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceQueue2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceQueue2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceQueue2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceQueue2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceQueue2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceQueue2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceQueue2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceQueue2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceQueue2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceQueue2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyValidationCacheEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyValidationCacheEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyValidationCacheEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyValidationCacheEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyValidationCacheEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyValidationCacheEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyValidationCacheEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyValidationCacheEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyValidationCacheEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyValidationCacheEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyValidationCacheEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyValidationCacheEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyValidationCacheEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyValidationCacheEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyValidationCacheEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyValidationCacheEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetValidationCacheDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetValidationCacheDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetValidationCacheDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetValidationCacheDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetValidationCacheDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetValidationCacheDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetValidationCacheDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetValidationCacheDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetValidationCacheDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetValidationCacheDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetValidationCacheDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetValidationCacheDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetValidationCacheDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetValidationCacheDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetValidationCacheDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetValidationCacheDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetValidationCacheDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetValidationCacheDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetValidationCacheDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetValidationCacheDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetValidationCacheDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetValidationCacheDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetValidationCacheDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetValidationCacheDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetValidationCacheDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetValidationCacheDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetValidationCacheDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetValidationCacheDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetValidationCacheDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetValidationCacheDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetValidationCacheDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetValidationCacheDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetValidationCacheDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetValidationCacheDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetValidationCacheDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetValidationCacheDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetValidationCacheDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetValidationCacheDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetValidationCacheDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetValidationCacheDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetValidationCacheDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetValidationCacheDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetValidationCacheDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetValidationCacheDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetValidationCacheDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetValidationCacheDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetValidationCacheDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetValidationCacheDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetValidationCacheDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetValidationCacheDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetValidationCacheDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetValidationCacheDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetValidationCacheDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetValidationCacheDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetValidationCacheDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetValidationCacheDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetValidationCacheDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetValidationCacheDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetValidationCacheDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetValidationCacheDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetValidationCacheDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetValidationCacheDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetValidationCacheDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetValidationCacheDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetValidationCacheDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetValidationCacheDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetValidationCacheDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetValidationCacheDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetValidationCacheDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetValidationCacheDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetValidationCacheDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetValidationCacheDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetValidationCacheDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetValidationCacheDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetValidationCacheDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetValidationCacheDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetValidationCacheDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetValidationCacheDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetValidationCacheDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetValidationCacheDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetValidationCacheDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetValidationCacheDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetValidationCacheDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetValidationCacheDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetValidationCacheDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetValidationCacheDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetValidationCacheDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetValidationCacheDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetValidationCacheDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetValidationCacheDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetValidationCacheDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetValidationCacheDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetValidationCacheDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetValidationCacheDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetValidationCacheDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetValidationCacheDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetValidationCacheDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetValidationCacheDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetValidationCacheDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetValidationCacheDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetValidationCacheDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetValidationCacheDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetValidationCacheDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetValidationCacheDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetValidationCacheDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetValidationCacheDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetValidationCacheDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetValidationCacheDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetValidationCacheDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetValidationCacheDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetValidationCacheDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetValidationCacheDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetValidationCacheDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetValidationCacheDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetValidationCacheDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetValidationCacheDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetValidationCacheDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetValidationCacheDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetValidationCacheDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetValidationCacheDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetValidationCacheDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetValidationCacheDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetValidationCacheDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetValidationCacheDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetValidationCacheDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetValidationCacheDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetValidationCacheDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetValidationCacheDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetValidationCacheDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetValidationCacheDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetValidationCacheDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetValidationCacheDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetValidationCacheDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetValidationCacheDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetValidationCacheDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetValidationCacheDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetValidationCacheDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetValidationCacheDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetValidationCacheDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetValidationCacheDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetValidationCacheDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetValidationCacheDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetValidationCacheDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetValidationCacheDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetValidationCacheDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetValidationCacheDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetValidationCacheDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetValidationCacheDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetValidationCacheDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetValidationCacheDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetValidationCacheDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetValidationCacheDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetValidationCacheDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetValidationCacheDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetValidationCacheDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetValidationCacheDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetValidationCacheDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetValidationCacheDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetValidationCacheDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetValidationCacheDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetValidationCacheDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetValidationCacheDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetValidationCacheDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetValidationCacheDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetValidationCacheDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetValidationCacheDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetValidationCacheDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetValidationCacheDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetValidationCacheDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetValidationCacheDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetValidationCacheDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetValidationCacheDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetValidationCacheDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetValidationCacheDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetValidationCacheDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetValidationCacheDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetValidationCacheDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetValidationCacheDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetValidationCacheDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetValidationCacheDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetValidationCacheDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetValidationCacheDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetValidationCacheDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetValidationCacheDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetValidationCacheDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetValidationCacheDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetValidationCacheDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetValidationCacheDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetValidationCacheDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetValidationCacheDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetValidationCacheDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetValidationCacheDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetValidationCacheDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetValidationCacheDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetValidationCacheDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetValidationCacheDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetValidationCacheDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetValidationCacheDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetValidationCacheDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetValidationCacheDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetValidationCacheDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetValidationCacheDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetValidationCacheDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetValidationCacheDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetValidationCacheDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetValidationCacheDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetValidationCacheDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetValidationCacheDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_MergeValidationCachesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergeValidationCachesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergeValidationCachesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergeValidationCachesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergeValidationCachesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergeValidationCachesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergeValidationCachesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergeValidationCachesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergeValidationCachesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergeValidationCachesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergeValidationCachesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergeValidationCachesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergeValidationCachesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MergeValidationCachesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_MergeValidationCachesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_MergeValidationCachesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_MergeValidationCachesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergeValidationCachesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergeValidationCachesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergeValidationCachesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergeValidationCachesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergeValidationCachesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergeValidationCachesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergeValidationCachesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergeValidationCachesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergeValidationCachesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergeValidationCachesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergeValidationCachesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergeValidationCachesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MergeValidationCachesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_MergeValidationCachesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_MergeValidationCachesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_MergeValidationCachesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergeValidationCachesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergeValidationCachesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergeValidationCachesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergeValidationCachesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergeValidationCachesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergeValidationCachesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergeValidationCachesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergeValidationCachesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergeValidationCachesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergeValidationCachesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergeValidationCachesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergeValidationCachesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MergeValidationCachesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_MergeValidationCachesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_MergeValidationCachesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_MergeValidationCachesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergeValidationCachesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergeValidationCachesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergeValidationCachesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergeValidationCachesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergeValidationCachesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergeValidationCachesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergeValidationCachesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergeValidationCachesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergeValidationCachesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergeValidationCachesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergeValidationCachesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergeValidationCachesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MergeValidationCachesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_MergeValidationCachesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_MergeValidationCachesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_MergeValidationCachesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergeValidationCachesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergeValidationCachesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergeValidationCachesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergeValidationCachesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergeValidationCachesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergeValidationCachesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergeValidationCachesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergeValidationCachesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergeValidationCachesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergeValidationCachesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergeValidationCachesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergeValidationCachesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MergeValidationCachesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_MergeValidationCachesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_MergeValidationCachesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_MergeValidationCachesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergeValidationCachesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergeValidationCachesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergeValidationCachesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergeValidationCachesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergeValidationCachesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergeValidationCachesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergeValidationCachesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergeValidationCachesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergeValidationCachesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergeValidationCachesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergeValidationCachesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergeValidationCachesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MergeValidationCachesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_MergeValidationCachesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_MergeValidationCachesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_MergeValidationCachesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergeValidationCachesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergeValidationCachesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergeValidationCachesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergeValidationCachesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergeValidationCachesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergeValidationCachesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergeValidationCachesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergeValidationCachesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergeValidationCachesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergeValidationCachesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergeValidationCachesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergeValidationCachesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MergeValidationCachesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_MergeValidationCachesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_MergeValidationCachesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_MergeValidationCachesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergeValidationCachesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergeValidationCachesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergeValidationCachesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergeValidationCachesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergeValidationCachesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergeValidationCachesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergeValidationCachesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergeValidationCachesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergeValidationCachesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergeValidationCachesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergeValidationCachesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergeValidationCachesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MergeValidationCachesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_MergeValidationCachesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_MergeValidationCachesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_MergeValidationCachesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergeValidationCachesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergeValidationCachesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergeValidationCachesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergeValidationCachesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergeValidationCachesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergeValidationCachesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergeValidationCachesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergeValidationCachesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergeValidationCachesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergeValidationCachesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergeValidationCachesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergeValidationCachesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MergeValidationCachesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_MergeValidationCachesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_MergeValidationCachesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_MergeValidationCachesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergeValidationCachesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergeValidationCachesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergeValidationCachesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergeValidationCachesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergeValidationCachesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergeValidationCachesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergeValidationCachesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergeValidationCachesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergeValidationCachesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergeValidationCachesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergeValidationCachesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergeValidationCachesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MergeValidationCachesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_MergeValidationCachesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_MergeValidationCachesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergeValidationCachesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergeValidationCachesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergeValidationCachesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergeValidationCachesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergeValidationCachesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergeValidationCachesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergeValidationCachesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergeValidationCachesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergeValidationCachesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergeValidationCachesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergeValidationCachesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergeValidationCachesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergeValidationCachesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MergeValidationCachesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_MergeValidationCachesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_MergeValidationCachesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_MergeValidationCachesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergeValidationCachesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergeValidationCachesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergeValidationCachesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergeValidationCachesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergeValidationCachesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergeValidationCachesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergeValidationCachesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergeValidationCachesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergeValidationCachesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergeValidationCachesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergeValidationCachesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergeValidationCachesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MergeValidationCachesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_MergeValidationCachesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_MergeValidationCachesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_MergeValidationCachesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergeValidationCachesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergeValidationCachesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergeValidationCachesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergeValidationCachesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergeValidationCachesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergeValidationCachesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergeValidationCachesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergeValidationCachesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergeValidationCachesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergeValidationCachesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergeValidationCachesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergeValidationCachesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MergeValidationCachesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_MergeValidationCachesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_MergeValidationCachesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDescriptorSetLayoutSupport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDescriptorSetLayoutSupport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDescriptorSetLayoutSupport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDescriptorSetLayoutSupport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDescriptorSetLayoutSupport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDescriptorSetLayoutSupport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDescriptorSetLayoutSupport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDescriptorSetLayoutSupport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDescriptorSetLayoutSupport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDescriptorSetLayoutSupport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDescriptorSetLayoutSupport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDescriptorSetLayoutSupport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDescriptorSetLayoutSupport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDescriptorSetLayoutSupport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDescriptorSetLayoutSupport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDescriptorSetLayoutSupport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDescriptorSetLayoutSupport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDescriptorSetLayoutSupport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDescriptorSetLayoutSupport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDescriptorSetLayoutSupport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDescriptorSetLayoutSupport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDescriptorSetLayoutSupport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDescriptorSetLayoutSupport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDescriptorSetLayoutSupport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupport@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupport@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupport@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupport@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupport@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupport@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupport@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupport@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupport@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupport@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupport@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupport@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupport@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupport@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDescriptorSetLayoutSupport=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDescriptorSetLayoutSupport=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDescriptorSetLayoutSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDescriptorSetLayoutSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDescriptorSetLayoutSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDescriptorSetLayoutSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDescriptorSetLayoutSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDescriptorSetLayoutSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDescriptorSetLayoutSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDescriptorSetLayoutSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDescriptorSetLayoutSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDescriptorSetLayoutSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDescriptorSetLayoutSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDescriptorSetLayoutSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDescriptorSetLayoutSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDescriptorSetLayoutSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDescriptorSetLayoutSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDescriptorSetLayoutSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDescriptorSetLayoutSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDescriptorSetLayoutSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDescriptorSetLayoutSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDescriptorSetLayoutSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDescriptorSetLayoutSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDescriptorSetLayoutSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDescriptorSetLayoutSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDescriptorSetLayoutSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupportKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupportKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupportKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupportKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupportKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupportKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupportKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupportKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupportKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupportKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupportKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupportKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupportKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSupportKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDescriptorSetLayoutSupportKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDescriptorSetLayoutSupportKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetSwapchainGrallocUsageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetSwapchainGrallocUsageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetSwapchainGrallocUsageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetSwapchainGrallocUsageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetSwapchainGrallocUsageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetSwapchainGrallocUsageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetSwapchainGrallocUsageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetSwapchainGrallocUsageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetSwapchainGrallocUsageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetSwapchainGrallocUsageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetSwapchainGrallocUsageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetSwapchainGrallocUsageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetSwapchainGrallocUsageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetSwapchainGrallocUsageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetSwapchainGrallocUsageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetSwapchainGrallocUsageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetSwapchainGrallocUsageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetSwapchainGrallocUsageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetSwapchainGrallocUsageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetSwapchainGrallocUsageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetSwapchainGrallocUsageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetSwapchainGrallocUsageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetSwapchainGrallocUsageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetSwapchainGrallocUsageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetSwapchainGrallocUsageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetSwapchainGrallocUsageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsage2ANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsage2ANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsage2ANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsage2ANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsage2ANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsage2ANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsage2ANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsage2ANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsage2ANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsage2ANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsage2ANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsage2ANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsage2ANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainGrallocUsage2ANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetSwapchainGrallocUsage2ANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetSwapchainGrallocUsage2ANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsage2ANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsage2ANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsage2ANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsage2ANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsage2ANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsage2ANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsage2ANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsage2ANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsage2ANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsage2ANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsage2ANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsage2ANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsage2ANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainGrallocUsage2ANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetSwapchainGrallocUsage2ANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetSwapchainGrallocUsage2ANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsage2ANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsage2ANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsage2ANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsage2ANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsage2ANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsage2ANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsage2ANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsage2ANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsage2ANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsage2ANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsage2ANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsage2ANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsage2ANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainGrallocUsage2ANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetSwapchainGrallocUsage2ANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetSwapchainGrallocUsage2ANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsage2ANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsage2ANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsage2ANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsage2ANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsage2ANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsage2ANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsage2ANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsage2ANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsage2ANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsage2ANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsage2ANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsage2ANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsage2ANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainGrallocUsage2ANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetSwapchainGrallocUsage2ANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetSwapchainGrallocUsage2ANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsage2ANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsage2ANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsage2ANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsage2ANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsage2ANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsage2ANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsage2ANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsage2ANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsage2ANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsage2ANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsage2ANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsage2ANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsage2ANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainGrallocUsage2ANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetSwapchainGrallocUsage2ANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetSwapchainGrallocUsage2ANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsage2ANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsage2ANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsage2ANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsage2ANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsage2ANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsage2ANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsage2ANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsage2ANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsage2ANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsage2ANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsage2ANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsage2ANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsage2ANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainGrallocUsage2ANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetSwapchainGrallocUsage2ANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetSwapchainGrallocUsage2ANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsage2ANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsage2ANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsage2ANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsage2ANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsage2ANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsage2ANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsage2ANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsage2ANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsage2ANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsage2ANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsage2ANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsage2ANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsage2ANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainGrallocUsage2ANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetSwapchainGrallocUsage2ANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetSwapchainGrallocUsage2ANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsage2ANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsage2ANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsage2ANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsage2ANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsage2ANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsage2ANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsage2ANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsage2ANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsage2ANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsage2ANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsage2ANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsage2ANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsage2ANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainGrallocUsage2ANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetSwapchainGrallocUsage2ANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetSwapchainGrallocUsage2ANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsage2ANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsage2ANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsage2ANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsage2ANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsage2ANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsage2ANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsage2ANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsage2ANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsage2ANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsage2ANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsage2ANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsage2ANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsage2ANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainGrallocUsage2ANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetSwapchainGrallocUsage2ANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetSwapchainGrallocUsage2ANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsage2ANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsage2ANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsage2ANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsage2ANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsage2ANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsage2ANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsage2ANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsage2ANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsage2ANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsage2ANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsage2ANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsage2ANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsage2ANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainGrallocUsage2ANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetSwapchainGrallocUsage2ANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetSwapchainGrallocUsage2ANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsage2ANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsage2ANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsage2ANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsage2ANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsage2ANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsage2ANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsage2ANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsage2ANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsage2ANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsage2ANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsage2ANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsage2ANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsage2ANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainGrallocUsage2ANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetSwapchainGrallocUsage2ANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetSwapchainGrallocUsage2ANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsage2ANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsage2ANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsage2ANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsage2ANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsage2ANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsage2ANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsage2ANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsage2ANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsage2ANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsage2ANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsage2ANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsage2ANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsage2ANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainGrallocUsage2ANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetSwapchainGrallocUsage2ANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetSwapchainGrallocUsage2ANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsage2ANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsage2ANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsage2ANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsage2ANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsage2ANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsage2ANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsage2ANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsage2ANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsage2ANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsage2ANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsage2ANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsage2ANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsage2ANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainGrallocUsage2ANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetSwapchainGrallocUsage2ANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetSwapchainGrallocUsage2ANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_AcquireImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_AcquireImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_AcquireImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_AcquireImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_AcquireImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_AcquireImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_AcquireImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_AcquireImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_AcquireImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_AcquireImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_AcquireImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_AcquireImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_AcquireImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_AcquireImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_AcquireImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_AcquireImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_AcquireImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_AcquireImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_AcquireImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_AcquireImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_AcquireImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_AcquireImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_AcquireImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_AcquireImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_AcquireImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_AcquireImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_AcquireImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_AcquireImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_AcquireImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_AcquireImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_AcquireImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_AcquireImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_AcquireImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_AcquireImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_AcquireImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_QueueSignalReleaseImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSignalReleaseImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSignalReleaseImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSignalReleaseImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSignalReleaseImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSignalReleaseImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSignalReleaseImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSignalReleaseImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSignalReleaseImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSignalReleaseImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSignalReleaseImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSignalReleaseImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSignalReleaseImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSignalReleaseImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_QueueSignalReleaseImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_QueueSignalReleaseImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_QueueSignalReleaseImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSignalReleaseImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSignalReleaseImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSignalReleaseImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSignalReleaseImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSignalReleaseImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSignalReleaseImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSignalReleaseImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSignalReleaseImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSignalReleaseImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSignalReleaseImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSignalReleaseImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSignalReleaseImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSignalReleaseImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_QueueSignalReleaseImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_QueueSignalReleaseImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_QueueSignalReleaseImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSignalReleaseImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSignalReleaseImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSignalReleaseImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSignalReleaseImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSignalReleaseImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSignalReleaseImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSignalReleaseImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSignalReleaseImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSignalReleaseImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSignalReleaseImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSignalReleaseImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSignalReleaseImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSignalReleaseImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_QueueSignalReleaseImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_QueueSignalReleaseImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_QueueSignalReleaseImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSignalReleaseImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSignalReleaseImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSignalReleaseImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSignalReleaseImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSignalReleaseImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSignalReleaseImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSignalReleaseImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSignalReleaseImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSignalReleaseImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSignalReleaseImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSignalReleaseImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSignalReleaseImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSignalReleaseImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_QueueSignalReleaseImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_QueueSignalReleaseImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_QueueSignalReleaseImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSignalReleaseImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSignalReleaseImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSignalReleaseImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSignalReleaseImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSignalReleaseImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSignalReleaseImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSignalReleaseImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSignalReleaseImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSignalReleaseImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSignalReleaseImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSignalReleaseImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSignalReleaseImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSignalReleaseImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_QueueSignalReleaseImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_QueueSignalReleaseImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_QueueSignalReleaseImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSignalReleaseImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSignalReleaseImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSignalReleaseImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSignalReleaseImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSignalReleaseImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSignalReleaseImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSignalReleaseImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSignalReleaseImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSignalReleaseImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSignalReleaseImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSignalReleaseImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSignalReleaseImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSignalReleaseImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_QueueSignalReleaseImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_QueueSignalReleaseImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_QueueSignalReleaseImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSignalReleaseImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSignalReleaseImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSignalReleaseImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSignalReleaseImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSignalReleaseImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSignalReleaseImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSignalReleaseImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSignalReleaseImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSignalReleaseImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSignalReleaseImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSignalReleaseImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSignalReleaseImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSignalReleaseImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_QueueSignalReleaseImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_QueueSignalReleaseImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_QueueSignalReleaseImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSignalReleaseImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSignalReleaseImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSignalReleaseImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSignalReleaseImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSignalReleaseImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSignalReleaseImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSignalReleaseImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSignalReleaseImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSignalReleaseImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSignalReleaseImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSignalReleaseImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSignalReleaseImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSignalReleaseImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_QueueSignalReleaseImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_QueueSignalReleaseImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSignalReleaseImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSignalReleaseImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSignalReleaseImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSignalReleaseImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSignalReleaseImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSignalReleaseImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSignalReleaseImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSignalReleaseImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSignalReleaseImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSignalReleaseImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSignalReleaseImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSignalReleaseImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSignalReleaseImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSignalReleaseImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_QueueSignalReleaseImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_QueueSignalReleaseImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSignalReleaseImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSignalReleaseImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSignalReleaseImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSignalReleaseImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSignalReleaseImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSignalReleaseImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSignalReleaseImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSignalReleaseImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSignalReleaseImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSignalReleaseImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSignalReleaseImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSignalReleaseImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSignalReleaseImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSignalReleaseImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_QueueSignalReleaseImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_QueueSignalReleaseImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSignalReleaseImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSignalReleaseImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSignalReleaseImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSignalReleaseImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSignalReleaseImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSignalReleaseImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSignalReleaseImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSignalReleaseImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSignalReleaseImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSignalReleaseImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSignalReleaseImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSignalReleaseImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSignalReleaseImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSignalReleaseImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_QueueSignalReleaseImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_QueueSignalReleaseImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_QueueSignalReleaseImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSignalReleaseImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSignalReleaseImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSignalReleaseImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSignalReleaseImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSignalReleaseImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSignalReleaseImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSignalReleaseImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSignalReleaseImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSignalReleaseImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSignalReleaseImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSignalReleaseImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSignalReleaseImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSignalReleaseImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_QueueSignalReleaseImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_QueueSignalReleaseImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSignalReleaseImageANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSignalReleaseImageANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSignalReleaseImageANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSignalReleaseImageANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSignalReleaseImageANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSignalReleaseImageANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSignalReleaseImageANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSignalReleaseImageANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSignalReleaseImageANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSignalReleaseImageANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSignalReleaseImageANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSignalReleaseImageANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSignalReleaseImageANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSignalReleaseImageANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_QueueSignalReleaseImageANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_QueueSignalReleaseImageANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetShaderInfoAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInfoAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInfoAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInfoAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInfoAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInfoAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInfoAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInfoAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInfoAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInfoAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInfoAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInfoAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInfoAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInfoAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetShaderInfoAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetShaderInfoAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInfoAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInfoAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInfoAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInfoAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInfoAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInfoAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInfoAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInfoAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInfoAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInfoAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInfoAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInfoAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInfoAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInfoAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetShaderInfoAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetShaderInfoAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInfoAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInfoAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInfoAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInfoAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInfoAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInfoAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInfoAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInfoAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInfoAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInfoAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInfoAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInfoAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInfoAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInfoAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetShaderInfoAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetShaderInfoAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInfoAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInfoAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInfoAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInfoAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInfoAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInfoAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInfoAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInfoAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInfoAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInfoAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInfoAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInfoAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInfoAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInfoAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetShaderInfoAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetShaderInfoAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInfoAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInfoAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInfoAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInfoAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInfoAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInfoAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInfoAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInfoAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInfoAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInfoAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInfoAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInfoAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInfoAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInfoAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetShaderInfoAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetShaderInfoAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInfoAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInfoAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInfoAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInfoAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInfoAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInfoAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInfoAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInfoAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInfoAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInfoAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInfoAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInfoAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInfoAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInfoAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetShaderInfoAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetShaderInfoAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInfoAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInfoAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInfoAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInfoAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInfoAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInfoAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInfoAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInfoAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInfoAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInfoAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInfoAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInfoAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInfoAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInfoAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetShaderInfoAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetShaderInfoAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInfoAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInfoAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInfoAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInfoAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInfoAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInfoAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInfoAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInfoAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInfoAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInfoAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInfoAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInfoAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInfoAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInfoAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetShaderInfoAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetShaderInfoAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInfoAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInfoAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInfoAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInfoAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInfoAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInfoAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInfoAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInfoAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInfoAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInfoAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInfoAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInfoAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInfoAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInfoAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetShaderInfoAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetShaderInfoAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInfoAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInfoAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInfoAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInfoAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInfoAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInfoAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInfoAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInfoAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInfoAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInfoAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInfoAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInfoAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInfoAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInfoAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetShaderInfoAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetShaderInfoAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInfoAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInfoAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInfoAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInfoAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInfoAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInfoAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInfoAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInfoAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInfoAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInfoAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInfoAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInfoAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInfoAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInfoAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetShaderInfoAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetShaderInfoAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInfoAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInfoAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInfoAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInfoAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInfoAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInfoAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInfoAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInfoAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInfoAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInfoAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInfoAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInfoAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInfoAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInfoAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetShaderInfoAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetShaderInfoAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInfoAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInfoAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInfoAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInfoAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInfoAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInfoAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInfoAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInfoAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInfoAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInfoAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInfoAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInfoAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInfoAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInfoAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetShaderInfoAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetShaderInfoAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_SetLocalDimmingAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLocalDimmingAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLocalDimmingAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLocalDimmingAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLocalDimmingAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLocalDimmingAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLocalDimmingAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLocalDimmingAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLocalDimmingAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLocalDimmingAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLocalDimmingAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLocalDimmingAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLocalDimmingAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLocalDimmingAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_SetLocalDimmingAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_SetLocalDimmingAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_SetLocalDimmingAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLocalDimmingAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLocalDimmingAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLocalDimmingAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLocalDimmingAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLocalDimmingAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLocalDimmingAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLocalDimmingAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLocalDimmingAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLocalDimmingAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLocalDimmingAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLocalDimmingAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLocalDimmingAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLocalDimmingAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_SetLocalDimmingAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_SetLocalDimmingAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_SetLocalDimmingAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLocalDimmingAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLocalDimmingAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLocalDimmingAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLocalDimmingAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLocalDimmingAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLocalDimmingAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLocalDimmingAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLocalDimmingAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLocalDimmingAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLocalDimmingAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLocalDimmingAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLocalDimmingAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLocalDimmingAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_SetLocalDimmingAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_SetLocalDimmingAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_SetLocalDimmingAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLocalDimmingAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLocalDimmingAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLocalDimmingAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLocalDimmingAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLocalDimmingAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLocalDimmingAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLocalDimmingAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLocalDimmingAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLocalDimmingAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLocalDimmingAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLocalDimmingAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLocalDimmingAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLocalDimmingAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_SetLocalDimmingAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_SetLocalDimmingAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_SetLocalDimmingAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLocalDimmingAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLocalDimmingAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLocalDimmingAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLocalDimmingAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLocalDimmingAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLocalDimmingAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLocalDimmingAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLocalDimmingAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLocalDimmingAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLocalDimmingAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLocalDimmingAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLocalDimmingAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLocalDimmingAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_SetLocalDimmingAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_SetLocalDimmingAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_SetLocalDimmingAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLocalDimmingAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLocalDimmingAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLocalDimmingAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLocalDimmingAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLocalDimmingAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLocalDimmingAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLocalDimmingAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLocalDimmingAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLocalDimmingAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLocalDimmingAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLocalDimmingAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLocalDimmingAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLocalDimmingAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_SetLocalDimmingAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_SetLocalDimmingAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_SetLocalDimmingAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLocalDimmingAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLocalDimmingAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLocalDimmingAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLocalDimmingAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLocalDimmingAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLocalDimmingAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLocalDimmingAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLocalDimmingAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLocalDimmingAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLocalDimmingAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLocalDimmingAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLocalDimmingAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLocalDimmingAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_SetLocalDimmingAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_SetLocalDimmingAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_SetLocalDimmingAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLocalDimmingAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLocalDimmingAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLocalDimmingAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLocalDimmingAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLocalDimmingAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLocalDimmingAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLocalDimmingAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLocalDimmingAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLocalDimmingAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLocalDimmingAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLocalDimmingAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLocalDimmingAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLocalDimmingAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_SetLocalDimmingAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_SetLocalDimmingAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLocalDimmingAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLocalDimmingAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLocalDimmingAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLocalDimmingAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLocalDimmingAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLocalDimmingAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLocalDimmingAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLocalDimmingAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLocalDimmingAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLocalDimmingAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLocalDimmingAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLocalDimmingAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLocalDimmingAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLocalDimmingAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_SetLocalDimmingAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_SetLocalDimmingAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLocalDimmingAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLocalDimmingAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLocalDimmingAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLocalDimmingAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLocalDimmingAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLocalDimmingAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLocalDimmingAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLocalDimmingAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLocalDimmingAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLocalDimmingAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLocalDimmingAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLocalDimmingAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLocalDimmingAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLocalDimmingAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_SetLocalDimmingAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_SetLocalDimmingAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLocalDimmingAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLocalDimmingAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLocalDimmingAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLocalDimmingAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLocalDimmingAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLocalDimmingAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLocalDimmingAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLocalDimmingAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLocalDimmingAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLocalDimmingAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLocalDimmingAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLocalDimmingAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLocalDimmingAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLocalDimmingAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_SetLocalDimmingAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_SetLocalDimmingAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_SetLocalDimmingAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLocalDimmingAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLocalDimmingAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLocalDimmingAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLocalDimmingAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLocalDimmingAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLocalDimmingAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLocalDimmingAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLocalDimmingAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLocalDimmingAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLocalDimmingAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLocalDimmingAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLocalDimmingAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLocalDimmingAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_SetLocalDimmingAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_SetLocalDimmingAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLocalDimmingAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLocalDimmingAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLocalDimmingAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLocalDimmingAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLocalDimmingAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLocalDimmingAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLocalDimmingAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLocalDimmingAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLocalDimmingAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLocalDimmingAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLocalDimmingAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLocalDimmingAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLocalDimmingAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLocalDimmingAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_SetLocalDimmingAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_SetLocalDimmingAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetCalibratedTimestampsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetCalibratedTimestampsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetCalibratedTimestampsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetCalibratedTimestampsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetCalibratedTimestampsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetCalibratedTimestampsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetCalibratedTimestampsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetCalibratedTimestampsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetCalibratedTimestampsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetCalibratedTimestampsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetCalibratedTimestampsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetCalibratedTimestampsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetCalibratedTimestampsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetCalibratedTimestampsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetCalibratedTimestampsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetCalibratedTimestampsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetCalibratedTimestampsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetCalibratedTimestampsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetCalibratedTimestampsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetCalibratedTimestampsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetCalibratedTimestampsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetCalibratedTimestampsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetCalibratedTimestampsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetCalibratedTimestampsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetCalibratedTimestampsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetCalibratedTimestampsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetCalibratedTimestampsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetCalibratedTimestampsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetCalibratedTimestampsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetCalibratedTimestampsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetCalibratedTimestampsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetCalibratedTimestampsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetCalibratedTimestampsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetCalibratedTimestampsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetCalibratedTimestampsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetCalibratedTimestampsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetCalibratedTimestampsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetCalibratedTimestampsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetCalibratedTimestampsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetCalibratedTimestampsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetCalibratedTimestampsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetCalibratedTimestampsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetCalibratedTimestampsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetCalibratedTimestampsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetCalibratedTimestampsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetCalibratedTimestampsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetCalibratedTimestampsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetCalibratedTimestampsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetCalibratedTimestampsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetCalibratedTimestampsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetCalibratedTimestampsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetCalibratedTimestampsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetCalibratedTimestampsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetCalibratedTimestampsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetCalibratedTimestampsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetCalibratedTimestampsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetCalibratedTimestampsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetCalibratedTimestampsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetCalibratedTimestampsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetCalibratedTimestampsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetCalibratedTimestampsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetCalibratedTimestampsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetCalibratedTimestampsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetCalibratedTimestampsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetCalibratedTimestampsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_SetDebugUtilsObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_SetDebugUtilsObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_SetDebugUtilsObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_SetDebugUtilsObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_SetDebugUtilsObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_SetDebugUtilsObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_SetDebugUtilsObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_SetDebugUtilsObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_SetDebugUtilsObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_SetDebugUtilsObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_SetDebugUtilsObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_SetDebugUtilsObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_SetDebugUtilsObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_SetDebugUtilsObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_SetDebugUtilsObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_SetDebugUtilsObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_SetDebugUtilsObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_SetDebugUtilsObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_SetDebugUtilsObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_SetDebugUtilsObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_SetDebugUtilsObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_SetDebugUtilsObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_SetDebugUtilsObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_SetDebugUtilsObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectNameEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectNameEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectNameEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectNameEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectNameEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectNameEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectNameEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectNameEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectNameEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectNameEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectNameEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectNameEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectNameEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectNameEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_SetDebugUtilsObjectNameEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_SetDebugUtilsObjectNameEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDebugUtilsObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_SetDebugUtilsObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_SetDebugUtilsObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDebugUtilsObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_SetDebugUtilsObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_SetDebugUtilsObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDebugUtilsObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_SetDebugUtilsObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_SetDebugUtilsObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDebugUtilsObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_SetDebugUtilsObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_SetDebugUtilsObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDebugUtilsObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_SetDebugUtilsObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_SetDebugUtilsObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDebugUtilsObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_SetDebugUtilsObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_SetDebugUtilsObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDebugUtilsObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_SetDebugUtilsObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_SetDebugUtilsObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDebugUtilsObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_SetDebugUtilsObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_SetDebugUtilsObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDebugUtilsObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_SetDebugUtilsObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_SetDebugUtilsObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDebugUtilsObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_SetDebugUtilsObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_SetDebugUtilsObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDebugUtilsObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_SetDebugUtilsObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_SetDebugUtilsObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDebugUtilsObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_SetDebugUtilsObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_SetDebugUtilsObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectTagEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectTagEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectTagEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectTagEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectTagEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectTagEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectTagEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectTagEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectTagEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectTagEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectTagEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectTagEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectTagEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDebugUtilsObjectTagEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_SetDebugUtilsObjectTagEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_SetDebugUtilsObjectTagEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_QueueBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_QueueBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_QueueBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_QueueBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_QueueBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_QueueBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_QueueBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_QueueBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_QueueBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_QueueBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_QueueBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_QueueBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_QueueBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_QueueBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_QueueBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_QueueBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_QueueBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_QueueBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_QueueBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_QueueBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_QueueBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_QueueBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_QueueBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_QueueBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_QueueBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_QueueBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_QueueBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_QueueBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_QueueBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_QueueBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_QueueBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_QueueBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_QueueBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_QueueBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_QueueBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_QueueEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_QueueEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_QueueEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_QueueEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_QueueEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_QueueEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_QueueEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_QueueEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_QueueEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_QueueEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_QueueEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_QueueEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_QueueEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_QueueEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_QueueEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_QueueEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_QueueEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_QueueEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_QueueEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_QueueEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_QueueEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_QueueEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_QueueEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_QueueEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_QueueEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_QueueEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_QueueEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_QueueEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_QueueEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_QueueEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_QueueEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_QueueEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_QueueEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_QueueEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_QueueEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_QueueInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_QueueInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_QueueInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_QueueInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_QueueInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_QueueInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_QueueInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_QueueInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_QueueInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_QueueInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_QueueInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_QueueInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_QueueInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_QueueInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_QueueInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_QueueInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_QueueInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_QueueInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_QueueInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_QueueInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_QueueInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_QueueInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_QueueInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_QueueInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_QueueInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_QueueInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_QueueInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_QueueInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_QueueInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_QueueInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_QueueInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_QueueInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_QueueInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_QueueInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_QueueInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBeginDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBeginDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdEndDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdEndDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdInsertDebugUtilsLabelEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdInsertDebugUtilsLabelEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdInsertDebugUtilsLabelEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdInsertDebugUtilsLabelEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdInsertDebugUtilsLabelEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdInsertDebugUtilsLabelEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdInsertDebugUtilsLabelEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdInsertDebugUtilsLabelEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdInsertDebugUtilsLabelEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdInsertDebugUtilsLabelEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdInsertDebugUtilsLabelEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdInsertDebugUtilsLabelEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdInsertDebugUtilsLabelEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdInsertDebugUtilsLabelEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdInsertDebugUtilsLabelEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdInsertDebugUtilsLabelEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetMemoryHostPointerPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryHostPointerPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryHostPointerPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryHostPointerPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryHostPointerPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryHostPointerPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryHostPointerPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryHostPointerPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryHostPointerPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryHostPointerPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryHostPointerPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryHostPointerPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryHostPointerPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryHostPointerPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetMemoryHostPointerPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetMemoryHostPointerPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryHostPointerPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryHostPointerPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryHostPointerPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryHostPointerPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryHostPointerPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryHostPointerPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryHostPointerPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryHostPointerPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryHostPointerPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryHostPointerPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryHostPointerPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryHostPointerPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryHostPointerPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryHostPointerPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetMemoryHostPointerPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetMemoryHostPointerPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryHostPointerPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryHostPointerPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryHostPointerPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryHostPointerPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryHostPointerPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryHostPointerPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryHostPointerPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryHostPointerPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryHostPointerPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryHostPointerPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryHostPointerPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryHostPointerPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryHostPointerPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryHostPointerPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetMemoryHostPointerPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetMemoryHostPointerPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryHostPointerPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryHostPointerPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryHostPointerPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryHostPointerPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryHostPointerPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryHostPointerPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryHostPointerPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryHostPointerPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryHostPointerPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryHostPointerPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryHostPointerPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryHostPointerPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryHostPointerPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryHostPointerPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetMemoryHostPointerPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetMemoryHostPointerPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryHostPointerPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryHostPointerPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryHostPointerPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryHostPointerPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryHostPointerPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryHostPointerPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryHostPointerPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryHostPointerPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryHostPointerPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryHostPointerPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryHostPointerPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryHostPointerPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryHostPointerPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryHostPointerPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetMemoryHostPointerPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetMemoryHostPointerPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryHostPointerPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryHostPointerPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryHostPointerPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryHostPointerPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryHostPointerPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryHostPointerPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryHostPointerPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryHostPointerPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryHostPointerPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryHostPointerPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryHostPointerPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryHostPointerPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryHostPointerPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryHostPointerPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetMemoryHostPointerPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetMemoryHostPointerPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryHostPointerPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryHostPointerPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryHostPointerPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryHostPointerPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryHostPointerPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryHostPointerPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryHostPointerPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryHostPointerPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryHostPointerPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryHostPointerPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryHostPointerPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryHostPointerPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryHostPointerPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryHostPointerPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetMemoryHostPointerPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetMemoryHostPointerPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryHostPointerPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryHostPointerPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryHostPointerPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryHostPointerPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryHostPointerPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryHostPointerPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryHostPointerPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryHostPointerPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryHostPointerPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryHostPointerPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryHostPointerPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryHostPointerPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryHostPointerPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryHostPointerPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetMemoryHostPointerPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetMemoryHostPointerPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryHostPointerPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryHostPointerPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryHostPointerPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryHostPointerPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryHostPointerPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryHostPointerPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryHostPointerPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryHostPointerPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryHostPointerPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryHostPointerPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryHostPointerPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryHostPointerPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryHostPointerPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryHostPointerPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetMemoryHostPointerPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetMemoryHostPointerPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryHostPointerPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryHostPointerPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryHostPointerPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryHostPointerPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryHostPointerPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryHostPointerPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryHostPointerPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryHostPointerPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryHostPointerPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryHostPointerPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryHostPointerPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryHostPointerPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryHostPointerPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryHostPointerPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetMemoryHostPointerPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetMemoryHostPointerPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryHostPointerPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryHostPointerPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryHostPointerPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryHostPointerPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryHostPointerPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryHostPointerPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryHostPointerPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryHostPointerPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryHostPointerPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryHostPointerPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryHostPointerPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryHostPointerPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryHostPointerPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryHostPointerPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetMemoryHostPointerPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetMemoryHostPointerPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryHostPointerPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryHostPointerPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryHostPointerPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryHostPointerPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryHostPointerPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryHostPointerPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryHostPointerPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryHostPointerPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryHostPointerPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryHostPointerPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryHostPointerPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryHostPointerPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryHostPointerPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryHostPointerPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetMemoryHostPointerPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetMemoryHostPointerPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryHostPointerPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryHostPointerPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryHostPointerPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryHostPointerPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryHostPointerPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryHostPointerPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryHostPointerPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryHostPointerPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryHostPointerPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryHostPointerPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryHostPointerPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryHostPointerPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryHostPointerPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryHostPointerPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetMemoryHostPointerPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetMemoryHostPointerPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarkerAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarkerAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarkerAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarkerAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarkerAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarkerAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarkerAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarkerAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarkerAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarkerAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarkerAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarkerAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarkerAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarkerAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdWriteBufferMarkerAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdWriteBufferMarkerAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarkerAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarkerAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarkerAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarkerAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarkerAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarkerAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarkerAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarkerAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarkerAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarkerAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarkerAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarkerAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarkerAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarkerAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdWriteBufferMarkerAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdWriteBufferMarkerAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarkerAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarkerAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarkerAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarkerAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarkerAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarkerAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarkerAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarkerAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarkerAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarkerAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarkerAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarkerAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarkerAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarkerAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdWriteBufferMarkerAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdWriteBufferMarkerAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarkerAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarkerAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarkerAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarkerAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarkerAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarkerAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarkerAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarkerAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarkerAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarkerAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarkerAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarkerAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarkerAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarkerAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdWriteBufferMarkerAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdWriteBufferMarkerAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarkerAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarkerAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarkerAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarkerAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarkerAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarkerAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarkerAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarkerAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarkerAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarkerAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarkerAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarkerAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarkerAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarkerAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdWriteBufferMarkerAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdWriteBufferMarkerAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarkerAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarkerAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarkerAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarkerAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarkerAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarkerAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarkerAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarkerAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarkerAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarkerAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarkerAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarkerAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarkerAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarkerAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdWriteBufferMarkerAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdWriteBufferMarkerAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarkerAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarkerAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarkerAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarkerAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarkerAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarkerAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarkerAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarkerAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarkerAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarkerAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarkerAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarkerAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarkerAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarkerAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdWriteBufferMarkerAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdWriteBufferMarkerAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarkerAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarkerAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarkerAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarkerAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarkerAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarkerAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarkerAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarkerAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarkerAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarkerAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarkerAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarkerAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarkerAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarkerAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdWriteBufferMarkerAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdWriteBufferMarkerAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarkerAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarkerAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarkerAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarkerAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarkerAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarkerAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarkerAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarkerAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarkerAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarkerAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarkerAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarkerAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarkerAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarkerAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdWriteBufferMarkerAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdWriteBufferMarkerAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarkerAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarkerAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarkerAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarkerAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarkerAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarkerAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarkerAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarkerAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarkerAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarkerAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarkerAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarkerAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarkerAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarkerAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdWriteBufferMarkerAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdWriteBufferMarkerAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarkerAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarkerAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarkerAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarkerAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarkerAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarkerAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarkerAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarkerAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarkerAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarkerAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarkerAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarkerAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarkerAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarkerAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdWriteBufferMarkerAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdWriteBufferMarkerAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarkerAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarkerAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarkerAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarkerAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarkerAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarkerAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarkerAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarkerAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarkerAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarkerAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarkerAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarkerAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarkerAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarkerAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdWriteBufferMarkerAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdWriteBufferMarkerAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarkerAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarkerAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarkerAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarkerAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarkerAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarkerAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarkerAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarkerAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarkerAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarkerAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarkerAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarkerAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarkerAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarkerAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdWriteBufferMarkerAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdWriteBufferMarkerAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBeginRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBeginRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBeginRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBeginRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBeginRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBeginRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBeginRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBeginRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBeginRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBeginRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBeginRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBeginRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBeginRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBeginRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBeginRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBeginRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBeginRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBeginRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBeginRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBeginRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBeginRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBeginRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBeginRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBeginRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBeginRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBeginRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBeginRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBeginRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBeginRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBeginRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBeginRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBeginRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBeginRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBeginRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBeginRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBeginRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBeginRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBeginRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBeginRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBeginRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBeginRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBeginRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBeginRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBeginRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBeginRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBeginRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBeginRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBeginRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBeginRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBeginRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBeginRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBeginRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdNextSubpass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdNextSubpass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdNextSubpass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdNextSubpass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdNextSubpass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdNextSubpass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdNextSubpass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdNextSubpass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdNextSubpass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdNextSubpass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdNextSubpass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdNextSubpass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdNextSubpass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdNextSubpass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdNextSubpass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdNextSubpass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdNextSubpass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdNextSubpass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdNextSubpass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdNextSubpass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdNextSubpass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdNextSubpass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdNextSubpass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdNextSubpass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdNextSubpass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdNextSubpass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdNextSubpass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdNextSubpass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdNextSubpass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdNextSubpass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdNextSubpass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdNextSubpass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdNextSubpass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdNextSubpass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdNextSubpass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdNextSubpass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdNextSubpass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdNextSubpass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdNextSubpass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdNextSubpass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdNextSubpass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdNextSubpass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdNextSubpass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdNextSubpass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdNextSubpass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdNextSubpass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdNextSubpass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdNextSubpass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdNextSubpass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdNextSubpass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdNextSubpass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdNextSubpass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdNextSubpass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdNextSubpass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdNextSubpass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdNextSubpass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdNextSubpass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdNextSubpass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdNextSubpass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdNextSubpass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdNextSubpass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdNextSubpass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdNextSubpass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdNextSubpass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdNextSubpass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdEndRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdEndRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdEndRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdEndRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdEndRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdEndRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdEndRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdEndRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdEndRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdEndRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdEndRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdEndRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdEndRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdEndRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdEndRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdEndRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdEndRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdEndRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdEndRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdEndRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdEndRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdEndRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdEndRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdEndRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdEndRenderPass2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdEndRenderPass2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdEndRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdEndRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdEndRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdEndRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdEndRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdEndRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdEndRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdEndRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdEndRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdEndRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdEndRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdEndRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdEndRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdEndRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdEndRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdEndRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdEndRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdEndRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdEndRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdEndRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdEndRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdEndRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdEndRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdEndRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderPass2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdEndRenderPass2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdEndRenderPass2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetSemaphoreCounterValue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetSemaphoreCounterValue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetSemaphoreCounterValue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetSemaphoreCounterValue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetSemaphoreCounterValue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetSemaphoreCounterValue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetSemaphoreCounterValue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetSemaphoreCounterValue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetSemaphoreCounterValue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetSemaphoreCounterValue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetSemaphoreCounterValue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetSemaphoreCounterValue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetSemaphoreCounterValue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetSemaphoreCounterValue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetSemaphoreCounterValue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetSemaphoreCounterValue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetSemaphoreCounterValue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetSemaphoreCounterValue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetSemaphoreCounterValue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetSemaphoreCounterValue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetSemaphoreCounterValue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetSemaphoreCounterValue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetSemaphoreCounterValue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetSemaphoreCounterValue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValue@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValue@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValue@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValue@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValue@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValue@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValue@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValue@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValue@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValue@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValue@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValue@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValue@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValue@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetSemaphoreCounterValue=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetSemaphoreCounterValue=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValueKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValueKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValueKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValueKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValueKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValueKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValueKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValueKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValueKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValueKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValueKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValueKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValueKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSemaphoreCounterValueKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetSemaphoreCounterValueKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetSemaphoreCounterValueKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValueKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValueKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValueKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValueKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValueKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValueKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValueKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValueKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValueKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValueKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValueKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValueKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValueKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSemaphoreCounterValueKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetSemaphoreCounterValueKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetSemaphoreCounterValueKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValueKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValueKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValueKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValueKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValueKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValueKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValueKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValueKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValueKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValueKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValueKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValueKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValueKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSemaphoreCounterValueKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetSemaphoreCounterValueKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetSemaphoreCounterValueKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValueKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValueKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValueKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValueKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValueKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValueKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValueKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValueKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValueKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValueKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValueKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValueKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValueKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSemaphoreCounterValueKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetSemaphoreCounterValueKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetSemaphoreCounterValueKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValueKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValueKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValueKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValueKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValueKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValueKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValueKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValueKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValueKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValueKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValueKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValueKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValueKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSemaphoreCounterValueKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetSemaphoreCounterValueKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetSemaphoreCounterValueKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValueKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValueKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValueKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValueKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValueKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValueKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValueKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValueKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValueKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValueKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValueKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValueKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValueKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSemaphoreCounterValueKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetSemaphoreCounterValueKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetSemaphoreCounterValueKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValueKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValueKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValueKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValueKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValueKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValueKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValueKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValueKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValueKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValueKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValueKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValueKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValueKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSemaphoreCounterValueKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetSemaphoreCounterValueKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetSemaphoreCounterValueKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValueKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValueKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValueKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValueKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValueKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValueKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValueKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValueKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValueKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValueKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValueKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValueKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValueKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSemaphoreCounterValueKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetSemaphoreCounterValueKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetSemaphoreCounterValueKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValueKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValueKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValueKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValueKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValueKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValueKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValueKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValueKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValueKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValueKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValueKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValueKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValueKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSemaphoreCounterValueKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetSemaphoreCounterValueKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetSemaphoreCounterValueKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValueKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValueKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValueKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValueKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValueKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValueKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValueKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValueKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValueKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValueKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValueKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValueKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValueKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSemaphoreCounterValueKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetSemaphoreCounterValueKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetSemaphoreCounterValueKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValueKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValueKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValueKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValueKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValueKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValueKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValueKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValueKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValueKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValueKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValueKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValueKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValueKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSemaphoreCounterValueKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetSemaphoreCounterValueKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetSemaphoreCounterValueKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValueKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValueKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValueKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValueKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValueKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValueKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValueKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValueKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValueKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValueKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValueKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValueKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValueKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSemaphoreCounterValueKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetSemaphoreCounterValueKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetSemaphoreCounterValueKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValueKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValueKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValueKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValueKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValueKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValueKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValueKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValueKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValueKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValueKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValueKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValueKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValueKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSemaphoreCounterValueKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetSemaphoreCounterValueKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetSemaphoreCounterValueKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphores@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphores@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphores@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphores@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphores@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphores@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphores@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphores@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphores@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphores@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphores@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphores@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphores@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphores@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_WaitSemaphores=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_WaitSemaphores=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphores@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphores@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphores@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphores@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphores@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphores@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphores@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphores@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphores@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphores@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphores@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphores@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphores@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphores@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_WaitSemaphores=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_WaitSemaphores=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphores@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphores@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphores@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphores@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphores@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphores@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphores@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphores@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphores@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphores@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphores@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphores@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphores@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphores@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_WaitSemaphores=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_WaitSemaphores=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphores@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphores@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphores@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphores@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphores@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphores@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphores@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphores@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphores@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphores@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphores@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphores@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphores@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphores@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_WaitSemaphores=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_WaitSemaphores=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphores@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphores@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphores@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphores@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphores@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphores@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphores@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphores@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphores@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphores@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphores@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphores@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphores@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphores@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_WaitSemaphores=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_WaitSemaphores=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphores@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphores@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphores@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphores@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphores@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphores@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphores@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphores@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphores@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphores@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphores@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphores@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphores@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphores@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_WaitSemaphores=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_WaitSemaphores=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphores@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphores@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphores@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphores@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphores@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphores@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphores@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphores@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphores@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphores@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphores@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphores@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphores@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphores@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_WaitSemaphores=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_WaitSemaphores=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphores@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphores@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphores@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphores@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphores@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphores@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphores@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphores@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphores@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphores@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphores@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphores@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphores@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphores@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_WaitSemaphores=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_WaitSemaphores=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphores@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphores@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphores@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphores@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphores@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphores@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphores@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphores@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphores@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphores@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphores@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphores@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphores@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphores@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_WaitSemaphores=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_WaitSemaphores=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphores@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphores@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphores@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphores@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphores@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphores@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphores@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphores@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphores@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphores@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphores@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphores@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphores@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphores@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_WaitSemaphores=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_WaitSemaphores=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphores@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphores@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphores@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphores@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphores@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphores@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphores@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphores@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphores@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphores@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphores@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphores@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphores@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphores@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_WaitSemaphores=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_WaitSemaphores=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphores@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphores@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphores@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphores@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphores@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphores@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphores@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphores@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphores@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphores@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphores@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphores@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphores@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphores@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_WaitSemaphores=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_WaitSemaphores=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphores@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphores@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphores@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphores@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphores@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphores@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphores@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphores@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphores@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphores@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphores@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphores@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphores@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphores@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_WaitSemaphores=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_WaitSemaphores=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphoresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphoresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphoresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphoresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphoresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphoresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphoresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphoresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphoresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphoresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphoresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphoresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphoresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitSemaphoresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_WaitSemaphoresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_WaitSemaphoresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphoresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphoresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphoresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphoresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphoresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphoresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphoresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphoresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphoresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphoresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphoresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphoresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphoresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitSemaphoresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_WaitSemaphoresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_WaitSemaphoresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphoresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphoresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphoresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphoresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphoresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphoresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphoresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphoresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphoresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphoresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphoresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphoresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphoresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitSemaphoresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_WaitSemaphoresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_WaitSemaphoresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphoresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphoresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphoresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphoresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphoresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphoresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphoresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphoresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphoresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphoresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphoresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphoresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphoresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitSemaphoresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_WaitSemaphoresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_WaitSemaphoresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphoresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphoresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphoresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphoresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphoresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphoresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphoresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphoresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphoresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphoresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphoresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphoresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphoresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitSemaphoresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_WaitSemaphoresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_WaitSemaphoresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphoresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphoresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphoresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphoresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphoresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphoresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphoresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphoresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphoresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphoresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphoresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphoresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphoresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitSemaphoresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_WaitSemaphoresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_WaitSemaphoresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphoresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphoresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphoresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphoresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphoresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphoresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphoresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphoresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphoresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphoresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphoresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphoresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphoresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitSemaphoresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_WaitSemaphoresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_WaitSemaphoresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphoresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphoresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphoresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphoresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphoresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphoresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphoresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphoresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphoresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphoresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphoresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphoresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphoresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitSemaphoresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_WaitSemaphoresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_WaitSemaphoresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphoresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphoresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphoresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphoresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphoresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphoresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphoresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphoresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphoresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphoresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphoresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphoresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphoresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitSemaphoresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_WaitSemaphoresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_WaitSemaphoresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphoresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphoresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphoresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphoresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphoresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphoresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphoresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphoresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphoresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphoresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphoresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphoresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphoresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitSemaphoresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_WaitSemaphoresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_WaitSemaphoresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphoresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphoresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphoresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphoresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphoresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphoresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphoresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphoresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphoresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphoresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphoresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphoresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphoresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitSemaphoresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_WaitSemaphoresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_WaitSemaphoresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphoresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphoresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphoresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphoresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphoresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphoresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphoresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphoresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphoresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphoresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphoresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphoresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphoresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitSemaphoresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_WaitSemaphoresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_WaitSemaphoresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphoresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphoresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphoresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphoresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphoresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphoresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphoresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphoresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphoresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphoresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphoresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphoresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphoresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitSemaphoresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_WaitSemaphoresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_WaitSemaphoresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_SignalSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_SignalSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_SignalSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_SignalSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_SignalSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_SignalSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_SignalSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_SignalSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_SignalSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_SignalSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_SignalSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_SignalSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_SignalSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_SignalSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_SignalSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_SignalSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_SignalSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_SignalSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_SignalSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_SignalSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_SignalSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_SignalSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_SignalSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_SignalSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphore@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphore@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphore@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphore@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphore@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphore@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphore@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphore@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphore@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphore@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphore@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphore@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphore@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphore@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_SignalSemaphore=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_SignalSemaphore=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphoreKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphoreKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphoreKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphoreKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphoreKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphoreKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphoreKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphoreKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphoreKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphoreKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphoreKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphoreKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphoreKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SignalSemaphoreKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_SignalSemaphoreKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_SignalSemaphoreKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphoreKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphoreKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphoreKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphoreKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphoreKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphoreKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphoreKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphoreKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphoreKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphoreKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphoreKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphoreKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphoreKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SignalSemaphoreKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_SignalSemaphoreKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_SignalSemaphoreKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphoreKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphoreKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphoreKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphoreKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphoreKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphoreKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphoreKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphoreKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphoreKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphoreKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphoreKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphoreKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphoreKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SignalSemaphoreKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_SignalSemaphoreKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_SignalSemaphoreKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphoreKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphoreKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphoreKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphoreKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphoreKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphoreKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphoreKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphoreKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphoreKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphoreKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphoreKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphoreKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphoreKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SignalSemaphoreKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_SignalSemaphoreKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_SignalSemaphoreKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphoreKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphoreKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphoreKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphoreKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphoreKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphoreKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphoreKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphoreKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphoreKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphoreKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphoreKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphoreKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphoreKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SignalSemaphoreKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_SignalSemaphoreKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_SignalSemaphoreKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphoreKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphoreKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphoreKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphoreKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphoreKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphoreKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphoreKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphoreKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphoreKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphoreKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphoreKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphoreKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphoreKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SignalSemaphoreKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_SignalSemaphoreKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_SignalSemaphoreKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphoreKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphoreKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphoreKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphoreKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphoreKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphoreKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphoreKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphoreKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphoreKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphoreKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphoreKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphoreKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphoreKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SignalSemaphoreKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_SignalSemaphoreKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_SignalSemaphoreKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphoreKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphoreKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphoreKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphoreKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphoreKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphoreKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphoreKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphoreKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphoreKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphoreKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphoreKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphoreKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphoreKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SignalSemaphoreKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_SignalSemaphoreKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_SignalSemaphoreKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphoreKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphoreKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphoreKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphoreKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphoreKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphoreKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphoreKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphoreKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphoreKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphoreKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphoreKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphoreKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphoreKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SignalSemaphoreKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_SignalSemaphoreKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_SignalSemaphoreKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphoreKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphoreKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphoreKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphoreKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphoreKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphoreKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphoreKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphoreKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphoreKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphoreKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphoreKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphoreKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphoreKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SignalSemaphoreKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_SignalSemaphoreKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_SignalSemaphoreKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphoreKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphoreKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphoreKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphoreKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphoreKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphoreKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphoreKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphoreKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphoreKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphoreKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphoreKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphoreKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphoreKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SignalSemaphoreKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_SignalSemaphoreKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_SignalSemaphoreKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphoreKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphoreKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphoreKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphoreKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphoreKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphoreKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphoreKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphoreKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphoreKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphoreKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphoreKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphoreKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphoreKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SignalSemaphoreKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_SignalSemaphoreKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_SignalSemaphoreKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphoreKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphoreKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphoreKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphoreKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphoreKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphoreKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphoreKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphoreKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphoreKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphoreKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphoreKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphoreKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphoreKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SignalSemaphoreKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_SignalSemaphoreKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_SignalSemaphoreKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetAndroidHardwareBufferPropertiesANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAndroidHardwareBufferPropertiesANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAndroidHardwareBufferPropertiesANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAndroidHardwareBufferPropertiesANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAndroidHardwareBufferPropertiesANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAndroidHardwareBufferPropertiesANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAndroidHardwareBufferPropertiesANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAndroidHardwareBufferPropertiesANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAndroidHardwareBufferPropertiesANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAndroidHardwareBufferPropertiesANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAndroidHardwareBufferPropertiesANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAndroidHardwareBufferPropertiesANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAndroidHardwareBufferPropertiesANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAndroidHardwareBufferPropertiesANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetAndroidHardwareBufferPropertiesANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetAndroidHardwareBufferPropertiesANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetAndroidHardwareBufferPropertiesANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAndroidHardwareBufferPropertiesANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAndroidHardwareBufferPropertiesANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAndroidHardwareBufferPropertiesANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAndroidHardwareBufferPropertiesANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAndroidHardwareBufferPropertiesANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAndroidHardwareBufferPropertiesANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAndroidHardwareBufferPropertiesANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAndroidHardwareBufferPropertiesANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAndroidHardwareBufferPropertiesANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAndroidHardwareBufferPropertiesANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAndroidHardwareBufferPropertiesANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAndroidHardwareBufferPropertiesANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAndroidHardwareBufferPropertiesANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetAndroidHardwareBufferPropertiesANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetAndroidHardwareBufferPropertiesANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetAndroidHardwareBufferPropertiesANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAndroidHardwareBufferPropertiesANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAndroidHardwareBufferPropertiesANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAndroidHardwareBufferPropertiesANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAndroidHardwareBufferPropertiesANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAndroidHardwareBufferPropertiesANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAndroidHardwareBufferPropertiesANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAndroidHardwareBufferPropertiesANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAndroidHardwareBufferPropertiesANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAndroidHardwareBufferPropertiesANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAndroidHardwareBufferPropertiesANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAndroidHardwareBufferPropertiesANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAndroidHardwareBufferPropertiesANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAndroidHardwareBufferPropertiesANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetAndroidHardwareBufferPropertiesANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetAndroidHardwareBufferPropertiesANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetAndroidHardwareBufferPropertiesANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAndroidHardwareBufferPropertiesANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAndroidHardwareBufferPropertiesANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAndroidHardwareBufferPropertiesANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAndroidHardwareBufferPropertiesANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAndroidHardwareBufferPropertiesANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAndroidHardwareBufferPropertiesANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAndroidHardwareBufferPropertiesANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAndroidHardwareBufferPropertiesANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAndroidHardwareBufferPropertiesANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAndroidHardwareBufferPropertiesANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAndroidHardwareBufferPropertiesANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAndroidHardwareBufferPropertiesANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAndroidHardwareBufferPropertiesANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetAndroidHardwareBufferPropertiesANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetAndroidHardwareBufferPropertiesANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetAndroidHardwareBufferPropertiesANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAndroidHardwareBufferPropertiesANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAndroidHardwareBufferPropertiesANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAndroidHardwareBufferPropertiesANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAndroidHardwareBufferPropertiesANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAndroidHardwareBufferPropertiesANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAndroidHardwareBufferPropertiesANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAndroidHardwareBufferPropertiesANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAndroidHardwareBufferPropertiesANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAndroidHardwareBufferPropertiesANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAndroidHardwareBufferPropertiesANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAndroidHardwareBufferPropertiesANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAndroidHardwareBufferPropertiesANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAndroidHardwareBufferPropertiesANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetAndroidHardwareBufferPropertiesANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetAndroidHardwareBufferPropertiesANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetAndroidHardwareBufferPropertiesANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAndroidHardwareBufferPropertiesANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAndroidHardwareBufferPropertiesANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAndroidHardwareBufferPropertiesANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAndroidHardwareBufferPropertiesANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAndroidHardwareBufferPropertiesANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAndroidHardwareBufferPropertiesANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAndroidHardwareBufferPropertiesANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAndroidHardwareBufferPropertiesANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAndroidHardwareBufferPropertiesANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAndroidHardwareBufferPropertiesANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAndroidHardwareBufferPropertiesANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAndroidHardwareBufferPropertiesANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAndroidHardwareBufferPropertiesANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetAndroidHardwareBufferPropertiesANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetAndroidHardwareBufferPropertiesANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetAndroidHardwareBufferPropertiesANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAndroidHardwareBufferPropertiesANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAndroidHardwareBufferPropertiesANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAndroidHardwareBufferPropertiesANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAndroidHardwareBufferPropertiesANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAndroidHardwareBufferPropertiesANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAndroidHardwareBufferPropertiesANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAndroidHardwareBufferPropertiesANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAndroidHardwareBufferPropertiesANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAndroidHardwareBufferPropertiesANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAndroidHardwareBufferPropertiesANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAndroidHardwareBufferPropertiesANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAndroidHardwareBufferPropertiesANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAndroidHardwareBufferPropertiesANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetAndroidHardwareBufferPropertiesANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetAndroidHardwareBufferPropertiesANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetAndroidHardwareBufferPropertiesANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAndroidHardwareBufferPropertiesANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAndroidHardwareBufferPropertiesANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAndroidHardwareBufferPropertiesANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAndroidHardwareBufferPropertiesANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAndroidHardwareBufferPropertiesANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAndroidHardwareBufferPropertiesANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAndroidHardwareBufferPropertiesANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAndroidHardwareBufferPropertiesANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAndroidHardwareBufferPropertiesANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAndroidHardwareBufferPropertiesANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAndroidHardwareBufferPropertiesANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAndroidHardwareBufferPropertiesANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAndroidHardwareBufferPropertiesANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetAndroidHardwareBufferPropertiesANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetAndroidHardwareBufferPropertiesANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAndroidHardwareBufferPropertiesANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAndroidHardwareBufferPropertiesANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAndroidHardwareBufferPropertiesANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAndroidHardwareBufferPropertiesANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAndroidHardwareBufferPropertiesANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAndroidHardwareBufferPropertiesANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAndroidHardwareBufferPropertiesANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAndroidHardwareBufferPropertiesANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAndroidHardwareBufferPropertiesANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAndroidHardwareBufferPropertiesANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAndroidHardwareBufferPropertiesANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAndroidHardwareBufferPropertiesANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAndroidHardwareBufferPropertiesANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAndroidHardwareBufferPropertiesANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetAndroidHardwareBufferPropertiesANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetAndroidHardwareBufferPropertiesANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAndroidHardwareBufferPropertiesANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAndroidHardwareBufferPropertiesANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAndroidHardwareBufferPropertiesANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAndroidHardwareBufferPropertiesANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAndroidHardwareBufferPropertiesANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAndroidHardwareBufferPropertiesANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAndroidHardwareBufferPropertiesANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAndroidHardwareBufferPropertiesANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAndroidHardwareBufferPropertiesANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAndroidHardwareBufferPropertiesANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAndroidHardwareBufferPropertiesANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAndroidHardwareBufferPropertiesANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAndroidHardwareBufferPropertiesANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAndroidHardwareBufferPropertiesANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetAndroidHardwareBufferPropertiesANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetAndroidHardwareBufferPropertiesANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAndroidHardwareBufferPropertiesANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAndroidHardwareBufferPropertiesANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAndroidHardwareBufferPropertiesANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAndroidHardwareBufferPropertiesANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAndroidHardwareBufferPropertiesANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAndroidHardwareBufferPropertiesANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAndroidHardwareBufferPropertiesANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAndroidHardwareBufferPropertiesANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAndroidHardwareBufferPropertiesANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAndroidHardwareBufferPropertiesANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAndroidHardwareBufferPropertiesANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAndroidHardwareBufferPropertiesANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAndroidHardwareBufferPropertiesANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAndroidHardwareBufferPropertiesANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetAndroidHardwareBufferPropertiesANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetAndroidHardwareBufferPropertiesANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetAndroidHardwareBufferPropertiesANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAndroidHardwareBufferPropertiesANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAndroidHardwareBufferPropertiesANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAndroidHardwareBufferPropertiesANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAndroidHardwareBufferPropertiesANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAndroidHardwareBufferPropertiesANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAndroidHardwareBufferPropertiesANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAndroidHardwareBufferPropertiesANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAndroidHardwareBufferPropertiesANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAndroidHardwareBufferPropertiesANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAndroidHardwareBufferPropertiesANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAndroidHardwareBufferPropertiesANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAndroidHardwareBufferPropertiesANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAndroidHardwareBufferPropertiesANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetAndroidHardwareBufferPropertiesANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetAndroidHardwareBufferPropertiesANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAndroidHardwareBufferPropertiesANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAndroidHardwareBufferPropertiesANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAndroidHardwareBufferPropertiesANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAndroidHardwareBufferPropertiesANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAndroidHardwareBufferPropertiesANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAndroidHardwareBufferPropertiesANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAndroidHardwareBufferPropertiesANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAndroidHardwareBufferPropertiesANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAndroidHardwareBufferPropertiesANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAndroidHardwareBufferPropertiesANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAndroidHardwareBufferPropertiesANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAndroidHardwareBufferPropertiesANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAndroidHardwareBufferPropertiesANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAndroidHardwareBufferPropertiesANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetAndroidHardwareBufferPropertiesANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetAndroidHardwareBufferPropertiesANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetMemoryAndroidHardwareBufferANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryAndroidHardwareBufferANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryAndroidHardwareBufferANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryAndroidHardwareBufferANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryAndroidHardwareBufferANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryAndroidHardwareBufferANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryAndroidHardwareBufferANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryAndroidHardwareBufferANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryAndroidHardwareBufferANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryAndroidHardwareBufferANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryAndroidHardwareBufferANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryAndroidHardwareBufferANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryAndroidHardwareBufferANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryAndroidHardwareBufferANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetMemoryAndroidHardwareBufferANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetMemoryAndroidHardwareBufferANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryAndroidHardwareBufferANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryAndroidHardwareBufferANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryAndroidHardwareBufferANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryAndroidHardwareBufferANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryAndroidHardwareBufferANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryAndroidHardwareBufferANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryAndroidHardwareBufferANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryAndroidHardwareBufferANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryAndroidHardwareBufferANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryAndroidHardwareBufferANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryAndroidHardwareBufferANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryAndroidHardwareBufferANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryAndroidHardwareBufferANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryAndroidHardwareBufferANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetMemoryAndroidHardwareBufferANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetMemoryAndroidHardwareBufferANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryAndroidHardwareBufferANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryAndroidHardwareBufferANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryAndroidHardwareBufferANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryAndroidHardwareBufferANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryAndroidHardwareBufferANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryAndroidHardwareBufferANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryAndroidHardwareBufferANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryAndroidHardwareBufferANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryAndroidHardwareBufferANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryAndroidHardwareBufferANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryAndroidHardwareBufferANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryAndroidHardwareBufferANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryAndroidHardwareBufferANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryAndroidHardwareBufferANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetMemoryAndroidHardwareBufferANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetMemoryAndroidHardwareBufferANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryAndroidHardwareBufferANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryAndroidHardwareBufferANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryAndroidHardwareBufferANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryAndroidHardwareBufferANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryAndroidHardwareBufferANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryAndroidHardwareBufferANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryAndroidHardwareBufferANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryAndroidHardwareBufferANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryAndroidHardwareBufferANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryAndroidHardwareBufferANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryAndroidHardwareBufferANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryAndroidHardwareBufferANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryAndroidHardwareBufferANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryAndroidHardwareBufferANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetMemoryAndroidHardwareBufferANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetMemoryAndroidHardwareBufferANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryAndroidHardwareBufferANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryAndroidHardwareBufferANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryAndroidHardwareBufferANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryAndroidHardwareBufferANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryAndroidHardwareBufferANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryAndroidHardwareBufferANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryAndroidHardwareBufferANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryAndroidHardwareBufferANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryAndroidHardwareBufferANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryAndroidHardwareBufferANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryAndroidHardwareBufferANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryAndroidHardwareBufferANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryAndroidHardwareBufferANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryAndroidHardwareBufferANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetMemoryAndroidHardwareBufferANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetMemoryAndroidHardwareBufferANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryAndroidHardwareBufferANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryAndroidHardwareBufferANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryAndroidHardwareBufferANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryAndroidHardwareBufferANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryAndroidHardwareBufferANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryAndroidHardwareBufferANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryAndroidHardwareBufferANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryAndroidHardwareBufferANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryAndroidHardwareBufferANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryAndroidHardwareBufferANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryAndroidHardwareBufferANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryAndroidHardwareBufferANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryAndroidHardwareBufferANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryAndroidHardwareBufferANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetMemoryAndroidHardwareBufferANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetMemoryAndroidHardwareBufferANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryAndroidHardwareBufferANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryAndroidHardwareBufferANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryAndroidHardwareBufferANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryAndroidHardwareBufferANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryAndroidHardwareBufferANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryAndroidHardwareBufferANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryAndroidHardwareBufferANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryAndroidHardwareBufferANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryAndroidHardwareBufferANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryAndroidHardwareBufferANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryAndroidHardwareBufferANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryAndroidHardwareBufferANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryAndroidHardwareBufferANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryAndroidHardwareBufferANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetMemoryAndroidHardwareBufferANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetMemoryAndroidHardwareBufferANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryAndroidHardwareBufferANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryAndroidHardwareBufferANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryAndroidHardwareBufferANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryAndroidHardwareBufferANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryAndroidHardwareBufferANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryAndroidHardwareBufferANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryAndroidHardwareBufferANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryAndroidHardwareBufferANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryAndroidHardwareBufferANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryAndroidHardwareBufferANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryAndroidHardwareBufferANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryAndroidHardwareBufferANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryAndroidHardwareBufferANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryAndroidHardwareBufferANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetMemoryAndroidHardwareBufferANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetMemoryAndroidHardwareBufferANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryAndroidHardwareBufferANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryAndroidHardwareBufferANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryAndroidHardwareBufferANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryAndroidHardwareBufferANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryAndroidHardwareBufferANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryAndroidHardwareBufferANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryAndroidHardwareBufferANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryAndroidHardwareBufferANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryAndroidHardwareBufferANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryAndroidHardwareBufferANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryAndroidHardwareBufferANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryAndroidHardwareBufferANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryAndroidHardwareBufferANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryAndroidHardwareBufferANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetMemoryAndroidHardwareBufferANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetMemoryAndroidHardwareBufferANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryAndroidHardwareBufferANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryAndroidHardwareBufferANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryAndroidHardwareBufferANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryAndroidHardwareBufferANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryAndroidHardwareBufferANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryAndroidHardwareBufferANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryAndroidHardwareBufferANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryAndroidHardwareBufferANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryAndroidHardwareBufferANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryAndroidHardwareBufferANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryAndroidHardwareBufferANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryAndroidHardwareBufferANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryAndroidHardwareBufferANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryAndroidHardwareBufferANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetMemoryAndroidHardwareBufferANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetMemoryAndroidHardwareBufferANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryAndroidHardwareBufferANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryAndroidHardwareBufferANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryAndroidHardwareBufferANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryAndroidHardwareBufferANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryAndroidHardwareBufferANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryAndroidHardwareBufferANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryAndroidHardwareBufferANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryAndroidHardwareBufferANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryAndroidHardwareBufferANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryAndroidHardwareBufferANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryAndroidHardwareBufferANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryAndroidHardwareBufferANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryAndroidHardwareBufferANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryAndroidHardwareBufferANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetMemoryAndroidHardwareBufferANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetMemoryAndroidHardwareBufferANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryAndroidHardwareBufferANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryAndroidHardwareBufferANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryAndroidHardwareBufferANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryAndroidHardwareBufferANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryAndroidHardwareBufferANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryAndroidHardwareBufferANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryAndroidHardwareBufferANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryAndroidHardwareBufferANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryAndroidHardwareBufferANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryAndroidHardwareBufferANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryAndroidHardwareBufferANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryAndroidHardwareBufferANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryAndroidHardwareBufferANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryAndroidHardwareBufferANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetMemoryAndroidHardwareBufferANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetMemoryAndroidHardwareBufferANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryAndroidHardwareBufferANDROID@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryAndroidHardwareBufferANDROID@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryAndroidHardwareBufferANDROID@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryAndroidHardwareBufferANDROID@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryAndroidHardwareBufferANDROID@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryAndroidHardwareBufferANDROID@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryAndroidHardwareBufferANDROID@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryAndroidHardwareBufferANDROID@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryAndroidHardwareBufferANDROID@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryAndroidHardwareBufferANDROID@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryAndroidHardwareBufferANDROID@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryAndroidHardwareBufferANDROID@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryAndroidHardwareBufferANDROID@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryAndroidHardwareBufferANDROID@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetMemoryAndroidHardwareBufferANDROID=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetMemoryAndroidHardwareBufferANDROID=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawIndexedIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawIndexedIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawIndexedIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawIndexedIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawIndexedIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawIndexedIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawIndexedIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawIndexedIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawIndexedIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawIndexedIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawIndexedIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawIndexedIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawIndexedIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawIndexedIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawIndexedIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawIndexedIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawIndexedIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawIndexedIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawIndexedIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawIndexedIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawIndexedIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawIndexedIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawIndexedIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawIndexedIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawIndexedIndirectCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawIndexedIndirectCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawIndexedIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawIndexedIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawIndexedIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawIndexedIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawIndexedIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawIndexedIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawIndexedIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawIndexedIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawIndexedIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawIndexedIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawIndexedIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawIndexedIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawIndexedIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawIndexedIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawIndexedIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawIndexedIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawIndexedIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawIndexedIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawIndexedIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawIndexedIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawIndexedIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawIndexedIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawIndexedIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawIndexedIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawIndexedIndirectCountKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawIndexedIndirectCountKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawIndexedIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawIndexedIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawIndexedIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawIndexedIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawIndexedIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawIndexedIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawIndexedIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawIndexedIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawIndexedIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawIndexedIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawIndexedIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawIndexedIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawIndexedIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawIndexedIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawIndexedIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawIndexedIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawIndexedIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawIndexedIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawIndexedIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawIndexedIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawIndexedIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawIndexedIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawIndexedIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawIndexedIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCountAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawIndexedIndirectCountAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawIndexedIndirectCountAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetCheckpointNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCheckpointNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCheckpointNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCheckpointNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCheckpointNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCheckpointNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCheckpointNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCheckpointNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCheckpointNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCheckpointNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCheckpointNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCheckpointNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCheckpointNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCheckpointNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetCheckpointNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetCheckpointNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCheckpointNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCheckpointNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCheckpointNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCheckpointNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCheckpointNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCheckpointNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCheckpointNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCheckpointNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCheckpointNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCheckpointNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCheckpointNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCheckpointNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCheckpointNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCheckpointNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetCheckpointNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetCheckpointNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCheckpointNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCheckpointNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCheckpointNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCheckpointNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCheckpointNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCheckpointNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCheckpointNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCheckpointNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCheckpointNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCheckpointNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCheckpointNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCheckpointNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCheckpointNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCheckpointNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetCheckpointNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetCheckpointNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCheckpointNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCheckpointNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCheckpointNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCheckpointNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCheckpointNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCheckpointNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCheckpointNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCheckpointNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCheckpointNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCheckpointNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCheckpointNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCheckpointNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCheckpointNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCheckpointNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetCheckpointNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetCheckpointNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCheckpointNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCheckpointNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCheckpointNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCheckpointNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCheckpointNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCheckpointNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCheckpointNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCheckpointNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCheckpointNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCheckpointNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCheckpointNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCheckpointNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCheckpointNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCheckpointNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetCheckpointNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetCheckpointNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCheckpointNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCheckpointNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCheckpointNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCheckpointNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCheckpointNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCheckpointNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCheckpointNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCheckpointNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCheckpointNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCheckpointNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCheckpointNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCheckpointNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCheckpointNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCheckpointNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetCheckpointNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetCheckpointNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCheckpointNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCheckpointNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCheckpointNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCheckpointNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCheckpointNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCheckpointNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCheckpointNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCheckpointNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCheckpointNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCheckpointNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCheckpointNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCheckpointNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCheckpointNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCheckpointNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetCheckpointNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetCheckpointNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCheckpointNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCheckpointNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCheckpointNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCheckpointNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCheckpointNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCheckpointNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCheckpointNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCheckpointNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCheckpointNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCheckpointNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCheckpointNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCheckpointNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCheckpointNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCheckpointNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetCheckpointNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetCheckpointNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCheckpointNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCheckpointNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCheckpointNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCheckpointNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCheckpointNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCheckpointNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCheckpointNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCheckpointNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCheckpointNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCheckpointNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCheckpointNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCheckpointNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCheckpointNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCheckpointNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetCheckpointNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetCheckpointNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCheckpointNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCheckpointNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCheckpointNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCheckpointNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCheckpointNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCheckpointNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCheckpointNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCheckpointNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCheckpointNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCheckpointNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCheckpointNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCheckpointNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCheckpointNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCheckpointNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetCheckpointNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetCheckpointNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCheckpointNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCheckpointNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCheckpointNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCheckpointNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCheckpointNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCheckpointNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCheckpointNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCheckpointNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCheckpointNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCheckpointNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCheckpointNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCheckpointNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCheckpointNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCheckpointNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetCheckpointNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetCheckpointNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCheckpointNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCheckpointNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCheckpointNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCheckpointNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCheckpointNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCheckpointNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCheckpointNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCheckpointNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCheckpointNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCheckpointNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCheckpointNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCheckpointNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCheckpointNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCheckpointNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetCheckpointNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetCheckpointNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCheckpointNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCheckpointNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCheckpointNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCheckpointNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCheckpointNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCheckpointNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCheckpointNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCheckpointNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCheckpointNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCheckpointNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCheckpointNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCheckpointNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCheckpointNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCheckpointNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetCheckpointNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetCheckpointNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetQueueCheckpointDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetQueueCheckpointDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetQueueCheckpointDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetQueueCheckpointDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetQueueCheckpointDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetQueueCheckpointDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetQueueCheckpointDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetQueueCheckpointDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetQueueCheckpointDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetQueueCheckpointDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetQueueCheckpointDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetQueueCheckpointDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetQueueCheckpointDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetQueueCheckpointDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetQueueCheckpointDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetQueueCheckpointDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetQueueCheckpointDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetQueueCheckpointDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetQueueCheckpointDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetQueueCheckpointDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetQueueCheckpointDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetQueueCheckpointDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetQueueCheckpointDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetQueueCheckpointDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointDataNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointDataNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointDataNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointDataNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointDataNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointDataNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointDataNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointDataNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointDataNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointDataNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointDataNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointDataNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointDataNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointDataNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetQueueCheckpointDataNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetQueueCheckpointDataNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindTransformFeedbackBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindTransformFeedbackBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindTransformFeedbackBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindTransformFeedbackBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindTransformFeedbackBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindTransformFeedbackBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindTransformFeedbackBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindTransformFeedbackBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindTransformFeedbackBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindTransformFeedbackBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindTransformFeedbackBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindTransformFeedbackBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindTransformFeedbackBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindTransformFeedbackBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindTransformFeedbackBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindTransformFeedbackBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindTransformFeedbackBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindTransformFeedbackBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindTransformFeedbackBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindTransformFeedbackBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindTransformFeedbackBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindTransformFeedbackBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindTransformFeedbackBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindTransformFeedbackBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindTransformFeedbackBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindTransformFeedbackBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBeginTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBeginTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBeginTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBeginTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBeginTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBeginTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBeginTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBeginTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBeginTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBeginTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBeginTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBeginTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBeginTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBeginTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBeginTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBeginTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBeginTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBeginTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBeginTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBeginTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBeginTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBeginTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBeginTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBeginTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBeginTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBeginTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdEndTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdEndTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdEndTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdEndTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdEndTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdEndTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdEndTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdEndTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdEndTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdEndTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdEndTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdEndTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdEndTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdEndTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdEndTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdEndTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdEndTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdEndTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdEndTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdEndTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdEndTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdEndTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdEndTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdEndTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedbackEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedbackEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedbackEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedbackEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedbackEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedbackEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedbackEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedbackEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedbackEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedbackEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedbackEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedbackEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedbackEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedbackEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdEndTransformFeedbackEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdEndTransformFeedbackEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBeginQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBeginQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBeginQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBeginQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBeginQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBeginQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBeginQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBeginQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBeginQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBeginQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBeginQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBeginQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBeginQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBeginQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBeginQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBeginQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBeginQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBeginQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBeginQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBeginQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBeginQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBeginQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBeginQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBeginQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBeginQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBeginQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdEndQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdEndQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdEndQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdEndQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdEndQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdEndQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdEndQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdEndQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdEndQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdEndQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdEndQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdEndQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdEndQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdEndQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdEndQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdEndQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdEndQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdEndQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdEndQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdEndQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdEndQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdEndQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdEndQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdEndQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdEndQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQueryIndexedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQueryIndexedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQueryIndexedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQueryIndexedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQueryIndexedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQueryIndexedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQueryIndexedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQueryIndexedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQueryIndexedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQueryIndexedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQueryIndexedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQueryIndexedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQueryIndexedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndQueryIndexedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdEndQueryIndexedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdEndQueryIndexedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawIndirectByteCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawIndirectByteCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawIndirectByteCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawIndirectByteCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawIndirectByteCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawIndirectByteCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawIndirectByteCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawIndirectByteCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawIndirectByteCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawIndirectByteCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawIndirectByteCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawIndirectByteCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawIndirectByteCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawIndirectByteCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawIndirectByteCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawIndirectByteCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawIndirectByteCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawIndirectByteCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawIndirectByteCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawIndirectByteCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawIndirectByteCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawIndirectByteCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawIndirectByteCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawIndirectByteCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawIndirectByteCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawIndirectByteCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetExclusiveScissorNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetExclusiveScissorNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetExclusiveScissorNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetExclusiveScissorNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetExclusiveScissorNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetExclusiveScissorNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetExclusiveScissorNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetExclusiveScissorNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetExclusiveScissorNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetExclusiveScissorNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetExclusiveScissorNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetExclusiveScissorNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetExclusiveScissorNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetExclusiveScissorNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetExclusiveScissorNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetExclusiveScissorNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetExclusiveScissorNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetExclusiveScissorNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetExclusiveScissorNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetExclusiveScissorNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetExclusiveScissorNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetExclusiveScissorNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetExclusiveScissorNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetExclusiveScissorNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetExclusiveScissorNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetExclusiveScissorNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExclusiveScissorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetExclusiveScissorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetExclusiveScissorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExclusiveScissorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetExclusiveScissorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetExclusiveScissorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExclusiveScissorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetExclusiveScissorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetExclusiveScissorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExclusiveScissorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetExclusiveScissorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetExclusiveScissorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExclusiveScissorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetExclusiveScissorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetExclusiveScissorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExclusiveScissorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetExclusiveScissorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetExclusiveScissorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExclusiveScissorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetExclusiveScissorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetExclusiveScissorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExclusiveScissorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetExclusiveScissorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetExclusiveScissorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExclusiveScissorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetExclusiveScissorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetExclusiveScissorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExclusiveScissorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetExclusiveScissorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetExclusiveScissorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExclusiveScissorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetExclusiveScissorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetExclusiveScissorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExclusiveScissorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetExclusiveScissorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetExclusiveScissorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExclusiveScissorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetExclusiveScissorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetExclusiveScissorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadingRateImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadingRateImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadingRateImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadingRateImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadingRateImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadingRateImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadingRateImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadingRateImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadingRateImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadingRateImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadingRateImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadingRateImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadingRateImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadingRateImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindShadingRateImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindShadingRateImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadingRateImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadingRateImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadingRateImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadingRateImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadingRateImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadingRateImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadingRateImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadingRateImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadingRateImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadingRateImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadingRateImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadingRateImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadingRateImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadingRateImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindShadingRateImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindShadingRateImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadingRateImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadingRateImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadingRateImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadingRateImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadingRateImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadingRateImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadingRateImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadingRateImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadingRateImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadingRateImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadingRateImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadingRateImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadingRateImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadingRateImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindShadingRateImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindShadingRateImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadingRateImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadingRateImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadingRateImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadingRateImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadingRateImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadingRateImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadingRateImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadingRateImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadingRateImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadingRateImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadingRateImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadingRateImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadingRateImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadingRateImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindShadingRateImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindShadingRateImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadingRateImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadingRateImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadingRateImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadingRateImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadingRateImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadingRateImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadingRateImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadingRateImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadingRateImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadingRateImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadingRateImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadingRateImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadingRateImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadingRateImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindShadingRateImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindShadingRateImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadingRateImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadingRateImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadingRateImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadingRateImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadingRateImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadingRateImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadingRateImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadingRateImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadingRateImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadingRateImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadingRateImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadingRateImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadingRateImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadingRateImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindShadingRateImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindShadingRateImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadingRateImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadingRateImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadingRateImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadingRateImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadingRateImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadingRateImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadingRateImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadingRateImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadingRateImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadingRateImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadingRateImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadingRateImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadingRateImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadingRateImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindShadingRateImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindShadingRateImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadingRateImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadingRateImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadingRateImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadingRateImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadingRateImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadingRateImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadingRateImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadingRateImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadingRateImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadingRateImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadingRateImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadingRateImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadingRateImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadingRateImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindShadingRateImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindShadingRateImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadingRateImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadingRateImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadingRateImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadingRateImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadingRateImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadingRateImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadingRateImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadingRateImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadingRateImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadingRateImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadingRateImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadingRateImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadingRateImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadingRateImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindShadingRateImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindShadingRateImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadingRateImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadingRateImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadingRateImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadingRateImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadingRateImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadingRateImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadingRateImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadingRateImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadingRateImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadingRateImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadingRateImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadingRateImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadingRateImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadingRateImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindShadingRateImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindShadingRateImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadingRateImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadingRateImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadingRateImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadingRateImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadingRateImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadingRateImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadingRateImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadingRateImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadingRateImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadingRateImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadingRateImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadingRateImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadingRateImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadingRateImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindShadingRateImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindShadingRateImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadingRateImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadingRateImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadingRateImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadingRateImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadingRateImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadingRateImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadingRateImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadingRateImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadingRateImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadingRateImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadingRateImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadingRateImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadingRateImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadingRateImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindShadingRateImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindShadingRateImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadingRateImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadingRateImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadingRateImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadingRateImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadingRateImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadingRateImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadingRateImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadingRateImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadingRateImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadingRateImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadingRateImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadingRateImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadingRateImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadingRateImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindShadingRateImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindShadingRateImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportShadingRatePaletteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportShadingRatePaletteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportShadingRatePaletteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportShadingRatePaletteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportShadingRatePaletteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportShadingRatePaletteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportShadingRatePaletteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportShadingRatePaletteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportShadingRatePaletteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportShadingRatePaletteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportShadingRatePaletteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportShadingRatePaletteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportShadingRatePaletteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportShadingRatePaletteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetViewportShadingRatePaletteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetViewportShadingRatePaletteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportShadingRatePaletteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportShadingRatePaletteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportShadingRatePaletteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportShadingRatePaletteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportShadingRatePaletteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportShadingRatePaletteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportShadingRatePaletteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportShadingRatePaletteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportShadingRatePaletteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportShadingRatePaletteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportShadingRatePaletteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportShadingRatePaletteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportShadingRatePaletteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportShadingRatePaletteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetViewportShadingRatePaletteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetViewportShadingRatePaletteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportShadingRatePaletteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportShadingRatePaletteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportShadingRatePaletteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportShadingRatePaletteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportShadingRatePaletteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportShadingRatePaletteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportShadingRatePaletteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportShadingRatePaletteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportShadingRatePaletteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportShadingRatePaletteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportShadingRatePaletteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportShadingRatePaletteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportShadingRatePaletteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportShadingRatePaletteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetViewportShadingRatePaletteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetViewportShadingRatePaletteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportShadingRatePaletteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportShadingRatePaletteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportShadingRatePaletteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportShadingRatePaletteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportShadingRatePaletteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportShadingRatePaletteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportShadingRatePaletteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportShadingRatePaletteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportShadingRatePaletteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportShadingRatePaletteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportShadingRatePaletteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportShadingRatePaletteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportShadingRatePaletteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportShadingRatePaletteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetViewportShadingRatePaletteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetViewportShadingRatePaletteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportShadingRatePaletteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportShadingRatePaletteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportShadingRatePaletteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportShadingRatePaletteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportShadingRatePaletteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportShadingRatePaletteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportShadingRatePaletteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportShadingRatePaletteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportShadingRatePaletteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportShadingRatePaletteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportShadingRatePaletteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportShadingRatePaletteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportShadingRatePaletteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportShadingRatePaletteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetViewportShadingRatePaletteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetViewportShadingRatePaletteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportShadingRatePaletteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportShadingRatePaletteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportShadingRatePaletteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportShadingRatePaletteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportShadingRatePaletteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportShadingRatePaletteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportShadingRatePaletteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportShadingRatePaletteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportShadingRatePaletteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportShadingRatePaletteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportShadingRatePaletteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportShadingRatePaletteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportShadingRatePaletteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportShadingRatePaletteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetViewportShadingRatePaletteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetViewportShadingRatePaletteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportShadingRatePaletteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportShadingRatePaletteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportShadingRatePaletteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportShadingRatePaletteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportShadingRatePaletteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportShadingRatePaletteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportShadingRatePaletteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportShadingRatePaletteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportShadingRatePaletteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportShadingRatePaletteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportShadingRatePaletteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportShadingRatePaletteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportShadingRatePaletteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportShadingRatePaletteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetViewportShadingRatePaletteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetViewportShadingRatePaletteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportShadingRatePaletteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportShadingRatePaletteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportShadingRatePaletteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportShadingRatePaletteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportShadingRatePaletteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportShadingRatePaletteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportShadingRatePaletteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportShadingRatePaletteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportShadingRatePaletteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportShadingRatePaletteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportShadingRatePaletteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportShadingRatePaletteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportShadingRatePaletteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportShadingRatePaletteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetViewportShadingRatePaletteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetViewportShadingRatePaletteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportShadingRatePaletteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportShadingRatePaletteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportShadingRatePaletteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportShadingRatePaletteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportShadingRatePaletteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportShadingRatePaletteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportShadingRatePaletteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportShadingRatePaletteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportShadingRatePaletteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportShadingRatePaletteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportShadingRatePaletteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportShadingRatePaletteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportShadingRatePaletteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportShadingRatePaletteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetViewportShadingRatePaletteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetViewportShadingRatePaletteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportShadingRatePaletteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportShadingRatePaletteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportShadingRatePaletteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportShadingRatePaletteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportShadingRatePaletteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportShadingRatePaletteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportShadingRatePaletteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportShadingRatePaletteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportShadingRatePaletteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportShadingRatePaletteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportShadingRatePaletteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportShadingRatePaletteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportShadingRatePaletteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportShadingRatePaletteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetViewportShadingRatePaletteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetViewportShadingRatePaletteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportShadingRatePaletteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportShadingRatePaletteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportShadingRatePaletteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportShadingRatePaletteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportShadingRatePaletteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportShadingRatePaletteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportShadingRatePaletteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportShadingRatePaletteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportShadingRatePaletteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportShadingRatePaletteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportShadingRatePaletteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportShadingRatePaletteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportShadingRatePaletteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportShadingRatePaletteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetViewportShadingRatePaletteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetViewportShadingRatePaletteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportShadingRatePaletteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportShadingRatePaletteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportShadingRatePaletteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportShadingRatePaletteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportShadingRatePaletteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportShadingRatePaletteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportShadingRatePaletteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportShadingRatePaletteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportShadingRatePaletteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportShadingRatePaletteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportShadingRatePaletteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportShadingRatePaletteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportShadingRatePaletteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportShadingRatePaletteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetViewportShadingRatePaletteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetViewportShadingRatePaletteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportShadingRatePaletteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportShadingRatePaletteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportShadingRatePaletteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportShadingRatePaletteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportShadingRatePaletteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportShadingRatePaletteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportShadingRatePaletteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportShadingRatePaletteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportShadingRatePaletteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportShadingRatePaletteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportShadingRatePaletteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportShadingRatePaletteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportShadingRatePaletteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportShadingRatePaletteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetViewportShadingRatePaletteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetViewportShadingRatePaletteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoarseSampleOrderNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoarseSampleOrderNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoarseSampleOrderNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoarseSampleOrderNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoarseSampleOrderNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoarseSampleOrderNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoarseSampleOrderNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoarseSampleOrderNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoarseSampleOrderNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoarseSampleOrderNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoarseSampleOrderNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoarseSampleOrderNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoarseSampleOrderNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoarseSampleOrderNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetCoarseSampleOrderNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetCoarseSampleOrderNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoarseSampleOrderNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoarseSampleOrderNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoarseSampleOrderNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoarseSampleOrderNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoarseSampleOrderNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoarseSampleOrderNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoarseSampleOrderNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoarseSampleOrderNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoarseSampleOrderNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoarseSampleOrderNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoarseSampleOrderNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoarseSampleOrderNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoarseSampleOrderNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoarseSampleOrderNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetCoarseSampleOrderNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetCoarseSampleOrderNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoarseSampleOrderNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoarseSampleOrderNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoarseSampleOrderNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoarseSampleOrderNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoarseSampleOrderNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoarseSampleOrderNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoarseSampleOrderNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoarseSampleOrderNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoarseSampleOrderNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoarseSampleOrderNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoarseSampleOrderNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoarseSampleOrderNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoarseSampleOrderNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoarseSampleOrderNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetCoarseSampleOrderNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetCoarseSampleOrderNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoarseSampleOrderNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoarseSampleOrderNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoarseSampleOrderNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoarseSampleOrderNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoarseSampleOrderNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoarseSampleOrderNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoarseSampleOrderNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoarseSampleOrderNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoarseSampleOrderNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoarseSampleOrderNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoarseSampleOrderNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoarseSampleOrderNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoarseSampleOrderNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoarseSampleOrderNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetCoarseSampleOrderNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetCoarseSampleOrderNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoarseSampleOrderNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoarseSampleOrderNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoarseSampleOrderNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoarseSampleOrderNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoarseSampleOrderNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoarseSampleOrderNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoarseSampleOrderNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoarseSampleOrderNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoarseSampleOrderNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoarseSampleOrderNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoarseSampleOrderNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoarseSampleOrderNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoarseSampleOrderNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoarseSampleOrderNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetCoarseSampleOrderNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetCoarseSampleOrderNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoarseSampleOrderNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoarseSampleOrderNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoarseSampleOrderNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoarseSampleOrderNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoarseSampleOrderNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoarseSampleOrderNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoarseSampleOrderNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoarseSampleOrderNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoarseSampleOrderNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoarseSampleOrderNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoarseSampleOrderNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoarseSampleOrderNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoarseSampleOrderNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoarseSampleOrderNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetCoarseSampleOrderNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetCoarseSampleOrderNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoarseSampleOrderNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoarseSampleOrderNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoarseSampleOrderNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoarseSampleOrderNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoarseSampleOrderNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoarseSampleOrderNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoarseSampleOrderNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoarseSampleOrderNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoarseSampleOrderNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoarseSampleOrderNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoarseSampleOrderNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoarseSampleOrderNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoarseSampleOrderNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoarseSampleOrderNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetCoarseSampleOrderNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetCoarseSampleOrderNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoarseSampleOrderNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoarseSampleOrderNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoarseSampleOrderNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoarseSampleOrderNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoarseSampleOrderNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoarseSampleOrderNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoarseSampleOrderNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoarseSampleOrderNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoarseSampleOrderNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoarseSampleOrderNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoarseSampleOrderNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoarseSampleOrderNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoarseSampleOrderNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoarseSampleOrderNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetCoarseSampleOrderNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetCoarseSampleOrderNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoarseSampleOrderNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoarseSampleOrderNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoarseSampleOrderNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoarseSampleOrderNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoarseSampleOrderNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoarseSampleOrderNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoarseSampleOrderNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoarseSampleOrderNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoarseSampleOrderNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoarseSampleOrderNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoarseSampleOrderNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoarseSampleOrderNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoarseSampleOrderNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoarseSampleOrderNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetCoarseSampleOrderNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetCoarseSampleOrderNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoarseSampleOrderNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoarseSampleOrderNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoarseSampleOrderNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoarseSampleOrderNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoarseSampleOrderNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoarseSampleOrderNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoarseSampleOrderNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoarseSampleOrderNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoarseSampleOrderNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoarseSampleOrderNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoarseSampleOrderNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoarseSampleOrderNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoarseSampleOrderNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoarseSampleOrderNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetCoarseSampleOrderNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetCoarseSampleOrderNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoarseSampleOrderNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoarseSampleOrderNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoarseSampleOrderNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoarseSampleOrderNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoarseSampleOrderNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoarseSampleOrderNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoarseSampleOrderNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoarseSampleOrderNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoarseSampleOrderNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoarseSampleOrderNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoarseSampleOrderNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoarseSampleOrderNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoarseSampleOrderNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoarseSampleOrderNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetCoarseSampleOrderNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetCoarseSampleOrderNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoarseSampleOrderNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoarseSampleOrderNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoarseSampleOrderNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoarseSampleOrderNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoarseSampleOrderNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoarseSampleOrderNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoarseSampleOrderNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoarseSampleOrderNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoarseSampleOrderNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoarseSampleOrderNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoarseSampleOrderNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoarseSampleOrderNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoarseSampleOrderNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoarseSampleOrderNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetCoarseSampleOrderNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetCoarseSampleOrderNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoarseSampleOrderNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoarseSampleOrderNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoarseSampleOrderNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoarseSampleOrderNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoarseSampleOrderNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoarseSampleOrderNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoarseSampleOrderNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoarseSampleOrderNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoarseSampleOrderNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoarseSampleOrderNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoarseSampleOrderNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoarseSampleOrderNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoarseSampleOrderNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoarseSampleOrderNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetCoarseSampleOrderNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetCoarseSampleOrderNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawMeshTasksNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawMeshTasksNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawMeshTasksNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawMeshTasksNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawMeshTasksNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawMeshTasksNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawMeshTasksNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawMeshTasksNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawMeshTasksNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawMeshTasksNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawMeshTasksNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawMeshTasksNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawMeshTasksNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawMeshTasksNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawMeshTasksNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawMeshTasksNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawMeshTasksNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawMeshTasksNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawMeshTasksNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawMeshTasksNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawMeshTasksNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawMeshTasksNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawMeshTasksNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawMeshTasksNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawMeshTasksNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawMeshTasksNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawMeshTasksIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawMeshTasksIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawMeshTasksIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawMeshTasksIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawMeshTasksIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawMeshTasksIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawMeshTasksIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawMeshTasksIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawMeshTasksIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawMeshTasksIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawMeshTasksIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawMeshTasksIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawMeshTasksIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawMeshTasksIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawMeshTasksIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawMeshTasksIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawMeshTasksIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawMeshTasksIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawMeshTasksIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawMeshTasksIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawMeshTasksIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawMeshTasksIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawMeshTasksIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawMeshTasksIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawMeshTasksIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawMeshTasksIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawMeshTasksIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawMeshTasksIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawMeshTasksIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawMeshTasksIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawMeshTasksIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawMeshTasksIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawMeshTasksIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawMeshTasksIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawMeshTasksIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawMeshTasksIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawMeshTasksIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawMeshTasksIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawMeshTasksIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawMeshTasksIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawMeshTasksIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawMeshTasksIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawMeshTasksIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawMeshTasksIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawMeshTasksIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawMeshTasksIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawMeshTasksIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawMeshTasksIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawMeshTasksIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawMeshTasksIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawMeshTasksIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawMeshTasksIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawMeshTasksEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawMeshTasksEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawMeshTasksEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawMeshTasksEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawMeshTasksEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawMeshTasksEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawMeshTasksEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawMeshTasksEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawMeshTasksEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawMeshTasksEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawMeshTasksEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawMeshTasksEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawMeshTasksEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawMeshTasksEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawMeshTasksEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawMeshTasksEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawMeshTasksEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawMeshTasksEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawMeshTasksEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawMeshTasksEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawMeshTasksEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawMeshTasksEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawMeshTasksEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawMeshTasksEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawMeshTasksEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawMeshTasksEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawMeshTasksIndirectEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawMeshTasksIndirectEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawMeshTasksIndirectEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawMeshTasksIndirectEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawMeshTasksIndirectEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawMeshTasksIndirectEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawMeshTasksIndirectEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawMeshTasksIndirectEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawMeshTasksIndirectEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawMeshTasksIndirectEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawMeshTasksIndirectEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawMeshTasksIndirectEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawMeshTasksIndirectEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawMeshTasksIndirectEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawMeshTasksIndirectEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawMeshTasksIndirectEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawMeshTasksIndirectEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawMeshTasksIndirectEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawMeshTasksIndirectEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawMeshTasksIndirectEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawMeshTasksIndirectEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawMeshTasksIndirectEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawMeshTasksIndirectEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawMeshTasksIndirectEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawMeshTasksIndirectEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawMeshTasksIndirectEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawMeshTasksIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawMeshTasksIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawMeshTasksIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawMeshTasksIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawMeshTasksIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawMeshTasksIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawMeshTasksIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawMeshTasksIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawMeshTasksIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawMeshTasksIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawMeshTasksIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawMeshTasksIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawMeshTasksIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawMeshTasksIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawMeshTasksIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawMeshTasksIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawMeshTasksIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawMeshTasksIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawMeshTasksIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawMeshTasksIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawMeshTasksIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawMeshTasksIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawMeshTasksIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawMeshTasksIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawMeshTasksIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawMeshTasksIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CompileDeferredNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CompileDeferredNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CompileDeferredNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CompileDeferredNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CompileDeferredNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CompileDeferredNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CompileDeferredNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CompileDeferredNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CompileDeferredNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CompileDeferredNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CompileDeferredNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CompileDeferredNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CompileDeferredNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CompileDeferredNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CompileDeferredNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CompileDeferredNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CompileDeferredNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CompileDeferredNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CompileDeferredNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CompileDeferredNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CompileDeferredNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CompileDeferredNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CompileDeferredNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CompileDeferredNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CompileDeferredNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CompileDeferredNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CompileDeferredNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CompileDeferredNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CompileDeferredNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CompileDeferredNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CompileDeferredNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CompileDeferredNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CompileDeferredNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CompileDeferredNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CompileDeferredNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CompileDeferredNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CompileDeferredNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CompileDeferredNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CompileDeferredNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CompileDeferredNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CompileDeferredNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CompileDeferredNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CompileDeferredNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CompileDeferredNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CompileDeferredNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CompileDeferredNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CompileDeferredNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CompileDeferredNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CompileDeferredNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CompileDeferredNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CompileDeferredNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CompileDeferredNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CompileDeferredNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CompileDeferredNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CompileDeferredNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CompileDeferredNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CompileDeferredNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CompileDeferredNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CompileDeferredNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CompileDeferredNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CompileDeferredNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CompileDeferredNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CompileDeferredNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CompileDeferredNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CompileDeferredNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CompileDeferredNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CompileDeferredNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CompileDeferredNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CompileDeferredNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CompileDeferredNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CompileDeferredNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CompileDeferredNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CompileDeferredNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CompileDeferredNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CompileDeferredNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CompileDeferredNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CompileDeferredNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CompileDeferredNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CompileDeferredNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CompileDeferredNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CompileDeferredNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CompileDeferredNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CompileDeferredNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CompileDeferredNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CompileDeferredNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CompileDeferredNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CompileDeferredNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CompileDeferredNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CompileDeferredNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CompileDeferredNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CompileDeferredNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CompileDeferredNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CompileDeferredNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CompileDeferredNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CompileDeferredNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CompileDeferredNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CompileDeferredNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CompileDeferredNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CompileDeferredNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CompileDeferredNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CompileDeferredNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CompileDeferredNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CompileDeferredNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CompileDeferredNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CompileDeferredNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CompileDeferredNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CompileDeferredNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CompileDeferredNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CompileDeferredNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CompileDeferredNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CompileDeferredNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CompileDeferredNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CompileDeferredNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CompileDeferredNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CompileDeferredNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CompileDeferredNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CompileDeferredNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CompileDeferredNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CompileDeferredNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CompileDeferredNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CompileDeferredNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CompileDeferredNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CompileDeferredNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CompileDeferredNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CompileDeferredNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CompileDeferredNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CompileDeferredNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CompileDeferredNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CompileDeferredNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CompileDeferredNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CompileDeferredNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CompileDeferredNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CompileDeferredNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CompileDeferredNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CompileDeferredNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CompileDeferredNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CompileDeferredNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CompileDeferredNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CompileDeferredNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CompileDeferredNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CompileDeferredNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CompileDeferredNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CompileDeferredNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CompileDeferredNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CompileDeferredNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CompileDeferredNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CompileDeferredNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CompileDeferredNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CompileDeferredNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CompileDeferredNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CompileDeferredNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CompileDeferredNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CompileDeferredNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CompileDeferredNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CompileDeferredNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CompileDeferredNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CompileDeferredNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CompileDeferredNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CompileDeferredNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CompileDeferredNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CompileDeferredNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CompileDeferredNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CompileDeferredNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CompileDeferredNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CompileDeferredNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CompileDeferredNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CompileDeferredNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CompileDeferredNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CompileDeferredNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CompileDeferredNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CompileDeferredNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CompileDeferredNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CompileDeferredNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CompileDeferredNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CompileDeferredNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CompileDeferredNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CompileDeferredNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CompileDeferredNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CompileDeferredNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CompileDeferredNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CompileDeferredNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CompileDeferredNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CompileDeferredNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CompileDeferredNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CompileDeferredNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CompileDeferredNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CompileDeferredNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CompileDeferredNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CompileDeferredNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CompileDeferredNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CompileDeferredNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CompileDeferredNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CompileDeferredNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CompileDeferredNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CompileDeferredNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CompileDeferredNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CompileDeferredNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CompileDeferredNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CompileDeferredNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CompileDeferredNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CompileDeferredNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CompileDeferredNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CompileDeferredNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CompileDeferredNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CompileDeferredNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CompileDeferredNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CompileDeferredNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CompileDeferredNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindInvocationMaskHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindInvocationMaskHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindInvocationMaskHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindInvocationMaskHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindInvocationMaskHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindInvocationMaskHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindInvocationMaskHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindInvocationMaskHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindInvocationMaskHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindInvocationMaskHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindInvocationMaskHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindInvocationMaskHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindInvocationMaskHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindInvocationMaskHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindInvocationMaskHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindInvocationMaskHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindInvocationMaskHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindInvocationMaskHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindInvocationMaskHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindInvocationMaskHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindInvocationMaskHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindInvocationMaskHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindInvocationMaskHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindInvocationMaskHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindInvocationMaskHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindInvocationMaskHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindInvocationMaskHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindInvocationMaskHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindInvocationMaskHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindInvocationMaskHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindInvocationMaskHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindInvocationMaskHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindInvocationMaskHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindInvocationMaskHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindInvocationMaskHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindInvocationMaskHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindInvocationMaskHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindInvocationMaskHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindInvocationMaskHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindInvocationMaskHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindInvocationMaskHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindInvocationMaskHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindInvocationMaskHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindInvocationMaskHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindInvocationMaskHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindInvocationMaskHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindInvocationMaskHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindInvocationMaskHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindInvocationMaskHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindInvocationMaskHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindInvocationMaskHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindInvocationMaskHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindInvocationMaskHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindInvocationMaskHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindInvocationMaskHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindInvocationMaskHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindInvocationMaskHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindInvocationMaskHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindInvocationMaskHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindInvocationMaskHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindInvocationMaskHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindInvocationMaskHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindInvocationMaskHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindInvocationMaskHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindInvocationMaskHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindInvocationMaskHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindInvocationMaskHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindInvocationMaskHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindInvocationMaskHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindInvocationMaskHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindInvocationMaskHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindInvocationMaskHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindInvocationMaskHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindInvocationMaskHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindInvocationMaskHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindInvocationMaskHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindInvocationMaskHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindInvocationMaskHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindInvocationMaskHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindInvocationMaskHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindInvocationMaskHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindInvocationMaskHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindInvocationMaskHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindInvocationMaskHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindInvocationMaskHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindInvocationMaskHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindInvocationMaskHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindInvocationMaskHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindInvocationMaskHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindInvocationMaskHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindInvocationMaskHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindInvocationMaskHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindInvocationMaskHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindInvocationMaskHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindInvocationMaskHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindInvocationMaskHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindInvocationMaskHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindInvocationMaskHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindInvocationMaskHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindInvocationMaskHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindInvocationMaskHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindInvocationMaskHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindInvocationMaskHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindInvocationMaskHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindInvocationMaskHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindInvocationMaskHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindInvocationMaskHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindInvocationMaskHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindInvocationMaskHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindInvocationMaskHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindInvocationMaskHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindInvocationMaskHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindInvocationMaskHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindInvocationMaskHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindInvocationMaskHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindInvocationMaskHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindInvocationMaskHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindInvocationMaskHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindInvocationMaskHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindInvocationMaskHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindInvocationMaskHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindInvocationMaskHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindInvocationMaskHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindInvocationMaskHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindInvocationMaskHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindInvocationMaskHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindInvocationMaskHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindInvocationMaskHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindInvocationMaskHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindInvocationMaskHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindInvocationMaskHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindInvocationMaskHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindInvocationMaskHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindInvocationMaskHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindInvocationMaskHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindInvocationMaskHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindInvocationMaskHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindInvocationMaskHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindInvocationMaskHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindInvocationMaskHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindInvocationMaskHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindInvocationMaskHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindInvocationMaskHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindInvocationMaskHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindInvocationMaskHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindInvocationMaskHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindInvocationMaskHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindInvocationMaskHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindInvocationMaskHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindInvocationMaskHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindInvocationMaskHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindInvocationMaskHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindInvocationMaskHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindInvocationMaskHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindInvocationMaskHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindInvocationMaskHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindInvocationMaskHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindInvocationMaskHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindInvocationMaskHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindInvocationMaskHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindInvocationMaskHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindInvocationMaskHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindInvocationMaskHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindInvocationMaskHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindInvocationMaskHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindInvocationMaskHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindInvocationMaskHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindInvocationMaskHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindInvocationMaskHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindInvocationMaskHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindInvocationMaskHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindInvocationMaskHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindInvocationMaskHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindInvocationMaskHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindInvocationMaskHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindInvocationMaskHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindInvocationMaskHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindInvocationMaskHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindInvocationMaskHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindInvocationMaskHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindInvocationMaskHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindInvocationMaskHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindInvocationMaskHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindInvocationMaskHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindInvocationMaskHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindInvocationMaskHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindInvocationMaskHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindInvocationMaskHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindInvocationMaskHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindInvocationMaskHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindInvocationMaskHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindInvocationMaskHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindInvocationMaskHUAWEI@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindInvocationMaskHUAWEI@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindInvocationMaskHUAWEI@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindInvocationMaskHUAWEI@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindInvocationMaskHUAWEI@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindInvocationMaskHUAWEI@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindInvocationMaskHUAWEI@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindInvocationMaskHUAWEI@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindInvocationMaskHUAWEI@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindInvocationMaskHUAWEI@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindInvocationMaskHUAWEI@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindInvocationMaskHUAWEI@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindInvocationMaskHUAWEI@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindInvocationMaskHUAWEI@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindInvocationMaskHUAWEI=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindInvocationMaskHUAWEI=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetAccelerationStructureMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetAccelerationStructureMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetAccelerationStructureMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetAccelerationStructureMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetAccelerationStructureMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetAccelerationStructureMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetAccelerationStructureMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetAccelerationStructureMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetAccelerationStructureMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetAccelerationStructureMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetAccelerationStructureMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetAccelerationStructureMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetAccelerationStructureMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetAccelerationStructureMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetAccelerationStructureMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetAccelerationStructureMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetAccelerationStructureMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetAccelerationStructureMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetAccelerationStructureMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetAccelerationStructureMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetAccelerationStructureMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetAccelerationStructureMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetAccelerationStructureMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetAccelerationStructureMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetAccelerationStructureMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetAccelerationStructureMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_BindAccelerationStructureMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindAccelerationStructureMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindAccelerationStructureMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindAccelerationStructureMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindAccelerationStructureMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindAccelerationStructureMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindAccelerationStructureMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindAccelerationStructureMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindAccelerationStructureMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindAccelerationStructureMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindAccelerationStructureMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindAccelerationStructureMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindAccelerationStructureMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindAccelerationStructureMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_BindAccelerationStructureMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_BindAccelerationStructureMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_BindAccelerationStructureMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindAccelerationStructureMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindAccelerationStructureMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindAccelerationStructureMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindAccelerationStructureMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindAccelerationStructureMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindAccelerationStructureMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindAccelerationStructureMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindAccelerationStructureMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindAccelerationStructureMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindAccelerationStructureMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindAccelerationStructureMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindAccelerationStructureMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindAccelerationStructureMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_BindAccelerationStructureMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_BindAccelerationStructureMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_BindAccelerationStructureMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindAccelerationStructureMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindAccelerationStructureMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindAccelerationStructureMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindAccelerationStructureMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindAccelerationStructureMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindAccelerationStructureMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindAccelerationStructureMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindAccelerationStructureMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindAccelerationStructureMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindAccelerationStructureMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindAccelerationStructureMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindAccelerationStructureMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindAccelerationStructureMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_BindAccelerationStructureMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_BindAccelerationStructureMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_BindAccelerationStructureMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindAccelerationStructureMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindAccelerationStructureMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindAccelerationStructureMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindAccelerationStructureMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindAccelerationStructureMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindAccelerationStructureMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindAccelerationStructureMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindAccelerationStructureMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindAccelerationStructureMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindAccelerationStructureMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindAccelerationStructureMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindAccelerationStructureMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindAccelerationStructureMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_BindAccelerationStructureMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_BindAccelerationStructureMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_BindAccelerationStructureMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindAccelerationStructureMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindAccelerationStructureMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindAccelerationStructureMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindAccelerationStructureMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindAccelerationStructureMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindAccelerationStructureMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindAccelerationStructureMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindAccelerationStructureMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindAccelerationStructureMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindAccelerationStructureMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindAccelerationStructureMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindAccelerationStructureMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindAccelerationStructureMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_BindAccelerationStructureMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_BindAccelerationStructureMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_BindAccelerationStructureMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindAccelerationStructureMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindAccelerationStructureMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindAccelerationStructureMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindAccelerationStructureMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindAccelerationStructureMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindAccelerationStructureMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindAccelerationStructureMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindAccelerationStructureMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindAccelerationStructureMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindAccelerationStructureMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindAccelerationStructureMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindAccelerationStructureMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindAccelerationStructureMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_BindAccelerationStructureMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_BindAccelerationStructureMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_BindAccelerationStructureMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindAccelerationStructureMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindAccelerationStructureMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindAccelerationStructureMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindAccelerationStructureMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindAccelerationStructureMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindAccelerationStructureMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindAccelerationStructureMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindAccelerationStructureMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindAccelerationStructureMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindAccelerationStructureMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindAccelerationStructureMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindAccelerationStructureMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindAccelerationStructureMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_BindAccelerationStructureMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_BindAccelerationStructureMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_BindAccelerationStructureMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindAccelerationStructureMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindAccelerationStructureMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindAccelerationStructureMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindAccelerationStructureMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindAccelerationStructureMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindAccelerationStructureMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindAccelerationStructureMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindAccelerationStructureMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindAccelerationStructureMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindAccelerationStructureMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindAccelerationStructureMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindAccelerationStructureMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindAccelerationStructureMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_BindAccelerationStructureMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_BindAccelerationStructureMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_BindAccelerationStructureMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindAccelerationStructureMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindAccelerationStructureMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindAccelerationStructureMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindAccelerationStructureMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindAccelerationStructureMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindAccelerationStructureMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindAccelerationStructureMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindAccelerationStructureMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindAccelerationStructureMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindAccelerationStructureMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindAccelerationStructureMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindAccelerationStructureMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindAccelerationStructureMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_BindAccelerationStructureMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_BindAccelerationStructureMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_BindAccelerationStructureMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindAccelerationStructureMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindAccelerationStructureMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindAccelerationStructureMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindAccelerationStructureMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindAccelerationStructureMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindAccelerationStructureMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindAccelerationStructureMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindAccelerationStructureMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindAccelerationStructureMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindAccelerationStructureMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindAccelerationStructureMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindAccelerationStructureMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindAccelerationStructureMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_BindAccelerationStructureMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_BindAccelerationStructureMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindAccelerationStructureMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindAccelerationStructureMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindAccelerationStructureMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindAccelerationStructureMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindAccelerationStructureMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindAccelerationStructureMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindAccelerationStructureMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindAccelerationStructureMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindAccelerationStructureMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindAccelerationStructureMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindAccelerationStructureMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindAccelerationStructureMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindAccelerationStructureMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindAccelerationStructureMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_BindAccelerationStructureMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_BindAccelerationStructureMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_BindAccelerationStructureMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindAccelerationStructureMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindAccelerationStructureMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindAccelerationStructureMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindAccelerationStructureMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindAccelerationStructureMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindAccelerationStructureMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindAccelerationStructureMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindAccelerationStructureMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindAccelerationStructureMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindAccelerationStructureMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindAccelerationStructureMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindAccelerationStructureMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindAccelerationStructureMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_BindAccelerationStructureMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_BindAccelerationStructureMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_BindAccelerationStructureMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindAccelerationStructureMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindAccelerationStructureMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindAccelerationStructureMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindAccelerationStructureMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindAccelerationStructureMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindAccelerationStructureMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindAccelerationStructureMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindAccelerationStructureMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindAccelerationStructureMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindAccelerationStructureMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindAccelerationStructureMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindAccelerationStructureMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindAccelerationStructureMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_BindAccelerationStructureMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_BindAccelerationStructureMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CopyAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CopyAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyAccelerationStructureToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CopyAccelerationStructureToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CopyAccelerationStructureToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CopyMemoryToAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CopyMemoryToAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdWriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdWriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdWriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdWriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdWriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdWriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdWriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdWriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdWriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdWriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdWriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdWriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdWriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdWriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdWriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdWriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdWriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdWriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdWriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdWriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdWriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdWriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdWriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdWriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdWriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdWriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteAccelerationStructuresPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdWriteAccelerationStructuresPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdWriteAccelerationStructuresPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteAccelerationStructuresPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdWriteAccelerationStructuresPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdWriteAccelerationStructuresPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteAccelerationStructuresPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdWriteAccelerationStructuresPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdWriteAccelerationStructuresPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteAccelerationStructuresPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdWriteAccelerationStructuresPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdWriteAccelerationStructuresPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteAccelerationStructuresPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdWriteAccelerationStructuresPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdWriteAccelerationStructuresPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteAccelerationStructuresPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdWriteAccelerationStructuresPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdWriteAccelerationStructuresPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteAccelerationStructuresPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdWriteAccelerationStructuresPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdWriteAccelerationStructuresPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteAccelerationStructuresPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdWriteAccelerationStructuresPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdWriteAccelerationStructuresPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteAccelerationStructuresPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdWriteAccelerationStructuresPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdWriteAccelerationStructuresPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteAccelerationStructuresPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdWriteAccelerationStructuresPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdWriteAccelerationStructuresPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteAccelerationStructuresPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdWriteAccelerationStructuresPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdWriteAccelerationStructuresPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteAccelerationStructuresPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdWriteAccelerationStructuresPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdWriteAccelerationStructuresPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteAccelerationStructuresPropertiesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdWriteAccelerationStructuresPropertiesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdWriteAccelerationStructuresPropertiesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBuildAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBuildAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBuildAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBuildAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBuildAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBuildAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBuildAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBuildAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBuildAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBuildAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBuildAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBuildAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBuildAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBuildAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBuildAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBuildAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBuildAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBuildAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBuildAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBuildAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBuildAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBuildAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBuildAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBuildAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructureNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructureNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructureNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructureNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructureNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructureNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructureNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructureNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructureNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructureNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructureNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructureNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructureNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructureNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBuildAccelerationStructureNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBuildAccelerationStructureNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_WriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_WriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_WriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_WriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_WriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_WriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_WriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_WriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_WriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_WriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_WriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_WriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_WriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_WriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_WriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_WriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_WriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_WriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_WriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_WriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_WriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_WriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_WriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_WriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_WriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_WriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_WriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_WriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_WriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_WriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_WriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_WriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_WriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteAccelerationStructuresPropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteAccelerationStructuresPropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteAccelerationStructuresPropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteAccelerationStructuresPropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteAccelerationStructuresPropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteAccelerationStructuresPropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteAccelerationStructuresPropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteAccelerationStructuresPropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteAccelerationStructuresPropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteAccelerationStructuresPropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteAccelerationStructuresPropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteAccelerationStructuresPropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteAccelerationStructuresPropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteAccelerationStructuresPropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_WriteAccelerationStructuresPropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_WriteAccelerationStructuresPropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdTraceRaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdTraceRaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdTraceRaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdTraceRaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdTraceRaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdTraceRaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdTraceRaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdTraceRaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdTraceRaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdTraceRaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdTraceRaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdTraceRaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdTraceRaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdTraceRaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdTraceRaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdTraceRaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdTraceRaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdTraceRaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdTraceRaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdTraceRaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdTraceRaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdTraceRaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdTraceRaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdTraceRaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdTraceRaysKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdTraceRaysKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdTraceRaysNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdTraceRaysNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdTraceRaysNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdTraceRaysNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdTraceRaysNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdTraceRaysNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdTraceRaysNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdTraceRaysNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdTraceRaysNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdTraceRaysNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdTraceRaysNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdTraceRaysNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdTraceRaysNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdTraceRaysNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdTraceRaysNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdTraceRaysNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdTraceRaysNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdTraceRaysNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdTraceRaysNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdTraceRaysNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdTraceRaysNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdTraceRaysNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdTraceRaysNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdTraceRaysNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdTraceRaysNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdTraceRaysNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetRayTracingShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetRayTracingShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetRayTracingShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetRayTracingShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetRayTracingShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetRayTracingShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetRayTracingShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetRayTracingShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetRayTracingShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetRayTracingShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetRayTracingShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetRayTracingShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetRayTracingShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetRayTracingShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetRayTracingShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetRayTracingShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetRayTracingShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetRayTracingShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetRayTracingShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetRayTracingShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetRayTracingShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetRayTracingShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetRayTracingShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetRayTracingShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetRayTracingShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetRayTracingShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupHandlesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetRayTracingShaderGroupHandlesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetRayTracingShaderGroupHandlesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupHandlesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetRayTracingShaderGroupHandlesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetRayTracingShaderGroupHandlesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupHandlesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetRayTracingShaderGroupHandlesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetRayTracingShaderGroupHandlesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupHandlesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetRayTracingShaderGroupHandlesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetRayTracingShaderGroupHandlesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupHandlesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetRayTracingShaderGroupHandlesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetRayTracingShaderGroupHandlesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupHandlesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetRayTracingShaderGroupHandlesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetRayTracingShaderGroupHandlesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupHandlesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetRayTracingShaderGroupHandlesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetRayTracingShaderGroupHandlesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupHandlesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetRayTracingShaderGroupHandlesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetRayTracingShaderGroupHandlesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupHandlesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetRayTracingShaderGroupHandlesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetRayTracingShaderGroupHandlesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupHandlesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetRayTracingShaderGroupHandlesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetRayTracingShaderGroupHandlesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupHandlesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetRayTracingShaderGroupHandlesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetRayTracingShaderGroupHandlesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupHandlesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetRayTracingShaderGroupHandlesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetRayTracingShaderGroupHandlesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupHandlesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetRayTracingShaderGroupHandlesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetRayTracingShaderGroupHandlesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetRayTracingCaptureReplayShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetRayTracingCaptureReplayShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingCaptureReplayShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingCaptureReplayShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingCaptureReplayShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingCaptureReplayShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingCaptureReplayShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingCaptureReplayShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingCaptureReplayShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingCaptureReplayShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingCaptureReplayShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingCaptureReplayShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingCaptureReplayShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingCaptureReplayShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingCaptureReplayShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingCaptureReplayShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetRayTracingCaptureReplayShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetRayTracingCaptureReplayShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingCaptureReplayShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingCaptureReplayShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingCaptureReplayShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingCaptureReplayShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingCaptureReplayShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingCaptureReplayShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingCaptureReplayShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingCaptureReplayShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingCaptureReplayShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingCaptureReplayShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingCaptureReplayShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingCaptureReplayShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingCaptureReplayShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingCaptureReplayShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetRayTracingCaptureReplayShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetRayTracingCaptureReplayShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingCaptureReplayShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingCaptureReplayShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingCaptureReplayShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingCaptureReplayShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingCaptureReplayShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingCaptureReplayShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingCaptureReplayShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingCaptureReplayShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingCaptureReplayShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingCaptureReplayShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingCaptureReplayShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingCaptureReplayShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingCaptureReplayShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingCaptureReplayShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetRayTracingCaptureReplayShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetRayTracingCaptureReplayShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingCaptureReplayShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingCaptureReplayShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingCaptureReplayShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingCaptureReplayShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingCaptureReplayShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingCaptureReplayShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingCaptureReplayShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingCaptureReplayShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingCaptureReplayShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingCaptureReplayShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingCaptureReplayShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingCaptureReplayShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingCaptureReplayShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingCaptureReplayShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetRayTracingCaptureReplayShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetRayTracingCaptureReplayShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingCaptureReplayShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingCaptureReplayShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingCaptureReplayShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingCaptureReplayShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingCaptureReplayShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingCaptureReplayShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingCaptureReplayShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingCaptureReplayShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingCaptureReplayShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingCaptureReplayShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingCaptureReplayShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingCaptureReplayShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingCaptureReplayShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingCaptureReplayShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetRayTracingCaptureReplayShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetRayTracingCaptureReplayShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingCaptureReplayShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingCaptureReplayShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingCaptureReplayShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingCaptureReplayShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingCaptureReplayShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingCaptureReplayShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingCaptureReplayShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingCaptureReplayShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingCaptureReplayShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingCaptureReplayShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingCaptureReplayShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingCaptureReplayShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingCaptureReplayShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingCaptureReplayShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetRayTracingCaptureReplayShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetRayTracingCaptureReplayShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingCaptureReplayShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingCaptureReplayShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingCaptureReplayShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingCaptureReplayShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingCaptureReplayShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingCaptureReplayShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingCaptureReplayShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingCaptureReplayShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingCaptureReplayShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingCaptureReplayShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingCaptureReplayShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingCaptureReplayShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingCaptureReplayShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingCaptureReplayShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetRayTracingCaptureReplayShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetRayTracingCaptureReplayShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingCaptureReplayShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingCaptureReplayShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingCaptureReplayShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingCaptureReplayShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingCaptureReplayShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingCaptureReplayShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingCaptureReplayShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingCaptureReplayShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingCaptureReplayShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingCaptureReplayShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingCaptureReplayShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingCaptureReplayShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingCaptureReplayShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingCaptureReplayShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetRayTracingCaptureReplayShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetRayTracingCaptureReplayShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingCaptureReplayShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingCaptureReplayShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingCaptureReplayShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingCaptureReplayShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingCaptureReplayShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingCaptureReplayShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingCaptureReplayShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingCaptureReplayShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingCaptureReplayShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingCaptureReplayShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingCaptureReplayShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingCaptureReplayShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingCaptureReplayShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingCaptureReplayShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetRayTracingCaptureReplayShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetRayTracingCaptureReplayShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingCaptureReplayShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingCaptureReplayShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingCaptureReplayShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingCaptureReplayShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingCaptureReplayShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingCaptureReplayShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingCaptureReplayShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingCaptureReplayShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingCaptureReplayShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingCaptureReplayShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingCaptureReplayShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingCaptureReplayShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingCaptureReplayShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingCaptureReplayShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetRayTracingCaptureReplayShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetRayTracingCaptureReplayShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingCaptureReplayShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingCaptureReplayShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingCaptureReplayShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingCaptureReplayShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingCaptureReplayShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingCaptureReplayShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingCaptureReplayShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingCaptureReplayShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingCaptureReplayShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingCaptureReplayShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingCaptureReplayShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingCaptureReplayShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingCaptureReplayShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingCaptureReplayShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetRayTracingCaptureReplayShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetRayTracingCaptureReplayShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingCaptureReplayShaderGroupHandlesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetRayTracingCaptureReplayShaderGroupHandlesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetRayTracingCaptureReplayShaderGroupHandlesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureHandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureHandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureHandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureHandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureHandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureHandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureHandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureHandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureHandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureHandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureHandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureHandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureHandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureHandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetAccelerationStructureHandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetAccelerationStructureHandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureHandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureHandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureHandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureHandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureHandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureHandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureHandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureHandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureHandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureHandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureHandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureHandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureHandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureHandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetAccelerationStructureHandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetAccelerationStructureHandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureHandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureHandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureHandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureHandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureHandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureHandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureHandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureHandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureHandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureHandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureHandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureHandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureHandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureHandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetAccelerationStructureHandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetAccelerationStructureHandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureHandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureHandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureHandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureHandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureHandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureHandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureHandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureHandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureHandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureHandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureHandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureHandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureHandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureHandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetAccelerationStructureHandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetAccelerationStructureHandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureHandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureHandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureHandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureHandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureHandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureHandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureHandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureHandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureHandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureHandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureHandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureHandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureHandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureHandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetAccelerationStructureHandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetAccelerationStructureHandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureHandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureHandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureHandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureHandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureHandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureHandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureHandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureHandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureHandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureHandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureHandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureHandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureHandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureHandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetAccelerationStructureHandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetAccelerationStructureHandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureHandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureHandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureHandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureHandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureHandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureHandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureHandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureHandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureHandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureHandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureHandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureHandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureHandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureHandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetAccelerationStructureHandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetAccelerationStructureHandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureHandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureHandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureHandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureHandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureHandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureHandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureHandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureHandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureHandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureHandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureHandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureHandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureHandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureHandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetAccelerationStructureHandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetAccelerationStructureHandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureHandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureHandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureHandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureHandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureHandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureHandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureHandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureHandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureHandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureHandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureHandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureHandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureHandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureHandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetAccelerationStructureHandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetAccelerationStructureHandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureHandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureHandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureHandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureHandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureHandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureHandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureHandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureHandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureHandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureHandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureHandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureHandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureHandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureHandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetAccelerationStructureHandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetAccelerationStructureHandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureHandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureHandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureHandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureHandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureHandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureHandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureHandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureHandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureHandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureHandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureHandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureHandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureHandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureHandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetAccelerationStructureHandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetAccelerationStructureHandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureHandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureHandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureHandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureHandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureHandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureHandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureHandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureHandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureHandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureHandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureHandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureHandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureHandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureHandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetAccelerationStructureHandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetAccelerationStructureHandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureHandleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureHandleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureHandleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureHandleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureHandleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureHandleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureHandleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureHandleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureHandleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureHandleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureHandleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureHandleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureHandleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureHandleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetAccelerationStructureHandleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetAccelerationStructureHandleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateRayTracingPipelinesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateRayTracingPipelinesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateRayTracingPipelinesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateRayTracingPipelinesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateRayTracingPipelinesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateRayTracingPipelinesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateRayTracingPipelinesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateRayTracingPipelinesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateRayTracingPipelinesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateRayTracingPipelinesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateRayTracingPipelinesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateRayTracingPipelinesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateRayTracingPipelinesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateRayTracingPipelinesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateRayTracingPipelinesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateRayTracingPipelinesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateRayTracingPipelinesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateRayTracingPipelinesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateRayTracingPipelinesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateRayTracingPipelinesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateRayTracingPipelinesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateRayTracingPipelinesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateRayTracingPipelinesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateRayTracingPipelinesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateRayTracingPipelinesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateRayTracingPipelinesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateRayTracingPipelinesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateRayTracingPipelinesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateRayTracingPipelinesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateRayTracingPipelinesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateRayTracingPipelinesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateRayTracingPipelinesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateRayTracingPipelinesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateRayTracingPipelinesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateRayTracingPipelinesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateRayTracingPipelinesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateRayTracingPipelinesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateRayTracingPipelinesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateRayTracingPipelinesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateRayTracingPipelinesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateRayTracingPipelinesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateRayTracingPipelinesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateRayTracingPipelinesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateRayTracingPipelinesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateRayTracingPipelinesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateRayTracingPipelinesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateRayTracingPipelinesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateRayTracingPipelinesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateRayTracingPipelinesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateRayTracingPipelinesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateRayTracingPipelinesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateRayTracingPipelinesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateRayTracingPipelinesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateRayTracingPipelinesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateRayTracingPipelinesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateRayTracingPipelinesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateRayTracingPipelinesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateRayTracingPipelinesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateRayTracingPipelinesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateRayTracingPipelinesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateRayTracingPipelinesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateRayTracingPipelinesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateRayTracingPipelinesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateRayTracingPipelinesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateRayTracingPipelinesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdTraceRaysIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdTraceRaysIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdTraceRaysIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdTraceRaysIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdTraceRaysIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdTraceRaysIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdTraceRaysIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdTraceRaysIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdTraceRaysIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdTraceRaysIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdTraceRaysIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdTraceRaysIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdTraceRaysIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdTraceRaysIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdTraceRaysIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdTraceRaysIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdTraceRaysIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdTraceRaysIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdTraceRaysIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdTraceRaysIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdTraceRaysIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdTraceRaysIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdTraceRaysIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdTraceRaysIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdTraceRaysIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdTraceRaysIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdTraceRaysIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdTraceRaysIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdTraceRaysIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdTraceRaysIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdTraceRaysIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdTraceRaysIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdTraceRaysIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdTraceRaysIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdTraceRaysIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdTraceRaysIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdTraceRaysIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdTraceRaysIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdTraceRaysIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdTraceRaysIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdTraceRaysIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdTraceRaysIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdTraceRaysIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdTraceRaysIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdTraceRaysIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdTraceRaysIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdTraceRaysIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdTraceRaysIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdTraceRaysIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdTraceRaysIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdTraceRaysIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdTraceRaysIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdTraceRaysIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdTraceRaysIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdTraceRaysIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdTraceRaysIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdTraceRaysIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdTraceRaysIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdTraceRaysIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdTraceRaysIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdTraceRaysIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdTraceRaysIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdTraceRaysIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdTraceRaysIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdTraceRaysIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetClusterAccelerationStructureBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetClusterAccelerationStructureBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetClusterAccelerationStructureBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetClusterAccelerationStructureBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetClusterAccelerationStructureBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetClusterAccelerationStructureBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetClusterAccelerationStructureBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetClusterAccelerationStructureBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetClusterAccelerationStructureBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetClusterAccelerationStructureBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetClusterAccelerationStructureBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetClusterAccelerationStructureBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetClusterAccelerationStructureBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetClusterAccelerationStructureBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetClusterAccelerationStructureBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetClusterAccelerationStructureBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetClusterAccelerationStructureBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetClusterAccelerationStructureBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetClusterAccelerationStructureBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetClusterAccelerationStructureBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetClusterAccelerationStructureBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetClusterAccelerationStructureBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetClusterAccelerationStructureBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetClusterAccelerationStructureBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetClusterAccelerationStructureBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetClusterAccelerationStructureBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetClusterAccelerationStructureBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetClusterAccelerationStructureBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetClusterAccelerationStructureBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetClusterAccelerationStructureBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetClusterAccelerationStructureBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetClusterAccelerationStructureBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetClusterAccelerationStructureBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetClusterAccelerationStructureBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetClusterAccelerationStructureBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetClusterAccelerationStructureBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetClusterAccelerationStructureBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetClusterAccelerationStructureBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetClusterAccelerationStructureBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetClusterAccelerationStructureBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetClusterAccelerationStructureBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetClusterAccelerationStructureBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetClusterAccelerationStructureBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetClusterAccelerationStructureBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetClusterAccelerationStructureBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetClusterAccelerationStructureBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetClusterAccelerationStructureBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetClusterAccelerationStructureBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetClusterAccelerationStructureBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetClusterAccelerationStructureBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetClusterAccelerationStructureBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetClusterAccelerationStructureBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetClusterAccelerationStructureBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetClusterAccelerationStructureBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetClusterAccelerationStructureBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetClusterAccelerationStructureBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetClusterAccelerationStructureBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetClusterAccelerationStructureBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetClusterAccelerationStructureBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetClusterAccelerationStructureBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetClusterAccelerationStructureBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetClusterAccelerationStructureBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetClusterAccelerationStructureBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetClusterAccelerationStructureBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetClusterAccelerationStructureBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetClusterAccelerationStructureBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetClusterAccelerationStructureBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetClusterAccelerationStructureBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetClusterAccelerationStructureBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetClusterAccelerationStructureBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetClusterAccelerationStructureBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetClusterAccelerationStructureBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetClusterAccelerationStructureBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetClusterAccelerationStructureBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetClusterAccelerationStructureBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetClusterAccelerationStructureBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetClusterAccelerationStructureBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetClusterAccelerationStructureBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetClusterAccelerationStructureBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetClusterAccelerationStructureBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetClusterAccelerationStructureBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetClusterAccelerationStructureBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetClusterAccelerationStructureBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetClusterAccelerationStructureBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetClusterAccelerationStructureBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetClusterAccelerationStructureBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetClusterAccelerationStructureBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetClusterAccelerationStructureBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetClusterAccelerationStructureBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetClusterAccelerationStructureBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetClusterAccelerationStructureBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetClusterAccelerationStructureBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetClusterAccelerationStructureBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetClusterAccelerationStructureBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetClusterAccelerationStructureBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetClusterAccelerationStructureBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetClusterAccelerationStructureBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetClusterAccelerationStructureBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetClusterAccelerationStructureBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetClusterAccelerationStructureBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetClusterAccelerationStructureBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetClusterAccelerationStructureBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetClusterAccelerationStructureBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetClusterAccelerationStructureBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetClusterAccelerationStructureBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetClusterAccelerationStructureBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetClusterAccelerationStructureBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetClusterAccelerationStructureBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetClusterAccelerationStructureBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetClusterAccelerationStructureBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetClusterAccelerationStructureBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetClusterAccelerationStructureBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetClusterAccelerationStructureBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetClusterAccelerationStructureBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetClusterAccelerationStructureBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetClusterAccelerationStructureBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetClusterAccelerationStructureBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetClusterAccelerationStructureBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetClusterAccelerationStructureBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetClusterAccelerationStructureBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetClusterAccelerationStructureBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetClusterAccelerationStructureBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetClusterAccelerationStructureBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetClusterAccelerationStructureBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetClusterAccelerationStructureBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetClusterAccelerationStructureBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetClusterAccelerationStructureBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetClusterAccelerationStructureBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetClusterAccelerationStructureBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetClusterAccelerationStructureBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetClusterAccelerationStructureBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetClusterAccelerationStructureBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetClusterAccelerationStructureBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetClusterAccelerationStructureBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetClusterAccelerationStructureBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetClusterAccelerationStructureBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetClusterAccelerationStructureBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetClusterAccelerationStructureBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetClusterAccelerationStructureBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetClusterAccelerationStructureBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetClusterAccelerationStructureBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetClusterAccelerationStructureBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetClusterAccelerationStructureBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetClusterAccelerationStructureBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetClusterAccelerationStructureBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetClusterAccelerationStructureBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetClusterAccelerationStructureBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetClusterAccelerationStructureBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetClusterAccelerationStructureBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetClusterAccelerationStructureBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetClusterAccelerationStructureBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetClusterAccelerationStructureBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetClusterAccelerationStructureBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetClusterAccelerationStructureBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetClusterAccelerationStructureBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetClusterAccelerationStructureBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetClusterAccelerationStructureBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetClusterAccelerationStructureBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetClusterAccelerationStructureBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetClusterAccelerationStructureBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetClusterAccelerationStructureBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetClusterAccelerationStructureBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetClusterAccelerationStructureBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetClusterAccelerationStructureBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetClusterAccelerationStructureBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetClusterAccelerationStructureBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetClusterAccelerationStructureBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetClusterAccelerationStructureBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetClusterAccelerationStructureBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetClusterAccelerationStructureBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetClusterAccelerationStructureBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetClusterAccelerationStructureBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetClusterAccelerationStructureBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetClusterAccelerationStructureBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetClusterAccelerationStructureBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetClusterAccelerationStructureBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetClusterAccelerationStructureBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetClusterAccelerationStructureBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetClusterAccelerationStructureBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetClusterAccelerationStructureBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetClusterAccelerationStructureBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetClusterAccelerationStructureBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetClusterAccelerationStructureBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetClusterAccelerationStructureBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetClusterAccelerationStructureBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetClusterAccelerationStructureBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetClusterAccelerationStructureBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetClusterAccelerationStructureBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetClusterAccelerationStructureBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetClusterAccelerationStructureBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetClusterAccelerationStructureBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetClusterAccelerationStructureBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetClusterAccelerationStructureBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetClusterAccelerationStructureBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetClusterAccelerationStructureBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetClusterAccelerationStructureBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetClusterAccelerationStructureBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetClusterAccelerationStructureBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetClusterAccelerationStructureBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetClusterAccelerationStructureBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetClusterAccelerationStructureBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetClusterAccelerationStructureBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetClusterAccelerationStructureBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetClusterAccelerationStructureBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetClusterAccelerationStructureBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetClusterAccelerationStructureBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetClusterAccelerationStructureBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetClusterAccelerationStructureBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBuildClusterAccelerationStructureIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildClusterAccelerationStructureIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildClusterAccelerationStructureIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildClusterAccelerationStructureIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildClusterAccelerationStructureIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildClusterAccelerationStructureIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildClusterAccelerationStructureIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildClusterAccelerationStructureIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildClusterAccelerationStructureIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildClusterAccelerationStructureIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildClusterAccelerationStructureIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildClusterAccelerationStructureIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildClusterAccelerationStructureIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildClusterAccelerationStructureIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBuildClusterAccelerationStructureIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBuildClusterAccelerationStructureIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildClusterAccelerationStructureIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildClusterAccelerationStructureIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildClusterAccelerationStructureIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildClusterAccelerationStructureIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildClusterAccelerationStructureIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildClusterAccelerationStructureIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildClusterAccelerationStructureIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildClusterAccelerationStructureIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildClusterAccelerationStructureIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildClusterAccelerationStructureIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildClusterAccelerationStructureIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildClusterAccelerationStructureIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildClusterAccelerationStructureIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildClusterAccelerationStructureIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBuildClusterAccelerationStructureIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBuildClusterAccelerationStructureIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildClusterAccelerationStructureIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildClusterAccelerationStructureIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildClusterAccelerationStructureIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildClusterAccelerationStructureIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildClusterAccelerationStructureIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildClusterAccelerationStructureIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildClusterAccelerationStructureIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildClusterAccelerationStructureIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildClusterAccelerationStructureIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildClusterAccelerationStructureIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildClusterAccelerationStructureIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildClusterAccelerationStructureIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildClusterAccelerationStructureIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildClusterAccelerationStructureIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBuildClusterAccelerationStructureIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBuildClusterAccelerationStructureIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildClusterAccelerationStructureIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildClusterAccelerationStructureIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildClusterAccelerationStructureIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildClusterAccelerationStructureIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildClusterAccelerationStructureIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildClusterAccelerationStructureIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildClusterAccelerationStructureIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildClusterAccelerationStructureIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildClusterAccelerationStructureIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildClusterAccelerationStructureIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildClusterAccelerationStructureIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildClusterAccelerationStructureIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildClusterAccelerationStructureIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildClusterAccelerationStructureIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBuildClusterAccelerationStructureIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBuildClusterAccelerationStructureIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildClusterAccelerationStructureIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildClusterAccelerationStructureIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildClusterAccelerationStructureIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildClusterAccelerationStructureIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildClusterAccelerationStructureIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildClusterAccelerationStructureIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildClusterAccelerationStructureIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildClusterAccelerationStructureIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildClusterAccelerationStructureIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildClusterAccelerationStructureIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildClusterAccelerationStructureIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildClusterAccelerationStructureIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildClusterAccelerationStructureIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildClusterAccelerationStructureIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBuildClusterAccelerationStructureIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBuildClusterAccelerationStructureIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildClusterAccelerationStructureIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildClusterAccelerationStructureIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildClusterAccelerationStructureIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildClusterAccelerationStructureIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildClusterAccelerationStructureIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildClusterAccelerationStructureIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildClusterAccelerationStructureIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildClusterAccelerationStructureIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildClusterAccelerationStructureIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildClusterAccelerationStructureIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildClusterAccelerationStructureIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildClusterAccelerationStructureIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildClusterAccelerationStructureIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildClusterAccelerationStructureIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBuildClusterAccelerationStructureIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBuildClusterAccelerationStructureIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildClusterAccelerationStructureIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildClusterAccelerationStructureIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildClusterAccelerationStructureIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildClusterAccelerationStructureIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildClusterAccelerationStructureIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildClusterAccelerationStructureIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildClusterAccelerationStructureIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildClusterAccelerationStructureIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildClusterAccelerationStructureIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildClusterAccelerationStructureIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildClusterAccelerationStructureIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildClusterAccelerationStructureIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildClusterAccelerationStructureIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildClusterAccelerationStructureIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBuildClusterAccelerationStructureIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBuildClusterAccelerationStructureIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildClusterAccelerationStructureIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildClusterAccelerationStructureIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildClusterAccelerationStructureIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildClusterAccelerationStructureIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildClusterAccelerationStructureIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildClusterAccelerationStructureIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildClusterAccelerationStructureIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildClusterAccelerationStructureIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildClusterAccelerationStructureIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildClusterAccelerationStructureIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildClusterAccelerationStructureIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildClusterAccelerationStructureIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildClusterAccelerationStructureIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildClusterAccelerationStructureIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBuildClusterAccelerationStructureIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBuildClusterAccelerationStructureIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildClusterAccelerationStructureIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildClusterAccelerationStructureIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildClusterAccelerationStructureIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildClusterAccelerationStructureIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildClusterAccelerationStructureIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildClusterAccelerationStructureIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildClusterAccelerationStructureIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildClusterAccelerationStructureIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildClusterAccelerationStructureIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildClusterAccelerationStructureIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildClusterAccelerationStructureIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildClusterAccelerationStructureIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildClusterAccelerationStructureIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildClusterAccelerationStructureIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBuildClusterAccelerationStructureIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBuildClusterAccelerationStructureIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildClusterAccelerationStructureIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildClusterAccelerationStructureIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildClusterAccelerationStructureIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildClusterAccelerationStructureIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildClusterAccelerationStructureIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildClusterAccelerationStructureIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildClusterAccelerationStructureIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildClusterAccelerationStructureIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildClusterAccelerationStructureIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildClusterAccelerationStructureIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildClusterAccelerationStructureIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildClusterAccelerationStructureIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildClusterAccelerationStructureIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildClusterAccelerationStructureIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBuildClusterAccelerationStructureIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBuildClusterAccelerationStructureIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildClusterAccelerationStructureIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildClusterAccelerationStructureIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildClusterAccelerationStructureIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildClusterAccelerationStructureIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildClusterAccelerationStructureIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildClusterAccelerationStructureIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildClusterAccelerationStructureIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildClusterAccelerationStructureIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildClusterAccelerationStructureIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildClusterAccelerationStructureIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildClusterAccelerationStructureIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildClusterAccelerationStructureIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildClusterAccelerationStructureIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildClusterAccelerationStructureIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBuildClusterAccelerationStructureIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBuildClusterAccelerationStructureIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildClusterAccelerationStructureIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildClusterAccelerationStructureIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildClusterAccelerationStructureIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildClusterAccelerationStructureIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildClusterAccelerationStructureIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildClusterAccelerationStructureIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildClusterAccelerationStructureIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildClusterAccelerationStructureIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildClusterAccelerationStructureIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildClusterAccelerationStructureIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildClusterAccelerationStructureIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildClusterAccelerationStructureIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildClusterAccelerationStructureIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildClusterAccelerationStructureIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBuildClusterAccelerationStructureIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBuildClusterAccelerationStructureIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildClusterAccelerationStructureIndirectNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildClusterAccelerationStructureIndirectNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildClusterAccelerationStructureIndirectNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildClusterAccelerationStructureIndirectNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildClusterAccelerationStructureIndirectNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildClusterAccelerationStructureIndirectNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildClusterAccelerationStructureIndirectNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildClusterAccelerationStructureIndirectNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildClusterAccelerationStructureIndirectNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildClusterAccelerationStructureIndirectNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildClusterAccelerationStructureIndirectNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildClusterAccelerationStructureIndirectNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildClusterAccelerationStructureIndirectNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildClusterAccelerationStructureIndirectNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBuildClusterAccelerationStructureIndirectNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBuildClusterAccelerationStructureIndirectNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceAccelerationStructureCompatibilityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceAccelerationStructureCompatibilityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceAccelerationStructureCompatibilityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceAccelerationStructureCompatibilityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceAccelerationStructureCompatibilityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceAccelerationStructureCompatibilityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceAccelerationStructureCompatibilityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceAccelerationStructureCompatibilityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceAccelerationStructureCompatibilityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceAccelerationStructureCompatibilityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceAccelerationStructureCompatibilityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceAccelerationStructureCompatibilityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceAccelerationStructureCompatibilityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceAccelerationStructureCompatibilityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceAccelerationStructureCompatibilityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceAccelerationStructureCompatibilityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceAccelerationStructureCompatibilityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceAccelerationStructureCompatibilityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceAccelerationStructureCompatibilityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceAccelerationStructureCompatibilityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceAccelerationStructureCompatibilityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceAccelerationStructureCompatibilityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceAccelerationStructureCompatibilityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceAccelerationStructureCompatibilityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceAccelerationStructureCompatibilityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceAccelerationStructureCompatibilityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceAccelerationStructureCompatibilityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceAccelerationStructureCompatibilityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceAccelerationStructureCompatibilityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceAccelerationStructureCompatibilityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceAccelerationStructureCompatibilityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceAccelerationStructureCompatibilityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceAccelerationStructureCompatibilityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceAccelerationStructureCompatibilityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceAccelerationStructureCompatibilityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceAccelerationStructureCompatibilityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceAccelerationStructureCompatibilityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceAccelerationStructureCompatibilityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceAccelerationStructureCompatibilityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceAccelerationStructureCompatibilityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceAccelerationStructureCompatibilityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceAccelerationStructureCompatibilityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceAccelerationStructureCompatibilityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceAccelerationStructureCompatibilityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceAccelerationStructureCompatibilityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceAccelerationStructureCompatibilityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceAccelerationStructureCompatibilityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceAccelerationStructureCompatibilityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceAccelerationStructureCompatibilityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceAccelerationStructureCompatibilityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceAccelerationStructureCompatibilityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceAccelerationStructureCompatibilityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceAccelerationStructureCompatibilityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceAccelerationStructureCompatibilityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceAccelerationStructureCompatibilityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceAccelerationStructureCompatibilityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceAccelerationStructureCompatibilityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceAccelerationStructureCompatibilityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceAccelerationStructureCompatibilityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceAccelerationStructureCompatibilityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceAccelerationStructureCompatibilityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceAccelerationStructureCompatibilityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceAccelerationStructureCompatibilityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceAccelerationStructureCompatibilityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceAccelerationStructureCompatibilityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceAccelerationStructureCompatibilityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceAccelerationStructureCompatibilityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceAccelerationStructureCompatibilityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceAccelerationStructureCompatibilityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceAccelerationStructureCompatibilityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceAccelerationStructureCompatibilityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceAccelerationStructureCompatibilityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceAccelerationStructureCompatibilityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceAccelerationStructureCompatibilityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceAccelerationStructureCompatibilityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceAccelerationStructureCompatibilityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceAccelerationStructureCompatibilityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceAccelerationStructureCompatibilityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceAccelerationStructureCompatibilityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceAccelerationStructureCompatibilityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceAccelerationStructureCompatibilityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceAccelerationStructureCompatibilityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceAccelerationStructureCompatibilityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceAccelerationStructureCompatibilityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceAccelerationStructureCompatibilityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceAccelerationStructureCompatibilityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceAccelerationStructureCompatibilityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceAccelerationStructureCompatibilityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceAccelerationStructureCompatibilityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceAccelerationStructureCompatibilityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceAccelerationStructureCompatibilityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceAccelerationStructureCompatibilityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceAccelerationStructureCompatibilityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceAccelerationStructureCompatibilityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceAccelerationStructureCompatibilityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceAccelerationStructureCompatibilityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceAccelerationStructureCompatibilityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceAccelerationStructureCompatibilityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceAccelerationStructureCompatibilityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceAccelerationStructureCompatibilityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceAccelerationStructureCompatibilityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceAccelerationStructureCompatibilityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceAccelerationStructureCompatibilityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceAccelerationStructureCompatibilityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceAccelerationStructureCompatibilityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceAccelerationStructureCompatibilityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceAccelerationStructureCompatibilityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceAccelerationStructureCompatibilityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceAccelerationStructureCompatibilityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceAccelerationStructureCompatibilityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceAccelerationStructureCompatibilityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceAccelerationStructureCompatibilityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceAccelerationStructureCompatibilityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceAccelerationStructureCompatibilityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceAccelerationStructureCompatibilityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceAccelerationStructureCompatibilityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceAccelerationStructureCompatibilityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceAccelerationStructureCompatibilityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceAccelerationStructureCompatibilityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceAccelerationStructureCompatibilityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceAccelerationStructureCompatibilityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceAccelerationStructureCompatibilityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceAccelerationStructureCompatibilityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceAccelerationStructureCompatibilityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceAccelerationStructureCompatibilityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceAccelerationStructureCompatibilityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceAccelerationStructureCompatibilityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceAccelerationStructureCompatibilityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceAccelerationStructureCompatibilityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceAccelerationStructureCompatibilityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceAccelerationStructureCompatibilityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceAccelerationStructureCompatibilityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceAccelerationStructureCompatibilityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceAccelerationStructureCompatibilityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceAccelerationStructureCompatibilityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceAccelerationStructureCompatibilityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceAccelerationStructureCompatibilityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceAccelerationStructureCompatibilityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceAccelerationStructureCompatibilityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceAccelerationStructureCompatibilityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceAccelerationStructureCompatibilityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceAccelerationStructureCompatibilityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceAccelerationStructureCompatibilityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceAccelerationStructureCompatibilityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceAccelerationStructureCompatibilityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceAccelerationStructureCompatibilityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceAccelerationStructureCompatibilityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceAccelerationStructureCompatibilityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceAccelerationStructureCompatibilityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceAccelerationStructureCompatibilityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceAccelerationStructureCompatibilityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceAccelerationStructureCompatibilityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceAccelerationStructureCompatibilityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceAccelerationStructureCompatibilityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceAccelerationStructureCompatibilityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceAccelerationStructureCompatibilityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceAccelerationStructureCompatibilityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceAccelerationStructureCompatibilityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceAccelerationStructureCompatibilityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceAccelerationStructureCompatibilityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceAccelerationStructureCompatibilityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceAccelerationStructureCompatibilityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceAccelerationStructureCompatibilityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceAccelerationStructureCompatibilityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceAccelerationStructureCompatibilityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceAccelerationStructureCompatibilityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceAccelerationStructureCompatibilityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceAccelerationStructureCompatibilityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceAccelerationStructureCompatibilityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceAccelerationStructureCompatibilityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceAccelerationStructureCompatibilityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceAccelerationStructureCompatibilityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceAccelerationStructureCompatibilityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceAccelerationStructureCompatibilityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceAccelerationStructureCompatibilityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceAccelerationStructureCompatibilityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceAccelerationStructureCompatibilityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceAccelerationStructureCompatibilityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceAccelerationStructureCompatibilityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceAccelerationStructureCompatibilityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceAccelerationStructureCompatibilityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceAccelerationStructureCompatibilityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceAccelerationStructureCompatibilityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceAccelerationStructureCompatibilityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceAccelerationStructureCompatibilityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceAccelerationStructureCompatibilityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceAccelerationStructureCompatibilityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceAccelerationStructureCompatibilityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceAccelerationStructureCompatibilityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceAccelerationStructureCompatibilityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceAccelerationStructureCompatibilityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceAccelerationStructureCompatibilityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceAccelerationStructureCompatibilityKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceAccelerationStructureCompatibilityKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceAccelerationStructureCompatibilityKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceAccelerationStructureCompatibilityKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceAccelerationStructureCompatibilityKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceAccelerationStructureCompatibilityKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceAccelerationStructureCompatibilityKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceAccelerationStructureCompatibilityKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceAccelerationStructureCompatibilityKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceAccelerationStructureCompatibilityKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceAccelerationStructureCompatibilityKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceAccelerationStructureCompatibilityKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceAccelerationStructureCompatibilityKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceAccelerationStructureCompatibilityKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceAccelerationStructureCompatibilityKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceAccelerationStructureCompatibilityKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetRayTracingShaderGroupStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetRayTracingShaderGroupStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetRayTracingShaderGroupStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetRayTracingShaderGroupStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetRayTracingShaderGroupStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetRayTracingShaderGroupStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetRayTracingShaderGroupStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetRayTracingShaderGroupStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetRayTracingShaderGroupStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetRayTracingShaderGroupStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetRayTracingShaderGroupStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetRayTracingShaderGroupStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetRayTracingShaderGroupStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetRayTracingShaderGroupStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetRayTracingShaderGroupStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetRayTracingShaderGroupStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetRayTracingShaderGroupStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetRayTracingShaderGroupStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetRayTracingShaderGroupStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetRayTracingShaderGroupStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetRayTracingShaderGroupStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetRayTracingShaderGroupStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetRayTracingShaderGroupStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetRayTracingShaderGroupStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetRayTracingShaderGroupStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetRayTracingShaderGroupStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetRayTracingShaderGroupStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetRayTracingShaderGroupStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetRayTracingShaderGroupStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetRayTracingShaderGroupStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetRayTracingShaderGroupStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetRayTracingShaderGroupStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetRayTracingShaderGroupStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetRayTracingShaderGroupStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetRayTracingShaderGroupStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetRayTracingShaderGroupStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetRayTracingShaderGroupStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetRayTracingShaderGroupStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetRayTracingShaderGroupStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetRayTracingPipelineStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRayTracingPipelineStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRayTracingPipelineStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRayTracingPipelineStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRayTracingPipelineStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRayTracingPipelineStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRayTracingPipelineStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRayTracingPipelineStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRayTracingPipelineStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRayTracingPipelineStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRayTracingPipelineStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRayTracingPipelineStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRayTracingPipelineStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRayTracingPipelineStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetRayTracingPipelineStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetRayTracingPipelineStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRayTracingPipelineStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRayTracingPipelineStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRayTracingPipelineStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRayTracingPipelineStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRayTracingPipelineStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRayTracingPipelineStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRayTracingPipelineStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRayTracingPipelineStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRayTracingPipelineStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRayTracingPipelineStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRayTracingPipelineStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRayTracingPipelineStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRayTracingPipelineStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRayTracingPipelineStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetRayTracingPipelineStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetRayTracingPipelineStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRayTracingPipelineStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRayTracingPipelineStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRayTracingPipelineStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRayTracingPipelineStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRayTracingPipelineStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRayTracingPipelineStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRayTracingPipelineStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRayTracingPipelineStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRayTracingPipelineStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRayTracingPipelineStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRayTracingPipelineStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRayTracingPipelineStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRayTracingPipelineStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRayTracingPipelineStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetRayTracingPipelineStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetRayTracingPipelineStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRayTracingPipelineStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRayTracingPipelineStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRayTracingPipelineStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRayTracingPipelineStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRayTracingPipelineStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRayTracingPipelineStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRayTracingPipelineStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRayTracingPipelineStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRayTracingPipelineStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRayTracingPipelineStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRayTracingPipelineStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRayTracingPipelineStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRayTracingPipelineStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRayTracingPipelineStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetRayTracingPipelineStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetRayTracingPipelineStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRayTracingPipelineStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRayTracingPipelineStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRayTracingPipelineStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRayTracingPipelineStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRayTracingPipelineStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRayTracingPipelineStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRayTracingPipelineStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRayTracingPipelineStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRayTracingPipelineStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRayTracingPipelineStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRayTracingPipelineStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRayTracingPipelineStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRayTracingPipelineStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRayTracingPipelineStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetRayTracingPipelineStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetRayTracingPipelineStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRayTracingPipelineStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRayTracingPipelineStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRayTracingPipelineStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRayTracingPipelineStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRayTracingPipelineStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRayTracingPipelineStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRayTracingPipelineStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRayTracingPipelineStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRayTracingPipelineStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRayTracingPipelineStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRayTracingPipelineStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRayTracingPipelineStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRayTracingPipelineStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRayTracingPipelineStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetRayTracingPipelineStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetRayTracingPipelineStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRayTracingPipelineStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRayTracingPipelineStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRayTracingPipelineStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRayTracingPipelineStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRayTracingPipelineStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRayTracingPipelineStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRayTracingPipelineStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRayTracingPipelineStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRayTracingPipelineStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRayTracingPipelineStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRayTracingPipelineStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRayTracingPipelineStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRayTracingPipelineStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRayTracingPipelineStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetRayTracingPipelineStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetRayTracingPipelineStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRayTracingPipelineStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRayTracingPipelineStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRayTracingPipelineStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRayTracingPipelineStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRayTracingPipelineStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRayTracingPipelineStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRayTracingPipelineStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRayTracingPipelineStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRayTracingPipelineStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRayTracingPipelineStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRayTracingPipelineStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRayTracingPipelineStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRayTracingPipelineStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRayTracingPipelineStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetRayTracingPipelineStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetRayTracingPipelineStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRayTracingPipelineStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRayTracingPipelineStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRayTracingPipelineStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRayTracingPipelineStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRayTracingPipelineStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRayTracingPipelineStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRayTracingPipelineStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRayTracingPipelineStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRayTracingPipelineStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRayTracingPipelineStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRayTracingPipelineStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRayTracingPipelineStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRayTracingPipelineStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRayTracingPipelineStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetRayTracingPipelineStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetRayTracingPipelineStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRayTracingPipelineStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRayTracingPipelineStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRayTracingPipelineStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRayTracingPipelineStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRayTracingPipelineStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRayTracingPipelineStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRayTracingPipelineStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRayTracingPipelineStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRayTracingPipelineStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRayTracingPipelineStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRayTracingPipelineStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRayTracingPipelineStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRayTracingPipelineStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRayTracingPipelineStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetRayTracingPipelineStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetRayTracingPipelineStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRayTracingPipelineStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRayTracingPipelineStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRayTracingPipelineStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRayTracingPipelineStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRayTracingPipelineStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRayTracingPipelineStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRayTracingPipelineStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRayTracingPipelineStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRayTracingPipelineStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRayTracingPipelineStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRayTracingPipelineStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRayTracingPipelineStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRayTracingPipelineStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRayTracingPipelineStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetRayTracingPipelineStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetRayTracingPipelineStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRayTracingPipelineStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRayTracingPipelineStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRayTracingPipelineStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRayTracingPipelineStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRayTracingPipelineStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRayTracingPipelineStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRayTracingPipelineStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRayTracingPipelineStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRayTracingPipelineStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRayTracingPipelineStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRayTracingPipelineStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRayTracingPipelineStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRayTracingPipelineStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRayTracingPipelineStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetRayTracingPipelineStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetRayTracingPipelineStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRayTracingPipelineStackSizeKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRayTracingPipelineStackSizeKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRayTracingPipelineStackSizeKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRayTracingPipelineStackSizeKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRayTracingPipelineStackSizeKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRayTracingPipelineStackSizeKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRayTracingPipelineStackSizeKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRayTracingPipelineStackSizeKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRayTracingPipelineStackSizeKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRayTracingPipelineStackSizeKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRayTracingPipelineStackSizeKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRayTracingPipelineStackSizeKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRayTracingPipelineStackSizeKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRayTracingPipelineStackSizeKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetRayTracingPipelineStackSizeKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetRayTracingPipelineStackSizeKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetImageViewHandleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetImageViewHandleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetImageViewHandleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetImageViewHandleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetImageViewHandleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetImageViewHandleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetImageViewHandleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetImageViewHandleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetImageViewHandleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetImageViewHandleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetImageViewHandleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetImageViewHandleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetImageViewHandleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetImageViewHandleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetImageViewHandleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetImageViewHandleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetImageViewHandleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetImageViewHandleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetImageViewHandleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetImageViewHandleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetImageViewHandleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetImageViewHandleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetImageViewHandleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetImageViewHandleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetImageViewHandleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetImageViewHandleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandle64NVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandle64NVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandle64NVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandle64NVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandle64NVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandle64NVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandle64NVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandle64NVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandle64NVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandle64NVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandle64NVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandle64NVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandle64NVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewHandle64NVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetImageViewHandle64NVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetImageViewHandle64NVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandle64NVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandle64NVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandle64NVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandle64NVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandle64NVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandle64NVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandle64NVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandle64NVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandle64NVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandle64NVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandle64NVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandle64NVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandle64NVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewHandle64NVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetImageViewHandle64NVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetImageViewHandle64NVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandle64NVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandle64NVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandle64NVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandle64NVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandle64NVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandle64NVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandle64NVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandle64NVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandle64NVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandle64NVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandle64NVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandle64NVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandle64NVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewHandle64NVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetImageViewHandle64NVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetImageViewHandle64NVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandle64NVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandle64NVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandle64NVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandle64NVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandle64NVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandle64NVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandle64NVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandle64NVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandle64NVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandle64NVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandle64NVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandle64NVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandle64NVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewHandle64NVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetImageViewHandle64NVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetImageViewHandle64NVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandle64NVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandle64NVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandle64NVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandle64NVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandle64NVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandle64NVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandle64NVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandle64NVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandle64NVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandle64NVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandle64NVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandle64NVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandle64NVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewHandle64NVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetImageViewHandle64NVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetImageViewHandle64NVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandle64NVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandle64NVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandle64NVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandle64NVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandle64NVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandle64NVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandle64NVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandle64NVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandle64NVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandle64NVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandle64NVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandle64NVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandle64NVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewHandle64NVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetImageViewHandle64NVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetImageViewHandle64NVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandle64NVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandle64NVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandle64NVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandle64NVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandle64NVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandle64NVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandle64NVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandle64NVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandle64NVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandle64NVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandle64NVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandle64NVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandle64NVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewHandle64NVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetImageViewHandle64NVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetImageViewHandle64NVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandle64NVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandle64NVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandle64NVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandle64NVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandle64NVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandle64NVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandle64NVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandle64NVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandle64NVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandle64NVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandle64NVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandle64NVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandle64NVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewHandle64NVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetImageViewHandle64NVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetImageViewHandle64NVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandle64NVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandle64NVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandle64NVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandle64NVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandle64NVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandle64NVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandle64NVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandle64NVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandle64NVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandle64NVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandle64NVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandle64NVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandle64NVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewHandle64NVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetImageViewHandle64NVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetImageViewHandle64NVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandle64NVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandle64NVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandle64NVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandle64NVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandle64NVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandle64NVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandle64NVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandle64NVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandle64NVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandle64NVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandle64NVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandle64NVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandle64NVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewHandle64NVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetImageViewHandle64NVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetImageViewHandle64NVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandle64NVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandle64NVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandle64NVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandle64NVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandle64NVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandle64NVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandle64NVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandle64NVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandle64NVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandle64NVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandle64NVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandle64NVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandle64NVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewHandle64NVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetImageViewHandle64NVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetImageViewHandle64NVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandle64NVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandle64NVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandle64NVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandle64NVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandle64NVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandle64NVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandle64NVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandle64NVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandle64NVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandle64NVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandle64NVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandle64NVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandle64NVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewHandle64NVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetImageViewHandle64NVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetImageViewHandle64NVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandle64NVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandle64NVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandle64NVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandle64NVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandle64NVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandle64NVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandle64NVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandle64NVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandle64NVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandle64NVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandle64NVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandle64NVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandle64NVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewHandle64NVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetImageViewHandle64NVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetImageViewHandle64NVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetImageViewAddressNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewAddressNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewAddressNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewAddressNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewAddressNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewAddressNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewAddressNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewAddressNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewAddressNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewAddressNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewAddressNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewAddressNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewAddressNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewAddressNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetImageViewAddressNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetImageViewAddressNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewAddressNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewAddressNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewAddressNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewAddressNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewAddressNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewAddressNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewAddressNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewAddressNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewAddressNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewAddressNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewAddressNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewAddressNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewAddressNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewAddressNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetImageViewAddressNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetImageViewAddressNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewAddressNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewAddressNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewAddressNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewAddressNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewAddressNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewAddressNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewAddressNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewAddressNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewAddressNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewAddressNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewAddressNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewAddressNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewAddressNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewAddressNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetImageViewAddressNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetImageViewAddressNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewAddressNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewAddressNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewAddressNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewAddressNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewAddressNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewAddressNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewAddressNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewAddressNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewAddressNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewAddressNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewAddressNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewAddressNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewAddressNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewAddressNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetImageViewAddressNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetImageViewAddressNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewAddressNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewAddressNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewAddressNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewAddressNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewAddressNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewAddressNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewAddressNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewAddressNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewAddressNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewAddressNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewAddressNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewAddressNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewAddressNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewAddressNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetImageViewAddressNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetImageViewAddressNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewAddressNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewAddressNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewAddressNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewAddressNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewAddressNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewAddressNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewAddressNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewAddressNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewAddressNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewAddressNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewAddressNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewAddressNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewAddressNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewAddressNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetImageViewAddressNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetImageViewAddressNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewAddressNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewAddressNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewAddressNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewAddressNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewAddressNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewAddressNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewAddressNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewAddressNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewAddressNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewAddressNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewAddressNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewAddressNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewAddressNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewAddressNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetImageViewAddressNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetImageViewAddressNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewAddressNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewAddressNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewAddressNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewAddressNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewAddressNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewAddressNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewAddressNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewAddressNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewAddressNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewAddressNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewAddressNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewAddressNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewAddressNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewAddressNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetImageViewAddressNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetImageViewAddressNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewAddressNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewAddressNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewAddressNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewAddressNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewAddressNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewAddressNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewAddressNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewAddressNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewAddressNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewAddressNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewAddressNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewAddressNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewAddressNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewAddressNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetImageViewAddressNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetImageViewAddressNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewAddressNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewAddressNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewAddressNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewAddressNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewAddressNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewAddressNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewAddressNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewAddressNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewAddressNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewAddressNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewAddressNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewAddressNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewAddressNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewAddressNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetImageViewAddressNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetImageViewAddressNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewAddressNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewAddressNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewAddressNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewAddressNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewAddressNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewAddressNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewAddressNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewAddressNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewAddressNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewAddressNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewAddressNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewAddressNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewAddressNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewAddressNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetImageViewAddressNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetImageViewAddressNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewAddressNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewAddressNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewAddressNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewAddressNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewAddressNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewAddressNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewAddressNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewAddressNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewAddressNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewAddressNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewAddressNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewAddressNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewAddressNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewAddressNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetImageViewAddressNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetImageViewAddressNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewAddressNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewAddressNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewAddressNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewAddressNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewAddressNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewAddressNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewAddressNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewAddressNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewAddressNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewAddressNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewAddressNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewAddressNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewAddressNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewAddressNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetImageViewAddressNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetImageViewAddressNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceCombinedImageSamplerIndexNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceCombinedImageSamplerIndexNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceCombinedImageSamplerIndexNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceCombinedImageSamplerIndexNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceCombinedImageSamplerIndexNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceCombinedImageSamplerIndexNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceCombinedImageSamplerIndexNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceCombinedImageSamplerIndexNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceCombinedImageSamplerIndexNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceCombinedImageSamplerIndexNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceCombinedImageSamplerIndexNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceCombinedImageSamplerIndexNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceCombinedImageSamplerIndexNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceCombinedImageSamplerIndexNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceCombinedImageSamplerIndexNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceCombinedImageSamplerIndexNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceCombinedImageSamplerIndexNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceCombinedImageSamplerIndexNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceCombinedImageSamplerIndexNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceCombinedImageSamplerIndexNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceCombinedImageSamplerIndexNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceCombinedImageSamplerIndexNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceCombinedImageSamplerIndexNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceCombinedImageSamplerIndexNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceCombinedImageSamplerIndexNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceCombinedImageSamplerIndexNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceCombinedImageSamplerIndexNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceCombinedImageSamplerIndexNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceCombinedImageSamplerIndexNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceCombinedImageSamplerIndexNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceCombinedImageSamplerIndexNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceCombinedImageSamplerIndexNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceCombinedImageSamplerIndexNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceCombinedImageSamplerIndexNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceCombinedImageSamplerIndexNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceCombinedImageSamplerIndexNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceCombinedImageSamplerIndexNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceCombinedImageSamplerIndexNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceCombinedImageSamplerIndexNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceCombinedImageSamplerIndexNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceCombinedImageSamplerIndexNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceCombinedImageSamplerIndexNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceCombinedImageSamplerIndexNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceCombinedImageSamplerIndexNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceCombinedImageSamplerIndexNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceCombinedImageSamplerIndexNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceCombinedImageSamplerIndexNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceCombinedImageSamplerIndexNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceCombinedImageSamplerIndexNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceCombinedImageSamplerIndexNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceCombinedImageSamplerIndexNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceCombinedImageSamplerIndexNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceCombinedImageSamplerIndexNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceCombinedImageSamplerIndexNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceCombinedImageSamplerIndexNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceCombinedImageSamplerIndexNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceCombinedImageSamplerIndexNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceCombinedImageSamplerIndexNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceCombinedImageSamplerIndexNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceCombinedImageSamplerIndexNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceCombinedImageSamplerIndexNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceCombinedImageSamplerIndexNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceCombinedImageSamplerIndexNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceCombinedImageSamplerIndexNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceCombinedImageSamplerIndexNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceCombinedImageSamplerIndexNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceCombinedImageSamplerIndexNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceCombinedImageSamplerIndexNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceCombinedImageSamplerIndexNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceCombinedImageSamplerIndexNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceCombinedImageSamplerIndexNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceCombinedImageSamplerIndexNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceCombinedImageSamplerIndexNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceCombinedImageSamplerIndexNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceCombinedImageSamplerIndexNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceCombinedImageSamplerIndexNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceCombinedImageSamplerIndexNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceCombinedImageSamplerIndexNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceCombinedImageSamplerIndexNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceCombinedImageSamplerIndexNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceCombinedImageSamplerIndexNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceCombinedImageSamplerIndexNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceCombinedImageSamplerIndexNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceCombinedImageSamplerIndexNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceCombinedImageSamplerIndexNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceCombinedImageSamplerIndexNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceCombinedImageSamplerIndexNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceCombinedImageSamplerIndexNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceCombinedImageSamplerIndexNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceCombinedImageSamplerIndexNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceCombinedImageSamplerIndexNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceCombinedImageSamplerIndexNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceCombinedImageSamplerIndexNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceCombinedImageSamplerIndexNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceCombinedImageSamplerIndexNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceCombinedImageSamplerIndexNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceCombinedImageSamplerIndexNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceCombinedImageSamplerIndexNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceCombinedImageSamplerIndexNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceCombinedImageSamplerIndexNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceCombinedImageSamplerIndexNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceCombinedImageSamplerIndexNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceCombinedImageSamplerIndexNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceCombinedImageSamplerIndexNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceCombinedImageSamplerIndexNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceCombinedImageSamplerIndexNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceCombinedImageSamplerIndexNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceCombinedImageSamplerIndexNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceCombinedImageSamplerIndexNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceCombinedImageSamplerIndexNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceCombinedImageSamplerIndexNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceCombinedImageSamplerIndexNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceCombinedImageSamplerIndexNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceCombinedImageSamplerIndexNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceCombinedImageSamplerIndexNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceCombinedImageSamplerIndexNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceCombinedImageSamplerIndexNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceCombinedImageSamplerIndexNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceCombinedImageSamplerIndexNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceCombinedImageSamplerIndexNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceCombinedImageSamplerIndexNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceCombinedImageSamplerIndexNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceCombinedImageSamplerIndexNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceCombinedImageSamplerIndexNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceCombinedImageSamplerIndexNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceCombinedImageSamplerIndexNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceCombinedImageSamplerIndexNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceCombinedImageSamplerIndexNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceCombinedImageSamplerIndexNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceCombinedImageSamplerIndexNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceCombinedImageSamplerIndexNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceCombinedImageSamplerIndexNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceCombinedImageSamplerIndexNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceCombinedImageSamplerIndexNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceCombinedImageSamplerIndexNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceCombinedImageSamplerIndexNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceCombinedImageSamplerIndexNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceCombinedImageSamplerIndexNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceCombinedImageSamplerIndexNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceCombinedImageSamplerIndexNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceCombinedImageSamplerIndexNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceCombinedImageSamplerIndexNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceCombinedImageSamplerIndexNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceCombinedImageSamplerIndexNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceCombinedImageSamplerIndexNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceCombinedImageSamplerIndexNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceCombinedImageSamplerIndexNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceCombinedImageSamplerIndexNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceCombinedImageSamplerIndexNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceCombinedImageSamplerIndexNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceCombinedImageSamplerIndexNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceCombinedImageSamplerIndexNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceCombinedImageSamplerIndexNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceCombinedImageSamplerIndexNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceCombinedImageSamplerIndexNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceCombinedImageSamplerIndexNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceCombinedImageSamplerIndexNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceCombinedImageSamplerIndexNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceCombinedImageSamplerIndexNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceCombinedImageSamplerIndexNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceCombinedImageSamplerIndexNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceCombinedImageSamplerIndexNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceCombinedImageSamplerIndexNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceCombinedImageSamplerIndexNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceCombinedImageSamplerIndexNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceCombinedImageSamplerIndexNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceCombinedImageSamplerIndexNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceCombinedImageSamplerIndexNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceCombinedImageSamplerIndexNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceCombinedImageSamplerIndexNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceCombinedImageSamplerIndexNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceCombinedImageSamplerIndexNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceCombinedImageSamplerIndexNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceCombinedImageSamplerIndexNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceCombinedImageSamplerIndexNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceCombinedImageSamplerIndexNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceCombinedImageSamplerIndexNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceCombinedImageSamplerIndexNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceCombinedImageSamplerIndexNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceCombinedImageSamplerIndexNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceCombinedImageSamplerIndexNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceCombinedImageSamplerIndexNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceCombinedImageSamplerIndexNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceCombinedImageSamplerIndexNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceCombinedImageSamplerIndexNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceCombinedImageSamplerIndexNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceCombinedImageSamplerIndexNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceCombinedImageSamplerIndexNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceCombinedImageSamplerIndexNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceCombinedImageSamplerIndexNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceCombinedImageSamplerIndexNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceCombinedImageSamplerIndexNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceCombinedImageSamplerIndexNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceCombinedImageSamplerIndexNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceCombinedImageSamplerIndexNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceCombinedImageSamplerIndexNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceCombinedImageSamplerIndexNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceCombinedImageSamplerIndexNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceCombinedImageSamplerIndexNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceCombinedImageSamplerIndexNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceCombinedImageSamplerIndexNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceCombinedImageSamplerIndexNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceCombinedImageSamplerIndexNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceCombinedImageSamplerIndexNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceCombinedImageSamplerIndexNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceCombinedImageSamplerIndexNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceCombinedImageSamplerIndexNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceCombinedImageSamplerIndexNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceGroupSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceGroupSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceGroupSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceGroupSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceGroupSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceGroupSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceGroupSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceGroupSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceGroupSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceGroupSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceGroupSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceGroupSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceGroupSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceGroupSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceGroupSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceGroupSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceGroupSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceGroupSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceGroupSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceGroupSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceGroupSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceGroupSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceGroupSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceGroupSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceGroupSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceGroupSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceGroupSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceGroupSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceGroupSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceGroupSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceGroupSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceGroupSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceGroupSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceGroupSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceGroupSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceGroupSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModes2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModes2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModes2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModes2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModes2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModes2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModes2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModes2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModes2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModes2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModes2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModes2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModes2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceGroupSurfacePresentModes2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceGroupSurfacePresentModes2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceGroupSurfacePresentModes2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_AcquireFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_AcquireFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_AcquireFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_AcquireFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_AcquireFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_AcquireFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_AcquireFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_AcquireFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_AcquireFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_AcquireFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_AcquireFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_AcquireFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_AcquireFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_AcquireFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_AcquireFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_AcquireFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_AcquireFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_AcquireFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_AcquireFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_AcquireFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_AcquireFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_AcquireFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_AcquireFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_AcquireFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_AcquireFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_AcquireFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_AcquireFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_AcquireFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_AcquireFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_AcquireFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_AcquireFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_AcquireFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_AcquireFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_AcquireFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_AcquireFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ReleaseFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ReleaseFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ReleaseFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ReleaseFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ReleaseFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ReleaseFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ReleaseFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ReleaseFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ReleaseFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ReleaseFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ReleaseFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ReleaseFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ReleaseFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ReleaseFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ReleaseFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ReleaseFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ReleaseFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ReleaseFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ReleaseFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ReleaseFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ReleaseFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ReleaseFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ReleaseFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ReleaseFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ReleaseFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseFullScreenExclusiveModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseFullScreenExclusiveModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseFullScreenExclusiveModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseFullScreenExclusiveModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseFullScreenExclusiveModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseFullScreenExclusiveModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseFullScreenExclusiveModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseFullScreenExclusiveModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseFullScreenExclusiveModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseFullScreenExclusiveModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseFullScreenExclusiveModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseFullScreenExclusiveModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseFullScreenExclusiveModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseFullScreenExclusiveModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ReleaseFullScreenExclusiveModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ReleaseFullScreenExclusiveModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_AcquireProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquireProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_AcquireProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_AcquireProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_AcquireProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquireProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_AcquireProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_AcquireProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_AcquireProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquireProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_AcquireProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_AcquireProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_AcquireProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquireProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_AcquireProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_AcquireProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_AcquireProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquireProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_AcquireProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_AcquireProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_AcquireProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquireProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_AcquireProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_AcquireProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_AcquireProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquireProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_AcquireProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_AcquireProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_AcquireProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquireProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_AcquireProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_AcquireProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquireProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_AcquireProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_AcquireProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquireProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_AcquireProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_AcquireProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquireProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_AcquireProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_AcquireProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_AcquireProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquireProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_AcquireProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_AcquireProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquireProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_AcquireProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_AcquireProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ReleaseProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ReleaseProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ReleaseProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ReleaseProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ReleaseProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ReleaseProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ReleaseProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ReleaseProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ReleaseProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ReleaseProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ReleaseProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ReleaseProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ReleaseProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ReleaseProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ReleaseProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ReleaseProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ReleaseProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ReleaseProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ReleaseProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ReleaseProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ReleaseProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ReleaseProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ReleaseProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ReleaseProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ReleaseProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseProfilingLockKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseProfilingLockKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseProfilingLockKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseProfilingLockKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseProfilingLockKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseProfilingLockKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseProfilingLockKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseProfilingLockKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseProfilingLockKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseProfilingLockKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseProfilingLockKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseProfilingLockKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseProfilingLockKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseProfilingLockKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ReleaseProfilingLockKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ReleaseProfilingLockKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetImageDrmFormatModifierPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageDrmFormatModifierPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageDrmFormatModifierPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageDrmFormatModifierPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageDrmFormatModifierPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageDrmFormatModifierPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageDrmFormatModifierPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageDrmFormatModifierPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageDrmFormatModifierPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageDrmFormatModifierPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageDrmFormatModifierPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageDrmFormatModifierPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageDrmFormatModifierPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageDrmFormatModifierPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetImageDrmFormatModifierPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetImageDrmFormatModifierPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetImageDrmFormatModifierPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageDrmFormatModifierPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageDrmFormatModifierPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageDrmFormatModifierPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageDrmFormatModifierPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageDrmFormatModifierPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageDrmFormatModifierPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageDrmFormatModifierPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageDrmFormatModifierPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageDrmFormatModifierPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageDrmFormatModifierPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageDrmFormatModifierPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageDrmFormatModifierPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageDrmFormatModifierPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetImageDrmFormatModifierPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetImageDrmFormatModifierPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetImageDrmFormatModifierPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageDrmFormatModifierPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageDrmFormatModifierPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageDrmFormatModifierPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageDrmFormatModifierPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageDrmFormatModifierPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageDrmFormatModifierPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageDrmFormatModifierPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageDrmFormatModifierPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageDrmFormatModifierPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageDrmFormatModifierPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageDrmFormatModifierPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageDrmFormatModifierPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageDrmFormatModifierPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetImageDrmFormatModifierPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetImageDrmFormatModifierPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetImageDrmFormatModifierPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageDrmFormatModifierPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageDrmFormatModifierPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageDrmFormatModifierPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageDrmFormatModifierPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageDrmFormatModifierPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageDrmFormatModifierPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageDrmFormatModifierPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageDrmFormatModifierPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageDrmFormatModifierPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageDrmFormatModifierPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageDrmFormatModifierPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageDrmFormatModifierPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageDrmFormatModifierPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetImageDrmFormatModifierPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetImageDrmFormatModifierPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetImageDrmFormatModifierPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageDrmFormatModifierPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageDrmFormatModifierPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageDrmFormatModifierPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageDrmFormatModifierPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageDrmFormatModifierPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageDrmFormatModifierPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageDrmFormatModifierPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageDrmFormatModifierPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageDrmFormatModifierPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageDrmFormatModifierPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageDrmFormatModifierPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageDrmFormatModifierPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageDrmFormatModifierPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetImageDrmFormatModifierPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetImageDrmFormatModifierPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetImageDrmFormatModifierPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageDrmFormatModifierPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageDrmFormatModifierPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageDrmFormatModifierPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageDrmFormatModifierPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageDrmFormatModifierPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageDrmFormatModifierPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageDrmFormatModifierPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageDrmFormatModifierPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageDrmFormatModifierPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageDrmFormatModifierPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageDrmFormatModifierPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageDrmFormatModifierPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageDrmFormatModifierPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetImageDrmFormatModifierPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetImageDrmFormatModifierPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetImageDrmFormatModifierPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageDrmFormatModifierPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageDrmFormatModifierPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageDrmFormatModifierPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageDrmFormatModifierPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageDrmFormatModifierPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageDrmFormatModifierPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageDrmFormatModifierPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageDrmFormatModifierPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageDrmFormatModifierPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageDrmFormatModifierPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageDrmFormatModifierPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageDrmFormatModifierPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageDrmFormatModifierPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetImageDrmFormatModifierPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetImageDrmFormatModifierPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetImageDrmFormatModifierPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageDrmFormatModifierPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageDrmFormatModifierPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageDrmFormatModifierPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageDrmFormatModifierPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageDrmFormatModifierPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageDrmFormatModifierPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageDrmFormatModifierPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageDrmFormatModifierPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageDrmFormatModifierPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageDrmFormatModifierPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageDrmFormatModifierPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageDrmFormatModifierPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageDrmFormatModifierPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetImageDrmFormatModifierPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetImageDrmFormatModifierPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageDrmFormatModifierPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageDrmFormatModifierPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageDrmFormatModifierPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageDrmFormatModifierPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageDrmFormatModifierPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageDrmFormatModifierPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageDrmFormatModifierPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageDrmFormatModifierPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageDrmFormatModifierPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageDrmFormatModifierPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageDrmFormatModifierPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageDrmFormatModifierPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageDrmFormatModifierPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageDrmFormatModifierPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetImageDrmFormatModifierPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetImageDrmFormatModifierPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageDrmFormatModifierPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageDrmFormatModifierPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageDrmFormatModifierPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageDrmFormatModifierPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageDrmFormatModifierPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageDrmFormatModifierPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageDrmFormatModifierPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageDrmFormatModifierPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageDrmFormatModifierPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageDrmFormatModifierPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageDrmFormatModifierPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageDrmFormatModifierPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageDrmFormatModifierPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageDrmFormatModifierPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetImageDrmFormatModifierPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetImageDrmFormatModifierPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageDrmFormatModifierPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageDrmFormatModifierPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageDrmFormatModifierPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageDrmFormatModifierPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageDrmFormatModifierPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageDrmFormatModifierPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageDrmFormatModifierPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageDrmFormatModifierPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageDrmFormatModifierPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageDrmFormatModifierPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageDrmFormatModifierPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageDrmFormatModifierPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageDrmFormatModifierPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageDrmFormatModifierPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetImageDrmFormatModifierPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetImageDrmFormatModifierPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetImageDrmFormatModifierPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageDrmFormatModifierPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageDrmFormatModifierPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageDrmFormatModifierPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageDrmFormatModifierPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageDrmFormatModifierPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageDrmFormatModifierPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageDrmFormatModifierPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageDrmFormatModifierPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageDrmFormatModifierPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageDrmFormatModifierPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageDrmFormatModifierPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageDrmFormatModifierPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageDrmFormatModifierPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetImageDrmFormatModifierPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetImageDrmFormatModifierPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageDrmFormatModifierPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageDrmFormatModifierPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageDrmFormatModifierPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageDrmFormatModifierPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageDrmFormatModifierPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageDrmFormatModifierPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageDrmFormatModifierPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageDrmFormatModifierPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageDrmFormatModifierPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageDrmFormatModifierPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageDrmFormatModifierPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageDrmFormatModifierPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageDrmFormatModifierPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageDrmFormatModifierPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetImageDrmFormatModifierPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetImageDrmFormatModifierPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetBufferOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetBufferOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetBufferOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetBufferOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetBufferOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetBufferOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetBufferOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetBufferOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetBufferOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetBufferOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetBufferOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetBufferOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetBufferOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetBufferOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetBufferOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetBufferOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetBufferOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetBufferOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetBufferOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetBufferOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetBufferOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetBufferOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetBufferOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetBufferOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetBufferOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetBufferOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetBufferOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetBufferOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetBufferOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetBufferOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetBufferOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetBufferOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetBufferOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetBufferOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetBufferOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetBufferOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetBufferOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetBufferOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetBufferOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetBufferOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetBufferOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetBufferOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetBufferOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetBufferOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetBufferOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetBufferOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetBufferOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetBufferOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetBufferOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetBufferOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetBufferOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetBufferOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetBufferDeviceAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetBufferDeviceAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetBufferDeviceAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetBufferDeviceAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetBufferDeviceAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetBufferDeviceAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetBufferDeviceAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetBufferDeviceAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetBufferDeviceAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetBufferDeviceAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetBufferDeviceAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetBufferDeviceAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetBufferDeviceAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetBufferDeviceAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetBufferDeviceAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetBufferDeviceAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetBufferDeviceAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetBufferDeviceAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetBufferDeviceAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetBufferDeviceAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetBufferDeviceAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetBufferDeviceAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetBufferDeviceAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetBufferDeviceAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetBufferDeviceAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetBufferDeviceAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetBufferDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetBufferDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetBufferDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetBufferDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetBufferDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetBufferDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetBufferDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetBufferDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetBufferDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetBufferDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetBufferDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetBufferDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetBufferDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetBufferDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetBufferDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetBufferDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetBufferDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetBufferDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetBufferDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetBufferDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetBufferDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetBufferDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetBufferDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetBufferDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetBufferDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetBufferDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferDeviceAddressEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetBufferDeviceAddressEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetBufferDeviceAddressEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferDeviceAddressEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetBufferDeviceAddressEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetBufferDeviceAddressEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferDeviceAddressEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetBufferDeviceAddressEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetBufferDeviceAddressEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferDeviceAddressEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetBufferDeviceAddressEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetBufferDeviceAddressEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferDeviceAddressEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetBufferDeviceAddressEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetBufferDeviceAddressEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferDeviceAddressEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetBufferDeviceAddressEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetBufferDeviceAddressEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferDeviceAddressEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetBufferDeviceAddressEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetBufferDeviceAddressEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferDeviceAddressEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetBufferDeviceAddressEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetBufferDeviceAddressEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferDeviceAddressEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetBufferDeviceAddressEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetBufferDeviceAddressEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferDeviceAddressEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetBufferDeviceAddressEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetBufferDeviceAddressEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferDeviceAddressEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetBufferDeviceAddressEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetBufferDeviceAddressEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferDeviceAddressEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetBufferDeviceAddressEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetBufferDeviceAddressEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferDeviceAddressEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetBufferDeviceAddressEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetBufferDeviceAddressEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_InitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_InitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_InitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_InitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_InitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_InitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_InitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_InitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_InitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_InitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_InitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_InitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_InitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_InitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_InitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_InitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_InitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_InitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_InitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_InitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_InitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_InitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_InitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_InitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_InitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_InitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_InitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_InitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_InitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_InitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_InitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_InitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_InitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_InitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_InitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_InitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_InitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_InitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_InitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_InitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_InitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_InitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_InitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_InitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_InitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_InitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_InitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_InitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_InitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_InitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_InitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_InitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_UninitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UninitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UninitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UninitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UninitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UninitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UninitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UninitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UninitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UninitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UninitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UninitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UninitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UninitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_UninitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_UninitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_UninitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UninitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UninitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UninitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UninitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UninitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UninitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UninitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UninitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UninitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UninitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UninitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UninitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UninitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_UninitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_UninitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_UninitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UninitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UninitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UninitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UninitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UninitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UninitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UninitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UninitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UninitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UninitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UninitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UninitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UninitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_UninitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_UninitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_UninitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UninitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UninitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UninitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UninitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UninitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UninitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UninitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UninitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UninitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UninitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UninitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UninitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UninitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_UninitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_UninitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_UninitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UninitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UninitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UninitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UninitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UninitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UninitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UninitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UninitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UninitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UninitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UninitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UninitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UninitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_UninitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_UninitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_UninitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UninitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UninitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UninitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UninitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UninitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UninitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UninitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UninitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UninitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UninitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UninitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UninitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UninitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_UninitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_UninitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_UninitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UninitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UninitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UninitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UninitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UninitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UninitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UninitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UninitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UninitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UninitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UninitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UninitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UninitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_UninitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_UninitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_UninitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UninitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UninitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UninitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UninitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UninitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UninitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UninitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UninitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UninitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UninitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UninitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UninitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UninitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_UninitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_UninitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_UninitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UninitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UninitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UninitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UninitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UninitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UninitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UninitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UninitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UninitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UninitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UninitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UninitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UninitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_UninitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_UninitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_UninitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UninitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UninitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UninitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UninitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UninitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UninitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UninitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UninitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UninitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UninitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UninitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UninitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UninitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_UninitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_UninitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_UninitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UninitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UninitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UninitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UninitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UninitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UninitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UninitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UninitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UninitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UninitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UninitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UninitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UninitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_UninitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_UninitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_UninitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UninitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UninitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UninitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UninitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UninitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UninitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UninitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UninitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UninitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UninitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UninitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UninitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UninitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_UninitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_UninitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_UninitializePerformanceApiINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UninitializePerformanceApiINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UninitializePerformanceApiINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UninitializePerformanceApiINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UninitializePerformanceApiINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UninitializePerformanceApiINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UninitializePerformanceApiINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UninitializePerformanceApiINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UninitializePerformanceApiINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UninitializePerformanceApiINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UninitializePerformanceApiINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UninitializePerformanceApiINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UninitializePerformanceApiINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UninitializePerformanceApiINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_UninitializePerformanceApiINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_UninitializePerformanceApiINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetPerformanceMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetPerformanceMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetPerformanceMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetPerformanceMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetPerformanceMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetPerformanceMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetPerformanceMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetPerformanceMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetPerformanceMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetPerformanceMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetPerformanceMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetPerformanceMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetPerformanceMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetPerformanceMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetPerformanceMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetPerformanceMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetPerformanceMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetPerformanceMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetPerformanceMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetPerformanceMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetPerformanceMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetPerformanceMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetPerformanceMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetPerformanceMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetPerformanceMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetPerformanceMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceStreamMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceStreamMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceStreamMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceStreamMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceStreamMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceStreamMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceStreamMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceStreamMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceStreamMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceStreamMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceStreamMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceStreamMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceStreamMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceStreamMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetPerformanceStreamMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetPerformanceStreamMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceStreamMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceStreamMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceStreamMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceStreamMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceStreamMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceStreamMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceStreamMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceStreamMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceStreamMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceStreamMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceStreamMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceStreamMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceStreamMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceStreamMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetPerformanceStreamMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetPerformanceStreamMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceStreamMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceStreamMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceStreamMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceStreamMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceStreamMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceStreamMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceStreamMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceStreamMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceStreamMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceStreamMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceStreamMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceStreamMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceStreamMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceStreamMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetPerformanceStreamMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetPerformanceStreamMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceStreamMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceStreamMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceStreamMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceStreamMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceStreamMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceStreamMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceStreamMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceStreamMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceStreamMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceStreamMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceStreamMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceStreamMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceStreamMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceStreamMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetPerformanceStreamMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetPerformanceStreamMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceStreamMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceStreamMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceStreamMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceStreamMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceStreamMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceStreamMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceStreamMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceStreamMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceStreamMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceStreamMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceStreamMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceStreamMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceStreamMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceStreamMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetPerformanceStreamMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetPerformanceStreamMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceStreamMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceStreamMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceStreamMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceStreamMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceStreamMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceStreamMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceStreamMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceStreamMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceStreamMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceStreamMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceStreamMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceStreamMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceStreamMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceStreamMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetPerformanceStreamMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetPerformanceStreamMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceStreamMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceStreamMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceStreamMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceStreamMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceStreamMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceStreamMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceStreamMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceStreamMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceStreamMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceStreamMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceStreamMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceStreamMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceStreamMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceStreamMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetPerformanceStreamMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetPerformanceStreamMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceStreamMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceStreamMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceStreamMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceStreamMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceStreamMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceStreamMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceStreamMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceStreamMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceStreamMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceStreamMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceStreamMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceStreamMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceStreamMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceStreamMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetPerformanceStreamMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetPerformanceStreamMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceStreamMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceStreamMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceStreamMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceStreamMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceStreamMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceStreamMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceStreamMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceStreamMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceStreamMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceStreamMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceStreamMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceStreamMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceStreamMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceStreamMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetPerformanceStreamMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetPerformanceStreamMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceStreamMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceStreamMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceStreamMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceStreamMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceStreamMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceStreamMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceStreamMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceStreamMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceStreamMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceStreamMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceStreamMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceStreamMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceStreamMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceStreamMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetPerformanceStreamMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetPerformanceStreamMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceStreamMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceStreamMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceStreamMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceStreamMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceStreamMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceStreamMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceStreamMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceStreamMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceStreamMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceStreamMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceStreamMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceStreamMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceStreamMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceStreamMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetPerformanceStreamMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetPerformanceStreamMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceStreamMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceStreamMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceStreamMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceStreamMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceStreamMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceStreamMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceStreamMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceStreamMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceStreamMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceStreamMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceStreamMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceStreamMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceStreamMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceStreamMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetPerformanceStreamMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetPerformanceStreamMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceStreamMarkerINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceStreamMarkerINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceStreamMarkerINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceStreamMarkerINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceStreamMarkerINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceStreamMarkerINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceStreamMarkerINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceStreamMarkerINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceStreamMarkerINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceStreamMarkerINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceStreamMarkerINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceStreamMarkerINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceStreamMarkerINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceStreamMarkerINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetPerformanceStreamMarkerINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetPerformanceStreamMarkerINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceOverrideINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceOverrideINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceOverrideINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceOverrideINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceOverrideINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceOverrideINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceOverrideINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceOverrideINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceOverrideINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceOverrideINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceOverrideINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceOverrideINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceOverrideINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPerformanceOverrideINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetPerformanceOverrideINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetPerformanceOverrideINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceOverrideINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceOverrideINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceOverrideINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceOverrideINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceOverrideINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceOverrideINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceOverrideINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceOverrideINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceOverrideINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceOverrideINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceOverrideINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceOverrideINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceOverrideINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPerformanceOverrideINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetPerformanceOverrideINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetPerformanceOverrideINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceOverrideINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceOverrideINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceOverrideINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceOverrideINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceOverrideINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceOverrideINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceOverrideINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceOverrideINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceOverrideINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceOverrideINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceOverrideINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceOverrideINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceOverrideINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPerformanceOverrideINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetPerformanceOverrideINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetPerformanceOverrideINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceOverrideINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceOverrideINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceOverrideINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceOverrideINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceOverrideINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceOverrideINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceOverrideINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceOverrideINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceOverrideINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceOverrideINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceOverrideINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceOverrideINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceOverrideINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPerformanceOverrideINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetPerformanceOverrideINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetPerformanceOverrideINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceOverrideINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceOverrideINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceOverrideINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceOverrideINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceOverrideINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceOverrideINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceOverrideINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceOverrideINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceOverrideINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceOverrideINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceOverrideINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceOverrideINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceOverrideINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPerformanceOverrideINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetPerformanceOverrideINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetPerformanceOverrideINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceOverrideINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceOverrideINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceOverrideINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceOverrideINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceOverrideINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceOverrideINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceOverrideINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceOverrideINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceOverrideINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceOverrideINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceOverrideINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceOverrideINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceOverrideINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPerformanceOverrideINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetPerformanceOverrideINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetPerformanceOverrideINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceOverrideINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceOverrideINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceOverrideINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceOverrideINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceOverrideINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceOverrideINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceOverrideINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceOverrideINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceOverrideINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceOverrideINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceOverrideINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceOverrideINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceOverrideINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPerformanceOverrideINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetPerformanceOverrideINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetPerformanceOverrideINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceOverrideINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceOverrideINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceOverrideINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceOverrideINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceOverrideINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceOverrideINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceOverrideINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceOverrideINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceOverrideINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceOverrideINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceOverrideINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceOverrideINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceOverrideINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPerformanceOverrideINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetPerformanceOverrideINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetPerformanceOverrideINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceOverrideINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceOverrideINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceOverrideINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceOverrideINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceOverrideINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceOverrideINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceOverrideINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceOverrideINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceOverrideINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceOverrideINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceOverrideINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceOverrideINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceOverrideINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPerformanceOverrideINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetPerformanceOverrideINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetPerformanceOverrideINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceOverrideINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceOverrideINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceOverrideINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceOverrideINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceOverrideINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceOverrideINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceOverrideINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceOverrideINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceOverrideINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceOverrideINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceOverrideINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceOverrideINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceOverrideINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPerformanceOverrideINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetPerformanceOverrideINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetPerformanceOverrideINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceOverrideINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceOverrideINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceOverrideINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceOverrideINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceOverrideINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceOverrideINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceOverrideINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceOverrideINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceOverrideINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceOverrideINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceOverrideINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceOverrideINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceOverrideINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPerformanceOverrideINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetPerformanceOverrideINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetPerformanceOverrideINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceOverrideINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceOverrideINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceOverrideINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceOverrideINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceOverrideINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceOverrideINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceOverrideINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceOverrideINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceOverrideINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceOverrideINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceOverrideINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceOverrideINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceOverrideINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPerformanceOverrideINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetPerformanceOverrideINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetPerformanceOverrideINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceOverrideINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceOverrideINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceOverrideINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceOverrideINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceOverrideINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceOverrideINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceOverrideINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceOverrideINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceOverrideINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceOverrideINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceOverrideINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceOverrideINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceOverrideINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPerformanceOverrideINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetPerformanceOverrideINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetPerformanceOverrideINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_AcquirePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquirePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquirePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquirePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquirePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquirePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquirePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquirePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquirePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquirePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquirePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquirePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquirePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AcquirePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_AcquirePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_AcquirePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_AcquirePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquirePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquirePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquirePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquirePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquirePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquirePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquirePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquirePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquirePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquirePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquirePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquirePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AcquirePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_AcquirePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_AcquirePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_AcquirePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquirePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquirePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquirePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquirePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquirePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquirePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquirePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquirePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquirePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquirePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquirePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquirePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AcquirePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_AcquirePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_AcquirePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_AcquirePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquirePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquirePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquirePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquirePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquirePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquirePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquirePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquirePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquirePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquirePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquirePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquirePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AcquirePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_AcquirePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_AcquirePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_AcquirePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquirePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquirePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquirePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquirePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquirePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquirePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquirePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquirePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquirePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquirePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquirePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquirePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AcquirePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_AcquirePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_AcquirePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_AcquirePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquirePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquirePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquirePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquirePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquirePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquirePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquirePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquirePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquirePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquirePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquirePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquirePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AcquirePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_AcquirePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_AcquirePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_AcquirePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquirePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquirePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquirePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquirePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquirePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquirePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquirePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquirePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquirePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquirePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquirePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquirePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AcquirePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_AcquirePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_AcquirePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_AcquirePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquirePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquirePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquirePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquirePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquirePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquirePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquirePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquirePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquirePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquirePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquirePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquirePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AcquirePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_AcquirePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_AcquirePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquirePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquirePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquirePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquirePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquirePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquirePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquirePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquirePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquirePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquirePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquirePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquirePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquirePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AcquirePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_AcquirePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_AcquirePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquirePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquirePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquirePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquirePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquirePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquirePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquirePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquirePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquirePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquirePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquirePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquirePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquirePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AcquirePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_AcquirePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_AcquirePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquirePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquirePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquirePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquirePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquirePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquirePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquirePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquirePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquirePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquirePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquirePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquirePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquirePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AcquirePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_AcquirePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_AcquirePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_AcquirePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquirePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquirePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquirePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquirePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquirePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquirePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquirePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquirePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquirePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquirePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquirePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquirePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AcquirePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_AcquirePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_AcquirePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquirePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquirePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquirePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquirePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquirePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquirePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquirePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquirePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquirePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquirePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquirePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquirePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquirePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AcquirePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_AcquirePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_AcquirePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ReleasePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleasePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleasePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleasePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleasePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleasePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleasePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleasePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleasePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleasePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleasePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleasePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleasePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleasePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ReleasePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ReleasePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ReleasePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleasePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleasePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleasePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleasePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleasePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleasePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleasePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleasePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleasePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleasePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleasePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleasePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleasePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ReleasePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ReleasePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ReleasePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleasePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleasePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleasePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleasePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleasePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleasePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleasePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleasePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleasePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleasePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleasePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleasePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleasePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ReleasePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ReleasePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ReleasePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleasePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleasePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleasePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleasePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleasePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleasePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleasePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleasePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleasePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleasePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleasePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleasePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleasePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ReleasePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ReleasePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ReleasePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleasePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleasePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleasePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleasePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleasePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleasePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleasePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleasePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleasePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleasePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleasePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleasePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleasePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ReleasePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ReleasePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ReleasePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleasePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleasePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleasePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleasePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleasePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleasePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleasePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleasePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleasePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleasePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleasePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleasePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleasePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ReleasePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ReleasePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ReleasePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleasePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleasePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleasePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleasePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleasePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleasePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleasePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleasePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleasePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleasePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleasePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleasePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleasePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ReleasePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ReleasePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ReleasePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleasePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleasePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleasePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleasePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleasePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleasePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleasePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleasePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleasePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleasePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleasePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleasePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleasePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ReleasePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ReleasePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleasePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleasePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleasePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleasePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleasePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleasePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleasePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleasePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleasePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleasePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleasePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleasePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleasePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleasePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ReleasePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ReleasePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleasePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleasePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleasePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleasePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleasePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleasePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleasePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleasePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleasePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleasePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleasePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleasePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleasePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleasePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ReleasePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ReleasePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleasePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleasePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleasePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleasePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleasePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleasePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleasePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleasePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleasePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleasePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleasePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleasePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleasePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleasePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ReleasePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ReleasePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ReleasePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleasePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleasePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleasePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleasePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleasePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleasePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleasePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleasePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleasePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleasePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleasePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleasePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleasePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ReleasePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ReleasePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleasePerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleasePerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleasePerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleasePerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleasePerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleasePerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleasePerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleasePerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleasePerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleasePerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleasePerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleasePerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleasePerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleasePerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ReleasePerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ReleasePerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_QueueSetPerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_QueueSetPerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_QueueSetPerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_QueueSetPerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_QueueSetPerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_QueueSetPerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_QueueSetPerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_QueueSetPerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_QueueSetPerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_QueueSetPerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_QueueSetPerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_QueueSetPerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_QueueSetPerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_QueueSetPerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_QueueSetPerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_QueueSetPerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_QueueSetPerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_QueueSetPerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_QueueSetPerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_QueueSetPerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_QueueSetPerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_QueueSetPerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_QueueSetPerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_QueueSetPerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerformanceConfigurationINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerformanceConfigurationINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerformanceConfigurationINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerformanceConfigurationINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerformanceConfigurationINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerformanceConfigurationINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerformanceConfigurationINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerformanceConfigurationINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerformanceConfigurationINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerformanceConfigurationINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerformanceConfigurationINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerformanceConfigurationINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerformanceConfigurationINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerformanceConfigurationINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_QueueSetPerformanceConfigurationINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_QueueSetPerformanceConfigurationINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPerformanceParameterINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPerformanceParameterINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPerformanceParameterINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPerformanceParameterINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPerformanceParameterINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPerformanceParameterINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPerformanceParameterINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPerformanceParameterINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPerformanceParameterINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPerformanceParameterINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPerformanceParameterINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPerformanceParameterINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPerformanceParameterINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPerformanceParameterINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPerformanceParameterINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPerformanceParameterINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetPerformanceParameterINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPerformanceParameterINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPerformanceParameterINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPerformanceParameterINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPerformanceParameterINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPerformanceParameterINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPerformanceParameterINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPerformanceParameterINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPerformanceParameterINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPerformanceParameterINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPerformanceParameterINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPerformanceParameterINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPerformanceParameterINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPerformanceParameterINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetPerformanceParameterINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetPerformanceParameterINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetPerformanceParameterINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPerformanceParameterINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPerformanceParameterINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPerformanceParameterINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPerformanceParameterINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPerformanceParameterINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPerformanceParameterINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPerformanceParameterINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPerformanceParameterINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPerformanceParameterINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPerformanceParameterINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPerformanceParameterINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPerformanceParameterINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPerformanceParameterINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetPerformanceParameterINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetPerformanceParameterINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetPerformanceParameterINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPerformanceParameterINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPerformanceParameterINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPerformanceParameterINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPerformanceParameterINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPerformanceParameterINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPerformanceParameterINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPerformanceParameterINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPerformanceParameterINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPerformanceParameterINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPerformanceParameterINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPerformanceParameterINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPerformanceParameterINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPerformanceParameterINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetPerformanceParameterINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetPerformanceParameterINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetPerformanceParameterINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPerformanceParameterINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPerformanceParameterINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPerformanceParameterINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPerformanceParameterINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPerformanceParameterINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPerformanceParameterINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPerformanceParameterINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPerformanceParameterINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPerformanceParameterINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPerformanceParameterINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPerformanceParameterINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPerformanceParameterINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPerformanceParameterINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetPerformanceParameterINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetPerformanceParameterINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetPerformanceParameterINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPerformanceParameterINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPerformanceParameterINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPerformanceParameterINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPerformanceParameterINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPerformanceParameterINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPerformanceParameterINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPerformanceParameterINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPerformanceParameterINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPerformanceParameterINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPerformanceParameterINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPerformanceParameterINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPerformanceParameterINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPerformanceParameterINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetPerformanceParameterINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetPerformanceParameterINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetPerformanceParameterINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPerformanceParameterINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPerformanceParameterINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPerformanceParameterINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPerformanceParameterINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPerformanceParameterINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPerformanceParameterINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPerformanceParameterINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPerformanceParameterINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPerformanceParameterINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPerformanceParameterINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPerformanceParameterINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPerformanceParameterINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPerformanceParameterINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetPerformanceParameterINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetPerformanceParameterINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetPerformanceParameterINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPerformanceParameterINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPerformanceParameterINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPerformanceParameterINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPerformanceParameterINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPerformanceParameterINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPerformanceParameterINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPerformanceParameterINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPerformanceParameterINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPerformanceParameterINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPerformanceParameterINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPerformanceParameterINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPerformanceParameterINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPerformanceParameterINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetPerformanceParameterINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetPerformanceParameterINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPerformanceParameterINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPerformanceParameterINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPerformanceParameterINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPerformanceParameterINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPerformanceParameterINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPerformanceParameterINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPerformanceParameterINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPerformanceParameterINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPerformanceParameterINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPerformanceParameterINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPerformanceParameterINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPerformanceParameterINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPerformanceParameterINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPerformanceParameterINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetPerformanceParameterINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetPerformanceParameterINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPerformanceParameterINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPerformanceParameterINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPerformanceParameterINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPerformanceParameterINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPerformanceParameterINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPerformanceParameterINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPerformanceParameterINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPerformanceParameterINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPerformanceParameterINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPerformanceParameterINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPerformanceParameterINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPerformanceParameterINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPerformanceParameterINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPerformanceParameterINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetPerformanceParameterINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetPerformanceParameterINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPerformanceParameterINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPerformanceParameterINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPerformanceParameterINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPerformanceParameterINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPerformanceParameterINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPerformanceParameterINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPerformanceParameterINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPerformanceParameterINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPerformanceParameterINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPerformanceParameterINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPerformanceParameterINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPerformanceParameterINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPerformanceParameterINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPerformanceParameterINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetPerformanceParameterINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetPerformanceParameterINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetPerformanceParameterINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPerformanceParameterINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPerformanceParameterINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPerformanceParameterINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPerformanceParameterINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPerformanceParameterINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPerformanceParameterINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPerformanceParameterINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPerformanceParameterINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPerformanceParameterINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPerformanceParameterINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPerformanceParameterINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPerformanceParameterINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPerformanceParameterINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetPerformanceParameterINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetPerformanceParameterINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPerformanceParameterINTEL@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPerformanceParameterINTEL@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPerformanceParameterINTEL@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPerformanceParameterINTEL@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPerformanceParameterINTEL@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPerformanceParameterINTEL@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPerformanceParameterINTEL@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPerformanceParameterINTEL@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPerformanceParameterINTEL@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPerformanceParameterINTEL@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPerformanceParameterINTEL@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPerformanceParameterINTEL@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPerformanceParameterINTEL@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPerformanceParameterINTEL@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetPerformanceParameterINTEL=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetPerformanceParameterINTEL=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceMemoryOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceMemoryOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceMemoryOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceMemoryOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceMemoryOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceMemoryOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceMemoryOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceMemoryOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceMemoryOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceMemoryOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceMemoryOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceMemoryOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceMemoryOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceMemoryOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceMemoryOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceMemoryOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceMemoryOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceMemoryOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceMemoryOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceMemoryOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceMemoryOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceMemoryOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceMemoryOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceMemoryOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddress@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddress@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddress@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddress@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddress@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddress@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddress@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddress@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddress@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddress@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddress@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddress@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddress@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddress@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceMemoryOpaqueCaptureAddress=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceMemoryOpaqueCaptureAddress=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMemoryOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceMemoryOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceMemoryOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMemoryOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceMemoryOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceMemoryOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMemoryOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceMemoryOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceMemoryOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMemoryOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceMemoryOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceMemoryOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMemoryOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceMemoryOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceMemoryOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMemoryOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceMemoryOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceMemoryOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMemoryOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceMemoryOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceMemoryOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMemoryOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceMemoryOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceMemoryOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMemoryOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceMemoryOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceMemoryOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMemoryOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceMemoryOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceMemoryOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMemoryOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceMemoryOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceMemoryOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMemoryOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceMemoryOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceMemoryOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMemoryOpaqueCaptureAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceMemoryOpaqueCaptureAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceMemoryOpaqueCaptureAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutablePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutablePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutablePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutablePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutablePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutablePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutablePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutablePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutablePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutablePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutablePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutablePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutablePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutablePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPipelineExecutablePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPipelineExecutablePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutablePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutablePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutablePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutablePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutablePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutablePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutablePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutablePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutablePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutablePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutablePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutablePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutablePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutablePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetPipelineExecutablePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetPipelineExecutablePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutablePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutablePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutablePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutablePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutablePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutablePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutablePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutablePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutablePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutablePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutablePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutablePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutablePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutablePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetPipelineExecutablePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetPipelineExecutablePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutablePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutablePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutablePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutablePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutablePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutablePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutablePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutablePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutablePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutablePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutablePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutablePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutablePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutablePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetPipelineExecutablePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetPipelineExecutablePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutablePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutablePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutablePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutablePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutablePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutablePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutablePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutablePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutablePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutablePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutablePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutablePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutablePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutablePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetPipelineExecutablePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetPipelineExecutablePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutablePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutablePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutablePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutablePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutablePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutablePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutablePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutablePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutablePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutablePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutablePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutablePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutablePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutablePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetPipelineExecutablePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetPipelineExecutablePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutablePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutablePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutablePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutablePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutablePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutablePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutablePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutablePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutablePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutablePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutablePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutablePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutablePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutablePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetPipelineExecutablePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetPipelineExecutablePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutablePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutablePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutablePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutablePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutablePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutablePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutablePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutablePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutablePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutablePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutablePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutablePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutablePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutablePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetPipelineExecutablePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetPipelineExecutablePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutablePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutablePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutablePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutablePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutablePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutablePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutablePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutablePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutablePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutablePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutablePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutablePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutablePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutablePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetPipelineExecutablePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetPipelineExecutablePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutablePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutablePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutablePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutablePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutablePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutablePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutablePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutablePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutablePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutablePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutablePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutablePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutablePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutablePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetPipelineExecutablePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetPipelineExecutablePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutablePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutablePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutablePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutablePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutablePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutablePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutablePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutablePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutablePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutablePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutablePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutablePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutablePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutablePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetPipelineExecutablePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetPipelineExecutablePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutablePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutablePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutablePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutablePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutablePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutablePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutablePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutablePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutablePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutablePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutablePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutablePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutablePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutablePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetPipelineExecutablePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetPipelineExecutablePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutablePropertiesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutablePropertiesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutablePropertiesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutablePropertiesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutablePropertiesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutablePropertiesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutablePropertiesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutablePropertiesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutablePropertiesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutablePropertiesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutablePropertiesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutablePropertiesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutablePropertiesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutablePropertiesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetPipelineExecutablePropertiesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetPipelineExecutablePropertiesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableStatisticsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableStatisticsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableStatisticsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableStatisticsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableStatisticsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableStatisticsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableStatisticsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableStatisticsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableStatisticsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableStatisticsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableStatisticsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableStatisticsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableStatisticsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableStatisticsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPipelineExecutableStatisticsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPipelineExecutableStatisticsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableStatisticsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableStatisticsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableStatisticsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableStatisticsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableStatisticsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableStatisticsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableStatisticsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableStatisticsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableStatisticsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableStatisticsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableStatisticsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableStatisticsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableStatisticsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableStatisticsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetPipelineExecutableStatisticsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetPipelineExecutableStatisticsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableStatisticsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableStatisticsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableStatisticsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableStatisticsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableStatisticsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableStatisticsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableStatisticsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableStatisticsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableStatisticsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableStatisticsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableStatisticsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableStatisticsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableStatisticsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableStatisticsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetPipelineExecutableStatisticsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetPipelineExecutableStatisticsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableStatisticsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableStatisticsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableStatisticsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableStatisticsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableStatisticsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableStatisticsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableStatisticsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableStatisticsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableStatisticsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableStatisticsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableStatisticsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableStatisticsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableStatisticsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableStatisticsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetPipelineExecutableStatisticsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetPipelineExecutableStatisticsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableStatisticsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableStatisticsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableStatisticsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableStatisticsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableStatisticsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableStatisticsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableStatisticsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableStatisticsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableStatisticsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableStatisticsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableStatisticsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableStatisticsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableStatisticsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableStatisticsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetPipelineExecutableStatisticsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetPipelineExecutableStatisticsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableStatisticsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableStatisticsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableStatisticsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableStatisticsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableStatisticsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableStatisticsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableStatisticsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableStatisticsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableStatisticsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableStatisticsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableStatisticsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableStatisticsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableStatisticsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableStatisticsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetPipelineExecutableStatisticsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetPipelineExecutableStatisticsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableStatisticsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableStatisticsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableStatisticsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableStatisticsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableStatisticsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableStatisticsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableStatisticsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableStatisticsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableStatisticsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableStatisticsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableStatisticsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableStatisticsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableStatisticsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableStatisticsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetPipelineExecutableStatisticsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetPipelineExecutableStatisticsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableStatisticsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableStatisticsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableStatisticsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableStatisticsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableStatisticsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableStatisticsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableStatisticsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableStatisticsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableStatisticsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableStatisticsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableStatisticsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableStatisticsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableStatisticsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableStatisticsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetPipelineExecutableStatisticsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetPipelineExecutableStatisticsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableStatisticsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableStatisticsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableStatisticsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableStatisticsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableStatisticsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableStatisticsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableStatisticsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableStatisticsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableStatisticsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableStatisticsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableStatisticsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableStatisticsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableStatisticsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableStatisticsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetPipelineExecutableStatisticsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetPipelineExecutableStatisticsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableStatisticsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableStatisticsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableStatisticsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableStatisticsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableStatisticsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableStatisticsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableStatisticsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableStatisticsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableStatisticsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableStatisticsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableStatisticsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableStatisticsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableStatisticsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableStatisticsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetPipelineExecutableStatisticsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetPipelineExecutableStatisticsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableStatisticsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableStatisticsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableStatisticsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableStatisticsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableStatisticsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableStatisticsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableStatisticsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableStatisticsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableStatisticsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableStatisticsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableStatisticsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableStatisticsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableStatisticsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableStatisticsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetPipelineExecutableStatisticsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetPipelineExecutableStatisticsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableStatisticsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableStatisticsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableStatisticsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableStatisticsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableStatisticsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableStatisticsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableStatisticsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableStatisticsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableStatisticsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableStatisticsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableStatisticsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableStatisticsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableStatisticsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableStatisticsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetPipelineExecutableStatisticsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetPipelineExecutableStatisticsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableStatisticsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableStatisticsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableStatisticsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableStatisticsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableStatisticsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableStatisticsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableStatisticsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableStatisticsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableStatisticsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableStatisticsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableStatisticsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableStatisticsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableStatisticsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableStatisticsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetPipelineExecutableStatisticsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetPipelineExecutableStatisticsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableInternalRepresentationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableInternalRepresentationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableInternalRepresentationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableInternalRepresentationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableInternalRepresentationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableInternalRepresentationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableInternalRepresentationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableInternalRepresentationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableInternalRepresentationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableInternalRepresentationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableInternalRepresentationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableInternalRepresentationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableInternalRepresentationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineExecutableInternalRepresentationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPipelineExecutableInternalRepresentationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPipelineExecutableInternalRepresentationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableInternalRepresentationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableInternalRepresentationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableInternalRepresentationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableInternalRepresentationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableInternalRepresentationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableInternalRepresentationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableInternalRepresentationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableInternalRepresentationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableInternalRepresentationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableInternalRepresentationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableInternalRepresentationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableInternalRepresentationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableInternalRepresentationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineExecutableInternalRepresentationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetPipelineExecutableInternalRepresentationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetPipelineExecutableInternalRepresentationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableInternalRepresentationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableInternalRepresentationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableInternalRepresentationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableInternalRepresentationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableInternalRepresentationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableInternalRepresentationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableInternalRepresentationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableInternalRepresentationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableInternalRepresentationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableInternalRepresentationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableInternalRepresentationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableInternalRepresentationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableInternalRepresentationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineExecutableInternalRepresentationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetPipelineExecutableInternalRepresentationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetPipelineExecutableInternalRepresentationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableInternalRepresentationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableInternalRepresentationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableInternalRepresentationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableInternalRepresentationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableInternalRepresentationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableInternalRepresentationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableInternalRepresentationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableInternalRepresentationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableInternalRepresentationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableInternalRepresentationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableInternalRepresentationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableInternalRepresentationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableInternalRepresentationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineExecutableInternalRepresentationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetPipelineExecutableInternalRepresentationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetPipelineExecutableInternalRepresentationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableInternalRepresentationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableInternalRepresentationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableInternalRepresentationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableInternalRepresentationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableInternalRepresentationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableInternalRepresentationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableInternalRepresentationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableInternalRepresentationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableInternalRepresentationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableInternalRepresentationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableInternalRepresentationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableInternalRepresentationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableInternalRepresentationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineExecutableInternalRepresentationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetPipelineExecutableInternalRepresentationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetPipelineExecutableInternalRepresentationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableInternalRepresentationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableInternalRepresentationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableInternalRepresentationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableInternalRepresentationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableInternalRepresentationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableInternalRepresentationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableInternalRepresentationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableInternalRepresentationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableInternalRepresentationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableInternalRepresentationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableInternalRepresentationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableInternalRepresentationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableInternalRepresentationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineExecutableInternalRepresentationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetPipelineExecutableInternalRepresentationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetPipelineExecutableInternalRepresentationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableInternalRepresentationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableInternalRepresentationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableInternalRepresentationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableInternalRepresentationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableInternalRepresentationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableInternalRepresentationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableInternalRepresentationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableInternalRepresentationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableInternalRepresentationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableInternalRepresentationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableInternalRepresentationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableInternalRepresentationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableInternalRepresentationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineExecutableInternalRepresentationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetPipelineExecutableInternalRepresentationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetPipelineExecutableInternalRepresentationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableInternalRepresentationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableInternalRepresentationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableInternalRepresentationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableInternalRepresentationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableInternalRepresentationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableInternalRepresentationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableInternalRepresentationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableInternalRepresentationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableInternalRepresentationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableInternalRepresentationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableInternalRepresentationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableInternalRepresentationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableInternalRepresentationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineExecutableInternalRepresentationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetPipelineExecutableInternalRepresentationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetPipelineExecutableInternalRepresentationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableInternalRepresentationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableInternalRepresentationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableInternalRepresentationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableInternalRepresentationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableInternalRepresentationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableInternalRepresentationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableInternalRepresentationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableInternalRepresentationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableInternalRepresentationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableInternalRepresentationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableInternalRepresentationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableInternalRepresentationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableInternalRepresentationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineExecutableInternalRepresentationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetPipelineExecutableInternalRepresentationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetPipelineExecutableInternalRepresentationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableInternalRepresentationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableInternalRepresentationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableInternalRepresentationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableInternalRepresentationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableInternalRepresentationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableInternalRepresentationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableInternalRepresentationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableInternalRepresentationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableInternalRepresentationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableInternalRepresentationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableInternalRepresentationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableInternalRepresentationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableInternalRepresentationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineExecutableInternalRepresentationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetPipelineExecutableInternalRepresentationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetPipelineExecutableInternalRepresentationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableInternalRepresentationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableInternalRepresentationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableInternalRepresentationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableInternalRepresentationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableInternalRepresentationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableInternalRepresentationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableInternalRepresentationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableInternalRepresentationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableInternalRepresentationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableInternalRepresentationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableInternalRepresentationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableInternalRepresentationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableInternalRepresentationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineExecutableInternalRepresentationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetPipelineExecutableInternalRepresentationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetPipelineExecutableInternalRepresentationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableInternalRepresentationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableInternalRepresentationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableInternalRepresentationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableInternalRepresentationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableInternalRepresentationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableInternalRepresentationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableInternalRepresentationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableInternalRepresentationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableInternalRepresentationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableInternalRepresentationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableInternalRepresentationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableInternalRepresentationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableInternalRepresentationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineExecutableInternalRepresentationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetPipelineExecutableInternalRepresentationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetPipelineExecutableInternalRepresentationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableInternalRepresentationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableInternalRepresentationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableInternalRepresentationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableInternalRepresentationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableInternalRepresentationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableInternalRepresentationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableInternalRepresentationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableInternalRepresentationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableInternalRepresentationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableInternalRepresentationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableInternalRepresentationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableInternalRepresentationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableInternalRepresentationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineExecutableInternalRepresentationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetPipelineExecutableInternalRepresentationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetPipelineExecutableInternalRepresentationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStipple@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStipple@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStipple@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStipple@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStipple@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStipple@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStipple@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStipple@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStipple@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStipple@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStipple@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStipple@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStipple@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStipple@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetLineStipple=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetLineStipple=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStipple@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStipple@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStipple@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStipple@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStipple@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStipple@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStipple@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStipple@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStipple@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStipple@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStipple@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStipple@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStipple@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStipple@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetLineStipple=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetLineStipple=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStipple@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStipple@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStipple@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStipple@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStipple@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStipple@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStipple@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStipple@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStipple@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStipple@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStipple@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStipple@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStipple@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStipple@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetLineStipple=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetLineStipple=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStipple@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStipple@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStipple@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStipple@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStipple@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStipple@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStipple@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStipple@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStipple@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStipple@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStipple@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStipple@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStipple@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStipple@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetLineStipple=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetLineStipple=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStipple@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStipple@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStipple@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStipple@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStipple@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStipple@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStipple@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStipple@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStipple@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStipple@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStipple@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStipple@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStipple@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStipple@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetLineStipple=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetLineStipple=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStipple@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStipple@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStipple@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStipple@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStipple@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStipple@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStipple@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStipple@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStipple@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStipple@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStipple@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStipple@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStipple@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStipple@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetLineStipple=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetLineStipple=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStipple@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStipple@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStipple@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStipple@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStipple@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStipple@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStipple@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStipple@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStipple@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStipple@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStipple@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStipple@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStipple@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStipple@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetLineStipple=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetLineStipple=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStipple@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStipple@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStipple@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStipple@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStipple@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStipple@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStipple@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStipple@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStipple@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStipple@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStipple@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStipple@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStipple@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStipple@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetLineStipple=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetLineStipple=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStipple@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStipple@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStipple@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStipple@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStipple@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStipple@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStipple@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStipple@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStipple@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStipple@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStipple@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStipple@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStipple@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStipple@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetLineStipple=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetLineStipple=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStipple@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStipple@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStipple@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStipple@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStipple@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStipple@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStipple@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStipple@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStipple@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStipple@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStipple@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStipple@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStipple@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStipple@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetLineStipple=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetLineStipple=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStipple@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStipple@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStipple@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStipple@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStipple@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStipple@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStipple@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStipple@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStipple@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStipple@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStipple@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStipple@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStipple@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStipple@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetLineStipple=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetLineStipple=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStipple@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStipple@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStipple@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStipple@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStipple@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStipple@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStipple@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStipple@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStipple@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStipple@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStipple@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStipple@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStipple@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStipple@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetLineStipple=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetLineStipple=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStipple@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStipple@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStipple@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStipple@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStipple@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStipple@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStipple@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStipple@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStipple@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStipple@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStipple@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStipple@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStipple@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStipple@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetLineStipple=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetLineStipple=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetLineStippleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetLineStippleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetLineStippleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetLineStippleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetLineStippleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetLineStippleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetLineStippleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetLineStippleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetLineStippleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetLineStippleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetLineStippleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetLineStippleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetLineStippleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetLineStippleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetLineStippleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetLineStippleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetLineStippleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetLineStippleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetLineStippleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetLineStippleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetLineStippleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetLineStippleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetLineStippleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetLineStippleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetLineStippleKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetLineStippleKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetLineStippleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetLineStippleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetLineStippleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetLineStippleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetLineStippleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetLineStippleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetLineStippleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetLineStippleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetLineStippleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetLineStippleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetLineStippleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetLineStippleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetLineStippleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetLineStippleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetLineStippleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetLineStippleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetLineStippleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetLineStippleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetLineStippleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetLineStippleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetLineStippleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetLineStippleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetLineStippleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetLineStippleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetLineStippleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetLineStippleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructureKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateAccelerationStructureKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateAccelerationStructureKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildAccelerationStructuresIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBuildAccelerationStructuresIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBuildAccelerationStructuresIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildAccelerationStructuresIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBuildAccelerationStructuresIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBuildAccelerationStructuresIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildAccelerationStructuresIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBuildAccelerationStructuresIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBuildAccelerationStructuresIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildAccelerationStructuresIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBuildAccelerationStructuresIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBuildAccelerationStructuresIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildAccelerationStructuresIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBuildAccelerationStructuresIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBuildAccelerationStructuresIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildAccelerationStructuresIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBuildAccelerationStructuresIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBuildAccelerationStructuresIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildAccelerationStructuresIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBuildAccelerationStructuresIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBuildAccelerationStructuresIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildAccelerationStructuresIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBuildAccelerationStructuresIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBuildAccelerationStructuresIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildAccelerationStructuresIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBuildAccelerationStructuresIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBuildAccelerationStructuresIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildAccelerationStructuresIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBuildAccelerationStructuresIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBuildAccelerationStructuresIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildAccelerationStructuresIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBuildAccelerationStructuresIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBuildAccelerationStructuresIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildAccelerationStructuresIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBuildAccelerationStructuresIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBuildAccelerationStructuresIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresIndirectKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresIndirectKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresIndirectKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresIndirectKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresIndirectKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresIndirectKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresIndirectKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresIndirectKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresIndirectKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresIndirectKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresIndirectKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresIndirectKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresIndirectKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildAccelerationStructuresIndirectKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBuildAccelerationStructuresIndirectKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBuildAccelerationStructuresIndirectKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_BuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_BuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_BuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_BuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_BuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_BuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_BuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_BuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_BuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_BuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_BuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_BuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_BuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_BuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_BuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_BuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_BuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_BuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_BuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_BuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_BuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_BuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_BuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_BuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_BuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_BuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_BuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_BuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_BuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_BuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_BuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_BuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_BuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildAccelerationStructuresKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildAccelerationStructuresKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildAccelerationStructuresKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildAccelerationStructuresKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildAccelerationStructuresKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildAccelerationStructuresKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildAccelerationStructuresKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildAccelerationStructuresKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildAccelerationStructuresKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildAccelerationStructuresKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildAccelerationStructuresKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildAccelerationStructuresKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildAccelerationStructuresKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildAccelerationStructuresKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_BuildAccelerationStructuresKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_BuildAccelerationStructuresKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetAccelerationStructureDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetAccelerationStructureDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetAccelerationStructureDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetAccelerationStructureDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetAccelerationStructureDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetAccelerationStructureDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetAccelerationStructureDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetAccelerationStructureDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetAccelerationStructureDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetAccelerationStructureDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetAccelerationStructureDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetAccelerationStructureDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetAccelerationStructureDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetAccelerationStructureDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetAccelerationStructureDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetAccelerationStructureDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetAccelerationStructureDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetAccelerationStructureDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetAccelerationStructureDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetAccelerationStructureDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetAccelerationStructureDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetAccelerationStructureDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetAccelerationStructureDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetAccelerationStructureDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureDeviceAddressKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureDeviceAddressKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureDeviceAddressKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureDeviceAddressKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureDeviceAddressKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureDeviceAddressKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureDeviceAddressKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureDeviceAddressKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureDeviceAddressKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureDeviceAddressKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureDeviceAddressKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureDeviceAddressKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureDeviceAddressKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureDeviceAddressKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetAccelerationStructureDeviceAddressKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetAccelerationStructureDeviceAddressKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDeferredOperationKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDeferredOperationKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDeferredOperationKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDeferredOperationKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDeferredOperationKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDeferredOperationKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDeferredOperationKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDeferredOperationKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDeferredOperationKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDeferredOperationKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDeferredOperationKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDeferredOperationKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDeferredOperationKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDeferredOperationKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyDeferredOperationKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyDeferredOperationKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationMaxConcurrencyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationMaxConcurrencyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationMaxConcurrencyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationMaxConcurrencyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationMaxConcurrencyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationMaxConcurrencyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationMaxConcurrencyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationMaxConcurrencyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationMaxConcurrencyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationMaxConcurrencyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationMaxConcurrencyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationMaxConcurrencyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationMaxConcurrencyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationMaxConcurrencyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeferredOperationMaxConcurrencyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeferredOperationMaxConcurrencyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationMaxConcurrencyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationMaxConcurrencyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationMaxConcurrencyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationMaxConcurrencyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationMaxConcurrencyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationMaxConcurrencyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationMaxConcurrencyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationMaxConcurrencyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationMaxConcurrencyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationMaxConcurrencyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationMaxConcurrencyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationMaxConcurrencyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationMaxConcurrencyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationMaxConcurrencyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeferredOperationMaxConcurrencyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeferredOperationMaxConcurrencyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationMaxConcurrencyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationMaxConcurrencyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationMaxConcurrencyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationMaxConcurrencyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationMaxConcurrencyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationMaxConcurrencyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationMaxConcurrencyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationMaxConcurrencyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationMaxConcurrencyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationMaxConcurrencyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationMaxConcurrencyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationMaxConcurrencyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationMaxConcurrencyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationMaxConcurrencyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeferredOperationMaxConcurrencyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeferredOperationMaxConcurrencyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationMaxConcurrencyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationMaxConcurrencyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationMaxConcurrencyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationMaxConcurrencyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationMaxConcurrencyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationMaxConcurrencyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationMaxConcurrencyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationMaxConcurrencyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationMaxConcurrencyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationMaxConcurrencyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationMaxConcurrencyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationMaxConcurrencyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationMaxConcurrencyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationMaxConcurrencyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeferredOperationMaxConcurrencyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeferredOperationMaxConcurrencyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationMaxConcurrencyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationMaxConcurrencyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationMaxConcurrencyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationMaxConcurrencyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationMaxConcurrencyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationMaxConcurrencyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationMaxConcurrencyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationMaxConcurrencyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationMaxConcurrencyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationMaxConcurrencyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationMaxConcurrencyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationMaxConcurrencyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationMaxConcurrencyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationMaxConcurrencyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeferredOperationMaxConcurrencyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeferredOperationMaxConcurrencyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationMaxConcurrencyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationMaxConcurrencyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationMaxConcurrencyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationMaxConcurrencyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationMaxConcurrencyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationMaxConcurrencyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationMaxConcurrencyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationMaxConcurrencyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationMaxConcurrencyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationMaxConcurrencyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationMaxConcurrencyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationMaxConcurrencyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationMaxConcurrencyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationMaxConcurrencyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeferredOperationMaxConcurrencyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeferredOperationMaxConcurrencyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationMaxConcurrencyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationMaxConcurrencyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationMaxConcurrencyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationMaxConcurrencyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationMaxConcurrencyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationMaxConcurrencyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationMaxConcurrencyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationMaxConcurrencyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationMaxConcurrencyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationMaxConcurrencyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationMaxConcurrencyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationMaxConcurrencyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationMaxConcurrencyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationMaxConcurrencyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeferredOperationMaxConcurrencyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeferredOperationMaxConcurrencyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationMaxConcurrencyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationMaxConcurrencyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationMaxConcurrencyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationMaxConcurrencyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationMaxConcurrencyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationMaxConcurrencyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationMaxConcurrencyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationMaxConcurrencyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationMaxConcurrencyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationMaxConcurrencyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationMaxConcurrencyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationMaxConcurrencyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationMaxConcurrencyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationMaxConcurrencyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeferredOperationMaxConcurrencyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeferredOperationMaxConcurrencyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationMaxConcurrencyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationMaxConcurrencyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationMaxConcurrencyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationMaxConcurrencyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationMaxConcurrencyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationMaxConcurrencyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationMaxConcurrencyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationMaxConcurrencyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationMaxConcurrencyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationMaxConcurrencyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationMaxConcurrencyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationMaxConcurrencyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationMaxConcurrencyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationMaxConcurrencyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeferredOperationMaxConcurrencyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeferredOperationMaxConcurrencyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationMaxConcurrencyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationMaxConcurrencyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationMaxConcurrencyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationMaxConcurrencyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationMaxConcurrencyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationMaxConcurrencyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationMaxConcurrencyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationMaxConcurrencyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationMaxConcurrencyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationMaxConcurrencyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationMaxConcurrencyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationMaxConcurrencyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationMaxConcurrencyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationMaxConcurrencyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeferredOperationMaxConcurrencyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeferredOperationMaxConcurrencyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationMaxConcurrencyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationMaxConcurrencyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationMaxConcurrencyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationMaxConcurrencyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationMaxConcurrencyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationMaxConcurrencyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationMaxConcurrencyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationMaxConcurrencyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationMaxConcurrencyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationMaxConcurrencyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationMaxConcurrencyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationMaxConcurrencyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationMaxConcurrencyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationMaxConcurrencyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeferredOperationMaxConcurrencyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeferredOperationMaxConcurrencyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationMaxConcurrencyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationMaxConcurrencyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationMaxConcurrencyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationMaxConcurrencyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationMaxConcurrencyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationMaxConcurrencyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationMaxConcurrencyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationMaxConcurrencyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationMaxConcurrencyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationMaxConcurrencyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationMaxConcurrencyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationMaxConcurrencyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationMaxConcurrencyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationMaxConcurrencyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeferredOperationMaxConcurrencyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeferredOperationMaxConcurrencyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationMaxConcurrencyKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationMaxConcurrencyKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationMaxConcurrencyKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationMaxConcurrencyKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationMaxConcurrencyKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationMaxConcurrencyKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationMaxConcurrencyKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationMaxConcurrencyKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationMaxConcurrencyKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationMaxConcurrencyKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationMaxConcurrencyKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationMaxConcurrencyKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationMaxConcurrencyKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationMaxConcurrencyKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeferredOperationMaxConcurrencyKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeferredOperationMaxConcurrencyKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationResultKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationResultKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationResultKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationResultKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationResultKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationResultKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationResultKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationResultKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationResultKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationResultKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationResultKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationResultKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationResultKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeferredOperationResultKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeferredOperationResultKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeferredOperationResultKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationResultKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationResultKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationResultKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationResultKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationResultKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationResultKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationResultKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationResultKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationResultKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationResultKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationResultKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationResultKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationResultKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeferredOperationResultKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeferredOperationResultKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeferredOperationResultKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationResultKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationResultKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationResultKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationResultKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationResultKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationResultKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationResultKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationResultKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationResultKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationResultKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationResultKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationResultKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationResultKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeferredOperationResultKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeferredOperationResultKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeferredOperationResultKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationResultKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationResultKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationResultKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationResultKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationResultKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationResultKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationResultKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationResultKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationResultKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationResultKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationResultKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationResultKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationResultKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeferredOperationResultKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeferredOperationResultKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeferredOperationResultKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationResultKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationResultKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationResultKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationResultKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationResultKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationResultKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationResultKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationResultKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationResultKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationResultKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationResultKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationResultKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationResultKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeferredOperationResultKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeferredOperationResultKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeferredOperationResultKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationResultKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationResultKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationResultKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationResultKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationResultKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationResultKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationResultKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationResultKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationResultKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationResultKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationResultKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationResultKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationResultKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeferredOperationResultKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeferredOperationResultKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeferredOperationResultKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationResultKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationResultKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationResultKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationResultKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationResultKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationResultKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationResultKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationResultKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationResultKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationResultKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationResultKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationResultKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationResultKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeferredOperationResultKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeferredOperationResultKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeferredOperationResultKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationResultKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationResultKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationResultKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationResultKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationResultKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationResultKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationResultKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationResultKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationResultKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationResultKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationResultKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationResultKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationResultKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeferredOperationResultKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeferredOperationResultKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeferredOperationResultKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationResultKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationResultKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationResultKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationResultKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationResultKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationResultKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationResultKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationResultKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationResultKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationResultKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationResultKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationResultKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationResultKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeferredOperationResultKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeferredOperationResultKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeferredOperationResultKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationResultKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationResultKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationResultKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationResultKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationResultKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationResultKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationResultKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationResultKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationResultKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationResultKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationResultKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationResultKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationResultKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeferredOperationResultKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeferredOperationResultKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeferredOperationResultKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationResultKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationResultKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationResultKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationResultKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationResultKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationResultKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationResultKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationResultKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationResultKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationResultKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationResultKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationResultKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationResultKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeferredOperationResultKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeferredOperationResultKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeferredOperationResultKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationResultKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationResultKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationResultKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationResultKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationResultKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationResultKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationResultKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationResultKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationResultKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationResultKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationResultKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationResultKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationResultKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeferredOperationResultKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeferredOperationResultKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeferredOperationResultKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationResultKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationResultKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationResultKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationResultKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationResultKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationResultKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationResultKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationResultKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationResultKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationResultKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationResultKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationResultKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationResultKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeferredOperationResultKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeferredOperationResultKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeferredOperationResultKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DeferredOperationJoinKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeferredOperationJoinKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeferredOperationJoinKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeferredOperationJoinKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeferredOperationJoinKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeferredOperationJoinKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeferredOperationJoinKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeferredOperationJoinKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeferredOperationJoinKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeferredOperationJoinKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeferredOperationJoinKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeferredOperationJoinKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeferredOperationJoinKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DeferredOperationJoinKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DeferredOperationJoinKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DeferredOperationJoinKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DeferredOperationJoinKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeferredOperationJoinKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeferredOperationJoinKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeferredOperationJoinKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeferredOperationJoinKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeferredOperationJoinKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeferredOperationJoinKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeferredOperationJoinKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeferredOperationJoinKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeferredOperationJoinKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeferredOperationJoinKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeferredOperationJoinKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeferredOperationJoinKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DeferredOperationJoinKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DeferredOperationJoinKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DeferredOperationJoinKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DeferredOperationJoinKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeferredOperationJoinKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeferredOperationJoinKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeferredOperationJoinKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeferredOperationJoinKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeferredOperationJoinKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeferredOperationJoinKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeferredOperationJoinKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeferredOperationJoinKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeferredOperationJoinKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeferredOperationJoinKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeferredOperationJoinKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeferredOperationJoinKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DeferredOperationJoinKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DeferredOperationJoinKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DeferredOperationJoinKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DeferredOperationJoinKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeferredOperationJoinKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeferredOperationJoinKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeferredOperationJoinKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeferredOperationJoinKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeferredOperationJoinKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeferredOperationJoinKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeferredOperationJoinKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeferredOperationJoinKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeferredOperationJoinKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeferredOperationJoinKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeferredOperationJoinKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeferredOperationJoinKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DeferredOperationJoinKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DeferredOperationJoinKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DeferredOperationJoinKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DeferredOperationJoinKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeferredOperationJoinKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeferredOperationJoinKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeferredOperationJoinKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeferredOperationJoinKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeferredOperationJoinKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeferredOperationJoinKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeferredOperationJoinKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeferredOperationJoinKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeferredOperationJoinKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeferredOperationJoinKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeferredOperationJoinKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeferredOperationJoinKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DeferredOperationJoinKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DeferredOperationJoinKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DeferredOperationJoinKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DeferredOperationJoinKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeferredOperationJoinKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeferredOperationJoinKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeferredOperationJoinKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeferredOperationJoinKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeferredOperationJoinKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeferredOperationJoinKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeferredOperationJoinKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeferredOperationJoinKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeferredOperationJoinKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeferredOperationJoinKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeferredOperationJoinKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeferredOperationJoinKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DeferredOperationJoinKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DeferredOperationJoinKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DeferredOperationJoinKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DeferredOperationJoinKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeferredOperationJoinKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeferredOperationJoinKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeferredOperationJoinKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeferredOperationJoinKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeferredOperationJoinKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeferredOperationJoinKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeferredOperationJoinKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeferredOperationJoinKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeferredOperationJoinKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeferredOperationJoinKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeferredOperationJoinKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeferredOperationJoinKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DeferredOperationJoinKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DeferredOperationJoinKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DeferredOperationJoinKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DeferredOperationJoinKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeferredOperationJoinKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeferredOperationJoinKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeferredOperationJoinKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeferredOperationJoinKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeferredOperationJoinKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeferredOperationJoinKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeferredOperationJoinKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeferredOperationJoinKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeferredOperationJoinKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeferredOperationJoinKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeferredOperationJoinKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeferredOperationJoinKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DeferredOperationJoinKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DeferredOperationJoinKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DeferredOperationJoinKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DeferredOperationJoinKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeferredOperationJoinKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeferredOperationJoinKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeferredOperationJoinKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeferredOperationJoinKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeferredOperationJoinKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeferredOperationJoinKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeferredOperationJoinKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeferredOperationJoinKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeferredOperationJoinKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeferredOperationJoinKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeferredOperationJoinKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeferredOperationJoinKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DeferredOperationJoinKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DeferredOperationJoinKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DeferredOperationJoinKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DeferredOperationJoinKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeferredOperationJoinKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeferredOperationJoinKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeferredOperationJoinKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeferredOperationJoinKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeferredOperationJoinKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeferredOperationJoinKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeferredOperationJoinKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeferredOperationJoinKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeferredOperationJoinKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeferredOperationJoinKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeferredOperationJoinKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeferredOperationJoinKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DeferredOperationJoinKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DeferredOperationJoinKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DeferredOperationJoinKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeferredOperationJoinKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeferredOperationJoinKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeferredOperationJoinKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeferredOperationJoinKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeferredOperationJoinKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeferredOperationJoinKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeferredOperationJoinKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeferredOperationJoinKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeferredOperationJoinKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeferredOperationJoinKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeferredOperationJoinKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeferredOperationJoinKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeferredOperationJoinKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DeferredOperationJoinKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DeferredOperationJoinKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DeferredOperationJoinKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DeferredOperationJoinKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeferredOperationJoinKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeferredOperationJoinKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeferredOperationJoinKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeferredOperationJoinKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeferredOperationJoinKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeferredOperationJoinKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeferredOperationJoinKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeferredOperationJoinKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeferredOperationJoinKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeferredOperationJoinKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeferredOperationJoinKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeferredOperationJoinKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DeferredOperationJoinKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DeferredOperationJoinKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DeferredOperationJoinKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DeferredOperationJoinKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeferredOperationJoinKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeferredOperationJoinKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeferredOperationJoinKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeferredOperationJoinKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeferredOperationJoinKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeferredOperationJoinKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeferredOperationJoinKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeferredOperationJoinKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeferredOperationJoinKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeferredOperationJoinKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeferredOperationJoinKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeferredOperationJoinKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DeferredOperationJoinKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DeferredOperationJoinKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DeferredOperationJoinKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPipelineIndirectMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPipelineIndirectMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetPipelineIndirectMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetPipelineIndirectMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetPipelineIndirectMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetPipelineIndirectMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetPipelineIndirectMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetPipelineIndirectMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetPipelineIndirectMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetPipelineIndirectMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetPipelineIndirectMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetPipelineIndirectMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetPipelineIndirectMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetPipelineIndirectMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetPipelineIndirectMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetPipelineIndirectMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetPipelineIndirectMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetPipelineIndirectMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetPipelineIndirectMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetPipelineIndirectMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetPipelineIndirectMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetPipelineIndirectMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetPipelineIndirectMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetPipelineIndirectMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectMemoryRequirementsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectMemoryRequirementsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectMemoryRequirementsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectMemoryRequirementsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectMemoryRequirementsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectMemoryRequirementsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectMemoryRequirementsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectMemoryRequirementsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectMemoryRequirementsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectMemoryRequirementsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectMemoryRequirementsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectMemoryRequirementsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectMemoryRequirementsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectMemoryRequirementsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetPipelineIndirectMemoryRequirementsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetPipelineIndirectMemoryRequirementsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectDeviceAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectDeviceAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectDeviceAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectDeviceAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectDeviceAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectDeviceAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectDeviceAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectDeviceAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectDeviceAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectDeviceAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectDeviceAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectDeviceAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectDeviceAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelineIndirectDeviceAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPipelineIndirectDeviceAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPipelineIndirectDeviceAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectDeviceAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectDeviceAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectDeviceAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectDeviceAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectDeviceAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectDeviceAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectDeviceAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectDeviceAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectDeviceAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectDeviceAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectDeviceAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectDeviceAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectDeviceAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelineIndirectDeviceAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetPipelineIndirectDeviceAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetPipelineIndirectDeviceAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectDeviceAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectDeviceAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectDeviceAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectDeviceAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectDeviceAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectDeviceAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectDeviceAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectDeviceAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectDeviceAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectDeviceAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectDeviceAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectDeviceAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectDeviceAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelineIndirectDeviceAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetPipelineIndirectDeviceAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetPipelineIndirectDeviceAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectDeviceAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectDeviceAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectDeviceAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectDeviceAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectDeviceAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectDeviceAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectDeviceAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectDeviceAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectDeviceAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectDeviceAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectDeviceAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectDeviceAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectDeviceAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelineIndirectDeviceAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetPipelineIndirectDeviceAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetPipelineIndirectDeviceAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectDeviceAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectDeviceAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectDeviceAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectDeviceAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectDeviceAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectDeviceAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectDeviceAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectDeviceAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectDeviceAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectDeviceAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectDeviceAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectDeviceAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectDeviceAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelineIndirectDeviceAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetPipelineIndirectDeviceAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetPipelineIndirectDeviceAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectDeviceAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectDeviceAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectDeviceAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectDeviceAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectDeviceAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectDeviceAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectDeviceAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectDeviceAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectDeviceAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectDeviceAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectDeviceAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectDeviceAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectDeviceAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelineIndirectDeviceAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetPipelineIndirectDeviceAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetPipelineIndirectDeviceAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectDeviceAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectDeviceAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectDeviceAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectDeviceAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectDeviceAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectDeviceAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectDeviceAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectDeviceAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectDeviceAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectDeviceAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectDeviceAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectDeviceAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectDeviceAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelineIndirectDeviceAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetPipelineIndirectDeviceAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetPipelineIndirectDeviceAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectDeviceAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectDeviceAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectDeviceAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectDeviceAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectDeviceAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectDeviceAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectDeviceAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectDeviceAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectDeviceAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectDeviceAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectDeviceAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectDeviceAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectDeviceAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelineIndirectDeviceAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetPipelineIndirectDeviceAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetPipelineIndirectDeviceAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectDeviceAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectDeviceAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectDeviceAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectDeviceAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectDeviceAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectDeviceAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectDeviceAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectDeviceAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectDeviceAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectDeviceAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectDeviceAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectDeviceAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectDeviceAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelineIndirectDeviceAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetPipelineIndirectDeviceAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetPipelineIndirectDeviceAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectDeviceAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectDeviceAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectDeviceAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectDeviceAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectDeviceAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectDeviceAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectDeviceAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectDeviceAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectDeviceAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectDeviceAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectDeviceAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectDeviceAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectDeviceAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelineIndirectDeviceAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetPipelineIndirectDeviceAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetPipelineIndirectDeviceAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectDeviceAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectDeviceAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectDeviceAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectDeviceAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectDeviceAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectDeviceAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectDeviceAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectDeviceAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectDeviceAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectDeviceAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectDeviceAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectDeviceAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectDeviceAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelineIndirectDeviceAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetPipelineIndirectDeviceAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetPipelineIndirectDeviceAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectDeviceAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectDeviceAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectDeviceAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectDeviceAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectDeviceAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectDeviceAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectDeviceAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectDeviceAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectDeviceAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectDeviceAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectDeviceAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectDeviceAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectDeviceAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelineIndirectDeviceAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetPipelineIndirectDeviceAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetPipelineIndirectDeviceAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectDeviceAddressNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectDeviceAddressNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectDeviceAddressNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectDeviceAddressNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectDeviceAddressNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectDeviceAddressNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectDeviceAddressNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectDeviceAddressNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectDeviceAddressNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectDeviceAddressNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectDeviceAddressNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectDeviceAddressNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectDeviceAddressNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelineIndirectDeviceAddressNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetPipelineIndirectDeviceAddressNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetPipelineIndirectDeviceAddressNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_AntiLagUpdateAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AntiLagUpdateAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AntiLagUpdateAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AntiLagUpdateAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AntiLagUpdateAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AntiLagUpdateAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AntiLagUpdateAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AntiLagUpdateAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AntiLagUpdateAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AntiLagUpdateAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AntiLagUpdateAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AntiLagUpdateAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AntiLagUpdateAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_AntiLagUpdateAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_AntiLagUpdateAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_AntiLagUpdateAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_AntiLagUpdateAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AntiLagUpdateAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AntiLagUpdateAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AntiLagUpdateAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AntiLagUpdateAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AntiLagUpdateAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AntiLagUpdateAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AntiLagUpdateAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AntiLagUpdateAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AntiLagUpdateAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AntiLagUpdateAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AntiLagUpdateAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AntiLagUpdateAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_AntiLagUpdateAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_AntiLagUpdateAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_AntiLagUpdateAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_AntiLagUpdateAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AntiLagUpdateAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AntiLagUpdateAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AntiLagUpdateAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AntiLagUpdateAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AntiLagUpdateAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AntiLagUpdateAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AntiLagUpdateAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AntiLagUpdateAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AntiLagUpdateAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AntiLagUpdateAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AntiLagUpdateAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AntiLagUpdateAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_AntiLagUpdateAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_AntiLagUpdateAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_AntiLagUpdateAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_AntiLagUpdateAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AntiLagUpdateAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AntiLagUpdateAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AntiLagUpdateAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AntiLagUpdateAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AntiLagUpdateAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AntiLagUpdateAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AntiLagUpdateAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AntiLagUpdateAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AntiLagUpdateAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AntiLagUpdateAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AntiLagUpdateAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AntiLagUpdateAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_AntiLagUpdateAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_AntiLagUpdateAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_AntiLagUpdateAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_AntiLagUpdateAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AntiLagUpdateAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AntiLagUpdateAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AntiLagUpdateAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AntiLagUpdateAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AntiLagUpdateAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AntiLagUpdateAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AntiLagUpdateAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AntiLagUpdateAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AntiLagUpdateAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AntiLagUpdateAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AntiLagUpdateAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AntiLagUpdateAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_AntiLagUpdateAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_AntiLagUpdateAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_AntiLagUpdateAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_AntiLagUpdateAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AntiLagUpdateAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AntiLagUpdateAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AntiLagUpdateAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AntiLagUpdateAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AntiLagUpdateAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AntiLagUpdateAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AntiLagUpdateAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AntiLagUpdateAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AntiLagUpdateAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AntiLagUpdateAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AntiLagUpdateAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AntiLagUpdateAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_AntiLagUpdateAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_AntiLagUpdateAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_AntiLagUpdateAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_AntiLagUpdateAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AntiLagUpdateAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AntiLagUpdateAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AntiLagUpdateAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AntiLagUpdateAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AntiLagUpdateAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AntiLagUpdateAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AntiLagUpdateAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AntiLagUpdateAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AntiLagUpdateAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AntiLagUpdateAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AntiLagUpdateAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AntiLagUpdateAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_AntiLagUpdateAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_AntiLagUpdateAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_AntiLagUpdateAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_AntiLagUpdateAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AntiLagUpdateAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AntiLagUpdateAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AntiLagUpdateAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AntiLagUpdateAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AntiLagUpdateAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AntiLagUpdateAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AntiLagUpdateAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AntiLagUpdateAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AntiLagUpdateAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AntiLagUpdateAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AntiLagUpdateAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AntiLagUpdateAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_AntiLagUpdateAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_AntiLagUpdateAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_AntiLagUpdateAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_AntiLagUpdateAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AntiLagUpdateAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AntiLagUpdateAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AntiLagUpdateAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AntiLagUpdateAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AntiLagUpdateAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AntiLagUpdateAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AntiLagUpdateAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AntiLagUpdateAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AntiLagUpdateAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AntiLagUpdateAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AntiLagUpdateAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AntiLagUpdateAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_AntiLagUpdateAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_AntiLagUpdateAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_AntiLagUpdateAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_AntiLagUpdateAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AntiLagUpdateAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AntiLagUpdateAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AntiLagUpdateAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AntiLagUpdateAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AntiLagUpdateAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AntiLagUpdateAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AntiLagUpdateAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AntiLagUpdateAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AntiLagUpdateAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AntiLagUpdateAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AntiLagUpdateAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AntiLagUpdateAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_AntiLagUpdateAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_AntiLagUpdateAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_AntiLagUpdateAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_AntiLagUpdateAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AntiLagUpdateAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AntiLagUpdateAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AntiLagUpdateAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AntiLagUpdateAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AntiLagUpdateAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AntiLagUpdateAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AntiLagUpdateAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AntiLagUpdateAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AntiLagUpdateAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AntiLagUpdateAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AntiLagUpdateAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AntiLagUpdateAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_AntiLagUpdateAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_AntiLagUpdateAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_AntiLagUpdateAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_AntiLagUpdateAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AntiLagUpdateAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AntiLagUpdateAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AntiLagUpdateAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AntiLagUpdateAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AntiLagUpdateAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AntiLagUpdateAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AntiLagUpdateAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AntiLagUpdateAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AntiLagUpdateAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AntiLagUpdateAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AntiLagUpdateAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AntiLagUpdateAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_AntiLagUpdateAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_AntiLagUpdateAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_AntiLagUpdateAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_AntiLagUpdateAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AntiLagUpdateAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AntiLagUpdateAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AntiLagUpdateAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AntiLagUpdateAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AntiLagUpdateAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AntiLagUpdateAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AntiLagUpdateAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AntiLagUpdateAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AntiLagUpdateAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AntiLagUpdateAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AntiLagUpdateAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AntiLagUpdateAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_AntiLagUpdateAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_AntiLagUpdateAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_AntiLagUpdateAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullMode@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullMode@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullMode@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullMode@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullMode@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullMode@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullMode@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullMode@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullMode@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullMode@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullMode@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullMode@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullMode@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullMode@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetCullMode=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetCullMode=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullMode@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullMode@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullMode@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullMode@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullMode@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullMode@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullMode@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullMode@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullMode@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullMode@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullMode@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullMode@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullMode@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullMode@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetCullMode=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetCullMode=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullMode@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullMode@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullMode@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullMode@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullMode@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullMode@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullMode@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullMode@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullMode@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullMode@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullMode@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullMode@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullMode@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullMode@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetCullMode=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetCullMode=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullMode@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullMode@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullMode@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullMode@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullMode@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullMode@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullMode@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullMode@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullMode@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullMode@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullMode@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullMode@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullMode@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullMode@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetCullMode=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetCullMode=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullMode@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullMode@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullMode@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullMode@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullMode@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullMode@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullMode@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullMode@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullMode@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullMode@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullMode@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullMode@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullMode@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullMode@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetCullMode=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetCullMode=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullMode@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullMode@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullMode@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullMode@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullMode@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullMode@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullMode@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullMode@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullMode@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullMode@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullMode@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullMode@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullMode@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullMode@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetCullMode=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetCullMode=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullMode@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullMode@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullMode@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullMode@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullMode@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullMode@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullMode@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullMode@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullMode@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullMode@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullMode@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullMode@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullMode@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullMode@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetCullMode=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetCullMode=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullMode@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullMode@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullMode@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullMode@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullMode@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullMode@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullMode@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullMode@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullMode@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullMode@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullMode@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullMode@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullMode@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullMode@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetCullMode=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetCullMode=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullMode@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullMode@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullMode@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullMode@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullMode@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullMode@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullMode@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullMode@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullMode@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullMode@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullMode@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullMode@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullMode@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullMode@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetCullMode=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetCullMode=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullMode@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullMode@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullMode@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullMode@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullMode@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullMode@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullMode@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullMode@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullMode@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullMode@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullMode@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullMode@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullMode@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullMode@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetCullMode=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetCullMode=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullMode@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullMode@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullMode@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullMode@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullMode@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullMode@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullMode@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullMode@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullMode@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullMode@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullMode@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullMode@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullMode@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullMode@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetCullMode=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetCullMode=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullMode@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullMode@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullMode@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullMode@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullMode@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullMode@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullMode@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullMode@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullMode@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullMode@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullMode@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullMode@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullMode@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullMode@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetCullMode=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetCullMode=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullMode@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullMode@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullMode@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullMode@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullMode@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullMode@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullMode@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullMode@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullMode@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullMode@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullMode@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullMode@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullMode@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullMode@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetCullMode=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetCullMode=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCullModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetCullModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetCullModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCullModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetCullModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetCullModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCullModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetCullModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetCullModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCullModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetCullModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetCullModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCullModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetCullModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetCullModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCullModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetCullModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetCullModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCullModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetCullModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetCullModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCullModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetCullModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetCullModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCullModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetCullModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetCullModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCullModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetCullModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetCullModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCullModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetCullModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetCullModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCullModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetCullModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetCullModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCullModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetCullModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetCullModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFace@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFace@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFace@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFace@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFace@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFace@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFace@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFace@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFace@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFace@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFace@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFace@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFace@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFace@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetFrontFace=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetFrontFace=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFace@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFace@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFace@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFace@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFace@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFace@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFace@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFace@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFace@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFace@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFace@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFace@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFace@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFace@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetFrontFace=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetFrontFace=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFace@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFace@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFace@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFace@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFace@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFace@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFace@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFace@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFace@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFace@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFace@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFace@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFace@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFace@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetFrontFace=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetFrontFace=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFace@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFace@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFace@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFace@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFace@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFace@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFace@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFace@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFace@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFace@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFace@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFace@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFace@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFace@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetFrontFace=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetFrontFace=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFace@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFace@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFace@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFace@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFace@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFace@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFace@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFace@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFace@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFace@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFace@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFace@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFace@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFace@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetFrontFace=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetFrontFace=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFace@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFace@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFace@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFace@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFace@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFace@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFace@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFace@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFace@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFace@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFace@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFace@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFace@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFace@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetFrontFace=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetFrontFace=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFace@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFace@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFace@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFace@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFace@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFace@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFace@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFace@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFace@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFace@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFace@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFace@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFace@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFace@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetFrontFace=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetFrontFace=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFace@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFace@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFace@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFace@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFace@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFace@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFace@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFace@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFace@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFace@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFace@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFace@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFace@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFace@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetFrontFace=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetFrontFace=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFace@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFace@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFace@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFace@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFace@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFace@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFace@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFace@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFace@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFace@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFace@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFace@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFace@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFace@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetFrontFace=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetFrontFace=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFace@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFace@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFace@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFace@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFace@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFace@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFace@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFace@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFace@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFace@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFace@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFace@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFace@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFace@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetFrontFace=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetFrontFace=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFace@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFace@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFace@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFace@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFace@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFace@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFace@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFace@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFace@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFace@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFace@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFace@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFace@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFace@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetFrontFace=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetFrontFace=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFace@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFace@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFace@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFace@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFace@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFace@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFace@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFace@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFace@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFace@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFace@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFace@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFace@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFace@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetFrontFace=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetFrontFace=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFace@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFace@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFace@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFace@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFace@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFace@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFace@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFace@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFace@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFace@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFace@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFace@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFace@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFace@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetFrontFace=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetFrontFace=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFrontFaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetFrontFaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetFrontFaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFrontFaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetFrontFaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetFrontFaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFrontFaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetFrontFaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetFrontFaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFrontFaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetFrontFaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetFrontFaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFrontFaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetFrontFaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetFrontFaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFrontFaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetFrontFaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetFrontFaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFrontFaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetFrontFaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetFrontFaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFrontFaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetFrontFaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetFrontFaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFrontFaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetFrontFaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetFrontFaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFrontFaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetFrontFaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetFrontFaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFrontFaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetFrontFaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetFrontFaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFrontFaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetFrontFaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetFrontFaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFaceEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFaceEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFaceEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFaceEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFaceEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFaceEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFaceEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFaceEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFaceEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFaceEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFaceEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFaceEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFaceEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFrontFaceEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetFrontFaceEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetFrontFaceEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopology@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopology@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopology@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopology@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopology@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopology@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopology@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopology@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopology@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopology@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopology@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopology@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopology@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopology@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetPrimitiveTopology=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetPrimitiveTopology=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopology@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopology@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopology@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopology@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopology@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopology@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopology@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopology@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopology@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopology@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopology@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopology@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopology@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopology@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetPrimitiveTopology=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetPrimitiveTopology=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopology@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopology@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopology@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopology@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopology@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopology@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopology@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopology@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopology@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopology@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopology@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopology@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopology@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopology@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetPrimitiveTopology=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetPrimitiveTopology=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopology@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopology@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopology@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopology@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopology@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopology@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopology@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopology@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopology@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopology@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopology@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopology@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopology@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopology@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetPrimitiveTopology=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetPrimitiveTopology=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopology@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopology@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopology@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopology@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopology@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopology@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopology@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopology@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopology@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopology@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopology@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopology@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopology@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopology@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetPrimitiveTopology=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetPrimitiveTopology=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopology@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopology@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopology@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopology@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopology@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopology@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopology@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopology@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopology@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopology@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopology@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopology@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopology@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopology@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetPrimitiveTopology=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetPrimitiveTopology=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopology@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopology@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopology@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopology@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopology@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopology@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopology@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopology@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopology@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopology@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopology@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopology@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopology@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopology@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetPrimitiveTopology=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetPrimitiveTopology=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopology@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopology@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopology@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopology@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopology@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopology@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopology@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopology@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopology@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopology@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopology@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopology@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopology@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopology@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetPrimitiveTopology=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetPrimitiveTopology=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopology@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopology@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopology@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopology@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopology@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopology@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopology@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopology@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopology@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopology@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopology@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopology@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopology@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopology@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetPrimitiveTopology=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetPrimitiveTopology=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopology@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopology@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopology@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopology@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopology@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopology@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopology@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopology@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopology@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopology@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopology@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopology@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopology@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopology@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetPrimitiveTopology=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetPrimitiveTopology=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopology@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopology@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopology@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopology@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopology@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopology@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopology@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopology@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopology@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopology@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopology@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopology@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopology@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopology@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetPrimitiveTopology=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetPrimitiveTopology=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopology@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopology@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopology@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopology@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopology@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopology@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopology@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopology@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopology@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopology@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopology@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopology@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopology@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopology@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetPrimitiveTopology=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetPrimitiveTopology=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopology@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopology@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopology@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopology@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopology@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopology@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopology@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopology@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopology@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopology@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopology@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopology@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopology@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopology@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetPrimitiveTopology=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetPrimitiveTopology=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopologyEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopologyEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopologyEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopologyEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopologyEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopologyEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopologyEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopologyEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopologyEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopologyEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopologyEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopologyEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopologyEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveTopologyEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetPrimitiveTopologyEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetPrimitiveTopologyEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopologyEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopologyEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopologyEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopologyEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopologyEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopologyEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopologyEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopologyEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopologyEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopologyEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopologyEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopologyEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopologyEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveTopologyEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetPrimitiveTopologyEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetPrimitiveTopologyEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopologyEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopologyEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopologyEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopologyEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopologyEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopologyEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopologyEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopologyEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopologyEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopologyEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopologyEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopologyEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopologyEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveTopologyEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetPrimitiveTopologyEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetPrimitiveTopologyEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopologyEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopologyEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopologyEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopologyEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopologyEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopologyEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopologyEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopologyEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopologyEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopologyEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopologyEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopologyEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopologyEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveTopologyEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetPrimitiveTopologyEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetPrimitiveTopologyEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopologyEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopologyEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopologyEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopologyEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopologyEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopologyEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopologyEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopologyEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopologyEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopologyEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopologyEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopologyEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopologyEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveTopologyEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetPrimitiveTopologyEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetPrimitiveTopologyEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopologyEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopologyEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopologyEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopologyEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopologyEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopologyEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopologyEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopologyEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopologyEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopologyEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopologyEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopologyEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopologyEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveTopologyEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetPrimitiveTopologyEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetPrimitiveTopologyEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopologyEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopologyEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopologyEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopologyEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopologyEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopologyEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopologyEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopologyEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopologyEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopologyEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopologyEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopologyEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopologyEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveTopologyEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetPrimitiveTopologyEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetPrimitiveTopologyEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopologyEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopologyEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopologyEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopologyEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopologyEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopologyEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopologyEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopologyEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopologyEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopologyEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopologyEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopologyEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopologyEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveTopologyEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetPrimitiveTopologyEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetPrimitiveTopologyEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopologyEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopologyEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopologyEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopologyEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopologyEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopologyEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopologyEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopologyEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopologyEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopologyEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopologyEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopologyEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopologyEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveTopologyEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetPrimitiveTopologyEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetPrimitiveTopologyEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopologyEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopologyEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopologyEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopologyEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopologyEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopologyEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopologyEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopologyEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopologyEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopologyEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopologyEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopologyEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopologyEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveTopologyEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetPrimitiveTopologyEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetPrimitiveTopologyEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopologyEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopologyEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopologyEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopologyEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopologyEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopologyEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopologyEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopologyEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopologyEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopologyEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopologyEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopologyEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopologyEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveTopologyEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetPrimitiveTopologyEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetPrimitiveTopologyEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopologyEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopologyEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopologyEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopologyEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopologyEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopologyEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopologyEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopologyEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopologyEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopologyEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopologyEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopologyEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopologyEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveTopologyEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetPrimitiveTopologyEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetPrimitiveTopologyEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopologyEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopologyEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopologyEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopologyEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopologyEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopologyEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopologyEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopologyEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopologyEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopologyEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopologyEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopologyEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopologyEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveTopologyEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetPrimitiveTopologyEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetPrimitiveTopologyEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetViewportWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetViewportWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetViewportWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetViewportWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetViewportWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetViewportWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetViewportWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetViewportWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetViewportWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetViewportWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetViewportWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetViewportWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetViewportWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetViewportWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetViewportWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetViewportWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetViewportWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetViewportWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetViewportWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetViewportWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetViewportWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetViewportWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetViewportWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetViewportWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetViewportWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetViewportWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetViewportWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetViewportWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetViewportWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetViewportWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetViewportWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetViewportWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetViewportWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetViewportWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetViewportWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetViewportWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetViewportWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetViewportWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetViewportWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetViewportWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetViewportWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetViewportWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetViewportWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetViewportWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetViewportWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetViewportWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetViewportWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetViewportWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetViewportWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetViewportWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetViewportWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetViewportWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetScissorWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetScissorWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetScissorWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetScissorWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetScissorWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetScissorWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetScissorWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetScissorWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetScissorWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetScissorWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetScissorWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetScissorWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetScissorWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetScissorWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetScissorWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetScissorWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetScissorWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetScissorWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetScissorWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetScissorWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetScissorWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetScissorWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetScissorWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetScissorWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCount@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCount@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCount@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCount@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCount@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCount@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCount@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCount@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCount@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCount@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCount@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCount@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCount@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCount@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetScissorWithCount=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetScissorWithCount=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetScissorWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetScissorWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetScissorWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetScissorWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetScissorWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetScissorWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetScissorWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetScissorWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetScissorWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetScissorWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetScissorWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetScissorWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetScissorWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetScissorWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetScissorWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetScissorWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetScissorWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetScissorWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetScissorWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetScissorWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetScissorWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetScissorWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetScissorWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetScissorWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetScissorWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetScissorWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetScissorWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetScissorWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetScissorWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetScissorWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetScissorWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetScissorWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetScissorWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetScissorWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetScissorWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetScissorWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetScissorWithCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetScissorWithCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetScissorWithCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindIndexBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindIndexBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindIndexBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindIndexBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindIndexBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindIndexBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindIndexBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindIndexBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindIndexBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindIndexBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindIndexBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindIndexBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindIndexBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindIndexBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindIndexBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindIndexBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindIndexBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindIndexBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindIndexBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindIndexBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindIndexBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindIndexBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindIndexBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindIndexBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindIndexBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindIndexBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindIndexBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindIndexBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindIndexBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindIndexBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindIndexBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindIndexBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindIndexBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindIndexBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindIndexBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindIndexBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindIndexBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindIndexBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindIndexBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindIndexBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindIndexBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindIndexBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindIndexBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindIndexBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindIndexBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindIndexBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindIndexBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindIndexBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindIndexBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindIndexBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindIndexBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindIndexBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindVertexBuffers2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindVertexBuffers2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindVertexBuffers2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindVertexBuffers2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindVertexBuffers2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindVertexBuffers2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindVertexBuffers2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindVertexBuffers2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindVertexBuffers2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindVertexBuffers2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindVertexBuffers2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindVertexBuffers2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindVertexBuffers2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindVertexBuffers2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindVertexBuffers2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindVertexBuffers2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindVertexBuffers2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindVertexBuffers2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindVertexBuffers2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindVertexBuffers2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindVertexBuffers2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindVertexBuffers2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindVertexBuffers2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindVertexBuffers2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindVertexBuffers2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindVertexBuffers2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindVertexBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindVertexBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindVertexBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindVertexBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindVertexBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindVertexBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindVertexBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindVertexBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindVertexBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindVertexBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindVertexBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindVertexBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindVertexBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindVertexBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindVertexBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindVertexBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindVertexBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindVertexBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindVertexBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindVertexBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindVertexBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindVertexBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindVertexBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindVertexBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindVertexBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindVertexBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDepthTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDepthTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDepthTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDepthTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDepthTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDepthTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDepthTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDepthTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDepthTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDepthTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDepthTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDepthTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDepthTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDepthTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDepthTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDepthTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDepthTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDepthTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDepthTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDepthTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDepthTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDepthTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDepthTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDepthTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDepthTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDepthTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDepthTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDepthTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDepthTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDepthTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDepthTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDepthTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDepthTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDepthTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDepthTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDepthTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDepthTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDepthTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDepthTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDepthTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDepthTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDepthTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDepthTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDepthTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDepthTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDepthTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDepthTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDepthTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDepthTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDepthTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDepthTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDepthTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDepthWriteEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDepthWriteEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDepthWriteEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDepthWriteEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDepthWriteEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDepthWriteEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDepthWriteEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDepthWriteEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDepthWriteEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDepthWriteEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDepthWriteEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDepthWriteEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDepthWriteEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDepthWriteEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDepthWriteEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDepthWriteEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDepthWriteEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDepthWriteEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDepthWriteEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDepthWriteEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDepthWriteEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDepthWriteEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDepthWriteEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDepthWriteEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDepthWriteEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDepthWriteEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDepthWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDepthWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDepthWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDepthWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDepthWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDepthWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDepthWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDepthWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDepthWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDepthWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDepthWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDepthWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDepthWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDepthWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDepthWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDepthWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDepthWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDepthWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDepthWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDepthWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDepthWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDepthWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDepthWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDepthWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDepthWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDepthWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDepthCompareOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDepthCompareOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDepthCompareOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDepthCompareOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDepthCompareOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDepthCompareOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDepthCompareOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDepthCompareOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDepthCompareOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDepthCompareOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDepthCompareOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDepthCompareOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDepthCompareOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDepthCompareOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDepthCompareOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDepthCompareOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDepthCompareOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDepthCompareOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDepthCompareOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDepthCompareOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDepthCompareOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDepthCompareOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDepthCompareOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDepthCompareOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDepthCompareOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDepthCompareOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthCompareOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDepthCompareOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDepthCompareOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthCompareOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDepthCompareOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDepthCompareOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthCompareOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDepthCompareOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDepthCompareOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthCompareOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDepthCompareOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDepthCompareOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthCompareOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDepthCompareOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDepthCompareOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthCompareOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDepthCompareOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDepthCompareOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthCompareOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDepthCompareOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDepthCompareOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthCompareOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDepthCompareOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDepthCompareOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthCompareOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDepthCompareOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDepthCompareOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthCompareOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDepthCompareOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDepthCompareOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthCompareOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDepthCompareOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDepthCompareOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthCompareOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDepthCompareOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDepthCompareOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthCompareOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDepthCompareOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDepthCompareOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDepthBoundsTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDepthBoundsTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDepthBoundsTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDepthBoundsTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDepthBoundsTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDepthBoundsTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDepthBoundsTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDepthBoundsTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDepthBoundsTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDepthBoundsTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDepthBoundsTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDepthBoundsTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDepthBoundsTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDepthBoundsTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDepthBoundsTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDepthBoundsTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDepthBoundsTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDepthBoundsTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDepthBoundsTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDepthBoundsTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDepthBoundsTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDepthBoundsTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDepthBoundsTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDepthBoundsTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDepthBoundsTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDepthBoundsTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBoundsTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDepthBoundsTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDepthBoundsTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBoundsTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDepthBoundsTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDepthBoundsTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBoundsTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDepthBoundsTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDepthBoundsTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBoundsTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDepthBoundsTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDepthBoundsTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBoundsTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDepthBoundsTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDepthBoundsTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBoundsTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDepthBoundsTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDepthBoundsTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBoundsTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDepthBoundsTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDepthBoundsTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBoundsTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDepthBoundsTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDepthBoundsTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBoundsTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDepthBoundsTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDepthBoundsTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBoundsTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDepthBoundsTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDepthBoundsTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBoundsTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDepthBoundsTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDepthBoundsTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBoundsTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDepthBoundsTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDepthBoundsTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBoundsTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDepthBoundsTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDepthBoundsTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetStencilTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetStencilTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetStencilTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetStencilTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetStencilTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetStencilTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetStencilTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetStencilTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetStencilTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetStencilTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetStencilTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetStencilTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetStencilTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetStencilTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetStencilTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetStencilTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetStencilTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetStencilTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetStencilTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetStencilTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetStencilTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetStencilTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetStencilTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetStencilTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetStencilTestEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetStencilTestEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetStencilTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetStencilTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetStencilTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetStencilTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetStencilTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetStencilTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetStencilTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetStencilTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetStencilTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetStencilTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetStencilTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetStencilTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetStencilTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetStencilTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetStencilTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetStencilTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetStencilTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetStencilTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetStencilTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetStencilTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetStencilTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetStencilTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetStencilTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetStencilTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilTestEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetStencilTestEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetStencilTestEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetStencilOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetStencilOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetStencilOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetStencilOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetStencilOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetStencilOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetStencilOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetStencilOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetStencilOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetStencilOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetStencilOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetStencilOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetStencilOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetStencilOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetStencilOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetStencilOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetStencilOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetStencilOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetStencilOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetStencilOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetStencilOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetStencilOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetStencilOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetStencilOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOp@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOp@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOp@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOp@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOp@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOp@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOp@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOp@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOp@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOp@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOp@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOp@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOp@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOp@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetStencilOp=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetStencilOp=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetStencilOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetStencilOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetStencilOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetStencilOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetStencilOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetStencilOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetStencilOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetStencilOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetStencilOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetStencilOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetStencilOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetStencilOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetStencilOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetStencilOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetStencilOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetStencilOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetStencilOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetStencilOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetStencilOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetStencilOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetStencilOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetStencilOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetStencilOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetStencilOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetStencilOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetStencilOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetStencilOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetStencilOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetStencilOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetStencilOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetStencilOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetStencilOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetStencilOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetStencilOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetStencilOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetStencilOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetStencilOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetStencilOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetStencilOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetPatchControlPointsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPatchControlPointsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPatchControlPointsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPatchControlPointsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPatchControlPointsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPatchControlPointsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPatchControlPointsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPatchControlPointsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPatchControlPointsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPatchControlPointsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPatchControlPointsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPatchControlPointsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPatchControlPointsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPatchControlPointsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetPatchControlPointsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetPatchControlPointsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPatchControlPointsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPatchControlPointsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPatchControlPointsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPatchControlPointsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPatchControlPointsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPatchControlPointsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPatchControlPointsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPatchControlPointsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPatchControlPointsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPatchControlPointsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPatchControlPointsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPatchControlPointsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPatchControlPointsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPatchControlPointsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetPatchControlPointsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetPatchControlPointsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPatchControlPointsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPatchControlPointsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPatchControlPointsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPatchControlPointsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPatchControlPointsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPatchControlPointsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPatchControlPointsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPatchControlPointsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPatchControlPointsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPatchControlPointsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPatchControlPointsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPatchControlPointsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPatchControlPointsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPatchControlPointsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetPatchControlPointsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetPatchControlPointsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPatchControlPointsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPatchControlPointsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPatchControlPointsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPatchControlPointsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPatchControlPointsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPatchControlPointsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPatchControlPointsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPatchControlPointsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPatchControlPointsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPatchControlPointsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPatchControlPointsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPatchControlPointsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPatchControlPointsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPatchControlPointsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetPatchControlPointsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetPatchControlPointsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPatchControlPointsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPatchControlPointsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPatchControlPointsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPatchControlPointsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPatchControlPointsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPatchControlPointsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPatchControlPointsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPatchControlPointsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPatchControlPointsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPatchControlPointsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPatchControlPointsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPatchControlPointsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPatchControlPointsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPatchControlPointsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetPatchControlPointsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetPatchControlPointsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPatchControlPointsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPatchControlPointsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPatchControlPointsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPatchControlPointsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPatchControlPointsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPatchControlPointsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPatchControlPointsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPatchControlPointsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPatchControlPointsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPatchControlPointsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPatchControlPointsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPatchControlPointsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPatchControlPointsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPatchControlPointsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetPatchControlPointsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetPatchControlPointsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPatchControlPointsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPatchControlPointsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPatchControlPointsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPatchControlPointsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPatchControlPointsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPatchControlPointsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPatchControlPointsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPatchControlPointsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPatchControlPointsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPatchControlPointsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPatchControlPointsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPatchControlPointsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPatchControlPointsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPatchControlPointsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetPatchControlPointsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetPatchControlPointsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPatchControlPointsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPatchControlPointsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPatchControlPointsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPatchControlPointsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPatchControlPointsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPatchControlPointsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPatchControlPointsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPatchControlPointsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPatchControlPointsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPatchControlPointsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPatchControlPointsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPatchControlPointsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPatchControlPointsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPatchControlPointsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetPatchControlPointsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetPatchControlPointsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPatchControlPointsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPatchControlPointsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPatchControlPointsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPatchControlPointsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPatchControlPointsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPatchControlPointsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPatchControlPointsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPatchControlPointsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPatchControlPointsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPatchControlPointsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPatchControlPointsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPatchControlPointsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPatchControlPointsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPatchControlPointsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetPatchControlPointsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetPatchControlPointsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPatchControlPointsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPatchControlPointsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPatchControlPointsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPatchControlPointsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPatchControlPointsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPatchControlPointsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPatchControlPointsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPatchControlPointsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPatchControlPointsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPatchControlPointsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPatchControlPointsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPatchControlPointsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPatchControlPointsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPatchControlPointsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetPatchControlPointsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetPatchControlPointsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPatchControlPointsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPatchControlPointsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPatchControlPointsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPatchControlPointsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPatchControlPointsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPatchControlPointsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPatchControlPointsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPatchControlPointsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPatchControlPointsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPatchControlPointsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPatchControlPointsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPatchControlPointsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPatchControlPointsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPatchControlPointsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetPatchControlPointsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetPatchControlPointsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPatchControlPointsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPatchControlPointsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPatchControlPointsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPatchControlPointsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPatchControlPointsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPatchControlPointsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPatchControlPointsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPatchControlPointsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPatchControlPointsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPatchControlPointsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPatchControlPointsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPatchControlPointsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPatchControlPointsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPatchControlPointsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetPatchControlPointsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetPatchControlPointsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPatchControlPointsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPatchControlPointsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPatchControlPointsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPatchControlPointsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPatchControlPointsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPatchControlPointsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPatchControlPointsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPatchControlPointsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPatchControlPointsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPatchControlPointsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPatchControlPointsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPatchControlPointsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPatchControlPointsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPatchControlPointsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetPatchControlPointsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetPatchControlPointsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetRasterizerDiscardEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetRasterizerDiscardEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetRasterizerDiscardEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetRasterizerDiscardEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetRasterizerDiscardEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetRasterizerDiscardEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetRasterizerDiscardEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetRasterizerDiscardEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetRasterizerDiscardEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetRasterizerDiscardEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetRasterizerDiscardEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetRasterizerDiscardEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetRasterizerDiscardEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetRasterizerDiscardEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetRasterizerDiscardEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetRasterizerDiscardEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetRasterizerDiscardEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetRasterizerDiscardEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetRasterizerDiscardEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetRasterizerDiscardEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetRasterizerDiscardEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetRasterizerDiscardEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetRasterizerDiscardEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetRasterizerDiscardEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetRasterizerDiscardEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetRasterizerDiscardEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizerDiscardEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetRasterizerDiscardEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetRasterizerDiscardEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizerDiscardEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetRasterizerDiscardEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetRasterizerDiscardEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizerDiscardEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetRasterizerDiscardEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetRasterizerDiscardEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizerDiscardEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetRasterizerDiscardEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetRasterizerDiscardEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizerDiscardEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetRasterizerDiscardEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetRasterizerDiscardEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizerDiscardEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetRasterizerDiscardEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetRasterizerDiscardEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizerDiscardEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetRasterizerDiscardEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetRasterizerDiscardEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizerDiscardEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetRasterizerDiscardEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetRasterizerDiscardEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizerDiscardEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetRasterizerDiscardEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetRasterizerDiscardEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizerDiscardEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetRasterizerDiscardEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetRasterizerDiscardEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizerDiscardEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetRasterizerDiscardEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetRasterizerDiscardEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizerDiscardEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetRasterizerDiscardEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetRasterizerDiscardEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizerDiscardEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetRasterizerDiscardEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetRasterizerDiscardEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDepthBiasEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDepthBiasEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDepthBiasEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDepthBiasEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDepthBiasEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDepthBiasEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDepthBiasEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDepthBiasEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDepthBiasEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDepthBiasEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDepthBiasEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDepthBiasEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDepthBiasEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDepthBiasEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDepthBiasEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDepthBiasEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDepthBiasEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDepthBiasEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDepthBiasEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDepthBiasEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDepthBiasEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDepthBiasEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDepthBiasEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDepthBiasEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDepthBiasEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDepthBiasEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBiasEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDepthBiasEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDepthBiasEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBiasEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDepthBiasEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDepthBiasEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBiasEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDepthBiasEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDepthBiasEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBiasEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDepthBiasEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDepthBiasEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBiasEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDepthBiasEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDepthBiasEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBiasEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDepthBiasEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDepthBiasEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBiasEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDepthBiasEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDepthBiasEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBiasEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDepthBiasEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDepthBiasEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBiasEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDepthBiasEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDepthBiasEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBiasEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDepthBiasEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDepthBiasEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBiasEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDepthBiasEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDepthBiasEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBiasEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDepthBiasEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDepthBiasEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBiasEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDepthBiasEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDepthBiasEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetLogicOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetLogicOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetLogicOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetLogicOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetLogicOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetLogicOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetLogicOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetLogicOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetLogicOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetLogicOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetLogicOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetLogicOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetLogicOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetLogicOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetLogicOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetLogicOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetLogicOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetLogicOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetLogicOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetLogicOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetLogicOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetLogicOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetLogicOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetLogicOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetLogicOpEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetLogicOpEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetPrimitiveRestartEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetPrimitiveRestartEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetPrimitiveRestartEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetPrimitiveRestartEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetPrimitiveRestartEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetPrimitiveRestartEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetPrimitiveRestartEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetPrimitiveRestartEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetPrimitiveRestartEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetPrimitiveRestartEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetPrimitiveRestartEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetPrimitiveRestartEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetPrimitiveRestartEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetPrimitiveRestartEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetPrimitiveRestartEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetPrimitiveRestartEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetPrimitiveRestartEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetPrimitiveRestartEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetPrimitiveRestartEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetPrimitiveRestartEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetPrimitiveRestartEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetPrimitiveRestartEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetPrimitiveRestartEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetPrimitiveRestartEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnable@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnable@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnable@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnable@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnable@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnable@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnable@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnable@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnable@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnable@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnable@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnable@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnable@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnable@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetPrimitiveRestartEnable=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetPrimitiveRestartEnable=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPrimitiveRestartEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetPrimitiveRestartEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetPrimitiveRestartEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPrimitiveRestartEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetPrimitiveRestartEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetPrimitiveRestartEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPrimitiveRestartEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetPrimitiveRestartEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetPrimitiveRestartEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPrimitiveRestartEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetPrimitiveRestartEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetPrimitiveRestartEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPrimitiveRestartEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetPrimitiveRestartEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetPrimitiveRestartEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPrimitiveRestartEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetPrimitiveRestartEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetPrimitiveRestartEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPrimitiveRestartEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetPrimitiveRestartEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetPrimitiveRestartEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPrimitiveRestartEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetPrimitiveRestartEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetPrimitiveRestartEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPrimitiveRestartEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetPrimitiveRestartEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetPrimitiveRestartEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPrimitiveRestartEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetPrimitiveRestartEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetPrimitiveRestartEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPrimitiveRestartEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetPrimitiveRestartEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetPrimitiveRestartEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPrimitiveRestartEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetPrimitiveRestartEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetPrimitiveRestartEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPrimitiveRestartEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetPrimitiveRestartEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetPrimitiveRestartEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetTessellationDomainOriginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetTessellationDomainOriginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetTessellationDomainOriginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetTessellationDomainOriginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetTessellationDomainOriginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetTessellationDomainOriginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetTessellationDomainOriginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetTessellationDomainOriginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetTessellationDomainOriginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetTessellationDomainOriginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetTessellationDomainOriginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetTessellationDomainOriginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetTessellationDomainOriginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetTessellationDomainOriginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetTessellationDomainOriginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetTessellationDomainOriginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetTessellationDomainOriginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetTessellationDomainOriginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetTessellationDomainOriginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetTessellationDomainOriginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetTessellationDomainOriginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetTessellationDomainOriginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetTessellationDomainOriginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetTessellationDomainOriginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetTessellationDomainOriginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetTessellationDomainOriginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetTessellationDomainOriginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetTessellationDomainOriginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetTessellationDomainOriginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetTessellationDomainOriginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetTessellationDomainOriginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetTessellationDomainOriginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetTessellationDomainOriginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetTessellationDomainOriginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetTessellationDomainOriginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetTessellationDomainOriginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetTessellationDomainOriginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetTessellationDomainOriginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetTessellationDomainOriginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetTessellationDomainOriginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetTessellationDomainOriginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetTessellationDomainOriginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetTessellationDomainOriginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetTessellationDomainOriginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetTessellationDomainOriginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetTessellationDomainOriginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetTessellationDomainOriginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetTessellationDomainOriginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetTessellationDomainOriginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetTessellationDomainOriginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetTessellationDomainOriginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetTessellationDomainOriginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetTessellationDomainOriginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetTessellationDomainOriginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetTessellationDomainOriginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetTessellationDomainOriginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetTessellationDomainOriginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetTessellationDomainOriginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetTessellationDomainOriginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetTessellationDomainOriginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetTessellationDomainOriginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetTessellationDomainOriginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetTessellationDomainOriginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetTessellationDomainOriginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetTessellationDomainOriginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetTessellationDomainOriginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetTessellationDomainOriginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetTessellationDomainOriginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetTessellationDomainOriginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetTessellationDomainOriginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetTessellationDomainOriginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetTessellationDomainOriginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetTessellationDomainOriginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetTessellationDomainOriginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetTessellationDomainOriginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetTessellationDomainOriginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetTessellationDomainOriginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetTessellationDomainOriginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetTessellationDomainOriginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetTessellationDomainOriginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetTessellationDomainOriginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetTessellationDomainOriginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetTessellationDomainOriginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetTessellationDomainOriginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetTessellationDomainOriginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetTessellationDomainOriginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetTessellationDomainOriginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetTessellationDomainOriginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetTessellationDomainOriginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetTessellationDomainOriginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetTessellationDomainOriginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetTessellationDomainOriginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetTessellationDomainOriginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetTessellationDomainOriginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetTessellationDomainOriginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetTessellationDomainOriginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetTessellationDomainOriginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetTessellationDomainOriginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetTessellationDomainOriginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetTessellationDomainOriginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetTessellationDomainOriginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetTessellationDomainOriginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetTessellationDomainOriginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetTessellationDomainOriginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetTessellationDomainOriginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetTessellationDomainOriginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetTessellationDomainOriginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetTessellationDomainOriginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetTessellationDomainOriginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetTessellationDomainOriginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetTessellationDomainOriginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetTessellationDomainOriginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetTessellationDomainOriginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetTessellationDomainOriginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetTessellationDomainOriginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetTessellationDomainOriginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetTessellationDomainOriginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetTessellationDomainOriginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetTessellationDomainOriginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetTessellationDomainOriginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetTessellationDomainOriginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetTessellationDomainOriginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetTessellationDomainOriginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetTessellationDomainOriginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetTessellationDomainOriginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetTessellationDomainOriginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetTessellationDomainOriginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetTessellationDomainOriginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetTessellationDomainOriginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetTessellationDomainOriginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetTessellationDomainOriginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetTessellationDomainOriginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetTessellationDomainOriginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetTessellationDomainOriginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetTessellationDomainOriginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetTessellationDomainOriginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetTessellationDomainOriginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetTessellationDomainOriginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetTessellationDomainOriginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetTessellationDomainOriginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetTessellationDomainOriginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetTessellationDomainOriginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetTessellationDomainOriginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetTessellationDomainOriginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetTessellationDomainOriginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetTessellationDomainOriginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetTessellationDomainOriginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetTessellationDomainOriginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetTessellationDomainOriginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetTessellationDomainOriginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetTessellationDomainOriginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetTessellationDomainOriginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetTessellationDomainOriginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetTessellationDomainOriginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetTessellationDomainOriginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetTessellationDomainOriginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetTessellationDomainOriginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetTessellationDomainOriginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetTessellationDomainOriginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetTessellationDomainOriginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetTessellationDomainOriginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetTessellationDomainOriginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetTessellationDomainOriginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetTessellationDomainOriginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetTessellationDomainOriginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetTessellationDomainOriginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetTessellationDomainOriginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetTessellationDomainOriginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetTessellationDomainOriginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetTessellationDomainOriginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetTessellationDomainOriginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetTessellationDomainOriginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetTessellationDomainOriginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetTessellationDomainOriginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetTessellationDomainOriginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetTessellationDomainOriginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetTessellationDomainOriginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetTessellationDomainOriginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetTessellationDomainOriginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetTessellationDomainOriginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetTessellationDomainOriginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetTessellationDomainOriginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetTessellationDomainOriginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetTessellationDomainOriginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetTessellationDomainOriginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetTessellationDomainOriginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetTessellationDomainOriginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetTessellationDomainOriginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetTessellationDomainOriginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetTessellationDomainOriginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetTessellationDomainOriginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetTessellationDomainOriginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetTessellationDomainOriginEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetTessellationDomainOriginEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetTessellationDomainOriginEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetTessellationDomainOriginEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetTessellationDomainOriginEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetTessellationDomainOriginEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetTessellationDomainOriginEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetTessellationDomainOriginEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetTessellationDomainOriginEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetTessellationDomainOriginEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetTessellationDomainOriginEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetTessellationDomainOriginEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetTessellationDomainOriginEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetTessellationDomainOriginEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetTessellationDomainOriginEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetTessellationDomainOriginEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDepthClampEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDepthClampEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDepthClampEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDepthClampEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDepthClampEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDepthClampEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDepthClampEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDepthClampEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDepthClampEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDepthClampEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDepthClampEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDepthClampEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDepthClampEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDepthClampEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDepthClampEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDepthClampEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDepthClampEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDepthClampEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDepthClampEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDepthClampEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDepthClampEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDepthClampEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDepthClampEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDepthClampEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDepthClampEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDepthClampEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetPolygonModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPolygonModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPolygonModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPolygonModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPolygonModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPolygonModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPolygonModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPolygonModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPolygonModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPolygonModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPolygonModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPolygonModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPolygonModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetPolygonModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetPolygonModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetPolygonModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPolygonModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPolygonModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPolygonModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPolygonModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPolygonModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPolygonModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPolygonModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPolygonModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPolygonModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPolygonModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPolygonModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPolygonModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPolygonModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetPolygonModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetPolygonModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetPolygonModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPolygonModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPolygonModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPolygonModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPolygonModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPolygonModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPolygonModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPolygonModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPolygonModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPolygonModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPolygonModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPolygonModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPolygonModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPolygonModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetPolygonModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetPolygonModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetPolygonModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPolygonModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPolygonModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPolygonModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPolygonModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPolygonModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPolygonModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPolygonModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPolygonModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPolygonModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPolygonModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPolygonModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPolygonModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPolygonModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetPolygonModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetPolygonModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetPolygonModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPolygonModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPolygonModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPolygonModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPolygonModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPolygonModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPolygonModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPolygonModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPolygonModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPolygonModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPolygonModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPolygonModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPolygonModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPolygonModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetPolygonModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetPolygonModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetPolygonModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPolygonModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPolygonModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPolygonModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPolygonModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPolygonModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPolygonModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPolygonModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPolygonModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPolygonModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPolygonModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPolygonModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPolygonModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPolygonModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetPolygonModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetPolygonModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetPolygonModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPolygonModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPolygonModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPolygonModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPolygonModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPolygonModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPolygonModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPolygonModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPolygonModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPolygonModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPolygonModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPolygonModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPolygonModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPolygonModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetPolygonModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetPolygonModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetPolygonModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPolygonModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPolygonModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPolygonModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPolygonModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPolygonModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPolygonModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPolygonModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPolygonModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPolygonModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPolygonModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPolygonModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPolygonModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPolygonModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetPolygonModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetPolygonModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetPolygonModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPolygonModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPolygonModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPolygonModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPolygonModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPolygonModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPolygonModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPolygonModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPolygonModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPolygonModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPolygonModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPolygonModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPolygonModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPolygonModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetPolygonModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetPolygonModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetPolygonModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPolygonModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPolygonModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPolygonModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPolygonModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPolygonModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPolygonModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPolygonModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPolygonModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPolygonModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPolygonModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPolygonModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPolygonModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPolygonModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetPolygonModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetPolygonModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetPolygonModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPolygonModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPolygonModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPolygonModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPolygonModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPolygonModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPolygonModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPolygonModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPolygonModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPolygonModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPolygonModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPolygonModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPolygonModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPolygonModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetPolygonModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetPolygonModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetPolygonModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPolygonModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPolygonModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPolygonModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPolygonModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPolygonModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPolygonModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPolygonModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPolygonModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPolygonModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPolygonModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPolygonModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPolygonModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPolygonModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetPolygonModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetPolygonModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetPolygonModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPolygonModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPolygonModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPolygonModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPolygonModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPolygonModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPolygonModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPolygonModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPolygonModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPolygonModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPolygonModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPolygonModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPolygonModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPolygonModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetPolygonModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetPolygonModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetPolygonModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationSamplesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationSamplesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationSamplesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationSamplesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationSamplesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationSamplesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationSamplesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationSamplesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationSamplesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationSamplesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationSamplesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationSamplesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationSamplesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationSamplesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetRasterizationSamplesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetRasterizationSamplesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationSamplesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationSamplesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationSamplesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationSamplesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationSamplesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationSamplesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationSamplesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationSamplesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationSamplesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationSamplesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationSamplesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationSamplesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationSamplesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationSamplesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetRasterizationSamplesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetRasterizationSamplesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationSamplesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationSamplesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationSamplesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationSamplesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationSamplesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationSamplesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationSamplesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationSamplesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationSamplesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationSamplesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationSamplesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationSamplesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationSamplesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationSamplesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetRasterizationSamplesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetRasterizationSamplesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationSamplesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationSamplesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationSamplesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationSamplesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationSamplesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationSamplesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationSamplesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationSamplesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationSamplesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationSamplesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationSamplesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationSamplesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationSamplesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationSamplesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetRasterizationSamplesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetRasterizationSamplesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationSamplesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationSamplesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationSamplesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationSamplesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationSamplesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationSamplesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationSamplesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationSamplesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationSamplesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationSamplesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationSamplesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationSamplesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationSamplesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationSamplesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetRasterizationSamplesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetRasterizationSamplesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationSamplesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationSamplesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationSamplesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationSamplesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationSamplesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationSamplesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationSamplesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationSamplesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationSamplesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationSamplesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationSamplesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationSamplesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationSamplesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationSamplesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetRasterizationSamplesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetRasterizationSamplesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationSamplesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationSamplesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationSamplesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationSamplesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationSamplesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationSamplesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationSamplesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationSamplesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationSamplesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationSamplesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationSamplesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationSamplesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationSamplesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationSamplesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetRasterizationSamplesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetRasterizationSamplesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationSamplesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationSamplesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationSamplesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationSamplesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationSamplesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationSamplesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationSamplesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationSamplesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationSamplesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationSamplesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationSamplesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationSamplesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationSamplesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationSamplesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetRasterizationSamplesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetRasterizationSamplesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationSamplesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationSamplesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationSamplesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationSamplesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationSamplesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationSamplesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationSamplesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationSamplesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationSamplesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationSamplesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationSamplesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationSamplesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationSamplesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationSamplesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetRasterizationSamplesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetRasterizationSamplesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationSamplesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationSamplesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationSamplesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationSamplesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationSamplesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationSamplesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationSamplesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationSamplesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationSamplesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationSamplesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationSamplesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationSamplesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationSamplesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationSamplesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetRasterizationSamplesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetRasterizationSamplesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationSamplesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationSamplesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationSamplesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationSamplesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationSamplesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationSamplesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationSamplesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationSamplesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationSamplesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationSamplesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationSamplesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationSamplesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationSamplesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationSamplesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetRasterizationSamplesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetRasterizationSamplesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationSamplesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationSamplesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationSamplesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationSamplesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationSamplesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationSamplesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationSamplesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationSamplesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationSamplesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationSamplesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationSamplesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationSamplesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationSamplesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationSamplesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetRasterizationSamplesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetRasterizationSamplesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationSamplesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationSamplesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationSamplesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationSamplesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationSamplesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationSamplesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationSamplesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationSamplesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationSamplesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationSamplesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationSamplesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationSamplesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationSamplesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationSamplesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetRasterizationSamplesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetRasterizationSamplesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetSampleMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetSampleMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetSampleMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetSampleMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetSampleMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetSampleMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetSampleMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetSampleMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetSampleMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetSampleMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetSampleMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetSampleMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetSampleMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetSampleMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetSampleMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetSampleMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetSampleMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetSampleMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetSampleMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetSampleMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetSampleMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetSampleMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetSampleMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetSampleMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetSampleMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetSampleMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToCoverageEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToCoverageEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToCoverageEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToCoverageEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToCoverageEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToCoverageEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToCoverageEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToCoverageEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToCoverageEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToCoverageEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToCoverageEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToCoverageEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToCoverageEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToCoverageEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetAlphaToCoverageEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetAlphaToCoverageEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToCoverageEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToCoverageEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToCoverageEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToCoverageEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToCoverageEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToCoverageEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToCoverageEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToCoverageEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToCoverageEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToCoverageEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToCoverageEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToCoverageEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToCoverageEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToCoverageEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetAlphaToCoverageEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetAlphaToCoverageEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToCoverageEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToCoverageEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToCoverageEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToCoverageEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToCoverageEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToCoverageEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToCoverageEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToCoverageEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToCoverageEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToCoverageEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToCoverageEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToCoverageEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToCoverageEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToCoverageEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetAlphaToCoverageEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetAlphaToCoverageEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToCoverageEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToCoverageEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToCoverageEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToCoverageEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToCoverageEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToCoverageEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToCoverageEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToCoverageEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToCoverageEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToCoverageEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToCoverageEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToCoverageEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToCoverageEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToCoverageEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetAlphaToCoverageEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetAlphaToCoverageEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToCoverageEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToCoverageEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToCoverageEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToCoverageEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToCoverageEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToCoverageEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToCoverageEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToCoverageEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToCoverageEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToCoverageEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToCoverageEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToCoverageEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToCoverageEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToCoverageEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetAlphaToCoverageEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetAlphaToCoverageEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToCoverageEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToCoverageEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToCoverageEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToCoverageEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToCoverageEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToCoverageEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToCoverageEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToCoverageEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToCoverageEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToCoverageEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToCoverageEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToCoverageEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToCoverageEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToCoverageEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetAlphaToCoverageEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetAlphaToCoverageEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToCoverageEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToCoverageEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToCoverageEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToCoverageEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToCoverageEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToCoverageEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToCoverageEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToCoverageEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToCoverageEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToCoverageEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToCoverageEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToCoverageEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToCoverageEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToCoverageEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetAlphaToCoverageEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetAlphaToCoverageEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToCoverageEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToCoverageEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToCoverageEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToCoverageEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToCoverageEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToCoverageEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToCoverageEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToCoverageEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToCoverageEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToCoverageEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToCoverageEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToCoverageEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToCoverageEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToCoverageEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetAlphaToCoverageEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetAlphaToCoverageEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToCoverageEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToCoverageEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToCoverageEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToCoverageEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToCoverageEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToCoverageEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToCoverageEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToCoverageEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToCoverageEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToCoverageEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToCoverageEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToCoverageEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToCoverageEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToCoverageEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetAlphaToCoverageEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetAlphaToCoverageEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToCoverageEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToCoverageEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToCoverageEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToCoverageEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToCoverageEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToCoverageEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToCoverageEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToCoverageEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToCoverageEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToCoverageEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToCoverageEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToCoverageEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToCoverageEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToCoverageEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetAlphaToCoverageEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetAlphaToCoverageEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToCoverageEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToCoverageEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToCoverageEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToCoverageEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToCoverageEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToCoverageEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToCoverageEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToCoverageEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToCoverageEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToCoverageEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToCoverageEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToCoverageEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToCoverageEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToCoverageEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetAlphaToCoverageEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetAlphaToCoverageEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToCoverageEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToCoverageEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToCoverageEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToCoverageEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToCoverageEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToCoverageEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToCoverageEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToCoverageEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToCoverageEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToCoverageEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToCoverageEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToCoverageEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToCoverageEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToCoverageEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetAlphaToCoverageEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetAlphaToCoverageEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToCoverageEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToCoverageEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToCoverageEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToCoverageEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToCoverageEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToCoverageEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToCoverageEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToCoverageEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToCoverageEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToCoverageEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToCoverageEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToCoverageEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToCoverageEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToCoverageEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetAlphaToCoverageEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetAlphaToCoverageEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToOneEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToOneEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToOneEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToOneEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToOneEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToOneEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToOneEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToOneEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToOneEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToOneEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToOneEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToOneEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToOneEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetAlphaToOneEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetAlphaToOneEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetAlphaToOneEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToOneEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToOneEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToOneEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToOneEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToOneEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToOneEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToOneEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToOneEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToOneEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToOneEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToOneEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToOneEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToOneEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetAlphaToOneEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetAlphaToOneEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetAlphaToOneEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToOneEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToOneEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToOneEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToOneEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToOneEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToOneEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToOneEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToOneEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToOneEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToOneEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToOneEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToOneEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToOneEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetAlphaToOneEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetAlphaToOneEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetAlphaToOneEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToOneEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToOneEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToOneEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToOneEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToOneEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToOneEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToOneEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToOneEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToOneEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToOneEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToOneEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToOneEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToOneEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetAlphaToOneEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetAlphaToOneEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetAlphaToOneEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToOneEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToOneEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToOneEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToOneEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToOneEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToOneEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToOneEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToOneEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToOneEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToOneEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToOneEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToOneEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToOneEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetAlphaToOneEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetAlphaToOneEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetAlphaToOneEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToOneEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToOneEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToOneEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToOneEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToOneEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToOneEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToOneEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToOneEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToOneEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToOneEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToOneEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToOneEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToOneEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetAlphaToOneEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetAlphaToOneEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetAlphaToOneEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToOneEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToOneEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToOneEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToOneEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToOneEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToOneEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToOneEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToOneEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToOneEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToOneEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToOneEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToOneEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToOneEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetAlphaToOneEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetAlphaToOneEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetAlphaToOneEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToOneEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToOneEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToOneEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToOneEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToOneEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToOneEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToOneEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToOneEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToOneEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToOneEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToOneEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToOneEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToOneEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetAlphaToOneEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetAlphaToOneEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetAlphaToOneEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToOneEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToOneEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToOneEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToOneEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToOneEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToOneEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToOneEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToOneEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToOneEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToOneEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToOneEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToOneEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToOneEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetAlphaToOneEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetAlphaToOneEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetAlphaToOneEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToOneEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToOneEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToOneEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToOneEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToOneEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToOneEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToOneEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToOneEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToOneEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToOneEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToOneEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToOneEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToOneEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetAlphaToOneEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetAlphaToOneEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetAlphaToOneEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToOneEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToOneEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToOneEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToOneEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToOneEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToOneEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToOneEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToOneEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToOneEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToOneEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToOneEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToOneEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToOneEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetAlphaToOneEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetAlphaToOneEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetAlphaToOneEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToOneEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToOneEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToOneEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToOneEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToOneEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToOneEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToOneEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToOneEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToOneEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToOneEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToOneEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToOneEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToOneEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetAlphaToOneEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetAlphaToOneEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetAlphaToOneEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToOneEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToOneEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToOneEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToOneEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToOneEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToOneEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToOneEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToOneEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToOneEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToOneEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToOneEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToOneEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToOneEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetAlphaToOneEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetAlphaToOneEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetAlphaToOneEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLogicOpEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetLogicOpEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetLogicOpEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLogicOpEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetLogicOpEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetLogicOpEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLogicOpEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetLogicOpEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetLogicOpEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLogicOpEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetLogicOpEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetLogicOpEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLogicOpEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetLogicOpEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetLogicOpEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLogicOpEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetLogicOpEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetLogicOpEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLogicOpEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetLogicOpEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetLogicOpEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLogicOpEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetLogicOpEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetLogicOpEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLogicOpEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetLogicOpEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetLogicOpEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLogicOpEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetLogicOpEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetLogicOpEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLogicOpEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetLogicOpEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetLogicOpEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLogicOpEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetLogicOpEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetLogicOpEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLogicOpEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetLogicOpEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetLogicOpEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetColorBlendEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetColorBlendEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetColorBlendEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetColorBlendEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetColorBlendEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetColorBlendEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetColorBlendEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetColorBlendEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetColorBlendEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetColorBlendEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetColorBlendEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetColorBlendEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetColorBlendEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetColorBlendEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetColorBlendEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetColorBlendEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetColorBlendEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetColorBlendEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetColorBlendEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetColorBlendEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetColorBlendEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetColorBlendEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetColorBlendEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetColorBlendEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetColorBlendEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetColorBlendEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEquationEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEquationEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEquationEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEquationEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEquationEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEquationEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEquationEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEquationEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEquationEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEquationEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEquationEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEquationEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEquationEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendEquationEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetColorBlendEquationEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetColorBlendEquationEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEquationEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEquationEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEquationEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEquationEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEquationEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEquationEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEquationEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEquationEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEquationEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEquationEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEquationEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEquationEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEquationEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendEquationEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetColorBlendEquationEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetColorBlendEquationEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEquationEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEquationEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEquationEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEquationEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEquationEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEquationEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEquationEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEquationEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEquationEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEquationEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEquationEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEquationEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEquationEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendEquationEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetColorBlendEquationEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetColorBlendEquationEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEquationEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEquationEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEquationEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEquationEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEquationEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEquationEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEquationEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEquationEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEquationEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEquationEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEquationEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEquationEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEquationEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendEquationEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetColorBlendEquationEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetColorBlendEquationEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEquationEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEquationEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEquationEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEquationEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEquationEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEquationEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEquationEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEquationEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEquationEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEquationEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEquationEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEquationEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEquationEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendEquationEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetColorBlendEquationEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetColorBlendEquationEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEquationEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEquationEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEquationEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEquationEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEquationEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEquationEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEquationEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEquationEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEquationEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEquationEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEquationEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEquationEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEquationEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendEquationEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetColorBlendEquationEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetColorBlendEquationEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEquationEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEquationEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEquationEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEquationEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEquationEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEquationEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEquationEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEquationEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEquationEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEquationEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEquationEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEquationEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEquationEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendEquationEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetColorBlendEquationEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetColorBlendEquationEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEquationEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEquationEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEquationEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEquationEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEquationEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEquationEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEquationEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEquationEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEquationEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEquationEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEquationEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEquationEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEquationEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendEquationEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetColorBlendEquationEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetColorBlendEquationEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEquationEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEquationEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEquationEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEquationEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEquationEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEquationEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEquationEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEquationEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEquationEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEquationEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEquationEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEquationEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEquationEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendEquationEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetColorBlendEquationEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetColorBlendEquationEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEquationEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEquationEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEquationEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEquationEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEquationEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEquationEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEquationEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEquationEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEquationEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEquationEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEquationEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEquationEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEquationEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendEquationEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetColorBlendEquationEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetColorBlendEquationEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEquationEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEquationEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEquationEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEquationEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEquationEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEquationEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEquationEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEquationEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEquationEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEquationEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEquationEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEquationEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEquationEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendEquationEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetColorBlendEquationEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetColorBlendEquationEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEquationEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEquationEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEquationEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEquationEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEquationEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEquationEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEquationEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEquationEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEquationEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEquationEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEquationEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEquationEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEquationEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendEquationEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetColorBlendEquationEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetColorBlendEquationEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEquationEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEquationEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEquationEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEquationEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEquationEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEquationEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEquationEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEquationEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEquationEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEquationEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEquationEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEquationEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEquationEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendEquationEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetColorBlendEquationEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetColorBlendEquationEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetColorWriteMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetColorWriteMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetColorWriteMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetColorWriteMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetColorWriteMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetColorWriteMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetColorWriteMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetColorWriteMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetColorWriteMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetColorWriteMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetColorWriteMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetColorWriteMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetColorWriteMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetColorWriteMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetColorWriteMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetColorWriteMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetColorWriteMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetColorWriteMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetColorWriteMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetColorWriteMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetColorWriteMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetColorWriteMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetColorWriteMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetColorWriteMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteMaskEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteMaskEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteMaskEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteMaskEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteMaskEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteMaskEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteMaskEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteMaskEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteMaskEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteMaskEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteMaskEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteMaskEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteMaskEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteMaskEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetColorWriteMaskEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetColorWriteMaskEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationStreamEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationStreamEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationStreamEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationStreamEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationStreamEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationStreamEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationStreamEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationStreamEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationStreamEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationStreamEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationStreamEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationStreamEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationStreamEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRasterizationStreamEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetRasterizationStreamEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetRasterizationStreamEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationStreamEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationStreamEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationStreamEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationStreamEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationStreamEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationStreamEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationStreamEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationStreamEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationStreamEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationStreamEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationStreamEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationStreamEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationStreamEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRasterizationStreamEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetRasterizationStreamEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetRasterizationStreamEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationStreamEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationStreamEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationStreamEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationStreamEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationStreamEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationStreamEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationStreamEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationStreamEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationStreamEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationStreamEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationStreamEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationStreamEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationStreamEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRasterizationStreamEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetRasterizationStreamEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetRasterizationStreamEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationStreamEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationStreamEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationStreamEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationStreamEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationStreamEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationStreamEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationStreamEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationStreamEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationStreamEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationStreamEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationStreamEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationStreamEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationStreamEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRasterizationStreamEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetRasterizationStreamEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetRasterizationStreamEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationStreamEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationStreamEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationStreamEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationStreamEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationStreamEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationStreamEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationStreamEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationStreamEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationStreamEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationStreamEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationStreamEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationStreamEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationStreamEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRasterizationStreamEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetRasterizationStreamEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetRasterizationStreamEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationStreamEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationStreamEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationStreamEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationStreamEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationStreamEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationStreamEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationStreamEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationStreamEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationStreamEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationStreamEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationStreamEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationStreamEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationStreamEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRasterizationStreamEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetRasterizationStreamEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetRasterizationStreamEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationStreamEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationStreamEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationStreamEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationStreamEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationStreamEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationStreamEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationStreamEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationStreamEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationStreamEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationStreamEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationStreamEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationStreamEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationStreamEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRasterizationStreamEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetRasterizationStreamEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetRasterizationStreamEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationStreamEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationStreamEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationStreamEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationStreamEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationStreamEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationStreamEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationStreamEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationStreamEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationStreamEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationStreamEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationStreamEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationStreamEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationStreamEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRasterizationStreamEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetRasterizationStreamEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetRasterizationStreamEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationStreamEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationStreamEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationStreamEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationStreamEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationStreamEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationStreamEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationStreamEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationStreamEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationStreamEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationStreamEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationStreamEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationStreamEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationStreamEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRasterizationStreamEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetRasterizationStreamEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetRasterizationStreamEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationStreamEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationStreamEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationStreamEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationStreamEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationStreamEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationStreamEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationStreamEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationStreamEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationStreamEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationStreamEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationStreamEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationStreamEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationStreamEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRasterizationStreamEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetRasterizationStreamEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetRasterizationStreamEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationStreamEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationStreamEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationStreamEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationStreamEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationStreamEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationStreamEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationStreamEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationStreamEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationStreamEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationStreamEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationStreamEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationStreamEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationStreamEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRasterizationStreamEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetRasterizationStreamEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetRasterizationStreamEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationStreamEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationStreamEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationStreamEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationStreamEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationStreamEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationStreamEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationStreamEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationStreamEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationStreamEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationStreamEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationStreamEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationStreamEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationStreamEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRasterizationStreamEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetRasterizationStreamEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetRasterizationStreamEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationStreamEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationStreamEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationStreamEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationStreamEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationStreamEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationStreamEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationStreamEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationStreamEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationStreamEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationStreamEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationStreamEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationStreamEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationStreamEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRasterizationStreamEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetRasterizationStreamEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetRasterizationStreamEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetConservativeRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetConservativeRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetConservativeRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetConservativeRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetConservativeRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetConservativeRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetConservativeRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetConservativeRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetConservativeRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetConservativeRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetConservativeRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetConservativeRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetConservativeRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetConservativeRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetConservativeRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetConservativeRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetConservativeRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetConservativeRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetConservativeRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetConservativeRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetConservativeRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetConservativeRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetConservativeRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetConservativeRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetConservativeRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetConservativeRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetConservativeRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetConservativeRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetConservativeRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetConservativeRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetConservativeRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetConservativeRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetConservativeRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetConservativeRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetConservativeRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetConservativeRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetConservativeRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetConservativeRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetConservativeRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetConservativeRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetConservativeRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetConservativeRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetConservativeRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetConservativeRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetConservativeRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetConservativeRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetConservativeRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetConservativeRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetConservativeRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetConservativeRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetConservativeRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetConservativeRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetConservativeRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetConservativeRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetConservativeRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetConservativeRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetConservativeRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetConservativeRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetConservativeRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetConservativeRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetConservativeRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetConservativeRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetConservativeRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetConservativeRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetConservativeRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetConservativeRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetConservativeRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetConservativeRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetConservativeRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetConservativeRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetConservativeRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetConservativeRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetConservativeRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetConservativeRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetConservativeRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetConservativeRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetConservativeRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetConservativeRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetConservativeRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetConservativeRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetConservativeRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetConservativeRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetConservativeRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetConservativeRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetConservativeRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetConservativeRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetConservativeRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetConservativeRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetConservativeRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetConservativeRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetConservativeRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetConservativeRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetConservativeRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetConservativeRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetConservativeRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetConservativeRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetConservativeRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetConservativeRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetConservativeRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetConservativeRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetConservativeRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetConservativeRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetConservativeRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetConservativeRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetConservativeRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetConservativeRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetConservativeRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetConservativeRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetConservativeRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetConservativeRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetConservativeRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetConservativeRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetConservativeRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetConservativeRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetConservativeRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetConservativeRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetConservativeRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetConservativeRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetConservativeRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetConservativeRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetConservativeRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetConservativeRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetConservativeRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetConservativeRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetConservativeRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetConservativeRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetConservativeRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetConservativeRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetConservativeRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetConservativeRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetConservativeRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetConservativeRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetConservativeRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetConservativeRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetConservativeRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetConservativeRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetConservativeRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetConservativeRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetConservativeRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetConservativeRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetConservativeRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetConservativeRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetConservativeRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetConservativeRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetConservativeRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetConservativeRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetConservativeRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetConservativeRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetConservativeRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetConservativeRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetConservativeRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetConservativeRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetConservativeRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetConservativeRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetConservativeRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetConservativeRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetConservativeRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetConservativeRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetConservativeRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetConservativeRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetConservativeRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetConservativeRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetConservativeRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetConservativeRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetConservativeRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetConservativeRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetConservativeRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetConservativeRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetConservativeRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetConservativeRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetConservativeRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetConservativeRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetConservativeRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetConservativeRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetConservativeRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetConservativeRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetConservativeRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetConservativeRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetConservativeRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetConservativeRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetConservativeRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetConservativeRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetConservativeRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetConservativeRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetConservativeRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetConservativeRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetConservativeRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetConservativeRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetConservativeRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetConservativeRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetConservativeRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetConservativeRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetConservativeRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetConservativeRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetConservativeRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetConservativeRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetConservativeRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetConservativeRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetConservativeRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetConservativeRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetConservativeRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetConservativeRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetConservativeRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetConservativeRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetConservativeRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetConservativeRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetConservativeRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetConservativeRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetExtraPrimitiveOverestimationSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExtraPrimitiveOverestimationSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExtraPrimitiveOverestimationSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExtraPrimitiveOverestimationSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExtraPrimitiveOverestimationSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExtraPrimitiveOverestimationSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExtraPrimitiveOverestimationSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExtraPrimitiveOverestimationSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExtraPrimitiveOverestimationSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExtraPrimitiveOverestimationSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExtraPrimitiveOverestimationSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExtraPrimitiveOverestimationSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExtraPrimitiveOverestimationSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetExtraPrimitiveOverestimationSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetExtraPrimitiveOverestimationSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetExtraPrimitiveOverestimationSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExtraPrimitiveOverestimationSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExtraPrimitiveOverestimationSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExtraPrimitiveOverestimationSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExtraPrimitiveOverestimationSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExtraPrimitiveOverestimationSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExtraPrimitiveOverestimationSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExtraPrimitiveOverestimationSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExtraPrimitiveOverestimationSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExtraPrimitiveOverestimationSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExtraPrimitiveOverestimationSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExtraPrimitiveOverestimationSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExtraPrimitiveOverestimationSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExtraPrimitiveOverestimationSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetExtraPrimitiveOverestimationSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetExtraPrimitiveOverestimationSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetExtraPrimitiveOverestimationSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExtraPrimitiveOverestimationSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExtraPrimitiveOverestimationSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExtraPrimitiveOverestimationSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExtraPrimitiveOverestimationSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExtraPrimitiveOverestimationSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExtraPrimitiveOverestimationSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExtraPrimitiveOverestimationSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExtraPrimitiveOverestimationSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExtraPrimitiveOverestimationSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExtraPrimitiveOverestimationSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExtraPrimitiveOverestimationSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExtraPrimitiveOverestimationSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExtraPrimitiveOverestimationSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetExtraPrimitiveOverestimationSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetExtraPrimitiveOverestimationSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetExtraPrimitiveOverestimationSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExtraPrimitiveOverestimationSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExtraPrimitiveOverestimationSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExtraPrimitiveOverestimationSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExtraPrimitiveOverestimationSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExtraPrimitiveOverestimationSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExtraPrimitiveOverestimationSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExtraPrimitiveOverestimationSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExtraPrimitiveOverestimationSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExtraPrimitiveOverestimationSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExtraPrimitiveOverestimationSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExtraPrimitiveOverestimationSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExtraPrimitiveOverestimationSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExtraPrimitiveOverestimationSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetExtraPrimitiveOverestimationSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetExtraPrimitiveOverestimationSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetExtraPrimitiveOverestimationSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExtraPrimitiveOverestimationSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExtraPrimitiveOverestimationSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExtraPrimitiveOverestimationSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExtraPrimitiveOverestimationSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExtraPrimitiveOverestimationSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExtraPrimitiveOverestimationSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExtraPrimitiveOverestimationSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExtraPrimitiveOverestimationSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExtraPrimitiveOverestimationSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExtraPrimitiveOverestimationSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExtraPrimitiveOverestimationSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExtraPrimitiveOverestimationSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExtraPrimitiveOverestimationSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetExtraPrimitiveOverestimationSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetExtraPrimitiveOverestimationSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetExtraPrimitiveOverestimationSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExtraPrimitiveOverestimationSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExtraPrimitiveOverestimationSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExtraPrimitiveOverestimationSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExtraPrimitiveOverestimationSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExtraPrimitiveOverestimationSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExtraPrimitiveOverestimationSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExtraPrimitiveOverestimationSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExtraPrimitiveOverestimationSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExtraPrimitiveOverestimationSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExtraPrimitiveOverestimationSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExtraPrimitiveOverestimationSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExtraPrimitiveOverestimationSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExtraPrimitiveOverestimationSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetExtraPrimitiveOverestimationSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetExtraPrimitiveOverestimationSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetExtraPrimitiveOverestimationSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExtraPrimitiveOverestimationSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExtraPrimitiveOverestimationSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExtraPrimitiveOverestimationSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExtraPrimitiveOverestimationSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExtraPrimitiveOverestimationSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExtraPrimitiveOverestimationSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExtraPrimitiveOverestimationSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExtraPrimitiveOverestimationSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExtraPrimitiveOverestimationSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExtraPrimitiveOverestimationSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExtraPrimitiveOverestimationSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExtraPrimitiveOverestimationSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExtraPrimitiveOverestimationSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetExtraPrimitiveOverestimationSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetExtraPrimitiveOverestimationSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetExtraPrimitiveOverestimationSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExtraPrimitiveOverestimationSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExtraPrimitiveOverestimationSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExtraPrimitiveOverestimationSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExtraPrimitiveOverestimationSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExtraPrimitiveOverestimationSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExtraPrimitiveOverestimationSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExtraPrimitiveOverestimationSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExtraPrimitiveOverestimationSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExtraPrimitiveOverestimationSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExtraPrimitiveOverestimationSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExtraPrimitiveOverestimationSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExtraPrimitiveOverestimationSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExtraPrimitiveOverestimationSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetExtraPrimitiveOverestimationSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetExtraPrimitiveOverestimationSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetExtraPrimitiveOverestimationSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExtraPrimitiveOverestimationSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExtraPrimitiveOverestimationSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExtraPrimitiveOverestimationSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExtraPrimitiveOverestimationSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExtraPrimitiveOverestimationSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExtraPrimitiveOverestimationSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExtraPrimitiveOverestimationSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExtraPrimitiveOverestimationSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExtraPrimitiveOverestimationSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExtraPrimitiveOverestimationSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExtraPrimitiveOverestimationSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExtraPrimitiveOverestimationSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExtraPrimitiveOverestimationSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetExtraPrimitiveOverestimationSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetExtraPrimitiveOverestimationSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetExtraPrimitiveOverestimationSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExtraPrimitiveOverestimationSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExtraPrimitiveOverestimationSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExtraPrimitiveOverestimationSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExtraPrimitiveOverestimationSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExtraPrimitiveOverestimationSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExtraPrimitiveOverestimationSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExtraPrimitiveOverestimationSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExtraPrimitiveOverestimationSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExtraPrimitiveOverestimationSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExtraPrimitiveOverestimationSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExtraPrimitiveOverestimationSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExtraPrimitiveOverestimationSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExtraPrimitiveOverestimationSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetExtraPrimitiveOverestimationSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetExtraPrimitiveOverestimationSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetExtraPrimitiveOverestimationSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExtraPrimitiveOverestimationSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExtraPrimitiveOverestimationSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExtraPrimitiveOverestimationSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExtraPrimitiveOverestimationSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExtraPrimitiveOverestimationSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExtraPrimitiveOverestimationSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExtraPrimitiveOverestimationSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExtraPrimitiveOverestimationSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExtraPrimitiveOverestimationSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExtraPrimitiveOverestimationSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExtraPrimitiveOverestimationSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExtraPrimitiveOverestimationSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExtraPrimitiveOverestimationSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetExtraPrimitiveOverestimationSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetExtraPrimitiveOverestimationSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetExtraPrimitiveOverestimationSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExtraPrimitiveOverestimationSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExtraPrimitiveOverestimationSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExtraPrimitiveOverestimationSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExtraPrimitiveOverestimationSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExtraPrimitiveOverestimationSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExtraPrimitiveOverestimationSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExtraPrimitiveOverestimationSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExtraPrimitiveOverestimationSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExtraPrimitiveOverestimationSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExtraPrimitiveOverestimationSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExtraPrimitiveOverestimationSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExtraPrimitiveOverestimationSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExtraPrimitiveOverestimationSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetExtraPrimitiveOverestimationSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetExtraPrimitiveOverestimationSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetExtraPrimitiveOverestimationSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExtraPrimitiveOverestimationSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExtraPrimitiveOverestimationSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExtraPrimitiveOverestimationSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExtraPrimitiveOverestimationSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExtraPrimitiveOverestimationSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExtraPrimitiveOverestimationSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExtraPrimitiveOverestimationSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExtraPrimitiveOverestimationSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExtraPrimitiveOverestimationSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExtraPrimitiveOverestimationSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExtraPrimitiveOverestimationSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExtraPrimitiveOverestimationSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExtraPrimitiveOverestimationSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetExtraPrimitiveOverestimationSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetExtraPrimitiveOverestimationSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetExtraPrimitiveOverestimationSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDepthClipEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDepthClipEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDepthClipEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDepthClipEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDepthClipEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDepthClipEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDepthClipEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDepthClipEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDepthClipEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDepthClipEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDepthClipEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDepthClipEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDepthClipEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDepthClipEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDepthClipEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDepthClipEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDepthClipEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDepthClipEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDepthClipEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDepthClipEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDepthClipEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDepthClipEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDepthClipEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDepthClipEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDepthClipEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDepthClipEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetSampleLocationsEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetSampleLocationsEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetSampleLocationsEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetSampleLocationsEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetSampleLocationsEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetSampleLocationsEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetSampleLocationsEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetSampleLocationsEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetSampleLocationsEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetSampleLocationsEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetSampleLocationsEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetSampleLocationsEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetSampleLocationsEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetSampleLocationsEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetSampleLocationsEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetSampleLocationsEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetSampleLocationsEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetSampleLocationsEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetSampleLocationsEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetSampleLocationsEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetSampleLocationsEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetSampleLocationsEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetSampleLocationsEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetSampleLocationsEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetSampleLocationsEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetSampleLocationsEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetSampleLocationsEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetSampleLocationsEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetSampleLocationsEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetSampleLocationsEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetSampleLocationsEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetSampleLocationsEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetSampleLocationsEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetSampleLocationsEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetSampleLocationsEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetSampleLocationsEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetSampleLocationsEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetSampleLocationsEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetSampleLocationsEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendAdvancedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendAdvancedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendAdvancedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendAdvancedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendAdvancedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendAdvancedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendAdvancedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendAdvancedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendAdvancedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendAdvancedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendAdvancedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendAdvancedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendAdvancedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorBlendAdvancedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetColorBlendAdvancedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetColorBlendAdvancedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendAdvancedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendAdvancedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendAdvancedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendAdvancedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendAdvancedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendAdvancedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendAdvancedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendAdvancedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendAdvancedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendAdvancedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendAdvancedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendAdvancedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendAdvancedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorBlendAdvancedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetColorBlendAdvancedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetColorBlendAdvancedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendAdvancedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendAdvancedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendAdvancedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendAdvancedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendAdvancedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendAdvancedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendAdvancedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendAdvancedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendAdvancedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendAdvancedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendAdvancedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendAdvancedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendAdvancedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorBlendAdvancedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetColorBlendAdvancedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetColorBlendAdvancedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendAdvancedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendAdvancedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendAdvancedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendAdvancedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendAdvancedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendAdvancedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendAdvancedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendAdvancedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendAdvancedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendAdvancedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendAdvancedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendAdvancedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendAdvancedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorBlendAdvancedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetColorBlendAdvancedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetColorBlendAdvancedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendAdvancedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendAdvancedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendAdvancedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendAdvancedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendAdvancedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendAdvancedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendAdvancedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendAdvancedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendAdvancedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendAdvancedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendAdvancedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendAdvancedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendAdvancedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorBlendAdvancedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetColorBlendAdvancedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetColorBlendAdvancedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendAdvancedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendAdvancedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendAdvancedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendAdvancedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendAdvancedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendAdvancedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendAdvancedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendAdvancedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendAdvancedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendAdvancedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendAdvancedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendAdvancedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendAdvancedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorBlendAdvancedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetColorBlendAdvancedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetColorBlendAdvancedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendAdvancedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendAdvancedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendAdvancedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendAdvancedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendAdvancedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendAdvancedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendAdvancedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendAdvancedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendAdvancedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendAdvancedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendAdvancedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendAdvancedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendAdvancedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorBlendAdvancedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetColorBlendAdvancedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetColorBlendAdvancedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendAdvancedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendAdvancedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendAdvancedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendAdvancedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendAdvancedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendAdvancedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendAdvancedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendAdvancedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendAdvancedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendAdvancedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendAdvancedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendAdvancedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendAdvancedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorBlendAdvancedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetColorBlendAdvancedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetColorBlendAdvancedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendAdvancedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendAdvancedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendAdvancedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendAdvancedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendAdvancedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendAdvancedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendAdvancedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendAdvancedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendAdvancedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendAdvancedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendAdvancedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendAdvancedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendAdvancedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorBlendAdvancedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetColorBlendAdvancedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetColorBlendAdvancedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendAdvancedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendAdvancedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendAdvancedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendAdvancedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendAdvancedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendAdvancedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendAdvancedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendAdvancedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendAdvancedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendAdvancedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendAdvancedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendAdvancedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendAdvancedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorBlendAdvancedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetColorBlendAdvancedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetColorBlendAdvancedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendAdvancedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendAdvancedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendAdvancedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendAdvancedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendAdvancedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendAdvancedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendAdvancedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendAdvancedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendAdvancedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendAdvancedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendAdvancedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendAdvancedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendAdvancedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorBlendAdvancedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetColorBlendAdvancedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetColorBlendAdvancedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendAdvancedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendAdvancedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendAdvancedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendAdvancedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendAdvancedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendAdvancedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendAdvancedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendAdvancedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendAdvancedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendAdvancedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendAdvancedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendAdvancedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendAdvancedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorBlendAdvancedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetColorBlendAdvancedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetColorBlendAdvancedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendAdvancedEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendAdvancedEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendAdvancedEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendAdvancedEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendAdvancedEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendAdvancedEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendAdvancedEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendAdvancedEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendAdvancedEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendAdvancedEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendAdvancedEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendAdvancedEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendAdvancedEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorBlendAdvancedEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetColorBlendAdvancedEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetColorBlendAdvancedEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetProvokingVertexModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetProvokingVertexModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetProvokingVertexModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetProvokingVertexModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetProvokingVertexModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetProvokingVertexModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetProvokingVertexModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetProvokingVertexModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetProvokingVertexModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetProvokingVertexModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetProvokingVertexModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetProvokingVertexModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetProvokingVertexModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetProvokingVertexModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetProvokingVertexModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetProvokingVertexModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetProvokingVertexModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetProvokingVertexModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetProvokingVertexModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetProvokingVertexModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetProvokingVertexModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetProvokingVertexModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetProvokingVertexModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetProvokingVertexModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetProvokingVertexModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetProvokingVertexModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetProvokingVertexModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetProvokingVertexModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetProvokingVertexModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetProvokingVertexModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetProvokingVertexModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetProvokingVertexModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetProvokingVertexModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetProvokingVertexModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetProvokingVertexModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetProvokingVertexModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetProvokingVertexModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetProvokingVertexModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetProvokingVertexModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetProvokingVertexModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetProvokingVertexModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetProvokingVertexModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetProvokingVertexModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetProvokingVertexModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetProvokingVertexModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetProvokingVertexModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetProvokingVertexModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetProvokingVertexModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetProvokingVertexModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetProvokingVertexModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetProvokingVertexModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetProvokingVertexModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetProvokingVertexModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetProvokingVertexModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetProvokingVertexModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetProvokingVertexModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetProvokingVertexModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetProvokingVertexModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetProvokingVertexModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetProvokingVertexModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetProvokingVertexModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetProvokingVertexModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetProvokingVertexModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetProvokingVertexModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetProvokingVertexModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetProvokingVertexModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetProvokingVertexModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetProvokingVertexModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetProvokingVertexModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetProvokingVertexModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetProvokingVertexModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetProvokingVertexModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetProvokingVertexModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetProvokingVertexModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetProvokingVertexModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetProvokingVertexModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetProvokingVertexModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetProvokingVertexModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetProvokingVertexModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetProvokingVertexModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetProvokingVertexModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetProvokingVertexModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetProvokingVertexModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetProvokingVertexModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetProvokingVertexModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetProvokingVertexModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetProvokingVertexModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetProvokingVertexModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetProvokingVertexModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetProvokingVertexModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetProvokingVertexModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetProvokingVertexModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetProvokingVertexModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetProvokingVertexModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetProvokingVertexModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetProvokingVertexModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetProvokingVertexModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetProvokingVertexModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetProvokingVertexModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetProvokingVertexModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetProvokingVertexModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetProvokingVertexModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetProvokingVertexModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetProvokingVertexModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetProvokingVertexModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetProvokingVertexModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetProvokingVertexModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetProvokingVertexModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetProvokingVertexModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetProvokingVertexModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetProvokingVertexModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetProvokingVertexModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetProvokingVertexModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetProvokingVertexModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetProvokingVertexModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetProvokingVertexModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetProvokingVertexModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetProvokingVertexModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetProvokingVertexModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetProvokingVertexModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetProvokingVertexModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetProvokingVertexModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetProvokingVertexModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetProvokingVertexModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetProvokingVertexModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetProvokingVertexModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetProvokingVertexModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetProvokingVertexModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetProvokingVertexModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetProvokingVertexModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetProvokingVertexModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetProvokingVertexModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetProvokingVertexModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetProvokingVertexModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetProvokingVertexModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetProvokingVertexModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetProvokingVertexModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetProvokingVertexModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetProvokingVertexModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetProvokingVertexModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetProvokingVertexModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetProvokingVertexModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetProvokingVertexModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetProvokingVertexModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetProvokingVertexModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetProvokingVertexModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetProvokingVertexModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetProvokingVertexModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetProvokingVertexModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetProvokingVertexModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetProvokingVertexModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetProvokingVertexModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetProvokingVertexModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetProvokingVertexModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetProvokingVertexModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetProvokingVertexModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetProvokingVertexModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetProvokingVertexModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetProvokingVertexModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetProvokingVertexModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetProvokingVertexModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetProvokingVertexModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetProvokingVertexModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetProvokingVertexModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetProvokingVertexModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetProvokingVertexModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetProvokingVertexModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetProvokingVertexModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetProvokingVertexModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetProvokingVertexModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetProvokingVertexModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetProvokingVertexModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetProvokingVertexModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetProvokingVertexModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetProvokingVertexModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetProvokingVertexModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetProvokingVertexModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetProvokingVertexModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetProvokingVertexModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetProvokingVertexModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetProvokingVertexModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetProvokingVertexModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetProvokingVertexModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetProvokingVertexModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetProvokingVertexModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetProvokingVertexModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetProvokingVertexModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetProvokingVertexModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetProvokingVertexModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetProvokingVertexModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetProvokingVertexModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetProvokingVertexModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetProvokingVertexModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetProvokingVertexModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetProvokingVertexModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetProvokingVertexModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetProvokingVertexModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetProvokingVertexModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetProvokingVertexModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetProvokingVertexModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetProvokingVertexModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetProvokingVertexModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetProvokingVertexModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetProvokingVertexModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetProvokingVertexModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetProvokingVertexModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetProvokingVertexModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetProvokingVertexModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetLineRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetLineRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetLineRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetLineRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetLineRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetLineRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetLineRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetLineRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetLineRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetLineRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetLineRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetLineRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetLineRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetLineRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetLineRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetLineRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetLineRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetLineRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetLineRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetLineRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetLineRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetLineRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetLineRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetLineRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineRasterizationModeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineRasterizationModeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineRasterizationModeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineRasterizationModeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineRasterizationModeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineRasterizationModeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineRasterizationModeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineRasterizationModeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineRasterizationModeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineRasterizationModeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineRasterizationModeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineRasterizationModeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineRasterizationModeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineRasterizationModeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetLineRasterizationModeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetLineRasterizationModeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetLineStippleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetLineStippleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetLineStippleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetLineStippleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetLineStippleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetLineStippleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetLineStippleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetLineStippleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetLineStippleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetLineStippleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetLineStippleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetLineStippleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetLineStippleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetLineStippleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetLineStippleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetLineStippleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetLineStippleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetLineStippleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetLineStippleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetLineStippleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetLineStippleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetLineStippleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetLineStippleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetLineStippleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetLineStippleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetLineStippleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetLineStippleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetLineStippleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetLineStippleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetLineStippleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetLineStippleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetLineStippleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetLineStippleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetLineStippleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetLineStippleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetLineStippleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetLineStippleEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetLineStippleEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetLineStippleEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipNegativeOneToOneEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipNegativeOneToOneEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipNegativeOneToOneEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipNegativeOneToOneEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipNegativeOneToOneEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipNegativeOneToOneEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipNegativeOneToOneEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipNegativeOneToOneEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipNegativeOneToOneEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipNegativeOneToOneEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipNegativeOneToOneEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipNegativeOneToOneEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipNegativeOneToOneEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClipNegativeOneToOneEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDepthClipNegativeOneToOneEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDepthClipNegativeOneToOneEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipNegativeOneToOneEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipNegativeOneToOneEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipNegativeOneToOneEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipNegativeOneToOneEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipNegativeOneToOneEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipNegativeOneToOneEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipNegativeOneToOneEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipNegativeOneToOneEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipNegativeOneToOneEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipNegativeOneToOneEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipNegativeOneToOneEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipNegativeOneToOneEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipNegativeOneToOneEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClipNegativeOneToOneEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDepthClipNegativeOneToOneEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDepthClipNegativeOneToOneEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipNegativeOneToOneEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipNegativeOneToOneEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipNegativeOneToOneEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipNegativeOneToOneEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipNegativeOneToOneEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipNegativeOneToOneEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipNegativeOneToOneEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipNegativeOneToOneEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipNegativeOneToOneEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipNegativeOneToOneEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipNegativeOneToOneEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipNegativeOneToOneEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipNegativeOneToOneEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClipNegativeOneToOneEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDepthClipNegativeOneToOneEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDepthClipNegativeOneToOneEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipNegativeOneToOneEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipNegativeOneToOneEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipNegativeOneToOneEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipNegativeOneToOneEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipNegativeOneToOneEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipNegativeOneToOneEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipNegativeOneToOneEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipNegativeOneToOneEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipNegativeOneToOneEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipNegativeOneToOneEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipNegativeOneToOneEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipNegativeOneToOneEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipNegativeOneToOneEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClipNegativeOneToOneEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDepthClipNegativeOneToOneEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDepthClipNegativeOneToOneEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipNegativeOneToOneEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipNegativeOneToOneEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipNegativeOneToOneEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipNegativeOneToOneEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipNegativeOneToOneEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipNegativeOneToOneEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipNegativeOneToOneEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipNegativeOneToOneEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipNegativeOneToOneEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipNegativeOneToOneEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipNegativeOneToOneEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipNegativeOneToOneEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipNegativeOneToOneEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClipNegativeOneToOneEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDepthClipNegativeOneToOneEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDepthClipNegativeOneToOneEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipNegativeOneToOneEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipNegativeOneToOneEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipNegativeOneToOneEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipNegativeOneToOneEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipNegativeOneToOneEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipNegativeOneToOneEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipNegativeOneToOneEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipNegativeOneToOneEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipNegativeOneToOneEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipNegativeOneToOneEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipNegativeOneToOneEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipNegativeOneToOneEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipNegativeOneToOneEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClipNegativeOneToOneEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDepthClipNegativeOneToOneEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDepthClipNegativeOneToOneEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipNegativeOneToOneEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipNegativeOneToOneEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipNegativeOneToOneEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipNegativeOneToOneEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipNegativeOneToOneEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipNegativeOneToOneEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipNegativeOneToOneEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipNegativeOneToOneEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipNegativeOneToOneEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipNegativeOneToOneEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipNegativeOneToOneEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipNegativeOneToOneEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipNegativeOneToOneEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClipNegativeOneToOneEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDepthClipNegativeOneToOneEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDepthClipNegativeOneToOneEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipNegativeOneToOneEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipNegativeOneToOneEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipNegativeOneToOneEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipNegativeOneToOneEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipNegativeOneToOneEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipNegativeOneToOneEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipNegativeOneToOneEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipNegativeOneToOneEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipNegativeOneToOneEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipNegativeOneToOneEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipNegativeOneToOneEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipNegativeOneToOneEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipNegativeOneToOneEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClipNegativeOneToOneEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDepthClipNegativeOneToOneEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDepthClipNegativeOneToOneEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipNegativeOneToOneEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipNegativeOneToOneEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipNegativeOneToOneEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipNegativeOneToOneEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipNegativeOneToOneEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipNegativeOneToOneEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipNegativeOneToOneEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipNegativeOneToOneEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipNegativeOneToOneEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipNegativeOneToOneEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipNegativeOneToOneEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipNegativeOneToOneEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipNegativeOneToOneEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClipNegativeOneToOneEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDepthClipNegativeOneToOneEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDepthClipNegativeOneToOneEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipNegativeOneToOneEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipNegativeOneToOneEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipNegativeOneToOneEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipNegativeOneToOneEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipNegativeOneToOneEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipNegativeOneToOneEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipNegativeOneToOneEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipNegativeOneToOneEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipNegativeOneToOneEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipNegativeOneToOneEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipNegativeOneToOneEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipNegativeOneToOneEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipNegativeOneToOneEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClipNegativeOneToOneEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDepthClipNegativeOneToOneEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDepthClipNegativeOneToOneEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipNegativeOneToOneEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipNegativeOneToOneEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipNegativeOneToOneEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipNegativeOneToOneEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipNegativeOneToOneEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipNegativeOneToOneEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipNegativeOneToOneEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipNegativeOneToOneEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipNegativeOneToOneEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipNegativeOneToOneEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipNegativeOneToOneEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipNegativeOneToOneEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipNegativeOneToOneEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClipNegativeOneToOneEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDepthClipNegativeOneToOneEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDepthClipNegativeOneToOneEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipNegativeOneToOneEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipNegativeOneToOneEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipNegativeOneToOneEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipNegativeOneToOneEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipNegativeOneToOneEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipNegativeOneToOneEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipNegativeOneToOneEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipNegativeOneToOneEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipNegativeOneToOneEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipNegativeOneToOneEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipNegativeOneToOneEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipNegativeOneToOneEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipNegativeOneToOneEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClipNegativeOneToOneEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDepthClipNegativeOneToOneEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDepthClipNegativeOneToOneEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipNegativeOneToOneEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipNegativeOneToOneEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipNegativeOneToOneEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipNegativeOneToOneEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipNegativeOneToOneEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipNegativeOneToOneEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipNegativeOneToOneEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipNegativeOneToOneEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipNegativeOneToOneEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipNegativeOneToOneEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipNegativeOneToOneEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipNegativeOneToOneEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipNegativeOneToOneEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClipNegativeOneToOneEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDepthClipNegativeOneToOneEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDepthClipNegativeOneToOneEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportWScalingEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetViewportWScalingEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetViewportWScalingEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportWScalingEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetViewportWScalingEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetViewportWScalingEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportWScalingEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetViewportWScalingEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetViewportWScalingEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportWScalingEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetViewportWScalingEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetViewportWScalingEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportWScalingEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetViewportWScalingEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetViewportWScalingEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportWScalingEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetViewportWScalingEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetViewportWScalingEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportWScalingEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetViewportWScalingEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetViewportWScalingEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportWScalingEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetViewportWScalingEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetViewportWScalingEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportWScalingEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetViewportWScalingEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetViewportWScalingEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportWScalingEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetViewportWScalingEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetViewportWScalingEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportWScalingEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetViewportWScalingEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetViewportWScalingEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportWScalingEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetViewportWScalingEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetViewportWScalingEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportWScalingEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetViewportWScalingEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetViewportWScalingEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportSwizzleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportSwizzleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportSwizzleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportSwizzleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportSwizzleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportSwizzleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportSwizzleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportSwizzleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportSwizzleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportSwizzleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportSwizzleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportSwizzleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportSwizzleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetViewportSwizzleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetViewportSwizzleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetViewportSwizzleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportSwizzleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportSwizzleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportSwizzleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportSwizzleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportSwizzleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportSwizzleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportSwizzleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportSwizzleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportSwizzleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportSwizzleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportSwizzleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportSwizzleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportSwizzleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetViewportSwizzleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetViewportSwizzleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetViewportSwizzleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportSwizzleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportSwizzleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportSwizzleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportSwizzleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportSwizzleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportSwizzleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportSwizzleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportSwizzleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportSwizzleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportSwizzleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportSwizzleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportSwizzleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportSwizzleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetViewportSwizzleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetViewportSwizzleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetViewportSwizzleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportSwizzleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportSwizzleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportSwizzleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportSwizzleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportSwizzleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportSwizzleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportSwizzleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportSwizzleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportSwizzleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportSwizzleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportSwizzleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportSwizzleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportSwizzleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetViewportSwizzleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetViewportSwizzleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetViewportSwizzleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportSwizzleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportSwizzleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportSwizzleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportSwizzleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportSwizzleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportSwizzleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportSwizzleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportSwizzleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportSwizzleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportSwizzleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportSwizzleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportSwizzleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportSwizzleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetViewportSwizzleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetViewportSwizzleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetViewportSwizzleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportSwizzleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportSwizzleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportSwizzleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportSwizzleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportSwizzleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportSwizzleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportSwizzleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportSwizzleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportSwizzleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportSwizzleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportSwizzleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportSwizzleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportSwizzleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetViewportSwizzleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetViewportSwizzleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetViewportSwizzleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportSwizzleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportSwizzleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportSwizzleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportSwizzleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportSwizzleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportSwizzleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportSwizzleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportSwizzleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportSwizzleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportSwizzleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportSwizzleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportSwizzleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportSwizzleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetViewportSwizzleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetViewportSwizzleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetViewportSwizzleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportSwizzleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportSwizzleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportSwizzleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportSwizzleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportSwizzleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportSwizzleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportSwizzleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportSwizzleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportSwizzleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportSwizzleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportSwizzleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportSwizzleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportSwizzleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetViewportSwizzleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetViewportSwizzleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetViewportSwizzleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportSwizzleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportSwizzleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportSwizzleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportSwizzleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportSwizzleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportSwizzleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportSwizzleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportSwizzleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportSwizzleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportSwizzleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportSwizzleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportSwizzleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportSwizzleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetViewportSwizzleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetViewportSwizzleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetViewportSwizzleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportSwizzleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportSwizzleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportSwizzleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportSwizzleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportSwizzleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportSwizzleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportSwizzleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportSwizzleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportSwizzleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportSwizzleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportSwizzleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportSwizzleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportSwizzleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetViewportSwizzleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetViewportSwizzleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetViewportSwizzleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportSwizzleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportSwizzleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportSwizzleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportSwizzleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportSwizzleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportSwizzleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportSwizzleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportSwizzleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportSwizzleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportSwizzleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportSwizzleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportSwizzleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportSwizzleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetViewportSwizzleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetViewportSwizzleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetViewportSwizzleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportSwizzleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportSwizzleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportSwizzleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportSwizzleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportSwizzleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportSwizzleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportSwizzleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportSwizzleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportSwizzleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportSwizzleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportSwizzleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportSwizzleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportSwizzleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetViewportSwizzleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetViewportSwizzleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetViewportSwizzleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportSwizzleNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportSwizzleNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportSwizzleNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportSwizzleNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportSwizzleNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportSwizzleNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportSwizzleNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportSwizzleNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportSwizzleNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportSwizzleNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportSwizzleNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportSwizzleNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportSwizzleNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetViewportSwizzleNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetViewportSwizzleNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetViewportSwizzleNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetCoverageToColorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetCoverageToColorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetCoverageToColorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetCoverageToColorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetCoverageToColorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetCoverageToColorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetCoverageToColorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetCoverageToColorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetCoverageToColorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetCoverageToColorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetCoverageToColorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetCoverageToColorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetCoverageToColorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetCoverageToColorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetCoverageToColorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetCoverageToColorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetCoverageToColorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetCoverageToColorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetCoverageToColorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetCoverageToColorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetCoverageToColorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetCoverageToColorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetCoverageToColorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetCoverageToColorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetCoverageToColorEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetCoverageToColorEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorLocationNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorLocationNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorLocationNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorLocationNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorLocationNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorLocationNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorLocationNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorLocationNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorLocationNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorLocationNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorLocationNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorLocationNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorLocationNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageToColorLocationNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetCoverageToColorLocationNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetCoverageToColorLocationNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorLocationNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorLocationNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorLocationNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorLocationNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorLocationNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorLocationNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorLocationNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorLocationNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorLocationNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorLocationNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorLocationNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorLocationNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorLocationNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageToColorLocationNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetCoverageToColorLocationNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetCoverageToColorLocationNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorLocationNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorLocationNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorLocationNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorLocationNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorLocationNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorLocationNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorLocationNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorLocationNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorLocationNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorLocationNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorLocationNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorLocationNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorLocationNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageToColorLocationNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetCoverageToColorLocationNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetCoverageToColorLocationNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorLocationNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorLocationNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorLocationNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorLocationNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorLocationNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorLocationNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorLocationNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorLocationNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorLocationNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorLocationNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorLocationNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorLocationNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorLocationNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageToColorLocationNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetCoverageToColorLocationNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetCoverageToColorLocationNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorLocationNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorLocationNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorLocationNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorLocationNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorLocationNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorLocationNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorLocationNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorLocationNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorLocationNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorLocationNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorLocationNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorLocationNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorLocationNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageToColorLocationNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetCoverageToColorLocationNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetCoverageToColorLocationNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorLocationNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorLocationNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorLocationNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorLocationNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorLocationNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorLocationNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorLocationNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorLocationNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorLocationNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorLocationNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorLocationNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorLocationNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorLocationNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageToColorLocationNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetCoverageToColorLocationNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetCoverageToColorLocationNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorLocationNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorLocationNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorLocationNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorLocationNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorLocationNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorLocationNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorLocationNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorLocationNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorLocationNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorLocationNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorLocationNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorLocationNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorLocationNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageToColorLocationNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetCoverageToColorLocationNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetCoverageToColorLocationNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorLocationNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorLocationNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorLocationNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorLocationNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorLocationNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorLocationNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorLocationNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorLocationNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorLocationNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorLocationNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorLocationNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorLocationNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorLocationNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageToColorLocationNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetCoverageToColorLocationNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetCoverageToColorLocationNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorLocationNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorLocationNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorLocationNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorLocationNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorLocationNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorLocationNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorLocationNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorLocationNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorLocationNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorLocationNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorLocationNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorLocationNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorLocationNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageToColorLocationNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetCoverageToColorLocationNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetCoverageToColorLocationNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorLocationNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorLocationNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorLocationNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorLocationNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorLocationNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorLocationNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorLocationNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorLocationNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorLocationNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorLocationNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorLocationNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorLocationNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorLocationNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageToColorLocationNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetCoverageToColorLocationNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetCoverageToColorLocationNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorLocationNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorLocationNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorLocationNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorLocationNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorLocationNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorLocationNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorLocationNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorLocationNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorLocationNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorLocationNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorLocationNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorLocationNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorLocationNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageToColorLocationNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetCoverageToColorLocationNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetCoverageToColorLocationNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorLocationNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorLocationNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorLocationNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorLocationNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorLocationNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorLocationNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorLocationNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorLocationNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorLocationNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorLocationNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorLocationNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorLocationNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorLocationNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageToColorLocationNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetCoverageToColorLocationNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetCoverageToColorLocationNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorLocationNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorLocationNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorLocationNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorLocationNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorLocationNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorLocationNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorLocationNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorLocationNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorLocationNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorLocationNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorLocationNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorLocationNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorLocationNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageToColorLocationNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetCoverageToColorLocationNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetCoverageToColorLocationNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetCoverageModulationModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetCoverageModulationModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetCoverageModulationModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetCoverageModulationModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetCoverageModulationModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetCoverageModulationModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetCoverageModulationModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetCoverageModulationModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetCoverageModulationModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetCoverageModulationModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetCoverageModulationModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetCoverageModulationModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetCoverageModulationModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetCoverageModulationModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetCoverageModulationModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetCoverageModulationModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetCoverageModulationModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetCoverageModulationModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetCoverageModulationModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetCoverageModulationModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetCoverageModulationModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetCoverageModulationModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetCoverageModulationModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetCoverageModulationModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetCoverageModulationModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetCoverageModulationModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetCoverageModulationTableEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetCoverageModulationTableEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetCoverageModulationTableEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetCoverageModulationTableEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetCoverageModulationTableEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetCoverageModulationTableEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetCoverageModulationTableEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetCoverageModulationTableEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetCoverageModulationTableEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetCoverageModulationTableEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetCoverageModulationTableEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetCoverageModulationTableEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetCoverageModulationTableEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetCoverageModulationTableEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetCoverageModulationTableEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetCoverageModulationTableEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetCoverageModulationTableEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetCoverageModulationTableEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetCoverageModulationTableEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetCoverageModulationTableEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetCoverageModulationTableEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetCoverageModulationTableEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetCoverageModulationTableEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetCoverageModulationTableEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetCoverageModulationTableEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetCoverageModulationTableEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageModulationTableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetCoverageModulationTableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetCoverageModulationTableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageModulationTableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetCoverageModulationTableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetCoverageModulationTableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageModulationTableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetCoverageModulationTableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetCoverageModulationTableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageModulationTableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetCoverageModulationTableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetCoverageModulationTableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageModulationTableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetCoverageModulationTableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetCoverageModulationTableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageModulationTableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetCoverageModulationTableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetCoverageModulationTableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageModulationTableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetCoverageModulationTableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetCoverageModulationTableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageModulationTableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetCoverageModulationTableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetCoverageModulationTableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageModulationTableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetCoverageModulationTableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetCoverageModulationTableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageModulationTableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetCoverageModulationTableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetCoverageModulationTableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageModulationTableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetCoverageModulationTableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetCoverageModulationTableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageModulationTableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetCoverageModulationTableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetCoverageModulationTableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageModulationTableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetCoverageModulationTableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetCoverageModulationTableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetShadingRateImageEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetShadingRateImageEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetShadingRateImageEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetShadingRateImageEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetShadingRateImageEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetShadingRateImageEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetShadingRateImageEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetShadingRateImageEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetShadingRateImageEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetShadingRateImageEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetShadingRateImageEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetShadingRateImageEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetShadingRateImageEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetShadingRateImageEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetShadingRateImageEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetShadingRateImageEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetShadingRateImageEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetShadingRateImageEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetShadingRateImageEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetShadingRateImageEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetShadingRateImageEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetShadingRateImageEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetShadingRateImageEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetShadingRateImageEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetShadingRateImageEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetShadingRateImageEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetShadingRateImageEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetShadingRateImageEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetShadingRateImageEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetShadingRateImageEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetShadingRateImageEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetShadingRateImageEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetShadingRateImageEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetShadingRateImageEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetShadingRateImageEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetShadingRateImageEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetShadingRateImageEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetShadingRateImageEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetShadingRateImageEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetShadingRateImageEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetShadingRateImageEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetShadingRateImageEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetShadingRateImageEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetShadingRateImageEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetShadingRateImageEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetShadingRateImageEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetShadingRateImageEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetShadingRateImageEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetShadingRateImageEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetShadingRateImageEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetShadingRateImageEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetShadingRateImageEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetShadingRateImageEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetShadingRateImageEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetShadingRateImageEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetShadingRateImageEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetShadingRateImageEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetShadingRateImageEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetShadingRateImageEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetShadingRateImageEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetShadingRateImageEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetShadingRateImageEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetShadingRateImageEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetShadingRateImageEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetShadingRateImageEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetShadingRateImageEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetShadingRateImageEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetShadingRateImageEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetShadingRateImageEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetShadingRateImageEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetShadingRateImageEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetShadingRateImageEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetShadingRateImageEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetShadingRateImageEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetShadingRateImageEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetShadingRateImageEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetShadingRateImageEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetShadingRateImageEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetShadingRateImageEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetShadingRateImageEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetShadingRateImageEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetShadingRateImageEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetShadingRateImageEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetShadingRateImageEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetShadingRateImageEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetShadingRateImageEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetShadingRateImageEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetShadingRateImageEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetShadingRateImageEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetShadingRateImageEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetShadingRateImageEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetShadingRateImageEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetShadingRateImageEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetShadingRateImageEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetShadingRateImageEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetShadingRateImageEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetShadingRateImageEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetShadingRateImageEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetShadingRateImageEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetShadingRateImageEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetShadingRateImageEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetShadingRateImageEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetShadingRateImageEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetShadingRateImageEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetShadingRateImageEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetShadingRateImageEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetShadingRateImageEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetShadingRateImageEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetShadingRateImageEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetShadingRateImageEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetShadingRateImageEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetShadingRateImageEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetShadingRateImageEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetShadingRateImageEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetShadingRateImageEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetShadingRateImageEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetShadingRateImageEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetShadingRateImageEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetShadingRateImageEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetShadingRateImageEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetShadingRateImageEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetShadingRateImageEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetShadingRateImageEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetShadingRateImageEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetShadingRateImageEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetShadingRateImageEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetShadingRateImageEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetShadingRateImageEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetShadingRateImageEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetShadingRateImageEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetShadingRateImageEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetShadingRateImageEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetShadingRateImageEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetShadingRateImageEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetShadingRateImageEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetShadingRateImageEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetShadingRateImageEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetShadingRateImageEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetShadingRateImageEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetShadingRateImageEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetShadingRateImageEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetShadingRateImageEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetShadingRateImageEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetShadingRateImageEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetShadingRateImageEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetShadingRateImageEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetShadingRateImageEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetShadingRateImageEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetShadingRateImageEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetShadingRateImageEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetShadingRateImageEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetShadingRateImageEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetShadingRateImageEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetShadingRateImageEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetShadingRateImageEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetShadingRateImageEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetShadingRateImageEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetShadingRateImageEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetShadingRateImageEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetShadingRateImageEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetShadingRateImageEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetShadingRateImageEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetShadingRateImageEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetShadingRateImageEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetShadingRateImageEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetShadingRateImageEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetShadingRateImageEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetShadingRateImageEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetShadingRateImageEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetShadingRateImageEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetShadingRateImageEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetShadingRateImageEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetShadingRateImageEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetShadingRateImageEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetShadingRateImageEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetShadingRateImageEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetShadingRateImageEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetShadingRateImageEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetShadingRateImageEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetShadingRateImageEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetShadingRateImageEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetShadingRateImageEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetShadingRateImageEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetShadingRateImageEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetShadingRateImageEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetShadingRateImageEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetShadingRateImageEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetShadingRateImageEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetShadingRateImageEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetShadingRateImageEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetShadingRateImageEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetShadingRateImageEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetShadingRateImageEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetShadingRateImageEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetShadingRateImageEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetShadingRateImageEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetShadingRateImageEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetShadingRateImageEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetShadingRateImageEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetShadingRateImageEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetShadingRateImageEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetShadingRateImageEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetShadingRateImageEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetShadingRateImageEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetShadingRateImageEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetShadingRateImageEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetShadingRateImageEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetShadingRateImageEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageReductionModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageReductionModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageReductionModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageReductionModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageReductionModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageReductionModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageReductionModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageReductionModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageReductionModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageReductionModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageReductionModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageReductionModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageReductionModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetCoverageReductionModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetCoverageReductionModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetCoverageReductionModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageReductionModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageReductionModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageReductionModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageReductionModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageReductionModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageReductionModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageReductionModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageReductionModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageReductionModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageReductionModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageReductionModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageReductionModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageReductionModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetCoverageReductionModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetCoverageReductionModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetCoverageReductionModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageReductionModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageReductionModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageReductionModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageReductionModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageReductionModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageReductionModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageReductionModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageReductionModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageReductionModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageReductionModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageReductionModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageReductionModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageReductionModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetCoverageReductionModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetCoverageReductionModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetCoverageReductionModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageReductionModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageReductionModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageReductionModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageReductionModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageReductionModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageReductionModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageReductionModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageReductionModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageReductionModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageReductionModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageReductionModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageReductionModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageReductionModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetCoverageReductionModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetCoverageReductionModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetCoverageReductionModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageReductionModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageReductionModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageReductionModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageReductionModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageReductionModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageReductionModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageReductionModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageReductionModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageReductionModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageReductionModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageReductionModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageReductionModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageReductionModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetCoverageReductionModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetCoverageReductionModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetCoverageReductionModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageReductionModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageReductionModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageReductionModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageReductionModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageReductionModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageReductionModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageReductionModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageReductionModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageReductionModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageReductionModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageReductionModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageReductionModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageReductionModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetCoverageReductionModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetCoverageReductionModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetCoverageReductionModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageReductionModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageReductionModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageReductionModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageReductionModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageReductionModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageReductionModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageReductionModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageReductionModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageReductionModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageReductionModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageReductionModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageReductionModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageReductionModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetCoverageReductionModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetCoverageReductionModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetCoverageReductionModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageReductionModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageReductionModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageReductionModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageReductionModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageReductionModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageReductionModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageReductionModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageReductionModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageReductionModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageReductionModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageReductionModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageReductionModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageReductionModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetCoverageReductionModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetCoverageReductionModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetCoverageReductionModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageReductionModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageReductionModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageReductionModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageReductionModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageReductionModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageReductionModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageReductionModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageReductionModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageReductionModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageReductionModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageReductionModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageReductionModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageReductionModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetCoverageReductionModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetCoverageReductionModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetCoverageReductionModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageReductionModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageReductionModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageReductionModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageReductionModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageReductionModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageReductionModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageReductionModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageReductionModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageReductionModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageReductionModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageReductionModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageReductionModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageReductionModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetCoverageReductionModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetCoverageReductionModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetCoverageReductionModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageReductionModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageReductionModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageReductionModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageReductionModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageReductionModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageReductionModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageReductionModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageReductionModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageReductionModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageReductionModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageReductionModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageReductionModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageReductionModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetCoverageReductionModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetCoverageReductionModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetCoverageReductionModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageReductionModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageReductionModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageReductionModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageReductionModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageReductionModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageReductionModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageReductionModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageReductionModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageReductionModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageReductionModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageReductionModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageReductionModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageReductionModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetCoverageReductionModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetCoverageReductionModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetCoverageReductionModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageReductionModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageReductionModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageReductionModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageReductionModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageReductionModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageReductionModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageReductionModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageReductionModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageReductionModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageReductionModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageReductionModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageReductionModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageReductionModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetCoverageReductionModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetCoverageReductionModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetCoverageReductionModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetRepresentativeFragmentTestEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRepresentativeFragmentTestEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRepresentativeFragmentTestEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRepresentativeFragmentTestEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRepresentativeFragmentTestEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRepresentativeFragmentTestEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRepresentativeFragmentTestEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRepresentativeFragmentTestEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRepresentativeFragmentTestEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRepresentativeFragmentTestEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRepresentativeFragmentTestEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRepresentativeFragmentTestEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRepresentativeFragmentTestEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRepresentativeFragmentTestEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetRepresentativeFragmentTestEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetRepresentativeFragmentTestEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRepresentativeFragmentTestEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRepresentativeFragmentTestEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRepresentativeFragmentTestEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRepresentativeFragmentTestEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRepresentativeFragmentTestEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRepresentativeFragmentTestEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRepresentativeFragmentTestEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRepresentativeFragmentTestEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRepresentativeFragmentTestEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRepresentativeFragmentTestEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRepresentativeFragmentTestEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRepresentativeFragmentTestEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRepresentativeFragmentTestEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRepresentativeFragmentTestEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetRepresentativeFragmentTestEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetRepresentativeFragmentTestEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRepresentativeFragmentTestEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRepresentativeFragmentTestEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRepresentativeFragmentTestEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRepresentativeFragmentTestEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRepresentativeFragmentTestEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRepresentativeFragmentTestEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRepresentativeFragmentTestEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRepresentativeFragmentTestEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRepresentativeFragmentTestEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRepresentativeFragmentTestEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRepresentativeFragmentTestEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRepresentativeFragmentTestEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRepresentativeFragmentTestEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRepresentativeFragmentTestEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetRepresentativeFragmentTestEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetRepresentativeFragmentTestEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRepresentativeFragmentTestEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRepresentativeFragmentTestEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRepresentativeFragmentTestEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRepresentativeFragmentTestEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRepresentativeFragmentTestEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRepresentativeFragmentTestEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRepresentativeFragmentTestEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRepresentativeFragmentTestEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRepresentativeFragmentTestEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRepresentativeFragmentTestEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRepresentativeFragmentTestEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRepresentativeFragmentTestEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRepresentativeFragmentTestEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRepresentativeFragmentTestEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetRepresentativeFragmentTestEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetRepresentativeFragmentTestEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRepresentativeFragmentTestEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRepresentativeFragmentTestEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRepresentativeFragmentTestEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRepresentativeFragmentTestEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRepresentativeFragmentTestEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRepresentativeFragmentTestEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRepresentativeFragmentTestEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRepresentativeFragmentTestEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRepresentativeFragmentTestEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRepresentativeFragmentTestEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRepresentativeFragmentTestEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRepresentativeFragmentTestEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRepresentativeFragmentTestEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRepresentativeFragmentTestEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetRepresentativeFragmentTestEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetRepresentativeFragmentTestEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRepresentativeFragmentTestEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRepresentativeFragmentTestEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRepresentativeFragmentTestEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRepresentativeFragmentTestEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRepresentativeFragmentTestEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRepresentativeFragmentTestEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRepresentativeFragmentTestEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRepresentativeFragmentTestEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRepresentativeFragmentTestEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRepresentativeFragmentTestEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRepresentativeFragmentTestEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRepresentativeFragmentTestEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRepresentativeFragmentTestEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRepresentativeFragmentTestEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetRepresentativeFragmentTestEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetRepresentativeFragmentTestEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRepresentativeFragmentTestEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRepresentativeFragmentTestEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRepresentativeFragmentTestEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRepresentativeFragmentTestEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRepresentativeFragmentTestEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRepresentativeFragmentTestEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRepresentativeFragmentTestEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRepresentativeFragmentTestEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRepresentativeFragmentTestEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRepresentativeFragmentTestEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRepresentativeFragmentTestEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRepresentativeFragmentTestEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRepresentativeFragmentTestEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRepresentativeFragmentTestEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetRepresentativeFragmentTestEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetRepresentativeFragmentTestEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRepresentativeFragmentTestEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRepresentativeFragmentTestEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRepresentativeFragmentTestEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRepresentativeFragmentTestEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRepresentativeFragmentTestEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRepresentativeFragmentTestEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRepresentativeFragmentTestEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRepresentativeFragmentTestEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRepresentativeFragmentTestEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRepresentativeFragmentTestEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRepresentativeFragmentTestEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRepresentativeFragmentTestEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRepresentativeFragmentTestEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRepresentativeFragmentTestEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetRepresentativeFragmentTestEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetRepresentativeFragmentTestEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRepresentativeFragmentTestEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRepresentativeFragmentTestEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRepresentativeFragmentTestEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRepresentativeFragmentTestEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRepresentativeFragmentTestEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRepresentativeFragmentTestEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRepresentativeFragmentTestEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRepresentativeFragmentTestEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRepresentativeFragmentTestEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRepresentativeFragmentTestEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRepresentativeFragmentTestEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRepresentativeFragmentTestEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRepresentativeFragmentTestEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRepresentativeFragmentTestEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetRepresentativeFragmentTestEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetRepresentativeFragmentTestEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRepresentativeFragmentTestEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRepresentativeFragmentTestEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRepresentativeFragmentTestEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRepresentativeFragmentTestEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRepresentativeFragmentTestEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRepresentativeFragmentTestEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRepresentativeFragmentTestEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRepresentativeFragmentTestEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRepresentativeFragmentTestEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRepresentativeFragmentTestEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRepresentativeFragmentTestEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRepresentativeFragmentTestEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRepresentativeFragmentTestEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRepresentativeFragmentTestEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetRepresentativeFragmentTestEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetRepresentativeFragmentTestEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRepresentativeFragmentTestEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRepresentativeFragmentTestEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRepresentativeFragmentTestEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRepresentativeFragmentTestEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRepresentativeFragmentTestEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRepresentativeFragmentTestEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRepresentativeFragmentTestEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRepresentativeFragmentTestEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRepresentativeFragmentTestEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRepresentativeFragmentTestEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRepresentativeFragmentTestEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRepresentativeFragmentTestEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRepresentativeFragmentTestEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRepresentativeFragmentTestEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetRepresentativeFragmentTestEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetRepresentativeFragmentTestEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRepresentativeFragmentTestEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRepresentativeFragmentTestEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRepresentativeFragmentTestEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRepresentativeFragmentTestEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRepresentativeFragmentTestEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRepresentativeFragmentTestEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRepresentativeFragmentTestEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRepresentativeFragmentTestEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRepresentativeFragmentTestEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRepresentativeFragmentTestEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRepresentativeFragmentTestEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRepresentativeFragmentTestEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRepresentativeFragmentTestEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRepresentativeFragmentTestEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetRepresentativeFragmentTestEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetRepresentativeFragmentTestEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRepresentativeFragmentTestEnableNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRepresentativeFragmentTestEnableNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRepresentativeFragmentTestEnableNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRepresentativeFragmentTestEnableNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRepresentativeFragmentTestEnableNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRepresentativeFragmentTestEnableNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRepresentativeFragmentTestEnableNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRepresentativeFragmentTestEnableNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRepresentativeFragmentTestEnableNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRepresentativeFragmentTestEnableNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRepresentativeFragmentTestEnableNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRepresentativeFragmentTestEnableNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRepresentativeFragmentTestEnableNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRepresentativeFragmentTestEnableNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetRepresentativeFragmentTestEnableNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetRepresentativeFragmentTestEnableNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreatePrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreatePrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreatePrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreatePrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreatePrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreatePrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreatePrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreatePrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreatePrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreatePrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreatePrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreatePrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreatePrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreatePrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreatePrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreatePrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreatePrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreatePrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreatePrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreatePrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreatePrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreatePrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreatePrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreatePrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreatePrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreatePrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreatePrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreatePrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreatePrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreatePrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreatePrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreatePrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreatePrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreatePrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreatePrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreatePrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreatePrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreatePrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreatePrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreatePrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreatePrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreatePrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreatePrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreatePrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreatePrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreatePrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreatePrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreatePrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreatePrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreatePrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreatePrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreatePrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreatePrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreatePrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreatePrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreatePrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreatePrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreatePrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreatePrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreatePrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreatePrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreatePrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreatePrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreatePrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreatePrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyPrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyPrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyPrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyPrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyPrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyPrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyPrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyPrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyPrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyPrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyPrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyPrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyPrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyPrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyPrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyPrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyPrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyPrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyPrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyPrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyPrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyPrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyPrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyPrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlot@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlot@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlot@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlot@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlot@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlot@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlot@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlot@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlot@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlot@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlot@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlot@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlot@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlot@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyPrivateDataSlot=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyPrivateDataSlot=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyPrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyPrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyPrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyPrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyPrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyPrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyPrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyPrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyPrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyPrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyPrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyPrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyPrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyPrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyPrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyPrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyPrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyPrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyPrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyPrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyPrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyPrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyPrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyPrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyPrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyPrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyPrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyPrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyPrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyPrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyPrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyPrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyPrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyPrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyPrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyPrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlotEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlotEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlotEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlotEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlotEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlotEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlotEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlotEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlotEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlotEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlotEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlotEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlotEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyPrivateDataSlotEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyPrivateDataSlotEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyPrivateDataSlotEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_SetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_SetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_SetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_SetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_SetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_SetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_SetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_SetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_SetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_SetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_SetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_SetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_SetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_SetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_SetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_SetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_SetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_SetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_SetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_SetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_SetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_SetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_SetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_SetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_SetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_SetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_SetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_SetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_SetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_SetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_SetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_SetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_SetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_SetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_SetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_SetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_SetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_SetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_SetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_SetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_SetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_SetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_SetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_SetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_SetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_SetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_SetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_SetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_SetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_SetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_SetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_SetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_SetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_SetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateData@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateData@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateData@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateData@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateData@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateData@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateData@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateData@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateData@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateData@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateData@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateData@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateData@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateData@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetPrivateData=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetPrivateData=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPrivateDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetPrivateDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetPrivateDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBlitImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBlitImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBlitImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBlitImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBlitImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBlitImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBlitImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBlitImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBlitImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBlitImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBlitImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBlitImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBlitImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBlitImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBlitImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBlitImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBlitImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBlitImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBlitImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBlitImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBlitImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBlitImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBlitImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBlitImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBlitImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBlitImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBlitImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBlitImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBlitImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBlitImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBlitImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBlitImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBlitImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBlitImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBlitImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBlitImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBlitImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBlitImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBlitImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBlitImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBlitImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBlitImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBlitImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBlitImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBlitImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBlitImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBlitImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBlitImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBlitImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBlitImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBlitImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBlitImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBlitImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBlitImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBlitImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBlitImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBlitImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBlitImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBlitImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBlitImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBlitImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBlitImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBlitImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBlitImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBlitImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyBufferToImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyBufferToImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyBufferToImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyBufferToImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyBufferToImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyBufferToImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyBufferToImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyBufferToImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyBufferToImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyBufferToImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyBufferToImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyBufferToImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyBufferToImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyBufferToImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyBufferToImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyBufferToImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyBufferToImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyBufferToImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyBufferToImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyBufferToImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyBufferToImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyBufferToImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyBufferToImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyBufferToImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyBufferToImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyBufferToImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyBufferToImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyBufferToImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyBufferToImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyBufferToImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyBufferToImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyBufferToImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyBufferToImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyBufferToImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyBufferToImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyBufferToImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyBufferToImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyBufferToImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyBufferToImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyBufferToImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyBufferToImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyBufferToImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyBufferToImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyBufferToImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyBufferToImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyBufferToImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyBufferToImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyBufferToImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyBufferToImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyBufferToImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyBufferToImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyBufferToImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyBufferToImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyBufferToImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyBufferToImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyBufferToImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyBufferToImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyBufferToImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyBufferToImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyBufferToImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyBufferToImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyBufferToImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyBufferToImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyBufferToImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyBufferToImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyImageToBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyImageToBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyImageToBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyImageToBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyImageToBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyImageToBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyImageToBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyImageToBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyImageToBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyImageToBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyImageToBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyImageToBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyImageToBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyImageToBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyImageToBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyImageToBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyImageToBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyImageToBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyImageToBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyImageToBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyImageToBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyImageToBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyImageToBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyImageToBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyImageToBuffer2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyImageToBuffer2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyImageToBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyImageToBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyImageToBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyImageToBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyImageToBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyImageToBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyImageToBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyImageToBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyImageToBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyImageToBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyImageToBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyImageToBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyImageToBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyImageToBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyImageToBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyImageToBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyImageToBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyImageToBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyImageToBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyImageToBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyImageToBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyImageToBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyImageToBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyImageToBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToBuffer2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyImageToBuffer2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyImageToBuffer2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdResolveImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdResolveImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdResolveImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdResolveImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdResolveImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdResolveImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdResolveImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdResolveImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdResolveImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdResolveImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdResolveImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdResolveImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdResolveImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdResolveImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdResolveImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdResolveImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdResolveImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdResolveImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdResolveImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdResolveImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdResolveImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdResolveImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdResolveImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdResolveImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdResolveImage2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdResolveImage2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResolveImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdResolveImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdResolveImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResolveImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdResolveImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdResolveImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResolveImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdResolveImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdResolveImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResolveImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdResolveImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdResolveImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResolveImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdResolveImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdResolveImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResolveImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdResolveImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdResolveImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResolveImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdResolveImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdResolveImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResolveImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdResolveImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdResolveImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResolveImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdResolveImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdResolveImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResolveImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdResolveImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdResolveImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResolveImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdResolveImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdResolveImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResolveImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdResolveImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdResolveImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResolveImage2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdResolveImage2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdResolveImage2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetFragmentShadingRateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetFragmentShadingRateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetFragmentShadingRateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetFragmentShadingRateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetFragmentShadingRateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetFragmentShadingRateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetFragmentShadingRateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetFragmentShadingRateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetFragmentShadingRateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetFragmentShadingRateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetFragmentShadingRateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetFragmentShadingRateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetFragmentShadingRateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetFragmentShadingRateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetFragmentShadingRateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetFragmentShadingRateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetFragmentShadingRateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetFragmentShadingRateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetFragmentShadingRateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetFragmentShadingRateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetFragmentShadingRateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetFragmentShadingRateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetFragmentShadingRateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetFragmentShadingRateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetFragmentShadingRateKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetFragmentShadingRateKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateEnumNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateEnumNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateEnumNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateEnumNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateEnumNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateEnumNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateEnumNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateEnumNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateEnumNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateEnumNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateEnumNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateEnumNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateEnumNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetFragmentShadingRateEnumNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetFragmentShadingRateEnumNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetFragmentShadingRateEnumNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateEnumNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateEnumNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateEnumNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateEnumNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateEnumNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateEnumNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateEnumNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateEnumNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateEnumNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateEnumNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateEnumNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateEnumNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateEnumNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetFragmentShadingRateEnumNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetFragmentShadingRateEnumNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetFragmentShadingRateEnumNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateEnumNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateEnumNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateEnumNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateEnumNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateEnumNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateEnumNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateEnumNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateEnumNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateEnumNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateEnumNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateEnumNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateEnumNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateEnumNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetFragmentShadingRateEnumNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetFragmentShadingRateEnumNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetFragmentShadingRateEnumNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateEnumNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateEnumNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateEnumNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateEnumNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateEnumNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateEnumNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateEnumNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateEnumNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateEnumNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateEnumNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateEnumNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateEnumNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateEnumNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetFragmentShadingRateEnumNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetFragmentShadingRateEnumNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetFragmentShadingRateEnumNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateEnumNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateEnumNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateEnumNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateEnumNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateEnumNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateEnumNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateEnumNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateEnumNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateEnumNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateEnumNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateEnumNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateEnumNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateEnumNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetFragmentShadingRateEnumNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetFragmentShadingRateEnumNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetFragmentShadingRateEnumNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateEnumNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateEnumNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateEnumNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateEnumNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateEnumNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateEnumNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateEnumNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateEnumNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateEnumNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateEnumNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateEnumNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateEnumNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateEnumNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetFragmentShadingRateEnumNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetFragmentShadingRateEnumNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetFragmentShadingRateEnumNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateEnumNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateEnumNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateEnumNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateEnumNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateEnumNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateEnumNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateEnumNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateEnumNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateEnumNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateEnumNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateEnumNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateEnumNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateEnumNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetFragmentShadingRateEnumNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetFragmentShadingRateEnumNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetFragmentShadingRateEnumNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateEnumNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateEnumNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateEnumNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateEnumNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateEnumNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateEnumNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateEnumNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateEnumNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateEnumNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateEnumNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateEnumNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateEnumNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateEnumNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetFragmentShadingRateEnumNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetFragmentShadingRateEnumNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetFragmentShadingRateEnumNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateEnumNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateEnumNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateEnumNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateEnumNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateEnumNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateEnumNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateEnumNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateEnumNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateEnumNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateEnumNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateEnumNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateEnumNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateEnumNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetFragmentShadingRateEnumNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetFragmentShadingRateEnumNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetFragmentShadingRateEnumNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateEnumNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateEnumNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateEnumNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateEnumNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateEnumNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateEnumNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateEnumNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateEnumNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateEnumNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateEnumNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateEnumNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateEnumNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateEnumNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetFragmentShadingRateEnumNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetFragmentShadingRateEnumNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetFragmentShadingRateEnumNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateEnumNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateEnumNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateEnumNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateEnumNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateEnumNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateEnumNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateEnumNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateEnumNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateEnumNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateEnumNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateEnumNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateEnumNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateEnumNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetFragmentShadingRateEnumNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetFragmentShadingRateEnumNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetFragmentShadingRateEnumNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateEnumNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateEnumNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateEnumNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateEnumNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateEnumNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateEnumNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateEnumNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateEnumNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateEnumNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateEnumNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateEnumNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateEnumNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateEnumNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetFragmentShadingRateEnumNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetFragmentShadingRateEnumNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetFragmentShadingRateEnumNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateEnumNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateEnumNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateEnumNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateEnumNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateEnumNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateEnumNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateEnumNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateEnumNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateEnumNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateEnumNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateEnumNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateEnumNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateEnumNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetFragmentShadingRateEnumNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetFragmentShadingRateEnumNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetFragmentShadingRateEnumNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureBuildSizesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureBuildSizesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureBuildSizesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureBuildSizesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureBuildSizesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureBuildSizesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureBuildSizesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureBuildSizesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureBuildSizesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureBuildSizesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureBuildSizesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureBuildSizesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureBuildSizesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureBuildSizesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetAccelerationStructureBuildSizesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetAccelerationStructureBuildSizesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureBuildSizesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureBuildSizesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureBuildSizesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureBuildSizesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureBuildSizesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureBuildSizesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureBuildSizesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureBuildSizesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureBuildSizesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureBuildSizesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureBuildSizesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureBuildSizesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureBuildSizesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureBuildSizesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetAccelerationStructureBuildSizesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetAccelerationStructureBuildSizesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureBuildSizesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureBuildSizesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureBuildSizesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureBuildSizesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureBuildSizesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureBuildSizesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureBuildSizesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureBuildSizesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureBuildSizesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureBuildSizesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureBuildSizesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureBuildSizesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureBuildSizesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureBuildSizesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetAccelerationStructureBuildSizesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetAccelerationStructureBuildSizesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureBuildSizesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureBuildSizesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureBuildSizesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureBuildSizesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureBuildSizesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureBuildSizesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureBuildSizesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureBuildSizesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureBuildSizesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureBuildSizesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureBuildSizesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureBuildSizesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureBuildSizesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureBuildSizesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetAccelerationStructureBuildSizesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetAccelerationStructureBuildSizesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureBuildSizesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureBuildSizesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureBuildSizesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureBuildSizesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureBuildSizesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureBuildSizesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureBuildSizesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureBuildSizesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureBuildSizesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureBuildSizesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureBuildSizesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureBuildSizesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureBuildSizesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureBuildSizesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetAccelerationStructureBuildSizesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetAccelerationStructureBuildSizesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureBuildSizesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureBuildSizesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureBuildSizesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureBuildSizesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureBuildSizesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureBuildSizesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureBuildSizesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureBuildSizesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureBuildSizesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureBuildSizesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureBuildSizesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureBuildSizesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureBuildSizesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureBuildSizesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetAccelerationStructureBuildSizesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetAccelerationStructureBuildSizesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureBuildSizesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureBuildSizesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureBuildSizesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureBuildSizesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureBuildSizesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureBuildSizesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureBuildSizesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureBuildSizesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureBuildSizesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureBuildSizesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureBuildSizesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureBuildSizesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureBuildSizesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureBuildSizesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetAccelerationStructureBuildSizesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetAccelerationStructureBuildSizesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureBuildSizesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureBuildSizesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureBuildSizesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureBuildSizesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureBuildSizesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureBuildSizesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureBuildSizesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureBuildSizesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureBuildSizesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureBuildSizesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureBuildSizesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureBuildSizesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureBuildSizesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureBuildSizesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetAccelerationStructureBuildSizesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetAccelerationStructureBuildSizesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureBuildSizesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureBuildSizesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureBuildSizesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureBuildSizesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureBuildSizesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureBuildSizesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureBuildSizesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureBuildSizesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureBuildSizesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureBuildSizesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureBuildSizesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureBuildSizesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureBuildSizesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureBuildSizesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetAccelerationStructureBuildSizesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetAccelerationStructureBuildSizesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureBuildSizesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureBuildSizesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureBuildSizesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureBuildSizesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureBuildSizesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureBuildSizesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureBuildSizesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureBuildSizesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureBuildSizesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureBuildSizesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureBuildSizesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureBuildSizesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureBuildSizesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureBuildSizesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetAccelerationStructureBuildSizesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetAccelerationStructureBuildSizesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureBuildSizesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureBuildSizesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureBuildSizesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureBuildSizesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureBuildSizesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureBuildSizesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureBuildSizesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureBuildSizesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureBuildSizesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureBuildSizesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureBuildSizesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureBuildSizesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureBuildSizesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureBuildSizesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetAccelerationStructureBuildSizesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetAccelerationStructureBuildSizesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureBuildSizesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureBuildSizesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureBuildSizesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureBuildSizesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureBuildSizesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureBuildSizesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureBuildSizesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureBuildSizesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureBuildSizesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureBuildSizesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureBuildSizesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureBuildSizesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureBuildSizesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureBuildSizesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetAccelerationStructureBuildSizesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetAccelerationStructureBuildSizesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureBuildSizesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureBuildSizesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureBuildSizesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureBuildSizesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureBuildSizesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureBuildSizesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureBuildSizesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureBuildSizesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureBuildSizesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureBuildSizesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureBuildSizesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureBuildSizesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureBuildSizesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureBuildSizesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetAccelerationStructureBuildSizesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetAccelerationStructureBuildSizesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetVertexInputEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetVertexInputEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetVertexInputEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetVertexInputEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetVertexInputEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetVertexInputEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetVertexInputEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetVertexInputEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetVertexInputEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetVertexInputEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetVertexInputEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetVertexInputEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetVertexInputEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetVertexInputEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetVertexInputEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetVertexInputEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetVertexInputEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetVertexInputEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetVertexInputEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetVertexInputEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetVertexInputEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetVertexInputEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetVertexInputEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetVertexInputEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetVertexInputEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetVertexInputEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetVertexInputEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetVertexInputEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetVertexInputEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetVertexInputEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetVertexInputEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetVertexInputEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetVertexInputEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetVertexInputEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetVertexInputEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetVertexInputEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetVertexInputEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetVertexInputEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetVertexInputEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetVertexInputEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetVertexInputEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetVertexInputEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetVertexInputEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetVertexInputEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetVertexInputEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetVertexInputEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetVertexInputEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetVertexInputEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetVertexInputEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetVertexInputEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetVertexInputEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetVertexInputEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetVertexInputEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetVertexInputEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetVertexInputEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetVertexInputEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetVertexInputEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetVertexInputEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetVertexInputEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetVertexInputEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetVertexInputEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetVertexInputEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetVertexInputEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetVertexInputEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetVertexInputEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetVertexInputEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetVertexInputEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetVertexInputEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetVertexInputEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetVertexInputEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetVertexInputEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetVertexInputEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetVertexInputEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetVertexInputEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetVertexInputEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetVertexInputEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetVertexInputEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetVertexInputEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetVertexInputEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetVertexInputEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetVertexInputEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetVertexInputEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetVertexInputEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetVertexInputEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetVertexInputEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetVertexInputEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetVertexInputEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetVertexInputEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetVertexInputEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetVertexInputEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetVertexInputEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetVertexInputEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetVertexInputEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetVertexInputEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetVertexInputEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetVertexInputEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetVertexInputEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetVertexInputEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetVertexInputEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetVertexInputEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetVertexInputEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetVertexInputEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetVertexInputEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetVertexInputEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetVertexInputEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetVertexInputEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetVertexInputEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetVertexInputEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetVertexInputEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetVertexInputEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetVertexInputEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetVertexInputEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetVertexInputEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetVertexInputEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetVertexInputEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetVertexInputEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetVertexInputEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetVertexInputEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetVertexInputEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetVertexInputEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetVertexInputEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetVertexInputEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetVertexInputEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetVertexInputEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetVertexInputEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetVertexInputEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetVertexInputEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetVertexInputEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetVertexInputEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetVertexInputEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetVertexInputEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetVertexInputEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetVertexInputEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetVertexInputEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetVertexInputEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetVertexInputEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetVertexInputEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetVertexInputEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetVertexInputEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetVertexInputEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetVertexInputEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetVertexInputEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetVertexInputEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetVertexInputEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetVertexInputEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetVertexInputEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetVertexInputEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetVertexInputEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetVertexInputEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetVertexInputEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetVertexInputEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetVertexInputEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetVertexInputEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetVertexInputEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetVertexInputEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetVertexInputEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetVertexInputEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetVertexInputEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetVertexInputEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetVertexInputEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetVertexInputEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetVertexInputEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetVertexInputEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetVertexInputEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetVertexInputEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetVertexInputEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetVertexInputEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetVertexInputEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetVertexInputEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetVertexInputEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetVertexInputEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetVertexInputEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetVertexInputEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetVertexInputEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetVertexInputEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetVertexInputEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetVertexInputEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetVertexInputEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetVertexInputEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetVertexInputEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetVertexInputEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetVertexInputEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetVertexInputEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetVertexInputEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetVertexInputEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetVertexInputEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetVertexInputEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetVertexInputEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetVertexInputEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetVertexInputEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetVertexInputEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetVertexInputEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetVertexInputEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetVertexInputEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetVertexInputEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetVertexInputEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetVertexInputEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetVertexInputEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetVertexInputEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetVertexInputEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetVertexInputEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetVertexInputEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetVertexInputEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetVertexInputEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetVertexInputEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetVertexInputEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetVertexInputEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetVertexInputEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetColorWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetColorWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetColorWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetColorWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetColorWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetColorWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetColorWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetColorWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetColorWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetColorWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetColorWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetColorWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetColorWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetColorWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetColorWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetColorWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetColorWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetColorWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetColorWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetColorWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetColorWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetColorWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetColorWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetColorWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetColorWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetColorWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetColorWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetColorWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetColorWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetColorWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetColorWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetColorWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetColorWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetColorWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetColorWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetColorWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteEnableEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteEnableEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteEnableEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteEnableEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteEnableEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteEnableEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteEnableEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteEnableEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteEnableEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteEnableEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteEnableEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteEnableEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteEnableEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetColorWriteEnableEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetColorWriteEnableEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetColorWriteEnableEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdResetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdResetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdResetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdResetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdResetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdResetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdResetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdResetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdResetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdResetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdResetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdResetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdResetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdResetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdResetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdResetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdResetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdResetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdResetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdResetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdResetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdResetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdResetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdResetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdResetEvent2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdResetEvent2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdResetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdResetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdResetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdResetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdResetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdResetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdResetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdResetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdResetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdResetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdResetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdResetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdResetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdResetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdResetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdResetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdResetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdResetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdResetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdResetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdResetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdResetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdResetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdResetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdResetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdResetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdResetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdResetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdResetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdResetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdResetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdResetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdResetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdResetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdResetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdResetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdResetEvent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdResetEvent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdResetEvent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdWaitEvents2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdWaitEvents2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdWaitEvents2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdWaitEvents2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdWaitEvents2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdWaitEvents2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdWaitEvents2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdWaitEvents2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdWaitEvents2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdWaitEvents2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdWaitEvents2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdWaitEvents2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdWaitEvents2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdWaitEvents2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdWaitEvents2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdWaitEvents2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdWaitEvents2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdWaitEvents2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdWaitEvents2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdWaitEvents2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdWaitEvents2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdWaitEvents2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdWaitEvents2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdWaitEvents2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdWaitEvents2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdWaitEvents2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWaitEvents2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdWaitEvents2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdWaitEvents2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWaitEvents2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdWaitEvents2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdWaitEvents2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWaitEvents2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdWaitEvents2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdWaitEvents2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWaitEvents2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdWaitEvents2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdWaitEvents2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWaitEvents2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdWaitEvents2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdWaitEvents2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWaitEvents2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdWaitEvents2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdWaitEvents2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWaitEvents2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdWaitEvents2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdWaitEvents2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWaitEvents2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdWaitEvents2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdWaitEvents2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWaitEvents2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdWaitEvents2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdWaitEvents2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWaitEvents2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdWaitEvents2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdWaitEvents2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWaitEvents2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdWaitEvents2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdWaitEvents2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWaitEvents2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdWaitEvents2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdWaitEvents2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWaitEvents2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdWaitEvents2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdWaitEvents2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdPipelineBarrier2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdPipelineBarrier2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdPipelineBarrier2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdPipelineBarrier2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdPipelineBarrier2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdPipelineBarrier2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdPipelineBarrier2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdPipelineBarrier2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdPipelineBarrier2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdPipelineBarrier2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdPipelineBarrier2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdPipelineBarrier2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdPipelineBarrier2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdPipelineBarrier2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdPipelineBarrier2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdPipelineBarrier2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdPipelineBarrier2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdPipelineBarrier2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdPipelineBarrier2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdPipelineBarrier2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdPipelineBarrier2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdPipelineBarrier2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdPipelineBarrier2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdPipelineBarrier2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdPipelineBarrier2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdPipelineBarrier2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPipelineBarrier2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdPipelineBarrier2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdPipelineBarrier2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPipelineBarrier2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdPipelineBarrier2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdPipelineBarrier2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPipelineBarrier2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdPipelineBarrier2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdPipelineBarrier2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPipelineBarrier2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdPipelineBarrier2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdPipelineBarrier2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPipelineBarrier2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdPipelineBarrier2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdPipelineBarrier2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPipelineBarrier2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdPipelineBarrier2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdPipelineBarrier2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPipelineBarrier2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdPipelineBarrier2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdPipelineBarrier2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPipelineBarrier2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdPipelineBarrier2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdPipelineBarrier2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPipelineBarrier2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdPipelineBarrier2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdPipelineBarrier2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPipelineBarrier2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdPipelineBarrier2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdPipelineBarrier2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPipelineBarrier2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdPipelineBarrier2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdPipelineBarrier2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPipelineBarrier2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdPipelineBarrier2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdPipelineBarrier2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPipelineBarrier2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdPipelineBarrier2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdPipelineBarrier2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_QueueSubmit2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_QueueSubmit2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_QueueSubmit2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_QueueSubmit2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_QueueSubmit2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_QueueSubmit2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_QueueSubmit2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_QueueSubmit2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_QueueSubmit2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_QueueSubmit2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_QueueSubmit2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_QueueSubmit2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_QueueSubmit2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_QueueSubmit2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_QueueSubmit2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_QueueSubmit2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_QueueSubmit2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_QueueSubmit2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_QueueSubmit2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_QueueSubmit2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_QueueSubmit2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_QueueSubmit2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_QueueSubmit2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_QueueSubmit2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_QueueSubmit2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_QueueSubmit2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSubmit2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_QueueSubmit2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_QueueSubmit2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSubmit2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_QueueSubmit2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_QueueSubmit2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSubmit2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_QueueSubmit2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_QueueSubmit2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSubmit2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_QueueSubmit2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_QueueSubmit2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSubmit2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_QueueSubmit2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_QueueSubmit2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSubmit2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_QueueSubmit2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_QueueSubmit2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSubmit2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_QueueSubmit2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_QueueSubmit2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSubmit2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_QueueSubmit2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_QueueSubmit2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSubmit2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_QueueSubmit2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_QueueSubmit2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSubmit2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_QueueSubmit2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_QueueSubmit2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSubmit2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_QueueSubmit2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_QueueSubmit2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSubmit2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_QueueSubmit2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_QueueSubmit2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSubmit2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_QueueSubmit2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_QueueSubmit2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdWriteTimestamp2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdWriteTimestamp2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdWriteTimestamp2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdWriteTimestamp2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdWriteTimestamp2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdWriteTimestamp2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdWriteTimestamp2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdWriteTimestamp2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdWriteTimestamp2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdWriteTimestamp2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdWriteTimestamp2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdWriteTimestamp2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdWriteTimestamp2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdWriteTimestamp2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdWriteTimestamp2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdWriteTimestamp2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdWriteTimestamp2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdWriteTimestamp2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdWriteTimestamp2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdWriteTimestamp2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdWriteTimestamp2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdWriteTimestamp2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdWriteTimestamp2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdWriteTimestamp2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdWriteTimestamp2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdWriteTimestamp2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteTimestamp2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdWriteTimestamp2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdWriteTimestamp2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteTimestamp2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdWriteTimestamp2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdWriteTimestamp2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteTimestamp2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdWriteTimestamp2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdWriteTimestamp2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteTimestamp2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdWriteTimestamp2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdWriteTimestamp2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteTimestamp2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdWriteTimestamp2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdWriteTimestamp2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteTimestamp2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdWriteTimestamp2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdWriteTimestamp2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteTimestamp2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdWriteTimestamp2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdWriteTimestamp2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteTimestamp2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdWriteTimestamp2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdWriteTimestamp2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteTimestamp2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdWriteTimestamp2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdWriteTimestamp2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteTimestamp2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdWriteTimestamp2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdWriteTimestamp2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteTimestamp2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdWriteTimestamp2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdWriteTimestamp2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteTimestamp2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdWriteTimestamp2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdWriteTimestamp2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteTimestamp2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdWriteTimestamp2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdWriteTimestamp2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarker2AMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarker2AMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarker2AMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarker2AMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarker2AMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarker2AMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarker2AMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarker2AMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarker2AMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarker2AMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarker2AMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarker2AMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarker2AMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteBufferMarker2AMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdWriteBufferMarker2AMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdWriteBufferMarker2AMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarker2AMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarker2AMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarker2AMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarker2AMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarker2AMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarker2AMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarker2AMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarker2AMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarker2AMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarker2AMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarker2AMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarker2AMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarker2AMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteBufferMarker2AMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdWriteBufferMarker2AMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdWriteBufferMarker2AMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarker2AMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarker2AMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarker2AMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarker2AMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarker2AMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarker2AMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarker2AMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarker2AMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarker2AMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarker2AMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarker2AMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarker2AMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarker2AMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteBufferMarker2AMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdWriteBufferMarker2AMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdWriteBufferMarker2AMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarker2AMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarker2AMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarker2AMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarker2AMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarker2AMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarker2AMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarker2AMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarker2AMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarker2AMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarker2AMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarker2AMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarker2AMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarker2AMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteBufferMarker2AMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdWriteBufferMarker2AMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdWriteBufferMarker2AMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarker2AMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarker2AMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarker2AMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarker2AMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarker2AMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarker2AMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarker2AMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarker2AMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarker2AMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarker2AMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarker2AMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarker2AMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarker2AMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteBufferMarker2AMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdWriteBufferMarker2AMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdWriteBufferMarker2AMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarker2AMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarker2AMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarker2AMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarker2AMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarker2AMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarker2AMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarker2AMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarker2AMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarker2AMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarker2AMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarker2AMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarker2AMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarker2AMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteBufferMarker2AMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdWriteBufferMarker2AMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdWriteBufferMarker2AMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarker2AMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarker2AMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarker2AMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarker2AMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarker2AMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarker2AMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarker2AMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarker2AMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarker2AMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarker2AMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarker2AMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarker2AMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarker2AMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteBufferMarker2AMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdWriteBufferMarker2AMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdWriteBufferMarker2AMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarker2AMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarker2AMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarker2AMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarker2AMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarker2AMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarker2AMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarker2AMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarker2AMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarker2AMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarker2AMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarker2AMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarker2AMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarker2AMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteBufferMarker2AMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdWriteBufferMarker2AMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdWriteBufferMarker2AMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarker2AMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarker2AMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarker2AMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarker2AMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarker2AMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarker2AMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarker2AMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarker2AMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarker2AMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarker2AMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarker2AMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarker2AMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarker2AMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteBufferMarker2AMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdWriteBufferMarker2AMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdWriteBufferMarker2AMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarker2AMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarker2AMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarker2AMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarker2AMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarker2AMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarker2AMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarker2AMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarker2AMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarker2AMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarker2AMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarker2AMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarker2AMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarker2AMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteBufferMarker2AMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdWriteBufferMarker2AMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdWriteBufferMarker2AMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarker2AMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarker2AMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarker2AMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarker2AMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarker2AMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarker2AMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarker2AMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarker2AMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarker2AMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarker2AMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarker2AMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarker2AMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarker2AMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteBufferMarker2AMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdWriteBufferMarker2AMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdWriteBufferMarker2AMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarker2AMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarker2AMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarker2AMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarker2AMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarker2AMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarker2AMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarker2AMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarker2AMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarker2AMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarker2AMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarker2AMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarker2AMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarker2AMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteBufferMarker2AMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdWriteBufferMarker2AMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdWriteBufferMarker2AMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarker2AMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarker2AMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarker2AMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarker2AMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarker2AMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarker2AMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarker2AMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarker2AMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarker2AMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarker2AMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarker2AMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarker2AMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarker2AMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteBufferMarker2AMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdWriteBufferMarker2AMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdWriteBufferMarker2AMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointData2NV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointData2NV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointData2NV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointData2NV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointData2NV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointData2NV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointData2NV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointData2NV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointData2NV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointData2NV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointData2NV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointData2NV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointData2NV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetQueueCheckpointData2NV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetQueueCheckpointData2NV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetQueueCheckpointData2NV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointData2NV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointData2NV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointData2NV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointData2NV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointData2NV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointData2NV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointData2NV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointData2NV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointData2NV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointData2NV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointData2NV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointData2NV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointData2NV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetQueueCheckpointData2NV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetQueueCheckpointData2NV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetQueueCheckpointData2NV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointData2NV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointData2NV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointData2NV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointData2NV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointData2NV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointData2NV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointData2NV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointData2NV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointData2NV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointData2NV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointData2NV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointData2NV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointData2NV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetQueueCheckpointData2NV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetQueueCheckpointData2NV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetQueueCheckpointData2NV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointData2NV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointData2NV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointData2NV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointData2NV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointData2NV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointData2NV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointData2NV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointData2NV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointData2NV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointData2NV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointData2NV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointData2NV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointData2NV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetQueueCheckpointData2NV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetQueueCheckpointData2NV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetQueueCheckpointData2NV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointData2NV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointData2NV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointData2NV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointData2NV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointData2NV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointData2NV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointData2NV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointData2NV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointData2NV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointData2NV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointData2NV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointData2NV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointData2NV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetQueueCheckpointData2NV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetQueueCheckpointData2NV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetQueueCheckpointData2NV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointData2NV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointData2NV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointData2NV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointData2NV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointData2NV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointData2NV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointData2NV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointData2NV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointData2NV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointData2NV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointData2NV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointData2NV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointData2NV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetQueueCheckpointData2NV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetQueueCheckpointData2NV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetQueueCheckpointData2NV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointData2NV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointData2NV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointData2NV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointData2NV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointData2NV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointData2NV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointData2NV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointData2NV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointData2NV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointData2NV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointData2NV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointData2NV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointData2NV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetQueueCheckpointData2NV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetQueueCheckpointData2NV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetQueueCheckpointData2NV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointData2NV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointData2NV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointData2NV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointData2NV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointData2NV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointData2NV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointData2NV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointData2NV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointData2NV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointData2NV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointData2NV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointData2NV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointData2NV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetQueueCheckpointData2NV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetQueueCheckpointData2NV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetQueueCheckpointData2NV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointData2NV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointData2NV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointData2NV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointData2NV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointData2NV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointData2NV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointData2NV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointData2NV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointData2NV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointData2NV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointData2NV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointData2NV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointData2NV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetQueueCheckpointData2NV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetQueueCheckpointData2NV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetQueueCheckpointData2NV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointData2NV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointData2NV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointData2NV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointData2NV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointData2NV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointData2NV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointData2NV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointData2NV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointData2NV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointData2NV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointData2NV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointData2NV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointData2NV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetQueueCheckpointData2NV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetQueueCheckpointData2NV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetQueueCheckpointData2NV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointData2NV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointData2NV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointData2NV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointData2NV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointData2NV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointData2NV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointData2NV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointData2NV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointData2NV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointData2NV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointData2NV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointData2NV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointData2NV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetQueueCheckpointData2NV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetQueueCheckpointData2NV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetQueueCheckpointData2NV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointData2NV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointData2NV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointData2NV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointData2NV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointData2NV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointData2NV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointData2NV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointData2NV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointData2NV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointData2NV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointData2NV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointData2NV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointData2NV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetQueueCheckpointData2NV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetQueueCheckpointData2NV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetQueueCheckpointData2NV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointData2NV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointData2NV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointData2NV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointData2NV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointData2NV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointData2NV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointData2NV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointData2NV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointData2NV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointData2NV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointData2NV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointData2NV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointData2NV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetQueueCheckpointData2NV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetQueueCheckpointData2NV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetQueueCheckpointData2NV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CopyMemoryToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CopyMemoryToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CopyMemoryToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CopyMemoryToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CopyMemoryToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CopyMemoryToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CopyMemoryToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CopyMemoryToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CopyMemoryToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CopyMemoryToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CopyMemoryToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CopyMemoryToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CopyMemoryToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CopyMemoryToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CopyMemoryToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CopyMemoryToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CopyMemoryToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CopyMemoryToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CopyMemoryToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CopyMemoryToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CopyMemoryToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CopyMemoryToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CopyMemoryToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CopyMemoryToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CopyMemoryToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CopyMemoryToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CopyMemoryToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CopyMemoryToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CopyMemoryToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CopyMemoryToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CopyMemoryToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CopyMemoryToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CopyMemoryToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CopyMemoryToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CopyMemoryToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CopyMemoryToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CopyMemoryToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CopyMemoryToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CopyMemoryToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CopyMemoryToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CopyMemoryToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CopyMemoryToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CopyMemoryToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CopyMemoryToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CopyMemoryToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CopyMemoryToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CopyMemoryToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CopyMemoryToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CopyMemoryToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CopyMemoryToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CopyMemoryToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CopyMemoryToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CopyImageToMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CopyImageToMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CopyImageToMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CopyImageToMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CopyImageToMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CopyImageToMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CopyImageToMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CopyImageToMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CopyImageToMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CopyImageToMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CopyImageToMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CopyImageToMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CopyImageToMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CopyImageToMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CopyImageToMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CopyImageToMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CopyImageToMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CopyImageToMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CopyImageToMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CopyImageToMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CopyImageToMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CopyImageToMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CopyImageToMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CopyImageToMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemory@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemory@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemory@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemory@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemory@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemory@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemory@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemory@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemory@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemory@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemory@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemory@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemory@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemory@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CopyImageToMemory=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CopyImageToMemory=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CopyImageToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CopyImageToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CopyImageToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CopyImageToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CopyImageToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CopyImageToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CopyImageToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CopyImageToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CopyImageToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CopyImageToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CopyImageToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CopyImageToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CopyImageToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CopyImageToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CopyImageToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CopyImageToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CopyImageToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CopyImageToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CopyImageToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CopyImageToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CopyImageToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CopyImageToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CopyImageToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CopyImageToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CopyImageToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CopyImageToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CopyImageToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CopyImageToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CopyImageToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CopyImageToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CopyImageToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CopyImageToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CopyImageToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CopyImageToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CopyImageToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CopyImageToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CopyImageToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CopyImageToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CopyImageToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CopyImageToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CopyImageToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CopyImageToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CopyImageToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CopyImageToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CopyImageToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CopyImageToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CopyImageToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CopyImageToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CopyImageToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CopyImageToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImage@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImage@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImage@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImage@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImage@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImage@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImage@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImage@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImage@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImage@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImage@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImage@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImage@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImage@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CopyImageToImage=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CopyImageToImage=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyImageToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CopyImageToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CopyImageToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyImageToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CopyImageToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CopyImageToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyImageToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CopyImageToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CopyImageToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyImageToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CopyImageToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CopyImageToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyImageToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CopyImageToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CopyImageToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyImageToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CopyImageToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CopyImageToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyImageToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CopyImageToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CopyImageToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyImageToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CopyImageToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CopyImageToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyImageToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CopyImageToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CopyImageToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyImageToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CopyImageToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CopyImageToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyImageToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CopyImageToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CopyImageToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyImageToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CopyImageToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CopyImageToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImageEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImageEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImageEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImageEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImageEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImageEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImageEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImageEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImageEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImageEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImageEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImageEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImageEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyImageToImageEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CopyImageToImageEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CopyImageToImageEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_TransitionImageLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_TransitionImageLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_TransitionImageLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_TransitionImageLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_TransitionImageLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_TransitionImageLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_TransitionImageLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_TransitionImageLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_TransitionImageLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_TransitionImageLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_TransitionImageLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_TransitionImageLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_TransitionImageLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_TransitionImageLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_TransitionImageLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_TransitionImageLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_TransitionImageLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_TransitionImageLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_TransitionImageLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_TransitionImageLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_TransitionImageLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_TransitionImageLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_TransitionImageLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_TransitionImageLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_TransitionImageLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_TransitionImageLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_TransitionImageLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_TransitionImageLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_TransitionImageLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_TransitionImageLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_TransitionImageLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_TransitionImageLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_TransitionImageLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_TransitionImageLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_TransitionImageLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_TransitionImageLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_TransitionImageLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_TransitionImageLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_TransitionImageLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_TransitionImageLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_TransitionImageLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_TransitionImageLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_TransitionImageLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_TransitionImageLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_TransitionImageLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_TransitionImageLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_TransitionImageLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_TransitionImageLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_TransitionImageLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_TransitionImageLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_TransitionImageLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_TransitionImageLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_TransitionImageLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_TransitionImageLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_TransitionImageLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_TransitionImageLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_TransitionImageLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_TransitionImageLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_TransitionImageLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_TransitionImageLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_TransitionImageLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_TransitionImageLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayoutEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayoutEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayoutEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayoutEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayoutEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayoutEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayoutEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayoutEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayoutEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayoutEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayoutEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayoutEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayoutEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_TransitionImageLayoutEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_TransitionImageLayoutEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_TransitionImageLayoutEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyVideoSessionKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyVideoSessionKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_UpdateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UpdateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_UpdateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_UpdateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_UpdateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UpdateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_UpdateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_UpdateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_UpdateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UpdateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_UpdateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_UpdateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_UpdateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UpdateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_UpdateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_UpdateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_UpdateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UpdateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_UpdateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_UpdateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_UpdateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UpdateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_UpdateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_UpdateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_UpdateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UpdateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_UpdateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_UpdateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_UpdateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UpdateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_UpdateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_UpdateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UpdateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_UpdateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_UpdateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UpdateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_UpdateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_UpdateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UpdateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_UpdateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_UpdateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_UpdateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UpdateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_UpdateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_UpdateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UpdateVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_UpdateVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_UpdateVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetEncodedVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEncodedVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEncodedVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEncodedVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEncodedVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEncodedVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEncodedVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEncodedVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEncodedVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEncodedVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEncodedVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEncodedVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEncodedVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetEncodedVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetEncodedVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetEncodedVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetEncodedVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEncodedVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEncodedVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEncodedVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEncodedVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEncodedVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEncodedVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEncodedVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEncodedVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEncodedVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEncodedVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEncodedVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEncodedVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetEncodedVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetEncodedVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetEncodedVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetEncodedVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEncodedVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEncodedVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEncodedVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEncodedVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEncodedVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEncodedVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEncodedVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEncodedVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEncodedVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEncodedVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEncodedVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEncodedVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetEncodedVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetEncodedVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetEncodedVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetEncodedVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEncodedVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEncodedVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEncodedVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEncodedVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEncodedVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEncodedVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEncodedVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEncodedVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEncodedVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEncodedVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEncodedVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEncodedVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetEncodedVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetEncodedVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetEncodedVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetEncodedVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEncodedVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEncodedVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEncodedVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEncodedVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEncodedVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEncodedVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEncodedVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEncodedVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEncodedVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEncodedVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEncodedVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEncodedVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetEncodedVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetEncodedVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetEncodedVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetEncodedVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEncodedVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEncodedVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEncodedVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEncodedVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEncodedVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEncodedVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEncodedVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEncodedVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEncodedVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEncodedVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEncodedVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEncodedVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetEncodedVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetEncodedVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetEncodedVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetEncodedVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEncodedVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEncodedVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEncodedVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEncodedVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEncodedVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEncodedVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEncodedVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEncodedVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEncodedVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEncodedVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEncodedVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEncodedVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetEncodedVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetEncodedVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetEncodedVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetEncodedVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEncodedVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEncodedVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEncodedVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEncodedVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEncodedVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEncodedVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEncodedVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEncodedVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEncodedVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEncodedVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEncodedVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEncodedVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetEncodedVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetEncodedVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetEncodedVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEncodedVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEncodedVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEncodedVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEncodedVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEncodedVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEncodedVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEncodedVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEncodedVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEncodedVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEncodedVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEncodedVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEncodedVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEncodedVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetEncodedVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetEncodedVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetEncodedVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEncodedVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEncodedVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEncodedVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEncodedVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEncodedVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEncodedVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEncodedVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEncodedVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEncodedVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEncodedVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEncodedVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEncodedVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEncodedVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetEncodedVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetEncodedVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetEncodedVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEncodedVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEncodedVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEncodedVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEncodedVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEncodedVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEncodedVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEncodedVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEncodedVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEncodedVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEncodedVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEncodedVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEncodedVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEncodedVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetEncodedVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetEncodedVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetEncodedVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetEncodedVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEncodedVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEncodedVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEncodedVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEncodedVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEncodedVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEncodedVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEncodedVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEncodedVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEncodedVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEncodedVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEncodedVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEncodedVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetEncodedVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetEncodedVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetEncodedVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEncodedVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEncodedVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEncodedVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEncodedVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEncodedVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEncodedVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEncodedVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEncodedVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEncodedVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEncodedVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEncodedVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEncodedVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEncodedVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetEncodedVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetEncodedVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetEncodedVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionParametersKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionParametersKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionParametersKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionParametersKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionParametersKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionParametersKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionParametersKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionParametersKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionParametersKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionParametersKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionParametersKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionParametersKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionParametersKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyVideoSessionParametersKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyVideoSessionParametersKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyVideoSessionParametersKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetVideoSessionMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetVideoSessionMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetVideoSessionMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetVideoSessionMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetVideoSessionMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetVideoSessionMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetVideoSessionMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetVideoSessionMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetVideoSessionMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetVideoSessionMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetVideoSessionMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetVideoSessionMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetVideoSessionMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetVideoSessionMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetVideoSessionMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetVideoSessionMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetVideoSessionMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetVideoSessionMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetVideoSessionMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetVideoSessionMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetVideoSessionMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetVideoSessionMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetVideoSessionMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetVideoSessionMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetVideoSessionMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetVideoSessionMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetVideoSessionMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetVideoSessionMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetVideoSessionMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetVideoSessionMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetVideoSessionMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetVideoSessionMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetVideoSessionMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetVideoSessionMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetVideoSessionMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetVideoSessionMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetVideoSessionMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetVideoSessionMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetVideoSessionMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetVideoSessionMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetVideoSessionMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetVideoSessionMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetVideoSessionMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetVideoSessionMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetVideoSessionMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetVideoSessionMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetVideoSessionMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetVideoSessionMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetVideoSessionMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetVideoSessionMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetVideoSessionMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetVideoSessionMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetVideoSessionMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetVideoSessionMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetVideoSessionMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetVideoSessionMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetVideoSessionMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetVideoSessionMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetVideoSessionMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetVideoSessionMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetVideoSessionMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetVideoSessionMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetVideoSessionMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetVideoSessionMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetVideoSessionMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetVideoSessionMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetVideoSessionMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetVideoSessionMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetVideoSessionMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetVideoSessionMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetVideoSessionMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetVideoSessionMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetVideoSessionMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetVideoSessionMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetVideoSessionMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetVideoSessionMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetVideoSessionMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetVideoSessionMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetVideoSessionMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetVideoSessionMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetVideoSessionMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetVideoSessionMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetVideoSessionMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetVideoSessionMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetVideoSessionMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetVideoSessionMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetVideoSessionMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetVideoSessionMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetVideoSessionMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetVideoSessionMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetVideoSessionMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetVideoSessionMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetVideoSessionMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetVideoSessionMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetVideoSessionMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetVideoSessionMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetVideoSessionMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetVideoSessionMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetVideoSessionMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetVideoSessionMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetVideoSessionMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetVideoSessionMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetVideoSessionMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetVideoSessionMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetVideoSessionMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetVideoSessionMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetVideoSessionMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetVideoSessionMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetVideoSessionMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetVideoSessionMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetVideoSessionMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetVideoSessionMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetVideoSessionMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetVideoSessionMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetVideoSessionMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetVideoSessionMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetVideoSessionMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetVideoSessionMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetVideoSessionMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetVideoSessionMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetVideoSessionMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetVideoSessionMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetVideoSessionMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetVideoSessionMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetVideoSessionMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetVideoSessionMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetVideoSessionMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetVideoSessionMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetVideoSessionMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetVideoSessionMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetVideoSessionMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetVideoSessionMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetVideoSessionMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetVideoSessionMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetVideoSessionMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetVideoSessionMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetVideoSessionMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetVideoSessionMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetVideoSessionMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetVideoSessionMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetVideoSessionMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetVideoSessionMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetVideoSessionMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetVideoSessionMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetVideoSessionMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetVideoSessionMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetVideoSessionMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetVideoSessionMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetVideoSessionMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetVideoSessionMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetVideoSessionMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetVideoSessionMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetVideoSessionMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetVideoSessionMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetVideoSessionMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetVideoSessionMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetVideoSessionMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetVideoSessionMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetVideoSessionMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetVideoSessionMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetVideoSessionMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetVideoSessionMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetVideoSessionMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetVideoSessionMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetVideoSessionMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetVideoSessionMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetVideoSessionMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetVideoSessionMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetVideoSessionMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetVideoSessionMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetVideoSessionMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetVideoSessionMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetVideoSessionMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetVideoSessionMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetVideoSessionMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetVideoSessionMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetVideoSessionMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetVideoSessionMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetVideoSessionMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetVideoSessionMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetVideoSessionMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetVideoSessionMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetVideoSessionMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetVideoSessionMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetVideoSessionMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetVideoSessionMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetVideoSessionMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetVideoSessionMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetVideoSessionMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetVideoSessionMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetVideoSessionMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetVideoSessionMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetVideoSessionMemoryRequirementsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetVideoSessionMemoryRequirementsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetVideoSessionMemoryRequirementsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetVideoSessionMemoryRequirementsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetVideoSessionMemoryRequirementsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetVideoSessionMemoryRequirementsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetVideoSessionMemoryRequirementsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetVideoSessionMemoryRequirementsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetVideoSessionMemoryRequirementsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetVideoSessionMemoryRequirementsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetVideoSessionMemoryRequirementsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetVideoSessionMemoryRequirementsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetVideoSessionMemoryRequirementsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetVideoSessionMemoryRequirementsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetVideoSessionMemoryRequirementsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetVideoSessionMemoryRequirementsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_BindVideoSessionMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindVideoSessionMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindVideoSessionMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindVideoSessionMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindVideoSessionMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindVideoSessionMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindVideoSessionMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindVideoSessionMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindVideoSessionMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindVideoSessionMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindVideoSessionMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindVideoSessionMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindVideoSessionMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindVideoSessionMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_BindVideoSessionMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_BindVideoSessionMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_BindVideoSessionMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindVideoSessionMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindVideoSessionMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindVideoSessionMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindVideoSessionMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindVideoSessionMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindVideoSessionMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindVideoSessionMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindVideoSessionMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindVideoSessionMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindVideoSessionMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindVideoSessionMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindVideoSessionMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindVideoSessionMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_BindVideoSessionMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_BindVideoSessionMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_BindVideoSessionMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindVideoSessionMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindVideoSessionMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindVideoSessionMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindVideoSessionMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindVideoSessionMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindVideoSessionMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindVideoSessionMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindVideoSessionMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindVideoSessionMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindVideoSessionMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindVideoSessionMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindVideoSessionMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindVideoSessionMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_BindVideoSessionMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_BindVideoSessionMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_BindVideoSessionMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindVideoSessionMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindVideoSessionMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindVideoSessionMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindVideoSessionMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindVideoSessionMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindVideoSessionMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindVideoSessionMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindVideoSessionMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindVideoSessionMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindVideoSessionMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindVideoSessionMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindVideoSessionMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindVideoSessionMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_BindVideoSessionMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_BindVideoSessionMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_BindVideoSessionMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindVideoSessionMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindVideoSessionMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindVideoSessionMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindVideoSessionMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindVideoSessionMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindVideoSessionMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindVideoSessionMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindVideoSessionMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindVideoSessionMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindVideoSessionMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindVideoSessionMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindVideoSessionMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindVideoSessionMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_BindVideoSessionMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_BindVideoSessionMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_BindVideoSessionMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindVideoSessionMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindVideoSessionMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindVideoSessionMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindVideoSessionMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindVideoSessionMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindVideoSessionMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindVideoSessionMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindVideoSessionMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindVideoSessionMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindVideoSessionMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindVideoSessionMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindVideoSessionMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindVideoSessionMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_BindVideoSessionMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_BindVideoSessionMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_BindVideoSessionMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindVideoSessionMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindVideoSessionMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindVideoSessionMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindVideoSessionMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindVideoSessionMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindVideoSessionMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindVideoSessionMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindVideoSessionMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindVideoSessionMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindVideoSessionMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindVideoSessionMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindVideoSessionMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindVideoSessionMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_BindVideoSessionMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_BindVideoSessionMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_BindVideoSessionMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindVideoSessionMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindVideoSessionMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindVideoSessionMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindVideoSessionMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindVideoSessionMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindVideoSessionMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindVideoSessionMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindVideoSessionMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindVideoSessionMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindVideoSessionMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindVideoSessionMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindVideoSessionMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindVideoSessionMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_BindVideoSessionMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_BindVideoSessionMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_BindVideoSessionMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindVideoSessionMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindVideoSessionMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindVideoSessionMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindVideoSessionMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindVideoSessionMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindVideoSessionMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindVideoSessionMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindVideoSessionMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindVideoSessionMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindVideoSessionMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindVideoSessionMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindVideoSessionMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindVideoSessionMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_BindVideoSessionMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_BindVideoSessionMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_BindVideoSessionMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindVideoSessionMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindVideoSessionMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindVideoSessionMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindVideoSessionMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindVideoSessionMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindVideoSessionMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindVideoSessionMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindVideoSessionMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindVideoSessionMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindVideoSessionMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindVideoSessionMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindVideoSessionMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindVideoSessionMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_BindVideoSessionMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_BindVideoSessionMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindVideoSessionMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindVideoSessionMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindVideoSessionMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindVideoSessionMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindVideoSessionMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindVideoSessionMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindVideoSessionMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindVideoSessionMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindVideoSessionMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindVideoSessionMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindVideoSessionMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindVideoSessionMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindVideoSessionMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindVideoSessionMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_BindVideoSessionMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_BindVideoSessionMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_BindVideoSessionMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindVideoSessionMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindVideoSessionMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindVideoSessionMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindVideoSessionMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindVideoSessionMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindVideoSessionMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindVideoSessionMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindVideoSessionMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindVideoSessionMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindVideoSessionMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindVideoSessionMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindVideoSessionMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindVideoSessionMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_BindVideoSessionMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_BindVideoSessionMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_BindVideoSessionMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindVideoSessionMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindVideoSessionMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindVideoSessionMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindVideoSessionMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindVideoSessionMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindVideoSessionMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindVideoSessionMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindVideoSessionMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindVideoSessionMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindVideoSessionMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindVideoSessionMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindVideoSessionMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindVideoSessionMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_BindVideoSessionMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_BindVideoSessionMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDecodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDecodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDecodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDecodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDecodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDecodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDecodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDecodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDecodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDecodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDecodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDecodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDecodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDecodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDecodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDecodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDecodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDecodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDecodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDecodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDecodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDecodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDecodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDecodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDecodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDecodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDecodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBeginVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBeginVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBeginVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBeginVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBeginVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBeginVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBeginVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBeginVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBeginVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBeginVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBeginVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBeginVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBeginVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBeginVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBeginVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBeginVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBeginVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBeginVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBeginVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBeginVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBeginVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBeginVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBeginVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBeginVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBeginVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBeginVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBeginVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdControlVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdControlVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdControlVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdControlVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdControlVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdControlVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdControlVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdControlVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdControlVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdControlVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdControlVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdControlVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdControlVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdControlVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdControlVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdControlVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdControlVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdControlVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdControlVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdControlVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdControlVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdControlVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdControlVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdControlVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdControlVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdControlVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdControlVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdControlVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdControlVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdControlVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdControlVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdControlVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdControlVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdControlVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdControlVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdControlVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdControlVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdControlVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdControlVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdControlVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdControlVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdControlVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdControlVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdControlVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdControlVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdControlVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdControlVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdControlVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdControlVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdControlVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdControlVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdControlVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdControlVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdControlVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdControlVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdControlVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdControlVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdControlVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdControlVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdControlVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdControlVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdControlVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdControlVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdControlVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdControlVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdControlVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdControlVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdControlVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdControlVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdControlVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdControlVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdControlVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdControlVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdControlVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdControlVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdControlVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdControlVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdControlVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdControlVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdControlVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdControlVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdControlVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdControlVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdControlVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdControlVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdControlVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdControlVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdControlVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdControlVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdControlVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdControlVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdControlVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdControlVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdControlVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdControlVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdControlVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdControlVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdControlVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdControlVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdControlVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdControlVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdControlVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdControlVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdControlVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdControlVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdControlVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdControlVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdControlVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdControlVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdControlVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdControlVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdControlVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdControlVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdControlVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdControlVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdControlVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdControlVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdControlVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdControlVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdControlVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdControlVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdControlVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdControlVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdControlVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdControlVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdControlVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdControlVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdControlVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdControlVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdControlVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdControlVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdControlVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdControlVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdControlVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdControlVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdControlVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdControlVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdControlVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdControlVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdControlVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdControlVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdControlVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdControlVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdControlVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdControlVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdControlVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdControlVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdControlVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdControlVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdControlVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdControlVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdControlVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdControlVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdControlVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdControlVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdControlVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdControlVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdControlVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdControlVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdControlVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdControlVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdControlVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdControlVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdControlVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdControlVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdControlVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdControlVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdControlVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdControlVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdControlVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdControlVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdControlVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdControlVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdControlVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdControlVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdControlVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdControlVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdControlVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdControlVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdControlVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdControlVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdControlVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdControlVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdControlVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdControlVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdControlVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdControlVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdControlVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdControlVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdControlVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdControlVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdControlVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdControlVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdControlVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdControlVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdControlVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdControlVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdControlVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdControlVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdControlVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdControlVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdControlVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdControlVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdControlVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdControlVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdControlVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdControlVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdControlVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdEndVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdEndVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdEndVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdEndVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdEndVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdEndVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdEndVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdEndVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdEndVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdEndVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdEndVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdEndVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdEndVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdEndVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdEndVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdEndVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdEndVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdEndVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdEndVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdEndVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdEndVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdEndVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdEndVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdEndVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdEndVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndVideoCodingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndVideoCodingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndVideoCodingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndVideoCodingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndVideoCodingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndVideoCodingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndVideoCodingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndVideoCodingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndVideoCodingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndVideoCodingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndVideoCodingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndVideoCodingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndVideoCodingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndVideoCodingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdEndVideoCodingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdEndVideoCodingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdEncodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEncodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEncodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEncodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEncodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEncodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEncodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEncodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEncodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEncodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEncodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEncodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEncodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEncodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdEncodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdEncodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdEncodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEncodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEncodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEncodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEncodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEncodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEncodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEncodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEncodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEncodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEncodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEncodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEncodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEncodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdEncodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdEncodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdEncodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEncodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEncodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEncodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEncodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEncodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEncodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEncodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEncodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEncodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEncodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEncodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEncodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEncodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdEncodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdEncodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdEncodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEncodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEncodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEncodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEncodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEncodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEncodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEncodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEncodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEncodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEncodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEncodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEncodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEncodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdEncodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdEncodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdEncodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEncodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEncodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEncodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEncodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEncodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEncodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEncodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEncodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEncodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEncodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEncodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEncodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEncodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdEncodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdEncodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdEncodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEncodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEncodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEncodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEncodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEncodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEncodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEncodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEncodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEncodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEncodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEncodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEncodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEncodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdEncodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdEncodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdEncodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEncodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEncodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEncodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEncodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEncodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEncodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEncodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEncodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEncodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEncodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEncodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEncodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEncodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdEncodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdEncodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdEncodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEncodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEncodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEncodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEncodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEncodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEncodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEncodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEncodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEncodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEncodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEncodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEncodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEncodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdEncodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdEncodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEncodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEncodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEncodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEncodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEncodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEncodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEncodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEncodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEncodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEncodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEncodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEncodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEncodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEncodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdEncodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdEncodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEncodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEncodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEncodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEncodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEncodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEncodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEncodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEncodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEncodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEncodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEncodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEncodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEncodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEncodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdEncodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdEncodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEncodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEncodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEncodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEncodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEncodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEncodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEncodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEncodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEncodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEncodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEncodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEncodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEncodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEncodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdEncodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdEncodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdEncodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEncodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEncodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEncodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEncodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEncodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEncodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEncodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEncodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEncodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEncodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEncodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEncodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEncodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdEncodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdEncodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEncodeVideoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEncodeVideoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEncodeVideoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEncodeVideoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEncodeVideoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEncodeVideoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEncodeVideoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEncodeVideoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEncodeVideoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEncodeVideoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEncodeVideoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEncodeVideoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEncodeVideoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEncodeVideoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdEncodeVideoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdEncodeVideoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDecompressMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDecompressMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDecompressMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDecompressMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDecompressMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDecompressMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDecompressMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDecompressMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDecompressMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDecompressMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDecompressMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDecompressMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDecompressMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDecompressMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDecompressMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDecompressMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDecompressMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDecompressMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDecompressMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDecompressMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDecompressMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDecompressMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDecompressMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDecompressMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDecompressMemoryNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDecompressMemoryNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDecompressMemoryIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDecompressMemoryIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDecompressMemoryIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDecompressMemoryIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDecompressMemoryIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDecompressMemoryIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDecompressMemoryIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDecompressMemoryIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDecompressMemoryIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDecompressMemoryIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDecompressMemoryIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDecompressMemoryIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDecompressMemoryIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDecompressMemoryIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDecompressMemoryIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDecompressMemoryIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDecompressMemoryIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDecompressMemoryIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDecompressMemoryIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDecompressMemoryIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDecompressMemoryIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDecompressMemoryIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDecompressMemoryIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDecompressMemoryIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDecompressMemoryIndirectCountNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDecompressMemoryIndirectCountNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPartitionedAccelerationStructuresBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPartitionedAccelerationStructuresBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPartitionedAccelerationStructuresBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPartitionedAccelerationStructuresBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPartitionedAccelerationStructuresBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPartitionedAccelerationStructuresBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPartitionedAccelerationStructuresBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPartitionedAccelerationStructuresBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPartitionedAccelerationStructuresBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPartitionedAccelerationStructuresBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPartitionedAccelerationStructuresBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPartitionedAccelerationStructuresBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPartitionedAccelerationStructuresBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPartitionedAccelerationStructuresBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPartitionedAccelerationStructuresBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPartitionedAccelerationStructuresBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetPartitionedAccelerationStructuresBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPartitionedAccelerationStructuresBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPartitionedAccelerationStructuresBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPartitionedAccelerationStructuresBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPartitionedAccelerationStructuresBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPartitionedAccelerationStructuresBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPartitionedAccelerationStructuresBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPartitionedAccelerationStructuresBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPartitionedAccelerationStructuresBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPartitionedAccelerationStructuresBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPartitionedAccelerationStructuresBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPartitionedAccelerationStructuresBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPartitionedAccelerationStructuresBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPartitionedAccelerationStructuresBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetPartitionedAccelerationStructuresBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetPartitionedAccelerationStructuresBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetPartitionedAccelerationStructuresBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPartitionedAccelerationStructuresBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPartitionedAccelerationStructuresBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPartitionedAccelerationStructuresBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPartitionedAccelerationStructuresBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPartitionedAccelerationStructuresBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPartitionedAccelerationStructuresBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPartitionedAccelerationStructuresBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPartitionedAccelerationStructuresBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPartitionedAccelerationStructuresBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPartitionedAccelerationStructuresBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPartitionedAccelerationStructuresBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPartitionedAccelerationStructuresBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPartitionedAccelerationStructuresBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetPartitionedAccelerationStructuresBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetPartitionedAccelerationStructuresBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetPartitionedAccelerationStructuresBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPartitionedAccelerationStructuresBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPartitionedAccelerationStructuresBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPartitionedAccelerationStructuresBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPartitionedAccelerationStructuresBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPartitionedAccelerationStructuresBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPartitionedAccelerationStructuresBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPartitionedAccelerationStructuresBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPartitionedAccelerationStructuresBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPartitionedAccelerationStructuresBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPartitionedAccelerationStructuresBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPartitionedAccelerationStructuresBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPartitionedAccelerationStructuresBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPartitionedAccelerationStructuresBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetPartitionedAccelerationStructuresBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetPartitionedAccelerationStructuresBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetPartitionedAccelerationStructuresBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPartitionedAccelerationStructuresBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPartitionedAccelerationStructuresBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPartitionedAccelerationStructuresBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPartitionedAccelerationStructuresBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPartitionedAccelerationStructuresBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPartitionedAccelerationStructuresBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPartitionedAccelerationStructuresBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPartitionedAccelerationStructuresBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPartitionedAccelerationStructuresBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPartitionedAccelerationStructuresBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPartitionedAccelerationStructuresBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPartitionedAccelerationStructuresBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPartitionedAccelerationStructuresBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetPartitionedAccelerationStructuresBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetPartitionedAccelerationStructuresBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetPartitionedAccelerationStructuresBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPartitionedAccelerationStructuresBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPartitionedAccelerationStructuresBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPartitionedAccelerationStructuresBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPartitionedAccelerationStructuresBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPartitionedAccelerationStructuresBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPartitionedAccelerationStructuresBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPartitionedAccelerationStructuresBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPartitionedAccelerationStructuresBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPartitionedAccelerationStructuresBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPartitionedAccelerationStructuresBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPartitionedAccelerationStructuresBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPartitionedAccelerationStructuresBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPartitionedAccelerationStructuresBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetPartitionedAccelerationStructuresBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetPartitionedAccelerationStructuresBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetPartitionedAccelerationStructuresBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPartitionedAccelerationStructuresBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPartitionedAccelerationStructuresBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPartitionedAccelerationStructuresBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPartitionedAccelerationStructuresBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPartitionedAccelerationStructuresBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPartitionedAccelerationStructuresBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPartitionedAccelerationStructuresBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPartitionedAccelerationStructuresBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPartitionedAccelerationStructuresBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPartitionedAccelerationStructuresBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPartitionedAccelerationStructuresBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPartitionedAccelerationStructuresBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPartitionedAccelerationStructuresBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetPartitionedAccelerationStructuresBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetPartitionedAccelerationStructuresBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetPartitionedAccelerationStructuresBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPartitionedAccelerationStructuresBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPartitionedAccelerationStructuresBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPartitionedAccelerationStructuresBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPartitionedAccelerationStructuresBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPartitionedAccelerationStructuresBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPartitionedAccelerationStructuresBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPartitionedAccelerationStructuresBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPartitionedAccelerationStructuresBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPartitionedAccelerationStructuresBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPartitionedAccelerationStructuresBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPartitionedAccelerationStructuresBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPartitionedAccelerationStructuresBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPartitionedAccelerationStructuresBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetPartitionedAccelerationStructuresBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetPartitionedAccelerationStructuresBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPartitionedAccelerationStructuresBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPartitionedAccelerationStructuresBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPartitionedAccelerationStructuresBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPartitionedAccelerationStructuresBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPartitionedAccelerationStructuresBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPartitionedAccelerationStructuresBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPartitionedAccelerationStructuresBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPartitionedAccelerationStructuresBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPartitionedAccelerationStructuresBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPartitionedAccelerationStructuresBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPartitionedAccelerationStructuresBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPartitionedAccelerationStructuresBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPartitionedAccelerationStructuresBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPartitionedAccelerationStructuresBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetPartitionedAccelerationStructuresBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetPartitionedAccelerationStructuresBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPartitionedAccelerationStructuresBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPartitionedAccelerationStructuresBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPartitionedAccelerationStructuresBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPartitionedAccelerationStructuresBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPartitionedAccelerationStructuresBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPartitionedAccelerationStructuresBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPartitionedAccelerationStructuresBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPartitionedAccelerationStructuresBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPartitionedAccelerationStructuresBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPartitionedAccelerationStructuresBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPartitionedAccelerationStructuresBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPartitionedAccelerationStructuresBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPartitionedAccelerationStructuresBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPartitionedAccelerationStructuresBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetPartitionedAccelerationStructuresBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetPartitionedAccelerationStructuresBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPartitionedAccelerationStructuresBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPartitionedAccelerationStructuresBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPartitionedAccelerationStructuresBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPartitionedAccelerationStructuresBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPartitionedAccelerationStructuresBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPartitionedAccelerationStructuresBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPartitionedAccelerationStructuresBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPartitionedAccelerationStructuresBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPartitionedAccelerationStructuresBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPartitionedAccelerationStructuresBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPartitionedAccelerationStructuresBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPartitionedAccelerationStructuresBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPartitionedAccelerationStructuresBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPartitionedAccelerationStructuresBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetPartitionedAccelerationStructuresBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetPartitionedAccelerationStructuresBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetPartitionedAccelerationStructuresBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPartitionedAccelerationStructuresBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPartitionedAccelerationStructuresBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPartitionedAccelerationStructuresBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPartitionedAccelerationStructuresBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPartitionedAccelerationStructuresBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPartitionedAccelerationStructuresBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPartitionedAccelerationStructuresBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPartitionedAccelerationStructuresBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPartitionedAccelerationStructuresBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPartitionedAccelerationStructuresBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPartitionedAccelerationStructuresBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPartitionedAccelerationStructuresBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPartitionedAccelerationStructuresBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetPartitionedAccelerationStructuresBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetPartitionedAccelerationStructuresBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPartitionedAccelerationStructuresBuildSizesNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPartitionedAccelerationStructuresBuildSizesNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPartitionedAccelerationStructuresBuildSizesNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPartitionedAccelerationStructuresBuildSizesNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPartitionedAccelerationStructuresBuildSizesNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPartitionedAccelerationStructuresBuildSizesNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPartitionedAccelerationStructuresBuildSizesNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPartitionedAccelerationStructuresBuildSizesNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPartitionedAccelerationStructuresBuildSizesNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPartitionedAccelerationStructuresBuildSizesNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPartitionedAccelerationStructuresBuildSizesNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPartitionedAccelerationStructuresBuildSizesNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPartitionedAccelerationStructuresBuildSizesNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPartitionedAccelerationStructuresBuildSizesNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetPartitionedAccelerationStructuresBuildSizesNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetPartitionedAccelerationStructuresBuildSizesNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBuildPartitionedAccelerationStructuresNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildPartitionedAccelerationStructuresNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildPartitionedAccelerationStructuresNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildPartitionedAccelerationStructuresNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildPartitionedAccelerationStructuresNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildPartitionedAccelerationStructuresNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildPartitionedAccelerationStructuresNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildPartitionedAccelerationStructuresNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildPartitionedAccelerationStructuresNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildPartitionedAccelerationStructuresNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildPartitionedAccelerationStructuresNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildPartitionedAccelerationStructuresNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildPartitionedAccelerationStructuresNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildPartitionedAccelerationStructuresNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBuildPartitionedAccelerationStructuresNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBuildPartitionedAccelerationStructuresNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildPartitionedAccelerationStructuresNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildPartitionedAccelerationStructuresNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildPartitionedAccelerationStructuresNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildPartitionedAccelerationStructuresNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildPartitionedAccelerationStructuresNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildPartitionedAccelerationStructuresNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildPartitionedAccelerationStructuresNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildPartitionedAccelerationStructuresNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildPartitionedAccelerationStructuresNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildPartitionedAccelerationStructuresNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildPartitionedAccelerationStructuresNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildPartitionedAccelerationStructuresNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildPartitionedAccelerationStructuresNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildPartitionedAccelerationStructuresNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBuildPartitionedAccelerationStructuresNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBuildPartitionedAccelerationStructuresNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildPartitionedAccelerationStructuresNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildPartitionedAccelerationStructuresNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildPartitionedAccelerationStructuresNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildPartitionedAccelerationStructuresNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildPartitionedAccelerationStructuresNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildPartitionedAccelerationStructuresNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildPartitionedAccelerationStructuresNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildPartitionedAccelerationStructuresNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildPartitionedAccelerationStructuresNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildPartitionedAccelerationStructuresNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildPartitionedAccelerationStructuresNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildPartitionedAccelerationStructuresNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildPartitionedAccelerationStructuresNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildPartitionedAccelerationStructuresNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBuildPartitionedAccelerationStructuresNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBuildPartitionedAccelerationStructuresNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildPartitionedAccelerationStructuresNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildPartitionedAccelerationStructuresNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildPartitionedAccelerationStructuresNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildPartitionedAccelerationStructuresNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildPartitionedAccelerationStructuresNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildPartitionedAccelerationStructuresNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildPartitionedAccelerationStructuresNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildPartitionedAccelerationStructuresNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildPartitionedAccelerationStructuresNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildPartitionedAccelerationStructuresNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildPartitionedAccelerationStructuresNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildPartitionedAccelerationStructuresNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildPartitionedAccelerationStructuresNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildPartitionedAccelerationStructuresNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBuildPartitionedAccelerationStructuresNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBuildPartitionedAccelerationStructuresNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildPartitionedAccelerationStructuresNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildPartitionedAccelerationStructuresNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildPartitionedAccelerationStructuresNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildPartitionedAccelerationStructuresNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildPartitionedAccelerationStructuresNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildPartitionedAccelerationStructuresNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildPartitionedAccelerationStructuresNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildPartitionedAccelerationStructuresNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildPartitionedAccelerationStructuresNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildPartitionedAccelerationStructuresNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildPartitionedAccelerationStructuresNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildPartitionedAccelerationStructuresNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildPartitionedAccelerationStructuresNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildPartitionedAccelerationStructuresNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBuildPartitionedAccelerationStructuresNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBuildPartitionedAccelerationStructuresNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildPartitionedAccelerationStructuresNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildPartitionedAccelerationStructuresNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildPartitionedAccelerationStructuresNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildPartitionedAccelerationStructuresNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildPartitionedAccelerationStructuresNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildPartitionedAccelerationStructuresNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildPartitionedAccelerationStructuresNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildPartitionedAccelerationStructuresNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildPartitionedAccelerationStructuresNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildPartitionedAccelerationStructuresNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildPartitionedAccelerationStructuresNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildPartitionedAccelerationStructuresNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildPartitionedAccelerationStructuresNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildPartitionedAccelerationStructuresNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBuildPartitionedAccelerationStructuresNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBuildPartitionedAccelerationStructuresNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildPartitionedAccelerationStructuresNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildPartitionedAccelerationStructuresNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildPartitionedAccelerationStructuresNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildPartitionedAccelerationStructuresNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildPartitionedAccelerationStructuresNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildPartitionedAccelerationStructuresNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildPartitionedAccelerationStructuresNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildPartitionedAccelerationStructuresNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildPartitionedAccelerationStructuresNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildPartitionedAccelerationStructuresNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildPartitionedAccelerationStructuresNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildPartitionedAccelerationStructuresNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildPartitionedAccelerationStructuresNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildPartitionedAccelerationStructuresNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBuildPartitionedAccelerationStructuresNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBuildPartitionedAccelerationStructuresNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildPartitionedAccelerationStructuresNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildPartitionedAccelerationStructuresNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildPartitionedAccelerationStructuresNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildPartitionedAccelerationStructuresNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildPartitionedAccelerationStructuresNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildPartitionedAccelerationStructuresNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildPartitionedAccelerationStructuresNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildPartitionedAccelerationStructuresNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildPartitionedAccelerationStructuresNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildPartitionedAccelerationStructuresNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildPartitionedAccelerationStructuresNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildPartitionedAccelerationStructuresNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildPartitionedAccelerationStructuresNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildPartitionedAccelerationStructuresNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBuildPartitionedAccelerationStructuresNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBuildPartitionedAccelerationStructuresNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildPartitionedAccelerationStructuresNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildPartitionedAccelerationStructuresNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildPartitionedAccelerationStructuresNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildPartitionedAccelerationStructuresNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildPartitionedAccelerationStructuresNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildPartitionedAccelerationStructuresNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildPartitionedAccelerationStructuresNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildPartitionedAccelerationStructuresNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildPartitionedAccelerationStructuresNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildPartitionedAccelerationStructuresNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildPartitionedAccelerationStructuresNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildPartitionedAccelerationStructuresNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildPartitionedAccelerationStructuresNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildPartitionedAccelerationStructuresNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBuildPartitionedAccelerationStructuresNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBuildPartitionedAccelerationStructuresNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildPartitionedAccelerationStructuresNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildPartitionedAccelerationStructuresNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildPartitionedAccelerationStructuresNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildPartitionedAccelerationStructuresNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildPartitionedAccelerationStructuresNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildPartitionedAccelerationStructuresNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildPartitionedAccelerationStructuresNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildPartitionedAccelerationStructuresNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildPartitionedAccelerationStructuresNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildPartitionedAccelerationStructuresNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildPartitionedAccelerationStructuresNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildPartitionedAccelerationStructuresNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildPartitionedAccelerationStructuresNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildPartitionedAccelerationStructuresNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBuildPartitionedAccelerationStructuresNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBuildPartitionedAccelerationStructuresNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildPartitionedAccelerationStructuresNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildPartitionedAccelerationStructuresNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildPartitionedAccelerationStructuresNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildPartitionedAccelerationStructuresNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildPartitionedAccelerationStructuresNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildPartitionedAccelerationStructuresNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildPartitionedAccelerationStructuresNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildPartitionedAccelerationStructuresNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildPartitionedAccelerationStructuresNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildPartitionedAccelerationStructuresNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildPartitionedAccelerationStructuresNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildPartitionedAccelerationStructuresNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildPartitionedAccelerationStructuresNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildPartitionedAccelerationStructuresNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBuildPartitionedAccelerationStructuresNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBuildPartitionedAccelerationStructuresNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildPartitionedAccelerationStructuresNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildPartitionedAccelerationStructuresNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildPartitionedAccelerationStructuresNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildPartitionedAccelerationStructuresNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildPartitionedAccelerationStructuresNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildPartitionedAccelerationStructuresNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildPartitionedAccelerationStructuresNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildPartitionedAccelerationStructuresNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildPartitionedAccelerationStructuresNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildPartitionedAccelerationStructuresNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildPartitionedAccelerationStructuresNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildPartitionedAccelerationStructuresNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildPartitionedAccelerationStructuresNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildPartitionedAccelerationStructuresNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBuildPartitionedAccelerationStructuresNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBuildPartitionedAccelerationStructuresNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildPartitionedAccelerationStructuresNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildPartitionedAccelerationStructuresNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildPartitionedAccelerationStructuresNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildPartitionedAccelerationStructuresNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildPartitionedAccelerationStructuresNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildPartitionedAccelerationStructuresNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildPartitionedAccelerationStructuresNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildPartitionedAccelerationStructuresNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildPartitionedAccelerationStructuresNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildPartitionedAccelerationStructuresNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildPartitionedAccelerationStructuresNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildPartitionedAccelerationStructuresNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildPartitionedAccelerationStructuresNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildPartitionedAccelerationStructuresNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBuildPartitionedAccelerationStructuresNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBuildPartitionedAccelerationStructuresNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDecompressMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDecompressMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDecompressMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDecompressMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDecompressMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDecompressMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDecompressMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDecompressMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDecompressMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDecompressMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDecompressMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDecompressMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDecompressMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDecompressMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDecompressMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDecompressMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDecompressMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDecompressMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDecompressMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDecompressMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDecompressMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDecompressMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDecompressMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDecompressMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDecompressMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDecompressMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDecompressMemoryIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDecompressMemoryIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDecompressMemoryIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDecompressMemoryIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDecompressMemoryIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDecompressMemoryIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDecompressMemoryIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDecompressMemoryIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDecompressMemoryIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDecompressMemoryIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDecompressMemoryIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDecompressMemoryIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDecompressMemoryIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDecompressMemoryIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDecompressMemoryIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDecompressMemoryIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDecompressMemoryIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDecompressMemoryIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDecompressMemoryIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDecompressMemoryIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDecompressMemoryIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDecompressMemoryIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDecompressMemoryIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDecompressMemoryIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDecompressMemoryIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDecompressMemoryIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDecompressMemoryIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDecompressMemoryIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDecompressMemoryIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDecompressMemoryIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDecompressMemoryIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDecompressMemoryIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDecompressMemoryIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDecompressMemoryIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDecompressMemoryIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDecompressMemoryIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDecompressMemoryIndirectCountEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDecompressMemoryIndirectCountEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDecompressMemoryIndirectCountEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuModuleNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuModuleNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuModuleNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuModuleNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuModuleNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuModuleNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuModuleNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuModuleNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuModuleNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuModuleNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuModuleNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuModuleNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuModuleNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuModuleNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyCuModuleNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyCuModuleNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuFunctionNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuFunctionNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuFunctionNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuFunctionNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuFunctionNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuFunctionNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuFunctionNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuFunctionNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuFunctionNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuFunctionNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuFunctionNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuFunctionNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuFunctionNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyCuFunctionNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyCuFunctionNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyCuFunctionNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCuLaunchKernelNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCuLaunchKernelNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCuLaunchKernelNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCuLaunchKernelNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCuLaunchKernelNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCuLaunchKernelNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCuLaunchKernelNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCuLaunchKernelNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCuLaunchKernelNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCuLaunchKernelNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCuLaunchKernelNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCuLaunchKernelNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCuLaunchKernelNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCuLaunchKernelNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCuLaunchKernelNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCuLaunchKernelNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCuLaunchKernelNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCuLaunchKernelNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCuLaunchKernelNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCuLaunchKernelNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCuLaunchKernelNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCuLaunchKernelNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCuLaunchKernelNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCuLaunchKernelNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCuLaunchKernelNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCuLaunchKernelNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCuLaunchKernelNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCuLaunchKernelNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCuLaunchKernelNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCuLaunchKernelNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCuLaunchKernelNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCuLaunchKernelNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCuLaunchKernelNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCuLaunchKernelNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCuLaunchKernelNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCuLaunchKernelNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCuLaunchKernelNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCuLaunchKernelNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCuLaunchKernelNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCuLaunchKernelNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCuLaunchKernelNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCuLaunchKernelNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCuLaunchKernelNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCuLaunchKernelNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCuLaunchKernelNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCuLaunchKernelNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCuLaunchKernelNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCuLaunchKernelNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCuLaunchKernelNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCuLaunchKernelNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCuLaunchKernelNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCuLaunchKernelNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCuLaunchKernelNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCuLaunchKernelNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCuLaunchKernelNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCuLaunchKernelNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCuLaunchKernelNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCuLaunchKernelNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCuLaunchKernelNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCuLaunchKernelNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCuLaunchKernelNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCuLaunchKernelNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCuLaunchKernelNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCuLaunchKernelNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCuLaunchKernelNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCuLaunchKernelNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCuLaunchKernelNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCuLaunchKernelNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCuLaunchKernelNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCuLaunchKernelNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCuLaunchKernelNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCuLaunchKernelNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCuLaunchKernelNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCuLaunchKernelNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCuLaunchKernelNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCuLaunchKernelNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCuLaunchKernelNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCuLaunchKernelNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCuLaunchKernelNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCuLaunchKernelNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCuLaunchKernelNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCuLaunchKernelNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCuLaunchKernelNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCuLaunchKernelNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCuLaunchKernelNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCuLaunchKernelNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCuLaunchKernelNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCuLaunchKernelNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCuLaunchKernelNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCuLaunchKernelNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCuLaunchKernelNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCuLaunchKernelNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCuLaunchKernelNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCuLaunchKernelNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCuLaunchKernelNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCuLaunchKernelNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCuLaunchKernelNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCuLaunchKernelNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCuLaunchKernelNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCuLaunchKernelNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCuLaunchKernelNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCuLaunchKernelNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCuLaunchKernelNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCuLaunchKernelNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCuLaunchKernelNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCuLaunchKernelNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCuLaunchKernelNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCuLaunchKernelNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCuLaunchKernelNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCuLaunchKernelNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCuLaunchKernelNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCuLaunchKernelNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCuLaunchKernelNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCuLaunchKernelNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCuLaunchKernelNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCuLaunchKernelNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCuLaunchKernelNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCuLaunchKernelNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCuLaunchKernelNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCuLaunchKernelNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCuLaunchKernelNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCuLaunchKernelNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCuLaunchKernelNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCuLaunchKernelNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCuLaunchKernelNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCuLaunchKernelNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCuLaunchKernelNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCuLaunchKernelNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCuLaunchKernelNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCuLaunchKernelNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCuLaunchKernelNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCuLaunchKernelNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCuLaunchKernelNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCuLaunchKernelNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCuLaunchKernelNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCuLaunchKernelNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCuLaunchKernelNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCuLaunchKernelNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCuLaunchKernelNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCuLaunchKernelNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCuLaunchKernelNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCuLaunchKernelNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCuLaunchKernelNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCuLaunchKernelNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCuLaunchKernelNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCuLaunchKernelNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCuLaunchKernelNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCuLaunchKernelNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCuLaunchKernelNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCuLaunchKernelNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCuLaunchKernelNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCuLaunchKernelNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCuLaunchKernelNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCuLaunchKernelNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCuLaunchKernelNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCuLaunchKernelNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCuLaunchKernelNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCuLaunchKernelNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCuLaunchKernelNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCuLaunchKernelNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCuLaunchKernelNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCuLaunchKernelNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCuLaunchKernelNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCuLaunchKernelNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCuLaunchKernelNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCuLaunchKernelNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCuLaunchKernelNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCuLaunchKernelNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCuLaunchKernelNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCuLaunchKernelNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCuLaunchKernelNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCuLaunchKernelNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCuLaunchKernelNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCuLaunchKernelNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCuLaunchKernelNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCuLaunchKernelNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCuLaunchKernelNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCuLaunchKernelNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCuLaunchKernelNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCuLaunchKernelNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCuLaunchKernelNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCuLaunchKernelNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCuLaunchKernelNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCuLaunchKernelNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCuLaunchKernelNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCuLaunchKernelNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCuLaunchKernelNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCuLaunchKernelNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCuLaunchKernelNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCuLaunchKernelNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCuLaunchKernelNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCuLaunchKernelNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCuLaunchKernelNVX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCuLaunchKernelNVX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCuLaunchKernelNVX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCuLaunchKernelNVX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCuLaunchKernelNVX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCuLaunchKernelNVX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCuLaunchKernelNVX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCuLaunchKernelNVX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCuLaunchKernelNVX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCuLaunchKernelNVX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCuLaunchKernelNVX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCuLaunchKernelNVX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCuLaunchKernelNVX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCuLaunchKernelNVX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCuLaunchKernelNVX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCuLaunchKernelNVX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDescriptorSetLayoutSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDescriptorSetLayoutSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDescriptorSetLayoutSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDescriptorSetLayoutSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDescriptorSetLayoutSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDescriptorSetLayoutSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDescriptorSetLayoutSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDescriptorSetLayoutSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDescriptorSetLayoutSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDescriptorSetLayoutSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDescriptorSetLayoutSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDescriptorSetLayoutSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDescriptorSetLayoutSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDescriptorSetLayoutSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDescriptorSetLayoutSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDescriptorSetLayoutSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDescriptorSetLayoutSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDescriptorSetLayoutSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDescriptorSetLayoutSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDescriptorSetLayoutSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDescriptorSetLayoutSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDescriptorSetLayoutSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDescriptorSetLayoutSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDescriptorSetLayoutSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDescriptorSetLayoutSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDescriptorSetLayoutSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutBindingOffsetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutBindingOffsetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutBindingOffsetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutBindingOffsetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutBindingOffsetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutBindingOffsetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutBindingOffsetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutBindingOffsetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutBindingOffsetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutBindingOffsetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutBindingOffsetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutBindingOffsetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutBindingOffsetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutBindingOffsetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDescriptorSetLayoutBindingOffsetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDescriptorSetLayoutBindingOffsetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutBindingOffsetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutBindingOffsetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutBindingOffsetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutBindingOffsetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutBindingOffsetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutBindingOffsetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutBindingOffsetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutBindingOffsetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutBindingOffsetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutBindingOffsetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutBindingOffsetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutBindingOffsetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutBindingOffsetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutBindingOffsetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDescriptorSetLayoutBindingOffsetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDescriptorSetLayoutBindingOffsetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutBindingOffsetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutBindingOffsetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutBindingOffsetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutBindingOffsetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutBindingOffsetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutBindingOffsetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutBindingOffsetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutBindingOffsetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutBindingOffsetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutBindingOffsetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutBindingOffsetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutBindingOffsetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutBindingOffsetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutBindingOffsetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDescriptorSetLayoutBindingOffsetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDescriptorSetLayoutBindingOffsetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutBindingOffsetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutBindingOffsetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutBindingOffsetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutBindingOffsetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutBindingOffsetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutBindingOffsetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutBindingOffsetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutBindingOffsetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutBindingOffsetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutBindingOffsetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutBindingOffsetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutBindingOffsetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutBindingOffsetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutBindingOffsetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDescriptorSetLayoutBindingOffsetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDescriptorSetLayoutBindingOffsetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutBindingOffsetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutBindingOffsetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutBindingOffsetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutBindingOffsetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutBindingOffsetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutBindingOffsetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutBindingOffsetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutBindingOffsetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutBindingOffsetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutBindingOffsetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutBindingOffsetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutBindingOffsetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutBindingOffsetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutBindingOffsetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDescriptorSetLayoutBindingOffsetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDescriptorSetLayoutBindingOffsetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutBindingOffsetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutBindingOffsetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutBindingOffsetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutBindingOffsetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutBindingOffsetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutBindingOffsetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutBindingOffsetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutBindingOffsetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutBindingOffsetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutBindingOffsetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutBindingOffsetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutBindingOffsetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutBindingOffsetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutBindingOffsetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDescriptorSetLayoutBindingOffsetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDescriptorSetLayoutBindingOffsetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutBindingOffsetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutBindingOffsetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutBindingOffsetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutBindingOffsetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutBindingOffsetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutBindingOffsetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutBindingOffsetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutBindingOffsetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutBindingOffsetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutBindingOffsetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutBindingOffsetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutBindingOffsetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutBindingOffsetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutBindingOffsetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDescriptorSetLayoutBindingOffsetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDescriptorSetLayoutBindingOffsetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutBindingOffsetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutBindingOffsetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutBindingOffsetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutBindingOffsetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutBindingOffsetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutBindingOffsetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutBindingOffsetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutBindingOffsetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutBindingOffsetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutBindingOffsetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutBindingOffsetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutBindingOffsetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutBindingOffsetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutBindingOffsetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDescriptorSetLayoutBindingOffsetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDescriptorSetLayoutBindingOffsetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutBindingOffsetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutBindingOffsetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutBindingOffsetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutBindingOffsetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutBindingOffsetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutBindingOffsetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutBindingOffsetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutBindingOffsetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutBindingOffsetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutBindingOffsetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutBindingOffsetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutBindingOffsetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutBindingOffsetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutBindingOffsetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDescriptorSetLayoutBindingOffsetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDescriptorSetLayoutBindingOffsetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutBindingOffsetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutBindingOffsetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutBindingOffsetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutBindingOffsetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutBindingOffsetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutBindingOffsetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutBindingOffsetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutBindingOffsetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutBindingOffsetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutBindingOffsetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutBindingOffsetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutBindingOffsetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutBindingOffsetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutBindingOffsetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDescriptorSetLayoutBindingOffsetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDescriptorSetLayoutBindingOffsetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutBindingOffsetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutBindingOffsetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutBindingOffsetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutBindingOffsetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutBindingOffsetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutBindingOffsetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutBindingOffsetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutBindingOffsetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutBindingOffsetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutBindingOffsetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutBindingOffsetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutBindingOffsetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutBindingOffsetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutBindingOffsetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDescriptorSetLayoutBindingOffsetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDescriptorSetLayoutBindingOffsetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutBindingOffsetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutBindingOffsetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutBindingOffsetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutBindingOffsetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutBindingOffsetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutBindingOffsetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutBindingOffsetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutBindingOffsetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutBindingOffsetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutBindingOffsetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutBindingOffsetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutBindingOffsetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutBindingOffsetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutBindingOffsetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDescriptorSetLayoutBindingOffsetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDescriptorSetLayoutBindingOffsetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutBindingOffsetEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutBindingOffsetEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutBindingOffsetEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutBindingOffsetEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutBindingOffsetEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutBindingOffsetEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutBindingOffsetEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutBindingOffsetEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutBindingOffsetEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutBindingOffsetEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutBindingOffsetEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutBindingOffsetEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutBindingOffsetEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutBindingOffsetEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDescriptorSetLayoutBindingOffsetEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDescriptorSetLayoutBindingOffsetEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDescriptorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDescriptorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDescriptorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDescriptorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDescriptorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDescriptorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDescriptorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDescriptorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDescriptorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDescriptorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDescriptorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDescriptorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDescriptorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDescriptorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDescriptorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDescriptorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDescriptorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDescriptorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDescriptorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDescriptorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDescriptorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDescriptorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDescriptorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDescriptorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDescriptorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDescriptorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindDescriptorBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindDescriptorBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindDescriptorBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindDescriptorBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindDescriptorBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindDescriptorBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindDescriptorBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindDescriptorBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindDescriptorBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindDescriptorBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindDescriptorBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindDescriptorBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindDescriptorBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindDescriptorBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindDescriptorBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindDescriptorBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindDescriptorBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindDescriptorBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindDescriptorBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindDescriptorBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindDescriptorBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindDescriptorBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindDescriptorBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindDescriptorBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBuffersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBuffersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBuffersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBuffersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBuffersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBuffersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBuffersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBuffersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBuffersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBuffersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBuffersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBuffersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBuffersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBuffersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindDescriptorBuffersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindDescriptorBuffersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsetsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsetsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsetsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsetsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsetsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsetsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsetsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsetsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsetsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsetsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsetsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsetsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsetsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsetsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDescriptorBufferOffsetsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDescriptorBufferOffsetsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsetsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsetsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsetsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsetsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsetsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsetsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsetsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsetsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsetsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsetsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsetsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsetsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsetsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsetsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDescriptorBufferOffsetsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDescriptorBufferOffsetsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsetsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsetsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsetsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsetsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsetsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsetsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsetsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsetsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsetsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsetsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsetsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsetsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsetsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsetsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDescriptorBufferOffsetsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDescriptorBufferOffsetsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsetsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsetsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsetsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsetsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsetsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsetsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsetsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsetsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsetsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsetsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsetsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsetsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsetsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsetsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDescriptorBufferOffsetsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDescriptorBufferOffsetsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsetsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsetsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsetsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsetsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsetsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsetsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsetsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsetsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsetsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsetsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsetsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsetsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsetsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsetsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDescriptorBufferOffsetsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDescriptorBufferOffsetsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsetsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsetsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsetsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsetsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsetsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsetsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsetsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsetsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsetsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsetsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsetsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsetsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsetsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsetsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDescriptorBufferOffsetsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDescriptorBufferOffsetsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsetsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsetsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsetsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsetsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsetsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsetsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsetsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsetsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsetsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsetsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsetsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsetsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsetsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsetsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDescriptorBufferOffsetsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDescriptorBufferOffsetsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsetsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsetsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsetsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsetsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsetsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsetsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsetsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsetsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsetsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsetsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsetsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsetsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsetsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsetsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDescriptorBufferOffsetsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDescriptorBufferOffsetsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsetsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsetsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsetsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsetsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsetsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsetsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsetsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsetsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsetsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsetsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsetsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsetsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsetsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsetsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDescriptorBufferOffsetsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDescriptorBufferOffsetsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsetsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsetsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsetsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsetsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsetsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsetsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsetsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsetsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsetsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsetsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsetsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsetsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsetsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsetsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDescriptorBufferOffsetsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDescriptorBufferOffsetsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsetsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsetsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsetsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsetsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsetsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsetsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsetsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsetsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsetsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsetsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsetsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsetsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsetsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsetsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDescriptorBufferOffsetsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDescriptorBufferOffsetsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsetsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsetsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsetsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsetsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsetsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsetsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsetsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsetsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsetsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsetsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsetsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsetsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsetsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsetsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDescriptorBufferOffsetsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDescriptorBufferOffsetsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsetsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsetsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsetsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsetsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsetsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsetsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsetsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsetsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsetsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsetsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsetsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsetsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsetsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsetsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDescriptorBufferOffsetsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDescriptorBufferOffsetsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindDescriptorBufferEmbeddedSamplersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindDescriptorBufferEmbeddedSamplersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindDescriptorBufferEmbeddedSamplersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindDescriptorBufferEmbeddedSamplersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindDescriptorBufferEmbeddedSamplersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindDescriptorBufferEmbeddedSamplersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindDescriptorBufferEmbeddedSamplersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindDescriptorBufferEmbeddedSamplersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindDescriptorBufferEmbeddedSamplersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindDescriptorBufferEmbeddedSamplersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindDescriptorBufferEmbeddedSamplersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindDescriptorBufferEmbeddedSamplersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindDescriptorBufferEmbeddedSamplersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindDescriptorBufferEmbeddedSamplersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindDescriptorBufferEmbeddedSamplersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindDescriptorBufferEmbeddedSamplersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindDescriptorBufferEmbeddedSamplersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindDescriptorBufferEmbeddedSamplersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindDescriptorBufferEmbeddedSamplersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindDescriptorBufferEmbeddedSamplersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindDescriptorBufferEmbeddedSamplersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindDescriptorBufferEmbeddedSamplersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindDescriptorBufferEmbeddedSamplersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindDescriptorBufferEmbeddedSamplersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetBufferOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetBufferOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetBufferOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetBufferOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetBufferOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetBufferOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetBufferOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetBufferOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetBufferOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetBufferOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetBufferOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetBufferOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetBufferOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetBufferOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetBufferOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetBufferOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetBufferOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetBufferOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetBufferOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetBufferOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetBufferOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetBufferOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetBufferOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetBufferOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetBufferOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetBufferOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetImageOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetImageOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetImageOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetImageOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetImageOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetImageOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetImageOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetImageOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetImageOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetImageOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetImageOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetImageOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetImageOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetImageOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetImageOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetImageOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetImageOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetImageOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetImageOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetImageOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetImageOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetImageOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetImageOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetImageOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetImageOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetImageOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetImageViewOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageViewOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetImageViewOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetImageViewOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageViewOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetImageViewOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetImageViewOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageViewOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetImageViewOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetImageViewOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageViewOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetImageViewOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetImageViewOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageViewOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetImageViewOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetImageViewOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageViewOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetImageViewOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetImageViewOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageViewOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetImageViewOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetImageViewOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageViewOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetImageViewOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetImageViewOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageViewOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetImageViewOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetImageViewOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageViewOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetImageViewOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetImageViewOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageViewOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetImageViewOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetImageViewOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageViewOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetImageViewOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetImageViewOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageViewOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetImageViewOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetImageViewOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetSamplerOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSamplerOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSamplerOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSamplerOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSamplerOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSamplerOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSamplerOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSamplerOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSamplerOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSamplerOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSamplerOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSamplerOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSamplerOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSamplerOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetSamplerOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetSamplerOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetSamplerOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSamplerOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSamplerOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSamplerOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSamplerOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSamplerOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSamplerOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSamplerOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSamplerOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSamplerOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSamplerOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSamplerOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSamplerOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSamplerOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetSamplerOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetSamplerOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetSamplerOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSamplerOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSamplerOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSamplerOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSamplerOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSamplerOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSamplerOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSamplerOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSamplerOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSamplerOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSamplerOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSamplerOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSamplerOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSamplerOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetSamplerOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetSamplerOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetSamplerOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSamplerOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSamplerOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSamplerOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSamplerOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSamplerOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSamplerOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSamplerOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSamplerOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSamplerOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSamplerOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSamplerOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSamplerOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSamplerOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetSamplerOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetSamplerOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetSamplerOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSamplerOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSamplerOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSamplerOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSamplerOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSamplerOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSamplerOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSamplerOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSamplerOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSamplerOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSamplerOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSamplerOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSamplerOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSamplerOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetSamplerOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetSamplerOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetSamplerOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSamplerOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSamplerOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSamplerOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSamplerOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSamplerOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSamplerOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSamplerOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSamplerOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSamplerOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSamplerOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSamplerOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSamplerOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSamplerOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetSamplerOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetSamplerOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetSamplerOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSamplerOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSamplerOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSamplerOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSamplerOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSamplerOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSamplerOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSamplerOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSamplerOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSamplerOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSamplerOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSamplerOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSamplerOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSamplerOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetSamplerOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetSamplerOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetSamplerOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSamplerOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSamplerOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSamplerOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSamplerOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSamplerOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSamplerOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSamplerOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSamplerOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSamplerOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSamplerOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSamplerOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSamplerOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSamplerOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetSamplerOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetSamplerOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSamplerOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSamplerOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSamplerOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSamplerOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSamplerOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSamplerOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSamplerOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSamplerOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSamplerOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSamplerOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSamplerOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSamplerOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSamplerOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSamplerOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetSamplerOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetSamplerOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSamplerOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSamplerOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSamplerOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSamplerOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSamplerOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSamplerOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSamplerOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSamplerOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSamplerOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSamplerOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSamplerOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSamplerOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSamplerOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSamplerOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetSamplerOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetSamplerOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSamplerOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSamplerOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSamplerOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSamplerOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSamplerOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSamplerOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSamplerOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSamplerOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSamplerOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSamplerOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSamplerOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSamplerOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSamplerOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSamplerOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetSamplerOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetSamplerOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetSamplerOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSamplerOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSamplerOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSamplerOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSamplerOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSamplerOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSamplerOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSamplerOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSamplerOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSamplerOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSamplerOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSamplerOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSamplerOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSamplerOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetSamplerOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetSamplerOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSamplerOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSamplerOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSamplerOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSamplerOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSamplerOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSamplerOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSamplerOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSamplerOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSamplerOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSamplerOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSamplerOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSamplerOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSamplerOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSamplerOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetSamplerOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetSamplerOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_SetDeviceMemoryPriorityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDeviceMemoryPriorityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDeviceMemoryPriorityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDeviceMemoryPriorityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDeviceMemoryPriorityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDeviceMemoryPriorityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDeviceMemoryPriorityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDeviceMemoryPriorityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDeviceMemoryPriorityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDeviceMemoryPriorityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDeviceMemoryPriorityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDeviceMemoryPriorityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDeviceMemoryPriorityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetDeviceMemoryPriorityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_SetDeviceMemoryPriorityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_SetDeviceMemoryPriorityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_SetDeviceMemoryPriorityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDeviceMemoryPriorityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDeviceMemoryPriorityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDeviceMemoryPriorityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDeviceMemoryPriorityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDeviceMemoryPriorityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDeviceMemoryPriorityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDeviceMemoryPriorityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDeviceMemoryPriorityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDeviceMemoryPriorityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDeviceMemoryPriorityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDeviceMemoryPriorityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDeviceMemoryPriorityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetDeviceMemoryPriorityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_SetDeviceMemoryPriorityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_SetDeviceMemoryPriorityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_SetDeviceMemoryPriorityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDeviceMemoryPriorityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDeviceMemoryPriorityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDeviceMemoryPriorityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDeviceMemoryPriorityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDeviceMemoryPriorityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDeviceMemoryPriorityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDeviceMemoryPriorityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDeviceMemoryPriorityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDeviceMemoryPriorityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDeviceMemoryPriorityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDeviceMemoryPriorityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDeviceMemoryPriorityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetDeviceMemoryPriorityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_SetDeviceMemoryPriorityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_SetDeviceMemoryPriorityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_SetDeviceMemoryPriorityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDeviceMemoryPriorityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDeviceMemoryPriorityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDeviceMemoryPriorityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDeviceMemoryPriorityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDeviceMemoryPriorityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDeviceMemoryPriorityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDeviceMemoryPriorityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDeviceMemoryPriorityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDeviceMemoryPriorityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDeviceMemoryPriorityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDeviceMemoryPriorityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDeviceMemoryPriorityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetDeviceMemoryPriorityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_SetDeviceMemoryPriorityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_SetDeviceMemoryPriorityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_SetDeviceMemoryPriorityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDeviceMemoryPriorityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDeviceMemoryPriorityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDeviceMemoryPriorityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDeviceMemoryPriorityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDeviceMemoryPriorityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDeviceMemoryPriorityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDeviceMemoryPriorityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDeviceMemoryPriorityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDeviceMemoryPriorityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDeviceMemoryPriorityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDeviceMemoryPriorityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDeviceMemoryPriorityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetDeviceMemoryPriorityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_SetDeviceMemoryPriorityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_SetDeviceMemoryPriorityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_SetDeviceMemoryPriorityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDeviceMemoryPriorityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDeviceMemoryPriorityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDeviceMemoryPriorityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDeviceMemoryPriorityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDeviceMemoryPriorityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDeviceMemoryPriorityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDeviceMemoryPriorityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDeviceMemoryPriorityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDeviceMemoryPriorityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDeviceMemoryPriorityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDeviceMemoryPriorityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDeviceMemoryPriorityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetDeviceMemoryPriorityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_SetDeviceMemoryPriorityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_SetDeviceMemoryPriorityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_SetDeviceMemoryPriorityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDeviceMemoryPriorityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDeviceMemoryPriorityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDeviceMemoryPriorityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDeviceMemoryPriorityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDeviceMemoryPriorityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDeviceMemoryPriorityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDeviceMemoryPriorityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDeviceMemoryPriorityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDeviceMemoryPriorityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDeviceMemoryPriorityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDeviceMemoryPriorityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDeviceMemoryPriorityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetDeviceMemoryPriorityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_SetDeviceMemoryPriorityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_SetDeviceMemoryPriorityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_SetDeviceMemoryPriorityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDeviceMemoryPriorityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDeviceMemoryPriorityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDeviceMemoryPriorityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDeviceMemoryPriorityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDeviceMemoryPriorityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDeviceMemoryPriorityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDeviceMemoryPriorityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDeviceMemoryPriorityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDeviceMemoryPriorityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDeviceMemoryPriorityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDeviceMemoryPriorityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDeviceMemoryPriorityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetDeviceMemoryPriorityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_SetDeviceMemoryPriorityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_SetDeviceMemoryPriorityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDeviceMemoryPriorityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDeviceMemoryPriorityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDeviceMemoryPriorityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDeviceMemoryPriorityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDeviceMemoryPriorityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDeviceMemoryPriorityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDeviceMemoryPriorityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDeviceMemoryPriorityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDeviceMemoryPriorityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDeviceMemoryPriorityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDeviceMemoryPriorityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDeviceMemoryPriorityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDeviceMemoryPriorityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetDeviceMemoryPriorityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_SetDeviceMemoryPriorityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_SetDeviceMemoryPriorityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDeviceMemoryPriorityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDeviceMemoryPriorityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDeviceMemoryPriorityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDeviceMemoryPriorityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDeviceMemoryPriorityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDeviceMemoryPriorityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDeviceMemoryPriorityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDeviceMemoryPriorityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDeviceMemoryPriorityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDeviceMemoryPriorityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDeviceMemoryPriorityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDeviceMemoryPriorityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDeviceMemoryPriorityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetDeviceMemoryPriorityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_SetDeviceMemoryPriorityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_SetDeviceMemoryPriorityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDeviceMemoryPriorityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDeviceMemoryPriorityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDeviceMemoryPriorityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDeviceMemoryPriorityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDeviceMemoryPriorityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDeviceMemoryPriorityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDeviceMemoryPriorityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDeviceMemoryPriorityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDeviceMemoryPriorityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDeviceMemoryPriorityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDeviceMemoryPriorityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDeviceMemoryPriorityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDeviceMemoryPriorityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetDeviceMemoryPriorityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_SetDeviceMemoryPriorityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_SetDeviceMemoryPriorityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_SetDeviceMemoryPriorityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDeviceMemoryPriorityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDeviceMemoryPriorityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDeviceMemoryPriorityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDeviceMemoryPriorityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDeviceMemoryPriorityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDeviceMemoryPriorityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDeviceMemoryPriorityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDeviceMemoryPriorityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDeviceMemoryPriorityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDeviceMemoryPriorityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDeviceMemoryPriorityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDeviceMemoryPriorityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetDeviceMemoryPriorityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_SetDeviceMemoryPriorityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_SetDeviceMemoryPriorityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDeviceMemoryPriorityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDeviceMemoryPriorityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDeviceMemoryPriorityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDeviceMemoryPriorityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDeviceMemoryPriorityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDeviceMemoryPriorityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDeviceMemoryPriorityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDeviceMemoryPriorityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDeviceMemoryPriorityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDeviceMemoryPriorityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDeviceMemoryPriorityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDeviceMemoryPriorityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDeviceMemoryPriorityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetDeviceMemoryPriorityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_SetDeviceMemoryPriorityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_SetDeviceMemoryPriorityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_WaitForPresent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_WaitForPresent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_WaitForPresent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_WaitForPresent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_WaitForPresent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_WaitForPresent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_WaitForPresent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_WaitForPresent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_WaitForPresent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_WaitForPresent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_WaitForPresent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_WaitForPresent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_WaitForPresent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_WaitForPresent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_WaitForPresent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_WaitForPresent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_WaitForPresent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_WaitForPresent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_WaitForPresent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_WaitForPresent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_WaitForPresent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_WaitForPresent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_WaitForPresent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_WaitForPresent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_WaitForPresent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresent2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresent2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresent2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresent2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresent2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresent2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresent2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresent2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresent2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresent2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresent2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresent2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresent2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresent2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_WaitForPresent2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_WaitForPresent2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_WaitForPresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WaitForPresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_WaitForPresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_WaitForPresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WaitForPresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_WaitForPresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_WaitForPresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WaitForPresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_WaitForPresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_WaitForPresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WaitForPresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_WaitForPresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_WaitForPresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WaitForPresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_WaitForPresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_WaitForPresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WaitForPresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_WaitForPresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_WaitForPresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WaitForPresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_WaitForPresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_WaitForPresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WaitForPresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_WaitForPresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_WaitForPresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WaitForPresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_WaitForPresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_WaitForPresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WaitForPresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_WaitForPresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_WaitForPresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WaitForPresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_WaitForPresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_WaitForPresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WaitForPresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_WaitForPresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_WaitForPresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresentKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresentKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresentKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresentKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresentKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresentKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresentKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresentKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresentKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresentKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresentKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresentKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresentKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WaitForPresentKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_WaitForPresentKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_WaitForPresentKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionBufferConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionBufferConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionBufferConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionBufferConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionBufferConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionBufferConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionBufferConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionBufferConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionBufferConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionBufferConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionBufferConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionBufferConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionBufferConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionBufferConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_SetBufferCollectionBufferConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_SetBufferCollectionBufferConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionBufferConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionBufferConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionBufferConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionBufferConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionBufferConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionBufferConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionBufferConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionBufferConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionBufferConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionBufferConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionBufferConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionBufferConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionBufferConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionBufferConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_SetBufferCollectionBufferConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_SetBufferCollectionBufferConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionBufferConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionBufferConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionBufferConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionBufferConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionBufferConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionBufferConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionBufferConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionBufferConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionBufferConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionBufferConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionBufferConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionBufferConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionBufferConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionBufferConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_SetBufferCollectionBufferConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_SetBufferCollectionBufferConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionBufferConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionBufferConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionBufferConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionBufferConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionBufferConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionBufferConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionBufferConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionBufferConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionBufferConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionBufferConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionBufferConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionBufferConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionBufferConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionBufferConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_SetBufferCollectionBufferConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_SetBufferCollectionBufferConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionBufferConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionBufferConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionBufferConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionBufferConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionBufferConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionBufferConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionBufferConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionBufferConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionBufferConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionBufferConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionBufferConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionBufferConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionBufferConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionBufferConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_SetBufferCollectionBufferConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_SetBufferCollectionBufferConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionBufferConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionBufferConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionBufferConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionBufferConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionBufferConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionBufferConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionBufferConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionBufferConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionBufferConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionBufferConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionBufferConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionBufferConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionBufferConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionBufferConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_SetBufferCollectionBufferConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_SetBufferCollectionBufferConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionBufferConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionBufferConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionBufferConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionBufferConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionBufferConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionBufferConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionBufferConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionBufferConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionBufferConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionBufferConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionBufferConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionBufferConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionBufferConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionBufferConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_SetBufferCollectionBufferConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_SetBufferCollectionBufferConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionBufferConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionBufferConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionBufferConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionBufferConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionBufferConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionBufferConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionBufferConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionBufferConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionBufferConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionBufferConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionBufferConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionBufferConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionBufferConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionBufferConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_SetBufferCollectionBufferConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_SetBufferCollectionBufferConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionBufferConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionBufferConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionBufferConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionBufferConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionBufferConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionBufferConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionBufferConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionBufferConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionBufferConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionBufferConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionBufferConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionBufferConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionBufferConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionBufferConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_SetBufferCollectionBufferConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_SetBufferCollectionBufferConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionBufferConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionBufferConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionBufferConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionBufferConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionBufferConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionBufferConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionBufferConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionBufferConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionBufferConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionBufferConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionBufferConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionBufferConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionBufferConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionBufferConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_SetBufferCollectionBufferConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_SetBufferCollectionBufferConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionBufferConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionBufferConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionBufferConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionBufferConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionBufferConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionBufferConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionBufferConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionBufferConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionBufferConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionBufferConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionBufferConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionBufferConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionBufferConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionBufferConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_SetBufferCollectionBufferConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_SetBufferCollectionBufferConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionBufferConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionBufferConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionBufferConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionBufferConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionBufferConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionBufferConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionBufferConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionBufferConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionBufferConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionBufferConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionBufferConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionBufferConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionBufferConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionBufferConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_SetBufferCollectionBufferConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_SetBufferCollectionBufferConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionBufferConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionBufferConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionBufferConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionBufferConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionBufferConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionBufferConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionBufferConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionBufferConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionBufferConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionBufferConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionBufferConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionBufferConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionBufferConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionBufferConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_SetBufferCollectionBufferConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_SetBufferCollectionBufferConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionImageConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionImageConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionImageConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionImageConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionImageConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionImageConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionImageConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionImageConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionImageConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionImageConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionImageConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionImageConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionImageConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetBufferCollectionImageConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_SetBufferCollectionImageConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_SetBufferCollectionImageConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionImageConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionImageConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionImageConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionImageConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionImageConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionImageConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionImageConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionImageConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionImageConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionImageConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionImageConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionImageConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionImageConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetBufferCollectionImageConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_SetBufferCollectionImageConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_SetBufferCollectionImageConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionImageConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionImageConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionImageConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionImageConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionImageConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionImageConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionImageConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionImageConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionImageConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionImageConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionImageConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionImageConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionImageConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetBufferCollectionImageConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_SetBufferCollectionImageConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_SetBufferCollectionImageConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionImageConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionImageConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionImageConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionImageConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionImageConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionImageConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionImageConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionImageConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionImageConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionImageConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionImageConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionImageConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionImageConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetBufferCollectionImageConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_SetBufferCollectionImageConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_SetBufferCollectionImageConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionImageConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionImageConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionImageConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionImageConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionImageConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionImageConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionImageConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionImageConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionImageConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionImageConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionImageConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionImageConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionImageConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetBufferCollectionImageConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_SetBufferCollectionImageConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_SetBufferCollectionImageConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionImageConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionImageConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionImageConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionImageConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionImageConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionImageConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionImageConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionImageConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionImageConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionImageConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionImageConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionImageConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionImageConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetBufferCollectionImageConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_SetBufferCollectionImageConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_SetBufferCollectionImageConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionImageConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionImageConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionImageConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionImageConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionImageConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionImageConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionImageConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionImageConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionImageConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionImageConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionImageConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionImageConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionImageConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetBufferCollectionImageConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_SetBufferCollectionImageConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_SetBufferCollectionImageConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionImageConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionImageConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionImageConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionImageConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionImageConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionImageConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionImageConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionImageConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionImageConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionImageConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionImageConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionImageConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionImageConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetBufferCollectionImageConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_SetBufferCollectionImageConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_SetBufferCollectionImageConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionImageConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionImageConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionImageConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionImageConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionImageConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionImageConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionImageConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionImageConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionImageConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionImageConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionImageConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionImageConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionImageConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetBufferCollectionImageConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_SetBufferCollectionImageConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_SetBufferCollectionImageConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionImageConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionImageConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionImageConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionImageConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionImageConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionImageConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionImageConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionImageConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionImageConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionImageConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionImageConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionImageConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionImageConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetBufferCollectionImageConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_SetBufferCollectionImageConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_SetBufferCollectionImageConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionImageConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionImageConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionImageConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionImageConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionImageConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionImageConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionImageConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionImageConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionImageConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionImageConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionImageConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionImageConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionImageConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetBufferCollectionImageConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_SetBufferCollectionImageConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_SetBufferCollectionImageConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionImageConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionImageConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionImageConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionImageConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionImageConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionImageConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionImageConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionImageConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionImageConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionImageConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionImageConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionImageConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionImageConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetBufferCollectionImageConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_SetBufferCollectionImageConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_SetBufferCollectionImageConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionImageConstraintsFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionImageConstraintsFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionImageConstraintsFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionImageConstraintsFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionImageConstraintsFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionImageConstraintsFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionImageConstraintsFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionImageConstraintsFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionImageConstraintsFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionImageConstraintsFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionImageConstraintsFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionImageConstraintsFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionImageConstraintsFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetBufferCollectionImageConstraintsFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_SetBufferCollectionImageConstraintsFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_SetBufferCollectionImageConstraintsFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferCollectionFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferCollectionFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferCollectionFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferCollectionFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferCollectionFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferCollectionFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferCollectionFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferCollectionFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferCollectionFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferCollectionFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferCollectionFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferCollectionFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferCollectionFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyBufferCollectionFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyBufferCollectionFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyBufferCollectionFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetBufferCollectionPropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferCollectionPropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferCollectionPropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferCollectionPropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferCollectionPropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferCollectionPropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferCollectionPropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferCollectionPropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferCollectionPropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferCollectionPropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferCollectionPropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferCollectionPropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferCollectionPropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetBufferCollectionPropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetBufferCollectionPropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetBufferCollectionPropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferCollectionPropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferCollectionPropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferCollectionPropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferCollectionPropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferCollectionPropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferCollectionPropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferCollectionPropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferCollectionPropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferCollectionPropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferCollectionPropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferCollectionPropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferCollectionPropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferCollectionPropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetBufferCollectionPropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetBufferCollectionPropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetBufferCollectionPropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferCollectionPropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferCollectionPropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferCollectionPropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferCollectionPropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferCollectionPropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferCollectionPropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferCollectionPropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferCollectionPropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferCollectionPropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferCollectionPropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferCollectionPropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferCollectionPropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferCollectionPropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetBufferCollectionPropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetBufferCollectionPropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetBufferCollectionPropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferCollectionPropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferCollectionPropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferCollectionPropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferCollectionPropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferCollectionPropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferCollectionPropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferCollectionPropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferCollectionPropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferCollectionPropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferCollectionPropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferCollectionPropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferCollectionPropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferCollectionPropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetBufferCollectionPropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetBufferCollectionPropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetBufferCollectionPropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferCollectionPropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferCollectionPropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferCollectionPropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferCollectionPropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferCollectionPropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferCollectionPropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferCollectionPropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferCollectionPropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferCollectionPropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferCollectionPropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferCollectionPropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferCollectionPropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferCollectionPropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetBufferCollectionPropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetBufferCollectionPropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetBufferCollectionPropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferCollectionPropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferCollectionPropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferCollectionPropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferCollectionPropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferCollectionPropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferCollectionPropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferCollectionPropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferCollectionPropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferCollectionPropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferCollectionPropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferCollectionPropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferCollectionPropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferCollectionPropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetBufferCollectionPropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetBufferCollectionPropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetBufferCollectionPropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferCollectionPropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferCollectionPropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferCollectionPropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferCollectionPropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferCollectionPropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferCollectionPropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferCollectionPropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferCollectionPropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferCollectionPropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferCollectionPropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferCollectionPropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferCollectionPropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferCollectionPropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetBufferCollectionPropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetBufferCollectionPropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetBufferCollectionPropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferCollectionPropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferCollectionPropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferCollectionPropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferCollectionPropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferCollectionPropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferCollectionPropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferCollectionPropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferCollectionPropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferCollectionPropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferCollectionPropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferCollectionPropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferCollectionPropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferCollectionPropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetBufferCollectionPropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetBufferCollectionPropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetBufferCollectionPropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferCollectionPropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferCollectionPropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferCollectionPropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferCollectionPropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferCollectionPropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferCollectionPropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferCollectionPropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferCollectionPropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferCollectionPropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferCollectionPropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferCollectionPropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferCollectionPropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferCollectionPropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetBufferCollectionPropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetBufferCollectionPropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetBufferCollectionPropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferCollectionPropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferCollectionPropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferCollectionPropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferCollectionPropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferCollectionPropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferCollectionPropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferCollectionPropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferCollectionPropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferCollectionPropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferCollectionPropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferCollectionPropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferCollectionPropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferCollectionPropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetBufferCollectionPropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetBufferCollectionPropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetBufferCollectionPropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferCollectionPropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferCollectionPropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferCollectionPropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferCollectionPropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferCollectionPropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferCollectionPropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferCollectionPropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferCollectionPropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferCollectionPropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferCollectionPropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferCollectionPropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferCollectionPropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferCollectionPropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetBufferCollectionPropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetBufferCollectionPropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetBufferCollectionPropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferCollectionPropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferCollectionPropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferCollectionPropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferCollectionPropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferCollectionPropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferCollectionPropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferCollectionPropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferCollectionPropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferCollectionPropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferCollectionPropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferCollectionPropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferCollectionPropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferCollectionPropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetBufferCollectionPropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetBufferCollectionPropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetBufferCollectionPropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferCollectionPropertiesFUCHSIA@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferCollectionPropertiesFUCHSIA@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferCollectionPropertiesFUCHSIA@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferCollectionPropertiesFUCHSIA@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferCollectionPropertiesFUCHSIA@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferCollectionPropertiesFUCHSIA@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferCollectionPropertiesFUCHSIA@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferCollectionPropertiesFUCHSIA@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferCollectionPropertiesFUCHSIA@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferCollectionPropertiesFUCHSIA@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferCollectionPropertiesFUCHSIA@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferCollectionPropertiesFUCHSIA@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferCollectionPropertiesFUCHSIA@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetBufferCollectionPropertiesFUCHSIA@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetBufferCollectionPropertiesFUCHSIA=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetBufferCollectionPropertiesFUCHSIA=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBeginRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBeginRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBeginRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBeginRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBeginRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBeginRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBeginRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBeginRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBeginRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBeginRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBeginRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBeginRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBeginRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBeginRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBeginRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBeginRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBeginRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBeginRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBeginRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBeginRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBeginRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBeginRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBeginRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBeginRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBeginRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBeginRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBeginRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBeginRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBeginRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBeginRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBeginRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBeginRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBeginRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBeginRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBeginRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBeginRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBeginRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBeginRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBeginRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBeginRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBeginRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBeginRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBeginRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBeginRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBeginRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBeginRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBeginRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBeginRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBeginRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBeginRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBeginRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBeginRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdEndRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdEndRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdEndRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdEndRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdEndRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdEndRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdEndRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdEndRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdEndRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdEndRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdEndRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdEndRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdEndRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdEndRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdEndRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdEndRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdEndRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdEndRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdEndRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdEndRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdEndRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdEndRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdEndRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdEndRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdEndRendering=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdEndRendering=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdEndRendering2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdEndRendering2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdEndRendering2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdEndRendering2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdEndRendering2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdEndRendering2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdEndRendering2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdEndRendering2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdEndRendering2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdEndRendering2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdEndRendering2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdEndRendering2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdEndRendering2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdEndRendering2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdEndRendering2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdEndRendering2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdEndRendering2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdEndRendering2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdEndRendering2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdEndRendering2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdEndRendering2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdEndRendering2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdEndRendering2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdEndRendering2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdEndRendering2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdEndRendering2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdEndRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdEndRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdEndRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdEndRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdEndRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdEndRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdEndRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdEndRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdEndRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdEndRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdEndRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdEndRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdEndRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdEndRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdEndRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdEndRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdEndRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdEndRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdEndRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdEndRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdEndRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdEndRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdEndRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdEndRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdEndRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdEndRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdEndRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdEndRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdEndRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdEndRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdEndRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdEndRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdEndRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdEndRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdEndRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdEndRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdEndRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdEndRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdEndRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdEndRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdEndRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdEndRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdEndRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdEndRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdEndRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdEndRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdEndRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdEndRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdEndRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdEndRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderingKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderingKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderingKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderingKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderingKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderingKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderingKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderingKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderingKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderingKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderingKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderingKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderingKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndRenderingKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdEndRenderingKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdEndRenderingKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutHostMappingInfoVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutHostMappingInfoVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutHostMappingInfoVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutHostMappingInfoVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutHostMappingInfoVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutHostMappingInfoVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutHostMappingInfoVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutHostMappingInfoVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutHostMappingInfoVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutHostMappingInfoVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutHostMappingInfoVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutHostMappingInfoVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutHostMappingInfoVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetLayoutHostMappingInfoVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDescriptorSetLayoutHostMappingInfoVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDescriptorSetLayoutHostMappingInfoVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutHostMappingInfoVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutHostMappingInfoVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutHostMappingInfoVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutHostMappingInfoVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutHostMappingInfoVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutHostMappingInfoVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutHostMappingInfoVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutHostMappingInfoVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutHostMappingInfoVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutHostMappingInfoVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutHostMappingInfoVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutHostMappingInfoVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutHostMappingInfoVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetLayoutHostMappingInfoVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDescriptorSetLayoutHostMappingInfoVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDescriptorSetLayoutHostMappingInfoVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutHostMappingInfoVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutHostMappingInfoVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutHostMappingInfoVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutHostMappingInfoVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutHostMappingInfoVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutHostMappingInfoVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutHostMappingInfoVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutHostMappingInfoVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutHostMappingInfoVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutHostMappingInfoVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutHostMappingInfoVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutHostMappingInfoVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutHostMappingInfoVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetLayoutHostMappingInfoVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDescriptorSetLayoutHostMappingInfoVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDescriptorSetLayoutHostMappingInfoVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutHostMappingInfoVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutHostMappingInfoVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutHostMappingInfoVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutHostMappingInfoVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutHostMappingInfoVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutHostMappingInfoVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutHostMappingInfoVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutHostMappingInfoVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutHostMappingInfoVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutHostMappingInfoVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutHostMappingInfoVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutHostMappingInfoVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutHostMappingInfoVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetLayoutHostMappingInfoVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDescriptorSetLayoutHostMappingInfoVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDescriptorSetLayoutHostMappingInfoVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutHostMappingInfoVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutHostMappingInfoVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutHostMappingInfoVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutHostMappingInfoVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutHostMappingInfoVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutHostMappingInfoVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutHostMappingInfoVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutHostMappingInfoVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutHostMappingInfoVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutHostMappingInfoVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutHostMappingInfoVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutHostMappingInfoVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutHostMappingInfoVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetLayoutHostMappingInfoVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDescriptorSetLayoutHostMappingInfoVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDescriptorSetLayoutHostMappingInfoVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutHostMappingInfoVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutHostMappingInfoVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutHostMappingInfoVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutHostMappingInfoVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutHostMappingInfoVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutHostMappingInfoVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutHostMappingInfoVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutHostMappingInfoVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutHostMappingInfoVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutHostMappingInfoVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutHostMappingInfoVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutHostMappingInfoVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutHostMappingInfoVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetLayoutHostMappingInfoVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDescriptorSetLayoutHostMappingInfoVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDescriptorSetLayoutHostMappingInfoVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutHostMappingInfoVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutHostMappingInfoVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutHostMappingInfoVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutHostMappingInfoVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutHostMappingInfoVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutHostMappingInfoVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutHostMappingInfoVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutHostMappingInfoVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutHostMappingInfoVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutHostMappingInfoVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutHostMappingInfoVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutHostMappingInfoVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutHostMappingInfoVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetLayoutHostMappingInfoVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDescriptorSetLayoutHostMappingInfoVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDescriptorSetLayoutHostMappingInfoVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutHostMappingInfoVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutHostMappingInfoVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutHostMappingInfoVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutHostMappingInfoVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutHostMappingInfoVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutHostMappingInfoVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutHostMappingInfoVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutHostMappingInfoVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutHostMappingInfoVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutHostMappingInfoVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutHostMappingInfoVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutHostMappingInfoVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutHostMappingInfoVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetLayoutHostMappingInfoVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDescriptorSetLayoutHostMappingInfoVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDescriptorSetLayoutHostMappingInfoVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutHostMappingInfoVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutHostMappingInfoVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutHostMappingInfoVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutHostMappingInfoVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutHostMappingInfoVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutHostMappingInfoVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutHostMappingInfoVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutHostMappingInfoVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutHostMappingInfoVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutHostMappingInfoVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutHostMappingInfoVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutHostMappingInfoVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutHostMappingInfoVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetLayoutHostMappingInfoVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDescriptorSetLayoutHostMappingInfoVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDescriptorSetLayoutHostMappingInfoVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutHostMappingInfoVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutHostMappingInfoVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutHostMappingInfoVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutHostMappingInfoVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutHostMappingInfoVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutHostMappingInfoVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutHostMappingInfoVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutHostMappingInfoVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutHostMappingInfoVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutHostMappingInfoVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutHostMappingInfoVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutHostMappingInfoVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutHostMappingInfoVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetLayoutHostMappingInfoVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDescriptorSetLayoutHostMappingInfoVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDescriptorSetLayoutHostMappingInfoVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutHostMappingInfoVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutHostMappingInfoVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutHostMappingInfoVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutHostMappingInfoVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutHostMappingInfoVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutHostMappingInfoVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutHostMappingInfoVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutHostMappingInfoVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutHostMappingInfoVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutHostMappingInfoVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutHostMappingInfoVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutHostMappingInfoVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutHostMappingInfoVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetLayoutHostMappingInfoVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDescriptorSetLayoutHostMappingInfoVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDescriptorSetLayoutHostMappingInfoVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutHostMappingInfoVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutHostMappingInfoVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutHostMappingInfoVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutHostMappingInfoVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutHostMappingInfoVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutHostMappingInfoVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutHostMappingInfoVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutHostMappingInfoVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutHostMappingInfoVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutHostMappingInfoVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutHostMappingInfoVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutHostMappingInfoVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutHostMappingInfoVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetLayoutHostMappingInfoVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDescriptorSetLayoutHostMappingInfoVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDescriptorSetLayoutHostMappingInfoVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutHostMappingInfoVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutHostMappingInfoVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutHostMappingInfoVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutHostMappingInfoVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutHostMappingInfoVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutHostMappingInfoVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutHostMappingInfoVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutHostMappingInfoVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutHostMappingInfoVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutHostMappingInfoVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutHostMappingInfoVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutHostMappingInfoVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutHostMappingInfoVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetLayoutHostMappingInfoVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDescriptorSetLayoutHostMappingInfoVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDescriptorSetLayoutHostMappingInfoVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetHostMappingVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetHostMappingVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetHostMappingVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetHostMappingVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetHostMappingVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetHostMappingVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetHostMappingVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetHostMappingVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetHostMappingVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetHostMappingVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetHostMappingVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetHostMappingVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetHostMappingVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDescriptorSetHostMappingVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDescriptorSetHostMappingVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDescriptorSetHostMappingVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetHostMappingVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetHostMappingVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetHostMappingVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetHostMappingVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetHostMappingVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetHostMappingVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetHostMappingVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetHostMappingVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetHostMappingVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetHostMappingVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetHostMappingVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetHostMappingVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetHostMappingVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDescriptorSetHostMappingVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDescriptorSetHostMappingVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDescriptorSetHostMappingVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetHostMappingVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetHostMappingVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetHostMappingVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetHostMappingVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetHostMappingVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetHostMappingVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetHostMappingVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetHostMappingVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetHostMappingVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetHostMappingVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetHostMappingVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetHostMappingVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetHostMappingVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDescriptorSetHostMappingVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDescriptorSetHostMappingVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDescriptorSetHostMappingVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetHostMappingVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetHostMappingVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetHostMappingVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetHostMappingVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetHostMappingVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetHostMappingVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetHostMappingVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetHostMappingVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetHostMappingVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetHostMappingVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetHostMappingVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetHostMappingVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetHostMappingVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDescriptorSetHostMappingVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDescriptorSetHostMappingVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDescriptorSetHostMappingVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetHostMappingVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetHostMappingVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetHostMappingVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetHostMappingVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetHostMappingVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetHostMappingVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetHostMappingVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetHostMappingVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetHostMappingVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetHostMappingVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetHostMappingVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetHostMappingVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetHostMappingVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDescriptorSetHostMappingVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDescriptorSetHostMappingVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDescriptorSetHostMappingVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetHostMappingVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetHostMappingVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetHostMappingVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetHostMappingVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetHostMappingVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetHostMappingVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetHostMappingVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetHostMappingVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetHostMappingVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetHostMappingVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetHostMappingVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetHostMappingVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetHostMappingVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDescriptorSetHostMappingVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDescriptorSetHostMappingVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDescriptorSetHostMappingVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetHostMappingVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetHostMappingVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetHostMappingVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetHostMappingVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetHostMappingVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetHostMappingVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetHostMappingVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetHostMappingVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetHostMappingVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetHostMappingVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetHostMappingVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetHostMappingVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetHostMappingVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDescriptorSetHostMappingVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDescriptorSetHostMappingVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDescriptorSetHostMappingVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetHostMappingVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetHostMappingVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetHostMappingVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetHostMappingVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetHostMappingVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetHostMappingVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetHostMappingVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetHostMappingVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetHostMappingVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetHostMappingVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetHostMappingVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetHostMappingVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetHostMappingVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDescriptorSetHostMappingVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDescriptorSetHostMappingVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDescriptorSetHostMappingVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetHostMappingVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetHostMappingVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetHostMappingVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetHostMappingVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetHostMappingVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetHostMappingVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetHostMappingVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetHostMappingVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetHostMappingVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetHostMappingVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetHostMappingVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetHostMappingVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetHostMappingVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDescriptorSetHostMappingVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDescriptorSetHostMappingVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDescriptorSetHostMappingVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetHostMappingVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetHostMappingVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetHostMappingVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetHostMappingVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetHostMappingVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetHostMappingVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetHostMappingVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetHostMappingVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetHostMappingVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetHostMappingVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetHostMappingVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetHostMappingVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetHostMappingVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDescriptorSetHostMappingVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDescriptorSetHostMappingVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDescriptorSetHostMappingVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetHostMappingVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetHostMappingVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetHostMappingVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetHostMappingVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetHostMappingVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetHostMappingVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetHostMappingVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetHostMappingVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetHostMappingVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetHostMappingVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetHostMappingVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetHostMappingVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetHostMappingVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDescriptorSetHostMappingVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDescriptorSetHostMappingVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDescriptorSetHostMappingVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetHostMappingVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetHostMappingVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetHostMappingVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetHostMappingVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetHostMappingVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetHostMappingVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetHostMappingVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetHostMappingVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetHostMappingVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetHostMappingVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetHostMappingVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetHostMappingVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetHostMappingVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDescriptorSetHostMappingVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDescriptorSetHostMappingVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDescriptorSetHostMappingVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetHostMappingVALVE@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetHostMappingVALVE@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetHostMappingVALVE@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetHostMappingVALVE@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetHostMappingVALVE@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetHostMappingVALVE@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetHostMappingVALVE@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetHostMappingVALVE@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetHostMappingVALVE@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetHostMappingVALVE@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetHostMappingVALVE@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetHostMappingVALVE@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetHostMappingVALVE@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDescriptorSetHostMappingVALVE@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDescriptorSetHostMappingVALVE=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDescriptorSetHostMappingVALVE=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_BuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_BuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_BuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_BuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_BuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_BuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_BuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_BuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_BuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_BuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_BuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_BuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_BuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_BuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_BuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_BuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_BuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_BuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_BuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_BuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_BuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_BuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_BuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_BuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_BuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_BuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_BuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_BuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_BuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_BuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_BuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_BuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_BuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildMicromapsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildMicromapsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildMicromapsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildMicromapsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildMicromapsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildMicromapsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildMicromapsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildMicromapsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildMicromapsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildMicromapsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildMicromapsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildMicromapsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildMicromapsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BuildMicromapsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_BuildMicromapsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_BuildMicromapsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CopyMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CopyMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapToMemoryEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapToMemoryEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapToMemoryEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapToMemoryEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapToMemoryEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapToMemoryEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapToMemoryEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapToMemoryEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapToMemoryEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapToMemoryEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapToMemoryEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapToMemoryEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapToMemoryEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMicromapToMemoryEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CopyMicromapToMemoryEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CopyMicromapToMemoryEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToMicromapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToMicromapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToMicromapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToMicromapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToMicromapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToMicromapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToMicromapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToMicromapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToMicromapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToMicromapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToMicromapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToMicromapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToMicromapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CopyMemoryToMicromapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CopyMemoryToMicromapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CopyMemoryToMicromapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdWriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdWriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdWriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdWriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdWriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdWriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdWriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdWriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdWriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdWriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdWriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdWriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdWriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdWriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdWriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdWriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdWriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdWriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdWriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdWriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdWriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdWriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdWriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdWriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdWriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdWriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_WriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_WriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_WriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_WriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_WriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_WriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_WriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_WriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_WriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_WriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_WriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_WriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_WriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_WriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_WriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_WriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_WriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_WriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_WriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_WriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_WriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_WriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_WriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_WriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_WriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_WriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_WriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_WriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_WriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_WriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_WriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_WriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_WriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteMicromapsPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteMicromapsPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteMicromapsPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteMicromapsPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteMicromapsPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteMicromapsPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteMicromapsPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteMicromapsPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteMicromapsPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteMicromapsPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteMicromapsPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteMicromapsPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteMicromapsPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteMicromapsPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_WriteMicromapsPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_WriteMicromapsPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMicromapCompatibilityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMicromapCompatibilityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMicromapCompatibilityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMicromapCompatibilityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMicromapCompatibilityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMicromapCompatibilityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMicromapCompatibilityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMicromapCompatibilityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMicromapCompatibilityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMicromapCompatibilityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMicromapCompatibilityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMicromapCompatibilityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMicromapCompatibilityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceMicromapCompatibilityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceMicromapCompatibilityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceMicromapCompatibilityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMicromapCompatibilityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMicromapCompatibilityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMicromapCompatibilityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMicromapCompatibilityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMicromapCompatibilityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMicromapCompatibilityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMicromapCompatibilityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMicromapCompatibilityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMicromapCompatibilityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMicromapCompatibilityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMicromapCompatibilityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMicromapCompatibilityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMicromapCompatibilityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceMicromapCompatibilityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceMicromapCompatibilityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceMicromapCompatibilityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMicromapCompatibilityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMicromapCompatibilityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMicromapCompatibilityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMicromapCompatibilityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMicromapCompatibilityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMicromapCompatibilityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMicromapCompatibilityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMicromapCompatibilityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMicromapCompatibilityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMicromapCompatibilityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMicromapCompatibilityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMicromapCompatibilityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMicromapCompatibilityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceMicromapCompatibilityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceMicromapCompatibilityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceMicromapCompatibilityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMicromapCompatibilityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMicromapCompatibilityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMicromapCompatibilityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMicromapCompatibilityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMicromapCompatibilityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMicromapCompatibilityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMicromapCompatibilityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMicromapCompatibilityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMicromapCompatibilityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMicromapCompatibilityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMicromapCompatibilityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMicromapCompatibilityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMicromapCompatibilityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceMicromapCompatibilityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceMicromapCompatibilityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceMicromapCompatibilityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMicromapCompatibilityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMicromapCompatibilityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMicromapCompatibilityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMicromapCompatibilityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMicromapCompatibilityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMicromapCompatibilityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMicromapCompatibilityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMicromapCompatibilityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMicromapCompatibilityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMicromapCompatibilityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMicromapCompatibilityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMicromapCompatibilityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMicromapCompatibilityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceMicromapCompatibilityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceMicromapCompatibilityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceMicromapCompatibilityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMicromapCompatibilityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMicromapCompatibilityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMicromapCompatibilityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMicromapCompatibilityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMicromapCompatibilityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMicromapCompatibilityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMicromapCompatibilityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMicromapCompatibilityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMicromapCompatibilityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMicromapCompatibilityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMicromapCompatibilityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMicromapCompatibilityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMicromapCompatibilityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceMicromapCompatibilityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceMicromapCompatibilityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceMicromapCompatibilityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMicromapCompatibilityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMicromapCompatibilityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMicromapCompatibilityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMicromapCompatibilityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMicromapCompatibilityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMicromapCompatibilityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMicromapCompatibilityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMicromapCompatibilityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMicromapCompatibilityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMicromapCompatibilityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMicromapCompatibilityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMicromapCompatibilityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMicromapCompatibilityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceMicromapCompatibilityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceMicromapCompatibilityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceMicromapCompatibilityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMicromapCompatibilityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMicromapCompatibilityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMicromapCompatibilityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMicromapCompatibilityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMicromapCompatibilityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMicromapCompatibilityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMicromapCompatibilityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMicromapCompatibilityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMicromapCompatibilityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMicromapCompatibilityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMicromapCompatibilityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMicromapCompatibilityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMicromapCompatibilityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceMicromapCompatibilityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceMicromapCompatibilityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceMicromapCompatibilityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMicromapCompatibilityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMicromapCompatibilityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMicromapCompatibilityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMicromapCompatibilityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMicromapCompatibilityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMicromapCompatibilityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMicromapCompatibilityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMicromapCompatibilityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMicromapCompatibilityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMicromapCompatibilityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMicromapCompatibilityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMicromapCompatibilityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMicromapCompatibilityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceMicromapCompatibilityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceMicromapCompatibilityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceMicromapCompatibilityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMicromapCompatibilityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMicromapCompatibilityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMicromapCompatibilityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMicromapCompatibilityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMicromapCompatibilityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMicromapCompatibilityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMicromapCompatibilityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMicromapCompatibilityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMicromapCompatibilityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMicromapCompatibilityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMicromapCompatibilityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMicromapCompatibilityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMicromapCompatibilityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceMicromapCompatibilityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceMicromapCompatibilityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceMicromapCompatibilityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMicromapCompatibilityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMicromapCompatibilityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMicromapCompatibilityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMicromapCompatibilityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMicromapCompatibilityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMicromapCompatibilityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMicromapCompatibilityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMicromapCompatibilityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMicromapCompatibilityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMicromapCompatibilityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMicromapCompatibilityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMicromapCompatibilityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMicromapCompatibilityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceMicromapCompatibilityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceMicromapCompatibilityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceMicromapCompatibilityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMicromapCompatibilityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMicromapCompatibilityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMicromapCompatibilityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMicromapCompatibilityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMicromapCompatibilityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMicromapCompatibilityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMicromapCompatibilityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMicromapCompatibilityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMicromapCompatibilityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMicromapCompatibilityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMicromapCompatibilityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMicromapCompatibilityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMicromapCompatibilityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceMicromapCompatibilityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceMicromapCompatibilityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceMicromapCompatibilityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMicromapCompatibilityEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMicromapCompatibilityEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMicromapCompatibilityEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMicromapCompatibilityEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMicromapCompatibilityEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMicromapCompatibilityEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMicromapCompatibilityEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMicromapCompatibilityEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMicromapCompatibilityEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMicromapCompatibilityEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMicromapCompatibilityEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMicromapCompatibilityEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMicromapCompatibilityEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceMicromapCompatibilityEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceMicromapCompatibilityEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceMicromapCompatibilityEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetMicromapBuildSizesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMicromapBuildSizesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMicromapBuildSizesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMicromapBuildSizesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMicromapBuildSizesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMicromapBuildSizesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMicromapBuildSizesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMicromapBuildSizesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMicromapBuildSizesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMicromapBuildSizesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMicromapBuildSizesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMicromapBuildSizesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMicromapBuildSizesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMicromapBuildSizesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetMicromapBuildSizesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetMicromapBuildSizesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetMicromapBuildSizesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMicromapBuildSizesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMicromapBuildSizesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMicromapBuildSizesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMicromapBuildSizesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMicromapBuildSizesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMicromapBuildSizesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMicromapBuildSizesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMicromapBuildSizesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMicromapBuildSizesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMicromapBuildSizesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMicromapBuildSizesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMicromapBuildSizesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMicromapBuildSizesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetMicromapBuildSizesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetMicromapBuildSizesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetMicromapBuildSizesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMicromapBuildSizesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMicromapBuildSizesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMicromapBuildSizesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMicromapBuildSizesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMicromapBuildSizesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMicromapBuildSizesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMicromapBuildSizesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMicromapBuildSizesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMicromapBuildSizesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMicromapBuildSizesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMicromapBuildSizesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMicromapBuildSizesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMicromapBuildSizesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetMicromapBuildSizesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetMicromapBuildSizesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetMicromapBuildSizesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMicromapBuildSizesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMicromapBuildSizesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMicromapBuildSizesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMicromapBuildSizesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMicromapBuildSizesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMicromapBuildSizesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMicromapBuildSizesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMicromapBuildSizesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMicromapBuildSizesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMicromapBuildSizesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMicromapBuildSizesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMicromapBuildSizesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMicromapBuildSizesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetMicromapBuildSizesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetMicromapBuildSizesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetMicromapBuildSizesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMicromapBuildSizesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMicromapBuildSizesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMicromapBuildSizesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMicromapBuildSizesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMicromapBuildSizesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMicromapBuildSizesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMicromapBuildSizesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMicromapBuildSizesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMicromapBuildSizesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMicromapBuildSizesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMicromapBuildSizesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMicromapBuildSizesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMicromapBuildSizesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetMicromapBuildSizesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetMicromapBuildSizesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetMicromapBuildSizesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMicromapBuildSizesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMicromapBuildSizesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMicromapBuildSizesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMicromapBuildSizesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMicromapBuildSizesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMicromapBuildSizesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMicromapBuildSizesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMicromapBuildSizesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMicromapBuildSizesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMicromapBuildSizesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMicromapBuildSizesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMicromapBuildSizesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMicromapBuildSizesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetMicromapBuildSizesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetMicromapBuildSizesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetMicromapBuildSizesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMicromapBuildSizesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMicromapBuildSizesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMicromapBuildSizesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMicromapBuildSizesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMicromapBuildSizesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMicromapBuildSizesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMicromapBuildSizesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMicromapBuildSizesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMicromapBuildSizesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMicromapBuildSizesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMicromapBuildSizesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMicromapBuildSizesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMicromapBuildSizesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetMicromapBuildSizesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetMicromapBuildSizesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetMicromapBuildSizesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMicromapBuildSizesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMicromapBuildSizesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMicromapBuildSizesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMicromapBuildSizesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMicromapBuildSizesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMicromapBuildSizesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMicromapBuildSizesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMicromapBuildSizesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMicromapBuildSizesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMicromapBuildSizesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMicromapBuildSizesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMicromapBuildSizesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMicromapBuildSizesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetMicromapBuildSizesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetMicromapBuildSizesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMicromapBuildSizesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMicromapBuildSizesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMicromapBuildSizesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMicromapBuildSizesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMicromapBuildSizesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMicromapBuildSizesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMicromapBuildSizesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMicromapBuildSizesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMicromapBuildSizesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMicromapBuildSizesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMicromapBuildSizesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMicromapBuildSizesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMicromapBuildSizesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMicromapBuildSizesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetMicromapBuildSizesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetMicromapBuildSizesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMicromapBuildSizesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMicromapBuildSizesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMicromapBuildSizesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMicromapBuildSizesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMicromapBuildSizesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMicromapBuildSizesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMicromapBuildSizesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMicromapBuildSizesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMicromapBuildSizesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMicromapBuildSizesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMicromapBuildSizesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMicromapBuildSizesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMicromapBuildSizesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMicromapBuildSizesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetMicromapBuildSizesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetMicromapBuildSizesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMicromapBuildSizesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMicromapBuildSizesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMicromapBuildSizesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMicromapBuildSizesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMicromapBuildSizesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMicromapBuildSizesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMicromapBuildSizesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMicromapBuildSizesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMicromapBuildSizesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMicromapBuildSizesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMicromapBuildSizesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMicromapBuildSizesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMicromapBuildSizesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMicromapBuildSizesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetMicromapBuildSizesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetMicromapBuildSizesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetMicromapBuildSizesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMicromapBuildSizesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMicromapBuildSizesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMicromapBuildSizesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMicromapBuildSizesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMicromapBuildSizesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMicromapBuildSizesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMicromapBuildSizesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMicromapBuildSizesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMicromapBuildSizesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMicromapBuildSizesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMicromapBuildSizesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMicromapBuildSizesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMicromapBuildSizesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetMicromapBuildSizesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetMicromapBuildSizesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMicromapBuildSizesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMicromapBuildSizesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMicromapBuildSizesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMicromapBuildSizesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMicromapBuildSizesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMicromapBuildSizesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMicromapBuildSizesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMicromapBuildSizesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMicromapBuildSizesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMicromapBuildSizesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMicromapBuildSizesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMicromapBuildSizesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMicromapBuildSizesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMicromapBuildSizesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetMicromapBuildSizesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetMicromapBuildSizesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetShaderModuleIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetShaderModuleIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetShaderModuleIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetShaderModuleIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetShaderModuleIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetShaderModuleIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetShaderModuleIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetShaderModuleIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetShaderModuleIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetShaderModuleIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetShaderModuleIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetShaderModuleIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetShaderModuleIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetShaderModuleIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetShaderModuleIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetShaderModuleIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetShaderModuleIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetShaderModuleIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetShaderModuleIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetShaderModuleIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetShaderModuleIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetShaderModuleIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetShaderModuleIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetShaderModuleIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetShaderModuleIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetShaderModuleIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleCreateInfoIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleCreateInfoIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleCreateInfoIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleCreateInfoIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleCreateInfoIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleCreateInfoIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleCreateInfoIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleCreateInfoIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleCreateInfoIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleCreateInfoIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleCreateInfoIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleCreateInfoIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleCreateInfoIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderModuleCreateInfoIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetShaderModuleCreateInfoIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetShaderModuleCreateInfoIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleCreateInfoIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleCreateInfoIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleCreateInfoIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleCreateInfoIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleCreateInfoIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleCreateInfoIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleCreateInfoIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleCreateInfoIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleCreateInfoIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleCreateInfoIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleCreateInfoIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleCreateInfoIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleCreateInfoIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderModuleCreateInfoIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetShaderModuleCreateInfoIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetShaderModuleCreateInfoIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleCreateInfoIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleCreateInfoIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleCreateInfoIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleCreateInfoIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleCreateInfoIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleCreateInfoIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleCreateInfoIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleCreateInfoIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleCreateInfoIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleCreateInfoIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleCreateInfoIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleCreateInfoIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleCreateInfoIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderModuleCreateInfoIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetShaderModuleCreateInfoIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetShaderModuleCreateInfoIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleCreateInfoIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleCreateInfoIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleCreateInfoIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleCreateInfoIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleCreateInfoIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleCreateInfoIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleCreateInfoIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleCreateInfoIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleCreateInfoIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleCreateInfoIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleCreateInfoIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleCreateInfoIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleCreateInfoIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderModuleCreateInfoIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetShaderModuleCreateInfoIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetShaderModuleCreateInfoIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleCreateInfoIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleCreateInfoIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleCreateInfoIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleCreateInfoIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleCreateInfoIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleCreateInfoIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleCreateInfoIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleCreateInfoIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleCreateInfoIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleCreateInfoIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleCreateInfoIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleCreateInfoIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleCreateInfoIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderModuleCreateInfoIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetShaderModuleCreateInfoIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetShaderModuleCreateInfoIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleCreateInfoIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleCreateInfoIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleCreateInfoIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleCreateInfoIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleCreateInfoIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleCreateInfoIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleCreateInfoIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleCreateInfoIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleCreateInfoIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleCreateInfoIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleCreateInfoIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleCreateInfoIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleCreateInfoIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderModuleCreateInfoIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetShaderModuleCreateInfoIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetShaderModuleCreateInfoIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleCreateInfoIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleCreateInfoIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleCreateInfoIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleCreateInfoIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleCreateInfoIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleCreateInfoIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleCreateInfoIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleCreateInfoIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleCreateInfoIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleCreateInfoIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleCreateInfoIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleCreateInfoIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleCreateInfoIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderModuleCreateInfoIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetShaderModuleCreateInfoIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetShaderModuleCreateInfoIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleCreateInfoIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleCreateInfoIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleCreateInfoIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleCreateInfoIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleCreateInfoIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleCreateInfoIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleCreateInfoIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleCreateInfoIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleCreateInfoIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleCreateInfoIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleCreateInfoIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleCreateInfoIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleCreateInfoIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderModuleCreateInfoIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetShaderModuleCreateInfoIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetShaderModuleCreateInfoIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleCreateInfoIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleCreateInfoIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleCreateInfoIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleCreateInfoIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleCreateInfoIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleCreateInfoIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleCreateInfoIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleCreateInfoIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleCreateInfoIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleCreateInfoIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleCreateInfoIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleCreateInfoIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleCreateInfoIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderModuleCreateInfoIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetShaderModuleCreateInfoIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetShaderModuleCreateInfoIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleCreateInfoIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleCreateInfoIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleCreateInfoIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleCreateInfoIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleCreateInfoIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleCreateInfoIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleCreateInfoIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleCreateInfoIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleCreateInfoIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleCreateInfoIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleCreateInfoIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleCreateInfoIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleCreateInfoIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderModuleCreateInfoIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetShaderModuleCreateInfoIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetShaderModuleCreateInfoIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleCreateInfoIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleCreateInfoIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleCreateInfoIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleCreateInfoIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleCreateInfoIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleCreateInfoIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleCreateInfoIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleCreateInfoIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleCreateInfoIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleCreateInfoIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleCreateInfoIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleCreateInfoIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleCreateInfoIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderModuleCreateInfoIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetShaderModuleCreateInfoIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetShaderModuleCreateInfoIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleCreateInfoIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleCreateInfoIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleCreateInfoIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleCreateInfoIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleCreateInfoIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleCreateInfoIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleCreateInfoIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleCreateInfoIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleCreateInfoIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleCreateInfoIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleCreateInfoIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleCreateInfoIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleCreateInfoIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderModuleCreateInfoIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetShaderModuleCreateInfoIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetShaderModuleCreateInfoIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleCreateInfoIdentifierEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleCreateInfoIdentifierEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleCreateInfoIdentifierEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleCreateInfoIdentifierEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleCreateInfoIdentifierEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleCreateInfoIdentifierEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleCreateInfoIdentifierEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleCreateInfoIdentifierEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleCreateInfoIdentifierEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleCreateInfoIdentifierEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleCreateInfoIdentifierEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleCreateInfoIdentifierEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleCreateInfoIdentifierEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderModuleCreateInfoIdentifierEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetShaderModuleCreateInfoIdentifierEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetShaderModuleCreateInfoIdentifierEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetImageSubresourceLayout2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetImageSubresourceLayout2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetImageSubresourceLayout2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetImageSubresourceLayout2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetImageSubresourceLayout2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetImageSubresourceLayout2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetImageSubresourceLayout2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetImageSubresourceLayout2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetImageSubresourceLayout2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetImageSubresourceLayout2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetImageSubresourceLayout2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetImageSubresourceLayout2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetImageSubresourceLayout2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetImageSubresourceLayout2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetImageSubresourceLayout2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetImageSubresourceLayout2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetImageSubresourceLayout2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetImageSubresourceLayout2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetImageSubresourceLayout2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetImageSubresourceLayout2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetImageSubresourceLayout2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetImageSubresourceLayout2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetImageSubresourceLayout2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetImageSubresourceLayout2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetImageSubresourceLayout2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetImageSubresourceLayout2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetImageSubresourceLayout2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetImageSubresourceLayout2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetImageSubresourceLayout2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetImageSubresourceLayout2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetImageSubresourceLayout2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetImageSubresourceLayout2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetImageSubresourceLayout2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetImageSubresourceLayout2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetImageSubresourceLayout2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetImageSubresourceLayout2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetImageSubresourceLayout2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetImageSubresourceLayout2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetImageSubresourceLayout2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetImageSubresourceLayout2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetImageSubresourceLayout2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetImageSubresourceLayout2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetImageSubresourceLayout2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetImageSubresourceLayout2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetImageSubresourceLayout2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetImageSubresourceLayout2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetImageSubresourceLayout2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetImageSubresourceLayout2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetImageSubresourceLayout2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetImageSubresourceLayout2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetImageSubresourceLayout2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetImageSubresourceLayout2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageSubresourceLayout2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetImageSubresourceLayout2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetImageSubresourceLayout2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageSubresourceLayout2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetImageSubresourceLayout2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetImageSubresourceLayout2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageSubresourceLayout2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetImageSubresourceLayout2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetImageSubresourceLayout2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageSubresourceLayout2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetImageSubresourceLayout2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetImageSubresourceLayout2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageSubresourceLayout2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetImageSubresourceLayout2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetImageSubresourceLayout2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageSubresourceLayout2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetImageSubresourceLayout2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetImageSubresourceLayout2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageSubresourceLayout2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetImageSubresourceLayout2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetImageSubresourceLayout2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageSubresourceLayout2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetImageSubresourceLayout2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetImageSubresourceLayout2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageSubresourceLayout2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetImageSubresourceLayout2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetImageSubresourceLayout2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageSubresourceLayout2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetImageSubresourceLayout2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetImageSubresourceLayout2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageSubresourceLayout2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetImageSubresourceLayout2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetImageSubresourceLayout2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageSubresourceLayout2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetImageSubresourceLayout2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetImageSubresourceLayout2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageSubresourceLayout2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetImageSubresourceLayout2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetImageSubresourceLayout2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPipelinePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelinePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelinePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelinePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelinePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelinePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelinePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelinePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelinePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelinePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelinePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelinePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelinePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPipelinePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPipelinePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPipelinePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelinePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelinePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelinePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelinePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelinePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelinePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelinePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelinePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelinePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelinePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelinePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelinePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelinePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPipelinePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetPipelinePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetPipelinePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelinePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelinePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelinePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelinePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelinePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelinePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelinePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelinePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelinePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelinePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelinePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelinePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelinePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPipelinePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetPipelinePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetPipelinePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelinePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelinePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelinePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelinePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelinePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelinePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelinePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelinePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelinePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelinePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelinePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelinePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelinePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPipelinePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetPipelinePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetPipelinePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelinePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelinePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelinePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelinePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelinePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelinePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelinePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelinePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelinePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelinePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelinePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelinePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelinePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPipelinePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetPipelinePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetPipelinePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelinePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelinePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelinePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelinePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelinePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelinePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelinePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelinePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelinePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelinePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelinePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelinePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelinePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPipelinePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetPipelinePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetPipelinePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelinePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelinePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelinePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelinePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelinePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelinePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelinePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelinePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelinePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelinePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelinePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelinePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelinePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPipelinePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetPipelinePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetPipelinePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelinePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelinePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelinePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelinePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelinePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelinePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelinePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelinePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelinePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelinePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelinePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelinePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelinePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPipelinePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetPipelinePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetPipelinePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelinePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelinePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelinePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelinePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelinePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelinePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelinePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelinePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelinePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelinePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelinePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelinePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelinePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPipelinePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetPipelinePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetPipelinePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelinePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelinePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelinePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelinePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelinePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelinePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelinePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelinePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelinePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelinePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelinePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelinePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelinePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPipelinePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetPipelinePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetPipelinePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelinePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelinePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelinePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelinePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelinePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelinePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelinePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelinePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelinePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelinePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelinePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelinePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelinePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPipelinePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetPipelinePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetPipelinePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelinePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelinePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelinePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelinePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelinePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelinePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelinePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelinePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelinePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelinePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelinePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelinePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelinePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPipelinePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetPipelinePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetPipelinePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelinePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelinePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelinePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelinePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelinePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelinePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelinePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelinePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelinePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelinePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelinePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelinePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelinePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPipelinePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetPipelinePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetPipelinePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ExportMetalObjectsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ExportMetalObjectsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ExportMetalObjectsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ExportMetalObjectsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ExportMetalObjectsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ExportMetalObjectsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ExportMetalObjectsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ExportMetalObjectsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ExportMetalObjectsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ExportMetalObjectsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ExportMetalObjectsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ExportMetalObjectsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ExportMetalObjectsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ExportMetalObjectsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ExportMetalObjectsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ExportMetalObjectsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ExportMetalObjectsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ExportMetalObjectsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ExportMetalObjectsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ExportMetalObjectsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ExportMetalObjectsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ExportMetalObjectsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ExportMetalObjectsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ExportMetalObjectsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ExportMetalObjectsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ExportMetalObjectsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ExportMetalObjectsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ExportMetalObjectsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ExportMetalObjectsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ExportMetalObjectsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ExportMetalObjectsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ExportMetalObjectsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ExportMetalObjectsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ExportMetalObjectsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ExportMetalObjectsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ExportMetalObjectsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ExportMetalObjectsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ExportMetalObjectsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ExportMetalObjectsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ExportMetalObjectsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ExportMetalObjectsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ExportMetalObjectsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ExportMetalObjectsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ExportMetalObjectsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ExportMetalObjectsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ExportMetalObjectsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ExportMetalObjectsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ExportMetalObjectsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ExportMetalObjectsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ExportMetalObjectsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ExportMetalObjectsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ExportMetalObjectsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ExportMetalObjectsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ExportMetalObjectsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ExportMetalObjectsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ExportMetalObjectsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ExportMetalObjectsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ExportMetalObjectsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ExportMetalObjectsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ExportMetalObjectsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ExportMetalObjectsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ExportMetalObjectsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ExportMetalObjectsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ExportMetalObjectsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ExportMetalObjectsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ExportMetalObjectsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ExportMetalObjectsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ExportMetalObjectsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ExportMetalObjectsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ExportMetalObjectsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ExportMetalObjectsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ExportMetalObjectsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ExportMetalObjectsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ExportMetalObjectsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ExportMetalObjectsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ExportMetalObjectsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ExportMetalObjectsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ExportMetalObjectsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ExportMetalObjectsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ExportMetalObjectsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ExportMetalObjectsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ExportMetalObjectsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ExportMetalObjectsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ExportMetalObjectsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ExportMetalObjectsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ExportMetalObjectsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ExportMetalObjectsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ExportMetalObjectsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ExportMetalObjectsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ExportMetalObjectsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ExportMetalObjectsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ExportMetalObjectsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ExportMetalObjectsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ExportMetalObjectsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ExportMetalObjectsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ExportMetalObjectsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ExportMetalObjectsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ExportMetalObjectsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ExportMetalObjectsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ExportMetalObjectsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ExportMetalObjectsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ExportMetalObjectsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ExportMetalObjectsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ExportMetalObjectsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ExportMetalObjectsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ExportMetalObjectsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ExportMetalObjectsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ExportMetalObjectsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ExportMetalObjectsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ExportMetalObjectsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ExportMetalObjectsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ExportMetalObjectsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ExportMetalObjectsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ExportMetalObjectsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ExportMetalObjectsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ExportMetalObjectsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ExportMetalObjectsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ExportMetalObjectsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ExportMetalObjectsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ExportMetalObjectsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ExportMetalObjectsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ExportMetalObjectsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ExportMetalObjectsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ExportMetalObjectsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ExportMetalObjectsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ExportMetalObjectsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ExportMetalObjectsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ExportMetalObjectsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ExportMetalObjectsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ExportMetalObjectsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ExportMetalObjectsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ExportMetalObjectsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ExportMetalObjectsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ExportMetalObjectsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ExportMetalObjectsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ExportMetalObjectsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ExportMetalObjectsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ExportMetalObjectsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ExportMetalObjectsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ExportMetalObjectsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ExportMetalObjectsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ExportMetalObjectsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ExportMetalObjectsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ExportMetalObjectsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ExportMetalObjectsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ExportMetalObjectsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ExportMetalObjectsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ExportMetalObjectsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ExportMetalObjectsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ExportMetalObjectsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ExportMetalObjectsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ExportMetalObjectsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ExportMetalObjectsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ExportMetalObjectsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ExportMetalObjectsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ExportMetalObjectsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ExportMetalObjectsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ExportMetalObjectsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ExportMetalObjectsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ExportMetalObjectsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ExportMetalObjectsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ExportMetalObjectsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ExportMetalObjectsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ExportMetalObjectsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ExportMetalObjectsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ExportMetalObjectsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ExportMetalObjectsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ExportMetalObjectsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ExportMetalObjectsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ExportMetalObjectsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ExportMetalObjectsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ExportMetalObjectsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ExportMetalObjectsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ExportMetalObjectsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ExportMetalObjectsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ExportMetalObjectsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ExportMetalObjectsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ExportMetalObjectsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ExportMetalObjectsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ExportMetalObjectsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ExportMetalObjectsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ExportMetalObjectsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ExportMetalObjectsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ExportMetalObjectsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ExportMetalObjectsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ExportMetalObjectsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ExportMetalObjectsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ExportMetalObjectsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ExportMetalObjectsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ExportMetalObjectsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ExportMetalObjectsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ExportMetalObjectsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ExportMetalObjectsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ExportMetalObjectsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ExportMetalObjectsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ExportMetalObjectsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ExportMetalObjectsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ExportMetalObjectsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ExportMetalObjectsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ExportMetalObjectsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ExportMetalObjectsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ExportMetalObjectsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ExportMetalObjectsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ExportMetalObjectsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ExportMetalObjectsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ExportMetalObjectsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ExportMetalObjectsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ExportMetalObjectsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindTileMemoryQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTileMemoryQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTileMemoryQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTileMemoryQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTileMemoryQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTileMemoryQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTileMemoryQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTileMemoryQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTileMemoryQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTileMemoryQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTileMemoryQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTileMemoryQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTileMemoryQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTileMemoryQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindTileMemoryQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindTileMemoryQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTileMemoryQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTileMemoryQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTileMemoryQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTileMemoryQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTileMemoryQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTileMemoryQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTileMemoryQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTileMemoryQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTileMemoryQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTileMemoryQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTileMemoryQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTileMemoryQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTileMemoryQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTileMemoryQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindTileMemoryQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindTileMemoryQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTileMemoryQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTileMemoryQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTileMemoryQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTileMemoryQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTileMemoryQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTileMemoryQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTileMemoryQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTileMemoryQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTileMemoryQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTileMemoryQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTileMemoryQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTileMemoryQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTileMemoryQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTileMemoryQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindTileMemoryQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindTileMemoryQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTileMemoryQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTileMemoryQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTileMemoryQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTileMemoryQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTileMemoryQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTileMemoryQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTileMemoryQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTileMemoryQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTileMemoryQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTileMemoryQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTileMemoryQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTileMemoryQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTileMemoryQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTileMemoryQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindTileMemoryQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindTileMemoryQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTileMemoryQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTileMemoryQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTileMemoryQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTileMemoryQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTileMemoryQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTileMemoryQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTileMemoryQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTileMemoryQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTileMemoryQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTileMemoryQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTileMemoryQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTileMemoryQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTileMemoryQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTileMemoryQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindTileMemoryQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindTileMemoryQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTileMemoryQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTileMemoryQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTileMemoryQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTileMemoryQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTileMemoryQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTileMemoryQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTileMemoryQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTileMemoryQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTileMemoryQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTileMemoryQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTileMemoryQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTileMemoryQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTileMemoryQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTileMemoryQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindTileMemoryQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindTileMemoryQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTileMemoryQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTileMemoryQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTileMemoryQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTileMemoryQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTileMemoryQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTileMemoryQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTileMemoryQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTileMemoryQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTileMemoryQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTileMemoryQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTileMemoryQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTileMemoryQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTileMemoryQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTileMemoryQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindTileMemoryQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindTileMemoryQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTileMemoryQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTileMemoryQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTileMemoryQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTileMemoryQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTileMemoryQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTileMemoryQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTileMemoryQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTileMemoryQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTileMemoryQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTileMemoryQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTileMemoryQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTileMemoryQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTileMemoryQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTileMemoryQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindTileMemoryQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindTileMemoryQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTileMemoryQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTileMemoryQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTileMemoryQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTileMemoryQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTileMemoryQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTileMemoryQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTileMemoryQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTileMemoryQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTileMemoryQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTileMemoryQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTileMemoryQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTileMemoryQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTileMemoryQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTileMemoryQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindTileMemoryQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindTileMemoryQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTileMemoryQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTileMemoryQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTileMemoryQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTileMemoryQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTileMemoryQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTileMemoryQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTileMemoryQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTileMemoryQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTileMemoryQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTileMemoryQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTileMemoryQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTileMemoryQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTileMemoryQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTileMemoryQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindTileMemoryQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindTileMemoryQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTileMemoryQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTileMemoryQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTileMemoryQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTileMemoryQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTileMemoryQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTileMemoryQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTileMemoryQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTileMemoryQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTileMemoryQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTileMemoryQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTileMemoryQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTileMemoryQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTileMemoryQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTileMemoryQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindTileMemoryQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindTileMemoryQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTileMemoryQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTileMemoryQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTileMemoryQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTileMemoryQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTileMemoryQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTileMemoryQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTileMemoryQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTileMemoryQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTileMemoryQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTileMemoryQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTileMemoryQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTileMemoryQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTileMemoryQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTileMemoryQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindTileMemoryQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindTileMemoryQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTileMemoryQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTileMemoryQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTileMemoryQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTileMemoryQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTileMemoryQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTileMemoryQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTileMemoryQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTileMemoryQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTileMemoryQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTileMemoryQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTileMemoryQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTileMemoryQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTileMemoryQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTileMemoryQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindTileMemoryQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindTileMemoryQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetFramebufferTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFramebufferTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFramebufferTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFramebufferTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFramebufferTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFramebufferTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFramebufferTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFramebufferTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFramebufferTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFramebufferTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFramebufferTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFramebufferTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFramebufferTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetFramebufferTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetFramebufferTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetFramebufferTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetFramebufferTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFramebufferTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFramebufferTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFramebufferTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFramebufferTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFramebufferTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFramebufferTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFramebufferTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFramebufferTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFramebufferTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFramebufferTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFramebufferTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFramebufferTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetFramebufferTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetFramebufferTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetFramebufferTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetFramebufferTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFramebufferTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFramebufferTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFramebufferTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFramebufferTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFramebufferTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFramebufferTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFramebufferTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFramebufferTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFramebufferTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFramebufferTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFramebufferTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFramebufferTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetFramebufferTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetFramebufferTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetFramebufferTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetFramebufferTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFramebufferTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFramebufferTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFramebufferTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFramebufferTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFramebufferTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFramebufferTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFramebufferTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFramebufferTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFramebufferTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFramebufferTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFramebufferTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFramebufferTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetFramebufferTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetFramebufferTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetFramebufferTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetFramebufferTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFramebufferTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFramebufferTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFramebufferTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFramebufferTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFramebufferTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFramebufferTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFramebufferTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFramebufferTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFramebufferTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFramebufferTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFramebufferTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFramebufferTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetFramebufferTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetFramebufferTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetFramebufferTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetFramebufferTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFramebufferTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFramebufferTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFramebufferTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFramebufferTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFramebufferTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFramebufferTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFramebufferTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFramebufferTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFramebufferTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFramebufferTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFramebufferTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFramebufferTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetFramebufferTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetFramebufferTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetFramebufferTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetFramebufferTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFramebufferTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFramebufferTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFramebufferTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFramebufferTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFramebufferTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFramebufferTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFramebufferTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFramebufferTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFramebufferTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFramebufferTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFramebufferTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFramebufferTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetFramebufferTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetFramebufferTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetFramebufferTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetFramebufferTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFramebufferTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFramebufferTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFramebufferTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFramebufferTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFramebufferTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFramebufferTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFramebufferTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFramebufferTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFramebufferTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFramebufferTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFramebufferTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFramebufferTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetFramebufferTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetFramebufferTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetFramebufferTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFramebufferTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFramebufferTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFramebufferTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFramebufferTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFramebufferTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFramebufferTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFramebufferTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFramebufferTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFramebufferTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFramebufferTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFramebufferTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFramebufferTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFramebufferTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetFramebufferTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetFramebufferTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetFramebufferTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFramebufferTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFramebufferTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFramebufferTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFramebufferTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFramebufferTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFramebufferTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFramebufferTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFramebufferTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFramebufferTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFramebufferTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFramebufferTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFramebufferTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFramebufferTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetFramebufferTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetFramebufferTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetFramebufferTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFramebufferTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFramebufferTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFramebufferTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFramebufferTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFramebufferTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFramebufferTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFramebufferTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFramebufferTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFramebufferTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFramebufferTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFramebufferTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFramebufferTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFramebufferTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetFramebufferTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetFramebufferTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetFramebufferTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetFramebufferTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFramebufferTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFramebufferTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFramebufferTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFramebufferTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFramebufferTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFramebufferTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFramebufferTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFramebufferTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFramebufferTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFramebufferTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFramebufferTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFramebufferTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetFramebufferTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetFramebufferTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetFramebufferTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFramebufferTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFramebufferTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFramebufferTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFramebufferTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFramebufferTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFramebufferTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFramebufferTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFramebufferTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFramebufferTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFramebufferTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFramebufferTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFramebufferTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFramebufferTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetFramebufferTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetFramebufferTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetFramebufferTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDynamicRenderingTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDynamicRenderingTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDynamicRenderingTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDynamicRenderingTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDynamicRenderingTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDynamicRenderingTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDynamicRenderingTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDynamicRenderingTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDynamicRenderingTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDynamicRenderingTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDynamicRenderingTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDynamicRenderingTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDynamicRenderingTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDynamicRenderingTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDynamicRenderingTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDynamicRenderingTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDynamicRenderingTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDynamicRenderingTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDynamicRenderingTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDynamicRenderingTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDynamicRenderingTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDynamicRenderingTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDynamicRenderingTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDynamicRenderingTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDynamicRenderingTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDynamicRenderingTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDynamicRenderingTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDynamicRenderingTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDynamicRenderingTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDynamicRenderingTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDynamicRenderingTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDynamicRenderingTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDynamicRenderingTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDynamicRenderingTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDynamicRenderingTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDynamicRenderingTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDynamicRenderingTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDynamicRenderingTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDynamicRenderingTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDynamicRenderingTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDynamicRenderingTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDynamicRenderingTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDynamicRenderingTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDynamicRenderingTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDynamicRenderingTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDynamicRenderingTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDynamicRenderingTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDynamicRenderingTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDynamicRenderingTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDynamicRenderingTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDynamicRenderingTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDynamicRenderingTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDynamicRenderingTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDynamicRenderingTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDynamicRenderingTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDynamicRenderingTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDynamicRenderingTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDynamicRenderingTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDynamicRenderingTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDynamicRenderingTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDynamicRenderingTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDynamicRenderingTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDynamicRenderingTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDynamicRenderingTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDynamicRenderingTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDynamicRenderingTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDynamicRenderingTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDynamicRenderingTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDynamicRenderingTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDynamicRenderingTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDynamicRenderingTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDynamicRenderingTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDynamicRenderingTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDynamicRenderingTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDynamicRenderingTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDynamicRenderingTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDynamicRenderingTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDynamicRenderingTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDynamicRenderingTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDynamicRenderingTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDynamicRenderingTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDynamicRenderingTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDynamicRenderingTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDynamicRenderingTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDynamicRenderingTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDynamicRenderingTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDynamicRenderingTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDynamicRenderingTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDynamicRenderingTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDynamicRenderingTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDynamicRenderingTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDynamicRenderingTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDynamicRenderingTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDynamicRenderingTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDynamicRenderingTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDynamicRenderingTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDynamicRenderingTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDynamicRenderingTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDynamicRenderingTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDynamicRenderingTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDynamicRenderingTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDynamicRenderingTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDynamicRenderingTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDynamicRenderingTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDynamicRenderingTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDynamicRenderingTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDynamicRenderingTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDynamicRenderingTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDynamicRenderingTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDynamicRenderingTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDynamicRenderingTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDynamicRenderingTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDynamicRenderingTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDynamicRenderingTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDynamicRenderingTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDynamicRenderingTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDynamicRenderingTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDynamicRenderingTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDynamicRenderingTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDynamicRenderingTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDynamicRenderingTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDynamicRenderingTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDynamicRenderingTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDynamicRenderingTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDynamicRenderingTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDynamicRenderingTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDynamicRenderingTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDynamicRenderingTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDynamicRenderingTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDynamicRenderingTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDynamicRenderingTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDynamicRenderingTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDynamicRenderingTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDynamicRenderingTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDynamicRenderingTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDynamicRenderingTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDynamicRenderingTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDynamicRenderingTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDynamicRenderingTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDynamicRenderingTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDynamicRenderingTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDynamicRenderingTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDynamicRenderingTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDynamicRenderingTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDynamicRenderingTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDynamicRenderingTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDynamicRenderingTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDynamicRenderingTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDynamicRenderingTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDynamicRenderingTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDynamicRenderingTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDynamicRenderingTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDynamicRenderingTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDynamicRenderingTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDynamicRenderingTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDynamicRenderingTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDynamicRenderingTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDynamicRenderingTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDynamicRenderingTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDynamicRenderingTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDynamicRenderingTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDynamicRenderingTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDynamicRenderingTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDynamicRenderingTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDynamicRenderingTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDynamicRenderingTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDynamicRenderingTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDynamicRenderingTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDynamicRenderingTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDynamicRenderingTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDynamicRenderingTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDynamicRenderingTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDynamicRenderingTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDynamicRenderingTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDynamicRenderingTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDynamicRenderingTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDynamicRenderingTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDynamicRenderingTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDynamicRenderingTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDynamicRenderingTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDynamicRenderingTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDynamicRenderingTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDynamicRenderingTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDynamicRenderingTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDynamicRenderingTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDynamicRenderingTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDynamicRenderingTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDynamicRenderingTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDynamicRenderingTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDynamicRenderingTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDynamicRenderingTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDynamicRenderingTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDynamicRenderingTilePropertiesQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDynamicRenderingTilePropertiesQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDynamicRenderingTilePropertiesQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDynamicRenderingTilePropertiesQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDynamicRenderingTilePropertiesQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDynamicRenderingTilePropertiesQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDynamicRenderingTilePropertiesQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDynamicRenderingTilePropertiesQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDynamicRenderingTilePropertiesQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDynamicRenderingTilePropertiesQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDynamicRenderingTilePropertiesQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDynamicRenderingTilePropertiesQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDynamicRenderingTilePropertiesQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDynamicRenderingTilePropertiesQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDynamicRenderingTilePropertiesQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDynamicRenderingTilePropertiesQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyOpticalFlowSessionNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyOpticalFlowSessionNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyOpticalFlowSessionNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyOpticalFlowSessionNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyOpticalFlowSessionNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyOpticalFlowSessionNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyOpticalFlowSessionNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyOpticalFlowSessionNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyOpticalFlowSessionNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyOpticalFlowSessionNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyOpticalFlowSessionNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyOpticalFlowSessionNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyOpticalFlowSessionNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyOpticalFlowSessionNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyOpticalFlowSessionNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyOpticalFlowSessionNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_BindOpticalFlowSessionImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindOpticalFlowSessionImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindOpticalFlowSessionImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindOpticalFlowSessionImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindOpticalFlowSessionImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindOpticalFlowSessionImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindOpticalFlowSessionImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindOpticalFlowSessionImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindOpticalFlowSessionImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindOpticalFlowSessionImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindOpticalFlowSessionImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindOpticalFlowSessionImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindOpticalFlowSessionImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindOpticalFlowSessionImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_BindOpticalFlowSessionImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_BindOpticalFlowSessionImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_BindOpticalFlowSessionImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindOpticalFlowSessionImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindOpticalFlowSessionImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindOpticalFlowSessionImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindOpticalFlowSessionImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindOpticalFlowSessionImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindOpticalFlowSessionImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindOpticalFlowSessionImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindOpticalFlowSessionImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindOpticalFlowSessionImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindOpticalFlowSessionImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindOpticalFlowSessionImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindOpticalFlowSessionImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindOpticalFlowSessionImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_BindOpticalFlowSessionImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_BindOpticalFlowSessionImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_BindOpticalFlowSessionImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindOpticalFlowSessionImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindOpticalFlowSessionImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindOpticalFlowSessionImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindOpticalFlowSessionImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindOpticalFlowSessionImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindOpticalFlowSessionImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindOpticalFlowSessionImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindOpticalFlowSessionImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindOpticalFlowSessionImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindOpticalFlowSessionImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindOpticalFlowSessionImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindOpticalFlowSessionImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindOpticalFlowSessionImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_BindOpticalFlowSessionImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_BindOpticalFlowSessionImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_BindOpticalFlowSessionImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindOpticalFlowSessionImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindOpticalFlowSessionImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindOpticalFlowSessionImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindOpticalFlowSessionImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindOpticalFlowSessionImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindOpticalFlowSessionImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindOpticalFlowSessionImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindOpticalFlowSessionImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindOpticalFlowSessionImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindOpticalFlowSessionImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindOpticalFlowSessionImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindOpticalFlowSessionImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindOpticalFlowSessionImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_BindOpticalFlowSessionImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_BindOpticalFlowSessionImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_BindOpticalFlowSessionImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindOpticalFlowSessionImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindOpticalFlowSessionImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindOpticalFlowSessionImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindOpticalFlowSessionImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindOpticalFlowSessionImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindOpticalFlowSessionImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindOpticalFlowSessionImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindOpticalFlowSessionImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindOpticalFlowSessionImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindOpticalFlowSessionImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindOpticalFlowSessionImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindOpticalFlowSessionImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindOpticalFlowSessionImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_BindOpticalFlowSessionImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_BindOpticalFlowSessionImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_BindOpticalFlowSessionImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindOpticalFlowSessionImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindOpticalFlowSessionImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindOpticalFlowSessionImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindOpticalFlowSessionImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindOpticalFlowSessionImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindOpticalFlowSessionImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindOpticalFlowSessionImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindOpticalFlowSessionImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindOpticalFlowSessionImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindOpticalFlowSessionImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindOpticalFlowSessionImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindOpticalFlowSessionImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindOpticalFlowSessionImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_BindOpticalFlowSessionImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_BindOpticalFlowSessionImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_BindOpticalFlowSessionImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindOpticalFlowSessionImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindOpticalFlowSessionImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindOpticalFlowSessionImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindOpticalFlowSessionImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindOpticalFlowSessionImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindOpticalFlowSessionImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindOpticalFlowSessionImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindOpticalFlowSessionImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindOpticalFlowSessionImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindOpticalFlowSessionImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindOpticalFlowSessionImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindOpticalFlowSessionImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindOpticalFlowSessionImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_BindOpticalFlowSessionImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_BindOpticalFlowSessionImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_BindOpticalFlowSessionImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindOpticalFlowSessionImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindOpticalFlowSessionImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindOpticalFlowSessionImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindOpticalFlowSessionImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindOpticalFlowSessionImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindOpticalFlowSessionImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindOpticalFlowSessionImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindOpticalFlowSessionImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindOpticalFlowSessionImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindOpticalFlowSessionImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindOpticalFlowSessionImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindOpticalFlowSessionImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindOpticalFlowSessionImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_BindOpticalFlowSessionImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_BindOpticalFlowSessionImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_BindOpticalFlowSessionImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindOpticalFlowSessionImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindOpticalFlowSessionImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindOpticalFlowSessionImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindOpticalFlowSessionImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindOpticalFlowSessionImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindOpticalFlowSessionImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindOpticalFlowSessionImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindOpticalFlowSessionImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindOpticalFlowSessionImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindOpticalFlowSessionImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindOpticalFlowSessionImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindOpticalFlowSessionImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindOpticalFlowSessionImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_BindOpticalFlowSessionImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_BindOpticalFlowSessionImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_BindOpticalFlowSessionImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindOpticalFlowSessionImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindOpticalFlowSessionImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindOpticalFlowSessionImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindOpticalFlowSessionImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindOpticalFlowSessionImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindOpticalFlowSessionImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindOpticalFlowSessionImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindOpticalFlowSessionImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindOpticalFlowSessionImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindOpticalFlowSessionImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindOpticalFlowSessionImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindOpticalFlowSessionImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindOpticalFlowSessionImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_BindOpticalFlowSessionImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_BindOpticalFlowSessionImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindOpticalFlowSessionImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindOpticalFlowSessionImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindOpticalFlowSessionImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindOpticalFlowSessionImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindOpticalFlowSessionImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindOpticalFlowSessionImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindOpticalFlowSessionImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindOpticalFlowSessionImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindOpticalFlowSessionImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindOpticalFlowSessionImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindOpticalFlowSessionImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindOpticalFlowSessionImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindOpticalFlowSessionImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindOpticalFlowSessionImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_BindOpticalFlowSessionImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_BindOpticalFlowSessionImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_BindOpticalFlowSessionImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindOpticalFlowSessionImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindOpticalFlowSessionImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindOpticalFlowSessionImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindOpticalFlowSessionImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindOpticalFlowSessionImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindOpticalFlowSessionImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindOpticalFlowSessionImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindOpticalFlowSessionImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindOpticalFlowSessionImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindOpticalFlowSessionImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindOpticalFlowSessionImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindOpticalFlowSessionImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindOpticalFlowSessionImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_BindOpticalFlowSessionImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_BindOpticalFlowSessionImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_BindOpticalFlowSessionImageNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindOpticalFlowSessionImageNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindOpticalFlowSessionImageNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindOpticalFlowSessionImageNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindOpticalFlowSessionImageNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindOpticalFlowSessionImageNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindOpticalFlowSessionImageNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindOpticalFlowSessionImageNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindOpticalFlowSessionImageNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindOpticalFlowSessionImageNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindOpticalFlowSessionImageNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindOpticalFlowSessionImageNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindOpticalFlowSessionImageNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindOpticalFlowSessionImageNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_BindOpticalFlowSessionImageNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_BindOpticalFlowSessionImageNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdOpticalFlowExecuteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdOpticalFlowExecuteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdOpticalFlowExecuteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdOpticalFlowExecuteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdOpticalFlowExecuteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdOpticalFlowExecuteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdOpticalFlowExecuteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdOpticalFlowExecuteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdOpticalFlowExecuteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdOpticalFlowExecuteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdOpticalFlowExecuteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdOpticalFlowExecuteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdOpticalFlowExecuteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdOpticalFlowExecuteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdOpticalFlowExecuteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdOpticalFlowExecuteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdOpticalFlowExecuteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdOpticalFlowExecuteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdOpticalFlowExecuteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdOpticalFlowExecuteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdOpticalFlowExecuteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdOpticalFlowExecuteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdOpticalFlowExecuteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdOpticalFlowExecuteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdOpticalFlowExecuteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdOpticalFlowExecuteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdOpticalFlowExecuteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdOpticalFlowExecuteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdOpticalFlowExecuteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdOpticalFlowExecuteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdOpticalFlowExecuteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdOpticalFlowExecuteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdOpticalFlowExecuteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdOpticalFlowExecuteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdOpticalFlowExecuteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdOpticalFlowExecuteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdOpticalFlowExecuteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdOpticalFlowExecuteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdOpticalFlowExecuteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdOpticalFlowExecuteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdOpticalFlowExecuteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdOpticalFlowExecuteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdOpticalFlowExecuteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdOpticalFlowExecuteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdOpticalFlowExecuteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdOpticalFlowExecuteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdOpticalFlowExecuteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdOpticalFlowExecuteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdOpticalFlowExecuteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdOpticalFlowExecuteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdOpticalFlowExecuteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdOpticalFlowExecuteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdOpticalFlowExecuteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdOpticalFlowExecuteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdOpticalFlowExecuteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdOpticalFlowExecuteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdOpticalFlowExecuteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdOpticalFlowExecuteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdOpticalFlowExecuteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdOpticalFlowExecuteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdOpticalFlowExecuteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdOpticalFlowExecuteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdOpticalFlowExecuteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdOpticalFlowExecuteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdOpticalFlowExecuteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdOpticalFlowExecuteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdOpticalFlowExecuteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdOpticalFlowExecuteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdOpticalFlowExecuteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdOpticalFlowExecuteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdOpticalFlowExecuteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdOpticalFlowExecuteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdOpticalFlowExecuteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdOpticalFlowExecuteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdOpticalFlowExecuteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdOpticalFlowExecuteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdOpticalFlowExecuteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdOpticalFlowExecuteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdOpticalFlowExecuteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdOpticalFlowExecuteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdOpticalFlowExecuteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdOpticalFlowExecuteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdOpticalFlowExecuteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdOpticalFlowExecuteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdOpticalFlowExecuteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdOpticalFlowExecuteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdOpticalFlowExecuteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdOpticalFlowExecuteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdOpticalFlowExecuteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdOpticalFlowExecuteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdOpticalFlowExecuteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdOpticalFlowExecuteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdOpticalFlowExecuteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdOpticalFlowExecuteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdOpticalFlowExecuteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdOpticalFlowExecuteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdOpticalFlowExecuteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdOpticalFlowExecuteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdOpticalFlowExecuteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdOpticalFlowExecuteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdOpticalFlowExecuteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdOpticalFlowExecuteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdOpticalFlowExecuteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdOpticalFlowExecuteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdOpticalFlowExecuteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdOpticalFlowExecuteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdOpticalFlowExecuteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdOpticalFlowExecuteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdOpticalFlowExecuteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdOpticalFlowExecuteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdOpticalFlowExecuteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdOpticalFlowExecuteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdOpticalFlowExecuteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdOpticalFlowExecuteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdOpticalFlowExecuteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdOpticalFlowExecuteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdOpticalFlowExecuteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdOpticalFlowExecuteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdOpticalFlowExecuteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdOpticalFlowExecuteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdOpticalFlowExecuteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdOpticalFlowExecuteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdOpticalFlowExecuteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdOpticalFlowExecuteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdOpticalFlowExecuteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdOpticalFlowExecuteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdOpticalFlowExecuteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdOpticalFlowExecuteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdOpticalFlowExecuteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdOpticalFlowExecuteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdOpticalFlowExecuteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdOpticalFlowExecuteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdOpticalFlowExecuteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdOpticalFlowExecuteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdOpticalFlowExecuteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdOpticalFlowExecuteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdOpticalFlowExecuteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdOpticalFlowExecuteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdOpticalFlowExecuteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdOpticalFlowExecuteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdOpticalFlowExecuteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdOpticalFlowExecuteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdOpticalFlowExecuteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdOpticalFlowExecuteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdOpticalFlowExecuteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdOpticalFlowExecuteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdOpticalFlowExecuteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdOpticalFlowExecuteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdOpticalFlowExecuteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdOpticalFlowExecuteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdOpticalFlowExecuteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdOpticalFlowExecuteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdOpticalFlowExecuteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdOpticalFlowExecuteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdOpticalFlowExecuteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdOpticalFlowExecuteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdOpticalFlowExecuteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdOpticalFlowExecuteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdOpticalFlowExecuteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdOpticalFlowExecuteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdOpticalFlowExecuteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdOpticalFlowExecuteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdOpticalFlowExecuteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdOpticalFlowExecuteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdOpticalFlowExecuteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdOpticalFlowExecuteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdOpticalFlowExecuteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdOpticalFlowExecuteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdOpticalFlowExecuteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdOpticalFlowExecuteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdOpticalFlowExecuteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdOpticalFlowExecuteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdOpticalFlowExecuteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdOpticalFlowExecuteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdOpticalFlowExecuteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdOpticalFlowExecuteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdOpticalFlowExecuteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdOpticalFlowExecuteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdOpticalFlowExecuteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdOpticalFlowExecuteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdOpticalFlowExecuteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdOpticalFlowExecuteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdOpticalFlowExecuteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdOpticalFlowExecuteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdOpticalFlowExecuteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdOpticalFlowExecuteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdOpticalFlowExecuteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdOpticalFlowExecuteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdOpticalFlowExecuteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdOpticalFlowExecuteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdOpticalFlowExecuteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdOpticalFlowExecuteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdOpticalFlowExecuteNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdOpticalFlowExecuteNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdOpticalFlowExecuteNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdOpticalFlowExecuteNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdOpticalFlowExecuteNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdOpticalFlowExecuteNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdOpticalFlowExecuteNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdOpticalFlowExecuteNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdOpticalFlowExecuteNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdOpticalFlowExecuteNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdOpticalFlowExecuteNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdOpticalFlowExecuteNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdOpticalFlowExecuteNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdOpticalFlowExecuteNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdOpticalFlowExecuteNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdOpticalFlowExecuteNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultInfoEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultInfoEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultInfoEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultInfoEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultInfoEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultInfoEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultInfoEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultInfoEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultInfoEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultInfoEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultInfoEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultInfoEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultInfoEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultInfoEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceFaultInfoEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceFaultInfoEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultInfoEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultInfoEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultInfoEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultInfoEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultInfoEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultInfoEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultInfoEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultInfoEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultInfoEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultInfoEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultInfoEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultInfoEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultInfoEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultInfoEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceFaultInfoEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceFaultInfoEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultInfoEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultInfoEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultInfoEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultInfoEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultInfoEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultInfoEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultInfoEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultInfoEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultInfoEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultInfoEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultInfoEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultInfoEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultInfoEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultInfoEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceFaultInfoEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceFaultInfoEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultInfoEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultInfoEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultInfoEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultInfoEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultInfoEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultInfoEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultInfoEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultInfoEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultInfoEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultInfoEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultInfoEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultInfoEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultInfoEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultInfoEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceFaultInfoEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceFaultInfoEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultInfoEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultInfoEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultInfoEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultInfoEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultInfoEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultInfoEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultInfoEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultInfoEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultInfoEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultInfoEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultInfoEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultInfoEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultInfoEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultInfoEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceFaultInfoEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceFaultInfoEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultInfoEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultInfoEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultInfoEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultInfoEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultInfoEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultInfoEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultInfoEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultInfoEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultInfoEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultInfoEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultInfoEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultInfoEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultInfoEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultInfoEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceFaultInfoEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceFaultInfoEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultInfoEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultInfoEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultInfoEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultInfoEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultInfoEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultInfoEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultInfoEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultInfoEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultInfoEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultInfoEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultInfoEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultInfoEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultInfoEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultInfoEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceFaultInfoEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceFaultInfoEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultInfoEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultInfoEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultInfoEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultInfoEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultInfoEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultInfoEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultInfoEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultInfoEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultInfoEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultInfoEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultInfoEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultInfoEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultInfoEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultInfoEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceFaultInfoEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceFaultInfoEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultInfoEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultInfoEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultInfoEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultInfoEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultInfoEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultInfoEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultInfoEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultInfoEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultInfoEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultInfoEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultInfoEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultInfoEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultInfoEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultInfoEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceFaultInfoEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceFaultInfoEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultInfoEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultInfoEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultInfoEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultInfoEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultInfoEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultInfoEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultInfoEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultInfoEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultInfoEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultInfoEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultInfoEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultInfoEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultInfoEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultInfoEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceFaultInfoEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceFaultInfoEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultInfoEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultInfoEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultInfoEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultInfoEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultInfoEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultInfoEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultInfoEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultInfoEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultInfoEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultInfoEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultInfoEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultInfoEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultInfoEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultInfoEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceFaultInfoEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceFaultInfoEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultInfoEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultInfoEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultInfoEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultInfoEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultInfoEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultInfoEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultInfoEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultInfoEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultInfoEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultInfoEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultInfoEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultInfoEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultInfoEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultInfoEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceFaultInfoEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceFaultInfoEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultInfoEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultInfoEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultInfoEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultInfoEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultInfoEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultInfoEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultInfoEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultInfoEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultInfoEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultInfoEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultInfoEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultInfoEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultInfoEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultInfoEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceFaultInfoEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceFaultInfoEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultReportsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultReportsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultReportsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultReportsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultReportsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultReportsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultReportsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultReportsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultReportsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultReportsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultReportsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultReportsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultReportsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultReportsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceFaultReportsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceFaultReportsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultReportsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultReportsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultReportsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultReportsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultReportsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultReportsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultReportsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultReportsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultReportsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultReportsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultReportsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultReportsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultReportsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultReportsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceFaultReportsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceFaultReportsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultReportsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultReportsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultReportsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultReportsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultReportsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultReportsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultReportsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultReportsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultReportsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultReportsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultReportsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultReportsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultReportsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultReportsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceFaultReportsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceFaultReportsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultReportsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultReportsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultReportsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultReportsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultReportsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultReportsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultReportsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultReportsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultReportsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultReportsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultReportsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultReportsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultReportsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultReportsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceFaultReportsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceFaultReportsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultReportsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultReportsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultReportsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultReportsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultReportsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultReportsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultReportsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultReportsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultReportsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultReportsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultReportsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultReportsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultReportsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultReportsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceFaultReportsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceFaultReportsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultReportsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultReportsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultReportsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultReportsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultReportsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultReportsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultReportsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultReportsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultReportsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultReportsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultReportsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultReportsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultReportsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultReportsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceFaultReportsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceFaultReportsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultReportsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultReportsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultReportsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultReportsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultReportsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultReportsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultReportsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultReportsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultReportsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultReportsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultReportsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultReportsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultReportsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultReportsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceFaultReportsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceFaultReportsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultReportsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultReportsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultReportsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultReportsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultReportsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultReportsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultReportsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultReportsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultReportsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultReportsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultReportsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultReportsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultReportsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultReportsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceFaultReportsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceFaultReportsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultReportsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultReportsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultReportsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultReportsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultReportsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultReportsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultReportsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultReportsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultReportsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultReportsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultReportsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultReportsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultReportsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultReportsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceFaultReportsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceFaultReportsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultReportsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultReportsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultReportsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultReportsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultReportsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultReportsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultReportsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultReportsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultReportsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultReportsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultReportsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultReportsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultReportsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultReportsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceFaultReportsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceFaultReportsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultReportsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultReportsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultReportsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultReportsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultReportsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultReportsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultReportsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultReportsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultReportsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultReportsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultReportsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultReportsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultReportsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultReportsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceFaultReportsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceFaultReportsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultReportsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultReportsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultReportsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultReportsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultReportsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultReportsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultReportsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultReportsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultReportsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultReportsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultReportsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultReportsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultReportsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultReportsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceFaultReportsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceFaultReportsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultReportsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultReportsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultReportsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultReportsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultReportsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultReportsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultReportsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultReportsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultReportsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultReportsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultReportsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultReportsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultReportsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultReportsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceFaultReportsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceFaultReportsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultDebugInfoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultDebugInfoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultDebugInfoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultDebugInfoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultDebugInfoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultDebugInfoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultDebugInfoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultDebugInfoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultDebugInfoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultDebugInfoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultDebugInfoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultDebugInfoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultDebugInfoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceFaultDebugInfoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceFaultDebugInfoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceFaultDebugInfoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultDebugInfoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultDebugInfoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultDebugInfoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultDebugInfoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultDebugInfoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultDebugInfoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultDebugInfoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultDebugInfoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultDebugInfoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultDebugInfoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultDebugInfoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultDebugInfoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultDebugInfoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceFaultDebugInfoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceFaultDebugInfoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceFaultDebugInfoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultDebugInfoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultDebugInfoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultDebugInfoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultDebugInfoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultDebugInfoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultDebugInfoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultDebugInfoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultDebugInfoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultDebugInfoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultDebugInfoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultDebugInfoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultDebugInfoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultDebugInfoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceFaultDebugInfoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceFaultDebugInfoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceFaultDebugInfoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultDebugInfoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultDebugInfoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultDebugInfoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultDebugInfoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultDebugInfoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultDebugInfoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultDebugInfoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultDebugInfoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultDebugInfoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultDebugInfoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultDebugInfoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultDebugInfoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultDebugInfoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceFaultDebugInfoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceFaultDebugInfoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceFaultDebugInfoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultDebugInfoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultDebugInfoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultDebugInfoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultDebugInfoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultDebugInfoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultDebugInfoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultDebugInfoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultDebugInfoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultDebugInfoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultDebugInfoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultDebugInfoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultDebugInfoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultDebugInfoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceFaultDebugInfoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceFaultDebugInfoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceFaultDebugInfoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultDebugInfoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultDebugInfoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultDebugInfoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultDebugInfoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultDebugInfoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultDebugInfoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultDebugInfoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultDebugInfoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultDebugInfoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultDebugInfoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultDebugInfoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultDebugInfoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultDebugInfoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceFaultDebugInfoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceFaultDebugInfoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceFaultDebugInfoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultDebugInfoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultDebugInfoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultDebugInfoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultDebugInfoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultDebugInfoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultDebugInfoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultDebugInfoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultDebugInfoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultDebugInfoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultDebugInfoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultDebugInfoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultDebugInfoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultDebugInfoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceFaultDebugInfoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceFaultDebugInfoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceFaultDebugInfoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultDebugInfoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultDebugInfoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultDebugInfoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultDebugInfoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultDebugInfoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultDebugInfoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultDebugInfoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultDebugInfoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultDebugInfoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultDebugInfoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultDebugInfoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultDebugInfoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultDebugInfoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceFaultDebugInfoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceFaultDebugInfoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceFaultDebugInfoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultDebugInfoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultDebugInfoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultDebugInfoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultDebugInfoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultDebugInfoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultDebugInfoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultDebugInfoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultDebugInfoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultDebugInfoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultDebugInfoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultDebugInfoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultDebugInfoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultDebugInfoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceFaultDebugInfoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceFaultDebugInfoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceFaultDebugInfoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultDebugInfoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultDebugInfoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultDebugInfoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultDebugInfoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultDebugInfoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultDebugInfoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultDebugInfoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultDebugInfoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultDebugInfoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultDebugInfoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultDebugInfoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultDebugInfoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultDebugInfoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceFaultDebugInfoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceFaultDebugInfoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceFaultDebugInfoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultDebugInfoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultDebugInfoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultDebugInfoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultDebugInfoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultDebugInfoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultDebugInfoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultDebugInfoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultDebugInfoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultDebugInfoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultDebugInfoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultDebugInfoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultDebugInfoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultDebugInfoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceFaultDebugInfoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceFaultDebugInfoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceFaultDebugInfoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultDebugInfoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultDebugInfoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultDebugInfoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultDebugInfoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultDebugInfoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultDebugInfoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultDebugInfoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultDebugInfoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultDebugInfoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultDebugInfoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultDebugInfoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultDebugInfoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultDebugInfoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceFaultDebugInfoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceFaultDebugInfoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceFaultDebugInfoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultDebugInfoKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultDebugInfoKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultDebugInfoKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultDebugInfoKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultDebugInfoKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultDebugInfoKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultDebugInfoKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultDebugInfoKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultDebugInfoKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultDebugInfoKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultDebugInfoKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultDebugInfoKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultDebugInfoKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceFaultDebugInfoKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceFaultDebugInfoKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceFaultDebugInfoKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthBias2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDepthBias2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDepthBias2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthBias2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDepthBias2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDepthBias2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthBias2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDepthBias2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDepthBias2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthBias2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDepthBias2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDepthBias2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthBias2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDepthBias2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDepthBias2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthBias2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDepthBias2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDepthBias2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthBias2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDepthBias2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDepthBias2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthBias2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDepthBias2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDepthBias2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthBias2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDepthBias2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDepthBias2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthBias2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDepthBias2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDepthBias2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthBias2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDepthBias2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDepthBias2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthBias2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDepthBias2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDepthBias2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthBias2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDepthBias2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDepthBias2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ReleaseSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ReleaseSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ReleaseSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ReleaseSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ReleaseSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ReleaseSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ReleaseSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ReleaseSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ReleaseSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ReleaseSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ReleaseSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ReleaseSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ReleaseSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ReleaseSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ReleaseSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ReleaseSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ReleaseSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ReleaseSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ReleaseSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ReleaseSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ReleaseSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ReleaseSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ReleaseSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ReleaseSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ReleaseSwapchainImagesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ReleaseSwapchainImagesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ReleaseSwapchainImagesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ReleaseSwapchainImagesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ReleaseSwapchainImagesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ReleaseSwapchainImagesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ReleaseSwapchainImagesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ReleaseSwapchainImagesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ReleaseSwapchainImagesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ReleaseSwapchainImagesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ReleaseSwapchainImagesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ReleaseSwapchainImagesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ReleaseSwapchainImagesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ReleaseSwapchainImagesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ReleaseSwapchainImagesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ReleaseSwapchainImagesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ReleaseSwapchainImagesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ReleaseSwapchainImagesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ReleaseSwapchainImagesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ReleaseSwapchainImagesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ReleaseSwapchainImagesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ReleaseSwapchainImagesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ReleaseSwapchainImagesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ReleaseSwapchainImagesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ReleaseSwapchainImagesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ReleaseSwapchainImagesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ReleaseSwapchainImagesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ReleaseSwapchainImagesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ReleaseSwapchainImagesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ReleaseSwapchainImagesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ReleaseSwapchainImagesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ReleaseSwapchainImagesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ReleaseSwapchainImagesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ReleaseSwapchainImagesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ReleaseSwapchainImagesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ReleaseSwapchainImagesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ReleaseSwapchainImagesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ReleaseSwapchainImagesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ReleaseSwapchainImagesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ReleaseSwapchainImagesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ReleaseSwapchainImagesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayout@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayout@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayout@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayout@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayout@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayout@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayout@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayout@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayout@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayout@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayout@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayout@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayout@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayout@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceImageSubresourceLayout=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceImageSubresourceLayout=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayoutKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayoutKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayoutKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayoutKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayoutKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayoutKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayoutKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayoutKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayoutKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayoutKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayoutKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayoutKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayoutKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceImageSubresourceLayoutKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceImageSubresourceLayoutKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceImageSubresourceLayoutKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayoutKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayoutKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayoutKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayoutKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayoutKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayoutKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayoutKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayoutKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayoutKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayoutKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayoutKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayoutKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayoutKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceImageSubresourceLayoutKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceImageSubresourceLayoutKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceImageSubresourceLayoutKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayoutKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayoutKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayoutKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayoutKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayoutKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayoutKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayoutKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayoutKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayoutKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayoutKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayoutKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayoutKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayoutKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceImageSubresourceLayoutKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceImageSubresourceLayoutKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceImageSubresourceLayoutKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayoutKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayoutKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayoutKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayoutKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayoutKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayoutKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayoutKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayoutKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayoutKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayoutKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayoutKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayoutKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayoutKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceImageSubresourceLayoutKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceImageSubresourceLayoutKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceImageSubresourceLayoutKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayoutKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayoutKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayoutKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayoutKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayoutKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayoutKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayoutKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayoutKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayoutKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayoutKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayoutKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayoutKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayoutKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceImageSubresourceLayoutKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceImageSubresourceLayoutKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceImageSubresourceLayoutKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayoutKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayoutKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayoutKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayoutKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayoutKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayoutKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayoutKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayoutKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayoutKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayoutKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayoutKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayoutKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayoutKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceImageSubresourceLayoutKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceImageSubresourceLayoutKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceImageSubresourceLayoutKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayoutKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayoutKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayoutKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayoutKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayoutKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayoutKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayoutKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayoutKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayoutKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayoutKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayoutKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayoutKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayoutKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceImageSubresourceLayoutKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceImageSubresourceLayoutKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceImageSubresourceLayoutKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayoutKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayoutKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayoutKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayoutKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayoutKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayoutKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayoutKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayoutKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayoutKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayoutKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayoutKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayoutKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayoutKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceImageSubresourceLayoutKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceImageSubresourceLayoutKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceImageSubresourceLayoutKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayoutKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayoutKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayoutKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayoutKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayoutKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayoutKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayoutKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayoutKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayoutKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayoutKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayoutKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayoutKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayoutKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceImageSubresourceLayoutKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceImageSubresourceLayoutKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceImageSubresourceLayoutKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayoutKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayoutKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayoutKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayoutKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayoutKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayoutKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayoutKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayoutKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayoutKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayoutKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayoutKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayoutKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayoutKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceImageSubresourceLayoutKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceImageSubresourceLayoutKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceImageSubresourceLayoutKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayoutKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayoutKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayoutKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayoutKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayoutKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayoutKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayoutKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayoutKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayoutKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayoutKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayoutKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayoutKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayoutKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceImageSubresourceLayoutKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceImageSubresourceLayoutKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceImageSubresourceLayoutKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayoutKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayoutKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayoutKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayoutKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayoutKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayoutKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayoutKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayoutKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayoutKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayoutKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayoutKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayoutKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayoutKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceImageSubresourceLayoutKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceImageSubresourceLayoutKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceImageSubresourceLayoutKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayoutKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayoutKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayoutKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayoutKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayoutKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayoutKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayoutKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayoutKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayoutKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayoutKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayoutKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayoutKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayoutKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceImageSubresourceLayoutKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceImageSubresourceLayoutKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceImageSubresourceLayoutKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_MapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_MapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_MapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_MapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_MapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_MapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_MapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_MapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_MapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_MapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_MapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_MapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_MapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_MapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_MapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_MapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_MapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_MapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_MapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_MapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_MapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_MapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_MapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_MapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_MapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_MapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_MapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_MapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_MapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_MapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_MapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_MapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_MapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_MapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_MapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_MapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_MapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_MapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_MapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_MapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_MapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_MapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_MapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_MapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_MapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_MapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_MapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_MapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_MapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_MapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_MapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_MapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_MapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_MapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_MapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_MapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_MapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_MapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_MapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_MapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_MapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_MapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_MapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_MapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_MapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_MapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_MapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_UnmapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_UnmapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_UnmapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_UnmapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_UnmapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_UnmapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_UnmapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_UnmapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_UnmapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_UnmapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_UnmapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_UnmapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_UnmapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_UnmapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_UnmapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_UnmapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_UnmapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_UnmapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_UnmapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_UnmapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_UnmapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_UnmapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_UnmapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_UnmapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_UnmapMemory2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_UnmapMemory2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnmapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_UnmapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_UnmapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnmapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_UnmapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_UnmapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnmapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_UnmapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_UnmapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnmapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_UnmapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_UnmapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnmapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_UnmapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_UnmapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnmapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_UnmapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_UnmapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnmapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_UnmapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_UnmapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnmapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_UnmapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_UnmapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnmapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_UnmapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_UnmapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnmapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_UnmapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_UnmapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnmapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_UnmapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_UnmapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnmapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_UnmapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_UnmapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnmapMemory2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_UnmapMemory2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_UnmapMemory2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyShaderEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyShaderEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetShaderBinaryDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderBinaryDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderBinaryDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderBinaryDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderBinaryDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderBinaryDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderBinaryDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderBinaryDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderBinaryDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderBinaryDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderBinaryDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderBinaryDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderBinaryDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderBinaryDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetShaderBinaryDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetShaderBinaryDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderBinaryDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderBinaryDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderBinaryDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderBinaryDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderBinaryDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderBinaryDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderBinaryDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderBinaryDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderBinaryDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderBinaryDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderBinaryDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderBinaryDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderBinaryDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderBinaryDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetShaderBinaryDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetShaderBinaryDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderBinaryDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderBinaryDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderBinaryDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderBinaryDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderBinaryDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderBinaryDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderBinaryDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderBinaryDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderBinaryDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderBinaryDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderBinaryDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderBinaryDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderBinaryDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderBinaryDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetShaderBinaryDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetShaderBinaryDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderBinaryDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderBinaryDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderBinaryDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderBinaryDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderBinaryDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderBinaryDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderBinaryDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderBinaryDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderBinaryDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderBinaryDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderBinaryDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderBinaryDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderBinaryDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderBinaryDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetShaderBinaryDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetShaderBinaryDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderBinaryDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderBinaryDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderBinaryDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderBinaryDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderBinaryDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderBinaryDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderBinaryDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderBinaryDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderBinaryDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderBinaryDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderBinaryDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderBinaryDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderBinaryDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderBinaryDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetShaderBinaryDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetShaderBinaryDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderBinaryDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderBinaryDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderBinaryDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderBinaryDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderBinaryDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderBinaryDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderBinaryDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderBinaryDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderBinaryDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderBinaryDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderBinaryDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderBinaryDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderBinaryDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderBinaryDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetShaderBinaryDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetShaderBinaryDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderBinaryDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderBinaryDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderBinaryDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderBinaryDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderBinaryDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderBinaryDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderBinaryDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderBinaryDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderBinaryDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderBinaryDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderBinaryDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderBinaryDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderBinaryDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderBinaryDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetShaderBinaryDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetShaderBinaryDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderBinaryDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderBinaryDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderBinaryDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderBinaryDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderBinaryDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderBinaryDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderBinaryDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderBinaryDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderBinaryDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderBinaryDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderBinaryDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderBinaryDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderBinaryDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderBinaryDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetShaderBinaryDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetShaderBinaryDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderBinaryDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderBinaryDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderBinaryDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderBinaryDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderBinaryDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderBinaryDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderBinaryDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderBinaryDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderBinaryDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderBinaryDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderBinaryDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderBinaryDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderBinaryDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderBinaryDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetShaderBinaryDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetShaderBinaryDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderBinaryDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderBinaryDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderBinaryDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderBinaryDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderBinaryDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderBinaryDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderBinaryDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderBinaryDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderBinaryDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderBinaryDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderBinaryDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderBinaryDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderBinaryDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderBinaryDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetShaderBinaryDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetShaderBinaryDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderBinaryDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderBinaryDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderBinaryDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderBinaryDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderBinaryDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderBinaryDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderBinaryDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderBinaryDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderBinaryDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderBinaryDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderBinaryDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderBinaryDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderBinaryDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderBinaryDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetShaderBinaryDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetShaderBinaryDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderBinaryDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderBinaryDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderBinaryDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderBinaryDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderBinaryDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderBinaryDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderBinaryDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderBinaryDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderBinaryDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderBinaryDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderBinaryDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderBinaryDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderBinaryDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderBinaryDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetShaderBinaryDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetShaderBinaryDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderBinaryDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderBinaryDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderBinaryDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderBinaryDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderBinaryDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderBinaryDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderBinaryDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderBinaryDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderBinaryDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderBinaryDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderBinaryDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderBinaryDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderBinaryDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderBinaryDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetShaderBinaryDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetShaderBinaryDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadersEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadersEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadersEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadersEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadersEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadersEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadersEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadersEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadersEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadersEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadersEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadersEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadersEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindShadersEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindShadersEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindShadersEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_SetSwapchainPresentTimingQueueSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetSwapchainPresentTimingQueueSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetSwapchainPresentTimingQueueSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetSwapchainPresentTimingQueueSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetSwapchainPresentTimingQueueSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetSwapchainPresentTimingQueueSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetSwapchainPresentTimingQueueSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetSwapchainPresentTimingQueueSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetSwapchainPresentTimingQueueSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetSwapchainPresentTimingQueueSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetSwapchainPresentTimingQueueSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetSwapchainPresentTimingQueueSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetSwapchainPresentTimingQueueSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetSwapchainPresentTimingQueueSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_SetSwapchainPresentTimingQueueSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_SetSwapchainPresentTimingQueueSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_SetSwapchainPresentTimingQueueSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetSwapchainPresentTimingQueueSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetSwapchainPresentTimingQueueSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetSwapchainPresentTimingQueueSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetSwapchainPresentTimingQueueSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetSwapchainPresentTimingQueueSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetSwapchainPresentTimingQueueSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetSwapchainPresentTimingQueueSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetSwapchainPresentTimingQueueSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetSwapchainPresentTimingQueueSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetSwapchainPresentTimingQueueSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetSwapchainPresentTimingQueueSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetSwapchainPresentTimingQueueSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetSwapchainPresentTimingQueueSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_SetSwapchainPresentTimingQueueSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_SetSwapchainPresentTimingQueueSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_SetSwapchainPresentTimingQueueSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetSwapchainPresentTimingQueueSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetSwapchainPresentTimingQueueSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetSwapchainPresentTimingQueueSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetSwapchainPresentTimingQueueSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetSwapchainPresentTimingQueueSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetSwapchainPresentTimingQueueSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetSwapchainPresentTimingQueueSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetSwapchainPresentTimingQueueSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetSwapchainPresentTimingQueueSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetSwapchainPresentTimingQueueSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetSwapchainPresentTimingQueueSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetSwapchainPresentTimingQueueSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetSwapchainPresentTimingQueueSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_SetSwapchainPresentTimingQueueSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_SetSwapchainPresentTimingQueueSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_SetSwapchainPresentTimingQueueSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetSwapchainPresentTimingQueueSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetSwapchainPresentTimingQueueSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetSwapchainPresentTimingQueueSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetSwapchainPresentTimingQueueSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetSwapchainPresentTimingQueueSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetSwapchainPresentTimingQueueSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetSwapchainPresentTimingQueueSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetSwapchainPresentTimingQueueSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetSwapchainPresentTimingQueueSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetSwapchainPresentTimingQueueSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetSwapchainPresentTimingQueueSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetSwapchainPresentTimingQueueSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetSwapchainPresentTimingQueueSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_SetSwapchainPresentTimingQueueSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_SetSwapchainPresentTimingQueueSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_SetSwapchainPresentTimingQueueSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetSwapchainPresentTimingQueueSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetSwapchainPresentTimingQueueSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetSwapchainPresentTimingQueueSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetSwapchainPresentTimingQueueSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetSwapchainPresentTimingQueueSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetSwapchainPresentTimingQueueSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetSwapchainPresentTimingQueueSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetSwapchainPresentTimingQueueSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetSwapchainPresentTimingQueueSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetSwapchainPresentTimingQueueSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetSwapchainPresentTimingQueueSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetSwapchainPresentTimingQueueSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetSwapchainPresentTimingQueueSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_SetSwapchainPresentTimingQueueSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_SetSwapchainPresentTimingQueueSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_SetSwapchainPresentTimingQueueSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetSwapchainPresentTimingQueueSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetSwapchainPresentTimingQueueSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetSwapchainPresentTimingQueueSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetSwapchainPresentTimingQueueSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetSwapchainPresentTimingQueueSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetSwapchainPresentTimingQueueSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetSwapchainPresentTimingQueueSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetSwapchainPresentTimingQueueSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetSwapchainPresentTimingQueueSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetSwapchainPresentTimingQueueSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetSwapchainPresentTimingQueueSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetSwapchainPresentTimingQueueSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetSwapchainPresentTimingQueueSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_SetSwapchainPresentTimingQueueSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_SetSwapchainPresentTimingQueueSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_SetSwapchainPresentTimingQueueSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetSwapchainPresentTimingQueueSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetSwapchainPresentTimingQueueSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetSwapchainPresentTimingQueueSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetSwapchainPresentTimingQueueSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetSwapchainPresentTimingQueueSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetSwapchainPresentTimingQueueSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetSwapchainPresentTimingQueueSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetSwapchainPresentTimingQueueSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetSwapchainPresentTimingQueueSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetSwapchainPresentTimingQueueSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetSwapchainPresentTimingQueueSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetSwapchainPresentTimingQueueSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetSwapchainPresentTimingQueueSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_SetSwapchainPresentTimingQueueSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_SetSwapchainPresentTimingQueueSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_SetSwapchainPresentTimingQueueSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetSwapchainPresentTimingQueueSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetSwapchainPresentTimingQueueSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetSwapchainPresentTimingQueueSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetSwapchainPresentTimingQueueSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetSwapchainPresentTimingQueueSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetSwapchainPresentTimingQueueSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetSwapchainPresentTimingQueueSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetSwapchainPresentTimingQueueSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetSwapchainPresentTimingQueueSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetSwapchainPresentTimingQueueSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetSwapchainPresentTimingQueueSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetSwapchainPresentTimingQueueSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetSwapchainPresentTimingQueueSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_SetSwapchainPresentTimingQueueSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_SetSwapchainPresentTimingQueueSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_SetSwapchainPresentTimingQueueSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetSwapchainPresentTimingQueueSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetSwapchainPresentTimingQueueSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetSwapchainPresentTimingQueueSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetSwapchainPresentTimingQueueSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetSwapchainPresentTimingQueueSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetSwapchainPresentTimingQueueSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetSwapchainPresentTimingQueueSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetSwapchainPresentTimingQueueSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetSwapchainPresentTimingQueueSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetSwapchainPresentTimingQueueSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetSwapchainPresentTimingQueueSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetSwapchainPresentTimingQueueSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetSwapchainPresentTimingQueueSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_SetSwapchainPresentTimingQueueSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_SetSwapchainPresentTimingQueueSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_SetSwapchainPresentTimingQueueSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetSwapchainPresentTimingQueueSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetSwapchainPresentTimingQueueSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetSwapchainPresentTimingQueueSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetSwapchainPresentTimingQueueSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetSwapchainPresentTimingQueueSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetSwapchainPresentTimingQueueSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetSwapchainPresentTimingQueueSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetSwapchainPresentTimingQueueSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetSwapchainPresentTimingQueueSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetSwapchainPresentTimingQueueSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetSwapchainPresentTimingQueueSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetSwapchainPresentTimingQueueSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetSwapchainPresentTimingQueueSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_SetSwapchainPresentTimingQueueSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_SetSwapchainPresentTimingQueueSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetSwapchainPresentTimingQueueSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetSwapchainPresentTimingQueueSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetSwapchainPresentTimingQueueSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetSwapchainPresentTimingQueueSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetSwapchainPresentTimingQueueSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetSwapchainPresentTimingQueueSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetSwapchainPresentTimingQueueSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetSwapchainPresentTimingQueueSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetSwapchainPresentTimingQueueSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetSwapchainPresentTimingQueueSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetSwapchainPresentTimingQueueSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetSwapchainPresentTimingQueueSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetSwapchainPresentTimingQueueSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetSwapchainPresentTimingQueueSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_SetSwapchainPresentTimingQueueSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_SetSwapchainPresentTimingQueueSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_SetSwapchainPresentTimingQueueSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetSwapchainPresentTimingQueueSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetSwapchainPresentTimingQueueSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetSwapchainPresentTimingQueueSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetSwapchainPresentTimingQueueSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetSwapchainPresentTimingQueueSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetSwapchainPresentTimingQueueSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetSwapchainPresentTimingQueueSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetSwapchainPresentTimingQueueSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetSwapchainPresentTimingQueueSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetSwapchainPresentTimingQueueSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetSwapchainPresentTimingQueueSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetSwapchainPresentTimingQueueSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetSwapchainPresentTimingQueueSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_SetSwapchainPresentTimingQueueSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_SetSwapchainPresentTimingQueueSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_SetSwapchainPresentTimingQueueSizeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetSwapchainPresentTimingQueueSizeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetSwapchainPresentTimingQueueSizeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetSwapchainPresentTimingQueueSizeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetSwapchainPresentTimingQueueSizeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetSwapchainPresentTimingQueueSizeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetSwapchainPresentTimingQueueSizeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetSwapchainPresentTimingQueueSizeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetSwapchainPresentTimingQueueSizeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetSwapchainPresentTimingQueueSizeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetSwapchainPresentTimingQueueSizeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetSwapchainPresentTimingQueueSizeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetSwapchainPresentTimingQueueSizeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetSwapchainPresentTimingQueueSizeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_SetSwapchainPresentTimingQueueSizeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_SetSwapchainPresentTimingQueueSizeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimingPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimingPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimingPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimingPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimingPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimingPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimingPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimingPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimingPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimingPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimingPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimingPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimingPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimingPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetSwapchainTimingPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetSwapchainTimingPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimingPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimingPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimingPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimingPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimingPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimingPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimingPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimingPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimingPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimingPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimingPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimingPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimingPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimingPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetSwapchainTimingPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetSwapchainTimingPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimingPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimingPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimingPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimingPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimingPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimingPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimingPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimingPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimingPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimingPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimingPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimingPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimingPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimingPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetSwapchainTimingPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetSwapchainTimingPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimingPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimingPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimingPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimingPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimingPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimingPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimingPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimingPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimingPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimingPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimingPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimingPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimingPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimingPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetSwapchainTimingPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetSwapchainTimingPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimingPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimingPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimingPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimingPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimingPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimingPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimingPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimingPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimingPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimingPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimingPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimingPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimingPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimingPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetSwapchainTimingPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetSwapchainTimingPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimingPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimingPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimingPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimingPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimingPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimingPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimingPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimingPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimingPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimingPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimingPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimingPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimingPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimingPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetSwapchainTimingPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetSwapchainTimingPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimingPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimingPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimingPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimingPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimingPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimingPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimingPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimingPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimingPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimingPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimingPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimingPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimingPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimingPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetSwapchainTimingPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetSwapchainTimingPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimingPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimingPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimingPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimingPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimingPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimingPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimingPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimingPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimingPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimingPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimingPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimingPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimingPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimingPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetSwapchainTimingPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetSwapchainTimingPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimingPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimingPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimingPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimingPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimingPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimingPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimingPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimingPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimingPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimingPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimingPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimingPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimingPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimingPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetSwapchainTimingPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetSwapchainTimingPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimingPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimingPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimingPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimingPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimingPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimingPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimingPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimingPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimingPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimingPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimingPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimingPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimingPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimingPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetSwapchainTimingPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetSwapchainTimingPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimingPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimingPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimingPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimingPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimingPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimingPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimingPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimingPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimingPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimingPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimingPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimingPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimingPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimingPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetSwapchainTimingPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetSwapchainTimingPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimingPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimingPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimingPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimingPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimingPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimingPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimingPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimingPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimingPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimingPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimingPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimingPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimingPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimingPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetSwapchainTimingPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetSwapchainTimingPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimingPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimingPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimingPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimingPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimingPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimingPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimingPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimingPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimingPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimingPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimingPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimingPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimingPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimingPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetSwapchainTimingPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetSwapchainTimingPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimeDomainPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimeDomainPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimeDomainPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimeDomainPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimeDomainPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimeDomainPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimeDomainPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimeDomainPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimeDomainPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimeDomainPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimeDomainPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimeDomainPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimeDomainPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetSwapchainTimeDomainPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetSwapchainTimeDomainPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetSwapchainTimeDomainPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimeDomainPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimeDomainPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimeDomainPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimeDomainPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimeDomainPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimeDomainPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimeDomainPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimeDomainPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimeDomainPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimeDomainPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimeDomainPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimeDomainPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimeDomainPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetSwapchainTimeDomainPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetSwapchainTimeDomainPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetSwapchainTimeDomainPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimeDomainPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimeDomainPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimeDomainPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimeDomainPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimeDomainPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimeDomainPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimeDomainPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimeDomainPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimeDomainPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimeDomainPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimeDomainPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimeDomainPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimeDomainPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetSwapchainTimeDomainPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetSwapchainTimeDomainPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetSwapchainTimeDomainPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimeDomainPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimeDomainPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimeDomainPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimeDomainPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimeDomainPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimeDomainPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimeDomainPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimeDomainPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimeDomainPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimeDomainPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimeDomainPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimeDomainPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimeDomainPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetSwapchainTimeDomainPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetSwapchainTimeDomainPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetSwapchainTimeDomainPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimeDomainPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimeDomainPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimeDomainPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimeDomainPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimeDomainPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimeDomainPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimeDomainPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimeDomainPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimeDomainPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimeDomainPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimeDomainPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimeDomainPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimeDomainPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetSwapchainTimeDomainPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetSwapchainTimeDomainPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetSwapchainTimeDomainPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimeDomainPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimeDomainPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimeDomainPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimeDomainPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimeDomainPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimeDomainPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimeDomainPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimeDomainPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimeDomainPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimeDomainPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimeDomainPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimeDomainPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimeDomainPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetSwapchainTimeDomainPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetSwapchainTimeDomainPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetSwapchainTimeDomainPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimeDomainPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimeDomainPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimeDomainPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimeDomainPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimeDomainPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimeDomainPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimeDomainPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimeDomainPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimeDomainPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimeDomainPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimeDomainPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimeDomainPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimeDomainPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetSwapchainTimeDomainPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetSwapchainTimeDomainPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetSwapchainTimeDomainPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimeDomainPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimeDomainPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimeDomainPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimeDomainPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimeDomainPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimeDomainPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimeDomainPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimeDomainPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimeDomainPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimeDomainPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimeDomainPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimeDomainPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimeDomainPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetSwapchainTimeDomainPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetSwapchainTimeDomainPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetSwapchainTimeDomainPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimeDomainPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimeDomainPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimeDomainPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimeDomainPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimeDomainPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimeDomainPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimeDomainPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimeDomainPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimeDomainPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimeDomainPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimeDomainPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimeDomainPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimeDomainPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetSwapchainTimeDomainPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetSwapchainTimeDomainPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetSwapchainTimeDomainPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimeDomainPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimeDomainPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimeDomainPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimeDomainPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimeDomainPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimeDomainPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimeDomainPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimeDomainPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimeDomainPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimeDomainPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimeDomainPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimeDomainPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimeDomainPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetSwapchainTimeDomainPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetSwapchainTimeDomainPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetSwapchainTimeDomainPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimeDomainPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimeDomainPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimeDomainPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimeDomainPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimeDomainPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimeDomainPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimeDomainPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimeDomainPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimeDomainPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimeDomainPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimeDomainPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimeDomainPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimeDomainPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetSwapchainTimeDomainPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetSwapchainTimeDomainPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetSwapchainTimeDomainPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimeDomainPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimeDomainPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimeDomainPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimeDomainPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimeDomainPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimeDomainPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimeDomainPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimeDomainPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimeDomainPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimeDomainPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimeDomainPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimeDomainPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimeDomainPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetSwapchainTimeDomainPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetSwapchainTimeDomainPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetSwapchainTimeDomainPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimeDomainPropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimeDomainPropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimeDomainPropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimeDomainPropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimeDomainPropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimeDomainPropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimeDomainPropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimeDomainPropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimeDomainPropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimeDomainPropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimeDomainPropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimeDomainPropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimeDomainPropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetSwapchainTimeDomainPropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetSwapchainTimeDomainPropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetSwapchainTimeDomainPropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetPastPresentationTimingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetPastPresentationTimingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetPastPresentationTimingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetPastPresentationTimingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetPastPresentationTimingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetPastPresentationTimingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetPastPresentationTimingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetPastPresentationTimingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetPastPresentationTimingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetPastPresentationTimingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetPastPresentationTimingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetPastPresentationTimingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetPastPresentationTimingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetPastPresentationTimingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetPastPresentationTimingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetPastPresentationTimingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetPastPresentationTimingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetPastPresentationTimingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetPastPresentationTimingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetPastPresentationTimingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetPastPresentationTimingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetPastPresentationTimingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetPastPresentationTimingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetPastPresentationTimingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetPastPresentationTimingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetPastPresentationTimingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetPastPresentationTimingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetPastPresentationTimingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetPastPresentationTimingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetPastPresentationTimingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetPastPresentationTimingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetPastPresentationTimingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetPastPresentationTimingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetPastPresentationTimingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetPastPresentationTimingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetPastPresentationTimingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetPastPresentationTimingEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetPastPresentationTimingEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetPastPresentationTimingEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetScreenBufferPropertiesQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetScreenBufferPropertiesQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetScreenBufferPropertiesQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetScreenBufferPropertiesQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetScreenBufferPropertiesQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetScreenBufferPropertiesQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetScreenBufferPropertiesQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetScreenBufferPropertiesQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetScreenBufferPropertiesQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetScreenBufferPropertiesQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetScreenBufferPropertiesQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetScreenBufferPropertiesQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetScreenBufferPropertiesQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetScreenBufferPropertiesQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetScreenBufferPropertiesQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetScreenBufferPropertiesQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetScreenBufferPropertiesQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetScreenBufferPropertiesQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetScreenBufferPropertiesQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetScreenBufferPropertiesQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetScreenBufferPropertiesQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetScreenBufferPropertiesQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetScreenBufferPropertiesQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetScreenBufferPropertiesQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetScreenBufferPropertiesQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetScreenBufferPropertiesQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetScreenBufferPropertiesQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetScreenBufferPropertiesQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetScreenBufferPropertiesQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetScreenBufferPropertiesQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetScreenBufferPropertiesQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetScreenBufferPropertiesQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetScreenBufferPropertiesQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetScreenBufferPropertiesQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetScreenBufferPropertiesQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetScreenBufferPropertiesQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetScreenBufferPropertiesQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetScreenBufferPropertiesQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetScreenBufferPropertiesQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetScreenBufferPropertiesQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetScreenBufferPropertiesQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetScreenBufferPropertiesQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetScreenBufferPropertiesQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetScreenBufferPropertiesQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetScreenBufferPropertiesQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetScreenBufferPropertiesQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetScreenBufferPropertiesQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetScreenBufferPropertiesQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetScreenBufferPropertiesQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetScreenBufferPropertiesQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetScreenBufferPropertiesQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetScreenBufferPropertiesQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetScreenBufferPropertiesQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetScreenBufferPropertiesQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetScreenBufferPropertiesQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetScreenBufferPropertiesQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetScreenBufferPropertiesQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetScreenBufferPropertiesQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetScreenBufferPropertiesQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetScreenBufferPropertiesQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetScreenBufferPropertiesQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetScreenBufferPropertiesQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetScreenBufferPropertiesQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetScreenBufferPropertiesQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetScreenBufferPropertiesQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetScreenBufferPropertiesQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetScreenBufferPropertiesQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetScreenBufferPropertiesQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetScreenBufferPropertiesQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetScreenBufferPropertiesQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetScreenBufferPropertiesQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetScreenBufferPropertiesQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetScreenBufferPropertiesQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetScreenBufferPropertiesQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetScreenBufferPropertiesQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetScreenBufferPropertiesQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetScreenBufferPropertiesQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetScreenBufferPropertiesQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetScreenBufferPropertiesQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetScreenBufferPropertiesQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetScreenBufferPropertiesQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetScreenBufferPropertiesQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetScreenBufferPropertiesQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetScreenBufferPropertiesQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetScreenBufferPropertiesQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetScreenBufferPropertiesQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetScreenBufferPropertiesQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetScreenBufferPropertiesQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetScreenBufferPropertiesQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetScreenBufferPropertiesQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetScreenBufferPropertiesQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetScreenBufferPropertiesQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetScreenBufferPropertiesQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetScreenBufferPropertiesQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetScreenBufferPropertiesQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetScreenBufferPropertiesQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetScreenBufferPropertiesQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetScreenBufferPropertiesQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetScreenBufferPropertiesQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetScreenBufferPropertiesQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetScreenBufferPropertiesQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetScreenBufferPropertiesQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetScreenBufferPropertiesQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetScreenBufferPropertiesQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetScreenBufferPropertiesQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetScreenBufferPropertiesQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetScreenBufferPropertiesQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetScreenBufferPropertiesQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetScreenBufferPropertiesQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetScreenBufferPropertiesQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetScreenBufferPropertiesQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetScreenBufferPropertiesQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetScreenBufferPropertiesQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetScreenBufferPropertiesQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetScreenBufferPropertiesQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetScreenBufferPropertiesQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetScreenBufferPropertiesQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetScreenBufferPropertiesQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetScreenBufferPropertiesQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetScreenBufferPropertiesQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetScreenBufferPropertiesQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetScreenBufferPropertiesQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetScreenBufferPropertiesQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetScreenBufferPropertiesQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetScreenBufferPropertiesQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetScreenBufferPropertiesQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetScreenBufferPropertiesQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetScreenBufferPropertiesQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetScreenBufferPropertiesQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetScreenBufferPropertiesQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetScreenBufferPropertiesQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetScreenBufferPropertiesQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetScreenBufferPropertiesQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetScreenBufferPropertiesQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetScreenBufferPropertiesQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetScreenBufferPropertiesQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetScreenBufferPropertiesQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetScreenBufferPropertiesQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetScreenBufferPropertiesQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetScreenBufferPropertiesQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetScreenBufferPropertiesQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetScreenBufferPropertiesQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetScreenBufferPropertiesQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetScreenBufferPropertiesQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetScreenBufferPropertiesQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetScreenBufferPropertiesQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetScreenBufferPropertiesQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetScreenBufferPropertiesQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetScreenBufferPropertiesQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetScreenBufferPropertiesQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetScreenBufferPropertiesQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetScreenBufferPropertiesQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetScreenBufferPropertiesQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetScreenBufferPropertiesQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetScreenBufferPropertiesQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetScreenBufferPropertiesQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetScreenBufferPropertiesQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetScreenBufferPropertiesQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetScreenBufferPropertiesQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetScreenBufferPropertiesQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetScreenBufferPropertiesQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetScreenBufferPropertiesQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetScreenBufferPropertiesQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetScreenBufferPropertiesQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetScreenBufferPropertiesQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetScreenBufferPropertiesQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetScreenBufferPropertiesQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetScreenBufferPropertiesQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetScreenBufferPropertiesQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetScreenBufferPropertiesQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetScreenBufferPropertiesQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetScreenBufferPropertiesQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetScreenBufferPropertiesQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetScreenBufferPropertiesQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetScreenBufferPropertiesQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetScreenBufferPropertiesQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetScreenBufferPropertiesQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetScreenBufferPropertiesQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetScreenBufferPropertiesQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetScreenBufferPropertiesQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetScreenBufferPropertiesQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetScreenBufferPropertiesQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetScreenBufferPropertiesQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetScreenBufferPropertiesQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetScreenBufferPropertiesQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetScreenBufferPropertiesQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetScreenBufferPropertiesQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetScreenBufferPropertiesQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetScreenBufferPropertiesQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetScreenBufferPropertiesQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetScreenBufferPropertiesQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetScreenBufferPropertiesQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetScreenBufferPropertiesQNX@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetScreenBufferPropertiesQNX@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetScreenBufferPropertiesQNX@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetScreenBufferPropertiesQNX@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetScreenBufferPropertiesQNX@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetScreenBufferPropertiesQNX@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetScreenBufferPropertiesQNX@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetScreenBufferPropertiesQNX@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetScreenBufferPropertiesQNX@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetScreenBufferPropertiesQNX@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetScreenBufferPropertiesQNX@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetScreenBufferPropertiesQNX@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetScreenBufferPropertiesQNX@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetScreenBufferPropertiesQNX@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetScreenBufferPropertiesQNX=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetScreenBufferPropertiesQNX=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_SCREEN_QNX
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindDescriptorSets2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindDescriptorSets2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindDescriptorSets2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindDescriptorSets2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindDescriptorSets2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindDescriptorSets2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindDescriptorSets2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindDescriptorSets2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindDescriptorSets2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindDescriptorSets2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindDescriptorSets2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindDescriptorSets2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindDescriptorSets2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindDescriptorSets2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindDescriptorSets2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindDescriptorSets2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindDescriptorSets2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindDescriptorSets2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindDescriptorSets2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindDescriptorSets2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindDescriptorSets2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindDescriptorSets2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindDescriptorSets2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindDescriptorSets2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindDescriptorSets2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindDescriptorSets2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorSets2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindDescriptorSets2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindDescriptorSets2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorSets2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindDescriptorSets2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindDescriptorSets2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorSets2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindDescriptorSets2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindDescriptorSets2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorSets2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindDescriptorSets2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindDescriptorSets2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorSets2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindDescriptorSets2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindDescriptorSets2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorSets2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindDescriptorSets2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindDescriptorSets2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorSets2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindDescriptorSets2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindDescriptorSets2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorSets2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindDescriptorSets2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindDescriptorSets2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorSets2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindDescriptorSets2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindDescriptorSets2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorSets2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindDescriptorSets2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindDescriptorSets2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorSets2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindDescriptorSets2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindDescriptorSets2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorSets2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindDescriptorSets2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindDescriptorSets2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorSets2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindDescriptorSets2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindDescriptorSets2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdPushConstants2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdPushConstants2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdPushConstants2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdPushConstants2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdPushConstants2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdPushConstants2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdPushConstants2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdPushConstants2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdPushConstants2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdPushConstants2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdPushConstants2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdPushConstants2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdPushConstants2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdPushConstants2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdPushConstants2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdPushConstants2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdPushConstants2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdPushConstants2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdPushConstants2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdPushConstants2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdPushConstants2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdPushConstants2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdPushConstants2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdPushConstants2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdPushConstants2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdPushConstants2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushConstants2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdPushConstants2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdPushConstants2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushConstants2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdPushConstants2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdPushConstants2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushConstants2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdPushConstants2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdPushConstants2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushConstants2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdPushConstants2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdPushConstants2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushConstants2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdPushConstants2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdPushConstants2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushConstants2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdPushConstants2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdPushConstants2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushConstants2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdPushConstants2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdPushConstants2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushConstants2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdPushConstants2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdPushConstants2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushConstants2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdPushConstants2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdPushConstants2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushConstants2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdPushConstants2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdPushConstants2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushConstants2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdPushConstants2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdPushConstants2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushConstants2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdPushConstants2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdPushConstants2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushConstants2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdPushConstants2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdPushConstants2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdPushDescriptorSet2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdPushDescriptorSet2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdPushDescriptorSet2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdPushDescriptorSet2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdPushDescriptorSet2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdPushDescriptorSet2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdPushDescriptorSet2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdPushDescriptorSet2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdPushDescriptorSet2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdPushDescriptorSet2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdPushDescriptorSet2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdPushDescriptorSet2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdPushDescriptorSet2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdPushDescriptorSet2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdPushDescriptorSet2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdPushDescriptorSet2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdPushDescriptorSet2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdPushDescriptorSet2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdPushDescriptorSet2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdPushDescriptorSet2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdPushDescriptorSet2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdPushDescriptorSet2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdPushDescriptorSet2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdPushDescriptorSet2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdPushDescriptorSet2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdPushDescriptorSet2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSet2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdPushDescriptorSet2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdPushDescriptorSet2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSet2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdPushDescriptorSet2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdPushDescriptorSet2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSet2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdPushDescriptorSet2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdPushDescriptorSet2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSet2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdPushDescriptorSet2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdPushDescriptorSet2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSet2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdPushDescriptorSet2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdPushDescriptorSet2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSet2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdPushDescriptorSet2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdPushDescriptorSet2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSet2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdPushDescriptorSet2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdPushDescriptorSet2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSet2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdPushDescriptorSet2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdPushDescriptorSet2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSet2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdPushDescriptorSet2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdPushDescriptorSet2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSet2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdPushDescriptorSet2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdPushDescriptorSet2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSet2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdPushDescriptorSet2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdPushDescriptorSet2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSet2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdPushDescriptorSet2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdPushDescriptorSet2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSet2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdPushDescriptorSet2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdPushDescriptorSet2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdPushDescriptorSetWithTemplate2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdPushDescriptorSetWithTemplate2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdPushDescriptorSetWithTemplate2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdPushDescriptorSetWithTemplate2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdPushDescriptorSetWithTemplate2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdPushDescriptorSetWithTemplate2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdPushDescriptorSetWithTemplate2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdPushDescriptorSetWithTemplate2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdPushDescriptorSetWithTemplate2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdPushDescriptorSetWithTemplate2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdPushDescriptorSetWithTemplate2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdPushDescriptorSetWithTemplate2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdPushDescriptorSetWithTemplate2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdPushDescriptorSetWithTemplate2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdPushDescriptorSetWithTemplate2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdPushDescriptorSetWithTemplate2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdPushDescriptorSetWithTemplate2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdPushDescriptorSetWithTemplate2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdPushDescriptorSetWithTemplate2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdPushDescriptorSetWithTemplate2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdPushDescriptorSetWithTemplate2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdPushDescriptorSetWithTemplate2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdPushDescriptorSetWithTemplate2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdPushDescriptorSetWithTemplate2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdPushDescriptorSetWithTemplate2=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdPushDescriptorSetWithTemplate2=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDescriptorSetWithTemplate2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdPushDescriptorSetWithTemplate2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdPushDescriptorSetWithTemplate2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDescriptorSetWithTemplate2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdPushDescriptorSetWithTemplate2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdPushDescriptorSetWithTemplate2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDescriptorSetWithTemplate2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdPushDescriptorSetWithTemplate2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdPushDescriptorSetWithTemplate2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDescriptorSetWithTemplate2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdPushDescriptorSetWithTemplate2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdPushDescriptorSetWithTemplate2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDescriptorSetWithTemplate2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdPushDescriptorSetWithTemplate2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdPushDescriptorSetWithTemplate2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDescriptorSetWithTemplate2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdPushDescriptorSetWithTemplate2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdPushDescriptorSetWithTemplate2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDescriptorSetWithTemplate2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdPushDescriptorSetWithTemplate2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdPushDescriptorSetWithTemplate2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDescriptorSetWithTemplate2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdPushDescriptorSetWithTemplate2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdPushDescriptorSetWithTemplate2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDescriptorSetWithTemplate2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdPushDescriptorSetWithTemplate2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdPushDescriptorSetWithTemplate2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDescriptorSetWithTemplate2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdPushDescriptorSetWithTemplate2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdPushDescriptorSetWithTemplate2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDescriptorSetWithTemplate2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdPushDescriptorSetWithTemplate2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdPushDescriptorSetWithTemplate2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDescriptorSetWithTemplate2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdPushDescriptorSetWithTemplate2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdPushDescriptorSetWithTemplate2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDescriptorSetWithTemplate2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdPushDescriptorSetWithTemplate2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdPushDescriptorSetWithTemplate2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsets2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsets2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsets2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsets2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsets2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsets2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsets2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsets2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsets2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsets2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsets2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsets2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsets2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDescriptorBufferOffsets2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDescriptorBufferOffsets2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDescriptorBufferOffsets2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsets2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsets2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsets2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsets2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsets2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsets2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsets2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsets2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsets2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsets2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsets2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsets2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsets2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDescriptorBufferOffsets2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDescriptorBufferOffsets2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDescriptorBufferOffsets2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsets2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsets2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsets2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsets2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsets2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsets2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsets2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsets2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsets2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsets2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsets2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsets2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsets2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDescriptorBufferOffsets2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDescriptorBufferOffsets2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDescriptorBufferOffsets2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsets2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsets2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsets2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsets2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsets2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsets2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsets2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsets2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsets2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsets2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsets2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsets2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsets2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDescriptorBufferOffsets2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDescriptorBufferOffsets2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDescriptorBufferOffsets2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsets2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsets2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsets2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsets2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsets2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsets2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsets2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsets2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsets2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsets2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsets2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsets2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsets2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDescriptorBufferOffsets2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDescriptorBufferOffsets2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDescriptorBufferOffsets2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsets2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsets2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsets2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsets2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsets2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsets2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsets2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsets2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsets2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsets2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsets2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsets2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsets2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDescriptorBufferOffsets2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDescriptorBufferOffsets2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDescriptorBufferOffsets2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsets2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsets2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsets2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsets2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsets2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsets2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsets2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsets2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsets2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsets2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsets2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsets2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsets2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDescriptorBufferOffsets2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDescriptorBufferOffsets2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDescriptorBufferOffsets2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsets2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsets2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsets2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsets2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsets2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsets2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsets2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsets2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsets2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsets2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsets2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsets2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsets2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDescriptorBufferOffsets2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDescriptorBufferOffsets2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDescriptorBufferOffsets2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsets2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsets2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsets2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsets2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsets2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsets2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsets2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsets2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsets2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsets2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsets2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsets2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsets2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDescriptorBufferOffsets2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDescriptorBufferOffsets2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDescriptorBufferOffsets2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsets2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsets2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsets2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsets2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsets2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsets2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsets2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsets2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsets2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsets2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsets2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsets2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsets2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDescriptorBufferOffsets2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDescriptorBufferOffsets2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDescriptorBufferOffsets2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsets2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsets2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsets2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsets2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsets2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsets2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsets2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsets2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsets2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsets2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsets2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsets2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsets2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDescriptorBufferOffsets2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDescriptorBufferOffsets2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDescriptorBufferOffsets2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsets2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsets2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsets2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsets2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsets2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsets2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsets2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsets2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsets2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsets2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsets2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsets2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsets2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDescriptorBufferOffsets2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDescriptorBufferOffsets2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDescriptorBufferOffsets2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsets2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsets2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsets2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsets2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsets2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsets2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsets2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsets2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsets2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsets2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsets2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsets2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsets2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDescriptorBufferOffsets2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDescriptorBufferOffsets2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDescriptorBufferOffsets2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindDescriptorBufferEmbeddedSamplers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindDescriptorBufferEmbeddedSamplers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindDescriptorBufferEmbeddedSamplers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindDescriptorBufferEmbeddedSamplers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindDescriptorBufferEmbeddedSamplers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindDescriptorBufferEmbeddedSamplers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindDescriptorBufferEmbeddedSamplers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindDescriptorBufferEmbeddedSamplers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindDescriptorBufferEmbeddedSamplers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindDescriptorBufferEmbeddedSamplers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindDescriptorBufferEmbeddedSamplers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindDescriptorBufferEmbeddedSamplers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindDescriptorBufferEmbeddedSamplers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindDescriptorBufferEmbeddedSamplers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindDescriptorBufferEmbeddedSamplers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindDescriptorBufferEmbeddedSamplers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindDescriptorBufferEmbeddedSamplers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindDescriptorBufferEmbeddedSamplers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindDescriptorBufferEmbeddedSamplers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindDescriptorBufferEmbeddedSamplers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindDescriptorBufferEmbeddedSamplers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindDescriptorBufferEmbeddedSamplers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindDescriptorBufferEmbeddedSamplers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindDescriptorBufferEmbeddedSamplers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindDescriptorBufferEmbeddedSamplers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindDescriptorBufferEmbeddedSamplers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindDescriptorBufferEmbeddedSamplers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindDescriptorBufferEmbeddedSamplers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindDescriptorBufferEmbeddedSamplers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindDescriptorBufferEmbeddedSamplers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindDescriptorBufferEmbeddedSamplers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindDescriptorBufferEmbeddedSamplers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindDescriptorBufferEmbeddedSamplers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindDescriptorBufferEmbeddedSamplers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindDescriptorBufferEmbeddedSamplers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_SetLatencySleepModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencySleepModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencySleepModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencySleepModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencySleepModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencySleepModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencySleepModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencySleepModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencySleepModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencySleepModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencySleepModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencySleepModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencySleepModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencySleepModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_SetLatencySleepModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_SetLatencySleepModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencySleepModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencySleepModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencySleepModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencySleepModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencySleepModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencySleepModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencySleepModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencySleepModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencySleepModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencySleepModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencySleepModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencySleepModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencySleepModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencySleepModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_SetLatencySleepModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_SetLatencySleepModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencySleepModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencySleepModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencySleepModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencySleepModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencySleepModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencySleepModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencySleepModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencySleepModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencySleepModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencySleepModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencySleepModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencySleepModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencySleepModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencySleepModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_SetLatencySleepModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_SetLatencySleepModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencySleepModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencySleepModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencySleepModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencySleepModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencySleepModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencySleepModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencySleepModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencySleepModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencySleepModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencySleepModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencySleepModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencySleepModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencySleepModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencySleepModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_SetLatencySleepModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_SetLatencySleepModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencySleepModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencySleepModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencySleepModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencySleepModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencySleepModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencySleepModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencySleepModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencySleepModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencySleepModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencySleepModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencySleepModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencySleepModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencySleepModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencySleepModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_SetLatencySleepModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_SetLatencySleepModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencySleepModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencySleepModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencySleepModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencySleepModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencySleepModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencySleepModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencySleepModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencySleepModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencySleepModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencySleepModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencySleepModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencySleepModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencySleepModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencySleepModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_SetLatencySleepModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_SetLatencySleepModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencySleepModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencySleepModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencySleepModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencySleepModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencySleepModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencySleepModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencySleepModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencySleepModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencySleepModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencySleepModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencySleepModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencySleepModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencySleepModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencySleepModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_SetLatencySleepModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_SetLatencySleepModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencySleepModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencySleepModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencySleepModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencySleepModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencySleepModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencySleepModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencySleepModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencySleepModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencySleepModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencySleepModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencySleepModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencySleepModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencySleepModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencySleepModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_SetLatencySleepModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_SetLatencySleepModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencySleepModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencySleepModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencySleepModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencySleepModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencySleepModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencySleepModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencySleepModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencySleepModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencySleepModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencySleepModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencySleepModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencySleepModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencySleepModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencySleepModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_SetLatencySleepModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_SetLatencySleepModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencySleepModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencySleepModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencySleepModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencySleepModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencySleepModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencySleepModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencySleepModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencySleepModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencySleepModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencySleepModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencySleepModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencySleepModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencySleepModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencySleepModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_SetLatencySleepModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_SetLatencySleepModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencySleepModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencySleepModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencySleepModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencySleepModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencySleepModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencySleepModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencySleepModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencySleepModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencySleepModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencySleepModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencySleepModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencySleepModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencySleepModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencySleepModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_SetLatencySleepModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_SetLatencySleepModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencySleepModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencySleepModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencySleepModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencySleepModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencySleepModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencySleepModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencySleepModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencySleepModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencySleepModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencySleepModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencySleepModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencySleepModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencySleepModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencySleepModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_SetLatencySleepModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_SetLatencySleepModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencySleepModeNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencySleepModeNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencySleepModeNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencySleepModeNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencySleepModeNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencySleepModeNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencySleepModeNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencySleepModeNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencySleepModeNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencySleepModeNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencySleepModeNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencySleepModeNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencySleepModeNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencySleepModeNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_SetLatencySleepModeNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_SetLatencySleepModeNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_LatencySleepNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_LatencySleepNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_LatencySleepNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_LatencySleepNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_LatencySleepNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_LatencySleepNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_LatencySleepNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_LatencySleepNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_LatencySleepNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_LatencySleepNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_LatencySleepNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_LatencySleepNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_LatencySleepNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_LatencySleepNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_LatencySleepNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_LatencySleepNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_LatencySleepNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_LatencySleepNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_LatencySleepNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_LatencySleepNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_LatencySleepNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_LatencySleepNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_LatencySleepNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_LatencySleepNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_LatencySleepNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_LatencySleepNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_LatencySleepNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_LatencySleepNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_LatencySleepNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_LatencySleepNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_LatencySleepNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_LatencySleepNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_LatencySleepNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_LatencySleepNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_LatencySleepNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_LatencySleepNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_LatencySleepNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_LatencySleepNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_LatencySleepNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_LatencySleepNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_LatencySleepNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_LatencySleepNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_LatencySleepNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_LatencySleepNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_LatencySleepNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_LatencySleepNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_LatencySleepNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_LatencySleepNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_LatencySleepNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_LatencySleepNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_LatencySleepNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_LatencySleepNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_LatencySleepNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_LatencySleepNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_LatencySleepNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_LatencySleepNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_LatencySleepNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_LatencySleepNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_LatencySleepNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_LatencySleepNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_LatencySleepNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_LatencySleepNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_LatencySleepNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_LatencySleepNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_LatencySleepNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_LatencySleepNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_LatencySleepNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_LatencySleepNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_LatencySleepNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_LatencySleepNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_LatencySleepNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_LatencySleepNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_LatencySleepNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_LatencySleepNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_LatencySleepNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_LatencySleepNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_LatencySleepNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_LatencySleepNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_LatencySleepNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_LatencySleepNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_LatencySleepNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_LatencySleepNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_LatencySleepNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_LatencySleepNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_LatencySleepNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_LatencySleepNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_LatencySleepNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_LatencySleepNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_LatencySleepNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_LatencySleepNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_LatencySleepNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_LatencySleepNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_LatencySleepNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_LatencySleepNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_LatencySleepNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_LatencySleepNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_LatencySleepNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_LatencySleepNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_LatencySleepNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_LatencySleepNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_LatencySleepNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_LatencySleepNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_LatencySleepNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_LatencySleepNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_LatencySleepNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_LatencySleepNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_LatencySleepNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_LatencySleepNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_LatencySleepNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_LatencySleepNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_LatencySleepNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_LatencySleepNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_LatencySleepNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_LatencySleepNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_LatencySleepNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_LatencySleepNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_LatencySleepNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_LatencySleepNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_LatencySleepNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_LatencySleepNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_LatencySleepNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_LatencySleepNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_LatencySleepNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_LatencySleepNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_LatencySleepNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_LatencySleepNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_LatencySleepNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_LatencySleepNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_LatencySleepNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_LatencySleepNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_LatencySleepNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_LatencySleepNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_LatencySleepNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_LatencySleepNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_LatencySleepNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_LatencySleepNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_LatencySleepNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_LatencySleepNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_LatencySleepNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_LatencySleepNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_LatencySleepNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_LatencySleepNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_LatencySleepNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_LatencySleepNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_LatencySleepNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_LatencySleepNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_LatencySleepNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_LatencySleepNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_LatencySleepNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_LatencySleepNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_LatencySleepNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_LatencySleepNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_LatencySleepNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_LatencySleepNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_LatencySleepNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_LatencySleepNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_LatencySleepNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_LatencySleepNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_LatencySleepNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_LatencySleepNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_LatencySleepNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_LatencySleepNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_LatencySleepNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_LatencySleepNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_LatencySleepNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_LatencySleepNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_LatencySleepNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_LatencySleepNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_LatencySleepNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_LatencySleepNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_LatencySleepNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_LatencySleepNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_LatencySleepNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_LatencySleepNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_LatencySleepNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_LatencySleepNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_LatencySleepNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_LatencySleepNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_LatencySleepNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_LatencySleepNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_LatencySleepNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_LatencySleepNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_LatencySleepNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_LatencySleepNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_LatencySleepNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_LatencySleepNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_LatencySleepNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_LatencySleepNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_LatencySleepNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_LatencySleepNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_LatencySleepNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_LatencySleepNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_LatencySleepNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_LatencySleepNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_LatencySleepNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_LatencySleepNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_LatencySleepNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_LatencySleepNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_LatencySleepNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_LatencySleepNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_LatencySleepNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_LatencySleepNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_LatencySleepNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_LatencySleepNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_LatencySleepNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_LatencySleepNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_LatencySleepNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_LatencySleepNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_SetLatencyMarkerNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencyMarkerNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencyMarkerNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencyMarkerNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencyMarkerNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencyMarkerNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencyMarkerNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencyMarkerNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencyMarkerNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencyMarkerNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencyMarkerNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencyMarkerNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencyMarkerNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_SetLatencyMarkerNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_SetLatencyMarkerNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_SetLatencyMarkerNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencyMarkerNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencyMarkerNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencyMarkerNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencyMarkerNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencyMarkerNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencyMarkerNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencyMarkerNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencyMarkerNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencyMarkerNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencyMarkerNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencyMarkerNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencyMarkerNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencyMarkerNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_SetLatencyMarkerNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_SetLatencyMarkerNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_SetLatencyMarkerNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencyMarkerNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencyMarkerNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencyMarkerNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencyMarkerNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencyMarkerNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencyMarkerNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencyMarkerNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencyMarkerNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencyMarkerNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencyMarkerNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencyMarkerNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencyMarkerNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencyMarkerNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_SetLatencyMarkerNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_SetLatencyMarkerNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_SetLatencyMarkerNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencyMarkerNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencyMarkerNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencyMarkerNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencyMarkerNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencyMarkerNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencyMarkerNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencyMarkerNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencyMarkerNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencyMarkerNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencyMarkerNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencyMarkerNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencyMarkerNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencyMarkerNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_SetLatencyMarkerNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_SetLatencyMarkerNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_SetLatencyMarkerNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencyMarkerNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencyMarkerNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencyMarkerNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencyMarkerNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencyMarkerNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencyMarkerNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencyMarkerNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencyMarkerNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencyMarkerNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencyMarkerNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencyMarkerNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencyMarkerNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencyMarkerNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_SetLatencyMarkerNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_SetLatencyMarkerNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_SetLatencyMarkerNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencyMarkerNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencyMarkerNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencyMarkerNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencyMarkerNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencyMarkerNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencyMarkerNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencyMarkerNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencyMarkerNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencyMarkerNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencyMarkerNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencyMarkerNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencyMarkerNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencyMarkerNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_SetLatencyMarkerNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_SetLatencyMarkerNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_SetLatencyMarkerNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencyMarkerNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencyMarkerNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencyMarkerNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencyMarkerNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencyMarkerNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencyMarkerNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencyMarkerNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencyMarkerNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencyMarkerNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencyMarkerNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencyMarkerNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencyMarkerNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencyMarkerNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_SetLatencyMarkerNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_SetLatencyMarkerNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_SetLatencyMarkerNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencyMarkerNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencyMarkerNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencyMarkerNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencyMarkerNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencyMarkerNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencyMarkerNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencyMarkerNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencyMarkerNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencyMarkerNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencyMarkerNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencyMarkerNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencyMarkerNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencyMarkerNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_SetLatencyMarkerNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_SetLatencyMarkerNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_SetLatencyMarkerNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencyMarkerNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencyMarkerNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencyMarkerNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencyMarkerNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencyMarkerNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencyMarkerNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencyMarkerNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencyMarkerNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencyMarkerNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencyMarkerNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencyMarkerNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencyMarkerNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencyMarkerNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_SetLatencyMarkerNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_SetLatencyMarkerNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_SetLatencyMarkerNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencyMarkerNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencyMarkerNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencyMarkerNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencyMarkerNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencyMarkerNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencyMarkerNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencyMarkerNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencyMarkerNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencyMarkerNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencyMarkerNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencyMarkerNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencyMarkerNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencyMarkerNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_SetLatencyMarkerNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_SetLatencyMarkerNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_SetLatencyMarkerNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencyMarkerNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencyMarkerNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencyMarkerNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencyMarkerNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencyMarkerNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencyMarkerNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencyMarkerNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencyMarkerNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencyMarkerNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencyMarkerNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencyMarkerNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencyMarkerNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencyMarkerNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_SetLatencyMarkerNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_SetLatencyMarkerNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_SetLatencyMarkerNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencyMarkerNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencyMarkerNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencyMarkerNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencyMarkerNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencyMarkerNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencyMarkerNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencyMarkerNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencyMarkerNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencyMarkerNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencyMarkerNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencyMarkerNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencyMarkerNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencyMarkerNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_SetLatencyMarkerNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_SetLatencyMarkerNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_SetLatencyMarkerNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencyMarkerNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencyMarkerNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencyMarkerNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencyMarkerNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencyMarkerNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencyMarkerNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencyMarkerNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencyMarkerNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencyMarkerNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencyMarkerNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencyMarkerNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencyMarkerNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencyMarkerNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_SetLatencyMarkerNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_SetLatencyMarkerNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_SetLatencyMarkerNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetLatencyTimingsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetLatencyTimingsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetLatencyTimingsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetLatencyTimingsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetLatencyTimingsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetLatencyTimingsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetLatencyTimingsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetLatencyTimingsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetLatencyTimingsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetLatencyTimingsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetLatencyTimingsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetLatencyTimingsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetLatencyTimingsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetLatencyTimingsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetLatencyTimingsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetLatencyTimingsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetLatencyTimingsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetLatencyTimingsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetLatencyTimingsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetLatencyTimingsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetLatencyTimingsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetLatencyTimingsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetLatencyTimingsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetLatencyTimingsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetLatencyTimingsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetLatencyTimingsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetLatencyTimingsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetLatencyTimingsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetLatencyTimingsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetLatencyTimingsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetLatencyTimingsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetLatencyTimingsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetLatencyTimingsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetLatencyTimingsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetLatencyTimingsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetLatencyTimingsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetLatencyTimingsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetLatencyTimingsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetLatencyTimingsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetLatencyTimingsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetLatencyTimingsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetLatencyTimingsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetLatencyTimingsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetLatencyTimingsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetLatencyTimingsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetLatencyTimingsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetLatencyTimingsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetLatencyTimingsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetLatencyTimingsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetLatencyTimingsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetLatencyTimingsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetLatencyTimingsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetLatencyTimingsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetLatencyTimingsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetLatencyTimingsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetLatencyTimingsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetLatencyTimingsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetLatencyTimingsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetLatencyTimingsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetLatencyTimingsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetLatencyTimingsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetLatencyTimingsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetLatencyTimingsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetLatencyTimingsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetLatencyTimingsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetLatencyTimingsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetLatencyTimingsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetLatencyTimingsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetLatencyTimingsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetLatencyTimingsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetLatencyTimingsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetLatencyTimingsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetLatencyTimingsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetLatencyTimingsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetLatencyTimingsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetLatencyTimingsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetLatencyTimingsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetLatencyTimingsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetLatencyTimingsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetLatencyTimingsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetLatencyTimingsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetLatencyTimingsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetLatencyTimingsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetLatencyTimingsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetLatencyTimingsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetLatencyTimingsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetLatencyTimingsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetLatencyTimingsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetLatencyTimingsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetLatencyTimingsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetLatencyTimingsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetLatencyTimingsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetLatencyTimingsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetLatencyTimingsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetLatencyTimingsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetLatencyTimingsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetLatencyTimingsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetLatencyTimingsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetLatencyTimingsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetLatencyTimingsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetLatencyTimingsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetLatencyTimingsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetLatencyTimingsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetLatencyTimingsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetLatencyTimingsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetLatencyTimingsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetLatencyTimingsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetLatencyTimingsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetLatencyTimingsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetLatencyTimingsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetLatencyTimingsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetLatencyTimingsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetLatencyTimingsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetLatencyTimingsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetLatencyTimingsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetLatencyTimingsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetLatencyTimingsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetLatencyTimingsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetLatencyTimingsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetLatencyTimingsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetLatencyTimingsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetLatencyTimingsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetLatencyTimingsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetLatencyTimingsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetLatencyTimingsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetLatencyTimingsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetLatencyTimingsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetLatencyTimingsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetLatencyTimingsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetLatencyTimingsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetLatencyTimingsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetLatencyTimingsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetLatencyTimingsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetLatencyTimingsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetLatencyTimingsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetLatencyTimingsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetLatencyTimingsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetLatencyTimingsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetLatencyTimingsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetLatencyTimingsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetLatencyTimingsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetLatencyTimingsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetLatencyTimingsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetLatencyTimingsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetLatencyTimingsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetLatencyTimingsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetLatencyTimingsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetLatencyTimingsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetLatencyTimingsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetLatencyTimingsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetLatencyTimingsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetLatencyTimingsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetLatencyTimingsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetLatencyTimingsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetLatencyTimingsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetLatencyTimingsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetLatencyTimingsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetLatencyTimingsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetLatencyTimingsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetLatencyTimingsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetLatencyTimingsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetLatencyTimingsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetLatencyTimingsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetLatencyTimingsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetLatencyTimingsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetLatencyTimingsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetLatencyTimingsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetLatencyTimingsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetLatencyTimingsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetLatencyTimingsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetLatencyTimingsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetLatencyTimingsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetLatencyTimingsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetLatencyTimingsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetLatencyTimingsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetLatencyTimingsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetLatencyTimingsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetLatencyTimingsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetLatencyTimingsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetLatencyTimingsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetLatencyTimingsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetLatencyTimingsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetLatencyTimingsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetLatencyTimingsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetLatencyTimingsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetLatencyTimingsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetLatencyTimingsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetLatencyTimingsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetLatencyTimingsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetLatencyTimingsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetLatencyTimingsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetLatencyTimingsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetLatencyTimingsNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetLatencyTimingsNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetLatencyTimingsNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetLatencyTimingsNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetLatencyTimingsNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetLatencyTimingsNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetLatencyTimingsNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetLatencyTimingsNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetLatencyTimingsNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetLatencyTimingsNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetLatencyTimingsNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetLatencyTimingsNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetLatencyTimingsNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetLatencyTimingsNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetLatencyTimingsNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetLatencyTimingsNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_QueueNotifyOutOfBandNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueNotifyOutOfBandNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueNotifyOutOfBandNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueNotifyOutOfBandNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueNotifyOutOfBandNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueNotifyOutOfBandNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueNotifyOutOfBandNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueNotifyOutOfBandNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueNotifyOutOfBandNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueNotifyOutOfBandNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueNotifyOutOfBandNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueNotifyOutOfBandNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueNotifyOutOfBandNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueNotifyOutOfBandNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_QueueNotifyOutOfBandNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_QueueNotifyOutOfBandNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_QueueNotifyOutOfBandNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueNotifyOutOfBandNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueNotifyOutOfBandNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueNotifyOutOfBandNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueNotifyOutOfBandNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueNotifyOutOfBandNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueNotifyOutOfBandNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueNotifyOutOfBandNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueNotifyOutOfBandNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueNotifyOutOfBandNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueNotifyOutOfBandNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueNotifyOutOfBandNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueNotifyOutOfBandNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueNotifyOutOfBandNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_QueueNotifyOutOfBandNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_QueueNotifyOutOfBandNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_QueueNotifyOutOfBandNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueNotifyOutOfBandNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueNotifyOutOfBandNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueNotifyOutOfBandNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueNotifyOutOfBandNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueNotifyOutOfBandNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueNotifyOutOfBandNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueNotifyOutOfBandNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueNotifyOutOfBandNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueNotifyOutOfBandNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueNotifyOutOfBandNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueNotifyOutOfBandNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueNotifyOutOfBandNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueNotifyOutOfBandNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_QueueNotifyOutOfBandNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_QueueNotifyOutOfBandNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_QueueNotifyOutOfBandNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueNotifyOutOfBandNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueNotifyOutOfBandNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueNotifyOutOfBandNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueNotifyOutOfBandNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueNotifyOutOfBandNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueNotifyOutOfBandNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueNotifyOutOfBandNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueNotifyOutOfBandNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueNotifyOutOfBandNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueNotifyOutOfBandNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueNotifyOutOfBandNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueNotifyOutOfBandNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueNotifyOutOfBandNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_QueueNotifyOutOfBandNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_QueueNotifyOutOfBandNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_QueueNotifyOutOfBandNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueNotifyOutOfBandNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueNotifyOutOfBandNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueNotifyOutOfBandNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueNotifyOutOfBandNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueNotifyOutOfBandNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueNotifyOutOfBandNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueNotifyOutOfBandNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueNotifyOutOfBandNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueNotifyOutOfBandNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueNotifyOutOfBandNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueNotifyOutOfBandNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueNotifyOutOfBandNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueNotifyOutOfBandNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_QueueNotifyOutOfBandNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_QueueNotifyOutOfBandNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_QueueNotifyOutOfBandNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueNotifyOutOfBandNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueNotifyOutOfBandNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueNotifyOutOfBandNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueNotifyOutOfBandNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueNotifyOutOfBandNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueNotifyOutOfBandNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueNotifyOutOfBandNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueNotifyOutOfBandNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueNotifyOutOfBandNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueNotifyOutOfBandNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueNotifyOutOfBandNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueNotifyOutOfBandNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueNotifyOutOfBandNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_QueueNotifyOutOfBandNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_QueueNotifyOutOfBandNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_QueueNotifyOutOfBandNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueNotifyOutOfBandNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueNotifyOutOfBandNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueNotifyOutOfBandNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueNotifyOutOfBandNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueNotifyOutOfBandNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueNotifyOutOfBandNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueNotifyOutOfBandNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueNotifyOutOfBandNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueNotifyOutOfBandNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueNotifyOutOfBandNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueNotifyOutOfBandNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueNotifyOutOfBandNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueNotifyOutOfBandNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_QueueNotifyOutOfBandNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_QueueNotifyOutOfBandNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_QueueNotifyOutOfBandNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueNotifyOutOfBandNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueNotifyOutOfBandNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueNotifyOutOfBandNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueNotifyOutOfBandNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueNotifyOutOfBandNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueNotifyOutOfBandNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueNotifyOutOfBandNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueNotifyOutOfBandNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueNotifyOutOfBandNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueNotifyOutOfBandNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueNotifyOutOfBandNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueNotifyOutOfBandNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueNotifyOutOfBandNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_QueueNotifyOutOfBandNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_QueueNotifyOutOfBandNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueNotifyOutOfBandNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueNotifyOutOfBandNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueNotifyOutOfBandNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueNotifyOutOfBandNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueNotifyOutOfBandNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueNotifyOutOfBandNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueNotifyOutOfBandNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueNotifyOutOfBandNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueNotifyOutOfBandNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueNotifyOutOfBandNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueNotifyOutOfBandNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueNotifyOutOfBandNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueNotifyOutOfBandNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueNotifyOutOfBandNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_QueueNotifyOutOfBandNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_QueueNotifyOutOfBandNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueNotifyOutOfBandNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueNotifyOutOfBandNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueNotifyOutOfBandNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueNotifyOutOfBandNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueNotifyOutOfBandNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueNotifyOutOfBandNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueNotifyOutOfBandNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueNotifyOutOfBandNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueNotifyOutOfBandNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueNotifyOutOfBandNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueNotifyOutOfBandNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueNotifyOutOfBandNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueNotifyOutOfBandNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueNotifyOutOfBandNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_QueueNotifyOutOfBandNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_QueueNotifyOutOfBandNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueNotifyOutOfBandNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueNotifyOutOfBandNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueNotifyOutOfBandNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueNotifyOutOfBandNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueNotifyOutOfBandNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueNotifyOutOfBandNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueNotifyOutOfBandNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueNotifyOutOfBandNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueNotifyOutOfBandNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueNotifyOutOfBandNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueNotifyOutOfBandNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueNotifyOutOfBandNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueNotifyOutOfBandNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueNotifyOutOfBandNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_QueueNotifyOutOfBandNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_QueueNotifyOutOfBandNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_QueueNotifyOutOfBandNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueNotifyOutOfBandNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueNotifyOutOfBandNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueNotifyOutOfBandNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueNotifyOutOfBandNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueNotifyOutOfBandNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueNotifyOutOfBandNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueNotifyOutOfBandNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueNotifyOutOfBandNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueNotifyOutOfBandNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueNotifyOutOfBandNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueNotifyOutOfBandNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueNotifyOutOfBandNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueNotifyOutOfBandNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_QueueNotifyOutOfBandNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_QueueNotifyOutOfBandNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueNotifyOutOfBandNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueNotifyOutOfBandNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueNotifyOutOfBandNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueNotifyOutOfBandNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueNotifyOutOfBandNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueNotifyOutOfBandNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueNotifyOutOfBandNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueNotifyOutOfBandNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueNotifyOutOfBandNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueNotifyOutOfBandNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueNotifyOutOfBandNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueNotifyOutOfBandNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueNotifyOutOfBandNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueNotifyOutOfBandNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_QueueNotifyOutOfBandNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_QueueNotifyOutOfBandNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocations@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocations@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocations@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocations@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocations@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocations@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocations@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocations@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocations@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocations@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocations@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocations@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocations@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocations@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetRenderingAttachmentLocations=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetRenderingAttachmentLocations=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocations@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocations@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocations@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocations@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocations@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocations@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocations@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocations@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocations@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocations@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocations@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocations@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocations@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocations@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetRenderingAttachmentLocations=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetRenderingAttachmentLocations=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocations@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocations@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocations@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocations@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocations@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocations@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocations@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocations@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocations@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocations@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocations@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocations@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocations@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocations@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetRenderingAttachmentLocations=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetRenderingAttachmentLocations=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocations@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocations@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocations@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocations@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocations@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocations@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocations@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocations@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocations@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocations@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocations@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocations@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocations@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocations@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetRenderingAttachmentLocations=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetRenderingAttachmentLocations=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocations@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocations@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocations@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocations@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocations@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocations@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocations@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocations@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocations@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocations@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocations@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocations@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocations@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocations@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetRenderingAttachmentLocations=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetRenderingAttachmentLocations=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocations@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocations@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocations@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocations@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocations@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocations@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocations@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocations@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocations@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocations@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocations@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocations@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocations@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocations@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetRenderingAttachmentLocations=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetRenderingAttachmentLocations=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocations@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocations@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocations@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocations@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocations@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocations@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocations@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocations@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocations@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocations@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocations@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocations@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocations@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocations@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetRenderingAttachmentLocations=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetRenderingAttachmentLocations=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocations@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocations@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocations@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocations@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocations@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocations@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocations@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocations@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocations@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocations@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocations@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocations@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocations@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocations@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetRenderingAttachmentLocations=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetRenderingAttachmentLocations=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocations@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocations@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocations@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocations@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocations@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocations@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocations@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocations@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocations@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocations@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocations@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocations@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocations@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocations@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetRenderingAttachmentLocations=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetRenderingAttachmentLocations=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocations@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocations@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocations@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocations@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocations@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocations@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocations@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocations@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocations@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocations@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocations@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocations@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocations@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocations@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetRenderingAttachmentLocations=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetRenderingAttachmentLocations=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocations@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocations@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocations@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocations@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocations@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocations@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocations@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocations@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocations@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocations@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocations@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocations@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocations@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocations@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetRenderingAttachmentLocations=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetRenderingAttachmentLocations=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocations@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocations@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocations@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocations@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocations@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocations@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocations@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocations@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocations@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocations@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocations@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocations@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocations@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocations@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetRenderingAttachmentLocations=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetRenderingAttachmentLocations=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocations@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocations@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocations@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocations@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocations@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocations@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocations@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocations@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocations@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocations@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocations@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocations@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocations@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocations@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetRenderingAttachmentLocations=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetRenderingAttachmentLocations=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingAttachmentLocationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetRenderingAttachmentLocationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetRenderingAttachmentLocationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingAttachmentLocationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetRenderingAttachmentLocationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetRenderingAttachmentLocationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingAttachmentLocationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetRenderingAttachmentLocationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetRenderingAttachmentLocationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingAttachmentLocationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetRenderingAttachmentLocationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetRenderingAttachmentLocationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingAttachmentLocationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetRenderingAttachmentLocationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetRenderingAttachmentLocationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingAttachmentLocationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetRenderingAttachmentLocationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetRenderingAttachmentLocationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingAttachmentLocationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetRenderingAttachmentLocationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetRenderingAttachmentLocationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingAttachmentLocationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetRenderingAttachmentLocationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetRenderingAttachmentLocationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingAttachmentLocationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetRenderingAttachmentLocationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetRenderingAttachmentLocationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingAttachmentLocationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetRenderingAttachmentLocationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetRenderingAttachmentLocationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingAttachmentLocationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetRenderingAttachmentLocationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetRenderingAttachmentLocationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingAttachmentLocationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetRenderingAttachmentLocationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetRenderingAttachmentLocationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocationsKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocationsKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocationsKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocationsKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocationsKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocationsKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocationsKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocationsKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocationsKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocationsKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocationsKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocationsKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocationsKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingAttachmentLocationsKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetRenderingAttachmentLocationsKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetRenderingAttachmentLocationsKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetRenderingInputAttachmentIndices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetRenderingInputAttachmentIndices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetRenderingInputAttachmentIndices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetRenderingInputAttachmentIndices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetRenderingInputAttachmentIndices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetRenderingInputAttachmentIndices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetRenderingInputAttachmentIndices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetRenderingInputAttachmentIndices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetRenderingInputAttachmentIndices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetRenderingInputAttachmentIndices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetRenderingInputAttachmentIndices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetRenderingInputAttachmentIndices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetRenderingInputAttachmentIndices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetRenderingInputAttachmentIndices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetRenderingInputAttachmentIndices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetRenderingInputAttachmentIndices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetRenderingInputAttachmentIndices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetRenderingInputAttachmentIndices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetRenderingInputAttachmentIndices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetRenderingInputAttachmentIndices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetRenderingInputAttachmentIndices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetRenderingInputAttachmentIndices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetRenderingInputAttachmentIndices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetRenderingInputAttachmentIndices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndices@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndices@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndices@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndices@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndices@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndices@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndices@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndices@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndices@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndices@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndices@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndices@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndices@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndices@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetRenderingInputAttachmentIndices=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetRenderingInputAttachmentIndices=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndicesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndicesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndicesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndicesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndicesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndicesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndicesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndicesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndicesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndicesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndicesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndicesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndicesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetRenderingInputAttachmentIndicesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetRenderingInputAttachmentIndicesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetRenderingInputAttachmentIndicesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndicesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndicesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndicesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndicesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndicesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndicesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndicesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndicesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndicesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndicesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndicesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndicesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndicesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetRenderingInputAttachmentIndicesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetRenderingInputAttachmentIndicesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetRenderingInputAttachmentIndicesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndicesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndicesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndicesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndicesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndicesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndicesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndicesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndicesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndicesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndicesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndicesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndicesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndicesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetRenderingInputAttachmentIndicesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetRenderingInputAttachmentIndicesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetRenderingInputAttachmentIndicesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndicesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndicesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndicesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndicesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndicesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndicesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndicesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndicesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndicesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndicesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndicesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndicesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndicesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetRenderingInputAttachmentIndicesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetRenderingInputAttachmentIndicesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetRenderingInputAttachmentIndicesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndicesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndicesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndicesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndicesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndicesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndicesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndicesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndicesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndicesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndicesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndicesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndicesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndicesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetRenderingInputAttachmentIndicesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetRenderingInputAttachmentIndicesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetRenderingInputAttachmentIndicesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndicesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndicesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndicesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndicesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndicesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndicesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndicesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndicesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndicesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndicesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndicesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndicesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndicesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetRenderingInputAttachmentIndicesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetRenderingInputAttachmentIndicesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetRenderingInputAttachmentIndicesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndicesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndicesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndicesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndicesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndicesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndicesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndicesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndicesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndicesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndicesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndicesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndicesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndicesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetRenderingInputAttachmentIndicesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetRenderingInputAttachmentIndicesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetRenderingInputAttachmentIndicesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndicesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndicesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndicesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndicesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndicesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndicesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndicesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndicesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndicesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndicesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndicesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndicesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndicesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetRenderingInputAttachmentIndicesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetRenderingInputAttachmentIndicesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetRenderingInputAttachmentIndicesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndicesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndicesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndicesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndicesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndicesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndicesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndicesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndicesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndicesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndicesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndicesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndicesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndicesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetRenderingInputAttachmentIndicesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetRenderingInputAttachmentIndicesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetRenderingInputAttachmentIndicesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndicesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndicesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndicesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndicesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndicesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndicesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndicesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndicesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndicesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndicesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndicesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndicesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndicesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetRenderingInputAttachmentIndicesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetRenderingInputAttachmentIndicesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetRenderingInputAttachmentIndicesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndicesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndicesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndicesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndicesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndicesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndicesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndicesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndicesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndicesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndicesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndicesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndicesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndicesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetRenderingInputAttachmentIndicesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetRenderingInputAttachmentIndicesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetRenderingInputAttachmentIndicesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndicesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndicesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndicesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndicesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndicesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndicesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndicesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndicesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndicesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndicesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndicesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndicesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndicesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetRenderingInputAttachmentIndicesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetRenderingInputAttachmentIndicesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetRenderingInputAttachmentIndicesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndicesKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndicesKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndicesKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndicesKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndicesKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndicesKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndicesKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndicesKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndicesKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndicesKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndicesKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndicesKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndicesKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetRenderingInputAttachmentIndicesKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetRenderingInputAttachmentIndicesKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetRenderingInputAttachmentIndicesKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampRangeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampRangeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampRangeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampRangeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampRangeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampRangeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampRangeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampRangeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampRangeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampRangeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampRangeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampRangeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampRangeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDepthClampRangeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDepthClampRangeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDepthClampRangeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampRangeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampRangeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampRangeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampRangeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampRangeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampRangeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampRangeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampRangeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampRangeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampRangeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampRangeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampRangeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampRangeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDepthClampRangeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDepthClampRangeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDepthClampRangeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampRangeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampRangeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampRangeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampRangeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampRangeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampRangeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampRangeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampRangeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampRangeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampRangeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampRangeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampRangeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampRangeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDepthClampRangeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDepthClampRangeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDepthClampRangeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampRangeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampRangeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampRangeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampRangeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampRangeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampRangeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampRangeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampRangeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampRangeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampRangeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampRangeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampRangeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampRangeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDepthClampRangeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDepthClampRangeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDepthClampRangeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampRangeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampRangeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampRangeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampRangeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampRangeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampRangeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampRangeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampRangeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampRangeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampRangeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampRangeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampRangeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampRangeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDepthClampRangeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDepthClampRangeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDepthClampRangeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampRangeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampRangeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampRangeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampRangeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampRangeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampRangeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampRangeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampRangeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampRangeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampRangeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampRangeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampRangeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampRangeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDepthClampRangeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDepthClampRangeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDepthClampRangeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampRangeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampRangeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampRangeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampRangeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampRangeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampRangeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampRangeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampRangeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampRangeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampRangeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampRangeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampRangeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampRangeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDepthClampRangeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDepthClampRangeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDepthClampRangeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampRangeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampRangeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampRangeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampRangeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampRangeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampRangeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampRangeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampRangeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampRangeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampRangeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampRangeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampRangeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampRangeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDepthClampRangeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDepthClampRangeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDepthClampRangeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampRangeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampRangeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampRangeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampRangeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampRangeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampRangeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampRangeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampRangeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampRangeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampRangeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampRangeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampRangeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampRangeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDepthClampRangeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDepthClampRangeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDepthClampRangeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampRangeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampRangeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampRangeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampRangeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampRangeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampRangeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampRangeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampRangeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampRangeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampRangeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampRangeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampRangeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampRangeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDepthClampRangeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDepthClampRangeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDepthClampRangeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampRangeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampRangeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampRangeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampRangeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampRangeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampRangeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampRangeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampRangeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampRangeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampRangeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampRangeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampRangeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampRangeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDepthClampRangeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDepthClampRangeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDepthClampRangeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampRangeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampRangeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampRangeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampRangeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampRangeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampRangeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampRangeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampRangeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampRangeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampRangeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampRangeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampRangeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampRangeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDepthClampRangeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDepthClampRangeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDepthClampRangeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampRangeEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampRangeEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampRangeEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampRangeEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampRangeEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampRangeEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampRangeEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampRangeEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampRangeEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampRangeEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampRangeEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampRangeEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampRangeEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDepthClampRangeEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDepthClampRangeEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDepthClampRangeEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetMemoryMetalHandleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetMemoryMetalHandleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetMemoryMetalHandleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetMemoryMetalHandleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetMemoryMetalHandleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetMemoryMetalHandleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetMemoryMetalHandleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetMemoryMetalHandleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetMemoryMetalHandleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetMemoryMetalHandleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetMemoryMetalHandleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetMemoryMetalHandleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetMemoryMetalHandleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetMemoryMetalHandleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetMemoryMetalHandleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetMemoryMetalHandleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetMemoryMetalHandleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetMemoryMetalHandleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetMemoryMetalHandleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetMemoryMetalHandleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetMemoryMetalHandleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetMemoryMetalHandleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetMemoryMetalHandleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetMemoryMetalHandleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandleEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandleEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandleEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandleEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandleEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandleEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandleEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandleEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandleEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandleEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandleEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandleEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandleEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandleEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetMemoryMetalHandleEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetMemoryMetalHandleEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandlePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandlePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandlePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandlePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandlePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandlePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandlePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandlePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandlePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandlePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandlePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandlePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandlePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryMetalHandlePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetMemoryMetalHandlePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetMemoryMetalHandlePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandlePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandlePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandlePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandlePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandlePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandlePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandlePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandlePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandlePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandlePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandlePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandlePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandlePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryMetalHandlePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetMemoryMetalHandlePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetMemoryMetalHandlePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandlePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandlePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandlePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandlePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandlePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandlePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandlePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandlePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandlePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandlePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandlePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandlePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandlePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryMetalHandlePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetMemoryMetalHandlePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetMemoryMetalHandlePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandlePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandlePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandlePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandlePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandlePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandlePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandlePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandlePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandlePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandlePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandlePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandlePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandlePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryMetalHandlePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetMemoryMetalHandlePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetMemoryMetalHandlePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandlePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandlePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandlePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandlePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandlePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandlePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandlePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandlePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandlePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandlePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandlePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandlePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandlePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryMetalHandlePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetMemoryMetalHandlePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetMemoryMetalHandlePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandlePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandlePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandlePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandlePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandlePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandlePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandlePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandlePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandlePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandlePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandlePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandlePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandlePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryMetalHandlePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetMemoryMetalHandlePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetMemoryMetalHandlePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandlePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandlePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandlePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandlePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandlePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandlePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandlePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandlePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandlePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandlePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandlePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandlePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandlePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryMetalHandlePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetMemoryMetalHandlePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetMemoryMetalHandlePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandlePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandlePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandlePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandlePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandlePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandlePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandlePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandlePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandlePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandlePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandlePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandlePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandlePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryMetalHandlePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetMemoryMetalHandlePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetMemoryMetalHandlePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandlePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandlePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandlePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandlePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandlePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandlePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandlePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandlePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandlePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandlePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandlePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandlePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandlePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryMetalHandlePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetMemoryMetalHandlePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetMemoryMetalHandlePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandlePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandlePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandlePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandlePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandlePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandlePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandlePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandlePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandlePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandlePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandlePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandlePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandlePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryMetalHandlePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetMemoryMetalHandlePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetMemoryMetalHandlePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandlePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandlePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandlePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandlePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandlePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandlePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandlePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandlePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandlePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandlePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandlePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandlePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandlePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryMetalHandlePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetMemoryMetalHandlePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetMemoryMetalHandlePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandlePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandlePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandlePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandlePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandlePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandlePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandlePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandlePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandlePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandlePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandlePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandlePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandlePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryMetalHandlePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetMemoryMetalHandlePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetMemoryMetalHandlePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandlePropertiesEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandlePropertiesEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandlePropertiesEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandlePropertiesEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandlePropertiesEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandlePropertiesEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandlePropertiesEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandlePropertiesEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandlePropertiesEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandlePropertiesEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandlePropertiesEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandlePropertiesEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandlePropertiesEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryMetalHandlePropertiesEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetMemoryMetalHandlePropertiesEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetMemoryMetalHandlePropertiesEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdConvertCooperativeVectorMatrixNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdConvertCooperativeVectorMatrixNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdConvertCooperativeVectorMatrixNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdConvertCooperativeVectorMatrixNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdConvertCooperativeVectorMatrixNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdConvertCooperativeVectorMatrixNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdConvertCooperativeVectorMatrixNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdConvertCooperativeVectorMatrixNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdConvertCooperativeVectorMatrixNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdConvertCooperativeVectorMatrixNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdConvertCooperativeVectorMatrixNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdConvertCooperativeVectorMatrixNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdConvertCooperativeVectorMatrixNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdConvertCooperativeVectorMatrixNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdConvertCooperativeVectorMatrixNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdConvertCooperativeVectorMatrixNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchTileQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchTileQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchTileQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchTileQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchTileQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchTileQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchTileQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchTileQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchTileQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchTileQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchTileQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchTileQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchTileQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchTileQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDispatchTileQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDispatchTileQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchTileQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchTileQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchTileQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchTileQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchTileQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchTileQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchTileQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchTileQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchTileQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchTileQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchTileQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchTileQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchTileQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchTileQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDispatchTileQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDispatchTileQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchTileQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchTileQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchTileQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchTileQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchTileQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchTileQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchTileQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchTileQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchTileQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchTileQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchTileQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchTileQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchTileQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchTileQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDispatchTileQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDispatchTileQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchTileQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchTileQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchTileQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchTileQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchTileQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchTileQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchTileQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchTileQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchTileQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchTileQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchTileQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchTileQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchTileQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchTileQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDispatchTileQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDispatchTileQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchTileQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchTileQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchTileQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchTileQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchTileQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchTileQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchTileQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchTileQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchTileQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchTileQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchTileQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchTileQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchTileQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchTileQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDispatchTileQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDispatchTileQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchTileQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchTileQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchTileQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchTileQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchTileQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchTileQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchTileQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchTileQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchTileQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchTileQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchTileQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchTileQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchTileQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchTileQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDispatchTileQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDispatchTileQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchTileQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchTileQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchTileQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchTileQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchTileQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchTileQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchTileQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchTileQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchTileQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchTileQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchTileQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchTileQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchTileQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchTileQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDispatchTileQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDispatchTileQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchTileQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchTileQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchTileQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchTileQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchTileQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchTileQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchTileQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchTileQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchTileQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchTileQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchTileQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchTileQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchTileQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchTileQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDispatchTileQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDispatchTileQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchTileQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchTileQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchTileQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchTileQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchTileQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchTileQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchTileQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchTileQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchTileQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchTileQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchTileQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchTileQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchTileQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchTileQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDispatchTileQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDispatchTileQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchTileQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchTileQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchTileQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchTileQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchTileQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchTileQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchTileQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchTileQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchTileQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchTileQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchTileQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchTileQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchTileQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchTileQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDispatchTileQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDispatchTileQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchTileQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchTileQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchTileQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchTileQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchTileQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchTileQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchTileQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchTileQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchTileQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchTileQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchTileQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchTileQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchTileQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchTileQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDispatchTileQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDispatchTileQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchTileQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchTileQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchTileQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchTileQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchTileQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchTileQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchTileQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchTileQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchTileQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchTileQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchTileQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchTileQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchTileQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchTileQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDispatchTileQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDispatchTileQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchTileQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchTileQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchTileQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchTileQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchTileQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchTileQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchTileQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchTileQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchTileQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchTileQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchTileQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchTileQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchTileQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchTileQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDispatchTileQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDispatchTileQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBeginPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBeginPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBeginPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBeginPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBeginPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBeginPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBeginPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBeginPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBeginPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBeginPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBeginPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBeginPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBeginPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBeginPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBeginPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBeginPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBeginPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBeginPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBeginPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBeginPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBeginPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBeginPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBeginPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBeginPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBeginPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBeginPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBeginPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdEndPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdEndPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdEndPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdEndPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdEndPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdEndPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdEndPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdEndPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdEndPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdEndPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdEndPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdEndPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdEndPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdEndPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdEndPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdEndPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdEndPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdEndPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdEndPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdEndPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdEndPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdEndPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdEndPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdEndPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdEndPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndPerTileExecutionQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndPerTileExecutionQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndPerTileExecutionQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndPerTileExecutionQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndPerTileExecutionQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndPerTileExecutionQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndPerTileExecutionQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndPerTileExecutionQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndPerTileExecutionQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndPerTileExecutionQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndPerTileExecutionQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndPerTileExecutionQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndPerTileExecutionQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndPerTileExecutionQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdEndPerTileExecutionQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdEndPerTileExecutionQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyExternalComputeQueueNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyExternalComputeQueueNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyExternalComputeQueueNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyExternalComputeQueueNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyExternalComputeQueueNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyExternalComputeQueueNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyExternalComputeQueueNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyExternalComputeQueueNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyExternalComputeQueueNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyExternalComputeQueueNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyExternalComputeQueueNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyExternalComputeQueueNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyExternalComputeQueueNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyExternalComputeQueueNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyExternalComputeQueueNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyExternalComputeQueueNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBeginShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBeginShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBeginShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBeginShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBeginShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBeginShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBeginShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBeginShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBeginShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBeginShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBeginShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBeginShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBeginShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBeginShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBeginShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBeginShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBeginShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBeginShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBeginShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBeginShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBeginShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBeginShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBeginShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBeginShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBeginShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBeginShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBeginShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdEndShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdEndShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdEndShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdEndShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdEndShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdEndShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdEndShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdEndShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdEndShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdEndShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdEndShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdEndShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdEndShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdEndShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdEndShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdEndShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdEndShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdEndShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdEndShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdEndShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdEndShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdEndShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdEndShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdEndShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdEndShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndShaderInstrumentationARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndShaderInstrumentationARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndShaderInstrumentationARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndShaderInstrumentationARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndShaderInstrumentationARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndShaderInstrumentationARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndShaderInstrumentationARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndShaderInstrumentationARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndShaderInstrumentationARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndShaderInstrumentationARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndShaderInstrumentationARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndShaderInstrumentationARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndShaderInstrumentationARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndShaderInstrumentationARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdEndShaderInstrumentationARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdEndShaderInstrumentationARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetShaderInstrumentationValuesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInstrumentationValuesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInstrumentationValuesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInstrumentationValuesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInstrumentationValuesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInstrumentationValuesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInstrumentationValuesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInstrumentationValuesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInstrumentationValuesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInstrumentationValuesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInstrumentationValuesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInstrumentationValuesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInstrumentationValuesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetShaderInstrumentationValuesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetShaderInstrumentationValuesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetShaderInstrumentationValuesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInstrumentationValuesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInstrumentationValuesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInstrumentationValuesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInstrumentationValuesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInstrumentationValuesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInstrumentationValuesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInstrumentationValuesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInstrumentationValuesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInstrumentationValuesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInstrumentationValuesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInstrumentationValuesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInstrumentationValuesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInstrumentationValuesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetShaderInstrumentationValuesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetShaderInstrumentationValuesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetShaderInstrumentationValuesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInstrumentationValuesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInstrumentationValuesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInstrumentationValuesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInstrumentationValuesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInstrumentationValuesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInstrumentationValuesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInstrumentationValuesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInstrumentationValuesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInstrumentationValuesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInstrumentationValuesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInstrumentationValuesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInstrumentationValuesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInstrumentationValuesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetShaderInstrumentationValuesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetShaderInstrumentationValuesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetShaderInstrumentationValuesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInstrumentationValuesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInstrumentationValuesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInstrumentationValuesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInstrumentationValuesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInstrumentationValuesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInstrumentationValuesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInstrumentationValuesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInstrumentationValuesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInstrumentationValuesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInstrumentationValuesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInstrumentationValuesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInstrumentationValuesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInstrumentationValuesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetShaderInstrumentationValuesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetShaderInstrumentationValuesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetShaderInstrumentationValuesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInstrumentationValuesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInstrumentationValuesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInstrumentationValuesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInstrumentationValuesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInstrumentationValuesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInstrumentationValuesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInstrumentationValuesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInstrumentationValuesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInstrumentationValuesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInstrumentationValuesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInstrumentationValuesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInstrumentationValuesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInstrumentationValuesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetShaderInstrumentationValuesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetShaderInstrumentationValuesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetShaderInstrumentationValuesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInstrumentationValuesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInstrumentationValuesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInstrumentationValuesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInstrumentationValuesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInstrumentationValuesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInstrumentationValuesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInstrumentationValuesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInstrumentationValuesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInstrumentationValuesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInstrumentationValuesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInstrumentationValuesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInstrumentationValuesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInstrumentationValuesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetShaderInstrumentationValuesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetShaderInstrumentationValuesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetShaderInstrumentationValuesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInstrumentationValuesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInstrumentationValuesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInstrumentationValuesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInstrumentationValuesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInstrumentationValuesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInstrumentationValuesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInstrumentationValuesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInstrumentationValuesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInstrumentationValuesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInstrumentationValuesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInstrumentationValuesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInstrumentationValuesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInstrumentationValuesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetShaderInstrumentationValuesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetShaderInstrumentationValuesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetShaderInstrumentationValuesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInstrumentationValuesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInstrumentationValuesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInstrumentationValuesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInstrumentationValuesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInstrumentationValuesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInstrumentationValuesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInstrumentationValuesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInstrumentationValuesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInstrumentationValuesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInstrumentationValuesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInstrumentationValuesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInstrumentationValuesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInstrumentationValuesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetShaderInstrumentationValuesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetShaderInstrumentationValuesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetShaderInstrumentationValuesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInstrumentationValuesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInstrumentationValuesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInstrumentationValuesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInstrumentationValuesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInstrumentationValuesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInstrumentationValuesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInstrumentationValuesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInstrumentationValuesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInstrumentationValuesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInstrumentationValuesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInstrumentationValuesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInstrumentationValuesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInstrumentationValuesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetShaderInstrumentationValuesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetShaderInstrumentationValuesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetShaderInstrumentationValuesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInstrumentationValuesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInstrumentationValuesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInstrumentationValuesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInstrumentationValuesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInstrumentationValuesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInstrumentationValuesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInstrumentationValuesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInstrumentationValuesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInstrumentationValuesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInstrumentationValuesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInstrumentationValuesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInstrumentationValuesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInstrumentationValuesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetShaderInstrumentationValuesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetShaderInstrumentationValuesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetShaderInstrumentationValuesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInstrumentationValuesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInstrumentationValuesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInstrumentationValuesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInstrumentationValuesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInstrumentationValuesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInstrumentationValuesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInstrumentationValuesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInstrumentationValuesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInstrumentationValuesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInstrumentationValuesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInstrumentationValuesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInstrumentationValuesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInstrumentationValuesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetShaderInstrumentationValuesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetShaderInstrumentationValuesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetShaderInstrumentationValuesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInstrumentationValuesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInstrumentationValuesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInstrumentationValuesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInstrumentationValuesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInstrumentationValuesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInstrumentationValuesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInstrumentationValuesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInstrumentationValuesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInstrumentationValuesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInstrumentationValuesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInstrumentationValuesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInstrumentationValuesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInstrumentationValuesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetShaderInstrumentationValuesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetShaderInstrumentationValuesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetShaderInstrumentationValuesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInstrumentationValuesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInstrumentationValuesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInstrumentationValuesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInstrumentationValuesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInstrumentationValuesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInstrumentationValuesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInstrumentationValuesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInstrumentationValuesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInstrumentationValuesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInstrumentationValuesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInstrumentationValuesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInstrumentationValuesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInstrumentationValuesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetShaderInstrumentationValuesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetShaderInstrumentationValuesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetShaderInstrumentationValuesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_ClearShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ClearShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ClearShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ClearShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ClearShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ClearShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ClearShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ClearShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ClearShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ClearShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ClearShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ClearShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ClearShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_ClearShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_ClearShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_ClearShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_ClearShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ClearShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ClearShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ClearShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ClearShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ClearShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ClearShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ClearShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ClearShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ClearShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ClearShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ClearShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ClearShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_ClearShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_ClearShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_ClearShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_ClearShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ClearShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ClearShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ClearShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ClearShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ClearShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ClearShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ClearShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ClearShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ClearShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ClearShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ClearShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ClearShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_ClearShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_ClearShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_ClearShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_ClearShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ClearShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ClearShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ClearShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ClearShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ClearShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ClearShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ClearShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ClearShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ClearShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ClearShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ClearShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ClearShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_ClearShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_ClearShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_ClearShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_ClearShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ClearShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ClearShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ClearShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ClearShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ClearShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ClearShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ClearShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ClearShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ClearShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ClearShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ClearShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ClearShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_ClearShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_ClearShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_ClearShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_ClearShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ClearShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ClearShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ClearShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ClearShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ClearShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ClearShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ClearShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ClearShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ClearShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ClearShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ClearShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ClearShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_ClearShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_ClearShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_ClearShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_ClearShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ClearShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ClearShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ClearShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ClearShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ClearShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ClearShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ClearShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ClearShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ClearShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ClearShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ClearShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ClearShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_ClearShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_ClearShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_ClearShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_ClearShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ClearShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ClearShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ClearShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ClearShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ClearShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ClearShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ClearShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ClearShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ClearShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ClearShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ClearShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ClearShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_ClearShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_ClearShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_ClearShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_ClearShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ClearShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ClearShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ClearShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ClearShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ClearShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ClearShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ClearShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ClearShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ClearShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ClearShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ClearShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ClearShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_ClearShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_ClearShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_ClearShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_ClearShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ClearShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ClearShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ClearShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ClearShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ClearShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ClearShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ClearShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ClearShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ClearShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ClearShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ClearShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ClearShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_ClearShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_ClearShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_ClearShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_ClearShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ClearShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ClearShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ClearShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ClearShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ClearShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ClearShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ClearShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ClearShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ClearShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ClearShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ClearShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ClearShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_ClearShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_ClearShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_ClearShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_ClearShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ClearShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ClearShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ClearShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ClearShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ClearShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ClearShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ClearShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ClearShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ClearShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ClearShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ClearShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ClearShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_ClearShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_ClearShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_ClearShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_ClearShaderInstrumentationMetricsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ClearShaderInstrumentationMetricsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ClearShaderInstrumentationMetricsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ClearShaderInstrumentationMetricsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ClearShaderInstrumentationMetricsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ClearShaderInstrumentationMetricsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ClearShaderInstrumentationMetricsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ClearShaderInstrumentationMetricsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ClearShaderInstrumentationMetricsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ClearShaderInstrumentationMetricsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ClearShaderInstrumentationMetricsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ClearShaderInstrumentationMetricsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ClearShaderInstrumentationMetricsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_ClearShaderInstrumentationMetricsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_ClearShaderInstrumentationMetricsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_ClearShaderInstrumentationMetricsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorViewARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorViewARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorViewARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorViewARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorViewARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorViewARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorViewARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorViewARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorViewARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorViewARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorViewARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorViewARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorViewARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyTensorViewARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyTensorViewARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyTensorViewARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_BindTensorMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindTensorMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindTensorMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindTensorMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindTensorMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindTensorMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindTensorMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindTensorMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindTensorMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindTensorMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindTensorMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindTensorMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindTensorMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindTensorMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_BindTensorMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_BindTensorMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_BindTensorMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindTensorMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindTensorMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindTensorMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindTensorMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindTensorMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindTensorMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindTensorMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindTensorMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindTensorMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindTensorMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindTensorMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindTensorMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindTensorMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_BindTensorMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_BindTensorMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_BindTensorMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindTensorMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindTensorMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindTensorMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindTensorMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindTensorMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindTensorMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindTensorMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindTensorMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindTensorMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindTensorMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindTensorMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindTensorMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindTensorMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_BindTensorMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_BindTensorMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_BindTensorMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindTensorMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindTensorMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindTensorMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindTensorMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindTensorMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindTensorMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindTensorMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindTensorMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindTensorMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindTensorMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindTensorMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindTensorMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindTensorMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_BindTensorMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_BindTensorMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_BindTensorMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindTensorMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindTensorMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindTensorMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindTensorMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindTensorMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindTensorMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindTensorMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindTensorMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindTensorMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindTensorMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindTensorMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindTensorMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindTensorMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_BindTensorMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_BindTensorMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_BindTensorMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindTensorMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindTensorMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindTensorMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindTensorMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindTensorMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindTensorMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindTensorMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindTensorMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindTensorMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindTensorMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindTensorMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindTensorMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindTensorMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_BindTensorMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_BindTensorMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_BindTensorMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindTensorMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindTensorMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindTensorMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindTensorMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindTensorMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindTensorMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindTensorMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindTensorMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindTensorMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindTensorMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindTensorMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindTensorMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindTensorMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_BindTensorMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_BindTensorMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_BindTensorMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindTensorMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindTensorMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindTensorMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindTensorMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindTensorMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindTensorMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindTensorMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindTensorMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindTensorMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindTensorMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindTensorMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindTensorMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindTensorMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_BindTensorMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_BindTensorMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_BindTensorMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindTensorMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindTensorMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindTensorMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindTensorMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindTensorMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindTensorMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindTensorMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindTensorMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindTensorMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindTensorMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindTensorMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindTensorMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindTensorMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_BindTensorMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_BindTensorMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_BindTensorMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindTensorMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindTensorMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindTensorMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindTensorMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindTensorMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindTensorMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindTensorMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindTensorMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindTensorMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindTensorMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindTensorMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindTensorMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindTensorMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_BindTensorMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_BindTensorMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindTensorMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindTensorMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindTensorMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindTensorMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindTensorMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindTensorMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindTensorMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindTensorMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindTensorMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindTensorMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindTensorMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindTensorMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindTensorMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindTensorMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_BindTensorMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_BindTensorMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_BindTensorMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindTensorMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindTensorMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindTensorMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindTensorMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindTensorMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindTensorMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindTensorMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindTensorMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindTensorMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindTensorMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindTensorMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindTensorMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindTensorMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_BindTensorMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_BindTensorMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_BindTensorMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindTensorMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindTensorMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindTensorMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindTensorMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindTensorMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindTensorMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindTensorMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindTensorMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindTensorMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindTensorMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindTensorMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindTensorMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindTensorMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_BindTensorMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_BindTensorMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDeviceTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDeviceTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDeviceTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDeviceTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDeviceTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDeviceTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDeviceTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDeviceTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDeviceTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDeviceTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDeviceTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDeviceTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDeviceTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDeviceTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDeviceTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDeviceTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDeviceTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDeviceTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDeviceTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDeviceTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDeviceTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDeviceTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDeviceTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDeviceTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDeviceTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDeviceTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDeviceTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDeviceTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDeviceTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDeviceTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDeviceTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDeviceTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDeviceTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDeviceTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDeviceTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDeviceTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDeviceTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceTensorMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceTensorMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceTensorMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceTensorMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceTensorMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceTensorMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceTensorMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceTensorMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceTensorMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceTensorMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceTensorMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceTensorMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceTensorMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDeviceTensorMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDeviceTensorMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDeviceTensorMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyTensorARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyTensorARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyTensorARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyTensorARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyTensorARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyTensorARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyTensorARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyTensorARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyTensorARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyTensorARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyTensorARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyTensorARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyTensorARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyTensorARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyTensorARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyTensorARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetTensorOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetTensorOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetTensorOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetTensorOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetTensorOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetTensorOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetTensorOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetTensorOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetTensorOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetTensorOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetTensorOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetTensorOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetTensorOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetTensorOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetTensorOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetTensorOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetTensorOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetTensorOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetTensorOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetTensorOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetTensorOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetTensorOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetTensorOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetTensorOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetTensorOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetTensorOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetTensorViewOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorViewOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorViewOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorViewOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorViewOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorViewOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorViewOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorViewOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorViewOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorViewOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorViewOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorViewOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorViewOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorViewOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetTensorViewOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetTensorViewOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorViewOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorViewOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorViewOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorViewOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorViewOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorViewOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorViewOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorViewOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorViewOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorViewOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorViewOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorViewOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorViewOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorViewOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetTensorViewOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetTensorViewOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorViewOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorViewOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorViewOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorViewOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorViewOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorViewOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorViewOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorViewOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorViewOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorViewOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorViewOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorViewOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorViewOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorViewOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetTensorViewOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetTensorViewOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorViewOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorViewOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorViewOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorViewOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorViewOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorViewOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorViewOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorViewOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorViewOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorViewOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorViewOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorViewOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorViewOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorViewOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetTensorViewOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetTensorViewOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorViewOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorViewOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorViewOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorViewOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorViewOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorViewOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorViewOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorViewOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorViewOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorViewOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorViewOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorViewOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorViewOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorViewOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetTensorViewOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetTensorViewOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorViewOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorViewOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorViewOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorViewOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorViewOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorViewOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorViewOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorViewOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorViewOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorViewOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorViewOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorViewOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorViewOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorViewOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetTensorViewOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetTensorViewOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorViewOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorViewOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorViewOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorViewOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorViewOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorViewOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorViewOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorViewOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorViewOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorViewOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorViewOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorViewOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorViewOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorViewOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetTensorViewOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetTensorViewOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorViewOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorViewOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorViewOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorViewOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorViewOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorViewOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorViewOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorViewOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorViewOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorViewOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorViewOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorViewOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorViewOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorViewOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetTensorViewOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetTensorViewOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorViewOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorViewOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorViewOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorViewOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorViewOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorViewOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorViewOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorViewOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorViewOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorViewOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorViewOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorViewOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorViewOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorViewOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetTensorViewOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetTensorViewOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorViewOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorViewOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorViewOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorViewOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorViewOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorViewOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorViewOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorViewOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorViewOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorViewOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorViewOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorViewOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorViewOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorViewOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetTensorViewOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetTensorViewOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorViewOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorViewOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorViewOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorViewOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorViewOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorViewOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorViewOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorViewOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorViewOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorViewOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorViewOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorViewOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorViewOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorViewOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetTensorViewOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetTensorViewOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorViewOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorViewOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorViewOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorViewOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorViewOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorViewOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorViewOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorViewOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorViewOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorViewOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorViewOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorViewOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorViewOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorViewOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetTensorViewOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetTensorViewOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorViewOpaqueCaptureDescriptorDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorViewOpaqueCaptureDescriptorDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorViewOpaqueCaptureDescriptorDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorViewOpaqueCaptureDescriptorDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorViewOpaqueCaptureDescriptorDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorViewOpaqueCaptureDescriptorDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorViewOpaqueCaptureDescriptorDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorViewOpaqueCaptureDescriptorDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorViewOpaqueCaptureDescriptorDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorViewOpaqueCaptureDescriptorDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorViewOpaqueCaptureDescriptorDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorViewOpaqueCaptureDescriptorDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorViewOpaqueCaptureDescriptorDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorViewOpaqueCaptureDescriptorDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetTensorViewOpaqueCaptureDescriptorDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetTensorViewOpaqueCaptureDescriptorDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelinesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelinesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelinesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelinesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelinesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelinesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelinesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelinesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelinesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelinesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelinesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelinesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelinesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelinesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateDataGraphPipelinesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateDataGraphPipelinesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelinesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelinesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelinesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelinesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelinesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelinesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelinesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelinesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelinesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelinesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelinesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelinesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelinesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelinesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateDataGraphPipelinesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateDataGraphPipelinesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelinesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelinesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelinesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelinesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelinesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelinesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelinesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelinesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelinesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelinesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelinesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelinesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelinesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelinesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateDataGraphPipelinesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateDataGraphPipelinesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelinesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelinesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelinesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelinesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelinesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelinesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelinesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelinesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelinesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelinesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelinesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelinesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelinesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelinesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateDataGraphPipelinesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateDataGraphPipelinesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelinesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelinesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelinesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelinesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelinesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelinesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelinesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelinesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelinesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelinesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelinesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelinesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelinesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelinesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateDataGraphPipelinesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateDataGraphPipelinesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelinesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelinesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelinesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelinesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelinesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelinesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelinesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelinesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelinesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelinesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelinesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelinesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelinesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelinesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateDataGraphPipelinesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateDataGraphPipelinesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelinesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelinesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelinesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelinesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelinesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelinesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelinesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelinesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelinesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelinesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelinesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelinesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelinesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelinesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateDataGraphPipelinesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateDataGraphPipelinesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelinesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelinesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelinesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelinesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelinesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelinesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelinesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelinesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelinesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelinesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelinesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelinesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelinesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelinesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateDataGraphPipelinesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateDataGraphPipelinesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelinesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelinesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelinesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelinesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelinesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelinesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelinesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelinesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelinesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelinesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelinesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelinesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelinesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelinesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateDataGraphPipelinesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateDataGraphPipelinesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelinesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelinesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelinesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelinesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelinesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelinesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelinesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelinesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelinesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelinesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelinesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelinesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelinesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelinesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateDataGraphPipelinesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateDataGraphPipelinesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelinesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelinesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelinesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelinesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelinesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelinesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelinesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelinesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelinesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelinesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelinesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelinesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelinesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelinesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateDataGraphPipelinesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateDataGraphPipelinesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelinesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelinesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelinesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelinesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelinesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelinesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelinesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelinesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelinesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelinesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelinesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelinesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelinesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelinesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateDataGraphPipelinesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateDataGraphPipelinesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelinesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelinesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelinesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelinesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelinesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelinesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelinesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelinesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelinesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelinesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelinesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelinesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelinesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelinesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateDataGraphPipelinesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateDataGraphPipelinesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionBindPointRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionBindPointRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionBindPointRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionBindPointRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionBindPointRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionBindPointRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionBindPointRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionBindPointRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionBindPointRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionBindPointRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionBindPointRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionBindPointRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionBindPointRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionBindPointRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDataGraphPipelineSessionBindPointRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDataGraphPipelineSessionBindPointRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionBindPointRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionBindPointRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionBindPointRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionBindPointRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionBindPointRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionBindPointRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionBindPointRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionBindPointRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionBindPointRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionBindPointRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionBindPointRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionBindPointRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionBindPointRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionBindPointRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDataGraphPipelineSessionBindPointRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDataGraphPipelineSessionBindPointRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionBindPointRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionBindPointRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionBindPointRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionBindPointRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionBindPointRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionBindPointRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionBindPointRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionBindPointRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionBindPointRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionBindPointRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionBindPointRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionBindPointRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionBindPointRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionBindPointRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDataGraphPipelineSessionBindPointRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDataGraphPipelineSessionBindPointRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionBindPointRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionBindPointRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionBindPointRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionBindPointRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionBindPointRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionBindPointRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionBindPointRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionBindPointRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionBindPointRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionBindPointRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionBindPointRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionBindPointRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionBindPointRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionBindPointRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDataGraphPipelineSessionBindPointRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDataGraphPipelineSessionBindPointRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionBindPointRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionBindPointRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionBindPointRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionBindPointRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionBindPointRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionBindPointRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionBindPointRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionBindPointRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionBindPointRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionBindPointRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionBindPointRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionBindPointRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionBindPointRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionBindPointRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDataGraphPipelineSessionBindPointRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDataGraphPipelineSessionBindPointRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionBindPointRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionBindPointRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionBindPointRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionBindPointRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionBindPointRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionBindPointRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionBindPointRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionBindPointRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionBindPointRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionBindPointRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionBindPointRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionBindPointRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionBindPointRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionBindPointRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDataGraphPipelineSessionBindPointRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDataGraphPipelineSessionBindPointRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionBindPointRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionBindPointRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionBindPointRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionBindPointRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionBindPointRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionBindPointRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionBindPointRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionBindPointRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionBindPointRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionBindPointRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionBindPointRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionBindPointRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionBindPointRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionBindPointRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDataGraphPipelineSessionBindPointRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDataGraphPipelineSessionBindPointRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionBindPointRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionBindPointRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionBindPointRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionBindPointRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionBindPointRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionBindPointRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionBindPointRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionBindPointRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionBindPointRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionBindPointRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionBindPointRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionBindPointRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionBindPointRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionBindPointRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDataGraphPipelineSessionBindPointRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDataGraphPipelineSessionBindPointRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionBindPointRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionBindPointRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionBindPointRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionBindPointRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionBindPointRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionBindPointRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionBindPointRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionBindPointRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionBindPointRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionBindPointRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionBindPointRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionBindPointRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionBindPointRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionBindPointRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDataGraphPipelineSessionBindPointRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDataGraphPipelineSessionBindPointRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionBindPointRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionBindPointRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionBindPointRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionBindPointRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionBindPointRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionBindPointRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionBindPointRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionBindPointRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionBindPointRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionBindPointRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionBindPointRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionBindPointRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionBindPointRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionBindPointRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDataGraphPipelineSessionBindPointRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDataGraphPipelineSessionBindPointRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionBindPointRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionBindPointRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionBindPointRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionBindPointRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionBindPointRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionBindPointRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionBindPointRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionBindPointRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionBindPointRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionBindPointRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionBindPointRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionBindPointRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionBindPointRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionBindPointRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDataGraphPipelineSessionBindPointRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDataGraphPipelineSessionBindPointRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionBindPointRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionBindPointRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionBindPointRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionBindPointRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionBindPointRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionBindPointRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionBindPointRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionBindPointRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionBindPointRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionBindPointRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionBindPointRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionBindPointRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionBindPointRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionBindPointRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDataGraphPipelineSessionBindPointRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDataGraphPipelineSessionBindPointRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionBindPointRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionBindPointRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionBindPointRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionBindPointRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionBindPointRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionBindPointRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionBindPointRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionBindPointRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionBindPointRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionBindPointRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionBindPointRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionBindPointRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionBindPointRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionBindPointRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDataGraphPipelineSessionBindPointRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDataGraphPipelineSessionBindPointRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineSessionMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDataGraphPipelineSessionMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDataGraphPipelineSessionMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineSessionMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDataGraphPipelineSessionMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDataGraphPipelineSessionMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineSessionMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDataGraphPipelineSessionMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDataGraphPipelineSessionMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineSessionMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDataGraphPipelineSessionMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDataGraphPipelineSessionMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineSessionMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDataGraphPipelineSessionMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDataGraphPipelineSessionMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineSessionMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDataGraphPipelineSessionMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDataGraphPipelineSessionMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineSessionMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDataGraphPipelineSessionMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDataGraphPipelineSessionMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineSessionMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDataGraphPipelineSessionMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDataGraphPipelineSessionMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineSessionMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDataGraphPipelineSessionMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDataGraphPipelineSessionMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineSessionMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDataGraphPipelineSessionMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDataGraphPipelineSessionMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineSessionMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDataGraphPipelineSessionMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDataGraphPipelineSessionMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineSessionMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDataGraphPipelineSessionMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDataGraphPipelineSessionMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionMemoryRequirementsARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionMemoryRequirementsARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionMemoryRequirementsARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionMemoryRequirementsARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionMemoryRequirementsARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionMemoryRequirementsARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionMemoryRequirementsARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionMemoryRequirementsARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionMemoryRequirementsARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionMemoryRequirementsARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionMemoryRequirementsARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionMemoryRequirementsARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionMemoryRequirementsARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineSessionMemoryRequirementsARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDataGraphPipelineSessionMemoryRequirementsARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDataGraphPipelineSessionMemoryRequirementsARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_BindDataGraphPipelineSessionMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindDataGraphPipelineSessionMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindDataGraphPipelineSessionMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindDataGraphPipelineSessionMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindDataGraphPipelineSessionMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindDataGraphPipelineSessionMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindDataGraphPipelineSessionMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindDataGraphPipelineSessionMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindDataGraphPipelineSessionMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindDataGraphPipelineSessionMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindDataGraphPipelineSessionMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindDataGraphPipelineSessionMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindDataGraphPipelineSessionMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_BindDataGraphPipelineSessionMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_BindDataGraphPipelineSessionMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_BindDataGraphPipelineSessionMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_BindDataGraphPipelineSessionMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindDataGraphPipelineSessionMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindDataGraphPipelineSessionMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindDataGraphPipelineSessionMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindDataGraphPipelineSessionMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindDataGraphPipelineSessionMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindDataGraphPipelineSessionMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindDataGraphPipelineSessionMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindDataGraphPipelineSessionMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindDataGraphPipelineSessionMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindDataGraphPipelineSessionMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindDataGraphPipelineSessionMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindDataGraphPipelineSessionMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_BindDataGraphPipelineSessionMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_BindDataGraphPipelineSessionMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_BindDataGraphPipelineSessionMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_BindDataGraphPipelineSessionMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindDataGraphPipelineSessionMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindDataGraphPipelineSessionMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindDataGraphPipelineSessionMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindDataGraphPipelineSessionMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindDataGraphPipelineSessionMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindDataGraphPipelineSessionMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindDataGraphPipelineSessionMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindDataGraphPipelineSessionMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindDataGraphPipelineSessionMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindDataGraphPipelineSessionMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindDataGraphPipelineSessionMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindDataGraphPipelineSessionMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_BindDataGraphPipelineSessionMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_BindDataGraphPipelineSessionMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_BindDataGraphPipelineSessionMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_BindDataGraphPipelineSessionMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindDataGraphPipelineSessionMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindDataGraphPipelineSessionMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindDataGraphPipelineSessionMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindDataGraphPipelineSessionMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindDataGraphPipelineSessionMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindDataGraphPipelineSessionMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindDataGraphPipelineSessionMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindDataGraphPipelineSessionMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindDataGraphPipelineSessionMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindDataGraphPipelineSessionMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindDataGraphPipelineSessionMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindDataGraphPipelineSessionMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_BindDataGraphPipelineSessionMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_BindDataGraphPipelineSessionMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_BindDataGraphPipelineSessionMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_BindDataGraphPipelineSessionMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindDataGraphPipelineSessionMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindDataGraphPipelineSessionMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindDataGraphPipelineSessionMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindDataGraphPipelineSessionMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindDataGraphPipelineSessionMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindDataGraphPipelineSessionMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindDataGraphPipelineSessionMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindDataGraphPipelineSessionMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindDataGraphPipelineSessionMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindDataGraphPipelineSessionMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindDataGraphPipelineSessionMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindDataGraphPipelineSessionMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_BindDataGraphPipelineSessionMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_BindDataGraphPipelineSessionMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_BindDataGraphPipelineSessionMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_BindDataGraphPipelineSessionMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindDataGraphPipelineSessionMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindDataGraphPipelineSessionMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindDataGraphPipelineSessionMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindDataGraphPipelineSessionMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindDataGraphPipelineSessionMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindDataGraphPipelineSessionMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindDataGraphPipelineSessionMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindDataGraphPipelineSessionMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindDataGraphPipelineSessionMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindDataGraphPipelineSessionMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindDataGraphPipelineSessionMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindDataGraphPipelineSessionMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_BindDataGraphPipelineSessionMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_BindDataGraphPipelineSessionMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_BindDataGraphPipelineSessionMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_BindDataGraphPipelineSessionMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindDataGraphPipelineSessionMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindDataGraphPipelineSessionMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindDataGraphPipelineSessionMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindDataGraphPipelineSessionMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindDataGraphPipelineSessionMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindDataGraphPipelineSessionMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindDataGraphPipelineSessionMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindDataGraphPipelineSessionMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindDataGraphPipelineSessionMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindDataGraphPipelineSessionMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindDataGraphPipelineSessionMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindDataGraphPipelineSessionMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_BindDataGraphPipelineSessionMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_BindDataGraphPipelineSessionMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_BindDataGraphPipelineSessionMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_BindDataGraphPipelineSessionMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindDataGraphPipelineSessionMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindDataGraphPipelineSessionMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindDataGraphPipelineSessionMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindDataGraphPipelineSessionMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindDataGraphPipelineSessionMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindDataGraphPipelineSessionMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindDataGraphPipelineSessionMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindDataGraphPipelineSessionMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindDataGraphPipelineSessionMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindDataGraphPipelineSessionMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindDataGraphPipelineSessionMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindDataGraphPipelineSessionMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_BindDataGraphPipelineSessionMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_BindDataGraphPipelineSessionMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_BindDataGraphPipelineSessionMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_BindDataGraphPipelineSessionMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindDataGraphPipelineSessionMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindDataGraphPipelineSessionMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindDataGraphPipelineSessionMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindDataGraphPipelineSessionMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindDataGraphPipelineSessionMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindDataGraphPipelineSessionMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindDataGraphPipelineSessionMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindDataGraphPipelineSessionMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindDataGraphPipelineSessionMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindDataGraphPipelineSessionMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindDataGraphPipelineSessionMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindDataGraphPipelineSessionMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_BindDataGraphPipelineSessionMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_BindDataGraphPipelineSessionMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_BindDataGraphPipelineSessionMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_BindDataGraphPipelineSessionMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindDataGraphPipelineSessionMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindDataGraphPipelineSessionMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindDataGraphPipelineSessionMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindDataGraphPipelineSessionMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindDataGraphPipelineSessionMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindDataGraphPipelineSessionMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindDataGraphPipelineSessionMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindDataGraphPipelineSessionMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindDataGraphPipelineSessionMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindDataGraphPipelineSessionMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindDataGraphPipelineSessionMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindDataGraphPipelineSessionMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_BindDataGraphPipelineSessionMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_BindDataGraphPipelineSessionMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_BindDataGraphPipelineSessionMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindDataGraphPipelineSessionMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindDataGraphPipelineSessionMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindDataGraphPipelineSessionMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindDataGraphPipelineSessionMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindDataGraphPipelineSessionMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindDataGraphPipelineSessionMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindDataGraphPipelineSessionMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindDataGraphPipelineSessionMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindDataGraphPipelineSessionMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindDataGraphPipelineSessionMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindDataGraphPipelineSessionMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindDataGraphPipelineSessionMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindDataGraphPipelineSessionMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_BindDataGraphPipelineSessionMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_BindDataGraphPipelineSessionMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_BindDataGraphPipelineSessionMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_BindDataGraphPipelineSessionMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindDataGraphPipelineSessionMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindDataGraphPipelineSessionMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindDataGraphPipelineSessionMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindDataGraphPipelineSessionMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindDataGraphPipelineSessionMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindDataGraphPipelineSessionMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindDataGraphPipelineSessionMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindDataGraphPipelineSessionMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindDataGraphPipelineSessionMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindDataGraphPipelineSessionMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindDataGraphPipelineSessionMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindDataGraphPipelineSessionMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_BindDataGraphPipelineSessionMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_BindDataGraphPipelineSessionMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_BindDataGraphPipelineSessionMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_BindDataGraphPipelineSessionMemoryARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindDataGraphPipelineSessionMemoryARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindDataGraphPipelineSessionMemoryARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindDataGraphPipelineSessionMemoryARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindDataGraphPipelineSessionMemoryARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindDataGraphPipelineSessionMemoryARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindDataGraphPipelineSessionMemoryARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindDataGraphPipelineSessionMemoryARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindDataGraphPipelineSessionMemoryARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindDataGraphPipelineSessionMemoryARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindDataGraphPipelineSessionMemoryARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindDataGraphPipelineSessionMemoryARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindDataGraphPipelineSessionMemoryARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_BindDataGraphPipelineSessionMemoryARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_BindDataGraphPipelineSessionMemoryARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_BindDataGraphPipelineSessionMemoryARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_DestroyDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_DestroyDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_DestroyDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_DestroyDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_DestroyDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_DestroyDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_DestroyDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_DestroyDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_DestroyDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_DestroyDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_DestroyDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_DestroyDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_DestroyDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_DestroyDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_DestroyDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_DestroyDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_DestroyDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_DestroyDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_DestroyDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_DestroyDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_DestroyDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_DestroyDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_DestroyDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_DestroyDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_DestroyDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_DestroyDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_DestroyDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_DestroyDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_DestroyDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_DestroyDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_DestroyDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_DestroyDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_DestroyDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_DestroyDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_DestroyDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_DestroyDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_DestroyDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDataGraphPipelineSessionARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDataGraphPipelineSessionARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDataGraphPipelineSessionARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDataGraphPipelineSessionARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDataGraphPipelineSessionARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDataGraphPipelineSessionARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDataGraphPipelineSessionARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDataGraphPipelineSessionARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDataGraphPipelineSessionARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDataGraphPipelineSessionARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDataGraphPipelineSessionARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDataGraphPipelineSessionARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDataGraphPipelineSessionARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_DestroyDataGraphPipelineSessionARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_DestroyDataGraphPipelineSessionARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_DestroyDataGraphPipelineSessionARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchDataGraphARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchDataGraphARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchDataGraphARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchDataGraphARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchDataGraphARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchDataGraphARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchDataGraphARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchDataGraphARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchDataGraphARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchDataGraphARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchDataGraphARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchDataGraphARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchDataGraphARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchDataGraphARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDispatchDataGraphARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDispatchDataGraphARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchDataGraphARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchDataGraphARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchDataGraphARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchDataGraphARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchDataGraphARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchDataGraphARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchDataGraphARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchDataGraphARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchDataGraphARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchDataGraphARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchDataGraphARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchDataGraphARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchDataGraphARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchDataGraphARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDispatchDataGraphARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDispatchDataGraphARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchDataGraphARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchDataGraphARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchDataGraphARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchDataGraphARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchDataGraphARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchDataGraphARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchDataGraphARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchDataGraphARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchDataGraphARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchDataGraphARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchDataGraphARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchDataGraphARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchDataGraphARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchDataGraphARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDispatchDataGraphARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDispatchDataGraphARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchDataGraphARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchDataGraphARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchDataGraphARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchDataGraphARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchDataGraphARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchDataGraphARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchDataGraphARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchDataGraphARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchDataGraphARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchDataGraphARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchDataGraphARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchDataGraphARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchDataGraphARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchDataGraphARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDispatchDataGraphARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDispatchDataGraphARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchDataGraphARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchDataGraphARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchDataGraphARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchDataGraphARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchDataGraphARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchDataGraphARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchDataGraphARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchDataGraphARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchDataGraphARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchDataGraphARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchDataGraphARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchDataGraphARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchDataGraphARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchDataGraphARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDispatchDataGraphARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDispatchDataGraphARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchDataGraphARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchDataGraphARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchDataGraphARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchDataGraphARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchDataGraphARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchDataGraphARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchDataGraphARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchDataGraphARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchDataGraphARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchDataGraphARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchDataGraphARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchDataGraphARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchDataGraphARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchDataGraphARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDispatchDataGraphARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDispatchDataGraphARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchDataGraphARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchDataGraphARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchDataGraphARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchDataGraphARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchDataGraphARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchDataGraphARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchDataGraphARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchDataGraphARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchDataGraphARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchDataGraphARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchDataGraphARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchDataGraphARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchDataGraphARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchDataGraphARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDispatchDataGraphARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDispatchDataGraphARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchDataGraphARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchDataGraphARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchDataGraphARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchDataGraphARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchDataGraphARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchDataGraphARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchDataGraphARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchDataGraphARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchDataGraphARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchDataGraphARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchDataGraphARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchDataGraphARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchDataGraphARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchDataGraphARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDispatchDataGraphARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDispatchDataGraphARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchDataGraphARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchDataGraphARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchDataGraphARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchDataGraphARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchDataGraphARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchDataGraphARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchDataGraphARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchDataGraphARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchDataGraphARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchDataGraphARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchDataGraphARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchDataGraphARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchDataGraphARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchDataGraphARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDispatchDataGraphARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDispatchDataGraphARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchDataGraphARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchDataGraphARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchDataGraphARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchDataGraphARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchDataGraphARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchDataGraphARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchDataGraphARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchDataGraphARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchDataGraphARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchDataGraphARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchDataGraphARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchDataGraphARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchDataGraphARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchDataGraphARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDispatchDataGraphARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDispatchDataGraphARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchDataGraphARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchDataGraphARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchDataGraphARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchDataGraphARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchDataGraphARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchDataGraphARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchDataGraphARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchDataGraphARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchDataGraphARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchDataGraphARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchDataGraphARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchDataGraphARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchDataGraphARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchDataGraphARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDispatchDataGraphARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDispatchDataGraphARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchDataGraphARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchDataGraphARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchDataGraphARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchDataGraphARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchDataGraphARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchDataGraphARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchDataGraphARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchDataGraphARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchDataGraphARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchDataGraphARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchDataGraphARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchDataGraphARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchDataGraphARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchDataGraphARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDispatchDataGraphARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDispatchDataGraphARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchDataGraphARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchDataGraphARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchDataGraphARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchDataGraphARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchDataGraphARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchDataGraphARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchDataGraphARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchDataGraphARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchDataGraphARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchDataGraphARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchDataGraphARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchDataGraphARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchDataGraphARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchDataGraphARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDispatchDataGraphARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDispatchDataGraphARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineAvailablePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineAvailablePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineAvailablePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineAvailablePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineAvailablePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineAvailablePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineAvailablePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineAvailablePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineAvailablePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineAvailablePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineAvailablePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineAvailablePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineAvailablePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelineAvailablePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDataGraphPipelineAvailablePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDataGraphPipelineAvailablePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineAvailablePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineAvailablePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineAvailablePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineAvailablePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineAvailablePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineAvailablePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineAvailablePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineAvailablePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineAvailablePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineAvailablePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineAvailablePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineAvailablePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineAvailablePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelineAvailablePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDataGraphPipelineAvailablePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDataGraphPipelineAvailablePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineAvailablePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineAvailablePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineAvailablePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineAvailablePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineAvailablePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineAvailablePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineAvailablePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineAvailablePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineAvailablePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineAvailablePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineAvailablePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineAvailablePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineAvailablePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelineAvailablePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDataGraphPipelineAvailablePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDataGraphPipelineAvailablePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineAvailablePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineAvailablePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineAvailablePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineAvailablePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineAvailablePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineAvailablePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineAvailablePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineAvailablePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineAvailablePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineAvailablePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineAvailablePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineAvailablePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineAvailablePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelineAvailablePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDataGraphPipelineAvailablePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDataGraphPipelineAvailablePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineAvailablePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineAvailablePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineAvailablePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineAvailablePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineAvailablePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineAvailablePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineAvailablePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineAvailablePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineAvailablePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineAvailablePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineAvailablePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineAvailablePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineAvailablePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelineAvailablePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDataGraphPipelineAvailablePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDataGraphPipelineAvailablePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineAvailablePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineAvailablePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineAvailablePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineAvailablePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineAvailablePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineAvailablePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineAvailablePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineAvailablePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineAvailablePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineAvailablePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineAvailablePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineAvailablePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineAvailablePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelineAvailablePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDataGraphPipelineAvailablePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDataGraphPipelineAvailablePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineAvailablePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineAvailablePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineAvailablePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineAvailablePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineAvailablePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineAvailablePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineAvailablePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineAvailablePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineAvailablePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineAvailablePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineAvailablePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineAvailablePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineAvailablePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelineAvailablePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDataGraphPipelineAvailablePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDataGraphPipelineAvailablePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineAvailablePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineAvailablePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineAvailablePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineAvailablePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineAvailablePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineAvailablePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineAvailablePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineAvailablePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineAvailablePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineAvailablePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineAvailablePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineAvailablePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineAvailablePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelineAvailablePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDataGraphPipelineAvailablePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDataGraphPipelineAvailablePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineAvailablePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineAvailablePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineAvailablePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineAvailablePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineAvailablePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineAvailablePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineAvailablePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineAvailablePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineAvailablePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineAvailablePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineAvailablePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineAvailablePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineAvailablePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelineAvailablePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDataGraphPipelineAvailablePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDataGraphPipelineAvailablePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineAvailablePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineAvailablePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineAvailablePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineAvailablePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineAvailablePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineAvailablePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineAvailablePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineAvailablePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineAvailablePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineAvailablePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineAvailablePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineAvailablePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineAvailablePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelineAvailablePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDataGraphPipelineAvailablePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDataGraphPipelineAvailablePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineAvailablePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineAvailablePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineAvailablePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineAvailablePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineAvailablePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineAvailablePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineAvailablePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineAvailablePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineAvailablePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineAvailablePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineAvailablePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineAvailablePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineAvailablePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelineAvailablePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDataGraphPipelineAvailablePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDataGraphPipelineAvailablePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineAvailablePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineAvailablePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineAvailablePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineAvailablePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineAvailablePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineAvailablePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineAvailablePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineAvailablePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineAvailablePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineAvailablePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineAvailablePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineAvailablePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineAvailablePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelineAvailablePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDataGraphPipelineAvailablePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDataGraphPipelineAvailablePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineAvailablePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineAvailablePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineAvailablePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineAvailablePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineAvailablePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineAvailablePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineAvailablePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineAvailablePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineAvailablePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineAvailablePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineAvailablePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineAvailablePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineAvailablePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelineAvailablePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDataGraphPipelineAvailablePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDataGraphPipelineAvailablePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelinePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelinePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelinePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelinePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelinePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelinePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelinePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelinePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelinePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelinePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelinePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelinePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelinePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetDataGraphPipelinePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetDataGraphPipelinePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetDataGraphPipelinePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelinePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelinePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelinePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelinePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelinePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelinePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelinePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelinePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelinePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelinePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelinePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelinePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelinePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetDataGraphPipelinePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetDataGraphPipelinePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetDataGraphPipelinePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelinePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelinePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelinePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelinePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelinePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelinePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelinePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelinePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelinePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelinePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelinePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelinePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelinePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetDataGraphPipelinePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetDataGraphPipelinePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetDataGraphPipelinePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelinePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelinePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelinePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelinePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelinePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelinePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelinePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelinePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelinePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelinePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelinePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelinePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelinePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetDataGraphPipelinePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetDataGraphPipelinePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetDataGraphPipelinePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelinePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelinePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelinePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelinePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelinePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelinePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelinePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelinePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelinePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelinePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelinePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelinePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelinePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetDataGraphPipelinePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetDataGraphPipelinePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetDataGraphPipelinePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelinePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelinePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelinePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelinePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelinePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelinePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelinePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelinePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelinePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelinePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelinePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelinePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelinePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetDataGraphPipelinePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetDataGraphPipelinePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetDataGraphPipelinePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelinePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelinePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelinePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelinePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelinePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelinePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelinePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelinePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelinePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelinePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelinePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelinePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelinePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetDataGraphPipelinePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetDataGraphPipelinePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetDataGraphPipelinePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelinePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelinePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelinePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelinePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelinePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelinePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelinePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelinePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelinePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelinePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelinePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelinePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelinePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetDataGraphPipelinePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetDataGraphPipelinePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetDataGraphPipelinePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelinePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelinePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelinePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelinePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelinePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelinePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelinePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelinePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelinePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelinePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelinePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelinePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelinePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetDataGraphPipelinePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetDataGraphPipelinePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetDataGraphPipelinePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelinePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelinePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelinePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelinePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelinePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelinePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelinePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelinePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelinePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelinePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelinePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelinePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelinePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetDataGraphPipelinePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetDataGraphPipelinePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetDataGraphPipelinePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelinePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelinePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelinePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelinePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelinePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelinePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelinePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelinePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelinePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelinePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelinePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelinePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelinePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetDataGraphPipelinePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetDataGraphPipelinePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetDataGraphPipelinePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelinePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelinePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelinePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelinePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelinePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelinePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelinePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelinePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelinePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelinePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelinePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelinePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelinePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetDataGraphPipelinePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetDataGraphPipelinePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetDataGraphPipelinePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelinePropertiesARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelinePropertiesARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelinePropertiesARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelinePropertiesARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelinePropertiesARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelinePropertiesARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelinePropertiesARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelinePropertiesARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelinePropertiesARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelinePropertiesARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelinePropertiesARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelinePropertiesARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelinePropertiesARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetDataGraphPipelinePropertiesARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetDataGraphPipelinePropertiesARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetDataGraphPipelinePropertiesARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef VK_USE_PLATFORM_OHOS
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetNativeBufferPropertiesOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetNativeBufferPropertiesOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetNativeBufferPropertiesOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetNativeBufferPropertiesOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetNativeBufferPropertiesOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetNativeBufferPropertiesOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetNativeBufferPropertiesOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetNativeBufferPropertiesOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetNativeBufferPropertiesOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetNativeBufferPropertiesOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetNativeBufferPropertiesOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetNativeBufferPropertiesOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetNativeBufferPropertiesOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetNativeBufferPropertiesOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetNativeBufferPropertiesOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetNativeBufferPropertiesOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetNativeBufferPropertiesOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetNativeBufferPropertiesOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetNativeBufferPropertiesOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetNativeBufferPropertiesOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetNativeBufferPropertiesOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetNativeBufferPropertiesOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetNativeBufferPropertiesOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetNativeBufferPropertiesOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetNativeBufferPropertiesOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetNativeBufferPropertiesOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetNativeBufferPropertiesOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetNativeBufferPropertiesOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetNativeBufferPropertiesOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetNativeBufferPropertiesOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetNativeBufferPropertiesOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetNativeBufferPropertiesOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetNativeBufferPropertiesOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetNativeBufferPropertiesOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetNativeBufferPropertiesOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetNativeBufferPropertiesOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetNativeBufferPropertiesOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetNativeBufferPropertiesOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetNativeBufferPropertiesOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetNativeBufferPropertiesOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetNativeBufferPropertiesOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetNativeBufferPropertiesOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetNativeBufferPropertiesOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetNativeBufferPropertiesOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetNativeBufferPropertiesOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetNativeBufferPropertiesOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetNativeBufferPropertiesOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetNativeBufferPropertiesOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetNativeBufferPropertiesOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetNativeBufferPropertiesOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetNativeBufferPropertiesOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetNativeBufferPropertiesOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetNativeBufferPropertiesOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetNativeBufferPropertiesOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetNativeBufferPropertiesOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetNativeBufferPropertiesOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetNativeBufferPropertiesOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetNativeBufferPropertiesOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetNativeBufferPropertiesOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetNativeBufferPropertiesOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetNativeBufferPropertiesOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetNativeBufferPropertiesOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetNativeBufferPropertiesOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetNativeBufferPropertiesOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetNativeBufferPropertiesOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetNativeBufferPropertiesOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetNativeBufferPropertiesOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetNativeBufferPropertiesOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetNativeBufferPropertiesOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetNativeBufferPropertiesOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetNativeBufferPropertiesOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetNativeBufferPropertiesOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetNativeBufferPropertiesOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetNativeBufferPropertiesOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetNativeBufferPropertiesOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetNativeBufferPropertiesOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetNativeBufferPropertiesOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetNativeBufferPropertiesOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetNativeBufferPropertiesOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetNativeBufferPropertiesOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetNativeBufferPropertiesOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetNativeBufferPropertiesOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetNativeBufferPropertiesOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetNativeBufferPropertiesOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetNativeBufferPropertiesOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetNativeBufferPropertiesOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetNativeBufferPropertiesOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetNativeBufferPropertiesOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetNativeBufferPropertiesOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetNativeBufferPropertiesOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetNativeBufferPropertiesOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetNativeBufferPropertiesOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetNativeBufferPropertiesOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetNativeBufferPropertiesOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetNativeBufferPropertiesOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetNativeBufferPropertiesOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetNativeBufferPropertiesOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetNativeBufferPropertiesOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetNativeBufferPropertiesOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetNativeBufferPropertiesOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetNativeBufferPropertiesOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetNativeBufferPropertiesOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetNativeBufferPropertiesOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetNativeBufferPropertiesOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetNativeBufferPropertiesOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetNativeBufferPropertiesOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetNativeBufferPropertiesOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetNativeBufferPropertiesOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetNativeBufferPropertiesOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetNativeBufferPropertiesOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetNativeBufferPropertiesOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetNativeBufferPropertiesOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetNativeBufferPropertiesOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetNativeBufferPropertiesOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetNativeBufferPropertiesOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetNativeBufferPropertiesOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetNativeBufferPropertiesOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetNativeBufferPropertiesOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetNativeBufferPropertiesOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetNativeBufferPropertiesOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetNativeBufferPropertiesOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetNativeBufferPropertiesOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetNativeBufferPropertiesOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetNativeBufferPropertiesOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetNativeBufferPropertiesOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetNativeBufferPropertiesOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetNativeBufferPropertiesOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetNativeBufferPropertiesOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetNativeBufferPropertiesOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetNativeBufferPropertiesOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetNativeBufferPropertiesOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetNativeBufferPropertiesOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetNativeBufferPropertiesOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetNativeBufferPropertiesOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetNativeBufferPropertiesOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetNativeBufferPropertiesOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetNativeBufferPropertiesOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetNativeBufferPropertiesOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetNativeBufferPropertiesOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetNativeBufferPropertiesOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetNativeBufferPropertiesOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetNativeBufferPropertiesOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetNativeBufferPropertiesOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetNativeBufferPropertiesOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetNativeBufferPropertiesOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetNativeBufferPropertiesOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetNativeBufferPropertiesOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetNativeBufferPropertiesOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetNativeBufferPropertiesOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetNativeBufferPropertiesOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetNativeBufferPropertiesOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetNativeBufferPropertiesOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetNativeBufferPropertiesOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetNativeBufferPropertiesOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetNativeBufferPropertiesOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetNativeBufferPropertiesOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetNativeBufferPropertiesOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetNativeBufferPropertiesOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetNativeBufferPropertiesOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetNativeBufferPropertiesOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetNativeBufferPropertiesOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetNativeBufferPropertiesOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetNativeBufferPropertiesOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetNativeBufferPropertiesOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetNativeBufferPropertiesOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetNativeBufferPropertiesOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetNativeBufferPropertiesOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetNativeBufferPropertiesOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetNativeBufferPropertiesOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetNativeBufferPropertiesOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetNativeBufferPropertiesOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetNativeBufferPropertiesOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetNativeBufferPropertiesOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetNativeBufferPropertiesOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetNativeBufferPropertiesOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetNativeBufferPropertiesOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetNativeBufferPropertiesOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetNativeBufferPropertiesOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetNativeBufferPropertiesOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetNativeBufferPropertiesOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetNativeBufferPropertiesOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetNativeBufferPropertiesOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetNativeBufferPropertiesOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetNativeBufferPropertiesOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetNativeBufferPropertiesOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetNativeBufferPropertiesOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetNativeBufferPropertiesOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetNativeBufferPropertiesOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetNativeBufferPropertiesOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetNativeBufferPropertiesOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetNativeBufferPropertiesOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetNativeBufferPropertiesOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetNativeBufferPropertiesOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetNativeBufferPropertiesOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetNativeBufferPropertiesOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetNativeBufferPropertiesOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetNativeBufferPropertiesOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetNativeBufferPropertiesOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetNativeBufferPropertiesOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetNativeBufferPropertiesOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetNativeBufferPropertiesOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetNativeBufferPropertiesOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetNativeBufferPropertiesOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetNativeBufferPropertiesOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetNativeBufferPropertiesOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetNativeBufferPropertiesOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetNativeBufferPropertiesOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetNativeBufferPropertiesOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetMemoryNativeBufferOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryNativeBufferOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryNativeBufferOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryNativeBufferOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryNativeBufferOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryNativeBufferOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryNativeBufferOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryNativeBufferOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryNativeBufferOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryNativeBufferOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryNativeBufferOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryNativeBufferOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryNativeBufferOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetMemoryNativeBufferOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetMemoryNativeBufferOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetMemoryNativeBufferOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryNativeBufferOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryNativeBufferOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryNativeBufferOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryNativeBufferOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryNativeBufferOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryNativeBufferOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryNativeBufferOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryNativeBufferOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryNativeBufferOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryNativeBufferOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryNativeBufferOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryNativeBufferOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryNativeBufferOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetMemoryNativeBufferOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetMemoryNativeBufferOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetMemoryNativeBufferOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryNativeBufferOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryNativeBufferOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryNativeBufferOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryNativeBufferOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryNativeBufferOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryNativeBufferOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryNativeBufferOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryNativeBufferOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryNativeBufferOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryNativeBufferOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryNativeBufferOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryNativeBufferOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryNativeBufferOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetMemoryNativeBufferOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetMemoryNativeBufferOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetMemoryNativeBufferOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryNativeBufferOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryNativeBufferOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryNativeBufferOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryNativeBufferOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryNativeBufferOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryNativeBufferOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryNativeBufferOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryNativeBufferOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryNativeBufferOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryNativeBufferOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryNativeBufferOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryNativeBufferOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryNativeBufferOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetMemoryNativeBufferOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetMemoryNativeBufferOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetMemoryNativeBufferOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryNativeBufferOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryNativeBufferOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryNativeBufferOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryNativeBufferOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryNativeBufferOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryNativeBufferOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryNativeBufferOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryNativeBufferOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryNativeBufferOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryNativeBufferOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryNativeBufferOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryNativeBufferOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryNativeBufferOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetMemoryNativeBufferOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetMemoryNativeBufferOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetMemoryNativeBufferOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryNativeBufferOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryNativeBufferOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryNativeBufferOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryNativeBufferOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryNativeBufferOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryNativeBufferOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryNativeBufferOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryNativeBufferOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryNativeBufferOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryNativeBufferOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryNativeBufferOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryNativeBufferOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryNativeBufferOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetMemoryNativeBufferOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetMemoryNativeBufferOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetMemoryNativeBufferOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryNativeBufferOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryNativeBufferOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryNativeBufferOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryNativeBufferOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryNativeBufferOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryNativeBufferOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryNativeBufferOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryNativeBufferOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryNativeBufferOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryNativeBufferOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryNativeBufferOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryNativeBufferOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryNativeBufferOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetMemoryNativeBufferOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetMemoryNativeBufferOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetMemoryNativeBufferOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryNativeBufferOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryNativeBufferOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryNativeBufferOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryNativeBufferOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryNativeBufferOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryNativeBufferOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryNativeBufferOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryNativeBufferOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryNativeBufferOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryNativeBufferOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryNativeBufferOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryNativeBufferOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryNativeBufferOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetMemoryNativeBufferOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetMemoryNativeBufferOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetMemoryNativeBufferOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryNativeBufferOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryNativeBufferOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryNativeBufferOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryNativeBufferOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryNativeBufferOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryNativeBufferOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryNativeBufferOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryNativeBufferOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryNativeBufferOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryNativeBufferOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryNativeBufferOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryNativeBufferOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryNativeBufferOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetMemoryNativeBufferOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetMemoryNativeBufferOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetMemoryNativeBufferOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryNativeBufferOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryNativeBufferOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryNativeBufferOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryNativeBufferOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryNativeBufferOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryNativeBufferOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryNativeBufferOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryNativeBufferOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryNativeBufferOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryNativeBufferOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryNativeBufferOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryNativeBufferOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryNativeBufferOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetMemoryNativeBufferOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetMemoryNativeBufferOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetMemoryNativeBufferOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryNativeBufferOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryNativeBufferOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryNativeBufferOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryNativeBufferOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryNativeBufferOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryNativeBufferOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryNativeBufferOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryNativeBufferOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryNativeBufferOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryNativeBufferOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryNativeBufferOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryNativeBufferOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryNativeBufferOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetMemoryNativeBufferOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetMemoryNativeBufferOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetMemoryNativeBufferOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryNativeBufferOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryNativeBufferOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryNativeBufferOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryNativeBufferOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryNativeBufferOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryNativeBufferOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryNativeBufferOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryNativeBufferOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryNativeBufferOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryNativeBufferOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryNativeBufferOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryNativeBufferOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryNativeBufferOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetMemoryNativeBufferOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetMemoryNativeBufferOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetMemoryNativeBufferOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryNativeBufferOHOS@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryNativeBufferOHOS@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryNativeBufferOHOS@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryNativeBufferOHOS@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryNativeBufferOHOS@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryNativeBufferOHOS@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryNativeBufferOHOS@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryNativeBufferOHOS@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryNativeBufferOHOS@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryNativeBufferOHOS@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryNativeBufferOHOS@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryNativeBufferOHOS@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryNativeBufferOHOS@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetMemoryNativeBufferOHOS@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetMemoryNativeBufferOHOS=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetMemoryNativeBufferOHOS=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#endif // VK_USE_PLATFORM_OHOS
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerfHintQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerfHintQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerfHintQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerfHintQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerfHintQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerfHintQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerfHintQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerfHintQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerfHintQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerfHintQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerfHintQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerfHintQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerfHintQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_QueueSetPerfHintQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_QueueSetPerfHintQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_QueueSetPerfHintQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerfHintQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerfHintQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerfHintQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerfHintQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerfHintQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerfHintQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerfHintQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerfHintQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerfHintQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerfHintQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerfHintQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerfHintQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerfHintQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_QueueSetPerfHintQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_QueueSetPerfHintQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_QueueSetPerfHintQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerfHintQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerfHintQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerfHintQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerfHintQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerfHintQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerfHintQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerfHintQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerfHintQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerfHintQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerfHintQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerfHintQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerfHintQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerfHintQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_QueueSetPerfHintQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_QueueSetPerfHintQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_QueueSetPerfHintQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerfHintQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerfHintQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerfHintQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerfHintQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerfHintQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerfHintQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerfHintQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerfHintQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerfHintQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerfHintQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerfHintQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerfHintQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerfHintQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_QueueSetPerfHintQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_QueueSetPerfHintQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_QueueSetPerfHintQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerfHintQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerfHintQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerfHintQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerfHintQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerfHintQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerfHintQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerfHintQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerfHintQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerfHintQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerfHintQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerfHintQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerfHintQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerfHintQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_QueueSetPerfHintQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_QueueSetPerfHintQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_QueueSetPerfHintQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerfHintQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerfHintQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerfHintQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerfHintQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerfHintQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerfHintQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerfHintQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerfHintQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerfHintQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerfHintQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerfHintQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerfHintQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerfHintQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_QueueSetPerfHintQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_QueueSetPerfHintQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_QueueSetPerfHintQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerfHintQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerfHintQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerfHintQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerfHintQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerfHintQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerfHintQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerfHintQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerfHintQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerfHintQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerfHintQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerfHintQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerfHintQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerfHintQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_QueueSetPerfHintQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_QueueSetPerfHintQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_QueueSetPerfHintQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerfHintQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerfHintQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerfHintQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerfHintQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerfHintQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerfHintQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerfHintQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerfHintQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerfHintQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerfHintQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerfHintQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerfHintQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerfHintQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_QueueSetPerfHintQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_QueueSetPerfHintQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_QueueSetPerfHintQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerfHintQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerfHintQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerfHintQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerfHintQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerfHintQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerfHintQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerfHintQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerfHintQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerfHintQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerfHintQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerfHintQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerfHintQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerfHintQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_QueueSetPerfHintQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_QueueSetPerfHintQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_QueueSetPerfHintQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerfHintQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerfHintQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerfHintQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerfHintQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerfHintQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerfHintQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerfHintQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerfHintQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerfHintQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerfHintQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerfHintQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerfHintQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerfHintQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_QueueSetPerfHintQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_QueueSetPerfHintQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_QueueSetPerfHintQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerfHintQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerfHintQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerfHintQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerfHintQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerfHintQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerfHintQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerfHintQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerfHintQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerfHintQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerfHintQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerfHintQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerfHintQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerfHintQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_QueueSetPerfHintQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_QueueSetPerfHintQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_QueueSetPerfHintQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerfHintQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerfHintQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerfHintQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerfHintQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerfHintQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerfHintQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerfHintQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerfHintQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerfHintQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerfHintQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerfHintQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerfHintQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerfHintQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_QueueSetPerfHintQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_QueueSetPerfHintQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_QueueSetPerfHintQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerfHintQCOM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerfHintQCOM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerfHintQCOM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerfHintQCOM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerfHintQCOM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerfHintQCOM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerfHintQCOM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerfHintQCOM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerfHintQCOM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerfHintQCOM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerfHintQCOM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerfHintQCOM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerfHintQCOM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_QueueSetPerfHintQCOM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_QueueSetPerfHintQCOM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_QueueSetPerfHintQCOM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetComputeOccupancyPriorityNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetComputeOccupancyPriorityNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetComputeOccupancyPriorityNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetComputeOccupancyPriorityNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetComputeOccupancyPriorityNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetComputeOccupancyPriorityNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetComputeOccupancyPriorityNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetComputeOccupancyPriorityNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetComputeOccupancyPriorityNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetComputeOccupancyPriorityNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetComputeOccupancyPriorityNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetComputeOccupancyPriorityNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetComputeOccupancyPriorityNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetComputeOccupancyPriorityNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetComputeOccupancyPriorityNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetComputeOccupancyPriorityNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetComputeOccupancyPriorityNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetComputeOccupancyPriorityNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetComputeOccupancyPriorityNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetComputeOccupancyPriorityNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetComputeOccupancyPriorityNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetComputeOccupancyPriorityNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetComputeOccupancyPriorityNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetComputeOccupancyPriorityNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetComputeOccupancyPriorityNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetComputeOccupancyPriorityNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetComputeOccupancyPriorityNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetComputeOccupancyPriorityNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetComputeOccupancyPriorityNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetComputeOccupancyPriorityNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetComputeOccupancyPriorityNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetComputeOccupancyPriorityNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetComputeOccupancyPriorityNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetComputeOccupancyPriorityNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetComputeOccupancyPriorityNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetComputeOccupancyPriorityNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetComputeOccupancyPriorityNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetComputeOccupancyPriorityNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetComputeOccupancyPriorityNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetComputeOccupancyPriorityNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetComputeOccupancyPriorityNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetComputeOccupancyPriorityNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetComputeOccupancyPriorityNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetComputeOccupancyPriorityNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetComputeOccupancyPriorityNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetComputeOccupancyPriorityNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetComputeOccupancyPriorityNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetComputeOccupancyPriorityNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetComputeOccupancyPriorityNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetComputeOccupancyPriorityNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetComputeOccupancyPriorityNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetComputeOccupancyPriorityNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetComputeOccupancyPriorityNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetComputeOccupancyPriorityNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetComputeOccupancyPriorityNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetComputeOccupancyPriorityNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetComputeOccupancyPriorityNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetComputeOccupancyPriorityNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetComputeOccupancyPriorityNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetComputeOccupancyPriorityNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetComputeOccupancyPriorityNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetComputeOccupancyPriorityNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetComputeOccupancyPriorityNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetComputeOccupancyPriorityNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetComputeOccupancyPriorityNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetComputeOccupancyPriorityNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetComputeOccupancyPriorityNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetComputeOccupancyPriorityNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetComputeOccupancyPriorityNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetComputeOccupancyPriorityNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetComputeOccupancyPriorityNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetComputeOccupancyPriorityNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetComputeOccupancyPriorityNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetComputeOccupancyPriorityNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetComputeOccupancyPriorityNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetComputeOccupancyPriorityNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetComputeOccupancyPriorityNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetComputeOccupancyPriorityNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetComputeOccupancyPriorityNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetComputeOccupancyPriorityNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetComputeOccupancyPriorityNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetComputeOccupancyPriorityNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetComputeOccupancyPriorityNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetComputeOccupancyPriorityNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetComputeOccupancyPriorityNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetComputeOccupancyPriorityNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetComputeOccupancyPriorityNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetComputeOccupancyPriorityNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetComputeOccupancyPriorityNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetComputeOccupancyPriorityNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetComputeOccupancyPriorityNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetComputeOccupancyPriorityNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetComputeOccupancyPriorityNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetComputeOccupancyPriorityNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetComputeOccupancyPriorityNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetComputeOccupancyPriorityNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetComputeOccupancyPriorityNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetComputeOccupancyPriorityNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetComputeOccupancyPriorityNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetComputeOccupancyPriorityNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetComputeOccupancyPriorityNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetComputeOccupancyPriorityNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetComputeOccupancyPriorityNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetComputeOccupancyPriorityNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetComputeOccupancyPriorityNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetComputeOccupancyPriorityNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetComputeOccupancyPriorityNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetComputeOccupancyPriorityNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetComputeOccupancyPriorityNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetComputeOccupancyPriorityNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetComputeOccupancyPriorityNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetComputeOccupancyPriorityNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetComputeOccupancyPriorityNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetComputeOccupancyPriorityNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetComputeOccupancyPriorityNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetComputeOccupancyPriorityNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetComputeOccupancyPriorityNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetComputeOccupancyPriorityNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetComputeOccupancyPriorityNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetComputeOccupancyPriorityNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetComputeOccupancyPriorityNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetComputeOccupancyPriorityNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetComputeOccupancyPriorityNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetComputeOccupancyPriorityNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetComputeOccupancyPriorityNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetComputeOccupancyPriorityNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetComputeOccupancyPriorityNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetComputeOccupancyPriorityNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetComputeOccupancyPriorityNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetComputeOccupancyPriorityNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetComputeOccupancyPriorityNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetComputeOccupancyPriorityNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetComputeOccupancyPriorityNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetComputeOccupancyPriorityNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetComputeOccupancyPriorityNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetComputeOccupancyPriorityNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetComputeOccupancyPriorityNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetComputeOccupancyPriorityNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetComputeOccupancyPriorityNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetComputeOccupancyPriorityNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetComputeOccupancyPriorityNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetComputeOccupancyPriorityNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetComputeOccupancyPriorityNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetComputeOccupancyPriorityNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetComputeOccupancyPriorityNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetComputeOccupancyPriorityNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetComputeOccupancyPriorityNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetComputeOccupancyPriorityNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetComputeOccupancyPriorityNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetComputeOccupancyPriorityNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetComputeOccupancyPriorityNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetComputeOccupancyPriorityNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetComputeOccupancyPriorityNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetComputeOccupancyPriorityNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetComputeOccupancyPriorityNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetComputeOccupancyPriorityNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetComputeOccupancyPriorityNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetComputeOccupancyPriorityNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetComputeOccupancyPriorityNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetComputeOccupancyPriorityNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetComputeOccupancyPriorityNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetComputeOccupancyPriorityNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetComputeOccupancyPriorityNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetComputeOccupancyPriorityNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetComputeOccupancyPriorityNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetComputeOccupancyPriorityNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetComputeOccupancyPriorityNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetComputeOccupancyPriorityNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetComputeOccupancyPriorityNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetComputeOccupancyPriorityNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetComputeOccupancyPriorityNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetComputeOccupancyPriorityNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetComputeOccupancyPriorityNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetComputeOccupancyPriorityNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetComputeOccupancyPriorityNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetComputeOccupancyPriorityNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetComputeOccupancyPriorityNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetComputeOccupancyPriorityNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetComputeOccupancyPriorityNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetComputeOccupancyPriorityNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetComputeOccupancyPriorityNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetComputeOccupancyPriorityNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetComputeOccupancyPriorityNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetComputeOccupancyPriorityNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetComputeOccupancyPriorityNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetComputeOccupancyPriorityNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetComputeOccupancyPriorityNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetComputeOccupancyPriorityNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetComputeOccupancyPriorityNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetComputeOccupancyPriorityNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetComputeOccupancyPriorityNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetComputeOccupancyPriorityNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetComputeOccupancyPriorityNV@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetComputeOccupancyPriorityNV@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetComputeOccupancyPriorityNV@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetComputeOccupancyPriorityNV@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetComputeOccupancyPriorityNV@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetComputeOccupancyPriorityNV@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetComputeOccupancyPriorityNV@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetComputeOccupancyPriorityNV@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetComputeOccupancyPriorityNV@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetComputeOccupancyPriorityNV@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetComputeOccupancyPriorityNV@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetComputeOccupancyPriorityNV@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetComputeOccupancyPriorityNV@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetComputeOccupancyPriorityNV@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetComputeOccupancyPriorityNV=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetComputeOccupancyPriorityNV=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_WriteSamplerDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteSamplerDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteSamplerDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteSamplerDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteSamplerDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteSamplerDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteSamplerDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteSamplerDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteSamplerDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteSamplerDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteSamplerDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteSamplerDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteSamplerDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteSamplerDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_WriteSamplerDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_WriteSamplerDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_WriteSamplerDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteSamplerDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteSamplerDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteSamplerDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteSamplerDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteSamplerDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteSamplerDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteSamplerDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteSamplerDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteSamplerDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteSamplerDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteSamplerDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteSamplerDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteSamplerDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_WriteSamplerDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_WriteSamplerDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_WriteSamplerDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteSamplerDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteSamplerDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteSamplerDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteSamplerDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteSamplerDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteSamplerDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteSamplerDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteSamplerDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteSamplerDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteSamplerDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteSamplerDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteSamplerDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteSamplerDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_WriteSamplerDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_WriteSamplerDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_WriteSamplerDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteSamplerDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteSamplerDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteSamplerDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteSamplerDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteSamplerDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteSamplerDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteSamplerDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteSamplerDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteSamplerDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteSamplerDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteSamplerDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteSamplerDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteSamplerDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_WriteSamplerDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_WriteSamplerDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_WriteSamplerDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteSamplerDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteSamplerDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteSamplerDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteSamplerDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteSamplerDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteSamplerDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteSamplerDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteSamplerDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteSamplerDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteSamplerDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteSamplerDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteSamplerDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteSamplerDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_WriteSamplerDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_WriteSamplerDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_WriteSamplerDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteSamplerDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteSamplerDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteSamplerDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteSamplerDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteSamplerDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteSamplerDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteSamplerDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteSamplerDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteSamplerDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteSamplerDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteSamplerDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteSamplerDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteSamplerDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_WriteSamplerDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_WriteSamplerDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_WriteSamplerDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteSamplerDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteSamplerDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteSamplerDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteSamplerDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteSamplerDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteSamplerDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteSamplerDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteSamplerDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteSamplerDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteSamplerDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteSamplerDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteSamplerDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteSamplerDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_WriteSamplerDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_WriteSamplerDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_WriteSamplerDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteSamplerDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteSamplerDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteSamplerDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteSamplerDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteSamplerDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteSamplerDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteSamplerDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteSamplerDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteSamplerDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteSamplerDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteSamplerDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteSamplerDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteSamplerDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_WriteSamplerDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_WriteSamplerDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteSamplerDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteSamplerDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteSamplerDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteSamplerDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteSamplerDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteSamplerDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteSamplerDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteSamplerDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteSamplerDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteSamplerDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteSamplerDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteSamplerDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteSamplerDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteSamplerDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_WriteSamplerDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_WriteSamplerDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteSamplerDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteSamplerDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteSamplerDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteSamplerDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteSamplerDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteSamplerDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteSamplerDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteSamplerDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteSamplerDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteSamplerDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteSamplerDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteSamplerDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteSamplerDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteSamplerDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_WriteSamplerDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_WriteSamplerDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteSamplerDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteSamplerDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteSamplerDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteSamplerDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteSamplerDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteSamplerDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteSamplerDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteSamplerDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteSamplerDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteSamplerDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteSamplerDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteSamplerDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteSamplerDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteSamplerDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_WriteSamplerDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_WriteSamplerDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_WriteSamplerDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteSamplerDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteSamplerDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteSamplerDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteSamplerDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteSamplerDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteSamplerDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteSamplerDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteSamplerDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteSamplerDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteSamplerDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteSamplerDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteSamplerDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteSamplerDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_WriteSamplerDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_WriteSamplerDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteSamplerDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteSamplerDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteSamplerDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteSamplerDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteSamplerDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteSamplerDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteSamplerDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteSamplerDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteSamplerDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteSamplerDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteSamplerDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteSamplerDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteSamplerDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteSamplerDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_WriteSamplerDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_WriteSamplerDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_WriteResourceDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteResourceDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteResourceDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteResourceDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteResourceDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteResourceDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteResourceDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteResourceDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteResourceDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteResourceDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteResourceDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteResourceDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteResourceDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_WriteResourceDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_WriteResourceDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_WriteResourceDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_WriteResourceDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteResourceDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteResourceDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteResourceDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteResourceDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteResourceDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteResourceDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteResourceDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteResourceDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteResourceDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteResourceDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteResourceDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteResourceDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_WriteResourceDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_WriteResourceDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_WriteResourceDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_WriteResourceDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteResourceDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteResourceDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteResourceDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteResourceDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteResourceDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteResourceDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteResourceDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteResourceDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteResourceDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteResourceDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteResourceDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteResourceDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_WriteResourceDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_WriteResourceDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_WriteResourceDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_WriteResourceDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteResourceDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteResourceDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteResourceDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteResourceDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteResourceDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteResourceDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteResourceDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteResourceDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteResourceDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteResourceDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteResourceDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteResourceDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_WriteResourceDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_WriteResourceDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_WriteResourceDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_WriteResourceDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteResourceDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteResourceDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteResourceDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteResourceDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteResourceDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteResourceDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteResourceDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteResourceDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteResourceDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteResourceDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteResourceDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteResourceDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_WriteResourceDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_WriteResourceDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_WriteResourceDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_WriteResourceDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteResourceDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteResourceDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteResourceDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteResourceDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteResourceDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteResourceDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteResourceDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteResourceDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteResourceDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteResourceDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteResourceDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteResourceDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_WriteResourceDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_WriteResourceDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_WriteResourceDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_WriteResourceDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteResourceDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteResourceDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteResourceDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteResourceDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteResourceDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteResourceDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteResourceDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteResourceDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteResourceDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteResourceDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteResourceDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteResourceDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_WriteResourceDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_WriteResourceDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_WriteResourceDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_WriteResourceDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteResourceDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteResourceDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteResourceDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteResourceDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteResourceDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteResourceDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteResourceDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteResourceDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteResourceDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteResourceDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteResourceDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteResourceDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_WriteResourceDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_WriteResourceDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_WriteResourceDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteResourceDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteResourceDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteResourceDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteResourceDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteResourceDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteResourceDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteResourceDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteResourceDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteResourceDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteResourceDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteResourceDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteResourceDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteResourceDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_WriteResourceDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_WriteResourceDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_WriteResourceDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteResourceDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteResourceDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteResourceDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteResourceDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteResourceDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteResourceDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteResourceDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteResourceDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteResourceDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteResourceDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteResourceDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteResourceDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteResourceDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_WriteResourceDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_WriteResourceDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_WriteResourceDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteResourceDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteResourceDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteResourceDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteResourceDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteResourceDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteResourceDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteResourceDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteResourceDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteResourceDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteResourceDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteResourceDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteResourceDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteResourceDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_WriteResourceDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_WriteResourceDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_WriteResourceDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_WriteResourceDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteResourceDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteResourceDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteResourceDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteResourceDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteResourceDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteResourceDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteResourceDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteResourceDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteResourceDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteResourceDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteResourceDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteResourceDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_WriteResourceDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_WriteResourceDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_WriteResourceDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteResourceDescriptorsEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteResourceDescriptorsEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteResourceDescriptorsEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteResourceDescriptorsEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteResourceDescriptorsEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteResourceDescriptorsEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteResourceDescriptorsEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteResourceDescriptorsEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteResourceDescriptorsEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteResourceDescriptorsEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteResourceDescriptorsEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteResourceDescriptorsEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteResourceDescriptorsEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_WriteResourceDescriptorsEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_WriteResourceDescriptorsEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_WriteResourceDescriptorsEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindSamplerHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindSamplerHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindSamplerHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindSamplerHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindSamplerHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindSamplerHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindSamplerHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindSamplerHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindSamplerHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindSamplerHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindSamplerHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindSamplerHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindSamplerHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindSamplerHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindSamplerHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindSamplerHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindSamplerHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindSamplerHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindSamplerHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindSamplerHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindSamplerHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindSamplerHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindSamplerHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindSamplerHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindSamplerHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindSamplerHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindSamplerHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindSamplerHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindSamplerHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindSamplerHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindSamplerHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindSamplerHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindSamplerHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindSamplerHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindSamplerHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindSamplerHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindSamplerHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindSamplerHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindSamplerHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindSamplerHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindSamplerHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindSamplerHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindSamplerHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindSamplerHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindSamplerHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindSamplerHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindSamplerHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindSamplerHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindSamplerHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindSamplerHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindSamplerHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindSamplerHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindSamplerHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindSamplerHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindSamplerHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindSamplerHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindSamplerHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindSamplerHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindSamplerHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindSamplerHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindSamplerHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindSamplerHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindSamplerHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindSamplerHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindSamplerHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindSamplerHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindSamplerHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindSamplerHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindSamplerHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindSamplerHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindSamplerHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindSamplerHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindSamplerHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindSamplerHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindSamplerHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindSamplerHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindSamplerHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindSamplerHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindSamplerHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindSamplerHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindSamplerHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindSamplerHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindSamplerHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindSamplerHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindSamplerHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindSamplerHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindSamplerHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindSamplerHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindSamplerHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindSamplerHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindSamplerHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindSamplerHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindSamplerHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindSamplerHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindSamplerHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindSamplerHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindSamplerHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindSamplerHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindSamplerHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindSamplerHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindSamplerHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindSamplerHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindSamplerHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindSamplerHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindSamplerHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindSamplerHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindSamplerHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindSamplerHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindSamplerHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindSamplerHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindSamplerHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindSamplerHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindSamplerHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindSamplerHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindSamplerHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindSamplerHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindSamplerHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindSamplerHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindSamplerHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindSamplerHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindSamplerHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindSamplerHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindSamplerHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindSamplerHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindSamplerHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindSamplerHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindSamplerHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindSamplerHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindSamplerHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindSamplerHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindSamplerHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindSamplerHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindSamplerHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindSamplerHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindSamplerHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindSamplerHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindSamplerHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindSamplerHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindSamplerHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindSamplerHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindSamplerHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindSamplerHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindSamplerHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindSamplerHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindSamplerHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindSamplerHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindSamplerHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindSamplerHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindSamplerHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindSamplerHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindSamplerHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindSamplerHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindSamplerHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindSamplerHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindSamplerHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindSamplerHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindSamplerHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindSamplerHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindSamplerHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindSamplerHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindSamplerHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindSamplerHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindSamplerHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindSamplerHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindSamplerHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindSamplerHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindSamplerHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindSamplerHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindSamplerHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindSamplerHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindSamplerHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindSamplerHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindSamplerHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindSamplerHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindSamplerHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindSamplerHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindSamplerHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindSamplerHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindSamplerHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindSamplerHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindSamplerHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindSamplerHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindSamplerHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindSamplerHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindSamplerHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindSamplerHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindSamplerHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindSamplerHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindSamplerHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindSamplerHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindSamplerHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindSamplerHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindSamplerHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindSamplerHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindSamplerHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindSamplerHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindSamplerHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindSamplerHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindSamplerHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindSamplerHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindSamplerHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindSamplerHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindSamplerHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindSamplerHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindSamplerHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindSamplerHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindSamplerHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindSamplerHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindResourceHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindResourceHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindResourceHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindResourceHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindResourceHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindResourceHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindResourceHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindResourceHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindResourceHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindResourceHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindResourceHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindResourceHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindResourceHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindResourceHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindResourceHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindResourceHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindResourceHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindResourceHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindResourceHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindResourceHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindResourceHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindResourceHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindResourceHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindResourceHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindResourceHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindResourceHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindResourceHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindResourceHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindResourceHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindResourceHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindResourceHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindResourceHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindResourceHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindResourceHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindResourceHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindResourceHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindResourceHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindResourceHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindResourceHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindResourceHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindResourceHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindResourceHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindResourceHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindResourceHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindResourceHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindResourceHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindResourceHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindResourceHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindResourceHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindResourceHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindResourceHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindResourceHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindResourceHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindResourceHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindResourceHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindResourceHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindResourceHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindResourceHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindResourceHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindResourceHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindResourceHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindResourceHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindResourceHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindResourceHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindResourceHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindResourceHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindResourceHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindResourceHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindResourceHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindResourceHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindResourceHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindResourceHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindResourceHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindResourceHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindResourceHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindResourceHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindResourceHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindResourceHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindResourceHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindResourceHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindResourceHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindResourceHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindResourceHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindResourceHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindResourceHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindResourceHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindResourceHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindResourceHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindResourceHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindResourceHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindResourceHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindResourceHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindResourceHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindResourceHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindResourceHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindResourceHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindResourceHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindResourceHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindResourceHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindResourceHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindResourceHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindResourceHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindResourceHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindResourceHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindResourceHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindResourceHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindResourceHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindResourceHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindResourceHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindResourceHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindResourceHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindResourceHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindResourceHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindResourceHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindResourceHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindResourceHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindResourceHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindResourceHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindResourceHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindResourceHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindResourceHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindResourceHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindResourceHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindResourceHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindResourceHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindResourceHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindResourceHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindResourceHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindResourceHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindResourceHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindResourceHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindResourceHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindResourceHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindResourceHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindResourceHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindResourceHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindResourceHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindResourceHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindResourceHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindResourceHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindResourceHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindResourceHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindResourceHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindResourceHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindResourceHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindResourceHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindResourceHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindResourceHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindResourceHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindResourceHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindResourceHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindResourceHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindResourceHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindResourceHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindResourceHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindResourceHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindResourceHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindResourceHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindResourceHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindResourceHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindResourceHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindResourceHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindResourceHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindResourceHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindResourceHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindResourceHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindResourceHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindResourceHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindResourceHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindResourceHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindResourceHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindResourceHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindResourceHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindResourceHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindResourceHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindResourceHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindResourceHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindResourceHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindResourceHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindResourceHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindResourceHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindResourceHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindResourceHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindResourceHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindResourceHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindResourceHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindResourceHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindResourceHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindResourceHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindResourceHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindResourceHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindResourceHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindResourceHeapEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindResourceHeapEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindResourceHeapEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindResourceHeapEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindResourceHeapEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindResourceHeapEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindResourceHeapEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindResourceHeapEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindResourceHeapEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindResourceHeapEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindResourceHeapEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindResourceHeapEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindResourceHeapEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindResourceHeapEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindResourceHeapEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindResourceHeapEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdPushDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdPushDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdPushDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdPushDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdPushDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdPushDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdPushDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdPushDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdPushDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdPushDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdPushDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdPushDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdPushDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdPushDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdPushDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdPushDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdPushDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdPushDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdPushDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdPushDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdPushDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdPushDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdPushDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdPushDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdPushDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdPushDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdPushDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdPushDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdPushDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdPushDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdPushDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdPushDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdPushDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdPushDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdPushDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdPushDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdPushDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdPushDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdPushDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdPushDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_RegisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_RegisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_RegisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_RegisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_RegisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_RegisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_RegisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_RegisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_RegisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_RegisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_RegisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_RegisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_RegisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_RegisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_RegisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_RegisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_RegisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_RegisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_RegisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_RegisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_RegisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_RegisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_RegisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_RegisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_RegisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_RegisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_RegisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_RegisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_RegisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_RegisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_RegisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_RegisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_RegisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_RegisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_RegisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_RegisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_RegisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_RegisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_RegisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_RegisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_RegisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_RegisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_RegisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_RegisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_RegisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_RegisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_RegisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_RegisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_UnregisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnregisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnregisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnregisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnregisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnregisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnregisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnregisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnregisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnregisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnregisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnregisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnregisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_UnregisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_UnregisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_UnregisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_UnregisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnregisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnregisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnregisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnregisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnregisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnregisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnregisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnregisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnregisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnregisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnregisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnregisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_UnregisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_UnregisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_UnregisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_UnregisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnregisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnregisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnregisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnregisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnregisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnregisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnregisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnregisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnregisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnregisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnregisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnregisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_UnregisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_UnregisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_UnregisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_UnregisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnregisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnregisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnregisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnregisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnregisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnregisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnregisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnregisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnregisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnregisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnregisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnregisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_UnregisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_UnregisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_UnregisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_UnregisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnregisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnregisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnregisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnregisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnregisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnregisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnregisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnregisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnregisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnregisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnregisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnregisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_UnregisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_UnregisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_UnregisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_UnregisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnregisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnregisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnregisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnregisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnregisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnregisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnregisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnregisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnregisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnregisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnregisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnregisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_UnregisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_UnregisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_UnregisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_UnregisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnregisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnregisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnregisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnregisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnregisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnregisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnregisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnregisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnregisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnregisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnregisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnregisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_UnregisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_UnregisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_UnregisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_UnregisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnregisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnregisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnregisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnregisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnregisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnregisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnregisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnregisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnregisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnregisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnregisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnregisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_UnregisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_UnregisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_UnregisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_UnregisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnregisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnregisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnregisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnregisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnregisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnregisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnregisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnregisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnregisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnregisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnregisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnregisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_UnregisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_UnregisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_UnregisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_UnregisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnregisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnregisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnregisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnregisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnregisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnregisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnregisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnregisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnregisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnregisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnregisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnregisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_UnregisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_UnregisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_UnregisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnregisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnregisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnregisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnregisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnregisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnregisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnregisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnregisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnregisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnregisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnregisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnregisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnregisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_UnregisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_UnregisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_UnregisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_UnregisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnregisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnregisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnregisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnregisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnregisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnregisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnregisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnregisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnregisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnregisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnregisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnregisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_UnregisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_UnregisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_UnregisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_UnregisterCustomBorderColorEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnregisterCustomBorderColorEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnregisterCustomBorderColorEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnregisterCustomBorderColorEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnregisterCustomBorderColorEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnregisterCustomBorderColorEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnregisterCustomBorderColorEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnregisterCustomBorderColorEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnregisterCustomBorderColorEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnregisterCustomBorderColorEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnregisterCustomBorderColorEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnregisterCustomBorderColorEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnregisterCustomBorderColorEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_UnregisterCustomBorderColorEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_UnregisterCustomBorderColorEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_UnregisterCustomBorderColorEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetImageOpaqueCaptureDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetImageOpaqueCaptureDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetImageOpaqueCaptureDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetImageOpaqueCaptureDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetImageOpaqueCaptureDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetImageOpaqueCaptureDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetImageOpaqueCaptureDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetImageOpaqueCaptureDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetImageOpaqueCaptureDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetImageOpaqueCaptureDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetImageOpaqueCaptureDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetImageOpaqueCaptureDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetImageOpaqueCaptureDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetImageOpaqueCaptureDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetImageOpaqueCaptureDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetImageOpaqueCaptureDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetImageOpaqueCaptureDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetImageOpaqueCaptureDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetImageOpaqueCaptureDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetImageOpaqueCaptureDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetImageOpaqueCaptureDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetImageOpaqueCaptureDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetImageOpaqueCaptureDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetImageOpaqueCaptureDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetImageOpaqueCaptureDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetImageOpaqueCaptureDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetImageOpaqueCaptureDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetImageOpaqueCaptureDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetImageOpaqueCaptureDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetImageOpaqueCaptureDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetImageOpaqueCaptureDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetImageOpaqueCaptureDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetImageOpaqueCaptureDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetImageOpaqueCaptureDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetImageOpaqueCaptureDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetImageOpaqueCaptureDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDataEXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDataEXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDataEXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDataEXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDataEXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDataEXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDataEXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDataEXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDataEXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDataEXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDataEXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDataEXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDataEXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetImageOpaqueCaptureDataEXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetImageOpaqueCaptureDataEXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetImageOpaqueCaptureDataEXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_GetTensorOpaqueCaptureDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_GetTensorOpaqueCaptureDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_GetTensorOpaqueCaptureDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_GetTensorOpaqueCaptureDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_GetTensorOpaqueCaptureDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_GetTensorOpaqueCaptureDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_GetTensorOpaqueCaptureDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_GetTensorOpaqueCaptureDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_GetTensorOpaqueCaptureDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_GetTensorOpaqueCaptureDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_GetTensorOpaqueCaptureDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_GetTensorOpaqueCaptureDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_GetTensorOpaqueCaptureDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_GetTensorOpaqueCaptureDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_GetTensorOpaqueCaptureDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_GetTensorOpaqueCaptureDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_GetTensorOpaqueCaptureDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_GetTensorOpaqueCaptureDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_GetTensorOpaqueCaptureDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_GetTensorOpaqueCaptureDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_GetTensorOpaqueCaptureDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_GetTensorOpaqueCaptureDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_GetTensorOpaqueCaptureDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_GetTensorOpaqueCaptureDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_GetTensorOpaqueCaptureDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_GetTensorOpaqueCaptureDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_GetTensorOpaqueCaptureDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_GetTensorOpaqueCaptureDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_GetTensorOpaqueCaptureDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_GetTensorOpaqueCaptureDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_GetTensorOpaqueCaptureDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_GetTensorOpaqueCaptureDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_GetTensorOpaqueCaptureDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_GetTensorOpaqueCaptureDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_GetTensorOpaqueCaptureDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_GetTensorOpaqueCaptureDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDataARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDataARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDataARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDataARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDataARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDataARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDataARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDataARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDataARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDataARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDataARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDataARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDataARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_GetTensorOpaqueCaptureDataARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_GetTensorOpaqueCaptureDataARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_GetTensorOpaqueCaptureDataARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyMemoryToImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyMemoryToImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyMemoryToImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyMemoryToImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyMemoryToImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyMemoryToImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyMemoryToImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyMemoryToImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyMemoryToImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyMemoryToImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyMemoryToImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyMemoryToImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyMemoryToImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyMemoryToImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyMemoryToImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyMemoryToImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyMemoryToImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyMemoryToImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyMemoryToImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyMemoryToImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyMemoryToImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyMemoryToImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyMemoryToImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyMemoryToImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyMemoryToImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyMemoryToImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyMemoryToImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyMemoryToImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyMemoryToImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyMemoryToImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyMemoryToImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyMemoryToImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyMemoryToImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyMemoryToImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyMemoryToImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyMemoryToImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyMemoryToImageKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyMemoryToImageKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyMemoryToImageKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyImageToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyImageToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyImageToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyImageToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyImageToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyImageToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyImageToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyImageToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyImageToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyImageToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyImageToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyImageToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyImageToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyImageToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyImageToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyImageToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyImageToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyImageToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyImageToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyImageToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyImageToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyImageToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyImageToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyImageToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyImageToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyImageToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyImageToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyImageToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyImageToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyImageToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyImageToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyImageToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyImageToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyImageToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyImageToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyImageToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyImageToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyImageToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyImageToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdUpdateMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdUpdateMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdUpdateMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdUpdateMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdUpdateMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdUpdateMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdUpdateMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdUpdateMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdUpdateMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdUpdateMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdUpdateMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdUpdateMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdUpdateMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdUpdateMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdUpdateMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdUpdateMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdUpdateMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdUpdateMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdUpdateMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdUpdateMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdUpdateMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdUpdateMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdUpdateMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdUpdateMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdUpdateMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdUpdateMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdUpdateMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdUpdateMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdUpdateMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdUpdateMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdUpdateMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdUpdateMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdUpdateMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdUpdateMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdUpdateMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdUpdateMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdUpdateMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdUpdateMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdUpdateMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdFillMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdFillMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdFillMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdFillMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdFillMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdFillMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdFillMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdFillMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdFillMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdFillMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdFillMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdFillMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdFillMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdFillMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdFillMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdFillMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdFillMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdFillMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdFillMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdFillMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdFillMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdFillMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdFillMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdFillMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdFillMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdFillMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdFillMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdFillMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdFillMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdFillMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdFillMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdFillMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdFillMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdFillMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdFillMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdFillMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdFillMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdFillMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdFillMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdFillMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResultsToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResultsToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResultsToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResultsToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResultsToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResultsToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResultsToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResultsToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResultsToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResultsToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResultsToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResultsToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResultsToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdCopyQueryPoolResultsToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdCopyQueryPoolResultsToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdCopyQueryPoolResultsToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResultsToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResultsToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResultsToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResultsToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResultsToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResultsToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResultsToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResultsToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResultsToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResultsToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResultsToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResultsToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResultsToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdCopyQueryPoolResultsToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdCopyQueryPoolResultsToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdCopyQueryPoolResultsToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResultsToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResultsToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResultsToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResultsToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResultsToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResultsToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResultsToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResultsToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResultsToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResultsToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResultsToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResultsToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResultsToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdCopyQueryPoolResultsToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdCopyQueryPoolResultsToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdCopyQueryPoolResultsToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResultsToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResultsToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResultsToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResultsToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResultsToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResultsToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResultsToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResultsToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResultsToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResultsToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResultsToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResultsToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResultsToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdCopyQueryPoolResultsToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdCopyQueryPoolResultsToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdCopyQueryPoolResultsToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResultsToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResultsToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResultsToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResultsToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResultsToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResultsToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResultsToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResultsToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResultsToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResultsToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResultsToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResultsToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResultsToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdCopyQueryPoolResultsToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdCopyQueryPoolResultsToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdCopyQueryPoolResultsToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResultsToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResultsToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResultsToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResultsToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResultsToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResultsToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResultsToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResultsToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResultsToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResultsToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResultsToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResultsToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResultsToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdCopyQueryPoolResultsToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdCopyQueryPoolResultsToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdCopyQueryPoolResultsToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResultsToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResultsToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResultsToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResultsToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResultsToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResultsToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResultsToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResultsToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResultsToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResultsToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResultsToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResultsToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResultsToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdCopyQueryPoolResultsToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdCopyQueryPoolResultsToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdCopyQueryPoolResultsToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResultsToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResultsToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResultsToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResultsToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResultsToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResultsToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResultsToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResultsToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResultsToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResultsToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResultsToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResultsToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResultsToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdCopyQueryPoolResultsToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdCopyQueryPoolResultsToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdCopyQueryPoolResultsToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResultsToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResultsToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResultsToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResultsToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResultsToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResultsToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResultsToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResultsToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResultsToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResultsToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResultsToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResultsToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResultsToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdCopyQueryPoolResultsToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdCopyQueryPoolResultsToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdCopyQueryPoolResultsToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResultsToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResultsToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResultsToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResultsToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResultsToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResultsToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResultsToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResultsToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResultsToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResultsToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResultsToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResultsToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResultsToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdCopyQueryPoolResultsToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdCopyQueryPoolResultsToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdCopyQueryPoolResultsToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResultsToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResultsToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResultsToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResultsToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResultsToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResultsToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResultsToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResultsToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResultsToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResultsToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResultsToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResultsToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResultsToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdCopyQueryPoolResultsToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdCopyQueryPoolResultsToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdCopyQueryPoolResultsToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResultsToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResultsToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResultsToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResultsToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResultsToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResultsToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResultsToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResultsToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResultsToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResultsToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResultsToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResultsToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResultsToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdCopyQueryPoolResultsToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdCopyQueryPoolResultsToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdCopyQueryPoolResultsToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResultsToMemoryKHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResultsToMemoryKHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResultsToMemoryKHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResultsToMemoryKHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResultsToMemoryKHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResultsToMemoryKHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResultsToMemoryKHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResultsToMemoryKHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResultsToMemoryKHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResultsToMemoryKHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResultsToMemoryKHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResultsToMemoryKHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResultsToMemoryKHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdCopyQueryPoolResultsToMemoryKHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdCopyQueryPoolResultsToMemoryKHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdCopyQueryPoolResultsToMemoryKHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginConditionalRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBeginConditionalRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBeginConditionalRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginConditionalRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBeginConditionalRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBeginConditionalRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginConditionalRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBeginConditionalRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBeginConditionalRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginConditionalRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBeginConditionalRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBeginConditionalRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginConditionalRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBeginConditionalRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBeginConditionalRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginConditionalRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBeginConditionalRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBeginConditionalRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginConditionalRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBeginConditionalRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBeginConditionalRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginConditionalRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBeginConditionalRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBeginConditionalRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginConditionalRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBeginConditionalRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBeginConditionalRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginConditionalRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBeginConditionalRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBeginConditionalRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginConditionalRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBeginConditionalRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBeginConditionalRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginConditionalRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBeginConditionalRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBeginConditionalRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRendering2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRendering2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRendering2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRendering2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRendering2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRendering2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRendering2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRendering2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRendering2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRendering2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRendering2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRendering2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRendering2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginConditionalRendering2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBeginConditionalRendering2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBeginConditionalRendering2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindTransformFeedbackBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindTransformFeedbackBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindTransformFeedbackBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindTransformFeedbackBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindTransformFeedbackBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindTransformFeedbackBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindTransformFeedbackBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindTransformFeedbackBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindTransformFeedbackBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindTransformFeedbackBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindTransformFeedbackBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindTransformFeedbackBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindTransformFeedbackBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindTransformFeedbackBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindTransformFeedbackBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindTransformFeedbackBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindTransformFeedbackBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindTransformFeedbackBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindTransformFeedbackBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindTransformFeedbackBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindTransformFeedbackBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindTransformFeedbackBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindTransformFeedbackBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindTransformFeedbackBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindTransformFeedbackBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindTransformFeedbackBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindTransformFeedbackBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindTransformFeedbackBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindTransformFeedbackBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindTransformFeedbackBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindTransformFeedbackBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindTransformFeedbackBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindTransformFeedbackBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindTransformFeedbackBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindTransformFeedbackBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindTransformFeedbackBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffers2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffers2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffers2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffers2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffers2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffers2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffers2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffers2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffers2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffers2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffers2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffers2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffers2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindTransformFeedbackBuffers2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindTransformFeedbackBuffers2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindTransformFeedbackBuffers2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBeginTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBeginTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBeginTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBeginTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBeginTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBeginTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBeginTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBeginTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBeginTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBeginTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBeginTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBeginTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBeginTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBeginTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBeginTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBeginTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBeginTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBeginTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBeginTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBeginTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBeginTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBeginTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBeginTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBeginTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBeginTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBeginTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBeginTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBeginTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBeginTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBeginTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBeginTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBeginTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBeginTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBeginTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBeginTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBeginTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBeginTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBeginTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBeginTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdEndTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdEndTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdEndTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdEndTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdEndTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdEndTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdEndTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdEndTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdEndTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdEndTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdEndTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdEndTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdEndTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdEndTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdEndTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdEndTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdEndTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdEndTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdEndTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdEndTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdEndTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdEndTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdEndTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdEndTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdEndTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdEndTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdEndTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdEndTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdEndTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdEndTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdEndTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdEndTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdEndTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdEndTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdEndTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdEndTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedback2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedback2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedback2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedback2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedback2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedback2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedback2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedback2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedback2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedback2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedback2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedback2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedback2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdEndTransformFeedback2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdEndTransformFeedback2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdEndTransformFeedback2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectByteCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawIndirectByteCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawIndirectByteCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectByteCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawIndirectByteCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawIndirectByteCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectByteCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawIndirectByteCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawIndirectByteCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectByteCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawIndirectByteCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawIndirectByteCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectByteCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawIndirectByteCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawIndirectByteCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectByteCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawIndirectByteCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawIndirectByteCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectByteCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawIndirectByteCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawIndirectByteCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectByteCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawIndirectByteCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawIndirectByteCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectByteCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawIndirectByteCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawIndirectByteCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectByteCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawIndirectByteCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawIndirectByteCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectByteCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawIndirectByteCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawIndirectByteCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectByteCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawIndirectByteCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawIndirectByteCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectByteCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawIndirectByteCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawIndirectByteCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMarkerToMemoryAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMarkerToMemoryAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMarkerToMemoryAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMarkerToMemoryAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMarkerToMemoryAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMarkerToMemoryAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMarkerToMemoryAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMarkerToMemoryAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMarkerToMemoryAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMarkerToMemoryAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMarkerToMemoryAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMarkerToMemoryAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMarkerToMemoryAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdWriteMarkerToMemoryAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdWriteMarkerToMemoryAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdWriteMarkerToMemoryAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMarkerToMemoryAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMarkerToMemoryAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMarkerToMemoryAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMarkerToMemoryAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMarkerToMemoryAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMarkerToMemoryAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMarkerToMemoryAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMarkerToMemoryAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMarkerToMemoryAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMarkerToMemoryAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMarkerToMemoryAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMarkerToMemoryAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMarkerToMemoryAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdWriteMarkerToMemoryAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdWriteMarkerToMemoryAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdWriteMarkerToMemoryAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMarkerToMemoryAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMarkerToMemoryAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMarkerToMemoryAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMarkerToMemoryAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMarkerToMemoryAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMarkerToMemoryAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMarkerToMemoryAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMarkerToMemoryAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMarkerToMemoryAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMarkerToMemoryAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMarkerToMemoryAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMarkerToMemoryAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMarkerToMemoryAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdWriteMarkerToMemoryAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdWriteMarkerToMemoryAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdWriteMarkerToMemoryAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMarkerToMemoryAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMarkerToMemoryAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMarkerToMemoryAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMarkerToMemoryAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMarkerToMemoryAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMarkerToMemoryAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMarkerToMemoryAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMarkerToMemoryAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMarkerToMemoryAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMarkerToMemoryAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMarkerToMemoryAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMarkerToMemoryAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMarkerToMemoryAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdWriteMarkerToMemoryAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdWriteMarkerToMemoryAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdWriteMarkerToMemoryAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMarkerToMemoryAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMarkerToMemoryAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMarkerToMemoryAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMarkerToMemoryAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMarkerToMemoryAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMarkerToMemoryAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMarkerToMemoryAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMarkerToMemoryAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMarkerToMemoryAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMarkerToMemoryAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMarkerToMemoryAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMarkerToMemoryAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMarkerToMemoryAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdWriteMarkerToMemoryAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdWriteMarkerToMemoryAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdWriteMarkerToMemoryAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMarkerToMemoryAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMarkerToMemoryAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMarkerToMemoryAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMarkerToMemoryAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMarkerToMemoryAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMarkerToMemoryAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMarkerToMemoryAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMarkerToMemoryAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMarkerToMemoryAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMarkerToMemoryAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMarkerToMemoryAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMarkerToMemoryAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMarkerToMemoryAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdWriteMarkerToMemoryAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdWriteMarkerToMemoryAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdWriteMarkerToMemoryAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMarkerToMemoryAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMarkerToMemoryAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMarkerToMemoryAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMarkerToMemoryAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMarkerToMemoryAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMarkerToMemoryAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMarkerToMemoryAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMarkerToMemoryAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMarkerToMemoryAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMarkerToMemoryAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMarkerToMemoryAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMarkerToMemoryAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMarkerToMemoryAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdWriteMarkerToMemoryAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdWriteMarkerToMemoryAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdWriteMarkerToMemoryAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMarkerToMemoryAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMarkerToMemoryAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMarkerToMemoryAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMarkerToMemoryAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMarkerToMemoryAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMarkerToMemoryAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMarkerToMemoryAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMarkerToMemoryAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMarkerToMemoryAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMarkerToMemoryAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMarkerToMemoryAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMarkerToMemoryAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMarkerToMemoryAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdWriteMarkerToMemoryAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdWriteMarkerToMemoryAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdWriteMarkerToMemoryAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMarkerToMemoryAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMarkerToMemoryAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMarkerToMemoryAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMarkerToMemoryAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMarkerToMemoryAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMarkerToMemoryAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMarkerToMemoryAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMarkerToMemoryAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMarkerToMemoryAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMarkerToMemoryAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMarkerToMemoryAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMarkerToMemoryAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMarkerToMemoryAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdWriteMarkerToMemoryAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdWriteMarkerToMemoryAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdWriteMarkerToMemoryAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMarkerToMemoryAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMarkerToMemoryAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMarkerToMemoryAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMarkerToMemoryAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMarkerToMemoryAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMarkerToMemoryAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMarkerToMemoryAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMarkerToMemoryAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMarkerToMemoryAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMarkerToMemoryAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMarkerToMemoryAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMarkerToMemoryAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMarkerToMemoryAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdWriteMarkerToMemoryAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdWriteMarkerToMemoryAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdWriteMarkerToMemoryAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMarkerToMemoryAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMarkerToMemoryAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMarkerToMemoryAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMarkerToMemoryAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMarkerToMemoryAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMarkerToMemoryAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMarkerToMemoryAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMarkerToMemoryAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMarkerToMemoryAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMarkerToMemoryAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMarkerToMemoryAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMarkerToMemoryAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMarkerToMemoryAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdWriteMarkerToMemoryAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdWriteMarkerToMemoryAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdWriteMarkerToMemoryAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMarkerToMemoryAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMarkerToMemoryAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMarkerToMemoryAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMarkerToMemoryAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMarkerToMemoryAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMarkerToMemoryAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMarkerToMemoryAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMarkerToMemoryAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMarkerToMemoryAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMarkerToMemoryAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMarkerToMemoryAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMarkerToMemoryAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMarkerToMemoryAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdWriteMarkerToMemoryAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdWriteMarkerToMemoryAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdWriteMarkerToMemoryAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMarkerToMemoryAMD@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMarkerToMemoryAMD@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMarkerToMemoryAMD@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMarkerToMemoryAMD@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMarkerToMemoryAMD@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMarkerToMemoryAMD@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMarkerToMemoryAMD@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMarkerToMemoryAMD@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMarkerToMemoryAMD@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMarkerToMemoryAMD@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMarkerToMemoryAMD@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMarkerToMemoryAMD@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMarkerToMemoryAMD@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdWriteMarkerToMemoryAMD@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdWriteMarkerToMemoryAMD=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdWriteMarkerToMemoryAMD=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindIndexBuffer3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindIndexBuffer3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindIndexBuffer3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindIndexBuffer3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindIndexBuffer3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindIndexBuffer3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindIndexBuffer3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindIndexBuffer3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindIndexBuffer3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindIndexBuffer3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindIndexBuffer3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindIndexBuffer3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindIndexBuffer3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindIndexBuffer3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindIndexBuffer3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindIndexBuffer3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindIndexBuffer3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindIndexBuffer3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindIndexBuffer3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindIndexBuffer3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindIndexBuffer3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindIndexBuffer3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindIndexBuffer3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindIndexBuffer3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindIndexBuffer3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindIndexBuffer3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindIndexBuffer3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindIndexBuffer3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindIndexBuffer3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindIndexBuffer3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindIndexBuffer3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindIndexBuffer3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindIndexBuffer3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindIndexBuffer3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindIndexBuffer3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindIndexBuffer3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindIndexBuffer3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindIndexBuffer3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindIndexBuffer3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdBindVertexBuffers3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdBindVertexBuffers3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdBindVertexBuffers3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdBindVertexBuffers3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdBindVertexBuffers3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdBindVertexBuffers3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdBindVertexBuffers3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdBindVertexBuffers3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdBindVertexBuffers3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdBindVertexBuffers3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdBindVertexBuffers3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdBindVertexBuffers3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdBindVertexBuffers3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdBindVertexBuffers3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdBindVertexBuffers3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdBindVertexBuffers3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdBindVertexBuffers3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdBindVertexBuffers3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdBindVertexBuffers3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdBindVertexBuffers3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdBindVertexBuffers3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdBindVertexBuffers3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdBindVertexBuffers3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdBindVertexBuffers3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdBindVertexBuffers3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdBindVertexBuffers3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdBindVertexBuffers3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdBindVertexBuffers3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdBindVertexBuffers3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdBindVertexBuffers3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdBindVertexBuffers3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdBindVertexBuffers3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdBindVertexBuffers3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdBindVertexBuffers3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdBindVertexBuffers3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdBindVertexBuffers3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers3KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers3KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers3KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers3KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers3KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers3KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers3KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers3KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers3KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers3KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers3KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers3KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers3KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdBindVertexBuffers3KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdBindVertexBuffers3KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdBindVertexBuffers3KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawIndexedIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawIndexedIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawIndexedIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawIndexedIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawIndexedIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawIndexedIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawIndexedIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawIndexedIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawIndexedIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawIndexedIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawIndexedIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawIndexedIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawIndexedIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawIndexedIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawIndexedIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawIndexedIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawIndexedIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawIndexedIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawIndexedIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawIndexedIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawIndexedIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawIndexedIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawIndexedIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawIndexedIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawIndexedIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawIndexedIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawIndexedIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawIndexedIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawIndexedIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawIndexedIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawIndexedIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawIndexedIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawIndexedIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawIndexedIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawIndexedIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawIndexedIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawIndexedIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawIndexedIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawIndexedIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawIndexedIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawIndexedIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawIndexedIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawIndexedIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawIndexedIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawIndexedIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawIndexedIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawIndexedIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawIndexedIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawIndexedIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawIndexedIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawIndexedIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawIndexedIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawIndexedIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawIndexedIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawIndexedIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawIndexedIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawIndexedIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawIndexedIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawIndexedIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawIndexedIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawIndexedIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawIndexedIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawIndexedIndirectCount2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawIndexedIndirectCount2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawIndexedIndirectCount2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirect2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirect2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirect2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirect2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirect2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirect2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirect2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirect2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirect2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirect2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirect2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirect2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirect2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirect2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawMeshTasksIndirect2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawMeshTasksIndirect2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirect2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirect2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirect2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirect2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirect2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirect2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirect2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirect2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirect2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirect2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirect2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirect2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirect2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirect2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawMeshTasksIndirect2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawMeshTasksIndirect2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirect2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirect2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirect2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirect2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirect2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirect2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirect2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirect2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirect2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirect2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirect2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirect2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirect2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirect2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawMeshTasksIndirect2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawMeshTasksIndirect2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirect2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirect2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirect2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirect2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirect2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirect2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirect2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirect2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirect2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirect2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirect2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirect2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirect2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirect2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawMeshTasksIndirect2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawMeshTasksIndirect2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirect2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirect2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirect2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirect2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirect2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirect2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirect2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirect2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirect2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirect2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirect2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirect2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirect2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirect2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawMeshTasksIndirect2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawMeshTasksIndirect2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirect2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirect2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirect2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirect2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirect2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirect2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirect2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirect2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirect2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirect2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirect2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirect2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirect2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirect2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawMeshTasksIndirect2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawMeshTasksIndirect2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirect2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirect2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirect2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirect2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirect2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirect2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirect2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirect2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirect2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirect2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirect2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirect2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirect2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirect2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawMeshTasksIndirect2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawMeshTasksIndirect2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirect2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirect2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirect2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirect2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirect2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirect2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirect2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirect2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirect2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirect2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirect2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirect2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirect2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirect2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawMeshTasksIndirect2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawMeshTasksIndirect2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirect2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirect2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirect2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirect2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirect2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirect2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirect2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirect2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirect2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirect2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirect2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirect2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirect2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirect2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawMeshTasksIndirect2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawMeshTasksIndirect2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirect2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirect2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirect2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirect2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirect2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirect2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirect2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirect2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirect2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirect2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirect2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirect2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirect2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirect2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawMeshTasksIndirect2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawMeshTasksIndirect2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirect2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirect2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirect2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirect2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirect2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirect2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirect2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirect2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirect2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirect2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirect2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirect2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirect2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirect2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawMeshTasksIndirect2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawMeshTasksIndirect2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirect2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirect2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirect2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirect2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirect2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirect2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirect2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirect2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirect2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirect2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirect2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirect2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirect2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirect2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawMeshTasksIndirect2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawMeshTasksIndirect2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirect2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirect2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirect2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirect2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirect2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirect2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirect2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirect2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirect2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirect2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirect2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirect2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirect2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirect2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawMeshTasksIndirect2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawMeshTasksIndirect2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDrawMeshTasksIndirectCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDrawMeshTasksIndirectCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDrawMeshTasksIndirectCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDrawMeshTasksIndirectCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDrawMeshTasksIndirectCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDrawMeshTasksIndirectCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDrawMeshTasksIndirectCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDrawMeshTasksIndirectCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDrawMeshTasksIndirectCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDrawMeshTasksIndirectCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDrawMeshTasksIndirectCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDrawMeshTasksIndirectCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDrawMeshTasksIndirectCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDrawMeshTasksIndirectCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDrawMeshTasksIndirectCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDrawMeshTasksIndirectCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDrawMeshTasksIndirectCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDrawMeshTasksIndirectCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDrawMeshTasksIndirectCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDrawMeshTasksIndirectCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDrawMeshTasksIndirectCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDrawMeshTasksIndirectCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDrawMeshTasksIndirectCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDrawMeshTasksIndirectCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDrawMeshTasksIndirectCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDrawMeshTasksIndirectCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDrawMeshTasksIndirectCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDrawMeshTasksIndirectCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDrawMeshTasksIndirectCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDrawMeshTasksIndirectCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDrawMeshTasksIndirectCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDrawMeshTasksIndirectCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDrawMeshTasksIndirectCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDrawMeshTasksIndirectCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDrawMeshTasksIndirectCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDrawMeshTasksIndirectCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCount2EXT@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCount2EXT@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCount2EXT@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCount2EXT@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCount2EXT@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCount2EXT@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCount2EXT@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCount2EXT@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCount2EXT@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCount2EXT@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCount2EXT@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCount2EXT@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCount2EXT@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDrawMeshTasksIndirectCount2EXT@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDrawMeshTasksIndirectCount2EXT=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDrawMeshTasksIndirectCount2EXT=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdDispatchIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdDispatchIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdDispatchIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdDispatchIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdDispatchIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdDispatchIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdDispatchIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdDispatchIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdDispatchIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdDispatchIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdDispatchIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdDispatchIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdDispatchIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdDispatchIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdDispatchIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdDispatchIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdDispatchIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdDispatchIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdDispatchIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdDispatchIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdDispatchIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdDispatchIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdDispatchIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdDispatchIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdDispatchIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdDispatchIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdDispatchIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdDispatchIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdDispatchIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdDispatchIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdDispatchIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdDispatchIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdDispatchIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdDispatchIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdDispatchIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdDispatchIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdDispatchIndirect2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdDispatchIndirect2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdDispatchIndirect2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructure2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructure2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructure2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructure2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructure2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructure2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructure2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructure2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructure2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructure2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructure2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructure2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructure2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CreateAccelerationStructure2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CreateAccelerationStructure2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CreateAccelerationStructure2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructure2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructure2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructure2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructure2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructure2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructure2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructure2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructure2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructure2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructure2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructure2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructure2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructure2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CreateAccelerationStructure2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CreateAccelerationStructure2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CreateAccelerationStructure2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructure2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructure2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructure2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructure2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructure2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructure2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructure2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructure2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructure2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructure2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructure2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructure2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructure2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CreateAccelerationStructure2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CreateAccelerationStructure2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CreateAccelerationStructure2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructure2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructure2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructure2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructure2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructure2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructure2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructure2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructure2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructure2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructure2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructure2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructure2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructure2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CreateAccelerationStructure2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CreateAccelerationStructure2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CreateAccelerationStructure2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructure2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructure2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructure2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructure2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructure2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructure2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructure2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructure2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructure2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructure2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructure2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructure2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructure2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CreateAccelerationStructure2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CreateAccelerationStructure2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CreateAccelerationStructure2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructure2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructure2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructure2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructure2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructure2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructure2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructure2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructure2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructure2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructure2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructure2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructure2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructure2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CreateAccelerationStructure2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CreateAccelerationStructure2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CreateAccelerationStructure2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructure2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructure2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructure2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructure2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructure2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructure2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructure2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructure2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructure2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructure2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructure2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructure2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructure2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CreateAccelerationStructure2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CreateAccelerationStructure2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CreateAccelerationStructure2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructure2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructure2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructure2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructure2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructure2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructure2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructure2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructure2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructure2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructure2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructure2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructure2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructure2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CreateAccelerationStructure2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CreateAccelerationStructure2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CreateAccelerationStructure2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructure2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructure2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructure2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructure2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructure2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructure2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructure2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructure2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructure2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructure2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructure2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructure2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructure2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CreateAccelerationStructure2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CreateAccelerationStructure2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CreateAccelerationStructure2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructure2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructure2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructure2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructure2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructure2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructure2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructure2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructure2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructure2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructure2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructure2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructure2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructure2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CreateAccelerationStructure2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CreateAccelerationStructure2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CreateAccelerationStructure2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructure2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructure2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructure2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructure2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructure2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructure2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructure2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructure2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructure2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructure2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructure2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructure2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructure2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CreateAccelerationStructure2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CreateAccelerationStructure2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CreateAccelerationStructure2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructure2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructure2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructure2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructure2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructure2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructure2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructure2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructure2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructure2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructure2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructure2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructure2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructure2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CreateAccelerationStructure2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CreateAccelerationStructure2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CreateAccelerationStructure2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructure2KHR@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructure2KHR@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructure2KHR@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructure2KHR@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructure2KHR@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructure2KHR@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructure2KHR@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructure2KHR@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructure2KHR@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructure2KHR@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructure2KHR@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructure2KHR@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructure2KHR@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CreateAccelerationStructure2KHR@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CreateAccelerationStructure2KHR=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CreateAccelerationStructure2KHR=#vk_entrypoint_stub")
#endif
#endif
#else
#endif

#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_CmdSetDispatchParametersARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDispatchParametersARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDispatchParametersARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDispatchParametersARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDispatchParametersARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDispatchParametersARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDispatchParametersARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDispatchParametersARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDispatchParametersARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDispatchParametersARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDispatchParametersARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDispatchParametersARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDispatchParametersARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_CmdSetDispatchParametersARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_CmdSetDispatchParametersARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_CmdSetDispatchParametersARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDispatchParametersARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDispatchParametersARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDispatchParametersARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDispatchParametersARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDispatchParametersARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDispatchParametersARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDispatchParametersARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDispatchParametersARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDispatchParametersARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDispatchParametersARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDispatchParametersARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDispatchParametersARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDispatchParametersARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx9_CmdSetDispatchParametersARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx9_CmdSetDispatchParametersARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx9_CmdSetDispatchParametersARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDispatchParametersARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDispatchParametersARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDispatchParametersARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDispatchParametersARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDispatchParametersARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDispatchParametersARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDispatchParametersARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDispatchParametersARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDispatchParametersARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDispatchParametersARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDispatchParametersARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDispatchParametersARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDispatchParametersARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx11_CmdSetDispatchParametersARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx11_CmdSetDispatchParametersARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx11_CmdSetDispatchParametersARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDispatchParametersARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDispatchParametersARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDispatchParametersARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDispatchParametersARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDispatchParametersARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDispatchParametersARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDispatchParametersARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDispatchParametersARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDispatchParametersARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDispatchParametersARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDispatchParametersARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDispatchParametersARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDispatchParametersARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx12_CmdSetDispatchParametersARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx12_CmdSetDispatchParametersARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx12_CmdSetDispatchParametersARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDispatchParametersARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDispatchParametersARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDispatchParametersARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDispatchParametersARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDispatchParametersARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDispatchParametersARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDispatchParametersARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDispatchParametersARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDispatchParametersARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDispatchParametersARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDispatchParametersARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDispatchParametersARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDispatchParametersARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx125_CmdSetDispatchParametersARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx125_CmdSetDispatchParametersARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx125_CmdSetDispatchParametersARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDispatchParametersARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDispatchParametersARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDispatchParametersARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDispatchParametersARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDispatchParametersARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDispatchParametersARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDispatchParametersARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDispatchParametersARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDispatchParametersARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDispatchParametersARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDispatchParametersARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDispatchParametersARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDispatchParametersARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx20_CmdSetDispatchParametersARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx20_CmdSetDispatchParametersARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx20_CmdSetDispatchParametersARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDispatchParametersARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDispatchParametersARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDispatchParametersARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDispatchParametersARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDispatchParametersARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDispatchParametersARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDispatchParametersARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDispatchParametersARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDispatchParametersARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDispatchParametersARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDispatchParametersARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDispatchParametersARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDispatchParametersARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx30_CmdSetDispatchParametersARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx30_CmdSetDispatchParametersARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx30_CmdSetDispatchParametersARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDispatchParametersARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDispatchParametersARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDispatchParametersARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDispatchParametersARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDispatchParametersARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDispatchParametersARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDispatchParametersARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDispatchParametersARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDispatchParametersARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDispatchParametersARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDispatchParametersARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDispatchParametersARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDispatchParametersARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_gfx35_CmdSetDispatchParametersARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:gfx35_CmdSetDispatchParametersARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#gfx35_CmdSetDispatchParametersARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDispatchParametersARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDispatchParametersARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDispatchParametersARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDispatchParametersARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDispatchParametersARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDispatchParametersARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDispatchParametersARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDispatchParametersARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDispatchParametersARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDispatchParametersARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDispatchParametersARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDispatchParametersARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDispatchParametersARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_doom64_CmdSetDispatchParametersARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_doom64_CmdSetDispatchParametersARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_doom64_CmdSetDispatchParametersARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDispatchParametersARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDispatchParametersARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDispatchParametersARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDispatchParametersARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDispatchParametersARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDispatchParametersARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDispatchParametersARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDispatchParametersARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDispatchParametersARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDispatchParametersARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDispatchParametersARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDispatchParametersARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDispatchParametersARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_furmark_CmdSetDispatchParametersARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_furmark_CmdSetDispatchParametersARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_furmark_CmdSetDispatchParametersARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDispatchParametersARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDispatchParametersARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDispatchParametersARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDispatchParametersARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDispatchParametersARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDispatchParametersARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDispatchParametersARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDispatchParametersARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDispatchParametersARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDispatchParametersARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDispatchParametersARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDispatchParametersARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDispatchParametersARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_hitman3_CmdSetDispatchParametersARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_hitman3_CmdSetDispatchParametersARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_hitman3_CmdSetDispatchParametersARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDispatchParametersARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDispatchParametersARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDispatchParametersARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDispatchParametersARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDispatchParametersARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDispatchParametersARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDispatchParametersARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDispatchParametersARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDispatchParametersARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDispatchParametersARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDispatchParametersARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDispatchParametersARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDispatchParametersARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_android_CmdSetDispatchParametersARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_android_CmdSetDispatchParametersARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_android_CmdSetDispatchParametersARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif
#ifdef _MSC_VER
#ifdef _M_IX86
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDispatchParametersARM@4=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDispatchParametersARM@8=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDispatchParametersARM@12=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDispatchParametersARM@16=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDispatchParametersARM@20=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDispatchParametersARM@24=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDispatchParametersARM@28=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDispatchParametersARM@32=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDispatchParametersARM@36=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDispatchParametersARM@40=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDispatchParametersARM@44=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDispatchParametersARM@48=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDispatchParametersARM@60=_vk_entrypoint_stub@0")
    #pragma comment(linker, "/alternatename:_anv_rmv_CmdSetDispatchParametersARM@104=_vk_entrypoint_stub@0")
#else
    #pragma comment(linker, "/alternatename:anv_rmv_CmdSetDispatchParametersARM=vk_entrypoint_stub")
#if defined(_M_ARM64EC)
    #pragma comment(linker, "/alternatename:#anv_rmv_CmdSetDispatchParametersARM=#vk_entrypoint_stub")
#endif
#endif
#else
#endif


const struct vk_device_entrypoint_table anv_device_entrypoints = {
    .GetDeviceProcAddr = anv_GetDeviceProcAddr,
    .DestroyDevice = anv_DestroyDevice,
    .GetDeviceQueue = anv_GetDeviceQueue,
    .QueueSubmit = anv_QueueSubmit,
    .QueueWaitIdle = anv_QueueWaitIdle,
    .DeviceWaitIdle = anv_DeviceWaitIdle,
    .AllocateMemory = anv_AllocateMemory,
    .FreeMemory = anv_FreeMemory,
    .MapMemory = anv_MapMemory,
    .UnmapMemory = anv_UnmapMemory,
    .FlushMappedMemoryRanges = anv_FlushMappedMemoryRanges,
    .InvalidateMappedMemoryRanges = anv_InvalidateMappedMemoryRanges,
    .GetDeviceMemoryCommitment = anv_GetDeviceMemoryCommitment,
    .GetBufferMemoryRequirements = anv_GetBufferMemoryRequirements,
    .BindBufferMemory = anv_BindBufferMemory,
    .GetImageMemoryRequirements = anv_GetImageMemoryRequirements,
    .BindImageMemory = anv_BindImageMemory,
    .GetImageSparseMemoryRequirements = anv_GetImageSparseMemoryRequirements,
    .QueueBindSparse = anv_QueueBindSparse,
    .CreateFence = anv_CreateFence,
    .DestroyFence = anv_DestroyFence,
    .ResetFences = anv_ResetFences,
    .GetFenceStatus = anv_GetFenceStatus,
    .WaitForFences = anv_WaitForFences,
    .CreateSemaphore = anv_CreateSemaphore,
    .DestroySemaphore = anv_DestroySemaphore,
    .CreateEvent = anv_CreateEvent,
    .DestroyEvent = anv_DestroyEvent,
    .GetEventStatus = anv_GetEventStatus,
    .SetEvent = anv_SetEvent,
    .ResetEvent = anv_ResetEvent,
    .CreateQueryPool = anv_CreateQueryPool,
    .DestroyQueryPool = anv_DestroyQueryPool,
    .GetQueryPoolResults = anv_GetQueryPoolResults,
    .ResetQueryPool = anv_ResetQueryPool,
    .ResetQueryPoolEXT = anv_ResetQueryPoolEXT,
    .CreateBuffer = anv_CreateBuffer,
    .DestroyBuffer = anv_DestroyBuffer,
    .CreateBufferView = anv_CreateBufferView,
    .DestroyBufferView = anv_DestroyBufferView,
    .CreateImage = anv_CreateImage,
    .DestroyImage = anv_DestroyImage,
    .GetImageSubresourceLayout = anv_GetImageSubresourceLayout,
    .CreateImageView = anv_CreateImageView,
    .DestroyImageView = anv_DestroyImageView,
    .CreateShaderModule = anv_CreateShaderModule,
    .DestroyShaderModule = anv_DestroyShaderModule,
    .CreatePipelineCache = anv_CreatePipelineCache,
    .DestroyPipelineCache = anv_DestroyPipelineCache,
    .GetPipelineCacheData = anv_GetPipelineCacheData,
    .MergePipelineCaches = anv_MergePipelineCaches,
    .CreatePipelineBinariesKHR = anv_CreatePipelineBinariesKHR,
    .DestroyPipelineBinaryKHR = anv_DestroyPipelineBinaryKHR,
    .GetPipelineKeyKHR = anv_GetPipelineKeyKHR,
    .GetPipelineBinaryDataKHR = anv_GetPipelineBinaryDataKHR,
    .ReleaseCapturedPipelineDataKHR = anv_ReleaseCapturedPipelineDataKHR,
    .CreateGraphicsPipelines = anv_CreateGraphicsPipelines,
    .CreateComputePipelines = anv_CreateComputePipelines,
    .GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI = anv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI,
    .DestroyPipeline = anv_DestroyPipeline,
    .CreatePipelineLayout = anv_CreatePipelineLayout,
    .DestroyPipelineLayout = anv_DestroyPipelineLayout,
    .CreateSampler = anv_CreateSampler,
    .DestroySampler = anv_DestroySampler,
    .CreateDescriptorSetLayout = anv_CreateDescriptorSetLayout,
    .DestroyDescriptorSetLayout = anv_DestroyDescriptorSetLayout,
    .CreateDescriptorPool = anv_CreateDescriptorPool,
    .DestroyDescriptorPool = anv_DestroyDescriptorPool,
    .ResetDescriptorPool = anv_ResetDescriptorPool,
    .AllocateDescriptorSets = anv_AllocateDescriptorSets,
    .FreeDescriptorSets = anv_FreeDescriptorSets,
    .UpdateDescriptorSets = anv_UpdateDescriptorSets,
    .CreateFramebuffer = anv_CreateFramebuffer,
    .DestroyFramebuffer = anv_DestroyFramebuffer,
    .CreateRenderPass = anv_CreateRenderPass,
    .DestroyRenderPass = anv_DestroyRenderPass,
    .GetRenderAreaGranularity = anv_GetRenderAreaGranularity,
    .GetRenderingAreaGranularity = anv_GetRenderingAreaGranularity,
    .GetRenderingAreaGranularityKHR = anv_GetRenderingAreaGranularityKHR,
    .CreateCommandPool = anv_CreateCommandPool,
    .DestroyCommandPool = anv_DestroyCommandPool,
    .ResetCommandPool = anv_ResetCommandPool,
    .AllocateCommandBuffers = anv_AllocateCommandBuffers,
    .FreeCommandBuffers = anv_FreeCommandBuffers,
    .BeginCommandBuffer = anv_BeginCommandBuffer,
    .EndCommandBuffer = anv_EndCommandBuffer,
    .ResetCommandBuffer = anv_ResetCommandBuffer,
    .CmdBindPipeline = anv_CmdBindPipeline,
    .CmdSetPrimitiveRestartIndexEXT = anv_CmdSetPrimitiveRestartIndexEXT,
    .CmdSetAttachmentFeedbackLoopEnableEXT = anv_CmdSetAttachmentFeedbackLoopEnableEXT,
    .CmdSetViewport = anv_CmdSetViewport,
    .CmdSetScissor = anv_CmdSetScissor,
    .CmdSetLineWidth = anv_CmdSetLineWidth,
    .CmdSetDepthBias = anv_CmdSetDepthBias,
    .CmdSetBlendConstants = anv_CmdSetBlendConstants,
    .CmdSetDepthBounds = anv_CmdSetDepthBounds,
    .CmdSetStencilCompareMask = anv_CmdSetStencilCompareMask,
    .CmdSetStencilWriteMask = anv_CmdSetStencilWriteMask,
    .CmdSetStencilReference = anv_CmdSetStencilReference,
    .CmdBindDescriptorSets = anv_CmdBindDescriptorSets,
    .CmdBindIndexBuffer = anv_CmdBindIndexBuffer,
    .CmdBindVertexBuffers = anv_CmdBindVertexBuffers,
    .CmdDraw = anv_CmdDraw,
    .CmdDrawIndexed = anv_CmdDrawIndexed,
    .CmdDrawMultiEXT = anv_CmdDrawMultiEXT,
    .CmdDrawMultiIndexedEXT = anv_CmdDrawMultiIndexedEXT,
    .CmdDrawIndirect = anv_CmdDrawIndirect,
    .CmdDrawIndexedIndirect = anv_CmdDrawIndexedIndirect,
    .CmdDispatch = anv_CmdDispatch,
    .CmdDispatchIndirect = anv_CmdDispatchIndirect,
    .CmdSubpassShadingHUAWEI = anv_CmdSubpassShadingHUAWEI,
    .CmdDrawClusterHUAWEI = anv_CmdDrawClusterHUAWEI,
    .CmdDrawClusterIndirectHUAWEI = anv_CmdDrawClusterIndirectHUAWEI,
    .CmdUpdatePipelineIndirectBufferNV = anv_CmdUpdatePipelineIndirectBufferNV,
    .CmdCopyBuffer = anv_CmdCopyBuffer,
    .CmdCopyImage = anv_CmdCopyImage,
    .CmdBlitImage = anv_CmdBlitImage,
    .CmdCopyBufferToImage = anv_CmdCopyBufferToImage,
    .CmdCopyImageToBuffer = anv_CmdCopyImageToBuffer,
    .CmdCopyMemoryIndirectNV = anv_CmdCopyMemoryIndirectNV,
    .CmdCopyMemoryIndirectKHR = anv_CmdCopyMemoryIndirectKHR,
    .CmdCopyMemoryToImageIndirectNV = anv_CmdCopyMemoryToImageIndirectNV,
    .CmdCopyMemoryToImageIndirectKHR = anv_CmdCopyMemoryToImageIndirectKHR,
    .CmdUpdateBuffer = anv_CmdUpdateBuffer,
    .CmdFillBuffer = anv_CmdFillBuffer,
    .CmdClearColorImage = anv_CmdClearColorImage,
    .CmdClearDepthStencilImage = anv_CmdClearDepthStencilImage,
    .CmdClearAttachments = anv_CmdClearAttachments,
    .CmdResolveImage = anv_CmdResolveImage,
    .CmdSetEvent = anv_CmdSetEvent,
    .CmdResetEvent = anv_CmdResetEvent,
    .CmdWaitEvents = anv_CmdWaitEvents,
    .CmdPipelineBarrier = anv_CmdPipelineBarrier,
    .CmdBeginQuery = anv_CmdBeginQuery,
    .CmdEndQuery = anv_CmdEndQuery,
    .CmdBeginConditionalRenderingEXT = anv_CmdBeginConditionalRenderingEXT,
    .CmdEndConditionalRenderingEXT = anv_CmdEndConditionalRenderingEXT,
    .CmdBeginCustomResolveEXT = anv_CmdBeginCustomResolveEXT,
    .CmdResetQueryPool = anv_CmdResetQueryPool,
    .CmdWriteTimestamp = anv_CmdWriteTimestamp,
    .CmdCopyQueryPoolResults = anv_CmdCopyQueryPoolResults,
    .CmdPushConstants = anv_CmdPushConstants,
    .CmdBeginRenderPass = anv_CmdBeginRenderPass,
    .CmdNextSubpass = anv_CmdNextSubpass,
    .CmdEndRenderPass = anv_CmdEndRenderPass,
    .CmdExecuteCommands = anv_CmdExecuteCommands,
    .CreateSharedSwapchainsKHR = anv_CreateSharedSwapchainsKHR,
    .CreateSwapchainKHR = anv_CreateSwapchainKHR,
    .DestroySwapchainKHR = anv_DestroySwapchainKHR,
    .GetSwapchainImagesKHR = anv_GetSwapchainImagesKHR,
    .AcquireNextImageKHR = anv_AcquireNextImageKHR,
    .QueuePresentKHR = anv_QueuePresentKHR,
    .DebugMarkerSetObjectNameEXT = anv_DebugMarkerSetObjectNameEXT,
    .DebugMarkerSetObjectTagEXT = anv_DebugMarkerSetObjectTagEXT,
    .CmdDebugMarkerBeginEXT = anv_CmdDebugMarkerBeginEXT,
    .CmdDebugMarkerEndEXT = anv_CmdDebugMarkerEndEXT,
    .CmdDebugMarkerInsertEXT = anv_CmdDebugMarkerInsertEXT,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleNV = anv_GetMemoryWin32HandleNV,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .CmdExecuteGeneratedCommandsNV = anv_CmdExecuteGeneratedCommandsNV,
    .CmdPreprocessGeneratedCommandsNV = anv_CmdPreprocessGeneratedCommandsNV,
    .CmdBindPipelineShaderGroupNV = anv_CmdBindPipelineShaderGroupNV,
    .GetGeneratedCommandsMemoryRequirementsNV = anv_GetGeneratedCommandsMemoryRequirementsNV,
    .CreateIndirectCommandsLayoutNV = anv_CreateIndirectCommandsLayoutNV,
    .DestroyIndirectCommandsLayoutNV = anv_DestroyIndirectCommandsLayoutNV,
    .CmdExecuteGeneratedCommandsEXT = anv_CmdExecuteGeneratedCommandsEXT,
    .CmdPreprocessGeneratedCommandsEXT = anv_CmdPreprocessGeneratedCommandsEXT,
    .GetGeneratedCommandsMemoryRequirementsEXT = anv_GetGeneratedCommandsMemoryRequirementsEXT,
    .CreateIndirectCommandsLayoutEXT = anv_CreateIndirectCommandsLayoutEXT,
    .DestroyIndirectCommandsLayoutEXT = anv_DestroyIndirectCommandsLayoutEXT,
    .CreateIndirectExecutionSetEXT = anv_CreateIndirectExecutionSetEXT,
    .DestroyIndirectExecutionSetEXT = anv_DestroyIndirectExecutionSetEXT,
    .UpdateIndirectExecutionSetPipelineEXT = anv_UpdateIndirectExecutionSetPipelineEXT,
    .UpdateIndirectExecutionSetShaderEXT = anv_UpdateIndirectExecutionSetShaderEXT,
    .CmdPushDescriptorSet = anv_CmdPushDescriptorSet,
    .CmdPushDescriptorSetKHR = anv_CmdPushDescriptorSetKHR,
    .TrimCommandPool = anv_TrimCommandPool,
    .TrimCommandPoolKHR = anv_TrimCommandPoolKHR,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleKHR = anv_GetMemoryWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandlePropertiesKHR = anv_GetMemoryWin32HandlePropertiesKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandlePropertiesKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryFdKHR = anv_GetMemoryFdKHR,
    .GetMemoryFdPropertiesKHR = anv_GetMemoryFdPropertiesKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandleFUCHSIA = anv_GetMemoryZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandlePropertiesFUCHSIA = anv_GetMemoryZirconHandlePropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandlePropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .GetMemoryRemoteAddressNV = anv_GetMemoryRemoteAddressNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreWin32HandleKHR = anv_GetSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportSemaphoreWin32HandleKHR = anv_ImportSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreFdKHR = anv_GetSemaphoreFdKHR,
    .ImportSemaphoreFdKHR = anv_ImportSemaphoreFdKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetSemaphoreZirconHandleFUCHSIA = anv_GetSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .ImportSemaphoreZirconHandleFUCHSIA = anv_ImportSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .ImportSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetFenceWin32HandleKHR = anv_GetFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportFenceWin32HandleKHR = anv_ImportFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetFenceFdKHR = anv_GetFenceFdKHR,
    .ImportFenceFdKHR = anv_ImportFenceFdKHR,
    .DisplayPowerControlEXT = anv_DisplayPowerControlEXT,
    .RegisterDeviceEventEXT = anv_RegisterDeviceEventEXT,
    .RegisterDisplayEventEXT = anv_RegisterDisplayEventEXT,
    .GetSwapchainCounterEXT = anv_GetSwapchainCounterEXT,
    .GetDeviceGroupPeerMemoryFeatures = anv_GetDeviceGroupPeerMemoryFeatures,
    .GetDeviceGroupPeerMemoryFeaturesKHR = anv_GetDeviceGroupPeerMemoryFeaturesKHR,
    .BindBufferMemory2 = anv_BindBufferMemory2,
    .BindBufferMemory2KHR = anv_BindBufferMemory2KHR,
    .BindImageMemory2 = anv_BindImageMemory2,
    .BindImageMemory2KHR = anv_BindImageMemory2KHR,
    .CmdSetDeviceMask = anv_CmdSetDeviceMask,
    .CmdSetDeviceMaskKHR = anv_CmdSetDeviceMaskKHR,
    .GetDeviceGroupPresentCapabilitiesKHR = anv_GetDeviceGroupPresentCapabilitiesKHR,
    .GetDeviceGroupSurfacePresentModesKHR = anv_GetDeviceGroupSurfacePresentModesKHR,
    .AcquireNextImage2KHR = anv_AcquireNextImage2KHR,
    .CmdDispatchBase = anv_CmdDispatchBase,
    .CmdDispatchBaseKHR = anv_CmdDispatchBaseKHR,
    .CreateDescriptorUpdateTemplate = anv_CreateDescriptorUpdateTemplate,
    .CreateDescriptorUpdateTemplateKHR = anv_CreateDescriptorUpdateTemplateKHR,
    .DestroyDescriptorUpdateTemplate = anv_DestroyDescriptorUpdateTemplate,
    .DestroyDescriptorUpdateTemplateKHR = anv_DestroyDescriptorUpdateTemplateKHR,
    .UpdateDescriptorSetWithTemplate = anv_UpdateDescriptorSetWithTemplate,
    .UpdateDescriptorSetWithTemplateKHR = anv_UpdateDescriptorSetWithTemplateKHR,
    .CmdPushDescriptorSetWithTemplate = anv_CmdPushDescriptorSetWithTemplate,
    .CmdPushDescriptorSetWithTemplateKHR = anv_CmdPushDescriptorSetWithTemplateKHR,
    .SetHdrMetadataEXT = anv_SetHdrMetadataEXT,
    .GetSwapchainStatusKHR = anv_GetSwapchainStatusKHR,
    .GetRefreshCycleDurationGOOGLE = anv_GetRefreshCycleDurationGOOGLE,
    .GetPastPresentationTimingGOOGLE = anv_GetPastPresentationTimingGOOGLE,
    .CmdSetViewportWScalingNV = anv_CmdSetViewportWScalingNV,
    .CmdSetDiscardRectangleEXT = anv_CmdSetDiscardRectangleEXT,
    .CmdSetDiscardRectangleEnableEXT = anv_CmdSetDiscardRectangleEnableEXT,
    .CmdSetDiscardRectangleModeEXT = anv_CmdSetDiscardRectangleModeEXT,
    .CmdSetSampleLocationsEXT = anv_CmdSetSampleLocationsEXT,
    .GetBufferMemoryRequirements2 = anv_GetBufferMemoryRequirements2,
    .GetBufferMemoryRequirements2KHR = anv_GetBufferMemoryRequirements2KHR,
    .GetImageMemoryRequirements2 = anv_GetImageMemoryRequirements2,
    .GetImageMemoryRequirements2KHR = anv_GetImageMemoryRequirements2KHR,
    .GetImageSparseMemoryRequirements2 = anv_GetImageSparseMemoryRequirements2,
    .GetImageSparseMemoryRequirements2KHR = anv_GetImageSparseMemoryRequirements2KHR,
    .GetDeviceBufferMemoryRequirements = anv_GetDeviceBufferMemoryRequirements,
    .GetDeviceBufferMemoryRequirementsKHR = anv_GetDeviceBufferMemoryRequirementsKHR,
    .GetDeviceImageMemoryRequirements = anv_GetDeviceImageMemoryRequirements,
    .GetDeviceImageMemoryRequirementsKHR = anv_GetDeviceImageMemoryRequirementsKHR,
    .GetDeviceImageSparseMemoryRequirements = anv_GetDeviceImageSparseMemoryRequirements,
    .GetDeviceImageSparseMemoryRequirementsKHR = anv_GetDeviceImageSparseMemoryRequirementsKHR,
    .CreateSamplerYcbcrConversion = anv_CreateSamplerYcbcrConversion,
    .CreateSamplerYcbcrConversionKHR = anv_CreateSamplerYcbcrConversionKHR,
    .DestroySamplerYcbcrConversion = anv_DestroySamplerYcbcrConversion,
    .DestroySamplerYcbcrConversionKHR = anv_DestroySamplerYcbcrConversionKHR,
    .GetDeviceQueue2 = anv_GetDeviceQueue2,
    .CreateValidationCacheEXT = anv_CreateValidationCacheEXT,
    .DestroyValidationCacheEXT = anv_DestroyValidationCacheEXT,
    .GetValidationCacheDataEXT = anv_GetValidationCacheDataEXT,
    .MergeValidationCachesEXT = anv_MergeValidationCachesEXT,
    .GetDescriptorSetLayoutSupport = anv_GetDescriptorSetLayoutSupport,
    .GetDescriptorSetLayoutSupportKHR = anv_GetDescriptorSetLayoutSupportKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsageANDROID = anv_GetSwapchainGrallocUsageANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsage2ANDROID = anv_GetSwapchainGrallocUsage2ANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsage2ANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .AcquireImageANDROID = anv_AcquireImageANDROID,
#elif defined(_MSC_VER)
    .AcquireImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .QueueSignalReleaseImageANDROID = anv_QueueSignalReleaseImageANDROID,
#elif defined(_MSC_VER)
    .QueueSignalReleaseImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .GetShaderInfoAMD = anv_GetShaderInfoAMD,
    .SetLocalDimmingAMD = anv_SetLocalDimmingAMD,
    .GetCalibratedTimestampsKHR = anv_GetCalibratedTimestampsKHR,
    .GetCalibratedTimestampsEXT = anv_GetCalibratedTimestampsEXT,
    .SetDebugUtilsObjectNameEXT = anv_SetDebugUtilsObjectNameEXT,
    .SetDebugUtilsObjectTagEXT = anv_SetDebugUtilsObjectTagEXT,
    .QueueBeginDebugUtilsLabelEXT = anv_QueueBeginDebugUtilsLabelEXT,
    .QueueEndDebugUtilsLabelEXT = anv_QueueEndDebugUtilsLabelEXT,
    .QueueInsertDebugUtilsLabelEXT = anv_QueueInsertDebugUtilsLabelEXT,
    .CmdBeginDebugUtilsLabelEXT = anv_CmdBeginDebugUtilsLabelEXT,
    .CmdEndDebugUtilsLabelEXT = anv_CmdEndDebugUtilsLabelEXT,
    .CmdInsertDebugUtilsLabelEXT = anv_CmdInsertDebugUtilsLabelEXT,
    .GetMemoryHostPointerPropertiesEXT = anv_GetMemoryHostPointerPropertiesEXT,
    .CmdWriteBufferMarkerAMD = anv_CmdWriteBufferMarkerAMD,
    .CreateRenderPass2 = anv_CreateRenderPass2,
    .CreateRenderPass2KHR = anv_CreateRenderPass2KHR,
    .CmdBeginRenderPass2 = anv_CmdBeginRenderPass2,
    .CmdBeginRenderPass2KHR = anv_CmdBeginRenderPass2KHR,
    .CmdNextSubpass2 = anv_CmdNextSubpass2,
    .CmdNextSubpass2KHR = anv_CmdNextSubpass2KHR,
    .CmdEndRenderPass2 = anv_CmdEndRenderPass2,
    .CmdEndRenderPass2KHR = anv_CmdEndRenderPass2KHR,
    .GetSemaphoreCounterValue = anv_GetSemaphoreCounterValue,
    .GetSemaphoreCounterValueKHR = anv_GetSemaphoreCounterValueKHR,
    .WaitSemaphores = anv_WaitSemaphores,
    .WaitSemaphoresKHR = anv_WaitSemaphoresKHR,
    .SignalSemaphore = anv_SignalSemaphore,
    .SignalSemaphoreKHR = anv_SignalSemaphoreKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetAndroidHardwareBufferPropertiesANDROID = anv_GetAndroidHardwareBufferPropertiesANDROID,
#elif defined(_MSC_VER)
    .GetAndroidHardwareBufferPropertiesANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetMemoryAndroidHardwareBufferANDROID = anv_GetMemoryAndroidHardwareBufferANDROID,
#elif defined(_MSC_VER)
    .GetMemoryAndroidHardwareBufferANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .CmdDrawIndirectCount = anv_CmdDrawIndirectCount,
    .CmdDrawIndirectCountKHR = anv_CmdDrawIndirectCountKHR,
    .CmdDrawIndirectCountAMD = anv_CmdDrawIndirectCountAMD,
    .CmdDrawIndexedIndirectCount = anv_CmdDrawIndexedIndirectCount,
    .CmdDrawIndexedIndirectCountKHR = anv_CmdDrawIndexedIndirectCountKHR,
    .CmdDrawIndexedIndirectCountAMD = anv_CmdDrawIndexedIndirectCountAMD,
    .CmdSetCheckpointNV = anv_CmdSetCheckpointNV,
    .GetQueueCheckpointDataNV = anv_GetQueueCheckpointDataNV,
    .CmdBindTransformFeedbackBuffersEXT = anv_CmdBindTransformFeedbackBuffersEXT,
    .CmdBeginTransformFeedbackEXT = anv_CmdBeginTransformFeedbackEXT,
    .CmdEndTransformFeedbackEXT = anv_CmdEndTransformFeedbackEXT,
    .CmdBeginQueryIndexedEXT = anv_CmdBeginQueryIndexedEXT,
    .CmdEndQueryIndexedEXT = anv_CmdEndQueryIndexedEXT,
    .CmdDrawIndirectByteCountEXT = anv_CmdDrawIndirectByteCountEXT,
    .CmdSetExclusiveScissorNV = anv_CmdSetExclusiveScissorNV,
    .CmdSetExclusiveScissorEnableNV = anv_CmdSetExclusiveScissorEnableNV,
    .CmdBindShadingRateImageNV = anv_CmdBindShadingRateImageNV,
    .CmdSetViewportShadingRatePaletteNV = anv_CmdSetViewportShadingRatePaletteNV,
    .CmdSetCoarseSampleOrderNV = anv_CmdSetCoarseSampleOrderNV,
    .CmdDrawMeshTasksNV = anv_CmdDrawMeshTasksNV,
    .CmdDrawMeshTasksIndirectNV = anv_CmdDrawMeshTasksIndirectNV,
    .CmdDrawMeshTasksIndirectCountNV = anv_CmdDrawMeshTasksIndirectCountNV,
    .CmdDrawMeshTasksEXT = anv_CmdDrawMeshTasksEXT,
    .CmdDrawMeshTasksIndirectEXT = anv_CmdDrawMeshTasksIndirectEXT,
    .CmdDrawMeshTasksIndirectCountEXT = anv_CmdDrawMeshTasksIndirectCountEXT,
    .CompileDeferredNV = anv_CompileDeferredNV,
    .CreateAccelerationStructureNV = anv_CreateAccelerationStructureNV,
    .CmdBindInvocationMaskHUAWEI = anv_CmdBindInvocationMaskHUAWEI,
    .DestroyAccelerationStructureKHR = anv_DestroyAccelerationStructureKHR,
    .DestroyAccelerationStructureNV = anv_DestroyAccelerationStructureNV,
    .GetAccelerationStructureMemoryRequirementsNV = anv_GetAccelerationStructureMemoryRequirementsNV,
    .BindAccelerationStructureMemoryNV = anv_BindAccelerationStructureMemoryNV,
    .CmdCopyAccelerationStructureNV = anv_CmdCopyAccelerationStructureNV,
    .CmdCopyAccelerationStructureKHR = anv_CmdCopyAccelerationStructureKHR,
    .CopyAccelerationStructureKHR = anv_CopyAccelerationStructureKHR,
    .CmdCopyAccelerationStructureToMemoryKHR = anv_CmdCopyAccelerationStructureToMemoryKHR,
    .CopyAccelerationStructureToMemoryKHR = anv_CopyAccelerationStructureToMemoryKHR,
    .CmdCopyMemoryToAccelerationStructureKHR = anv_CmdCopyMemoryToAccelerationStructureKHR,
    .CopyMemoryToAccelerationStructureKHR = anv_CopyMemoryToAccelerationStructureKHR,
    .CmdWriteAccelerationStructuresPropertiesKHR = anv_CmdWriteAccelerationStructuresPropertiesKHR,
    .CmdWriteAccelerationStructuresPropertiesNV = anv_CmdWriteAccelerationStructuresPropertiesNV,
    .CmdBuildAccelerationStructureNV = anv_CmdBuildAccelerationStructureNV,
    .WriteAccelerationStructuresPropertiesKHR = anv_WriteAccelerationStructuresPropertiesKHR,
    .CmdTraceRaysKHR = anv_CmdTraceRaysKHR,
    .CmdTraceRaysNV = anv_CmdTraceRaysNV,
    .GetRayTracingShaderGroupHandlesKHR = anv_GetRayTracingShaderGroupHandlesKHR,
    .GetRayTracingShaderGroupHandlesNV = anv_GetRayTracingShaderGroupHandlesNV,
    .GetRayTracingCaptureReplayShaderGroupHandlesKHR = anv_GetRayTracingCaptureReplayShaderGroupHandlesKHR,
    .GetAccelerationStructureHandleNV = anv_GetAccelerationStructureHandleNV,
    .CreateRayTracingPipelinesNV = anv_CreateRayTracingPipelinesNV,
    .CreateRayTracingPipelinesKHR = anv_CreateRayTracingPipelinesKHR,
    .CmdTraceRaysIndirectKHR = anv_CmdTraceRaysIndirectKHR,
    .CmdTraceRaysIndirect2KHR = anv_CmdTraceRaysIndirect2KHR,
    .GetClusterAccelerationStructureBuildSizesNV = anv_GetClusterAccelerationStructureBuildSizesNV,
    .CmdBuildClusterAccelerationStructureIndirectNV = anv_CmdBuildClusterAccelerationStructureIndirectNV,
    .GetDeviceAccelerationStructureCompatibilityKHR = anv_GetDeviceAccelerationStructureCompatibilityKHR,
    .GetRayTracingShaderGroupStackSizeKHR = anv_GetRayTracingShaderGroupStackSizeKHR,
    .CmdSetRayTracingPipelineStackSizeKHR = anv_CmdSetRayTracingPipelineStackSizeKHR,
    .GetImageViewHandleNVX = anv_GetImageViewHandleNVX,
    .GetImageViewHandle64NVX = anv_GetImageViewHandle64NVX,
    .GetImageViewAddressNVX = anv_GetImageViewAddressNVX,
    .GetDeviceCombinedImageSamplerIndexNVX = anv_GetDeviceCombinedImageSamplerIndexNVX,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetDeviceGroupSurfacePresentModes2EXT = anv_GetDeviceGroupSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetDeviceGroupSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireFullScreenExclusiveModeEXT = anv_AcquireFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .AcquireFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ReleaseFullScreenExclusiveModeEXT = anv_ReleaseFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .ReleaseFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .AcquireProfilingLockKHR = anv_AcquireProfilingLockKHR,
    .ReleaseProfilingLockKHR = anv_ReleaseProfilingLockKHR,
    .GetImageDrmFormatModifierPropertiesEXT = anv_GetImageDrmFormatModifierPropertiesEXT,
    .GetBufferOpaqueCaptureAddress = anv_GetBufferOpaqueCaptureAddress,
    .GetBufferOpaqueCaptureAddressKHR = anv_GetBufferOpaqueCaptureAddressKHR,
    .GetBufferDeviceAddress = anv_GetBufferDeviceAddress,
    .GetBufferDeviceAddressKHR = anv_GetBufferDeviceAddressKHR,
    .GetBufferDeviceAddressEXT = anv_GetBufferDeviceAddressEXT,
    .InitializePerformanceApiINTEL = anv_InitializePerformanceApiINTEL,
    .UninitializePerformanceApiINTEL = anv_UninitializePerformanceApiINTEL,
    .CmdSetPerformanceMarkerINTEL = anv_CmdSetPerformanceMarkerINTEL,
    .CmdSetPerformanceStreamMarkerINTEL = anv_CmdSetPerformanceStreamMarkerINTEL,
    .CmdSetPerformanceOverrideINTEL = anv_CmdSetPerformanceOverrideINTEL,
    .AcquirePerformanceConfigurationINTEL = anv_AcquirePerformanceConfigurationINTEL,
    .ReleasePerformanceConfigurationINTEL = anv_ReleasePerformanceConfigurationINTEL,
    .QueueSetPerformanceConfigurationINTEL = anv_QueueSetPerformanceConfigurationINTEL,
    .GetPerformanceParameterINTEL = anv_GetPerformanceParameterINTEL,
    .GetDeviceMemoryOpaqueCaptureAddress = anv_GetDeviceMemoryOpaqueCaptureAddress,
    .GetDeviceMemoryOpaqueCaptureAddressKHR = anv_GetDeviceMemoryOpaqueCaptureAddressKHR,
    .GetPipelineExecutablePropertiesKHR = anv_GetPipelineExecutablePropertiesKHR,
    .GetPipelineExecutableStatisticsKHR = anv_GetPipelineExecutableStatisticsKHR,
    .GetPipelineExecutableInternalRepresentationsKHR = anv_GetPipelineExecutableInternalRepresentationsKHR,
    .CmdSetLineStipple = anv_CmdSetLineStipple,
    .CmdSetLineStippleKHR = anv_CmdSetLineStippleKHR,
    .CmdSetLineStippleEXT = anv_CmdSetLineStippleEXT,
    .CreateAccelerationStructureKHR = anv_CreateAccelerationStructureKHR,
    .CmdBuildAccelerationStructuresKHR = anv_CmdBuildAccelerationStructuresKHR,
    .CmdBuildAccelerationStructuresIndirectKHR = anv_CmdBuildAccelerationStructuresIndirectKHR,
    .BuildAccelerationStructuresKHR = anv_BuildAccelerationStructuresKHR,
    .GetAccelerationStructureDeviceAddressKHR = anv_GetAccelerationStructureDeviceAddressKHR,
    .CreateDeferredOperationKHR = anv_CreateDeferredOperationKHR,
    .DestroyDeferredOperationKHR = anv_DestroyDeferredOperationKHR,
    .GetDeferredOperationMaxConcurrencyKHR = anv_GetDeferredOperationMaxConcurrencyKHR,
    .GetDeferredOperationResultKHR = anv_GetDeferredOperationResultKHR,
    .DeferredOperationJoinKHR = anv_DeferredOperationJoinKHR,
    .GetPipelineIndirectMemoryRequirementsNV = anv_GetPipelineIndirectMemoryRequirementsNV,
    .GetPipelineIndirectDeviceAddressNV = anv_GetPipelineIndirectDeviceAddressNV,
    .AntiLagUpdateAMD = anv_AntiLagUpdateAMD,
    .CmdSetCullMode = anv_CmdSetCullMode,
    .CmdSetCullModeEXT = anv_CmdSetCullModeEXT,
    .CmdSetFrontFace = anv_CmdSetFrontFace,
    .CmdSetFrontFaceEXT = anv_CmdSetFrontFaceEXT,
    .CmdSetPrimitiveTopology = anv_CmdSetPrimitiveTopology,
    .CmdSetPrimitiveTopologyEXT = anv_CmdSetPrimitiveTopologyEXT,
    .CmdSetViewportWithCount = anv_CmdSetViewportWithCount,
    .CmdSetViewportWithCountEXT = anv_CmdSetViewportWithCountEXT,
    .CmdSetScissorWithCount = anv_CmdSetScissorWithCount,
    .CmdSetScissorWithCountEXT = anv_CmdSetScissorWithCountEXT,
    .CmdBindIndexBuffer2 = anv_CmdBindIndexBuffer2,
    .CmdBindIndexBuffer2KHR = anv_CmdBindIndexBuffer2KHR,
    .CmdBindVertexBuffers2 = anv_CmdBindVertexBuffers2,
    .CmdBindVertexBuffers2EXT = anv_CmdBindVertexBuffers2EXT,
    .CmdSetDepthTestEnable = anv_CmdSetDepthTestEnable,
    .CmdSetDepthTestEnableEXT = anv_CmdSetDepthTestEnableEXT,
    .CmdSetDepthWriteEnable = anv_CmdSetDepthWriteEnable,
    .CmdSetDepthWriteEnableEXT = anv_CmdSetDepthWriteEnableEXT,
    .CmdSetDepthCompareOp = anv_CmdSetDepthCompareOp,
    .CmdSetDepthCompareOpEXT = anv_CmdSetDepthCompareOpEXT,
    .CmdSetDepthBoundsTestEnable = anv_CmdSetDepthBoundsTestEnable,
    .CmdSetDepthBoundsTestEnableEXT = anv_CmdSetDepthBoundsTestEnableEXT,
    .CmdSetStencilTestEnable = anv_CmdSetStencilTestEnable,
    .CmdSetStencilTestEnableEXT = anv_CmdSetStencilTestEnableEXT,
    .CmdSetStencilOp = anv_CmdSetStencilOp,
    .CmdSetStencilOpEXT = anv_CmdSetStencilOpEXT,
    .CmdSetPatchControlPointsEXT = anv_CmdSetPatchControlPointsEXT,
    .CmdSetRasterizerDiscardEnable = anv_CmdSetRasterizerDiscardEnable,
    .CmdSetRasterizerDiscardEnableEXT = anv_CmdSetRasterizerDiscardEnableEXT,
    .CmdSetDepthBiasEnable = anv_CmdSetDepthBiasEnable,
    .CmdSetDepthBiasEnableEXT = anv_CmdSetDepthBiasEnableEXT,
    .CmdSetLogicOpEXT = anv_CmdSetLogicOpEXT,
    .CmdSetPrimitiveRestartEnable = anv_CmdSetPrimitiveRestartEnable,
    .CmdSetPrimitiveRestartEnableEXT = anv_CmdSetPrimitiveRestartEnableEXT,
    .CmdSetTessellationDomainOriginEXT = anv_CmdSetTessellationDomainOriginEXT,
    .CmdSetDepthClampEnableEXT = anv_CmdSetDepthClampEnableEXT,
    .CmdSetPolygonModeEXT = anv_CmdSetPolygonModeEXT,
    .CmdSetRasterizationSamplesEXT = anv_CmdSetRasterizationSamplesEXT,
    .CmdSetSampleMaskEXT = anv_CmdSetSampleMaskEXT,
    .CmdSetAlphaToCoverageEnableEXT = anv_CmdSetAlphaToCoverageEnableEXT,
    .CmdSetAlphaToOneEnableEXT = anv_CmdSetAlphaToOneEnableEXT,
    .CmdSetLogicOpEnableEXT = anv_CmdSetLogicOpEnableEXT,
    .CmdSetColorBlendEnableEXT = anv_CmdSetColorBlendEnableEXT,
    .CmdSetColorBlendEquationEXT = anv_CmdSetColorBlendEquationEXT,
    .CmdSetColorWriteMaskEXT = anv_CmdSetColorWriteMaskEXT,
    .CmdSetRasterizationStreamEXT = anv_CmdSetRasterizationStreamEXT,
    .CmdSetConservativeRasterizationModeEXT = anv_CmdSetConservativeRasterizationModeEXT,
    .CmdSetExtraPrimitiveOverestimationSizeEXT = anv_CmdSetExtraPrimitiveOverestimationSizeEXT,
    .CmdSetDepthClipEnableEXT = anv_CmdSetDepthClipEnableEXT,
    .CmdSetSampleLocationsEnableEXT = anv_CmdSetSampleLocationsEnableEXT,
    .CmdSetColorBlendAdvancedEXT = anv_CmdSetColorBlendAdvancedEXT,
    .CmdSetProvokingVertexModeEXT = anv_CmdSetProvokingVertexModeEXT,
    .CmdSetLineRasterizationModeEXT = anv_CmdSetLineRasterizationModeEXT,
    .CmdSetLineStippleEnableEXT = anv_CmdSetLineStippleEnableEXT,
    .CmdSetDepthClipNegativeOneToOneEXT = anv_CmdSetDepthClipNegativeOneToOneEXT,
    .CmdSetViewportWScalingEnableNV = anv_CmdSetViewportWScalingEnableNV,
    .CmdSetViewportSwizzleNV = anv_CmdSetViewportSwizzleNV,
    .CmdSetCoverageToColorEnableNV = anv_CmdSetCoverageToColorEnableNV,
    .CmdSetCoverageToColorLocationNV = anv_CmdSetCoverageToColorLocationNV,
    .CmdSetCoverageModulationModeNV = anv_CmdSetCoverageModulationModeNV,
    .CmdSetCoverageModulationTableEnableNV = anv_CmdSetCoverageModulationTableEnableNV,
    .CmdSetCoverageModulationTableNV = anv_CmdSetCoverageModulationTableNV,
    .CmdSetShadingRateImageEnableNV = anv_CmdSetShadingRateImageEnableNV,
    .CmdSetCoverageReductionModeNV = anv_CmdSetCoverageReductionModeNV,
    .CmdSetRepresentativeFragmentTestEnableNV = anv_CmdSetRepresentativeFragmentTestEnableNV,
    .CreatePrivateDataSlot = anv_CreatePrivateDataSlot,
    .CreatePrivateDataSlotEXT = anv_CreatePrivateDataSlotEXT,
    .DestroyPrivateDataSlot = anv_DestroyPrivateDataSlot,
    .DestroyPrivateDataSlotEXT = anv_DestroyPrivateDataSlotEXT,
    .SetPrivateData = anv_SetPrivateData,
    .SetPrivateDataEXT = anv_SetPrivateDataEXT,
    .GetPrivateData = anv_GetPrivateData,
    .GetPrivateDataEXT = anv_GetPrivateDataEXT,
    .CmdCopyBuffer2 = anv_CmdCopyBuffer2,
    .CmdCopyBuffer2KHR = anv_CmdCopyBuffer2KHR,
    .CmdCopyImage2 = anv_CmdCopyImage2,
    .CmdCopyImage2KHR = anv_CmdCopyImage2KHR,
    .CmdBlitImage2 = anv_CmdBlitImage2,
    .CmdBlitImage2KHR = anv_CmdBlitImage2KHR,
    .CmdCopyBufferToImage2 = anv_CmdCopyBufferToImage2,
    .CmdCopyBufferToImage2KHR = anv_CmdCopyBufferToImage2KHR,
    .CmdCopyImageToBuffer2 = anv_CmdCopyImageToBuffer2,
    .CmdCopyImageToBuffer2KHR = anv_CmdCopyImageToBuffer2KHR,
    .CmdResolveImage2 = anv_CmdResolveImage2,
    .CmdResolveImage2KHR = anv_CmdResolveImage2KHR,
    .CmdSetFragmentShadingRateKHR = anv_CmdSetFragmentShadingRateKHR,
    .CmdSetFragmentShadingRateEnumNV = anv_CmdSetFragmentShadingRateEnumNV,
    .GetAccelerationStructureBuildSizesKHR = anv_GetAccelerationStructureBuildSizesKHR,
    .CmdSetVertexInputEXT = anv_CmdSetVertexInputEXT,
    .CmdSetColorWriteEnableEXT = anv_CmdSetColorWriteEnableEXT,
    .CmdSetEvent2 = anv_CmdSetEvent2,
    .CmdSetEvent2KHR = anv_CmdSetEvent2KHR,
    .CmdResetEvent2 = anv_CmdResetEvent2,
    .CmdResetEvent2KHR = anv_CmdResetEvent2KHR,
    .CmdWaitEvents2 = anv_CmdWaitEvents2,
    .CmdWaitEvents2KHR = anv_CmdWaitEvents2KHR,
    .CmdPipelineBarrier2 = anv_CmdPipelineBarrier2,
    .CmdPipelineBarrier2KHR = anv_CmdPipelineBarrier2KHR,
    .QueueSubmit2 = anv_QueueSubmit2,
    .QueueSubmit2KHR = anv_QueueSubmit2KHR,
    .CmdWriteTimestamp2 = anv_CmdWriteTimestamp2,
    .CmdWriteTimestamp2KHR = anv_CmdWriteTimestamp2KHR,
    .CmdWriteBufferMarker2AMD = anv_CmdWriteBufferMarker2AMD,
    .GetQueueCheckpointData2NV = anv_GetQueueCheckpointData2NV,
    .CopyMemoryToImage = anv_CopyMemoryToImage,
    .CopyMemoryToImageEXT = anv_CopyMemoryToImageEXT,
    .CopyImageToMemory = anv_CopyImageToMemory,
    .CopyImageToMemoryEXT = anv_CopyImageToMemoryEXT,
    .CopyImageToImage = anv_CopyImageToImage,
    .CopyImageToImageEXT = anv_CopyImageToImageEXT,
    .TransitionImageLayout = anv_TransitionImageLayout,
    .TransitionImageLayoutEXT = anv_TransitionImageLayoutEXT,
    .CreateVideoSessionKHR = anv_CreateVideoSessionKHR,
    .DestroyVideoSessionKHR = anv_DestroyVideoSessionKHR,
    .CreateVideoSessionParametersKHR = anv_CreateVideoSessionParametersKHR,
    .UpdateVideoSessionParametersKHR = anv_UpdateVideoSessionParametersKHR,
    .GetEncodedVideoSessionParametersKHR = anv_GetEncodedVideoSessionParametersKHR,
    .DestroyVideoSessionParametersKHR = anv_DestroyVideoSessionParametersKHR,
    .GetVideoSessionMemoryRequirementsKHR = anv_GetVideoSessionMemoryRequirementsKHR,
    .BindVideoSessionMemoryKHR = anv_BindVideoSessionMemoryKHR,
    .CmdDecodeVideoKHR = anv_CmdDecodeVideoKHR,
    .CmdBeginVideoCodingKHR = anv_CmdBeginVideoCodingKHR,
    .CmdControlVideoCodingKHR = anv_CmdControlVideoCodingKHR,
    .CmdEndVideoCodingKHR = anv_CmdEndVideoCodingKHR,
    .CmdEncodeVideoKHR = anv_CmdEncodeVideoKHR,
    .CmdDecompressMemoryNV = anv_CmdDecompressMemoryNV,
    .CmdDecompressMemoryIndirectCountNV = anv_CmdDecompressMemoryIndirectCountNV,
    .GetPartitionedAccelerationStructuresBuildSizesNV = anv_GetPartitionedAccelerationStructuresBuildSizesNV,
    .CmdBuildPartitionedAccelerationStructuresNV = anv_CmdBuildPartitionedAccelerationStructuresNV,
    .CmdDecompressMemoryEXT = anv_CmdDecompressMemoryEXT,
    .CmdDecompressMemoryIndirectCountEXT = anv_CmdDecompressMemoryIndirectCountEXT,
    .CreateCuModuleNVX = anv_CreateCuModuleNVX,
    .CreateCuFunctionNVX = anv_CreateCuFunctionNVX,
    .DestroyCuModuleNVX = anv_DestroyCuModuleNVX,
    .DestroyCuFunctionNVX = anv_DestroyCuFunctionNVX,
    .CmdCuLaunchKernelNVX = anv_CmdCuLaunchKernelNVX,
    .GetDescriptorSetLayoutSizeEXT = anv_GetDescriptorSetLayoutSizeEXT,
    .GetDescriptorSetLayoutBindingOffsetEXT = anv_GetDescriptorSetLayoutBindingOffsetEXT,
    .GetDescriptorEXT = anv_GetDescriptorEXT,
    .CmdBindDescriptorBuffersEXT = anv_CmdBindDescriptorBuffersEXT,
    .CmdSetDescriptorBufferOffsetsEXT = anv_CmdSetDescriptorBufferOffsetsEXT,
    .CmdBindDescriptorBufferEmbeddedSamplersEXT = anv_CmdBindDescriptorBufferEmbeddedSamplersEXT,
    .GetBufferOpaqueCaptureDescriptorDataEXT = anv_GetBufferOpaqueCaptureDescriptorDataEXT,
    .GetImageOpaqueCaptureDescriptorDataEXT = anv_GetImageOpaqueCaptureDescriptorDataEXT,
    .GetImageViewOpaqueCaptureDescriptorDataEXT = anv_GetImageViewOpaqueCaptureDescriptorDataEXT,
    .GetSamplerOpaqueCaptureDescriptorDataEXT = anv_GetSamplerOpaqueCaptureDescriptorDataEXT,
    .GetAccelerationStructureOpaqueCaptureDescriptorDataEXT = anv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT,
    .SetDeviceMemoryPriorityEXT = anv_SetDeviceMemoryPriorityEXT,
    .WaitForPresent2KHR = anv_WaitForPresent2KHR,
    .WaitForPresentKHR = anv_WaitForPresentKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateBufferCollectionFUCHSIA = anv_CreateBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .CreateBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionBufferConstraintsFUCHSIA = anv_SetBufferCollectionBufferConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionBufferConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionImageConstraintsFUCHSIA = anv_SetBufferCollectionImageConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionImageConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .DestroyBufferCollectionFUCHSIA = anv_DestroyBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .DestroyBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetBufferCollectionPropertiesFUCHSIA = anv_GetBufferCollectionPropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetBufferCollectionPropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .CmdBeginRendering = anv_CmdBeginRendering,
    .CmdBeginRenderingKHR = anv_CmdBeginRenderingKHR,
    .CmdEndRendering = anv_CmdEndRendering,
    .CmdEndRendering2KHR = anv_CmdEndRendering2KHR,
    .CmdEndRendering2EXT = anv_CmdEndRendering2EXT,
    .CmdEndRenderingKHR = anv_CmdEndRenderingKHR,
    .GetDescriptorSetLayoutHostMappingInfoVALVE = anv_GetDescriptorSetLayoutHostMappingInfoVALVE,
    .GetDescriptorSetHostMappingVALVE = anv_GetDescriptorSetHostMappingVALVE,
    .CreateMicromapEXT = anv_CreateMicromapEXT,
    .CmdBuildMicromapsEXT = anv_CmdBuildMicromapsEXT,
    .BuildMicromapsEXT = anv_BuildMicromapsEXT,
    .DestroyMicromapEXT = anv_DestroyMicromapEXT,
    .CmdCopyMicromapEXT = anv_CmdCopyMicromapEXT,
    .CopyMicromapEXT = anv_CopyMicromapEXT,
    .CmdCopyMicromapToMemoryEXT = anv_CmdCopyMicromapToMemoryEXT,
    .CopyMicromapToMemoryEXT = anv_CopyMicromapToMemoryEXT,
    .CmdCopyMemoryToMicromapEXT = anv_CmdCopyMemoryToMicromapEXT,
    .CopyMemoryToMicromapEXT = anv_CopyMemoryToMicromapEXT,
    .CmdWriteMicromapsPropertiesEXT = anv_CmdWriteMicromapsPropertiesEXT,
    .WriteMicromapsPropertiesEXT = anv_WriteMicromapsPropertiesEXT,
    .GetDeviceMicromapCompatibilityEXT = anv_GetDeviceMicromapCompatibilityEXT,
    .GetMicromapBuildSizesEXT = anv_GetMicromapBuildSizesEXT,
    .GetShaderModuleIdentifierEXT = anv_GetShaderModuleIdentifierEXT,
    .GetShaderModuleCreateInfoIdentifierEXT = anv_GetShaderModuleCreateInfoIdentifierEXT,
    .GetImageSubresourceLayout2 = anv_GetImageSubresourceLayout2,
    .GetImageSubresourceLayout2KHR = anv_GetImageSubresourceLayout2KHR,
    .GetImageSubresourceLayout2EXT = anv_GetImageSubresourceLayout2EXT,
    .GetPipelinePropertiesEXT = anv_GetPipelinePropertiesEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .ExportMetalObjectsEXT = anv_ExportMetalObjectsEXT,
#elif defined(_MSC_VER)
    .ExportMetalObjectsEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CmdBindTileMemoryQCOM = anv_CmdBindTileMemoryQCOM,
    .GetFramebufferTilePropertiesQCOM = anv_GetFramebufferTilePropertiesQCOM,
    .GetDynamicRenderingTilePropertiesQCOM = anv_GetDynamicRenderingTilePropertiesQCOM,
    .CreateOpticalFlowSessionNV = anv_CreateOpticalFlowSessionNV,
    .DestroyOpticalFlowSessionNV = anv_DestroyOpticalFlowSessionNV,
    .BindOpticalFlowSessionImageNV = anv_BindOpticalFlowSessionImageNV,
    .CmdOpticalFlowExecuteNV = anv_CmdOpticalFlowExecuteNV,
    .GetDeviceFaultInfoEXT = anv_GetDeviceFaultInfoEXT,
    .GetDeviceFaultReportsKHR = anv_GetDeviceFaultReportsKHR,
    .GetDeviceFaultDebugInfoKHR = anv_GetDeviceFaultDebugInfoKHR,
    .CmdSetDepthBias2EXT = anv_CmdSetDepthBias2EXT,
    .ReleaseSwapchainImagesKHR = anv_ReleaseSwapchainImagesKHR,
    .ReleaseSwapchainImagesEXT = anv_ReleaseSwapchainImagesEXT,
    .GetDeviceImageSubresourceLayout = anv_GetDeviceImageSubresourceLayout,
    .GetDeviceImageSubresourceLayoutKHR = anv_GetDeviceImageSubresourceLayoutKHR,
    .MapMemory2 = anv_MapMemory2,
    .MapMemory2KHR = anv_MapMemory2KHR,
    .UnmapMemory2 = anv_UnmapMemory2,
    .UnmapMemory2KHR = anv_UnmapMemory2KHR,
    .CreateShadersEXT = anv_CreateShadersEXT,
    .DestroyShaderEXT = anv_DestroyShaderEXT,
    .GetShaderBinaryDataEXT = anv_GetShaderBinaryDataEXT,
    .CmdBindShadersEXT = anv_CmdBindShadersEXT,
    .SetSwapchainPresentTimingQueueSizeEXT = anv_SetSwapchainPresentTimingQueueSizeEXT,
    .GetSwapchainTimingPropertiesEXT = anv_GetSwapchainTimingPropertiesEXT,
    .GetSwapchainTimeDomainPropertiesEXT = anv_GetSwapchainTimeDomainPropertiesEXT,
    .GetPastPresentationTimingEXT = anv_GetPastPresentationTimingEXT,
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetScreenBufferPropertiesQNX = anv_GetScreenBufferPropertiesQNX,
#elif defined(_MSC_VER)
    .GetScreenBufferPropertiesQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CmdBindDescriptorSets2 = anv_CmdBindDescriptorSets2,
    .CmdBindDescriptorSets2KHR = anv_CmdBindDescriptorSets2KHR,
    .CmdPushConstants2 = anv_CmdPushConstants2,
    .CmdPushConstants2KHR = anv_CmdPushConstants2KHR,
    .CmdPushDescriptorSet2 = anv_CmdPushDescriptorSet2,
    .CmdPushDescriptorSet2KHR = anv_CmdPushDescriptorSet2KHR,
    .CmdPushDescriptorSetWithTemplate2 = anv_CmdPushDescriptorSetWithTemplate2,
    .CmdPushDescriptorSetWithTemplate2KHR = anv_CmdPushDescriptorSetWithTemplate2KHR,
    .CmdSetDescriptorBufferOffsets2EXT = anv_CmdSetDescriptorBufferOffsets2EXT,
    .CmdBindDescriptorBufferEmbeddedSamplers2EXT = anv_CmdBindDescriptorBufferEmbeddedSamplers2EXT,
    .SetLatencySleepModeNV = anv_SetLatencySleepModeNV,
    .LatencySleepNV = anv_LatencySleepNV,
    .SetLatencyMarkerNV = anv_SetLatencyMarkerNV,
    .GetLatencyTimingsNV = anv_GetLatencyTimingsNV,
    .QueueNotifyOutOfBandNV = anv_QueueNotifyOutOfBandNV,
    .CmdSetRenderingAttachmentLocations = anv_CmdSetRenderingAttachmentLocations,
    .CmdSetRenderingAttachmentLocationsKHR = anv_CmdSetRenderingAttachmentLocationsKHR,
    .CmdSetRenderingInputAttachmentIndices = anv_CmdSetRenderingInputAttachmentIndices,
    .CmdSetRenderingInputAttachmentIndicesKHR = anv_CmdSetRenderingInputAttachmentIndicesKHR,
    .CmdSetDepthClampRangeEXT = anv_CmdSetDepthClampRangeEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandleEXT = anv_GetMemoryMetalHandleEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandleEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandlePropertiesEXT = anv_GetMemoryMetalHandlePropertiesEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandlePropertiesEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .ConvertCooperativeVectorMatrixNV = anv_ConvertCooperativeVectorMatrixNV,
    .CmdConvertCooperativeVectorMatrixNV = anv_CmdConvertCooperativeVectorMatrixNV,
    .CmdDispatchTileQCOM = anv_CmdDispatchTileQCOM,
    .CmdBeginPerTileExecutionQCOM = anv_CmdBeginPerTileExecutionQCOM,
    .CmdEndPerTileExecutionQCOM = anv_CmdEndPerTileExecutionQCOM,
    .CreateExternalComputeQueueNV = anv_CreateExternalComputeQueueNV,
    .DestroyExternalComputeQueueNV = anv_DestroyExternalComputeQueueNV,
    .CreateShaderInstrumentationARM = anv_CreateShaderInstrumentationARM,
    .DestroyShaderInstrumentationARM = anv_DestroyShaderInstrumentationARM,
    .CmdBeginShaderInstrumentationARM = anv_CmdBeginShaderInstrumentationARM,
    .CmdEndShaderInstrumentationARM = anv_CmdEndShaderInstrumentationARM,
    .GetShaderInstrumentationValuesARM = anv_GetShaderInstrumentationValuesARM,
    .ClearShaderInstrumentationMetricsARM = anv_ClearShaderInstrumentationMetricsARM,
    .CreateTensorARM = anv_CreateTensorARM,
    .DestroyTensorARM = anv_DestroyTensorARM,
    .CreateTensorViewARM = anv_CreateTensorViewARM,
    .DestroyTensorViewARM = anv_DestroyTensorViewARM,
    .GetTensorMemoryRequirementsARM = anv_GetTensorMemoryRequirementsARM,
    .BindTensorMemoryARM = anv_BindTensorMemoryARM,
    .GetDeviceTensorMemoryRequirementsARM = anv_GetDeviceTensorMemoryRequirementsARM,
    .CmdCopyTensorARM = anv_CmdCopyTensorARM,
    .GetTensorOpaqueCaptureDescriptorDataARM = anv_GetTensorOpaqueCaptureDescriptorDataARM,
    .GetTensorViewOpaqueCaptureDescriptorDataARM = anv_GetTensorViewOpaqueCaptureDescriptorDataARM,
    .CreateDataGraphPipelinesARM = anv_CreateDataGraphPipelinesARM,
    .CreateDataGraphPipelineSessionARM = anv_CreateDataGraphPipelineSessionARM,
    .GetDataGraphPipelineSessionBindPointRequirementsARM = anv_GetDataGraphPipelineSessionBindPointRequirementsARM,
    .GetDataGraphPipelineSessionMemoryRequirementsARM = anv_GetDataGraphPipelineSessionMemoryRequirementsARM,
    .BindDataGraphPipelineSessionMemoryARM = anv_BindDataGraphPipelineSessionMemoryARM,
    .DestroyDataGraphPipelineSessionARM = anv_DestroyDataGraphPipelineSessionARM,
    .CmdDispatchDataGraphARM = anv_CmdDispatchDataGraphARM,
    .GetDataGraphPipelineAvailablePropertiesARM = anv_GetDataGraphPipelineAvailablePropertiesARM,
    .GetDataGraphPipelinePropertiesARM = anv_GetDataGraphPipelinePropertiesARM,
#ifdef VK_USE_PLATFORM_OHOS
    .GetNativeBufferPropertiesOHOS = anv_GetNativeBufferPropertiesOHOS,
#elif defined(_MSC_VER)
    .GetNativeBufferPropertiesOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
    .GetMemoryNativeBufferOHOS = anv_GetMemoryNativeBufferOHOS,
#elif defined(_MSC_VER)
    .GetMemoryNativeBufferOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .QueueSetPerfHintQCOM = anv_QueueSetPerfHintQCOM,
    .CmdSetComputeOccupancyPriorityNV = anv_CmdSetComputeOccupancyPriorityNV,
    .WriteSamplerDescriptorsEXT = anv_WriteSamplerDescriptorsEXT,
    .WriteResourceDescriptorsEXT = anv_WriteResourceDescriptorsEXT,
    .CmdBindSamplerHeapEXT = anv_CmdBindSamplerHeapEXT,
    .CmdBindResourceHeapEXT = anv_CmdBindResourceHeapEXT,
    .CmdPushDataEXT = anv_CmdPushDataEXT,
    .RegisterCustomBorderColorEXT = anv_RegisterCustomBorderColorEXT,
    .UnregisterCustomBorderColorEXT = anv_UnregisterCustomBorderColorEXT,
    .GetImageOpaqueCaptureDataEXT = anv_GetImageOpaqueCaptureDataEXT,
    .GetTensorOpaqueCaptureDataARM = anv_GetTensorOpaqueCaptureDataARM,
    .CmdCopyMemoryKHR = anv_CmdCopyMemoryKHR,
    .CmdCopyMemoryToImageKHR = anv_CmdCopyMemoryToImageKHR,
    .CmdCopyImageToMemoryKHR = anv_CmdCopyImageToMemoryKHR,
    .CmdUpdateMemoryKHR = anv_CmdUpdateMemoryKHR,
    .CmdFillMemoryKHR = anv_CmdFillMemoryKHR,
    .CmdCopyQueryPoolResultsToMemoryKHR = anv_CmdCopyQueryPoolResultsToMemoryKHR,
    .CmdBeginConditionalRendering2EXT = anv_CmdBeginConditionalRendering2EXT,
    .CmdBindTransformFeedbackBuffers2EXT = anv_CmdBindTransformFeedbackBuffers2EXT,
    .CmdBeginTransformFeedback2EXT = anv_CmdBeginTransformFeedback2EXT,
    .CmdEndTransformFeedback2EXT = anv_CmdEndTransformFeedback2EXT,
    .CmdDrawIndirectByteCount2EXT = anv_CmdDrawIndirectByteCount2EXT,
    .CmdWriteMarkerToMemoryAMD = anv_CmdWriteMarkerToMemoryAMD,
    .CmdBindIndexBuffer3KHR = anv_CmdBindIndexBuffer3KHR,
    .CmdBindVertexBuffers3KHR = anv_CmdBindVertexBuffers3KHR,
    .CmdDrawIndirect2KHR = anv_CmdDrawIndirect2KHR,
    .CmdDrawIndexedIndirect2KHR = anv_CmdDrawIndexedIndirect2KHR,
    .CmdDrawIndirectCount2KHR = anv_CmdDrawIndirectCount2KHR,
    .CmdDrawIndexedIndirectCount2KHR = anv_CmdDrawIndexedIndirectCount2KHR,
    .CmdDrawMeshTasksIndirect2EXT = anv_CmdDrawMeshTasksIndirect2EXT,
    .CmdDrawMeshTasksIndirectCount2EXT = anv_CmdDrawMeshTasksIndirectCount2EXT,
    .CmdDispatchIndirect2KHR = anv_CmdDispatchIndirect2KHR,
    .CreateAccelerationStructure2KHR = anv_CreateAccelerationStructure2KHR,
    .CmdSetDispatchParametersARM = anv_CmdSetDispatchParametersARM,
};
const struct vk_device_entrypoint_table gfx9_device_entrypoints = {
    .GetDeviceProcAddr = gfx9_GetDeviceProcAddr,
    .DestroyDevice = gfx9_DestroyDevice,
    .GetDeviceQueue = gfx9_GetDeviceQueue,
    .QueueSubmit = gfx9_QueueSubmit,
    .QueueWaitIdle = gfx9_QueueWaitIdle,
    .DeviceWaitIdle = gfx9_DeviceWaitIdle,
    .AllocateMemory = gfx9_AllocateMemory,
    .FreeMemory = gfx9_FreeMemory,
    .MapMemory = gfx9_MapMemory,
    .UnmapMemory = gfx9_UnmapMemory,
    .FlushMappedMemoryRanges = gfx9_FlushMappedMemoryRanges,
    .InvalidateMappedMemoryRanges = gfx9_InvalidateMappedMemoryRanges,
    .GetDeviceMemoryCommitment = gfx9_GetDeviceMemoryCommitment,
    .GetBufferMemoryRequirements = gfx9_GetBufferMemoryRequirements,
    .BindBufferMemory = gfx9_BindBufferMemory,
    .GetImageMemoryRequirements = gfx9_GetImageMemoryRequirements,
    .BindImageMemory = gfx9_BindImageMemory,
    .GetImageSparseMemoryRequirements = gfx9_GetImageSparseMemoryRequirements,
    .QueueBindSparse = gfx9_QueueBindSparse,
    .CreateFence = gfx9_CreateFence,
    .DestroyFence = gfx9_DestroyFence,
    .ResetFences = gfx9_ResetFences,
    .GetFenceStatus = gfx9_GetFenceStatus,
    .WaitForFences = gfx9_WaitForFences,
    .CreateSemaphore = gfx9_CreateSemaphore,
    .DestroySemaphore = gfx9_DestroySemaphore,
    .CreateEvent = gfx9_CreateEvent,
    .DestroyEvent = gfx9_DestroyEvent,
    .GetEventStatus = gfx9_GetEventStatus,
    .SetEvent = gfx9_SetEvent,
    .ResetEvent = gfx9_ResetEvent,
    .CreateQueryPool = gfx9_CreateQueryPool,
    .DestroyQueryPool = gfx9_DestroyQueryPool,
    .GetQueryPoolResults = gfx9_GetQueryPoolResults,
    .ResetQueryPool = gfx9_ResetQueryPool,
    .ResetQueryPoolEXT = gfx9_ResetQueryPoolEXT,
    .CreateBuffer = gfx9_CreateBuffer,
    .DestroyBuffer = gfx9_DestroyBuffer,
    .CreateBufferView = gfx9_CreateBufferView,
    .DestroyBufferView = gfx9_DestroyBufferView,
    .CreateImage = gfx9_CreateImage,
    .DestroyImage = gfx9_DestroyImage,
    .GetImageSubresourceLayout = gfx9_GetImageSubresourceLayout,
    .CreateImageView = gfx9_CreateImageView,
    .DestroyImageView = gfx9_DestroyImageView,
    .CreateShaderModule = gfx9_CreateShaderModule,
    .DestroyShaderModule = gfx9_DestroyShaderModule,
    .CreatePipelineCache = gfx9_CreatePipelineCache,
    .DestroyPipelineCache = gfx9_DestroyPipelineCache,
    .GetPipelineCacheData = gfx9_GetPipelineCacheData,
    .MergePipelineCaches = gfx9_MergePipelineCaches,
    .CreatePipelineBinariesKHR = gfx9_CreatePipelineBinariesKHR,
    .DestroyPipelineBinaryKHR = gfx9_DestroyPipelineBinaryKHR,
    .GetPipelineKeyKHR = gfx9_GetPipelineKeyKHR,
    .GetPipelineBinaryDataKHR = gfx9_GetPipelineBinaryDataKHR,
    .ReleaseCapturedPipelineDataKHR = gfx9_ReleaseCapturedPipelineDataKHR,
    .CreateGraphicsPipelines = gfx9_CreateGraphicsPipelines,
    .CreateComputePipelines = gfx9_CreateComputePipelines,
    .GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI = gfx9_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI,
    .DestroyPipeline = gfx9_DestroyPipeline,
    .CreatePipelineLayout = gfx9_CreatePipelineLayout,
    .DestroyPipelineLayout = gfx9_DestroyPipelineLayout,
    .CreateSampler = gfx9_CreateSampler,
    .DestroySampler = gfx9_DestroySampler,
    .CreateDescriptorSetLayout = gfx9_CreateDescriptorSetLayout,
    .DestroyDescriptorSetLayout = gfx9_DestroyDescriptorSetLayout,
    .CreateDescriptorPool = gfx9_CreateDescriptorPool,
    .DestroyDescriptorPool = gfx9_DestroyDescriptorPool,
    .ResetDescriptorPool = gfx9_ResetDescriptorPool,
    .AllocateDescriptorSets = gfx9_AllocateDescriptorSets,
    .FreeDescriptorSets = gfx9_FreeDescriptorSets,
    .UpdateDescriptorSets = gfx9_UpdateDescriptorSets,
    .CreateFramebuffer = gfx9_CreateFramebuffer,
    .DestroyFramebuffer = gfx9_DestroyFramebuffer,
    .CreateRenderPass = gfx9_CreateRenderPass,
    .DestroyRenderPass = gfx9_DestroyRenderPass,
    .GetRenderAreaGranularity = gfx9_GetRenderAreaGranularity,
    .GetRenderingAreaGranularity = gfx9_GetRenderingAreaGranularity,
    .GetRenderingAreaGranularityKHR = gfx9_GetRenderingAreaGranularityKHR,
    .CreateCommandPool = gfx9_CreateCommandPool,
    .DestroyCommandPool = gfx9_DestroyCommandPool,
    .ResetCommandPool = gfx9_ResetCommandPool,
    .AllocateCommandBuffers = gfx9_AllocateCommandBuffers,
    .FreeCommandBuffers = gfx9_FreeCommandBuffers,
    .BeginCommandBuffer = gfx9_BeginCommandBuffer,
    .EndCommandBuffer = gfx9_EndCommandBuffer,
    .ResetCommandBuffer = gfx9_ResetCommandBuffer,
    .CmdBindPipeline = gfx9_CmdBindPipeline,
    .CmdSetPrimitiveRestartIndexEXT = gfx9_CmdSetPrimitiveRestartIndexEXT,
    .CmdSetAttachmentFeedbackLoopEnableEXT = gfx9_CmdSetAttachmentFeedbackLoopEnableEXT,
    .CmdSetViewport = gfx9_CmdSetViewport,
    .CmdSetScissor = gfx9_CmdSetScissor,
    .CmdSetLineWidth = gfx9_CmdSetLineWidth,
    .CmdSetDepthBias = gfx9_CmdSetDepthBias,
    .CmdSetBlendConstants = gfx9_CmdSetBlendConstants,
    .CmdSetDepthBounds = gfx9_CmdSetDepthBounds,
    .CmdSetStencilCompareMask = gfx9_CmdSetStencilCompareMask,
    .CmdSetStencilWriteMask = gfx9_CmdSetStencilWriteMask,
    .CmdSetStencilReference = gfx9_CmdSetStencilReference,
    .CmdBindDescriptorSets = gfx9_CmdBindDescriptorSets,
    .CmdBindIndexBuffer = gfx9_CmdBindIndexBuffer,
    .CmdBindVertexBuffers = gfx9_CmdBindVertexBuffers,
    .CmdDraw = gfx9_CmdDraw,
    .CmdDrawIndexed = gfx9_CmdDrawIndexed,
    .CmdDrawMultiEXT = gfx9_CmdDrawMultiEXT,
    .CmdDrawMultiIndexedEXT = gfx9_CmdDrawMultiIndexedEXT,
    .CmdDrawIndirect = gfx9_CmdDrawIndirect,
    .CmdDrawIndexedIndirect = gfx9_CmdDrawIndexedIndirect,
    .CmdDispatch = gfx9_CmdDispatch,
    .CmdDispatchIndirect = gfx9_CmdDispatchIndirect,
    .CmdSubpassShadingHUAWEI = gfx9_CmdSubpassShadingHUAWEI,
    .CmdDrawClusterHUAWEI = gfx9_CmdDrawClusterHUAWEI,
    .CmdDrawClusterIndirectHUAWEI = gfx9_CmdDrawClusterIndirectHUAWEI,
    .CmdUpdatePipelineIndirectBufferNV = gfx9_CmdUpdatePipelineIndirectBufferNV,
    .CmdCopyBuffer = gfx9_CmdCopyBuffer,
    .CmdCopyImage = gfx9_CmdCopyImage,
    .CmdBlitImage = gfx9_CmdBlitImage,
    .CmdCopyBufferToImage = gfx9_CmdCopyBufferToImage,
    .CmdCopyImageToBuffer = gfx9_CmdCopyImageToBuffer,
    .CmdCopyMemoryIndirectNV = gfx9_CmdCopyMemoryIndirectNV,
    .CmdCopyMemoryIndirectKHR = gfx9_CmdCopyMemoryIndirectKHR,
    .CmdCopyMemoryToImageIndirectNV = gfx9_CmdCopyMemoryToImageIndirectNV,
    .CmdCopyMemoryToImageIndirectKHR = gfx9_CmdCopyMemoryToImageIndirectKHR,
    .CmdUpdateBuffer = gfx9_CmdUpdateBuffer,
    .CmdFillBuffer = gfx9_CmdFillBuffer,
    .CmdClearColorImage = gfx9_CmdClearColorImage,
    .CmdClearDepthStencilImage = gfx9_CmdClearDepthStencilImage,
    .CmdClearAttachments = gfx9_CmdClearAttachments,
    .CmdResolveImage = gfx9_CmdResolveImage,
    .CmdSetEvent = gfx9_CmdSetEvent,
    .CmdResetEvent = gfx9_CmdResetEvent,
    .CmdWaitEvents = gfx9_CmdWaitEvents,
    .CmdPipelineBarrier = gfx9_CmdPipelineBarrier,
    .CmdBeginQuery = gfx9_CmdBeginQuery,
    .CmdEndQuery = gfx9_CmdEndQuery,
    .CmdBeginConditionalRenderingEXT = gfx9_CmdBeginConditionalRenderingEXT,
    .CmdEndConditionalRenderingEXT = gfx9_CmdEndConditionalRenderingEXT,
    .CmdBeginCustomResolveEXT = gfx9_CmdBeginCustomResolveEXT,
    .CmdResetQueryPool = gfx9_CmdResetQueryPool,
    .CmdWriteTimestamp = gfx9_CmdWriteTimestamp,
    .CmdCopyQueryPoolResults = gfx9_CmdCopyQueryPoolResults,
    .CmdPushConstants = gfx9_CmdPushConstants,
    .CmdBeginRenderPass = gfx9_CmdBeginRenderPass,
    .CmdNextSubpass = gfx9_CmdNextSubpass,
    .CmdEndRenderPass = gfx9_CmdEndRenderPass,
    .CmdExecuteCommands = gfx9_CmdExecuteCommands,
    .CreateSharedSwapchainsKHR = gfx9_CreateSharedSwapchainsKHR,
    .CreateSwapchainKHR = gfx9_CreateSwapchainKHR,
    .DestroySwapchainKHR = gfx9_DestroySwapchainKHR,
    .GetSwapchainImagesKHR = gfx9_GetSwapchainImagesKHR,
    .AcquireNextImageKHR = gfx9_AcquireNextImageKHR,
    .QueuePresentKHR = gfx9_QueuePresentKHR,
    .DebugMarkerSetObjectNameEXT = gfx9_DebugMarkerSetObjectNameEXT,
    .DebugMarkerSetObjectTagEXT = gfx9_DebugMarkerSetObjectTagEXT,
    .CmdDebugMarkerBeginEXT = gfx9_CmdDebugMarkerBeginEXT,
    .CmdDebugMarkerEndEXT = gfx9_CmdDebugMarkerEndEXT,
    .CmdDebugMarkerInsertEXT = gfx9_CmdDebugMarkerInsertEXT,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleNV = gfx9_GetMemoryWin32HandleNV,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .CmdExecuteGeneratedCommandsNV = gfx9_CmdExecuteGeneratedCommandsNV,
    .CmdPreprocessGeneratedCommandsNV = gfx9_CmdPreprocessGeneratedCommandsNV,
    .CmdBindPipelineShaderGroupNV = gfx9_CmdBindPipelineShaderGroupNV,
    .GetGeneratedCommandsMemoryRequirementsNV = gfx9_GetGeneratedCommandsMemoryRequirementsNV,
    .CreateIndirectCommandsLayoutNV = gfx9_CreateIndirectCommandsLayoutNV,
    .DestroyIndirectCommandsLayoutNV = gfx9_DestroyIndirectCommandsLayoutNV,
    .CmdExecuteGeneratedCommandsEXT = gfx9_CmdExecuteGeneratedCommandsEXT,
    .CmdPreprocessGeneratedCommandsEXT = gfx9_CmdPreprocessGeneratedCommandsEXT,
    .GetGeneratedCommandsMemoryRequirementsEXT = gfx9_GetGeneratedCommandsMemoryRequirementsEXT,
    .CreateIndirectCommandsLayoutEXT = gfx9_CreateIndirectCommandsLayoutEXT,
    .DestroyIndirectCommandsLayoutEXT = gfx9_DestroyIndirectCommandsLayoutEXT,
    .CreateIndirectExecutionSetEXT = gfx9_CreateIndirectExecutionSetEXT,
    .DestroyIndirectExecutionSetEXT = gfx9_DestroyIndirectExecutionSetEXT,
    .UpdateIndirectExecutionSetPipelineEXT = gfx9_UpdateIndirectExecutionSetPipelineEXT,
    .UpdateIndirectExecutionSetShaderEXT = gfx9_UpdateIndirectExecutionSetShaderEXT,
    .CmdPushDescriptorSet = gfx9_CmdPushDescriptorSet,
    .CmdPushDescriptorSetKHR = gfx9_CmdPushDescriptorSetKHR,
    .TrimCommandPool = gfx9_TrimCommandPool,
    .TrimCommandPoolKHR = gfx9_TrimCommandPoolKHR,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleKHR = gfx9_GetMemoryWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandlePropertiesKHR = gfx9_GetMemoryWin32HandlePropertiesKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandlePropertiesKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryFdKHR = gfx9_GetMemoryFdKHR,
    .GetMemoryFdPropertiesKHR = gfx9_GetMemoryFdPropertiesKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandleFUCHSIA = gfx9_GetMemoryZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandlePropertiesFUCHSIA = gfx9_GetMemoryZirconHandlePropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandlePropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .GetMemoryRemoteAddressNV = gfx9_GetMemoryRemoteAddressNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreWin32HandleKHR = gfx9_GetSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportSemaphoreWin32HandleKHR = gfx9_ImportSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreFdKHR = gfx9_GetSemaphoreFdKHR,
    .ImportSemaphoreFdKHR = gfx9_ImportSemaphoreFdKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetSemaphoreZirconHandleFUCHSIA = gfx9_GetSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .ImportSemaphoreZirconHandleFUCHSIA = gfx9_ImportSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .ImportSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetFenceWin32HandleKHR = gfx9_GetFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportFenceWin32HandleKHR = gfx9_ImportFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetFenceFdKHR = gfx9_GetFenceFdKHR,
    .ImportFenceFdKHR = gfx9_ImportFenceFdKHR,
    .DisplayPowerControlEXT = gfx9_DisplayPowerControlEXT,
    .RegisterDeviceEventEXT = gfx9_RegisterDeviceEventEXT,
    .RegisterDisplayEventEXT = gfx9_RegisterDisplayEventEXT,
    .GetSwapchainCounterEXT = gfx9_GetSwapchainCounterEXT,
    .GetDeviceGroupPeerMemoryFeatures = gfx9_GetDeviceGroupPeerMemoryFeatures,
    .GetDeviceGroupPeerMemoryFeaturesKHR = gfx9_GetDeviceGroupPeerMemoryFeaturesKHR,
    .BindBufferMemory2 = gfx9_BindBufferMemory2,
    .BindBufferMemory2KHR = gfx9_BindBufferMemory2KHR,
    .BindImageMemory2 = gfx9_BindImageMemory2,
    .BindImageMemory2KHR = gfx9_BindImageMemory2KHR,
    .CmdSetDeviceMask = gfx9_CmdSetDeviceMask,
    .CmdSetDeviceMaskKHR = gfx9_CmdSetDeviceMaskKHR,
    .GetDeviceGroupPresentCapabilitiesKHR = gfx9_GetDeviceGroupPresentCapabilitiesKHR,
    .GetDeviceGroupSurfacePresentModesKHR = gfx9_GetDeviceGroupSurfacePresentModesKHR,
    .AcquireNextImage2KHR = gfx9_AcquireNextImage2KHR,
    .CmdDispatchBase = gfx9_CmdDispatchBase,
    .CmdDispatchBaseKHR = gfx9_CmdDispatchBaseKHR,
    .CreateDescriptorUpdateTemplate = gfx9_CreateDescriptorUpdateTemplate,
    .CreateDescriptorUpdateTemplateKHR = gfx9_CreateDescriptorUpdateTemplateKHR,
    .DestroyDescriptorUpdateTemplate = gfx9_DestroyDescriptorUpdateTemplate,
    .DestroyDescriptorUpdateTemplateKHR = gfx9_DestroyDescriptorUpdateTemplateKHR,
    .UpdateDescriptorSetWithTemplate = gfx9_UpdateDescriptorSetWithTemplate,
    .UpdateDescriptorSetWithTemplateKHR = gfx9_UpdateDescriptorSetWithTemplateKHR,
    .CmdPushDescriptorSetWithTemplate = gfx9_CmdPushDescriptorSetWithTemplate,
    .CmdPushDescriptorSetWithTemplateKHR = gfx9_CmdPushDescriptorSetWithTemplateKHR,
    .SetHdrMetadataEXT = gfx9_SetHdrMetadataEXT,
    .GetSwapchainStatusKHR = gfx9_GetSwapchainStatusKHR,
    .GetRefreshCycleDurationGOOGLE = gfx9_GetRefreshCycleDurationGOOGLE,
    .GetPastPresentationTimingGOOGLE = gfx9_GetPastPresentationTimingGOOGLE,
    .CmdSetViewportWScalingNV = gfx9_CmdSetViewportWScalingNV,
    .CmdSetDiscardRectangleEXT = gfx9_CmdSetDiscardRectangleEXT,
    .CmdSetDiscardRectangleEnableEXT = gfx9_CmdSetDiscardRectangleEnableEXT,
    .CmdSetDiscardRectangleModeEXT = gfx9_CmdSetDiscardRectangleModeEXT,
    .CmdSetSampleLocationsEXT = gfx9_CmdSetSampleLocationsEXT,
    .GetBufferMemoryRequirements2 = gfx9_GetBufferMemoryRequirements2,
    .GetBufferMemoryRequirements2KHR = gfx9_GetBufferMemoryRequirements2KHR,
    .GetImageMemoryRequirements2 = gfx9_GetImageMemoryRequirements2,
    .GetImageMemoryRequirements2KHR = gfx9_GetImageMemoryRequirements2KHR,
    .GetImageSparseMemoryRequirements2 = gfx9_GetImageSparseMemoryRequirements2,
    .GetImageSparseMemoryRequirements2KHR = gfx9_GetImageSparseMemoryRequirements2KHR,
    .GetDeviceBufferMemoryRequirements = gfx9_GetDeviceBufferMemoryRequirements,
    .GetDeviceBufferMemoryRequirementsKHR = gfx9_GetDeviceBufferMemoryRequirementsKHR,
    .GetDeviceImageMemoryRequirements = gfx9_GetDeviceImageMemoryRequirements,
    .GetDeviceImageMemoryRequirementsKHR = gfx9_GetDeviceImageMemoryRequirementsKHR,
    .GetDeviceImageSparseMemoryRequirements = gfx9_GetDeviceImageSparseMemoryRequirements,
    .GetDeviceImageSparseMemoryRequirementsKHR = gfx9_GetDeviceImageSparseMemoryRequirementsKHR,
    .CreateSamplerYcbcrConversion = gfx9_CreateSamplerYcbcrConversion,
    .CreateSamplerYcbcrConversionKHR = gfx9_CreateSamplerYcbcrConversionKHR,
    .DestroySamplerYcbcrConversion = gfx9_DestroySamplerYcbcrConversion,
    .DestroySamplerYcbcrConversionKHR = gfx9_DestroySamplerYcbcrConversionKHR,
    .GetDeviceQueue2 = gfx9_GetDeviceQueue2,
    .CreateValidationCacheEXT = gfx9_CreateValidationCacheEXT,
    .DestroyValidationCacheEXT = gfx9_DestroyValidationCacheEXT,
    .GetValidationCacheDataEXT = gfx9_GetValidationCacheDataEXT,
    .MergeValidationCachesEXT = gfx9_MergeValidationCachesEXT,
    .GetDescriptorSetLayoutSupport = gfx9_GetDescriptorSetLayoutSupport,
    .GetDescriptorSetLayoutSupportKHR = gfx9_GetDescriptorSetLayoutSupportKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsageANDROID = gfx9_GetSwapchainGrallocUsageANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsage2ANDROID = gfx9_GetSwapchainGrallocUsage2ANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsage2ANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .AcquireImageANDROID = gfx9_AcquireImageANDROID,
#elif defined(_MSC_VER)
    .AcquireImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .QueueSignalReleaseImageANDROID = gfx9_QueueSignalReleaseImageANDROID,
#elif defined(_MSC_VER)
    .QueueSignalReleaseImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .GetShaderInfoAMD = gfx9_GetShaderInfoAMD,
    .SetLocalDimmingAMD = gfx9_SetLocalDimmingAMD,
    .GetCalibratedTimestampsKHR = gfx9_GetCalibratedTimestampsKHR,
    .GetCalibratedTimestampsEXT = gfx9_GetCalibratedTimestampsEXT,
    .SetDebugUtilsObjectNameEXT = gfx9_SetDebugUtilsObjectNameEXT,
    .SetDebugUtilsObjectTagEXT = gfx9_SetDebugUtilsObjectTagEXT,
    .QueueBeginDebugUtilsLabelEXT = gfx9_QueueBeginDebugUtilsLabelEXT,
    .QueueEndDebugUtilsLabelEXT = gfx9_QueueEndDebugUtilsLabelEXT,
    .QueueInsertDebugUtilsLabelEXT = gfx9_QueueInsertDebugUtilsLabelEXT,
    .CmdBeginDebugUtilsLabelEXT = gfx9_CmdBeginDebugUtilsLabelEXT,
    .CmdEndDebugUtilsLabelEXT = gfx9_CmdEndDebugUtilsLabelEXT,
    .CmdInsertDebugUtilsLabelEXT = gfx9_CmdInsertDebugUtilsLabelEXT,
    .GetMemoryHostPointerPropertiesEXT = gfx9_GetMemoryHostPointerPropertiesEXT,
    .CmdWriteBufferMarkerAMD = gfx9_CmdWriteBufferMarkerAMD,
    .CreateRenderPass2 = gfx9_CreateRenderPass2,
    .CreateRenderPass2KHR = gfx9_CreateRenderPass2KHR,
    .CmdBeginRenderPass2 = gfx9_CmdBeginRenderPass2,
    .CmdBeginRenderPass2KHR = gfx9_CmdBeginRenderPass2KHR,
    .CmdNextSubpass2 = gfx9_CmdNextSubpass2,
    .CmdNextSubpass2KHR = gfx9_CmdNextSubpass2KHR,
    .CmdEndRenderPass2 = gfx9_CmdEndRenderPass2,
    .CmdEndRenderPass2KHR = gfx9_CmdEndRenderPass2KHR,
    .GetSemaphoreCounterValue = gfx9_GetSemaphoreCounterValue,
    .GetSemaphoreCounterValueKHR = gfx9_GetSemaphoreCounterValueKHR,
    .WaitSemaphores = gfx9_WaitSemaphores,
    .WaitSemaphoresKHR = gfx9_WaitSemaphoresKHR,
    .SignalSemaphore = gfx9_SignalSemaphore,
    .SignalSemaphoreKHR = gfx9_SignalSemaphoreKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetAndroidHardwareBufferPropertiesANDROID = gfx9_GetAndroidHardwareBufferPropertiesANDROID,
#elif defined(_MSC_VER)
    .GetAndroidHardwareBufferPropertiesANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetMemoryAndroidHardwareBufferANDROID = gfx9_GetMemoryAndroidHardwareBufferANDROID,
#elif defined(_MSC_VER)
    .GetMemoryAndroidHardwareBufferANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .CmdDrawIndirectCount = gfx9_CmdDrawIndirectCount,
    .CmdDrawIndirectCountKHR = gfx9_CmdDrawIndirectCountKHR,
    .CmdDrawIndirectCountAMD = gfx9_CmdDrawIndirectCountAMD,
    .CmdDrawIndexedIndirectCount = gfx9_CmdDrawIndexedIndirectCount,
    .CmdDrawIndexedIndirectCountKHR = gfx9_CmdDrawIndexedIndirectCountKHR,
    .CmdDrawIndexedIndirectCountAMD = gfx9_CmdDrawIndexedIndirectCountAMD,
    .CmdSetCheckpointNV = gfx9_CmdSetCheckpointNV,
    .GetQueueCheckpointDataNV = gfx9_GetQueueCheckpointDataNV,
    .CmdBindTransformFeedbackBuffersEXT = gfx9_CmdBindTransformFeedbackBuffersEXT,
    .CmdBeginTransformFeedbackEXT = gfx9_CmdBeginTransformFeedbackEXT,
    .CmdEndTransformFeedbackEXT = gfx9_CmdEndTransformFeedbackEXT,
    .CmdBeginQueryIndexedEXT = gfx9_CmdBeginQueryIndexedEXT,
    .CmdEndQueryIndexedEXT = gfx9_CmdEndQueryIndexedEXT,
    .CmdDrawIndirectByteCountEXT = gfx9_CmdDrawIndirectByteCountEXT,
    .CmdSetExclusiveScissorNV = gfx9_CmdSetExclusiveScissorNV,
    .CmdSetExclusiveScissorEnableNV = gfx9_CmdSetExclusiveScissorEnableNV,
    .CmdBindShadingRateImageNV = gfx9_CmdBindShadingRateImageNV,
    .CmdSetViewportShadingRatePaletteNV = gfx9_CmdSetViewportShadingRatePaletteNV,
    .CmdSetCoarseSampleOrderNV = gfx9_CmdSetCoarseSampleOrderNV,
    .CmdDrawMeshTasksNV = gfx9_CmdDrawMeshTasksNV,
    .CmdDrawMeshTasksIndirectNV = gfx9_CmdDrawMeshTasksIndirectNV,
    .CmdDrawMeshTasksIndirectCountNV = gfx9_CmdDrawMeshTasksIndirectCountNV,
    .CmdDrawMeshTasksEXT = gfx9_CmdDrawMeshTasksEXT,
    .CmdDrawMeshTasksIndirectEXT = gfx9_CmdDrawMeshTasksIndirectEXT,
    .CmdDrawMeshTasksIndirectCountEXT = gfx9_CmdDrawMeshTasksIndirectCountEXT,
    .CompileDeferredNV = gfx9_CompileDeferredNV,
    .CreateAccelerationStructureNV = gfx9_CreateAccelerationStructureNV,
    .CmdBindInvocationMaskHUAWEI = gfx9_CmdBindInvocationMaskHUAWEI,
    .DestroyAccelerationStructureKHR = gfx9_DestroyAccelerationStructureKHR,
    .DestroyAccelerationStructureNV = gfx9_DestroyAccelerationStructureNV,
    .GetAccelerationStructureMemoryRequirementsNV = gfx9_GetAccelerationStructureMemoryRequirementsNV,
    .BindAccelerationStructureMemoryNV = gfx9_BindAccelerationStructureMemoryNV,
    .CmdCopyAccelerationStructureNV = gfx9_CmdCopyAccelerationStructureNV,
    .CmdCopyAccelerationStructureKHR = gfx9_CmdCopyAccelerationStructureKHR,
    .CopyAccelerationStructureKHR = gfx9_CopyAccelerationStructureKHR,
    .CmdCopyAccelerationStructureToMemoryKHR = gfx9_CmdCopyAccelerationStructureToMemoryKHR,
    .CopyAccelerationStructureToMemoryKHR = gfx9_CopyAccelerationStructureToMemoryKHR,
    .CmdCopyMemoryToAccelerationStructureKHR = gfx9_CmdCopyMemoryToAccelerationStructureKHR,
    .CopyMemoryToAccelerationStructureKHR = gfx9_CopyMemoryToAccelerationStructureKHR,
    .CmdWriteAccelerationStructuresPropertiesKHR = gfx9_CmdWriteAccelerationStructuresPropertiesKHR,
    .CmdWriteAccelerationStructuresPropertiesNV = gfx9_CmdWriteAccelerationStructuresPropertiesNV,
    .CmdBuildAccelerationStructureNV = gfx9_CmdBuildAccelerationStructureNV,
    .WriteAccelerationStructuresPropertiesKHR = gfx9_WriteAccelerationStructuresPropertiesKHR,
    .CmdTraceRaysKHR = gfx9_CmdTraceRaysKHR,
    .CmdTraceRaysNV = gfx9_CmdTraceRaysNV,
    .GetRayTracingShaderGroupHandlesKHR = gfx9_GetRayTracingShaderGroupHandlesKHR,
    .GetRayTracingShaderGroupHandlesNV = gfx9_GetRayTracingShaderGroupHandlesNV,
    .GetRayTracingCaptureReplayShaderGroupHandlesKHR = gfx9_GetRayTracingCaptureReplayShaderGroupHandlesKHR,
    .GetAccelerationStructureHandleNV = gfx9_GetAccelerationStructureHandleNV,
    .CreateRayTracingPipelinesNV = gfx9_CreateRayTracingPipelinesNV,
    .CreateRayTracingPipelinesKHR = gfx9_CreateRayTracingPipelinesKHR,
    .CmdTraceRaysIndirectKHR = gfx9_CmdTraceRaysIndirectKHR,
    .CmdTraceRaysIndirect2KHR = gfx9_CmdTraceRaysIndirect2KHR,
    .GetClusterAccelerationStructureBuildSizesNV = gfx9_GetClusterAccelerationStructureBuildSizesNV,
    .CmdBuildClusterAccelerationStructureIndirectNV = gfx9_CmdBuildClusterAccelerationStructureIndirectNV,
    .GetDeviceAccelerationStructureCompatibilityKHR = gfx9_GetDeviceAccelerationStructureCompatibilityKHR,
    .GetRayTracingShaderGroupStackSizeKHR = gfx9_GetRayTracingShaderGroupStackSizeKHR,
    .CmdSetRayTracingPipelineStackSizeKHR = gfx9_CmdSetRayTracingPipelineStackSizeKHR,
    .GetImageViewHandleNVX = gfx9_GetImageViewHandleNVX,
    .GetImageViewHandle64NVX = gfx9_GetImageViewHandle64NVX,
    .GetImageViewAddressNVX = gfx9_GetImageViewAddressNVX,
    .GetDeviceCombinedImageSamplerIndexNVX = gfx9_GetDeviceCombinedImageSamplerIndexNVX,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetDeviceGroupSurfacePresentModes2EXT = gfx9_GetDeviceGroupSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetDeviceGroupSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireFullScreenExclusiveModeEXT = gfx9_AcquireFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .AcquireFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ReleaseFullScreenExclusiveModeEXT = gfx9_ReleaseFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .ReleaseFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .AcquireProfilingLockKHR = gfx9_AcquireProfilingLockKHR,
    .ReleaseProfilingLockKHR = gfx9_ReleaseProfilingLockKHR,
    .GetImageDrmFormatModifierPropertiesEXT = gfx9_GetImageDrmFormatModifierPropertiesEXT,
    .GetBufferOpaqueCaptureAddress = gfx9_GetBufferOpaqueCaptureAddress,
    .GetBufferOpaqueCaptureAddressKHR = gfx9_GetBufferOpaqueCaptureAddressKHR,
    .GetBufferDeviceAddress = gfx9_GetBufferDeviceAddress,
    .GetBufferDeviceAddressKHR = gfx9_GetBufferDeviceAddressKHR,
    .GetBufferDeviceAddressEXT = gfx9_GetBufferDeviceAddressEXT,
    .InitializePerformanceApiINTEL = gfx9_InitializePerformanceApiINTEL,
    .UninitializePerformanceApiINTEL = gfx9_UninitializePerformanceApiINTEL,
    .CmdSetPerformanceMarkerINTEL = gfx9_CmdSetPerformanceMarkerINTEL,
    .CmdSetPerformanceStreamMarkerINTEL = gfx9_CmdSetPerformanceStreamMarkerINTEL,
    .CmdSetPerformanceOverrideINTEL = gfx9_CmdSetPerformanceOverrideINTEL,
    .AcquirePerformanceConfigurationINTEL = gfx9_AcquirePerformanceConfigurationINTEL,
    .ReleasePerformanceConfigurationINTEL = gfx9_ReleasePerformanceConfigurationINTEL,
    .QueueSetPerformanceConfigurationINTEL = gfx9_QueueSetPerformanceConfigurationINTEL,
    .GetPerformanceParameterINTEL = gfx9_GetPerformanceParameterINTEL,
    .GetDeviceMemoryOpaqueCaptureAddress = gfx9_GetDeviceMemoryOpaqueCaptureAddress,
    .GetDeviceMemoryOpaqueCaptureAddressKHR = gfx9_GetDeviceMemoryOpaqueCaptureAddressKHR,
    .GetPipelineExecutablePropertiesKHR = gfx9_GetPipelineExecutablePropertiesKHR,
    .GetPipelineExecutableStatisticsKHR = gfx9_GetPipelineExecutableStatisticsKHR,
    .GetPipelineExecutableInternalRepresentationsKHR = gfx9_GetPipelineExecutableInternalRepresentationsKHR,
    .CmdSetLineStipple = gfx9_CmdSetLineStipple,
    .CmdSetLineStippleKHR = gfx9_CmdSetLineStippleKHR,
    .CmdSetLineStippleEXT = gfx9_CmdSetLineStippleEXT,
    .CreateAccelerationStructureKHR = gfx9_CreateAccelerationStructureKHR,
    .CmdBuildAccelerationStructuresKHR = gfx9_CmdBuildAccelerationStructuresKHR,
    .CmdBuildAccelerationStructuresIndirectKHR = gfx9_CmdBuildAccelerationStructuresIndirectKHR,
    .BuildAccelerationStructuresKHR = gfx9_BuildAccelerationStructuresKHR,
    .GetAccelerationStructureDeviceAddressKHR = gfx9_GetAccelerationStructureDeviceAddressKHR,
    .CreateDeferredOperationKHR = gfx9_CreateDeferredOperationKHR,
    .DestroyDeferredOperationKHR = gfx9_DestroyDeferredOperationKHR,
    .GetDeferredOperationMaxConcurrencyKHR = gfx9_GetDeferredOperationMaxConcurrencyKHR,
    .GetDeferredOperationResultKHR = gfx9_GetDeferredOperationResultKHR,
    .DeferredOperationJoinKHR = gfx9_DeferredOperationJoinKHR,
    .GetPipelineIndirectMemoryRequirementsNV = gfx9_GetPipelineIndirectMemoryRequirementsNV,
    .GetPipelineIndirectDeviceAddressNV = gfx9_GetPipelineIndirectDeviceAddressNV,
    .AntiLagUpdateAMD = gfx9_AntiLagUpdateAMD,
    .CmdSetCullMode = gfx9_CmdSetCullMode,
    .CmdSetCullModeEXT = gfx9_CmdSetCullModeEXT,
    .CmdSetFrontFace = gfx9_CmdSetFrontFace,
    .CmdSetFrontFaceEXT = gfx9_CmdSetFrontFaceEXT,
    .CmdSetPrimitiveTopology = gfx9_CmdSetPrimitiveTopology,
    .CmdSetPrimitiveTopologyEXT = gfx9_CmdSetPrimitiveTopologyEXT,
    .CmdSetViewportWithCount = gfx9_CmdSetViewportWithCount,
    .CmdSetViewportWithCountEXT = gfx9_CmdSetViewportWithCountEXT,
    .CmdSetScissorWithCount = gfx9_CmdSetScissorWithCount,
    .CmdSetScissorWithCountEXT = gfx9_CmdSetScissorWithCountEXT,
    .CmdBindIndexBuffer2 = gfx9_CmdBindIndexBuffer2,
    .CmdBindIndexBuffer2KHR = gfx9_CmdBindIndexBuffer2KHR,
    .CmdBindVertexBuffers2 = gfx9_CmdBindVertexBuffers2,
    .CmdBindVertexBuffers2EXT = gfx9_CmdBindVertexBuffers2EXT,
    .CmdSetDepthTestEnable = gfx9_CmdSetDepthTestEnable,
    .CmdSetDepthTestEnableEXT = gfx9_CmdSetDepthTestEnableEXT,
    .CmdSetDepthWriteEnable = gfx9_CmdSetDepthWriteEnable,
    .CmdSetDepthWriteEnableEXT = gfx9_CmdSetDepthWriteEnableEXT,
    .CmdSetDepthCompareOp = gfx9_CmdSetDepthCompareOp,
    .CmdSetDepthCompareOpEXT = gfx9_CmdSetDepthCompareOpEXT,
    .CmdSetDepthBoundsTestEnable = gfx9_CmdSetDepthBoundsTestEnable,
    .CmdSetDepthBoundsTestEnableEXT = gfx9_CmdSetDepthBoundsTestEnableEXT,
    .CmdSetStencilTestEnable = gfx9_CmdSetStencilTestEnable,
    .CmdSetStencilTestEnableEXT = gfx9_CmdSetStencilTestEnableEXT,
    .CmdSetStencilOp = gfx9_CmdSetStencilOp,
    .CmdSetStencilOpEXT = gfx9_CmdSetStencilOpEXT,
    .CmdSetPatchControlPointsEXT = gfx9_CmdSetPatchControlPointsEXT,
    .CmdSetRasterizerDiscardEnable = gfx9_CmdSetRasterizerDiscardEnable,
    .CmdSetRasterizerDiscardEnableEXT = gfx9_CmdSetRasterizerDiscardEnableEXT,
    .CmdSetDepthBiasEnable = gfx9_CmdSetDepthBiasEnable,
    .CmdSetDepthBiasEnableEXT = gfx9_CmdSetDepthBiasEnableEXT,
    .CmdSetLogicOpEXT = gfx9_CmdSetLogicOpEXT,
    .CmdSetPrimitiveRestartEnable = gfx9_CmdSetPrimitiveRestartEnable,
    .CmdSetPrimitiveRestartEnableEXT = gfx9_CmdSetPrimitiveRestartEnableEXT,
    .CmdSetTessellationDomainOriginEXT = gfx9_CmdSetTessellationDomainOriginEXT,
    .CmdSetDepthClampEnableEXT = gfx9_CmdSetDepthClampEnableEXT,
    .CmdSetPolygonModeEXT = gfx9_CmdSetPolygonModeEXT,
    .CmdSetRasterizationSamplesEXT = gfx9_CmdSetRasterizationSamplesEXT,
    .CmdSetSampleMaskEXT = gfx9_CmdSetSampleMaskEXT,
    .CmdSetAlphaToCoverageEnableEXT = gfx9_CmdSetAlphaToCoverageEnableEXT,
    .CmdSetAlphaToOneEnableEXT = gfx9_CmdSetAlphaToOneEnableEXT,
    .CmdSetLogicOpEnableEXT = gfx9_CmdSetLogicOpEnableEXT,
    .CmdSetColorBlendEnableEXT = gfx9_CmdSetColorBlendEnableEXT,
    .CmdSetColorBlendEquationEXT = gfx9_CmdSetColorBlendEquationEXT,
    .CmdSetColorWriteMaskEXT = gfx9_CmdSetColorWriteMaskEXT,
    .CmdSetRasterizationStreamEXT = gfx9_CmdSetRasterizationStreamEXT,
    .CmdSetConservativeRasterizationModeEXT = gfx9_CmdSetConservativeRasterizationModeEXT,
    .CmdSetExtraPrimitiveOverestimationSizeEXT = gfx9_CmdSetExtraPrimitiveOverestimationSizeEXT,
    .CmdSetDepthClipEnableEXT = gfx9_CmdSetDepthClipEnableEXT,
    .CmdSetSampleLocationsEnableEXT = gfx9_CmdSetSampleLocationsEnableEXT,
    .CmdSetColorBlendAdvancedEXT = gfx9_CmdSetColorBlendAdvancedEXT,
    .CmdSetProvokingVertexModeEXT = gfx9_CmdSetProvokingVertexModeEXT,
    .CmdSetLineRasterizationModeEXT = gfx9_CmdSetLineRasterizationModeEXT,
    .CmdSetLineStippleEnableEXT = gfx9_CmdSetLineStippleEnableEXT,
    .CmdSetDepthClipNegativeOneToOneEXT = gfx9_CmdSetDepthClipNegativeOneToOneEXT,
    .CmdSetViewportWScalingEnableNV = gfx9_CmdSetViewportWScalingEnableNV,
    .CmdSetViewportSwizzleNV = gfx9_CmdSetViewportSwizzleNV,
    .CmdSetCoverageToColorEnableNV = gfx9_CmdSetCoverageToColorEnableNV,
    .CmdSetCoverageToColorLocationNV = gfx9_CmdSetCoverageToColorLocationNV,
    .CmdSetCoverageModulationModeNV = gfx9_CmdSetCoverageModulationModeNV,
    .CmdSetCoverageModulationTableEnableNV = gfx9_CmdSetCoverageModulationTableEnableNV,
    .CmdSetCoverageModulationTableNV = gfx9_CmdSetCoverageModulationTableNV,
    .CmdSetShadingRateImageEnableNV = gfx9_CmdSetShadingRateImageEnableNV,
    .CmdSetCoverageReductionModeNV = gfx9_CmdSetCoverageReductionModeNV,
    .CmdSetRepresentativeFragmentTestEnableNV = gfx9_CmdSetRepresentativeFragmentTestEnableNV,
    .CreatePrivateDataSlot = gfx9_CreatePrivateDataSlot,
    .CreatePrivateDataSlotEXT = gfx9_CreatePrivateDataSlotEXT,
    .DestroyPrivateDataSlot = gfx9_DestroyPrivateDataSlot,
    .DestroyPrivateDataSlotEXT = gfx9_DestroyPrivateDataSlotEXT,
    .SetPrivateData = gfx9_SetPrivateData,
    .SetPrivateDataEXT = gfx9_SetPrivateDataEXT,
    .GetPrivateData = gfx9_GetPrivateData,
    .GetPrivateDataEXT = gfx9_GetPrivateDataEXT,
    .CmdCopyBuffer2 = gfx9_CmdCopyBuffer2,
    .CmdCopyBuffer2KHR = gfx9_CmdCopyBuffer2KHR,
    .CmdCopyImage2 = gfx9_CmdCopyImage2,
    .CmdCopyImage2KHR = gfx9_CmdCopyImage2KHR,
    .CmdBlitImage2 = gfx9_CmdBlitImage2,
    .CmdBlitImage2KHR = gfx9_CmdBlitImage2KHR,
    .CmdCopyBufferToImage2 = gfx9_CmdCopyBufferToImage2,
    .CmdCopyBufferToImage2KHR = gfx9_CmdCopyBufferToImage2KHR,
    .CmdCopyImageToBuffer2 = gfx9_CmdCopyImageToBuffer2,
    .CmdCopyImageToBuffer2KHR = gfx9_CmdCopyImageToBuffer2KHR,
    .CmdResolveImage2 = gfx9_CmdResolveImage2,
    .CmdResolveImage2KHR = gfx9_CmdResolveImage2KHR,
    .CmdSetFragmentShadingRateKHR = gfx9_CmdSetFragmentShadingRateKHR,
    .CmdSetFragmentShadingRateEnumNV = gfx9_CmdSetFragmentShadingRateEnumNV,
    .GetAccelerationStructureBuildSizesKHR = gfx9_GetAccelerationStructureBuildSizesKHR,
    .CmdSetVertexInputEXT = gfx9_CmdSetVertexInputEXT,
    .CmdSetColorWriteEnableEXT = gfx9_CmdSetColorWriteEnableEXT,
    .CmdSetEvent2 = gfx9_CmdSetEvent2,
    .CmdSetEvent2KHR = gfx9_CmdSetEvent2KHR,
    .CmdResetEvent2 = gfx9_CmdResetEvent2,
    .CmdResetEvent2KHR = gfx9_CmdResetEvent2KHR,
    .CmdWaitEvents2 = gfx9_CmdWaitEvents2,
    .CmdWaitEvents2KHR = gfx9_CmdWaitEvents2KHR,
    .CmdPipelineBarrier2 = gfx9_CmdPipelineBarrier2,
    .CmdPipelineBarrier2KHR = gfx9_CmdPipelineBarrier2KHR,
    .QueueSubmit2 = gfx9_QueueSubmit2,
    .QueueSubmit2KHR = gfx9_QueueSubmit2KHR,
    .CmdWriteTimestamp2 = gfx9_CmdWriteTimestamp2,
    .CmdWriteTimestamp2KHR = gfx9_CmdWriteTimestamp2KHR,
    .CmdWriteBufferMarker2AMD = gfx9_CmdWriteBufferMarker2AMD,
    .GetQueueCheckpointData2NV = gfx9_GetQueueCheckpointData2NV,
    .CopyMemoryToImage = gfx9_CopyMemoryToImage,
    .CopyMemoryToImageEXT = gfx9_CopyMemoryToImageEXT,
    .CopyImageToMemory = gfx9_CopyImageToMemory,
    .CopyImageToMemoryEXT = gfx9_CopyImageToMemoryEXT,
    .CopyImageToImage = gfx9_CopyImageToImage,
    .CopyImageToImageEXT = gfx9_CopyImageToImageEXT,
    .TransitionImageLayout = gfx9_TransitionImageLayout,
    .TransitionImageLayoutEXT = gfx9_TransitionImageLayoutEXT,
    .CreateVideoSessionKHR = gfx9_CreateVideoSessionKHR,
    .DestroyVideoSessionKHR = gfx9_DestroyVideoSessionKHR,
    .CreateVideoSessionParametersKHR = gfx9_CreateVideoSessionParametersKHR,
    .UpdateVideoSessionParametersKHR = gfx9_UpdateVideoSessionParametersKHR,
    .GetEncodedVideoSessionParametersKHR = gfx9_GetEncodedVideoSessionParametersKHR,
    .DestroyVideoSessionParametersKHR = gfx9_DestroyVideoSessionParametersKHR,
    .GetVideoSessionMemoryRequirementsKHR = gfx9_GetVideoSessionMemoryRequirementsKHR,
    .BindVideoSessionMemoryKHR = gfx9_BindVideoSessionMemoryKHR,
    .CmdDecodeVideoKHR = gfx9_CmdDecodeVideoKHR,
    .CmdBeginVideoCodingKHR = gfx9_CmdBeginVideoCodingKHR,
    .CmdControlVideoCodingKHR = gfx9_CmdControlVideoCodingKHR,
    .CmdEndVideoCodingKHR = gfx9_CmdEndVideoCodingKHR,
    .CmdEncodeVideoKHR = gfx9_CmdEncodeVideoKHR,
    .CmdDecompressMemoryNV = gfx9_CmdDecompressMemoryNV,
    .CmdDecompressMemoryIndirectCountNV = gfx9_CmdDecompressMemoryIndirectCountNV,
    .GetPartitionedAccelerationStructuresBuildSizesNV = gfx9_GetPartitionedAccelerationStructuresBuildSizesNV,
    .CmdBuildPartitionedAccelerationStructuresNV = gfx9_CmdBuildPartitionedAccelerationStructuresNV,
    .CmdDecompressMemoryEXT = gfx9_CmdDecompressMemoryEXT,
    .CmdDecompressMemoryIndirectCountEXT = gfx9_CmdDecompressMemoryIndirectCountEXT,
    .CreateCuModuleNVX = gfx9_CreateCuModuleNVX,
    .CreateCuFunctionNVX = gfx9_CreateCuFunctionNVX,
    .DestroyCuModuleNVX = gfx9_DestroyCuModuleNVX,
    .DestroyCuFunctionNVX = gfx9_DestroyCuFunctionNVX,
    .CmdCuLaunchKernelNVX = gfx9_CmdCuLaunchKernelNVX,
    .GetDescriptorSetLayoutSizeEXT = gfx9_GetDescriptorSetLayoutSizeEXT,
    .GetDescriptorSetLayoutBindingOffsetEXT = gfx9_GetDescriptorSetLayoutBindingOffsetEXT,
    .GetDescriptorEXT = gfx9_GetDescriptorEXT,
    .CmdBindDescriptorBuffersEXT = gfx9_CmdBindDescriptorBuffersEXT,
    .CmdSetDescriptorBufferOffsetsEXT = gfx9_CmdSetDescriptorBufferOffsetsEXT,
    .CmdBindDescriptorBufferEmbeddedSamplersEXT = gfx9_CmdBindDescriptorBufferEmbeddedSamplersEXT,
    .GetBufferOpaqueCaptureDescriptorDataEXT = gfx9_GetBufferOpaqueCaptureDescriptorDataEXT,
    .GetImageOpaqueCaptureDescriptorDataEXT = gfx9_GetImageOpaqueCaptureDescriptorDataEXT,
    .GetImageViewOpaqueCaptureDescriptorDataEXT = gfx9_GetImageViewOpaqueCaptureDescriptorDataEXT,
    .GetSamplerOpaqueCaptureDescriptorDataEXT = gfx9_GetSamplerOpaqueCaptureDescriptorDataEXT,
    .GetAccelerationStructureOpaqueCaptureDescriptorDataEXT = gfx9_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT,
    .SetDeviceMemoryPriorityEXT = gfx9_SetDeviceMemoryPriorityEXT,
    .WaitForPresent2KHR = gfx9_WaitForPresent2KHR,
    .WaitForPresentKHR = gfx9_WaitForPresentKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateBufferCollectionFUCHSIA = gfx9_CreateBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .CreateBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionBufferConstraintsFUCHSIA = gfx9_SetBufferCollectionBufferConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionBufferConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionImageConstraintsFUCHSIA = gfx9_SetBufferCollectionImageConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionImageConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .DestroyBufferCollectionFUCHSIA = gfx9_DestroyBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .DestroyBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetBufferCollectionPropertiesFUCHSIA = gfx9_GetBufferCollectionPropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetBufferCollectionPropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .CmdBeginRendering = gfx9_CmdBeginRendering,
    .CmdBeginRenderingKHR = gfx9_CmdBeginRenderingKHR,
    .CmdEndRendering = gfx9_CmdEndRendering,
    .CmdEndRendering2KHR = gfx9_CmdEndRendering2KHR,
    .CmdEndRendering2EXT = gfx9_CmdEndRendering2EXT,
    .CmdEndRenderingKHR = gfx9_CmdEndRenderingKHR,
    .GetDescriptorSetLayoutHostMappingInfoVALVE = gfx9_GetDescriptorSetLayoutHostMappingInfoVALVE,
    .GetDescriptorSetHostMappingVALVE = gfx9_GetDescriptorSetHostMappingVALVE,
    .CreateMicromapEXT = gfx9_CreateMicromapEXT,
    .CmdBuildMicromapsEXT = gfx9_CmdBuildMicromapsEXT,
    .BuildMicromapsEXT = gfx9_BuildMicromapsEXT,
    .DestroyMicromapEXT = gfx9_DestroyMicromapEXT,
    .CmdCopyMicromapEXT = gfx9_CmdCopyMicromapEXT,
    .CopyMicromapEXT = gfx9_CopyMicromapEXT,
    .CmdCopyMicromapToMemoryEXT = gfx9_CmdCopyMicromapToMemoryEXT,
    .CopyMicromapToMemoryEXT = gfx9_CopyMicromapToMemoryEXT,
    .CmdCopyMemoryToMicromapEXT = gfx9_CmdCopyMemoryToMicromapEXT,
    .CopyMemoryToMicromapEXT = gfx9_CopyMemoryToMicromapEXT,
    .CmdWriteMicromapsPropertiesEXT = gfx9_CmdWriteMicromapsPropertiesEXT,
    .WriteMicromapsPropertiesEXT = gfx9_WriteMicromapsPropertiesEXT,
    .GetDeviceMicromapCompatibilityEXT = gfx9_GetDeviceMicromapCompatibilityEXT,
    .GetMicromapBuildSizesEXT = gfx9_GetMicromapBuildSizesEXT,
    .GetShaderModuleIdentifierEXT = gfx9_GetShaderModuleIdentifierEXT,
    .GetShaderModuleCreateInfoIdentifierEXT = gfx9_GetShaderModuleCreateInfoIdentifierEXT,
    .GetImageSubresourceLayout2 = gfx9_GetImageSubresourceLayout2,
    .GetImageSubresourceLayout2KHR = gfx9_GetImageSubresourceLayout2KHR,
    .GetImageSubresourceLayout2EXT = gfx9_GetImageSubresourceLayout2EXT,
    .GetPipelinePropertiesEXT = gfx9_GetPipelinePropertiesEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .ExportMetalObjectsEXT = gfx9_ExportMetalObjectsEXT,
#elif defined(_MSC_VER)
    .ExportMetalObjectsEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CmdBindTileMemoryQCOM = gfx9_CmdBindTileMemoryQCOM,
    .GetFramebufferTilePropertiesQCOM = gfx9_GetFramebufferTilePropertiesQCOM,
    .GetDynamicRenderingTilePropertiesQCOM = gfx9_GetDynamicRenderingTilePropertiesQCOM,
    .CreateOpticalFlowSessionNV = gfx9_CreateOpticalFlowSessionNV,
    .DestroyOpticalFlowSessionNV = gfx9_DestroyOpticalFlowSessionNV,
    .BindOpticalFlowSessionImageNV = gfx9_BindOpticalFlowSessionImageNV,
    .CmdOpticalFlowExecuteNV = gfx9_CmdOpticalFlowExecuteNV,
    .GetDeviceFaultInfoEXT = gfx9_GetDeviceFaultInfoEXT,
    .GetDeviceFaultReportsKHR = gfx9_GetDeviceFaultReportsKHR,
    .GetDeviceFaultDebugInfoKHR = gfx9_GetDeviceFaultDebugInfoKHR,
    .CmdSetDepthBias2EXT = gfx9_CmdSetDepthBias2EXT,
    .ReleaseSwapchainImagesKHR = gfx9_ReleaseSwapchainImagesKHR,
    .ReleaseSwapchainImagesEXT = gfx9_ReleaseSwapchainImagesEXT,
    .GetDeviceImageSubresourceLayout = gfx9_GetDeviceImageSubresourceLayout,
    .GetDeviceImageSubresourceLayoutKHR = gfx9_GetDeviceImageSubresourceLayoutKHR,
    .MapMemory2 = gfx9_MapMemory2,
    .MapMemory2KHR = gfx9_MapMemory2KHR,
    .UnmapMemory2 = gfx9_UnmapMemory2,
    .UnmapMemory2KHR = gfx9_UnmapMemory2KHR,
    .CreateShadersEXT = gfx9_CreateShadersEXT,
    .DestroyShaderEXT = gfx9_DestroyShaderEXT,
    .GetShaderBinaryDataEXT = gfx9_GetShaderBinaryDataEXT,
    .CmdBindShadersEXT = gfx9_CmdBindShadersEXT,
    .SetSwapchainPresentTimingQueueSizeEXT = gfx9_SetSwapchainPresentTimingQueueSizeEXT,
    .GetSwapchainTimingPropertiesEXT = gfx9_GetSwapchainTimingPropertiesEXT,
    .GetSwapchainTimeDomainPropertiesEXT = gfx9_GetSwapchainTimeDomainPropertiesEXT,
    .GetPastPresentationTimingEXT = gfx9_GetPastPresentationTimingEXT,
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetScreenBufferPropertiesQNX = gfx9_GetScreenBufferPropertiesQNX,
#elif defined(_MSC_VER)
    .GetScreenBufferPropertiesQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CmdBindDescriptorSets2 = gfx9_CmdBindDescriptorSets2,
    .CmdBindDescriptorSets2KHR = gfx9_CmdBindDescriptorSets2KHR,
    .CmdPushConstants2 = gfx9_CmdPushConstants2,
    .CmdPushConstants2KHR = gfx9_CmdPushConstants2KHR,
    .CmdPushDescriptorSet2 = gfx9_CmdPushDescriptorSet2,
    .CmdPushDescriptorSet2KHR = gfx9_CmdPushDescriptorSet2KHR,
    .CmdPushDescriptorSetWithTemplate2 = gfx9_CmdPushDescriptorSetWithTemplate2,
    .CmdPushDescriptorSetWithTemplate2KHR = gfx9_CmdPushDescriptorSetWithTemplate2KHR,
    .CmdSetDescriptorBufferOffsets2EXT = gfx9_CmdSetDescriptorBufferOffsets2EXT,
    .CmdBindDescriptorBufferEmbeddedSamplers2EXT = gfx9_CmdBindDescriptorBufferEmbeddedSamplers2EXT,
    .SetLatencySleepModeNV = gfx9_SetLatencySleepModeNV,
    .LatencySleepNV = gfx9_LatencySleepNV,
    .SetLatencyMarkerNV = gfx9_SetLatencyMarkerNV,
    .GetLatencyTimingsNV = gfx9_GetLatencyTimingsNV,
    .QueueNotifyOutOfBandNV = gfx9_QueueNotifyOutOfBandNV,
    .CmdSetRenderingAttachmentLocations = gfx9_CmdSetRenderingAttachmentLocations,
    .CmdSetRenderingAttachmentLocationsKHR = gfx9_CmdSetRenderingAttachmentLocationsKHR,
    .CmdSetRenderingInputAttachmentIndices = gfx9_CmdSetRenderingInputAttachmentIndices,
    .CmdSetRenderingInputAttachmentIndicesKHR = gfx9_CmdSetRenderingInputAttachmentIndicesKHR,
    .CmdSetDepthClampRangeEXT = gfx9_CmdSetDepthClampRangeEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandleEXT = gfx9_GetMemoryMetalHandleEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandleEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandlePropertiesEXT = gfx9_GetMemoryMetalHandlePropertiesEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandlePropertiesEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .ConvertCooperativeVectorMatrixNV = gfx9_ConvertCooperativeVectorMatrixNV,
    .CmdConvertCooperativeVectorMatrixNV = gfx9_CmdConvertCooperativeVectorMatrixNV,
    .CmdDispatchTileQCOM = gfx9_CmdDispatchTileQCOM,
    .CmdBeginPerTileExecutionQCOM = gfx9_CmdBeginPerTileExecutionQCOM,
    .CmdEndPerTileExecutionQCOM = gfx9_CmdEndPerTileExecutionQCOM,
    .CreateExternalComputeQueueNV = gfx9_CreateExternalComputeQueueNV,
    .DestroyExternalComputeQueueNV = gfx9_DestroyExternalComputeQueueNV,
    .CreateShaderInstrumentationARM = gfx9_CreateShaderInstrumentationARM,
    .DestroyShaderInstrumentationARM = gfx9_DestroyShaderInstrumentationARM,
    .CmdBeginShaderInstrumentationARM = gfx9_CmdBeginShaderInstrumentationARM,
    .CmdEndShaderInstrumentationARM = gfx9_CmdEndShaderInstrumentationARM,
    .GetShaderInstrumentationValuesARM = gfx9_GetShaderInstrumentationValuesARM,
    .ClearShaderInstrumentationMetricsARM = gfx9_ClearShaderInstrumentationMetricsARM,
    .CreateTensorARM = gfx9_CreateTensorARM,
    .DestroyTensorARM = gfx9_DestroyTensorARM,
    .CreateTensorViewARM = gfx9_CreateTensorViewARM,
    .DestroyTensorViewARM = gfx9_DestroyTensorViewARM,
    .GetTensorMemoryRequirementsARM = gfx9_GetTensorMemoryRequirementsARM,
    .BindTensorMemoryARM = gfx9_BindTensorMemoryARM,
    .GetDeviceTensorMemoryRequirementsARM = gfx9_GetDeviceTensorMemoryRequirementsARM,
    .CmdCopyTensorARM = gfx9_CmdCopyTensorARM,
    .GetTensorOpaqueCaptureDescriptorDataARM = gfx9_GetTensorOpaqueCaptureDescriptorDataARM,
    .GetTensorViewOpaqueCaptureDescriptorDataARM = gfx9_GetTensorViewOpaqueCaptureDescriptorDataARM,
    .CreateDataGraphPipelinesARM = gfx9_CreateDataGraphPipelinesARM,
    .CreateDataGraphPipelineSessionARM = gfx9_CreateDataGraphPipelineSessionARM,
    .GetDataGraphPipelineSessionBindPointRequirementsARM = gfx9_GetDataGraphPipelineSessionBindPointRequirementsARM,
    .GetDataGraphPipelineSessionMemoryRequirementsARM = gfx9_GetDataGraphPipelineSessionMemoryRequirementsARM,
    .BindDataGraphPipelineSessionMemoryARM = gfx9_BindDataGraphPipelineSessionMemoryARM,
    .DestroyDataGraphPipelineSessionARM = gfx9_DestroyDataGraphPipelineSessionARM,
    .CmdDispatchDataGraphARM = gfx9_CmdDispatchDataGraphARM,
    .GetDataGraphPipelineAvailablePropertiesARM = gfx9_GetDataGraphPipelineAvailablePropertiesARM,
    .GetDataGraphPipelinePropertiesARM = gfx9_GetDataGraphPipelinePropertiesARM,
#ifdef VK_USE_PLATFORM_OHOS
    .GetNativeBufferPropertiesOHOS = gfx9_GetNativeBufferPropertiesOHOS,
#elif defined(_MSC_VER)
    .GetNativeBufferPropertiesOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
    .GetMemoryNativeBufferOHOS = gfx9_GetMemoryNativeBufferOHOS,
#elif defined(_MSC_VER)
    .GetMemoryNativeBufferOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .QueueSetPerfHintQCOM = gfx9_QueueSetPerfHintQCOM,
    .CmdSetComputeOccupancyPriorityNV = gfx9_CmdSetComputeOccupancyPriorityNV,
    .WriteSamplerDescriptorsEXT = gfx9_WriteSamplerDescriptorsEXT,
    .WriteResourceDescriptorsEXT = gfx9_WriteResourceDescriptorsEXT,
    .CmdBindSamplerHeapEXT = gfx9_CmdBindSamplerHeapEXT,
    .CmdBindResourceHeapEXT = gfx9_CmdBindResourceHeapEXT,
    .CmdPushDataEXT = gfx9_CmdPushDataEXT,
    .RegisterCustomBorderColorEXT = gfx9_RegisterCustomBorderColorEXT,
    .UnregisterCustomBorderColorEXT = gfx9_UnregisterCustomBorderColorEXT,
    .GetImageOpaqueCaptureDataEXT = gfx9_GetImageOpaqueCaptureDataEXT,
    .GetTensorOpaqueCaptureDataARM = gfx9_GetTensorOpaqueCaptureDataARM,
    .CmdCopyMemoryKHR = gfx9_CmdCopyMemoryKHR,
    .CmdCopyMemoryToImageKHR = gfx9_CmdCopyMemoryToImageKHR,
    .CmdCopyImageToMemoryKHR = gfx9_CmdCopyImageToMemoryKHR,
    .CmdUpdateMemoryKHR = gfx9_CmdUpdateMemoryKHR,
    .CmdFillMemoryKHR = gfx9_CmdFillMemoryKHR,
    .CmdCopyQueryPoolResultsToMemoryKHR = gfx9_CmdCopyQueryPoolResultsToMemoryKHR,
    .CmdBeginConditionalRendering2EXT = gfx9_CmdBeginConditionalRendering2EXT,
    .CmdBindTransformFeedbackBuffers2EXT = gfx9_CmdBindTransformFeedbackBuffers2EXT,
    .CmdBeginTransformFeedback2EXT = gfx9_CmdBeginTransformFeedback2EXT,
    .CmdEndTransformFeedback2EXT = gfx9_CmdEndTransformFeedback2EXT,
    .CmdDrawIndirectByteCount2EXT = gfx9_CmdDrawIndirectByteCount2EXT,
    .CmdWriteMarkerToMemoryAMD = gfx9_CmdWriteMarkerToMemoryAMD,
    .CmdBindIndexBuffer3KHR = gfx9_CmdBindIndexBuffer3KHR,
    .CmdBindVertexBuffers3KHR = gfx9_CmdBindVertexBuffers3KHR,
    .CmdDrawIndirect2KHR = gfx9_CmdDrawIndirect2KHR,
    .CmdDrawIndexedIndirect2KHR = gfx9_CmdDrawIndexedIndirect2KHR,
    .CmdDrawIndirectCount2KHR = gfx9_CmdDrawIndirectCount2KHR,
    .CmdDrawIndexedIndirectCount2KHR = gfx9_CmdDrawIndexedIndirectCount2KHR,
    .CmdDrawMeshTasksIndirect2EXT = gfx9_CmdDrawMeshTasksIndirect2EXT,
    .CmdDrawMeshTasksIndirectCount2EXT = gfx9_CmdDrawMeshTasksIndirectCount2EXT,
    .CmdDispatchIndirect2KHR = gfx9_CmdDispatchIndirect2KHR,
    .CreateAccelerationStructure2KHR = gfx9_CreateAccelerationStructure2KHR,
    .CmdSetDispatchParametersARM = gfx9_CmdSetDispatchParametersARM,
};
const struct vk_device_entrypoint_table gfx11_device_entrypoints = {
    .GetDeviceProcAddr = gfx11_GetDeviceProcAddr,
    .DestroyDevice = gfx11_DestroyDevice,
    .GetDeviceQueue = gfx11_GetDeviceQueue,
    .QueueSubmit = gfx11_QueueSubmit,
    .QueueWaitIdle = gfx11_QueueWaitIdle,
    .DeviceWaitIdle = gfx11_DeviceWaitIdle,
    .AllocateMemory = gfx11_AllocateMemory,
    .FreeMemory = gfx11_FreeMemory,
    .MapMemory = gfx11_MapMemory,
    .UnmapMemory = gfx11_UnmapMemory,
    .FlushMappedMemoryRanges = gfx11_FlushMappedMemoryRanges,
    .InvalidateMappedMemoryRanges = gfx11_InvalidateMappedMemoryRanges,
    .GetDeviceMemoryCommitment = gfx11_GetDeviceMemoryCommitment,
    .GetBufferMemoryRequirements = gfx11_GetBufferMemoryRequirements,
    .BindBufferMemory = gfx11_BindBufferMemory,
    .GetImageMemoryRequirements = gfx11_GetImageMemoryRequirements,
    .BindImageMemory = gfx11_BindImageMemory,
    .GetImageSparseMemoryRequirements = gfx11_GetImageSparseMemoryRequirements,
    .QueueBindSparse = gfx11_QueueBindSparse,
    .CreateFence = gfx11_CreateFence,
    .DestroyFence = gfx11_DestroyFence,
    .ResetFences = gfx11_ResetFences,
    .GetFenceStatus = gfx11_GetFenceStatus,
    .WaitForFences = gfx11_WaitForFences,
    .CreateSemaphore = gfx11_CreateSemaphore,
    .DestroySemaphore = gfx11_DestroySemaphore,
    .CreateEvent = gfx11_CreateEvent,
    .DestroyEvent = gfx11_DestroyEvent,
    .GetEventStatus = gfx11_GetEventStatus,
    .SetEvent = gfx11_SetEvent,
    .ResetEvent = gfx11_ResetEvent,
    .CreateQueryPool = gfx11_CreateQueryPool,
    .DestroyQueryPool = gfx11_DestroyQueryPool,
    .GetQueryPoolResults = gfx11_GetQueryPoolResults,
    .ResetQueryPool = gfx11_ResetQueryPool,
    .ResetQueryPoolEXT = gfx11_ResetQueryPoolEXT,
    .CreateBuffer = gfx11_CreateBuffer,
    .DestroyBuffer = gfx11_DestroyBuffer,
    .CreateBufferView = gfx11_CreateBufferView,
    .DestroyBufferView = gfx11_DestroyBufferView,
    .CreateImage = gfx11_CreateImage,
    .DestroyImage = gfx11_DestroyImage,
    .GetImageSubresourceLayout = gfx11_GetImageSubresourceLayout,
    .CreateImageView = gfx11_CreateImageView,
    .DestroyImageView = gfx11_DestroyImageView,
    .CreateShaderModule = gfx11_CreateShaderModule,
    .DestroyShaderModule = gfx11_DestroyShaderModule,
    .CreatePipelineCache = gfx11_CreatePipelineCache,
    .DestroyPipelineCache = gfx11_DestroyPipelineCache,
    .GetPipelineCacheData = gfx11_GetPipelineCacheData,
    .MergePipelineCaches = gfx11_MergePipelineCaches,
    .CreatePipelineBinariesKHR = gfx11_CreatePipelineBinariesKHR,
    .DestroyPipelineBinaryKHR = gfx11_DestroyPipelineBinaryKHR,
    .GetPipelineKeyKHR = gfx11_GetPipelineKeyKHR,
    .GetPipelineBinaryDataKHR = gfx11_GetPipelineBinaryDataKHR,
    .ReleaseCapturedPipelineDataKHR = gfx11_ReleaseCapturedPipelineDataKHR,
    .CreateGraphicsPipelines = gfx11_CreateGraphicsPipelines,
    .CreateComputePipelines = gfx11_CreateComputePipelines,
    .GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI = gfx11_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI,
    .DestroyPipeline = gfx11_DestroyPipeline,
    .CreatePipelineLayout = gfx11_CreatePipelineLayout,
    .DestroyPipelineLayout = gfx11_DestroyPipelineLayout,
    .CreateSampler = gfx11_CreateSampler,
    .DestroySampler = gfx11_DestroySampler,
    .CreateDescriptorSetLayout = gfx11_CreateDescriptorSetLayout,
    .DestroyDescriptorSetLayout = gfx11_DestroyDescriptorSetLayout,
    .CreateDescriptorPool = gfx11_CreateDescriptorPool,
    .DestroyDescriptorPool = gfx11_DestroyDescriptorPool,
    .ResetDescriptorPool = gfx11_ResetDescriptorPool,
    .AllocateDescriptorSets = gfx11_AllocateDescriptorSets,
    .FreeDescriptorSets = gfx11_FreeDescriptorSets,
    .UpdateDescriptorSets = gfx11_UpdateDescriptorSets,
    .CreateFramebuffer = gfx11_CreateFramebuffer,
    .DestroyFramebuffer = gfx11_DestroyFramebuffer,
    .CreateRenderPass = gfx11_CreateRenderPass,
    .DestroyRenderPass = gfx11_DestroyRenderPass,
    .GetRenderAreaGranularity = gfx11_GetRenderAreaGranularity,
    .GetRenderingAreaGranularity = gfx11_GetRenderingAreaGranularity,
    .GetRenderingAreaGranularityKHR = gfx11_GetRenderingAreaGranularityKHR,
    .CreateCommandPool = gfx11_CreateCommandPool,
    .DestroyCommandPool = gfx11_DestroyCommandPool,
    .ResetCommandPool = gfx11_ResetCommandPool,
    .AllocateCommandBuffers = gfx11_AllocateCommandBuffers,
    .FreeCommandBuffers = gfx11_FreeCommandBuffers,
    .BeginCommandBuffer = gfx11_BeginCommandBuffer,
    .EndCommandBuffer = gfx11_EndCommandBuffer,
    .ResetCommandBuffer = gfx11_ResetCommandBuffer,
    .CmdBindPipeline = gfx11_CmdBindPipeline,
    .CmdSetPrimitiveRestartIndexEXT = gfx11_CmdSetPrimitiveRestartIndexEXT,
    .CmdSetAttachmentFeedbackLoopEnableEXT = gfx11_CmdSetAttachmentFeedbackLoopEnableEXT,
    .CmdSetViewport = gfx11_CmdSetViewport,
    .CmdSetScissor = gfx11_CmdSetScissor,
    .CmdSetLineWidth = gfx11_CmdSetLineWidth,
    .CmdSetDepthBias = gfx11_CmdSetDepthBias,
    .CmdSetBlendConstants = gfx11_CmdSetBlendConstants,
    .CmdSetDepthBounds = gfx11_CmdSetDepthBounds,
    .CmdSetStencilCompareMask = gfx11_CmdSetStencilCompareMask,
    .CmdSetStencilWriteMask = gfx11_CmdSetStencilWriteMask,
    .CmdSetStencilReference = gfx11_CmdSetStencilReference,
    .CmdBindDescriptorSets = gfx11_CmdBindDescriptorSets,
    .CmdBindIndexBuffer = gfx11_CmdBindIndexBuffer,
    .CmdBindVertexBuffers = gfx11_CmdBindVertexBuffers,
    .CmdDraw = gfx11_CmdDraw,
    .CmdDrawIndexed = gfx11_CmdDrawIndexed,
    .CmdDrawMultiEXT = gfx11_CmdDrawMultiEXT,
    .CmdDrawMultiIndexedEXT = gfx11_CmdDrawMultiIndexedEXT,
    .CmdDrawIndirect = gfx11_CmdDrawIndirect,
    .CmdDrawIndexedIndirect = gfx11_CmdDrawIndexedIndirect,
    .CmdDispatch = gfx11_CmdDispatch,
    .CmdDispatchIndirect = gfx11_CmdDispatchIndirect,
    .CmdSubpassShadingHUAWEI = gfx11_CmdSubpassShadingHUAWEI,
    .CmdDrawClusterHUAWEI = gfx11_CmdDrawClusterHUAWEI,
    .CmdDrawClusterIndirectHUAWEI = gfx11_CmdDrawClusterIndirectHUAWEI,
    .CmdUpdatePipelineIndirectBufferNV = gfx11_CmdUpdatePipelineIndirectBufferNV,
    .CmdCopyBuffer = gfx11_CmdCopyBuffer,
    .CmdCopyImage = gfx11_CmdCopyImage,
    .CmdBlitImage = gfx11_CmdBlitImage,
    .CmdCopyBufferToImage = gfx11_CmdCopyBufferToImage,
    .CmdCopyImageToBuffer = gfx11_CmdCopyImageToBuffer,
    .CmdCopyMemoryIndirectNV = gfx11_CmdCopyMemoryIndirectNV,
    .CmdCopyMemoryIndirectKHR = gfx11_CmdCopyMemoryIndirectKHR,
    .CmdCopyMemoryToImageIndirectNV = gfx11_CmdCopyMemoryToImageIndirectNV,
    .CmdCopyMemoryToImageIndirectKHR = gfx11_CmdCopyMemoryToImageIndirectKHR,
    .CmdUpdateBuffer = gfx11_CmdUpdateBuffer,
    .CmdFillBuffer = gfx11_CmdFillBuffer,
    .CmdClearColorImage = gfx11_CmdClearColorImage,
    .CmdClearDepthStencilImage = gfx11_CmdClearDepthStencilImage,
    .CmdClearAttachments = gfx11_CmdClearAttachments,
    .CmdResolveImage = gfx11_CmdResolveImage,
    .CmdSetEvent = gfx11_CmdSetEvent,
    .CmdResetEvent = gfx11_CmdResetEvent,
    .CmdWaitEvents = gfx11_CmdWaitEvents,
    .CmdPipelineBarrier = gfx11_CmdPipelineBarrier,
    .CmdBeginQuery = gfx11_CmdBeginQuery,
    .CmdEndQuery = gfx11_CmdEndQuery,
    .CmdBeginConditionalRenderingEXT = gfx11_CmdBeginConditionalRenderingEXT,
    .CmdEndConditionalRenderingEXT = gfx11_CmdEndConditionalRenderingEXT,
    .CmdBeginCustomResolveEXT = gfx11_CmdBeginCustomResolveEXT,
    .CmdResetQueryPool = gfx11_CmdResetQueryPool,
    .CmdWriteTimestamp = gfx11_CmdWriteTimestamp,
    .CmdCopyQueryPoolResults = gfx11_CmdCopyQueryPoolResults,
    .CmdPushConstants = gfx11_CmdPushConstants,
    .CmdBeginRenderPass = gfx11_CmdBeginRenderPass,
    .CmdNextSubpass = gfx11_CmdNextSubpass,
    .CmdEndRenderPass = gfx11_CmdEndRenderPass,
    .CmdExecuteCommands = gfx11_CmdExecuteCommands,
    .CreateSharedSwapchainsKHR = gfx11_CreateSharedSwapchainsKHR,
    .CreateSwapchainKHR = gfx11_CreateSwapchainKHR,
    .DestroySwapchainKHR = gfx11_DestroySwapchainKHR,
    .GetSwapchainImagesKHR = gfx11_GetSwapchainImagesKHR,
    .AcquireNextImageKHR = gfx11_AcquireNextImageKHR,
    .QueuePresentKHR = gfx11_QueuePresentKHR,
    .DebugMarkerSetObjectNameEXT = gfx11_DebugMarkerSetObjectNameEXT,
    .DebugMarkerSetObjectTagEXT = gfx11_DebugMarkerSetObjectTagEXT,
    .CmdDebugMarkerBeginEXT = gfx11_CmdDebugMarkerBeginEXT,
    .CmdDebugMarkerEndEXT = gfx11_CmdDebugMarkerEndEXT,
    .CmdDebugMarkerInsertEXT = gfx11_CmdDebugMarkerInsertEXT,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleNV = gfx11_GetMemoryWin32HandleNV,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .CmdExecuteGeneratedCommandsNV = gfx11_CmdExecuteGeneratedCommandsNV,
    .CmdPreprocessGeneratedCommandsNV = gfx11_CmdPreprocessGeneratedCommandsNV,
    .CmdBindPipelineShaderGroupNV = gfx11_CmdBindPipelineShaderGroupNV,
    .GetGeneratedCommandsMemoryRequirementsNV = gfx11_GetGeneratedCommandsMemoryRequirementsNV,
    .CreateIndirectCommandsLayoutNV = gfx11_CreateIndirectCommandsLayoutNV,
    .DestroyIndirectCommandsLayoutNV = gfx11_DestroyIndirectCommandsLayoutNV,
    .CmdExecuteGeneratedCommandsEXT = gfx11_CmdExecuteGeneratedCommandsEXT,
    .CmdPreprocessGeneratedCommandsEXT = gfx11_CmdPreprocessGeneratedCommandsEXT,
    .GetGeneratedCommandsMemoryRequirementsEXT = gfx11_GetGeneratedCommandsMemoryRequirementsEXT,
    .CreateIndirectCommandsLayoutEXT = gfx11_CreateIndirectCommandsLayoutEXT,
    .DestroyIndirectCommandsLayoutEXT = gfx11_DestroyIndirectCommandsLayoutEXT,
    .CreateIndirectExecutionSetEXT = gfx11_CreateIndirectExecutionSetEXT,
    .DestroyIndirectExecutionSetEXT = gfx11_DestroyIndirectExecutionSetEXT,
    .UpdateIndirectExecutionSetPipelineEXT = gfx11_UpdateIndirectExecutionSetPipelineEXT,
    .UpdateIndirectExecutionSetShaderEXT = gfx11_UpdateIndirectExecutionSetShaderEXT,
    .CmdPushDescriptorSet = gfx11_CmdPushDescriptorSet,
    .CmdPushDescriptorSetKHR = gfx11_CmdPushDescriptorSetKHR,
    .TrimCommandPool = gfx11_TrimCommandPool,
    .TrimCommandPoolKHR = gfx11_TrimCommandPoolKHR,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleKHR = gfx11_GetMemoryWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandlePropertiesKHR = gfx11_GetMemoryWin32HandlePropertiesKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandlePropertiesKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryFdKHR = gfx11_GetMemoryFdKHR,
    .GetMemoryFdPropertiesKHR = gfx11_GetMemoryFdPropertiesKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandleFUCHSIA = gfx11_GetMemoryZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandlePropertiesFUCHSIA = gfx11_GetMemoryZirconHandlePropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandlePropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .GetMemoryRemoteAddressNV = gfx11_GetMemoryRemoteAddressNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreWin32HandleKHR = gfx11_GetSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportSemaphoreWin32HandleKHR = gfx11_ImportSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreFdKHR = gfx11_GetSemaphoreFdKHR,
    .ImportSemaphoreFdKHR = gfx11_ImportSemaphoreFdKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetSemaphoreZirconHandleFUCHSIA = gfx11_GetSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .ImportSemaphoreZirconHandleFUCHSIA = gfx11_ImportSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .ImportSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetFenceWin32HandleKHR = gfx11_GetFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportFenceWin32HandleKHR = gfx11_ImportFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetFenceFdKHR = gfx11_GetFenceFdKHR,
    .ImportFenceFdKHR = gfx11_ImportFenceFdKHR,
    .DisplayPowerControlEXT = gfx11_DisplayPowerControlEXT,
    .RegisterDeviceEventEXT = gfx11_RegisterDeviceEventEXT,
    .RegisterDisplayEventEXT = gfx11_RegisterDisplayEventEXT,
    .GetSwapchainCounterEXT = gfx11_GetSwapchainCounterEXT,
    .GetDeviceGroupPeerMemoryFeatures = gfx11_GetDeviceGroupPeerMemoryFeatures,
    .GetDeviceGroupPeerMemoryFeaturesKHR = gfx11_GetDeviceGroupPeerMemoryFeaturesKHR,
    .BindBufferMemory2 = gfx11_BindBufferMemory2,
    .BindBufferMemory2KHR = gfx11_BindBufferMemory2KHR,
    .BindImageMemory2 = gfx11_BindImageMemory2,
    .BindImageMemory2KHR = gfx11_BindImageMemory2KHR,
    .CmdSetDeviceMask = gfx11_CmdSetDeviceMask,
    .CmdSetDeviceMaskKHR = gfx11_CmdSetDeviceMaskKHR,
    .GetDeviceGroupPresentCapabilitiesKHR = gfx11_GetDeviceGroupPresentCapabilitiesKHR,
    .GetDeviceGroupSurfacePresentModesKHR = gfx11_GetDeviceGroupSurfacePresentModesKHR,
    .AcquireNextImage2KHR = gfx11_AcquireNextImage2KHR,
    .CmdDispatchBase = gfx11_CmdDispatchBase,
    .CmdDispatchBaseKHR = gfx11_CmdDispatchBaseKHR,
    .CreateDescriptorUpdateTemplate = gfx11_CreateDescriptorUpdateTemplate,
    .CreateDescriptorUpdateTemplateKHR = gfx11_CreateDescriptorUpdateTemplateKHR,
    .DestroyDescriptorUpdateTemplate = gfx11_DestroyDescriptorUpdateTemplate,
    .DestroyDescriptorUpdateTemplateKHR = gfx11_DestroyDescriptorUpdateTemplateKHR,
    .UpdateDescriptorSetWithTemplate = gfx11_UpdateDescriptorSetWithTemplate,
    .UpdateDescriptorSetWithTemplateKHR = gfx11_UpdateDescriptorSetWithTemplateKHR,
    .CmdPushDescriptorSetWithTemplate = gfx11_CmdPushDescriptorSetWithTemplate,
    .CmdPushDescriptorSetWithTemplateKHR = gfx11_CmdPushDescriptorSetWithTemplateKHR,
    .SetHdrMetadataEXT = gfx11_SetHdrMetadataEXT,
    .GetSwapchainStatusKHR = gfx11_GetSwapchainStatusKHR,
    .GetRefreshCycleDurationGOOGLE = gfx11_GetRefreshCycleDurationGOOGLE,
    .GetPastPresentationTimingGOOGLE = gfx11_GetPastPresentationTimingGOOGLE,
    .CmdSetViewportWScalingNV = gfx11_CmdSetViewportWScalingNV,
    .CmdSetDiscardRectangleEXT = gfx11_CmdSetDiscardRectangleEXT,
    .CmdSetDiscardRectangleEnableEXT = gfx11_CmdSetDiscardRectangleEnableEXT,
    .CmdSetDiscardRectangleModeEXT = gfx11_CmdSetDiscardRectangleModeEXT,
    .CmdSetSampleLocationsEXT = gfx11_CmdSetSampleLocationsEXT,
    .GetBufferMemoryRequirements2 = gfx11_GetBufferMemoryRequirements2,
    .GetBufferMemoryRequirements2KHR = gfx11_GetBufferMemoryRequirements2KHR,
    .GetImageMemoryRequirements2 = gfx11_GetImageMemoryRequirements2,
    .GetImageMemoryRequirements2KHR = gfx11_GetImageMemoryRequirements2KHR,
    .GetImageSparseMemoryRequirements2 = gfx11_GetImageSparseMemoryRequirements2,
    .GetImageSparseMemoryRequirements2KHR = gfx11_GetImageSparseMemoryRequirements2KHR,
    .GetDeviceBufferMemoryRequirements = gfx11_GetDeviceBufferMemoryRequirements,
    .GetDeviceBufferMemoryRequirementsKHR = gfx11_GetDeviceBufferMemoryRequirementsKHR,
    .GetDeviceImageMemoryRequirements = gfx11_GetDeviceImageMemoryRequirements,
    .GetDeviceImageMemoryRequirementsKHR = gfx11_GetDeviceImageMemoryRequirementsKHR,
    .GetDeviceImageSparseMemoryRequirements = gfx11_GetDeviceImageSparseMemoryRequirements,
    .GetDeviceImageSparseMemoryRequirementsKHR = gfx11_GetDeviceImageSparseMemoryRequirementsKHR,
    .CreateSamplerYcbcrConversion = gfx11_CreateSamplerYcbcrConversion,
    .CreateSamplerYcbcrConversionKHR = gfx11_CreateSamplerYcbcrConversionKHR,
    .DestroySamplerYcbcrConversion = gfx11_DestroySamplerYcbcrConversion,
    .DestroySamplerYcbcrConversionKHR = gfx11_DestroySamplerYcbcrConversionKHR,
    .GetDeviceQueue2 = gfx11_GetDeviceQueue2,
    .CreateValidationCacheEXT = gfx11_CreateValidationCacheEXT,
    .DestroyValidationCacheEXT = gfx11_DestroyValidationCacheEXT,
    .GetValidationCacheDataEXT = gfx11_GetValidationCacheDataEXT,
    .MergeValidationCachesEXT = gfx11_MergeValidationCachesEXT,
    .GetDescriptorSetLayoutSupport = gfx11_GetDescriptorSetLayoutSupport,
    .GetDescriptorSetLayoutSupportKHR = gfx11_GetDescriptorSetLayoutSupportKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsageANDROID = gfx11_GetSwapchainGrallocUsageANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsage2ANDROID = gfx11_GetSwapchainGrallocUsage2ANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsage2ANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .AcquireImageANDROID = gfx11_AcquireImageANDROID,
#elif defined(_MSC_VER)
    .AcquireImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .QueueSignalReleaseImageANDROID = gfx11_QueueSignalReleaseImageANDROID,
#elif defined(_MSC_VER)
    .QueueSignalReleaseImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .GetShaderInfoAMD = gfx11_GetShaderInfoAMD,
    .SetLocalDimmingAMD = gfx11_SetLocalDimmingAMD,
    .GetCalibratedTimestampsKHR = gfx11_GetCalibratedTimestampsKHR,
    .GetCalibratedTimestampsEXT = gfx11_GetCalibratedTimestampsEXT,
    .SetDebugUtilsObjectNameEXT = gfx11_SetDebugUtilsObjectNameEXT,
    .SetDebugUtilsObjectTagEXT = gfx11_SetDebugUtilsObjectTagEXT,
    .QueueBeginDebugUtilsLabelEXT = gfx11_QueueBeginDebugUtilsLabelEXT,
    .QueueEndDebugUtilsLabelEXT = gfx11_QueueEndDebugUtilsLabelEXT,
    .QueueInsertDebugUtilsLabelEXT = gfx11_QueueInsertDebugUtilsLabelEXT,
    .CmdBeginDebugUtilsLabelEXT = gfx11_CmdBeginDebugUtilsLabelEXT,
    .CmdEndDebugUtilsLabelEXT = gfx11_CmdEndDebugUtilsLabelEXT,
    .CmdInsertDebugUtilsLabelEXT = gfx11_CmdInsertDebugUtilsLabelEXT,
    .GetMemoryHostPointerPropertiesEXT = gfx11_GetMemoryHostPointerPropertiesEXT,
    .CmdWriteBufferMarkerAMD = gfx11_CmdWriteBufferMarkerAMD,
    .CreateRenderPass2 = gfx11_CreateRenderPass2,
    .CreateRenderPass2KHR = gfx11_CreateRenderPass2KHR,
    .CmdBeginRenderPass2 = gfx11_CmdBeginRenderPass2,
    .CmdBeginRenderPass2KHR = gfx11_CmdBeginRenderPass2KHR,
    .CmdNextSubpass2 = gfx11_CmdNextSubpass2,
    .CmdNextSubpass2KHR = gfx11_CmdNextSubpass2KHR,
    .CmdEndRenderPass2 = gfx11_CmdEndRenderPass2,
    .CmdEndRenderPass2KHR = gfx11_CmdEndRenderPass2KHR,
    .GetSemaphoreCounterValue = gfx11_GetSemaphoreCounterValue,
    .GetSemaphoreCounterValueKHR = gfx11_GetSemaphoreCounterValueKHR,
    .WaitSemaphores = gfx11_WaitSemaphores,
    .WaitSemaphoresKHR = gfx11_WaitSemaphoresKHR,
    .SignalSemaphore = gfx11_SignalSemaphore,
    .SignalSemaphoreKHR = gfx11_SignalSemaphoreKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetAndroidHardwareBufferPropertiesANDROID = gfx11_GetAndroidHardwareBufferPropertiesANDROID,
#elif defined(_MSC_VER)
    .GetAndroidHardwareBufferPropertiesANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetMemoryAndroidHardwareBufferANDROID = gfx11_GetMemoryAndroidHardwareBufferANDROID,
#elif defined(_MSC_VER)
    .GetMemoryAndroidHardwareBufferANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .CmdDrawIndirectCount = gfx11_CmdDrawIndirectCount,
    .CmdDrawIndirectCountKHR = gfx11_CmdDrawIndirectCountKHR,
    .CmdDrawIndirectCountAMD = gfx11_CmdDrawIndirectCountAMD,
    .CmdDrawIndexedIndirectCount = gfx11_CmdDrawIndexedIndirectCount,
    .CmdDrawIndexedIndirectCountKHR = gfx11_CmdDrawIndexedIndirectCountKHR,
    .CmdDrawIndexedIndirectCountAMD = gfx11_CmdDrawIndexedIndirectCountAMD,
    .CmdSetCheckpointNV = gfx11_CmdSetCheckpointNV,
    .GetQueueCheckpointDataNV = gfx11_GetQueueCheckpointDataNV,
    .CmdBindTransformFeedbackBuffersEXT = gfx11_CmdBindTransformFeedbackBuffersEXT,
    .CmdBeginTransformFeedbackEXT = gfx11_CmdBeginTransformFeedbackEXT,
    .CmdEndTransformFeedbackEXT = gfx11_CmdEndTransformFeedbackEXT,
    .CmdBeginQueryIndexedEXT = gfx11_CmdBeginQueryIndexedEXT,
    .CmdEndQueryIndexedEXT = gfx11_CmdEndQueryIndexedEXT,
    .CmdDrawIndirectByteCountEXT = gfx11_CmdDrawIndirectByteCountEXT,
    .CmdSetExclusiveScissorNV = gfx11_CmdSetExclusiveScissorNV,
    .CmdSetExclusiveScissorEnableNV = gfx11_CmdSetExclusiveScissorEnableNV,
    .CmdBindShadingRateImageNV = gfx11_CmdBindShadingRateImageNV,
    .CmdSetViewportShadingRatePaletteNV = gfx11_CmdSetViewportShadingRatePaletteNV,
    .CmdSetCoarseSampleOrderNV = gfx11_CmdSetCoarseSampleOrderNV,
    .CmdDrawMeshTasksNV = gfx11_CmdDrawMeshTasksNV,
    .CmdDrawMeshTasksIndirectNV = gfx11_CmdDrawMeshTasksIndirectNV,
    .CmdDrawMeshTasksIndirectCountNV = gfx11_CmdDrawMeshTasksIndirectCountNV,
    .CmdDrawMeshTasksEXT = gfx11_CmdDrawMeshTasksEXT,
    .CmdDrawMeshTasksIndirectEXT = gfx11_CmdDrawMeshTasksIndirectEXT,
    .CmdDrawMeshTasksIndirectCountEXT = gfx11_CmdDrawMeshTasksIndirectCountEXT,
    .CompileDeferredNV = gfx11_CompileDeferredNV,
    .CreateAccelerationStructureNV = gfx11_CreateAccelerationStructureNV,
    .CmdBindInvocationMaskHUAWEI = gfx11_CmdBindInvocationMaskHUAWEI,
    .DestroyAccelerationStructureKHR = gfx11_DestroyAccelerationStructureKHR,
    .DestroyAccelerationStructureNV = gfx11_DestroyAccelerationStructureNV,
    .GetAccelerationStructureMemoryRequirementsNV = gfx11_GetAccelerationStructureMemoryRequirementsNV,
    .BindAccelerationStructureMemoryNV = gfx11_BindAccelerationStructureMemoryNV,
    .CmdCopyAccelerationStructureNV = gfx11_CmdCopyAccelerationStructureNV,
    .CmdCopyAccelerationStructureKHR = gfx11_CmdCopyAccelerationStructureKHR,
    .CopyAccelerationStructureKHR = gfx11_CopyAccelerationStructureKHR,
    .CmdCopyAccelerationStructureToMemoryKHR = gfx11_CmdCopyAccelerationStructureToMemoryKHR,
    .CopyAccelerationStructureToMemoryKHR = gfx11_CopyAccelerationStructureToMemoryKHR,
    .CmdCopyMemoryToAccelerationStructureKHR = gfx11_CmdCopyMemoryToAccelerationStructureKHR,
    .CopyMemoryToAccelerationStructureKHR = gfx11_CopyMemoryToAccelerationStructureKHR,
    .CmdWriteAccelerationStructuresPropertiesKHR = gfx11_CmdWriteAccelerationStructuresPropertiesKHR,
    .CmdWriteAccelerationStructuresPropertiesNV = gfx11_CmdWriteAccelerationStructuresPropertiesNV,
    .CmdBuildAccelerationStructureNV = gfx11_CmdBuildAccelerationStructureNV,
    .WriteAccelerationStructuresPropertiesKHR = gfx11_WriteAccelerationStructuresPropertiesKHR,
    .CmdTraceRaysKHR = gfx11_CmdTraceRaysKHR,
    .CmdTraceRaysNV = gfx11_CmdTraceRaysNV,
    .GetRayTracingShaderGroupHandlesKHR = gfx11_GetRayTracingShaderGroupHandlesKHR,
    .GetRayTracingShaderGroupHandlesNV = gfx11_GetRayTracingShaderGroupHandlesNV,
    .GetRayTracingCaptureReplayShaderGroupHandlesKHR = gfx11_GetRayTracingCaptureReplayShaderGroupHandlesKHR,
    .GetAccelerationStructureHandleNV = gfx11_GetAccelerationStructureHandleNV,
    .CreateRayTracingPipelinesNV = gfx11_CreateRayTracingPipelinesNV,
    .CreateRayTracingPipelinesKHR = gfx11_CreateRayTracingPipelinesKHR,
    .CmdTraceRaysIndirectKHR = gfx11_CmdTraceRaysIndirectKHR,
    .CmdTraceRaysIndirect2KHR = gfx11_CmdTraceRaysIndirect2KHR,
    .GetClusterAccelerationStructureBuildSizesNV = gfx11_GetClusterAccelerationStructureBuildSizesNV,
    .CmdBuildClusterAccelerationStructureIndirectNV = gfx11_CmdBuildClusterAccelerationStructureIndirectNV,
    .GetDeviceAccelerationStructureCompatibilityKHR = gfx11_GetDeviceAccelerationStructureCompatibilityKHR,
    .GetRayTracingShaderGroupStackSizeKHR = gfx11_GetRayTracingShaderGroupStackSizeKHR,
    .CmdSetRayTracingPipelineStackSizeKHR = gfx11_CmdSetRayTracingPipelineStackSizeKHR,
    .GetImageViewHandleNVX = gfx11_GetImageViewHandleNVX,
    .GetImageViewHandle64NVX = gfx11_GetImageViewHandle64NVX,
    .GetImageViewAddressNVX = gfx11_GetImageViewAddressNVX,
    .GetDeviceCombinedImageSamplerIndexNVX = gfx11_GetDeviceCombinedImageSamplerIndexNVX,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetDeviceGroupSurfacePresentModes2EXT = gfx11_GetDeviceGroupSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetDeviceGroupSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireFullScreenExclusiveModeEXT = gfx11_AcquireFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .AcquireFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ReleaseFullScreenExclusiveModeEXT = gfx11_ReleaseFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .ReleaseFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .AcquireProfilingLockKHR = gfx11_AcquireProfilingLockKHR,
    .ReleaseProfilingLockKHR = gfx11_ReleaseProfilingLockKHR,
    .GetImageDrmFormatModifierPropertiesEXT = gfx11_GetImageDrmFormatModifierPropertiesEXT,
    .GetBufferOpaqueCaptureAddress = gfx11_GetBufferOpaqueCaptureAddress,
    .GetBufferOpaqueCaptureAddressKHR = gfx11_GetBufferOpaqueCaptureAddressKHR,
    .GetBufferDeviceAddress = gfx11_GetBufferDeviceAddress,
    .GetBufferDeviceAddressKHR = gfx11_GetBufferDeviceAddressKHR,
    .GetBufferDeviceAddressEXT = gfx11_GetBufferDeviceAddressEXT,
    .InitializePerformanceApiINTEL = gfx11_InitializePerformanceApiINTEL,
    .UninitializePerformanceApiINTEL = gfx11_UninitializePerformanceApiINTEL,
    .CmdSetPerformanceMarkerINTEL = gfx11_CmdSetPerformanceMarkerINTEL,
    .CmdSetPerformanceStreamMarkerINTEL = gfx11_CmdSetPerformanceStreamMarkerINTEL,
    .CmdSetPerformanceOverrideINTEL = gfx11_CmdSetPerformanceOverrideINTEL,
    .AcquirePerformanceConfigurationINTEL = gfx11_AcquirePerformanceConfigurationINTEL,
    .ReleasePerformanceConfigurationINTEL = gfx11_ReleasePerformanceConfigurationINTEL,
    .QueueSetPerformanceConfigurationINTEL = gfx11_QueueSetPerformanceConfigurationINTEL,
    .GetPerformanceParameterINTEL = gfx11_GetPerformanceParameterINTEL,
    .GetDeviceMemoryOpaqueCaptureAddress = gfx11_GetDeviceMemoryOpaqueCaptureAddress,
    .GetDeviceMemoryOpaqueCaptureAddressKHR = gfx11_GetDeviceMemoryOpaqueCaptureAddressKHR,
    .GetPipelineExecutablePropertiesKHR = gfx11_GetPipelineExecutablePropertiesKHR,
    .GetPipelineExecutableStatisticsKHR = gfx11_GetPipelineExecutableStatisticsKHR,
    .GetPipelineExecutableInternalRepresentationsKHR = gfx11_GetPipelineExecutableInternalRepresentationsKHR,
    .CmdSetLineStipple = gfx11_CmdSetLineStipple,
    .CmdSetLineStippleKHR = gfx11_CmdSetLineStippleKHR,
    .CmdSetLineStippleEXT = gfx11_CmdSetLineStippleEXT,
    .CreateAccelerationStructureKHR = gfx11_CreateAccelerationStructureKHR,
    .CmdBuildAccelerationStructuresKHR = gfx11_CmdBuildAccelerationStructuresKHR,
    .CmdBuildAccelerationStructuresIndirectKHR = gfx11_CmdBuildAccelerationStructuresIndirectKHR,
    .BuildAccelerationStructuresKHR = gfx11_BuildAccelerationStructuresKHR,
    .GetAccelerationStructureDeviceAddressKHR = gfx11_GetAccelerationStructureDeviceAddressKHR,
    .CreateDeferredOperationKHR = gfx11_CreateDeferredOperationKHR,
    .DestroyDeferredOperationKHR = gfx11_DestroyDeferredOperationKHR,
    .GetDeferredOperationMaxConcurrencyKHR = gfx11_GetDeferredOperationMaxConcurrencyKHR,
    .GetDeferredOperationResultKHR = gfx11_GetDeferredOperationResultKHR,
    .DeferredOperationJoinKHR = gfx11_DeferredOperationJoinKHR,
    .GetPipelineIndirectMemoryRequirementsNV = gfx11_GetPipelineIndirectMemoryRequirementsNV,
    .GetPipelineIndirectDeviceAddressNV = gfx11_GetPipelineIndirectDeviceAddressNV,
    .AntiLagUpdateAMD = gfx11_AntiLagUpdateAMD,
    .CmdSetCullMode = gfx11_CmdSetCullMode,
    .CmdSetCullModeEXT = gfx11_CmdSetCullModeEXT,
    .CmdSetFrontFace = gfx11_CmdSetFrontFace,
    .CmdSetFrontFaceEXT = gfx11_CmdSetFrontFaceEXT,
    .CmdSetPrimitiveTopology = gfx11_CmdSetPrimitiveTopology,
    .CmdSetPrimitiveTopologyEXT = gfx11_CmdSetPrimitiveTopologyEXT,
    .CmdSetViewportWithCount = gfx11_CmdSetViewportWithCount,
    .CmdSetViewportWithCountEXT = gfx11_CmdSetViewportWithCountEXT,
    .CmdSetScissorWithCount = gfx11_CmdSetScissorWithCount,
    .CmdSetScissorWithCountEXT = gfx11_CmdSetScissorWithCountEXT,
    .CmdBindIndexBuffer2 = gfx11_CmdBindIndexBuffer2,
    .CmdBindIndexBuffer2KHR = gfx11_CmdBindIndexBuffer2KHR,
    .CmdBindVertexBuffers2 = gfx11_CmdBindVertexBuffers2,
    .CmdBindVertexBuffers2EXT = gfx11_CmdBindVertexBuffers2EXT,
    .CmdSetDepthTestEnable = gfx11_CmdSetDepthTestEnable,
    .CmdSetDepthTestEnableEXT = gfx11_CmdSetDepthTestEnableEXT,
    .CmdSetDepthWriteEnable = gfx11_CmdSetDepthWriteEnable,
    .CmdSetDepthWriteEnableEXT = gfx11_CmdSetDepthWriteEnableEXT,
    .CmdSetDepthCompareOp = gfx11_CmdSetDepthCompareOp,
    .CmdSetDepthCompareOpEXT = gfx11_CmdSetDepthCompareOpEXT,
    .CmdSetDepthBoundsTestEnable = gfx11_CmdSetDepthBoundsTestEnable,
    .CmdSetDepthBoundsTestEnableEXT = gfx11_CmdSetDepthBoundsTestEnableEXT,
    .CmdSetStencilTestEnable = gfx11_CmdSetStencilTestEnable,
    .CmdSetStencilTestEnableEXT = gfx11_CmdSetStencilTestEnableEXT,
    .CmdSetStencilOp = gfx11_CmdSetStencilOp,
    .CmdSetStencilOpEXT = gfx11_CmdSetStencilOpEXT,
    .CmdSetPatchControlPointsEXT = gfx11_CmdSetPatchControlPointsEXT,
    .CmdSetRasterizerDiscardEnable = gfx11_CmdSetRasterizerDiscardEnable,
    .CmdSetRasterizerDiscardEnableEXT = gfx11_CmdSetRasterizerDiscardEnableEXT,
    .CmdSetDepthBiasEnable = gfx11_CmdSetDepthBiasEnable,
    .CmdSetDepthBiasEnableEXT = gfx11_CmdSetDepthBiasEnableEXT,
    .CmdSetLogicOpEXT = gfx11_CmdSetLogicOpEXT,
    .CmdSetPrimitiveRestartEnable = gfx11_CmdSetPrimitiveRestartEnable,
    .CmdSetPrimitiveRestartEnableEXT = gfx11_CmdSetPrimitiveRestartEnableEXT,
    .CmdSetTessellationDomainOriginEXT = gfx11_CmdSetTessellationDomainOriginEXT,
    .CmdSetDepthClampEnableEXT = gfx11_CmdSetDepthClampEnableEXT,
    .CmdSetPolygonModeEXT = gfx11_CmdSetPolygonModeEXT,
    .CmdSetRasterizationSamplesEXT = gfx11_CmdSetRasterizationSamplesEXT,
    .CmdSetSampleMaskEXT = gfx11_CmdSetSampleMaskEXT,
    .CmdSetAlphaToCoverageEnableEXT = gfx11_CmdSetAlphaToCoverageEnableEXT,
    .CmdSetAlphaToOneEnableEXT = gfx11_CmdSetAlphaToOneEnableEXT,
    .CmdSetLogicOpEnableEXT = gfx11_CmdSetLogicOpEnableEXT,
    .CmdSetColorBlendEnableEXT = gfx11_CmdSetColorBlendEnableEXT,
    .CmdSetColorBlendEquationEXT = gfx11_CmdSetColorBlendEquationEXT,
    .CmdSetColorWriteMaskEXT = gfx11_CmdSetColorWriteMaskEXT,
    .CmdSetRasterizationStreamEXT = gfx11_CmdSetRasterizationStreamEXT,
    .CmdSetConservativeRasterizationModeEXT = gfx11_CmdSetConservativeRasterizationModeEXT,
    .CmdSetExtraPrimitiveOverestimationSizeEXT = gfx11_CmdSetExtraPrimitiveOverestimationSizeEXT,
    .CmdSetDepthClipEnableEXT = gfx11_CmdSetDepthClipEnableEXT,
    .CmdSetSampleLocationsEnableEXT = gfx11_CmdSetSampleLocationsEnableEXT,
    .CmdSetColorBlendAdvancedEXT = gfx11_CmdSetColorBlendAdvancedEXT,
    .CmdSetProvokingVertexModeEXT = gfx11_CmdSetProvokingVertexModeEXT,
    .CmdSetLineRasterizationModeEXT = gfx11_CmdSetLineRasterizationModeEXT,
    .CmdSetLineStippleEnableEXT = gfx11_CmdSetLineStippleEnableEXT,
    .CmdSetDepthClipNegativeOneToOneEXT = gfx11_CmdSetDepthClipNegativeOneToOneEXT,
    .CmdSetViewportWScalingEnableNV = gfx11_CmdSetViewportWScalingEnableNV,
    .CmdSetViewportSwizzleNV = gfx11_CmdSetViewportSwizzleNV,
    .CmdSetCoverageToColorEnableNV = gfx11_CmdSetCoverageToColorEnableNV,
    .CmdSetCoverageToColorLocationNV = gfx11_CmdSetCoverageToColorLocationNV,
    .CmdSetCoverageModulationModeNV = gfx11_CmdSetCoverageModulationModeNV,
    .CmdSetCoverageModulationTableEnableNV = gfx11_CmdSetCoverageModulationTableEnableNV,
    .CmdSetCoverageModulationTableNV = gfx11_CmdSetCoverageModulationTableNV,
    .CmdSetShadingRateImageEnableNV = gfx11_CmdSetShadingRateImageEnableNV,
    .CmdSetCoverageReductionModeNV = gfx11_CmdSetCoverageReductionModeNV,
    .CmdSetRepresentativeFragmentTestEnableNV = gfx11_CmdSetRepresentativeFragmentTestEnableNV,
    .CreatePrivateDataSlot = gfx11_CreatePrivateDataSlot,
    .CreatePrivateDataSlotEXT = gfx11_CreatePrivateDataSlotEXT,
    .DestroyPrivateDataSlot = gfx11_DestroyPrivateDataSlot,
    .DestroyPrivateDataSlotEXT = gfx11_DestroyPrivateDataSlotEXT,
    .SetPrivateData = gfx11_SetPrivateData,
    .SetPrivateDataEXT = gfx11_SetPrivateDataEXT,
    .GetPrivateData = gfx11_GetPrivateData,
    .GetPrivateDataEXT = gfx11_GetPrivateDataEXT,
    .CmdCopyBuffer2 = gfx11_CmdCopyBuffer2,
    .CmdCopyBuffer2KHR = gfx11_CmdCopyBuffer2KHR,
    .CmdCopyImage2 = gfx11_CmdCopyImage2,
    .CmdCopyImage2KHR = gfx11_CmdCopyImage2KHR,
    .CmdBlitImage2 = gfx11_CmdBlitImage2,
    .CmdBlitImage2KHR = gfx11_CmdBlitImage2KHR,
    .CmdCopyBufferToImage2 = gfx11_CmdCopyBufferToImage2,
    .CmdCopyBufferToImage2KHR = gfx11_CmdCopyBufferToImage2KHR,
    .CmdCopyImageToBuffer2 = gfx11_CmdCopyImageToBuffer2,
    .CmdCopyImageToBuffer2KHR = gfx11_CmdCopyImageToBuffer2KHR,
    .CmdResolveImage2 = gfx11_CmdResolveImage2,
    .CmdResolveImage2KHR = gfx11_CmdResolveImage2KHR,
    .CmdSetFragmentShadingRateKHR = gfx11_CmdSetFragmentShadingRateKHR,
    .CmdSetFragmentShadingRateEnumNV = gfx11_CmdSetFragmentShadingRateEnumNV,
    .GetAccelerationStructureBuildSizesKHR = gfx11_GetAccelerationStructureBuildSizesKHR,
    .CmdSetVertexInputEXT = gfx11_CmdSetVertexInputEXT,
    .CmdSetColorWriteEnableEXT = gfx11_CmdSetColorWriteEnableEXT,
    .CmdSetEvent2 = gfx11_CmdSetEvent2,
    .CmdSetEvent2KHR = gfx11_CmdSetEvent2KHR,
    .CmdResetEvent2 = gfx11_CmdResetEvent2,
    .CmdResetEvent2KHR = gfx11_CmdResetEvent2KHR,
    .CmdWaitEvents2 = gfx11_CmdWaitEvents2,
    .CmdWaitEvents2KHR = gfx11_CmdWaitEvents2KHR,
    .CmdPipelineBarrier2 = gfx11_CmdPipelineBarrier2,
    .CmdPipelineBarrier2KHR = gfx11_CmdPipelineBarrier2KHR,
    .QueueSubmit2 = gfx11_QueueSubmit2,
    .QueueSubmit2KHR = gfx11_QueueSubmit2KHR,
    .CmdWriteTimestamp2 = gfx11_CmdWriteTimestamp2,
    .CmdWriteTimestamp2KHR = gfx11_CmdWriteTimestamp2KHR,
    .CmdWriteBufferMarker2AMD = gfx11_CmdWriteBufferMarker2AMD,
    .GetQueueCheckpointData2NV = gfx11_GetQueueCheckpointData2NV,
    .CopyMemoryToImage = gfx11_CopyMemoryToImage,
    .CopyMemoryToImageEXT = gfx11_CopyMemoryToImageEXT,
    .CopyImageToMemory = gfx11_CopyImageToMemory,
    .CopyImageToMemoryEXT = gfx11_CopyImageToMemoryEXT,
    .CopyImageToImage = gfx11_CopyImageToImage,
    .CopyImageToImageEXT = gfx11_CopyImageToImageEXT,
    .TransitionImageLayout = gfx11_TransitionImageLayout,
    .TransitionImageLayoutEXT = gfx11_TransitionImageLayoutEXT,
    .CreateVideoSessionKHR = gfx11_CreateVideoSessionKHR,
    .DestroyVideoSessionKHR = gfx11_DestroyVideoSessionKHR,
    .CreateVideoSessionParametersKHR = gfx11_CreateVideoSessionParametersKHR,
    .UpdateVideoSessionParametersKHR = gfx11_UpdateVideoSessionParametersKHR,
    .GetEncodedVideoSessionParametersKHR = gfx11_GetEncodedVideoSessionParametersKHR,
    .DestroyVideoSessionParametersKHR = gfx11_DestroyVideoSessionParametersKHR,
    .GetVideoSessionMemoryRequirementsKHR = gfx11_GetVideoSessionMemoryRequirementsKHR,
    .BindVideoSessionMemoryKHR = gfx11_BindVideoSessionMemoryKHR,
    .CmdDecodeVideoKHR = gfx11_CmdDecodeVideoKHR,
    .CmdBeginVideoCodingKHR = gfx11_CmdBeginVideoCodingKHR,
    .CmdControlVideoCodingKHR = gfx11_CmdControlVideoCodingKHR,
    .CmdEndVideoCodingKHR = gfx11_CmdEndVideoCodingKHR,
    .CmdEncodeVideoKHR = gfx11_CmdEncodeVideoKHR,
    .CmdDecompressMemoryNV = gfx11_CmdDecompressMemoryNV,
    .CmdDecompressMemoryIndirectCountNV = gfx11_CmdDecompressMemoryIndirectCountNV,
    .GetPartitionedAccelerationStructuresBuildSizesNV = gfx11_GetPartitionedAccelerationStructuresBuildSizesNV,
    .CmdBuildPartitionedAccelerationStructuresNV = gfx11_CmdBuildPartitionedAccelerationStructuresNV,
    .CmdDecompressMemoryEXT = gfx11_CmdDecompressMemoryEXT,
    .CmdDecompressMemoryIndirectCountEXT = gfx11_CmdDecompressMemoryIndirectCountEXT,
    .CreateCuModuleNVX = gfx11_CreateCuModuleNVX,
    .CreateCuFunctionNVX = gfx11_CreateCuFunctionNVX,
    .DestroyCuModuleNVX = gfx11_DestroyCuModuleNVX,
    .DestroyCuFunctionNVX = gfx11_DestroyCuFunctionNVX,
    .CmdCuLaunchKernelNVX = gfx11_CmdCuLaunchKernelNVX,
    .GetDescriptorSetLayoutSizeEXT = gfx11_GetDescriptorSetLayoutSizeEXT,
    .GetDescriptorSetLayoutBindingOffsetEXT = gfx11_GetDescriptorSetLayoutBindingOffsetEXT,
    .GetDescriptorEXT = gfx11_GetDescriptorEXT,
    .CmdBindDescriptorBuffersEXT = gfx11_CmdBindDescriptorBuffersEXT,
    .CmdSetDescriptorBufferOffsetsEXT = gfx11_CmdSetDescriptorBufferOffsetsEXT,
    .CmdBindDescriptorBufferEmbeddedSamplersEXT = gfx11_CmdBindDescriptorBufferEmbeddedSamplersEXT,
    .GetBufferOpaqueCaptureDescriptorDataEXT = gfx11_GetBufferOpaqueCaptureDescriptorDataEXT,
    .GetImageOpaqueCaptureDescriptorDataEXT = gfx11_GetImageOpaqueCaptureDescriptorDataEXT,
    .GetImageViewOpaqueCaptureDescriptorDataEXT = gfx11_GetImageViewOpaqueCaptureDescriptorDataEXT,
    .GetSamplerOpaqueCaptureDescriptorDataEXT = gfx11_GetSamplerOpaqueCaptureDescriptorDataEXT,
    .GetAccelerationStructureOpaqueCaptureDescriptorDataEXT = gfx11_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT,
    .SetDeviceMemoryPriorityEXT = gfx11_SetDeviceMemoryPriorityEXT,
    .WaitForPresent2KHR = gfx11_WaitForPresent2KHR,
    .WaitForPresentKHR = gfx11_WaitForPresentKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateBufferCollectionFUCHSIA = gfx11_CreateBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .CreateBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionBufferConstraintsFUCHSIA = gfx11_SetBufferCollectionBufferConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionBufferConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionImageConstraintsFUCHSIA = gfx11_SetBufferCollectionImageConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionImageConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .DestroyBufferCollectionFUCHSIA = gfx11_DestroyBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .DestroyBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetBufferCollectionPropertiesFUCHSIA = gfx11_GetBufferCollectionPropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetBufferCollectionPropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .CmdBeginRendering = gfx11_CmdBeginRendering,
    .CmdBeginRenderingKHR = gfx11_CmdBeginRenderingKHR,
    .CmdEndRendering = gfx11_CmdEndRendering,
    .CmdEndRendering2KHR = gfx11_CmdEndRendering2KHR,
    .CmdEndRendering2EXT = gfx11_CmdEndRendering2EXT,
    .CmdEndRenderingKHR = gfx11_CmdEndRenderingKHR,
    .GetDescriptorSetLayoutHostMappingInfoVALVE = gfx11_GetDescriptorSetLayoutHostMappingInfoVALVE,
    .GetDescriptorSetHostMappingVALVE = gfx11_GetDescriptorSetHostMappingVALVE,
    .CreateMicromapEXT = gfx11_CreateMicromapEXT,
    .CmdBuildMicromapsEXT = gfx11_CmdBuildMicromapsEXT,
    .BuildMicromapsEXT = gfx11_BuildMicromapsEXT,
    .DestroyMicromapEXT = gfx11_DestroyMicromapEXT,
    .CmdCopyMicromapEXT = gfx11_CmdCopyMicromapEXT,
    .CopyMicromapEXT = gfx11_CopyMicromapEXT,
    .CmdCopyMicromapToMemoryEXT = gfx11_CmdCopyMicromapToMemoryEXT,
    .CopyMicromapToMemoryEXT = gfx11_CopyMicromapToMemoryEXT,
    .CmdCopyMemoryToMicromapEXT = gfx11_CmdCopyMemoryToMicromapEXT,
    .CopyMemoryToMicromapEXT = gfx11_CopyMemoryToMicromapEXT,
    .CmdWriteMicromapsPropertiesEXT = gfx11_CmdWriteMicromapsPropertiesEXT,
    .WriteMicromapsPropertiesEXT = gfx11_WriteMicromapsPropertiesEXT,
    .GetDeviceMicromapCompatibilityEXT = gfx11_GetDeviceMicromapCompatibilityEXT,
    .GetMicromapBuildSizesEXT = gfx11_GetMicromapBuildSizesEXT,
    .GetShaderModuleIdentifierEXT = gfx11_GetShaderModuleIdentifierEXT,
    .GetShaderModuleCreateInfoIdentifierEXT = gfx11_GetShaderModuleCreateInfoIdentifierEXT,
    .GetImageSubresourceLayout2 = gfx11_GetImageSubresourceLayout2,
    .GetImageSubresourceLayout2KHR = gfx11_GetImageSubresourceLayout2KHR,
    .GetImageSubresourceLayout2EXT = gfx11_GetImageSubresourceLayout2EXT,
    .GetPipelinePropertiesEXT = gfx11_GetPipelinePropertiesEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .ExportMetalObjectsEXT = gfx11_ExportMetalObjectsEXT,
#elif defined(_MSC_VER)
    .ExportMetalObjectsEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CmdBindTileMemoryQCOM = gfx11_CmdBindTileMemoryQCOM,
    .GetFramebufferTilePropertiesQCOM = gfx11_GetFramebufferTilePropertiesQCOM,
    .GetDynamicRenderingTilePropertiesQCOM = gfx11_GetDynamicRenderingTilePropertiesQCOM,
    .CreateOpticalFlowSessionNV = gfx11_CreateOpticalFlowSessionNV,
    .DestroyOpticalFlowSessionNV = gfx11_DestroyOpticalFlowSessionNV,
    .BindOpticalFlowSessionImageNV = gfx11_BindOpticalFlowSessionImageNV,
    .CmdOpticalFlowExecuteNV = gfx11_CmdOpticalFlowExecuteNV,
    .GetDeviceFaultInfoEXT = gfx11_GetDeviceFaultInfoEXT,
    .GetDeviceFaultReportsKHR = gfx11_GetDeviceFaultReportsKHR,
    .GetDeviceFaultDebugInfoKHR = gfx11_GetDeviceFaultDebugInfoKHR,
    .CmdSetDepthBias2EXT = gfx11_CmdSetDepthBias2EXT,
    .ReleaseSwapchainImagesKHR = gfx11_ReleaseSwapchainImagesKHR,
    .ReleaseSwapchainImagesEXT = gfx11_ReleaseSwapchainImagesEXT,
    .GetDeviceImageSubresourceLayout = gfx11_GetDeviceImageSubresourceLayout,
    .GetDeviceImageSubresourceLayoutKHR = gfx11_GetDeviceImageSubresourceLayoutKHR,
    .MapMemory2 = gfx11_MapMemory2,
    .MapMemory2KHR = gfx11_MapMemory2KHR,
    .UnmapMemory2 = gfx11_UnmapMemory2,
    .UnmapMemory2KHR = gfx11_UnmapMemory2KHR,
    .CreateShadersEXT = gfx11_CreateShadersEXT,
    .DestroyShaderEXT = gfx11_DestroyShaderEXT,
    .GetShaderBinaryDataEXT = gfx11_GetShaderBinaryDataEXT,
    .CmdBindShadersEXT = gfx11_CmdBindShadersEXT,
    .SetSwapchainPresentTimingQueueSizeEXT = gfx11_SetSwapchainPresentTimingQueueSizeEXT,
    .GetSwapchainTimingPropertiesEXT = gfx11_GetSwapchainTimingPropertiesEXT,
    .GetSwapchainTimeDomainPropertiesEXT = gfx11_GetSwapchainTimeDomainPropertiesEXT,
    .GetPastPresentationTimingEXT = gfx11_GetPastPresentationTimingEXT,
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetScreenBufferPropertiesQNX = gfx11_GetScreenBufferPropertiesQNX,
#elif defined(_MSC_VER)
    .GetScreenBufferPropertiesQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CmdBindDescriptorSets2 = gfx11_CmdBindDescriptorSets2,
    .CmdBindDescriptorSets2KHR = gfx11_CmdBindDescriptorSets2KHR,
    .CmdPushConstants2 = gfx11_CmdPushConstants2,
    .CmdPushConstants2KHR = gfx11_CmdPushConstants2KHR,
    .CmdPushDescriptorSet2 = gfx11_CmdPushDescriptorSet2,
    .CmdPushDescriptorSet2KHR = gfx11_CmdPushDescriptorSet2KHR,
    .CmdPushDescriptorSetWithTemplate2 = gfx11_CmdPushDescriptorSetWithTemplate2,
    .CmdPushDescriptorSetWithTemplate2KHR = gfx11_CmdPushDescriptorSetWithTemplate2KHR,
    .CmdSetDescriptorBufferOffsets2EXT = gfx11_CmdSetDescriptorBufferOffsets2EXT,
    .CmdBindDescriptorBufferEmbeddedSamplers2EXT = gfx11_CmdBindDescriptorBufferEmbeddedSamplers2EXT,
    .SetLatencySleepModeNV = gfx11_SetLatencySleepModeNV,
    .LatencySleepNV = gfx11_LatencySleepNV,
    .SetLatencyMarkerNV = gfx11_SetLatencyMarkerNV,
    .GetLatencyTimingsNV = gfx11_GetLatencyTimingsNV,
    .QueueNotifyOutOfBandNV = gfx11_QueueNotifyOutOfBandNV,
    .CmdSetRenderingAttachmentLocations = gfx11_CmdSetRenderingAttachmentLocations,
    .CmdSetRenderingAttachmentLocationsKHR = gfx11_CmdSetRenderingAttachmentLocationsKHR,
    .CmdSetRenderingInputAttachmentIndices = gfx11_CmdSetRenderingInputAttachmentIndices,
    .CmdSetRenderingInputAttachmentIndicesKHR = gfx11_CmdSetRenderingInputAttachmentIndicesKHR,
    .CmdSetDepthClampRangeEXT = gfx11_CmdSetDepthClampRangeEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandleEXT = gfx11_GetMemoryMetalHandleEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandleEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandlePropertiesEXT = gfx11_GetMemoryMetalHandlePropertiesEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandlePropertiesEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .ConvertCooperativeVectorMatrixNV = gfx11_ConvertCooperativeVectorMatrixNV,
    .CmdConvertCooperativeVectorMatrixNV = gfx11_CmdConvertCooperativeVectorMatrixNV,
    .CmdDispatchTileQCOM = gfx11_CmdDispatchTileQCOM,
    .CmdBeginPerTileExecutionQCOM = gfx11_CmdBeginPerTileExecutionQCOM,
    .CmdEndPerTileExecutionQCOM = gfx11_CmdEndPerTileExecutionQCOM,
    .CreateExternalComputeQueueNV = gfx11_CreateExternalComputeQueueNV,
    .DestroyExternalComputeQueueNV = gfx11_DestroyExternalComputeQueueNV,
    .CreateShaderInstrumentationARM = gfx11_CreateShaderInstrumentationARM,
    .DestroyShaderInstrumentationARM = gfx11_DestroyShaderInstrumentationARM,
    .CmdBeginShaderInstrumentationARM = gfx11_CmdBeginShaderInstrumentationARM,
    .CmdEndShaderInstrumentationARM = gfx11_CmdEndShaderInstrumentationARM,
    .GetShaderInstrumentationValuesARM = gfx11_GetShaderInstrumentationValuesARM,
    .ClearShaderInstrumentationMetricsARM = gfx11_ClearShaderInstrumentationMetricsARM,
    .CreateTensorARM = gfx11_CreateTensorARM,
    .DestroyTensorARM = gfx11_DestroyTensorARM,
    .CreateTensorViewARM = gfx11_CreateTensorViewARM,
    .DestroyTensorViewARM = gfx11_DestroyTensorViewARM,
    .GetTensorMemoryRequirementsARM = gfx11_GetTensorMemoryRequirementsARM,
    .BindTensorMemoryARM = gfx11_BindTensorMemoryARM,
    .GetDeviceTensorMemoryRequirementsARM = gfx11_GetDeviceTensorMemoryRequirementsARM,
    .CmdCopyTensorARM = gfx11_CmdCopyTensorARM,
    .GetTensorOpaqueCaptureDescriptorDataARM = gfx11_GetTensorOpaqueCaptureDescriptorDataARM,
    .GetTensorViewOpaqueCaptureDescriptorDataARM = gfx11_GetTensorViewOpaqueCaptureDescriptorDataARM,
    .CreateDataGraphPipelinesARM = gfx11_CreateDataGraphPipelinesARM,
    .CreateDataGraphPipelineSessionARM = gfx11_CreateDataGraphPipelineSessionARM,
    .GetDataGraphPipelineSessionBindPointRequirementsARM = gfx11_GetDataGraphPipelineSessionBindPointRequirementsARM,
    .GetDataGraphPipelineSessionMemoryRequirementsARM = gfx11_GetDataGraphPipelineSessionMemoryRequirementsARM,
    .BindDataGraphPipelineSessionMemoryARM = gfx11_BindDataGraphPipelineSessionMemoryARM,
    .DestroyDataGraphPipelineSessionARM = gfx11_DestroyDataGraphPipelineSessionARM,
    .CmdDispatchDataGraphARM = gfx11_CmdDispatchDataGraphARM,
    .GetDataGraphPipelineAvailablePropertiesARM = gfx11_GetDataGraphPipelineAvailablePropertiesARM,
    .GetDataGraphPipelinePropertiesARM = gfx11_GetDataGraphPipelinePropertiesARM,
#ifdef VK_USE_PLATFORM_OHOS
    .GetNativeBufferPropertiesOHOS = gfx11_GetNativeBufferPropertiesOHOS,
#elif defined(_MSC_VER)
    .GetNativeBufferPropertiesOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
    .GetMemoryNativeBufferOHOS = gfx11_GetMemoryNativeBufferOHOS,
#elif defined(_MSC_VER)
    .GetMemoryNativeBufferOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .QueueSetPerfHintQCOM = gfx11_QueueSetPerfHintQCOM,
    .CmdSetComputeOccupancyPriorityNV = gfx11_CmdSetComputeOccupancyPriorityNV,
    .WriteSamplerDescriptorsEXT = gfx11_WriteSamplerDescriptorsEXT,
    .WriteResourceDescriptorsEXT = gfx11_WriteResourceDescriptorsEXT,
    .CmdBindSamplerHeapEXT = gfx11_CmdBindSamplerHeapEXT,
    .CmdBindResourceHeapEXT = gfx11_CmdBindResourceHeapEXT,
    .CmdPushDataEXT = gfx11_CmdPushDataEXT,
    .RegisterCustomBorderColorEXT = gfx11_RegisterCustomBorderColorEXT,
    .UnregisterCustomBorderColorEXT = gfx11_UnregisterCustomBorderColorEXT,
    .GetImageOpaqueCaptureDataEXT = gfx11_GetImageOpaqueCaptureDataEXT,
    .GetTensorOpaqueCaptureDataARM = gfx11_GetTensorOpaqueCaptureDataARM,
    .CmdCopyMemoryKHR = gfx11_CmdCopyMemoryKHR,
    .CmdCopyMemoryToImageKHR = gfx11_CmdCopyMemoryToImageKHR,
    .CmdCopyImageToMemoryKHR = gfx11_CmdCopyImageToMemoryKHR,
    .CmdUpdateMemoryKHR = gfx11_CmdUpdateMemoryKHR,
    .CmdFillMemoryKHR = gfx11_CmdFillMemoryKHR,
    .CmdCopyQueryPoolResultsToMemoryKHR = gfx11_CmdCopyQueryPoolResultsToMemoryKHR,
    .CmdBeginConditionalRendering2EXT = gfx11_CmdBeginConditionalRendering2EXT,
    .CmdBindTransformFeedbackBuffers2EXT = gfx11_CmdBindTransformFeedbackBuffers2EXT,
    .CmdBeginTransformFeedback2EXT = gfx11_CmdBeginTransformFeedback2EXT,
    .CmdEndTransformFeedback2EXT = gfx11_CmdEndTransformFeedback2EXT,
    .CmdDrawIndirectByteCount2EXT = gfx11_CmdDrawIndirectByteCount2EXT,
    .CmdWriteMarkerToMemoryAMD = gfx11_CmdWriteMarkerToMemoryAMD,
    .CmdBindIndexBuffer3KHR = gfx11_CmdBindIndexBuffer3KHR,
    .CmdBindVertexBuffers3KHR = gfx11_CmdBindVertexBuffers3KHR,
    .CmdDrawIndirect2KHR = gfx11_CmdDrawIndirect2KHR,
    .CmdDrawIndexedIndirect2KHR = gfx11_CmdDrawIndexedIndirect2KHR,
    .CmdDrawIndirectCount2KHR = gfx11_CmdDrawIndirectCount2KHR,
    .CmdDrawIndexedIndirectCount2KHR = gfx11_CmdDrawIndexedIndirectCount2KHR,
    .CmdDrawMeshTasksIndirect2EXT = gfx11_CmdDrawMeshTasksIndirect2EXT,
    .CmdDrawMeshTasksIndirectCount2EXT = gfx11_CmdDrawMeshTasksIndirectCount2EXT,
    .CmdDispatchIndirect2KHR = gfx11_CmdDispatchIndirect2KHR,
    .CreateAccelerationStructure2KHR = gfx11_CreateAccelerationStructure2KHR,
    .CmdSetDispatchParametersARM = gfx11_CmdSetDispatchParametersARM,
};
const struct vk_device_entrypoint_table gfx12_device_entrypoints = {
    .GetDeviceProcAddr = gfx12_GetDeviceProcAddr,
    .DestroyDevice = gfx12_DestroyDevice,
    .GetDeviceQueue = gfx12_GetDeviceQueue,
    .QueueSubmit = gfx12_QueueSubmit,
    .QueueWaitIdle = gfx12_QueueWaitIdle,
    .DeviceWaitIdle = gfx12_DeviceWaitIdle,
    .AllocateMemory = gfx12_AllocateMemory,
    .FreeMemory = gfx12_FreeMemory,
    .MapMemory = gfx12_MapMemory,
    .UnmapMemory = gfx12_UnmapMemory,
    .FlushMappedMemoryRanges = gfx12_FlushMappedMemoryRanges,
    .InvalidateMappedMemoryRanges = gfx12_InvalidateMappedMemoryRanges,
    .GetDeviceMemoryCommitment = gfx12_GetDeviceMemoryCommitment,
    .GetBufferMemoryRequirements = gfx12_GetBufferMemoryRequirements,
    .BindBufferMemory = gfx12_BindBufferMemory,
    .GetImageMemoryRequirements = gfx12_GetImageMemoryRequirements,
    .BindImageMemory = gfx12_BindImageMemory,
    .GetImageSparseMemoryRequirements = gfx12_GetImageSparseMemoryRequirements,
    .QueueBindSparse = gfx12_QueueBindSparse,
    .CreateFence = gfx12_CreateFence,
    .DestroyFence = gfx12_DestroyFence,
    .ResetFences = gfx12_ResetFences,
    .GetFenceStatus = gfx12_GetFenceStatus,
    .WaitForFences = gfx12_WaitForFences,
    .CreateSemaphore = gfx12_CreateSemaphore,
    .DestroySemaphore = gfx12_DestroySemaphore,
    .CreateEvent = gfx12_CreateEvent,
    .DestroyEvent = gfx12_DestroyEvent,
    .GetEventStatus = gfx12_GetEventStatus,
    .SetEvent = gfx12_SetEvent,
    .ResetEvent = gfx12_ResetEvent,
    .CreateQueryPool = gfx12_CreateQueryPool,
    .DestroyQueryPool = gfx12_DestroyQueryPool,
    .GetQueryPoolResults = gfx12_GetQueryPoolResults,
    .ResetQueryPool = gfx12_ResetQueryPool,
    .ResetQueryPoolEXT = gfx12_ResetQueryPoolEXT,
    .CreateBuffer = gfx12_CreateBuffer,
    .DestroyBuffer = gfx12_DestroyBuffer,
    .CreateBufferView = gfx12_CreateBufferView,
    .DestroyBufferView = gfx12_DestroyBufferView,
    .CreateImage = gfx12_CreateImage,
    .DestroyImage = gfx12_DestroyImage,
    .GetImageSubresourceLayout = gfx12_GetImageSubresourceLayout,
    .CreateImageView = gfx12_CreateImageView,
    .DestroyImageView = gfx12_DestroyImageView,
    .CreateShaderModule = gfx12_CreateShaderModule,
    .DestroyShaderModule = gfx12_DestroyShaderModule,
    .CreatePipelineCache = gfx12_CreatePipelineCache,
    .DestroyPipelineCache = gfx12_DestroyPipelineCache,
    .GetPipelineCacheData = gfx12_GetPipelineCacheData,
    .MergePipelineCaches = gfx12_MergePipelineCaches,
    .CreatePipelineBinariesKHR = gfx12_CreatePipelineBinariesKHR,
    .DestroyPipelineBinaryKHR = gfx12_DestroyPipelineBinaryKHR,
    .GetPipelineKeyKHR = gfx12_GetPipelineKeyKHR,
    .GetPipelineBinaryDataKHR = gfx12_GetPipelineBinaryDataKHR,
    .ReleaseCapturedPipelineDataKHR = gfx12_ReleaseCapturedPipelineDataKHR,
    .CreateGraphicsPipelines = gfx12_CreateGraphicsPipelines,
    .CreateComputePipelines = gfx12_CreateComputePipelines,
    .GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI = gfx12_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI,
    .DestroyPipeline = gfx12_DestroyPipeline,
    .CreatePipelineLayout = gfx12_CreatePipelineLayout,
    .DestroyPipelineLayout = gfx12_DestroyPipelineLayout,
    .CreateSampler = gfx12_CreateSampler,
    .DestroySampler = gfx12_DestroySampler,
    .CreateDescriptorSetLayout = gfx12_CreateDescriptorSetLayout,
    .DestroyDescriptorSetLayout = gfx12_DestroyDescriptorSetLayout,
    .CreateDescriptorPool = gfx12_CreateDescriptorPool,
    .DestroyDescriptorPool = gfx12_DestroyDescriptorPool,
    .ResetDescriptorPool = gfx12_ResetDescriptorPool,
    .AllocateDescriptorSets = gfx12_AllocateDescriptorSets,
    .FreeDescriptorSets = gfx12_FreeDescriptorSets,
    .UpdateDescriptorSets = gfx12_UpdateDescriptorSets,
    .CreateFramebuffer = gfx12_CreateFramebuffer,
    .DestroyFramebuffer = gfx12_DestroyFramebuffer,
    .CreateRenderPass = gfx12_CreateRenderPass,
    .DestroyRenderPass = gfx12_DestroyRenderPass,
    .GetRenderAreaGranularity = gfx12_GetRenderAreaGranularity,
    .GetRenderingAreaGranularity = gfx12_GetRenderingAreaGranularity,
    .GetRenderingAreaGranularityKHR = gfx12_GetRenderingAreaGranularityKHR,
    .CreateCommandPool = gfx12_CreateCommandPool,
    .DestroyCommandPool = gfx12_DestroyCommandPool,
    .ResetCommandPool = gfx12_ResetCommandPool,
    .AllocateCommandBuffers = gfx12_AllocateCommandBuffers,
    .FreeCommandBuffers = gfx12_FreeCommandBuffers,
    .BeginCommandBuffer = gfx12_BeginCommandBuffer,
    .EndCommandBuffer = gfx12_EndCommandBuffer,
    .ResetCommandBuffer = gfx12_ResetCommandBuffer,
    .CmdBindPipeline = gfx12_CmdBindPipeline,
    .CmdSetPrimitiveRestartIndexEXT = gfx12_CmdSetPrimitiveRestartIndexEXT,
    .CmdSetAttachmentFeedbackLoopEnableEXT = gfx12_CmdSetAttachmentFeedbackLoopEnableEXT,
    .CmdSetViewport = gfx12_CmdSetViewport,
    .CmdSetScissor = gfx12_CmdSetScissor,
    .CmdSetLineWidth = gfx12_CmdSetLineWidth,
    .CmdSetDepthBias = gfx12_CmdSetDepthBias,
    .CmdSetBlendConstants = gfx12_CmdSetBlendConstants,
    .CmdSetDepthBounds = gfx12_CmdSetDepthBounds,
    .CmdSetStencilCompareMask = gfx12_CmdSetStencilCompareMask,
    .CmdSetStencilWriteMask = gfx12_CmdSetStencilWriteMask,
    .CmdSetStencilReference = gfx12_CmdSetStencilReference,
    .CmdBindDescriptorSets = gfx12_CmdBindDescriptorSets,
    .CmdBindIndexBuffer = gfx12_CmdBindIndexBuffer,
    .CmdBindVertexBuffers = gfx12_CmdBindVertexBuffers,
    .CmdDraw = gfx12_CmdDraw,
    .CmdDrawIndexed = gfx12_CmdDrawIndexed,
    .CmdDrawMultiEXT = gfx12_CmdDrawMultiEXT,
    .CmdDrawMultiIndexedEXT = gfx12_CmdDrawMultiIndexedEXT,
    .CmdDrawIndirect = gfx12_CmdDrawIndirect,
    .CmdDrawIndexedIndirect = gfx12_CmdDrawIndexedIndirect,
    .CmdDispatch = gfx12_CmdDispatch,
    .CmdDispatchIndirect = gfx12_CmdDispatchIndirect,
    .CmdSubpassShadingHUAWEI = gfx12_CmdSubpassShadingHUAWEI,
    .CmdDrawClusterHUAWEI = gfx12_CmdDrawClusterHUAWEI,
    .CmdDrawClusterIndirectHUAWEI = gfx12_CmdDrawClusterIndirectHUAWEI,
    .CmdUpdatePipelineIndirectBufferNV = gfx12_CmdUpdatePipelineIndirectBufferNV,
    .CmdCopyBuffer = gfx12_CmdCopyBuffer,
    .CmdCopyImage = gfx12_CmdCopyImage,
    .CmdBlitImage = gfx12_CmdBlitImage,
    .CmdCopyBufferToImage = gfx12_CmdCopyBufferToImage,
    .CmdCopyImageToBuffer = gfx12_CmdCopyImageToBuffer,
    .CmdCopyMemoryIndirectNV = gfx12_CmdCopyMemoryIndirectNV,
    .CmdCopyMemoryIndirectKHR = gfx12_CmdCopyMemoryIndirectKHR,
    .CmdCopyMemoryToImageIndirectNV = gfx12_CmdCopyMemoryToImageIndirectNV,
    .CmdCopyMemoryToImageIndirectKHR = gfx12_CmdCopyMemoryToImageIndirectKHR,
    .CmdUpdateBuffer = gfx12_CmdUpdateBuffer,
    .CmdFillBuffer = gfx12_CmdFillBuffer,
    .CmdClearColorImage = gfx12_CmdClearColorImage,
    .CmdClearDepthStencilImage = gfx12_CmdClearDepthStencilImage,
    .CmdClearAttachments = gfx12_CmdClearAttachments,
    .CmdResolveImage = gfx12_CmdResolveImage,
    .CmdSetEvent = gfx12_CmdSetEvent,
    .CmdResetEvent = gfx12_CmdResetEvent,
    .CmdWaitEvents = gfx12_CmdWaitEvents,
    .CmdPipelineBarrier = gfx12_CmdPipelineBarrier,
    .CmdBeginQuery = gfx12_CmdBeginQuery,
    .CmdEndQuery = gfx12_CmdEndQuery,
    .CmdBeginConditionalRenderingEXT = gfx12_CmdBeginConditionalRenderingEXT,
    .CmdEndConditionalRenderingEXT = gfx12_CmdEndConditionalRenderingEXT,
    .CmdBeginCustomResolveEXT = gfx12_CmdBeginCustomResolveEXT,
    .CmdResetQueryPool = gfx12_CmdResetQueryPool,
    .CmdWriteTimestamp = gfx12_CmdWriteTimestamp,
    .CmdCopyQueryPoolResults = gfx12_CmdCopyQueryPoolResults,
    .CmdPushConstants = gfx12_CmdPushConstants,
    .CmdBeginRenderPass = gfx12_CmdBeginRenderPass,
    .CmdNextSubpass = gfx12_CmdNextSubpass,
    .CmdEndRenderPass = gfx12_CmdEndRenderPass,
    .CmdExecuteCommands = gfx12_CmdExecuteCommands,
    .CreateSharedSwapchainsKHR = gfx12_CreateSharedSwapchainsKHR,
    .CreateSwapchainKHR = gfx12_CreateSwapchainKHR,
    .DestroySwapchainKHR = gfx12_DestroySwapchainKHR,
    .GetSwapchainImagesKHR = gfx12_GetSwapchainImagesKHR,
    .AcquireNextImageKHR = gfx12_AcquireNextImageKHR,
    .QueuePresentKHR = gfx12_QueuePresentKHR,
    .DebugMarkerSetObjectNameEXT = gfx12_DebugMarkerSetObjectNameEXT,
    .DebugMarkerSetObjectTagEXT = gfx12_DebugMarkerSetObjectTagEXT,
    .CmdDebugMarkerBeginEXT = gfx12_CmdDebugMarkerBeginEXT,
    .CmdDebugMarkerEndEXT = gfx12_CmdDebugMarkerEndEXT,
    .CmdDebugMarkerInsertEXT = gfx12_CmdDebugMarkerInsertEXT,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleNV = gfx12_GetMemoryWin32HandleNV,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .CmdExecuteGeneratedCommandsNV = gfx12_CmdExecuteGeneratedCommandsNV,
    .CmdPreprocessGeneratedCommandsNV = gfx12_CmdPreprocessGeneratedCommandsNV,
    .CmdBindPipelineShaderGroupNV = gfx12_CmdBindPipelineShaderGroupNV,
    .GetGeneratedCommandsMemoryRequirementsNV = gfx12_GetGeneratedCommandsMemoryRequirementsNV,
    .CreateIndirectCommandsLayoutNV = gfx12_CreateIndirectCommandsLayoutNV,
    .DestroyIndirectCommandsLayoutNV = gfx12_DestroyIndirectCommandsLayoutNV,
    .CmdExecuteGeneratedCommandsEXT = gfx12_CmdExecuteGeneratedCommandsEXT,
    .CmdPreprocessGeneratedCommandsEXT = gfx12_CmdPreprocessGeneratedCommandsEXT,
    .GetGeneratedCommandsMemoryRequirementsEXT = gfx12_GetGeneratedCommandsMemoryRequirementsEXT,
    .CreateIndirectCommandsLayoutEXT = gfx12_CreateIndirectCommandsLayoutEXT,
    .DestroyIndirectCommandsLayoutEXT = gfx12_DestroyIndirectCommandsLayoutEXT,
    .CreateIndirectExecutionSetEXT = gfx12_CreateIndirectExecutionSetEXT,
    .DestroyIndirectExecutionSetEXT = gfx12_DestroyIndirectExecutionSetEXT,
    .UpdateIndirectExecutionSetPipelineEXT = gfx12_UpdateIndirectExecutionSetPipelineEXT,
    .UpdateIndirectExecutionSetShaderEXT = gfx12_UpdateIndirectExecutionSetShaderEXT,
    .CmdPushDescriptorSet = gfx12_CmdPushDescriptorSet,
    .CmdPushDescriptorSetKHR = gfx12_CmdPushDescriptorSetKHR,
    .TrimCommandPool = gfx12_TrimCommandPool,
    .TrimCommandPoolKHR = gfx12_TrimCommandPoolKHR,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleKHR = gfx12_GetMemoryWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandlePropertiesKHR = gfx12_GetMemoryWin32HandlePropertiesKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandlePropertiesKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryFdKHR = gfx12_GetMemoryFdKHR,
    .GetMemoryFdPropertiesKHR = gfx12_GetMemoryFdPropertiesKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandleFUCHSIA = gfx12_GetMemoryZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandlePropertiesFUCHSIA = gfx12_GetMemoryZirconHandlePropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandlePropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .GetMemoryRemoteAddressNV = gfx12_GetMemoryRemoteAddressNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreWin32HandleKHR = gfx12_GetSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportSemaphoreWin32HandleKHR = gfx12_ImportSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreFdKHR = gfx12_GetSemaphoreFdKHR,
    .ImportSemaphoreFdKHR = gfx12_ImportSemaphoreFdKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetSemaphoreZirconHandleFUCHSIA = gfx12_GetSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .ImportSemaphoreZirconHandleFUCHSIA = gfx12_ImportSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .ImportSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetFenceWin32HandleKHR = gfx12_GetFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportFenceWin32HandleKHR = gfx12_ImportFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetFenceFdKHR = gfx12_GetFenceFdKHR,
    .ImportFenceFdKHR = gfx12_ImportFenceFdKHR,
    .DisplayPowerControlEXT = gfx12_DisplayPowerControlEXT,
    .RegisterDeviceEventEXT = gfx12_RegisterDeviceEventEXT,
    .RegisterDisplayEventEXT = gfx12_RegisterDisplayEventEXT,
    .GetSwapchainCounterEXT = gfx12_GetSwapchainCounterEXT,
    .GetDeviceGroupPeerMemoryFeatures = gfx12_GetDeviceGroupPeerMemoryFeatures,
    .GetDeviceGroupPeerMemoryFeaturesKHR = gfx12_GetDeviceGroupPeerMemoryFeaturesKHR,
    .BindBufferMemory2 = gfx12_BindBufferMemory2,
    .BindBufferMemory2KHR = gfx12_BindBufferMemory2KHR,
    .BindImageMemory2 = gfx12_BindImageMemory2,
    .BindImageMemory2KHR = gfx12_BindImageMemory2KHR,
    .CmdSetDeviceMask = gfx12_CmdSetDeviceMask,
    .CmdSetDeviceMaskKHR = gfx12_CmdSetDeviceMaskKHR,
    .GetDeviceGroupPresentCapabilitiesKHR = gfx12_GetDeviceGroupPresentCapabilitiesKHR,
    .GetDeviceGroupSurfacePresentModesKHR = gfx12_GetDeviceGroupSurfacePresentModesKHR,
    .AcquireNextImage2KHR = gfx12_AcquireNextImage2KHR,
    .CmdDispatchBase = gfx12_CmdDispatchBase,
    .CmdDispatchBaseKHR = gfx12_CmdDispatchBaseKHR,
    .CreateDescriptorUpdateTemplate = gfx12_CreateDescriptorUpdateTemplate,
    .CreateDescriptorUpdateTemplateKHR = gfx12_CreateDescriptorUpdateTemplateKHR,
    .DestroyDescriptorUpdateTemplate = gfx12_DestroyDescriptorUpdateTemplate,
    .DestroyDescriptorUpdateTemplateKHR = gfx12_DestroyDescriptorUpdateTemplateKHR,
    .UpdateDescriptorSetWithTemplate = gfx12_UpdateDescriptorSetWithTemplate,
    .UpdateDescriptorSetWithTemplateKHR = gfx12_UpdateDescriptorSetWithTemplateKHR,
    .CmdPushDescriptorSetWithTemplate = gfx12_CmdPushDescriptorSetWithTemplate,
    .CmdPushDescriptorSetWithTemplateKHR = gfx12_CmdPushDescriptorSetWithTemplateKHR,
    .SetHdrMetadataEXT = gfx12_SetHdrMetadataEXT,
    .GetSwapchainStatusKHR = gfx12_GetSwapchainStatusKHR,
    .GetRefreshCycleDurationGOOGLE = gfx12_GetRefreshCycleDurationGOOGLE,
    .GetPastPresentationTimingGOOGLE = gfx12_GetPastPresentationTimingGOOGLE,
    .CmdSetViewportWScalingNV = gfx12_CmdSetViewportWScalingNV,
    .CmdSetDiscardRectangleEXT = gfx12_CmdSetDiscardRectangleEXT,
    .CmdSetDiscardRectangleEnableEXT = gfx12_CmdSetDiscardRectangleEnableEXT,
    .CmdSetDiscardRectangleModeEXT = gfx12_CmdSetDiscardRectangleModeEXT,
    .CmdSetSampleLocationsEXT = gfx12_CmdSetSampleLocationsEXT,
    .GetBufferMemoryRequirements2 = gfx12_GetBufferMemoryRequirements2,
    .GetBufferMemoryRequirements2KHR = gfx12_GetBufferMemoryRequirements2KHR,
    .GetImageMemoryRequirements2 = gfx12_GetImageMemoryRequirements2,
    .GetImageMemoryRequirements2KHR = gfx12_GetImageMemoryRequirements2KHR,
    .GetImageSparseMemoryRequirements2 = gfx12_GetImageSparseMemoryRequirements2,
    .GetImageSparseMemoryRequirements2KHR = gfx12_GetImageSparseMemoryRequirements2KHR,
    .GetDeviceBufferMemoryRequirements = gfx12_GetDeviceBufferMemoryRequirements,
    .GetDeviceBufferMemoryRequirementsKHR = gfx12_GetDeviceBufferMemoryRequirementsKHR,
    .GetDeviceImageMemoryRequirements = gfx12_GetDeviceImageMemoryRequirements,
    .GetDeviceImageMemoryRequirementsKHR = gfx12_GetDeviceImageMemoryRequirementsKHR,
    .GetDeviceImageSparseMemoryRequirements = gfx12_GetDeviceImageSparseMemoryRequirements,
    .GetDeviceImageSparseMemoryRequirementsKHR = gfx12_GetDeviceImageSparseMemoryRequirementsKHR,
    .CreateSamplerYcbcrConversion = gfx12_CreateSamplerYcbcrConversion,
    .CreateSamplerYcbcrConversionKHR = gfx12_CreateSamplerYcbcrConversionKHR,
    .DestroySamplerYcbcrConversion = gfx12_DestroySamplerYcbcrConversion,
    .DestroySamplerYcbcrConversionKHR = gfx12_DestroySamplerYcbcrConversionKHR,
    .GetDeviceQueue2 = gfx12_GetDeviceQueue2,
    .CreateValidationCacheEXT = gfx12_CreateValidationCacheEXT,
    .DestroyValidationCacheEXT = gfx12_DestroyValidationCacheEXT,
    .GetValidationCacheDataEXT = gfx12_GetValidationCacheDataEXT,
    .MergeValidationCachesEXT = gfx12_MergeValidationCachesEXT,
    .GetDescriptorSetLayoutSupport = gfx12_GetDescriptorSetLayoutSupport,
    .GetDescriptorSetLayoutSupportKHR = gfx12_GetDescriptorSetLayoutSupportKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsageANDROID = gfx12_GetSwapchainGrallocUsageANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsage2ANDROID = gfx12_GetSwapchainGrallocUsage2ANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsage2ANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .AcquireImageANDROID = gfx12_AcquireImageANDROID,
#elif defined(_MSC_VER)
    .AcquireImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .QueueSignalReleaseImageANDROID = gfx12_QueueSignalReleaseImageANDROID,
#elif defined(_MSC_VER)
    .QueueSignalReleaseImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .GetShaderInfoAMD = gfx12_GetShaderInfoAMD,
    .SetLocalDimmingAMD = gfx12_SetLocalDimmingAMD,
    .GetCalibratedTimestampsKHR = gfx12_GetCalibratedTimestampsKHR,
    .GetCalibratedTimestampsEXT = gfx12_GetCalibratedTimestampsEXT,
    .SetDebugUtilsObjectNameEXT = gfx12_SetDebugUtilsObjectNameEXT,
    .SetDebugUtilsObjectTagEXT = gfx12_SetDebugUtilsObjectTagEXT,
    .QueueBeginDebugUtilsLabelEXT = gfx12_QueueBeginDebugUtilsLabelEXT,
    .QueueEndDebugUtilsLabelEXT = gfx12_QueueEndDebugUtilsLabelEXT,
    .QueueInsertDebugUtilsLabelEXT = gfx12_QueueInsertDebugUtilsLabelEXT,
    .CmdBeginDebugUtilsLabelEXT = gfx12_CmdBeginDebugUtilsLabelEXT,
    .CmdEndDebugUtilsLabelEXT = gfx12_CmdEndDebugUtilsLabelEXT,
    .CmdInsertDebugUtilsLabelEXT = gfx12_CmdInsertDebugUtilsLabelEXT,
    .GetMemoryHostPointerPropertiesEXT = gfx12_GetMemoryHostPointerPropertiesEXT,
    .CmdWriteBufferMarkerAMD = gfx12_CmdWriteBufferMarkerAMD,
    .CreateRenderPass2 = gfx12_CreateRenderPass2,
    .CreateRenderPass2KHR = gfx12_CreateRenderPass2KHR,
    .CmdBeginRenderPass2 = gfx12_CmdBeginRenderPass2,
    .CmdBeginRenderPass2KHR = gfx12_CmdBeginRenderPass2KHR,
    .CmdNextSubpass2 = gfx12_CmdNextSubpass2,
    .CmdNextSubpass2KHR = gfx12_CmdNextSubpass2KHR,
    .CmdEndRenderPass2 = gfx12_CmdEndRenderPass2,
    .CmdEndRenderPass2KHR = gfx12_CmdEndRenderPass2KHR,
    .GetSemaphoreCounterValue = gfx12_GetSemaphoreCounterValue,
    .GetSemaphoreCounterValueKHR = gfx12_GetSemaphoreCounterValueKHR,
    .WaitSemaphores = gfx12_WaitSemaphores,
    .WaitSemaphoresKHR = gfx12_WaitSemaphoresKHR,
    .SignalSemaphore = gfx12_SignalSemaphore,
    .SignalSemaphoreKHR = gfx12_SignalSemaphoreKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetAndroidHardwareBufferPropertiesANDROID = gfx12_GetAndroidHardwareBufferPropertiesANDROID,
#elif defined(_MSC_VER)
    .GetAndroidHardwareBufferPropertiesANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetMemoryAndroidHardwareBufferANDROID = gfx12_GetMemoryAndroidHardwareBufferANDROID,
#elif defined(_MSC_VER)
    .GetMemoryAndroidHardwareBufferANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .CmdDrawIndirectCount = gfx12_CmdDrawIndirectCount,
    .CmdDrawIndirectCountKHR = gfx12_CmdDrawIndirectCountKHR,
    .CmdDrawIndirectCountAMD = gfx12_CmdDrawIndirectCountAMD,
    .CmdDrawIndexedIndirectCount = gfx12_CmdDrawIndexedIndirectCount,
    .CmdDrawIndexedIndirectCountKHR = gfx12_CmdDrawIndexedIndirectCountKHR,
    .CmdDrawIndexedIndirectCountAMD = gfx12_CmdDrawIndexedIndirectCountAMD,
    .CmdSetCheckpointNV = gfx12_CmdSetCheckpointNV,
    .GetQueueCheckpointDataNV = gfx12_GetQueueCheckpointDataNV,
    .CmdBindTransformFeedbackBuffersEXT = gfx12_CmdBindTransformFeedbackBuffersEXT,
    .CmdBeginTransformFeedbackEXT = gfx12_CmdBeginTransformFeedbackEXT,
    .CmdEndTransformFeedbackEXT = gfx12_CmdEndTransformFeedbackEXT,
    .CmdBeginQueryIndexedEXT = gfx12_CmdBeginQueryIndexedEXT,
    .CmdEndQueryIndexedEXT = gfx12_CmdEndQueryIndexedEXT,
    .CmdDrawIndirectByteCountEXT = gfx12_CmdDrawIndirectByteCountEXT,
    .CmdSetExclusiveScissorNV = gfx12_CmdSetExclusiveScissorNV,
    .CmdSetExclusiveScissorEnableNV = gfx12_CmdSetExclusiveScissorEnableNV,
    .CmdBindShadingRateImageNV = gfx12_CmdBindShadingRateImageNV,
    .CmdSetViewportShadingRatePaletteNV = gfx12_CmdSetViewportShadingRatePaletteNV,
    .CmdSetCoarseSampleOrderNV = gfx12_CmdSetCoarseSampleOrderNV,
    .CmdDrawMeshTasksNV = gfx12_CmdDrawMeshTasksNV,
    .CmdDrawMeshTasksIndirectNV = gfx12_CmdDrawMeshTasksIndirectNV,
    .CmdDrawMeshTasksIndirectCountNV = gfx12_CmdDrawMeshTasksIndirectCountNV,
    .CmdDrawMeshTasksEXT = gfx12_CmdDrawMeshTasksEXT,
    .CmdDrawMeshTasksIndirectEXT = gfx12_CmdDrawMeshTasksIndirectEXT,
    .CmdDrawMeshTasksIndirectCountEXT = gfx12_CmdDrawMeshTasksIndirectCountEXT,
    .CompileDeferredNV = gfx12_CompileDeferredNV,
    .CreateAccelerationStructureNV = gfx12_CreateAccelerationStructureNV,
    .CmdBindInvocationMaskHUAWEI = gfx12_CmdBindInvocationMaskHUAWEI,
    .DestroyAccelerationStructureKHR = gfx12_DestroyAccelerationStructureKHR,
    .DestroyAccelerationStructureNV = gfx12_DestroyAccelerationStructureNV,
    .GetAccelerationStructureMemoryRequirementsNV = gfx12_GetAccelerationStructureMemoryRequirementsNV,
    .BindAccelerationStructureMemoryNV = gfx12_BindAccelerationStructureMemoryNV,
    .CmdCopyAccelerationStructureNV = gfx12_CmdCopyAccelerationStructureNV,
    .CmdCopyAccelerationStructureKHR = gfx12_CmdCopyAccelerationStructureKHR,
    .CopyAccelerationStructureKHR = gfx12_CopyAccelerationStructureKHR,
    .CmdCopyAccelerationStructureToMemoryKHR = gfx12_CmdCopyAccelerationStructureToMemoryKHR,
    .CopyAccelerationStructureToMemoryKHR = gfx12_CopyAccelerationStructureToMemoryKHR,
    .CmdCopyMemoryToAccelerationStructureKHR = gfx12_CmdCopyMemoryToAccelerationStructureKHR,
    .CopyMemoryToAccelerationStructureKHR = gfx12_CopyMemoryToAccelerationStructureKHR,
    .CmdWriteAccelerationStructuresPropertiesKHR = gfx12_CmdWriteAccelerationStructuresPropertiesKHR,
    .CmdWriteAccelerationStructuresPropertiesNV = gfx12_CmdWriteAccelerationStructuresPropertiesNV,
    .CmdBuildAccelerationStructureNV = gfx12_CmdBuildAccelerationStructureNV,
    .WriteAccelerationStructuresPropertiesKHR = gfx12_WriteAccelerationStructuresPropertiesKHR,
    .CmdTraceRaysKHR = gfx12_CmdTraceRaysKHR,
    .CmdTraceRaysNV = gfx12_CmdTraceRaysNV,
    .GetRayTracingShaderGroupHandlesKHR = gfx12_GetRayTracingShaderGroupHandlesKHR,
    .GetRayTracingShaderGroupHandlesNV = gfx12_GetRayTracingShaderGroupHandlesNV,
    .GetRayTracingCaptureReplayShaderGroupHandlesKHR = gfx12_GetRayTracingCaptureReplayShaderGroupHandlesKHR,
    .GetAccelerationStructureHandleNV = gfx12_GetAccelerationStructureHandleNV,
    .CreateRayTracingPipelinesNV = gfx12_CreateRayTracingPipelinesNV,
    .CreateRayTracingPipelinesKHR = gfx12_CreateRayTracingPipelinesKHR,
    .CmdTraceRaysIndirectKHR = gfx12_CmdTraceRaysIndirectKHR,
    .CmdTraceRaysIndirect2KHR = gfx12_CmdTraceRaysIndirect2KHR,
    .GetClusterAccelerationStructureBuildSizesNV = gfx12_GetClusterAccelerationStructureBuildSizesNV,
    .CmdBuildClusterAccelerationStructureIndirectNV = gfx12_CmdBuildClusterAccelerationStructureIndirectNV,
    .GetDeviceAccelerationStructureCompatibilityKHR = gfx12_GetDeviceAccelerationStructureCompatibilityKHR,
    .GetRayTracingShaderGroupStackSizeKHR = gfx12_GetRayTracingShaderGroupStackSizeKHR,
    .CmdSetRayTracingPipelineStackSizeKHR = gfx12_CmdSetRayTracingPipelineStackSizeKHR,
    .GetImageViewHandleNVX = gfx12_GetImageViewHandleNVX,
    .GetImageViewHandle64NVX = gfx12_GetImageViewHandle64NVX,
    .GetImageViewAddressNVX = gfx12_GetImageViewAddressNVX,
    .GetDeviceCombinedImageSamplerIndexNVX = gfx12_GetDeviceCombinedImageSamplerIndexNVX,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetDeviceGroupSurfacePresentModes2EXT = gfx12_GetDeviceGroupSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetDeviceGroupSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireFullScreenExclusiveModeEXT = gfx12_AcquireFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .AcquireFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ReleaseFullScreenExclusiveModeEXT = gfx12_ReleaseFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .ReleaseFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .AcquireProfilingLockKHR = gfx12_AcquireProfilingLockKHR,
    .ReleaseProfilingLockKHR = gfx12_ReleaseProfilingLockKHR,
    .GetImageDrmFormatModifierPropertiesEXT = gfx12_GetImageDrmFormatModifierPropertiesEXT,
    .GetBufferOpaqueCaptureAddress = gfx12_GetBufferOpaqueCaptureAddress,
    .GetBufferOpaqueCaptureAddressKHR = gfx12_GetBufferOpaqueCaptureAddressKHR,
    .GetBufferDeviceAddress = gfx12_GetBufferDeviceAddress,
    .GetBufferDeviceAddressKHR = gfx12_GetBufferDeviceAddressKHR,
    .GetBufferDeviceAddressEXT = gfx12_GetBufferDeviceAddressEXT,
    .InitializePerformanceApiINTEL = gfx12_InitializePerformanceApiINTEL,
    .UninitializePerformanceApiINTEL = gfx12_UninitializePerformanceApiINTEL,
    .CmdSetPerformanceMarkerINTEL = gfx12_CmdSetPerformanceMarkerINTEL,
    .CmdSetPerformanceStreamMarkerINTEL = gfx12_CmdSetPerformanceStreamMarkerINTEL,
    .CmdSetPerformanceOverrideINTEL = gfx12_CmdSetPerformanceOverrideINTEL,
    .AcquirePerformanceConfigurationINTEL = gfx12_AcquirePerformanceConfigurationINTEL,
    .ReleasePerformanceConfigurationINTEL = gfx12_ReleasePerformanceConfigurationINTEL,
    .QueueSetPerformanceConfigurationINTEL = gfx12_QueueSetPerformanceConfigurationINTEL,
    .GetPerformanceParameterINTEL = gfx12_GetPerformanceParameterINTEL,
    .GetDeviceMemoryOpaqueCaptureAddress = gfx12_GetDeviceMemoryOpaqueCaptureAddress,
    .GetDeviceMemoryOpaqueCaptureAddressKHR = gfx12_GetDeviceMemoryOpaqueCaptureAddressKHR,
    .GetPipelineExecutablePropertiesKHR = gfx12_GetPipelineExecutablePropertiesKHR,
    .GetPipelineExecutableStatisticsKHR = gfx12_GetPipelineExecutableStatisticsKHR,
    .GetPipelineExecutableInternalRepresentationsKHR = gfx12_GetPipelineExecutableInternalRepresentationsKHR,
    .CmdSetLineStipple = gfx12_CmdSetLineStipple,
    .CmdSetLineStippleKHR = gfx12_CmdSetLineStippleKHR,
    .CmdSetLineStippleEXT = gfx12_CmdSetLineStippleEXT,
    .CreateAccelerationStructureKHR = gfx12_CreateAccelerationStructureKHR,
    .CmdBuildAccelerationStructuresKHR = gfx12_CmdBuildAccelerationStructuresKHR,
    .CmdBuildAccelerationStructuresIndirectKHR = gfx12_CmdBuildAccelerationStructuresIndirectKHR,
    .BuildAccelerationStructuresKHR = gfx12_BuildAccelerationStructuresKHR,
    .GetAccelerationStructureDeviceAddressKHR = gfx12_GetAccelerationStructureDeviceAddressKHR,
    .CreateDeferredOperationKHR = gfx12_CreateDeferredOperationKHR,
    .DestroyDeferredOperationKHR = gfx12_DestroyDeferredOperationKHR,
    .GetDeferredOperationMaxConcurrencyKHR = gfx12_GetDeferredOperationMaxConcurrencyKHR,
    .GetDeferredOperationResultKHR = gfx12_GetDeferredOperationResultKHR,
    .DeferredOperationJoinKHR = gfx12_DeferredOperationJoinKHR,
    .GetPipelineIndirectMemoryRequirementsNV = gfx12_GetPipelineIndirectMemoryRequirementsNV,
    .GetPipelineIndirectDeviceAddressNV = gfx12_GetPipelineIndirectDeviceAddressNV,
    .AntiLagUpdateAMD = gfx12_AntiLagUpdateAMD,
    .CmdSetCullMode = gfx12_CmdSetCullMode,
    .CmdSetCullModeEXT = gfx12_CmdSetCullModeEXT,
    .CmdSetFrontFace = gfx12_CmdSetFrontFace,
    .CmdSetFrontFaceEXT = gfx12_CmdSetFrontFaceEXT,
    .CmdSetPrimitiveTopology = gfx12_CmdSetPrimitiveTopology,
    .CmdSetPrimitiveTopologyEXT = gfx12_CmdSetPrimitiveTopologyEXT,
    .CmdSetViewportWithCount = gfx12_CmdSetViewportWithCount,
    .CmdSetViewportWithCountEXT = gfx12_CmdSetViewportWithCountEXT,
    .CmdSetScissorWithCount = gfx12_CmdSetScissorWithCount,
    .CmdSetScissorWithCountEXT = gfx12_CmdSetScissorWithCountEXT,
    .CmdBindIndexBuffer2 = gfx12_CmdBindIndexBuffer2,
    .CmdBindIndexBuffer2KHR = gfx12_CmdBindIndexBuffer2KHR,
    .CmdBindVertexBuffers2 = gfx12_CmdBindVertexBuffers2,
    .CmdBindVertexBuffers2EXT = gfx12_CmdBindVertexBuffers2EXT,
    .CmdSetDepthTestEnable = gfx12_CmdSetDepthTestEnable,
    .CmdSetDepthTestEnableEXT = gfx12_CmdSetDepthTestEnableEXT,
    .CmdSetDepthWriteEnable = gfx12_CmdSetDepthWriteEnable,
    .CmdSetDepthWriteEnableEXT = gfx12_CmdSetDepthWriteEnableEXT,
    .CmdSetDepthCompareOp = gfx12_CmdSetDepthCompareOp,
    .CmdSetDepthCompareOpEXT = gfx12_CmdSetDepthCompareOpEXT,
    .CmdSetDepthBoundsTestEnable = gfx12_CmdSetDepthBoundsTestEnable,
    .CmdSetDepthBoundsTestEnableEXT = gfx12_CmdSetDepthBoundsTestEnableEXT,
    .CmdSetStencilTestEnable = gfx12_CmdSetStencilTestEnable,
    .CmdSetStencilTestEnableEXT = gfx12_CmdSetStencilTestEnableEXT,
    .CmdSetStencilOp = gfx12_CmdSetStencilOp,
    .CmdSetStencilOpEXT = gfx12_CmdSetStencilOpEXT,
    .CmdSetPatchControlPointsEXT = gfx12_CmdSetPatchControlPointsEXT,
    .CmdSetRasterizerDiscardEnable = gfx12_CmdSetRasterizerDiscardEnable,
    .CmdSetRasterizerDiscardEnableEXT = gfx12_CmdSetRasterizerDiscardEnableEXT,
    .CmdSetDepthBiasEnable = gfx12_CmdSetDepthBiasEnable,
    .CmdSetDepthBiasEnableEXT = gfx12_CmdSetDepthBiasEnableEXT,
    .CmdSetLogicOpEXT = gfx12_CmdSetLogicOpEXT,
    .CmdSetPrimitiveRestartEnable = gfx12_CmdSetPrimitiveRestartEnable,
    .CmdSetPrimitiveRestartEnableEXT = gfx12_CmdSetPrimitiveRestartEnableEXT,
    .CmdSetTessellationDomainOriginEXT = gfx12_CmdSetTessellationDomainOriginEXT,
    .CmdSetDepthClampEnableEXT = gfx12_CmdSetDepthClampEnableEXT,
    .CmdSetPolygonModeEXT = gfx12_CmdSetPolygonModeEXT,
    .CmdSetRasterizationSamplesEXT = gfx12_CmdSetRasterizationSamplesEXT,
    .CmdSetSampleMaskEXT = gfx12_CmdSetSampleMaskEXT,
    .CmdSetAlphaToCoverageEnableEXT = gfx12_CmdSetAlphaToCoverageEnableEXT,
    .CmdSetAlphaToOneEnableEXT = gfx12_CmdSetAlphaToOneEnableEXT,
    .CmdSetLogicOpEnableEXT = gfx12_CmdSetLogicOpEnableEXT,
    .CmdSetColorBlendEnableEXT = gfx12_CmdSetColorBlendEnableEXT,
    .CmdSetColorBlendEquationEXT = gfx12_CmdSetColorBlendEquationEXT,
    .CmdSetColorWriteMaskEXT = gfx12_CmdSetColorWriteMaskEXT,
    .CmdSetRasterizationStreamEXT = gfx12_CmdSetRasterizationStreamEXT,
    .CmdSetConservativeRasterizationModeEXT = gfx12_CmdSetConservativeRasterizationModeEXT,
    .CmdSetExtraPrimitiveOverestimationSizeEXT = gfx12_CmdSetExtraPrimitiveOverestimationSizeEXT,
    .CmdSetDepthClipEnableEXT = gfx12_CmdSetDepthClipEnableEXT,
    .CmdSetSampleLocationsEnableEXT = gfx12_CmdSetSampleLocationsEnableEXT,
    .CmdSetColorBlendAdvancedEXT = gfx12_CmdSetColorBlendAdvancedEXT,
    .CmdSetProvokingVertexModeEXT = gfx12_CmdSetProvokingVertexModeEXT,
    .CmdSetLineRasterizationModeEXT = gfx12_CmdSetLineRasterizationModeEXT,
    .CmdSetLineStippleEnableEXT = gfx12_CmdSetLineStippleEnableEXT,
    .CmdSetDepthClipNegativeOneToOneEXT = gfx12_CmdSetDepthClipNegativeOneToOneEXT,
    .CmdSetViewportWScalingEnableNV = gfx12_CmdSetViewportWScalingEnableNV,
    .CmdSetViewportSwizzleNV = gfx12_CmdSetViewportSwizzleNV,
    .CmdSetCoverageToColorEnableNV = gfx12_CmdSetCoverageToColorEnableNV,
    .CmdSetCoverageToColorLocationNV = gfx12_CmdSetCoverageToColorLocationNV,
    .CmdSetCoverageModulationModeNV = gfx12_CmdSetCoverageModulationModeNV,
    .CmdSetCoverageModulationTableEnableNV = gfx12_CmdSetCoverageModulationTableEnableNV,
    .CmdSetCoverageModulationTableNV = gfx12_CmdSetCoverageModulationTableNV,
    .CmdSetShadingRateImageEnableNV = gfx12_CmdSetShadingRateImageEnableNV,
    .CmdSetCoverageReductionModeNV = gfx12_CmdSetCoverageReductionModeNV,
    .CmdSetRepresentativeFragmentTestEnableNV = gfx12_CmdSetRepresentativeFragmentTestEnableNV,
    .CreatePrivateDataSlot = gfx12_CreatePrivateDataSlot,
    .CreatePrivateDataSlotEXT = gfx12_CreatePrivateDataSlotEXT,
    .DestroyPrivateDataSlot = gfx12_DestroyPrivateDataSlot,
    .DestroyPrivateDataSlotEXT = gfx12_DestroyPrivateDataSlotEXT,
    .SetPrivateData = gfx12_SetPrivateData,
    .SetPrivateDataEXT = gfx12_SetPrivateDataEXT,
    .GetPrivateData = gfx12_GetPrivateData,
    .GetPrivateDataEXT = gfx12_GetPrivateDataEXT,
    .CmdCopyBuffer2 = gfx12_CmdCopyBuffer2,
    .CmdCopyBuffer2KHR = gfx12_CmdCopyBuffer2KHR,
    .CmdCopyImage2 = gfx12_CmdCopyImage2,
    .CmdCopyImage2KHR = gfx12_CmdCopyImage2KHR,
    .CmdBlitImage2 = gfx12_CmdBlitImage2,
    .CmdBlitImage2KHR = gfx12_CmdBlitImage2KHR,
    .CmdCopyBufferToImage2 = gfx12_CmdCopyBufferToImage2,
    .CmdCopyBufferToImage2KHR = gfx12_CmdCopyBufferToImage2KHR,
    .CmdCopyImageToBuffer2 = gfx12_CmdCopyImageToBuffer2,
    .CmdCopyImageToBuffer2KHR = gfx12_CmdCopyImageToBuffer2KHR,
    .CmdResolveImage2 = gfx12_CmdResolveImage2,
    .CmdResolveImage2KHR = gfx12_CmdResolveImage2KHR,
    .CmdSetFragmentShadingRateKHR = gfx12_CmdSetFragmentShadingRateKHR,
    .CmdSetFragmentShadingRateEnumNV = gfx12_CmdSetFragmentShadingRateEnumNV,
    .GetAccelerationStructureBuildSizesKHR = gfx12_GetAccelerationStructureBuildSizesKHR,
    .CmdSetVertexInputEXT = gfx12_CmdSetVertexInputEXT,
    .CmdSetColorWriteEnableEXT = gfx12_CmdSetColorWriteEnableEXT,
    .CmdSetEvent2 = gfx12_CmdSetEvent2,
    .CmdSetEvent2KHR = gfx12_CmdSetEvent2KHR,
    .CmdResetEvent2 = gfx12_CmdResetEvent2,
    .CmdResetEvent2KHR = gfx12_CmdResetEvent2KHR,
    .CmdWaitEvents2 = gfx12_CmdWaitEvents2,
    .CmdWaitEvents2KHR = gfx12_CmdWaitEvents2KHR,
    .CmdPipelineBarrier2 = gfx12_CmdPipelineBarrier2,
    .CmdPipelineBarrier2KHR = gfx12_CmdPipelineBarrier2KHR,
    .QueueSubmit2 = gfx12_QueueSubmit2,
    .QueueSubmit2KHR = gfx12_QueueSubmit2KHR,
    .CmdWriteTimestamp2 = gfx12_CmdWriteTimestamp2,
    .CmdWriteTimestamp2KHR = gfx12_CmdWriteTimestamp2KHR,
    .CmdWriteBufferMarker2AMD = gfx12_CmdWriteBufferMarker2AMD,
    .GetQueueCheckpointData2NV = gfx12_GetQueueCheckpointData2NV,
    .CopyMemoryToImage = gfx12_CopyMemoryToImage,
    .CopyMemoryToImageEXT = gfx12_CopyMemoryToImageEXT,
    .CopyImageToMemory = gfx12_CopyImageToMemory,
    .CopyImageToMemoryEXT = gfx12_CopyImageToMemoryEXT,
    .CopyImageToImage = gfx12_CopyImageToImage,
    .CopyImageToImageEXT = gfx12_CopyImageToImageEXT,
    .TransitionImageLayout = gfx12_TransitionImageLayout,
    .TransitionImageLayoutEXT = gfx12_TransitionImageLayoutEXT,
    .CreateVideoSessionKHR = gfx12_CreateVideoSessionKHR,
    .DestroyVideoSessionKHR = gfx12_DestroyVideoSessionKHR,
    .CreateVideoSessionParametersKHR = gfx12_CreateVideoSessionParametersKHR,
    .UpdateVideoSessionParametersKHR = gfx12_UpdateVideoSessionParametersKHR,
    .GetEncodedVideoSessionParametersKHR = gfx12_GetEncodedVideoSessionParametersKHR,
    .DestroyVideoSessionParametersKHR = gfx12_DestroyVideoSessionParametersKHR,
    .GetVideoSessionMemoryRequirementsKHR = gfx12_GetVideoSessionMemoryRequirementsKHR,
    .BindVideoSessionMemoryKHR = gfx12_BindVideoSessionMemoryKHR,
    .CmdDecodeVideoKHR = gfx12_CmdDecodeVideoKHR,
    .CmdBeginVideoCodingKHR = gfx12_CmdBeginVideoCodingKHR,
    .CmdControlVideoCodingKHR = gfx12_CmdControlVideoCodingKHR,
    .CmdEndVideoCodingKHR = gfx12_CmdEndVideoCodingKHR,
    .CmdEncodeVideoKHR = gfx12_CmdEncodeVideoKHR,
    .CmdDecompressMemoryNV = gfx12_CmdDecompressMemoryNV,
    .CmdDecompressMemoryIndirectCountNV = gfx12_CmdDecompressMemoryIndirectCountNV,
    .GetPartitionedAccelerationStructuresBuildSizesNV = gfx12_GetPartitionedAccelerationStructuresBuildSizesNV,
    .CmdBuildPartitionedAccelerationStructuresNV = gfx12_CmdBuildPartitionedAccelerationStructuresNV,
    .CmdDecompressMemoryEXT = gfx12_CmdDecompressMemoryEXT,
    .CmdDecompressMemoryIndirectCountEXT = gfx12_CmdDecompressMemoryIndirectCountEXT,
    .CreateCuModuleNVX = gfx12_CreateCuModuleNVX,
    .CreateCuFunctionNVX = gfx12_CreateCuFunctionNVX,
    .DestroyCuModuleNVX = gfx12_DestroyCuModuleNVX,
    .DestroyCuFunctionNVX = gfx12_DestroyCuFunctionNVX,
    .CmdCuLaunchKernelNVX = gfx12_CmdCuLaunchKernelNVX,
    .GetDescriptorSetLayoutSizeEXT = gfx12_GetDescriptorSetLayoutSizeEXT,
    .GetDescriptorSetLayoutBindingOffsetEXT = gfx12_GetDescriptorSetLayoutBindingOffsetEXT,
    .GetDescriptorEXT = gfx12_GetDescriptorEXT,
    .CmdBindDescriptorBuffersEXT = gfx12_CmdBindDescriptorBuffersEXT,
    .CmdSetDescriptorBufferOffsetsEXT = gfx12_CmdSetDescriptorBufferOffsetsEXT,
    .CmdBindDescriptorBufferEmbeddedSamplersEXT = gfx12_CmdBindDescriptorBufferEmbeddedSamplersEXT,
    .GetBufferOpaqueCaptureDescriptorDataEXT = gfx12_GetBufferOpaqueCaptureDescriptorDataEXT,
    .GetImageOpaqueCaptureDescriptorDataEXT = gfx12_GetImageOpaqueCaptureDescriptorDataEXT,
    .GetImageViewOpaqueCaptureDescriptorDataEXT = gfx12_GetImageViewOpaqueCaptureDescriptorDataEXT,
    .GetSamplerOpaqueCaptureDescriptorDataEXT = gfx12_GetSamplerOpaqueCaptureDescriptorDataEXT,
    .GetAccelerationStructureOpaqueCaptureDescriptorDataEXT = gfx12_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT,
    .SetDeviceMemoryPriorityEXT = gfx12_SetDeviceMemoryPriorityEXT,
    .WaitForPresent2KHR = gfx12_WaitForPresent2KHR,
    .WaitForPresentKHR = gfx12_WaitForPresentKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateBufferCollectionFUCHSIA = gfx12_CreateBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .CreateBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionBufferConstraintsFUCHSIA = gfx12_SetBufferCollectionBufferConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionBufferConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionImageConstraintsFUCHSIA = gfx12_SetBufferCollectionImageConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionImageConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .DestroyBufferCollectionFUCHSIA = gfx12_DestroyBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .DestroyBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetBufferCollectionPropertiesFUCHSIA = gfx12_GetBufferCollectionPropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetBufferCollectionPropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .CmdBeginRendering = gfx12_CmdBeginRendering,
    .CmdBeginRenderingKHR = gfx12_CmdBeginRenderingKHR,
    .CmdEndRendering = gfx12_CmdEndRendering,
    .CmdEndRendering2KHR = gfx12_CmdEndRendering2KHR,
    .CmdEndRendering2EXT = gfx12_CmdEndRendering2EXT,
    .CmdEndRenderingKHR = gfx12_CmdEndRenderingKHR,
    .GetDescriptorSetLayoutHostMappingInfoVALVE = gfx12_GetDescriptorSetLayoutHostMappingInfoVALVE,
    .GetDescriptorSetHostMappingVALVE = gfx12_GetDescriptorSetHostMappingVALVE,
    .CreateMicromapEXT = gfx12_CreateMicromapEXT,
    .CmdBuildMicromapsEXT = gfx12_CmdBuildMicromapsEXT,
    .BuildMicromapsEXT = gfx12_BuildMicromapsEXT,
    .DestroyMicromapEXT = gfx12_DestroyMicromapEXT,
    .CmdCopyMicromapEXT = gfx12_CmdCopyMicromapEXT,
    .CopyMicromapEXT = gfx12_CopyMicromapEXT,
    .CmdCopyMicromapToMemoryEXT = gfx12_CmdCopyMicromapToMemoryEXT,
    .CopyMicromapToMemoryEXT = gfx12_CopyMicromapToMemoryEXT,
    .CmdCopyMemoryToMicromapEXT = gfx12_CmdCopyMemoryToMicromapEXT,
    .CopyMemoryToMicromapEXT = gfx12_CopyMemoryToMicromapEXT,
    .CmdWriteMicromapsPropertiesEXT = gfx12_CmdWriteMicromapsPropertiesEXT,
    .WriteMicromapsPropertiesEXT = gfx12_WriteMicromapsPropertiesEXT,
    .GetDeviceMicromapCompatibilityEXT = gfx12_GetDeviceMicromapCompatibilityEXT,
    .GetMicromapBuildSizesEXT = gfx12_GetMicromapBuildSizesEXT,
    .GetShaderModuleIdentifierEXT = gfx12_GetShaderModuleIdentifierEXT,
    .GetShaderModuleCreateInfoIdentifierEXT = gfx12_GetShaderModuleCreateInfoIdentifierEXT,
    .GetImageSubresourceLayout2 = gfx12_GetImageSubresourceLayout2,
    .GetImageSubresourceLayout2KHR = gfx12_GetImageSubresourceLayout2KHR,
    .GetImageSubresourceLayout2EXT = gfx12_GetImageSubresourceLayout2EXT,
    .GetPipelinePropertiesEXT = gfx12_GetPipelinePropertiesEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .ExportMetalObjectsEXT = gfx12_ExportMetalObjectsEXT,
#elif defined(_MSC_VER)
    .ExportMetalObjectsEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CmdBindTileMemoryQCOM = gfx12_CmdBindTileMemoryQCOM,
    .GetFramebufferTilePropertiesQCOM = gfx12_GetFramebufferTilePropertiesQCOM,
    .GetDynamicRenderingTilePropertiesQCOM = gfx12_GetDynamicRenderingTilePropertiesQCOM,
    .CreateOpticalFlowSessionNV = gfx12_CreateOpticalFlowSessionNV,
    .DestroyOpticalFlowSessionNV = gfx12_DestroyOpticalFlowSessionNV,
    .BindOpticalFlowSessionImageNV = gfx12_BindOpticalFlowSessionImageNV,
    .CmdOpticalFlowExecuteNV = gfx12_CmdOpticalFlowExecuteNV,
    .GetDeviceFaultInfoEXT = gfx12_GetDeviceFaultInfoEXT,
    .GetDeviceFaultReportsKHR = gfx12_GetDeviceFaultReportsKHR,
    .GetDeviceFaultDebugInfoKHR = gfx12_GetDeviceFaultDebugInfoKHR,
    .CmdSetDepthBias2EXT = gfx12_CmdSetDepthBias2EXT,
    .ReleaseSwapchainImagesKHR = gfx12_ReleaseSwapchainImagesKHR,
    .ReleaseSwapchainImagesEXT = gfx12_ReleaseSwapchainImagesEXT,
    .GetDeviceImageSubresourceLayout = gfx12_GetDeviceImageSubresourceLayout,
    .GetDeviceImageSubresourceLayoutKHR = gfx12_GetDeviceImageSubresourceLayoutKHR,
    .MapMemory2 = gfx12_MapMemory2,
    .MapMemory2KHR = gfx12_MapMemory2KHR,
    .UnmapMemory2 = gfx12_UnmapMemory2,
    .UnmapMemory2KHR = gfx12_UnmapMemory2KHR,
    .CreateShadersEXT = gfx12_CreateShadersEXT,
    .DestroyShaderEXT = gfx12_DestroyShaderEXT,
    .GetShaderBinaryDataEXT = gfx12_GetShaderBinaryDataEXT,
    .CmdBindShadersEXT = gfx12_CmdBindShadersEXT,
    .SetSwapchainPresentTimingQueueSizeEXT = gfx12_SetSwapchainPresentTimingQueueSizeEXT,
    .GetSwapchainTimingPropertiesEXT = gfx12_GetSwapchainTimingPropertiesEXT,
    .GetSwapchainTimeDomainPropertiesEXT = gfx12_GetSwapchainTimeDomainPropertiesEXT,
    .GetPastPresentationTimingEXT = gfx12_GetPastPresentationTimingEXT,
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetScreenBufferPropertiesQNX = gfx12_GetScreenBufferPropertiesQNX,
#elif defined(_MSC_VER)
    .GetScreenBufferPropertiesQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CmdBindDescriptorSets2 = gfx12_CmdBindDescriptorSets2,
    .CmdBindDescriptorSets2KHR = gfx12_CmdBindDescriptorSets2KHR,
    .CmdPushConstants2 = gfx12_CmdPushConstants2,
    .CmdPushConstants2KHR = gfx12_CmdPushConstants2KHR,
    .CmdPushDescriptorSet2 = gfx12_CmdPushDescriptorSet2,
    .CmdPushDescriptorSet2KHR = gfx12_CmdPushDescriptorSet2KHR,
    .CmdPushDescriptorSetWithTemplate2 = gfx12_CmdPushDescriptorSetWithTemplate2,
    .CmdPushDescriptorSetWithTemplate2KHR = gfx12_CmdPushDescriptorSetWithTemplate2KHR,
    .CmdSetDescriptorBufferOffsets2EXT = gfx12_CmdSetDescriptorBufferOffsets2EXT,
    .CmdBindDescriptorBufferEmbeddedSamplers2EXT = gfx12_CmdBindDescriptorBufferEmbeddedSamplers2EXT,
    .SetLatencySleepModeNV = gfx12_SetLatencySleepModeNV,
    .LatencySleepNV = gfx12_LatencySleepNV,
    .SetLatencyMarkerNV = gfx12_SetLatencyMarkerNV,
    .GetLatencyTimingsNV = gfx12_GetLatencyTimingsNV,
    .QueueNotifyOutOfBandNV = gfx12_QueueNotifyOutOfBandNV,
    .CmdSetRenderingAttachmentLocations = gfx12_CmdSetRenderingAttachmentLocations,
    .CmdSetRenderingAttachmentLocationsKHR = gfx12_CmdSetRenderingAttachmentLocationsKHR,
    .CmdSetRenderingInputAttachmentIndices = gfx12_CmdSetRenderingInputAttachmentIndices,
    .CmdSetRenderingInputAttachmentIndicesKHR = gfx12_CmdSetRenderingInputAttachmentIndicesKHR,
    .CmdSetDepthClampRangeEXT = gfx12_CmdSetDepthClampRangeEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandleEXT = gfx12_GetMemoryMetalHandleEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandleEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandlePropertiesEXT = gfx12_GetMemoryMetalHandlePropertiesEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandlePropertiesEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .ConvertCooperativeVectorMatrixNV = gfx12_ConvertCooperativeVectorMatrixNV,
    .CmdConvertCooperativeVectorMatrixNV = gfx12_CmdConvertCooperativeVectorMatrixNV,
    .CmdDispatchTileQCOM = gfx12_CmdDispatchTileQCOM,
    .CmdBeginPerTileExecutionQCOM = gfx12_CmdBeginPerTileExecutionQCOM,
    .CmdEndPerTileExecutionQCOM = gfx12_CmdEndPerTileExecutionQCOM,
    .CreateExternalComputeQueueNV = gfx12_CreateExternalComputeQueueNV,
    .DestroyExternalComputeQueueNV = gfx12_DestroyExternalComputeQueueNV,
    .CreateShaderInstrumentationARM = gfx12_CreateShaderInstrumentationARM,
    .DestroyShaderInstrumentationARM = gfx12_DestroyShaderInstrumentationARM,
    .CmdBeginShaderInstrumentationARM = gfx12_CmdBeginShaderInstrumentationARM,
    .CmdEndShaderInstrumentationARM = gfx12_CmdEndShaderInstrumentationARM,
    .GetShaderInstrumentationValuesARM = gfx12_GetShaderInstrumentationValuesARM,
    .ClearShaderInstrumentationMetricsARM = gfx12_ClearShaderInstrumentationMetricsARM,
    .CreateTensorARM = gfx12_CreateTensorARM,
    .DestroyTensorARM = gfx12_DestroyTensorARM,
    .CreateTensorViewARM = gfx12_CreateTensorViewARM,
    .DestroyTensorViewARM = gfx12_DestroyTensorViewARM,
    .GetTensorMemoryRequirementsARM = gfx12_GetTensorMemoryRequirementsARM,
    .BindTensorMemoryARM = gfx12_BindTensorMemoryARM,
    .GetDeviceTensorMemoryRequirementsARM = gfx12_GetDeviceTensorMemoryRequirementsARM,
    .CmdCopyTensorARM = gfx12_CmdCopyTensorARM,
    .GetTensorOpaqueCaptureDescriptorDataARM = gfx12_GetTensorOpaqueCaptureDescriptorDataARM,
    .GetTensorViewOpaqueCaptureDescriptorDataARM = gfx12_GetTensorViewOpaqueCaptureDescriptorDataARM,
    .CreateDataGraphPipelinesARM = gfx12_CreateDataGraphPipelinesARM,
    .CreateDataGraphPipelineSessionARM = gfx12_CreateDataGraphPipelineSessionARM,
    .GetDataGraphPipelineSessionBindPointRequirementsARM = gfx12_GetDataGraphPipelineSessionBindPointRequirementsARM,
    .GetDataGraphPipelineSessionMemoryRequirementsARM = gfx12_GetDataGraphPipelineSessionMemoryRequirementsARM,
    .BindDataGraphPipelineSessionMemoryARM = gfx12_BindDataGraphPipelineSessionMemoryARM,
    .DestroyDataGraphPipelineSessionARM = gfx12_DestroyDataGraphPipelineSessionARM,
    .CmdDispatchDataGraphARM = gfx12_CmdDispatchDataGraphARM,
    .GetDataGraphPipelineAvailablePropertiesARM = gfx12_GetDataGraphPipelineAvailablePropertiesARM,
    .GetDataGraphPipelinePropertiesARM = gfx12_GetDataGraphPipelinePropertiesARM,
#ifdef VK_USE_PLATFORM_OHOS
    .GetNativeBufferPropertiesOHOS = gfx12_GetNativeBufferPropertiesOHOS,
#elif defined(_MSC_VER)
    .GetNativeBufferPropertiesOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
    .GetMemoryNativeBufferOHOS = gfx12_GetMemoryNativeBufferOHOS,
#elif defined(_MSC_VER)
    .GetMemoryNativeBufferOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .QueueSetPerfHintQCOM = gfx12_QueueSetPerfHintQCOM,
    .CmdSetComputeOccupancyPriorityNV = gfx12_CmdSetComputeOccupancyPriorityNV,
    .WriteSamplerDescriptorsEXT = gfx12_WriteSamplerDescriptorsEXT,
    .WriteResourceDescriptorsEXT = gfx12_WriteResourceDescriptorsEXT,
    .CmdBindSamplerHeapEXT = gfx12_CmdBindSamplerHeapEXT,
    .CmdBindResourceHeapEXT = gfx12_CmdBindResourceHeapEXT,
    .CmdPushDataEXT = gfx12_CmdPushDataEXT,
    .RegisterCustomBorderColorEXT = gfx12_RegisterCustomBorderColorEXT,
    .UnregisterCustomBorderColorEXT = gfx12_UnregisterCustomBorderColorEXT,
    .GetImageOpaqueCaptureDataEXT = gfx12_GetImageOpaqueCaptureDataEXT,
    .GetTensorOpaqueCaptureDataARM = gfx12_GetTensorOpaqueCaptureDataARM,
    .CmdCopyMemoryKHR = gfx12_CmdCopyMemoryKHR,
    .CmdCopyMemoryToImageKHR = gfx12_CmdCopyMemoryToImageKHR,
    .CmdCopyImageToMemoryKHR = gfx12_CmdCopyImageToMemoryKHR,
    .CmdUpdateMemoryKHR = gfx12_CmdUpdateMemoryKHR,
    .CmdFillMemoryKHR = gfx12_CmdFillMemoryKHR,
    .CmdCopyQueryPoolResultsToMemoryKHR = gfx12_CmdCopyQueryPoolResultsToMemoryKHR,
    .CmdBeginConditionalRendering2EXT = gfx12_CmdBeginConditionalRendering2EXT,
    .CmdBindTransformFeedbackBuffers2EXT = gfx12_CmdBindTransformFeedbackBuffers2EXT,
    .CmdBeginTransformFeedback2EXT = gfx12_CmdBeginTransformFeedback2EXT,
    .CmdEndTransformFeedback2EXT = gfx12_CmdEndTransformFeedback2EXT,
    .CmdDrawIndirectByteCount2EXT = gfx12_CmdDrawIndirectByteCount2EXT,
    .CmdWriteMarkerToMemoryAMD = gfx12_CmdWriteMarkerToMemoryAMD,
    .CmdBindIndexBuffer3KHR = gfx12_CmdBindIndexBuffer3KHR,
    .CmdBindVertexBuffers3KHR = gfx12_CmdBindVertexBuffers3KHR,
    .CmdDrawIndirect2KHR = gfx12_CmdDrawIndirect2KHR,
    .CmdDrawIndexedIndirect2KHR = gfx12_CmdDrawIndexedIndirect2KHR,
    .CmdDrawIndirectCount2KHR = gfx12_CmdDrawIndirectCount2KHR,
    .CmdDrawIndexedIndirectCount2KHR = gfx12_CmdDrawIndexedIndirectCount2KHR,
    .CmdDrawMeshTasksIndirect2EXT = gfx12_CmdDrawMeshTasksIndirect2EXT,
    .CmdDrawMeshTasksIndirectCount2EXT = gfx12_CmdDrawMeshTasksIndirectCount2EXT,
    .CmdDispatchIndirect2KHR = gfx12_CmdDispatchIndirect2KHR,
    .CreateAccelerationStructure2KHR = gfx12_CreateAccelerationStructure2KHR,
    .CmdSetDispatchParametersARM = gfx12_CmdSetDispatchParametersARM,
};
const struct vk_device_entrypoint_table gfx125_device_entrypoints = {
    .GetDeviceProcAddr = gfx125_GetDeviceProcAddr,
    .DestroyDevice = gfx125_DestroyDevice,
    .GetDeviceQueue = gfx125_GetDeviceQueue,
    .QueueSubmit = gfx125_QueueSubmit,
    .QueueWaitIdle = gfx125_QueueWaitIdle,
    .DeviceWaitIdle = gfx125_DeviceWaitIdle,
    .AllocateMemory = gfx125_AllocateMemory,
    .FreeMemory = gfx125_FreeMemory,
    .MapMemory = gfx125_MapMemory,
    .UnmapMemory = gfx125_UnmapMemory,
    .FlushMappedMemoryRanges = gfx125_FlushMappedMemoryRanges,
    .InvalidateMappedMemoryRanges = gfx125_InvalidateMappedMemoryRanges,
    .GetDeviceMemoryCommitment = gfx125_GetDeviceMemoryCommitment,
    .GetBufferMemoryRequirements = gfx125_GetBufferMemoryRequirements,
    .BindBufferMemory = gfx125_BindBufferMemory,
    .GetImageMemoryRequirements = gfx125_GetImageMemoryRequirements,
    .BindImageMemory = gfx125_BindImageMemory,
    .GetImageSparseMemoryRequirements = gfx125_GetImageSparseMemoryRequirements,
    .QueueBindSparse = gfx125_QueueBindSparse,
    .CreateFence = gfx125_CreateFence,
    .DestroyFence = gfx125_DestroyFence,
    .ResetFences = gfx125_ResetFences,
    .GetFenceStatus = gfx125_GetFenceStatus,
    .WaitForFences = gfx125_WaitForFences,
    .CreateSemaphore = gfx125_CreateSemaphore,
    .DestroySemaphore = gfx125_DestroySemaphore,
    .CreateEvent = gfx125_CreateEvent,
    .DestroyEvent = gfx125_DestroyEvent,
    .GetEventStatus = gfx125_GetEventStatus,
    .SetEvent = gfx125_SetEvent,
    .ResetEvent = gfx125_ResetEvent,
    .CreateQueryPool = gfx125_CreateQueryPool,
    .DestroyQueryPool = gfx125_DestroyQueryPool,
    .GetQueryPoolResults = gfx125_GetQueryPoolResults,
    .ResetQueryPool = gfx125_ResetQueryPool,
    .ResetQueryPoolEXT = gfx125_ResetQueryPoolEXT,
    .CreateBuffer = gfx125_CreateBuffer,
    .DestroyBuffer = gfx125_DestroyBuffer,
    .CreateBufferView = gfx125_CreateBufferView,
    .DestroyBufferView = gfx125_DestroyBufferView,
    .CreateImage = gfx125_CreateImage,
    .DestroyImage = gfx125_DestroyImage,
    .GetImageSubresourceLayout = gfx125_GetImageSubresourceLayout,
    .CreateImageView = gfx125_CreateImageView,
    .DestroyImageView = gfx125_DestroyImageView,
    .CreateShaderModule = gfx125_CreateShaderModule,
    .DestroyShaderModule = gfx125_DestroyShaderModule,
    .CreatePipelineCache = gfx125_CreatePipelineCache,
    .DestroyPipelineCache = gfx125_DestroyPipelineCache,
    .GetPipelineCacheData = gfx125_GetPipelineCacheData,
    .MergePipelineCaches = gfx125_MergePipelineCaches,
    .CreatePipelineBinariesKHR = gfx125_CreatePipelineBinariesKHR,
    .DestroyPipelineBinaryKHR = gfx125_DestroyPipelineBinaryKHR,
    .GetPipelineKeyKHR = gfx125_GetPipelineKeyKHR,
    .GetPipelineBinaryDataKHR = gfx125_GetPipelineBinaryDataKHR,
    .ReleaseCapturedPipelineDataKHR = gfx125_ReleaseCapturedPipelineDataKHR,
    .CreateGraphicsPipelines = gfx125_CreateGraphicsPipelines,
    .CreateComputePipelines = gfx125_CreateComputePipelines,
    .GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI = gfx125_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI,
    .DestroyPipeline = gfx125_DestroyPipeline,
    .CreatePipelineLayout = gfx125_CreatePipelineLayout,
    .DestroyPipelineLayout = gfx125_DestroyPipelineLayout,
    .CreateSampler = gfx125_CreateSampler,
    .DestroySampler = gfx125_DestroySampler,
    .CreateDescriptorSetLayout = gfx125_CreateDescriptorSetLayout,
    .DestroyDescriptorSetLayout = gfx125_DestroyDescriptorSetLayout,
    .CreateDescriptorPool = gfx125_CreateDescriptorPool,
    .DestroyDescriptorPool = gfx125_DestroyDescriptorPool,
    .ResetDescriptorPool = gfx125_ResetDescriptorPool,
    .AllocateDescriptorSets = gfx125_AllocateDescriptorSets,
    .FreeDescriptorSets = gfx125_FreeDescriptorSets,
    .UpdateDescriptorSets = gfx125_UpdateDescriptorSets,
    .CreateFramebuffer = gfx125_CreateFramebuffer,
    .DestroyFramebuffer = gfx125_DestroyFramebuffer,
    .CreateRenderPass = gfx125_CreateRenderPass,
    .DestroyRenderPass = gfx125_DestroyRenderPass,
    .GetRenderAreaGranularity = gfx125_GetRenderAreaGranularity,
    .GetRenderingAreaGranularity = gfx125_GetRenderingAreaGranularity,
    .GetRenderingAreaGranularityKHR = gfx125_GetRenderingAreaGranularityKHR,
    .CreateCommandPool = gfx125_CreateCommandPool,
    .DestroyCommandPool = gfx125_DestroyCommandPool,
    .ResetCommandPool = gfx125_ResetCommandPool,
    .AllocateCommandBuffers = gfx125_AllocateCommandBuffers,
    .FreeCommandBuffers = gfx125_FreeCommandBuffers,
    .BeginCommandBuffer = gfx125_BeginCommandBuffer,
    .EndCommandBuffer = gfx125_EndCommandBuffer,
    .ResetCommandBuffer = gfx125_ResetCommandBuffer,
    .CmdBindPipeline = gfx125_CmdBindPipeline,
    .CmdSetPrimitiveRestartIndexEXT = gfx125_CmdSetPrimitiveRestartIndexEXT,
    .CmdSetAttachmentFeedbackLoopEnableEXT = gfx125_CmdSetAttachmentFeedbackLoopEnableEXT,
    .CmdSetViewport = gfx125_CmdSetViewport,
    .CmdSetScissor = gfx125_CmdSetScissor,
    .CmdSetLineWidth = gfx125_CmdSetLineWidth,
    .CmdSetDepthBias = gfx125_CmdSetDepthBias,
    .CmdSetBlendConstants = gfx125_CmdSetBlendConstants,
    .CmdSetDepthBounds = gfx125_CmdSetDepthBounds,
    .CmdSetStencilCompareMask = gfx125_CmdSetStencilCompareMask,
    .CmdSetStencilWriteMask = gfx125_CmdSetStencilWriteMask,
    .CmdSetStencilReference = gfx125_CmdSetStencilReference,
    .CmdBindDescriptorSets = gfx125_CmdBindDescriptorSets,
    .CmdBindIndexBuffer = gfx125_CmdBindIndexBuffer,
    .CmdBindVertexBuffers = gfx125_CmdBindVertexBuffers,
    .CmdDraw = gfx125_CmdDraw,
    .CmdDrawIndexed = gfx125_CmdDrawIndexed,
    .CmdDrawMultiEXT = gfx125_CmdDrawMultiEXT,
    .CmdDrawMultiIndexedEXT = gfx125_CmdDrawMultiIndexedEXT,
    .CmdDrawIndirect = gfx125_CmdDrawIndirect,
    .CmdDrawIndexedIndirect = gfx125_CmdDrawIndexedIndirect,
    .CmdDispatch = gfx125_CmdDispatch,
    .CmdDispatchIndirect = gfx125_CmdDispatchIndirect,
    .CmdSubpassShadingHUAWEI = gfx125_CmdSubpassShadingHUAWEI,
    .CmdDrawClusterHUAWEI = gfx125_CmdDrawClusterHUAWEI,
    .CmdDrawClusterIndirectHUAWEI = gfx125_CmdDrawClusterIndirectHUAWEI,
    .CmdUpdatePipelineIndirectBufferNV = gfx125_CmdUpdatePipelineIndirectBufferNV,
    .CmdCopyBuffer = gfx125_CmdCopyBuffer,
    .CmdCopyImage = gfx125_CmdCopyImage,
    .CmdBlitImage = gfx125_CmdBlitImage,
    .CmdCopyBufferToImage = gfx125_CmdCopyBufferToImage,
    .CmdCopyImageToBuffer = gfx125_CmdCopyImageToBuffer,
    .CmdCopyMemoryIndirectNV = gfx125_CmdCopyMemoryIndirectNV,
    .CmdCopyMemoryIndirectKHR = gfx125_CmdCopyMemoryIndirectKHR,
    .CmdCopyMemoryToImageIndirectNV = gfx125_CmdCopyMemoryToImageIndirectNV,
    .CmdCopyMemoryToImageIndirectKHR = gfx125_CmdCopyMemoryToImageIndirectKHR,
    .CmdUpdateBuffer = gfx125_CmdUpdateBuffer,
    .CmdFillBuffer = gfx125_CmdFillBuffer,
    .CmdClearColorImage = gfx125_CmdClearColorImage,
    .CmdClearDepthStencilImage = gfx125_CmdClearDepthStencilImage,
    .CmdClearAttachments = gfx125_CmdClearAttachments,
    .CmdResolveImage = gfx125_CmdResolveImage,
    .CmdSetEvent = gfx125_CmdSetEvent,
    .CmdResetEvent = gfx125_CmdResetEvent,
    .CmdWaitEvents = gfx125_CmdWaitEvents,
    .CmdPipelineBarrier = gfx125_CmdPipelineBarrier,
    .CmdBeginQuery = gfx125_CmdBeginQuery,
    .CmdEndQuery = gfx125_CmdEndQuery,
    .CmdBeginConditionalRenderingEXT = gfx125_CmdBeginConditionalRenderingEXT,
    .CmdEndConditionalRenderingEXT = gfx125_CmdEndConditionalRenderingEXT,
    .CmdBeginCustomResolveEXT = gfx125_CmdBeginCustomResolveEXT,
    .CmdResetQueryPool = gfx125_CmdResetQueryPool,
    .CmdWriteTimestamp = gfx125_CmdWriteTimestamp,
    .CmdCopyQueryPoolResults = gfx125_CmdCopyQueryPoolResults,
    .CmdPushConstants = gfx125_CmdPushConstants,
    .CmdBeginRenderPass = gfx125_CmdBeginRenderPass,
    .CmdNextSubpass = gfx125_CmdNextSubpass,
    .CmdEndRenderPass = gfx125_CmdEndRenderPass,
    .CmdExecuteCommands = gfx125_CmdExecuteCommands,
    .CreateSharedSwapchainsKHR = gfx125_CreateSharedSwapchainsKHR,
    .CreateSwapchainKHR = gfx125_CreateSwapchainKHR,
    .DestroySwapchainKHR = gfx125_DestroySwapchainKHR,
    .GetSwapchainImagesKHR = gfx125_GetSwapchainImagesKHR,
    .AcquireNextImageKHR = gfx125_AcquireNextImageKHR,
    .QueuePresentKHR = gfx125_QueuePresentKHR,
    .DebugMarkerSetObjectNameEXT = gfx125_DebugMarkerSetObjectNameEXT,
    .DebugMarkerSetObjectTagEXT = gfx125_DebugMarkerSetObjectTagEXT,
    .CmdDebugMarkerBeginEXT = gfx125_CmdDebugMarkerBeginEXT,
    .CmdDebugMarkerEndEXT = gfx125_CmdDebugMarkerEndEXT,
    .CmdDebugMarkerInsertEXT = gfx125_CmdDebugMarkerInsertEXT,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleNV = gfx125_GetMemoryWin32HandleNV,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .CmdExecuteGeneratedCommandsNV = gfx125_CmdExecuteGeneratedCommandsNV,
    .CmdPreprocessGeneratedCommandsNV = gfx125_CmdPreprocessGeneratedCommandsNV,
    .CmdBindPipelineShaderGroupNV = gfx125_CmdBindPipelineShaderGroupNV,
    .GetGeneratedCommandsMemoryRequirementsNV = gfx125_GetGeneratedCommandsMemoryRequirementsNV,
    .CreateIndirectCommandsLayoutNV = gfx125_CreateIndirectCommandsLayoutNV,
    .DestroyIndirectCommandsLayoutNV = gfx125_DestroyIndirectCommandsLayoutNV,
    .CmdExecuteGeneratedCommandsEXT = gfx125_CmdExecuteGeneratedCommandsEXT,
    .CmdPreprocessGeneratedCommandsEXT = gfx125_CmdPreprocessGeneratedCommandsEXT,
    .GetGeneratedCommandsMemoryRequirementsEXT = gfx125_GetGeneratedCommandsMemoryRequirementsEXT,
    .CreateIndirectCommandsLayoutEXT = gfx125_CreateIndirectCommandsLayoutEXT,
    .DestroyIndirectCommandsLayoutEXT = gfx125_DestroyIndirectCommandsLayoutEXT,
    .CreateIndirectExecutionSetEXT = gfx125_CreateIndirectExecutionSetEXT,
    .DestroyIndirectExecutionSetEXT = gfx125_DestroyIndirectExecutionSetEXT,
    .UpdateIndirectExecutionSetPipelineEXT = gfx125_UpdateIndirectExecutionSetPipelineEXT,
    .UpdateIndirectExecutionSetShaderEXT = gfx125_UpdateIndirectExecutionSetShaderEXT,
    .CmdPushDescriptorSet = gfx125_CmdPushDescriptorSet,
    .CmdPushDescriptorSetKHR = gfx125_CmdPushDescriptorSetKHR,
    .TrimCommandPool = gfx125_TrimCommandPool,
    .TrimCommandPoolKHR = gfx125_TrimCommandPoolKHR,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleKHR = gfx125_GetMemoryWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandlePropertiesKHR = gfx125_GetMemoryWin32HandlePropertiesKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandlePropertiesKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryFdKHR = gfx125_GetMemoryFdKHR,
    .GetMemoryFdPropertiesKHR = gfx125_GetMemoryFdPropertiesKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandleFUCHSIA = gfx125_GetMemoryZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandlePropertiesFUCHSIA = gfx125_GetMemoryZirconHandlePropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandlePropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .GetMemoryRemoteAddressNV = gfx125_GetMemoryRemoteAddressNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreWin32HandleKHR = gfx125_GetSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportSemaphoreWin32HandleKHR = gfx125_ImportSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreFdKHR = gfx125_GetSemaphoreFdKHR,
    .ImportSemaphoreFdKHR = gfx125_ImportSemaphoreFdKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetSemaphoreZirconHandleFUCHSIA = gfx125_GetSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .ImportSemaphoreZirconHandleFUCHSIA = gfx125_ImportSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .ImportSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetFenceWin32HandleKHR = gfx125_GetFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportFenceWin32HandleKHR = gfx125_ImportFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetFenceFdKHR = gfx125_GetFenceFdKHR,
    .ImportFenceFdKHR = gfx125_ImportFenceFdKHR,
    .DisplayPowerControlEXT = gfx125_DisplayPowerControlEXT,
    .RegisterDeviceEventEXT = gfx125_RegisterDeviceEventEXT,
    .RegisterDisplayEventEXT = gfx125_RegisterDisplayEventEXT,
    .GetSwapchainCounterEXT = gfx125_GetSwapchainCounterEXT,
    .GetDeviceGroupPeerMemoryFeatures = gfx125_GetDeviceGroupPeerMemoryFeatures,
    .GetDeviceGroupPeerMemoryFeaturesKHR = gfx125_GetDeviceGroupPeerMemoryFeaturesKHR,
    .BindBufferMemory2 = gfx125_BindBufferMemory2,
    .BindBufferMemory2KHR = gfx125_BindBufferMemory2KHR,
    .BindImageMemory2 = gfx125_BindImageMemory2,
    .BindImageMemory2KHR = gfx125_BindImageMemory2KHR,
    .CmdSetDeviceMask = gfx125_CmdSetDeviceMask,
    .CmdSetDeviceMaskKHR = gfx125_CmdSetDeviceMaskKHR,
    .GetDeviceGroupPresentCapabilitiesKHR = gfx125_GetDeviceGroupPresentCapabilitiesKHR,
    .GetDeviceGroupSurfacePresentModesKHR = gfx125_GetDeviceGroupSurfacePresentModesKHR,
    .AcquireNextImage2KHR = gfx125_AcquireNextImage2KHR,
    .CmdDispatchBase = gfx125_CmdDispatchBase,
    .CmdDispatchBaseKHR = gfx125_CmdDispatchBaseKHR,
    .CreateDescriptorUpdateTemplate = gfx125_CreateDescriptorUpdateTemplate,
    .CreateDescriptorUpdateTemplateKHR = gfx125_CreateDescriptorUpdateTemplateKHR,
    .DestroyDescriptorUpdateTemplate = gfx125_DestroyDescriptorUpdateTemplate,
    .DestroyDescriptorUpdateTemplateKHR = gfx125_DestroyDescriptorUpdateTemplateKHR,
    .UpdateDescriptorSetWithTemplate = gfx125_UpdateDescriptorSetWithTemplate,
    .UpdateDescriptorSetWithTemplateKHR = gfx125_UpdateDescriptorSetWithTemplateKHR,
    .CmdPushDescriptorSetWithTemplate = gfx125_CmdPushDescriptorSetWithTemplate,
    .CmdPushDescriptorSetWithTemplateKHR = gfx125_CmdPushDescriptorSetWithTemplateKHR,
    .SetHdrMetadataEXT = gfx125_SetHdrMetadataEXT,
    .GetSwapchainStatusKHR = gfx125_GetSwapchainStatusKHR,
    .GetRefreshCycleDurationGOOGLE = gfx125_GetRefreshCycleDurationGOOGLE,
    .GetPastPresentationTimingGOOGLE = gfx125_GetPastPresentationTimingGOOGLE,
    .CmdSetViewportWScalingNV = gfx125_CmdSetViewportWScalingNV,
    .CmdSetDiscardRectangleEXT = gfx125_CmdSetDiscardRectangleEXT,
    .CmdSetDiscardRectangleEnableEXT = gfx125_CmdSetDiscardRectangleEnableEXT,
    .CmdSetDiscardRectangleModeEXT = gfx125_CmdSetDiscardRectangleModeEXT,
    .CmdSetSampleLocationsEXT = gfx125_CmdSetSampleLocationsEXT,
    .GetBufferMemoryRequirements2 = gfx125_GetBufferMemoryRequirements2,
    .GetBufferMemoryRequirements2KHR = gfx125_GetBufferMemoryRequirements2KHR,
    .GetImageMemoryRequirements2 = gfx125_GetImageMemoryRequirements2,
    .GetImageMemoryRequirements2KHR = gfx125_GetImageMemoryRequirements2KHR,
    .GetImageSparseMemoryRequirements2 = gfx125_GetImageSparseMemoryRequirements2,
    .GetImageSparseMemoryRequirements2KHR = gfx125_GetImageSparseMemoryRequirements2KHR,
    .GetDeviceBufferMemoryRequirements = gfx125_GetDeviceBufferMemoryRequirements,
    .GetDeviceBufferMemoryRequirementsKHR = gfx125_GetDeviceBufferMemoryRequirementsKHR,
    .GetDeviceImageMemoryRequirements = gfx125_GetDeviceImageMemoryRequirements,
    .GetDeviceImageMemoryRequirementsKHR = gfx125_GetDeviceImageMemoryRequirementsKHR,
    .GetDeviceImageSparseMemoryRequirements = gfx125_GetDeviceImageSparseMemoryRequirements,
    .GetDeviceImageSparseMemoryRequirementsKHR = gfx125_GetDeviceImageSparseMemoryRequirementsKHR,
    .CreateSamplerYcbcrConversion = gfx125_CreateSamplerYcbcrConversion,
    .CreateSamplerYcbcrConversionKHR = gfx125_CreateSamplerYcbcrConversionKHR,
    .DestroySamplerYcbcrConversion = gfx125_DestroySamplerYcbcrConversion,
    .DestroySamplerYcbcrConversionKHR = gfx125_DestroySamplerYcbcrConversionKHR,
    .GetDeviceQueue2 = gfx125_GetDeviceQueue2,
    .CreateValidationCacheEXT = gfx125_CreateValidationCacheEXT,
    .DestroyValidationCacheEXT = gfx125_DestroyValidationCacheEXT,
    .GetValidationCacheDataEXT = gfx125_GetValidationCacheDataEXT,
    .MergeValidationCachesEXT = gfx125_MergeValidationCachesEXT,
    .GetDescriptorSetLayoutSupport = gfx125_GetDescriptorSetLayoutSupport,
    .GetDescriptorSetLayoutSupportKHR = gfx125_GetDescriptorSetLayoutSupportKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsageANDROID = gfx125_GetSwapchainGrallocUsageANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsage2ANDROID = gfx125_GetSwapchainGrallocUsage2ANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsage2ANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .AcquireImageANDROID = gfx125_AcquireImageANDROID,
#elif defined(_MSC_VER)
    .AcquireImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .QueueSignalReleaseImageANDROID = gfx125_QueueSignalReleaseImageANDROID,
#elif defined(_MSC_VER)
    .QueueSignalReleaseImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .GetShaderInfoAMD = gfx125_GetShaderInfoAMD,
    .SetLocalDimmingAMD = gfx125_SetLocalDimmingAMD,
    .GetCalibratedTimestampsKHR = gfx125_GetCalibratedTimestampsKHR,
    .GetCalibratedTimestampsEXT = gfx125_GetCalibratedTimestampsEXT,
    .SetDebugUtilsObjectNameEXT = gfx125_SetDebugUtilsObjectNameEXT,
    .SetDebugUtilsObjectTagEXT = gfx125_SetDebugUtilsObjectTagEXT,
    .QueueBeginDebugUtilsLabelEXT = gfx125_QueueBeginDebugUtilsLabelEXT,
    .QueueEndDebugUtilsLabelEXT = gfx125_QueueEndDebugUtilsLabelEXT,
    .QueueInsertDebugUtilsLabelEXT = gfx125_QueueInsertDebugUtilsLabelEXT,
    .CmdBeginDebugUtilsLabelEXT = gfx125_CmdBeginDebugUtilsLabelEXT,
    .CmdEndDebugUtilsLabelEXT = gfx125_CmdEndDebugUtilsLabelEXT,
    .CmdInsertDebugUtilsLabelEXT = gfx125_CmdInsertDebugUtilsLabelEXT,
    .GetMemoryHostPointerPropertiesEXT = gfx125_GetMemoryHostPointerPropertiesEXT,
    .CmdWriteBufferMarkerAMD = gfx125_CmdWriteBufferMarkerAMD,
    .CreateRenderPass2 = gfx125_CreateRenderPass2,
    .CreateRenderPass2KHR = gfx125_CreateRenderPass2KHR,
    .CmdBeginRenderPass2 = gfx125_CmdBeginRenderPass2,
    .CmdBeginRenderPass2KHR = gfx125_CmdBeginRenderPass2KHR,
    .CmdNextSubpass2 = gfx125_CmdNextSubpass2,
    .CmdNextSubpass2KHR = gfx125_CmdNextSubpass2KHR,
    .CmdEndRenderPass2 = gfx125_CmdEndRenderPass2,
    .CmdEndRenderPass2KHR = gfx125_CmdEndRenderPass2KHR,
    .GetSemaphoreCounterValue = gfx125_GetSemaphoreCounterValue,
    .GetSemaphoreCounterValueKHR = gfx125_GetSemaphoreCounterValueKHR,
    .WaitSemaphores = gfx125_WaitSemaphores,
    .WaitSemaphoresKHR = gfx125_WaitSemaphoresKHR,
    .SignalSemaphore = gfx125_SignalSemaphore,
    .SignalSemaphoreKHR = gfx125_SignalSemaphoreKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetAndroidHardwareBufferPropertiesANDROID = gfx125_GetAndroidHardwareBufferPropertiesANDROID,
#elif defined(_MSC_VER)
    .GetAndroidHardwareBufferPropertiesANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetMemoryAndroidHardwareBufferANDROID = gfx125_GetMemoryAndroidHardwareBufferANDROID,
#elif defined(_MSC_VER)
    .GetMemoryAndroidHardwareBufferANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .CmdDrawIndirectCount = gfx125_CmdDrawIndirectCount,
    .CmdDrawIndirectCountKHR = gfx125_CmdDrawIndirectCountKHR,
    .CmdDrawIndirectCountAMD = gfx125_CmdDrawIndirectCountAMD,
    .CmdDrawIndexedIndirectCount = gfx125_CmdDrawIndexedIndirectCount,
    .CmdDrawIndexedIndirectCountKHR = gfx125_CmdDrawIndexedIndirectCountKHR,
    .CmdDrawIndexedIndirectCountAMD = gfx125_CmdDrawIndexedIndirectCountAMD,
    .CmdSetCheckpointNV = gfx125_CmdSetCheckpointNV,
    .GetQueueCheckpointDataNV = gfx125_GetQueueCheckpointDataNV,
    .CmdBindTransformFeedbackBuffersEXT = gfx125_CmdBindTransformFeedbackBuffersEXT,
    .CmdBeginTransformFeedbackEXT = gfx125_CmdBeginTransformFeedbackEXT,
    .CmdEndTransformFeedbackEXT = gfx125_CmdEndTransformFeedbackEXT,
    .CmdBeginQueryIndexedEXT = gfx125_CmdBeginQueryIndexedEXT,
    .CmdEndQueryIndexedEXT = gfx125_CmdEndQueryIndexedEXT,
    .CmdDrawIndirectByteCountEXT = gfx125_CmdDrawIndirectByteCountEXT,
    .CmdSetExclusiveScissorNV = gfx125_CmdSetExclusiveScissorNV,
    .CmdSetExclusiveScissorEnableNV = gfx125_CmdSetExclusiveScissorEnableNV,
    .CmdBindShadingRateImageNV = gfx125_CmdBindShadingRateImageNV,
    .CmdSetViewportShadingRatePaletteNV = gfx125_CmdSetViewportShadingRatePaletteNV,
    .CmdSetCoarseSampleOrderNV = gfx125_CmdSetCoarseSampleOrderNV,
    .CmdDrawMeshTasksNV = gfx125_CmdDrawMeshTasksNV,
    .CmdDrawMeshTasksIndirectNV = gfx125_CmdDrawMeshTasksIndirectNV,
    .CmdDrawMeshTasksIndirectCountNV = gfx125_CmdDrawMeshTasksIndirectCountNV,
    .CmdDrawMeshTasksEXT = gfx125_CmdDrawMeshTasksEXT,
    .CmdDrawMeshTasksIndirectEXT = gfx125_CmdDrawMeshTasksIndirectEXT,
    .CmdDrawMeshTasksIndirectCountEXT = gfx125_CmdDrawMeshTasksIndirectCountEXT,
    .CompileDeferredNV = gfx125_CompileDeferredNV,
    .CreateAccelerationStructureNV = gfx125_CreateAccelerationStructureNV,
    .CmdBindInvocationMaskHUAWEI = gfx125_CmdBindInvocationMaskHUAWEI,
    .DestroyAccelerationStructureKHR = gfx125_DestroyAccelerationStructureKHR,
    .DestroyAccelerationStructureNV = gfx125_DestroyAccelerationStructureNV,
    .GetAccelerationStructureMemoryRequirementsNV = gfx125_GetAccelerationStructureMemoryRequirementsNV,
    .BindAccelerationStructureMemoryNV = gfx125_BindAccelerationStructureMemoryNV,
    .CmdCopyAccelerationStructureNV = gfx125_CmdCopyAccelerationStructureNV,
    .CmdCopyAccelerationStructureKHR = gfx125_CmdCopyAccelerationStructureKHR,
    .CopyAccelerationStructureKHR = gfx125_CopyAccelerationStructureKHR,
    .CmdCopyAccelerationStructureToMemoryKHR = gfx125_CmdCopyAccelerationStructureToMemoryKHR,
    .CopyAccelerationStructureToMemoryKHR = gfx125_CopyAccelerationStructureToMemoryKHR,
    .CmdCopyMemoryToAccelerationStructureKHR = gfx125_CmdCopyMemoryToAccelerationStructureKHR,
    .CopyMemoryToAccelerationStructureKHR = gfx125_CopyMemoryToAccelerationStructureKHR,
    .CmdWriteAccelerationStructuresPropertiesKHR = gfx125_CmdWriteAccelerationStructuresPropertiesKHR,
    .CmdWriteAccelerationStructuresPropertiesNV = gfx125_CmdWriteAccelerationStructuresPropertiesNV,
    .CmdBuildAccelerationStructureNV = gfx125_CmdBuildAccelerationStructureNV,
    .WriteAccelerationStructuresPropertiesKHR = gfx125_WriteAccelerationStructuresPropertiesKHR,
    .CmdTraceRaysKHR = gfx125_CmdTraceRaysKHR,
    .CmdTraceRaysNV = gfx125_CmdTraceRaysNV,
    .GetRayTracingShaderGroupHandlesKHR = gfx125_GetRayTracingShaderGroupHandlesKHR,
    .GetRayTracingShaderGroupHandlesNV = gfx125_GetRayTracingShaderGroupHandlesNV,
    .GetRayTracingCaptureReplayShaderGroupHandlesKHR = gfx125_GetRayTracingCaptureReplayShaderGroupHandlesKHR,
    .GetAccelerationStructureHandleNV = gfx125_GetAccelerationStructureHandleNV,
    .CreateRayTracingPipelinesNV = gfx125_CreateRayTracingPipelinesNV,
    .CreateRayTracingPipelinesKHR = gfx125_CreateRayTracingPipelinesKHR,
    .CmdTraceRaysIndirectKHR = gfx125_CmdTraceRaysIndirectKHR,
    .CmdTraceRaysIndirect2KHR = gfx125_CmdTraceRaysIndirect2KHR,
    .GetClusterAccelerationStructureBuildSizesNV = gfx125_GetClusterAccelerationStructureBuildSizesNV,
    .CmdBuildClusterAccelerationStructureIndirectNV = gfx125_CmdBuildClusterAccelerationStructureIndirectNV,
    .GetDeviceAccelerationStructureCompatibilityKHR = gfx125_GetDeviceAccelerationStructureCompatibilityKHR,
    .GetRayTracingShaderGroupStackSizeKHR = gfx125_GetRayTracingShaderGroupStackSizeKHR,
    .CmdSetRayTracingPipelineStackSizeKHR = gfx125_CmdSetRayTracingPipelineStackSizeKHR,
    .GetImageViewHandleNVX = gfx125_GetImageViewHandleNVX,
    .GetImageViewHandle64NVX = gfx125_GetImageViewHandle64NVX,
    .GetImageViewAddressNVX = gfx125_GetImageViewAddressNVX,
    .GetDeviceCombinedImageSamplerIndexNVX = gfx125_GetDeviceCombinedImageSamplerIndexNVX,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetDeviceGroupSurfacePresentModes2EXT = gfx125_GetDeviceGroupSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetDeviceGroupSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireFullScreenExclusiveModeEXT = gfx125_AcquireFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .AcquireFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ReleaseFullScreenExclusiveModeEXT = gfx125_ReleaseFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .ReleaseFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .AcquireProfilingLockKHR = gfx125_AcquireProfilingLockKHR,
    .ReleaseProfilingLockKHR = gfx125_ReleaseProfilingLockKHR,
    .GetImageDrmFormatModifierPropertiesEXT = gfx125_GetImageDrmFormatModifierPropertiesEXT,
    .GetBufferOpaqueCaptureAddress = gfx125_GetBufferOpaqueCaptureAddress,
    .GetBufferOpaqueCaptureAddressKHR = gfx125_GetBufferOpaqueCaptureAddressKHR,
    .GetBufferDeviceAddress = gfx125_GetBufferDeviceAddress,
    .GetBufferDeviceAddressKHR = gfx125_GetBufferDeviceAddressKHR,
    .GetBufferDeviceAddressEXT = gfx125_GetBufferDeviceAddressEXT,
    .InitializePerformanceApiINTEL = gfx125_InitializePerformanceApiINTEL,
    .UninitializePerformanceApiINTEL = gfx125_UninitializePerformanceApiINTEL,
    .CmdSetPerformanceMarkerINTEL = gfx125_CmdSetPerformanceMarkerINTEL,
    .CmdSetPerformanceStreamMarkerINTEL = gfx125_CmdSetPerformanceStreamMarkerINTEL,
    .CmdSetPerformanceOverrideINTEL = gfx125_CmdSetPerformanceOverrideINTEL,
    .AcquirePerformanceConfigurationINTEL = gfx125_AcquirePerformanceConfigurationINTEL,
    .ReleasePerformanceConfigurationINTEL = gfx125_ReleasePerformanceConfigurationINTEL,
    .QueueSetPerformanceConfigurationINTEL = gfx125_QueueSetPerformanceConfigurationINTEL,
    .GetPerformanceParameterINTEL = gfx125_GetPerformanceParameterINTEL,
    .GetDeviceMemoryOpaqueCaptureAddress = gfx125_GetDeviceMemoryOpaqueCaptureAddress,
    .GetDeviceMemoryOpaqueCaptureAddressKHR = gfx125_GetDeviceMemoryOpaqueCaptureAddressKHR,
    .GetPipelineExecutablePropertiesKHR = gfx125_GetPipelineExecutablePropertiesKHR,
    .GetPipelineExecutableStatisticsKHR = gfx125_GetPipelineExecutableStatisticsKHR,
    .GetPipelineExecutableInternalRepresentationsKHR = gfx125_GetPipelineExecutableInternalRepresentationsKHR,
    .CmdSetLineStipple = gfx125_CmdSetLineStipple,
    .CmdSetLineStippleKHR = gfx125_CmdSetLineStippleKHR,
    .CmdSetLineStippleEXT = gfx125_CmdSetLineStippleEXT,
    .CreateAccelerationStructureKHR = gfx125_CreateAccelerationStructureKHR,
    .CmdBuildAccelerationStructuresKHR = gfx125_CmdBuildAccelerationStructuresKHR,
    .CmdBuildAccelerationStructuresIndirectKHR = gfx125_CmdBuildAccelerationStructuresIndirectKHR,
    .BuildAccelerationStructuresKHR = gfx125_BuildAccelerationStructuresKHR,
    .GetAccelerationStructureDeviceAddressKHR = gfx125_GetAccelerationStructureDeviceAddressKHR,
    .CreateDeferredOperationKHR = gfx125_CreateDeferredOperationKHR,
    .DestroyDeferredOperationKHR = gfx125_DestroyDeferredOperationKHR,
    .GetDeferredOperationMaxConcurrencyKHR = gfx125_GetDeferredOperationMaxConcurrencyKHR,
    .GetDeferredOperationResultKHR = gfx125_GetDeferredOperationResultKHR,
    .DeferredOperationJoinKHR = gfx125_DeferredOperationJoinKHR,
    .GetPipelineIndirectMemoryRequirementsNV = gfx125_GetPipelineIndirectMemoryRequirementsNV,
    .GetPipelineIndirectDeviceAddressNV = gfx125_GetPipelineIndirectDeviceAddressNV,
    .AntiLagUpdateAMD = gfx125_AntiLagUpdateAMD,
    .CmdSetCullMode = gfx125_CmdSetCullMode,
    .CmdSetCullModeEXT = gfx125_CmdSetCullModeEXT,
    .CmdSetFrontFace = gfx125_CmdSetFrontFace,
    .CmdSetFrontFaceEXT = gfx125_CmdSetFrontFaceEXT,
    .CmdSetPrimitiveTopology = gfx125_CmdSetPrimitiveTopology,
    .CmdSetPrimitiveTopologyEXT = gfx125_CmdSetPrimitiveTopologyEXT,
    .CmdSetViewportWithCount = gfx125_CmdSetViewportWithCount,
    .CmdSetViewportWithCountEXT = gfx125_CmdSetViewportWithCountEXT,
    .CmdSetScissorWithCount = gfx125_CmdSetScissorWithCount,
    .CmdSetScissorWithCountEXT = gfx125_CmdSetScissorWithCountEXT,
    .CmdBindIndexBuffer2 = gfx125_CmdBindIndexBuffer2,
    .CmdBindIndexBuffer2KHR = gfx125_CmdBindIndexBuffer2KHR,
    .CmdBindVertexBuffers2 = gfx125_CmdBindVertexBuffers2,
    .CmdBindVertexBuffers2EXT = gfx125_CmdBindVertexBuffers2EXT,
    .CmdSetDepthTestEnable = gfx125_CmdSetDepthTestEnable,
    .CmdSetDepthTestEnableEXT = gfx125_CmdSetDepthTestEnableEXT,
    .CmdSetDepthWriteEnable = gfx125_CmdSetDepthWriteEnable,
    .CmdSetDepthWriteEnableEXT = gfx125_CmdSetDepthWriteEnableEXT,
    .CmdSetDepthCompareOp = gfx125_CmdSetDepthCompareOp,
    .CmdSetDepthCompareOpEXT = gfx125_CmdSetDepthCompareOpEXT,
    .CmdSetDepthBoundsTestEnable = gfx125_CmdSetDepthBoundsTestEnable,
    .CmdSetDepthBoundsTestEnableEXT = gfx125_CmdSetDepthBoundsTestEnableEXT,
    .CmdSetStencilTestEnable = gfx125_CmdSetStencilTestEnable,
    .CmdSetStencilTestEnableEXT = gfx125_CmdSetStencilTestEnableEXT,
    .CmdSetStencilOp = gfx125_CmdSetStencilOp,
    .CmdSetStencilOpEXT = gfx125_CmdSetStencilOpEXT,
    .CmdSetPatchControlPointsEXT = gfx125_CmdSetPatchControlPointsEXT,
    .CmdSetRasterizerDiscardEnable = gfx125_CmdSetRasterizerDiscardEnable,
    .CmdSetRasterizerDiscardEnableEXT = gfx125_CmdSetRasterizerDiscardEnableEXT,
    .CmdSetDepthBiasEnable = gfx125_CmdSetDepthBiasEnable,
    .CmdSetDepthBiasEnableEXT = gfx125_CmdSetDepthBiasEnableEXT,
    .CmdSetLogicOpEXT = gfx125_CmdSetLogicOpEXT,
    .CmdSetPrimitiveRestartEnable = gfx125_CmdSetPrimitiveRestartEnable,
    .CmdSetPrimitiveRestartEnableEXT = gfx125_CmdSetPrimitiveRestartEnableEXT,
    .CmdSetTessellationDomainOriginEXT = gfx125_CmdSetTessellationDomainOriginEXT,
    .CmdSetDepthClampEnableEXT = gfx125_CmdSetDepthClampEnableEXT,
    .CmdSetPolygonModeEXT = gfx125_CmdSetPolygonModeEXT,
    .CmdSetRasterizationSamplesEXT = gfx125_CmdSetRasterizationSamplesEXT,
    .CmdSetSampleMaskEXT = gfx125_CmdSetSampleMaskEXT,
    .CmdSetAlphaToCoverageEnableEXT = gfx125_CmdSetAlphaToCoverageEnableEXT,
    .CmdSetAlphaToOneEnableEXT = gfx125_CmdSetAlphaToOneEnableEXT,
    .CmdSetLogicOpEnableEXT = gfx125_CmdSetLogicOpEnableEXT,
    .CmdSetColorBlendEnableEXT = gfx125_CmdSetColorBlendEnableEXT,
    .CmdSetColorBlendEquationEXT = gfx125_CmdSetColorBlendEquationEXT,
    .CmdSetColorWriteMaskEXT = gfx125_CmdSetColorWriteMaskEXT,
    .CmdSetRasterizationStreamEXT = gfx125_CmdSetRasterizationStreamEXT,
    .CmdSetConservativeRasterizationModeEXT = gfx125_CmdSetConservativeRasterizationModeEXT,
    .CmdSetExtraPrimitiveOverestimationSizeEXT = gfx125_CmdSetExtraPrimitiveOverestimationSizeEXT,
    .CmdSetDepthClipEnableEXT = gfx125_CmdSetDepthClipEnableEXT,
    .CmdSetSampleLocationsEnableEXT = gfx125_CmdSetSampleLocationsEnableEXT,
    .CmdSetColorBlendAdvancedEXT = gfx125_CmdSetColorBlendAdvancedEXT,
    .CmdSetProvokingVertexModeEXT = gfx125_CmdSetProvokingVertexModeEXT,
    .CmdSetLineRasterizationModeEXT = gfx125_CmdSetLineRasterizationModeEXT,
    .CmdSetLineStippleEnableEXT = gfx125_CmdSetLineStippleEnableEXT,
    .CmdSetDepthClipNegativeOneToOneEXT = gfx125_CmdSetDepthClipNegativeOneToOneEXT,
    .CmdSetViewportWScalingEnableNV = gfx125_CmdSetViewportWScalingEnableNV,
    .CmdSetViewportSwizzleNV = gfx125_CmdSetViewportSwizzleNV,
    .CmdSetCoverageToColorEnableNV = gfx125_CmdSetCoverageToColorEnableNV,
    .CmdSetCoverageToColorLocationNV = gfx125_CmdSetCoverageToColorLocationNV,
    .CmdSetCoverageModulationModeNV = gfx125_CmdSetCoverageModulationModeNV,
    .CmdSetCoverageModulationTableEnableNV = gfx125_CmdSetCoverageModulationTableEnableNV,
    .CmdSetCoverageModulationTableNV = gfx125_CmdSetCoverageModulationTableNV,
    .CmdSetShadingRateImageEnableNV = gfx125_CmdSetShadingRateImageEnableNV,
    .CmdSetCoverageReductionModeNV = gfx125_CmdSetCoverageReductionModeNV,
    .CmdSetRepresentativeFragmentTestEnableNV = gfx125_CmdSetRepresentativeFragmentTestEnableNV,
    .CreatePrivateDataSlot = gfx125_CreatePrivateDataSlot,
    .CreatePrivateDataSlotEXT = gfx125_CreatePrivateDataSlotEXT,
    .DestroyPrivateDataSlot = gfx125_DestroyPrivateDataSlot,
    .DestroyPrivateDataSlotEXT = gfx125_DestroyPrivateDataSlotEXT,
    .SetPrivateData = gfx125_SetPrivateData,
    .SetPrivateDataEXT = gfx125_SetPrivateDataEXT,
    .GetPrivateData = gfx125_GetPrivateData,
    .GetPrivateDataEXT = gfx125_GetPrivateDataEXT,
    .CmdCopyBuffer2 = gfx125_CmdCopyBuffer2,
    .CmdCopyBuffer2KHR = gfx125_CmdCopyBuffer2KHR,
    .CmdCopyImage2 = gfx125_CmdCopyImage2,
    .CmdCopyImage2KHR = gfx125_CmdCopyImage2KHR,
    .CmdBlitImage2 = gfx125_CmdBlitImage2,
    .CmdBlitImage2KHR = gfx125_CmdBlitImage2KHR,
    .CmdCopyBufferToImage2 = gfx125_CmdCopyBufferToImage2,
    .CmdCopyBufferToImage2KHR = gfx125_CmdCopyBufferToImage2KHR,
    .CmdCopyImageToBuffer2 = gfx125_CmdCopyImageToBuffer2,
    .CmdCopyImageToBuffer2KHR = gfx125_CmdCopyImageToBuffer2KHR,
    .CmdResolveImage2 = gfx125_CmdResolveImage2,
    .CmdResolveImage2KHR = gfx125_CmdResolveImage2KHR,
    .CmdSetFragmentShadingRateKHR = gfx125_CmdSetFragmentShadingRateKHR,
    .CmdSetFragmentShadingRateEnumNV = gfx125_CmdSetFragmentShadingRateEnumNV,
    .GetAccelerationStructureBuildSizesKHR = gfx125_GetAccelerationStructureBuildSizesKHR,
    .CmdSetVertexInputEXT = gfx125_CmdSetVertexInputEXT,
    .CmdSetColorWriteEnableEXT = gfx125_CmdSetColorWriteEnableEXT,
    .CmdSetEvent2 = gfx125_CmdSetEvent2,
    .CmdSetEvent2KHR = gfx125_CmdSetEvent2KHR,
    .CmdResetEvent2 = gfx125_CmdResetEvent2,
    .CmdResetEvent2KHR = gfx125_CmdResetEvent2KHR,
    .CmdWaitEvents2 = gfx125_CmdWaitEvents2,
    .CmdWaitEvents2KHR = gfx125_CmdWaitEvents2KHR,
    .CmdPipelineBarrier2 = gfx125_CmdPipelineBarrier2,
    .CmdPipelineBarrier2KHR = gfx125_CmdPipelineBarrier2KHR,
    .QueueSubmit2 = gfx125_QueueSubmit2,
    .QueueSubmit2KHR = gfx125_QueueSubmit2KHR,
    .CmdWriteTimestamp2 = gfx125_CmdWriteTimestamp2,
    .CmdWriteTimestamp2KHR = gfx125_CmdWriteTimestamp2KHR,
    .CmdWriteBufferMarker2AMD = gfx125_CmdWriteBufferMarker2AMD,
    .GetQueueCheckpointData2NV = gfx125_GetQueueCheckpointData2NV,
    .CopyMemoryToImage = gfx125_CopyMemoryToImage,
    .CopyMemoryToImageEXT = gfx125_CopyMemoryToImageEXT,
    .CopyImageToMemory = gfx125_CopyImageToMemory,
    .CopyImageToMemoryEXT = gfx125_CopyImageToMemoryEXT,
    .CopyImageToImage = gfx125_CopyImageToImage,
    .CopyImageToImageEXT = gfx125_CopyImageToImageEXT,
    .TransitionImageLayout = gfx125_TransitionImageLayout,
    .TransitionImageLayoutEXT = gfx125_TransitionImageLayoutEXT,
    .CreateVideoSessionKHR = gfx125_CreateVideoSessionKHR,
    .DestroyVideoSessionKHR = gfx125_DestroyVideoSessionKHR,
    .CreateVideoSessionParametersKHR = gfx125_CreateVideoSessionParametersKHR,
    .UpdateVideoSessionParametersKHR = gfx125_UpdateVideoSessionParametersKHR,
    .GetEncodedVideoSessionParametersKHR = gfx125_GetEncodedVideoSessionParametersKHR,
    .DestroyVideoSessionParametersKHR = gfx125_DestroyVideoSessionParametersKHR,
    .GetVideoSessionMemoryRequirementsKHR = gfx125_GetVideoSessionMemoryRequirementsKHR,
    .BindVideoSessionMemoryKHR = gfx125_BindVideoSessionMemoryKHR,
    .CmdDecodeVideoKHR = gfx125_CmdDecodeVideoKHR,
    .CmdBeginVideoCodingKHR = gfx125_CmdBeginVideoCodingKHR,
    .CmdControlVideoCodingKHR = gfx125_CmdControlVideoCodingKHR,
    .CmdEndVideoCodingKHR = gfx125_CmdEndVideoCodingKHR,
    .CmdEncodeVideoKHR = gfx125_CmdEncodeVideoKHR,
    .CmdDecompressMemoryNV = gfx125_CmdDecompressMemoryNV,
    .CmdDecompressMemoryIndirectCountNV = gfx125_CmdDecompressMemoryIndirectCountNV,
    .GetPartitionedAccelerationStructuresBuildSizesNV = gfx125_GetPartitionedAccelerationStructuresBuildSizesNV,
    .CmdBuildPartitionedAccelerationStructuresNV = gfx125_CmdBuildPartitionedAccelerationStructuresNV,
    .CmdDecompressMemoryEXT = gfx125_CmdDecompressMemoryEXT,
    .CmdDecompressMemoryIndirectCountEXT = gfx125_CmdDecompressMemoryIndirectCountEXT,
    .CreateCuModuleNVX = gfx125_CreateCuModuleNVX,
    .CreateCuFunctionNVX = gfx125_CreateCuFunctionNVX,
    .DestroyCuModuleNVX = gfx125_DestroyCuModuleNVX,
    .DestroyCuFunctionNVX = gfx125_DestroyCuFunctionNVX,
    .CmdCuLaunchKernelNVX = gfx125_CmdCuLaunchKernelNVX,
    .GetDescriptorSetLayoutSizeEXT = gfx125_GetDescriptorSetLayoutSizeEXT,
    .GetDescriptorSetLayoutBindingOffsetEXT = gfx125_GetDescriptorSetLayoutBindingOffsetEXT,
    .GetDescriptorEXT = gfx125_GetDescriptorEXT,
    .CmdBindDescriptorBuffersEXT = gfx125_CmdBindDescriptorBuffersEXT,
    .CmdSetDescriptorBufferOffsetsEXT = gfx125_CmdSetDescriptorBufferOffsetsEXT,
    .CmdBindDescriptorBufferEmbeddedSamplersEXT = gfx125_CmdBindDescriptorBufferEmbeddedSamplersEXT,
    .GetBufferOpaqueCaptureDescriptorDataEXT = gfx125_GetBufferOpaqueCaptureDescriptorDataEXT,
    .GetImageOpaqueCaptureDescriptorDataEXT = gfx125_GetImageOpaqueCaptureDescriptorDataEXT,
    .GetImageViewOpaqueCaptureDescriptorDataEXT = gfx125_GetImageViewOpaqueCaptureDescriptorDataEXT,
    .GetSamplerOpaqueCaptureDescriptorDataEXT = gfx125_GetSamplerOpaqueCaptureDescriptorDataEXT,
    .GetAccelerationStructureOpaqueCaptureDescriptorDataEXT = gfx125_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT,
    .SetDeviceMemoryPriorityEXT = gfx125_SetDeviceMemoryPriorityEXT,
    .WaitForPresent2KHR = gfx125_WaitForPresent2KHR,
    .WaitForPresentKHR = gfx125_WaitForPresentKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateBufferCollectionFUCHSIA = gfx125_CreateBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .CreateBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionBufferConstraintsFUCHSIA = gfx125_SetBufferCollectionBufferConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionBufferConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionImageConstraintsFUCHSIA = gfx125_SetBufferCollectionImageConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionImageConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .DestroyBufferCollectionFUCHSIA = gfx125_DestroyBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .DestroyBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetBufferCollectionPropertiesFUCHSIA = gfx125_GetBufferCollectionPropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetBufferCollectionPropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .CmdBeginRendering = gfx125_CmdBeginRendering,
    .CmdBeginRenderingKHR = gfx125_CmdBeginRenderingKHR,
    .CmdEndRendering = gfx125_CmdEndRendering,
    .CmdEndRendering2KHR = gfx125_CmdEndRendering2KHR,
    .CmdEndRendering2EXT = gfx125_CmdEndRendering2EXT,
    .CmdEndRenderingKHR = gfx125_CmdEndRenderingKHR,
    .GetDescriptorSetLayoutHostMappingInfoVALVE = gfx125_GetDescriptorSetLayoutHostMappingInfoVALVE,
    .GetDescriptorSetHostMappingVALVE = gfx125_GetDescriptorSetHostMappingVALVE,
    .CreateMicromapEXT = gfx125_CreateMicromapEXT,
    .CmdBuildMicromapsEXT = gfx125_CmdBuildMicromapsEXT,
    .BuildMicromapsEXT = gfx125_BuildMicromapsEXT,
    .DestroyMicromapEXT = gfx125_DestroyMicromapEXT,
    .CmdCopyMicromapEXT = gfx125_CmdCopyMicromapEXT,
    .CopyMicromapEXT = gfx125_CopyMicromapEXT,
    .CmdCopyMicromapToMemoryEXT = gfx125_CmdCopyMicromapToMemoryEXT,
    .CopyMicromapToMemoryEXT = gfx125_CopyMicromapToMemoryEXT,
    .CmdCopyMemoryToMicromapEXT = gfx125_CmdCopyMemoryToMicromapEXT,
    .CopyMemoryToMicromapEXT = gfx125_CopyMemoryToMicromapEXT,
    .CmdWriteMicromapsPropertiesEXT = gfx125_CmdWriteMicromapsPropertiesEXT,
    .WriteMicromapsPropertiesEXT = gfx125_WriteMicromapsPropertiesEXT,
    .GetDeviceMicromapCompatibilityEXT = gfx125_GetDeviceMicromapCompatibilityEXT,
    .GetMicromapBuildSizesEXT = gfx125_GetMicromapBuildSizesEXT,
    .GetShaderModuleIdentifierEXT = gfx125_GetShaderModuleIdentifierEXT,
    .GetShaderModuleCreateInfoIdentifierEXT = gfx125_GetShaderModuleCreateInfoIdentifierEXT,
    .GetImageSubresourceLayout2 = gfx125_GetImageSubresourceLayout2,
    .GetImageSubresourceLayout2KHR = gfx125_GetImageSubresourceLayout2KHR,
    .GetImageSubresourceLayout2EXT = gfx125_GetImageSubresourceLayout2EXT,
    .GetPipelinePropertiesEXT = gfx125_GetPipelinePropertiesEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .ExportMetalObjectsEXT = gfx125_ExportMetalObjectsEXT,
#elif defined(_MSC_VER)
    .ExportMetalObjectsEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CmdBindTileMemoryQCOM = gfx125_CmdBindTileMemoryQCOM,
    .GetFramebufferTilePropertiesQCOM = gfx125_GetFramebufferTilePropertiesQCOM,
    .GetDynamicRenderingTilePropertiesQCOM = gfx125_GetDynamicRenderingTilePropertiesQCOM,
    .CreateOpticalFlowSessionNV = gfx125_CreateOpticalFlowSessionNV,
    .DestroyOpticalFlowSessionNV = gfx125_DestroyOpticalFlowSessionNV,
    .BindOpticalFlowSessionImageNV = gfx125_BindOpticalFlowSessionImageNV,
    .CmdOpticalFlowExecuteNV = gfx125_CmdOpticalFlowExecuteNV,
    .GetDeviceFaultInfoEXT = gfx125_GetDeviceFaultInfoEXT,
    .GetDeviceFaultReportsKHR = gfx125_GetDeviceFaultReportsKHR,
    .GetDeviceFaultDebugInfoKHR = gfx125_GetDeviceFaultDebugInfoKHR,
    .CmdSetDepthBias2EXT = gfx125_CmdSetDepthBias2EXT,
    .ReleaseSwapchainImagesKHR = gfx125_ReleaseSwapchainImagesKHR,
    .ReleaseSwapchainImagesEXT = gfx125_ReleaseSwapchainImagesEXT,
    .GetDeviceImageSubresourceLayout = gfx125_GetDeviceImageSubresourceLayout,
    .GetDeviceImageSubresourceLayoutKHR = gfx125_GetDeviceImageSubresourceLayoutKHR,
    .MapMemory2 = gfx125_MapMemory2,
    .MapMemory2KHR = gfx125_MapMemory2KHR,
    .UnmapMemory2 = gfx125_UnmapMemory2,
    .UnmapMemory2KHR = gfx125_UnmapMemory2KHR,
    .CreateShadersEXT = gfx125_CreateShadersEXT,
    .DestroyShaderEXT = gfx125_DestroyShaderEXT,
    .GetShaderBinaryDataEXT = gfx125_GetShaderBinaryDataEXT,
    .CmdBindShadersEXT = gfx125_CmdBindShadersEXT,
    .SetSwapchainPresentTimingQueueSizeEXT = gfx125_SetSwapchainPresentTimingQueueSizeEXT,
    .GetSwapchainTimingPropertiesEXT = gfx125_GetSwapchainTimingPropertiesEXT,
    .GetSwapchainTimeDomainPropertiesEXT = gfx125_GetSwapchainTimeDomainPropertiesEXT,
    .GetPastPresentationTimingEXT = gfx125_GetPastPresentationTimingEXT,
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetScreenBufferPropertiesQNX = gfx125_GetScreenBufferPropertiesQNX,
#elif defined(_MSC_VER)
    .GetScreenBufferPropertiesQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CmdBindDescriptorSets2 = gfx125_CmdBindDescriptorSets2,
    .CmdBindDescriptorSets2KHR = gfx125_CmdBindDescriptorSets2KHR,
    .CmdPushConstants2 = gfx125_CmdPushConstants2,
    .CmdPushConstants2KHR = gfx125_CmdPushConstants2KHR,
    .CmdPushDescriptorSet2 = gfx125_CmdPushDescriptorSet2,
    .CmdPushDescriptorSet2KHR = gfx125_CmdPushDescriptorSet2KHR,
    .CmdPushDescriptorSetWithTemplate2 = gfx125_CmdPushDescriptorSetWithTemplate2,
    .CmdPushDescriptorSetWithTemplate2KHR = gfx125_CmdPushDescriptorSetWithTemplate2KHR,
    .CmdSetDescriptorBufferOffsets2EXT = gfx125_CmdSetDescriptorBufferOffsets2EXT,
    .CmdBindDescriptorBufferEmbeddedSamplers2EXT = gfx125_CmdBindDescriptorBufferEmbeddedSamplers2EXT,
    .SetLatencySleepModeNV = gfx125_SetLatencySleepModeNV,
    .LatencySleepNV = gfx125_LatencySleepNV,
    .SetLatencyMarkerNV = gfx125_SetLatencyMarkerNV,
    .GetLatencyTimingsNV = gfx125_GetLatencyTimingsNV,
    .QueueNotifyOutOfBandNV = gfx125_QueueNotifyOutOfBandNV,
    .CmdSetRenderingAttachmentLocations = gfx125_CmdSetRenderingAttachmentLocations,
    .CmdSetRenderingAttachmentLocationsKHR = gfx125_CmdSetRenderingAttachmentLocationsKHR,
    .CmdSetRenderingInputAttachmentIndices = gfx125_CmdSetRenderingInputAttachmentIndices,
    .CmdSetRenderingInputAttachmentIndicesKHR = gfx125_CmdSetRenderingInputAttachmentIndicesKHR,
    .CmdSetDepthClampRangeEXT = gfx125_CmdSetDepthClampRangeEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandleEXT = gfx125_GetMemoryMetalHandleEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandleEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandlePropertiesEXT = gfx125_GetMemoryMetalHandlePropertiesEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandlePropertiesEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .ConvertCooperativeVectorMatrixNV = gfx125_ConvertCooperativeVectorMatrixNV,
    .CmdConvertCooperativeVectorMatrixNV = gfx125_CmdConvertCooperativeVectorMatrixNV,
    .CmdDispatchTileQCOM = gfx125_CmdDispatchTileQCOM,
    .CmdBeginPerTileExecutionQCOM = gfx125_CmdBeginPerTileExecutionQCOM,
    .CmdEndPerTileExecutionQCOM = gfx125_CmdEndPerTileExecutionQCOM,
    .CreateExternalComputeQueueNV = gfx125_CreateExternalComputeQueueNV,
    .DestroyExternalComputeQueueNV = gfx125_DestroyExternalComputeQueueNV,
    .CreateShaderInstrumentationARM = gfx125_CreateShaderInstrumentationARM,
    .DestroyShaderInstrumentationARM = gfx125_DestroyShaderInstrumentationARM,
    .CmdBeginShaderInstrumentationARM = gfx125_CmdBeginShaderInstrumentationARM,
    .CmdEndShaderInstrumentationARM = gfx125_CmdEndShaderInstrumentationARM,
    .GetShaderInstrumentationValuesARM = gfx125_GetShaderInstrumentationValuesARM,
    .ClearShaderInstrumentationMetricsARM = gfx125_ClearShaderInstrumentationMetricsARM,
    .CreateTensorARM = gfx125_CreateTensorARM,
    .DestroyTensorARM = gfx125_DestroyTensorARM,
    .CreateTensorViewARM = gfx125_CreateTensorViewARM,
    .DestroyTensorViewARM = gfx125_DestroyTensorViewARM,
    .GetTensorMemoryRequirementsARM = gfx125_GetTensorMemoryRequirementsARM,
    .BindTensorMemoryARM = gfx125_BindTensorMemoryARM,
    .GetDeviceTensorMemoryRequirementsARM = gfx125_GetDeviceTensorMemoryRequirementsARM,
    .CmdCopyTensorARM = gfx125_CmdCopyTensorARM,
    .GetTensorOpaqueCaptureDescriptorDataARM = gfx125_GetTensorOpaqueCaptureDescriptorDataARM,
    .GetTensorViewOpaqueCaptureDescriptorDataARM = gfx125_GetTensorViewOpaqueCaptureDescriptorDataARM,
    .CreateDataGraphPipelinesARM = gfx125_CreateDataGraphPipelinesARM,
    .CreateDataGraphPipelineSessionARM = gfx125_CreateDataGraphPipelineSessionARM,
    .GetDataGraphPipelineSessionBindPointRequirementsARM = gfx125_GetDataGraphPipelineSessionBindPointRequirementsARM,
    .GetDataGraphPipelineSessionMemoryRequirementsARM = gfx125_GetDataGraphPipelineSessionMemoryRequirementsARM,
    .BindDataGraphPipelineSessionMemoryARM = gfx125_BindDataGraphPipelineSessionMemoryARM,
    .DestroyDataGraphPipelineSessionARM = gfx125_DestroyDataGraphPipelineSessionARM,
    .CmdDispatchDataGraphARM = gfx125_CmdDispatchDataGraphARM,
    .GetDataGraphPipelineAvailablePropertiesARM = gfx125_GetDataGraphPipelineAvailablePropertiesARM,
    .GetDataGraphPipelinePropertiesARM = gfx125_GetDataGraphPipelinePropertiesARM,
#ifdef VK_USE_PLATFORM_OHOS
    .GetNativeBufferPropertiesOHOS = gfx125_GetNativeBufferPropertiesOHOS,
#elif defined(_MSC_VER)
    .GetNativeBufferPropertiesOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
    .GetMemoryNativeBufferOHOS = gfx125_GetMemoryNativeBufferOHOS,
#elif defined(_MSC_VER)
    .GetMemoryNativeBufferOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .QueueSetPerfHintQCOM = gfx125_QueueSetPerfHintQCOM,
    .CmdSetComputeOccupancyPriorityNV = gfx125_CmdSetComputeOccupancyPriorityNV,
    .WriteSamplerDescriptorsEXT = gfx125_WriteSamplerDescriptorsEXT,
    .WriteResourceDescriptorsEXT = gfx125_WriteResourceDescriptorsEXT,
    .CmdBindSamplerHeapEXT = gfx125_CmdBindSamplerHeapEXT,
    .CmdBindResourceHeapEXT = gfx125_CmdBindResourceHeapEXT,
    .CmdPushDataEXT = gfx125_CmdPushDataEXT,
    .RegisterCustomBorderColorEXT = gfx125_RegisterCustomBorderColorEXT,
    .UnregisterCustomBorderColorEXT = gfx125_UnregisterCustomBorderColorEXT,
    .GetImageOpaqueCaptureDataEXT = gfx125_GetImageOpaqueCaptureDataEXT,
    .GetTensorOpaqueCaptureDataARM = gfx125_GetTensorOpaqueCaptureDataARM,
    .CmdCopyMemoryKHR = gfx125_CmdCopyMemoryKHR,
    .CmdCopyMemoryToImageKHR = gfx125_CmdCopyMemoryToImageKHR,
    .CmdCopyImageToMemoryKHR = gfx125_CmdCopyImageToMemoryKHR,
    .CmdUpdateMemoryKHR = gfx125_CmdUpdateMemoryKHR,
    .CmdFillMemoryKHR = gfx125_CmdFillMemoryKHR,
    .CmdCopyQueryPoolResultsToMemoryKHR = gfx125_CmdCopyQueryPoolResultsToMemoryKHR,
    .CmdBeginConditionalRendering2EXT = gfx125_CmdBeginConditionalRendering2EXT,
    .CmdBindTransformFeedbackBuffers2EXT = gfx125_CmdBindTransformFeedbackBuffers2EXT,
    .CmdBeginTransformFeedback2EXT = gfx125_CmdBeginTransformFeedback2EXT,
    .CmdEndTransformFeedback2EXT = gfx125_CmdEndTransformFeedback2EXT,
    .CmdDrawIndirectByteCount2EXT = gfx125_CmdDrawIndirectByteCount2EXT,
    .CmdWriteMarkerToMemoryAMD = gfx125_CmdWriteMarkerToMemoryAMD,
    .CmdBindIndexBuffer3KHR = gfx125_CmdBindIndexBuffer3KHR,
    .CmdBindVertexBuffers3KHR = gfx125_CmdBindVertexBuffers3KHR,
    .CmdDrawIndirect2KHR = gfx125_CmdDrawIndirect2KHR,
    .CmdDrawIndexedIndirect2KHR = gfx125_CmdDrawIndexedIndirect2KHR,
    .CmdDrawIndirectCount2KHR = gfx125_CmdDrawIndirectCount2KHR,
    .CmdDrawIndexedIndirectCount2KHR = gfx125_CmdDrawIndexedIndirectCount2KHR,
    .CmdDrawMeshTasksIndirect2EXT = gfx125_CmdDrawMeshTasksIndirect2EXT,
    .CmdDrawMeshTasksIndirectCount2EXT = gfx125_CmdDrawMeshTasksIndirectCount2EXT,
    .CmdDispatchIndirect2KHR = gfx125_CmdDispatchIndirect2KHR,
    .CreateAccelerationStructure2KHR = gfx125_CreateAccelerationStructure2KHR,
    .CmdSetDispatchParametersARM = gfx125_CmdSetDispatchParametersARM,
};
const struct vk_device_entrypoint_table gfx20_device_entrypoints = {
    .GetDeviceProcAddr = gfx20_GetDeviceProcAddr,
    .DestroyDevice = gfx20_DestroyDevice,
    .GetDeviceQueue = gfx20_GetDeviceQueue,
    .QueueSubmit = gfx20_QueueSubmit,
    .QueueWaitIdle = gfx20_QueueWaitIdle,
    .DeviceWaitIdle = gfx20_DeviceWaitIdle,
    .AllocateMemory = gfx20_AllocateMemory,
    .FreeMemory = gfx20_FreeMemory,
    .MapMemory = gfx20_MapMemory,
    .UnmapMemory = gfx20_UnmapMemory,
    .FlushMappedMemoryRanges = gfx20_FlushMappedMemoryRanges,
    .InvalidateMappedMemoryRanges = gfx20_InvalidateMappedMemoryRanges,
    .GetDeviceMemoryCommitment = gfx20_GetDeviceMemoryCommitment,
    .GetBufferMemoryRequirements = gfx20_GetBufferMemoryRequirements,
    .BindBufferMemory = gfx20_BindBufferMemory,
    .GetImageMemoryRequirements = gfx20_GetImageMemoryRequirements,
    .BindImageMemory = gfx20_BindImageMemory,
    .GetImageSparseMemoryRequirements = gfx20_GetImageSparseMemoryRequirements,
    .QueueBindSparse = gfx20_QueueBindSparse,
    .CreateFence = gfx20_CreateFence,
    .DestroyFence = gfx20_DestroyFence,
    .ResetFences = gfx20_ResetFences,
    .GetFenceStatus = gfx20_GetFenceStatus,
    .WaitForFences = gfx20_WaitForFences,
    .CreateSemaphore = gfx20_CreateSemaphore,
    .DestroySemaphore = gfx20_DestroySemaphore,
    .CreateEvent = gfx20_CreateEvent,
    .DestroyEvent = gfx20_DestroyEvent,
    .GetEventStatus = gfx20_GetEventStatus,
    .SetEvent = gfx20_SetEvent,
    .ResetEvent = gfx20_ResetEvent,
    .CreateQueryPool = gfx20_CreateQueryPool,
    .DestroyQueryPool = gfx20_DestroyQueryPool,
    .GetQueryPoolResults = gfx20_GetQueryPoolResults,
    .ResetQueryPool = gfx20_ResetQueryPool,
    .ResetQueryPoolEXT = gfx20_ResetQueryPoolEXT,
    .CreateBuffer = gfx20_CreateBuffer,
    .DestroyBuffer = gfx20_DestroyBuffer,
    .CreateBufferView = gfx20_CreateBufferView,
    .DestroyBufferView = gfx20_DestroyBufferView,
    .CreateImage = gfx20_CreateImage,
    .DestroyImage = gfx20_DestroyImage,
    .GetImageSubresourceLayout = gfx20_GetImageSubresourceLayout,
    .CreateImageView = gfx20_CreateImageView,
    .DestroyImageView = gfx20_DestroyImageView,
    .CreateShaderModule = gfx20_CreateShaderModule,
    .DestroyShaderModule = gfx20_DestroyShaderModule,
    .CreatePipelineCache = gfx20_CreatePipelineCache,
    .DestroyPipelineCache = gfx20_DestroyPipelineCache,
    .GetPipelineCacheData = gfx20_GetPipelineCacheData,
    .MergePipelineCaches = gfx20_MergePipelineCaches,
    .CreatePipelineBinariesKHR = gfx20_CreatePipelineBinariesKHR,
    .DestroyPipelineBinaryKHR = gfx20_DestroyPipelineBinaryKHR,
    .GetPipelineKeyKHR = gfx20_GetPipelineKeyKHR,
    .GetPipelineBinaryDataKHR = gfx20_GetPipelineBinaryDataKHR,
    .ReleaseCapturedPipelineDataKHR = gfx20_ReleaseCapturedPipelineDataKHR,
    .CreateGraphicsPipelines = gfx20_CreateGraphicsPipelines,
    .CreateComputePipelines = gfx20_CreateComputePipelines,
    .GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI = gfx20_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI,
    .DestroyPipeline = gfx20_DestroyPipeline,
    .CreatePipelineLayout = gfx20_CreatePipelineLayout,
    .DestroyPipelineLayout = gfx20_DestroyPipelineLayout,
    .CreateSampler = gfx20_CreateSampler,
    .DestroySampler = gfx20_DestroySampler,
    .CreateDescriptorSetLayout = gfx20_CreateDescriptorSetLayout,
    .DestroyDescriptorSetLayout = gfx20_DestroyDescriptorSetLayout,
    .CreateDescriptorPool = gfx20_CreateDescriptorPool,
    .DestroyDescriptorPool = gfx20_DestroyDescriptorPool,
    .ResetDescriptorPool = gfx20_ResetDescriptorPool,
    .AllocateDescriptorSets = gfx20_AllocateDescriptorSets,
    .FreeDescriptorSets = gfx20_FreeDescriptorSets,
    .UpdateDescriptorSets = gfx20_UpdateDescriptorSets,
    .CreateFramebuffer = gfx20_CreateFramebuffer,
    .DestroyFramebuffer = gfx20_DestroyFramebuffer,
    .CreateRenderPass = gfx20_CreateRenderPass,
    .DestroyRenderPass = gfx20_DestroyRenderPass,
    .GetRenderAreaGranularity = gfx20_GetRenderAreaGranularity,
    .GetRenderingAreaGranularity = gfx20_GetRenderingAreaGranularity,
    .GetRenderingAreaGranularityKHR = gfx20_GetRenderingAreaGranularityKHR,
    .CreateCommandPool = gfx20_CreateCommandPool,
    .DestroyCommandPool = gfx20_DestroyCommandPool,
    .ResetCommandPool = gfx20_ResetCommandPool,
    .AllocateCommandBuffers = gfx20_AllocateCommandBuffers,
    .FreeCommandBuffers = gfx20_FreeCommandBuffers,
    .BeginCommandBuffer = gfx20_BeginCommandBuffer,
    .EndCommandBuffer = gfx20_EndCommandBuffer,
    .ResetCommandBuffer = gfx20_ResetCommandBuffer,
    .CmdBindPipeline = gfx20_CmdBindPipeline,
    .CmdSetPrimitiveRestartIndexEXT = gfx20_CmdSetPrimitiveRestartIndexEXT,
    .CmdSetAttachmentFeedbackLoopEnableEXT = gfx20_CmdSetAttachmentFeedbackLoopEnableEXT,
    .CmdSetViewport = gfx20_CmdSetViewport,
    .CmdSetScissor = gfx20_CmdSetScissor,
    .CmdSetLineWidth = gfx20_CmdSetLineWidth,
    .CmdSetDepthBias = gfx20_CmdSetDepthBias,
    .CmdSetBlendConstants = gfx20_CmdSetBlendConstants,
    .CmdSetDepthBounds = gfx20_CmdSetDepthBounds,
    .CmdSetStencilCompareMask = gfx20_CmdSetStencilCompareMask,
    .CmdSetStencilWriteMask = gfx20_CmdSetStencilWriteMask,
    .CmdSetStencilReference = gfx20_CmdSetStencilReference,
    .CmdBindDescriptorSets = gfx20_CmdBindDescriptorSets,
    .CmdBindIndexBuffer = gfx20_CmdBindIndexBuffer,
    .CmdBindVertexBuffers = gfx20_CmdBindVertexBuffers,
    .CmdDraw = gfx20_CmdDraw,
    .CmdDrawIndexed = gfx20_CmdDrawIndexed,
    .CmdDrawMultiEXT = gfx20_CmdDrawMultiEXT,
    .CmdDrawMultiIndexedEXT = gfx20_CmdDrawMultiIndexedEXT,
    .CmdDrawIndirect = gfx20_CmdDrawIndirect,
    .CmdDrawIndexedIndirect = gfx20_CmdDrawIndexedIndirect,
    .CmdDispatch = gfx20_CmdDispatch,
    .CmdDispatchIndirect = gfx20_CmdDispatchIndirect,
    .CmdSubpassShadingHUAWEI = gfx20_CmdSubpassShadingHUAWEI,
    .CmdDrawClusterHUAWEI = gfx20_CmdDrawClusterHUAWEI,
    .CmdDrawClusterIndirectHUAWEI = gfx20_CmdDrawClusterIndirectHUAWEI,
    .CmdUpdatePipelineIndirectBufferNV = gfx20_CmdUpdatePipelineIndirectBufferNV,
    .CmdCopyBuffer = gfx20_CmdCopyBuffer,
    .CmdCopyImage = gfx20_CmdCopyImage,
    .CmdBlitImage = gfx20_CmdBlitImage,
    .CmdCopyBufferToImage = gfx20_CmdCopyBufferToImage,
    .CmdCopyImageToBuffer = gfx20_CmdCopyImageToBuffer,
    .CmdCopyMemoryIndirectNV = gfx20_CmdCopyMemoryIndirectNV,
    .CmdCopyMemoryIndirectKHR = gfx20_CmdCopyMemoryIndirectKHR,
    .CmdCopyMemoryToImageIndirectNV = gfx20_CmdCopyMemoryToImageIndirectNV,
    .CmdCopyMemoryToImageIndirectKHR = gfx20_CmdCopyMemoryToImageIndirectKHR,
    .CmdUpdateBuffer = gfx20_CmdUpdateBuffer,
    .CmdFillBuffer = gfx20_CmdFillBuffer,
    .CmdClearColorImage = gfx20_CmdClearColorImage,
    .CmdClearDepthStencilImage = gfx20_CmdClearDepthStencilImage,
    .CmdClearAttachments = gfx20_CmdClearAttachments,
    .CmdResolveImage = gfx20_CmdResolveImage,
    .CmdSetEvent = gfx20_CmdSetEvent,
    .CmdResetEvent = gfx20_CmdResetEvent,
    .CmdWaitEvents = gfx20_CmdWaitEvents,
    .CmdPipelineBarrier = gfx20_CmdPipelineBarrier,
    .CmdBeginQuery = gfx20_CmdBeginQuery,
    .CmdEndQuery = gfx20_CmdEndQuery,
    .CmdBeginConditionalRenderingEXT = gfx20_CmdBeginConditionalRenderingEXT,
    .CmdEndConditionalRenderingEXT = gfx20_CmdEndConditionalRenderingEXT,
    .CmdBeginCustomResolveEXT = gfx20_CmdBeginCustomResolveEXT,
    .CmdResetQueryPool = gfx20_CmdResetQueryPool,
    .CmdWriteTimestamp = gfx20_CmdWriteTimestamp,
    .CmdCopyQueryPoolResults = gfx20_CmdCopyQueryPoolResults,
    .CmdPushConstants = gfx20_CmdPushConstants,
    .CmdBeginRenderPass = gfx20_CmdBeginRenderPass,
    .CmdNextSubpass = gfx20_CmdNextSubpass,
    .CmdEndRenderPass = gfx20_CmdEndRenderPass,
    .CmdExecuteCommands = gfx20_CmdExecuteCommands,
    .CreateSharedSwapchainsKHR = gfx20_CreateSharedSwapchainsKHR,
    .CreateSwapchainKHR = gfx20_CreateSwapchainKHR,
    .DestroySwapchainKHR = gfx20_DestroySwapchainKHR,
    .GetSwapchainImagesKHR = gfx20_GetSwapchainImagesKHR,
    .AcquireNextImageKHR = gfx20_AcquireNextImageKHR,
    .QueuePresentKHR = gfx20_QueuePresentKHR,
    .DebugMarkerSetObjectNameEXT = gfx20_DebugMarkerSetObjectNameEXT,
    .DebugMarkerSetObjectTagEXT = gfx20_DebugMarkerSetObjectTagEXT,
    .CmdDebugMarkerBeginEXT = gfx20_CmdDebugMarkerBeginEXT,
    .CmdDebugMarkerEndEXT = gfx20_CmdDebugMarkerEndEXT,
    .CmdDebugMarkerInsertEXT = gfx20_CmdDebugMarkerInsertEXT,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleNV = gfx20_GetMemoryWin32HandleNV,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .CmdExecuteGeneratedCommandsNV = gfx20_CmdExecuteGeneratedCommandsNV,
    .CmdPreprocessGeneratedCommandsNV = gfx20_CmdPreprocessGeneratedCommandsNV,
    .CmdBindPipelineShaderGroupNV = gfx20_CmdBindPipelineShaderGroupNV,
    .GetGeneratedCommandsMemoryRequirementsNV = gfx20_GetGeneratedCommandsMemoryRequirementsNV,
    .CreateIndirectCommandsLayoutNV = gfx20_CreateIndirectCommandsLayoutNV,
    .DestroyIndirectCommandsLayoutNV = gfx20_DestroyIndirectCommandsLayoutNV,
    .CmdExecuteGeneratedCommandsEXT = gfx20_CmdExecuteGeneratedCommandsEXT,
    .CmdPreprocessGeneratedCommandsEXT = gfx20_CmdPreprocessGeneratedCommandsEXT,
    .GetGeneratedCommandsMemoryRequirementsEXT = gfx20_GetGeneratedCommandsMemoryRequirementsEXT,
    .CreateIndirectCommandsLayoutEXT = gfx20_CreateIndirectCommandsLayoutEXT,
    .DestroyIndirectCommandsLayoutEXT = gfx20_DestroyIndirectCommandsLayoutEXT,
    .CreateIndirectExecutionSetEXT = gfx20_CreateIndirectExecutionSetEXT,
    .DestroyIndirectExecutionSetEXT = gfx20_DestroyIndirectExecutionSetEXT,
    .UpdateIndirectExecutionSetPipelineEXT = gfx20_UpdateIndirectExecutionSetPipelineEXT,
    .UpdateIndirectExecutionSetShaderEXT = gfx20_UpdateIndirectExecutionSetShaderEXT,
    .CmdPushDescriptorSet = gfx20_CmdPushDescriptorSet,
    .CmdPushDescriptorSetKHR = gfx20_CmdPushDescriptorSetKHR,
    .TrimCommandPool = gfx20_TrimCommandPool,
    .TrimCommandPoolKHR = gfx20_TrimCommandPoolKHR,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleKHR = gfx20_GetMemoryWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandlePropertiesKHR = gfx20_GetMemoryWin32HandlePropertiesKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandlePropertiesKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryFdKHR = gfx20_GetMemoryFdKHR,
    .GetMemoryFdPropertiesKHR = gfx20_GetMemoryFdPropertiesKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandleFUCHSIA = gfx20_GetMemoryZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandlePropertiesFUCHSIA = gfx20_GetMemoryZirconHandlePropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandlePropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .GetMemoryRemoteAddressNV = gfx20_GetMemoryRemoteAddressNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreWin32HandleKHR = gfx20_GetSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportSemaphoreWin32HandleKHR = gfx20_ImportSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreFdKHR = gfx20_GetSemaphoreFdKHR,
    .ImportSemaphoreFdKHR = gfx20_ImportSemaphoreFdKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetSemaphoreZirconHandleFUCHSIA = gfx20_GetSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .ImportSemaphoreZirconHandleFUCHSIA = gfx20_ImportSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .ImportSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetFenceWin32HandleKHR = gfx20_GetFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportFenceWin32HandleKHR = gfx20_ImportFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetFenceFdKHR = gfx20_GetFenceFdKHR,
    .ImportFenceFdKHR = gfx20_ImportFenceFdKHR,
    .DisplayPowerControlEXT = gfx20_DisplayPowerControlEXT,
    .RegisterDeviceEventEXT = gfx20_RegisterDeviceEventEXT,
    .RegisterDisplayEventEXT = gfx20_RegisterDisplayEventEXT,
    .GetSwapchainCounterEXT = gfx20_GetSwapchainCounterEXT,
    .GetDeviceGroupPeerMemoryFeatures = gfx20_GetDeviceGroupPeerMemoryFeatures,
    .GetDeviceGroupPeerMemoryFeaturesKHR = gfx20_GetDeviceGroupPeerMemoryFeaturesKHR,
    .BindBufferMemory2 = gfx20_BindBufferMemory2,
    .BindBufferMemory2KHR = gfx20_BindBufferMemory2KHR,
    .BindImageMemory2 = gfx20_BindImageMemory2,
    .BindImageMemory2KHR = gfx20_BindImageMemory2KHR,
    .CmdSetDeviceMask = gfx20_CmdSetDeviceMask,
    .CmdSetDeviceMaskKHR = gfx20_CmdSetDeviceMaskKHR,
    .GetDeviceGroupPresentCapabilitiesKHR = gfx20_GetDeviceGroupPresentCapabilitiesKHR,
    .GetDeviceGroupSurfacePresentModesKHR = gfx20_GetDeviceGroupSurfacePresentModesKHR,
    .AcquireNextImage2KHR = gfx20_AcquireNextImage2KHR,
    .CmdDispatchBase = gfx20_CmdDispatchBase,
    .CmdDispatchBaseKHR = gfx20_CmdDispatchBaseKHR,
    .CreateDescriptorUpdateTemplate = gfx20_CreateDescriptorUpdateTemplate,
    .CreateDescriptorUpdateTemplateKHR = gfx20_CreateDescriptorUpdateTemplateKHR,
    .DestroyDescriptorUpdateTemplate = gfx20_DestroyDescriptorUpdateTemplate,
    .DestroyDescriptorUpdateTemplateKHR = gfx20_DestroyDescriptorUpdateTemplateKHR,
    .UpdateDescriptorSetWithTemplate = gfx20_UpdateDescriptorSetWithTemplate,
    .UpdateDescriptorSetWithTemplateKHR = gfx20_UpdateDescriptorSetWithTemplateKHR,
    .CmdPushDescriptorSetWithTemplate = gfx20_CmdPushDescriptorSetWithTemplate,
    .CmdPushDescriptorSetWithTemplateKHR = gfx20_CmdPushDescriptorSetWithTemplateKHR,
    .SetHdrMetadataEXT = gfx20_SetHdrMetadataEXT,
    .GetSwapchainStatusKHR = gfx20_GetSwapchainStatusKHR,
    .GetRefreshCycleDurationGOOGLE = gfx20_GetRefreshCycleDurationGOOGLE,
    .GetPastPresentationTimingGOOGLE = gfx20_GetPastPresentationTimingGOOGLE,
    .CmdSetViewportWScalingNV = gfx20_CmdSetViewportWScalingNV,
    .CmdSetDiscardRectangleEXT = gfx20_CmdSetDiscardRectangleEXT,
    .CmdSetDiscardRectangleEnableEXT = gfx20_CmdSetDiscardRectangleEnableEXT,
    .CmdSetDiscardRectangleModeEXT = gfx20_CmdSetDiscardRectangleModeEXT,
    .CmdSetSampleLocationsEXT = gfx20_CmdSetSampleLocationsEXT,
    .GetBufferMemoryRequirements2 = gfx20_GetBufferMemoryRequirements2,
    .GetBufferMemoryRequirements2KHR = gfx20_GetBufferMemoryRequirements2KHR,
    .GetImageMemoryRequirements2 = gfx20_GetImageMemoryRequirements2,
    .GetImageMemoryRequirements2KHR = gfx20_GetImageMemoryRequirements2KHR,
    .GetImageSparseMemoryRequirements2 = gfx20_GetImageSparseMemoryRequirements2,
    .GetImageSparseMemoryRequirements2KHR = gfx20_GetImageSparseMemoryRequirements2KHR,
    .GetDeviceBufferMemoryRequirements = gfx20_GetDeviceBufferMemoryRequirements,
    .GetDeviceBufferMemoryRequirementsKHR = gfx20_GetDeviceBufferMemoryRequirementsKHR,
    .GetDeviceImageMemoryRequirements = gfx20_GetDeviceImageMemoryRequirements,
    .GetDeviceImageMemoryRequirementsKHR = gfx20_GetDeviceImageMemoryRequirementsKHR,
    .GetDeviceImageSparseMemoryRequirements = gfx20_GetDeviceImageSparseMemoryRequirements,
    .GetDeviceImageSparseMemoryRequirementsKHR = gfx20_GetDeviceImageSparseMemoryRequirementsKHR,
    .CreateSamplerYcbcrConversion = gfx20_CreateSamplerYcbcrConversion,
    .CreateSamplerYcbcrConversionKHR = gfx20_CreateSamplerYcbcrConversionKHR,
    .DestroySamplerYcbcrConversion = gfx20_DestroySamplerYcbcrConversion,
    .DestroySamplerYcbcrConversionKHR = gfx20_DestroySamplerYcbcrConversionKHR,
    .GetDeviceQueue2 = gfx20_GetDeviceQueue2,
    .CreateValidationCacheEXT = gfx20_CreateValidationCacheEXT,
    .DestroyValidationCacheEXT = gfx20_DestroyValidationCacheEXT,
    .GetValidationCacheDataEXT = gfx20_GetValidationCacheDataEXT,
    .MergeValidationCachesEXT = gfx20_MergeValidationCachesEXT,
    .GetDescriptorSetLayoutSupport = gfx20_GetDescriptorSetLayoutSupport,
    .GetDescriptorSetLayoutSupportKHR = gfx20_GetDescriptorSetLayoutSupportKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsageANDROID = gfx20_GetSwapchainGrallocUsageANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsage2ANDROID = gfx20_GetSwapchainGrallocUsage2ANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsage2ANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .AcquireImageANDROID = gfx20_AcquireImageANDROID,
#elif defined(_MSC_VER)
    .AcquireImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .QueueSignalReleaseImageANDROID = gfx20_QueueSignalReleaseImageANDROID,
#elif defined(_MSC_VER)
    .QueueSignalReleaseImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .GetShaderInfoAMD = gfx20_GetShaderInfoAMD,
    .SetLocalDimmingAMD = gfx20_SetLocalDimmingAMD,
    .GetCalibratedTimestampsKHR = gfx20_GetCalibratedTimestampsKHR,
    .GetCalibratedTimestampsEXT = gfx20_GetCalibratedTimestampsEXT,
    .SetDebugUtilsObjectNameEXT = gfx20_SetDebugUtilsObjectNameEXT,
    .SetDebugUtilsObjectTagEXT = gfx20_SetDebugUtilsObjectTagEXT,
    .QueueBeginDebugUtilsLabelEXT = gfx20_QueueBeginDebugUtilsLabelEXT,
    .QueueEndDebugUtilsLabelEXT = gfx20_QueueEndDebugUtilsLabelEXT,
    .QueueInsertDebugUtilsLabelEXT = gfx20_QueueInsertDebugUtilsLabelEXT,
    .CmdBeginDebugUtilsLabelEXT = gfx20_CmdBeginDebugUtilsLabelEXT,
    .CmdEndDebugUtilsLabelEXT = gfx20_CmdEndDebugUtilsLabelEXT,
    .CmdInsertDebugUtilsLabelEXT = gfx20_CmdInsertDebugUtilsLabelEXT,
    .GetMemoryHostPointerPropertiesEXT = gfx20_GetMemoryHostPointerPropertiesEXT,
    .CmdWriteBufferMarkerAMD = gfx20_CmdWriteBufferMarkerAMD,
    .CreateRenderPass2 = gfx20_CreateRenderPass2,
    .CreateRenderPass2KHR = gfx20_CreateRenderPass2KHR,
    .CmdBeginRenderPass2 = gfx20_CmdBeginRenderPass2,
    .CmdBeginRenderPass2KHR = gfx20_CmdBeginRenderPass2KHR,
    .CmdNextSubpass2 = gfx20_CmdNextSubpass2,
    .CmdNextSubpass2KHR = gfx20_CmdNextSubpass2KHR,
    .CmdEndRenderPass2 = gfx20_CmdEndRenderPass2,
    .CmdEndRenderPass2KHR = gfx20_CmdEndRenderPass2KHR,
    .GetSemaphoreCounterValue = gfx20_GetSemaphoreCounterValue,
    .GetSemaphoreCounterValueKHR = gfx20_GetSemaphoreCounterValueKHR,
    .WaitSemaphores = gfx20_WaitSemaphores,
    .WaitSemaphoresKHR = gfx20_WaitSemaphoresKHR,
    .SignalSemaphore = gfx20_SignalSemaphore,
    .SignalSemaphoreKHR = gfx20_SignalSemaphoreKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetAndroidHardwareBufferPropertiesANDROID = gfx20_GetAndroidHardwareBufferPropertiesANDROID,
#elif defined(_MSC_VER)
    .GetAndroidHardwareBufferPropertiesANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetMemoryAndroidHardwareBufferANDROID = gfx20_GetMemoryAndroidHardwareBufferANDROID,
#elif defined(_MSC_VER)
    .GetMemoryAndroidHardwareBufferANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .CmdDrawIndirectCount = gfx20_CmdDrawIndirectCount,
    .CmdDrawIndirectCountKHR = gfx20_CmdDrawIndirectCountKHR,
    .CmdDrawIndirectCountAMD = gfx20_CmdDrawIndirectCountAMD,
    .CmdDrawIndexedIndirectCount = gfx20_CmdDrawIndexedIndirectCount,
    .CmdDrawIndexedIndirectCountKHR = gfx20_CmdDrawIndexedIndirectCountKHR,
    .CmdDrawIndexedIndirectCountAMD = gfx20_CmdDrawIndexedIndirectCountAMD,
    .CmdSetCheckpointNV = gfx20_CmdSetCheckpointNV,
    .GetQueueCheckpointDataNV = gfx20_GetQueueCheckpointDataNV,
    .CmdBindTransformFeedbackBuffersEXT = gfx20_CmdBindTransformFeedbackBuffersEXT,
    .CmdBeginTransformFeedbackEXT = gfx20_CmdBeginTransformFeedbackEXT,
    .CmdEndTransformFeedbackEXT = gfx20_CmdEndTransformFeedbackEXT,
    .CmdBeginQueryIndexedEXT = gfx20_CmdBeginQueryIndexedEXT,
    .CmdEndQueryIndexedEXT = gfx20_CmdEndQueryIndexedEXT,
    .CmdDrawIndirectByteCountEXT = gfx20_CmdDrawIndirectByteCountEXT,
    .CmdSetExclusiveScissorNV = gfx20_CmdSetExclusiveScissorNV,
    .CmdSetExclusiveScissorEnableNV = gfx20_CmdSetExclusiveScissorEnableNV,
    .CmdBindShadingRateImageNV = gfx20_CmdBindShadingRateImageNV,
    .CmdSetViewportShadingRatePaletteNV = gfx20_CmdSetViewportShadingRatePaletteNV,
    .CmdSetCoarseSampleOrderNV = gfx20_CmdSetCoarseSampleOrderNV,
    .CmdDrawMeshTasksNV = gfx20_CmdDrawMeshTasksNV,
    .CmdDrawMeshTasksIndirectNV = gfx20_CmdDrawMeshTasksIndirectNV,
    .CmdDrawMeshTasksIndirectCountNV = gfx20_CmdDrawMeshTasksIndirectCountNV,
    .CmdDrawMeshTasksEXT = gfx20_CmdDrawMeshTasksEXT,
    .CmdDrawMeshTasksIndirectEXT = gfx20_CmdDrawMeshTasksIndirectEXT,
    .CmdDrawMeshTasksIndirectCountEXT = gfx20_CmdDrawMeshTasksIndirectCountEXT,
    .CompileDeferredNV = gfx20_CompileDeferredNV,
    .CreateAccelerationStructureNV = gfx20_CreateAccelerationStructureNV,
    .CmdBindInvocationMaskHUAWEI = gfx20_CmdBindInvocationMaskHUAWEI,
    .DestroyAccelerationStructureKHR = gfx20_DestroyAccelerationStructureKHR,
    .DestroyAccelerationStructureNV = gfx20_DestroyAccelerationStructureNV,
    .GetAccelerationStructureMemoryRequirementsNV = gfx20_GetAccelerationStructureMemoryRequirementsNV,
    .BindAccelerationStructureMemoryNV = gfx20_BindAccelerationStructureMemoryNV,
    .CmdCopyAccelerationStructureNV = gfx20_CmdCopyAccelerationStructureNV,
    .CmdCopyAccelerationStructureKHR = gfx20_CmdCopyAccelerationStructureKHR,
    .CopyAccelerationStructureKHR = gfx20_CopyAccelerationStructureKHR,
    .CmdCopyAccelerationStructureToMemoryKHR = gfx20_CmdCopyAccelerationStructureToMemoryKHR,
    .CopyAccelerationStructureToMemoryKHR = gfx20_CopyAccelerationStructureToMemoryKHR,
    .CmdCopyMemoryToAccelerationStructureKHR = gfx20_CmdCopyMemoryToAccelerationStructureKHR,
    .CopyMemoryToAccelerationStructureKHR = gfx20_CopyMemoryToAccelerationStructureKHR,
    .CmdWriteAccelerationStructuresPropertiesKHR = gfx20_CmdWriteAccelerationStructuresPropertiesKHR,
    .CmdWriteAccelerationStructuresPropertiesNV = gfx20_CmdWriteAccelerationStructuresPropertiesNV,
    .CmdBuildAccelerationStructureNV = gfx20_CmdBuildAccelerationStructureNV,
    .WriteAccelerationStructuresPropertiesKHR = gfx20_WriteAccelerationStructuresPropertiesKHR,
    .CmdTraceRaysKHR = gfx20_CmdTraceRaysKHR,
    .CmdTraceRaysNV = gfx20_CmdTraceRaysNV,
    .GetRayTracingShaderGroupHandlesKHR = gfx20_GetRayTracingShaderGroupHandlesKHR,
    .GetRayTracingShaderGroupHandlesNV = gfx20_GetRayTracingShaderGroupHandlesNV,
    .GetRayTracingCaptureReplayShaderGroupHandlesKHR = gfx20_GetRayTracingCaptureReplayShaderGroupHandlesKHR,
    .GetAccelerationStructureHandleNV = gfx20_GetAccelerationStructureHandleNV,
    .CreateRayTracingPipelinesNV = gfx20_CreateRayTracingPipelinesNV,
    .CreateRayTracingPipelinesKHR = gfx20_CreateRayTracingPipelinesKHR,
    .CmdTraceRaysIndirectKHR = gfx20_CmdTraceRaysIndirectKHR,
    .CmdTraceRaysIndirect2KHR = gfx20_CmdTraceRaysIndirect2KHR,
    .GetClusterAccelerationStructureBuildSizesNV = gfx20_GetClusterAccelerationStructureBuildSizesNV,
    .CmdBuildClusterAccelerationStructureIndirectNV = gfx20_CmdBuildClusterAccelerationStructureIndirectNV,
    .GetDeviceAccelerationStructureCompatibilityKHR = gfx20_GetDeviceAccelerationStructureCompatibilityKHR,
    .GetRayTracingShaderGroupStackSizeKHR = gfx20_GetRayTracingShaderGroupStackSizeKHR,
    .CmdSetRayTracingPipelineStackSizeKHR = gfx20_CmdSetRayTracingPipelineStackSizeKHR,
    .GetImageViewHandleNVX = gfx20_GetImageViewHandleNVX,
    .GetImageViewHandle64NVX = gfx20_GetImageViewHandle64NVX,
    .GetImageViewAddressNVX = gfx20_GetImageViewAddressNVX,
    .GetDeviceCombinedImageSamplerIndexNVX = gfx20_GetDeviceCombinedImageSamplerIndexNVX,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetDeviceGroupSurfacePresentModes2EXT = gfx20_GetDeviceGroupSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetDeviceGroupSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireFullScreenExclusiveModeEXT = gfx20_AcquireFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .AcquireFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ReleaseFullScreenExclusiveModeEXT = gfx20_ReleaseFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .ReleaseFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .AcquireProfilingLockKHR = gfx20_AcquireProfilingLockKHR,
    .ReleaseProfilingLockKHR = gfx20_ReleaseProfilingLockKHR,
    .GetImageDrmFormatModifierPropertiesEXT = gfx20_GetImageDrmFormatModifierPropertiesEXT,
    .GetBufferOpaqueCaptureAddress = gfx20_GetBufferOpaqueCaptureAddress,
    .GetBufferOpaqueCaptureAddressKHR = gfx20_GetBufferOpaqueCaptureAddressKHR,
    .GetBufferDeviceAddress = gfx20_GetBufferDeviceAddress,
    .GetBufferDeviceAddressKHR = gfx20_GetBufferDeviceAddressKHR,
    .GetBufferDeviceAddressEXT = gfx20_GetBufferDeviceAddressEXT,
    .InitializePerformanceApiINTEL = gfx20_InitializePerformanceApiINTEL,
    .UninitializePerformanceApiINTEL = gfx20_UninitializePerformanceApiINTEL,
    .CmdSetPerformanceMarkerINTEL = gfx20_CmdSetPerformanceMarkerINTEL,
    .CmdSetPerformanceStreamMarkerINTEL = gfx20_CmdSetPerformanceStreamMarkerINTEL,
    .CmdSetPerformanceOverrideINTEL = gfx20_CmdSetPerformanceOverrideINTEL,
    .AcquirePerformanceConfigurationINTEL = gfx20_AcquirePerformanceConfigurationINTEL,
    .ReleasePerformanceConfigurationINTEL = gfx20_ReleasePerformanceConfigurationINTEL,
    .QueueSetPerformanceConfigurationINTEL = gfx20_QueueSetPerformanceConfigurationINTEL,
    .GetPerformanceParameterINTEL = gfx20_GetPerformanceParameterINTEL,
    .GetDeviceMemoryOpaqueCaptureAddress = gfx20_GetDeviceMemoryOpaqueCaptureAddress,
    .GetDeviceMemoryOpaqueCaptureAddressKHR = gfx20_GetDeviceMemoryOpaqueCaptureAddressKHR,
    .GetPipelineExecutablePropertiesKHR = gfx20_GetPipelineExecutablePropertiesKHR,
    .GetPipelineExecutableStatisticsKHR = gfx20_GetPipelineExecutableStatisticsKHR,
    .GetPipelineExecutableInternalRepresentationsKHR = gfx20_GetPipelineExecutableInternalRepresentationsKHR,
    .CmdSetLineStipple = gfx20_CmdSetLineStipple,
    .CmdSetLineStippleKHR = gfx20_CmdSetLineStippleKHR,
    .CmdSetLineStippleEXT = gfx20_CmdSetLineStippleEXT,
    .CreateAccelerationStructureKHR = gfx20_CreateAccelerationStructureKHR,
    .CmdBuildAccelerationStructuresKHR = gfx20_CmdBuildAccelerationStructuresKHR,
    .CmdBuildAccelerationStructuresIndirectKHR = gfx20_CmdBuildAccelerationStructuresIndirectKHR,
    .BuildAccelerationStructuresKHR = gfx20_BuildAccelerationStructuresKHR,
    .GetAccelerationStructureDeviceAddressKHR = gfx20_GetAccelerationStructureDeviceAddressKHR,
    .CreateDeferredOperationKHR = gfx20_CreateDeferredOperationKHR,
    .DestroyDeferredOperationKHR = gfx20_DestroyDeferredOperationKHR,
    .GetDeferredOperationMaxConcurrencyKHR = gfx20_GetDeferredOperationMaxConcurrencyKHR,
    .GetDeferredOperationResultKHR = gfx20_GetDeferredOperationResultKHR,
    .DeferredOperationJoinKHR = gfx20_DeferredOperationJoinKHR,
    .GetPipelineIndirectMemoryRequirementsNV = gfx20_GetPipelineIndirectMemoryRequirementsNV,
    .GetPipelineIndirectDeviceAddressNV = gfx20_GetPipelineIndirectDeviceAddressNV,
    .AntiLagUpdateAMD = gfx20_AntiLagUpdateAMD,
    .CmdSetCullMode = gfx20_CmdSetCullMode,
    .CmdSetCullModeEXT = gfx20_CmdSetCullModeEXT,
    .CmdSetFrontFace = gfx20_CmdSetFrontFace,
    .CmdSetFrontFaceEXT = gfx20_CmdSetFrontFaceEXT,
    .CmdSetPrimitiveTopology = gfx20_CmdSetPrimitiveTopology,
    .CmdSetPrimitiveTopologyEXT = gfx20_CmdSetPrimitiveTopologyEXT,
    .CmdSetViewportWithCount = gfx20_CmdSetViewportWithCount,
    .CmdSetViewportWithCountEXT = gfx20_CmdSetViewportWithCountEXT,
    .CmdSetScissorWithCount = gfx20_CmdSetScissorWithCount,
    .CmdSetScissorWithCountEXT = gfx20_CmdSetScissorWithCountEXT,
    .CmdBindIndexBuffer2 = gfx20_CmdBindIndexBuffer2,
    .CmdBindIndexBuffer2KHR = gfx20_CmdBindIndexBuffer2KHR,
    .CmdBindVertexBuffers2 = gfx20_CmdBindVertexBuffers2,
    .CmdBindVertexBuffers2EXT = gfx20_CmdBindVertexBuffers2EXT,
    .CmdSetDepthTestEnable = gfx20_CmdSetDepthTestEnable,
    .CmdSetDepthTestEnableEXT = gfx20_CmdSetDepthTestEnableEXT,
    .CmdSetDepthWriteEnable = gfx20_CmdSetDepthWriteEnable,
    .CmdSetDepthWriteEnableEXT = gfx20_CmdSetDepthWriteEnableEXT,
    .CmdSetDepthCompareOp = gfx20_CmdSetDepthCompareOp,
    .CmdSetDepthCompareOpEXT = gfx20_CmdSetDepthCompareOpEXT,
    .CmdSetDepthBoundsTestEnable = gfx20_CmdSetDepthBoundsTestEnable,
    .CmdSetDepthBoundsTestEnableEXT = gfx20_CmdSetDepthBoundsTestEnableEXT,
    .CmdSetStencilTestEnable = gfx20_CmdSetStencilTestEnable,
    .CmdSetStencilTestEnableEXT = gfx20_CmdSetStencilTestEnableEXT,
    .CmdSetStencilOp = gfx20_CmdSetStencilOp,
    .CmdSetStencilOpEXT = gfx20_CmdSetStencilOpEXT,
    .CmdSetPatchControlPointsEXT = gfx20_CmdSetPatchControlPointsEXT,
    .CmdSetRasterizerDiscardEnable = gfx20_CmdSetRasterizerDiscardEnable,
    .CmdSetRasterizerDiscardEnableEXT = gfx20_CmdSetRasterizerDiscardEnableEXT,
    .CmdSetDepthBiasEnable = gfx20_CmdSetDepthBiasEnable,
    .CmdSetDepthBiasEnableEXT = gfx20_CmdSetDepthBiasEnableEXT,
    .CmdSetLogicOpEXT = gfx20_CmdSetLogicOpEXT,
    .CmdSetPrimitiveRestartEnable = gfx20_CmdSetPrimitiveRestartEnable,
    .CmdSetPrimitiveRestartEnableEXT = gfx20_CmdSetPrimitiveRestartEnableEXT,
    .CmdSetTessellationDomainOriginEXT = gfx20_CmdSetTessellationDomainOriginEXT,
    .CmdSetDepthClampEnableEXT = gfx20_CmdSetDepthClampEnableEXT,
    .CmdSetPolygonModeEXT = gfx20_CmdSetPolygonModeEXT,
    .CmdSetRasterizationSamplesEXT = gfx20_CmdSetRasterizationSamplesEXT,
    .CmdSetSampleMaskEXT = gfx20_CmdSetSampleMaskEXT,
    .CmdSetAlphaToCoverageEnableEXT = gfx20_CmdSetAlphaToCoverageEnableEXT,
    .CmdSetAlphaToOneEnableEXT = gfx20_CmdSetAlphaToOneEnableEXT,
    .CmdSetLogicOpEnableEXT = gfx20_CmdSetLogicOpEnableEXT,
    .CmdSetColorBlendEnableEXT = gfx20_CmdSetColorBlendEnableEXT,
    .CmdSetColorBlendEquationEXT = gfx20_CmdSetColorBlendEquationEXT,
    .CmdSetColorWriteMaskEXT = gfx20_CmdSetColorWriteMaskEXT,
    .CmdSetRasterizationStreamEXT = gfx20_CmdSetRasterizationStreamEXT,
    .CmdSetConservativeRasterizationModeEXT = gfx20_CmdSetConservativeRasterizationModeEXT,
    .CmdSetExtraPrimitiveOverestimationSizeEXT = gfx20_CmdSetExtraPrimitiveOverestimationSizeEXT,
    .CmdSetDepthClipEnableEXT = gfx20_CmdSetDepthClipEnableEXT,
    .CmdSetSampleLocationsEnableEXT = gfx20_CmdSetSampleLocationsEnableEXT,
    .CmdSetColorBlendAdvancedEXT = gfx20_CmdSetColorBlendAdvancedEXT,
    .CmdSetProvokingVertexModeEXT = gfx20_CmdSetProvokingVertexModeEXT,
    .CmdSetLineRasterizationModeEXT = gfx20_CmdSetLineRasterizationModeEXT,
    .CmdSetLineStippleEnableEXT = gfx20_CmdSetLineStippleEnableEXT,
    .CmdSetDepthClipNegativeOneToOneEXT = gfx20_CmdSetDepthClipNegativeOneToOneEXT,
    .CmdSetViewportWScalingEnableNV = gfx20_CmdSetViewportWScalingEnableNV,
    .CmdSetViewportSwizzleNV = gfx20_CmdSetViewportSwizzleNV,
    .CmdSetCoverageToColorEnableNV = gfx20_CmdSetCoverageToColorEnableNV,
    .CmdSetCoverageToColorLocationNV = gfx20_CmdSetCoverageToColorLocationNV,
    .CmdSetCoverageModulationModeNV = gfx20_CmdSetCoverageModulationModeNV,
    .CmdSetCoverageModulationTableEnableNV = gfx20_CmdSetCoverageModulationTableEnableNV,
    .CmdSetCoverageModulationTableNV = gfx20_CmdSetCoverageModulationTableNV,
    .CmdSetShadingRateImageEnableNV = gfx20_CmdSetShadingRateImageEnableNV,
    .CmdSetCoverageReductionModeNV = gfx20_CmdSetCoverageReductionModeNV,
    .CmdSetRepresentativeFragmentTestEnableNV = gfx20_CmdSetRepresentativeFragmentTestEnableNV,
    .CreatePrivateDataSlot = gfx20_CreatePrivateDataSlot,
    .CreatePrivateDataSlotEXT = gfx20_CreatePrivateDataSlotEXT,
    .DestroyPrivateDataSlot = gfx20_DestroyPrivateDataSlot,
    .DestroyPrivateDataSlotEXT = gfx20_DestroyPrivateDataSlotEXT,
    .SetPrivateData = gfx20_SetPrivateData,
    .SetPrivateDataEXT = gfx20_SetPrivateDataEXT,
    .GetPrivateData = gfx20_GetPrivateData,
    .GetPrivateDataEXT = gfx20_GetPrivateDataEXT,
    .CmdCopyBuffer2 = gfx20_CmdCopyBuffer2,
    .CmdCopyBuffer2KHR = gfx20_CmdCopyBuffer2KHR,
    .CmdCopyImage2 = gfx20_CmdCopyImage2,
    .CmdCopyImage2KHR = gfx20_CmdCopyImage2KHR,
    .CmdBlitImage2 = gfx20_CmdBlitImage2,
    .CmdBlitImage2KHR = gfx20_CmdBlitImage2KHR,
    .CmdCopyBufferToImage2 = gfx20_CmdCopyBufferToImage2,
    .CmdCopyBufferToImage2KHR = gfx20_CmdCopyBufferToImage2KHR,
    .CmdCopyImageToBuffer2 = gfx20_CmdCopyImageToBuffer2,
    .CmdCopyImageToBuffer2KHR = gfx20_CmdCopyImageToBuffer2KHR,
    .CmdResolveImage2 = gfx20_CmdResolveImage2,
    .CmdResolveImage2KHR = gfx20_CmdResolveImage2KHR,
    .CmdSetFragmentShadingRateKHR = gfx20_CmdSetFragmentShadingRateKHR,
    .CmdSetFragmentShadingRateEnumNV = gfx20_CmdSetFragmentShadingRateEnumNV,
    .GetAccelerationStructureBuildSizesKHR = gfx20_GetAccelerationStructureBuildSizesKHR,
    .CmdSetVertexInputEXT = gfx20_CmdSetVertexInputEXT,
    .CmdSetColorWriteEnableEXT = gfx20_CmdSetColorWriteEnableEXT,
    .CmdSetEvent2 = gfx20_CmdSetEvent2,
    .CmdSetEvent2KHR = gfx20_CmdSetEvent2KHR,
    .CmdResetEvent2 = gfx20_CmdResetEvent2,
    .CmdResetEvent2KHR = gfx20_CmdResetEvent2KHR,
    .CmdWaitEvents2 = gfx20_CmdWaitEvents2,
    .CmdWaitEvents2KHR = gfx20_CmdWaitEvents2KHR,
    .CmdPipelineBarrier2 = gfx20_CmdPipelineBarrier2,
    .CmdPipelineBarrier2KHR = gfx20_CmdPipelineBarrier2KHR,
    .QueueSubmit2 = gfx20_QueueSubmit2,
    .QueueSubmit2KHR = gfx20_QueueSubmit2KHR,
    .CmdWriteTimestamp2 = gfx20_CmdWriteTimestamp2,
    .CmdWriteTimestamp2KHR = gfx20_CmdWriteTimestamp2KHR,
    .CmdWriteBufferMarker2AMD = gfx20_CmdWriteBufferMarker2AMD,
    .GetQueueCheckpointData2NV = gfx20_GetQueueCheckpointData2NV,
    .CopyMemoryToImage = gfx20_CopyMemoryToImage,
    .CopyMemoryToImageEXT = gfx20_CopyMemoryToImageEXT,
    .CopyImageToMemory = gfx20_CopyImageToMemory,
    .CopyImageToMemoryEXT = gfx20_CopyImageToMemoryEXT,
    .CopyImageToImage = gfx20_CopyImageToImage,
    .CopyImageToImageEXT = gfx20_CopyImageToImageEXT,
    .TransitionImageLayout = gfx20_TransitionImageLayout,
    .TransitionImageLayoutEXT = gfx20_TransitionImageLayoutEXT,
    .CreateVideoSessionKHR = gfx20_CreateVideoSessionKHR,
    .DestroyVideoSessionKHR = gfx20_DestroyVideoSessionKHR,
    .CreateVideoSessionParametersKHR = gfx20_CreateVideoSessionParametersKHR,
    .UpdateVideoSessionParametersKHR = gfx20_UpdateVideoSessionParametersKHR,
    .GetEncodedVideoSessionParametersKHR = gfx20_GetEncodedVideoSessionParametersKHR,
    .DestroyVideoSessionParametersKHR = gfx20_DestroyVideoSessionParametersKHR,
    .GetVideoSessionMemoryRequirementsKHR = gfx20_GetVideoSessionMemoryRequirementsKHR,
    .BindVideoSessionMemoryKHR = gfx20_BindVideoSessionMemoryKHR,
    .CmdDecodeVideoKHR = gfx20_CmdDecodeVideoKHR,
    .CmdBeginVideoCodingKHR = gfx20_CmdBeginVideoCodingKHR,
    .CmdControlVideoCodingKHR = gfx20_CmdControlVideoCodingKHR,
    .CmdEndVideoCodingKHR = gfx20_CmdEndVideoCodingKHR,
    .CmdEncodeVideoKHR = gfx20_CmdEncodeVideoKHR,
    .CmdDecompressMemoryNV = gfx20_CmdDecompressMemoryNV,
    .CmdDecompressMemoryIndirectCountNV = gfx20_CmdDecompressMemoryIndirectCountNV,
    .GetPartitionedAccelerationStructuresBuildSizesNV = gfx20_GetPartitionedAccelerationStructuresBuildSizesNV,
    .CmdBuildPartitionedAccelerationStructuresNV = gfx20_CmdBuildPartitionedAccelerationStructuresNV,
    .CmdDecompressMemoryEXT = gfx20_CmdDecompressMemoryEXT,
    .CmdDecompressMemoryIndirectCountEXT = gfx20_CmdDecompressMemoryIndirectCountEXT,
    .CreateCuModuleNVX = gfx20_CreateCuModuleNVX,
    .CreateCuFunctionNVX = gfx20_CreateCuFunctionNVX,
    .DestroyCuModuleNVX = gfx20_DestroyCuModuleNVX,
    .DestroyCuFunctionNVX = gfx20_DestroyCuFunctionNVX,
    .CmdCuLaunchKernelNVX = gfx20_CmdCuLaunchKernelNVX,
    .GetDescriptorSetLayoutSizeEXT = gfx20_GetDescriptorSetLayoutSizeEXT,
    .GetDescriptorSetLayoutBindingOffsetEXT = gfx20_GetDescriptorSetLayoutBindingOffsetEXT,
    .GetDescriptorEXT = gfx20_GetDescriptorEXT,
    .CmdBindDescriptorBuffersEXT = gfx20_CmdBindDescriptorBuffersEXT,
    .CmdSetDescriptorBufferOffsetsEXT = gfx20_CmdSetDescriptorBufferOffsetsEXT,
    .CmdBindDescriptorBufferEmbeddedSamplersEXT = gfx20_CmdBindDescriptorBufferEmbeddedSamplersEXT,
    .GetBufferOpaqueCaptureDescriptorDataEXT = gfx20_GetBufferOpaqueCaptureDescriptorDataEXT,
    .GetImageOpaqueCaptureDescriptorDataEXT = gfx20_GetImageOpaqueCaptureDescriptorDataEXT,
    .GetImageViewOpaqueCaptureDescriptorDataEXT = gfx20_GetImageViewOpaqueCaptureDescriptorDataEXT,
    .GetSamplerOpaqueCaptureDescriptorDataEXT = gfx20_GetSamplerOpaqueCaptureDescriptorDataEXT,
    .GetAccelerationStructureOpaqueCaptureDescriptorDataEXT = gfx20_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT,
    .SetDeviceMemoryPriorityEXT = gfx20_SetDeviceMemoryPriorityEXT,
    .WaitForPresent2KHR = gfx20_WaitForPresent2KHR,
    .WaitForPresentKHR = gfx20_WaitForPresentKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateBufferCollectionFUCHSIA = gfx20_CreateBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .CreateBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionBufferConstraintsFUCHSIA = gfx20_SetBufferCollectionBufferConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionBufferConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionImageConstraintsFUCHSIA = gfx20_SetBufferCollectionImageConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionImageConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .DestroyBufferCollectionFUCHSIA = gfx20_DestroyBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .DestroyBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetBufferCollectionPropertiesFUCHSIA = gfx20_GetBufferCollectionPropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetBufferCollectionPropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .CmdBeginRendering = gfx20_CmdBeginRendering,
    .CmdBeginRenderingKHR = gfx20_CmdBeginRenderingKHR,
    .CmdEndRendering = gfx20_CmdEndRendering,
    .CmdEndRendering2KHR = gfx20_CmdEndRendering2KHR,
    .CmdEndRendering2EXT = gfx20_CmdEndRendering2EXT,
    .CmdEndRenderingKHR = gfx20_CmdEndRenderingKHR,
    .GetDescriptorSetLayoutHostMappingInfoVALVE = gfx20_GetDescriptorSetLayoutHostMappingInfoVALVE,
    .GetDescriptorSetHostMappingVALVE = gfx20_GetDescriptorSetHostMappingVALVE,
    .CreateMicromapEXT = gfx20_CreateMicromapEXT,
    .CmdBuildMicromapsEXT = gfx20_CmdBuildMicromapsEXT,
    .BuildMicromapsEXT = gfx20_BuildMicromapsEXT,
    .DestroyMicromapEXT = gfx20_DestroyMicromapEXT,
    .CmdCopyMicromapEXT = gfx20_CmdCopyMicromapEXT,
    .CopyMicromapEXT = gfx20_CopyMicromapEXT,
    .CmdCopyMicromapToMemoryEXT = gfx20_CmdCopyMicromapToMemoryEXT,
    .CopyMicromapToMemoryEXT = gfx20_CopyMicromapToMemoryEXT,
    .CmdCopyMemoryToMicromapEXT = gfx20_CmdCopyMemoryToMicromapEXT,
    .CopyMemoryToMicromapEXT = gfx20_CopyMemoryToMicromapEXT,
    .CmdWriteMicromapsPropertiesEXT = gfx20_CmdWriteMicromapsPropertiesEXT,
    .WriteMicromapsPropertiesEXT = gfx20_WriteMicromapsPropertiesEXT,
    .GetDeviceMicromapCompatibilityEXT = gfx20_GetDeviceMicromapCompatibilityEXT,
    .GetMicromapBuildSizesEXT = gfx20_GetMicromapBuildSizesEXT,
    .GetShaderModuleIdentifierEXT = gfx20_GetShaderModuleIdentifierEXT,
    .GetShaderModuleCreateInfoIdentifierEXT = gfx20_GetShaderModuleCreateInfoIdentifierEXT,
    .GetImageSubresourceLayout2 = gfx20_GetImageSubresourceLayout2,
    .GetImageSubresourceLayout2KHR = gfx20_GetImageSubresourceLayout2KHR,
    .GetImageSubresourceLayout2EXT = gfx20_GetImageSubresourceLayout2EXT,
    .GetPipelinePropertiesEXT = gfx20_GetPipelinePropertiesEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .ExportMetalObjectsEXT = gfx20_ExportMetalObjectsEXT,
#elif defined(_MSC_VER)
    .ExportMetalObjectsEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CmdBindTileMemoryQCOM = gfx20_CmdBindTileMemoryQCOM,
    .GetFramebufferTilePropertiesQCOM = gfx20_GetFramebufferTilePropertiesQCOM,
    .GetDynamicRenderingTilePropertiesQCOM = gfx20_GetDynamicRenderingTilePropertiesQCOM,
    .CreateOpticalFlowSessionNV = gfx20_CreateOpticalFlowSessionNV,
    .DestroyOpticalFlowSessionNV = gfx20_DestroyOpticalFlowSessionNV,
    .BindOpticalFlowSessionImageNV = gfx20_BindOpticalFlowSessionImageNV,
    .CmdOpticalFlowExecuteNV = gfx20_CmdOpticalFlowExecuteNV,
    .GetDeviceFaultInfoEXT = gfx20_GetDeviceFaultInfoEXT,
    .GetDeviceFaultReportsKHR = gfx20_GetDeviceFaultReportsKHR,
    .GetDeviceFaultDebugInfoKHR = gfx20_GetDeviceFaultDebugInfoKHR,
    .CmdSetDepthBias2EXT = gfx20_CmdSetDepthBias2EXT,
    .ReleaseSwapchainImagesKHR = gfx20_ReleaseSwapchainImagesKHR,
    .ReleaseSwapchainImagesEXT = gfx20_ReleaseSwapchainImagesEXT,
    .GetDeviceImageSubresourceLayout = gfx20_GetDeviceImageSubresourceLayout,
    .GetDeviceImageSubresourceLayoutKHR = gfx20_GetDeviceImageSubresourceLayoutKHR,
    .MapMemory2 = gfx20_MapMemory2,
    .MapMemory2KHR = gfx20_MapMemory2KHR,
    .UnmapMemory2 = gfx20_UnmapMemory2,
    .UnmapMemory2KHR = gfx20_UnmapMemory2KHR,
    .CreateShadersEXT = gfx20_CreateShadersEXT,
    .DestroyShaderEXT = gfx20_DestroyShaderEXT,
    .GetShaderBinaryDataEXT = gfx20_GetShaderBinaryDataEXT,
    .CmdBindShadersEXT = gfx20_CmdBindShadersEXT,
    .SetSwapchainPresentTimingQueueSizeEXT = gfx20_SetSwapchainPresentTimingQueueSizeEXT,
    .GetSwapchainTimingPropertiesEXT = gfx20_GetSwapchainTimingPropertiesEXT,
    .GetSwapchainTimeDomainPropertiesEXT = gfx20_GetSwapchainTimeDomainPropertiesEXT,
    .GetPastPresentationTimingEXT = gfx20_GetPastPresentationTimingEXT,
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetScreenBufferPropertiesQNX = gfx20_GetScreenBufferPropertiesQNX,
#elif defined(_MSC_VER)
    .GetScreenBufferPropertiesQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CmdBindDescriptorSets2 = gfx20_CmdBindDescriptorSets2,
    .CmdBindDescriptorSets2KHR = gfx20_CmdBindDescriptorSets2KHR,
    .CmdPushConstants2 = gfx20_CmdPushConstants2,
    .CmdPushConstants2KHR = gfx20_CmdPushConstants2KHR,
    .CmdPushDescriptorSet2 = gfx20_CmdPushDescriptorSet2,
    .CmdPushDescriptorSet2KHR = gfx20_CmdPushDescriptorSet2KHR,
    .CmdPushDescriptorSetWithTemplate2 = gfx20_CmdPushDescriptorSetWithTemplate2,
    .CmdPushDescriptorSetWithTemplate2KHR = gfx20_CmdPushDescriptorSetWithTemplate2KHR,
    .CmdSetDescriptorBufferOffsets2EXT = gfx20_CmdSetDescriptorBufferOffsets2EXT,
    .CmdBindDescriptorBufferEmbeddedSamplers2EXT = gfx20_CmdBindDescriptorBufferEmbeddedSamplers2EXT,
    .SetLatencySleepModeNV = gfx20_SetLatencySleepModeNV,
    .LatencySleepNV = gfx20_LatencySleepNV,
    .SetLatencyMarkerNV = gfx20_SetLatencyMarkerNV,
    .GetLatencyTimingsNV = gfx20_GetLatencyTimingsNV,
    .QueueNotifyOutOfBandNV = gfx20_QueueNotifyOutOfBandNV,
    .CmdSetRenderingAttachmentLocations = gfx20_CmdSetRenderingAttachmentLocations,
    .CmdSetRenderingAttachmentLocationsKHR = gfx20_CmdSetRenderingAttachmentLocationsKHR,
    .CmdSetRenderingInputAttachmentIndices = gfx20_CmdSetRenderingInputAttachmentIndices,
    .CmdSetRenderingInputAttachmentIndicesKHR = gfx20_CmdSetRenderingInputAttachmentIndicesKHR,
    .CmdSetDepthClampRangeEXT = gfx20_CmdSetDepthClampRangeEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandleEXT = gfx20_GetMemoryMetalHandleEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandleEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandlePropertiesEXT = gfx20_GetMemoryMetalHandlePropertiesEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandlePropertiesEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .ConvertCooperativeVectorMatrixNV = gfx20_ConvertCooperativeVectorMatrixNV,
    .CmdConvertCooperativeVectorMatrixNV = gfx20_CmdConvertCooperativeVectorMatrixNV,
    .CmdDispatchTileQCOM = gfx20_CmdDispatchTileQCOM,
    .CmdBeginPerTileExecutionQCOM = gfx20_CmdBeginPerTileExecutionQCOM,
    .CmdEndPerTileExecutionQCOM = gfx20_CmdEndPerTileExecutionQCOM,
    .CreateExternalComputeQueueNV = gfx20_CreateExternalComputeQueueNV,
    .DestroyExternalComputeQueueNV = gfx20_DestroyExternalComputeQueueNV,
    .CreateShaderInstrumentationARM = gfx20_CreateShaderInstrumentationARM,
    .DestroyShaderInstrumentationARM = gfx20_DestroyShaderInstrumentationARM,
    .CmdBeginShaderInstrumentationARM = gfx20_CmdBeginShaderInstrumentationARM,
    .CmdEndShaderInstrumentationARM = gfx20_CmdEndShaderInstrumentationARM,
    .GetShaderInstrumentationValuesARM = gfx20_GetShaderInstrumentationValuesARM,
    .ClearShaderInstrumentationMetricsARM = gfx20_ClearShaderInstrumentationMetricsARM,
    .CreateTensorARM = gfx20_CreateTensorARM,
    .DestroyTensorARM = gfx20_DestroyTensorARM,
    .CreateTensorViewARM = gfx20_CreateTensorViewARM,
    .DestroyTensorViewARM = gfx20_DestroyTensorViewARM,
    .GetTensorMemoryRequirementsARM = gfx20_GetTensorMemoryRequirementsARM,
    .BindTensorMemoryARM = gfx20_BindTensorMemoryARM,
    .GetDeviceTensorMemoryRequirementsARM = gfx20_GetDeviceTensorMemoryRequirementsARM,
    .CmdCopyTensorARM = gfx20_CmdCopyTensorARM,
    .GetTensorOpaqueCaptureDescriptorDataARM = gfx20_GetTensorOpaqueCaptureDescriptorDataARM,
    .GetTensorViewOpaqueCaptureDescriptorDataARM = gfx20_GetTensorViewOpaqueCaptureDescriptorDataARM,
    .CreateDataGraphPipelinesARM = gfx20_CreateDataGraphPipelinesARM,
    .CreateDataGraphPipelineSessionARM = gfx20_CreateDataGraphPipelineSessionARM,
    .GetDataGraphPipelineSessionBindPointRequirementsARM = gfx20_GetDataGraphPipelineSessionBindPointRequirementsARM,
    .GetDataGraphPipelineSessionMemoryRequirementsARM = gfx20_GetDataGraphPipelineSessionMemoryRequirementsARM,
    .BindDataGraphPipelineSessionMemoryARM = gfx20_BindDataGraphPipelineSessionMemoryARM,
    .DestroyDataGraphPipelineSessionARM = gfx20_DestroyDataGraphPipelineSessionARM,
    .CmdDispatchDataGraphARM = gfx20_CmdDispatchDataGraphARM,
    .GetDataGraphPipelineAvailablePropertiesARM = gfx20_GetDataGraphPipelineAvailablePropertiesARM,
    .GetDataGraphPipelinePropertiesARM = gfx20_GetDataGraphPipelinePropertiesARM,
#ifdef VK_USE_PLATFORM_OHOS
    .GetNativeBufferPropertiesOHOS = gfx20_GetNativeBufferPropertiesOHOS,
#elif defined(_MSC_VER)
    .GetNativeBufferPropertiesOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
    .GetMemoryNativeBufferOHOS = gfx20_GetMemoryNativeBufferOHOS,
#elif defined(_MSC_VER)
    .GetMemoryNativeBufferOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .QueueSetPerfHintQCOM = gfx20_QueueSetPerfHintQCOM,
    .CmdSetComputeOccupancyPriorityNV = gfx20_CmdSetComputeOccupancyPriorityNV,
    .WriteSamplerDescriptorsEXT = gfx20_WriteSamplerDescriptorsEXT,
    .WriteResourceDescriptorsEXT = gfx20_WriteResourceDescriptorsEXT,
    .CmdBindSamplerHeapEXT = gfx20_CmdBindSamplerHeapEXT,
    .CmdBindResourceHeapEXT = gfx20_CmdBindResourceHeapEXT,
    .CmdPushDataEXT = gfx20_CmdPushDataEXT,
    .RegisterCustomBorderColorEXT = gfx20_RegisterCustomBorderColorEXT,
    .UnregisterCustomBorderColorEXT = gfx20_UnregisterCustomBorderColorEXT,
    .GetImageOpaqueCaptureDataEXT = gfx20_GetImageOpaqueCaptureDataEXT,
    .GetTensorOpaqueCaptureDataARM = gfx20_GetTensorOpaqueCaptureDataARM,
    .CmdCopyMemoryKHR = gfx20_CmdCopyMemoryKHR,
    .CmdCopyMemoryToImageKHR = gfx20_CmdCopyMemoryToImageKHR,
    .CmdCopyImageToMemoryKHR = gfx20_CmdCopyImageToMemoryKHR,
    .CmdUpdateMemoryKHR = gfx20_CmdUpdateMemoryKHR,
    .CmdFillMemoryKHR = gfx20_CmdFillMemoryKHR,
    .CmdCopyQueryPoolResultsToMemoryKHR = gfx20_CmdCopyQueryPoolResultsToMemoryKHR,
    .CmdBeginConditionalRendering2EXT = gfx20_CmdBeginConditionalRendering2EXT,
    .CmdBindTransformFeedbackBuffers2EXT = gfx20_CmdBindTransformFeedbackBuffers2EXT,
    .CmdBeginTransformFeedback2EXT = gfx20_CmdBeginTransformFeedback2EXT,
    .CmdEndTransformFeedback2EXT = gfx20_CmdEndTransformFeedback2EXT,
    .CmdDrawIndirectByteCount2EXT = gfx20_CmdDrawIndirectByteCount2EXT,
    .CmdWriteMarkerToMemoryAMD = gfx20_CmdWriteMarkerToMemoryAMD,
    .CmdBindIndexBuffer3KHR = gfx20_CmdBindIndexBuffer3KHR,
    .CmdBindVertexBuffers3KHR = gfx20_CmdBindVertexBuffers3KHR,
    .CmdDrawIndirect2KHR = gfx20_CmdDrawIndirect2KHR,
    .CmdDrawIndexedIndirect2KHR = gfx20_CmdDrawIndexedIndirect2KHR,
    .CmdDrawIndirectCount2KHR = gfx20_CmdDrawIndirectCount2KHR,
    .CmdDrawIndexedIndirectCount2KHR = gfx20_CmdDrawIndexedIndirectCount2KHR,
    .CmdDrawMeshTasksIndirect2EXT = gfx20_CmdDrawMeshTasksIndirect2EXT,
    .CmdDrawMeshTasksIndirectCount2EXT = gfx20_CmdDrawMeshTasksIndirectCount2EXT,
    .CmdDispatchIndirect2KHR = gfx20_CmdDispatchIndirect2KHR,
    .CreateAccelerationStructure2KHR = gfx20_CreateAccelerationStructure2KHR,
    .CmdSetDispatchParametersARM = gfx20_CmdSetDispatchParametersARM,
};
const struct vk_device_entrypoint_table gfx30_device_entrypoints = {
    .GetDeviceProcAddr = gfx30_GetDeviceProcAddr,
    .DestroyDevice = gfx30_DestroyDevice,
    .GetDeviceQueue = gfx30_GetDeviceQueue,
    .QueueSubmit = gfx30_QueueSubmit,
    .QueueWaitIdle = gfx30_QueueWaitIdle,
    .DeviceWaitIdle = gfx30_DeviceWaitIdle,
    .AllocateMemory = gfx30_AllocateMemory,
    .FreeMemory = gfx30_FreeMemory,
    .MapMemory = gfx30_MapMemory,
    .UnmapMemory = gfx30_UnmapMemory,
    .FlushMappedMemoryRanges = gfx30_FlushMappedMemoryRanges,
    .InvalidateMappedMemoryRanges = gfx30_InvalidateMappedMemoryRanges,
    .GetDeviceMemoryCommitment = gfx30_GetDeviceMemoryCommitment,
    .GetBufferMemoryRequirements = gfx30_GetBufferMemoryRequirements,
    .BindBufferMemory = gfx30_BindBufferMemory,
    .GetImageMemoryRequirements = gfx30_GetImageMemoryRequirements,
    .BindImageMemory = gfx30_BindImageMemory,
    .GetImageSparseMemoryRequirements = gfx30_GetImageSparseMemoryRequirements,
    .QueueBindSparse = gfx30_QueueBindSparse,
    .CreateFence = gfx30_CreateFence,
    .DestroyFence = gfx30_DestroyFence,
    .ResetFences = gfx30_ResetFences,
    .GetFenceStatus = gfx30_GetFenceStatus,
    .WaitForFences = gfx30_WaitForFences,
    .CreateSemaphore = gfx30_CreateSemaphore,
    .DestroySemaphore = gfx30_DestroySemaphore,
    .CreateEvent = gfx30_CreateEvent,
    .DestroyEvent = gfx30_DestroyEvent,
    .GetEventStatus = gfx30_GetEventStatus,
    .SetEvent = gfx30_SetEvent,
    .ResetEvent = gfx30_ResetEvent,
    .CreateQueryPool = gfx30_CreateQueryPool,
    .DestroyQueryPool = gfx30_DestroyQueryPool,
    .GetQueryPoolResults = gfx30_GetQueryPoolResults,
    .ResetQueryPool = gfx30_ResetQueryPool,
    .ResetQueryPoolEXT = gfx30_ResetQueryPoolEXT,
    .CreateBuffer = gfx30_CreateBuffer,
    .DestroyBuffer = gfx30_DestroyBuffer,
    .CreateBufferView = gfx30_CreateBufferView,
    .DestroyBufferView = gfx30_DestroyBufferView,
    .CreateImage = gfx30_CreateImage,
    .DestroyImage = gfx30_DestroyImage,
    .GetImageSubresourceLayout = gfx30_GetImageSubresourceLayout,
    .CreateImageView = gfx30_CreateImageView,
    .DestroyImageView = gfx30_DestroyImageView,
    .CreateShaderModule = gfx30_CreateShaderModule,
    .DestroyShaderModule = gfx30_DestroyShaderModule,
    .CreatePipelineCache = gfx30_CreatePipelineCache,
    .DestroyPipelineCache = gfx30_DestroyPipelineCache,
    .GetPipelineCacheData = gfx30_GetPipelineCacheData,
    .MergePipelineCaches = gfx30_MergePipelineCaches,
    .CreatePipelineBinariesKHR = gfx30_CreatePipelineBinariesKHR,
    .DestroyPipelineBinaryKHR = gfx30_DestroyPipelineBinaryKHR,
    .GetPipelineKeyKHR = gfx30_GetPipelineKeyKHR,
    .GetPipelineBinaryDataKHR = gfx30_GetPipelineBinaryDataKHR,
    .ReleaseCapturedPipelineDataKHR = gfx30_ReleaseCapturedPipelineDataKHR,
    .CreateGraphicsPipelines = gfx30_CreateGraphicsPipelines,
    .CreateComputePipelines = gfx30_CreateComputePipelines,
    .GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI = gfx30_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI,
    .DestroyPipeline = gfx30_DestroyPipeline,
    .CreatePipelineLayout = gfx30_CreatePipelineLayout,
    .DestroyPipelineLayout = gfx30_DestroyPipelineLayout,
    .CreateSampler = gfx30_CreateSampler,
    .DestroySampler = gfx30_DestroySampler,
    .CreateDescriptorSetLayout = gfx30_CreateDescriptorSetLayout,
    .DestroyDescriptorSetLayout = gfx30_DestroyDescriptorSetLayout,
    .CreateDescriptorPool = gfx30_CreateDescriptorPool,
    .DestroyDescriptorPool = gfx30_DestroyDescriptorPool,
    .ResetDescriptorPool = gfx30_ResetDescriptorPool,
    .AllocateDescriptorSets = gfx30_AllocateDescriptorSets,
    .FreeDescriptorSets = gfx30_FreeDescriptorSets,
    .UpdateDescriptorSets = gfx30_UpdateDescriptorSets,
    .CreateFramebuffer = gfx30_CreateFramebuffer,
    .DestroyFramebuffer = gfx30_DestroyFramebuffer,
    .CreateRenderPass = gfx30_CreateRenderPass,
    .DestroyRenderPass = gfx30_DestroyRenderPass,
    .GetRenderAreaGranularity = gfx30_GetRenderAreaGranularity,
    .GetRenderingAreaGranularity = gfx30_GetRenderingAreaGranularity,
    .GetRenderingAreaGranularityKHR = gfx30_GetRenderingAreaGranularityKHR,
    .CreateCommandPool = gfx30_CreateCommandPool,
    .DestroyCommandPool = gfx30_DestroyCommandPool,
    .ResetCommandPool = gfx30_ResetCommandPool,
    .AllocateCommandBuffers = gfx30_AllocateCommandBuffers,
    .FreeCommandBuffers = gfx30_FreeCommandBuffers,
    .BeginCommandBuffer = gfx30_BeginCommandBuffer,
    .EndCommandBuffer = gfx30_EndCommandBuffer,
    .ResetCommandBuffer = gfx30_ResetCommandBuffer,
    .CmdBindPipeline = gfx30_CmdBindPipeline,
    .CmdSetPrimitiveRestartIndexEXT = gfx30_CmdSetPrimitiveRestartIndexEXT,
    .CmdSetAttachmentFeedbackLoopEnableEXT = gfx30_CmdSetAttachmentFeedbackLoopEnableEXT,
    .CmdSetViewport = gfx30_CmdSetViewport,
    .CmdSetScissor = gfx30_CmdSetScissor,
    .CmdSetLineWidth = gfx30_CmdSetLineWidth,
    .CmdSetDepthBias = gfx30_CmdSetDepthBias,
    .CmdSetBlendConstants = gfx30_CmdSetBlendConstants,
    .CmdSetDepthBounds = gfx30_CmdSetDepthBounds,
    .CmdSetStencilCompareMask = gfx30_CmdSetStencilCompareMask,
    .CmdSetStencilWriteMask = gfx30_CmdSetStencilWriteMask,
    .CmdSetStencilReference = gfx30_CmdSetStencilReference,
    .CmdBindDescriptorSets = gfx30_CmdBindDescriptorSets,
    .CmdBindIndexBuffer = gfx30_CmdBindIndexBuffer,
    .CmdBindVertexBuffers = gfx30_CmdBindVertexBuffers,
    .CmdDraw = gfx30_CmdDraw,
    .CmdDrawIndexed = gfx30_CmdDrawIndexed,
    .CmdDrawMultiEXT = gfx30_CmdDrawMultiEXT,
    .CmdDrawMultiIndexedEXT = gfx30_CmdDrawMultiIndexedEXT,
    .CmdDrawIndirect = gfx30_CmdDrawIndirect,
    .CmdDrawIndexedIndirect = gfx30_CmdDrawIndexedIndirect,
    .CmdDispatch = gfx30_CmdDispatch,
    .CmdDispatchIndirect = gfx30_CmdDispatchIndirect,
    .CmdSubpassShadingHUAWEI = gfx30_CmdSubpassShadingHUAWEI,
    .CmdDrawClusterHUAWEI = gfx30_CmdDrawClusterHUAWEI,
    .CmdDrawClusterIndirectHUAWEI = gfx30_CmdDrawClusterIndirectHUAWEI,
    .CmdUpdatePipelineIndirectBufferNV = gfx30_CmdUpdatePipelineIndirectBufferNV,
    .CmdCopyBuffer = gfx30_CmdCopyBuffer,
    .CmdCopyImage = gfx30_CmdCopyImage,
    .CmdBlitImage = gfx30_CmdBlitImage,
    .CmdCopyBufferToImage = gfx30_CmdCopyBufferToImage,
    .CmdCopyImageToBuffer = gfx30_CmdCopyImageToBuffer,
    .CmdCopyMemoryIndirectNV = gfx30_CmdCopyMemoryIndirectNV,
    .CmdCopyMemoryIndirectKHR = gfx30_CmdCopyMemoryIndirectKHR,
    .CmdCopyMemoryToImageIndirectNV = gfx30_CmdCopyMemoryToImageIndirectNV,
    .CmdCopyMemoryToImageIndirectKHR = gfx30_CmdCopyMemoryToImageIndirectKHR,
    .CmdUpdateBuffer = gfx30_CmdUpdateBuffer,
    .CmdFillBuffer = gfx30_CmdFillBuffer,
    .CmdClearColorImage = gfx30_CmdClearColorImage,
    .CmdClearDepthStencilImage = gfx30_CmdClearDepthStencilImage,
    .CmdClearAttachments = gfx30_CmdClearAttachments,
    .CmdResolveImage = gfx30_CmdResolveImage,
    .CmdSetEvent = gfx30_CmdSetEvent,
    .CmdResetEvent = gfx30_CmdResetEvent,
    .CmdWaitEvents = gfx30_CmdWaitEvents,
    .CmdPipelineBarrier = gfx30_CmdPipelineBarrier,
    .CmdBeginQuery = gfx30_CmdBeginQuery,
    .CmdEndQuery = gfx30_CmdEndQuery,
    .CmdBeginConditionalRenderingEXT = gfx30_CmdBeginConditionalRenderingEXT,
    .CmdEndConditionalRenderingEXT = gfx30_CmdEndConditionalRenderingEXT,
    .CmdBeginCustomResolveEXT = gfx30_CmdBeginCustomResolveEXT,
    .CmdResetQueryPool = gfx30_CmdResetQueryPool,
    .CmdWriteTimestamp = gfx30_CmdWriteTimestamp,
    .CmdCopyQueryPoolResults = gfx30_CmdCopyQueryPoolResults,
    .CmdPushConstants = gfx30_CmdPushConstants,
    .CmdBeginRenderPass = gfx30_CmdBeginRenderPass,
    .CmdNextSubpass = gfx30_CmdNextSubpass,
    .CmdEndRenderPass = gfx30_CmdEndRenderPass,
    .CmdExecuteCommands = gfx30_CmdExecuteCommands,
    .CreateSharedSwapchainsKHR = gfx30_CreateSharedSwapchainsKHR,
    .CreateSwapchainKHR = gfx30_CreateSwapchainKHR,
    .DestroySwapchainKHR = gfx30_DestroySwapchainKHR,
    .GetSwapchainImagesKHR = gfx30_GetSwapchainImagesKHR,
    .AcquireNextImageKHR = gfx30_AcquireNextImageKHR,
    .QueuePresentKHR = gfx30_QueuePresentKHR,
    .DebugMarkerSetObjectNameEXT = gfx30_DebugMarkerSetObjectNameEXT,
    .DebugMarkerSetObjectTagEXT = gfx30_DebugMarkerSetObjectTagEXT,
    .CmdDebugMarkerBeginEXT = gfx30_CmdDebugMarkerBeginEXT,
    .CmdDebugMarkerEndEXT = gfx30_CmdDebugMarkerEndEXT,
    .CmdDebugMarkerInsertEXT = gfx30_CmdDebugMarkerInsertEXT,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleNV = gfx30_GetMemoryWin32HandleNV,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .CmdExecuteGeneratedCommandsNV = gfx30_CmdExecuteGeneratedCommandsNV,
    .CmdPreprocessGeneratedCommandsNV = gfx30_CmdPreprocessGeneratedCommandsNV,
    .CmdBindPipelineShaderGroupNV = gfx30_CmdBindPipelineShaderGroupNV,
    .GetGeneratedCommandsMemoryRequirementsNV = gfx30_GetGeneratedCommandsMemoryRequirementsNV,
    .CreateIndirectCommandsLayoutNV = gfx30_CreateIndirectCommandsLayoutNV,
    .DestroyIndirectCommandsLayoutNV = gfx30_DestroyIndirectCommandsLayoutNV,
    .CmdExecuteGeneratedCommandsEXT = gfx30_CmdExecuteGeneratedCommandsEXT,
    .CmdPreprocessGeneratedCommandsEXT = gfx30_CmdPreprocessGeneratedCommandsEXT,
    .GetGeneratedCommandsMemoryRequirementsEXT = gfx30_GetGeneratedCommandsMemoryRequirementsEXT,
    .CreateIndirectCommandsLayoutEXT = gfx30_CreateIndirectCommandsLayoutEXT,
    .DestroyIndirectCommandsLayoutEXT = gfx30_DestroyIndirectCommandsLayoutEXT,
    .CreateIndirectExecutionSetEXT = gfx30_CreateIndirectExecutionSetEXT,
    .DestroyIndirectExecutionSetEXT = gfx30_DestroyIndirectExecutionSetEXT,
    .UpdateIndirectExecutionSetPipelineEXT = gfx30_UpdateIndirectExecutionSetPipelineEXT,
    .UpdateIndirectExecutionSetShaderEXT = gfx30_UpdateIndirectExecutionSetShaderEXT,
    .CmdPushDescriptorSet = gfx30_CmdPushDescriptorSet,
    .CmdPushDescriptorSetKHR = gfx30_CmdPushDescriptorSetKHR,
    .TrimCommandPool = gfx30_TrimCommandPool,
    .TrimCommandPoolKHR = gfx30_TrimCommandPoolKHR,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleKHR = gfx30_GetMemoryWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandlePropertiesKHR = gfx30_GetMemoryWin32HandlePropertiesKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandlePropertiesKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryFdKHR = gfx30_GetMemoryFdKHR,
    .GetMemoryFdPropertiesKHR = gfx30_GetMemoryFdPropertiesKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandleFUCHSIA = gfx30_GetMemoryZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandlePropertiesFUCHSIA = gfx30_GetMemoryZirconHandlePropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandlePropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .GetMemoryRemoteAddressNV = gfx30_GetMemoryRemoteAddressNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreWin32HandleKHR = gfx30_GetSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportSemaphoreWin32HandleKHR = gfx30_ImportSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreFdKHR = gfx30_GetSemaphoreFdKHR,
    .ImportSemaphoreFdKHR = gfx30_ImportSemaphoreFdKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetSemaphoreZirconHandleFUCHSIA = gfx30_GetSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .ImportSemaphoreZirconHandleFUCHSIA = gfx30_ImportSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .ImportSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetFenceWin32HandleKHR = gfx30_GetFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportFenceWin32HandleKHR = gfx30_ImportFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetFenceFdKHR = gfx30_GetFenceFdKHR,
    .ImportFenceFdKHR = gfx30_ImportFenceFdKHR,
    .DisplayPowerControlEXT = gfx30_DisplayPowerControlEXT,
    .RegisterDeviceEventEXT = gfx30_RegisterDeviceEventEXT,
    .RegisterDisplayEventEXT = gfx30_RegisterDisplayEventEXT,
    .GetSwapchainCounterEXT = gfx30_GetSwapchainCounterEXT,
    .GetDeviceGroupPeerMemoryFeatures = gfx30_GetDeviceGroupPeerMemoryFeatures,
    .GetDeviceGroupPeerMemoryFeaturesKHR = gfx30_GetDeviceGroupPeerMemoryFeaturesKHR,
    .BindBufferMemory2 = gfx30_BindBufferMemory2,
    .BindBufferMemory2KHR = gfx30_BindBufferMemory2KHR,
    .BindImageMemory2 = gfx30_BindImageMemory2,
    .BindImageMemory2KHR = gfx30_BindImageMemory2KHR,
    .CmdSetDeviceMask = gfx30_CmdSetDeviceMask,
    .CmdSetDeviceMaskKHR = gfx30_CmdSetDeviceMaskKHR,
    .GetDeviceGroupPresentCapabilitiesKHR = gfx30_GetDeviceGroupPresentCapabilitiesKHR,
    .GetDeviceGroupSurfacePresentModesKHR = gfx30_GetDeviceGroupSurfacePresentModesKHR,
    .AcquireNextImage2KHR = gfx30_AcquireNextImage2KHR,
    .CmdDispatchBase = gfx30_CmdDispatchBase,
    .CmdDispatchBaseKHR = gfx30_CmdDispatchBaseKHR,
    .CreateDescriptorUpdateTemplate = gfx30_CreateDescriptorUpdateTemplate,
    .CreateDescriptorUpdateTemplateKHR = gfx30_CreateDescriptorUpdateTemplateKHR,
    .DestroyDescriptorUpdateTemplate = gfx30_DestroyDescriptorUpdateTemplate,
    .DestroyDescriptorUpdateTemplateKHR = gfx30_DestroyDescriptorUpdateTemplateKHR,
    .UpdateDescriptorSetWithTemplate = gfx30_UpdateDescriptorSetWithTemplate,
    .UpdateDescriptorSetWithTemplateKHR = gfx30_UpdateDescriptorSetWithTemplateKHR,
    .CmdPushDescriptorSetWithTemplate = gfx30_CmdPushDescriptorSetWithTemplate,
    .CmdPushDescriptorSetWithTemplateKHR = gfx30_CmdPushDescriptorSetWithTemplateKHR,
    .SetHdrMetadataEXT = gfx30_SetHdrMetadataEXT,
    .GetSwapchainStatusKHR = gfx30_GetSwapchainStatusKHR,
    .GetRefreshCycleDurationGOOGLE = gfx30_GetRefreshCycleDurationGOOGLE,
    .GetPastPresentationTimingGOOGLE = gfx30_GetPastPresentationTimingGOOGLE,
    .CmdSetViewportWScalingNV = gfx30_CmdSetViewportWScalingNV,
    .CmdSetDiscardRectangleEXT = gfx30_CmdSetDiscardRectangleEXT,
    .CmdSetDiscardRectangleEnableEXT = gfx30_CmdSetDiscardRectangleEnableEXT,
    .CmdSetDiscardRectangleModeEXT = gfx30_CmdSetDiscardRectangleModeEXT,
    .CmdSetSampleLocationsEXT = gfx30_CmdSetSampleLocationsEXT,
    .GetBufferMemoryRequirements2 = gfx30_GetBufferMemoryRequirements2,
    .GetBufferMemoryRequirements2KHR = gfx30_GetBufferMemoryRequirements2KHR,
    .GetImageMemoryRequirements2 = gfx30_GetImageMemoryRequirements2,
    .GetImageMemoryRequirements2KHR = gfx30_GetImageMemoryRequirements2KHR,
    .GetImageSparseMemoryRequirements2 = gfx30_GetImageSparseMemoryRequirements2,
    .GetImageSparseMemoryRequirements2KHR = gfx30_GetImageSparseMemoryRequirements2KHR,
    .GetDeviceBufferMemoryRequirements = gfx30_GetDeviceBufferMemoryRequirements,
    .GetDeviceBufferMemoryRequirementsKHR = gfx30_GetDeviceBufferMemoryRequirementsKHR,
    .GetDeviceImageMemoryRequirements = gfx30_GetDeviceImageMemoryRequirements,
    .GetDeviceImageMemoryRequirementsKHR = gfx30_GetDeviceImageMemoryRequirementsKHR,
    .GetDeviceImageSparseMemoryRequirements = gfx30_GetDeviceImageSparseMemoryRequirements,
    .GetDeviceImageSparseMemoryRequirementsKHR = gfx30_GetDeviceImageSparseMemoryRequirementsKHR,
    .CreateSamplerYcbcrConversion = gfx30_CreateSamplerYcbcrConversion,
    .CreateSamplerYcbcrConversionKHR = gfx30_CreateSamplerYcbcrConversionKHR,
    .DestroySamplerYcbcrConversion = gfx30_DestroySamplerYcbcrConversion,
    .DestroySamplerYcbcrConversionKHR = gfx30_DestroySamplerYcbcrConversionKHR,
    .GetDeviceQueue2 = gfx30_GetDeviceQueue2,
    .CreateValidationCacheEXT = gfx30_CreateValidationCacheEXT,
    .DestroyValidationCacheEXT = gfx30_DestroyValidationCacheEXT,
    .GetValidationCacheDataEXT = gfx30_GetValidationCacheDataEXT,
    .MergeValidationCachesEXT = gfx30_MergeValidationCachesEXT,
    .GetDescriptorSetLayoutSupport = gfx30_GetDescriptorSetLayoutSupport,
    .GetDescriptorSetLayoutSupportKHR = gfx30_GetDescriptorSetLayoutSupportKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsageANDROID = gfx30_GetSwapchainGrallocUsageANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsage2ANDROID = gfx30_GetSwapchainGrallocUsage2ANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsage2ANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .AcquireImageANDROID = gfx30_AcquireImageANDROID,
#elif defined(_MSC_VER)
    .AcquireImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .QueueSignalReleaseImageANDROID = gfx30_QueueSignalReleaseImageANDROID,
#elif defined(_MSC_VER)
    .QueueSignalReleaseImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .GetShaderInfoAMD = gfx30_GetShaderInfoAMD,
    .SetLocalDimmingAMD = gfx30_SetLocalDimmingAMD,
    .GetCalibratedTimestampsKHR = gfx30_GetCalibratedTimestampsKHR,
    .GetCalibratedTimestampsEXT = gfx30_GetCalibratedTimestampsEXT,
    .SetDebugUtilsObjectNameEXT = gfx30_SetDebugUtilsObjectNameEXT,
    .SetDebugUtilsObjectTagEXT = gfx30_SetDebugUtilsObjectTagEXT,
    .QueueBeginDebugUtilsLabelEXT = gfx30_QueueBeginDebugUtilsLabelEXT,
    .QueueEndDebugUtilsLabelEXT = gfx30_QueueEndDebugUtilsLabelEXT,
    .QueueInsertDebugUtilsLabelEXT = gfx30_QueueInsertDebugUtilsLabelEXT,
    .CmdBeginDebugUtilsLabelEXT = gfx30_CmdBeginDebugUtilsLabelEXT,
    .CmdEndDebugUtilsLabelEXT = gfx30_CmdEndDebugUtilsLabelEXT,
    .CmdInsertDebugUtilsLabelEXT = gfx30_CmdInsertDebugUtilsLabelEXT,
    .GetMemoryHostPointerPropertiesEXT = gfx30_GetMemoryHostPointerPropertiesEXT,
    .CmdWriteBufferMarkerAMD = gfx30_CmdWriteBufferMarkerAMD,
    .CreateRenderPass2 = gfx30_CreateRenderPass2,
    .CreateRenderPass2KHR = gfx30_CreateRenderPass2KHR,
    .CmdBeginRenderPass2 = gfx30_CmdBeginRenderPass2,
    .CmdBeginRenderPass2KHR = gfx30_CmdBeginRenderPass2KHR,
    .CmdNextSubpass2 = gfx30_CmdNextSubpass2,
    .CmdNextSubpass2KHR = gfx30_CmdNextSubpass2KHR,
    .CmdEndRenderPass2 = gfx30_CmdEndRenderPass2,
    .CmdEndRenderPass2KHR = gfx30_CmdEndRenderPass2KHR,
    .GetSemaphoreCounterValue = gfx30_GetSemaphoreCounterValue,
    .GetSemaphoreCounterValueKHR = gfx30_GetSemaphoreCounterValueKHR,
    .WaitSemaphores = gfx30_WaitSemaphores,
    .WaitSemaphoresKHR = gfx30_WaitSemaphoresKHR,
    .SignalSemaphore = gfx30_SignalSemaphore,
    .SignalSemaphoreKHR = gfx30_SignalSemaphoreKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetAndroidHardwareBufferPropertiesANDROID = gfx30_GetAndroidHardwareBufferPropertiesANDROID,
#elif defined(_MSC_VER)
    .GetAndroidHardwareBufferPropertiesANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetMemoryAndroidHardwareBufferANDROID = gfx30_GetMemoryAndroidHardwareBufferANDROID,
#elif defined(_MSC_VER)
    .GetMemoryAndroidHardwareBufferANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .CmdDrawIndirectCount = gfx30_CmdDrawIndirectCount,
    .CmdDrawIndirectCountKHR = gfx30_CmdDrawIndirectCountKHR,
    .CmdDrawIndirectCountAMD = gfx30_CmdDrawIndirectCountAMD,
    .CmdDrawIndexedIndirectCount = gfx30_CmdDrawIndexedIndirectCount,
    .CmdDrawIndexedIndirectCountKHR = gfx30_CmdDrawIndexedIndirectCountKHR,
    .CmdDrawIndexedIndirectCountAMD = gfx30_CmdDrawIndexedIndirectCountAMD,
    .CmdSetCheckpointNV = gfx30_CmdSetCheckpointNV,
    .GetQueueCheckpointDataNV = gfx30_GetQueueCheckpointDataNV,
    .CmdBindTransformFeedbackBuffersEXT = gfx30_CmdBindTransformFeedbackBuffersEXT,
    .CmdBeginTransformFeedbackEXT = gfx30_CmdBeginTransformFeedbackEXT,
    .CmdEndTransformFeedbackEXT = gfx30_CmdEndTransformFeedbackEXT,
    .CmdBeginQueryIndexedEXT = gfx30_CmdBeginQueryIndexedEXT,
    .CmdEndQueryIndexedEXT = gfx30_CmdEndQueryIndexedEXT,
    .CmdDrawIndirectByteCountEXT = gfx30_CmdDrawIndirectByteCountEXT,
    .CmdSetExclusiveScissorNV = gfx30_CmdSetExclusiveScissorNV,
    .CmdSetExclusiveScissorEnableNV = gfx30_CmdSetExclusiveScissorEnableNV,
    .CmdBindShadingRateImageNV = gfx30_CmdBindShadingRateImageNV,
    .CmdSetViewportShadingRatePaletteNV = gfx30_CmdSetViewportShadingRatePaletteNV,
    .CmdSetCoarseSampleOrderNV = gfx30_CmdSetCoarseSampleOrderNV,
    .CmdDrawMeshTasksNV = gfx30_CmdDrawMeshTasksNV,
    .CmdDrawMeshTasksIndirectNV = gfx30_CmdDrawMeshTasksIndirectNV,
    .CmdDrawMeshTasksIndirectCountNV = gfx30_CmdDrawMeshTasksIndirectCountNV,
    .CmdDrawMeshTasksEXT = gfx30_CmdDrawMeshTasksEXT,
    .CmdDrawMeshTasksIndirectEXT = gfx30_CmdDrawMeshTasksIndirectEXT,
    .CmdDrawMeshTasksIndirectCountEXT = gfx30_CmdDrawMeshTasksIndirectCountEXT,
    .CompileDeferredNV = gfx30_CompileDeferredNV,
    .CreateAccelerationStructureNV = gfx30_CreateAccelerationStructureNV,
    .CmdBindInvocationMaskHUAWEI = gfx30_CmdBindInvocationMaskHUAWEI,
    .DestroyAccelerationStructureKHR = gfx30_DestroyAccelerationStructureKHR,
    .DestroyAccelerationStructureNV = gfx30_DestroyAccelerationStructureNV,
    .GetAccelerationStructureMemoryRequirementsNV = gfx30_GetAccelerationStructureMemoryRequirementsNV,
    .BindAccelerationStructureMemoryNV = gfx30_BindAccelerationStructureMemoryNV,
    .CmdCopyAccelerationStructureNV = gfx30_CmdCopyAccelerationStructureNV,
    .CmdCopyAccelerationStructureKHR = gfx30_CmdCopyAccelerationStructureKHR,
    .CopyAccelerationStructureKHR = gfx30_CopyAccelerationStructureKHR,
    .CmdCopyAccelerationStructureToMemoryKHR = gfx30_CmdCopyAccelerationStructureToMemoryKHR,
    .CopyAccelerationStructureToMemoryKHR = gfx30_CopyAccelerationStructureToMemoryKHR,
    .CmdCopyMemoryToAccelerationStructureKHR = gfx30_CmdCopyMemoryToAccelerationStructureKHR,
    .CopyMemoryToAccelerationStructureKHR = gfx30_CopyMemoryToAccelerationStructureKHR,
    .CmdWriteAccelerationStructuresPropertiesKHR = gfx30_CmdWriteAccelerationStructuresPropertiesKHR,
    .CmdWriteAccelerationStructuresPropertiesNV = gfx30_CmdWriteAccelerationStructuresPropertiesNV,
    .CmdBuildAccelerationStructureNV = gfx30_CmdBuildAccelerationStructureNV,
    .WriteAccelerationStructuresPropertiesKHR = gfx30_WriteAccelerationStructuresPropertiesKHR,
    .CmdTraceRaysKHR = gfx30_CmdTraceRaysKHR,
    .CmdTraceRaysNV = gfx30_CmdTraceRaysNV,
    .GetRayTracingShaderGroupHandlesKHR = gfx30_GetRayTracingShaderGroupHandlesKHR,
    .GetRayTracingShaderGroupHandlesNV = gfx30_GetRayTracingShaderGroupHandlesNV,
    .GetRayTracingCaptureReplayShaderGroupHandlesKHR = gfx30_GetRayTracingCaptureReplayShaderGroupHandlesKHR,
    .GetAccelerationStructureHandleNV = gfx30_GetAccelerationStructureHandleNV,
    .CreateRayTracingPipelinesNV = gfx30_CreateRayTracingPipelinesNV,
    .CreateRayTracingPipelinesKHR = gfx30_CreateRayTracingPipelinesKHR,
    .CmdTraceRaysIndirectKHR = gfx30_CmdTraceRaysIndirectKHR,
    .CmdTraceRaysIndirect2KHR = gfx30_CmdTraceRaysIndirect2KHR,
    .GetClusterAccelerationStructureBuildSizesNV = gfx30_GetClusterAccelerationStructureBuildSizesNV,
    .CmdBuildClusterAccelerationStructureIndirectNV = gfx30_CmdBuildClusterAccelerationStructureIndirectNV,
    .GetDeviceAccelerationStructureCompatibilityKHR = gfx30_GetDeviceAccelerationStructureCompatibilityKHR,
    .GetRayTracingShaderGroupStackSizeKHR = gfx30_GetRayTracingShaderGroupStackSizeKHR,
    .CmdSetRayTracingPipelineStackSizeKHR = gfx30_CmdSetRayTracingPipelineStackSizeKHR,
    .GetImageViewHandleNVX = gfx30_GetImageViewHandleNVX,
    .GetImageViewHandle64NVX = gfx30_GetImageViewHandle64NVX,
    .GetImageViewAddressNVX = gfx30_GetImageViewAddressNVX,
    .GetDeviceCombinedImageSamplerIndexNVX = gfx30_GetDeviceCombinedImageSamplerIndexNVX,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetDeviceGroupSurfacePresentModes2EXT = gfx30_GetDeviceGroupSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetDeviceGroupSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireFullScreenExclusiveModeEXT = gfx30_AcquireFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .AcquireFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ReleaseFullScreenExclusiveModeEXT = gfx30_ReleaseFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .ReleaseFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .AcquireProfilingLockKHR = gfx30_AcquireProfilingLockKHR,
    .ReleaseProfilingLockKHR = gfx30_ReleaseProfilingLockKHR,
    .GetImageDrmFormatModifierPropertiesEXT = gfx30_GetImageDrmFormatModifierPropertiesEXT,
    .GetBufferOpaqueCaptureAddress = gfx30_GetBufferOpaqueCaptureAddress,
    .GetBufferOpaqueCaptureAddressKHR = gfx30_GetBufferOpaqueCaptureAddressKHR,
    .GetBufferDeviceAddress = gfx30_GetBufferDeviceAddress,
    .GetBufferDeviceAddressKHR = gfx30_GetBufferDeviceAddressKHR,
    .GetBufferDeviceAddressEXT = gfx30_GetBufferDeviceAddressEXT,
    .InitializePerformanceApiINTEL = gfx30_InitializePerformanceApiINTEL,
    .UninitializePerformanceApiINTEL = gfx30_UninitializePerformanceApiINTEL,
    .CmdSetPerformanceMarkerINTEL = gfx30_CmdSetPerformanceMarkerINTEL,
    .CmdSetPerformanceStreamMarkerINTEL = gfx30_CmdSetPerformanceStreamMarkerINTEL,
    .CmdSetPerformanceOverrideINTEL = gfx30_CmdSetPerformanceOverrideINTEL,
    .AcquirePerformanceConfigurationINTEL = gfx30_AcquirePerformanceConfigurationINTEL,
    .ReleasePerformanceConfigurationINTEL = gfx30_ReleasePerformanceConfigurationINTEL,
    .QueueSetPerformanceConfigurationINTEL = gfx30_QueueSetPerformanceConfigurationINTEL,
    .GetPerformanceParameterINTEL = gfx30_GetPerformanceParameterINTEL,
    .GetDeviceMemoryOpaqueCaptureAddress = gfx30_GetDeviceMemoryOpaqueCaptureAddress,
    .GetDeviceMemoryOpaqueCaptureAddressKHR = gfx30_GetDeviceMemoryOpaqueCaptureAddressKHR,
    .GetPipelineExecutablePropertiesKHR = gfx30_GetPipelineExecutablePropertiesKHR,
    .GetPipelineExecutableStatisticsKHR = gfx30_GetPipelineExecutableStatisticsKHR,
    .GetPipelineExecutableInternalRepresentationsKHR = gfx30_GetPipelineExecutableInternalRepresentationsKHR,
    .CmdSetLineStipple = gfx30_CmdSetLineStipple,
    .CmdSetLineStippleKHR = gfx30_CmdSetLineStippleKHR,
    .CmdSetLineStippleEXT = gfx30_CmdSetLineStippleEXT,
    .CreateAccelerationStructureKHR = gfx30_CreateAccelerationStructureKHR,
    .CmdBuildAccelerationStructuresKHR = gfx30_CmdBuildAccelerationStructuresKHR,
    .CmdBuildAccelerationStructuresIndirectKHR = gfx30_CmdBuildAccelerationStructuresIndirectKHR,
    .BuildAccelerationStructuresKHR = gfx30_BuildAccelerationStructuresKHR,
    .GetAccelerationStructureDeviceAddressKHR = gfx30_GetAccelerationStructureDeviceAddressKHR,
    .CreateDeferredOperationKHR = gfx30_CreateDeferredOperationKHR,
    .DestroyDeferredOperationKHR = gfx30_DestroyDeferredOperationKHR,
    .GetDeferredOperationMaxConcurrencyKHR = gfx30_GetDeferredOperationMaxConcurrencyKHR,
    .GetDeferredOperationResultKHR = gfx30_GetDeferredOperationResultKHR,
    .DeferredOperationJoinKHR = gfx30_DeferredOperationJoinKHR,
    .GetPipelineIndirectMemoryRequirementsNV = gfx30_GetPipelineIndirectMemoryRequirementsNV,
    .GetPipelineIndirectDeviceAddressNV = gfx30_GetPipelineIndirectDeviceAddressNV,
    .AntiLagUpdateAMD = gfx30_AntiLagUpdateAMD,
    .CmdSetCullMode = gfx30_CmdSetCullMode,
    .CmdSetCullModeEXT = gfx30_CmdSetCullModeEXT,
    .CmdSetFrontFace = gfx30_CmdSetFrontFace,
    .CmdSetFrontFaceEXT = gfx30_CmdSetFrontFaceEXT,
    .CmdSetPrimitiveTopology = gfx30_CmdSetPrimitiveTopology,
    .CmdSetPrimitiveTopologyEXT = gfx30_CmdSetPrimitiveTopologyEXT,
    .CmdSetViewportWithCount = gfx30_CmdSetViewportWithCount,
    .CmdSetViewportWithCountEXT = gfx30_CmdSetViewportWithCountEXT,
    .CmdSetScissorWithCount = gfx30_CmdSetScissorWithCount,
    .CmdSetScissorWithCountEXT = gfx30_CmdSetScissorWithCountEXT,
    .CmdBindIndexBuffer2 = gfx30_CmdBindIndexBuffer2,
    .CmdBindIndexBuffer2KHR = gfx30_CmdBindIndexBuffer2KHR,
    .CmdBindVertexBuffers2 = gfx30_CmdBindVertexBuffers2,
    .CmdBindVertexBuffers2EXT = gfx30_CmdBindVertexBuffers2EXT,
    .CmdSetDepthTestEnable = gfx30_CmdSetDepthTestEnable,
    .CmdSetDepthTestEnableEXT = gfx30_CmdSetDepthTestEnableEXT,
    .CmdSetDepthWriteEnable = gfx30_CmdSetDepthWriteEnable,
    .CmdSetDepthWriteEnableEXT = gfx30_CmdSetDepthWriteEnableEXT,
    .CmdSetDepthCompareOp = gfx30_CmdSetDepthCompareOp,
    .CmdSetDepthCompareOpEXT = gfx30_CmdSetDepthCompareOpEXT,
    .CmdSetDepthBoundsTestEnable = gfx30_CmdSetDepthBoundsTestEnable,
    .CmdSetDepthBoundsTestEnableEXT = gfx30_CmdSetDepthBoundsTestEnableEXT,
    .CmdSetStencilTestEnable = gfx30_CmdSetStencilTestEnable,
    .CmdSetStencilTestEnableEXT = gfx30_CmdSetStencilTestEnableEXT,
    .CmdSetStencilOp = gfx30_CmdSetStencilOp,
    .CmdSetStencilOpEXT = gfx30_CmdSetStencilOpEXT,
    .CmdSetPatchControlPointsEXT = gfx30_CmdSetPatchControlPointsEXT,
    .CmdSetRasterizerDiscardEnable = gfx30_CmdSetRasterizerDiscardEnable,
    .CmdSetRasterizerDiscardEnableEXT = gfx30_CmdSetRasterizerDiscardEnableEXT,
    .CmdSetDepthBiasEnable = gfx30_CmdSetDepthBiasEnable,
    .CmdSetDepthBiasEnableEXT = gfx30_CmdSetDepthBiasEnableEXT,
    .CmdSetLogicOpEXT = gfx30_CmdSetLogicOpEXT,
    .CmdSetPrimitiveRestartEnable = gfx30_CmdSetPrimitiveRestartEnable,
    .CmdSetPrimitiveRestartEnableEXT = gfx30_CmdSetPrimitiveRestartEnableEXT,
    .CmdSetTessellationDomainOriginEXT = gfx30_CmdSetTessellationDomainOriginEXT,
    .CmdSetDepthClampEnableEXT = gfx30_CmdSetDepthClampEnableEXT,
    .CmdSetPolygonModeEXT = gfx30_CmdSetPolygonModeEXT,
    .CmdSetRasterizationSamplesEXT = gfx30_CmdSetRasterizationSamplesEXT,
    .CmdSetSampleMaskEXT = gfx30_CmdSetSampleMaskEXT,
    .CmdSetAlphaToCoverageEnableEXT = gfx30_CmdSetAlphaToCoverageEnableEXT,
    .CmdSetAlphaToOneEnableEXT = gfx30_CmdSetAlphaToOneEnableEXT,
    .CmdSetLogicOpEnableEXT = gfx30_CmdSetLogicOpEnableEXT,
    .CmdSetColorBlendEnableEXT = gfx30_CmdSetColorBlendEnableEXT,
    .CmdSetColorBlendEquationEXT = gfx30_CmdSetColorBlendEquationEXT,
    .CmdSetColorWriteMaskEXT = gfx30_CmdSetColorWriteMaskEXT,
    .CmdSetRasterizationStreamEXT = gfx30_CmdSetRasterizationStreamEXT,
    .CmdSetConservativeRasterizationModeEXT = gfx30_CmdSetConservativeRasterizationModeEXT,
    .CmdSetExtraPrimitiveOverestimationSizeEXT = gfx30_CmdSetExtraPrimitiveOverestimationSizeEXT,
    .CmdSetDepthClipEnableEXT = gfx30_CmdSetDepthClipEnableEXT,
    .CmdSetSampleLocationsEnableEXT = gfx30_CmdSetSampleLocationsEnableEXT,
    .CmdSetColorBlendAdvancedEXT = gfx30_CmdSetColorBlendAdvancedEXT,
    .CmdSetProvokingVertexModeEXT = gfx30_CmdSetProvokingVertexModeEXT,
    .CmdSetLineRasterizationModeEXT = gfx30_CmdSetLineRasterizationModeEXT,
    .CmdSetLineStippleEnableEXT = gfx30_CmdSetLineStippleEnableEXT,
    .CmdSetDepthClipNegativeOneToOneEXT = gfx30_CmdSetDepthClipNegativeOneToOneEXT,
    .CmdSetViewportWScalingEnableNV = gfx30_CmdSetViewportWScalingEnableNV,
    .CmdSetViewportSwizzleNV = gfx30_CmdSetViewportSwizzleNV,
    .CmdSetCoverageToColorEnableNV = gfx30_CmdSetCoverageToColorEnableNV,
    .CmdSetCoverageToColorLocationNV = gfx30_CmdSetCoverageToColorLocationNV,
    .CmdSetCoverageModulationModeNV = gfx30_CmdSetCoverageModulationModeNV,
    .CmdSetCoverageModulationTableEnableNV = gfx30_CmdSetCoverageModulationTableEnableNV,
    .CmdSetCoverageModulationTableNV = gfx30_CmdSetCoverageModulationTableNV,
    .CmdSetShadingRateImageEnableNV = gfx30_CmdSetShadingRateImageEnableNV,
    .CmdSetCoverageReductionModeNV = gfx30_CmdSetCoverageReductionModeNV,
    .CmdSetRepresentativeFragmentTestEnableNV = gfx30_CmdSetRepresentativeFragmentTestEnableNV,
    .CreatePrivateDataSlot = gfx30_CreatePrivateDataSlot,
    .CreatePrivateDataSlotEXT = gfx30_CreatePrivateDataSlotEXT,
    .DestroyPrivateDataSlot = gfx30_DestroyPrivateDataSlot,
    .DestroyPrivateDataSlotEXT = gfx30_DestroyPrivateDataSlotEXT,
    .SetPrivateData = gfx30_SetPrivateData,
    .SetPrivateDataEXT = gfx30_SetPrivateDataEXT,
    .GetPrivateData = gfx30_GetPrivateData,
    .GetPrivateDataEXT = gfx30_GetPrivateDataEXT,
    .CmdCopyBuffer2 = gfx30_CmdCopyBuffer2,
    .CmdCopyBuffer2KHR = gfx30_CmdCopyBuffer2KHR,
    .CmdCopyImage2 = gfx30_CmdCopyImage2,
    .CmdCopyImage2KHR = gfx30_CmdCopyImage2KHR,
    .CmdBlitImage2 = gfx30_CmdBlitImage2,
    .CmdBlitImage2KHR = gfx30_CmdBlitImage2KHR,
    .CmdCopyBufferToImage2 = gfx30_CmdCopyBufferToImage2,
    .CmdCopyBufferToImage2KHR = gfx30_CmdCopyBufferToImage2KHR,
    .CmdCopyImageToBuffer2 = gfx30_CmdCopyImageToBuffer2,
    .CmdCopyImageToBuffer2KHR = gfx30_CmdCopyImageToBuffer2KHR,
    .CmdResolveImage2 = gfx30_CmdResolveImage2,
    .CmdResolveImage2KHR = gfx30_CmdResolveImage2KHR,
    .CmdSetFragmentShadingRateKHR = gfx30_CmdSetFragmentShadingRateKHR,
    .CmdSetFragmentShadingRateEnumNV = gfx30_CmdSetFragmentShadingRateEnumNV,
    .GetAccelerationStructureBuildSizesKHR = gfx30_GetAccelerationStructureBuildSizesKHR,
    .CmdSetVertexInputEXT = gfx30_CmdSetVertexInputEXT,
    .CmdSetColorWriteEnableEXT = gfx30_CmdSetColorWriteEnableEXT,
    .CmdSetEvent2 = gfx30_CmdSetEvent2,
    .CmdSetEvent2KHR = gfx30_CmdSetEvent2KHR,
    .CmdResetEvent2 = gfx30_CmdResetEvent2,
    .CmdResetEvent2KHR = gfx30_CmdResetEvent2KHR,
    .CmdWaitEvents2 = gfx30_CmdWaitEvents2,
    .CmdWaitEvents2KHR = gfx30_CmdWaitEvents2KHR,
    .CmdPipelineBarrier2 = gfx30_CmdPipelineBarrier2,
    .CmdPipelineBarrier2KHR = gfx30_CmdPipelineBarrier2KHR,
    .QueueSubmit2 = gfx30_QueueSubmit2,
    .QueueSubmit2KHR = gfx30_QueueSubmit2KHR,
    .CmdWriteTimestamp2 = gfx30_CmdWriteTimestamp2,
    .CmdWriteTimestamp2KHR = gfx30_CmdWriteTimestamp2KHR,
    .CmdWriteBufferMarker2AMD = gfx30_CmdWriteBufferMarker2AMD,
    .GetQueueCheckpointData2NV = gfx30_GetQueueCheckpointData2NV,
    .CopyMemoryToImage = gfx30_CopyMemoryToImage,
    .CopyMemoryToImageEXT = gfx30_CopyMemoryToImageEXT,
    .CopyImageToMemory = gfx30_CopyImageToMemory,
    .CopyImageToMemoryEXT = gfx30_CopyImageToMemoryEXT,
    .CopyImageToImage = gfx30_CopyImageToImage,
    .CopyImageToImageEXT = gfx30_CopyImageToImageEXT,
    .TransitionImageLayout = gfx30_TransitionImageLayout,
    .TransitionImageLayoutEXT = gfx30_TransitionImageLayoutEXT,
    .CreateVideoSessionKHR = gfx30_CreateVideoSessionKHR,
    .DestroyVideoSessionKHR = gfx30_DestroyVideoSessionKHR,
    .CreateVideoSessionParametersKHR = gfx30_CreateVideoSessionParametersKHR,
    .UpdateVideoSessionParametersKHR = gfx30_UpdateVideoSessionParametersKHR,
    .GetEncodedVideoSessionParametersKHR = gfx30_GetEncodedVideoSessionParametersKHR,
    .DestroyVideoSessionParametersKHR = gfx30_DestroyVideoSessionParametersKHR,
    .GetVideoSessionMemoryRequirementsKHR = gfx30_GetVideoSessionMemoryRequirementsKHR,
    .BindVideoSessionMemoryKHR = gfx30_BindVideoSessionMemoryKHR,
    .CmdDecodeVideoKHR = gfx30_CmdDecodeVideoKHR,
    .CmdBeginVideoCodingKHR = gfx30_CmdBeginVideoCodingKHR,
    .CmdControlVideoCodingKHR = gfx30_CmdControlVideoCodingKHR,
    .CmdEndVideoCodingKHR = gfx30_CmdEndVideoCodingKHR,
    .CmdEncodeVideoKHR = gfx30_CmdEncodeVideoKHR,
    .CmdDecompressMemoryNV = gfx30_CmdDecompressMemoryNV,
    .CmdDecompressMemoryIndirectCountNV = gfx30_CmdDecompressMemoryIndirectCountNV,
    .GetPartitionedAccelerationStructuresBuildSizesNV = gfx30_GetPartitionedAccelerationStructuresBuildSizesNV,
    .CmdBuildPartitionedAccelerationStructuresNV = gfx30_CmdBuildPartitionedAccelerationStructuresNV,
    .CmdDecompressMemoryEXT = gfx30_CmdDecompressMemoryEXT,
    .CmdDecompressMemoryIndirectCountEXT = gfx30_CmdDecompressMemoryIndirectCountEXT,
    .CreateCuModuleNVX = gfx30_CreateCuModuleNVX,
    .CreateCuFunctionNVX = gfx30_CreateCuFunctionNVX,
    .DestroyCuModuleNVX = gfx30_DestroyCuModuleNVX,
    .DestroyCuFunctionNVX = gfx30_DestroyCuFunctionNVX,
    .CmdCuLaunchKernelNVX = gfx30_CmdCuLaunchKernelNVX,
    .GetDescriptorSetLayoutSizeEXT = gfx30_GetDescriptorSetLayoutSizeEXT,
    .GetDescriptorSetLayoutBindingOffsetEXT = gfx30_GetDescriptorSetLayoutBindingOffsetEXT,
    .GetDescriptorEXT = gfx30_GetDescriptorEXT,
    .CmdBindDescriptorBuffersEXT = gfx30_CmdBindDescriptorBuffersEXT,
    .CmdSetDescriptorBufferOffsetsEXT = gfx30_CmdSetDescriptorBufferOffsetsEXT,
    .CmdBindDescriptorBufferEmbeddedSamplersEXT = gfx30_CmdBindDescriptorBufferEmbeddedSamplersEXT,
    .GetBufferOpaqueCaptureDescriptorDataEXT = gfx30_GetBufferOpaqueCaptureDescriptorDataEXT,
    .GetImageOpaqueCaptureDescriptorDataEXT = gfx30_GetImageOpaqueCaptureDescriptorDataEXT,
    .GetImageViewOpaqueCaptureDescriptorDataEXT = gfx30_GetImageViewOpaqueCaptureDescriptorDataEXT,
    .GetSamplerOpaqueCaptureDescriptorDataEXT = gfx30_GetSamplerOpaqueCaptureDescriptorDataEXT,
    .GetAccelerationStructureOpaqueCaptureDescriptorDataEXT = gfx30_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT,
    .SetDeviceMemoryPriorityEXT = gfx30_SetDeviceMemoryPriorityEXT,
    .WaitForPresent2KHR = gfx30_WaitForPresent2KHR,
    .WaitForPresentKHR = gfx30_WaitForPresentKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateBufferCollectionFUCHSIA = gfx30_CreateBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .CreateBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionBufferConstraintsFUCHSIA = gfx30_SetBufferCollectionBufferConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionBufferConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionImageConstraintsFUCHSIA = gfx30_SetBufferCollectionImageConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionImageConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .DestroyBufferCollectionFUCHSIA = gfx30_DestroyBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .DestroyBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetBufferCollectionPropertiesFUCHSIA = gfx30_GetBufferCollectionPropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetBufferCollectionPropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .CmdBeginRendering = gfx30_CmdBeginRendering,
    .CmdBeginRenderingKHR = gfx30_CmdBeginRenderingKHR,
    .CmdEndRendering = gfx30_CmdEndRendering,
    .CmdEndRendering2KHR = gfx30_CmdEndRendering2KHR,
    .CmdEndRendering2EXT = gfx30_CmdEndRendering2EXT,
    .CmdEndRenderingKHR = gfx30_CmdEndRenderingKHR,
    .GetDescriptorSetLayoutHostMappingInfoVALVE = gfx30_GetDescriptorSetLayoutHostMappingInfoVALVE,
    .GetDescriptorSetHostMappingVALVE = gfx30_GetDescriptorSetHostMappingVALVE,
    .CreateMicromapEXT = gfx30_CreateMicromapEXT,
    .CmdBuildMicromapsEXT = gfx30_CmdBuildMicromapsEXT,
    .BuildMicromapsEXT = gfx30_BuildMicromapsEXT,
    .DestroyMicromapEXT = gfx30_DestroyMicromapEXT,
    .CmdCopyMicromapEXT = gfx30_CmdCopyMicromapEXT,
    .CopyMicromapEXT = gfx30_CopyMicromapEXT,
    .CmdCopyMicromapToMemoryEXT = gfx30_CmdCopyMicromapToMemoryEXT,
    .CopyMicromapToMemoryEXT = gfx30_CopyMicromapToMemoryEXT,
    .CmdCopyMemoryToMicromapEXT = gfx30_CmdCopyMemoryToMicromapEXT,
    .CopyMemoryToMicromapEXT = gfx30_CopyMemoryToMicromapEXT,
    .CmdWriteMicromapsPropertiesEXT = gfx30_CmdWriteMicromapsPropertiesEXT,
    .WriteMicromapsPropertiesEXT = gfx30_WriteMicromapsPropertiesEXT,
    .GetDeviceMicromapCompatibilityEXT = gfx30_GetDeviceMicromapCompatibilityEXT,
    .GetMicromapBuildSizesEXT = gfx30_GetMicromapBuildSizesEXT,
    .GetShaderModuleIdentifierEXT = gfx30_GetShaderModuleIdentifierEXT,
    .GetShaderModuleCreateInfoIdentifierEXT = gfx30_GetShaderModuleCreateInfoIdentifierEXT,
    .GetImageSubresourceLayout2 = gfx30_GetImageSubresourceLayout2,
    .GetImageSubresourceLayout2KHR = gfx30_GetImageSubresourceLayout2KHR,
    .GetImageSubresourceLayout2EXT = gfx30_GetImageSubresourceLayout2EXT,
    .GetPipelinePropertiesEXT = gfx30_GetPipelinePropertiesEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .ExportMetalObjectsEXT = gfx30_ExportMetalObjectsEXT,
#elif defined(_MSC_VER)
    .ExportMetalObjectsEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CmdBindTileMemoryQCOM = gfx30_CmdBindTileMemoryQCOM,
    .GetFramebufferTilePropertiesQCOM = gfx30_GetFramebufferTilePropertiesQCOM,
    .GetDynamicRenderingTilePropertiesQCOM = gfx30_GetDynamicRenderingTilePropertiesQCOM,
    .CreateOpticalFlowSessionNV = gfx30_CreateOpticalFlowSessionNV,
    .DestroyOpticalFlowSessionNV = gfx30_DestroyOpticalFlowSessionNV,
    .BindOpticalFlowSessionImageNV = gfx30_BindOpticalFlowSessionImageNV,
    .CmdOpticalFlowExecuteNV = gfx30_CmdOpticalFlowExecuteNV,
    .GetDeviceFaultInfoEXT = gfx30_GetDeviceFaultInfoEXT,
    .GetDeviceFaultReportsKHR = gfx30_GetDeviceFaultReportsKHR,
    .GetDeviceFaultDebugInfoKHR = gfx30_GetDeviceFaultDebugInfoKHR,
    .CmdSetDepthBias2EXT = gfx30_CmdSetDepthBias2EXT,
    .ReleaseSwapchainImagesKHR = gfx30_ReleaseSwapchainImagesKHR,
    .ReleaseSwapchainImagesEXT = gfx30_ReleaseSwapchainImagesEXT,
    .GetDeviceImageSubresourceLayout = gfx30_GetDeviceImageSubresourceLayout,
    .GetDeviceImageSubresourceLayoutKHR = gfx30_GetDeviceImageSubresourceLayoutKHR,
    .MapMemory2 = gfx30_MapMemory2,
    .MapMemory2KHR = gfx30_MapMemory2KHR,
    .UnmapMemory2 = gfx30_UnmapMemory2,
    .UnmapMemory2KHR = gfx30_UnmapMemory2KHR,
    .CreateShadersEXT = gfx30_CreateShadersEXT,
    .DestroyShaderEXT = gfx30_DestroyShaderEXT,
    .GetShaderBinaryDataEXT = gfx30_GetShaderBinaryDataEXT,
    .CmdBindShadersEXT = gfx30_CmdBindShadersEXT,
    .SetSwapchainPresentTimingQueueSizeEXT = gfx30_SetSwapchainPresentTimingQueueSizeEXT,
    .GetSwapchainTimingPropertiesEXT = gfx30_GetSwapchainTimingPropertiesEXT,
    .GetSwapchainTimeDomainPropertiesEXT = gfx30_GetSwapchainTimeDomainPropertiesEXT,
    .GetPastPresentationTimingEXT = gfx30_GetPastPresentationTimingEXT,
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetScreenBufferPropertiesQNX = gfx30_GetScreenBufferPropertiesQNX,
#elif defined(_MSC_VER)
    .GetScreenBufferPropertiesQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CmdBindDescriptorSets2 = gfx30_CmdBindDescriptorSets2,
    .CmdBindDescriptorSets2KHR = gfx30_CmdBindDescriptorSets2KHR,
    .CmdPushConstants2 = gfx30_CmdPushConstants2,
    .CmdPushConstants2KHR = gfx30_CmdPushConstants2KHR,
    .CmdPushDescriptorSet2 = gfx30_CmdPushDescriptorSet2,
    .CmdPushDescriptorSet2KHR = gfx30_CmdPushDescriptorSet2KHR,
    .CmdPushDescriptorSetWithTemplate2 = gfx30_CmdPushDescriptorSetWithTemplate2,
    .CmdPushDescriptorSetWithTemplate2KHR = gfx30_CmdPushDescriptorSetWithTemplate2KHR,
    .CmdSetDescriptorBufferOffsets2EXT = gfx30_CmdSetDescriptorBufferOffsets2EXT,
    .CmdBindDescriptorBufferEmbeddedSamplers2EXT = gfx30_CmdBindDescriptorBufferEmbeddedSamplers2EXT,
    .SetLatencySleepModeNV = gfx30_SetLatencySleepModeNV,
    .LatencySleepNV = gfx30_LatencySleepNV,
    .SetLatencyMarkerNV = gfx30_SetLatencyMarkerNV,
    .GetLatencyTimingsNV = gfx30_GetLatencyTimingsNV,
    .QueueNotifyOutOfBandNV = gfx30_QueueNotifyOutOfBandNV,
    .CmdSetRenderingAttachmentLocations = gfx30_CmdSetRenderingAttachmentLocations,
    .CmdSetRenderingAttachmentLocationsKHR = gfx30_CmdSetRenderingAttachmentLocationsKHR,
    .CmdSetRenderingInputAttachmentIndices = gfx30_CmdSetRenderingInputAttachmentIndices,
    .CmdSetRenderingInputAttachmentIndicesKHR = gfx30_CmdSetRenderingInputAttachmentIndicesKHR,
    .CmdSetDepthClampRangeEXT = gfx30_CmdSetDepthClampRangeEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandleEXT = gfx30_GetMemoryMetalHandleEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandleEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandlePropertiesEXT = gfx30_GetMemoryMetalHandlePropertiesEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandlePropertiesEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .ConvertCooperativeVectorMatrixNV = gfx30_ConvertCooperativeVectorMatrixNV,
    .CmdConvertCooperativeVectorMatrixNV = gfx30_CmdConvertCooperativeVectorMatrixNV,
    .CmdDispatchTileQCOM = gfx30_CmdDispatchTileQCOM,
    .CmdBeginPerTileExecutionQCOM = gfx30_CmdBeginPerTileExecutionQCOM,
    .CmdEndPerTileExecutionQCOM = gfx30_CmdEndPerTileExecutionQCOM,
    .CreateExternalComputeQueueNV = gfx30_CreateExternalComputeQueueNV,
    .DestroyExternalComputeQueueNV = gfx30_DestroyExternalComputeQueueNV,
    .CreateShaderInstrumentationARM = gfx30_CreateShaderInstrumentationARM,
    .DestroyShaderInstrumentationARM = gfx30_DestroyShaderInstrumentationARM,
    .CmdBeginShaderInstrumentationARM = gfx30_CmdBeginShaderInstrumentationARM,
    .CmdEndShaderInstrumentationARM = gfx30_CmdEndShaderInstrumentationARM,
    .GetShaderInstrumentationValuesARM = gfx30_GetShaderInstrumentationValuesARM,
    .ClearShaderInstrumentationMetricsARM = gfx30_ClearShaderInstrumentationMetricsARM,
    .CreateTensorARM = gfx30_CreateTensorARM,
    .DestroyTensorARM = gfx30_DestroyTensorARM,
    .CreateTensorViewARM = gfx30_CreateTensorViewARM,
    .DestroyTensorViewARM = gfx30_DestroyTensorViewARM,
    .GetTensorMemoryRequirementsARM = gfx30_GetTensorMemoryRequirementsARM,
    .BindTensorMemoryARM = gfx30_BindTensorMemoryARM,
    .GetDeviceTensorMemoryRequirementsARM = gfx30_GetDeviceTensorMemoryRequirementsARM,
    .CmdCopyTensorARM = gfx30_CmdCopyTensorARM,
    .GetTensorOpaqueCaptureDescriptorDataARM = gfx30_GetTensorOpaqueCaptureDescriptorDataARM,
    .GetTensorViewOpaqueCaptureDescriptorDataARM = gfx30_GetTensorViewOpaqueCaptureDescriptorDataARM,
    .CreateDataGraphPipelinesARM = gfx30_CreateDataGraphPipelinesARM,
    .CreateDataGraphPipelineSessionARM = gfx30_CreateDataGraphPipelineSessionARM,
    .GetDataGraphPipelineSessionBindPointRequirementsARM = gfx30_GetDataGraphPipelineSessionBindPointRequirementsARM,
    .GetDataGraphPipelineSessionMemoryRequirementsARM = gfx30_GetDataGraphPipelineSessionMemoryRequirementsARM,
    .BindDataGraphPipelineSessionMemoryARM = gfx30_BindDataGraphPipelineSessionMemoryARM,
    .DestroyDataGraphPipelineSessionARM = gfx30_DestroyDataGraphPipelineSessionARM,
    .CmdDispatchDataGraphARM = gfx30_CmdDispatchDataGraphARM,
    .GetDataGraphPipelineAvailablePropertiesARM = gfx30_GetDataGraphPipelineAvailablePropertiesARM,
    .GetDataGraphPipelinePropertiesARM = gfx30_GetDataGraphPipelinePropertiesARM,
#ifdef VK_USE_PLATFORM_OHOS
    .GetNativeBufferPropertiesOHOS = gfx30_GetNativeBufferPropertiesOHOS,
#elif defined(_MSC_VER)
    .GetNativeBufferPropertiesOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
    .GetMemoryNativeBufferOHOS = gfx30_GetMemoryNativeBufferOHOS,
#elif defined(_MSC_VER)
    .GetMemoryNativeBufferOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .QueueSetPerfHintQCOM = gfx30_QueueSetPerfHintQCOM,
    .CmdSetComputeOccupancyPriorityNV = gfx30_CmdSetComputeOccupancyPriorityNV,
    .WriteSamplerDescriptorsEXT = gfx30_WriteSamplerDescriptorsEXT,
    .WriteResourceDescriptorsEXT = gfx30_WriteResourceDescriptorsEXT,
    .CmdBindSamplerHeapEXT = gfx30_CmdBindSamplerHeapEXT,
    .CmdBindResourceHeapEXT = gfx30_CmdBindResourceHeapEXT,
    .CmdPushDataEXT = gfx30_CmdPushDataEXT,
    .RegisterCustomBorderColorEXT = gfx30_RegisterCustomBorderColorEXT,
    .UnregisterCustomBorderColorEXT = gfx30_UnregisterCustomBorderColorEXT,
    .GetImageOpaqueCaptureDataEXT = gfx30_GetImageOpaqueCaptureDataEXT,
    .GetTensorOpaqueCaptureDataARM = gfx30_GetTensorOpaqueCaptureDataARM,
    .CmdCopyMemoryKHR = gfx30_CmdCopyMemoryKHR,
    .CmdCopyMemoryToImageKHR = gfx30_CmdCopyMemoryToImageKHR,
    .CmdCopyImageToMemoryKHR = gfx30_CmdCopyImageToMemoryKHR,
    .CmdUpdateMemoryKHR = gfx30_CmdUpdateMemoryKHR,
    .CmdFillMemoryKHR = gfx30_CmdFillMemoryKHR,
    .CmdCopyQueryPoolResultsToMemoryKHR = gfx30_CmdCopyQueryPoolResultsToMemoryKHR,
    .CmdBeginConditionalRendering2EXT = gfx30_CmdBeginConditionalRendering2EXT,
    .CmdBindTransformFeedbackBuffers2EXT = gfx30_CmdBindTransformFeedbackBuffers2EXT,
    .CmdBeginTransformFeedback2EXT = gfx30_CmdBeginTransformFeedback2EXT,
    .CmdEndTransformFeedback2EXT = gfx30_CmdEndTransformFeedback2EXT,
    .CmdDrawIndirectByteCount2EXT = gfx30_CmdDrawIndirectByteCount2EXT,
    .CmdWriteMarkerToMemoryAMD = gfx30_CmdWriteMarkerToMemoryAMD,
    .CmdBindIndexBuffer3KHR = gfx30_CmdBindIndexBuffer3KHR,
    .CmdBindVertexBuffers3KHR = gfx30_CmdBindVertexBuffers3KHR,
    .CmdDrawIndirect2KHR = gfx30_CmdDrawIndirect2KHR,
    .CmdDrawIndexedIndirect2KHR = gfx30_CmdDrawIndexedIndirect2KHR,
    .CmdDrawIndirectCount2KHR = gfx30_CmdDrawIndirectCount2KHR,
    .CmdDrawIndexedIndirectCount2KHR = gfx30_CmdDrawIndexedIndirectCount2KHR,
    .CmdDrawMeshTasksIndirect2EXT = gfx30_CmdDrawMeshTasksIndirect2EXT,
    .CmdDrawMeshTasksIndirectCount2EXT = gfx30_CmdDrawMeshTasksIndirectCount2EXT,
    .CmdDispatchIndirect2KHR = gfx30_CmdDispatchIndirect2KHR,
    .CreateAccelerationStructure2KHR = gfx30_CreateAccelerationStructure2KHR,
    .CmdSetDispatchParametersARM = gfx30_CmdSetDispatchParametersARM,
};
const struct vk_device_entrypoint_table gfx35_device_entrypoints = {
    .GetDeviceProcAddr = gfx35_GetDeviceProcAddr,
    .DestroyDevice = gfx35_DestroyDevice,
    .GetDeviceQueue = gfx35_GetDeviceQueue,
    .QueueSubmit = gfx35_QueueSubmit,
    .QueueWaitIdle = gfx35_QueueWaitIdle,
    .DeviceWaitIdle = gfx35_DeviceWaitIdle,
    .AllocateMemory = gfx35_AllocateMemory,
    .FreeMemory = gfx35_FreeMemory,
    .MapMemory = gfx35_MapMemory,
    .UnmapMemory = gfx35_UnmapMemory,
    .FlushMappedMemoryRanges = gfx35_FlushMappedMemoryRanges,
    .InvalidateMappedMemoryRanges = gfx35_InvalidateMappedMemoryRanges,
    .GetDeviceMemoryCommitment = gfx35_GetDeviceMemoryCommitment,
    .GetBufferMemoryRequirements = gfx35_GetBufferMemoryRequirements,
    .BindBufferMemory = gfx35_BindBufferMemory,
    .GetImageMemoryRequirements = gfx35_GetImageMemoryRequirements,
    .BindImageMemory = gfx35_BindImageMemory,
    .GetImageSparseMemoryRequirements = gfx35_GetImageSparseMemoryRequirements,
    .QueueBindSparse = gfx35_QueueBindSparse,
    .CreateFence = gfx35_CreateFence,
    .DestroyFence = gfx35_DestroyFence,
    .ResetFences = gfx35_ResetFences,
    .GetFenceStatus = gfx35_GetFenceStatus,
    .WaitForFences = gfx35_WaitForFences,
    .CreateSemaphore = gfx35_CreateSemaphore,
    .DestroySemaphore = gfx35_DestroySemaphore,
    .CreateEvent = gfx35_CreateEvent,
    .DestroyEvent = gfx35_DestroyEvent,
    .GetEventStatus = gfx35_GetEventStatus,
    .SetEvent = gfx35_SetEvent,
    .ResetEvent = gfx35_ResetEvent,
    .CreateQueryPool = gfx35_CreateQueryPool,
    .DestroyQueryPool = gfx35_DestroyQueryPool,
    .GetQueryPoolResults = gfx35_GetQueryPoolResults,
    .ResetQueryPool = gfx35_ResetQueryPool,
    .ResetQueryPoolEXT = gfx35_ResetQueryPoolEXT,
    .CreateBuffer = gfx35_CreateBuffer,
    .DestroyBuffer = gfx35_DestroyBuffer,
    .CreateBufferView = gfx35_CreateBufferView,
    .DestroyBufferView = gfx35_DestroyBufferView,
    .CreateImage = gfx35_CreateImage,
    .DestroyImage = gfx35_DestroyImage,
    .GetImageSubresourceLayout = gfx35_GetImageSubresourceLayout,
    .CreateImageView = gfx35_CreateImageView,
    .DestroyImageView = gfx35_DestroyImageView,
    .CreateShaderModule = gfx35_CreateShaderModule,
    .DestroyShaderModule = gfx35_DestroyShaderModule,
    .CreatePipelineCache = gfx35_CreatePipelineCache,
    .DestroyPipelineCache = gfx35_DestroyPipelineCache,
    .GetPipelineCacheData = gfx35_GetPipelineCacheData,
    .MergePipelineCaches = gfx35_MergePipelineCaches,
    .CreatePipelineBinariesKHR = gfx35_CreatePipelineBinariesKHR,
    .DestroyPipelineBinaryKHR = gfx35_DestroyPipelineBinaryKHR,
    .GetPipelineKeyKHR = gfx35_GetPipelineKeyKHR,
    .GetPipelineBinaryDataKHR = gfx35_GetPipelineBinaryDataKHR,
    .ReleaseCapturedPipelineDataKHR = gfx35_ReleaseCapturedPipelineDataKHR,
    .CreateGraphicsPipelines = gfx35_CreateGraphicsPipelines,
    .CreateComputePipelines = gfx35_CreateComputePipelines,
    .GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI = gfx35_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI,
    .DestroyPipeline = gfx35_DestroyPipeline,
    .CreatePipelineLayout = gfx35_CreatePipelineLayout,
    .DestroyPipelineLayout = gfx35_DestroyPipelineLayout,
    .CreateSampler = gfx35_CreateSampler,
    .DestroySampler = gfx35_DestroySampler,
    .CreateDescriptorSetLayout = gfx35_CreateDescriptorSetLayout,
    .DestroyDescriptorSetLayout = gfx35_DestroyDescriptorSetLayout,
    .CreateDescriptorPool = gfx35_CreateDescriptorPool,
    .DestroyDescriptorPool = gfx35_DestroyDescriptorPool,
    .ResetDescriptorPool = gfx35_ResetDescriptorPool,
    .AllocateDescriptorSets = gfx35_AllocateDescriptorSets,
    .FreeDescriptorSets = gfx35_FreeDescriptorSets,
    .UpdateDescriptorSets = gfx35_UpdateDescriptorSets,
    .CreateFramebuffer = gfx35_CreateFramebuffer,
    .DestroyFramebuffer = gfx35_DestroyFramebuffer,
    .CreateRenderPass = gfx35_CreateRenderPass,
    .DestroyRenderPass = gfx35_DestroyRenderPass,
    .GetRenderAreaGranularity = gfx35_GetRenderAreaGranularity,
    .GetRenderingAreaGranularity = gfx35_GetRenderingAreaGranularity,
    .GetRenderingAreaGranularityKHR = gfx35_GetRenderingAreaGranularityKHR,
    .CreateCommandPool = gfx35_CreateCommandPool,
    .DestroyCommandPool = gfx35_DestroyCommandPool,
    .ResetCommandPool = gfx35_ResetCommandPool,
    .AllocateCommandBuffers = gfx35_AllocateCommandBuffers,
    .FreeCommandBuffers = gfx35_FreeCommandBuffers,
    .BeginCommandBuffer = gfx35_BeginCommandBuffer,
    .EndCommandBuffer = gfx35_EndCommandBuffer,
    .ResetCommandBuffer = gfx35_ResetCommandBuffer,
    .CmdBindPipeline = gfx35_CmdBindPipeline,
    .CmdSetPrimitiveRestartIndexEXT = gfx35_CmdSetPrimitiveRestartIndexEXT,
    .CmdSetAttachmentFeedbackLoopEnableEXT = gfx35_CmdSetAttachmentFeedbackLoopEnableEXT,
    .CmdSetViewport = gfx35_CmdSetViewport,
    .CmdSetScissor = gfx35_CmdSetScissor,
    .CmdSetLineWidth = gfx35_CmdSetLineWidth,
    .CmdSetDepthBias = gfx35_CmdSetDepthBias,
    .CmdSetBlendConstants = gfx35_CmdSetBlendConstants,
    .CmdSetDepthBounds = gfx35_CmdSetDepthBounds,
    .CmdSetStencilCompareMask = gfx35_CmdSetStencilCompareMask,
    .CmdSetStencilWriteMask = gfx35_CmdSetStencilWriteMask,
    .CmdSetStencilReference = gfx35_CmdSetStencilReference,
    .CmdBindDescriptorSets = gfx35_CmdBindDescriptorSets,
    .CmdBindIndexBuffer = gfx35_CmdBindIndexBuffer,
    .CmdBindVertexBuffers = gfx35_CmdBindVertexBuffers,
    .CmdDraw = gfx35_CmdDraw,
    .CmdDrawIndexed = gfx35_CmdDrawIndexed,
    .CmdDrawMultiEXT = gfx35_CmdDrawMultiEXT,
    .CmdDrawMultiIndexedEXT = gfx35_CmdDrawMultiIndexedEXT,
    .CmdDrawIndirect = gfx35_CmdDrawIndirect,
    .CmdDrawIndexedIndirect = gfx35_CmdDrawIndexedIndirect,
    .CmdDispatch = gfx35_CmdDispatch,
    .CmdDispatchIndirect = gfx35_CmdDispatchIndirect,
    .CmdSubpassShadingHUAWEI = gfx35_CmdSubpassShadingHUAWEI,
    .CmdDrawClusterHUAWEI = gfx35_CmdDrawClusterHUAWEI,
    .CmdDrawClusterIndirectHUAWEI = gfx35_CmdDrawClusterIndirectHUAWEI,
    .CmdUpdatePipelineIndirectBufferNV = gfx35_CmdUpdatePipelineIndirectBufferNV,
    .CmdCopyBuffer = gfx35_CmdCopyBuffer,
    .CmdCopyImage = gfx35_CmdCopyImage,
    .CmdBlitImage = gfx35_CmdBlitImage,
    .CmdCopyBufferToImage = gfx35_CmdCopyBufferToImage,
    .CmdCopyImageToBuffer = gfx35_CmdCopyImageToBuffer,
    .CmdCopyMemoryIndirectNV = gfx35_CmdCopyMemoryIndirectNV,
    .CmdCopyMemoryIndirectKHR = gfx35_CmdCopyMemoryIndirectKHR,
    .CmdCopyMemoryToImageIndirectNV = gfx35_CmdCopyMemoryToImageIndirectNV,
    .CmdCopyMemoryToImageIndirectKHR = gfx35_CmdCopyMemoryToImageIndirectKHR,
    .CmdUpdateBuffer = gfx35_CmdUpdateBuffer,
    .CmdFillBuffer = gfx35_CmdFillBuffer,
    .CmdClearColorImage = gfx35_CmdClearColorImage,
    .CmdClearDepthStencilImage = gfx35_CmdClearDepthStencilImage,
    .CmdClearAttachments = gfx35_CmdClearAttachments,
    .CmdResolveImage = gfx35_CmdResolveImage,
    .CmdSetEvent = gfx35_CmdSetEvent,
    .CmdResetEvent = gfx35_CmdResetEvent,
    .CmdWaitEvents = gfx35_CmdWaitEvents,
    .CmdPipelineBarrier = gfx35_CmdPipelineBarrier,
    .CmdBeginQuery = gfx35_CmdBeginQuery,
    .CmdEndQuery = gfx35_CmdEndQuery,
    .CmdBeginConditionalRenderingEXT = gfx35_CmdBeginConditionalRenderingEXT,
    .CmdEndConditionalRenderingEXT = gfx35_CmdEndConditionalRenderingEXT,
    .CmdBeginCustomResolveEXT = gfx35_CmdBeginCustomResolveEXT,
    .CmdResetQueryPool = gfx35_CmdResetQueryPool,
    .CmdWriteTimestamp = gfx35_CmdWriteTimestamp,
    .CmdCopyQueryPoolResults = gfx35_CmdCopyQueryPoolResults,
    .CmdPushConstants = gfx35_CmdPushConstants,
    .CmdBeginRenderPass = gfx35_CmdBeginRenderPass,
    .CmdNextSubpass = gfx35_CmdNextSubpass,
    .CmdEndRenderPass = gfx35_CmdEndRenderPass,
    .CmdExecuteCommands = gfx35_CmdExecuteCommands,
    .CreateSharedSwapchainsKHR = gfx35_CreateSharedSwapchainsKHR,
    .CreateSwapchainKHR = gfx35_CreateSwapchainKHR,
    .DestroySwapchainKHR = gfx35_DestroySwapchainKHR,
    .GetSwapchainImagesKHR = gfx35_GetSwapchainImagesKHR,
    .AcquireNextImageKHR = gfx35_AcquireNextImageKHR,
    .QueuePresentKHR = gfx35_QueuePresentKHR,
    .DebugMarkerSetObjectNameEXT = gfx35_DebugMarkerSetObjectNameEXT,
    .DebugMarkerSetObjectTagEXT = gfx35_DebugMarkerSetObjectTagEXT,
    .CmdDebugMarkerBeginEXT = gfx35_CmdDebugMarkerBeginEXT,
    .CmdDebugMarkerEndEXT = gfx35_CmdDebugMarkerEndEXT,
    .CmdDebugMarkerInsertEXT = gfx35_CmdDebugMarkerInsertEXT,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleNV = gfx35_GetMemoryWin32HandleNV,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .CmdExecuteGeneratedCommandsNV = gfx35_CmdExecuteGeneratedCommandsNV,
    .CmdPreprocessGeneratedCommandsNV = gfx35_CmdPreprocessGeneratedCommandsNV,
    .CmdBindPipelineShaderGroupNV = gfx35_CmdBindPipelineShaderGroupNV,
    .GetGeneratedCommandsMemoryRequirementsNV = gfx35_GetGeneratedCommandsMemoryRequirementsNV,
    .CreateIndirectCommandsLayoutNV = gfx35_CreateIndirectCommandsLayoutNV,
    .DestroyIndirectCommandsLayoutNV = gfx35_DestroyIndirectCommandsLayoutNV,
    .CmdExecuteGeneratedCommandsEXT = gfx35_CmdExecuteGeneratedCommandsEXT,
    .CmdPreprocessGeneratedCommandsEXT = gfx35_CmdPreprocessGeneratedCommandsEXT,
    .GetGeneratedCommandsMemoryRequirementsEXT = gfx35_GetGeneratedCommandsMemoryRequirementsEXT,
    .CreateIndirectCommandsLayoutEXT = gfx35_CreateIndirectCommandsLayoutEXT,
    .DestroyIndirectCommandsLayoutEXT = gfx35_DestroyIndirectCommandsLayoutEXT,
    .CreateIndirectExecutionSetEXT = gfx35_CreateIndirectExecutionSetEXT,
    .DestroyIndirectExecutionSetEXT = gfx35_DestroyIndirectExecutionSetEXT,
    .UpdateIndirectExecutionSetPipelineEXT = gfx35_UpdateIndirectExecutionSetPipelineEXT,
    .UpdateIndirectExecutionSetShaderEXT = gfx35_UpdateIndirectExecutionSetShaderEXT,
    .CmdPushDescriptorSet = gfx35_CmdPushDescriptorSet,
    .CmdPushDescriptorSetKHR = gfx35_CmdPushDescriptorSetKHR,
    .TrimCommandPool = gfx35_TrimCommandPool,
    .TrimCommandPoolKHR = gfx35_TrimCommandPoolKHR,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleKHR = gfx35_GetMemoryWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandlePropertiesKHR = gfx35_GetMemoryWin32HandlePropertiesKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandlePropertiesKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryFdKHR = gfx35_GetMemoryFdKHR,
    .GetMemoryFdPropertiesKHR = gfx35_GetMemoryFdPropertiesKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandleFUCHSIA = gfx35_GetMemoryZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandlePropertiesFUCHSIA = gfx35_GetMemoryZirconHandlePropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandlePropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .GetMemoryRemoteAddressNV = gfx35_GetMemoryRemoteAddressNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreWin32HandleKHR = gfx35_GetSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportSemaphoreWin32HandleKHR = gfx35_ImportSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreFdKHR = gfx35_GetSemaphoreFdKHR,
    .ImportSemaphoreFdKHR = gfx35_ImportSemaphoreFdKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetSemaphoreZirconHandleFUCHSIA = gfx35_GetSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .ImportSemaphoreZirconHandleFUCHSIA = gfx35_ImportSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .ImportSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetFenceWin32HandleKHR = gfx35_GetFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportFenceWin32HandleKHR = gfx35_ImportFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetFenceFdKHR = gfx35_GetFenceFdKHR,
    .ImportFenceFdKHR = gfx35_ImportFenceFdKHR,
    .DisplayPowerControlEXT = gfx35_DisplayPowerControlEXT,
    .RegisterDeviceEventEXT = gfx35_RegisterDeviceEventEXT,
    .RegisterDisplayEventEXT = gfx35_RegisterDisplayEventEXT,
    .GetSwapchainCounterEXT = gfx35_GetSwapchainCounterEXT,
    .GetDeviceGroupPeerMemoryFeatures = gfx35_GetDeviceGroupPeerMemoryFeatures,
    .GetDeviceGroupPeerMemoryFeaturesKHR = gfx35_GetDeviceGroupPeerMemoryFeaturesKHR,
    .BindBufferMemory2 = gfx35_BindBufferMemory2,
    .BindBufferMemory2KHR = gfx35_BindBufferMemory2KHR,
    .BindImageMemory2 = gfx35_BindImageMemory2,
    .BindImageMemory2KHR = gfx35_BindImageMemory2KHR,
    .CmdSetDeviceMask = gfx35_CmdSetDeviceMask,
    .CmdSetDeviceMaskKHR = gfx35_CmdSetDeviceMaskKHR,
    .GetDeviceGroupPresentCapabilitiesKHR = gfx35_GetDeviceGroupPresentCapabilitiesKHR,
    .GetDeviceGroupSurfacePresentModesKHR = gfx35_GetDeviceGroupSurfacePresentModesKHR,
    .AcquireNextImage2KHR = gfx35_AcquireNextImage2KHR,
    .CmdDispatchBase = gfx35_CmdDispatchBase,
    .CmdDispatchBaseKHR = gfx35_CmdDispatchBaseKHR,
    .CreateDescriptorUpdateTemplate = gfx35_CreateDescriptorUpdateTemplate,
    .CreateDescriptorUpdateTemplateKHR = gfx35_CreateDescriptorUpdateTemplateKHR,
    .DestroyDescriptorUpdateTemplate = gfx35_DestroyDescriptorUpdateTemplate,
    .DestroyDescriptorUpdateTemplateKHR = gfx35_DestroyDescriptorUpdateTemplateKHR,
    .UpdateDescriptorSetWithTemplate = gfx35_UpdateDescriptorSetWithTemplate,
    .UpdateDescriptorSetWithTemplateKHR = gfx35_UpdateDescriptorSetWithTemplateKHR,
    .CmdPushDescriptorSetWithTemplate = gfx35_CmdPushDescriptorSetWithTemplate,
    .CmdPushDescriptorSetWithTemplateKHR = gfx35_CmdPushDescriptorSetWithTemplateKHR,
    .SetHdrMetadataEXT = gfx35_SetHdrMetadataEXT,
    .GetSwapchainStatusKHR = gfx35_GetSwapchainStatusKHR,
    .GetRefreshCycleDurationGOOGLE = gfx35_GetRefreshCycleDurationGOOGLE,
    .GetPastPresentationTimingGOOGLE = gfx35_GetPastPresentationTimingGOOGLE,
    .CmdSetViewportWScalingNV = gfx35_CmdSetViewportWScalingNV,
    .CmdSetDiscardRectangleEXT = gfx35_CmdSetDiscardRectangleEXT,
    .CmdSetDiscardRectangleEnableEXT = gfx35_CmdSetDiscardRectangleEnableEXT,
    .CmdSetDiscardRectangleModeEXT = gfx35_CmdSetDiscardRectangleModeEXT,
    .CmdSetSampleLocationsEXT = gfx35_CmdSetSampleLocationsEXT,
    .GetBufferMemoryRequirements2 = gfx35_GetBufferMemoryRequirements2,
    .GetBufferMemoryRequirements2KHR = gfx35_GetBufferMemoryRequirements2KHR,
    .GetImageMemoryRequirements2 = gfx35_GetImageMemoryRequirements2,
    .GetImageMemoryRequirements2KHR = gfx35_GetImageMemoryRequirements2KHR,
    .GetImageSparseMemoryRequirements2 = gfx35_GetImageSparseMemoryRequirements2,
    .GetImageSparseMemoryRequirements2KHR = gfx35_GetImageSparseMemoryRequirements2KHR,
    .GetDeviceBufferMemoryRequirements = gfx35_GetDeviceBufferMemoryRequirements,
    .GetDeviceBufferMemoryRequirementsKHR = gfx35_GetDeviceBufferMemoryRequirementsKHR,
    .GetDeviceImageMemoryRequirements = gfx35_GetDeviceImageMemoryRequirements,
    .GetDeviceImageMemoryRequirementsKHR = gfx35_GetDeviceImageMemoryRequirementsKHR,
    .GetDeviceImageSparseMemoryRequirements = gfx35_GetDeviceImageSparseMemoryRequirements,
    .GetDeviceImageSparseMemoryRequirementsKHR = gfx35_GetDeviceImageSparseMemoryRequirementsKHR,
    .CreateSamplerYcbcrConversion = gfx35_CreateSamplerYcbcrConversion,
    .CreateSamplerYcbcrConversionKHR = gfx35_CreateSamplerYcbcrConversionKHR,
    .DestroySamplerYcbcrConversion = gfx35_DestroySamplerYcbcrConversion,
    .DestroySamplerYcbcrConversionKHR = gfx35_DestroySamplerYcbcrConversionKHR,
    .GetDeviceQueue2 = gfx35_GetDeviceQueue2,
    .CreateValidationCacheEXT = gfx35_CreateValidationCacheEXT,
    .DestroyValidationCacheEXT = gfx35_DestroyValidationCacheEXT,
    .GetValidationCacheDataEXT = gfx35_GetValidationCacheDataEXT,
    .MergeValidationCachesEXT = gfx35_MergeValidationCachesEXT,
    .GetDescriptorSetLayoutSupport = gfx35_GetDescriptorSetLayoutSupport,
    .GetDescriptorSetLayoutSupportKHR = gfx35_GetDescriptorSetLayoutSupportKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsageANDROID = gfx35_GetSwapchainGrallocUsageANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsage2ANDROID = gfx35_GetSwapchainGrallocUsage2ANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsage2ANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .AcquireImageANDROID = gfx35_AcquireImageANDROID,
#elif defined(_MSC_VER)
    .AcquireImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .QueueSignalReleaseImageANDROID = gfx35_QueueSignalReleaseImageANDROID,
#elif defined(_MSC_VER)
    .QueueSignalReleaseImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .GetShaderInfoAMD = gfx35_GetShaderInfoAMD,
    .SetLocalDimmingAMD = gfx35_SetLocalDimmingAMD,
    .GetCalibratedTimestampsKHR = gfx35_GetCalibratedTimestampsKHR,
    .GetCalibratedTimestampsEXT = gfx35_GetCalibratedTimestampsEXT,
    .SetDebugUtilsObjectNameEXT = gfx35_SetDebugUtilsObjectNameEXT,
    .SetDebugUtilsObjectTagEXT = gfx35_SetDebugUtilsObjectTagEXT,
    .QueueBeginDebugUtilsLabelEXT = gfx35_QueueBeginDebugUtilsLabelEXT,
    .QueueEndDebugUtilsLabelEXT = gfx35_QueueEndDebugUtilsLabelEXT,
    .QueueInsertDebugUtilsLabelEXT = gfx35_QueueInsertDebugUtilsLabelEXT,
    .CmdBeginDebugUtilsLabelEXT = gfx35_CmdBeginDebugUtilsLabelEXT,
    .CmdEndDebugUtilsLabelEXT = gfx35_CmdEndDebugUtilsLabelEXT,
    .CmdInsertDebugUtilsLabelEXT = gfx35_CmdInsertDebugUtilsLabelEXT,
    .GetMemoryHostPointerPropertiesEXT = gfx35_GetMemoryHostPointerPropertiesEXT,
    .CmdWriteBufferMarkerAMD = gfx35_CmdWriteBufferMarkerAMD,
    .CreateRenderPass2 = gfx35_CreateRenderPass2,
    .CreateRenderPass2KHR = gfx35_CreateRenderPass2KHR,
    .CmdBeginRenderPass2 = gfx35_CmdBeginRenderPass2,
    .CmdBeginRenderPass2KHR = gfx35_CmdBeginRenderPass2KHR,
    .CmdNextSubpass2 = gfx35_CmdNextSubpass2,
    .CmdNextSubpass2KHR = gfx35_CmdNextSubpass2KHR,
    .CmdEndRenderPass2 = gfx35_CmdEndRenderPass2,
    .CmdEndRenderPass2KHR = gfx35_CmdEndRenderPass2KHR,
    .GetSemaphoreCounterValue = gfx35_GetSemaphoreCounterValue,
    .GetSemaphoreCounterValueKHR = gfx35_GetSemaphoreCounterValueKHR,
    .WaitSemaphores = gfx35_WaitSemaphores,
    .WaitSemaphoresKHR = gfx35_WaitSemaphoresKHR,
    .SignalSemaphore = gfx35_SignalSemaphore,
    .SignalSemaphoreKHR = gfx35_SignalSemaphoreKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetAndroidHardwareBufferPropertiesANDROID = gfx35_GetAndroidHardwareBufferPropertiesANDROID,
#elif defined(_MSC_VER)
    .GetAndroidHardwareBufferPropertiesANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetMemoryAndroidHardwareBufferANDROID = gfx35_GetMemoryAndroidHardwareBufferANDROID,
#elif defined(_MSC_VER)
    .GetMemoryAndroidHardwareBufferANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .CmdDrawIndirectCount = gfx35_CmdDrawIndirectCount,
    .CmdDrawIndirectCountKHR = gfx35_CmdDrawIndirectCountKHR,
    .CmdDrawIndirectCountAMD = gfx35_CmdDrawIndirectCountAMD,
    .CmdDrawIndexedIndirectCount = gfx35_CmdDrawIndexedIndirectCount,
    .CmdDrawIndexedIndirectCountKHR = gfx35_CmdDrawIndexedIndirectCountKHR,
    .CmdDrawIndexedIndirectCountAMD = gfx35_CmdDrawIndexedIndirectCountAMD,
    .CmdSetCheckpointNV = gfx35_CmdSetCheckpointNV,
    .GetQueueCheckpointDataNV = gfx35_GetQueueCheckpointDataNV,
    .CmdBindTransformFeedbackBuffersEXT = gfx35_CmdBindTransformFeedbackBuffersEXT,
    .CmdBeginTransformFeedbackEXT = gfx35_CmdBeginTransformFeedbackEXT,
    .CmdEndTransformFeedbackEXT = gfx35_CmdEndTransformFeedbackEXT,
    .CmdBeginQueryIndexedEXT = gfx35_CmdBeginQueryIndexedEXT,
    .CmdEndQueryIndexedEXT = gfx35_CmdEndQueryIndexedEXT,
    .CmdDrawIndirectByteCountEXT = gfx35_CmdDrawIndirectByteCountEXT,
    .CmdSetExclusiveScissorNV = gfx35_CmdSetExclusiveScissorNV,
    .CmdSetExclusiveScissorEnableNV = gfx35_CmdSetExclusiveScissorEnableNV,
    .CmdBindShadingRateImageNV = gfx35_CmdBindShadingRateImageNV,
    .CmdSetViewportShadingRatePaletteNV = gfx35_CmdSetViewportShadingRatePaletteNV,
    .CmdSetCoarseSampleOrderNV = gfx35_CmdSetCoarseSampleOrderNV,
    .CmdDrawMeshTasksNV = gfx35_CmdDrawMeshTasksNV,
    .CmdDrawMeshTasksIndirectNV = gfx35_CmdDrawMeshTasksIndirectNV,
    .CmdDrawMeshTasksIndirectCountNV = gfx35_CmdDrawMeshTasksIndirectCountNV,
    .CmdDrawMeshTasksEXT = gfx35_CmdDrawMeshTasksEXT,
    .CmdDrawMeshTasksIndirectEXT = gfx35_CmdDrawMeshTasksIndirectEXT,
    .CmdDrawMeshTasksIndirectCountEXT = gfx35_CmdDrawMeshTasksIndirectCountEXT,
    .CompileDeferredNV = gfx35_CompileDeferredNV,
    .CreateAccelerationStructureNV = gfx35_CreateAccelerationStructureNV,
    .CmdBindInvocationMaskHUAWEI = gfx35_CmdBindInvocationMaskHUAWEI,
    .DestroyAccelerationStructureKHR = gfx35_DestroyAccelerationStructureKHR,
    .DestroyAccelerationStructureNV = gfx35_DestroyAccelerationStructureNV,
    .GetAccelerationStructureMemoryRequirementsNV = gfx35_GetAccelerationStructureMemoryRequirementsNV,
    .BindAccelerationStructureMemoryNV = gfx35_BindAccelerationStructureMemoryNV,
    .CmdCopyAccelerationStructureNV = gfx35_CmdCopyAccelerationStructureNV,
    .CmdCopyAccelerationStructureKHR = gfx35_CmdCopyAccelerationStructureKHR,
    .CopyAccelerationStructureKHR = gfx35_CopyAccelerationStructureKHR,
    .CmdCopyAccelerationStructureToMemoryKHR = gfx35_CmdCopyAccelerationStructureToMemoryKHR,
    .CopyAccelerationStructureToMemoryKHR = gfx35_CopyAccelerationStructureToMemoryKHR,
    .CmdCopyMemoryToAccelerationStructureKHR = gfx35_CmdCopyMemoryToAccelerationStructureKHR,
    .CopyMemoryToAccelerationStructureKHR = gfx35_CopyMemoryToAccelerationStructureKHR,
    .CmdWriteAccelerationStructuresPropertiesKHR = gfx35_CmdWriteAccelerationStructuresPropertiesKHR,
    .CmdWriteAccelerationStructuresPropertiesNV = gfx35_CmdWriteAccelerationStructuresPropertiesNV,
    .CmdBuildAccelerationStructureNV = gfx35_CmdBuildAccelerationStructureNV,
    .WriteAccelerationStructuresPropertiesKHR = gfx35_WriteAccelerationStructuresPropertiesKHR,
    .CmdTraceRaysKHR = gfx35_CmdTraceRaysKHR,
    .CmdTraceRaysNV = gfx35_CmdTraceRaysNV,
    .GetRayTracingShaderGroupHandlesKHR = gfx35_GetRayTracingShaderGroupHandlesKHR,
    .GetRayTracingShaderGroupHandlesNV = gfx35_GetRayTracingShaderGroupHandlesNV,
    .GetRayTracingCaptureReplayShaderGroupHandlesKHR = gfx35_GetRayTracingCaptureReplayShaderGroupHandlesKHR,
    .GetAccelerationStructureHandleNV = gfx35_GetAccelerationStructureHandleNV,
    .CreateRayTracingPipelinesNV = gfx35_CreateRayTracingPipelinesNV,
    .CreateRayTracingPipelinesKHR = gfx35_CreateRayTracingPipelinesKHR,
    .CmdTraceRaysIndirectKHR = gfx35_CmdTraceRaysIndirectKHR,
    .CmdTraceRaysIndirect2KHR = gfx35_CmdTraceRaysIndirect2KHR,
    .GetClusterAccelerationStructureBuildSizesNV = gfx35_GetClusterAccelerationStructureBuildSizesNV,
    .CmdBuildClusterAccelerationStructureIndirectNV = gfx35_CmdBuildClusterAccelerationStructureIndirectNV,
    .GetDeviceAccelerationStructureCompatibilityKHR = gfx35_GetDeviceAccelerationStructureCompatibilityKHR,
    .GetRayTracingShaderGroupStackSizeKHR = gfx35_GetRayTracingShaderGroupStackSizeKHR,
    .CmdSetRayTracingPipelineStackSizeKHR = gfx35_CmdSetRayTracingPipelineStackSizeKHR,
    .GetImageViewHandleNVX = gfx35_GetImageViewHandleNVX,
    .GetImageViewHandle64NVX = gfx35_GetImageViewHandle64NVX,
    .GetImageViewAddressNVX = gfx35_GetImageViewAddressNVX,
    .GetDeviceCombinedImageSamplerIndexNVX = gfx35_GetDeviceCombinedImageSamplerIndexNVX,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetDeviceGroupSurfacePresentModes2EXT = gfx35_GetDeviceGroupSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetDeviceGroupSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireFullScreenExclusiveModeEXT = gfx35_AcquireFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .AcquireFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ReleaseFullScreenExclusiveModeEXT = gfx35_ReleaseFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .ReleaseFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .AcquireProfilingLockKHR = gfx35_AcquireProfilingLockKHR,
    .ReleaseProfilingLockKHR = gfx35_ReleaseProfilingLockKHR,
    .GetImageDrmFormatModifierPropertiesEXT = gfx35_GetImageDrmFormatModifierPropertiesEXT,
    .GetBufferOpaqueCaptureAddress = gfx35_GetBufferOpaqueCaptureAddress,
    .GetBufferOpaqueCaptureAddressKHR = gfx35_GetBufferOpaqueCaptureAddressKHR,
    .GetBufferDeviceAddress = gfx35_GetBufferDeviceAddress,
    .GetBufferDeviceAddressKHR = gfx35_GetBufferDeviceAddressKHR,
    .GetBufferDeviceAddressEXT = gfx35_GetBufferDeviceAddressEXT,
    .InitializePerformanceApiINTEL = gfx35_InitializePerformanceApiINTEL,
    .UninitializePerformanceApiINTEL = gfx35_UninitializePerformanceApiINTEL,
    .CmdSetPerformanceMarkerINTEL = gfx35_CmdSetPerformanceMarkerINTEL,
    .CmdSetPerformanceStreamMarkerINTEL = gfx35_CmdSetPerformanceStreamMarkerINTEL,
    .CmdSetPerformanceOverrideINTEL = gfx35_CmdSetPerformanceOverrideINTEL,
    .AcquirePerformanceConfigurationINTEL = gfx35_AcquirePerformanceConfigurationINTEL,
    .ReleasePerformanceConfigurationINTEL = gfx35_ReleasePerformanceConfigurationINTEL,
    .QueueSetPerformanceConfigurationINTEL = gfx35_QueueSetPerformanceConfigurationINTEL,
    .GetPerformanceParameterINTEL = gfx35_GetPerformanceParameterINTEL,
    .GetDeviceMemoryOpaqueCaptureAddress = gfx35_GetDeviceMemoryOpaqueCaptureAddress,
    .GetDeviceMemoryOpaqueCaptureAddressKHR = gfx35_GetDeviceMemoryOpaqueCaptureAddressKHR,
    .GetPipelineExecutablePropertiesKHR = gfx35_GetPipelineExecutablePropertiesKHR,
    .GetPipelineExecutableStatisticsKHR = gfx35_GetPipelineExecutableStatisticsKHR,
    .GetPipelineExecutableInternalRepresentationsKHR = gfx35_GetPipelineExecutableInternalRepresentationsKHR,
    .CmdSetLineStipple = gfx35_CmdSetLineStipple,
    .CmdSetLineStippleKHR = gfx35_CmdSetLineStippleKHR,
    .CmdSetLineStippleEXT = gfx35_CmdSetLineStippleEXT,
    .CreateAccelerationStructureKHR = gfx35_CreateAccelerationStructureKHR,
    .CmdBuildAccelerationStructuresKHR = gfx35_CmdBuildAccelerationStructuresKHR,
    .CmdBuildAccelerationStructuresIndirectKHR = gfx35_CmdBuildAccelerationStructuresIndirectKHR,
    .BuildAccelerationStructuresKHR = gfx35_BuildAccelerationStructuresKHR,
    .GetAccelerationStructureDeviceAddressKHR = gfx35_GetAccelerationStructureDeviceAddressKHR,
    .CreateDeferredOperationKHR = gfx35_CreateDeferredOperationKHR,
    .DestroyDeferredOperationKHR = gfx35_DestroyDeferredOperationKHR,
    .GetDeferredOperationMaxConcurrencyKHR = gfx35_GetDeferredOperationMaxConcurrencyKHR,
    .GetDeferredOperationResultKHR = gfx35_GetDeferredOperationResultKHR,
    .DeferredOperationJoinKHR = gfx35_DeferredOperationJoinKHR,
    .GetPipelineIndirectMemoryRequirementsNV = gfx35_GetPipelineIndirectMemoryRequirementsNV,
    .GetPipelineIndirectDeviceAddressNV = gfx35_GetPipelineIndirectDeviceAddressNV,
    .AntiLagUpdateAMD = gfx35_AntiLagUpdateAMD,
    .CmdSetCullMode = gfx35_CmdSetCullMode,
    .CmdSetCullModeEXT = gfx35_CmdSetCullModeEXT,
    .CmdSetFrontFace = gfx35_CmdSetFrontFace,
    .CmdSetFrontFaceEXT = gfx35_CmdSetFrontFaceEXT,
    .CmdSetPrimitiveTopology = gfx35_CmdSetPrimitiveTopology,
    .CmdSetPrimitiveTopologyEXT = gfx35_CmdSetPrimitiveTopologyEXT,
    .CmdSetViewportWithCount = gfx35_CmdSetViewportWithCount,
    .CmdSetViewportWithCountEXT = gfx35_CmdSetViewportWithCountEXT,
    .CmdSetScissorWithCount = gfx35_CmdSetScissorWithCount,
    .CmdSetScissorWithCountEXT = gfx35_CmdSetScissorWithCountEXT,
    .CmdBindIndexBuffer2 = gfx35_CmdBindIndexBuffer2,
    .CmdBindIndexBuffer2KHR = gfx35_CmdBindIndexBuffer2KHR,
    .CmdBindVertexBuffers2 = gfx35_CmdBindVertexBuffers2,
    .CmdBindVertexBuffers2EXT = gfx35_CmdBindVertexBuffers2EXT,
    .CmdSetDepthTestEnable = gfx35_CmdSetDepthTestEnable,
    .CmdSetDepthTestEnableEXT = gfx35_CmdSetDepthTestEnableEXT,
    .CmdSetDepthWriteEnable = gfx35_CmdSetDepthWriteEnable,
    .CmdSetDepthWriteEnableEXT = gfx35_CmdSetDepthWriteEnableEXT,
    .CmdSetDepthCompareOp = gfx35_CmdSetDepthCompareOp,
    .CmdSetDepthCompareOpEXT = gfx35_CmdSetDepthCompareOpEXT,
    .CmdSetDepthBoundsTestEnable = gfx35_CmdSetDepthBoundsTestEnable,
    .CmdSetDepthBoundsTestEnableEXT = gfx35_CmdSetDepthBoundsTestEnableEXT,
    .CmdSetStencilTestEnable = gfx35_CmdSetStencilTestEnable,
    .CmdSetStencilTestEnableEXT = gfx35_CmdSetStencilTestEnableEXT,
    .CmdSetStencilOp = gfx35_CmdSetStencilOp,
    .CmdSetStencilOpEXT = gfx35_CmdSetStencilOpEXT,
    .CmdSetPatchControlPointsEXT = gfx35_CmdSetPatchControlPointsEXT,
    .CmdSetRasterizerDiscardEnable = gfx35_CmdSetRasterizerDiscardEnable,
    .CmdSetRasterizerDiscardEnableEXT = gfx35_CmdSetRasterizerDiscardEnableEXT,
    .CmdSetDepthBiasEnable = gfx35_CmdSetDepthBiasEnable,
    .CmdSetDepthBiasEnableEXT = gfx35_CmdSetDepthBiasEnableEXT,
    .CmdSetLogicOpEXT = gfx35_CmdSetLogicOpEXT,
    .CmdSetPrimitiveRestartEnable = gfx35_CmdSetPrimitiveRestartEnable,
    .CmdSetPrimitiveRestartEnableEXT = gfx35_CmdSetPrimitiveRestartEnableEXT,
    .CmdSetTessellationDomainOriginEXT = gfx35_CmdSetTessellationDomainOriginEXT,
    .CmdSetDepthClampEnableEXT = gfx35_CmdSetDepthClampEnableEXT,
    .CmdSetPolygonModeEXT = gfx35_CmdSetPolygonModeEXT,
    .CmdSetRasterizationSamplesEXT = gfx35_CmdSetRasterizationSamplesEXT,
    .CmdSetSampleMaskEXT = gfx35_CmdSetSampleMaskEXT,
    .CmdSetAlphaToCoverageEnableEXT = gfx35_CmdSetAlphaToCoverageEnableEXT,
    .CmdSetAlphaToOneEnableEXT = gfx35_CmdSetAlphaToOneEnableEXT,
    .CmdSetLogicOpEnableEXT = gfx35_CmdSetLogicOpEnableEXT,
    .CmdSetColorBlendEnableEXT = gfx35_CmdSetColorBlendEnableEXT,
    .CmdSetColorBlendEquationEXT = gfx35_CmdSetColorBlendEquationEXT,
    .CmdSetColorWriteMaskEXT = gfx35_CmdSetColorWriteMaskEXT,
    .CmdSetRasterizationStreamEXT = gfx35_CmdSetRasterizationStreamEXT,
    .CmdSetConservativeRasterizationModeEXT = gfx35_CmdSetConservativeRasterizationModeEXT,
    .CmdSetExtraPrimitiveOverestimationSizeEXT = gfx35_CmdSetExtraPrimitiveOverestimationSizeEXT,
    .CmdSetDepthClipEnableEXT = gfx35_CmdSetDepthClipEnableEXT,
    .CmdSetSampleLocationsEnableEXT = gfx35_CmdSetSampleLocationsEnableEXT,
    .CmdSetColorBlendAdvancedEXT = gfx35_CmdSetColorBlendAdvancedEXT,
    .CmdSetProvokingVertexModeEXT = gfx35_CmdSetProvokingVertexModeEXT,
    .CmdSetLineRasterizationModeEXT = gfx35_CmdSetLineRasterizationModeEXT,
    .CmdSetLineStippleEnableEXT = gfx35_CmdSetLineStippleEnableEXT,
    .CmdSetDepthClipNegativeOneToOneEXT = gfx35_CmdSetDepthClipNegativeOneToOneEXT,
    .CmdSetViewportWScalingEnableNV = gfx35_CmdSetViewportWScalingEnableNV,
    .CmdSetViewportSwizzleNV = gfx35_CmdSetViewportSwizzleNV,
    .CmdSetCoverageToColorEnableNV = gfx35_CmdSetCoverageToColorEnableNV,
    .CmdSetCoverageToColorLocationNV = gfx35_CmdSetCoverageToColorLocationNV,
    .CmdSetCoverageModulationModeNV = gfx35_CmdSetCoverageModulationModeNV,
    .CmdSetCoverageModulationTableEnableNV = gfx35_CmdSetCoverageModulationTableEnableNV,
    .CmdSetCoverageModulationTableNV = gfx35_CmdSetCoverageModulationTableNV,
    .CmdSetShadingRateImageEnableNV = gfx35_CmdSetShadingRateImageEnableNV,
    .CmdSetCoverageReductionModeNV = gfx35_CmdSetCoverageReductionModeNV,
    .CmdSetRepresentativeFragmentTestEnableNV = gfx35_CmdSetRepresentativeFragmentTestEnableNV,
    .CreatePrivateDataSlot = gfx35_CreatePrivateDataSlot,
    .CreatePrivateDataSlotEXT = gfx35_CreatePrivateDataSlotEXT,
    .DestroyPrivateDataSlot = gfx35_DestroyPrivateDataSlot,
    .DestroyPrivateDataSlotEXT = gfx35_DestroyPrivateDataSlotEXT,
    .SetPrivateData = gfx35_SetPrivateData,
    .SetPrivateDataEXT = gfx35_SetPrivateDataEXT,
    .GetPrivateData = gfx35_GetPrivateData,
    .GetPrivateDataEXT = gfx35_GetPrivateDataEXT,
    .CmdCopyBuffer2 = gfx35_CmdCopyBuffer2,
    .CmdCopyBuffer2KHR = gfx35_CmdCopyBuffer2KHR,
    .CmdCopyImage2 = gfx35_CmdCopyImage2,
    .CmdCopyImage2KHR = gfx35_CmdCopyImage2KHR,
    .CmdBlitImage2 = gfx35_CmdBlitImage2,
    .CmdBlitImage2KHR = gfx35_CmdBlitImage2KHR,
    .CmdCopyBufferToImage2 = gfx35_CmdCopyBufferToImage2,
    .CmdCopyBufferToImage2KHR = gfx35_CmdCopyBufferToImage2KHR,
    .CmdCopyImageToBuffer2 = gfx35_CmdCopyImageToBuffer2,
    .CmdCopyImageToBuffer2KHR = gfx35_CmdCopyImageToBuffer2KHR,
    .CmdResolveImage2 = gfx35_CmdResolveImage2,
    .CmdResolveImage2KHR = gfx35_CmdResolveImage2KHR,
    .CmdSetFragmentShadingRateKHR = gfx35_CmdSetFragmentShadingRateKHR,
    .CmdSetFragmentShadingRateEnumNV = gfx35_CmdSetFragmentShadingRateEnumNV,
    .GetAccelerationStructureBuildSizesKHR = gfx35_GetAccelerationStructureBuildSizesKHR,
    .CmdSetVertexInputEXT = gfx35_CmdSetVertexInputEXT,
    .CmdSetColorWriteEnableEXT = gfx35_CmdSetColorWriteEnableEXT,
    .CmdSetEvent2 = gfx35_CmdSetEvent2,
    .CmdSetEvent2KHR = gfx35_CmdSetEvent2KHR,
    .CmdResetEvent2 = gfx35_CmdResetEvent2,
    .CmdResetEvent2KHR = gfx35_CmdResetEvent2KHR,
    .CmdWaitEvents2 = gfx35_CmdWaitEvents2,
    .CmdWaitEvents2KHR = gfx35_CmdWaitEvents2KHR,
    .CmdPipelineBarrier2 = gfx35_CmdPipelineBarrier2,
    .CmdPipelineBarrier2KHR = gfx35_CmdPipelineBarrier2KHR,
    .QueueSubmit2 = gfx35_QueueSubmit2,
    .QueueSubmit2KHR = gfx35_QueueSubmit2KHR,
    .CmdWriteTimestamp2 = gfx35_CmdWriteTimestamp2,
    .CmdWriteTimestamp2KHR = gfx35_CmdWriteTimestamp2KHR,
    .CmdWriteBufferMarker2AMD = gfx35_CmdWriteBufferMarker2AMD,
    .GetQueueCheckpointData2NV = gfx35_GetQueueCheckpointData2NV,
    .CopyMemoryToImage = gfx35_CopyMemoryToImage,
    .CopyMemoryToImageEXT = gfx35_CopyMemoryToImageEXT,
    .CopyImageToMemory = gfx35_CopyImageToMemory,
    .CopyImageToMemoryEXT = gfx35_CopyImageToMemoryEXT,
    .CopyImageToImage = gfx35_CopyImageToImage,
    .CopyImageToImageEXT = gfx35_CopyImageToImageEXT,
    .TransitionImageLayout = gfx35_TransitionImageLayout,
    .TransitionImageLayoutEXT = gfx35_TransitionImageLayoutEXT,
    .CreateVideoSessionKHR = gfx35_CreateVideoSessionKHR,
    .DestroyVideoSessionKHR = gfx35_DestroyVideoSessionKHR,
    .CreateVideoSessionParametersKHR = gfx35_CreateVideoSessionParametersKHR,
    .UpdateVideoSessionParametersKHR = gfx35_UpdateVideoSessionParametersKHR,
    .GetEncodedVideoSessionParametersKHR = gfx35_GetEncodedVideoSessionParametersKHR,
    .DestroyVideoSessionParametersKHR = gfx35_DestroyVideoSessionParametersKHR,
    .GetVideoSessionMemoryRequirementsKHR = gfx35_GetVideoSessionMemoryRequirementsKHR,
    .BindVideoSessionMemoryKHR = gfx35_BindVideoSessionMemoryKHR,
    .CmdDecodeVideoKHR = gfx35_CmdDecodeVideoKHR,
    .CmdBeginVideoCodingKHR = gfx35_CmdBeginVideoCodingKHR,
    .CmdControlVideoCodingKHR = gfx35_CmdControlVideoCodingKHR,
    .CmdEndVideoCodingKHR = gfx35_CmdEndVideoCodingKHR,
    .CmdEncodeVideoKHR = gfx35_CmdEncodeVideoKHR,
    .CmdDecompressMemoryNV = gfx35_CmdDecompressMemoryNV,
    .CmdDecompressMemoryIndirectCountNV = gfx35_CmdDecompressMemoryIndirectCountNV,
    .GetPartitionedAccelerationStructuresBuildSizesNV = gfx35_GetPartitionedAccelerationStructuresBuildSizesNV,
    .CmdBuildPartitionedAccelerationStructuresNV = gfx35_CmdBuildPartitionedAccelerationStructuresNV,
    .CmdDecompressMemoryEXT = gfx35_CmdDecompressMemoryEXT,
    .CmdDecompressMemoryIndirectCountEXT = gfx35_CmdDecompressMemoryIndirectCountEXT,
    .CreateCuModuleNVX = gfx35_CreateCuModuleNVX,
    .CreateCuFunctionNVX = gfx35_CreateCuFunctionNVX,
    .DestroyCuModuleNVX = gfx35_DestroyCuModuleNVX,
    .DestroyCuFunctionNVX = gfx35_DestroyCuFunctionNVX,
    .CmdCuLaunchKernelNVX = gfx35_CmdCuLaunchKernelNVX,
    .GetDescriptorSetLayoutSizeEXT = gfx35_GetDescriptorSetLayoutSizeEXT,
    .GetDescriptorSetLayoutBindingOffsetEXT = gfx35_GetDescriptorSetLayoutBindingOffsetEXT,
    .GetDescriptorEXT = gfx35_GetDescriptorEXT,
    .CmdBindDescriptorBuffersEXT = gfx35_CmdBindDescriptorBuffersEXT,
    .CmdSetDescriptorBufferOffsetsEXT = gfx35_CmdSetDescriptorBufferOffsetsEXT,
    .CmdBindDescriptorBufferEmbeddedSamplersEXT = gfx35_CmdBindDescriptorBufferEmbeddedSamplersEXT,
    .GetBufferOpaqueCaptureDescriptorDataEXT = gfx35_GetBufferOpaqueCaptureDescriptorDataEXT,
    .GetImageOpaqueCaptureDescriptorDataEXT = gfx35_GetImageOpaqueCaptureDescriptorDataEXT,
    .GetImageViewOpaqueCaptureDescriptorDataEXT = gfx35_GetImageViewOpaqueCaptureDescriptorDataEXT,
    .GetSamplerOpaqueCaptureDescriptorDataEXT = gfx35_GetSamplerOpaqueCaptureDescriptorDataEXT,
    .GetAccelerationStructureOpaqueCaptureDescriptorDataEXT = gfx35_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT,
    .SetDeviceMemoryPriorityEXT = gfx35_SetDeviceMemoryPriorityEXT,
    .WaitForPresent2KHR = gfx35_WaitForPresent2KHR,
    .WaitForPresentKHR = gfx35_WaitForPresentKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateBufferCollectionFUCHSIA = gfx35_CreateBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .CreateBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionBufferConstraintsFUCHSIA = gfx35_SetBufferCollectionBufferConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionBufferConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionImageConstraintsFUCHSIA = gfx35_SetBufferCollectionImageConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionImageConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .DestroyBufferCollectionFUCHSIA = gfx35_DestroyBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .DestroyBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetBufferCollectionPropertiesFUCHSIA = gfx35_GetBufferCollectionPropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetBufferCollectionPropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .CmdBeginRendering = gfx35_CmdBeginRendering,
    .CmdBeginRenderingKHR = gfx35_CmdBeginRenderingKHR,
    .CmdEndRendering = gfx35_CmdEndRendering,
    .CmdEndRendering2KHR = gfx35_CmdEndRendering2KHR,
    .CmdEndRendering2EXT = gfx35_CmdEndRendering2EXT,
    .CmdEndRenderingKHR = gfx35_CmdEndRenderingKHR,
    .GetDescriptorSetLayoutHostMappingInfoVALVE = gfx35_GetDescriptorSetLayoutHostMappingInfoVALVE,
    .GetDescriptorSetHostMappingVALVE = gfx35_GetDescriptorSetHostMappingVALVE,
    .CreateMicromapEXT = gfx35_CreateMicromapEXT,
    .CmdBuildMicromapsEXT = gfx35_CmdBuildMicromapsEXT,
    .BuildMicromapsEXT = gfx35_BuildMicromapsEXT,
    .DestroyMicromapEXT = gfx35_DestroyMicromapEXT,
    .CmdCopyMicromapEXT = gfx35_CmdCopyMicromapEXT,
    .CopyMicromapEXT = gfx35_CopyMicromapEXT,
    .CmdCopyMicromapToMemoryEXT = gfx35_CmdCopyMicromapToMemoryEXT,
    .CopyMicromapToMemoryEXT = gfx35_CopyMicromapToMemoryEXT,
    .CmdCopyMemoryToMicromapEXT = gfx35_CmdCopyMemoryToMicromapEXT,
    .CopyMemoryToMicromapEXT = gfx35_CopyMemoryToMicromapEXT,
    .CmdWriteMicromapsPropertiesEXT = gfx35_CmdWriteMicromapsPropertiesEXT,
    .WriteMicromapsPropertiesEXT = gfx35_WriteMicromapsPropertiesEXT,
    .GetDeviceMicromapCompatibilityEXT = gfx35_GetDeviceMicromapCompatibilityEXT,
    .GetMicromapBuildSizesEXT = gfx35_GetMicromapBuildSizesEXT,
    .GetShaderModuleIdentifierEXT = gfx35_GetShaderModuleIdentifierEXT,
    .GetShaderModuleCreateInfoIdentifierEXT = gfx35_GetShaderModuleCreateInfoIdentifierEXT,
    .GetImageSubresourceLayout2 = gfx35_GetImageSubresourceLayout2,
    .GetImageSubresourceLayout2KHR = gfx35_GetImageSubresourceLayout2KHR,
    .GetImageSubresourceLayout2EXT = gfx35_GetImageSubresourceLayout2EXT,
    .GetPipelinePropertiesEXT = gfx35_GetPipelinePropertiesEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .ExportMetalObjectsEXT = gfx35_ExportMetalObjectsEXT,
#elif defined(_MSC_VER)
    .ExportMetalObjectsEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CmdBindTileMemoryQCOM = gfx35_CmdBindTileMemoryQCOM,
    .GetFramebufferTilePropertiesQCOM = gfx35_GetFramebufferTilePropertiesQCOM,
    .GetDynamicRenderingTilePropertiesQCOM = gfx35_GetDynamicRenderingTilePropertiesQCOM,
    .CreateOpticalFlowSessionNV = gfx35_CreateOpticalFlowSessionNV,
    .DestroyOpticalFlowSessionNV = gfx35_DestroyOpticalFlowSessionNV,
    .BindOpticalFlowSessionImageNV = gfx35_BindOpticalFlowSessionImageNV,
    .CmdOpticalFlowExecuteNV = gfx35_CmdOpticalFlowExecuteNV,
    .GetDeviceFaultInfoEXT = gfx35_GetDeviceFaultInfoEXT,
    .GetDeviceFaultReportsKHR = gfx35_GetDeviceFaultReportsKHR,
    .GetDeviceFaultDebugInfoKHR = gfx35_GetDeviceFaultDebugInfoKHR,
    .CmdSetDepthBias2EXT = gfx35_CmdSetDepthBias2EXT,
    .ReleaseSwapchainImagesKHR = gfx35_ReleaseSwapchainImagesKHR,
    .ReleaseSwapchainImagesEXT = gfx35_ReleaseSwapchainImagesEXT,
    .GetDeviceImageSubresourceLayout = gfx35_GetDeviceImageSubresourceLayout,
    .GetDeviceImageSubresourceLayoutKHR = gfx35_GetDeviceImageSubresourceLayoutKHR,
    .MapMemory2 = gfx35_MapMemory2,
    .MapMemory2KHR = gfx35_MapMemory2KHR,
    .UnmapMemory2 = gfx35_UnmapMemory2,
    .UnmapMemory2KHR = gfx35_UnmapMemory2KHR,
    .CreateShadersEXT = gfx35_CreateShadersEXT,
    .DestroyShaderEXT = gfx35_DestroyShaderEXT,
    .GetShaderBinaryDataEXT = gfx35_GetShaderBinaryDataEXT,
    .CmdBindShadersEXT = gfx35_CmdBindShadersEXT,
    .SetSwapchainPresentTimingQueueSizeEXT = gfx35_SetSwapchainPresentTimingQueueSizeEXT,
    .GetSwapchainTimingPropertiesEXT = gfx35_GetSwapchainTimingPropertiesEXT,
    .GetSwapchainTimeDomainPropertiesEXT = gfx35_GetSwapchainTimeDomainPropertiesEXT,
    .GetPastPresentationTimingEXT = gfx35_GetPastPresentationTimingEXT,
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetScreenBufferPropertiesQNX = gfx35_GetScreenBufferPropertiesQNX,
#elif defined(_MSC_VER)
    .GetScreenBufferPropertiesQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CmdBindDescriptorSets2 = gfx35_CmdBindDescriptorSets2,
    .CmdBindDescriptorSets2KHR = gfx35_CmdBindDescriptorSets2KHR,
    .CmdPushConstants2 = gfx35_CmdPushConstants2,
    .CmdPushConstants2KHR = gfx35_CmdPushConstants2KHR,
    .CmdPushDescriptorSet2 = gfx35_CmdPushDescriptorSet2,
    .CmdPushDescriptorSet2KHR = gfx35_CmdPushDescriptorSet2KHR,
    .CmdPushDescriptorSetWithTemplate2 = gfx35_CmdPushDescriptorSetWithTemplate2,
    .CmdPushDescriptorSetWithTemplate2KHR = gfx35_CmdPushDescriptorSetWithTemplate2KHR,
    .CmdSetDescriptorBufferOffsets2EXT = gfx35_CmdSetDescriptorBufferOffsets2EXT,
    .CmdBindDescriptorBufferEmbeddedSamplers2EXT = gfx35_CmdBindDescriptorBufferEmbeddedSamplers2EXT,
    .SetLatencySleepModeNV = gfx35_SetLatencySleepModeNV,
    .LatencySleepNV = gfx35_LatencySleepNV,
    .SetLatencyMarkerNV = gfx35_SetLatencyMarkerNV,
    .GetLatencyTimingsNV = gfx35_GetLatencyTimingsNV,
    .QueueNotifyOutOfBandNV = gfx35_QueueNotifyOutOfBandNV,
    .CmdSetRenderingAttachmentLocations = gfx35_CmdSetRenderingAttachmentLocations,
    .CmdSetRenderingAttachmentLocationsKHR = gfx35_CmdSetRenderingAttachmentLocationsKHR,
    .CmdSetRenderingInputAttachmentIndices = gfx35_CmdSetRenderingInputAttachmentIndices,
    .CmdSetRenderingInputAttachmentIndicesKHR = gfx35_CmdSetRenderingInputAttachmentIndicesKHR,
    .CmdSetDepthClampRangeEXT = gfx35_CmdSetDepthClampRangeEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandleEXT = gfx35_GetMemoryMetalHandleEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandleEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandlePropertiesEXT = gfx35_GetMemoryMetalHandlePropertiesEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandlePropertiesEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .ConvertCooperativeVectorMatrixNV = gfx35_ConvertCooperativeVectorMatrixNV,
    .CmdConvertCooperativeVectorMatrixNV = gfx35_CmdConvertCooperativeVectorMatrixNV,
    .CmdDispatchTileQCOM = gfx35_CmdDispatchTileQCOM,
    .CmdBeginPerTileExecutionQCOM = gfx35_CmdBeginPerTileExecutionQCOM,
    .CmdEndPerTileExecutionQCOM = gfx35_CmdEndPerTileExecutionQCOM,
    .CreateExternalComputeQueueNV = gfx35_CreateExternalComputeQueueNV,
    .DestroyExternalComputeQueueNV = gfx35_DestroyExternalComputeQueueNV,
    .CreateShaderInstrumentationARM = gfx35_CreateShaderInstrumentationARM,
    .DestroyShaderInstrumentationARM = gfx35_DestroyShaderInstrumentationARM,
    .CmdBeginShaderInstrumentationARM = gfx35_CmdBeginShaderInstrumentationARM,
    .CmdEndShaderInstrumentationARM = gfx35_CmdEndShaderInstrumentationARM,
    .GetShaderInstrumentationValuesARM = gfx35_GetShaderInstrumentationValuesARM,
    .ClearShaderInstrumentationMetricsARM = gfx35_ClearShaderInstrumentationMetricsARM,
    .CreateTensorARM = gfx35_CreateTensorARM,
    .DestroyTensorARM = gfx35_DestroyTensorARM,
    .CreateTensorViewARM = gfx35_CreateTensorViewARM,
    .DestroyTensorViewARM = gfx35_DestroyTensorViewARM,
    .GetTensorMemoryRequirementsARM = gfx35_GetTensorMemoryRequirementsARM,
    .BindTensorMemoryARM = gfx35_BindTensorMemoryARM,
    .GetDeviceTensorMemoryRequirementsARM = gfx35_GetDeviceTensorMemoryRequirementsARM,
    .CmdCopyTensorARM = gfx35_CmdCopyTensorARM,
    .GetTensorOpaqueCaptureDescriptorDataARM = gfx35_GetTensorOpaqueCaptureDescriptorDataARM,
    .GetTensorViewOpaqueCaptureDescriptorDataARM = gfx35_GetTensorViewOpaqueCaptureDescriptorDataARM,
    .CreateDataGraphPipelinesARM = gfx35_CreateDataGraphPipelinesARM,
    .CreateDataGraphPipelineSessionARM = gfx35_CreateDataGraphPipelineSessionARM,
    .GetDataGraphPipelineSessionBindPointRequirementsARM = gfx35_GetDataGraphPipelineSessionBindPointRequirementsARM,
    .GetDataGraphPipelineSessionMemoryRequirementsARM = gfx35_GetDataGraphPipelineSessionMemoryRequirementsARM,
    .BindDataGraphPipelineSessionMemoryARM = gfx35_BindDataGraphPipelineSessionMemoryARM,
    .DestroyDataGraphPipelineSessionARM = gfx35_DestroyDataGraphPipelineSessionARM,
    .CmdDispatchDataGraphARM = gfx35_CmdDispatchDataGraphARM,
    .GetDataGraphPipelineAvailablePropertiesARM = gfx35_GetDataGraphPipelineAvailablePropertiesARM,
    .GetDataGraphPipelinePropertiesARM = gfx35_GetDataGraphPipelinePropertiesARM,
#ifdef VK_USE_PLATFORM_OHOS
    .GetNativeBufferPropertiesOHOS = gfx35_GetNativeBufferPropertiesOHOS,
#elif defined(_MSC_VER)
    .GetNativeBufferPropertiesOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
    .GetMemoryNativeBufferOHOS = gfx35_GetMemoryNativeBufferOHOS,
#elif defined(_MSC_VER)
    .GetMemoryNativeBufferOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .QueueSetPerfHintQCOM = gfx35_QueueSetPerfHintQCOM,
    .CmdSetComputeOccupancyPriorityNV = gfx35_CmdSetComputeOccupancyPriorityNV,
    .WriteSamplerDescriptorsEXT = gfx35_WriteSamplerDescriptorsEXT,
    .WriteResourceDescriptorsEXT = gfx35_WriteResourceDescriptorsEXT,
    .CmdBindSamplerHeapEXT = gfx35_CmdBindSamplerHeapEXT,
    .CmdBindResourceHeapEXT = gfx35_CmdBindResourceHeapEXT,
    .CmdPushDataEXT = gfx35_CmdPushDataEXT,
    .RegisterCustomBorderColorEXT = gfx35_RegisterCustomBorderColorEXT,
    .UnregisterCustomBorderColorEXT = gfx35_UnregisterCustomBorderColorEXT,
    .GetImageOpaqueCaptureDataEXT = gfx35_GetImageOpaqueCaptureDataEXT,
    .GetTensorOpaqueCaptureDataARM = gfx35_GetTensorOpaqueCaptureDataARM,
    .CmdCopyMemoryKHR = gfx35_CmdCopyMemoryKHR,
    .CmdCopyMemoryToImageKHR = gfx35_CmdCopyMemoryToImageKHR,
    .CmdCopyImageToMemoryKHR = gfx35_CmdCopyImageToMemoryKHR,
    .CmdUpdateMemoryKHR = gfx35_CmdUpdateMemoryKHR,
    .CmdFillMemoryKHR = gfx35_CmdFillMemoryKHR,
    .CmdCopyQueryPoolResultsToMemoryKHR = gfx35_CmdCopyQueryPoolResultsToMemoryKHR,
    .CmdBeginConditionalRendering2EXT = gfx35_CmdBeginConditionalRendering2EXT,
    .CmdBindTransformFeedbackBuffers2EXT = gfx35_CmdBindTransformFeedbackBuffers2EXT,
    .CmdBeginTransformFeedback2EXT = gfx35_CmdBeginTransformFeedback2EXT,
    .CmdEndTransformFeedback2EXT = gfx35_CmdEndTransformFeedback2EXT,
    .CmdDrawIndirectByteCount2EXT = gfx35_CmdDrawIndirectByteCount2EXT,
    .CmdWriteMarkerToMemoryAMD = gfx35_CmdWriteMarkerToMemoryAMD,
    .CmdBindIndexBuffer3KHR = gfx35_CmdBindIndexBuffer3KHR,
    .CmdBindVertexBuffers3KHR = gfx35_CmdBindVertexBuffers3KHR,
    .CmdDrawIndirect2KHR = gfx35_CmdDrawIndirect2KHR,
    .CmdDrawIndexedIndirect2KHR = gfx35_CmdDrawIndexedIndirect2KHR,
    .CmdDrawIndirectCount2KHR = gfx35_CmdDrawIndirectCount2KHR,
    .CmdDrawIndexedIndirectCount2KHR = gfx35_CmdDrawIndexedIndirectCount2KHR,
    .CmdDrawMeshTasksIndirect2EXT = gfx35_CmdDrawMeshTasksIndirect2EXT,
    .CmdDrawMeshTasksIndirectCount2EXT = gfx35_CmdDrawMeshTasksIndirectCount2EXT,
    .CmdDispatchIndirect2KHR = gfx35_CmdDispatchIndirect2KHR,
    .CreateAccelerationStructure2KHR = gfx35_CreateAccelerationStructure2KHR,
    .CmdSetDispatchParametersARM = gfx35_CmdSetDispatchParametersARM,
};
const struct vk_device_entrypoint_table anv_doom64_device_entrypoints = {
    .GetDeviceProcAddr = anv_doom64_GetDeviceProcAddr,
    .DestroyDevice = anv_doom64_DestroyDevice,
    .GetDeviceQueue = anv_doom64_GetDeviceQueue,
    .QueueSubmit = anv_doom64_QueueSubmit,
    .QueueWaitIdle = anv_doom64_QueueWaitIdle,
    .DeviceWaitIdle = anv_doom64_DeviceWaitIdle,
    .AllocateMemory = anv_doom64_AllocateMemory,
    .FreeMemory = anv_doom64_FreeMemory,
    .MapMemory = anv_doom64_MapMemory,
    .UnmapMemory = anv_doom64_UnmapMemory,
    .FlushMappedMemoryRanges = anv_doom64_FlushMappedMemoryRanges,
    .InvalidateMappedMemoryRanges = anv_doom64_InvalidateMappedMemoryRanges,
    .GetDeviceMemoryCommitment = anv_doom64_GetDeviceMemoryCommitment,
    .GetBufferMemoryRequirements = anv_doom64_GetBufferMemoryRequirements,
    .BindBufferMemory = anv_doom64_BindBufferMemory,
    .GetImageMemoryRequirements = anv_doom64_GetImageMemoryRequirements,
    .BindImageMemory = anv_doom64_BindImageMemory,
    .GetImageSparseMemoryRequirements = anv_doom64_GetImageSparseMemoryRequirements,
    .QueueBindSparse = anv_doom64_QueueBindSparse,
    .CreateFence = anv_doom64_CreateFence,
    .DestroyFence = anv_doom64_DestroyFence,
    .ResetFences = anv_doom64_ResetFences,
    .GetFenceStatus = anv_doom64_GetFenceStatus,
    .WaitForFences = anv_doom64_WaitForFences,
    .CreateSemaphore = anv_doom64_CreateSemaphore,
    .DestroySemaphore = anv_doom64_DestroySemaphore,
    .CreateEvent = anv_doom64_CreateEvent,
    .DestroyEvent = anv_doom64_DestroyEvent,
    .GetEventStatus = anv_doom64_GetEventStatus,
    .SetEvent = anv_doom64_SetEvent,
    .ResetEvent = anv_doom64_ResetEvent,
    .CreateQueryPool = anv_doom64_CreateQueryPool,
    .DestroyQueryPool = anv_doom64_DestroyQueryPool,
    .GetQueryPoolResults = anv_doom64_GetQueryPoolResults,
    .ResetQueryPool = anv_doom64_ResetQueryPool,
    .ResetQueryPoolEXT = anv_doom64_ResetQueryPoolEXT,
    .CreateBuffer = anv_doom64_CreateBuffer,
    .DestroyBuffer = anv_doom64_DestroyBuffer,
    .CreateBufferView = anv_doom64_CreateBufferView,
    .DestroyBufferView = anv_doom64_DestroyBufferView,
    .CreateImage = anv_doom64_CreateImage,
    .DestroyImage = anv_doom64_DestroyImage,
    .GetImageSubresourceLayout = anv_doom64_GetImageSubresourceLayout,
    .CreateImageView = anv_doom64_CreateImageView,
    .DestroyImageView = anv_doom64_DestroyImageView,
    .CreateShaderModule = anv_doom64_CreateShaderModule,
    .DestroyShaderModule = anv_doom64_DestroyShaderModule,
    .CreatePipelineCache = anv_doom64_CreatePipelineCache,
    .DestroyPipelineCache = anv_doom64_DestroyPipelineCache,
    .GetPipelineCacheData = anv_doom64_GetPipelineCacheData,
    .MergePipelineCaches = anv_doom64_MergePipelineCaches,
    .CreatePipelineBinariesKHR = anv_doom64_CreatePipelineBinariesKHR,
    .DestroyPipelineBinaryKHR = anv_doom64_DestroyPipelineBinaryKHR,
    .GetPipelineKeyKHR = anv_doom64_GetPipelineKeyKHR,
    .GetPipelineBinaryDataKHR = anv_doom64_GetPipelineBinaryDataKHR,
    .ReleaseCapturedPipelineDataKHR = anv_doom64_ReleaseCapturedPipelineDataKHR,
    .CreateGraphicsPipelines = anv_doom64_CreateGraphicsPipelines,
    .CreateComputePipelines = anv_doom64_CreateComputePipelines,
    .GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI = anv_doom64_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI,
    .DestroyPipeline = anv_doom64_DestroyPipeline,
    .CreatePipelineLayout = anv_doom64_CreatePipelineLayout,
    .DestroyPipelineLayout = anv_doom64_DestroyPipelineLayout,
    .CreateSampler = anv_doom64_CreateSampler,
    .DestroySampler = anv_doom64_DestroySampler,
    .CreateDescriptorSetLayout = anv_doom64_CreateDescriptorSetLayout,
    .DestroyDescriptorSetLayout = anv_doom64_DestroyDescriptorSetLayout,
    .CreateDescriptorPool = anv_doom64_CreateDescriptorPool,
    .DestroyDescriptorPool = anv_doom64_DestroyDescriptorPool,
    .ResetDescriptorPool = anv_doom64_ResetDescriptorPool,
    .AllocateDescriptorSets = anv_doom64_AllocateDescriptorSets,
    .FreeDescriptorSets = anv_doom64_FreeDescriptorSets,
    .UpdateDescriptorSets = anv_doom64_UpdateDescriptorSets,
    .CreateFramebuffer = anv_doom64_CreateFramebuffer,
    .DestroyFramebuffer = anv_doom64_DestroyFramebuffer,
    .CreateRenderPass = anv_doom64_CreateRenderPass,
    .DestroyRenderPass = anv_doom64_DestroyRenderPass,
    .GetRenderAreaGranularity = anv_doom64_GetRenderAreaGranularity,
    .GetRenderingAreaGranularity = anv_doom64_GetRenderingAreaGranularity,
    .GetRenderingAreaGranularityKHR = anv_doom64_GetRenderingAreaGranularityKHR,
    .CreateCommandPool = anv_doom64_CreateCommandPool,
    .DestroyCommandPool = anv_doom64_DestroyCommandPool,
    .ResetCommandPool = anv_doom64_ResetCommandPool,
    .AllocateCommandBuffers = anv_doom64_AllocateCommandBuffers,
    .FreeCommandBuffers = anv_doom64_FreeCommandBuffers,
    .BeginCommandBuffer = anv_doom64_BeginCommandBuffer,
    .EndCommandBuffer = anv_doom64_EndCommandBuffer,
    .ResetCommandBuffer = anv_doom64_ResetCommandBuffer,
    .CmdBindPipeline = anv_doom64_CmdBindPipeline,
    .CmdSetPrimitiveRestartIndexEXT = anv_doom64_CmdSetPrimitiveRestartIndexEXT,
    .CmdSetAttachmentFeedbackLoopEnableEXT = anv_doom64_CmdSetAttachmentFeedbackLoopEnableEXT,
    .CmdSetViewport = anv_doom64_CmdSetViewport,
    .CmdSetScissor = anv_doom64_CmdSetScissor,
    .CmdSetLineWidth = anv_doom64_CmdSetLineWidth,
    .CmdSetDepthBias = anv_doom64_CmdSetDepthBias,
    .CmdSetBlendConstants = anv_doom64_CmdSetBlendConstants,
    .CmdSetDepthBounds = anv_doom64_CmdSetDepthBounds,
    .CmdSetStencilCompareMask = anv_doom64_CmdSetStencilCompareMask,
    .CmdSetStencilWriteMask = anv_doom64_CmdSetStencilWriteMask,
    .CmdSetStencilReference = anv_doom64_CmdSetStencilReference,
    .CmdBindDescriptorSets = anv_doom64_CmdBindDescriptorSets,
    .CmdBindIndexBuffer = anv_doom64_CmdBindIndexBuffer,
    .CmdBindVertexBuffers = anv_doom64_CmdBindVertexBuffers,
    .CmdDraw = anv_doom64_CmdDraw,
    .CmdDrawIndexed = anv_doom64_CmdDrawIndexed,
    .CmdDrawMultiEXT = anv_doom64_CmdDrawMultiEXT,
    .CmdDrawMultiIndexedEXT = anv_doom64_CmdDrawMultiIndexedEXT,
    .CmdDrawIndirect = anv_doom64_CmdDrawIndirect,
    .CmdDrawIndexedIndirect = anv_doom64_CmdDrawIndexedIndirect,
    .CmdDispatch = anv_doom64_CmdDispatch,
    .CmdDispatchIndirect = anv_doom64_CmdDispatchIndirect,
    .CmdSubpassShadingHUAWEI = anv_doom64_CmdSubpassShadingHUAWEI,
    .CmdDrawClusterHUAWEI = anv_doom64_CmdDrawClusterHUAWEI,
    .CmdDrawClusterIndirectHUAWEI = anv_doom64_CmdDrawClusterIndirectHUAWEI,
    .CmdUpdatePipelineIndirectBufferNV = anv_doom64_CmdUpdatePipelineIndirectBufferNV,
    .CmdCopyBuffer = anv_doom64_CmdCopyBuffer,
    .CmdCopyImage = anv_doom64_CmdCopyImage,
    .CmdBlitImage = anv_doom64_CmdBlitImage,
    .CmdCopyBufferToImage = anv_doom64_CmdCopyBufferToImage,
    .CmdCopyImageToBuffer = anv_doom64_CmdCopyImageToBuffer,
    .CmdCopyMemoryIndirectNV = anv_doom64_CmdCopyMemoryIndirectNV,
    .CmdCopyMemoryIndirectKHR = anv_doom64_CmdCopyMemoryIndirectKHR,
    .CmdCopyMemoryToImageIndirectNV = anv_doom64_CmdCopyMemoryToImageIndirectNV,
    .CmdCopyMemoryToImageIndirectKHR = anv_doom64_CmdCopyMemoryToImageIndirectKHR,
    .CmdUpdateBuffer = anv_doom64_CmdUpdateBuffer,
    .CmdFillBuffer = anv_doom64_CmdFillBuffer,
    .CmdClearColorImage = anv_doom64_CmdClearColorImage,
    .CmdClearDepthStencilImage = anv_doom64_CmdClearDepthStencilImage,
    .CmdClearAttachments = anv_doom64_CmdClearAttachments,
    .CmdResolveImage = anv_doom64_CmdResolveImage,
    .CmdSetEvent = anv_doom64_CmdSetEvent,
    .CmdResetEvent = anv_doom64_CmdResetEvent,
    .CmdWaitEvents = anv_doom64_CmdWaitEvents,
    .CmdPipelineBarrier = anv_doom64_CmdPipelineBarrier,
    .CmdBeginQuery = anv_doom64_CmdBeginQuery,
    .CmdEndQuery = anv_doom64_CmdEndQuery,
    .CmdBeginConditionalRenderingEXT = anv_doom64_CmdBeginConditionalRenderingEXT,
    .CmdEndConditionalRenderingEXT = anv_doom64_CmdEndConditionalRenderingEXT,
    .CmdBeginCustomResolveEXT = anv_doom64_CmdBeginCustomResolveEXT,
    .CmdResetQueryPool = anv_doom64_CmdResetQueryPool,
    .CmdWriteTimestamp = anv_doom64_CmdWriteTimestamp,
    .CmdCopyQueryPoolResults = anv_doom64_CmdCopyQueryPoolResults,
    .CmdPushConstants = anv_doom64_CmdPushConstants,
    .CmdBeginRenderPass = anv_doom64_CmdBeginRenderPass,
    .CmdNextSubpass = anv_doom64_CmdNextSubpass,
    .CmdEndRenderPass = anv_doom64_CmdEndRenderPass,
    .CmdExecuteCommands = anv_doom64_CmdExecuteCommands,
    .CreateSharedSwapchainsKHR = anv_doom64_CreateSharedSwapchainsKHR,
    .CreateSwapchainKHR = anv_doom64_CreateSwapchainKHR,
    .DestroySwapchainKHR = anv_doom64_DestroySwapchainKHR,
    .GetSwapchainImagesKHR = anv_doom64_GetSwapchainImagesKHR,
    .AcquireNextImageKHR = anv_doom64_AcquireNextImageKHR,
    .QueuePresentKHR = anv_doom64_QueuePresentKHR,
    .DebugMarkerSetObjectNameEXT = anv_doom64_DebugMarkerSetObjectNameEXT,
    .DebugMarkerSetObjectTagEXT = anv_doom64_DebugMarkerSetObjectTagEXT,
    .CmdDebugMarkerBeginEXT = anv_doom64_CmdDebugMarkerBeginEXT,
    .CmdDebugMarkerEndEXT = anv_doom64_CmdDebugMarkerEndEXT,
    .CmdDebugMarkerInsertEXT = anv_doom64_CmdDebugMarkerInsertEXT,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleNV = anv_doom64_GetMemoryWin32HandleNV,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .CmdExecuteGeneratedCommandsNV = anv_doom64_CmdExecuteGeneratedCommandsNV,
    .CmdPreprocessGeneratedCommandsNV = anv_doom64_CmdPreprocessGeneratedCommandsNV,
    .CmdBindPipelineShaderGroupNV = anv_doom64_CmdBindPipelineShaderGroupNV,
    .GetGeneratedCommandsMemoryRequirementsNV = anv_doom64_GetGeneratedCommandsMemoryRequirementsNV,
    .CreateIndirectCommandsLayoutNV = anv_doom64_CreateIndirectCommandsLayoutNV,
    .DestroyIndirectCommandsLayoutNV = anv_doom64_DestroyIndirectCommandsLayoutNV,
    .CmdExecuteGeneratedCommandsEXT = anv_doom64_CmdExecuteGeneratedCommandsEXT,
    .CmdPreprocessGeneratedCommandsEXT = anv_doom64_CmdPreprocessGeneratedCommandsEXT,
    .GetGeneratedCommandsMemoryRequirementsEXT = anv_doom64_GetGeneratedCommandsMemoryRequirementsEXT,
    .CreateIndirectCommandsLayoutEXT = anv_doom64_CreateIndirectCommandsLayoutEXT,
    .DestroyIndirectCommandsLayoutEXT = anv_doom64_DestroyIndirectCommandsLayoutEXT,
    .CreateIndirectExecutionSetEXT = anv_doom64_CreateIndirectExecutionSetEXT,
    .DestroyIndirectExecutionSetEXT = anv_doom64_DestroyIndirectExecutionSetEXT,
    .UpdateIndirectExecutionSetPipelineEXT = anv_doom64_UpdateIndirectExecutionSetPipelineEXT,
    .UpdateIndirectExecutionSetShaderEXT = anv_doom64_UpdateIndirectExecutionSetShaderEXT,
    .CmdPushDescriptorSet = anv_doom64_CmdPushDescriptorSet,
    .CmdPushDescriptorSetKHR = anv_doom64_CmdPushDescriptorSetKHR,
    .TrimCommandPool = anv_doom64_TrimCommandPool,
    .TrimCommandPoolKHR = anv_doom64_TrimCommandPoolKHR,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleKHR = anv_doom64_GetMemoryWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandlePropertiesKHR = anv_doom64_GetMemoryWin32HandlePropertiesKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandlePropertiesKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryFdKHR = anv_doom64_GetMemoryFdKHR,
    .GetMemoryFdPropertiesKHR = anv_doom64_GetMemoryFdPropertiesKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandleFUCHSIA = anv_doom64_GetMemoryZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandlePropertiesFUCHSIA = anv_doom64_GetMemoryZirconHandlePropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandlePropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .GetMemoryRemoteAddressNV = anv_doom64_GetMemoryRemoteAddressNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreWin32HandleKHR = anv_doom64_GetSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportSemaphoreWin32HandleKHR = anv_doom64_ImportSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreFdKHR = anv_doom64_GetSemaphoreFdKHR,
    .ImportSemaphoreFdKHR = anv_doom64_ImportSemaphoreFdKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetSemaphoreZirconHandleFUCHSIA = anv_doom64_GetSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .ImportSemaphoreZirconHandleFUCHSIA = anv_doom64_ImportSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .ImportSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetFenceWin32HandleKHR = anv_doom64_GetFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportFenceWin32HandleKHR = anv_doom64_ImportFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetFenceFdKHR = anv_doom64_GetFenceFdKHR,
    .ImportFenceFdKHR = anv_doom64_ImportFenceFdKHR,
    .DisplayPowerControlEXT = anv_doom64_DisplayPowerControlEXT,
    .RegisterDeviceEventEXT = anv_doom64_RegisterDeviceEventEXT,
    .RegisterDisplayEventEXT = anv_doom64_RegisterDisplayEventEXT,
    .GetSwapchainCounterEXT = anv_doom64_GetSwapchainCounterEXT,
    .GetDeviceGroupPeerMemoryFeatures = anv_doom64_GetDeviceGroupPeerMemoryFeatures,
    .GetDeviceGroupPeerMemoryFeaturesKHR = anv_doom64_GetDeviceGroupPeerMemoryFeaturesKHR,
    .BindBufferMemory2 = anv_doom64_BindBufferMemory2,
    .BindBufferMemory2KHR = anv_doom64_BindBufferMemory2KHR,
    .BindImageMemory2 = anv_doom64_BindImageMemory2,
    .BindImageMemory2KHR = anv_doom64_BindImageMemory2KHR,
    .CmdSetDeviceMask = anv_doom64_CmdSetDeviceMask,
    .CmdSetDeviceMaskKHR = anv_doom64_CmdSetDeviceMaskKHR,
    .GetDeviceGroupPresentCapabilitiesKHR = anv_doom64_GetDeviceGroupPresentCapabilitiesKHR,
    .GetDeviceGroupSurfacePresentModesKHR = anv_doom64_GetDeviceGroupSurfacePresentModesKHR,
    .AcquireNextImage2KHR = anv_doom64_AcquireNextImage2KHR,
    .CmdDispatchBase = anv_doom64_CmdDispatchBase,
    .CmdDispatchBaseKHR = anv_doom64_CmdDispatchBaseKHR,
    .CreateDescriptorUpdateTemplate = anv_doom64_CreateDescriptorUpdateTemplate,
    .CreateDescriptorUpdateTemplateKHR = anv_doom64_CreateDescriptorUpdateTemplateKHR,
    .DestroyDescriptorUpdateTemplate = anv_doom64_DestroyDescriptorUpdateTemplate,
    .DestroyDescriptorUpdateTemplateKHR = anv_doom64_DestroyDescriptorUpdateTemplateKHR,
    .UpdateDescriptorSetWithTemplate = anv_doom64_UpdateDescriptorSetWithTemplate,
    .UpdateDescriptorSetWithTemplateKHR = anv_doom64_UpdateDescriptorSetWithTemplateKHR,
    .CmdPushDescriptorSetWithTemplate = anv_doom64_CmdPushDescriptorSetWithTemplate,
    .CmdPushDescriptorSetWithTemplateKHR = anv_doom64_CmdPushDescriptorSetWithTemplateKHR,
    .SetHdrMetadataEXT = anv_doom64_SetHdrMetadataEXT,
    .GetSwapchainStatusKHR = anv_doom64_GetSwapchainStatusKHR,
    .GetRefreshCycleDurationGOOGLE = anv_doom64_GetRefreshCycleDurationGOOGLE,
    .GetPastPresentationTimingGOOGLE = anv_doom64_GetPastPresentationTimingGOOGLE,
    .CmdSetViewportWScalingNV = anv_doom64_CmdSetViewportWScalingNV,
    .CmdSetDiscardRectangleEXT = anv_doom64_CmdSetDiscardRectangleEXT,
    .CmdSetDiscardRectangleEnableEXT = anv_doom64_CmdSetDiscardRectangleEnableEXT,
    .CmdSetDiscardRectangleModeEXT = anv_doom64_CmdSetDiscardRectangleModeEXT,
    .CmdSetSampleLocationsEXT = anv_doom64_CmdSetSampleLocationsEXT,
    .GetBufferMemoryRequirements2 = anv_doom64_GetBufferMemoryRequirements2,
    .GetBufferMemoryRequirements2KHR = anv_doom64_GetBufferMemoryRequirements2KHR,
    .GetImageMemoryRequirements2 = anv_doom64_GetImageMemoryRequirements2,
    .GetImageMemoryRequirements2KHR = anv_doom64_GetImageMemoryRequirements2KHR,
    .GetImageSparseMemoryRequirements2 = anv_doom64_GetImageSparseMemoryRequirements2,
    .GetImageSparseMemoryRequirements2KHR = anv_doom64_GetImageSparseMemoryRequirements2KHR,
    .GetDeviceBufferMemoryRequirements = anv_doom64_GetDeviceBufferMemoryRequirements,
    .GetDeviceBufferMemoryRequirementsKHR = anv_doom64_GetDeviceBufferMemoryRequirementsKHR,
    .GetDeviceImageMemoryRequirements = anv_doom64_GetDeviceImageMemoryRequirements,
    .GetDeviceImageMemoryRequirementsKHR = anv_doom64_GetDeviceImageMemoryRequirementsKHR,
    .GetDeviceImageSparseMemoryRequirements = anv_doom64_GetDeviceImageSparseMemoryRequirements,
    .GetDeviceImageSparseMemoryRequirementsKHR = anv_doom64_GetDeviceImageSparseMemoryRequirementsKHR,
    .CreateSamplerYcbcrConversion = anv_doom64_CreateSamplerYcbcrConversion,
    .CreateSamplerYcbcrConversionKHR = anv_doom64_CreateSamplerYcbcrConversionKHR,
    .DestroySamplerYcbcrConversion = anv_doom64_DestroySamplerYcbcrConversion,
    .DestroySamplerYcbcrConversionKHR = anv_doom64_DestroySamplerYcbcrConversionKHR,
    .GetDeviceQueue2 = anv_doom64_GetDeviceQueue2,
    .CreateValidationCacheEXT = anv_doom64_CreateValidationCacheEXT,
    .DestroyValidationCacheEXT = anv_doom64_DestroyValidationCacheEXT,
    .GetValidationCacheDataEXT = anv_doom64_GetValidationCacheDataEXT,
    .MergeValidationCachesEXT = anv_doom64_MergeValidationCachesEXT,
    .GetDescriptorSetLayoutSupport = anv_doom64_GetDescriptorSetLayoutSupport,
    .GetDescriptorSetLayoutSupportKHR = anv_doom64_GetDescriptorSetLayoutSupportKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsageANDROID = anv_doom64_GetSwapchainGrallocUsageANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsage2ANDROID = anv_doom64_GetSwapchainGrallocUsage2ANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsage2ANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .AcquireImageANDROID = anv_doom64_AcquireImageANDROID,
#elif defined(_MSC_VER)
    .AcquireImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .QueueSignalReleaseImageANDROID = anv_doom64_QueueSignalReleaseImageANDROID,
#elif defined(_MSC_VER)
    .QueueSignalReleaseImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .GetShaderInfoAMD = anv_doom64_GetShaderInfoAMD,
    .SetLocalDimmingAMD = anv_doom64_SetLocalDimmingAMD,
    .GetCalibratedTimestampsKHR = anv_doom64_GetCalibratedTimestampsKHR,
    .GetCalibratedTimestampsEXT = anv_doom64_GetCalibratedTimestampsEXT,
    .SetDebugUtilsObjectNameEXT = anv_doom64_SetDebugUtilsObjectNameEXT,
    .SetDebugUtilsObjectTagEXT = anv_doom64_SetDebugUtilsObjectTagEXT,
    .QueueBeginDebugUtilsLabelEXT = anv_doom64_QueueBeginDebugUtilsLabelEXT,
    .QueueEndDebugUtilsLabelEXT = anv_doom64_QueueEndDebugUtilsLabelEXT,
    .QueueInsertDebugUtilsLabelEXT = anv_doom64_QueueInsertDebugUtilsLabelEXT,
    .CmdBeginDebugUtilsLabelEXT = anv_doom64_CmdBeginDebugUtilsLabelEXT,
    .CmdEndDebugUtilsLabelEXT = anv_doom64_CmdEndDebugUtilsLabelEXT,
    .CmdInsertDebugUtilsLabelEXT = anv_doom64_CmdInsertDebugUtilsLabelEXT,
    .GetMemoryHostPointerPropertiesEXT = anv_doom64_GetMemoryHostPointerPropertiesEXT,
    .CmdWriteBufferMarkerAMD = anv_doom64_CmdWriteBufferMarkerAMD,
    .CreateRenderPass2 = anv_doom64_CreateRenderPass2,
    .CreateRenderPass2KHR = anv_doom64_CreateRenderPass2KHR,
    .CmdBeginRenderPass2 = anv_doom64_CmdBeginRenderPass2,
    .CmdBeginRenderPass2KHR = anv_doom64_CmdBeginRenderPass2KHR,
    .CmdNextSubpass2 = anv_doom64_CmdNextSubpass2,
    .CmdNextSubpass2KHR = anv_doom64_CmdNextSubpass2KHR,
    .CmdEndRenderPass2 = anv_doom64_CmdEndRenderPass2,
    .CmdEndRenderPass2KHR = anv_doom64_CmdEndRenderPass2KHR,
    .GetSemaphoreCounterValue = anv_doom64_GetSemaphoreCounterValue,
    .GetSemaphoreCounterValueKHR = anv_doom64_GetSemaphoreCounterValueKHR,
    .WaitSemaphores = anv_doom64_WaitSemaphores,
    .WaitSemaphoresKHR = anv_doom64_WaitSemaphoresKHR,
    .SignalSemaphore = anv_doom64_SignalSemaphore,
    .SignalSemaphoreKHR = anv_doom64_SignalSemaphoreKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetAndroidHardwareBufferPropertiesANDROID = anv_doom64_GetAndroidHardwareBufferPropertiesANDROID,
#elif defined(_MSC_VER)
    .GetAndroidHardwareBufferPropertiesANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetMemoryAndroidHardwareBufferANDROID = anv_doom64_GetMemoryAndroidHardwareBufferANDROID,
#elif defined(_MSC_VER)
    .GetMemoryAndroidHardwareBufferANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .CmdDrawIndirectCount = anv_doom64_CmdDrawIndirectCount,
    .CmdDrawIndirectCountKHR = anv_doom64_CmdDrawIndirectCountKHR,
    .CmdDrawIndirectCountAMD = anv_doom64_CmdDrawIndirectCountAMD,
    .CmdDrawIndexedIndirectCount = anv_doom64_CmdDrawIndexedIndirectCount,
    .CmdDrawIndexedIndirectCountKHR = anv_doom64_CmdDrawIndexedIndirectCountKHR,
    .CmdDrawIndexedIndirectCountAMD = anv_doom64_CmdDrawIndexedIndirectCountAMD,
    .CmdSetCheckpointNV = anv_doom64_CmdSetCheckpointNV,
    .GetQueueCheckpointDataNV = anv_doom64_GetQueueCheckpointDataNV,
    .CmdBindTransformFeedbackBuffersEXT = anv_doom64_CmdBindTransformFeedbackBuffersEXT,
    .CmdBeginTransformFeedbackEXT = anv_doom64_CmdBeginTransformFeedbackEXT,
    .CmdEndTransformFeedbackEXT = anv_doom64_CmdEndTransformFeedbackEXT,
    .CmdBeginQueryIndexedEXT = anv_doom64_CmdBeginQueryIndexedEXT,
    .CmdEndQueryIndexedEXT = anv_doom64_CmdEndQueryIndexedEXT,
    .CmdDrawIndirectByteCountEXT = anv_doom64_CmdDrawIndirectByteCountEXT,
    .CmdSetExclusiveScissorNV = anv_doom64_CmdSetExclusiveScissorNV,
    .CmdSetExclusiveScissorEnableNV = anv_doom64_CmdSetExclusiveScissorEnableNV,
    .CmdBindShadingRateImageNV = anv_doom64_CmdBindShadingRateImageNV,
    .CmdSetViewportShadingRatePaletteNV = anv_doom64_CmdSetViewportShadingRatePaletteNV,
    .CmdSetCoarseSampleOrderNV = anv_doom64_CmdSetCoarseSampleOrderNV,
    .CmdDrawMeshTasksNV = anv_doom64_CmdDrawMeshTasksNV,
    .CmdDrawMeshTasksIndirectNV = anv_doom64_CmdDrawMeshTasksIndirectNV,
    .CmdDrawMeshTasksIndirectCountNV = anv_doom64_CmdDrawMeshTasksIndirectCountNV,
    .CmdDrawMeshTasksEXT = anv_doom64_CmdDrawMeshTasksEXT,
    .CmdDrawMeshTasksIndirectEXT = anv_doom64_CmdDrawMeshTasksIndirectEXT,
    .CmdDrawMeshTasksIndirectCountEXT = anv_doom64_CmdDrawMeshTasksIndirectCountEXT,
    .CompileDeferredNV = anv_doom64_CompileDeferredNV,
    .CreateAccelerationStructureNV = anv_doom64_CreateAccelerationStructureNV,
    .CmdBindInvocationMaskHUAWEI = anv_doom64_CmdBindInvocationMaskHUAWEI,
    .DestroyAccelerationStructureKHR = anv_doom64_DestroyAccelerationStructureKHR,
    .DestroyAccelerationStructureNV = anv_doom64_DestroyAccelerationStructureNV,
    .GetAccelerationStructureMemoryRequirementsNV = anv_doom64_GetAccelerationStructureMemoryRequirementsNV,
    .BindAccelerationStructureMemoryNV = anv_doom64_BindAccelerationStructureMemoryNV,
    .CmdCopyAccelerationStructureNV = anv_doom64_CmdCopyAccelerationStructureNV,
    .CmdCopyAccelerationStructureKHR = anv_doom64_CmdCopyAccelerationStructureKHR,
    .CopyAccelerationStructureKHR = anv_doom64_CopyAccelerationStructureKHR,
    .CmdCopyAccelerationStructureToMemoryKHR = anv_doom64_CmdCopyAccelerationStructureToMemoryKHR,
    .CopyAccelerationStructureToMemoryKHR = anv_doom64_CopyAccelerationStructureToMemoryKHR,
    .CmdCopyMemoryToAccelerationStructureKHR = anv_doom64_CmdCopyMemoryToAccelerationStructureKHR,
    .CopyMemoryToAccelerationStructureKHR = anv_doom64_CopyMemoryToAccelerationStructureKHR,
    .CmdWriteAccelerationStructuresPropertiesKHR = anv_doom64_CmdWriteAccelerationStructuresPropertiesKHR,
    .CmdWriteAccelerationStructuresPropertiesNV = anv_doom64_CmdWriteAccelerationStructuresPropertiesNV,
    .CmdBuildAccelerationStructureNV = anv_doom64_CmdBuildAccelerationStructureNV,
    .WriteAccelerationStructuresPropertiesKHR = anv_doom64_WriteAccelerationStructuresPropertiesKHR,
    .CmdTraceRaysKHR = anv_doom64_CmdTraceRaysKHR,
    .CmdTraceRaysNV = anv_doom64_CmdTraceRaysNV,
    .GetRayTracingShaderGroupHandlesKHR = anv_doom64_GetRayTracingShaderGroupHandlesKHR,
    .GetRayTracingShaderGroupHandlesNV = anv_doom64_GetRayTracingShaderGroupHandlesNV,
    .GetRayTracingCaptureReplayShaderGroupHandlesKHR = anv_doom64_GetRayTracingCaptureReplayShaderGroupHandlesKHR,
    .GetAccelerationStructureHandleNV = anv_doom64_GetAccelerationStructureHandleNV,
    .CreateRayTracingPipelinesNV = anv_doom64_CreateRayTracingPipelinesNV,
    .CreateRayTracingPipelinesKHR = anv_doom64_CreateRayTracingPipelinesKHR,
    .CmdTraceRaysIndirectKHR = anv_doom64_CmdTraceRaysIndirectKHR,
    .CmdTraceRaysIndirect2KHR = anv_doom64_CmdTraceRaysIndirect2KHR,
    .GetClusterAccelerationStructureBuildSizesNV = anv_doom64_GetClusterAccelerationStructureBuildSizesNV,
    .CmdBuildClusterAccelerationStructureIndirectNV = anv_doom64_CmdBuildClusterAccelerationStructureIndirectNV,
    .GetDeviceAccelerationStructureCompatibilityKHR = anv_doom64_GetDeviceAccelerationStructureCompatibilityKHR,
    .GetRayTracingShaderGroupStackSizeKHR = anv_doom64_GetRayTracingShaderGroupStackSizeKHR,
    .CmdSetRayTracingPipelineStackSizeKHR = anv_doom64_CmdSetRayTracingPipelineStackSizeKHR,
    .GetImageViewHandleNVX = anv_doom64_GetImageViewHandleNVX,
    .GetImageViewHandle64NVX = anv_doom64_GetImageViewHandle64NVX,
    .GetImageViewAddressNVX = anv_doom64_GetImageViewAddressNVX,
    .GetDeviceCombinedImageSamplerIndexNVX = anv_doom64_GetDeviceCombinedImageSamplerIndexNVX,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetDeviceGroupSurfacePresentModes2EXT = anv_doom64_GetDeviceGroupSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetDeviceGroupSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireFullScreenExclusiveModeEXT = anv_doom64_AcquireFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .AcquireFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ReleaseFullScreenExclusiveModeEXT = anv_doom64_ReleaseFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .ReleaseFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .AcquireProfilingLockKHR = anv_doom64_AcquireProfilingLockKHR,
    .ReleaseProfilingLockKHR = anv_doom64_ReleaseProfilingLockKHR,
    .GetImageDrmFormatModifierPropertiesEXT = anv_doom64_GetImageDrmFormatModifierPropertiesEXT,
    .GetBufferOpaqueCaptureAddress = anv_doom64_GetBufferOpaqueCaptureAddress,
    .GetBufferOpaqueCaptureAddressKHR = anv_doom64_GetBufferOpaqueCaptureAddressKHR,
    .GetBufferDeviceAddress = anv_doom64_GetBufferDeviceAddress,
    .GetBufferDeviceAddressKHR = anv_doom64_GetBufferDeviceAddressKHR,
    .GetBufferDeviceAddressEXT = anv_doom64_GetBufferDeviceAddressEXT,
    .InitializePerformanceApiINTEL = anv_doom64_InitializePerformanceApiINTEL,
    .UninitializePerformanceApiINTEL = anv_doom64_UninitializePerformanceApiINTEL,
    .CmdSetPerformanceMarkerINTEL = anv_doom64_CmdSetPerformanceMarkerINTEL,
    .CmdSetPerformanceStreamMarkerINTEL = anv_doom64_CmdSetPerformanceStreamMarkerINTEL,
    .CmdSetPerformanceOverrideINTEL = anv_doom64_CmdSetPerformanceOverrideINTEL,
    .AcquirePerformanceConfigurationINTEL = anv_doom64_AcquirePerformanceConfigurationINTEL,
    .ReleasePerformanceConfigurationINTEL = anv_doom64_ReleasePerformanceConfigurationINTEL,
    .QueueSetPerformanceConfigurationINTEL = anv_doom64_QueueSetPerformanceConfigurationINTEL,
    .GetPerformanceParameterINTEL = anv_doom64_GetPerformanceParameterINTEL,
    .GetDeviceMemoryOpaqueCaptureAddress = anv_doom64_GetDeviceMemoryOpaqueCaptureAddress,
    .GetDeviceMemoryOpaqueCaptureAddressKHR = anv_doom64_GetDeviceMemoryOpaqueCaptureAddressKHR,
    .GetPipelineExecutablePropertiesKHR = anv_doom64_GetPipelineExecutablePropertiesKHR,
    .GetPipelineExecutableStatisticsKHR = anv_doom64_GetPipelineExecutableStatisticsKHR,
    .GetPipelineExecutableInternalRepresentationsKHR = anv_doom64_GetPipelineExecutableInternalRepresentationsKHR,
    .CmdSetLineStipple = anv_doom64_CmdSetLineStipple,
    .CmdSetLineStippleKHR = anv_doom64_CmdSetLineStippleKHR,
    .CmdSetLineStippleEXT = anv_doom64_CmdSetLineStippleEXT,
    .CreateAccelerationStructureKHR = anv_doom64_CreateAccelerationStructureKHR,
    .CmdBuildAccelerationStructuresKHR = anv_doom64_CmdBuildAccelerationStructuresKHR,
    .CmdBuildAccelerationStructuresIndirectKHR = anv_doom64_CmdBuildAccelerationStructuresIndirectKHR,
    .BuildAccelerationStructuresKHR = anv_doom64_BuildAccelerationStructuresKHR,
    .GetAccelerationStructureDeviceAddressKHR = anv_doom64_GetAccelerationStructureDeviceAddressKHR,
    .CreateDeferredOperationKHR = anv_doom64_CreateDeferredOperationKHR,
    .DestroyDeferredOperationKHR = anv_doom64_DestroyDeferredOperationKHR,
    .GetDeferredOperationMaxConcurrencyKHR = anv_doom64_GetDeferredOperationMaxConcurrencyKHR,
    .GetDeferredOperationResultKHR = anv_doom64_GetDeferredOperationResultKHR,
    .DeferredOperationJoinKHR = anv_doom64_DeferredOperationJoinKHR,
    .GetPipelineIndirectMemoryRequirementsNV = anv_doom64_GetPipelineIndirectMemoryRequirementsNV,
    .GetPipelineIndirectDeviceAddressNV = anv_doom64_GetPipelineIndirectDeviceAddressNV,
    .AntiLagUpdateAMD = anv_doom64_AntiLagUpdateAMD,
    .CmdSetCullMode = anv_doom64_CmdSetCullMode,
    .CmdSetCullModeEXT = anv_doom64_CmdSetCullModeEXT,
    .CmdSetFrontFace = anv_doom64_CmdSetFrontFace,
    .CmdSetFrontFaceEXT = anv_doom64_CmdSetFrontFaceEXT,
    .CmdSetPrimitiveTopology = anv_doom64_CmdSetPrimitiveTopology,
    .CmdSetPrimitiveTopologyEXT = anv_doom64_CmdSetPrimitiveTopologyEXT,
    .CmdSetViewportWithCount = anv_doom64_CmdSetViewportWithCount,
    .CmdSetViewportWithCountEXT = anv_doom64_CmdSetViewportWithCountEXT,
    .CmdSetScissorWithCount = anv_doom64_CmdSetScissorWithCount,
    .CmdSetScissorWithCountEXT = anv_doom64_CmdSetScissorWithCountEXT,
    .CmdBindIndexBuffer2 = anv_doom64_CmdBindIndexBuffer2,
    .CmdBindIndexBuffer2KHR = anv_doom64_CmdBindIndexBuffer2KHR,
    .CmdBindVertexBuffers2 = anv_doom64_CmdBindVertexBuffers2,
    .CmdBindVertexBuffers2EXT = anv_doom64_CmdBindVertexBuffers2EXT,
    .CmdSetDepthTestEnable = anv_doom64_CmdSetDepthTestEnable,
    .CmdSetDepthTestEnableEXT = anv_doom64_CmdSetDepthTestEnableEXT,
    .CmdSetDepthWriteEnable = anv_doom64_CmdSetDepthWriteEnable,
    .CmdSetDepthWriteEnableEXT = anv_doom64_CmdSetDepthWriteEnableEXT,
    .CmdSetDepthCompareOp = anv_doom64_CmdSetDepthCompareOp,
    .CmdSetDepthCompareOpEXT = anv_doom64_CmdSetDepthCompareOpEXT,
    .CmdSetDepthBoundsTestEnable = anv_doom64_CmdSetDepthBoundsTestEnable,
    .CmdSetDepthBoundsTestEnableEXT = anv_doom64_CmdSetDepthBoundsTestEnableEXT,
    .CmdSetStencilTestEnable = anv_doom64_CmdSetStencilTestEnable,
    .CmdSetStencilTestEnableEXT = anv_doom64_CmdSetStencilTestEnableEXT,
    .CmdSetStencilOp = anv_doom64_CmdSetStencilOp,
    .CmdSetStencilOpEXT = anv_doom64_CmdSetStencilOpEXT,
    .CmdSetPatchControlPointsEXT = anv_doom64_CmdSetPatchControlPointsEXT,
    .CmdSetRasterizerDiscardEnable = anv_doom64_CmdSetRasterizerDiscardEnable,
    .CmdSetRasterizerDiscardEnableEXT = anv_doom64_CmdSetRasterizerDiscardEnableEXT,
    .CmdSetDepthBiasEnable = anv_doom64_CmdSetDepthBiasEnable,
    .CmdSetDepthBiasEnableEXT = anv_doom64_CmdSetDepthBiasEnableEXT,
    .CmdSetLogicOpEXT = anv_doom64_CmdSetLogicOpEXT,
    .CmdSetPrimitiveRestartEnable = anv_doom64_CmdSetPrimitiveRestartEnable,
    .CmdSetPrimitiveRestartEnableEXT = anv_doom64_CmdSetPrimitiveRestartEnableEXT,
    .CmdSetTessellationDomainOriginEXT = anv_doom64_CmdSetTessellationDomainOriginEXT,
    .CmdSetDepthClampEnableEXT = anv_doom64_CmdSetDepthClampEnableEXT,
    .CmdSetPolygonModeEXT = anv_doom64_CmdSetPolygonModeEXT,
    .CmdSetRasterizationSamplesEXT = anv_doom64_CmdSetRasterizationSamplesEXT,
    .CmdSetSampleMaskEXT = anv_doom64_CmdSetSampleMaskEXT,
    .CmdSetAlphaToCoverageEnableEXT = anv_doom64_CmdSetAlphaToCoverageEnableEXT,
    .CmdSetAlphaToOneEnableEXT = anv_doom64_CmdSetAlphaToOneEnableEXT,
    .CmdSetLogicOpEnableEXT = anv_doom64_CmdSetLogicOpEnableEXT,
    .CmdSetColorBlendEnableEXT = anv_doom64_CmdSetColorBlendEnableEXT,
    .CmdSetColorBlendEquationEXT = anv_doom64_CmdSetColorBlendEquationEXT,
    .CmdSetColorWriteMaskEXT = anv_doom64_CmdSetColorWriteMaskEXT,
    .CmdSetRasterizationStreamEXT = anv_doom64_CmdSetRasterizationStreamEXT,
    .CmdSetConservativeRasterizationModeEXT = anv_doom64_CmdSetConservativeRasterizationModeEXT,
    .CmdSetExtraPrimitiveOverestimationSizeEXT = anv_doom64_CmdSetExtraPrimitiveOverestimationSizeEXT,
    .CmdSetDepthClipEnableEXT = anv_doom64_CmdSetDepthClipEnableEXT,
    .CmdSetSampleLocationsEnableEXT = anv_doom64_CmdSetSampleLocationsEnableEXT,
    .CmdSetColorBlendAdvancedEXT = anv_doom64_CmdSetColorBlendAdvancedEXT,
    .CmdSetProvokingVertexModeEXT = anv_doom64_CmdSetProvokingVertexModeEXT,
    .CmdSetLineRasterizationModeEXT = anv_doom64_CmdSetLineRasterizationModeEXT,
    .CmdSetLineStippleEnableEXT = anv_doom64_CmdSetLineStippleEnableEXT,
    .CmdSetDepthClipNegativeOneToOneEXT = anv_doom64_CmdSetDepthClipNegativeOneToOneEXT,
    .CmdSetViewportWScalingEnableNV = anv_doom64_CmdSetViewportWScalingEnableNV,
    .CmdSetViewportSwizzleNV = anv_doom64_CmdSetViewportSwizzleNV,
    .CmdSetCoverageToColorEnableNV = anv_doom64_CmdSetCoverageToColorEnableNV,
    .CmdSetCoverageToColorLocationNV = anv_doom64_CmdSetCoverageToColorLocationNV,
    .CmdSetCoverageModulationModeNV = anv_doom64_CmdSetCoverageModulationModeNV,
    .CmdSetCoverageModulationTableEnableNV = anv_doom64_CmdSetCoverageModulationTableEnableNV,
    .CmdSetCoverageModulationTableNV = anv_doom64_CmdSetCoverageModulationTableNV,
    .CmdSetShadingRateImageEnableNV = anv_doom64_CmdSetShadingRateImageEnableNV,
    .CmdSetCoverageReductionModeNV = anv_doom64_CmdSetCoverageReductionModeNV,
    .CmdSetRepresentativeFragmentTestEnableNV = anv_doom64_CmdSetRepresentativeFragmentTestEnableNV,
    .CreatePrivateDataSlot = anv_doom64_CreatePrivateDataSlot,
    .CreatePrivateDataSlotEXT = anv_doom64_CreatePrivateDataSlotEXT,
    .DestroyPrivateDataSlot = anv_doom64_DestroyPrivateDataSlot,
    .DestroyPrivateDataSlotEXT = anv_doom64_DestroyPrivateDataSlotEXT,
    .SetPrivateData = anv_doom64_SetPrivateData,
    .SetPrivateDataEXT = anv_doom64_SetPrivateDataEXT,
    .GetPrivateData = anv_doom64_GetPrivateData,
    .GetPrivateDataEXT = anv_doom64_GetPrivateDataEXT,
    .CmdCopyBuffer2 = anv_doom64_CmdCopyBuffer2,
    .CmdCopyBuffer2KHR = anv_doom64_CmdCopyBuffer2KHR,
    .CmdCopyImage2 = anv_doom64_CmdCopyImage2,
    .CmdCopyImage2KHR = anv_doom64_CmdCopyImage2KHR,
    .CmdBlitImage2 = anv_doom64_CmdBlitImage2,
    .CmdBlitImage2KHR = anv_doom64_CmdBlitImage2KHR,
    .CmdCopyBufferToImage2 = anv_doom64_CmdCopyBufferToImage2,
    .CmdCopyBufferToImage2KHR = anv_doom64_CmdCopyBufferToImage2KHR,
    .CmdCopyImageToBuffer2 = anv_doom64_CmdCopyImageToBuffer2,
    .CmdCopyImageToBuffer2KHR = anv_doom64_CmdCopyImageToBuffer2KHR,
    .CmdResolveImage2 = anv_doom64_CmdResolveImage2,
    .CmdResolveImage2KHR = anv_doom64_CmdResolveImage2KHR,
    .CmdSetFragmentShadingRateKHR = anv_doom64_CmdSetFragmentShadingRateKHR,
    .CmdSetFragmentShadingRateEnumNV = anv_doom64_CmdSetFragmentShadingRateEnumNV,
    .GetAccelerationStructureBuildSizesKHR = anv_doom64_GetAccelerationStructureBuildSizesKHR,
    .CmdSetVertexInputEXT = anv_doom64_CmdSetVertexInputEXT,
    .CmdSetColorWriteEnableEXT = anv_doom64_CmdSetColorWriteEnableEXT,
    .CmdSetEvent2 = anv_doom64_CmdSetEvent2,
    .CmdSetEvent2KHR = anv_doom64_CmdSetEvent2KHR,
    .CmdResetEvent2 = anv_doom64_CmdResetEvent2,
    .CmdResetEvent2KHR = anv_doom64_CmdResetEvent2KHR,
    .CmdWaitEvents2 = anv_doom64_CmdWaitEvents2,
    .CmdWaitEvents2KHR = anv_doom64_CmdWaitEvents2KHR,
    .CmdPipelineBarrier2 = anv_doom64_CmdPipelineBarrier2,
    .CmdPipelineBarrier2KHR = anv_doom64_CmdPipelineBarrier2KHR,
    .QueueSubmit2 = anv_doom64_QueueSubmit2,
    .QueueSubmit2KHR = anv_doom64_QueueSubmit2KHR,
    .CmdWriteTimestamp2 = anv_doom64_CmdWriteTimestamp2,
    .CmdWriteTimestamp2KHR = anv_doom64_CmdWriteTimestamp2KHR,
    .CmdWriteBufferMarker2AMD = anv_doom64_CmdWriteBufferMarker2AMD,
    .GetQueueCheckpointData2NV = anv_doom64_GetQueueCheckpointData2NV,
    .CopyMemoryToImage = anv_doom64_CopyMemoryToImage,
    .CopyMemoryToImageEXT = anv_doom64_CopyMemoryToImageEXT,
    .CopyImageToMemory = anv_doom64_CopyImageToMemory,
    .CopyImageToMemoryEXT = anv_doom64_CopyImageToMemoryEXT,
    .CopyImageToImage = anv_doom64_CopyImageToImage,
    .CopyImageToImageEXT = anv_doom64_CopyImageToImageEXT,
    .TransitionImageLayout = anv_doom64_TransitionImageLayout,
    .TransitionImageLayoutEXT = anv_doom64_TransitionImageLayoutEXT,
    .CreateVideoSessionKHR = anv_doom64_CreateVideoSessionKHR,
    .DestroyVideoSessionKHR = anv_doom64_DestroyVideoSessionKHR,
    .CreateVideoSessionParametersKHR = anv_doom64_CreateVideoSessionParametersKHR,
    .UpdateVideoSessionParametersKHR = anv_doom64_UpdateVideoSessionParametersKHR,
    .GetEncodedVideoSessionParametersKHR = anv_doom64_GetEncodedVideoSessionParametersKHR,
    .DestroyVideoSessionParametersKHR = anv_doom64_DestroyVideoSessionParametersKHR,
    .GetVideoSessionMemoryRequirementsKHR = anv_doom64_GetVideoSessionMemoryRequirementsKHR,
    .BindVideoSessionMemoryKHR = anv_doom64_BindVideoSessionMemoryKHR,
    .CmdDecodeVideoKHR = anv_doom64_CmdDecodeVideoKHR,
    .CmdBeginVideoCodingKHR = anv_doom64_CmdBeginVideoCodingKHR,
    .CmdControlVideoCodingKHR = anv_doom64_CmdControlVideoCodingKHR,
    .CmdEndVideoCodingKHR = anv_doom64_CmdEndVideoCodingKHR,
    .CmdEncodeVideoKHR = anv_doom64_CmdEncodeVideoKHR,
    .CmdDecompressMemoryNV = anv_doom64_CmdDecompressMemoryNV,
    .CmdDecompressMemoryIndirectCountNV = anv_doom64_CmdDecompressMemoryIndirectCountNV,
    .GetPartitionedAccelerationStructuresBuildSizesNV = anv_doom64_GetPartitionedAccelerationStructuresBuildSizesNV,
    .CmdBuildPartitionedAccelerationStructuresNV = anv_doom64_CmdBuildPartitionedAccelerationStructuresNV,
    .CmdDecompressMemoryEXT = anv_doom64_CmdDecompressMemoryEXT,
    .CmdDecompressMemoryIndirectCountEXT = anv_doom64_CmdDecompressMemoryIndirectCountEXT,
    .CreateCuModuleNVX = anv_doom64_CreateCuModuleNVX,
    .CreateCuFunctionNVX = anv_doom64_CreateCuFunctionNVX,
    .DestroyCuModuleNVX = anv_doom64_DestroyCuModuleNVX,
    .DestroyCuFunctionNVX = anv_doom64_DestroyCuFunctionNVX,
    .CmdCuLaunchKernelNVX = anv_doom64_CmdCuLaunchKernelNVX,
    .GetDescriptorSetLayoutSizeEXT = anv_doom64_GetDescriptorSetLayoutSizeEXT,
    .GetDescriptorSetLayoutBindingOffsetEXT = anv_doom64_GetDescriptorSetLayoutBindingOffsetEXT,
    .GetDescriptorEXT = anv_doom64_GetDescriptorEXT,
    .CmdBindDescriptorBuffersEXT = anv_doom64_CmdBindDescriptorBuffersEXT,
    .CmdSetDescriptorBufferOffsetsEXT = anv_doom64_CmdSetDescriptorBufferOffsetsEXT,
    .CmdBindDescriptorBufferEmbeddedSamplersEXT = anv_doom64_CmdBindDescriptorBufferEmbeddedSamplersEXT,
    .GetBufferOpaqueCaptureDescriptorDataEXT = anv_doom64_GetBufferOpaqueCaptureDescriptorDataEXT,
    .GetImageOpaqueCaptureDescriptorDataEXT = anv_doom64_GetImageOpaqueCaptureDescriptorDataEXT,
    .GetImageViewOpaqueCaptureDescriptorDataEXT = anv_doom64_GetImageViewOpaqueCaptureDescriptorDataEXT,
    .GetSamplerOpaqueCaptureDescriptorDataEXT = anv_doom64_GetSamplerOpaqueCaptureDescriptorDataEXT,
    .GetAccelerationStructureOpaqueCaptureDescriptorDataEXT = anv_doom64_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT,
    .SetDeviceMemoryPriorityEXT = anv_doom64_SetDeviceMemoryPriorityEXT,
    .WaitForPresent2KHR = anv_doom64_WaitForPresent2KHR,
    .WaitForPresentKHR = anv_doom64_WaitForPresentKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateBufferCollectionFUCHSIA = anv_doom64_CreateBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .CreateBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionBufferConstraintsFUCHSIA = anv_doom64_SetBufferCollectionBufferConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionBufferConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionImageConstraintsFUCHSIA = anv_doom64_SetBufferCollectionImageConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionImageConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .DestroyBufferCollectionFUCHSIA = anv_doom64_DestroyBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .DestroyBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetBufferCollectionPropertiesFUCHSIA = anv_doom64_GetBufferCollectionPropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetBufferCollectionPropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .CmdBeginRendering = anv_doom64_CmdBeginRendering,
    .CmdBeginRenderingKHR = anv_doom64_CmdBeginRenderingKHR,
    .CmdEndRendering = anv_doom64_CmdEndRendering,
    .CmdEndRendering2KHR = anv_doom64_CmdEndRendering2KHR,
    .CmdEndRendering2EXT = anv_doom64_CmdEndRendering2EXT,
    .CmdEndRenderingKHR = anv_doom64_CmdEndRenderingKHR,
    .GetDescriptorSetLayoutHostMappingInfoVALVE = anv_doom64_GetDescriptorSetLayoutHostMappingInfoVALVE,
    .GetDescriptorSetHostMappingVALVE = anv_doom64_GetDescriptorSetHostMappingVALVE,
    .CreateMicromapEXT = anv_doom64_CreateMicromapEXT,
    .CmdBuildMicromapsEXT = anv_doom64_CmdBuildMicromapsEXT,
    .BuildMicromapsEXT = anv_doom64_BuildMicromapsEXT,
    .DestroyMicromapEXT = anv_doom64_DestroyMicromapEXT,
    .CmdCopyMicromapEXT = anv_doom64_CmdCopyMicromapEXT,
    .CopyMicromapEXT = anv_doom64_CopyMicromapEXT,
    .CmdCopyMicromapToMemoryEXT = anv_doom64_CmdCopyMicromapToMemoryEXT,
    .CopyMicromapToMemoryEXT = anv_doom64_CopyMicromapToMemoryEXT,
    .CmdCopyMemoryToMicromapEXT = anv_doom64_CmdCopyMemoryToMicromapEXT,
    .CopyMemoryToMicromapEXT = anv_doom64_CopyMemoryToMicromapEXT,
    .CmdWriteMicromapsPropertiesEXT = anv_doom64_CmdWriteMicromapsPropertiesEXT,
    .WriteMicromapsPropertiesEXT = anv_doom64_WriteMicromapsPropertiesEXT,
    .GetDeviceMicromapCompatibilityEXT = anv_doom64_GetDeviceMicromapCompatibilityEXT,
    .GetMicromapBuildSizesEXT = anv_doom64_GetMicromapBuildSizesEXT,
    .GetShaderModuleIdentifierEXT = anv_doom64_GetShaderModuleIdentifierEXT,
    .GetShaderModuleCreateInfoIdentifierEXT = anv_doom64_GetShaderModuleCreateInfoIdentifierEXT,
    .GetImageSubresourceLayout2 = anv_doom64_GetImageSubresourceLayout2,
    .GetImageSubresourceLayout2KHR = anv_doom64_GetImageSubresourceLayout2KHR,
    .GetImageSubresourceLayout2EXT = anv_doom64_GetImageSubresourceLayout2EXT,
    .GetPipelinePropertiesEXT = anv_doom64_GetPipelinePropertiesEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .ExportMetalObjectsEXT = anv_doom64_ExportMetalObjectsEXT,
#elif defined(_MSC_VER)
    .ExportMetalObjectsEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CmdBindTileMemoryQCOM = anv_doom64_CmdBindTileMemoryQCOM,
    .GetFramebufferTilePropertiesQCOM = anv_doom64_GetFramebufferTilePropertiesQCOM,
    .GetDynamicRenderingTilePropertiesQCOM = anv_doom64_GetDynamicRenderingTilePropertiesQCOM,
    .CreateOpticalFlowSessionNV = anv_doom64_CreateOpticalFlowSessionNV,
    .DestroyOpticalFlowSessionNV = anv_doom64_DestroyOpticalFlowSessionNV,
    .BindOpticalFlowSessionImageNV = anv_doom64_BindOpticalFlowSessionImageNV,
    .CmdOpticalFlowExecuteNV = anv_doom64_CmdOpticalFlowExecuteNV,
    .GetDeviceFaultInfoEXT = anv_doom64_GetDeviceFaultInfoEXT,
    .GetDeviceFaultReportsKHR = anv_doom64_GetDeviceFaultReportsKHR,
    .GetDeviceFaultDebugInfoKHR = anv_doom64_GetDeviceFaultDebugInfoKHR,
    .CmdSetDepthBias2EXT = anv_doom64_CmdSetDepthBias2EXT,
    .ReleaseSwapchainImagesKHR = anv_doom64_ReleaseSwapchainImagesKHR,
    .ReleaseSwapchainImagesEXT = anv_doom64_ReleaseSwapchainImagesEXT,
    .GetDeviceImageSubresourceLayout = anv_doom64_GetDeviceImageSubresourceLayout,
    .GetDeviceImageSubresourceLayoutKHR = anv_doom64_GetDeviceImageSubresourceLayoutKHR,
    .MapMemory2 = anv_doom64_MapMemory2,
    .MapMemory2KHR = anv_doom64_MapMemory2KHR,
    .UnmapMemory2 = anv_doom64_UnmapMemory2,
    .UnmapMemory2KHR = anv_doom64_UnmapMemory2KHR,
    .CreateShadersEXT = anv_doom64_CreateShadersEXT,
    .DestroyShaderEXT = anv_doom64_DestroyShaderEXT,
    .GetShaderBinaryDataEXT = anv_doom64_GetShaderBinaryDataEXT,
    .CmdBindShadersEXT = anv_doom64_CmdBindShadersEXT,
    .SetSwapchainPresentTimingQueueSizeEXT = anv_doom64_SetSwapchainPresentTimingQueueSizeEXT,
    .GetSwapchainTimingPropertiesEXT = anv_doom64_GetSwapchainTimingPropertiesEXT,
    .GetSwapchainTimeDomainPropertiesEXT = anv_doom64_GetSwapchainTimeDomainPropertiesEXT,
    .GetPastPresentationTimingEXT = anv_doom64_GetPastPresentationTimingEXT,
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetScreenBufferPropertiesQNX = anv_doom64_GetScreenBufferPropertiesQNX,
#elif defined(_MSC_VER)
    .GetScreenBufferPropertiesQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CmdBindDescriptorSets2 = anv_doom64_CmdBindDescriptorSets2,
    .CmdBindDescriptorSets2KHR = anv_doom64_CmdBindDescriptorSets2KHR,
    .CmdPushConstants2 = anv_doom64_CmdPushConstants2,
    .CmdPushConstants2KHR = anv_doom64_CmdPushConstants2KHR,
    .CmdPushDescriptorSet2 = anv_doom64_CmdPushDescriptorSet2,
    .CmdPushDescriptorSet2KHR = anv_doom64_CmdPushDescriptorSet2KHR,
    .CmdPushDescriptorSetWithTemplate2 = anv_doom64_CmdPushDescriptorSetWithTemplate2,
    .CmdPushDescriptorSetWithTemplate2KHR = anv_doom64_CmdPushDescriptorSetWithTemplate2KHR,
    .CmdSetDescriptorBufferOffsets2EXT = anv_doom64_CmdSetDescriptorBufferOffsets2EXT,
    .CmdBindDescriptorBufferEmbeddedSamplers2EXT = anv_doom64_CmdBindDescriptorBufferEmbeddedSamplers2EXT,
    .SetLatencySleepModeNV = anv_doom64_SetLatencySleepModeNV,
    .LatencySleepNV = anv_doom64_LatencySleepNV,
    .SetLatencyMarkerNV = anv_doom64_SetLatencyMarkerNV,
    .GetLatencyTimingsNV = anv_doom64_GetLatencyTimingsNV,
    .QueueNotifyOutOfBandNV = anv_doom64_QueueNotifyOutOfBandNV,
    .CmdSetRenderingAttachmentLocations = anv_doom64_CmdSetRenderingAttachmentLocations,
    .CmdSetRenderingAttachmentLocationsKHR = anv_doom64_CmdSetRenderingAttachmentLocationsKHR,
    .CmdSetRenderingInputAttachmentIndices = anv_doom64_CmdSetRenderingInputAttachmentIndices,
    .CmdSetRenderingInputAttachmentIndicesKHR = anv_doom64_CmdSetRenderingInputAttachmentIndicesKHR,
    .CmdSetDepthClampRangeEXT = anv_doom64_CmdSetDepthClampRangeEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandleEXT = anv_doom64_GetMemoryMetalHandleEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandleEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandlePropertiesEXT = anv_doom64_GetMemoryMetalHandlePropertiesEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandlePropertiesEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .ConvertCooperativeVectorMatrixNV = anv_doom64_ConvertCooperativeVectorMatrixNV,
    .CmdConvertCooperativeVectorMatrixNV = anv_doom64_CmdConvertCooperativeVectorMatrixNV,
    .CmdDispatchTileQCOM = anv_doom64_CmdDispatchTileQCOM,
    .CmdBeginPerTileExecutionQCOM = anv_doom64_CmdBeginPerTileExecutionQCOM,
    .CmdEndPerTileExecutionQCOM = anv_doom64_CmdEndPerTileExecutionQCOM,
    .CreateExternalComputeQueueNV = anv_doom64_CreateExternalComputeQueueNV,
    .DestroyExternalComputeQueueNV = anv_doom64_DestroyExternalComputeQueueNV,
    .CreateShaderInstrumentationARM = anv_doom64_CreateShaderInstrumentationARM,
    .DestroyShaderInstrumentationARM = anv_doom64_DestroyShaderInstrumentationARM,
    .CmdBeginShaderInstrumentationARM = anv_doom64_CmdBeginShaderInstrumentationARM,
    .CmdEndShaderInstrumentationARM = anv_doom64_CmdEndShaderInstrumentationARM,
    .GetShaderInstrumentationValuesARM = anv_doom64_GetShaderInstrumentationValuesARM,
    .ClearShaderInstrumentationMetricsARM = anv_doom64_ClearShaderInstrumentationMetricsARM,
    .CreateTensorARM = anv_doom64_CreateTensorARM,
    .DestroyTensorARM = anv_doom64_DestroyTensorARM,
    .CreateTensorViewARM = anv_doom64_CreateTensorViewARM,
    .DestroyTensorViewARM = anv_doom64_DestroyTensorViewARM,
    .GetTensorMemoryRequirementsARM = anv_doom64_GetTensorMemoryRequirementsARM,
    .BindTensorMemoryARM = anv_doom64_BindTensorMemoryARM,
    .GetDeviceTensorMemoryRequirementsARM = anv_doom64_GetDeviceTensorMemoryRequirementsARM,
    .CmdCopyTensorARM = anv_doom64_CmdCopyTensorARM,
    .GetTensorOpaqueCaptureDescriptorDataARM = anv_doom64_GetTensorOpaqueCaptureDescriptorDataARM,
    .GetTensorViewOpaqueCaptureDescriptorDataARM = anv_doom64_GetTensorViewOpaqueCaptureDescriptorDataARM,
    .CreateDataGraphPipelinesARM = anv_doom64_CreateDataGraphPipelinesARM,
    .CreateDataGraphPipelineSessionARM = anv_doom64_CreateDataGraphPipelineSessionARM,
    .GetDataGraphPipelineSessionBindPointRequirementsARM = anv_doom64_GetDataGraphPipelineSessionBindPointRequirementsARM,
    .GetDataGraphPipelineSessionMemoryRequirementsARM = anv_doom64_GetDataGraphPipelineSessionMemoryRequirementsARM,
    .BindDataGraphPipelineSessionMemoryARM = anv_doom64_BindDataGraphPipelineSessionMemoryARM,
    .DestroyDataGraphPipelineSessionARM = anv_doom64_DestroyDataGraphPipelineSessionARM,
    .CmdDispatchDataGraphARM = anv_doom64_CmdDispatchDataGraphARM,
    .GetDataGraphPipelineAvailablePropertiesARM = anv_doom64_GetDataGraphPipelineAvailablePropertiesARM,
    .GetDataGraphPipelinePropertiesARM = anv_doom64_GetDataGraphPipelinePropertiesARM,
#ifdef VK_USE_PLATFORM_OHOS
    .GetNativeBufferPropertiesOHOS = anv_doom64_GetNativeBufferPropertiesOHOS,
#elif defined(_MSC_VER)
    .GetNativeBufferPropertiesOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
    .GetMemoryNativeBufferOHOS = anv_doom64_GetMemoryNativeBufferOHOS,
#elif defined(_MSC_VER)
    .GetMemoryNativeBufferOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .QueueSetPerfHintQCOM = anv_doom64_QueueSetPerfHintQCOM,
    .CmdSetComputeOccupancyPriorityNV = anv_doom64_CmdSetComputeOccupancyPriorityNV,
    .WriteSamplerDescriptorsEXT = anv_doom64_WriteSamplerDescriptorsEXT,
    .WriteResourceDescriptorsEXT = anv_doom64_WriteResourceDescriptorsEXT,
    .CmdBindSamplerHeapEXT = anv_doom64_CmdBindSamplerHeapEXT,
    .CmdBindResourceHeapEXT = anv_doom64_CmdBindResourceHeapEXT,
    .CmdPushDataEXT = anv_doom64_CmdPushDataEXT,
    .RegisterCustomBorderColorEXT = anv_doom64_RegisterCustomBorderColorEXT,
    .UnregisterCustomBorderColorEXT = anv_doom64_UnregisterCustomBorderColorEXT,
    .GetImageOpaqueCaptureDataEXT = anv_doom64_GetImageOpaqueCaptureDataEXT,
    .GetTensorOpaqueCaptureDataARM = anv_doom64_GetTensorOpaqueCaptureDataARM,
    .CmdCopyMemoryKHR = anv_doom64_CmdCopyMemoryKHR,
    .CmdCopyMemoryToImageKHR = anv_doom64_CmdCopyMemoryToImageKHR,
    .CmdCopyImageToMemoryKHR = anv_doom64_CmdCopyImageToMemoryKHR,
    .CmdUpdateMemoryKHR = anv_doom64_CmdUpdateMemoryKHR,
    .CmdFillMemoryKHR = anv_doom64_CmdFillMemoryKHR,
    .CmdCopyQueryPoolResultsToMemoryKHR = anv_doom64_CmdCopyQueryPoolResultsToMemoryKHR,
    .CmdBeginConditionalRendering2EXT = anv_doom64_CmdBeginConditionalRendering2EXT,
    .CmdBindTransformFeedbackBuffers2EXT = anv_doom64_CmdBindTransformFeedbackBuffers2EXT,
    .CmdBeginTransformFeedback2EXT = anv_doom64_CmdBeginTransformFeedback2EXT,
    .CmdEndTransformFeedback2EXT = anv_doom64_CmdEndTransformFeedback2EXT,
    .CmdDrawIndirectByteCount2EXT = anv_doom64_CmdDrawIndirectByteCount2EXT,
    .CmdWriteMarkerToMemoryAMD = anv_doom64_CmdWriteMarkerToMemoryAMD,
    .CmdBindIndexBuffer3KHR = anv_doom64_CmdBindIndexBuffer3KHR,
    .CmdBindVertexBuffers3KHR = anv_doom64_CmdBindVertexBuffers3KHR,
    .CmdDrawIndirect2KHR = anv_doom64_CmdDrawIndirect2KHR,
    .CmdDrawIndexedIndirect2KHR = anv_doom64_CmdDrawIndexedIndirect2KHR,
    .CmdDrawIndirectCount2KHR = anv_doom64_CmdDrawIndirectCount2KHR,
    .CmdDrawIndexedIndirectCount2KHR = anv_doom64_CmdDrawIndexedIndirectCount2KHR,
    .CmdDrawMeshTasksIndirect2EXT = anv_doom64_CmdDrawMeshTasksIndirect2EXT,
    .CmdDrawMeshTasksIndirectCount2EXT = anv_doom64_CmdDrawMeshTasksIndirectCount2EXT,
    .CmdDispatchIndirect2KHR = anv_doom64_CmdDispatchIndirect2KHR,
    .CreateAccelerationStructure2KHR = anv_doom64_CreateAccelerationStructure2KHR,
    .CmdSetDispatchParametersARM = anv_doom64_CmdSetDispatchParametersARM,
};
const struct vk_device_entrypoint_table anv_furmark_device_entrypoints = {
    .GetDeviceProcAddr = anv_furmark_GetDeviceProcAddr,
    .DestroyDevice = anv_furmark_DestroyDevice,
    .GetDeviceQueue = anv_furmark_GetDeviceQueue,
    .QueueSubmit = anv_furmark_QueueSubmit,
    .QueueWaitIdle = anv_furmark_QueueWaitIdle,
    .DeviceWaitIdle = anv_furmark_DeviceWaitIdle,
    .AllocateMemory = anv_furmark_AllocateMemory,
    .FreeMemory = anv_furmark_FreeMemory,
    .MapMemory = anv_furmark_MapMemory,
    .UnmapMemory = anv_furmark_UnmapMemory,
    .FlushMappedMemoryRanges = anv_furmark_FlushMappedMemoryRanges,
    .InvalidateMappedMemoryRanges = anv_furmark_InvalidateMappedMemoryRanges,
    .GetDeviceMemoryCommitment = anv_furmark_GetDeviceMemoryCommitment,
    .GetBufferMemoryRequirements = anv_furmark_GetBufferMemoryRequirements,
    .BindBufferMemory = anv_furmark_BindBufferMemory,
    .GetImageMemoryRequirements = anv_furmark_GetImageMemoryRequirements,
    .BindImageMemory = anv_furmark_BindImageMemory,
    .GetImageSparseMemoryRequirements = anv_furmark_GetImageSparseMemoryRequirements,
    .QueueBindSparse = anv_furmark_QueueBindSparse,
    .CreateFence = anv_furmark_CreateFence,
    .DestroyFence = anv_furmark_DestroyFence,
    .ResetFences = anv_furmark_ResetFences,
    .GetFenceStatus = anv_furmark_GetFenceStatus,
    .WaitForFences = anv_furmark_WaitForFences,
    .CreateSemaphore = anv_furmark_CreateSemaphore,
    .DestroySemaphore = anv_furmark_DestroySemaphore,
    .CreateEvent = anv_furmark_CreateEvent,
    .DestroyEvent = anv_furmark_DestroyEvent,
    .GetEventStatus = anv_furmark_GetEventStatus,
    .SetEvent = anv_furmark_SetEvent,
    .ResetEvent = anv_furmark_ResetEvent,
    .CreateQueryPool = anv_furmark_CreateQueryPool,
    .DestroyQueryPool = anv_furmark_DestroyQueryPool,
    .GetQueryPoolResults = anv_furmark_GetQueryPoolResults,
    .ResetQueryPool = anv_furmark_ResetQueryPool,
    .ResetQueryPoolEXT = anv_furmark_ResetQueryPoolEXT,
    .CreateBuffer = anv_furmark_CreateBuffer,
    .DestroyBuffer = anv_furmark_DestroyBuffer,
    .CreateBufferView = anv_furmark_CreateBufferView,
    .DestroyBufferView = anv_furmark_DestroyBufferView,
    .CreateImage = anv_furmark_CreateImage,
    .DestroyImage = anv_furmark_DestroyImage,
    .GetImageSubresourceLayout = anv_furmark_GetImageSubresourceLayout,
    .CreateImageView = anv_furmark_CreateImageView,
    .DestroyImageView = anv_furmark_DestroyImageView,
    .CreateShaderModule = anv_furmark_CreateShaderModule,
    .DestroyShaderModule = anv_furmark_DestroyShaderModule,
    .CreatePipelineCache = anv_furmark_CreatePipelineCache,
    .DestroyPipelineCache = anv_furmark_DestroyPipelineCache,
    .GetPipelineCacheData = anv_furmark_GetPipelineCacheData,
    .MergePipelineCaches = anv_furmark_MergePipelineCaches,
    .CreatePipelineBinariesKHR = anv_furmark_CreatePipelineBinariesKHR,
    .DestroyPipelineBinaryKHR = anv_furmark_DestroyPipelineBinaryKHR,
    .GetPipelineKeyKHR = anv_furmark_GetPipelineKeyKHR,
    .GetPipelineBinaryDataKHR = anv_furmark_GetPipelineBinaryDataKHR,
    .ReleaseCapturedPipelineDataKHR = anv_furmark_ReleaseCapturedPipelineDataKHR,
    .CreateGraphicsPipelines = anv_furmark_CreateGraphicsPipelines,
    .CreateComputePipelines = anv_furmark_CreateComputePipelines,
    .GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI = anv_furmark_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI,
    .DestroyPipeline = anv_furmark_DestroyPipeline,
    .CreatePipelineLayout = anv_furmark_CreatePipelineLayout,
    .DestroyPipelineLayout = anv_furmark_DestroyPipelineLayout,
    .CreateSampler = anv_furmark_CreateSampler,
    .DestroySampler = anv_furmark_DestroySampler,
    .CreateDescriptorSetLayout = anv_furmark_CreateDescriptorSetLayout,
    .DestroyDescriptorSetLayout = anv_furmark_DestroyDescriptorSetLayout,
    .CreateDescriptorPool = anv_furmark_CreateDescriptorPool,
    .DestroyDescriptorPool = anv_furmark_DestroyDescriptorPool,
    .ResetDescriptorPool = anv_furmark_ResetDescriptorPool,
    .AllocateDescriptorSets = anv_furmark_AllocateDescriptorSets,
    .FreeDescriptorSets = anv_furmark_FreeDescriptorSets,
    .UpdateDescriptorSets = anv_furmark_UpdateDescriptorSets,
    .CreateFramebuffer = anv_furmark_CreateFramebuffer,
    .DestroyFramebuffer = anv_furmark_DestroyFramebuffer,
    .CreateRenderPass = anv_furmark_CreateRenderPass,
    .DestroyRenderPass = anv_furmark_DestroyRenderPass,
    .GetRenderAreaGranularity = anv_furmark_GetRenderAreaGranularity,
    .GetRenderingAreaGranularity = anv_furmark_GetRenderingAreaGranularity,
    .GetRenderingAreaGranularityKHR = anv_furmark_GetRenderingAreaGranularityKHR,
    .CreateCommandPool = anv_furmark_CreateCommandPool,
    .DestroyCommandPool = anv_furmark_DestroyCommandPool,
    .ResetCommandPool = anv_furmark_ResetCommandPool,
    .AllocateCommandBuffers = anv_furmark_AllocateCommandBuffers,
    .FreeCommandBuffers = anv_furmark_FreeCommandBuffers,
    .BeginCommandBuffer = anv_furmark_BeginCommandBuffer,
    .EndCommandBuffer = anv_furmark_EndCommandBuffer,
    .ResetCommandBuffer = anv_furmark_ResetCommandBuffer,
    .CmdBindPipeline = anv_furmark_CmdBindPipeline,
    .CmdSetPrimitiveRestartIndexEXT = anv_furmark_CmdSetPrimitiveRestartIndexEXT,
    .CmdSetAttachmentFeedbackLoopEnableEXT = anv_furmark_CmdSetAttachmentFeedbackLoopEnableEXT,
    .CmdSetViewport = anv_furmark_CmdSetViewport,
    .CmdSetScissor = anv_furmark_CmdSetScissor,
    .CmdSetLineWidth = anv_furmark_CmdSetLineWidth,
    .CmdSetDepthBias = anv_furmark_CmdSetDepthBias,
    .CmdSetBlendConstants = anv_furmark_CmdSetBlendConstants,
    .CmdSetDepthBounds = anv_furmark_CmdSetDepthBounds,
    .CmdSetStencilCompareMask = anv_furmark_CmdSetStencilCompareMask,
    .CmdSetStencilWriteMask = anv_furmark_CmdSetStencilWriteMask,
    .CmdSetStencilReference = anv_furmark_CmdSetStencilReference,
    .CmdBindDescriptorSets = anv_furmark_CmdBindDescriptorSets,
    .CmdBindIndexBuffer = anv_furmark_CmdBindIndexBuffer,
    .CmdBindVertexBuffers = anv_furmark_CmdBindVertexBuffers,
    .CmdDraw = anv_furmark_CmdDraw,
    .CmdDrawIndexed = anv_furmark_CmdDrawIndexed,
    .CmdDrawMultiEXT = anv_furmark_CmdDrawMultiEXT,
    .CmdDrawMultiIndexedEXT = anv_furmark_CmdDrawMultiIndexedEXT,
    .CmdDrawIndirect = anv_furmark_CmdDrawIndirect,
    .CmdDrawIndexedIndirect = anv_furmark_CmdDrawIndexedIndirect,
    .CmdDispatch = anv_furmark_CmdDispatch,
    .CmdDispatchIndirect = anv_furmark_CmdDispatchIndirect,
    .CmdSubpassShadingHUAWEI = anv_furmark_CmdSubpassShadingHUAWEI,
    .CmdDrawClusterHUAWEI = anv_furmark_CmdDrawClusterHUAWEI,
    .CmdDrawClusterIndirectHUAWEI = anv_furmark_CmdDrawClusterIndirectHUAWEI,
    .CmdUpdatePipelineIndirectBufferNV = anv_furmark_CmdUpdatePipelineIndirectBufferNV,
    .CmdCopyBuffer = anv_furmark_CmdCopyBuffer,
    .CmdCopyImage = anv_furmark_CmdCopyImage,
    .CmdBlitImage = anv_furmark_CmdBlitImage,
    .CmdCopyBufferToImage = anv_furmark_CmdCopyBufferToImage,
    .CmdCopyImageToBuffer = anv_furmark_CmdCopyImageToBuffer,
    .CmdCopyMemoryIndirectNV = anv_furmark_CmdCopyMemoryIndirectNV,
    .CmdCopyMemoryIndirectKHR = anv_furmark_CmdCopyMemoryIndirectKHR,
    .CmdCopyMemoryToImageIndirectNV = anv_furmark_CmdCopyMemoryToImageIndirectNV,
    .CmdCopyMemoryToImageIndirectKHR = anv_furmark_CmdCopyMemoryToImageIndirectKHR,
    .CmdUpdateBuffer = anv_furmark_CmdUpdateBuffer,
    .CmdFillBuffer = anv_furmark_CmdFillBuffer,
    .CmdClearColorImage = anv_furmark_CmdClearColorImage,
    .CmdClearDepthStencilImage = anv_furmark_CmdClearDepthStencilImage,
    .CmdClearAttachments = anv_furmark_CmdClearAttachments,
    .CmdResolveImage = anv_furmark_CmdResolveImage,
    .CmdSetEvent = anv_furmark_CmdSetEvent,
    .CmdResetEvent = anv_furmark_CmdResetEvent,
    .CmdWaitEvents = anv_furmark_CmdWaitEvents,
    .CmdPipelineBarrier = anv_furmark_CmdPipelineBarrier,
    .CmdBeginQuery = anv_furmark_CmdBeginQuery,
    .CmdEndQuery = anv_furmark_CmdEndQuery,
    .CmdBeginConditionalRenderingEXT = anv_furmark_CmdBeginConditionalRenderingEXT,
    .CmdEndConditionalRenderingEXT = anv_furmark_CmdEndConditionalRenderingEXT,
    .CmdBeginCustomResolveEXT = anv_furmark_CmdBeginCustomResolveEXT,
    .CmdResetQueryPool = anv_furmark_CmdResetQueryPool,
    .CmdWriteTimestamp = anv_furmark_CmdWriteTimestamp,
    .CmdCopyQueryPoolResults = anv_furmark_CmdCopyQueryPoolResults,
    .CmdPushConstants = anv_furmark_CmdPushConstants,
    .CmdBeginRenderPass = anv_furmark_CmdBeginRenderPass,
    .CmdNextSubpass = anv_furmark_CmdNextSubpass,
    .CmdEndRenderPass = anv_furmark_CmdEndRenderPass,
    .CmdExecuteCommands = anv_furmark_CmdExecuteCommands,
    .CreateSharedSwapchainsKHR = anv_furmark_CreateSharedSwapchainsKHR,
    .CreateSwapchainKHR = anv_furmark_CreateSwapchainKHR,
    .DestroySwapchainKHR = anv_furmark_DestroySwapchainKHR,
    .GetSwapchainImagesKHR = anv_furmark_GetSwapchainImagesKHR,
    .AcquireNextImageKHR = anv_furmark_AcquireNextImageKHR,
    .QueuePresentKHR = anv_furmark_QueuePresentKHR,
    .DebugMarkerSetObjectNameEXT = anv_furmark_DebugMarkerSetObjectNameEXT,
    .DebugMarkerSetObjectTagEXT = anv_furmark_DebugMarkerSetObjectTagEXT,
    .CmdDebugMarkerBeginEXT = anv_furmark_CmdDebugMarkerBeginEXT,
    .CmdDebugMarkerEndEXT = anv_furmark_CmdDebugMarkerEndEXT,
    .CmdDebugMarkerInsertEXT = anv_furmark_CmdDebugMarkerInsertEXT,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleNV = anv_furmark_GetMemoryWin32HandleNV,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .CmdExecuteGeneratedCommandsNV = anv_furmark_CmdExecuteGeneratedCommandsNV,
    .CmdPreprocessGeneratedCommandsNV = anv_furmark_CmdPreprocessGeneratedCommandsNV,
    .CmdBindPipelineShaderGroupNV = anv_furmark_CmdBindPipelineShaderGroupNV,
    .GetGeneratedCommandsMemoryRequirementsNV = anv_furmark_GetGeneratedCommandsMemoryRequirementsNV,
    .CreateIndirectCommandsLayoutNV = anv_furmark_CreateIndirectCommandsLayoutNV,
    .DestroyIndirectCommandsLayoutNV = anv_furmark_DestroyIndirectCommandsLayoutNV,
    .CmdExecuteGeneratedCommandsEXT = anv_furmark_CmdExecuteGeneratedCommandsEXT,
    .CmdPreprocessGeneratedCommandsEXT = anv_furmark_CmdPreprocessGeneratedCommandsEXT,
    .GetGeneratedCommandsMemoryRequirementsEXT = anv_furmark_GetGeneratedCommandsMemoryRequirementsEXT,
    .CreateIndirectCommandsLayoutEXT = anv_furmark_CreateIndirectCommandsLayoutEXT,
    .DestroyIndirectCommandsLayoutEXT = anv_furmark_DestroyIndirectCommandsLayoutEXT,
    .CreateIndirectExecutionSetEXT = anv_furmark_CreateIndirectExecutionSetEXT,
    .DestroyIndirectExecutionSetEXT = anv_furmark_DestroyIndirectExecutionSetEXT,
    .UpdateIndirectExecutionSetPipelineEXT = anv_furmark_UpdateIndirectExecutionSetPipelineEXT,
    .UpdateIndirectExecutionSetShaderEXT = anv_furmark_UpdateIndirectExecutionSetShaderEXT,
    .CmdPushDescriptorSet = anv_furmark_CmdPushDescriptorSet,
    .CmdPushDescriptorSetKHR = anv_furmark_CmdPushDescriptorSetKHR,
    .TrimCommandPool = anv_furmark_TrimCommandPool,
    .TrimCommandPoolKHR = anv_furmark_TrimCommandPoolKHR,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleKHR = anv_furmark_GetMemoryWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandlePropertiesKHR = anv_furmark_GetMemoryWin32HandlePropertiesKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandlePropertiesKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryFdKHR = anv_furmark_GetMemoryFdKHR,
    .GetMemoryFdPropertiesKHR = anv_furmark_GetMemoryFdPropertiesKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandleFUCHSIA = anv_furmark_GetMemoryZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandlePropertiesFUCHSIA = anv_furmark_GetMemoryZirconHandlePropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandlePropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .GetMemoryRemoteAddressNV = anv_furmark_GetMemoryRemoteAddressNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreWin32HandleKHR = anv_furmark_GetSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportSemaphoreWin32HandleKHR = anv_furmark_ImportSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreFdKHR = anv_furmark_GetSemaphoreFdKHR,
    .ImportSemaphoreFdKHR = anv_furmark_ImportSemaphoreFdKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetSemaphoreZirconHandleFUCHSIA = anv_furmark_GetSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .ImportSemaphoreZirconHandleFUCHSIA = anv_furmark_ImportSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .ImportSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetFenceWin32HandleKHR = anv_furmark_GetFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportFenceWin32HandleKHR = anv_furmark_ImportFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetFenceFdKHR = anv_furmark_GetFenceFdKHR,
    .ImportFenceFdKHR = anv_furmark_ImportFenceFdKHR,
    .DisplayPowerControlEXT = anv_furmark_DisplayPowerControlEXT,
    .RegisterDeviceEventEXT = anv_furmark_RegisterDeviceEventEXT,
    .RegisterDisplayEventEXT = anv_furmark_RegisterDisplayEventEXT,
    .GetSwapchainCounterEXT = anv_furmark_GetSwapchainCounterEXT,
    .GetDeviceGroupPeerMemoryFeatures = anv_furmark_GetDeviceGroupPeerMemoryFeatures,
    .GetDeviceGroupPeerMemoryFeaturesKHR = anv_furmark_GetDeviceGroupPeerMemoryFeaturesKHR,
    .BindBufferMemory2 = anv_furmark_BindBufferMemory2,
    .BindBufferMemory2KHR = anv_furmark_BindBufferMemory2KHR,
    .BindImageMemory2 = anv_furmark_BindImageMemory2,
    .BindImageMemory2KHR = anv_furmark_BindImageMemory2KHR,
    .CmdSetDeviceMask = anv_furmark_CmdSetDeviceMask,
    .CmdSetDeviceMaskKHR = anv_furmark_CmdSetDeviceMaskKHR,
    .GetDeviceGroupPresentCapabilitiesKHR = anv_furmark_GetDeviceGroupPresentCapabilitiesKHR,
    .GetDeviceGroupSurfacePresentModesKHR = anv_furmark_GetDeviceGroupSurfacePresentModesKHR,
    .AcquireNextImage2KHR = anv_furmark_AcquireNextImage2KHR,
    .CmdDispatchBase = anv_furmark_CmdDispatchBase,
    .CmdDispatchBaseKHR = anv_furmark_CmdDispatchBaseKHR,
    .CreateDescriptorUpdateTemplate = anv_furmark_CreateDescriptorUpdateTemplate,
    .CreateDescriptorUpdateTemplateKHR = anv_furmark_CreateDescriptorUpdateTemplateKHR,
    .DestroyDescriptorUpdateTemplate = anv_furmark_DestroyDescriptorUpdateTemplate,
    .DestroyDescriptorUpdateTemplateKHR = anv_furmark_DestroyDescriptorUpdateTemplateKHR,
    .UpdateDescriptorSetWithTemplate = anv_furmark_UpdateDescriptorSetWithTemplate,
    .UpdateDescriptorSetWithTemplateKHR = anv_furmark_UpdateDescriptorSetWithTemplateKHR,
    .CmdPushDescriptorSetWithTemplate = anv_furmark_CmdPushDescriptorSetWithTemplate,
    .CmdPushDescriptorSetWithTemplateKHR = anv_furmark_CmdPushDescriptorSetWithTemplateKHR,
    .SetHdrMetadataEXT = anv_furmark_SetHdrMetadataEXT,
    .GetSwapchainStatusKHR = anv_furmark_GetSwapchainStatusKHR,
    .GetRefreshCycleDurationGOOGLE = anv_furmark_GetRefreshCycleDurationGOOGLE,
    .GetPastPresentationTimingGOOGLE = anv_furmark_GetPastPresentationTimingGOOGLE,
    .CmdSetViewportWScalingNV = anv_furmark_CmdSetViewportWScalingNV,
    .CmdSetDiscardRectangleEXT = anv_furmark_CmdSetDiscardRectangleEXT,
    .CmdSetDiscardRectangleEnableEXT = anv_furmark_CmdSetDiscardRectangleEnableEXT,
    .CmdSetDiscardRectangleModeEXT = anv_furmark_CmdSetDiscardRectangleModeEXT,
    .CmdSetSampleLocationsEXT = anv_furmark_CmdSetSampleLocationsEXT,
    .GetBufferMemoryRequirements2 = anv_furmark_GetBufferMemoryRequirements2,
    .GetBufferMemoryRequirements2KHR = anv_furmark_GetBufferMemoryRequirements2KHR,
    .GetImageMemoryRequirements2 = anv_furmark_GetImageMemoryRequirements2,
    .GetImageMemoryRequirements2KHR = anv_furmark_GetImageMemoryRequirements2KHR,
    .GetImageSparseMemoryRequirements2 = anv_furmark_GetImageSparseMemoryRequirements2,
    .GetImageSparseMemoryRequirements2KHR = anv_furmark_GetImageSparseMemoryRequirements2KHR,
    .GetDeviceBufferMemoryRequirements = anv_furmark_GetDeviceBufferMemoryRequirements,
    .GetDeviceBufferMemoryRequirementsKHR = anv_furmark_GetDeviceBufferMemoryRequirementsKHR,
    .GetDeviceImageMemoryRequirements = anv_furmark_GetDeviceImageMemoryRequirements,
    .GetDeviceImageMemoryRequirementsKHR = anv_furmark_GetDeviceImageMemoryRequirementsKHR,
    .GetDeviceImageSparseMemoryRequirements = anv_furmark_GetDeviceImageSparseMemoryRequirements,
    .GetDeviceImageSparseMemoryRequirementsKHR = anv_furmark_GetDeviceImageSparseMemoryRequirementsKHR,
    .CreateSamplerYcbcrConversion = anv_furmark_CreateSamplerYcbcrConversion,
    .CreateSamplerYcbcrConversionKHR = anv_furmark_CreateSamplerYcbcrConversionKHR,
    .DestroySamplerYcbcrConversion = anv_furmark_DestroySamplerYcbcrConversion,
    .DestroySamplerYcbcrConversionKHR = anv_furmark_DestroySamplerYcbcrConversionKHR,
    .GetDeviceQueue2 = anv_furmark_GetDeviceQueue2,
    .CreateValidationCacheEXT = anv_furmark_CreateValidationCacheEXT,
    .DestroyValidationCacheEXT = anv_furmark_DestroyValidationCacheEXT,
    .GetValidationCacheDataEXT = anv_furmark_GetValidationCacheDataEXT,
    .MergeValidationCachesEXT = anv_furmark_MergeValidationCachesEXT,
    .GetDescriptorSetLayoutSupport = anv_furmark_GetDescriptorSetLayoutSupport,
    .GetDescriptorSetLayoutSupportKHR = anv_furmark_GetDescriptorSetLayoutSupportKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsageANDROID = anv_furmark_GetSwapchainGrallocUsageANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsage2ANDROID = anv_furmark_GetSwapchainGrallocUsage2ANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsage2ANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .AcquireImageANDROID = anv_furmark_AcquireImageANDROID,
#elif defined(_MSC_VER)
    .AcquireImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .QueueSignalReleaseImageANDROID = anv_furmark_QueueSignalReleaseImageANDROID,
#elif defined(_MSC_VER)
    .QueueSignalReleaseImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .GetShaderInfoAMD = anv_furmark_GetShaderInfoAMD,
    .SetLocalDimmingAMD = anv_furmark_SetLocalDimmingAMD,
    .GetCalibratedTimestampsKHR = anv_furmark_GetCalibratedTimestampsKHR,
    .GetCalibratedTimestampsEXT = anv_furmark_GetCalibratedTimestampsEXT,
    .SetDebugUtilsObjectNameEXT = anv_furmark_SetDebugUtilsObjectNameEXT,
    .SetDebugUtilsObjectTagEXT = anv_furmark_SetDebugUtilsObjectTagEXT,
    .QueueBeginDebugUtilsLabelEXT = anv_furmark_QueueBeginDebugUtilsLabelEXT,
    .QueueEndDebugUtilsLabelEXT = anv_furmark_QueueEndDebugUtilsLabelEXT,
    .QueueInsertDebugUtilsLabelEXT = anv_furmark_QueueInsertDebugUtilsLabelEXT,
    .CmdBeginDebugUtilsLabelEXT = anv_furmark_CmdBeginDebugUtilsLabelEXT,
    .CmdEndDebugUtilsLabelEXT = anv_furmark_CmdEndDebugUtilsLabelEXT,
    .CmdInsertDebugUtilsLabelEXT = anv_furmark_CmdInsertDebugUtilsLabelEXT,
    .GetMemoryHostPointerPropertiesEXT = anv_furmark_GetMemoryHostPointerPropertiesEXT,
    .CmdWriteBufferMarkerAMD = anv_furmark_CmdWriteBufferMarkerAMD,
    .CreateRenderPass2 = anv_furmark_CreateRenderPass2,
    .CreateRenderPass2KHR = anv_furmark_CreateRenderPass2KHR,
    .CmdBeginRenderPass2 = anv_furmark_CmdBeginRenderPass2,
    .CmdBeginRenderPass2KHR = anv_furmark_CmdBeginRenderPass2KHR,
    .CmdNextSubpass2 = anv_furmark_CmdNextSubpass2,
    .CmdNextSubpass2KHR = anv_furmark_CmdNextSubpass2KHR,
    .CmdEndRenderPass2 = anv_furmark_CmdEndRenderPass2,
    .CmdEndRenderPass2KHR = anv_furmark_CmdEndRenderPass2KHR,
    .GetSemaphoreCounterValue = anv_furmark_GetSemaphoreCounterValue,
    .GetSemaphoreCounterValueKHR = anv_furmark_GetSemaphoreCounterValueKHR,
    .WaitSemaphores = anv_furmark_WaitSemaphores,
    .WaitSemaphoresKHR = anv_furmark_WaitSemaphoresKHR,
    .SignalSemaphore = anv_furmark_SignalSemaphore,
    .SignalSemaphoreKHR = anv_furmark_SignalSemaphoreKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetAndroidHardwareBufferPropertiesANDROID = anv_furmark_GetAndroidHardwareBufferPropertiesANDROID,
#elif defined(_MSC_VER)
    .GetAndroidHardwareBufferPropertiesANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetMemoryAndroidHardwareBufferANDROID = anv_furmark_GetMemoryAndroidHardwareBufferANDROID,
#elif defined(_MSC_VER)
    .GetMemoryAndroidHardwareBufferANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .CmdDrawIndirectCount = anv_furmark_CmdDrawIndirectCount,
    .CmdDrawIndirectCountKHR = anv_furmark_CmdDrawIndirectCountKHR,
    .CmdDrawIndirectCountAMD = anv_furmark_CmdDrawIndirectCountAMD,
    .CmdDrawIndexedIndirectCount = anv_furmark_CmdDrawIndexedIndirectCount,
    .CmdDrawIndexedIndirectCountKHR = anv_furmark_CmdDrawIndexedIndirectCountKHR,
    .CmdDrawIndexedIndirectCountAMD = anv_furmark_CmdDrawIndexedIndirectCountAMD,
    .CmdSetCheckpointNV = anv_furmark_CmdSetCheckpointNV,
    .GetQueueCheckpointDataNV = anv_furmark_GetQueueCheckpointDataNV,
    .CmdBindTransformFeedbackBuffersEXT = anv_furmark_CmdBindTransformFeedbackBuffersEXT,
    .CmdBeginTransformFeedbackEXT = anv_furmark_CmdBeginTransformFeedbackEXT,
    .CmdEndTransformFeedbackEXT = anv_furmark_CmdEndTransformFeedbackEXT,
    .CmdBeginQueryIndexedEXT = anv_furmark_CmdBeginQueryIndexedEXT,
    .CmdEndQueryIndexedEXT = anv_furmark_CmdEndQueryIndexedEXT,
    .CmdDrawIndirectByteCountEXT = anv_furmark_CmdDrawIndirectByteCountEXT,
    .CmdSetExclusiveScissorNV = anv_furmark_CmdSetExclusiveScissorNV,
    .CmdSetExclusiveScissorEnableNV = anv_furmark_CmdSetExclusiveScissorEnableNV,
    .CmdBindShadingRateImageNV = anv_furmark_CmdBindShadingRateImageNV,
    .CmdSetViewportShadingRatePaletteNV = anv_furmark_CmdSetViewportShadingRatePaletteNV,
    .CmdSetCoarseSampleOrderNV = anv_furmark_CmdSetCoarseSampleOrderNV,
    .CmdDrawMeshTasksNV = anv_furmark_CmdDrawMeshTasksNV,
    .CmdDrawMeshTasksIndirectNV = anv_furmark_CmdDrawMeshTasksIndirectNV,
    .CmdDrawMeshTasksIndirectCountNV = anv_furmark_CmdDrawMeshTasksIndirectCountNV,
    .CmdDrawMeshTasksEXT = anv_furmark_CmdDrawMeshTasksEXT,
    .CmdDrawMeshTasksIndirectEXT = anv_furmark_CmdDrawMeshTasksIndirectEXT,
    .CmdDrawMeshTasksIndirectCountEXT = anv_furmark_CmdDrawMeshTasksIndirectCountEXT,
    .CompileDeferredNV = anv_furmark_CompileDeferredNV,
    .CreateAccelerationStructureNV = anv_furmark_CreateAccelerationStructureNV,
    .CmdBindInvocationMaskHUAWEI = anv_furmark_CmdBindInvocationMaskHUAWEI,
    .DestroyAccelerationStructureKHR = anv_furmark_DestroyAccelerationStructureKHR,
    .DestroyAccelerationStructureNV = anv_furmark_DestroyAccelerationStructureNV,
    .GetAccelerationStructureMemoryRequirementsNV = anv_furmark_GetAccelerationStructureMemoryRequirementsNV,
    .BindAccelerationStructureMemoryNV = anv_furmark_BindAccelerationStructureMemoryNV,
    .CmdCopyAccelerationStructureNV = anv_furmark_CmdCopyAccelerationStructureNV,
    .CmdCopyAccelerationStructureKHR = anv_furmark_CmdCopyAccelerationStructureKHR,
    .CopyAccelerationStructureKHR = anv_furmark_CopyAccelerationStructureKHR,
    .CmdCopyAccelerationStructureToMemoryKHR = anv_furmark_CmdCopyAccelerationStructureToMemoryKHR,
    .CopyAccelerationStructureToMemoryKHR = anv_furmark_CopyAccelerationStructureToMemoryKHR,
    .CmdCopyMemoryToAccelerationStructureKHR = anv_furmark_CmdCopyMemoryToAccelerationStructureKHR,
    .CopyMemoryToAccelerationStructureKHR = anv_furmark_CopyMemoryToAccelerationStructureKHR,
    .CmdWriteAccelerationStructuresPropertiesKHR = anv_furmark_CmdWriteAccelerationStructuresPropertiesKHR,
    .CmdWriteAccelerationStructuresPropertiesNV = anv_furmark_CmdWriteAccelerationStructuresPropertiesNV,
    .CmdBuildAccelerationStructureNV = anv_furmark_CmdBuildAccelerationStructureNV,
    .WriteAccelerationStructuresPropertiesKHR = anv_furmark_WriteAccelerationStructuresPropertiesKHR,
    .CmdTraceRaysKHR = anv_furmark_CmdTraceRaysKHR,
    .CmdTraceRaysNV = anv_furmark_CmdTraceRaysNV,
    .GetRayTracingShaderGroupHandlesKHR = anv_furmark_GetRayTracingShaderGroupHandlesKHR,
    .GetRayTracingShaderGroupHandlesNV = anv_furmark_GetRayTracingShaderGroupHandlesNV,
    .GetRayTracingCaptureReplayShaderGroupHandlesKHR = anv_furmark_GetRayTracingCaptureReplayShaderGroupHandlesKHR,
    .GetAccelerationStructureHandleNV = anv_furmark_GetAccelerationStructureHandleNV,
    .CreateRayTracingPipelinesNV = anv_furmark_CreateRayTracingPipelinesNV,
    .CreateRayTracingPipelinesKHR = anv_furmark_CreateRayTracingPipelinesKHR,
    .CmdTraceRaysIndirectKHR = anv_furmark_CmdTraceRaysIndirectKHR,
    .CmdTraceRaysIndirect2KHR = anv_furmark_CmdTraceRaysIndirect2KHR,
    .GetClusterAccelerationStructureBuildSizesNV = anv_furmark_GetClusterAccelerationStructureBuildSizesNV,
    .CmdBuildClusterAccelerationStructureIndirectNV = anv_furmark_CmdBuildClusterAccelerationStructureIndirectNV,
    .GetDeviceAccelerationStructureCompatibilityKHR = anv_furmark_GetDeviceAccelerationStructureCompatibilityKHR,
    .GetRayTracingShaderGroupStackSizeKHR = anv_furmark_GetRayTracingShaderGroupStackSizeKHR,
    .CmdSetRayTracingPipelineStackSizeKHR = anv_furmark_CmdSetRayTracingPipelineStackSizeKHR,
    .GetImageViewHandleNVX = anv_furmark_GetImageViewHandleNVX,
    .GetImageViewHandle64NVX = anv_furmark_GetImageViewHandle64NVX,
    .GetImageViewAddressNVX = anv_furmark_GetImageViewAddressNVX,
    .GetDeviceCombinedImageSamplerIndexNVX = anv_furmark_GetDeviceCombinedImageSamplerIndexNVX,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetDeviceGroupSurfacePresentModes2EXT = anv_furmark_GetDeviceGroupSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetDeviceGroupSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireFullScreenExclusiveModeEXT = anv_furmark_AcquireFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .AcquireFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ReleaseFullScreenExclusiveModeEXT = anv_furmark_ReleaseFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .ReleaseFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .AcquireProfilingLockKHR = anv_furmark_AcquireProfilingLockKHR,
    .ReleaseProfilingLockKHR = anv_furmark_ReleaseProfilingLockKHR,
    .GetImageDrmFormatModifierPropertiesEXT = anv_furmark_GetImageDrmFormatModifierPropertiesEXT,
    .GetBufferOpaqueCaptureAddress = anv_furmark_GetBufferOpaqueCaptureAddress,
    .GetBufferOpaqueCaptureAddressKHR = anv_furmark_GetBufferOpaqueCaptureAddressKHR,
    .GetBufferDeviceAddress = anv_furmark_GetBufferDeviceAddress,
    .GetBufferDeviceAddressKHR = anv_furmark_GetBufferDeviceAddressKHR,
    .GetBufferDeviceAddressEXT = anv_furmark_GetBufferDeviceAddressEXT,
    .InitializePerformanceApiINTEL = anv_furmark_InitializePerformanceApiINTEL,
    .UninitializePerformanceApiINTEL = anv_furmark_UninitializePerformanceApiINTEL,
    .CmdSetPerformanceMarkerINTEL = anv_furmark_CmdSetPerformanceMarkerINTEL,
    .CmdSetPerformanceStreamMarkerINTEL = anv_furmark_CmdSetPerformanceStreamMarkerINTEL,
    .CmdSetPerformanceOverrideINTEL = anv_furmark_CmdSetPerformanceOverrideINTEL,
    .AcquirePerformanceConfigurationINTEL = anv_furmark_AcquirePerformanceConfigurationINTEL,
    .ReleasePerformanceConfigurationINTEL = anv_furmark_ReleasePerformanceConfigurationINTEL,
    .QueueSetPerformanceConfigurationINTEL = anv_furmark_QueueSetPerformanceConfigurationINTEL,
    .GetPerformanceParameterINTEL = anv_furmark_GetPerformanceParameterINTEL,
    .GetDeviceMemoryOpaqueCaptureAddress = anv_furmark_GetDeviceMemoryOpaqueCaptureAddress,
    .GetDeviceMemoryOpaqueCaptureAddressKHR = anv_furmark_GetDeviceMemoryOpaqueCaptureAddressKHR,
    .GetPipelineExecutablePropertiesKHR = anv_furmark_GetPipelineExecutablePropertiesKHR,
    .GetPipelineExecutableStatisticsKHR = anv_furmark_GetPipelineExecutableStatisticsKHR,
    .GetPipelineExecutableInternalRepresentationsKHR = anv_furmark_GetPipelineExecutableInternalRepresentationsKHR,
    .CmdSetLineStipple = anv_furmark_CmdSetLineStipple,
    .CmdSetLineStippleKHR = anv_furmark_CmdSetLineStippleKHR,
    .CmdSetLineStippleEXT = anv_furmark_CmdSetLineStippleEXT,
    .CreateAccelerationStructureKHR = anv_furmark_CreateAccelerationStructureKHR,
    .CmdBuildAccelerationStructuresKHR = anv_furmark_CmdBuildAccelerationStructuresKHR,
    .CmdBuildAccelerationStructuresIndirectKHR = anv_furmark_CmdBuildAccelerationStructuresIndirectKHR,
    .BuildAccelerationStructuresKHR = anv_furmark_BuildAccelerationStructuresKHR,
    .GetAccelerationStructureDeviceAddressKHR = anv_furmark_GetAccelerationStructureDeviceAddressKHR,
    .CreateDeferredOperationKHR = anv_furmark_CreateDeferredOperationKHR,
    .DestroyDeferredOperationKHR = anv_furmark_DestroyDeferredOperationKHR,
    .GetDeferredOperationMaxConcurrencyKHR = anv_furmark_GetDeferredOperationMaxConcurrencyKHR,
    .GetDeferredOperationResultKHR = anv_furmark_GetDeferredOperationResultKHR,
    .DeferredOperationJoinKHR = anv_furmark_DeferredOperationJoinKHR,
    .GetPipelineIndirectMemoryRequirementsNV = anv_furmark_GetPipelineIndirectMemoryRequirementsNV,
    .GetPipelineIndirectDeviceAddressNV = anv_furmark_GetPipelineIndirectDeviceAddressNV,
    .AntiLagUpdateAMD = anv_furmark_AntiLagUpdateAMD,
    .CmdSetCullMode = anv_furmark_CmdSetCullMode,
    .CmdSetCullModeEXT = anv_furmark_CmdSetCullModeEXT,
    .CmdSetFrontFace = anv_furmark_CmdSetFrontFace,
    .CmdSetFrontFaceEXT = anv_furmark_CmdSetFrontFaceEXT,
    .CmdSetPrimitiveTopology = anv_furmark_CmdSetPrimitiveTopology,
    .CmdSetPrimitiveTopologyEXT = anv_furmark_CmdSetPrimitiveTopologyEXT,
    .CmdSetViewportWithCount = anv_furmark_CmdSetViewportWithCount,
    .CmdSetViewportWithCountEXT = anv_furmark_CmdSetViewportWithCountEXT,
    .CmdSetScissorWithCount = anv_furmark_CmdSetScissorWithCount,
    .CmdSetScissorWithCountEXT = anv_furmark_CmdSetScissorWithCountEXT,
    .CmdBindIndexBuffer2 = anv_furmark_CmdBindIndexBuffer2,
    .CmdBindIndexBuffer2KHR = anv_furmark_CmdBindIndexBuffer2KHR,
    .CmdBindVertexBuffers2 = anv_furmark_CmdBindVertexBuffers2,
    .CmdBindVertexBuffers2EXT = anv_furmark_CmdBindVertexBuffers2EXT,
    .CmdSetDepthTestEnable = anv_furmark_CmdSetDepthTestEnable,
    .CmdSetDepthTestEnableEXT = anv_furmark_CmdSetDepthTestEnableEXT,
    .CmdSetDepthWriteEnable = anv_furmark_CmdSetDepthWriteEnable,
    .CmdSetDepthWriteEnableEXT = anv_furmark_CmdSetDepthWriteEnableEXT,
    .CmdSetDepthCompareOp = anv_furmark_CmdSetDepthCompareOp,
    .CmdSetDepthCompareOpEXT = anv_furmark_CmdSetDepthCompareOpEXT,
    .CmdSetDepthBoundsTestEnable = anv_furmark_CmdSetDepthBoundsTestEnable,
    .CmdSetDepthBoundsTestEnableEXT = anv_furmark_CmdSetDepthBoundsTestEnableEXT,
    .CmdSetStencilTestEnable = anv_furmark_CmdSetStencilTestEnable,
    .CmdSetStencilTestEnableEXT = anv_furmark_CmdSetStencilTestEnableEXT,
    .CmdSetStencilOp = anv_furmark_CmdSetStencilOp,
    .CmdSetStencilOpEXT = anv_furmark_CmdSetStencilOpEXT,
    .CmdSetPatchControlPointsEXT = anv_furmark_CmdSetPatchControlPointsEXT,
    .CmdSetRasterizerDiscardEnable = anv_furmark_CmdSetRasterizerDiscardEnable,
    .CmdSetRasterizerDiscardEnableEXT = anv_furmark_CmdSetRasterizerDiscardEnableEXT,
    .CmdSetDepthBiasEnable = anv_furmark_CmdSetDepthBiasEnable,
    .CmdSetDepthBiasEnableEXT = anv_furmark_CmdSetDepthBiasEnableEXT,
    .CmdSetLogicOpEXT = anv_furmark_CmdSetLogicOpEXT,
    .CmdSetPrimitiveRestartEnable = anv_furmark_CmdSetPrimitiveRestartEnable,
    .CmdSetPrimitiveRestartEnableEXT = anv_furmark_CmdSetPrimitiveRestartEnableEXT,
    .CmdSetTessellationDomainOriginEXT = anv_furmark_CmdSetTessellationDomainOriginEXT,
    .CmdSetDepthClampEnableEXT = anv_furmark_CmdSetDepthClampEnableEXT,
    .CmdSetPolygonModeEXT = anv_furmark_CmdSetPolygonModeEXT,
    .CmdSetRasterizationSamplesEXT = anv_furmark_CmdSetRasterizationSamplesEXT,
    .CmdSetSampleMaskEXT = anv_furmark_CmdSetSampleMaskEXT,
    .CmdSetAlphaToCoverageEnableEXT = anv_furmark_CmdSetAlphaToCoverageEnableEXT,
    .CmdSetAlphaToOneEnableEXT = anv_furmark_CmdSetAlphaToOneEnableEXT,
    .CmdSetLogicOpEnableEXT = anv_furmark_CmdSetLogicOpEnableEXT,
    .CmdSetColorBlendEnableEXT = anv_furmark_CmdSetColorBlendEnableEXT,
    .CmdSetColorBlendEquationEXT = anv_furmark_CmdSetColorBlendEquationEXT,
    .CmdSetColorWriteMaskEXT = anv_furmark_CmdSetColorWriteMaskEXT,
    .CmdSetRasterizationStreamEXT = anv_furmark_CmdSetRasterizationStreamEXT,
    .CmdSetConservativeRasterizationModeEXT = anv_furmark_CmdSetConservativeRasterizationModeEXT,
    .CmdSetExtraPrimitiveOverestimationSizeEXT = anv_furmark_CmdSetExtraPrimitiveOverestimationSizeEXT,
    .CmdSetDepthClipEnableEXT = anv_furmark_CmdSetDepthClipEnableEXT,
    .CmdSetSampleLocationsEnableEXT = anv_furmark_CmdSetSampleLocationsEnableEXT,
    .CmdSetColorBlendAdvancedEXT = anv_furmark_CmdSetColorBlendAdvancedEXT,
    .CmdSetProvokingVertexModeEXT = anv_furmark_CmdSetProvokingVertexModeEXT,
    .CmdSetLineRasterizationModeEXT = anv_furmark_CmdSetLineRasterizationModeEXT,
    .CmdSetLineStippleEnableEXT = anv_furmark_CmdSetLineStippleEnableEXT,
    .CmdSetDepthClipNegativeOneToOneEXT = anv_furmark_CmdSetDepthClipNegativeOneToOneEXT,
    .CmdSetViewportWScalingEnableNV = anv_furmark_CmdSetViewportWScalingEnableNV,
    .CmdSetViewportSwizzleNV = anv_furmark_CmdSetViewportSwizzleNV,
    .CmdSetCoverageToColorEnableNV = anv_furmark_CmdSetCoverageToColorEnableNV,
    .CmdSetCoverageToColorLocationNV = anv_furmark_CmdSetCoverageToColorLocationNV,
    .CmdSetCoverageModulationModeNV = anv_furmark_CmdSetCoverageModulationModeNV,
    .CmdSetCoverageModulationTableEnableNV = anv_furmark_CmdSetCoverageModulationTableEnableNV,
    .CmdSetCoverageModulationTableNV = anv_furmark_CmdSetCoverageModulationTableNV,
    .CmdSetShadingRateImageEnableNV = anv_furmark_CmdSetShadingRateImageEnableNV,
    .CmdSetCoverageReductionModeNV = anv_furmark_CmdSetCoverageReductionModeNV,
    .CmdSetRepresentativeFragmentTestEnableNV = anv_furmark_CmdSetRepresentativeFragmentTestEnableNV,
    .CreatePrivateDataSlot = anv_furmark_CreatePrivateDataSlot,
    .CreatePrivateDataSlotEXT = anv_furmark_CreatePrivateDataSlotEXT,
    .DestroyPrivateDataSlot = anv_furmark_DestroyPrivateDataSlot,
    .DestroyPrivateDataSlotEXT = anv_furmark_DestroyPrivateDataSlotEXT,
    .SetPrivateData = anv_furmark_SetPrivateData,
    .SetPrivateDataEXT = anv_furmark_SetPrivateDataEXT,
    .GetPrivateData = anv_furmark_GetPrivateData,
    .GetPrivateDataEXT = anv_furmark_GetPrivateDataEXT,
    .CmdCopyBuffer2 = anv_furmark_CmdCopyBuffer2,
    .CmdCopyBuffer2KHR = anv_furmark_CmdCopyBuffer2KHR,
    .CmdCopyImage2 = anv_furmark_CmdCopyImage2,
    .CmdCopyImage2KHR = anv_furmark_CmdCopyImage2KHR,
    .CmdBlitImage2 = anv_furmark_CmdBlitImage2,
    .CmdBlitImage2KHR = anv_furmark_CmdBlitImage2KHR,
    .CmdCopyBufferToImage2 = anv_furmark_CmdCopyBufferToImage2,
    .CmdCopyBufferToImage2KHR = anv_furmark_CmdCopyBufferToImage2KHR,
    .CmdCopyImageToBuffer2 = anv_furmark_CmdCopyImageToBuffer2,
    .CmdCopyImageToBuffer2KHR = anv_furmark_CmdCopyImageToBuffer2KHR,
    .CmdResolveImage2 = anv_furmark_CmdResolveImage2,
    .CmdResolveImage2KHR = anv_furmark_CmdResolveImage2KHR,
    .CmdSetFragmentShadingRateKHR = anv_furmark_CmdSetFragmentShadingRateKHR,
    .CmdSetFragmentShadingRateEnumNV = anv_furmark_CmdSetFragmentShadingRateEnumNV,
    .GetAccelerationStructureBuildSizesKHR = anv_furmark_GetAccelerationStructureBuildSizesKHR,
    .CmdSetVertexInputEXT = anv_furmark_CmdSetVertexInputEXT,
    .CmdSetColorWriteEnableEXT = anv_furmark_CmdSetColorWriteEnableEXT,
    .CmdSetEvent2 = anv_furmark_CmdSetEvent2,
    .CmdSetEvent2KHR = anv_furmark_CmdSetEvent2KHR,
    .CmdResetEvent2 = anv_furmark_CmdResetEvent2,
    .CmdResetEvent2KHR = anv_furmark_CmdResetEvent2KHR,
    .CmdWaitEvents2 = anv_furmark_CmdWaitEvents2,
    .CmdWaitEvents2KHR = anv_furmark_CmdWaitEvents2KHR,
    .CmdPipelineBarrier2 = anv_furmark_CmdPipelineBarrier2,
    .CmdPipelineBarrier2KHR = anv_furmark_CmdPipelineBarrier2KHR,
    .QueueSubmit2 = anv_furmark_QueueSubmit2,
    .QueueSubmit2KHR = anv_furmark_QueueSubmit2KHR,
    .CmdWriteTimestamp2 = anv_furmark_CmdWriteTimestamp2,
    .CmdWriteTimestamp2KHR = anv_furmark_CmdWriteTimestamp2KHR,
    .CmdWriteBufferMarker2AMD = anv_furmark_CmdWriteBufferMarker2AMD,
    .GetQueueCheckpointData2NV = anv_furmark_GetQueueCheckpointData2NV,
    .CopyMemoryToImage = anv_furmark_CopyMemoryToImage,
    .CopyMemoryToImageEXT = anv_furmark_CopyMemoryToImageEXT,
    .CopyImageToMemory = anv_furmark_CopyImageToMemory,
    .CopyImageToMemoryEXT = anv_furmark_CopyImageToMemoryEXT,
    .CopyImageToImage = anv_furmark_CopyImageToImage,
    .CopyImageToImageEXT = anv_furmark_CopyImageToImageEXT,
    .TransitionImageLayout = anv_furmark_TransitionImageLayout,
    .TransitionImageLayoutEXT = anv_furmark_TransitionImageLayoutEXT,
    .CreateVideoSessionKHR = anv_furmark_CreateVideoSessionKHR,
    .DestroyVideoSessionKHR = anv_furmark_DestroyVideoSessionKHR,
    .CreateVideoSessionParametersKHR = anv_furmark_CreateVideoSessionParametersKHR,
    .UpdateVideoSessionParametersKHR = anv_furmark_UpdateVideoSessionParametersKHR,
    .GetEncodedVideoSessionParametersKHR = anv_furmark_GetEncodedVideoSessionParametersKHR,
    .DestroyVideoSessionParametersKHR = anv_furmark_DestroyVideoSessionParametersKHR,
    .GetVideoSessionMemoryRequirementsKHR = anv_furmark_GetVideoSessionMemoryRequirementsKHR,
    .BindVideoSessionMemoryKHR = anv_furmark_BindVideoSessionMemoryKHR,
    .CmdDecodeVideoKHR = anv_furmark_CmdDecodeVideoKHR,
    .CmdBeginVideoCodingKHR = anv_furmark_CmdBeginVideoCodingKHR,
    .CmdControlVideoCodingKHR = anv_furmark_CmdControlVideoCodingKHR,
    .CmdEndVideoCodingKHR = anv_furmark_CmdEndVideoCodingKHR,
    .CmdEncodeVideoKHR = anv_furmark_CmdEncodeVideoKHR,
    .CmdDecompressMemoryNV = anv_furmark_CmdDecompressMemoryNV,
    .CmdDecompressMemoryIndirectCountNV = anv_furmark_CmdDecompressMemoryIndirectCountNV,
    .GetPartitionedAccelerationStructuresBuildSizesNV = anv_furmark_GetPartitionedAccelerationStructuresBuildSizesNV,
    .CmdBuildPartitionedAccelerationStructuresNV = anv_furmark_CmdBuildPartitionedAccelerationStructuresNV,
    .CmdDecompressMemoryEXT = anv_furmark_CmdDecompressMemoryEXT,
    .CmdDecompressMemoryIndirectCountEXT = anv_furmark_CmdDecompressMemoryIndirectCountEXT,
    .CreateCuModuleNVX = anv_furmark_CreateCuModuleNVX,
    .CreateCuFunctionNVX = anv_furmark_CreateCuFunctionNVX,
    .DestroyCuModuleNVX = anv_furmark_DestroyCuModuleNVX,
    .DestroyCuFunctionNVX = anv_furmark_DestroyCuFunctionNVX,
    .CmdCuLaunchKernelNVX = anv_furmark_CmdCuLaunchKernelNVX,
    .GetDescriptorSetLayoutSizeEXT = anv_furmark_GetDescriptorSetLayoutSizeEXT,
    .GetDescriptorSetLayoutBindingOffsetEXT = anv_furmark_GetDescriptorSetLayoutBindingOffsetEXT,
    .GetDescriptorEXT = anv_furmark_GetDescriptorEXT,
    .CmdBindDescriptorBuffersEXT = anv_furmark_CmdBindDescriptorBuffersEXT,
    .CmdSetDescriptorBufferOffsetsEXT = anv_furmark_CmdSetDescriptorBufferOffsetsEXT,
    .CmdBindDescriptorBufferEmbeddedSamplersEXT = anv_furmark_CmdBindDescriptorBufferEmbeddedSamplersEXT,
    .GetBufferOpaqueCaptureDescriptorDataEXT = anv_furmark_GetBufferOpaqueCaptureDescriptorDataEXT,
    .GetImageOpaqueCaptureDescriptorDataEXT = anv_furmark_GetImageOpaqueCaptureDescriptorDataEXT,
    .GetImageViewOpaqueCaptureDescriptorDataEXT = anv_furmark_GetImageViewOpaqueCaptureDescriptorDataEXT,
    .GetSamplerOpaqueCaptureDescriptorDataEXT = anv_furmark_GetSamplerOpaqueCaptureDescriptorDataEXT,
    .GetAccelerationStructureOpaqueCaptureDescriptorDataEXT = anv_furmark_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT,
    .SetDeviceMemoryPriorityEXT = anv_furmark_SetDeviceMemoryPriorityEXT,
    .WaitForPresent2KHR = anv_furmark_WaitForPresent2KHR,
    .WaitForPresentKHR = anv_furmark_WaitForPresentKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateBufferCollectionFUCHSIA = anv_furmark_CreateBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .CreateBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionBufferConstraintsFUCHSIA = anv_furmark_SetBufferCollectionBufferConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionBufferConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionImageConstraintsFUCHSIA = anv_furmark_SetBufferCollectionImageConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionImageConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .DestroyBufferCollectionFUCHSIA = anv_furmark_DestroyBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .DestroyBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetBufferCollectionPropertiesFUCHSIA = anv_furmark_GetBufferCollectionPropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetBufferCollectionPropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .CmdBeginRendering = anv_furmark_CmdBeginRendering,
    .CmdBeginRenderingKHR = anv_furmark_CmdBeginRenderingKHR,
    .CmdEndRendering = anv_furmark_CmdEndRendering,
    .CmdEndRendering2KHR = anv_furmark_CmdEndRendering2KHR,
    .CmdEndRendering2EXT = anv_furmark_CmdEndRendering2EXT,
    .CmdEndRenderingKHR = anv_furmark_CmdEndRenderingKHR,
    .GetDescriptorSetLayoutHostMappingInfoVALVE = anv_furmark_GetDescriptorSetLayoutHostMappingInfoVALVE,
    .GetDescriptorSetHostMappingVALVE = anv_furmark_GetDescriptorSetHostMappingVALVE,
    .CreateMicromapEXT = anv_furmark_CreateMicromapEXT,
    .CmdBuildMicromapsEXT = anv_furmark_CmdBuildMicromapsEXT,
    .BuildMicromapsEXT = anv_furmark_BuildMicromapsEXT,
    .DestroyMicromapEXT = anv_furmark_DestroyMicromapEXT,
    .CmdCopyMicromapEXT = anv_furmark_CmdCopyMicromapEXT,
    .CopyMicromapEXT = anv_furmark_CopyMicromapEXT,
    .CmdCopyMicromapToMemoryEXT = anv_furmark_CmdCopyMicromapToMemoryEXT,
    .CopyMicromapToMemoryEXT = anv_furmark_CopyMicromapToMemoryEXT,
    .CmdCopyMemoryToMicromapEXT = anv_furmark_CmdCopyMemoryToMicromapEXT,
    .CopyMemoryToMicromapEXT = anv_furmark_CopyMemoryToMicromapEXT,
    .CmdWriteMicromapsPropertiesEXT = anv_furmark_CmdWriteMicromapsPropertiesEXT,
    .WriteMicromapsPropertiesEXT = anv_furmark_WriteMicromapsPropertiesEXT,
    .GetDeviceMicromapCompatibilityEXT = anv_furmark_GetDeviceMicromapCompatibilityEXT,
    .GetMicromapBuildSizesEXT = anv_furmark_GetMicromapBuildSizesEXT,
    .GetShaderModuleIdentifierEXT = anv_furmark_GetShaderModuleIdentifierEXT,
    .GetShaderModuleCreateInfoIdentifierEXT = anv_furmark_GetShaderModuleCreateInfoIdentifierEXT,
    .GetImageSubresourceLayout2 = anv_furmark_GetImageSubresourceLayout2,
    .GetImageSubresourceLayout2KHR = anv_furmark_GetImageSubresourceLayout2KHR,
    .GetImageSubresourceLayout2EXT = anv_furmark_GetImageSubresourceLayout2EXT,
    .GetPipelinePropertiesEXT = anv_furmark_GetPipelinePropertiesEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .ExportMetalObjectsEXT = anv_furmark_ExportMetalObjectsEXT,
#elif defined(_MSC_VER)
    .ExportMetalObjectsEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CmdBindTileMemoryQCOM = anv_furmark_CmdBindTileMemoryQCOM,
    .GetFramebufferTilePropertiesQCOM = anv_furmark_GetFramebufferTilePropertiesQCOM,
    .GetDynamicRenderingTilePropertiesQCOM = anv_furmark_GetDynamicRenderingTilePropertiesQCOM,
    .CreateOpticalFlowSessionNV = anv_furmark_CreateOpticalFlowSessionNV,
    .DestroyOpticalFlowSessionNV = anv_furmark_DestroyOpticalFlowSessionNV,
    .BindOpticalFlowSessionImageNV = anv_furmark_BindOpticalFlowSessionImageNV,
    .CmdOpticalFlowExecuteNV = anv_furmark_CmdOpticalFlowExecuteNV,
    .GetDeviceFaultInfoEXT = anv_furmark_GetDeviceFaultInfoEXT,
    .GetDeviceFaultReportsKHR = anv_furmark_GetDeviceFaultReportsKHR,
    .GetDeviceFaultDebugInfoKHR = anv_furmark_GetDeviceFaultDebugInfoKHR,
    .CmdSetDepthBias2EXT = anv_furmark_CmdSetDepthBias2EXT,
    .ReleaseSwapchainImagesKHR = anv_furmark_ReleaseSwapchainImagesKHR,
    .ReleaseSwapchainImagesEXT = anv_furmark_ReleaseSwapchainImagesEXT,
    .GetDeviceImageSubresourceLayout = anv_furmark_GetDeviceImageSubresourceLayout,
    .GetDeviceImageSubresourceLayoutKHR = anv_furmark_GetDeviceImageSubresourceLayoutKHR,
    .MapMemory2 = anv_furmark_MapMemory2,
    .MapMemory2KHR = anv_furmark_MapMemory2KHR,
    .UnmapMemory2 = anv_furmark_UnmapMemory2,
    .UnmapMemory2KHR = anv_furmark_UnmapMemory2KHR,
    .CreateShadersEXT = anv_furmark_CreateShadersEXT,
    .DestroyShaderEXT = anv_furmark_DestroyShaderEXT,
    .GetShaderBinaryDataEXT = anv_furmark_GetShaderBinaryDataEXT,
    .CmdBindShadersEXT = anv_furmark_CmdBindShadersEXT,
    .SetSwapchainPresentTimingQueueSizeEXT = anv_furmark_SetSwapchainPresentTimingQueueSizeEXT,
    .GetSwapchainTimingPropertiesEXT = anv_furmark_GetSwapchainTimingPropertiesEXT,
    .GetSwapchainTimeDomainPropertiesEXT = anv_furmark_GetSwapchainTimeDomainPropertiesEXT,
    .GetPastPresentationTimingEXT = anv_furmark_GetPastPresentationTimingEXT,
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetScreenBufferPropertiesQNX = anv_furmark_GetScreenBufferPropertiesQNX,
#elif defined(_MSC_VER)
    .GetScreenBufferPropertiesQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CmdBindDescriptorSets2 = anv_furmark_CmdBindDescriptorSets2,
    .CmdBindDescriptorSets2KHR = anv_furmark_CmdBindDescriptorSets2KHR,
    .CmdPushConstants2 = anv_furmark_CmdPushConstants2,
    .CmdPushConstants2KHR = anv_furmark_CmdPushConstants2KHR,
    .CmdPushDescriptorSet2 = anv_furmark_CmdPushDescriptorSet2,
    .CmdPushDescriptorSet2KHR = anv_furmark_CmdPushDescriptorSet2KHR,
    .CmdPushDescriptorSetWithTemplate2 = anv_furmark_CmdPushDescriptorSetWithTemplate2,
    .CmdPushDescriptorSetWithTemplate2KHR = anv_furmark_CmdPushDescriptorSetWithTemplate2KHR,
    .CmdSetDescriptorBufferOffsets2EXT = anv_furmark_CmdSetDescriptorBufferOffsets2EXT,
    .CmdBindDescriptorBufferEmbeddedSamplers2EXT = anv_furmark_CmdBindDescriptorBufferEmbeddedSamplers2EXT,
    .SetLatencySleepModeNV = anv_furmark_SetLatencySleepModeNV,
    .LatencySleepNV = anv_furmark_LatencySleepNV,
    .SetLatencyMarkerNV = anv_furmark_SetLatencyMarkerNV,
    .GetLatencyTimingsNV = anv_furmark_GetLatencyTimingsNV,
    .QueueNotifyOutOfBandNV = anv_furmark_QueueNotifyOutOfBandNV,
    .CmdSetRenderingAttachmentLocations = anv_furmark_CmdSetRenderingAttachmentLocations,
    .CmdSetRenderingAttachmentLocationsKHR = anv_furmark_CmdSetRenderingAttachmentLocationsKHR,
    .CmdSetRenderingInputAttachmentIndices = anv_furmark_CmdSetRenderingInputAttachmentIndices,
    .CmdSetRenderingInputAttachmentIndicesKHR = anv_furmark_CmdSetRenderingInputAttachmentIndicesKHR,
    .CmdSetDepthClampRangeEXT = anv_furmark_CmdSetDepthClampRangeEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandleEXT = anv_furmark_GetMemoryMetalHandleEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandleEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandlePropertiesEXT = anv_furmark_GetMemoryMetalHandlePropertiesEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandlePropertiesEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .ConvertCooperativeVectorMatrixNV = anv_furmark_ConvertCooperativeVectorMatrixNV,
    .CmdConvertCooperativeVectorMatrixNV = anv_furmark_CmdConvertCooperativeVectorMatrixNV,
    .CmdDispatchTileQCOM = anv_furmark_CmdDispatchTileQCOM,
    .CmdBeginPerTileExecutionQCOM = anv_furmark_CmdBeginPerTileExecutionQCOM,
    .CmdEndPerTileExecutionQCOM = anv_furmark_CmdEndPerTileExecutionQCOM,
    .CreateExternalComputeQueueNV = anv_furmark_CreateExternalComputeQueueNV,
    .DestroyExternalComputeQueueNV = anv_furmark_DestroyExternalComputeQueueNV,
    .CreateShaderInstrumentationARM = anv_furmark_CreateShaderInstrumentationARM,
    .DestroyShaderInstrumentationARM = anv_furmark_DestroyShaderInstrumentationARM,
    .CmdBeginShaderInstrumentationARM = anv_furmark_CmdBeginShaderInstrumentationARM,
    .CmdEndShaderInstrumentationARM = anv_furmark_CmdEndShaderInstrumentationARM,
    .GetShaderInstrumentationValuesARM = anv_furmark_GetShaderInstrumentationValuesARM,
    .ClearShaderInstrumentationMetricsARM = anv_furmark_ClearShaderInstrumentationMetricsARM,
    .CreateTensorARM = anv_furmark_CreateTensorARM,
    .DestroyTensorARM = anv_furmark_DestroyTensorARM,
    .CreateTensorViewARM = anv_furmark_CreateTensorViewARM,
    .DestroyTensorViewARM = anv_furmark_DestroyTensorViewARM,
    .GetTensorMemoryRequirementsARM = anv_furmark_GetTensorMemoryRequirementsARM,
    .BindTensorMemoryARM = anv_furmark_BindTensorMemoryARM,
    .GetDeviceTensorMemoryRequirementsARM = anv_furmark_GetDeviceTensorMemoryRequirementsARM,
    .CmdCopyTensorARM = anv_furmark_CmdCopyTensorARM,
    .GetTensorOpaqueCaptureDescriptorDataARM = anv_furmark_GetTensorOpaqueCaptureDescriptorDataARM,
    .GetTensorViewOpaqueCaptureDescriptorDataARM = anv_furmark_GetTensorViewOpaqueCaptureDescriptorDataARM,
    .CreateDataGraphPipelinesARM = anv_furmark_CreateDataGraphPipelinesARM,
    .CreateDataGraphPipelineSessionARM = anv_furmark_CreateDataGraphPipelineSessionARM,
    .GetDataGraphPipelineSessionBindPointRequirementsARM = anv_furmark_GetDataGraphPipelineSessionBindPointRequirementsARM,
    .GetDataGraphPipelineSessionMemoryRequirementsARM = anv_furmark_GetDataGraphPipelineSessionMemoryRequirementsARM,
    .BindDataGraphPipelineSessionMemoryARM = anv_furmark_BindDataGraphPipelineSessionMemoryARM,
    .DestroyDataGraphPipelineSessionARM = anv_furmark_DestroyDataGraphPipelineSessionARM,
    .CmdDispatchDataGraphARM = anv_furmark_CmdDispatchDataGraphARM,
    .GetDataGraphPipelineAvailablePropertiesARM = anv_furmark_GetDataGraphPipelineAvailablePropertiesARM,
    .GetDataGraphPipelinePropertiesARM = anv_furmark_GetDataGraphPipelinePropertiesARM,
#ifdef VK_USE_PLATFORM_OHOS
    .GetNativeBufferPropertiesOHOS = anv_furmark_GetNativeBufferPropertiesOHOS,
#elif defined(_MSC_VER)
    .GetNativeBufferPropertiesOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
    .GetMemoryNativeBufferOHOS = anv_furmark_GetMemoryNativeBufferOHOS,
#elif defined(_MSC_VER)
    .GetMemoryNativeBufferOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .QueueSetPerfHintQCOM = anv_furmark_QueueSetPerfHintQCOM,
    .CmdSetComputeOccupancyPriorityNV = anv_furmark_CmdSetComputeOccupancyPriorityNV,
    .WriteSamplerDescriptorsEXT = anv_furmark_WriteSamplerDescriptorsEXT,
    .WriteResourceDescriptorsEXT = anv_furmark_WriteResourceDescriptorsEXT,
    .CmdBindSamplerHeapEXT = anv_furmark_CmdBindSamplerHeapEXT,
    .CmdBindResourceHeapEXT = anv_furmark_CmdBindResourceHeapEXT,
    .CmdPushDataEXT = anv_furmark_CmdPushDataEXT,
    .RegisterCustomBorderColorEXT = anv_furmark_RegisterCustomBorderColorEXT,
    .UnregisterCustomBorderColorEXT = anv_furmark_UnregisterCustomBorderColorEXT,
    .GetImageOpaqueCaptureDataEXT = anv_furmark_GetImageOpaqueCaptureDataEXT,
    .GetTensorOpaqueCaptureDataARM = anv_furmark_GetTensorOpaqueCaptureDataARM,
    .CmdCopyMemoryKHR = anv_furmark_CmdCopyMemoryKHR,
    .CmdCopyMemoryToImageKHR = anv_furmark_CmdCopyMemoryToImageKHR,
    .CmdCopyImageToMemoryKHR = anv_furmark_CmdCopyImageToMemoryKHR,
    .CmdUpdateMemoryKHR = anv_furmark_CmdUpdateMemoryKHR,
    .CmdFillMemoryKHR = anv_furmark_CmdFillMemoryKHR,
    .CmdCopyQueryPoolResultsToMemoryKHR = anv_furmark_CmdCopyQueryPoolResultsToMemoryKHR,
    .CmdBeginConditionalRendering2EXT = anv_furmark_CmdBeginConditionalRendering2EXT,
    .CmdBindTransformFeedbackBuffers2EXT = anv_furmark_CmdBindTransformFeedbackBuffers2EXT,
    .CmdBeginTransformFeedback2EXT = anv_furmark_CmdBeginTransformFeedback2EXT,
    .CmdEndTransformFeedback2EXT = anv_furmark_CmdEndTransformFeedback2EXT,
    .CmdDrawIndirectByteCount2EXT = anv_furmark_CmdDrawIndirectByteCount2EXT,
    .CmdWriteMarkerToMemoryAMD = anv_furmark_CmdWriteMarkerToMemoryAMD,
    .CmdBindIndexBuffer3KHR = anv_furmark_CmdBindIndexBuffer3KHR,
    .CmdBindVertexBuffers3KHR = anv_furmark_CmdBindVertexBuffers3KHR,
    .CmdDrawIndirect2KHR = anv_furmark_CmdDrawIndirect2KHR,
    .CmdDrawIndexedIndirect2KHR = anv_furmark_CmdDrawIndexedIndirect2KHR,
    .CmdDrawIndirectCount2KHR = anv_furmark_CmdDrawIndirectCount2KHR,
    .CmdDrawIndexedIndirectCount2KHR = anv_furmark_CmdDrawIndexedIndirectCount2KHR,
    .CmdDrawMeshTasksIndirect2EXT = anv_furmark_CmdDrawMeshTasksIndirect2EXT,
    .CmdDrawMeshTasksIndirectCount2EXT = anv_furmark_CmdDrawMeshTasksIndirectCount2EXT,
    .CmdDispatchIndirect2KHR = anv_furmark_CmdDispatchIndirect2KHR,
    .CreateAccelerationStructure2KHR = anv_furmark_CreateAccelerationStructure2KHR,
    .CmdSetDispatchParametersARM = anv_furmark_CmdSetDispatchParametersARM,
};
const struct vk_device_entrypoint_table anv_hitman3_device_entrypoints = {
    .GetDeviceProcAddr = anv_hitman3_GetDeviceProcAddr,
    .DestroyDevice = anv_hitman3_DestroyDevice,
    .GetDeviceQueue = anv_hitman3_GetDeviceQueue,
    .QueueSubmit = anv_hitman3_QueueSubmit,
    .QueueWaitIdle = anv_hitman3_QueueWaitIdle,
    .DeviceWaitIdle = anv_hitman3_DeviceWaitIdle,
    .AllocateMemory = anv_hitman3_AllocateMemory,
    .FreeMemory = anv_hitman3_FreeMemory,
    .MapMemory = anv_hitman3_MapMemory,
    .UnmapMemory = anv_hitman3_UnmapMemory,
    .FlushMappedMemoryRanges = anv_hitman3_FlushMappedMemoryRanges,
    .InvalidateMappedMemoryRanges = anv_hitman3_InvalidateMappedMemoryRanges,
    .GetDeviceMemoryCommitment = anv_hitman3_GetDeviceMemoryCommitment,
    .GetBufferMemoryRequirements = anv_hitman3_GetBufferMemoryRequirements,
    .BindBufferMemory = anv_hitman3_BindBufferMemory,
    .GetImageMemoryRequirements = anv_hitman3_GetImageMemoryRequirements,
    .BindImageMemory = anv_hitman3_BindImageMemory,
    .GetImageSparseMemoryRequirements = anv_hitman3_GetImageSparseMemoryRequirements,
    .QueueBindSparse = anv_hitman3_QueueBindSparse,
    .CreateFence = anv_hitman3_CreateFence,
    .DestroyFence = anv_hitman3_DestroyFence,
    .ResetFences = anv_hitman3_ResetFences,
    .GetFenceStatus = anv_hitman3_GetFenceStatus,
    .WaitForFences = anv_hitman3_WaitForFences,
    .CreateSemaphore = anv_hitman3_CreateSemaphore,
    .DestroySemaphore = anv_hitman3_DestroySemaphore,
    .CreateEvent = anv_hitman3_CreateEvent,
    .DestroyEvent = anv_hitman3_DestroyEvent,
    .GetEventStatus = anv_hitman3_GetEventStatus,
    .SetEvent = anv_hitman3_SetEvent,
    .ResetEvent = anv_hitman3_ResetEvent,
    .CreateQueryPool = anv_hitman3_CreateQueryPool,
    .DestroyQueryPool = anv_hitman3_DestroyQueryPool,
    .GetQueryPoolResults = anv_hitman3_GetQueryPoolResults,
    .ResetQueryPool = anv_hitman3_ResetQueryPool,
    .ResetQueryPoolEXT = anv_hitman3_ResetQueryPoolEXT,
    .CreateBuffer = anv_hitman3_CreateBuffer,
    .DestroyBuffer = anv_hitman3_DestroyBuffer,
    .CreateBufferView = anv_hitman3_CreateBufferView,
    .DestroyBufferView = anv_hitman3_DestroyBufferView,
    .CreateImage = anv_hitman3_CreateImage,
    .DestroyImage = anv_hitman3_DestroyImage,
    .GetImageSubresourceLayout = anv_hitman3_GetImageSubresourceLayout,
    .CreateImageView = anv_hitman3_CreateImageView,
    .DestroyImageView = anv_hitman3_DestroyImageView,
    .CreateShaderModule = anv_hitman3_CreateShaderModule,
    .DestroyShaderModule = anv_hitman3_DestroyShaderModule,
    .CreatePipelineCache = anv_hitman3_CreatePipelineCache,
    .DestroyPipelineCache = anv_hitman3_DestroyPipelineCache,
    .GetPipelineCacheData = anv_hitman3_GetPipelineCacheData,
    .MergePipelineCaches = anv_hitman3_MergePipelineCaches,
    .CreatePipelineBinariesKHR = anv_hitman3_CreatePipelineBinariesKHR,
    .DestroyPipelineBinaryKHR = anv_hitman3_DestroyPipelineBinaryKHR,
    .GetPipelineKeyKHR = anv_hitman3_GetPipelineKeyKHR,
    .GetPipelineBinaryDataKHR = anv_hitman3_GetPipelineBinaryDataKHR,
    .ReleaseCapturedPipelineDataKHR = anv_hitman3_ReleaseCapturedPipelineDataKHR,
    .CreateGraphicsPipelines = anv_hitman3_CreateGraphicsPipelines,
    .CreateComputePipelines = anv_hitman3_CreateComputePipelines,
    .GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI = anv_hitman3_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI,
    .DestroyPipeline = anv_hitman3_DestroyPipeline,
    .CreatePipelineLayout = anv_hitman3_CreatePipelineLayout,
    .DestroyPipelineLayout = anv_hitman3_DestroyPipelineLayout,
    .CreateSampler = anv_hitman3_CreateSampler,
    .DestroySampler = anv_hitman3_DestroySampler,
    .CreateDescriptorSetLayout = anv_hitman3_CreateDescriptorSetLayout,
    .DestroyDescriptorSetLayout = anv_hitman3_DestroyDescriptorSetLayout,
    .CreateDescriptorPool = anv_hitman3_CreateDescriptorPool,
    .DestroyDescriptorPool = anv_hitman3_DestroyDescriptorPool,
    .ResetDescriptorPool = anv_hitman3_ResetDescriptorPool,
    .AllocateDescriptorSets = anv_hitman3_AllocateDescriptorSets,
    .FreeDescriptorSets = anv_hitman3_FreeDescriptorSets,
    .UpdateDescriptorSets = anv_hitman3_UpdateDescriptorSets,
    .CreateFramebuffer = anv_hitman3_CreateFramebuffer,
    .DestroyFramebuffer = anv_hitman3_DestroyFramebuffer,
    .CreateRenderPass = anv_hitman3_CreateRenderPass,
    .DestroyRenderPass = anv_hitman3_DestroyRenderPass,
    .GetRenderAreaGranularity = anv_hitman3_GetRenderAreaGranularity,
    .GetRenderingAreaGranularity = anv_hitman3_GetRenderingAreaGranularity,
    .GetRenderingAreaGranularityKHR = anv_hitman3_GetRenderingAreaGranularityKHR,
    .CreateCommandPool = anv_hitman3_CreateCommandPool,
    .DestroyCommandPool = anv_hitman3_DestroyCommandPool,
    .ResetCommandPool = anv_hitman3_ResetCommandPool,
    .AllocateCommandBuffers = anv_hitman3_AllocateCommandBuffers,
    .FreeCommandBuffers = anv_hitman3_FreeCommandBuffers,
    .BeginCommandBuffer = anv_hitman3_BeginCommandBuffer,
    .EndCommandBuffer = anv_hitman3_EndCommandBuffer,
    .ResetCommandBuffer = anv_hitman3_ResetCommandBuffer,
    .CmdBindPipeline = anv_hitman3_CmdBindPipeline,
    .CmdSetPrimitiveRestartIndexEXT = anv_hitman3_CmdSetPrimitiveRestartIndexEXT,
    .CmdSetAttachmentFeedbackLoopEnableEXT = anv_hitman3_CmdSetAttachmentFeedbackLoopEnableEXT,
    .CmdSetViewport = anv_hitman3_CmdSetViewport,
    .CmdSetScissor = anv_hitman3_CmdSetScissor,
    .CmdSetLineWidth = anv_hitman3_CmdSetLineWidth,
    .CmdSetDepthBias = anv_hitman3_CmdSetDepthBias,
    .CmdSetBlendConstants = anv_hitman3_CmdSetBlendConstants,
    .CmdSetDepthBounds = anv_hitman3_CmdSetDepthBounds,
    .CmdSetStencilCompareMask = anv_hitman3_CmdSetStencilCompareMask,
    .CmdSetStencilWriteMask = anv_hitman3_CmdSetStencilWriteMask,
    .CmdSetStencilReference = anv_hitman3_CmdSetStencilReference,
    .CmdBindDescriptorSets = anv_hitman3_CmdBindDescriptorSets,
    .CmdBindIndexBuffer = anv_hitman3_CmdBindIndexBuffer,
    .CmdBindVertexBuffers = anv_hitman3_CmdBindVertexBuffers,
    .CmdDraw = anv_hitman3_CmdDraw,
    .CmdDrawIndexed = anv_hitman3_CmdDrawIndexed,
    .CmdDrawMultiEXT = anv_hitman3_CmdDrawMultiEXT,
    .CmdDrawMultiIndexedEXT = anv_hitman3_CmdDrawMultiIndexedEXT,
    .CmdDrawIndirect = anv_hitman3_CmdDrawIndirect,
    .CmdDrawIndexedIndirect = anv_hitman3_CmdDrawIndexedIndirect,
    .CmdDispatch = anv_hitman3_CmdDispatch,
    .CmdDispatchIndirect = anv_hitman3_CmdDispatchIndirect,
    .CmdSubpassShadingHUAWEI = anv_hitman3_CmdSubpassShadingHUAWEI,
    .CmdDrawClusterHUAWEI = anv_hitman3_CmdDrawClusterHUAWEI,
    .CmdDrawClusterIndirectHUAWEI = anv_hitman3_CmdDrawClusterIndirectHUAWEI,
    .CmdUpdatePipelineIndirectBufferNV = anv_hitman3_CmdUpdatePipelineIndirectBufferNV,
    .CmdCopyBuffer = anv_hitman3_CmdCopyBuffer,
    .CmdCopyImage = anv_hitman3_CmdCopyImage,
    .CmdBlitImage = anv_hitman3_CmdBlitImage,
    .CmdCopyBufferToImage = anv_hitman3_CmdCopyBufferToImage,
    .CmdCopyImageToBuffer = anv_hitman3_CmdCopyImageToBuffer,
    .CmdCopyMemoryIndirectNV = anv_hitman3_CmdCopyMemoryIndirectNV,
    .CmdCopyMemoryIndirectKHR = anv_hitman3_CmdCopyMemoryIndirectKHR,
    .CmdCopyMemoryToImageIndirectNV = anv_hitman3_CmdCopyMemoryToImageIndirectNV,
    .CmdCopyMemoryToImageIndirectKHR = anv_hitman3_CmdCopyMemoryToImageIndirectKHR,
    .CmdUpdateBuffer = anv_hitman3_CmdUpdateBuffer,
    .CmdFillBuffer = anv_hitman3_CmdFillBuffer,
    .CmdClearColorImage = anv_hitman3_CmdClearColorImage,
    .CmdClearDepthStencilImage = anv_hitman3_CmdClearDepthStencilImage,
    .CmdClearAttachments = anv_hitman3_CmdClearAttachments,
    .CmdResolveImage = anv_hitman3_CmdResolveImage,
    .CmdSetEvent = anv_hitman3_CmdSetEvent,
    .CmdResetEvent = anv_hitman3_CmdResetEvent,
    .CmdWaitEvents = anv_hitman3_CmdWaitEvents,
    .CmdPipelineBarrier = anv_hitman3_CmdPipelineBarrier,
    .CmdBeginQuery = anv_hitman3_CmdBeginQuery,
    .CmdEndQuery = anv_hitman3_CmdEndQuery,
    .CmdBeginConditionalRenderingEXT = anv_hitman3_CmdBeginConditionalRenderingEXT,
    .CmdEndConditionalRenderingEXT = anv_hitman3_CmdEndConditionalRenderingEXT,
    .CmdBeginCustomResolveEXT = anv_hitman3_CmdBeginCustomResolveEXT,
    .CmdResetQueryPool = anv_hitman3_CmdResetQueryPool,
    .CmdWriteTimestamp = anv_hitman3_CmdWriteTimestamp,
    .CmdCopyQueryPoolResults = anv_hitman3_CmdCopyQueryPoolResults,
    .CmdPushConstants = anv_hitman3_CmdPushConstants,
    .CmdBeginRenderPass = anv_hitman3_CmdBeginRenderPass,
    .CmdNextSubpass = anv_hitman3_CmdNextSubpass,
    .CmdEndRenderPass = anv_hitman3_CmdEndRenderPass,
    .CmdExecuteCommands = anv_hitman3_CmdExecuteCommands,
    .CreateSharedSwapchainsKHR = anv_hitman3_CreateSharedSwapchainsKHR,
    .CreateSwapchainKHR = anv_hitman3_CreateSwapchainKHR,
    .DestroySwapchainKHR = anv_hitman3_DestroySwapchainKHR,
    .GetSwapchainImagesKHR = anv_hitman3_GetSwapchainImagesKHR,
    .AcquireNextImageKHR = anv_hitman3_AcquireNextImageKHR,
    .QueuePresentKHR = anv_hitman3_QueuePresentKHR,
    .DebugMarkerSetObjectNameEXT = anv_hitman3_DebugMarkerSetObjectNameEXT,
    .DebugMarkerSetObjectTagEXT = anv_hitman3_DebugMarkerSetObjectTagEXT,
    .CmdDebugMarkerBeginEXT = anv_hitman3_CmdDebugMarkerBeginEXT,
    .CmdDebugMarkerEndEXT = anv_hitman3_CmdDebugMarkerEndEXT,
    .CmdDebugMarkerInsertEXT = anv_hitman3_CmdDebugMarkerInsertEXT,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleNV = anv_hitman3_GetMemoryWin32HandleNV,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .CmdExecuteGeneratedCommandsNV = anv_hitman3_CmdExecuteGeneratedCommandsNV,
    .CmdPreprocessGeneratedCommandsNV = anv_hitman3_CmdPreprocessGeneratedCommandsNV,
    .CmdBindPipelineShaderGroupNV = anv_hitman3_CmdBindPipelineShaderGroupNV,
    .GetGeneratedCommandsMemoryRequirementsNV = anv_hitman3_GetGeneratedCommandsMemoryRequirementsNV,
    .CreateIndirectCommandsLayoutNV = anv_hitman3_CreateIndirectCommandsLayoutNV,
    .DestroyIndirectCommandsLayoutNV = anv_hitman3_DestroyIndirectCommandsLayoutNV,
    .CmdExecuteGeneratedCommandsEXT = anv_hitman3_CmdExecuteGeneratedCommandsEXT,
    .CmdPreprocessGeneratedCommandsEXT = anv_hitman3_CmdPreprocessGeneratedCommandsEXT,
    .GetGeneratedCommandsMemoryRequirementsEXT = anv_hitman3_GetGeneratedCommandsMemoryRequirementsEXT,
    .CreateIndirectCommandsLayoutEXT = anv_hitman3_CreateIndirectCommandsLayoutEXT,
    .DestroyIndirectCommandsLayoutEXT = anv_hitman3_DestroyIndirectCommandsLayoutEXT,
    .CreateIndirectExecutionSetEXT = anv_hitman3_CreateIndirectExecutionSetEXT,
    .DestroyIndirectExecutionSetEXT = anv_hitman3_DestroyIndirectExecutionSetEXT,
    .UpdateIndirectExecutionSetPipelineEXT = anv_hitman3_UpdateIndirectExecutionSetPipelineEXT,
    .UpdateIndirectExecutionSetShaderEXT = anv_hitman3_UpdateIndirectExecutionSetShaderEXT,
    .CmdPushDescriptorSet = anv_hitman3_CmdPushDescriptorSet,
    .CmdPushDescriptorSetKHR = anv_hitman3_CmdPushDescriptorSetKHR,
    .TrimCommandPool = anv_hitman3_TrimCommandPool,
    .TrimCommandPoolKHR = anv_hitman3_TrimCommandPoolKHR,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleKHR = anv_hitman3_GetMemoryWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandlePropertiesKHR = anv_hitman3_GetMemoryWin32HandlePropertiesKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandlePropertiesKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryFdKHR = anv_hitman3_GetMemoryFdKHR,
    .GetMemoryFdPropertiesKHR = anv_hitman3_GetMemoryFdPropertiesKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandleFUCHSIA = anv_hitman3_GetMemoryZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandlePropertiesFUCHSIA = anv_hitman3_GetMemoryZirconHandlePropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandlePropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .GetMemoryRemoteAddressNV = anv_hitman3_GetMemoryRemoteAddressNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreWin32HandleKHR = anv_hitman3_GetSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportSemaphoreWin32HandleKHR = anv_hitman3_ImportSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreFdKHR = anv_hitman3_GetSemaphoreFdKHR,
    .ImportSemaphoreFdKHR = anv_hitman3_ImportSemaphoreFdKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetSemaphoreZirconHandleFUCHSIA = anv_hitman3_GetSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .ImportSemaphoreZirconHandleFUCHSIA = anv_hitman3_ImportSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .ImportSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetFenceWin32HandleKHR = anv_hitman3_GetFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportFenceWin32HandleKHR = anv_hitman3_ImportFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetFenceFdKHR = anv_hitman3_GetFenceFdKHR,
    .ImportFenceFdKHR = anv_hitman3_ImportFenceFdKHR,
    .DisplayPowerControlEXT = anv_hitman3_DisplayPowerControlEXT,
    .RegisterDeviceEventEXT = anv_hitman3_RegisterDeviceEventEXT,
    .RegisterDisplayEventEXT = anv_hitman3_RegisterDisplayEventEXT,
    .GetSwapchainCounterEXT = anv_hitman3_GetSwapchainCounterEXT,
    .GetDeviceGroupPeerMemoryFeatures = anv_hitman3_GetDeviceGroupPeerMemoryFeatures,
    .GetDeviceGroupPeerMemoryFeaturesKHR = anv_hitman3_GetDeviceGroupPeerMemoryFeaturesKHR,
    .BindBufferMemory2 = anv_hitman3_BindBufferMemory2,
    .BindBufferMemory2KHR = anv_hitman3_BindBufferMemory2KHR,
    .BindImageMemory2 = anv_hitman3_BindImageMemory2,
    .BindImageMemory2KHR = anv_hitman3_BindImageMemory2KHR,
    .CmdSetDeviceMask = anv_hitman3_CmdSetDeviceMask,
    .CmdSetDeviceMaskKHR = anv_hitman3_CmdSetDeviceMaskKHR,
    .GetDeviceGroupPresentCapabilitiesKHR = anv_hitman3_GetDeviceGroupPresentCapabilitiesKHR,
    .GetDeviceGroupSurfacePresentModesKHR = anv_hitman3_GetDeviceGroupSurfacePresentModesKHR,
    .AcquireNextImage2KHR = anv_hitman3_AcquireNextImage2KHR,
    .CmdDispatchBase = anv_hitman3_CmdDispatchBase,
    .CmdDispatchBaseKHR = anv_hitman3_CmdDispatchBaseKHR,
    .CreateDescriptorUpdateTemplate = anv_hitman3_CreateDescriptorUpdateTemplate,
    .CreateDescriptorUpdateTemplateKHR = anv_hitman3_CreateDescriptorUpdateTemplateKHR,
    .DestroyDescriptorUpdateTemplate = anv_hitman3_DestroyDescriptorUpdateTemplate,
    .DestroyDescriptorUpdateTemplateKHR = anv_hitman3_DestroyDescriptorUpdateTemplateKHR,
    .UpdateDescriptorSetWithTemplate = anv_hitman3_UpdateDescriptorSetWithTemplate,
    .UpdateDescriptorSetWithTemplateKHR = anv_hitman3_UpdateDescriptorSetWithTemplateKHR,
    .CmdPushDescriptorSetWithTemplate = anv_hitman3_CmdPushDescriptorSetWithTemplate,
    .CmdPushDescriptorSetWithTemplateKHR = anv_hitman3_CmdPushDescriptorSetWithTemplateKHR,
    .SetHdrMetadataEXT = anv_hitman3_SetHdrMetadataEXT,
    .GetSwapchainStatusKHR = anv_hitman3_GetSwapchainStatusKHR,
    .GetRefreshCycleDurationGOOGLE = anv_hitman3_GetRefreshCycleDurationGOOGLE,
    .GetPastPresentationTimingGOOGLE = anv_hitman3_GetPastPresentationTimingGOOGLE,
    .CmdSetViewportWScalingNV = anv_hitman3_CmdSetViewportWScalingNV,
    .CmdSetDiscardRectangleEXT = anv_hitman3_CmdSetDiscardRectangleEXT,
    .CmdSetDiscardRectangleEnableEXT = anv_hitman3_CmdSetDiscardRectangleEnableEXT,
    .CmdSetDiscardRectangleModeEXT = anv_hitman3_CmdSetDiscardRectangleModeEXT,
    .CmdSetSampleLocationsEXT = anv_hitman3_CmdSetSampleLocationsEXT,
    .GetBufferMemoryRequirements2 = anv_hitman3_GetBufferMemoryRequirements2,
    .GetBufferMemoryRequirements2KHR = anv_hitman3_GetBufferMemoryRequirements2KHR,
    .GetImageMemoryRequirements2 = anv_hitman3_GetImageMemoryRequirements2,
    .GetImageMemoryRequirements2KHR = anv_hitman3_GetImageMemoryRequirements2KHR,
    .GetImageSparseMemoryRequirements2 = anv_hitman3_GetImageSparseMemoryRequirements2,
    .GetImageSparseMemoryRequirements2KHR = anv_hitman3_GetImageSparseMemoryRequirements2KHR,
    .GetDeviceBufferMemoryRequirements = anv_hitman3_GetDeviceBufferMemoryRequirements,
    .GetDeviceBufferMemoryRequirementsKHR = anv_hitman3_GetDeviceBufferMemoryRequirementsKHR,
    .GetDeviceImageMemoryRequirements = anv_hitman3_GetDeviceImageMemoryRequirements,
    .GetDeviceImageMemoryRequirementsKHR = anv_hitman3_GetDeviceImageMemoryRequirementsKHR,
    .GetDeviceImageSparseMemoryRequirements = anv_hitman3_GetDeviceImageSparseMemoryRequirements,
    .GetDeviceImageSparseMemoryRequirementsKHR = anv_hitman3_GetDeviceImageSparseMemoryRequirementsKHR,
    .CreateSamplerYcbcrConversion = anv_hitman3_CreateSamplerYcbcrConversion,
    .CreateSamplerYcbcrConversionKHR = anv_hitman3_CreateSamplerYcbcrConversionKHR,
    .DestroySamplerYcbcrConversion = anv_hitman3_DestroySamplerYcbcrConversion,
    .DestroySamplerYcbcrConversionKHR = anv_hitman3_DestroySamplerYcbcrConversionKHR,
    .GetDeviceQueue2 = anv_hitman3_GetDeviceQueue2,
    .CreateValidationCacheEXT = anv_hitman3_CreateValidationCacheEXT,
    .DestroyValidationCacheEXT = anv_hitman3_DestroyValidationCacheEXT,
    .GetValidationCacheDataEXT = anv_hitman3_GetValidationCacheDataEXT,
    .MergeValidationCachesEXT = anv_hitman3_MergeValidationCachesEXT,
    .GetDescriptorSetLayoutSupport = anv_hitman3_GetDescriptorSetLayoutSupport,
    .GetDescriptorSetLayoutSupportKHR = anv_hitman3_GetDescriptorSetLayoutSupportKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsageANDROID = anv_hitman3_GetSwapchainGrallocUsageANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsage2ANDROID = anv_hitman3_GetSwapchainGrallocUsage2ANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsage2ANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .AcquireImageANDROID = anv_hitman3_AcquireImageANDROID,
#elif defined(_MSC_VER)
    .AcquireImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .QueueSignalReleaseImageANDROID = anv_hitman3_QueueSignalReleaseImageANDROID,
#elif defined(_MSC_VER)
    .QueueSignalReleaseImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .GetShaderInfoAMD = anv_hitman3_GetShaderInfoAMD,
    .SetLocalDimmingAMD = anv_hitman3_SetLocalDimmingAMD,
    .GetCalibratedTimestampsKHR = anv_hitman3_GetCalibratedTimestampsKHR,
    .GetCalibratedTimestampsEXT = anv_hitman3_GetCalibratedTimestampsEXT,
    .SetDebugUtilsObjectNameEXT = anv_hitman3_SetDebugUtilsObjectNameEXT,
    .SetDebugUtilsObjectTagEXT = anv_hitman3_SetDebugUtilsObjectTagEXT,
    .QueueBeginDebugUtilsLabelEXT = anv_hitman3_QueueBeginDebugUtilsLabelEXT,
    .QueueEndDebugUtilsLabelEXT = anv_hitman3_QueueEndDebugUtilsLabelEXT,
    .QueueInsertDebugUtilsLabelEXT = anv_hitman3_QueueInsertDebugUtilsLabelEXT,
    .CmdBeginDebugUtilsLabelEXT = anv_hitman3_CmdBeginDebugUtilsLabelEXT,
    .CmdEndDebugUtilsLabelEXT = anv_hitman3_CmdEndDebugUtilsLabelEXT,
    .CmdInsertDebugUtilsLabelEXT = anv_hitman3_CmdInsertDebugUtilsLabelEXT,
    .GetMemoryHostPointerPropertiesEXT = anv_hitman3_GetMemoryHostPointerPropertiesEXT,
    .CmdWriteBufferMarkerAMD = anv_hitman3_CmdWriteBufferMarkerAMD,
    .CreateRenderPass2 = anv_hitman3_CreateRenderPass2,
    .CreateRenderPass2KHR = anv_hitman3_CreateRenderPass2KHR,
    .CmdBeginRenderPass2 = anv_hitman3_CmdBeginRenderPass2,
    .CmdBeginRenderPass2KHR = anv_hitman3_CmdBeginRenderPass2KHR,
    .CmdNextSubpass2 = anv_hitman3_CmdNextSubpass2,
    .CmdNextSubpass2KHR = anv_hitman3_CmdNextSubpass2KHR,
    .CmdEndRenderPass2 = anv_hitman3_CmdEndRenderPass2,
    .CmdEndRenderPass2KHR = anv_hitman3_CmdEndRenderPass2KHR,
    .GetSemaphoreCounterValue = anv_hitman3_GetSemaphoreCounterValue,
    .GetSemaphoreCounterValueKHR = anv_hitman3_GetSemaphoreCounterValueKHR,
    .WaitSemaphores = anv_hitman3_WaitSemaphores,
    .WaitSemaphoresKHR = anv_hitman3_WaitSemaphoresKHR,
    .SignalSemaphore = anv_hitman3_SignalSemaphore,
    .SignalSemaphoreKHR = anv_hitman3_SignalSemaphoreKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetAndroidHardwareBufferPropertiesANDROID = anv_hitman3_GetAndroidHardwareBufferPropertiesANDROID,
#elif defined(_MSC_VER)
    .GetAndroidHardwareBufferPropertiesANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetMemoryAndroidHardwareBufferANDROID = anv_hitman3_GetMemoryAndroidHardwareBufferANDROID,
#elif defined(_MSC_VER)
    .GetMemoryAndroidHardwareBufferANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .CmdDrawIndirectCount = anv_hitman3_CmdDrawIndirectCount,
    .CmdDrawIndirectCountKHR = anv_hitman3_CmdDrawIndirectCountKHR,
    .CmdDrawIndirectCountAMD = anv_hitman3_CmdDrawIndirectCountAMD,
    .CmdDrawIndexedIndirectCount = anv_hitman3_CmdDrawIndexedIndirectCount,
    .CmdDrawIndexedIndirectCountKHR = anv_hitman3_CmdDrawIndexedIndirectCountKHR,
    .CmdDrawIndexedIndirectCountAMD = anv_hitman3_CmdDrawIndexedIndirectCountAMD,
    .CmdSetCheckpointNV = anv_hitman3_CmdSetCheckpointNV,
    .GetQueueCheckpointDataNV = anv_hitman3_GetQueueCheckpointDataNV,
    .CmdBindTransformFeedbackBuffersEXT = anv_hitman3_CmdBindTransformFeedbackBuffersEXT,
    .CmdBeginTransformFeedbackEXT = anv_hitman3_CmdBeginTransformFeedbackEXT,
    .CmdEndTransformFeedbackEXT = anv_hitman3_CmdEndTransformFeedbackEXT,
    .CmdBeginQueryIndexedEXT = anv_hitman3_CmdBeginQueryIndexedEXT,
    .CmdEndQueryIndexedEXT = anv_hitman3_CmdEndQueryIndexedEXT,
    .CmdDrawIndirectByteCountEXT = anv_hitman3_CmdDrawIndirectByteCountEXT,
    .CmdSetExclusiveScissorNV = anv_hitman3_CmdSetExclusiveScissorNV,
    .CmdSetExclusiveScissorEnableNV = anv_hitman3_CmdSetExclusiveScissorEnableNV,
    .CmdBindShadingRateImageNV = anv_hitman3_CmdBindShadingRateImageNV,
    .CmdSetViewportShadingRatePaletteNV = anv_hitman3_CmdSetViewportShadingRatePaletteNV,
    .CmdSetCoarseSampleOrderNV = anv_hitman3_CmdSetCoarseSampleOrderNV,
    .CmdDrawMeshTasksNV = anv_hitman3_CmdDrawMeshTasksNV,
    .CmdDrawMeshTasksIndirectNV = anv_hitman3_CmdDrawMeshTasksIndirectNV,
    .CmdDrawMeshTasksIndirectCountNV = anv_hitman3_CmdDrawMeshTasksIndirectCountNV,
    .CmdDrawMeshTasksEXT = anv_hitman3_CmdDrawMeshTasksEXT,
    .CmdDrawMeshTasksIndirectEXT = anv_hitman3_CmdDrawMeshTasksIndirectEXT,
    .CmdDrawMeshTasksIndirectCountEXT = anv_hitman3_CmdDrawMeshTasksIndirectCountEXT,
    .CompileDeferredNV = anv_hitman3_CompileDeferredNV,
    .CreateAccelerationStructureNV = anv_hitman3_CreateAccelerationStructureNV,
    .CmdBindInvocationMaskHUAWEI = anv_hitman3_CmdBindInvocationMaskHUAWEI,
    .DestroyAccelerationStructureKHR = anv_hitman3_DestroyAccelerationStructureKHR,
    .DestroyAccelerationStructureNV = anv_hitman3_DestroyAccelerationStructureNV,
    .GetAccelerationStructureMemoryRequirementsNV = anv_hitman3_GetAccelerationStructureMemoryRequirementsNV,
    .BindAccelerationStructureMemoryNV = anv_hitman3_BindAccelerationStructureMemoryNV,
    .CmdCopyAccelerationStructureNV = anv_hitman3_CmdCopyAccelerationStructureNV,
    .CmdCopyAccelerationStructureKHR = anv_hitman3_CmdCopyAccelerationStructureKHR,
    .CopyAccelerationStructureKHR = anv_hitman3_CopyAccelerationStructureKHR,
    .CmdCopyAccelerationStructureToMemoryKHR = anv_hitman3_CmdCopyAccelerationStructureToMemoryKHR,
    .CopyAccelerationStructureToMemoryKHR = anv_hitman3_CopyAccelerationStructureToMemoryKHR,
    .CmdCopyMemoryToAccelerationStructureKHR = anv_hitman3_CmdCopyMemoryToAccelerationStructureKHR,
    .CopyMemoryToAccelerationStructureKHR = anv_hitman3_CopyMemoryToAccelerationStructureKHR,
    .CmdWriteAccelerationStructuresPropertiesKHR = anv_hitman3_CmdWriteAccelerationStructuresPropertiesKHR,
    .CmdWriteAccelerationStructuresPropertiesNV = anv_hitman3_CmdWriteAccelerationStructuresPropertiesNV,
    .CmdBuildAccelerationStructureNV = anv_hitman3_CmdBuildAccelerationStructureNV,
    .WriteAccelerationStructuresPropertiesKHR = anv_hitman3_WriteAccelerationStructuresPropertiesKHR,
    .CmdTraceRaysKHR = anv_hitman3_CmdTraceRaysKHR,
    .CmdTraceRaysNV = anv_hitman3_CmdTraceRaysNV,
    .GetRayTracingShaderGroupHandlesKHR = anv_hitman3_GetRayTracingShaderGroupHandlesKHR,
    .GetRayTracingShaderGroupHandlesNV = anv_hitman3_GetRayTracingShaderGroupHandlesNV,
    .GetRayTracingCaptureReplayShaderGroupHandlesKHR = anv_hitman3_GetRayTracingCaptureReplayShaderGroupHandlesKHR,
    .GetAccelerationStructureHandleNV = anv_hitman3_GetAccelerationStructureHandleNV,
    .CreateRayTracingPipelinesNV = anv_hitman3_CreateRayTracingPipelinesNV,
    .CreateRayTracingPipelinesKHR = anv_hitman3_CreateRayTracingPipelinesKHR,
    .CmdTraceRaysIndirectKHR = anv_hitman3_CmdTraceRaysIndirectKHR,
    .CmdTraceRaysIndirect2KHR = anv_hitman3_CmdTraceRaysIndirect2KHR,
    .GetClusterAccelerationStructureBuildSizesNV = anv_hitman3_GetClusterAccelerationStructureBuildSizesNV,
    .CmdBuildClusterAccelerationStructureIndirectNV = anv_hitman3_CmdBuildClusterAccelerationStructureIndirectNV,
    .GetDeviceAccelerationStructureCompatibilityKHR = anv_hitman3_GetDeviceAccelerationStructureCompatibilityKHR,
    .GetRayTracingShaderGroupStackSizeKHR = anv_hitman3_GetRayTracingShaderGroupStackSizeKHR,
    .CmdSetRayTracingPipelineStackSizeKHR = anv_hitman3_CmdSetRayTracingPipelineStackSizeKHR,
    .GetImageViewHandleNVX = anv_hitman3_GetImageViewHandleNVX,
    .GetImageViewHandle64NVX = anv_hitman3_GetImageViewHandle64NVX,
    .GetImageViewAddressNVX = anv_hitman3_GetImageViewAddressNVX,
    .GetDeviceCombinedImageSamplerIndexNVX = anv_hitman3_GetDeviceCombinedImageSamplerIndexNVX,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetDeviceGroupSurfacePresentModes2EXT = anv_hitman3_GetDeviceGroupSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetDeviceGroupSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireFullScreenExclusiveModeEXT = anv_hitman3_AcquireFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .AcquireFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ReleaseFullScreenExclusiveModeEXT = anv_hitman3_ReleaseFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .ReleaseFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .AcquireProfilingLockKHR = anv_hitman3_AcquireProfilingLockKHR,
    .ReleaseProfilingLockKHR = anv_hitman3_ReleaseProfilingLockKHR,
    .GetImageDrmFormatModifierPropertiesEXT = anv_hitman3_GetImageDrmFormatModifierPropertiesEXT,
    .GetBufferOpaqueCaptureAddress = anv_hitman3_GetBufferOpaqueCaptureAddress,
    .GetBufferOpaqueCaptureAddressKHR = anv_hitman3_GetBufferOpaqueCaptureAddressKHR,
    .GetBufferDeviceAddress = anv_hitman3_GetBufferDeviceAddress,
    .GetBufferDeviceAddressKHR = anv_hitman3_GetBufferDeviceAddressKHR,
    .GetBufferDeviceAddressEXT = anv_hitman3_GetBufferDeviceAddressEXT,
    .InitializePerformanceApiINTEL = anv_hitman3_InitializePerformanceApiINTEL,
    .UninitializePerformanceApiINTEL = anv_hitman3_UninitializePerformanceApiINTEL,
    .CmdSetPerformanceMarkerINTEL = anv_hitman3_CmdSetPerformanceMarkerINTEL,
    .CmdSetPerformanceStreamMarkerINTEL = anv_hitman3_CmdSetPerformanceStreamMarkerINTEL,
    .CmdSetPerformanceOverrideINTEL = anv_hitman3_CmdSetPerformanceOverrideINTEL,
    .AcquirePerformanceConfigurationINTEL = anv_hitman3_AcquirePerformanceConfigurationINTEL,
    .ReleasePerformanceConfigurationINTEL = anv_hitman3_ReleasePerformanceConfigurationINTEL,
    .QueueSetPerformanceConfigurationINTEL = anv_hitman3_QueueSetPerformanceConfigurationINTEL,
    .GetPerformanceParameterINTEL = anv_hitman3_GetPerformanceParameterINTEL,
    .GetDeviceMemoryOpaqueCaptureAddress = anv_hitman3_GetDeviceMemoryOpaqueCaptureAddress,
    .GetDeviceMemoryOpaqueCaptureAddressKHR = anv_hitman3_GetDeviceMemoryOpaqueCaptureAddressKHR,
    .GetPipelineExecutablePropertiesKHR = anv_hitman3_GetPipelineExecutablePropertiesKHR,
    .GetPipelineExecutableStatisticsKHR = anv_hitman3_GetPipelineExecutableStatisticsKHR,
    .GetPipelineExecutableInternalRepresentationsKHR = anv_hitman3_GetPipelineExecutableInternalRepresentationsKHR,
    .CmdSetLineStipple = anv_hitman3_CmdSetLineStipple,
    .CmdSetLineStippleKHR = anv_hitman3_CmdSetLineStippleKHR,
    .CmdSetLineStippleEXT = anv_hitman3_CmdSetLineStippleEXT,
    .CreateAccelerationStructureKHR = anv_hitman3_CreateAccelerationStructureKHR,
    .CmdBuildAccelerationStructuresKHR = anv_hitman3_CmdBuildAccelerationStructuresKHR,
    .CmdBuildAccelerationStructuresIndirectKHR = anv_hitman3_CmdBuildAccelerationStructuresIndirectKHR,
    .BuildAccelerationStructuresKHR = anv_hitman3_BuildAccelerationStructuresKHR,
    .GetAccelerationStructureDeviceAddressKHR = anv_hitman3_GetAccelerationStructureDeviceAddressKHR,
    .CreateDeferredOperationKHR = anv_hitman3_CreateDeferredOperationKHR,
    .DestroyDeferredOperationKHR = anv_hitman3_DestroyDeferredOperationKHR,
    .GetDeferredOperationMaxConcurrencyKHR = anv_hitman3_GetDeferredOperationMaxConcurrencyKHR,
    .GetDeferredOperationResultKHR = anv_hitman3_GetDeferredOperationResultKHR,
    .DeferredOperationJoinKHR = anv_hitman3_DeferredOperationJoinKHR,
    .GetPipelineIndirectMemoryRequirementsNV = anv_hitman3_GetPipelineIndirectMemoryRequirementsNV,
    .GetPipelineIndirectDeviceAddressNV = anv_hitman3_GetPipelineIndirectDeviceAddressNV,
    .AntiLagUpdateAMD = anv_hitman3_AntiLagUpdateAMD,
    .CmdSetCullMode = anv_hitman3_CmdSetCullMode,
    .CmdSetCullModeEXT = anv_hitman3_CmdSetCullModeEXT,
    .CmdSetFrontFace = anv_hitman3_CmdSetFrontFace,
    .CmdSetFrontFaceEXT = anv_hitman3_CmdSetFrontFaceEXT,
    .CmdSetPrimitiveTopology = anv_hitman3_CmdSetPrimitiveTopology,
    .CmdSetPrimitiveTopologyEXT = anv_hitman3_CmdSetPrimitiveTopologyEXT,
    .CmdSetViewportWithCount = anv_hitman3_CmdSetViewportWithCount,
    .CmdSetViewportWithCountEXT = anv_hitman3_CmdSetViewportWithCountEXT,
    .CmdSetScissorWithCount = anv_hitman3_CmdSetScissorWithCount,
    .CmdSetScissorWithCountEXT = anv_hitman3_CmdSetScissorWithCountEXT,
    .CmdBindIndexBuffer2 = anv_hitman3_CmdBindIndexBuffer2,
    .CmdBindIndexBuffer2KHR = anv_hitman3_CmdBindIndexBuffer2KHR,
    .CmdBindVertexBuffers2 = anv_hitman3_CmdBindVertexBuffers2,
    .CmdBindVertexBuffers2EXT = anv_hitman3_CmdBindVertexBuffers2EXT,
    .CmdSetDepthTestEnable = anv_hitman3_CmdSetDepthTestEnable,
    .CmdSetDepthTestEnableEXT = anv_hitman3_CmdSetDepthTestEnableEXT,
    .CmdSetDepthWriteEnable = anv_hitman3_CmdSetDepthWriteEnable,
    .CmdSetDepthWriteEnableEXT = anv_hitman3_CmdSetDepthWriteEnableEXT,
    .CmdSetDepthCompareOp = anv_hitman3_CmdSetDepthCompareOp,
    .CmdSetDepthCompareOpEXT = anv_hitman3_CmdSetDepthCompareOpEXT,
    .CmdSetDepthBoundsTestEnable = anv_hitman3_CmdSetDepthBoundsTestEnable,
    .CmdSetDepthBoundsTestEnableEXT = anv_hitman3_CmdSetDepthBoundsTestEnableEXT,
    .CmdSetStencilTestEnable = anv_hitman3_CmdSetStencilTestEnable,
    .CmdSetStencilTestEnableEXT = anv_hitman3_CmdSetStencilTestEnableEXT,
    .CmdSetStencilOp = anv_hitman3_CmdSetStencilOp,
    .CmdSetStencilOpEXT = anv_hitman3_CmdSetStencilOpEXT,
    .CmdSetPatchControlPointsEXT = anv_hitman3_CmdSetPatchControlPointsEXT,
    .CmdSetRasterizerDiscardEnable = anv_hitman3_CmdSetRasterizerDiscardEnable,
    .CmdSetRasterizerDiscardEnableEXT = anv_hitman3_CmdSetRasterizerDiscardEnableEXT,
    .CmdSetDepthBiasEnable = anv_hitman3_CmdSetDepthBiasEnable,
    .CmdSetDepthBiasEnableEXT = anv_hitman3_CmdSetDepthBiasEnableEXT,
    .CmdSetLogicOpEXT = anv_hitman3_CmdSetLogicOpEXT,
    .CmdSetPrimitiveRestartEnable = anv_hitman3_CmdSetPrimitiveRestartEnable,
    .CmdSetPrimitiveRestartEnableEXT = anv_hitman3_CmdSetPrimitiveRestartEnableEXT,
    .CmdSetTessellationDomainOriginEXT = anv_hitman3_CmdSetTessellationDomainOriginEXT,
    .CmdSetDepthClampEnableEXT = anv_hitman3_CmdSetDepthClampEnableEXT,
    .CmdSetPolygonModeEXT = anv_hitman3_CmdSetPolygonModeEXT,
    .CmdSetRasterizationSamplesEXT = anv_hitman3_CmdSetRasterizationSamplesEXT,
    .CmdSetSampleMaskEXT = anv_hitman3_CmdSetSampleMaskEXT,
    .CmdSetAlphaToCoverageEnableEXT = anv_hitman3_CmdSetAlphaToCoverageEnableEXT,
    .CmdSetAlphaToOneEnableEXT = anv_hitman3_CmdSetAlphaToOneEnableEXT,
    .CmdSetLogicOpEnableEXT = anv_hitman3_CmdSetLogicOpEnableEXT,
    .CmdSetColorBlendEnableEXT = anv_hitman3_CmdSetColorBlendEnableEXT,
    .CmdSetColorBlendEquationEXT = anv_hitman3_CmdSetColorBlendEquationEXT,
    .CmdSetColorWriteMaskEXT = anv_hitman3_CmdSetColorWriteMaskEXT,
    .CmdSetRasterizationStreamEXT = anv_hitman3_CmdSetRasterizationStreamEXT,
    .CmdSetConservativeRasterizationModeEXT = anv_hitman3_CmdSetConservativeRasterizationModeEXT,
    .CmdSetExtraPrimitiveOverestimationSizeEXT = anv_hitman3_CmdSetExtraPrimitiveOverestimationSizeEXT,
    .CmdSetDepthClipEnableEXT = anv_hitman3_CmdSetDepthClipEnableEXT,
    .CmdSetSampleLocationsEnableEXT = anv_hitman3_CmdSetSampleLocationsEnableEXT,
    .CmdSetColorBlendAdvancedEXT = anv_hitman3_CmdSetColorBlendAdvancedEXT,
    .CmdSetProvokingVertexModeEXT = anv_hitman3_CmdSetProvokingVertexModeEXT,
    .CmdSetLineRasterizationModeEXT = anv_hitman3_CmdSetLineRasterizationModeEXT,
    .CmdSetLineStippleEnableEXT = anv_hitman3_CmdSetLineStippleEnableEXT,
    .CmdSetDepthClipNegativeOneToOneEXT = anv_hitman3_CmdSetDepthClipNegativeOneToOneEXT,
    .CmdSetViewportWScalingEnableNV = anv_hitman3_CmdSetViewportWScalingEnableNV,
    .CmdSetViewportSwizzleNV = anv_hitman3_CmdSetViewportSwizzleNV,
    .CmdSetCoverageToColorEnableNV = anv_hitman3_CmdSetCoverageToColorEnableNV,
    .CmdSetCoverageToColorLocationNV = anv_hitman3_CmdSetCoverageToColorLocationNV,
    .CmdSetCoverageModulationModeNV = anv_hitman3_CmdSetCoverageModulationModeNV,
    .CmdSetCoverageModulationTableEnableNV = anv_hitman3_CmdSetCoverageModulationTableEnableNV,
    .CmdSetCoverageModulationTableNV = anv_hitman3_CmdSetCoverageModulationTableNV,
    .CmdSetShadingRateImageEnableNV = anv_hitman3_CmdSetShadingRateImageEnableNV,
    .CmdSetCoverageReductionModeNV = anv_hitman3_CmdSetCoverageReductionModeNV,
    .CmdSetRepresentativeFragmentTestEnableNV = anv_hitman3_CmdSetRepresentativeFragmentTestEnableNV,
    .CreatePrivateDataSlot = anv_hitman3_CreatePrivateDataSlot,
    .CreatePrivateDataSlotEXT = anv_hitman3_CreatePrivateDataSlotEXT,
    .DestroyPrivateDataSlot = anv_hitman3_DestroyPrivateDataSlot,
    .DestroyPrivateDataSlotEXT = anv_hitman3_DestroyPrivateDataSlotEXT,
    .SetPrivateData = anv_hitman3_SetPrivateData,
    .SetPrivateDataEXT = anv_hitman3_SetPrivateDataEXT,
    .GetPrivateData = anv_hitman3_GetPrivateData,
    .GetPrivateDataEXT = anv_hitman3_GetPrivateDataEXT,
    .CmdCopyBuffer2 = anv_hitman3_CmdCopyBuffer2,
    .CmdCopyBuffer2KHR = anv_hitman3_CmdCopyBuffer2KHR,
    .CmdCopyImage2 = anv_hitman3_CmdCopyImage2,
    .CmdCopyImage2KHR = anv_hitman3_CmdCopyImage2KHR,
    .CmdBlitImage2 = anv_hitman3_CmdBlitImage2,
    .CmdBlitImage2KHR = anv_hitman3_CmdBlitImage2KHR,
    .CmdCopyBufferToImage2 = anv_hitman3_CmdCopyBufferToImage2,
    .CmdCopyBufferToImage2KHR = anv_hitman3_CmdCopyBufferToImage2KHR,
    .CmdCopyImageToBuffer2 = anv_hitman3_CmdCopyImageToBuffer2,
    .CmdCopyImageToBuffer2KHR = anv_hitman3_CmdCopyImageToBuffer2KHR,
    .CmdResolveImage2 = anv_hitman3_CmdResolveImage2,
    .CmdResolveImage2KHR = anv_hitman3_CmdResolveImage2KHR,
    .CmdSetFragmentShadingRateKHR = anv_hitman3_CmdSetFragmentShadingRateKHR,
    .CmdSetFragmentShadingRateEnumNV = anv_hitman3_CmdSetFragmentShadingRateEnumNV,
    .GetAccelerationStructureBuildSizesKHR = anv_hitman3_GetAccelerationStructureBuildSizesKHR,
    .CmdSetVertexInputEXT = anv_hitman3_CmdSetVertexInputEXT,
    .CmdSetColorWriteEnableEXT = anv_hitman3_CmdSetColorWriteEnableEXT,
    .CmdSetEvent2 = anv_hitman3_CmdSetEvent2,
    .CmdSetEvent2KHR = anv_hitman3_CmdSetEvent2KHR,
    .CmdResetEvent2 = anv_hitman3_CmdResetEvent2,
    .CmdResetEvent2KHR = anv_hitman3_CmdResetEvent2KHR,
    .CmdWaitEvents2 = anv_hitman3_CmdWaitEvents2,
    .CmdWaitEvents2KHR = anv_hitman3_CmdWaitEvents2KHR,
    .CmdPipelineBarrier2 = anv_hitman3_CmdPipelineBarrier2,
    .CmdPipelineBarrier2KHR = anv_hitman3_CmdPipelineBarrier2KHR,
    .QueueSubmit2 = anv_hitman3_QueueSubmit2,
    .QueueSubmit2KHR = anv_hitman3_QueueSubmit2KHR,
    .CmdWriteTimestamp2 = anv_hitman3_CmdWriteTimestamp2,
    .CmdWriteTimestamp2KHR = anv_hitman3_CmdWriteTimestamp2KHR,
    .CmdWriteBufferMarker2AMD = anv_hitman3_CmdWriteBufferMarker2AMD,
    .GetQueueCheckpointData2NV = anv_hitman3_GetQueueCheckpointData2NV,
    .CopyMemoryToImage = anv_hitman3_CopyMemoryToImage,
    .CopyMemoryToImageEXT = anv_hitman3_CopyMemoryToImageEXT,
    .CopyImageToMemory = anv_hitman3_CopyImageToMemory,
    .CopyImageToMemoryEXT = anv_hitman3_CopyImageToMemoryEXT,
    .CopyImageToImage = anv_hitman3_CopyImageToImage,
    .CopyImageToImageEXT = anv_hitman3_CopyImageToImageEXT,
    .TransitionImageLayout = anv_hitman3_TransitionImageLayout,
    .TransitionImageLayoutEXT = anv_hitman3_TransitionImageLayoutEXT,
    .CreateVideoSessionKHR = anv_hitman3_CreateVideoSessionKHR,
    .DestroyVideoSessionKHR = anv_hitman3_DestroyVideoSessionKHR,
    .CreateVideoSessionParametersKHR = anv_hitman3_CreateVideoSessionParametersKHR,
    .UpdateVideoSessionParametersKHR = anv_hitman3_UpdateVideoSessionParametersKHR,
    .GetEncodedVideoSessionParametersKHR = anv_hitman3_GetEncodedVideoSessionParametersKHR,
    .DestroyVideoSessionParametersKHR = anv_hitman3_DestroyVideoSessionParametersKHR,
    .GetVideoSessionMemoryRequirementsKHR = anv_hitman3_GetVideoSessionMemoryRequirementsKHR,
    .BindVideoSessionMemoryKHR = anv_hitman3_BindVideoSessionMemoryKHR,
    .CmdDecodeVideoKHR = anv_hitman3_CmdDecodeVideoKHR,
    .CmdBeginVideoCodingKHR = anv_hitman3_CmdBeginVideoCodingKHR,
    .CmdControlVideoCodingKHR = anv_hitman3_CmdControlVideoCodingKHR,
    .CmdEndVideoCodingKHR = anv_hitman3_CmdEndVideoCodingKHR,
    .CmdEncodeVideoKHR = anv_hitman3_CmdEncodeVideoKHR,
    .CmdDecompressMemoryNV = anv_hitman3_CmdDecompressMemoryNV,
    .CmdDecompressMemoryIndirectCountNV = anv_hitman3_CmdDecompressMemoryIndirectCountNV,
    .GetPartitionedAccelerationStructuresBuildSizesNV = anv_hitman3_GetPartitionedAccelerationStructuresBuildSizesNV,
    .CmdBuildPartitionedAccelerationStructuresNV = anv_hitman3_CmdBuildPartitionedAccelerationStructuresNV,
    .CmdDecompressMemoryEXT = anv_hitman3_CmdDecompressMemoryEXT,
    .CmdDecompressMemoryIndirectCountEXT = anv_hitman3_CmdDecompressMemoryIndirectCountEXT,
    .CreateCuModuleNVX = anv_hitman3_CreateCuModuleNVX,
    .CreateCuFunctionNVX = anv_hitman3_CreateCuFunctionNVX,
    .DestroyCuModuleNVX = anv_hitman3_DestroyCuModuleNVX,
    .DestroyCuFunctionNVX = anv_hitman3_DestroyCuFunctionNVX,
    .CmdCuLaunchKernelNVX = anv_hitman3_CmdCuLaunchKernelNVX,
    .GetDescriptorSetLayoutSizeEXT = anv_hitman3_GetDescriptorSetLayoutSizeEXT,
    .GetDescriptorSetLayoutBindingOffsetEXT = anv_hitman3_GetDescriptorSetLayoutBindingOffsetEXT,
    .GetDescriptorEXT = anv_hitman3_GetDescriptorEXT,
    .CmdBindDescriptorBuffersEXT = anv_hitman3_CmdBindDescriptorBuffersEXT,
    .CmdSetDescriptorBufferOffsetsEXT = anv_hitman3_CmdSetDescriptorBufferOffsetsEXT,
    .CmdBindDescriptorBufferEmbeddedSamplersEXT = anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplersEXT,
    .GetBufferOpaqueCaptureDescriptorDataEXT = anv_hitman3_GetBufferOpaqueCaptureDescriptorDataEXT,
    .GetImageOpaqueCaptureDescriptorDataEXT = anv_hitman3_GetImageOpaqueCaptureDescriptorDataEXT,
    .GetImageViewOpaqueCaptureDescriptorDataEXT = anv_hitman3_GetImageViewOpaqueCaptureDescriptorDataEXT,
    .GetSamplerOpaqueCaptureDescriptorDataEXT = anv_hitman3_GetSamplerOpaqueCaptureDescriptorDataEXT,
    .GetAccelerationStructureOpaqueCaptureDescriptorDataEXT = anv_hitman3_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT,
    .SetDeviceMemoryPriorityEXT = anv_hitman3_SetDeviceMemoryPriorityEXT,
    .WaitForPresent2KHR = anv_hitman3_WaitForPresent2KHR,
    .WaitForPresentKHR = anv_hitman3_WaitForPresentKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateBufferCollectionFUCHSIA = anv_hitman3_CreateBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .CreateBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionBufferConstraintsFUCHSIA = anv_hitman3_SetBufferCollectionBufferConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionBufferConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionImageConstraintsFUCHSIA = anv_hitman3_SetBufferCollectionImageConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionImageConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .DestroyBufferCollectionFUCHSIA = anv_hitman3_DestroyBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .DestroyBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetBufferCollectionPropertiesFUCHSIA = anv_hitman3_GetBufferCollectionPropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetBufferCollectionPropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .CmdBeginRendering = anv_hitman3_CmdBeginRendering,
    .CmdBeginRenderingKHR = anv_hitman3_CmdBeginRenderingKHR,
    .CmdEndRendering = anv_hitman3_CmdEndRendering,
    .CmdEndRendering2KHR = anv_hitman3_CmdEndRendering2KHR,
    .CmdEndRendering2EXT = anv_hitman3_CmdEndRendering2EXT,
    .CmdEndRenderingKHR = anv_hitman3_CmdEndRenderingKHR,
    .GetDescriptorSetLayoutHostMappingInfoVALVE = anv_hitman3_GetDescriptorSetLayoutHostMappingInfoVALVE,
    .GetDescriptorSetHostMappingVALVE = anv_hitman3_GetDescriptorSetHostMappingVALVE,
    .CreateMicromapEXT = anv_hitman3_CreateMicromapEXT,
    .CmdBuildMicromapsEXT = anv_hitman3_CmdBuildMicromapsEXT,
    .BuildMicromapsEXT = anv_hitman3_BuildMicromapsEXT,
    .DestroyMicromapEXT = anv_hitman3_DestroyMicromapEXT,
    .CmdCopyMicromapEXT = anv_hitman3_CmdCopyMicromapEXT,
    .CopyMicromapEXT = anv_hitman3_CopyMicromapEXT,
    .CmdCopyMicromapToMemoryEXT = anv_hitman3_CmdCopyMicromapToMemoryEXT,
    .CopyMicromapToMemoryEXT = anv_hitman3_CopyMicromapToMemoryEXT,
    .CmdCopyMemoryToMicromapEXT = anv_hitman3_CmdCopyMemoryToMicromapEXT,
    .CopyMemoryToMicromapEXT = anv_hitman3_CopyMemoryToMicromapEXT,
    .CmdWriteMicromapsPropertiesEXT = anv_hitman3_CmdWriteMicromapsPropertiesEXT,
    .WriteMicromapsPropertiesEXT = anv_hitman3_WriteMicromapsPropertiesEXT,
    .GetDeviceMicromapCompatibilityEXT = anv_hitman3_GetDeviceMicromapCompatibilityEXT,
    .GetMicromapBuildSizesEXT = anv_hitman3_GetMicromapBuildSizesEXT,
    .GetShaderModuleIdentifierEXT = anv_hitman3_GetShaderModuleIdentifierEXT,
    .GetShaderModuleCreateInfoIdentifierEXT = anv_hitman3_GetShaderModuleCreateInfoIdentifierEXT,
    .GetImageSubresourceLayout2 = anv_hitman3_GetImageSubresourceLayout2,
    .GetImageSubresourceLayout2KHR = anv_hitman3_GetImageSubresourceLayout2KHR,
    .GetImageSubresourceLayout2EXT = anv_hitman3_GetImageSubresourceLayout2EXT,
    .GetPipelinePropertiesEXT = anv_hitman3_GetPipelinePropertiesEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .ExportMetalObjectsEXT = anv_hitman3_ExportMetalObjectsEXT,
#elif defined(_MSC_VER)
    .ExportMetalObjectsEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CmdBindTileMemoryQCOM = anv_hitman3_CmdBindTileMemoryQCOM,
    .GetFramebufferTilePropertiesQCOM = anv_hitman3_GetFramebufferTilePropertiesQCOM,
    .GetDynamicRenderingTilePropertiesQCOM = anv_hitman3_GetDynamicRenderingTilePropertiesQCOM,
    .CreateOpticalFlowSessionNV = anv_hitman3_CreateOpticalFlowSessionNV,
    .DestroyOpticalFlowSessionNV = anv_hitman3_DestroyOpticalFlowSessionNV,
    .BindOpticalFlowSessionImageNV = anv_hitman3_BindOpticalFlowSessionImageNV,
    .CmdOpticalFlowExecuteNV = anv_hitman3_CmdOpticalFlowExecuteNV,
    .GetDeviceFaultInfoEXT = anv_hitman3_GetDeviceFaultInfoEXT,
    .GetDeviceFaultReportsKHR = anv_hitman3_GetDeviceFaultReportsKHR,
    .GetDeviceFaultDebugInfoKHR = anv_hitman3_GetDeviceFaultDebugInfoKHR,
    .CmdSetDepthBias2EXT = anv_hitman3_CmdSetDepthBias2EXT,
    .ReleaseSwapchainImagesKHR = anv_hitman3_ReleaseSwapchainImagesKHR,
    .ReleaseSwapchainImagesEXT = anv_hitman3_ReleaseSwapchainImagesEXT,
    .GetDeviceImageSubresourceLayout = anv_hitman3_GetDeviceImageSubresourceLayout,
    .GetDeviceImageSubresourceLayoutKHR = anv_hitman3_GetDeviceImageSubresourceLayoutKHR,
    .MapMemory2 = anv_hitman3_MapMemory2,
    .MapMemory2KHR = anv_hitman3_MapMemory2KHR,
    .UnmapMemory2 = anv_hitman3_UnmapMemory2,
    .UnmapMemory2KHR = anv_hitman3_UnmapMemory2KHR,
    .CreateShadersEXT = anv_hitman3_CreateShadersEXT,
    .DestroyShaderEXT = anv_hitman3_DestroyShaderEXT,
    .GetShaderBinaryDataEXT = anv_hitman3_GetShaderBinaryDataEXT,
    .CmdBindShadersEXT = anv_hitman3_CmdBindShadersEXT,
    .SetSwapchainPresentTimingQueueSizeEXT = anv_hitman3_SetSwapchainPresentTimingQueueSizeEXT,
    .GetSwapchainTimingPropertiesEXT = anv_hitman3_GetSwapchainTimingPropertiesEXT,
    .GetSwapchainTimeDomainPropertiesEXT = anv_hitman3_GetSwapchainTimeDomainPropertiesEXT,
    .GetPastPresentationTimingEXT = anv_hitman3_GetPastPresentationTimingEXT,
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetScreenBufferPropertiesQNX = anv_hitman3_GetScreenBufferPropertiesQNX,
#elif defined(_MSC_VER)
    .GetScreenBufferPropertiesQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CmdBindDescriptorSets2 = anv_hitman3_CmdBindDescriptorSets2,
    .CmdBindDescriptorSets2KHR = anv_hitman3_CmdBindDescriptorSets2KHR,
    .CmdPushConstants2 = anv_hitman3_CmdPushConstants2,
    .CmdPushConstants2KHR = anv_hitman3_CmdPushConstants2KHR,
    .CmdPushDescriptorSet2 = anv_hitman3_CmdPushDescriptorSet2,
    .CmdPushDescriptorSet2KHR = anv_hitman3_CmdPushDescriptorSet2KHR,
    .CmdPushDescriptorSetWithTemplate2 = anv_hitman3_CmdPushDescriptorSetWithTemplate2,
    .CmdPushDescriptorSetWithTemplate2KHR = anv_hitman3_CmdPushDescriptorSetWithTemplate2KHR,
    .CmdSetDescriptorBufferOffsets2EXT = anv_hitman3_CmdSetDescriptorBufferOffsets2EXT,
    .CmdBindDescriptorBufferEmbeddedSamplers2EXT = anv_hitman3_CmdBindDescriptorBufferEmbeddedSamplers2EXT,
    .SetLatencySleepModeNV = anv_hitman3_SetLatencySleepModeNV,
    .LatencySleepNV = anv_hitman3_LatencySleepNV,
    .SetLatencyMarkerNV = anv_hitman3_SetLatencyMarkerNV,
    .GetLatencyTimingsNV = anv_hitman3_GetLatencyTimingsNV,
    .QueueNotifyOutOfBandNV = anv_hitman3_QueueNotifyOutOfBandNV,
    .CmdSetRenderingAttachmentLocations = anv_hitman3_CmdSetRenderingAttachmentLocations,
    .CmdSetRenderingAttachmentLocationsKHR = anv_hitman3_CmdSetRenderingAttachmentLocationsKHR,
    .CmdSetRenderingInputAttachmentIndices = anv_hitman3_CmdSetRenderingInputAttachmentIndices,
    .CmdSetRenderingInputAttachmentIndicesKHR = anv_hitman3_CmdSetRenderingInputAttachmentIndicesKHR,
    .CmdSetDepthClampRangeEXT = anv_hitman3_CmdSetDepthClampRangeEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandleEXT = anv_hitman3_GetMemoryMetalHandleEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandleEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandlePropertiesEXT = anv_hitman3_GetMemoryMetalHandlePropertiesEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandlePropertiesEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .ConvertCooperativeVectorMatrixNV = anv_hitman3_ConvertCooperativeVectorMatrixNV,
    .CmdConvertCooperativeVectorMatrixNV = anv_hitman3_CmdConvertCooperativeVectorMatrixNV,
    .CmdDispatchTileQCOM = anv_hitman3_CmdDispatchTileQCOM,
    .CmdBeginPerTileExecutionQCOM = anv_hitman3_CmdBeginPerTileExecutionQCOM,
    .CmdEndPerTileExecutionQCOM = anv_hitman3_CmdEndPerTileExecutionQCOM,
    .CreateExternalComputeQueueNV = anv_hitman3_CreateExternalComputeQueueNV,
    .DestroyExternalComputeQueueNV = anv_hitman3_DestroyExternalComputeQueueNV,
    .CreateShaderInstrumentationARM = anv_hitman3_CreateShaderInstrumentationARM,
    .DestroyShaderInstrumentationARM = anv_hitman3_DestroyShaderInstrumentationARM,
    .CmdBeginShaderInstrumentationARM = anv_hitman3_CmdBeginShaderInstrumentationARM,
    .CmdEndShaderInstrumentationARM = anv_hitman3_CmdEndShaderInstrumentationARM,
    .GetShaderInstrumentationValuesARM = anv_hitman3_GetShaderInstrumentationValuesARM,
    .ClearShaderInstrumentationMetricsARM = anv_hitman3_ClearShaderInstrumentationMetricsARM,
    .CreateTensorARM = anv_hitman3_CreateTensorARM,
    .DestroyTensorARM = anv_hitman3_DestroyTensorARM,
    .CreateTensorViewARM = anv_hitman3_CreateTensorViewARM,
    .DestroyTensorViewARM = anv_hitman3_DestroyTensorViewARM,
    .GetTensorMemoryRequirementsARM = anv_hitman3_GetTensorMemoryRequirementsARM,
    .BindTensorMemoryARM = anv_hitman3_BindTensorMemoryARM,
    .GetDeviceTensorMemoryRequirementsARM = anv_hitman3_GetDeviceTensorMemoryRequirementsARM,
    .CmdCopyTensorARM = anv_hitman3_CmdCopyTensorARM,
    .GetTensorOpaqueCaptureDescriptorDataARM = anv_hitman3_GetTensorOpaqueCaptureDescriptorDataARM,
    .GetTensorViewOpaqueCaptureDescriptorDataARM = anv_hitman3_GetTensorViewOpaqueCaptureDescriptorDataARM,
    .CreateDataGraphPipelinesARM = anv_hitman3_CreateDataGraphPipelinesARM,
    .CreateDataGraphPipelineSessionARM = anv_hitman3_CreateDataGraphPipelineSessionARM,
    .GetDataGraphPipelineSessionBindPointRequirementsARM = anv_hitman3_GetDataGraphPipelineSessionBindPointRequirementsARM,
    .GetDataGraphPipelineSessionMemoryRequirementsARM = anv_hitman3_GetDataGraphPipelineSessionMemoryRequirementsARM,
    .BindDataGraphPipelineSessionMemoryARM = anv_hitman3_BindDataGraphPipelineSessionMemoryARM,
    .DestroyDataGraphPipelineSessionARM = anv_hitman3_DestroyDataGraphPipelineSessionARM,
    .CmdDispatchDataGraphARM = anv_hitman3_CmdDispatchDataGraphARM,
    .GetDataGraphPipelineAvailablePropertiesARM = anv_hitman3_GetDataGraphPipelineAvailablePropertiesARM,
    .GetDataGraphPipelinePropertiesARM = anv_hitman3_GetDataGraphPipelinePropertiesARM,
#ifdef VK_USE_PLATFORM_OHOS
    .GetNativeBufferPropertiesOHOS = anv_hitman3_GetNativeBufferPropertiesOHOS,
#elif defined(_MSC_VER)
    .GetNativeBufferPropertiesOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
    .GetMemoryNativeBufferOHOS = anv_hitman3_GetMemoryNativeBufferOHOS,
#elif defined(_MSC_VER)
    .GetMemoryNativeBufferOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .QueueSetPerfHintQCOM = anv_hitman3_QueueSetPerfHintQCOM,
    .CmdSetComputeOccupancyPriorityNV = anv_hitman3_CmdSetComputeOccupancyPriorityNV,
    .WriteSamplerDescriptorsEXT = anv_hitman3_WriteSamplerDescriptorsEXT,
    .WriteResourceDescriptorsEXT = anv_hitman3_WriteResourceDescriptorsEXT,
    .CmdBindSamplerHeapEXT = anv_hitman3_CmdBindSamplerHeapEXT,
    .CmdBindResourceHeapEXT = anv_hitman3_CmdBindResourceHeapEXT,
    .CmdPushDataEXT = anv_hitman3_CmdPushDataEXT,
    .RegisterCustomBorderColorEXT = anv_hitman3_RegisterCustomBorderColorEXT,
    .UnregisterCustomBorderColorEXT = anv_hitman3_UnregisterCustomBorderColorEXT,
    .GetImageOpaqueCaptureDataEXT = anv_hitman3_GetImageOpaqueCaptureDataEXT,
    .GetTensorOpaqueCaptureDataARM = anv_hitman3_GetTensorOpaqueCaptureDataARM,
    .CmdCopyMemoryKHR = anv_hitman3_CmdCopyMemoryKHR,
    .CmdCopyMemoryToImageKHR = anv_hitman3_CmdCopyMemoryToImageKHR,
    .CmdCopyImageToMemoryKHR = anv_hitman3_CmdCopyImageToMemoryKHR,
    .CmdUpdateMemoryKHR = anv_hitman3_CmdUpdateMemoryKHR,
    .CmdFillMemoryKHR = anv_hitman3_CmdFillMemoryKHR,
    .CmdCopyQueryPoolResultsToMemoryKHR = anv_hitman3_CmdCopyQueryPoolResultsToMemoryKHR,
    .CmdBeginConditionalRendering2EXT = anv_hitman3_CmdBeginConditionalRendering2EXT,
    .CmdBindTransformFeedbackBuffers2EXT = anv_hitman3_CmdBindTransformFeedbackBuffers2EXT,
    .CmdBeginTransformFeedback2EXT = anv_hitman3_CmdBeginTransformFeedback2EXT,
    .CmdEndTransformFeedback2EXT = anv_hitman3_CmdEndTransformFeedback2EXT,
    .CmdDrawIndirectByteCount2EXT = anv_hitman3_CmdDrawIndirectByteCount2EXT,
    .CmdWriteMarkerToMemoryAMD = anv_hitman3_CmdWriteMarkerToMemoryAMD,
    .CmdBindIndexBuffer3KHR = anv_hitman3_CmdBindIndexBuffer3KHR,
    .CmdBindVertexBuffers3KHR = anv_hitman3_CmdBindVertexBuffers3KHR,
    .CmdDrawIndirect2KHR = anv_hitman3_CmdDrawIndirect2KHR,
    .CmdDrawIndexedIndirect2KHR = anv_hitman3_CmdDrawIndexedIndirect2KHR,
    .CmdDrawIndirectCount2KHR = anv_hitman3_CmdDrawIndirectCount2KHR,
    .CmdDrawIndexedIndirectCount2KHR = anv_hitman3_CmdDrawIndexedIndirectCount2KHR,
    .CmdDrawMeshTasksIndirect2EXT = anv_hitman3_CmdDrawMeshTasksIndirect2EXT,
    .CmdDrawMeshTasksIndirectCount2EXT = anv_hitman3_CmdDrawMeshTasksIndirectCount2EXT,
    .CmdDispatchIndirect2KHR = anv_hitman3_CmdDispatchIndirect2KHR,
    .CreateAccelerationStructure2KHR = anv_hitman3_CreateAccelerationStructure2KHR,
    .CmdSetDispatchParametersARM = anv_hitman3_CmdSetDispatchParametersARM,
};
const struct vk_device_entrypoint_table anv_android_device_entrypoints = {
    .GetDeviceProcAddr = anv_android_GetDeviceProcAddr,
    .DestroyDevice = anv_android_DestroyDevice,
    .GetDeviceQueue = anv_android_GetDeviceQueue,
    .QueueSubmit = anv_android_QueueSubmit,
    .QueueWaitIdle = anv_android_QueueWaitIdle,
    .DeviceWaitIdle = anv_android_DeviceWaitIdle,
    .AllocateMemory = anv_android_AllocateMemory,
    .FreeMemory = anv_android_FreeMemory,
    .MapMemory = anv_android_MapMemory,
    .UnmapMemory = anv_android_UnmapMemory,
    .FlushMappedMemoryRanges = anv_android_FlushMappedMemoryRanges,
    .InvalidateMappedMemoryRanges = anv_android_InvalidateMappedMemoryRanges,
    .GetDeviceMemoryCommitment = anv_android_GetDeviceMemoryCommitment,
    .GetBufferMemoryRequirements = anv_android_GetBufferMemoryRequirements,
    .BindBufferMemory = anv_android_BindBufferMemory,
    .GetImageMemoryRequirements = anv_android_GetImageMemoryRequirements,
    .BindImageMemory = anv_android_BindImageMemory,
    .GetImageSparseMemoryRequirements = anv_android_GetImageSparseMemoryRequirements,
    .QueueBindSparse = anv_android_QueueBindSparse,
    .CreateFence = anv_android_CreateFence,
    .DestroyFence = anv_android_DestroyFence,
    .ResetFences = anv_android_ResetFences,
    .GetFenceStatus = anv_android_GetFenceStatus,
    .WaitForFences = anv_android_WaitForFences,
    .CreateSemaphore = anv_android_CreateSemaphore,
    .DestroySemaphore = anv_android_DestroySemaphore,
    .CreateEvent = anv_android_CreateEvent,
    .DestroyEvent = anv_android_DestroyEvent,
    .GetEventStatus = anv_android_GetEventStatus,
    .SetEvent = anv_android_SetEvent,
    .ResetEvent = anv_android_ResetEvent,
    .CreateQueryPool = anv_android_CreateQueryPool,
    .DestroyQueryPool = anv_android_DestroyQueryPool,
    .GetQueryPoolResults = anv_android_GetQueryPoolResults,
    .ResetQueryPool = anv_android_ResetQueryPool,
    .ResetQueryPoolEXT = anv_android_ResetQueryPoolEXT,
    .CreateBuffer = anv_android_CreateBuffer,
    .DestroyBuffer = anv_android_DestroyBuffer,
    .CreateBufferView = anv_android_CreateBufferView,
    .DestroyBufferView = anv_android_DestroyBufferView,
    .CreateImage = anv_android_CreateImage,
    .DestroyImage = anv_android_DestroyImage,
    .GetImageSubresourceLayout = anv_android_GetImageSubresourceLayout,
    .CreateImageView = anv_android_CreateImageView,
    .DestroyImageView = anv_android_DestroyImageView,
    .CreateShaderModule = anv_android_CreateShaderModule,
    .DestroyShaderModule = anv_android_DestroyShaderModule,
    .CreatePipelineCache = anv_android_CreatePipelineCache,
    .DestroyPipelineCache = anv_android_DestroyPipelineCache,
    .GetPipelineCacheData = anv_android_GetPipelineCacheData,
    .MergePipelineCaches = anv_android_MergePipelineCaches,
    .CreatePipelineBinariesKHR = anv_android_CreatePipelineBinariesKHR,
    .DestroyPipelineBinaryKHR = anv_android_DestroyPipelineBinaryKHR,
    .GetPipelineKeyKHR = anv_android_GetPipelineKeyKHR,
    .GetPipelineBinaryDataKHR = anv_android_GetPipelineBinaryDataKHR,
    .ReleaseCapturedPipelineDataKHR = anv_android_ReleaseCapturedPipelineDataKHR,
    .CreateGraphicsPipelines = anv_android_CreateGraphicsPipelines,
    .CreateComputePipelines = anv_android_CreateComputePipelines,
    .GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI = anv_android_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI,
    .DestroyPipeline = anv_android_DestroyPipeline,
    .CreatePipelineLayout = anv_android_CreatePipelineLayout,
    .DestroyPipelineLayout = anv_android_DestroyPipelineLayout,
    .CreateSampler = anv_android_CreateSampler,
    .DestroySampler = anv_android_DestroySampler,
    .CreateDescriptorSetLayout = anv_android_CreateDescriptorSetLayout,
    .DestroyDescriptorSetLayout = anv_android_DestroyDescriptorSetLayout,
    .CreateDescriptorPool = anv_android_CreateDescriptorPool,
    .DestroyDescriptorPool = anv_android_DestroyDescriptorPool,
    .ResetDescriptorPool = anv_android_ResetDescriptorPool,
    .AllocateDescriptorSets = anv_android_AllocateDescriptorSets,
    .FreeDescriptorSets = anv_android_FreeDescriptorSets,
    .UpdateDescriptorSets = anv_android_UpdateDescriptorSets,
    .CreateFramebuffer = anv_android_CreateFramebuffer,
    .DestroyFramebuffer = anv_android_DestroyFramebuffer,
    .CreateRenderPass = anv_android_CreateRenderPass,
    .DestroyRenderPass = anv_android_DestroyRenderPass,
    .GetRenderAreaGranularity = anv_android_GetRenderAreaGranularity,
    .GetRenderingAreaGranularity = anv_android_GetRenderingAreaGranularity,
    .GetRenderingAreaGranularityKHR = anv_android_GetRenderingAreaGranularityKHR,
    .CreateCommandPool = anv_android_CreateCommandPool,
    .DestroyCommandPool = anv_android_DestroyCommandPool,
    .ResetCommandPool = anv_android_ResetCommandPool,
    .AllocateCommandBuffers = anv_android_AllocateCommandBuffers,
    .FreeCommandBuffers = anv_android_FreeCommandBuffers,
    .BeginCommandBuffer = anv_android_BeginCommandBuffer,
    .EndCommandBuffer = anv_android_EndCommandBuffer,
    .ResetCommandBuffer = anv_android_ResetCommandBuffer,
    .CmdBindPipeline = anv_android_CmdBindPipeline,
    .CmdSetPrimitiveRestartIndexEXT = anv_android_CmdSetPrimitiveRestartIndexEXT,
    .CmdSetAttachmentFeedbackLoopEnableEXT = anv_android_CmdSetAttachmentFeedbackLoopEnableEXT,
    .CmdSetViewport = anv_android_CmdSetViewport,
    .CmdSetScissor = anv_android_CmdSetScissor,
    .CmdSetLineWidth = anv_android_CmdSetLineWidth,
    .CmdSetDepthBias = anv_android_CmdSetDepthBias,
    .CmdSetBlendConstants = anv_android_CmdSetBlendConstants,
    .CmdSetDepthBounds = anv_android_CmdSetDepthBounds,
    .CmdSetStencilCompareMask = anv_android_CmdSetStencilCompareMask,
    .CmdSetStencilWriteMask = anv_android_CmdSetStencilWriteMask,
    .CmdSetStencilReference = anv_android_CmdSetStencilReference,
    .CmdBindDescriptorSets = anv_android_CmdBindDescriptorSets,
    .CmdBindIndexBuffer = anv_android_CmdBindIndexBuffer,
    .CmdBindVertexBuffers = anv_android_CmdBindVertexBuffers,
    .CmdDraw = anv_android_CmdDraw,
    .CmdDrawIndexed = anv_android_CmdDrawIndexed,
    .CmdDrawMultiEXT = anv_android_CmdDrawMultiEXT,
    .CmdDrawMultiIndexedEXT = anv_android_CmdDrawMultiIndexedEXT,
    .CmdDrawIndirect = anv_android_CmdDrawIndirect,
    .CmdDrawIndexedIndirect = anv_android_CmdDrawIndexedIndirect,
    .CmdDispatch = anv_android_CmdDispatch,
    .CmdDispatchIndirect = anv_android_CmdDispatchIndirect,
    .CmdSubpassShadingHUAWEI = anv_android_CmdSubpassShadingHUAWEI,
    .CmdDrawClusterHUAWEI = anv_android_CmdDrawClusterHUAWEI,
    .CmdDrawClusterIndirectHUAWEI = anv_android_CmdDrawClusterIndirectHUAWEI,
    .CmdUpdatePipelineIndirectBufferNV = anv_android_CmdUpdatePipelineIndirectBufferNV,
    .CmdCopyBuffer = anv_android_CmdCopyBuffer,
    .CmdCopyImage = anv_android_CmdCopyImage,
    .CmdBlitImage = anv_android_CmdBlitImage,
    .CmdCopyBufferToImage = anv_android_CmdCopyBufferToImage,
    .CmdCopyImageToBuffer = anv_android_CmdCopyImageToBuffer,
    .CmdCopyMemoryIndirectNV = anv_android_CmdCopyMemoryIndirectNV,
    .CmdCopyMemoryIndirectKHR = anv_android_CmdCopyMemoryIndirectKHR,
    .CmdCopyMemoryToImageIndirectNV = anv_android_CmdCopyMemoryToImageIndirectNV,
    .CmdCopyMemoryToImageIndirectKHR = anv_android_CmdCopyMemoryToImageIndirectKHR,
    .CmdUpdateBuffer = anv_android_CmdUpdateBuffer,
    .CmdFillBuffer = anv_android_CmdFillBuffer,
    .CmdClearColorImage = anv_android_CmdClearColorImage,
    .CmdClearDepthStencilImage = anv_android_CmdClearDepthStencilImage,
    .CmdClearAttachments = anv_android_CmdClearAttachments,
    .CmdResolveImage = anv_android_CmdResolveImage,
    .CmdSetEvent = anv_android_CmdSetEvent,
    .CmdResetEvent = anv_android_CmdResetEvent,
    .CmdWaitEvents = anv_android_CmdWaitEvents,
    .CmdPipelineBarrier = anv_android_CmdPipelineBarrier,
    .CmdBeginQuery = anv_android_CmdBeginQuery,
    .CmdEndQuery = anv_android_CmdEndQuery,
    .CmdBeginConditionalRenderingEXT = anv_android_CmdBeginConditionalRenderingEXT,
    .CmdEndConditionalRenderingEXT = anv_android_CmdEndConditionalRenderingEXT,
    .CmdBeginCustomResolveEXT = anv_android_CmdBeginCustomResolveEXT,
    .CmdResetQueryPool = anv_android_CmdResetQueryPool,
    .CmdWriteTimestamp = anv_android_CmdWriteTimestamp,
    .CmdCopyQueryPoolResults = anv_android_CmdCopyQueryPoolResults,
    .CmdPushConstants = anv_android_CmdPushConstants,
    .CmdBeginRenderPass = anv_android_CmdBeginRenderPass,
    .CmdNextSubpass = anv_android_CmdNextSubpass,
    .CmdEndRenderPass = anv_android_CmdEndRenderPass,
    .CmdExecuteCommands = anv_android_CmdExecuteCommands,
    .CreateSharedSwapchainsKHR = anv_android_CreateSharedSwapchainsKHR,
    .CreateSwapchainKHR = anv_android_CreateSwapchainKHR,
    .DestroySwapchainKHR = anv_android_DestroySwapchainKHR,
    .GetSwapchainImagesKHR = anv_android_GetSwapchainImagesKHR,
    .AcquireNextImageKHR = anv_android_AcquireNextImageKHR,
    .QueuePresentKHR = anv_android_QueuePresentKHR,
    .DebugMarkerSetObjectNameEXT = anv_android_DebugMarkerSetObjectNameEXT,
    .DebugMarkerSetObjectTagEXT = anv_android_DebugMarkerSetObjectTagEXT,
    .CmdDebugMarkerBeginEXT = anv_android_CmdDebugMarkerBeginEXT,
    .CmdDebugMarkerEndEXT = anv_android_CmdDebugMarkerEndEXT,
    .CmdDebugMarkerInsertEXT = anv_android_CmdDebugMarkerInsertEXT,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleNV = anv_android_GetMemoryWin32HandleNV,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .CmdExecuteGeneratedCommandsNV = anv_android_CmdExecuteGeneratedCommandsNV,
    .CmdPreprocessGeneratedCommandsNV = anv_android_CmdPreprocessGeneratedCommandsNV,
    .CmdBindPipelineShaderGroupNV = anv_android_CmdBindPipelineShaderGroupNV,
    .GetGeneratedCommandsMemoryRequirementsNV = anv_android_GetGeneratedCommandsMemoryRequirementsNV,
    .CreateIndirectCommandsLayoutNV = anv_android_CreateIndirectCommandsLayoutNV,
    .DestroyIndirectCommandsLayoutNV = anv_android_DestroyIndirectCommandsLayoutNV,
    .CmdExecuteGeneratedCommandsEXT = anv_android_CmdExecuteGeneratedCommandsEXT,
    .CmdPreprocessGeneratedCommandsEXT = anv_android_CmdPreprocessGeneratedCommandsEXT,
    .GetGeneratedCommandsMemoryRequirementsEXT = anv_android_GetGeneratedCommandsMemoryRequirementsEXT,
    .CreateIndirectCommandsLayoutEXT = anv_android_CreateIndirectCommandsLayoutEXT,
    .DestroyIndirectCommandsLayoutEXT = anv_android_DestroyIndirectCommandsLayoutEXT,
    .CreateIndirectExecutionSetEXT = anv_android_CreateIndirectExecutionSetEXT,
    .DestroyIndirectExecutionSetEXT = anv_android_DestroyIndirectExecutionSetEXT,
    .UpdateIndirectExecutionSetPipelineEXT = anv_android_UpdateIndirectExecutionSetPipelineEXT,
    .UpdateIndirectExecutionSetShaderEXT = anv_android_UpdateIndirectExecutionSetShaderEXT,
    .CmdPushDescriptorSet = anv_android_CmdPushDescriptorSet,
    .CmdPushDescriptorSetKHR = anv_android_CmdPushDescriptorSetKHR,
    .TrimCommandPool = anv_android_TrimCommandPool,
    .TrimCommandPoolKHR = anv_android_TrimCommandPoolKHR,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleKHR = anv_android_GetMemoryWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandlePropertiesKHR = anv_android_GetMemoryWin32HandlePropertiesKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandlePropertiesKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryFdKHR = anv_android_GetMemoryFdKHR,
    .GetMemoryFdPropertiesKHR = anv_android_GetMemoryFdPropertiesKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandleFUCHSIA = anv_android_GetMemoryZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandlePropertiesFUCHSIA = anv_android_GetMemoryZirconHandlePropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandlePropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .GetMemoryRemoteAddressNV = anv_android_GetMemoryRemoteAddressNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreWin32HandleKHR = anv_android_GetSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportSemaphoreWin32HandleKHR = anv_android_ImportSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreFdKHR = anv_android_GetSemaphoreFdKHR,
    .ImportSemaphoreFdKHR = anv_android_ImportSemaphoreFdKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetSemaphoreZirconHandleFUCHSIA = anv_android_GetSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .ImportSemaphoreZirconHandleFUCHSIA = anv_android_ImportSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .ImportSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetFenceWin32HandleKHR = anv_android_GetFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportFenceWin32HandleKHR = anv_android_ImportFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetFenceFdKHR = anv_android_GetFenceFdKHR,
    .ImportFenceFdKHR = anv_android_ImportFenceFdKHR,
    .DisplayPowerControlEXT = anv_android_DisplayPowerControlEXT,
    .RegisterDeviceEventEXT = anv_android_RegisterDeviceEventEXT,
    .RegisterDisplayEventEXT = anv_android_RegisterDisplayEventEXT,
    .GetSwapchainCounterEXT = anv_android_GetSwapchainCounterEXT,
    .GetDeviceGroupPeerMemoryFeatures = anv_android_GetDeviceGroupPeerMemoryFeatures,
    .GetDeviceGroupPeerMemoryFeaturesKHR = anv_android_GetDeviceGroupPeerMemoryFeaturesKHR,
    .BindBufferMemory2 = anv_android_BindBufferMemory2,
    .BindBufferMemory2KHR = anv_android_BindBufferMemory2KHR,
    .BindImageMemory2 = anv_android_BindImageMemory2,
    .BindImageMemory2KHR = anv_android_BindImageMemory2KHR,
    .CmdSetDeviceMask = anv_android_CmdSetDeviceMask,
    .CmdSetDeviceMaskKHR = anv_android_CmdSetDeviceMaskKHR,
    .GetDeviceGroupPresentCapabilitiesKHR = anv_android_GetDeviceGroupPresentCapabilitiesKHR,
    .GetDeviceGroupSurfacePresentModesKHR = anv_android_GetDeviceGroupSurfacePresentModesKHR,
    .AcquireNextImage2KHR = anv_android_AcquireNextImage2KHR,
    .CmdDispatchBase = anv_android_CmdDispatchBase,
    .CmdDispatchBaseKHR = anv_android_CmdDispatchBaseKHR,
    .CreateDescriptorUpdateTemplate = anv_android_CreateDescriptorUpdateTemplate,
    .CreateDescriptorUpdateTemplateKHR = anv_android_CreateDescriptorUpdateTemplateKHR,
    .DestroyDescriptorUpdateTemplate = anv_android_DestroyDescriptorUpdateTemplate,
    .DestroyDescriptorUpdateTemplateKHR = anv_android_DestroyDescriptorUpdateTemplateKHR,
    .UpdateDescriptorSetWithTemplate = anv_android_UpdateDescriptorSetWithTemplate,
    .UpdateDescriptorSetWithTemplateKHR = anv_android_UpdateDescriptorSetWithTemplateKHR,
    .CmdPushDescriptorSetWithTemplate = anv_android_CmdPushDescriptorSetWithTemplate,
    .CmdPushDescriptorSetWithTemplateKHR = anv_android_CmdPushDescriptorSetWithTemplateKHR,
    .SetHdrMetadataEXT = anv_android_SetHdrMetadataEXT,
    .GetSwapchainStatusKHR = anv_android_GetSwapchainStatusKHR,
    .GetRefreshCycleDurationGOOGLE = anv_android_GetRefreshCycleDurationGOOGLE,
    .GetPastPresentationTimingGOOGLE = anv_android_GetPastPresentationTimingGOOGLE,
    .CmdSetViewportWScalingNV = anv_android_CmdSetViewportWScalingNV,
    .CmdSetDiscardRectangleEXT = anv_android_CmdSetDiscardRectangleEXT,
    .CmdSetDiscardRectangleEnableEXT = anv_android_CmdSetDiscardRectangleEnableEXT,
    .CmdSetDiscardRectangleModeEXT = anv_android_CmdSetDiscardRectangleModeEXT,
    .CmdSetSampleLocationsEXT = anv_android_CmdSetSampleLocationsEXT,
    .GetBufferMemoryRequirements2 = anv_android_GetBufferMemoryRequirements2,
    .GetBufferMemoryRequirements2KHR = anv_android_GetBufferMemoryRequirements2KHR,
    .GetImageMemoryRequirements2 = anv_android_GetImageMemoryRequirements2,
    .GetImageMemoryRequirements2KHR = anv_android_GetImageMemoryRequirements2KHR,
    .GetImageSparseMemoryRequirements2 = anv_android_GetImageSparseMemoryRequirements2,
    .GetImageSparseMemoryRequirements2KHR = anv_android_GetImageSparseMemoryRequirements2KHR,
    .GetDeviceBufferMemoryRequirements = anv_android_GetDeviceBufferMemoryRequirements,
    .GetDeviceBufferMemoryRequirementsKHR = anv_android_GetDeviceBufferMemoryRequirementsKHR,
    .GetDeviceImageMemoryRequirements = anv_android_GetDeviceImageMemoryRequirements,
    .GetDeviceImageMemoryRequirementsKHR = anv_android_GetDeviceImageMemoryRequirementsKHR,
    .GetDeviceImageSparseMemoryRequirements = anv_android_GetDeviceImageSparseMemoryRequirements,
    .GetDeviceImageSparseMemoryRequirementsKHR = anv_android_GetDeviceImageSparseMemoryRequirementsKHR,
    .CreateSamplerYcbcrConversion = anv_android_CreateSamplerYcbcrConversion,
    .CreateSamplerYcbcrConversionKHR = anv_android_CreateSamplerYcbcrConversionKHR,
    .DestroySamplerYcbcrConversion = anv_android_DestroySamplerYcbcrConversion,
    .DestroySamplerYcbcrConversionKHR = anv_android_DestroySamplerYcbcrConversionKHR,
    .GetDeviceQueue2 = anv_android_GetDeviceQueue2,
    .CreateValidationCacheEXT = anv_android_CreateValidationCacheEXT,
    .DestroyValidationCacheEXT = anv_android_DestroyValidationCacheEXT,
    .GetValidationCacheDataEXT = anv_android_GetValidationCacheDataEXT,
    .MergeValidationCachesEXT = anv_android_MergeValidationCachesEXT,
    .GetDescriptorSetLayoutSupport = anv_android_GetDescriptorSetLayoutSupport,
    .GetDescriptorSetLayoutSupportKHR = anv_android_GetDescriptorSetLayoutSupportKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsageANDROID = anv_android_GetSwapchainGrallocUsageANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsage2ANDROID = anv_android_GetSwapchainGrallocUsage2ANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsage2ANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .AcquireImageANDROID = anv_android_AcquireImageANDROID,
#elif defined(_MSC_VER)
    .AcquireImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .QueueSignalReleaseImageANDROID = anv_android_QueueSignalReleaseImageANDROID,
#elif defined(_MSC_VER)
    .QueueSignalReleaseImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .GetShaderInfoAMD = anv_android_GetShaderInfoAMD,
    .SetLocalDimmingAMD = anv_android_SetLocalDimmingAMD,
    .GetCalibratedTimestampsKHR = anv_android_GetCalibratedTimestampsKHR,
    .GetCalibratedTimestampsEXT = anv_android_GetCalibratedTimestampsEXT,
    .SetDebugUtilsObjectNameEXT = anv_android_SetDebugUtilsObjectNameEXT,
    .SetDebugUtilsObjectTagEXT = anv_android_SetDebugUtilsObjectTagEXT,
    .QueueBeginDebugUtilsLabelEXT = anv_android_QueueBeginDebugUtilsLabelEXT,
    .QueueEndDebugUtilsLabelEXT = anv_android_QueueEndDebugUtilsLabelEXT,
    .QueueInsertDebugUtilsLabelEXT = anv_android_QueueInsertDebugUtilsLabelEXT,
    .CmdBeginDebugUtilsLabelEXT = anv_android_CmdBeginDebugUtilsLabelEXT,
    .CmdEndDebugUtilsLabelEXT = anv_android_CmdEndDebugUtilsLabelEXT,
    .CmdInsertDebugUtilsLabelEXT = anv_android_CmdInsertDebugUtilsLabelEXT,
    .GetMemoryHostPointerPropertiesEXT = anv_android_GetMemoryHostPointerPropertiesEXT,
    .CmdWriteBufferMarkerAMD = anv_android_CmdWriteBufferMarkerAMD,
    .CreateRenderPass2 = anv_android_CreateRenderPass2,
    .CreateRenderPass2KHR = anv_android_CreateRenderPass2KHR,
    .CmdBeginRenderPass2 = anv_android_CmdBeginRenderPass2,
    .CmdBeginRenderPass2KHR = anv_android_CmdBeginRenderPass2KHR,
    .CmdNextSubpass2 = anv_android_CmdNextSubpass2,
    .CmdNextSubpass2KHR = anv_android_CmdNextSubpass2KHR,
    .CmdEndRenderPass2 = anv_android_CmdEndRenderPass2,
    .CmdEndRenderPass2KHR = anv_android_CmdEndRenderPass2KHR,
    .GetSemaphoreCounterValue = anv_android_GetSemaphoreCounterValue,
    .GetSemaphoreCounterValueKHR = anv_android_GetSemaphoreCounterValueKHR,
    .WaitSemaphores = anv_android_WaitSemaphores,
    .WaitSemaphoresKHR = anv_android_WaitSemaphoresKHR,
    .SignalSemaphore = anv_android_SignalSemaphore,
    .SignalSemaphoreKHR = anv_android_SignalSemaphoreKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetAndroidHardwareBufferPropertiesANDROID = anv_android_GetAndroidHardwareBufferPropertiesANDROID,
#elif defined(_MSC_VER)
    .GetAndroidHardwareBufferPropertiesANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetMemoryAndroidHardwareBufferANDROID = anv_android_GetMemoryAndroidHardwareBufferANDROID,
#elif defined(_MSC_VER)
    .GetMemoryAndroidHardwareBufferANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .CmdDrawIndirectCount = anv_android_CmdDrawIndirectCount,
    .CmdDrawIndirectCountKHR = anv_android_CmdDrawIndirectCountKHR,
    .CmdDrawIndirectCountAMD = anv_android_CmdDrawIndirectCountAMD,
    .CmdDrawIndexedIndirectCount = anv_android_CmdDrawIndexedIndirectCount,
    .CmdDrawIndexedIndirectCountKHR = anv_android_CmdDrawIndexedIndirectCountKHR,
    .CmdDrawIndexedIndirectCountAMD = anv_android_CmdDrawIndexedIndirectCountAMD,
    .CmdSetCheckpointNV = anv_android_CmdSetCheckpointNV,
    .GetQueueCheckpointDataNV = anv_android_GetQueueCheckpointDataNV,
    .CmdBindTransformFeedbackBuffersEXT = anv_android_CmdBindTransformFeedbackBuffersEXT,
    .CmdBeginTransformFeedbackEXT = anv_android_CmdBeginTransformFeedbackEXT,
    .CmdEndTransformFeedbackEXT = anv_android_CmdEndTransformFeedbackEXT,
    .CmdBeginQueryIndexedEXT = anv_android_CmdBeginQueryIndexedEXT,
    .CmdEndQueryIndexedEXT = anv_android_CmdEndQueryIndexedEXT,
    .CmdDrawIndirectByteCountEXT = anv_android_CmdDrawIndirectByteCountEXT,
    .CmdSetExclusiveScissorNV = anv_android_CmdSetExclusiveScissorNV,
    .CmdSetExclusiveScissorEnableNV = anv_android_CmdSetExclusiveScissorEnableNV,
    .CmdBindShadingRateImageNV = anv_android_CmdBindShadingRateImageNV,
    .CmdSetViewportShadingRatePaletteNV = anv_android_CmdSetViewportShadingRatePaletteNV,
    .CmdSetCoarseSampleOrderNV = anv_android_CmdSetCoarseSampleOrderNV,
    .CmdDrawMeshTasksNV = anv_android_CmdDrawMeshTasksNV,
    .CmdDrawMeshTasksIndirectNV = anv_android_CmdDrawMeshTasksIndirectNV,
    .CmdDrawMeshTasksIndirectCountNV = anv_android_CmdDrawMeshTasksIndirectCountNV,
    .CmdDrawMeshTasksEXT = anv_android_CmdDrawMeshTasksEXT,
    .CmdDrawMeshTasksIndirectEXT = anv_android_CmdDrawMeshTasksIndirectEXT,
    .CmdDrawMeshTasksIndirectCountEXT = anv_android_CmdDrawMeshTasksIndirectCountEXT,
    .CompileDeferredNV = anv_android_CompileDeferredNV,
    .CreateAccelerationStructureNV = anv_android_CreateAccelerationStructureNV,
    .CmdBindInvocationMaskHUAWEI = anv_android_CmdBindInvocationMaskHUAWEI,
    .DestroyAccelerationStructureKHR = anv_android_DestroyAccelerationStructureKHR,
    .DestroyAccelerationStructureNV = anv_android_DestroyAccelerationStructureNV,
    .GetAccelerationStructureMemoryRequirementsNV = anv_android_GetAccelerationStructureMemoryRequirementsNV,
    .BindAccelerationStructureMemoryNV = anv_android_BindAccelerationStructureMemoryNV,
    .CmdCopyAccelerationStructureNV = anv_android_CmdCopyAccelerationStructureNV,
    .CmdCopyAccelerationStructureKHR = anv_android_CmdCopyAccelerationStructureKHR,
    .CopyAccelerationStructureKHR = anv_android_CopyAccelerationStructureKHR,
    .CmdCopyAccelerationStructureToMemoryKHR = anv_android_CmdCopyAccelerationStructureToMemoryKHR,
    .CopyAccelerationStructureToMemoryKHR = anv_android_CopyAccelerationStructureToMemoryKHR,
    .CmdCopyMemoryToAccelerationStructureKHR = anv_android_CmdCopyMemoryToAccelerationStructureKHR,
    .CopyMemoryToAccelerationStructureKHR = anv_android_CopyMemoryToAccelerationStructureKHR,
    .CmdWriteAccelerationStructuresPropertiesKHR = anv_android_CmdWriteAccelerationStructuresPropertiesKHR,
    .CmdWriteAccelerationStructuresPropertiesNV = anv_android_CmdWriteAccelerationStructuresPropertiesNV,
    .CmdBuildAccelerationStructureNV = anv_android_CmdBuildAccelerationStructureNV,
    .WriteAccelerationStructuresPropertiesKHR = anv_android_WriteAccelerationStructuresPropertiesKHR,
    .CmdTraceRaysKHR = anv_android_CmdTraceRaysKHR,
    .CmdTraceRaysNV = anv_android_CmdTraceRaysNV,
    .GetRayTracingShaderGroupHandlesKHR = anv_android_GetRayTracingShaderGroupHandlesKHR,
    .GetRayTracingShaderGroupHandlesNV = anv_android_GetRayTracingShaderGroupHandlesNV,
    .GetRayTracingCaptureReplayShaderGroupHandlesKHR = anv_android_GetRayTracingCaptureReplayShaderGroupHandlesKHR,
    .GetAccelerationStructureHandleNV = anv_android_GetAccelerationStructureHandleNV,
    .CreateRayTracingPipelinesNV = anv_android_CreateRayTracingPipelinesNV,
    .CreateRayTracingPipelinesKHR = anv_android_CreateRayTracingPipelinesKHR,
    .CmdTraceRaysIndirectKHR = anv_android_CmdTraceRaysIndirectKHR,
    .CmdTraceRaysIndirect2KHR = anv_android_CmdTraceRaysIndirect2KHR,
    .GetClusterAccelerationStructureBuildSizesNV = anv_android_GetClusterAccelerationStructureBuildSizesNV,
    .CmdBuildClusterAccelerationStructureIndirectNV = anv_android_CmdBuildClusterAccelerationStructureIndirectNV,
    .GetDeviceAccelerationStructureCompatibilityKHR = anv_android_GetDeviceAccelerationStructureCompatibilityKHR,
    .GetRayTracingShaderGroupStackSizeKHR = anv_android_GetRayTracingShaderGroupStackSizeKHR,
    .CmdSetRayTracingPipelineStackSizeKHR = anv_android_CmdSetRayTracingPipelineStackSizeKHR,
    .GetImageViewHandleNVX = anv_android_GetImageViewHandleNVX,
    .GetImageViewHandle64NVX = anv_android_GetImageViewHandle64NVX,
    .GetImageViewAddressNVX = anv_android_GetImageViewAddressNVX,
    .GetDeviceCombinedImageSamplerIndexNVX = anv_android_GetDeviceCombinedImageSamplerIndexNVX,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetDeviceGroupSurfacePresentModes2EXT = anv_android_GetDeviceGroupSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetDeviceGroupSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireFullScreenExclusiveModeEXT = anv_android_AcquireFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .AcquireFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ReleaseFullScreenExclusiveModeEXT = anv_android_ReleaseFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .ReleaseFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .AcquireProfilingLockKHR = anv_android_AcquireProfilingLockKHR,
    .ReleaseProfilingLockKHR = anv_android_ReleaseProfilingLockKHR,
    .GetImageDrmFormatModifierPropertiesEXT = anv_android_GetImageDrmFormatModifierPropertiesEXT,
    .GetBufferOpaqueCaptureAddress = anv_android_GetBufferOpaqueCaptureAddress,
    .GetBufferOpaqueCaptureAddressKHR = anv_android_GetBufferOpaqueCaptureAddressKHR,
    .GetBufferDeviceAddress = anv_android_GetBufferDeviceAddress,
    .GetBufferDeviceAddressKHR = anv_android_GetBufferDeviceAddressKHR,
    .GetBufferDeviceAddressEXT = anv_android_GetBufferDeviceAddressEXT,
    .InitializePerformanceApiINTEL = anv_android_InitializePerformanceApiINTEL,
    .UninitializePerformanceApiINTEL = anv_android_UninitializePerformanceApiINTEL,
    .CmdSetPerformanceMarkerINTEL = anv_android_CmdSetPerformanceMarkerINTEL,
    .CmdSetPerformanceStreamMarkerINTEL = anv_android_CmdSetPerformanceStreamMarkerINTEL,
    .CmdSetPerformanceOverrideINTEL = anv_android_CmdSetPerformanceOverrideINTEL,
    .AcquirePerformanceConfigurationINTEL = anv_android_AcquirePerformanceConfigurationINTEL,
    .ReleasePerformanceConfigurationINTEL = anv_android_ReleasePerformanceConfigurationINTEL,
    .QueueSetPerformanceConfigurationINTEL = anv_android_QueueSetPerformanceConfigurationINTEL,
    .GetPerformanceParameterINTEL = anv_android_GetPerformanceParameterINTEL,
    .GetDeviceMemoryOpaqueCaptureAddress = anv_android_GetDeviceMemoryOpaqueCaptureAddress,
    .GetDeviceMemoryOpaqueCaptureAddressKHR = anv_android_GetDeviceMemoryOpaqueCaptureAddressKHR,
    .GetPipelineExecutablePropertiesKHR = anv_android_GetPipelineExecutablePropertiesKHR,
    .GetPipelineExecutableStatisticsKHR = anv_android_GetPipelineExecutableStatisticsKHR,
    .GetPipelineExecutableInternalRepresentationsKHR = anv_android_GetPipelineExecutableInternalRepresentationsKHR,
    .CmdSetLineStipple = anv_android_CmdSetLineStipple,
    .CmdSetLineStippleKHR = anv_android_CmdSetLineStippleKHR,
    .CmdSetLineStippleEXT = anv_android_CmdSetLineStippleEXT,
    .CreateAccelerationStructureKHR = anv_android_CreateAccelerationStructureKHR,
    .CmdBuildAccelerationStructuresKHR = anv_android_CmdBuildAccelerationStructuresKHR,
    .CmdBuildAccelerationStructuresIndirectKHR = anv_android_CmdBuildAccelerationStructuresIndirectKHR,
    .BuildAccelerationStructuresKHR = anv_android_BuildAccelerationStructuresKHR,
    .GetAccelerationStructureDeviceAddressKHR = anv_android_GetAccelerationStructureDeviceAddressKHR,
    .CreateDeferredOperationKHR = anv_android_CreateDeferredOperationKHR,
    .DestroyDeferredOperationKHR = anv_android_DestroyDeferredOperationKHR,
    .GetDeferredOperationMaxConcurrencyKHR = anv_android_GetDeferredOperationMaxConcurrencyKHR,
    .GetDeferredOperationResultKHR = anv_android_GetDeferredOperationResultKHR,
    .DeferredOperationJoinKHR = anv_android_DeferredOperationJoinKHR,
    .GetPipelineIndirectMemoryRequirementsNV = anv_android_GetPipelineIndirectMemoryRequirementsNV,
    .GetPipelineIndirectDeviceAddressNV = anv_android_GetPipelineIndirectDeviceAddressNV,
    .AntiLagUpdateAMD = anv_android_AntiLagUpdateAMD,
    .CmdSetCullMode = anv_android_CmdSetCullMode,
    .CmdSetCullModeEXT = anv_android_CmdSetCullModeEXT,
    .CmdSetFrontFace = anv_android_CmdSetFrontFace,
    .CmdSetFrontFaceEXT = anv_android_CmdSetFrontFaceEXT,
    .CmdSetPrimitiveTopology = anv_android_CmdSetPrimitiveTopology,
    .CmdSetPrimitiveTopologyEXT = anv_android_CmdSetPrimitiveTopologyEXT,
    .CmdSetViewportWithCount = anv_android_CmdSetViewportWithCount,
    .CmdSetViewportWithCountEXT = anv_android_CmdSetViewportWithCountEXT,
    .CmdSetScissorWithCount = anv_android_CmdSetScissorWithCount,
    .CmdSetScissorWithCountEXT = anv_android_CmdSetScissorWithCountEXT,
    .CmdBindIndexBuffer2 = anv_android_CmdBindIndexBuffer2,
    .CmdBindIndexBuffer2KHR = anv_android_CmdBindIndexBuffer2KHR,
    .CmdBindVertexBuffers2 = anv_android_CmdBindVertexBuffers2,
    .CmdBindVertexBuffers2EXT = anv_android_CmdBindVertexBuffers2EXT,
    .CmdSetDepthTestEnable = anv_android_CmdSetDepthTestEnable,
    .CmdSetDepthTestEnableEXT = anv_android_CmdSetDepthTestEnableEXT,
    .CmdSetDepthWriteEnable = anv_android_CmdSetDepthWriteEnable,
    .CmdSetDepthWriteEnableEXT = anv_android_CmdSetDepthWriteEnableEXT,
    .CmdSetDepthCompareOp = anv_android_CmdSetDepthCompareOp,
    .CmdSetDepthCompareOpEXT = anv_android_CmdSetDepthCompareOpEXT,
    .CmdSetDepthBoundsTestEnable = anv_android_CmdSetDepthBoundsTestEnable,
    .CmdSetDepthBoundsTestEnableEXT = anv_android_CmdSetDepthBoundsTestEnableEXT,
    .CmdSetStencilTestEnable = anv_android_CmdSetStencilTestEnable,
    .CmdSetStencilTestEnableEXT = anv_android_CmdSetStencilTestEnableEXT,
    .CmdSetStencilOp = anv_android_CmdSetStencilOp,
    .CmdSetStencilOpEXT = anv_android_CmdSetStencilOpEXT,
    .CmdSetPatchControlPointsEXT = anv_android_CmdSetPatchControlPointsEXT,
    .CmdSetRasterizerDiscardEnable = anv_android_CmdSetRasterizerDiscardEnable,
    .CmdSetRasterizerDiscardEnableEXT = anv_android_CmdSetRasterizerDiscardEnableEXT,
    .CmdSetDepthBiasEnable = anv_android_CmdSetDepthBiasEnable,
    .CmdSetDepthBiasEnableEXT = anv_android_CmdSetDepthBiasEnableEXT,
    .CmdSetLogicOpEXT = anv_android_CmdSetLogicOpEXT,
    .CmdSetPrimitiveRestartEnable = anv_android_CmdSetPrimitiveRestartEnable,
    .CmdSetPrimitiveRestartEnableEXT = anv_android_CmdSetPrimitiveRestartEnableEXT,
    .CmdSetTessellationDomainOriginEXT = anv_android_CmdSetTessellationDomainOriginEXT,
    .CmdSetDepthClampEnableEXT = anv_android_CmdSetDepthClampEnableEXT,
    .CmdSetPolygonModeEXT = anv_android_CmdSetPolygonModeEXT,
    .CmdSetRasterizationSamplesEXT = anv_android_CmdSetRasterizationSamplesEXT,
    .CmdSetSampleMaskEXT = anv_android_CmdSetSampleMaskEXT,
    .CmdSetAlphaToCoverageEnableEXT = anv_android_CmdSetAlphaToCoverageEnableEXT,
    .CmdSetAlphaToOneEnableEXT = anv_android_CmdSetAlphaToOneEnableEXT,
    .CmdSetLogicOpEnableEXT = anv_android_CmdSetLogicOpEnableEXT,
    .CmdSetColorBlendEnableEXT = anv_android_CmdSetColorBlendEnableEXT,
    .CmdSetColorBlendEquationEXT = anv_android_CmdSetColorBlendEquationEXT,
    .CmdSetColorWriteMaskEXT = anv_android_CmdSetColorWriteMaskEXT,
    .CmdSetRasterizationStreamEXT = anv_android_CmdSetRasterizationStreamEXT,
    .CmdSetConservativeRasterizationModeEXT = anv_android_CmdSetConservativeRasterizationModeEXT,
    .CmdSetExtraPrimitiveOverestimationSizeEXT = anv_android_CmdSetExtraPrimitiveOverestimationSizeEXT,
    .CmdSetDepthClipEnableEXT = anv_android_CmdSetDepthClipEnableEXT,
    .CmdSetSampleLocationsEnableEXT = anv_android_CmdSetSampleLocationsEnableEXT,
    .CmdSetColorBlendAdvancedEXT = anv_android_CmdSetColorBlendAdvancedEXT,
    .CmdSetProvokingVertexModeEXT = anv_android_CmdSetProvokingVertexModeEXT,
    .CmdSetLineRasterizationModeEXT = anv_android_CmdSetLineRasterizationModeEXT,
    .CmdSetLineStippleEnableEXT = anv_android_CmdSetLineStippleEnableEXT,
    .CmdSetDepthClipNegativeOneToOneEXT = anv_android_CmdSetDepthClipNegativeOneToOneEXT,
    .CmdSetViewportWScalingEnableNV = anv_android_CmdSetViewportWScalingEnableNV,
    .CmdSetViewportSwizzleNV = anv_android_CmdSetViewportSwizzleNV,
    .CmdSetCoverageToColorEnableNV = anv_android_CmdSetCoverageToColorEnableNV,
    .CmdSetCoverageToColorLocationNV = anv_android_CmdSetCoverageToColorLocationNV,
    .CmdSetCoverageModulationModeNV = anv_android_CmdSetCoverageModulationModeNV,
    .CmdSetCoverageModulationTableEnableNV = anv_android_CmdSetCoverageModulationTableEnableNV,
    .CmdSetCoverageModulationTableNV = anv_android_CmdSetCoverageModulationTableNV,
    .CmdSetShadingRateImageEnableNV = anv_android_CmdSetShadingRateImageEnableNV,
    .CmdSetCoverageReductionModeNV = anv_android_CmdSetCoverageReductionModeNV,
    .CmdSetRepresentativeFragmentTestEnableNV = anv_android_CmdSetRepresentativeFragmentTestEnableNV,
    .CreatePrivateDataSlot = anv_android_CreatePrivateDataSlot,
    .CreatePrivateDataSlotEXT = anv_android_CreatePrivateDataSlotEXT,
    .DestroyPrivateDataSlot = anv_android_DestroyPrivateDataSlot,
    .DestroyPrivateDataSlotEXT = anv_android_DestroyPrivateDataSlotEXT,
    .SetPrivateData = anv_android_SetPrivateData,
    .SetPrivateDataEXT = anv_android_SetPrivateDataEXT,
    .GetPrivateData = anv_android_GetPrivateData,
    .GetPrivateDataEXT = anv_android_GetPrivateDataEXT,
    .CmdCopyBuffer2 = anv_android_CmdCopyBuffer2,
    .CmdCopyBuffer2KHR = anv_android_CmdCopyBuffer2KHR,
    .CmdCopyImage2 = anv_android_CmdCopyImage2,
    .CmdCopyImage2KHR = anv_android_CmdCopyImage2KHR,
    .CmdBlitImage2 = anv_android_CmdBlitImage2,
    .CmdBlitImage2KHR = anv_android_CmdBlitImage2KHR,
    .CmdCopyBufferToImage2 = anv_android_CmdCopyBufferToImage2,
    .CmdCopyBufferToImage2KHR = anv_android_CmdCopyBufferToImage2KHR,
    .CmdCopyImageToBuffer2 = anv_android_CmdCopyImageToBuffer2,
    .CmdCopyImageToBuffer2KHR = anv_android_CmdCopyImageToBuffer2KHR,
    .CmdResolveImage2 = anv_android_CmdResolveImage2,
    .CmdResolveImage2KHR = anv_android_CmdResolveImage2KHR,
    .CmdSetFragmentShadingRateKHR = anv_android_CmdSetFragmentShadingRateKHR,
    .CmdSetFragmentShadingRateEnumNV = anv_android_CmdSetFragmentShadingRateEnumNV,
    .GetAccelerationStructureBuildSizesKHR = anv_android_GetAccelerationStructureBuildSizesKHR,
    .CmdSetVertexInputEXT = anv_android_CmdSetVertexInputEXT,
    .CmdSetColorWriteEnableEXT = anv_android_CmdSetColorWriteEnableEXT,
    .CmdSetEvent2 = anv_android_CmdSetEvent2,
    .CmdSetEvent2KHR = anv_android_CmdSetEvent2KHR,
    .CmdResetEvent2 = anv_android_CmdResetEvent2,
    .CmdResetEvent2KHR = anv_android_CmdResetEvent2KHR,
    .CmdWaitEvents2 = anv_android_CmdWaitEvents2,
    .CmdWaitEvents2KHR = anv_android_CmdWaitEvents2KHR,
    .CmdPipelineBarrier2 = anv_android_CmdPipelineBarrier2,
    .CmdPipelineBarrier2KHR = anv_android_CmdPipelineBarrier2KHR,
    .QueueSubmit2 = anv_android_QueueSubmit2,
    .QueueSubmit2KHR = anv_android_QueueSubmit2KHR,
    .CmdWriteTimestamp2 = anv_android_CmdWriteTimestamp2,
    .CmdWriteTimestamp2KHR = anv_android_CmdWriteTimestamp2KHR,
    .CmdWriteBufferMarker2AMD = anv_android_CmdWriteBufferMarker2AMD,
    .GetQueueCheckpointData2NV = anv_android_GetQueueCheckpointData2NV,
    .CopyMemoryToImage = anv_android_CopyMemoryToImage,
    .CopyMemoryToImageEXT = anv_android_CopyMemoryToImageEXT,
    .CopyImageToMemory = anv_android_CopyImageToMemory,
    .CopyImageToMemoryEXT = anv_android_CopyImageToMemoryEXT,
    .CopyImageToImage = anv_android_CopyImageToImage,
    .CopyImageToImageEXT = anv_android_CopyImageToImageEXT,
    .TransitionImageLayout = anv_android_TransitionImageLayout,
    .TransitionImageLayoutEXT = anv_android_TransitionImageLayoutEXT,
    .CreateVideoSessionKHR = anv_android_CreateVideoSessionKHR,
    .DestroyVideoSessionKHR = anv_android_DestroyVideoSessionKHR,
    .CreateVideoSessionParametersKHR = anv_android_CreateVideoSessionParametersKHR,
    .UpdateVideoSessionParametersKHR = anv_android_UpdateVideoSessionParametersKHR,
    .GetEncodedVideoSessionParametersKHR = anv_android_GetEncodedVideoSessionParametersKHR,
    .DestroyVideoSessionParametersKHR = anv_android_DestroyVideoSessionParametersKHR,
    .GetVideoSessionMemoryRequirementsKHR = anv_android_GetVideoSessionMemoryRequirementsKHR,
    .BindVideoSessionMemoryKHR = anv_android_BindVideoSessionMemoryKHR,
    .CmdDecodeVideoKHR = anv_android_CmdDecodeVideoKHR,
    .CmdBeginVideoCodingKHR = anv_android_CmdBeginVideoCodingKHR,
    .CmdControlVideoCodingKHR = anv_android_CmdControlVideoCodingKHR,
    .CmdEndVideoCodingKHR = anv_android_CmdEndVideoCodingKHR,
    .CmdEncodeVideoKHR = anv_android_CmdEncodeVideoKHR,
    .CmdDecompressMemoryNV = anv_android_CmdDecompressMemoryNV,
    .CmdDecompressMemoryIndirectCountNV = anv_android_CmdDecompressMemoryIndirectCountNV,
    .GetPartitionedAccelerationStructuresBuildSizesNV = anv_android_GetPartitionedAccelerationStructuresBuildSizesNV,
    .CmdBuildPartitionedAccelerationStructuresNV = anv_android_CmdBuildPartitionedAccelerationStructuresNV,
    .CmdDecompressMemoryEXT = anv_android_CmdDecompressMemoryEXT,
    .CmdDecompressMemoryIndirectCountEXT = anv_android_CmdDecompressMemoryIndirectCountEXT,
    .CreateCuModuleNVX = anv_android_CreateCuModuleNVX,
    .CreateCuFunctionNVX = anv_android_CreateCuFunctionNVX,
    .DestroyCuModuleNVX = anv_android_DestroyCuModuleNVX,
    .DestroyCuFunctionNVX = anv_android_DestroyCuFunctionNVX,
    .CmdCuLaunchKernelNVX = anv_android_CmdCuLaunchKernelNVX,
    .GetDescriptorSetLayoutSizeEXT = anv_android_GetDescriptorSetLayoutSizeEXT,
    .GetDescriptorSetLayoutBindingOffsetEXT = anv_android_GetDescriptorSetLayoutBindingOffsetEXT,
    .GetDescriptorEXT = anv_android_GetDescriptorEXT,
    .CmdBindDescriptorBuffersEXT = anv_android_CmdBindDescriptorBuffersEXT,
    .CmdSetDescriptorBufferOffsetsEXT = anv_android_CmdSetDescriptorBufferOffsetsEXT,
    .CmdBindDescriptorBufferEmbeddedSamplersEXT = anv_android_CmdBindDescriptorBufferEmbeddedSamplersEXT,
    .GetBufferOpaqueCaptureDescriptorDataEXT = anv_android_GetBufferOpaqueCaptureDescriptorDataEXT,
    .GetImageOpaqueCaptureDescriptorDataEXT = anv_android_GetImageOpaqueCaptureDescriptorDataEXT,
    .GetImageViewOpaqueCaptureDescriptorDataEXT = anv_android_GetImageViewOpaqueCaptureDescriptorDataEXT,
    .GetSamplerOpaqueCaptureDescriptorDataEXT = anv_android_GetSamplerOpaqueCaptureDescriptorDataEXT,
    .GetAccelerationStructureOpaqueCaptureDescriptorDataEXT = anv_android_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT,
    .SetDeviceMemoryPriorityEXT = anv_android_SetDeviceMemoryPriorityEXT,
    .WaitForPresent2KHR = anv_android_WaitForPresent2KHR,
    .WaitForPresentKHR = anv_android_WaitForPresentKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateBufferCollectionFUCHSIA = anv_android_CreateBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .CreateBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionBufferConstraintsFUCHSIA = anv_android_SetBufferCollectionBufferConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionBufferConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionImageConstraintsFUCHSIA = anv_android_SetBufferCollectionImageConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionImageConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .DestroyBufferCollectionFUCHSIA = anv_android_DestroyBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .DestroyBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetBufferCollectionPropertiesFUCHSIA = anv_android_GetBufferCollectionPropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetBufferCollectionPropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .CmdBeginRendering = anv_android_CmdBeginRendering,
    .CmdBeginRenderingKHR = anv_android_CmdBeginRenderingKHR,
    .CmdEndRendering = anv_android_CmdEndRendering,
    .CmdEndRendering2KHR = anv_android_CmdEndRendering2KHR,
    .CmdEndRendering2EXT = anv_android_CmdEndRendering2EXT,
    .CmdEndRenderingKHR = anv_android_CmdEndRenderingKHR,
    .GetDescriptorSetLayoutHostMappingInfoVALVE = anv_android_GetDescriptorSetLayoutHostMappingInfoVALVE,
    .GetDescriptorSetHostMappingVALVE = anv_android_GetDescriptorSetHostMappingVALVE,
    .CreateMicromapEXT = anv_android_CreateMicromapEXT,
    .CmdBuildMicromapsEXT = anv_android_CmdBuildMicromapsEXT,
    .BuildMicromapsEXT = anv_android_BuildMicromapsEXT,
    .DestroyMicromapEXT = anv_android_DestroyMicromapEXT,
    .CmdCopyMicromapEXT = anv_android_CmdCopyMicromapEXT,
    .CopyMicromapEXT = anv_android_CopyMicromapEXT,
    .CmdCopyMicromapToMemoryEXT = anv_android_CmdCopyMicromapToMemoryEXT,
    .CopyMicromapToMemoryEXT = anv_android_CopyMicromapToMemoryEXT,
    .CmdCopyMemoryToMicromapEXT = anv_android_CmdCopyMemoryToMicromapEXT,
    .CopyMemoryToMicromapEXT = anv_android_CopyMemoryToMicromapEXT,
    .CmdWriteMicromapsPropertiesEXT = anv_android_CmdWriteMicromapsPropertiesEXT,
    .WriteMicromapsPropertiesEXT = anv_android_WriteMicromapsPropertiesEXT,
    .GetDeviceMicromapCompatibilityEXT = anv_android_GetDeviceMicromapCompatibilityEXT,
    .GetMicromapBuildSizesEXT = anv_android_GetMicromapBuildSizesEXT,
    .GetShaderModuleIdentifierEXT = anv_android_GetShaderModuleIdentifierEXT,
    .GetShaderModuleCreateInfoIdentifierEXT = anv_android_GetShaderModuleCreateInfoIdentifierEXT,
    .GetImageSubresourceLayout2 = anv_android_GetImageSubresourceLayout2,
    .GetImageSubresourceLayout2KHR = anv_android_GetImageSubresourceLayout2KHR,
    .GetImageSubresourceLayout2EXT = anv_android_GetImageSubresourceLayout2EXT,
    .GetPipelinePropertiesEXT = anv_android_GetPipelinePropertiesEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .ExportMetalObjectsEXT = anv_android_ExportMetalObjectsEXT,
#elif defined(_MSC_VER)
    .ExportMetalObjectsEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CmdBindTileMemoryQCOM = anv_android_CmdBindTileMemoryQCOM,
    .GetFramebufferTilePropertiesQCOM = anv_android_GetFramebufferTilePropertiesQCOM,
    .GetDynamicRenderingTilePropertiesQCOM = anv_android_GetDynamicRenderingTilePropertiesQCOM,
    .CreateOpticalFlowSessionNV = anv_android_CreateOpticalFlowSessionNV,
    .DestroyOpticalFlowSessionNV = anv_android_DestroyOpticalFlowSessionNV,
    .BindOpticalFlowSessionImageNV = anv_android_BindOpticalFlowSessionImageNV,
    .CmdOpticalFlowExecuteNV = anv_android_CmdOpticalFlowExecuteNV,
    .GetDeviceFaultInfoEXT = anv_android_GetDeviceFaultInfoEXT,
    .GetDeviceFaultReportsKHR = anv_android_GetDeviceFaultReportsKHR,
    .GetDeviceFaultDebugInfoKHR = anv_android_GetDeviceFaultDebugInfoKHR,
    .CmdSetDepthBias2EXT = anv_android_CmdSetDepthBias2EXT,
    .ReleaseSwapchainImagesKHR = anv_android_ReleaseSwapchainImagesKHR,
    .ReleaseSwapchainImagesEXT = anv_android_ReleaseSwapchainImagesEXT,
    .GetDeviceImageSubresourceLayout = anv_android_GetDeviceImageSubresourceLayout,
    .GetDeviceImageSubresourceLayoutKHR = anv_android_GetDeviceImageSubresourceLayoutKHR,
    .MapMemory2 = anv_android_MapMemory2,
    .MapMemory2KHR = anv_android_MapMemory2KHR,
    .UnmapMemory2 = anv_android_UnmapMemory2,
    .UnmapMemory2KHR = anv_android_UnmapMemory2KHR,
    .CreateShadersEXT = anv_android_CreateShadersEXT,
    .DestroyShaderEXT = anv_android_DestroyShaderEXT,
    .GetShaderBinaryDataEXT = anv_android_GetShaderBinaryDataEXT,
    .CmdBindShadersEXT = anv_android_CmdBindShadersEXT,
    .SetSwapchainPresentTimingQueueSizeEXT = anv_android_SetSwapchainPresentTimingQueueSizeEXT,
    .GetSwapchainTimingPropertiesEXT = anv_android_GetSwapchainTimingPropertiesEXT,
    .GetSwapchainTimeDomainPropertiesEXT = anv_android_GetSwapchainTimeDomainPropertiesEXT,
    .GetPastPresentationTimingEXT = anv_android_GetPastPresentationTimingEXT,
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetScreenBufferPropertiesQNX = anv_android_GetScreenBufferPropertiesQNX,
#elif defined(_MSC_VER)
    .GetScreenBufferPropertiesQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CmdBindDescriptorSets2 = anv_android_CmdBindDescriptorSets2,
    .CmdBindDescriptorSets2KHR = anv_android_CmdBindDescriptorSets2KHR,
    .CmdPushConstants2 = anv_android_CmdPushConstants2,
    .CmdPushConstants2KHR = anv_android_CmdPushConstants2KHR,
    .CmdPushDescriptorSet2 = anv_android_CmdPushDescriptorSet2,
    .CmdPushDescriptorSet2KHR = anv_android_CmdPushDescriptorSet2KHR,
    .CmdPushDescriptorSetWithTemplate2 = anv_android_CmdPushDescriptorSetWithTemplate2,
    .CmdPushDescriptorSetWithTemplate2KHR = anv_android_CmdPushDescriptorSetWithTemplate2KHR,
    .CmdSetDescriptorBufferOffsets2EXT = anv_android_CmdSetDescriptorBufferOffsets2EXT,
    .CmdBindDescriptorBufferEmbeddedSamplers2EXT = anv_android_CmdBindDescriptorBufferEmbeddedSamplers2EXT,
    .SetLatencySleepModeNV = anv_android_SetLatencySleepModeNV,
    .LatencySleepNV = anv_android_LatencySleepNV,
    .SetLatencyMarkerNV = anv_android_SetLatencyMarkerNV,
    .GetLatencyTimingsNV = anv_android_GetLatencyTimingsNV,
    .QueueNotifyOutOfBandNV = anv_android_QueueNotifyOutOfBandNV,
    .CmdSetRenderingAttachmentLocations = anv_android_CmdSetRenderingAttachmentLocations,
    .CmdSetRenderingAttachmentLocationsKHR = anv_android_CmdSetRenderingAttachmentLocationsKHR,
    .CmdSetRenderingInputAttachmentIndices = anv_android_CmdSetRenderingInputAttachmentIndices,
    .CmdSetRenderingInputAttachmentIndicesKHR = anv_android_CmdSetRenderingInputAttachmentIndicesKHR,
    .CmdSetDepthClampRangeEXT = anv_android_CmdSetDepthClampRangeEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandleEXT = anv_android_GetMemoryMetalHandleEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandleEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandlePropertiesEXT = anv_android_GetMemoryMetalHandlePropertiesEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandlePropertiesEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .ConvertCooperativeVectorMatrixNV = anv_android_ConvertCooperativeVectorMatrixNV,
    .CmdConvertCooperativeVectorMatrixNV = anv_android_CmdConvertCooperativeVectorMatrixNV,
    .CmdDispatchTileQCOM = anv_android_CmdDispatchTileQCOM,
    .CmdBeginPerTileExecutionQCOM = anv_android_CmdBeginPerTileExecutionQCOM,
    .CmdEndPerTileExecutionQCOM = anv_android_CmdEndPerTileExecutionQCOM,
    .CreateExternalComputeQueueNV = anv_android_CreateExternalComputeQueueNV,
    .DestroyExternalComputeQueueNV = anv_android_DestroyExternalComputeQueueNV,
    .CreateShaderInstrumentationARM = anv_android_CreateShaderInstrumentationARM,
    .DestroyShaderInstrumentationARM = anv_android_DestroyShaderInstrumentationARM,
    .CmdBeginShaderInstrumentationARM = anv_android_CmdBeginShaderInstrumentationARM,
    .CmdEndShaderInstrumentationARM = anv_android_CmdEndShaderInstrumentationARM,
    .GetShaderInstrumentationValuesARM = anv_android_GetShaderInstrumentationValuesARM,
    .ClearShaderInstrumentationMetricsARM = anv_android_ClearShaderInstrumentationMetricsARM,
    .CreateTensorARM = anv_android_CreateTensorARM,
    .DestroyTensorARM = anv_android_DestroyTensorARM,
    .CreateTensorViewARM = anv_android_CreateTensorViewARM,
    .DestroyTensorViewARM = anv_android_DestroyTensorViewARM,
    .GetTensorMemoryRequirementsARM = anv_android_GetTensorMemoryRequirementsARM,
    .BindTensorMemoryARM = anv_android_BindTensorMemoryARM,
    .GetDeviceTensorMemoryRequirementsARM = anv_android_GetDeviceTensorMemoryRequirementsARM,
    .CmdCopyTensorARM = anv_android_CmdCopyTensorARM,
    .GetTensorOpaqueCaptureDescriptorDataARM = anv_android_GetTensorOpaqueCaptureDescriptorDataARM,
    .GetTensorViewOpaqueCaptureDescriptorDataARM = anv_android_GetTensorViewOpaqueCaptureDescriptorDataARM,
    .CreateDataGraphPipelinesARM = anv_android_CreateDataGraphPipelinesARM,
    .CreateDataGraphPipelineSessionARM = anv_android_CreateDataGraphPipelineSessionARM,
    .GetDataGraphPipelineSessionBindPointRequirementsARM = anv_android_GetDataGraphPipelineSessionBindPointRequirementsARM,
    .GetDataGraphPipelineSessionMemoryRequirementsARM = anv_android_GetDataGraphPipelineSessionMemoryRequirementsARM,
    .BindDataGraphPipelineSessionMemoryARM = anv_android_BindDataGraphPipelineSessionMemoryARM,
    .DestroyDataGraphPipelineSessionARM = anv_android_DestroyDataGraphPipelineSessionARM,
    .CmdDispatchDataGraphARM = anv_android_CmdDispatchDataGraphARM,
    .GetDataGraphPipelineAvailablePropertiesARM = anv_android_GetDataGraphPipelineAvailablePropertiesARM,
    .GetDataGraphPipelinePropertiesARM = anv_android_GetDataGraphPipelinePropertiesARM,
#ifdef VK_USE_PLATFORM_OHOS
    .GetNativeBufferPropertiesOHOS = anv_android_GetNativeBufferPropertiesOHOS,
#elif defined(_MSC_VER)
    .GetNativeBufferPropertiesOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
    .GetMemoryNativeBufferOHOS = anv_android_GetMemoryNativeBufferOHOS,
#elif defined(_MSC_VER)
    .GetMemoryNativeBufferOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .QueueSetPerfHintQCOM = anv_android_QueueSetPerfHintQCOM,
    .CmdSetComputeOccupancyPriorityNV = anv_android_CmdSetComputeOccupancyPriorityNV,
    .WriteSamplerDescriptorsEXT = anv_android_WriteSamplerDescriptorsEXT,
    .WriteResourceDescriptorsEXT = anv_android_WriteResourceDescriptorsEXT,
    .CmdBindSamplerHeapEXT = anv_android_CmdBindSamplerHeapEXT,
    .CmdBindResourceHeapEXT = anv_android_CmdBindResourceHeapEXT,
    .CmdPushDataEXT = anv_android_CmdPushDataEXT,
    .RegisterCustomBorderColorEXT = anv_android_RegisterCustomBorderColorEXT,
    .UnregisterCustomBorderColorEXT = anv_android_UnregisterCustomBorderColorEXT,
    .GetImageOpaqueCaptureDataEXT = anv_android_GetImageOpaqueCaptureDataEXT,
    .GetTensorOpaqueCaptureDataARM = anv_android_GetTensorOpaqueCaptureDataARM,
    .CmdCopyMemoryKHR = anv_android_CmdCopyMemoryKHR,
    .CmdCopyMemoryToImageKHR = anv_android_CmdCopyMemoryToImageKHR,
    .CmdCopyImageToMemoryKHR = anv_android_CmdCopyImageToMemoryKHR,
    .CmdUpdateMemoryKHR = anv_android_CmdUpdateMemoryKHR,
    .CmdFillMemoryKHR = anv_android_CmdFillMemoryKHR,
    .CmdCopyQueryPoolResultsToMemoryKHR = anv_android_CmdCopyQueryPoolResultsToMemoryKHR,
    .CmdBeginConditionalRendering2EXT = anv_android_CmdBeginConditionalRendering2EXT,
    .CmdBindTransformFeedbackBuffers2EXT = anv_android_CmdBindTransformFeedbackBuffers2EXT,
    .CmdBeginTransformFeedback2EXT = anv_android_CmdBeginTransformFeedback2EXT,
    .CmdEndTransformFeedback2EXT = anv_android_CmdEndTransformFeedback2EXT,
    .CmdDrawIndirectByteCount2EXT = anv_android_CmdDrawIndirectByteCount2EXT,
    .CmdWriteMarkerToMemoryAMD = anv_android_CmdWriteMarkerToMemoryAMD,
    .CmdBindIndexBuffer3KHR = anv_android_CmdBindIndexBuffer3KHR,
    .CmdBindVertexBuffers3KHR = anv_android_CmdBindVertexBuffers3KHR,
    .CmdDrawIndirect2KHR = anv_android_CmdDrawIndirect2KHR,
    .CmdDrawIndexedIndirect2KHR = anv_android_CmdDrawIndexedIndirect2KHR,
    .CmdDrawIndirectCount2KHR = anv_android_CmdDrawIndirectCount2KHR,
    .CmdDrawIndexedIndirectCount2KHR = anv_android_CmdDrawIndexedIndirectCount2KHR,
    .CmdDrawMeshTasksIndirect2EXT = anv_android_CmdDrawMeshTasksIndirect2EXT,
    .CmdDrawMeshTasksIndirectCount2EXT = anv_android_CmdDrawMeshTasksIndirectCount2EXT,
    .CmdDispatchIndirect2KHR = anv_android_CmdDispatchIndirect2KHR,
    .CreateAccelerationStructure2KHR = anv_android_CreateAccelerationStructure2KHR,
    .CmdSetDispatchParametersARM = anv_android_CmdSetDispatchParametersARM,
};
const struct vk_device_entrypoint_table anv_rmv_device_entrypoints = {
    .GetDeviceProcAddr = anv_rmv_GetDeviceProcAddr,
    .DestroyDevice = anv_rmv_DestroyDevice,
    .GetDeviceQueue = anv_rmv_GetDeviceQueue,
    .QueueSubmit = anv_rmv_QueueSubmit,
    .QueueWaitIdle = anv_rmv_QueueWaitIdle,
    .DeviceWaitIdle = anv_rmv_DeviceWaitIdle,
    .AllocateMemory = anv_rmv_AllocateMemory,
    .FreeMemory = anv_rmv_FreeMemory,
    .MapMemory = anv_rmv_MapMemory,
    .UnmapMemory = anv_rmv_UnmapMemory,
    .FlushMappedMemoryRanges = anv_rmv_FlushMappedMemoryRanges,
    .InvalidateMappedMemoryRanges = anv_rmv_InvalidateMappedMemoryRanges,
    .GetDeviceMemoryCommitment = anv_rmv_GetDeviceMemoryCommitment,
    .GetBufferMemoryRequirements = anv_rmv_GetBufferMemoryRequirements,
    .BindBufferMemory = anv_rmv_BindBufferMemory,
    .GetImageMemoryRequirements = anv_rmv_GetImageMemoryRequirements,
    .BindImageMemory = anv_rmv_BindImageMemory,
    .GetImageSparseMemoryRequirements = anv_rmv_GetImageSparseMemoryRequirements,
    .QueueBindSparse = anv_rmv_QueueBindSparse,
    .CreateFence = anv_rmv_CreateFence,
    .DestroyFence = anv_rmv_DestroyFence,
    .ResetFences = anv_rmv_ResetFences,
    .GetFenceStatus = anv_rmv_GetFenceStatus,
    .WaitForFences = anv_rmv_WaitForFences,
    .CreateSemaphore = anv_rmv_CreateSemaphore,
    .DestroySemaphore = anv_rmv_DestroySemaphore,
    .CreateEvent = anv_rmv_CreateEvent,
    .DestroyEvent = anv_rmv_DestroyEvent,
    .GetEventStatus = anv_rmv_GetEventStatus,
    .SetEvent = anv_rmv_SetEvent,
    .ResetEvent = anv_rmv_ResetEvent,
    .CreateQueryPool = anv_rmv_CreateQueryPool,
    .DestroyQueryPool = anv_rmv_DestroyQueryPool,
    .GetQueryPoolResults = anv_rmv_GetQueryPoolResults,
    .ResetQueryPool = anv_rmv_ResetQueryPool,
    .ResetQueryPoolEXT = anv_rmv_ResetQueryPoolEXT,
    .CreateBuffer = anv_rmv_CreateBuffer,
    .DestroyBuffer = anv_rmv_DestroyBuffer,
    .CreateBufferView = anv_rmv_CreateBufferView,
    .DestroyBufferView = anv_rmv_DestroyBufferView,
    .CreateImage = anv_rmv_CreateImage,
    .DestroyImage = anv_rmv_DestroyImage,
    .GetImageSubresourceLayout = anv_rmv_GetImageSubresourceLayout,
    .CreateImageView = anv_rmv_CreateImageView,
    .DestroyImageView = anv_rmv_DestroyImageView,
    .CreateShaderModule = anv_rmv_CreateShaderModule,
    .DestroyShaderModule = anv_rmv_DestroyShaderModule,
    .CreatePipelineCache = anv_rmv_CreatePipelineCache,
    .DestroyPipelineCache = anv_rmv_DestroyPipelineCache,
    .GetPipelineCacheData = anv_rmv_GetPipelineCacheData,
    .MergePipelineCaches = anv_rmv_MergePipelineCaches,
    .CreatePipelineBinariesKHR = anv_rmv_CreatePipelineBinariesKHR,
    .DestroyPipelineBinaryKHR = anv_rmv_DestroyPipelineBinaryKHR,
    .GetPipelineKeyKHR = anv_rmv_GetPipelineKeyKHR,
    .GetPipelineBinaryDataKHR = anv_rmv_GetPipelineBinaryDataKHR,
    .ReleaseCapturedPipelineDataKHR = anv_rmv_ReleaseCapturedPipelineDataKHR,
    .CreateGraphicsPipelines = anv_rmv_CreateGraphicsPipelines,
    .CreateComputePipelines = anv_rmv_CreateComputePipelines,
    .GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI = anv_rmv_GetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI,
    .DestroyPipeline = anv_rmv_DestroyPipeline,
    .CreatePipelineLayout = anv_rmv_CreatePipelineLayout,
    .DestroyPipelineLayout = anv_rmv_DestroyPipelineLayout,
    .CreateSampler = anv_rmv_CreateSampler,
    .DestroySampler = anv_rmv_DestroySampler,
    .CreateDescriptorSetLayout = anv_rmv_CreateDescriptorSetLayout,
    .DestroyDescriptorSetLayout = anv_rmv_DestroyDescriptorSetLayout,
    .CreateDescriptorPool = anv_rmv_CreateDescriptorPool,
    .DestroyDescriptorPool = anv_rmv_DestroyDescriptorPool,
    .ResetDescriptorPool = anv_rmv_ResetDescriptorPool,
    .AllocateDescriptorSets = anv_rmv_AllocateDescriptorSets,
    .FreeDescriptorSets = anv_rmv_FreeDescriptorSets,
    .UpdateDescriptorSets = anv_rmv_UpdateDescriptorSets,
    .CreateFramebuffer = anv_rmv_CreateFramebuffer,
    .DestroyFramebuffer = anv_rmv_DestroyFramebuffer,
    .CreateRenderPass = anv_rmv_CreateRenderPass,
    .DestroyRenderPass = anv_rmv_DestroyRenderPass,
    .GetRenderAreaGranularity = anv_rmv_GetRenderAreaGranularity,
    .GetRenderingAreaGranularity = anv_rmv_GetRenderingAreaGranularity,
    .GetRenderingAreaGranularityKHR = anv_rmv_GetRenderingAreaGranularityKHR,
    .CreateCommandPool = anv_rmv_CreateCommandPool,
    .DestroyCommandPool = anv_rmv_DestroyCommandPool,
    .ResetCommandPool = anv_rmv_ResetCommandPool,
    .AllocateCommandBuffers = anv_rmv_AllocateCommandBuffers,
    .FreeCommandBuffers = anv_rmv_FreeCommandBuffers,
    .BeginCommandBuffer = anv_rmv_BeginCommandBuffer,
    .EndCommandBuffer = anv_rmv_EndCommandBuffer,
    .ResetCommandBuffer = anv_rmv_ResetCommandBuffer,
    .CmdBindPipeline = anv_rmv_CmdBindPipeline,
    .CmdSetPrimitiveRestartIndexEXT = anv_rmv_CmdSetPrimitiveRestartIndexEXT,
    .CmdSetAttachmentFeedbackLoopEnableEXT = anv_rmv_CmdSetAttachmentFeedbackLoopEnableEXT,
    .CmdSetViewport = anv_rmv_CmdSetViewport,
    .CmdSetScissor = anv_rmv_CmdSetScissor,
    .CmdSetLineWidth = anv_rmv_CmdSetLineWidth,
    .CmdSetDepthBias = anv_rmv_CmdSetDepthBias,
    .CmdSetBlendConstants = anv_rmv_CmdSetBlendConstants,
    .CmdSetDepthBounds = anv_rmv_CmdSetDepthBounds,
    .CmdSetStencilCompareMask = anv_rmv_CmdSetStencilCompareMask,
    .CmdSetStencilWriteMask = anv_rmv_CmdSetStencilWriteMask,
    .CmdSetStencilReference = anv_rmv_CmdSetStencilReference,
    .CmdBindDescriptorSets = anv_rmv_CmdBindDescriptorSets,
    .CmdBindIndexBuffer = anv_rmv_CmdBindIndexBuffer,
    .CmdBindVertexBuffers = anv_rmv_CmdBindVertexBuffers,
    .CmdDraw = anv_rmv_CmdDraw,
    .CmdDrawIndexed = anv_rmv_CmdDrawIndexed,
    .CmdDrawMultiEXT = anv_rmv_CmdDrawMultiEXT,
    .CmdDrawMultiIndexedEXT = anv_rmv_CmdDrawMultiIndexedEXT,
    .CmdDrawIndirect = anv_rmv_CmdDrawIndirect,
    .CmdDrawIndexedIndirect = anv_rmv_CmdDrawIndexedIndirect,
    .CmdDispatch = anv_rmv_CmdDispatch,
    .CmdDispatchIndirect = anv_rmv_CmdDispatchIndirect,
    .CmdSubpassShadingHUAWEI = anv_rmv_CmdSubpassShadingHUAWEI,
    .CmdDrawClusterHUAWEI = anv_rmv_CmdDrawClusterHUAWEI,
    .CmdDrawClusterIndirectHUAWEI = anv_rmv_CmdDrawClusterIndirectHUAWEI,
    .CmdUpdatePipelineIndirectBufferNV = anv_rmv_CmdUpdatePipelineIndirectBufferNV,
    .CmdCopyBuffer = anv_rmv_CmdCopyBuffer,
    .CmdCopyImage = anv_rmv_CmdCopyImage,
    .CmdBlitImage = anv_rmv_CmdBlitImage,
    .CmdCopyBufferToImage = anv_rmv_CmdCopyBufferToImage,
    .CmdCopyImageToBuffer = anv_rmv_CmdCopyImageToBuffer,
    .CmdCopyMemoryIndirectNV = anv_rmv_CmdCopyMemoryIndirectNV,
    .CmdCopyMemoryIndirectKHR = anv_rmv_CmdCopyMemoryIndirectKHR,
    .CmdCopyMemoryToImageIndirectNV = anv_rmv_CmdCopyMemoryToImageIndirectNV,
    .CmdCopyMemoryToImageIndirectKHR = anv_rmv_CmdCopyMemoryToImageIndirectKHR,
    .CmdUpdateBuffer = anv_rmv_CmdUpdateBuffer,
    .CmdFillBuffer = anv_rmv_CmdFillBuffer,
    .CmdClearColorImage = anv_rmv_CmdClearColorImage,
    .CmdClearDepthStencilImage = anv_rmv_CmdClearDepthStencilImage,
    .CmdClearAttachments = anv_rmv_CmdClearAttachments,
    .CmdResolveImage = anv_rmv_CmdResolveImage,
    .CmdSetEvent = anv_rmv_CmdSetEvent,
    .CmdResetEvent = anv_rmv_CmdResetEvent,
    .CmdWaitEvents = anv_rmv_CmdWaitEvents,
    .CmdPipelineBarrier = anv_rmv_CmdPipelineBarrier,
    .CmdBeginQuery = anv_rmv_CmdBeginQuery,
    .CmdEndQuery = anv_rmv_CmdEndQuery,
    .CmdBeginConditionalRenderingEXT = anv_rmv_CmdBeginConditionalRenderingEXT,
    .CmdEndConditionalRenderingEXT = anv_rmv_CmdEndConditionalRenderingEXT,
    .CmdBeginCustomResolveEXT = anv_rmv_CmdBeginCustomResolveEXT,
    .CmdResetQueryPool = anv_rmv_CmdResetQueryPool,
    .CmdWriteTimestamp = anv_rmv_CmdWriteTimestamp,
    .CmdCopyQueryPoolResults = anv_rmv_CmdCopyQueryPoolResults,
    .CmdPushConstants = anv_rmv_CmdPushConstants,
    .CmdBeginRenderPass = anv_rmv_CmdBeginRenderPass,
    .CmdNextSubpass = anv_rmv_CmdNextSubpass,
    .CmdEndRenderPass = anv_rmv_CmdEndRenderPass,
    .CmdExecuteCommands = anv_rmv_CmdExecuteCommands,
    .CreateSharedSwapchainsKHR = anv_rmv_CreateSharedSwapchainsKHR,
    .CreateSwapchainKHR = anv_rmv_CreateSwapchainKHR,
    .DestroySwapchainKHR = anv_rmv_DestroySwapchainKHR,
    .GetSwapchainImagesKHR = anv_rmv_GetSwapchainImagesKHR,
    .AcquireNextImageKHR = anv_rmv_AcquireNextImageKHR,
    .QueuePresentKHR = anv_rmv_QueuePresentKHR,
    .DebugMarkerSetObjectNameEXT = anv_rmv_DebugMarkerSetObjectNameEXT,
    .DebugMarkerSetObjectTagEXT = anv_rmv_DebugMarkerSetObjectTagEXT,
    .CmdDebugMarkerBeginEXT = anv_rmv_CmdDebugMarkerBeginEXT,
    .CmdDebugMarkerEndEXT = anv_rmv_CmdDebugMarkerEndEXT,
    .CmdDebugMarkerInsertEXT = anv_rmv_CmdDebugMarkerInsertEXT,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleNV = anv_rmv_GetMemoryWin32HandleNV,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleNV = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .CmdExecuteGeneratedCommandsNV = anv_rmv_CmdExecuteGeneratedCommandsNV,
    .CmdPreprocessGeneratedCommandsNV = anv_rmv_CmdPreprocessGeneratedCommandsNV,
    .CmdBindPipelineShaderGroupNV = anv_rmv_CmdBindPipelineShaderGroupNV,
    .GetGeneratedCommandsMemoryRequirementsNV = anv_rmv_GetGeneratedCommandsMemoryRequirementsNV,
    .CreateIndirectCommandsLayoutNV = anv_rmv_CreateIndirectCommandsLayoutNV,
    .DestroyIndirectCommandsLayoutNV = anv_rmv_DestroyIndirectCommandsLayoutNV,
    .CmdExecuteGeneratedCommandsEXT = anv_rmv_CmdExecuteGeneratedCommandsEXT,
    .CmdPreprocessGeneratedCommandsEXT = anv_rmv_CmdPreprocessGeneratedCommandsEXT,
    .GetGeneratedCommandsMemoryRequirementsEXT = anv_rmv_GetGeneratedCommandsMemoryRequirementsEXT,
    .CreateIndirectCommandsLayoutEXT = anv_rmv_CreateIndirectCommandsLayoutEXT,
    .DestroyIndirectCommandsLayoutEXT = anv_rmv_DestroyIndirectCommandsLayoutEXT,
    .CreateIndirectExecutionSetEXT = anv_rmv_CreateIndirectExecutionSetEXT,
    .DestroyIndirectExecutionSetEXT = anv_rmv_DestroyIndirectExecutionSetEXT,
    .UpdateIndirectExecutionSetPipelineEXT = anv_rmv_UpdateIndirectExecutionSetPipelineEXT,
    .UpdateIndirectExecutionSetShaderEXT = anv_rmv_UpdateIndirectExecutionSetShaderEXT,
    .CmdPushDescriptorSet = anv_rmv_CmdPushDescriptorSet,
    .CmdPushDescriptorSetKHR = anv_rmv_CmdPushDescriptorSetKHR,
    .TrimCommandPool = anv_rmv_TrimCommandPool,
    .TrimCommandPoolKHR = anv_rmv_TrimCommandPoolKHR,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandleKHR = anv_rmv_GetMemoryWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryWin32HandlePropertiesKHR = anv_rmv_GetMemoryWin32HandlePropertiesKHR,
#elif defined(_MSC_VER)
    .GetMemoryWin32HandlePropertiesKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetMemoryFdKHR = anv_rmv_GetMemoryFdKHR,
    .GetMemoryFdPropertiesKHR = anv_rmv_GetMemoryFdPropertiesKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandleFUCHSIA = anv_rmv_GetMemoryZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetMemoryZirconHandlePropertiesFUCHSIA = anv_rmv_GetMemoryZirconHandlePropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetMemoryZirconHandlePropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .GetMemoryRemoteAddressNV = anv_rmv_GetMemoryRemoteAddressNV,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreWin32HandleKHR = anv_rmv_GetSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportSemaphoreWin32HandleKHR = anv_rmv_ImportSemaphoreWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportSemaphoreWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetSemaphoreFdKHR = anv_rmv_GetSemaphoreFdKHR,
    .ImportSemaphoreFdKHR = anv_rmv_ImportSemaphoreFdKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetSemaphoreZirconHandleFUCHSIA = anv_rmv_GetSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .GetSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .ImportSemaphoreZirconHandleFUCHSIA = anv_rmv_ImportSemaphoreZirconHandleFUCHSIA,
#elif defined(_MSC_VER)
    .ImportSemaphoreZirconHandleFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetFenceWin32HandleKHR = anv_rmv_GetFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .GetFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ImportFenceWin32HandleKHR = anv_rmv_ImportFenceWin32HandleKHR,
#elif defined(_MSC_VER)
    .ImportFenceWin32HandleKHR = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .GetFenceFdKHR = anv_rmv_GetFenceFdKHR,
    .ImportFenceFdKHR = anv_rmv_ImportFenceFdKHR,
    .DisplayPowerControlEXT = anv_rmv_DisplayPowerControlEXT,
    .RegisterDeviceEventEXT = anv_rmv_RegisterDeviceEventEXT,
    .RegisterDisplayEventEXT = anv_rmv_RegisterDisplayEventEXT,
    .GetSwapchainCounterEXT = anv_rmv_GetSwapchainCounterEXT,
    .GetDeviceGroupPeerMemoryFeatures = anv_rmv_GetDeviceGroupPeerMemoryFeatures,
    .GetDeviceGroupPeerMemoryFeaturesKHR = anv_rmv_GetDeviceGroupPeerMemoryFeaturesKHR,
    .BindBufferMemory2 = anv_rmv_BindBufferMemory2,
    .BindBufferMemory2KHR = anv_rmv_BindBufferMemory2KHR,
    .BindImageMemory2 = anv_rmv_BindImageMemory2,
    .BindImageMemory2KHR = anv_rmv_BindImageMemory2KHR,
    .CmdSetDeviceMask = anv_rmv_CmdSetDeviceMask,
    .CmdSetDeviceMaskKHR = anv_rmv_CmdSetDeviceMaskKHR,
    .GetDeviceGroupPresentCapabilitiesKHR = anv_rmv_GetDeviceGroupPresentCapabilitiesKHR,
    .GetDeviceGroupSurfacePresentModesKHR = anv_rmv_GetDeviceGroupSurfacePresentModesKHR,
    .AcquireNextImage2KHR = anv_rmv_AcquireNextImage2KHR,
    .CmdDispatchBase = anv_rmv_CmdDispatchBase,
    .CmdDispatchBaseKHR = anv_rmv_CmdDispatchBaseKHR,
    .CreateDescriptorUpdateTemplate = anv_rmv_CreateDescriptorUpdateTemplate,
    .CreateDescriptorUpdateTemplateKHR = anv_rmv_CreateDescriptorUpdateTemplateKHR,
    .DestroyDescriptorUpdateTemplate = anv_rmv_DestroyDescriptorUpdateTemplate,
    .DestroyDescriptorUpdateTemplateKHR = anv_rmv_DestroyDescriptorUpdateTemplateKHR,
    .UpdateDescriptorSetWithTemplate = anv_rmv_UpdateDescriptorSetWithTemplate,
    .UpdateDescriptorSetWithTemplateKHR = anv_rmv_UpdateDescriptorSetWithTemplateKHR,
    .CmdPushDescriptorSetWithTemplate = anv_rmv_CmdPushDescriptorSetWithTemplate,
    .CmdPushDescriptorSetWithTemplateKHR = anv_rmv_CmdPushDescriptorSetWithTemplateKHR,
    .SetHdrMetadataEXT = anv_rmv_SetHdrMetadataEXT,
    .GetSwapchainStatusKHR = anv_rmv_GetSwapchainStatusKHR,
    .GetRefreshCycleDurationGOOGLE = anv_rmv_GetRefreshCycleDurationGOOGLE,
    .GetPastPresentationTimingGOOGLE = anv_rmv_GetPastPresentationTimingGOOGLE,
    .CmdSetViewportWScalingNV = anv_rmv_CmdSetViewportWScalingNV,
    .CmdSetDiscardRectangleEXT = anv_rmv_CmdSetDiscardRectangleEXT,
    .CmdSetDiscardRectangleEnableEXT = anv_rmv_CmdSetDiscardRectangleEnableEXT,
    .CmdSetDiscardRectangleModeEXT = anv_rmv_CmdSetDiscardRectangleModeEXT,
    .CmdSetSampleLocationsEXT = anv_rmv_CmdSetSampleLocationsEXT,
    .GetBufferMemoryRequirements2 = anv_rmv_GetBufferMemoryRequirements2,
    .GetBufferMemoryRequirements2KHR = anv_rmv_GetBufferMemoryRequirements2KHR,
    .GetImageMemoryRequirements2 = anv_rmv_GetImageMemoryRequirements2,
    .GetImageMemoryRequirements2KHR = anv_rmv_GetImageMemoryRequirements2KHR,
    .GetImageSparseMemoryRequirements2 = anv_rmv_GetImageSparseMemoryRequirements2,
    .GetImageSparseMemoryRequirements2KHR = anv_rmv_GetImageSparseMemoryRequirements2KHR,
    .GetDeviceBufferMemoryRequirements = anv_rmv_GetDeviceBufferMemoryRequirements,
    .GetDeviceBufferMemoryRequirementsKHR = anv_rmv_GetDeviceBufferMemoryRequirementsKHR,
    .GetDeviceImageMemoryRequirements = anv_rmv_GetDeviceImageMemoryRequirements,
    .GetDeviceImageMemoryRequirementsKHR = anv_rmv_GetDeviceImageMemoryRequirementsKHR,
    .GetDeviceImageSparseMemoryRequirements = anv_rmv_GetDeviceImageSparseMemoryRequirements,
    .GetDeviceImageSparseMemoryRequirementsKHR = anv_rmv_GetDeviceImageSparseMemoryRequirementsKHR,
    .CreateSamplerYcbcrConversion = anv_rmv_CreateSamplerYcbcrConversion,
    .CreateSamplerYcbcrConversionKHR = anv_rmv_CreateSamplerYcbcrConversionKHR,
    .DestroySamplerYcbcrConversion = anv_rmv_DestroySamplerYcbcrConversion,
    .DestroySamplerYcbcrConversionKHR = anv_rmv_DestroySamplerYcbcrConversionKHR,
    .GetDeviceQueue2 = anv_rmv_GetDeviceQueue2,
    .CreateValidationCacheEXT = anv_rmv_CreateValidationCacheEXT,
    .DestroyValidationCacheEXT = anv_rmv_DestroyValidationCacheEXT,
    .GetValidationCacheDataEXT = anv_rmv_GetValidationCacheDataEXT,
    .MergeValidationCachesEXT = anv_rmv_MergeValidationCachesEXT,
    .GetDescriptorSetLayoutSupport = anv_rmv_GetDescriptorSetLayoutSupport,
    .GetDescriptorSetLayoutSupportKHR = anv_rmv_GetDescriptorSetLayoutSupportKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsageANDROID = anv_rmv_GetSwapchainGrallocUsageANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetSwapchainGrallocUsage2ANDROID = anv_rmv_GetSwapchainGrallocUsage2ANDROID,
#elif defined(_MSC_VER)
    .GetSwapchainGrallocUsage2ANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .AcquireImageANDROID = anv_rmv_AcquireImageANDROID,
#elif defined(_MSC_VER)
    .AcquireImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .QueueSignalReleaseImageANDROID = anv_rmv_QueueSignalReleaseImageANDROID,
#elif defined(_MSC_VER)
    .QueueSignalReleaseImageANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .GetShaderInfoAMD = anv_rmv_GetShaderInfoAMD,
    .SetLocalDimmingAMD = anv_rmv_SetLocalDimmingAMD,
    .GetCalibratedTimestampsKHR = anv_rmv_GetCalibratedTimestampsKHR,
    .GetCalibratedTimestampsEXT = anv_rmv_GetCalibratedTimestampsEXT,
    .SetDebugUtilsObjectNameEXT = anv_rmv_SetDebugUtilsObjectNameEXT,
    .SetDebugUtilsObjectTagEXT = anv_rmv_SetDebugUtilsObjectTagEXT,
    .QueueBeginDebugUtilsLabelEXT = anv_rmv_QueueBeginDebugUtilsLabelEXT,
    .QueueEndDebugUtilsLabelEXT = anv_rmv_QueueEndDebugUtilsLabelEXT,
    .QueueInsertDebugUtilsLabelEXT = anv_rmv_QueueInsertDebugUtilsLabelEXT,
    .CmdBeginDebugUtilsLabelEXT = anv_rmv_CmdBeginDebugUtilsLabelEXT,
    .CmdEndDebugUtilsLabelEXT = anv_rmv_CmdEndDebugUtilsLabelEXT,
    .CmdInsertDebugUtilsLabelEXT = anv_rmv_CmdInsertDebugUtilsLabelEXT,
    .GetMemoryHostPointerPropertiesEXT = anv_rmv_GetMemoryHostPointerPropertiesEXT,
    .CmdWriteBufferMarkerAMD = anv_rmv_CmdWriteBufferMarkerAMD,
    .CreateRenderPass2 = anv_rmv_CreateRenderPass2,
    .CreateRenderPass2KHR = anv_rmv_CreateRenderPass2KHR,
    .CmdBeginRenderPass2 = anv_rmv_CmdBeginRenderPass2,
    .CmdBeginRenderPass2KHR = anv_rmv_CmdBeginRenderPass2KHR,
    .CmdNextSubpass2 = anv_rmv_CmdNextSubpass2,
    .CmdNextSubpass2KHR = anv_rmv_CmdNextSubpass2KHR,
    .CmdEndRenderPass2 = anv_rmv_CmdEndRenderPass2,
    .CmdEndRenderPass2KHR = anv_rmv_CmdEndRenderPass2KHR,
    .GetSemaphoreCounterValue = anv_rmv_GetSemaphoreCounterValue,
    .GetSemaphoreCounterValueKHR = anv_rmv_GetSemaphoreCounterValueKHR,
    .WaitSemaphores = anv_rmv_WaitSemaphores,
    .WaitSemaphoresKHR = anv_rmv_WaitSemaphoresKHR,
    .SignalSemaphore = anv_rmv_SignalSemaphore,
    .SignalSemaphoreKHR = anv_rmv_SignalSemaphoreKHR,
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetAndroidHardwareBufferPropertiesANDROID = anv_rmv_GetAndroidHardwareBufferPropertiesANDROID,
#elif defined(_MSC_VER)
    .GetAndroidHardwareBufferPropertiesANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
#ifdef VK_USE_PLATFORM_ANDROID_KHR
    .GetMemoryAndroidHardwareBufferANDROID = anv_rmv_GetMemoryAndroidHardwareBufferANDROID,
#elif defined(_MSC_VER)
    .GetMemoryAndroidHardwareBufferANDROID = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_ANDROID_KHR
    .CmdDrawIndirectCount = anv_rmv_CmdDrawIndirectCount,
    .CmdDrawIndirectCountKHR = anv_rmv_CmdDrawIndirectCountKHR,
    .CmdDrawIndirectCountAMD = anv_rmv_CmdDrawIndirectCountAMD,
    .CmdDrawIndexedIndirectCount = anv_rmv_CmdDrawIndexedIndirectCount,
    .CmdDrawIndexedIndirectCountKHR = anv_rmv_CmdDrawIndexedIndirectCountKHR,
    .CmdDrawIndexedIndirectCountAMD = anv_rmv_CmdDrawIndexedIndirectCountAMD,
    .CmdSetCheckpointNV = anv_rmv_CmdSetCheckpointNV,
    .GetQueueCheckpointDataNV = anv_rmv_GetQueueCheckpointDataNV,
    .CmdBindTransformFeedbackBuffersEXT = anv_rmv_CmdBindTransformFeedbackBuffersEXT,
    .CmdBeginTransformFeedbackEXT = anv_rmv_CmdBeginTransformFeedbackEXT,
    .CmdEndTransformFeedbackEXT = anv_rmv_CmdEndTransformFeedbackEXT,
    .CmdBeginQueryIndexedEXT = anv_rmv_CmdBeginQueryIndexedEXT,
    .CmdEndQueryIndexedEXT = anv_rmv_CmdEndQueryIndexedEXT,
    .CmdDrawIndirectByteCountEXT = anv_rmv_CmdDrawIndirectByteCountEXT,
    .CmdSetExclusiveScissorNV = anv_rmv_CmdSetExclusiveScissorNV,
    .CmdSetExclusiveScissorEnableNV = anv_rmv_CmdSetExclusiveScissorEnableNV,
    .CmdBindShadingRateImageNV = anv_rmv_CmdBindShadingRateImageNV,
    .CmdSetViewportShadingRatePaletteNV = anv_rmv_CmdSetViewportShadingRatePaletteNV,
    .CmdSetCoarseSampleOrderNV = anv_rmv_CmdSetCoarseSampleOrderNV,
    .CmdDrawMeshTasksNV = anv_rmv_CmdDrawMeshTasksNV,
    .CmdDrawMeshTasksIndirectNV = anv_rmv_CmdDrawMeshTasksIndirectNV,
    .CmdDrawMeshTasksIndirectCountNV = anv_rmv_CmdDrawMeshTasksIndirectCountNV,
    .CmdDrawMeshTasksEXT = anv_rmv_CmdDrawMeshTasksEXT,
    .CmdDrawMeshTasksIndirectEXT = anv_rmv_CmdDrawMeshTasksIndirectEXT,
    .CmdDrawMeshTasksIndirectCountEXT = anv_rmv_CmdDrawMeshTasksIndirectCountEXT,
    .CompileDeferredNV = anv_rmv_CompileDeferredNV,
    .CreateAccelerationStructureNV = anv_rmv_CreateAccelerationStructureNV,
    .CmdBindInvocationMaskHUAWEI = anv_rmv_CmdBindInvocationMaskHUAWEI,
    .DestroyAccelerationStructureKHR = anv_rmv_DestroyAccelerationStructureKHR,
    .DestroyAccelerationStructureNV = anv_rmv_DestroyAccelerationStructureNV,
    .GetAccelerationStructureMemoryRequirementsNV = anv_rmv_GetAccelerationStructureMemoryRequirementsNV,
    .BindAccelerationStructureMemoryNV = anv_rmv_BindAccelerationStructureMemoryNV,
    .CmdCopyAccelerationStructureNV = anv_rmv_CmdCopyAccelerationStructureNV,
    .CmdCopyAccelerationStructureKHR = anv_rmv_CmdCopyAccelerationStructureKHR,
    .CopyAccelerationStructureKHR = anv_rmv_CopyAccelerationStructureKHR,
    .CmdCopyAccelerationStructureToMemoryKHR = anv_rmv_CmdCopyAccelerationStructureToMemoryKHR,
    .CopyAccelerationStructureToMemoryKHR = anv_rmv_CopyAccelerationStructureToMemoryKHR,
    .CmdCopyMemoryToAccelerationStructureKHR = anv_rmv_CmdCopyMemoryToAccelerationStructureKHR,
    .CopyMemoryToAccelerationStructureKHR = anv_rmv_CopyMemoryToAccelerationStructureKHR,
    .CmdWriteAccelerationStructuresPropertiesKHR = anv_rmv_CmdWriteAccelerationStructuresPropertiesKHR,
    .CmdWriteAccelerationStructuresPropertiesNV = anv_rmv_CmdWriteAccelerationStructuresPropertiesNV,
    .CmdBuildAccelerationStructureNV = anv_rmv_CmdBuildAccelerationStructureNV,
    .WriteAccelerationStructuresPropertiesKHR = anv_rmv_WriteAccelerationStructuresPropertiesKHR,
    .CmdTraceRaysKHR = anv_rmv_CmdTraceRaysKHR,
    .CmdTraceRaysNV = anv_rmv_CmdTraceRaysNV,
    .GetRayTracingShaderGroupHandlesKHR = anv_rmv_GetRayTracingShaderGroupHandlesKHR,
    .GetRayTracingShaderGroupHandlesNV = anv_rmv_GetRayTracingShaderGroupHandlesNV,
    .GetRayTracingCaptureReplayShaderGroupHandlesKHR = anv_rmv_GetRayTracingCaptureReplayShaderGroupHandlesKHR,
    .GetAccelerationStructureHandleNV = anv_rmv_GetAccelerationStructureHandleNV,
    .CreateRayTracingPipelinesNV = anv_rmv_CreateRayTracingPipelinesNV,
    .CreateRayTracingPipelinesKHR = anv_rmv_CreateRayTracingPipelinesKHR,
    .CmdTraceRaysIndirectKHR = anv_rmv_CmdTraceRaysIndirectKHR,
    .CmdTraceRaysIndirect2KHR = anv_rmv_CmdTraceRaysIndirect2KHR,
    .GetClusterAccelerationStructureBuildSizesNV = anv_rmv_GetClusterAccelerationStructureBuildSizesNV,
    .CmdBuildClusterAccelerationStructureIndirectNV = anv_rmv_CmdBuildClusterAccelerationStructureIndirectNV,
    .GetDeviceAccelerationStructureCompatibilityKHR = anv_rmv_GetDeviceAccelerationStructureCompatibilityKHR,
    .GetRayTracingShaderGroupStackSizeKHR = anv_rmv_GetRayTracingShaderGroupStackSizeKHR,
    .CmdSetRayTracingPipelineStackSizeKHR = anv_rmv_CmdSetRayTracingPipelineStackSizeKHR,
    .GetImageViewHandleNVX = anv_rmv_GetImageViewHandleNVX,
    .GetImageViewHandle64NVX = anv_rmv_GetImageViewHandle64NVX,
    .GetImageViewAddressNVX = anv_rmv_GetImageViewAddressNVX,
    .GetDeviceCombinedImageSamplerIndexNVX = anv_rmv_GetDeviceCombinedImageSamplerIndexNVX,
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .GetDeviceGroupSurfacePresentModes2EXT = anv_rmv_GetDeviceGroupSurfacePresentModes2EXT,
#elif defined(_MSC_VER)
    .GetDeviceGroupSurfacePresentModes2EXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .AcquireFullScreenExclusiveModeEXT = anv_rmv_AcquireFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .AcquireFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
#ifdef VK_USE_PLATFORM_WIN32_KHR
    .ReleaseFullScreenExclusiveModeEXT = anv_rmv_ReleaseFullScreenExclusiveModeEXT,
#elif defined(_MSC_VER)
    .ReleaseFullScreenExclusiveModeEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_WIN32_KHR
    .AcquireProfilingLockKHR = anv_rmv_AcquireProfilingLockKHR,
    .ReleaseProfilingLockKHR = anv_rmv_ReleaseProfilingLockKHR,
    .GetImageDrmFormatModifierPropertiesEXT = anv_rmv_GetImageDrmFormatModifierPropertiesEXT,
    .GetBufferOpaqueCaptureAddress = anv_rmv_GetBufferOpaqueCaptureAddress,
    .GetBufferOpaqueCaptureAddressKHR = anv_rmv_GetBufferOpaqueCaptureAddressKHR,
    .GetBufferDeviceAddress = anv_rmv_GetBufferDeviceAddress,
    .GetBufferDeviceAddressKHR = anv_rmv_GetBufferDeviceAddressKHR,
    .GetBufferDeviceAddressEXT = anv_rmv_GetBufferDeviceAddressEXT,
    .InitializePerformanceApiINTEL = anv_rmv_InitializePerformanceApiINTEL,
    .UninitializePerformanceApiINTEL = anv_rmv_UninitializePerformanceApiINTEL,
    .CmdSetPerformanceMarkerINTEL = anv_rmv_CmdSetPerformanceMarkerINTEL,
    .CmdSetPerformanceStreamMarkerINTEL = anv_rmv_CmdSetPerformanceStreamMarkerINTEL,
    .CmdSetPerformanceOverrideINTEL = anv_rmv_CmdSetPerformanceOverrideINTEL,
    .AcquirePerformanceConfigurationINTEL = anv_rmv_AcquirePerformanceConfigurationINTEL,
    .ReleasePerformanceConfigurationINTEL = anv_rmv_ReleasePerformanceConfigurationINTEL,
    .QueueSetPerformanceConfigurationINTEL = anv_rmv_QueueSetPerformanceConfigurationINTEL,
    .GetPerformanceParameterINTEL = anv_rmv_GetPerformanceParameterINTEL,
    .GetDeviceMemoryOpaqueCaptureAddress = anv_rmv_GetDeviceMemoryOpaqueCaptureAddress,
    .GetDeviceMemoryOpaqueCaptureAddressKHR = anv_rmv_GetDeviceMemoryOpaqueCaptureAddressKHR,
    .GetPipelineExecutablePropertiesKHR = anv_rmv_GetPipelineExecutablePropertiesKHR,
    .GetPipelineExecutableStatisticsKHR = anv_rmv_GetPipelineExecutableStatisticsKHR,
    .GetPipelineExecutableInternalRepresentationsKHR = anv_rmv_GetPipelineExecutableInternalRepresentationsKHR,
    .CmdSetLineStipple = anv_rmv_CmdSetLineStipple,
    .CmdSetLineStippleKHR = anv_rmv_CmdSetLineStippleKHR,
    .CmdSetLineStippleEXT = anv_rmv_CmdSetLineStippleEXT,
    .CreateAccelerationStructureKHR = anv_rmv_CreateAccelerationStructureKHR,
    .CmdBuildAccelerationStructuresKHR = anv_rmv_CmdBuildAccelerationStructuresKHR,
    .CmdBuildAccelerationStructuresIndirectKHR = anv_rmv_CmdBuildAccelerationStructuresIndirectKHR,
    .BuildAccelerationStructuresKHR = anv_rmv_BuildAccelerationStructuresKHR,
    .GetAccelerationStructureDeviceAddressKHR = anv_rmv_GetAccelerationStructureDeviceAddressKHR,
    .CreateDeferredOperationKHR = anv_rmv_CreateDeferredOperationKHR,
    .DestroyDeferredOperationKHR = anv_rmv_DestroyDeferredOperationKHR,
    .GetDeferredOperationMaxConcurrencyKHR = anv_rmv_GetDeferredOperationMaxConcurrencyKHR,
    .GetDeferredOperationResultKHR = anv_rmv_GetDeferredOperationResultKHR,
    .DeferredOperationJoinKHR = anv_rmv_DeferredOperationJoinKHR,
    .GetPipelineIndirectMemoryRequirementsNV = anv_rmv_GetPipelineIndirectMemoryRequirementsNV,
    .GetPipelineIndirectDeviceAddressNV = anv_rmv_GetPipelineIndirectDeviceAddressNV,
    .AntiLagUpdateAMD = anv_rmv_AntiLagUpdateAMD,
    .CmdSetCullMode = anv_rmv_CmdSetCullMode,
    .CmdSetCullModeEXT = anv_rmv_CmdSetCullModeEXT,
    .CmdSetFrontFace = anv_rmv_CmdSetFrontFace,
    .CmdSetFrontFaceEXT = anv_rmv_CmdSetFrontFaceEXT,
    .CmdSetPrimitiveTopology = anv_rmv_CmdSetPrimitiveTopology,
    .CmdSetPrimitiveTopologyEXT = anv_rmv_CmdSetPrimitiveTopologyEXT,
    .CmdSetViewportWithCount = anv_rmv_CmdSetViewportWithCount,
    .CmdSetViewportWithCountEXT = anv_rmv_CmdSetViewportWithCountEXT,
    .CmdSetScissorWithCount = anv_rmv_CmdSetScissorWithCount,
    .CmdSetScissorWithCountEXT = anv_rmv_CmdSetScissorWithCountEXT,
    .CmdBindIndexBuffer2 = anv_rmv_CmdBindIndexBuffer2,
    .CmdBindIndexBuffer2KHR = anv_rmv_CmdBindIndexBuffer2KHR,
    .CmdBindVertexBuffers2 = anv_rmv_CmdBindVertexBuffers2,
    .CmdBindVertexBuffers2EXT = anv_rmv_CmdBindVertexBuffers2EXT,
    .CmdSetDepthTestEnable = anv_rmv_CmdSetDepthTestEnable,
    .CmdSetDepthTestEnableEXT = anv_rmv_CmdSetDepthTestEnableEXT,
    .CmdSetDepthWriteEnable = anv_rmv_CmdSetDepthWriteEnable,
    .CmdSetDepthWriteEnableEXT = anv_rmv_CmdSetDepthWriteEnableEXT,
    .CmdSetDepthCompareOp = anv_rmv_CmdSetDepthCompareOp,
    .CmdSetDepthCompareOpEXT = anv_rmv_CmdSetDepthCompareOpEXT,
    .CmdSetDepthBoundsTestEnable = anv_rmv_CmdSetDepthBoundsTestEnable,
    .CmdSetDepthBoundsTestEnableEXT = anv_rmv_CmdSetDepthBoundsTestEnableEXT,
    .CmdSetStencilTestEnable = anv_rmv_CmdSetStencilTestEnable,
    .CmdSetStencilTestEnableEXT = anv_rmv_CmdSetStencilTestEnableEXT,
    .CmdSetStencilOp = anv_rmv_CmdSetStencilOp,
    .CmdSetStencilOpEXT = anv_rmv_CmdSetStencilOpEXT,
    .CmdSetPatchControlPointsEXT = anv_rmv_CmdSetPatchControlPointsEXT,
    .CmdSetRasterizerDiscardEnable = anv_rmv_CmdSetRasterizerDiscardEnable,
    .CmdSetRasterizerDiscardEnableEXT = anv_rmv_CmdSetRasterizerDiscardEnableEXT,
    .CmdSetDepthBiasEnable = anv_rmv_CmdSetDepthBiasEnable,
    .CmdSetDepthBiasEnableEXT = anv_rmv_CmdSetDepthBiasEnableEXT,
    .CmdSetLogicOpEXT = anv_rmv_CmdSetLogicOpEXT,
    .CmdSetPrimitiveRestartEnable = anv_rmv_CmdSetPrimitiveRestartEnable,
    .CmdSetPrimitiveRestartEnableEXT = anv_rmv_CmdSetPrimitiveRestartEnableEXT,
    .CmdSetTessellationDomainOriginEXT = anv_rmv_CmdSetTessellationDomainOriginEXT,
    .CmdSetDepthClampEnableEXT = anv_rmv_CmdSetDepthClampEnableEXT,
    .CmdSetPolygonModeEXT = anv_rmv_CmdSetPolygonModeEXT,
    .CmdSetRasterizationSamplesEXT = anv_rmv_CmdSetRasterizationSamplesEXT,
    .CmdSetSampleMaskEXT = anv_rmv_CmdSetSampleMaskEXT,
    .CmdSetAlphaToCoverageEnableEXT = anv_rmv_CmdSetAlphaToCoverageEnableEXT,
    .CmdSetAlphaToOneEnableEXT = anv_rmv_CmdSetAlphaToOneEnableEXT,
    .CmdSetLogicOpEnableEXT = anv_rmv_CmdSetLogicOpEnableEXT,
    .CmdSetColorBlendEnableEXT = anv_rmv_CmdSetColorBlendEnableEXT,
    .CmdSetColorBlendEquationEXT = anv_rmv_CmdSetColorBlendEquationEXT,
    .CmdSetColorWriteMaskEXT = anv_rmv_CmdSetColorWriteMaskEXT,
    .CmdSetRasterizationStreamEXT = anv_rmv_CmdSetRasterizationStreamEXT,
    .CmdSetConservativeRasterizationModeEXT = anv_rmv_CmdSetConservativeRasterizationModeEXT,
    .CmdSetExtraPrimitiveOverestimationSizeEXT = anv_rmv_CmdSetExtraPrimitiveOverestimationSizeEXT,
    .CmdSetDepthClipEnableEXT = anv_rmv_CmdSetDepthClipEnableEXT,
    .CmdSetSampleLocationsEnableEXT = anv_rmv_CmdSetSampleLocationsEnableEXT,
    .CmdSetColorBlendAdvancedEXT = anv_rmv_CmdSetColorBlendAdvancedEXT,
    .CmdSetProvokingVertexModeEXT = anv_rmv_CmdSetProvokingVertexModeEXT,
    .CmdSetLineRasterizationModeEXT = anv_rmv_CmdSetLineRasterizationModeEXT,
    .CmdSetLineStippleEnableEXT = anv_rmv_CmdSetLineStippleEnableEXT,
    .CmdSetDepthClipNegativeOneToOneEXT = anv_rmv_CmdSetDepthClipNegativeOneToOneEXT,
    .CmdSetViewportWScalingEnableNV = anv_rmv_CmdSetViewportWScalingEnableNV,
    .CmdSetViewportSwizzleNV = anv_rmv_CmdSetViewportSwizzleNV,
    .CmdSetCoverageToColorEnableNV = anv_rmv_CmdSetCoverageToColorEnableNV,
    .CmdSetCoverageToColorLocationNV = anv_rmv_CmdSetCoverageToColorLocationNV,
    .CmdSetCoverageModulationModeNV = anv_rmv_CmdSetCoverageModulationModeNV,
    .CmdSetCoverageModulationTableEnableNV = anv_rmv_CmdSetCoverageModulationTableEnableNV,
    .CmdSetCoverageModulationTableNV = anv_rmv_CmdSetCoverageModulationTableNV,
    .CmdSetShadingRateImageEnableNV = anv_rmv_CmdSetShadingRateImageEnableNV,
    .CmdSetCoverageReductionModeNV = anv_rmv_CmdSetCoverageReductionModeNV,
    .CmdSetRepresentativeFragmentTestEnableNV = anv_rmv_CmdSetRepresentativeFragmentTestEnableNV,
    .CreatePrivateDataSlot = anv_rmv_CreatePrivateDataSlot,
    .CreatePrivateDataSlotEXT = anv_rmv_CreatePrivateDataSlotEXT,
    .DestroyPrivateDataSlot = anv_rmv_DestroyPrivateDataSlot,
    .DestroyPrivateDataSlotEXT = anv_rmv_DestroyPrivateDataSlotEXT,
    .SetPrivateData = anv_rmv_SetPrivateData,
    .SetPrivateDataEXT = anv_rmv_SetPrivateDataEXT,
    .GetPrivateData = anv_rmv_GetPrivateData,
    .GetPrivateDataEXT = anv_rmv_GetPrivateDataEXT,
    .CmdCopyBuffer2 = anv_rmv_CmdCopyBuffer2,
    .CmdCopyBuffer2KHR = anv_rmv_CmdCopyBuffer2KHR,
    .CmdCopyImage2 = anv_rmv_CmdCopyImage2,
    .CmdCopyImage2KHR = anv_rmv_CmdCopyImage2KHR,
    .CmdBlitImage2 = anv_rmv_CmdBlitImage2,
    .CmdBlitImage2KHR = anv_rmv_CmdBlitImage2KHR,
    .CmdCopyBufferToImage2 = anv_rmv_CmdCopyBufferToImage2,
    .CmdCopyBufferToImage2KHR = anv_rmv_CmdCopyBufferToImage2KHR,
    .CmdCopyImageToBuffer2 = anv_rmv_CmdCopyImageToBuffer2,
    .CmdCopyImageToBuffer2KHR = anv_rmv_CmdCopyImageToBuffer2KHR,
    .CmdResolveImage2 = anv_rmv_CmdResolveImage2,
    .CmdResolveImage2KHR = anv_rmv_CmdResolveImage2KHR,
    .CmdSetFragmentShadingRateKHR = anv_rmv_CmdSetFragmentShadingRateKHR,
    .CmdSetFragmentShadingRateEnumNV = anv_rmv_CmdSetFragmentShadingRateEnumNV,
    .GetAccelerationStructureBuildSizesKHR = anv_rmv_GetAccelerationStructureBuildSizesKHR,
    .CmdSetVertexInputEXT = anv_rmv_CmdSetVertexInputEXT,
    .CmdSetColorWriteEnableEXT = anv_rmv_CmdSetColorWriteEnableEXT,
    .CmdSetEvent2 = anv_rmv_CmdSetEvent2,
    .CmdSetEvent2KHR = anv_rmv_CmdSetEvent2KHR,
    .CmdResetEvent2 = anv_rmv_CmdResetEvent2,
    .CmdResetEvent2KHR = anv_rmv_CmdResetEvent2KHR,
    .CmdWaitEvents2 = anv_rmv_CmdWaitEvents2,
    .CmdWaitEvents2KHR = anv_rmv_CmdWaitEvents2KHR,
    .CmdPipelineBarrier2 = anv_rmv_CmdPipelineBarrier2,
    .CmdPipelineBarrier2KHR = anv_rmv_CmdPipelineBarrier2KHR,
    .QueueSubmit2 = anv_rmv_QueueSubmit2,
    .QueueSubmit2KHR = anv_rmv_QueueSubmit2KHR,
    .CmdWriteTimestamp2 = anv_rmv_CmdWriteTimestamp2,
    .CmdWriteTimestamp2KHR = anv_rmv_CmdWriteTimestamp2KHR,
    .CmdWriteBufferMarker2AMD = anv_rmv_CmdWriteBufferMarker2AMD,
    .GetQueueCheckpointData2NV = anv_rmv_GetQueueCheckpointData2NV,
    .CopyMemoryToImage = anv_rmv_CopyMemoryToImage,
    .CopyMemoryToImageEXT = anv_rmv_CopyMemoryToImageEXT,
    .CopyImageToMemory = anv_rmv_CopyImageToMemory,
    .CopyImageToMemoryEXT = anv_rmv_CopyImageToMemoryEXT,
    .CopyImageToImage = anv_rmv_CopyImageToImage,
    .CopyImageToImageEXT = anv_rmv_CopyImageToImageEXT,
    .TransitionImageLayout = anv_rmv_TransitionImageLayout,
    .TransitionImageLayoutEXT = anv_rmv_TransitionImageLayoutEXT,
    .CreateVideoSessionKHR = anv_rmv_CreateVideoSessionKHR,
    .DestroyVideoSessionKHR = anv_rmv_DestroyVideoSessionKHR,
    .CreateVideoSessionParametersKHR = anv_rmv_CreateVideoSessionParametersKHR,
    .UpdateVideoSessionParametersKHR = anv_rmv_UpdateVideoSessionParametersKHR,
    .GetEncodedVideoSessionParametersKHR = anv_rmv_GetEncodedVideoSessionParametersKHR,
    .DestroyVideoSessionParametersKHR = anv_rmv_DestroyVideoSessionParametersKHR,
    .GetVideoSessionMemoryRequirementsKHR = anv_rmv_GetVideoSessionMemoryRequirementsKHR,
    .BindVideoSessionMemoryKHR = anv_rmv_BindVideoSessionMemoryKHR,
    .CmdDecodeVideoKHR = anv_rmv_CmdDecodeVideoKHR,
    .CmdBeginVideoCodingKHR = anv_rmv_CmdBeginVideoCodingKHR,
    .CmdControlVideoCodingKHR = anv_rmv_CmdControlVideoCodingKHR,
    .CmdEndVideoCodingKHR = anv_rmv_CmdEndVideoCodingKHR,
    .CmdEncodeVideoKHR = anv_rmv_CmdEncodeVideoKHR,
    .CmdDecompressMemoryNV = anv_rmv_CmdDecompressMemoryNV,
    .CmdDecompressMemoryIndirectCountNV = anv_rmv_CmdDecompressMemoryIndirectCountNV,
    .GetPartitionedAccelerationStructuresBuildSizesNV = anv_rmv_GetPartitionedAccelerationStructuresBuildSizesNV,
    .CmdBuildPartitionedAccelerationStructuresNV = anv_rmv_CmdBuildPartitionedAccelerationStructuresNV,
    .CmdDecompressMemoryEXT = anv_rmv_CmdDecompressMemoryEXT,
    .CmdDecompressMemoryIndirectCountEXT = anv_rmv_CmdDecompressMemoryIndirectCountEXT,
    .CreateCuModuleNVX = anv_rmv_CreateCuModuleNVX,
    .CreateCuFunctionNVX = anv_rmv_CreateCuFunctionNVX,
    .DestroyCuModuleNVX = anv_rmv_DestroyCuModuleNVX,
    .DestroyCuFunctionNVX = anv_rmv_DestroyCuFunctionNVX,
    .CmdCuLaunchKernelNVX = anv_rmv_CmdCuLaunchKernelNVX,
    .GetDescriptorSetLayoutSizeEXT = anv_rmv_GetDescriptorSetLayoutSizeEXT,
    .GetDescriptorSetLayoutBindingOffsetEXT = anv_rmv_GetDescriptorSetLayoutBindingOffsetEXT,
    .GetDescriptorEXT = anv_rmv_GetDescriptorEXT,
    .CmdBindDescriptorBuffersEXT = anv_rmv_CmdBindDescriptorBuffersEXT,
    .CmdSetDescriptorBufferOffsetsEXT = anv_rmv_CmdSetDescriptorBufferOffsetsEXT,
    .CmdBindDescriptorBufferEmbeddedSamplersEXT = anv_rmv_CmdBindDescriptorBufferEmbeddedSamplersEXT,
    .GetBufferOpaqueCaptureDescriptorDataEXT = anv_rmv_GetBufferOpaqueCaptureDescriptorDataEXT,
    .GetImageOpaqueCaptureDescriptorDataEXT = anv_rmv_GetImageOpaqueCaptureDescriptorDataEXT,
    .GetImageViewOpaqueCaptureDescriptorDataEXT = anv_rmv_GetImageViewOpaqueCaptureDescriptorDataEXT,
    .GetSamplerOpaqueCaptureDescriptorDataEXT = anv_rmv_GetSamplerOpaqueCaptureDescriptorDataEXT,
    .GetAccelerationStructureOpaqueCaptureDescriptorDataEXT = anv_rmv_GetAccelerationStructureOpaqueCaptureDescriptorDataEXT,
    .SetDeviceMemoryPriorityEXT = anv_rmv_SetDeviceMemoryPriorityEXT,
    .WaitForPresent2KHR = anv_rmv_WaitForPresent2KHR,
    .WaitForPresentKHR = anv_rmv_WaitForPresentKHR,
#ifdef VK_USE_PLATFORM_FUCHSIA
    .CreateBufferCollectionFUCHSIA = anv_rmv_CreateBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .CreateBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionBufferConstraintsFUCHSIA = anv_rmv_SetBufferCollectionBufferConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionBufferConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .SetBufferCollectionImageConstraintsFUCHSIA = anv_rmv_SetBufferCollectionImageConstraintsFUCHSIA,
#elif defined(_MSC_VER)
    .SetBufferCollectionImageConstraintsFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .DestroyBufferCollectionFUCHSIA = anv_rmv_DestroyBufferCollectionFUCHSIA,
#elif defined(_MSC_VER)
    .DestroyBufferCollectionFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
#ifdef VK_USE_PLATFORM_FUCHSIA
    .GetBufferCollectionPropertiesFUCHSIA = anv_rmv_GetBufferCollectionPropertiesFUCHSIA,
#elif defined(_MSC_VER)
    .GetBufferCollectionPropertiesFUCHSIA = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_FUCHSIA
    .CmdBeginRendering = anv_rmv_CmdBeginRendering,
    .CmdBeginRenderingKHR = anv_rmv_CmdBeginRenderingKHR,
    .CmdEndRendering = anv_rmv_CmdEndRendering,
    .CmdEndRendering2KHR = anv_rmv_CmdEndRendering2KHR,
    .CmdEndRendering2EXT = anv_rmv_CmdEndRendering2EXT,
    .CmdEndRenderingKHR = anv_rmv_CmdEndRenderingKHR,
    .GetDescriptorSetLayoutHostMappingInfoVALVE = anv_rmv_GetDescriptorSetLayoutHostMappingInfoVALVE,
    .GetDescriptorSetHostMappingVALVE = anv_rmv_GetDescriptorSetHostMappingVALVE,
    .CreateMicromapEXT = anv_rmv_CreateMicromapEXT,
    .CmdBuildMicromapsEXT = anv_rmv_CmdBuildMicromapsEXT,
    .BuildMicromapsEXT = anv_rmv_BuildMicromapsEXT,
    .DestroyMicromapEXT = anv_rmv_DestroyMicromapEXT,
    .CmdCopyMicromapEXT = anv_rmv_CmdCopyMicromapEXT,
    .CopyMicromapEXT = anv_rmv_CopyMicromapEXT,
    .CmdCopyMicromapToMemoryEXT = anv_rmv_CmdCopyMicromapToMemoryEXT,
    .CopyMicromapToMemoryEXT = anv_rmv_CopyMicromapToMemoryEXT,
    .CmdCopyMemoryToMicromapEXT = anv_rmv_CmdCopyMemoryToMicromapEXT,
    .CopyMemoryToMicromapEXT = anv_rmv_CopyMemoryToMicromapEXT,
    .CmdWriteMicromapsPropertiesEXT = anv_rmv_CmdWriteMicromapsPropertiesEXT,
    .WriteMicromapsPropertiesEXT = anv_rmv_WriteMicromapsPropertiesEXT,
    .GetDeviceMicromapCompatibilityEXT = anv_rmv_GetDeviceMicromapCompatibilityEXT,
    .GetMicromapBuildSizesEXT = anv_rmv_GetMicromapBuildSizesEXT,
    .GetShaderModuleIdentifierEXT = anv_rmv_GetShaderModuleIdentifierEXT,
    .GetShaderModuleCreateInfoIdentifierEXT = anv_rmv_GetShaderModuleCreateInfoIdentifierEXT,
    .GetImageSubresourceLayout2 = anv_rmv_GetImageSubresourceLayout2,
    .GetImageSubresourceLayout2KHR = anv_rmv_GetImageSubresourceLayout2KHR,
    .GetImageSubresourceLayout2EXT = anv_rmv_GetImageSubresourceLayout2EXT,
    .GetPipelinePropertiesEXT = anv_rmv_GetPipelinePropertiesEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .ExportMetalObjectsEXT = anv_rmv_ExportMetalObjectsEXT,
#elif defined(_MSC_VER)
    .ExportMetalObjectsEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .CmdBindTileMemoryQCOM = anv_rmv_CmdBindTileMemoryQCOM,
    .GetFramebufferTilePropertiesQCOM = anv_rmv_GetFramebufferTilePropertiesQCOM,
    .GetDynamicRenderingTilePropertiesQCOM = anv_rmv_GetDynamicRenderingTilePropertiesQCOM,
    .CreateOpticalFlowSessionNV = anv_rmv_CreateOpticalFlowSessionNV,
    .DestroyOpticalFlowSessionNV = anv_rmv_DestroyOpticalFlowSessionNV,
    .BindOpticalFlowSessionImageNV = anv_rmv_BindOpticalFlowSessionImageNV,
    .CmdOpticalFlowExecuteNV = anv_rmv_CmdOpticalFlowExecuteNV,
    .GetDeviceFaultInfoEXT = anv_rmv_GetDeviceFaultInfoEXT,
    .GetDeviceFaultReportsKHR = anv_rmv_GetDeviceFaultReportsKHR,
    .GetDeviceFaultDebugInfoKHR = anv_rmv_GetDeviceFaultDebugInfoKHR,
    .CmdSetDepthBias2EXT = anv_rmv_CmdSetDepthBias2EXT,
    .ReleaseSwapchainImagesKHR = anv_rmv_ReleaseSwapchainImagesKHR,
    .ReleaseSwapchainImagesEXT = anv_rmv_ReleaseSwapchainImagesEXT,
    .GetDeviceImageSubresourceLayout = anv_rmv_GetDeviceImageSubresourceLayout,
    .GetDeviceImageSubresourceLayoutKHR = anv_rmv_GetDeviceImageSubresourceLayoutKHR,
    .MapMemory2 = anv_rmv_MapMemory2,
    .MapMemory2KHR = anv_rmv_MapMemory2KHR,
    .UnmapMemory2 = anv_rmv_UnmapMemory2,
    .UnmapMemory2KHR = anv_rmv_UnmapMemory2KHR,
    .CreateShadersEXT = anv_rmv_CreateShadersEXT,
    .DestroyShaderEXT = anv_rmv_DestroyShaderEXT,
    .GetShaderBinaryDataEXT = anv_rmv_GetShaderBinaryDataEXT,
    .CmdBindShadersEXT = anv_rmv_CmdBindShadersEXT,
    .SetSwapchainPresentTimingQueueSizeEXT = anv_rmv_SetSwapchainPresentTimingQueueSizeEXT,
    .GetSwapchainTimingPropertiesEXT = anv_rmv_GetSwapchainTimingPropertiesEXT,
    .GetSwapchainTimeDomainPropertiesEXT = anv_rmv_GetSwapchainTimeDomainPropertiesEXT,
    .GetPastPresentationTimingEXT = anv_rmv_GetPastPresentationTimingEXT,
#ifdef VK_USE_PLATFORM_SCREEN_QNX
    .GetScreenBufferPropertiesQNX = anv_rmv_GetScreenBufferPropertiesQNX,
#elif defined(_MSC_VER)
    .GetScreenBufferPropertiesQNX = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_SCREEN_QNX
    .CmdBindDescriptorSets2 = anv_rmv_CmdBindDescriptorSets2,
    .CmdBindDescriptorSets2KHR = anv_rmv_CmdBindDescriptorSets2KHR,
    .CmdPushConstants2 = anv_rmv_CmdPushConstants2,
    .CmdPushConstants2KHR = anv_rmv_CmdPushConstants2KHR,
    .CmdPushDescriptorSet2 = anv_rmv_CmdPushDescriptorSet2,
    .CmdPushDescriptorSet2KHR = anv_rmv_CmdPushDescriptorSet2KHR,
    .CmdPushDescriptorSetWithTemplate2 = anv_rmv_CmdPushDescriptorSetWithTemplate2,
    .CmdPushDescriptorSetWithTemplate2KHR = anv_rmv_CmdPushDescriptorSetWithTemplate2KHR,
    .CmdSetDescriptorBufferOffsets2EXT = anv_rmv_CmdSetDescriptorBufferOffsets2EXT,
    .CmdBindDescriptorBufferEmbeddedSamplers2EXT = anv_rmv_CmdBindDescriptorBufferEmbeddedSamplers2EXT,
    .SetLatencySleepModeNV = anv_rmv_SetLatencySleepModeNV,
    .LatencySleepNV = anv_rmv_LatencySleepNV,
    .SetLatencyMarkerNV = anv_rmv_SetLatencyMarkerNV,
    .GetLatencyTimingsNV = anv_rmv_GetLatencyTimingsNV,
    .QueueNotifyOutOfBandNV = anv_rmv_QueueNotifyOutOfBandNV,
    .CmdSetRenderingAttachmentLocations = anv_rmv_CmdSetRenderingAttachmentLocations,
    .CmdSetRenderingAttachmentLocationsKHR = anv_rmv_CmdSetRenderingAttachmentLocationsKHR,
    .CmdSetRenderingInputAttachmentIndices = anv_rmv_CmdSetRenderingInputAttachmentIndices,
    .CmdSetRenderingInputAttachmentIndicesKHR = anv_rmv_CmdSetRenderingInputAttachmentIndicesKHR,
    .CmdSetDepthClampRangeEXT = anv_rmv_CmdSetDepthClampRangeEXT,
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandleEXT = anv_rmv_GetMemoryMetalHandleEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandleEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
#ifdef VK_USE_PLATFORM_METAL_EXT
    .GetMemoryMetalHandlePropertiesEXT = anv_rmv_GetMemoryMetalHandlePropertiesEXT,
#elif defined(_MSC_VER)
    .GetMemoryMetalHandlePropertiesEXT = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_METAL_EXT
    .ConvertCooperativeVectorMatrixNV = anv_rmv_ConvertCooperativeVectorMatrixNV,
    .CmdConvertCooperativeVectorMatrixNV = anv_rmv_CmdConvertCooperativeVectorMatrixNV,
    .CmdDispatchTileQCOM = anv_rmv_CmdDispatchTileQCOM,
    .CmdBeginPerTileExecutionQCOM = anv_rmv_CmdBeginPerTileExecutionQCOM,
    .CmdEndPerTileExecutionQCOM = anv_rmv_CmdEndPerTileExecutionQCOM,
    .CreateExternalComputeQueueNV = anv_rmv_CreateExternalComputeQueueNV,
    .DestroyExternalComputeQueueNV = anv_rmv_DestroyExternalComputeQueueNV,
    .CreateShaderInstrumentationARM = anv_rmv_CreateShaderInstrumentationARM,
    .DestroyShaderInstrumentationARM = anv_rmv_DestroyShaderInstrumentationARM,
    .CmdBeginShaderInstrumentationARM = anv_rmv_CmdBeginShaderInstrumentationARM,
    .CmdEndShaderInstrumentationARM = anv_rmv_CmdEndShaderInstrumentationARM,
    .GetShaderInstrumentationValuesARM = anv_rmv_GetShaderInstrumentationValuesARM,
    .ClearShaderInstrumentationMetricsARM = anv_rmv_ClearShaderInstrumentationMetricsARM,
    .CreateTensorARM = anv_rmv_CreateTensorARM,
    .DestroyTensorARM = anv_rmv_DestroyTensorARM,
    .CreateTensorViewARM = anv_rmv_CreateTensorViewARM,
    .DestroyTensorViewARM = anv_rmv_DestroyTensorViewARM,
    .GetTensorMemoryRequirementsARM = anv_rmv_GetTensorMemoryRequirementsARM,
    .BindTensorMemoryARM = anv_rmv_BindTensorMemoryARM,
    .GetDeviceTensorMemoryRequirementsARM = anv_rmv_GetDeviceTensorMemoryRequirementsARM,
    .CmdCopyTensorARM = anv_rmv_CmdCopyTensorARM,
    .GetTensorOpaqueCaptureDescriptorDataARM = anv_rmv_GetTensorOpaqueCaptureDescriptorDataARM,
    .GetTensorViewOpaqueCaptureDescriptorDataARM = anv_rmv_GetTensorViewOpaqueCaptureDescriptorDataARM,
    .CreateDataGraphPipelinesARM = anv_rmv_CreateDataGraphPipelinesARM,
    .CreateDataGraphPipelineSessionARM = anv_rmv_CreateDataGraphPipelineSessionARM,
    .GetDataGraphPipelineSessionBindPointRequirementsARM = anv_rmv_GetDataGraphPipelineSessionBindPointRequirementsARM,
    .GetDataGraphPipelineSessionMemoryRequirementsARM = anv_rmv_GetDataGraphPipelineSessionMemoryRequirementsARM,
    .BindDataGraphPipelineSessionMemoryARM = anv_rmv_BindDataGraphPipelineSessionMemoryARM,
    .DestroyDataGraphPipelineSessionARM = anv_rmv_DestroyDataGraphPipelineSessionARM,
    .CmdDispatchDataGraphARM = anv_rmv_CmdDispatchDataGraphARM,
    .GetDataGraphPipelineAvailablePropertiesARM = anv_rmv_GetDataGraphPipelineAvailablePropertiesARM,
    .GetDataGraphPipelinePropertiesARM = anv_rmv_GetDataGraphPipelinePropertiesARM,
#ifdef VK_USE_PLATFORM_OHOS
    .GetNativeBufferPropertiesOHOS = anv_rmv_GetNativeBufferPropertiesOHOS,
#elif defined(_MSC_VER)
    .GetNativeBufferPropertiesOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
#ifdef VK_USE_PLATFORM_OHOS
    .GetMemoryNativeBufferOHOS = anv_rmv_GetMemoryNativeBufferOHOS,
#elif defined(_MSC_VER)
    .GetMemoryNativeBufferOHOS = (PFN_vkVoidFunction)vk_entrypoint_stub,
#endif // VK_USE_PLATFORM_OHOS
    .QueueSetPerfHintQCOM = anv_rmv_QueueSetPerfHintQCOM,
    .CmdSetComputeOccupancyPriorityNV = anv_rmv_CmdSetComputeOccupancyPriorityNV,
    .WriteSamplerDescriptorsEXT = anv_rmv_WriteSamplerDescriptorsEXT,
    .WriteResourceDescriptorsEXT = anv_rmv_WriteResourceDescriptorsEXT,
    .CmdBindSamplerHeapEXT = anv_rmv_CmdBindSamplerHeapEXT,
    .CmdBindResourceHeapEXT = anv_rmv_CmdBindResourceHeapEXT,
    .CmdPushDataEXT = anv_rmv_CmdPushDataEXT,
    .RegisterCustomBorderColorEXT = anv_rmv_RegisterCustomBorderColorEXT,
    .UnregisterCustomBorderColorEXT = anv_rmv_UnregisterCustomBorderColorEXT,
    .GetImageOpaqueCaptureDataEXT = anv_rmv_GetImageOpaqueCaptureDataEXT,
    .GetTensorOpaqueCaptureDataARM = anv_rmv_GetTensorOpaqueCaptureDataARM,
    .CmdCopyMemoryKHR = anv_rmv_CmdCopyMemoryKHR,
    .CmdCopyMemoryToImageKHR = anv_rmv_CmdCopyMemoryToImageKHR,
    .CmdCopyImageToMemoryKHR = anv_rmv_CmdCopyImageToMemoryKHR,
    .CmdUpdateMemoryKHR = anv_rmv_CmdUpdateMemoryKHR,
    .CmdFillMemoryKHR = anv_rmv_CmdFillMemoryKHR,
    .CmdCopyQueryPoolResultsToMemoryKHR = anv_rmv_CmdCopyQueryPoolResultsToMemoryKHR,
    .CmdBeginConditionalRendering2EXT = anv_rmv_CmdBeginConditionalRendering2EXT,
    .CmdBindTransformFeedbackBuffers2EXT = anv_rmv_CmdBindTransformFeedbackBuffers2EXT,
    .CmdBeginTransformFeedback2EXT = anv_rmv_CmdBeginTransformFeedback2EXT,
    .CmdEndTransformFeedback2EXT = anv_rmv_CmdEndTransformFeedback2EXT,
    .CmdDrawIndirectByteCount2EXT = anv_rmv_CmdDrawIndirectByteCount2EXT,
    .CmdWriteMarkerToMemoryAMD = anv_rmv_CmdWriteMarkerToMemoryAMD,
    .CmdBindIndexBuffer3KHR = anv_rmv_CmdBindIndexBuffer3KHR,
    .CmdBindVertexBuffers3KHR = anv_rmv_CmdBindVertexBuffers3KHR,
    .CmdDrawIndirect2KHR = anv_rmv_CmdDrawIndirect2KHR,
    .CmdDrawIndexedIndirect2KHR = anv_rmv_CmdDrawIndexedIndirect2KHR,
    .CmdDrawIndirectCount2KHR = anv_rmv_CmdDrawIndirectCount2KHR,
    .CmdDrawIndexedIndirectCount2KHR = anv_rmv_CmdDrawIndexedIndirectCount2KHR,
    .CmdDrawMeshTasksIndirect2EXT = anv_rmv_CmdDrawMeshTasksIndirect2EXT,
    .CmdDrawMeshTasksIndirectCount2EXT = anv_rmv_CmdDrawMeshTasksIndirectCount2EXT,
    .CmdDispatchIndirect2KHR = anv_rmv_CmdDispatchIndirect2KHR,
    .CreateAccelerationStructure2KHR = anv_rmv_CreateAccelerationStructure2KHR,
    .CmdSetDispatchParametersARM = anv_rmv_CmdSetDispatchParametersARM,
};


